program ex1_apply_q_bd
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines APPLY_Q_BD and APPLY_P_BD
!   in module SVD_Procedures for applying Householder transformations stored during the
!   bidiagonal reduction of a real matrix to other matrices.
!
!
! Further Details
! ===============
!
!    The program also shows the use of subroutines BD_CMP, BD_SVD and BD_INVITER
!    in module SVD_Procedures for computing a partial or full SVD of a real matrix.
!
!
! LATEST REVISION : 15/11/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, bd_inviter,     &
                         bd_cmp, bd_svd, apply_q_bd, apply_p_bd, norm, c50
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n AND m ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsing IS THE NUMBER OF THE LARGEST SINGULAR TRIPLETS WHICH MUST BE COMPUTED,
! maxiter IS THE NUMBER OF INVERSE ITERATIONS PERFORMED TO COMPUTE THE TOP nsing SINGULAR VECTORS.
!
    integer(i4b), parameter :: prtunit=6, n=1000, m=500, mn=min(m,n), nsing=20, maxiter=2
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of apply_q_bd'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                     :: err, eps, elapsed_time
    real(stnd), dimension(n,m)     :: a, a2
    real(stnd), dimension(n,nsing) :: leftvec
    real(stnd), dimension(m,nsing) :: rightvec
    real(stnd), dimension(mn)      :: s, d, e, e2, tauq, taup
!
    integer      :: istart, iend, irate, imax, itime
!
    logical(lgl) :: failure1, failure2, bd_is_upper
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : BIDIAGONAL REDUCTION AND PARTIAL SVD OF A n-by-m REAL MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
!
!   GENERATE A RANDOM DATA MATRIX a.
!
    call random_number( a )
!
!   SAVE RANDOM DATA MATRIX a .
!
    a2(:n,:m) = a(:n,:m)
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   REDUCE a TO LOWER OR UPPER BIDIAGONAL FORM.
!
    call bd_cmp( a, s, e2, tauq, taup )
!
!   THE DIAGONAL ELEMENTS ARE STORED IN s .
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN e2 .
!
!   SAVE BIDIAGONAL FORM OF a .
!
    e(:mn) = e2(:mn)
    d(:mn) = s(:mn)
!
    bd_is_upper = n>=m
!
!   COMPUTE SINGULAR VALUES OF BIDIAGONAL FORM OF a .
!
    call bd_svd( bd_is_upper, s, e2, failure1, sort=sort  )
!
!   COMPUTE THE FIRST nsing SINGULAR VECTORS OF a BY maxiter
!   INVERSE ITERATIONS ON THE INTERMEDIATE BIDIAGONAL MATRIX
!   d_e AND BACK-TRANSFORMATION.
!
    call bd_inviter( bd_is_upper, d, e, s(:nsing), leftvec(:mn,:nsing), rightvec(:mn,:nsing),  &
                     failure2, maxiter=maxiter )
!
!   COMPUTE SINGULAR VECTORS OF a BY BACK-TRANSFORMATION.
!
    if ( bd_is_upper ) then
        leftvec(mn+1_i4b:n,:nsing) = zero
    else
        rightvec(mn+1_i4b:m,:nsing) = zero
    end if
!
!   GENERATE LEFT SINGULAR VECTORS OF a IN leftvec BY BACK-TRANSFORMATION.
!
    call apply_q_bd( a, tauq, leftvec,  left=true, trans=false )
!
!   GENERATE RIGHT SINGULAR VECTORS OF a IN rightvec BY BACK-TRANSFORMATION.
!
    call apply_p_bd( a, taup, rightvec, left=true, trans=false )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*rightvec - leftvec*s(:nsing),
!   WHERE leftvec AND rightvec ARE THE LEFT AND RIGHT SINGULAR VECTORS.
!
    err =  norm( matmul(a2,rightvec) - leftvec*spread(s(:nsing),dim=1,ncopies=n) )/( sum( abs(s(:mn)) )*real(mn,stnd) )
!
    if ( err<=eps .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' FAILURE ( from bd_svd() )     = ', failure1
    write (prtunit,*) ' FAILURE ( from bd_inviter() ) = ', failure2
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and ', nsing, ' singular vectors of a ', &
       n, ' by ', m,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_apply_q_bd
! =============================
!
end program ex1_apply_q_bd
