program ex1_bd_inviter2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine BD_INVITER2
!   in module SVD_Procedures for computing a full or partial SVD of a real matrix using
!   the Chan-Golub-Reinsch bidiagonal reduction algorithm, a bisection algorithm for
!   singular values and the inverse iteration method for singular vectors.
!
!   The computations are parallelized if OpenMP is used and highly efficient variants of
!   the Chan-Golub-Reinsch bidiagonal reduction, bisection and inverse iteration
!   algorithms are used.
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutines BD_CMP and BD_SINGVAL in module SVD_Procedures.
!                                                                              
! LATEST REVISION : 05/12/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, safmin, bd_cmp, bd_inviter2, bd_singval, &
                         norm, unit_matrix, zero, one, seven, c30, c50, c1_5, c1_e6, merror,   &
                         allocate_error, random_seed_, random_number_, gen_random_mat
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n AND m ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX (WHICH MUST BE LESS OR EQUAL TO mn) FOR CASES GREATER THAN 0,
! nsing IS THE NUMBER OF THE LARGEST SINGULAR TRIPLETS WHICH MUST BE COMPUTED.
! maxiter IS THE NUMBER OF INVERSE ITERATIONS PERFORMED TO COMPUTE THE TOP nsing SINGULAR VECTORS.
!
    integer(i4b), parameter :: prtunit = 6, n=3000, m=3000, mn=min(m,n), nsvd0=3000, &
                               nsing=3000, maxiter=2
!   
! conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN CASES 5, 6 and 7,
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50, conda=c1_e6
!
    character(len=*), parameter :: name_proc='Example 1 of bd_inviter2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, tmp, tmp2, abstol, &
                                               abs_err, rel_err, anorm, elapsed_time
    real(stnd), allocatable, dimension(:)   :: s, s0, d, e, tauq, taup, tauo
    real(stnd), allocatable, dimension(:,:) :: a, a2, leftvec, rightvec, resid, rla
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: ns, mnthr, i, mat_type
!
    logical(lgl) :: failure1, failure2, do_test, two_stage
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : nsing SINGULAR VALUES AND VECTORS OF A n-BY-m REAL MATRIX
!               USING A BISECTION METHOD FOR THE SINGULAR VALUES AND THE INVERSE
!               ITERATION METHOD FOR THE SINGULAR VECTORS (e.g., A PARTIAL SVD DECOMPOSITION).
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type < 1  -> RANDOM MATRIX FROM THE UNIFORM DISTRIBUTION
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type = 4  -> VERY SLOW DECAY OF SINGULAR VALUES
!   mat_type = 5  -> STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda
!                    SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE
!   mat_type = 6  -> GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type = 7  -> ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type > 7  -> UNIFORM DISTRIBUTION OF SINGULAR VALUES
!
    mat_type = 3_i4b
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps  = fudge*epsilon( err )
    err  = zero
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   CHOOSE TUNING PARAMETERS FOR THE ALGORITHM.
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   a TO BIDIAGONAL FORM. IF max(n,m) EXCEEDS mnthr ,
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE
!   n-BY-m MATRIX a TO A TRIANGULAR FORM.
!    
    mnthr = int( real( mn, stnd )*c1_5, i4b )
!
    two_stage = max( n, m )>=mnthr
!    two_stage = true
!    two_stage = false
!
!   CHOOSE TUNING PARAMETERS FOR THE BISECTION ALGORITHM.
!
!   SINGULAR VALUES WILL BE COMPUTED MOST ACCURATELY WHEN abstol IS SET
!   TO THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD (E.G., sqrt(safmin) ).
!
    abstol = sqrt( safmin )
!
!   ALLOCATE WORK ARRAYS.
!
    if ( two_stage ) then
        allocate( a(n,m), leftvec(n,nsing), rightvec(m,nsing), rla(mn,mn),    &
                  s(mn), d(mn), e(mn), tauq(mn), taup(mn), tauo(mn), stat=iok )
    else
        allocate( a(n,m), leftvec(n,nsing), rightvec(m,nsing),      &
                  s(mn), d(mn), e(mn), tauq(mn), taup(mn), stat=iok )
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE A RANDOM UNIFORM MATRIX OR SINGULAR VALUES.
!
    select case( mat_type )
!
        case( :0_i4b )
!
!           RANDOM UNIFORM MATRIX.
!
            call random_number_( a )
!
!           COMPUTE THE FROBENIUS NORM OF THE RANDOM MATRIX.
!
            anorm = norm( a )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s(i) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
            end do
!
        case( 4_i4b )
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            tmp = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                s(i) = exp( -tmp2/tmp )
!        
            end do
!
        case( 5_i4b )
!
!           STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE.
!
            s(:nsvd0-1_i4b) = one
            s(nsvd0) = one/conda
!
        case( 6_i4b )
!
!           GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                s(i) = conda**( -tmp2/tmp )
!        
            end do
!
        case( 7_i4b )
!
!           ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
            tmp2 = one - one/conda
!
            do i = 1_i4b, nsvd0
!
                s(i) = one - (real( i - 1_i4b, stnd )/tmp)*tmp2
!        
            end do
!
        case default
!
!           UNIFORM DISTRIBUTION OF SINGULAR VALUES.
!
            call random_number_( s(:nsvd0) )
!
    end select
!
    if ( mat_type>0_i4b ) then
!
#ifdef _F2003
        if ( ieee_support_datatype( s ) ) then
!
            if ( .not.all( ieee_is_normal( s(:nsvd0) ) ) ) then
!
                call merror( name_proc//' : Exceptions occurred when generating the input matrix !'  )                    
!
            end if
!
        end if
#endif
!
!       GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!       OF SINGULAR VALUES AND A RANK EQUAL TO nsvd0.
!
        call gen_random_mat( s(:nsvd0), a )
!
!       COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
        anorm = norm( s(:nsvd0) )
!
    end if
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,m), resid(n,nsing), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX .
!
        a2(:n,:m) = a(:n,:m)
!
        if ( mat_type>0_i4b ) then
!
!           ALLOCATE WORK ARRAY.
!
            allocate( s0(mn), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           SAVE THE TRUE SINGULAR VALUES.
!
            s0(:nsvd0) = s(:nsvd0)
            s0(nsvd0+1_i4b:mn) = zero
!
        end if
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL n-BY-m MATRIX a
!   IS WRITTEN
!
!                       a = U * S * V**(t)
!
!   WHERE S IS AN n-BY-m MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   min(m,n) DIAGONAL ELEMENTS, U IS AN n-BY-n ORTHOGONAL MATRIX, AND
!   V IS AN m-BY-m ORTHOGONAL MATRIX.  THE DIAGONAL ELEMENTS OF S
!   ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!   THIS PROGRAM ILLUSTRATES HOW TO COMPUTE nsing SINGULAR VALUES OF a AND
!   THE ASSOCIATED LEFT AND RIGHT SINGULAR VECTORS OF a (E.G., A PARTIAL SVD
!   DECOMPOSITION OF a) IN THREE STEPS:
!
    if ( two_stage ) then
!
!       STEP1A : REDUCE THE MATRIX a TO BIDIAGONAL FORM BY A TWO-STAGE ALGORITHM WITH SUBROUTINE bd_cmp.
!       THE INTERMEDIATE BIDIAGONAL FORM OF a IS STORED IN d AND e.
!       THE ORTHOGONAL MATRICES USED TO REDUCE a TO BIDIAGONAL FORM
!       ARE STORED IN mat, tauq, taup, rla AND tauo.
!
        call bd_cmp( a(:n,:m), d(:mn), e(:mn), tauq(:mn), taup(:mn), rla(:mn,:mn), tauo=tauo(:mn) )
!        call bd_cmp( a(:n,:m), d(:mn), e(:mn), tauq(:mn), taup(:mn), rla(:mn,:mn) )
!
    else
!
!       STEP1B : REDUCE THE MATRIX a TO BIDIAGONAL FORM WITHOUT A QR OR LQ FACTORIZATION WITH SUBROUTINE bd_cmp.
!       THE INTERMEDIATE BIDIAGONAL FORM OF a IS STORED IN d AND e.
!       THE ORTHOGONAL MATRICES USED TO REDUCE a TO BIDIAGONAL FORM
!       ARE STORED IN mat, tauq, taup.
!
        call bd_cmp( a(:n,:m), d(:mn), e(:mn), tauq(:mn), taup(:mn) )
!
    end if
!
!   STEP2 : COMPUTE nsing SINGULAR VALUES OF THE BIDIAGONAL MATRIX (WHICH ARE ALSO THE
!   SINGULAR VALUES OF a) WITH SUBROUTINE bd_singval TO HIGH RELATIVE PRECISION.
!   THE SINGULAR VALUES ARE STORED IN s IN DECREASING ORDER (IF sort='d').
!
!   THE OPTIONAL ARGUMENT abstol OF bd_singval MAY BE USED TO INDICATE THE REQUIRED
!   PRECISION FOR THE SINGULAR VALUES. A SINGULAR VALUE IS CONSIDERED TO BE LOCATED
!   IF IT HAS BEEN DETERMINED TO LIE IN AN INTERVAL WHOSE WIDTH IS EQUAL OR LESS
!   THAN abstol. SINGULAR VALUES WILL BE COMPUTED MOST ACCURATELY WHEN abstol IS SET
!   TO THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD (E.G., sqrt(safmin) ).
!
    call bd_singval( d(:mn), e(:mn), ns, s(:mn), failure=failure1, sort=sort, &
                     abstol=abstol, ls=nsing )
!
!   ON EXIT OF bd_singval :
!
!   failure= false :  INDICATES SUCCESSFUL EXIT IN THE BISECTION ALGORITHM
!   failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                     THAT FULL ACCURACY WAS NOT ATTAINED IN THE
!                     COMPUTATION OF THE SINGULAR VALUES OF a.
!
!   ns GIVES THE NUMBER OF SINGULAR VALUES REALLY COMPUTED BY bd_singval.
!   ns MAY BE GREATER THAN ls IF CLUSTERS OF SINGULAR VALUES ARE PRESENT.
!   THE ns LARGEST SINGULAR VALUES ARE STORED IN THE FIRST ns POSITIONS OF s
!   IN DECREASING ORDER IF sort='d'.
!
!   THE SINGULAR VALUES ARE STORED IN s IN DECREASING ORDER (IF sort='d'). HERE, THIS
!   IS REQUIRED BY SUBROUTINE bd_inviter2 IN THE SECOND STEP.
!
!   STEP3 : COMPUTE THE FIRST nsing LEFT AND RIGHT SINGULAR VECTORS OF a BY maxiter
!   INVERSE ITERATIONS ON THE INTERMEDIATE BIDIAGONAL MATRIX AND BACK-TRANSFORMATION
!   WITH SUBROUTINE bd_inviter2 .
!
!   ON ENTRY OF bd_inviter2, PARAMETER s CONTAINS SELECTED SINGULAR VALUES OF THE BIDIAGONAL MATRIX
!   STORED IN VECTORS d AND e. THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
    if ( two_stage ) then
!
        call bd_inviter2( a, tauq, taup, rla, d, e, s(:nsing), leftvec, rightvec, failure=failure2, tauo=tauo, maxiter=maxiter )
!        call bd_inviter2( a, tauq, taup, rla, d, e, s(:nsing), leftvec, rightvec, failure=failure2, maxiter=maxiter )
!
    else
!
        call bd_inviter2( a, tauq, taup, d, e, s(:nsing), leftvec, rightvec, failure=failure2, maxiter=maxiter )
!
    end if
!
!   ON EXIT OF bd_inviter2 :
!
!         failure= false :  INDICATES SUCCESSFUL EXIT IN THE INVERSE ITERATION ALGORITHM
!         failure= true  :  INDICATES THAT THE ALGORITHM FAILS TO CONVERGE FOR SOME SINGULAR VECTORS
!
!   THE SINGULAR VECTORS OF THE BIDIAGONAL FORM ARE COMPUTED USING maxiter INVERSE ITERATIONS
!   FOR ALL THE SINGULAR VALUES AT ONE STEP. THE SINGULAR VECTORS OF ARE THEN ORTHOGONALIZED BY 
!   THE MODIFIED GRAM-SCHMIDT ALGORITHM IF NECESSARY. THE SINGULAR VECTORS OF a ARE FINALLY COMPUTED 
!   BY BACK TRANSFORMATIONS. ON EXIT, leftvec AND rightvec CONTAIN THE FIRST nsing LEFT AND RIGHT
!   SINGULAR VECTORS OF a, RESPECTIVELY.
!
!   NOTE THAT bd_inviter2 MAY FAIL IF SOME THE SINGULAR VALUES SPECIFIED IN PARAMETER s ARE NEARLY
!   IDENTICAL FOR SOME PATHOLOGICAL MATRICES.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
        if ( mat_type>0_i4b ) then
!
!           COMPUTE ERRORS FOR SINGULAR VALUES.
!
            where( s0(:nsing)/=zero )
                resid(:nsing,1_i4b) = s0(:nsing)
            elsewhere
                resid(:nsing,1_i4b) = one
            end where
!
!           ABSOLUTE ERRORS OF SINGULAR VALUES.
!
            abs_err = maxval( abs( s(:nsing) - s0(:nsing) ) )
!
!           RELATIVE ERRORS OF SINGULAR VALUES.
!
            rel_err = maxval( abs( (s(:nsing) - s0(:nsing))/resid(:nsing,1_i4b) ) )
!
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*V(:m,:nsing) - U(:n,:nsing)*S(:nsing,:nsing),
!       WHERE leftvec AND rightvec ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        resid(:n,:nsing) = matmul(a2,rightvec) - leftvec*spread(s(:nsing),dim=1,ncopies=n)
        a2(:nsing,1_i4b) = norm( resid(:n,:nsing), dim=2_i4b )
        err1 = maxval( a2(:nsing,1_i4b) )/( anorm*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U(:n,:nsing)**(t)*U(:n,:nsing).
!
        call unit_matrix( a2(:nsing,:nsing) )
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(leftvec), leftvec ) )
        err2 = maxval( resid(:nsing,:nsing) )/real(n,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V(:m,:nsing)**(t)*V(:m,:nsing).
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(rightvec), rightvec ) )
        err3 = maxval( resid(:nsing,:nsing) )/real(m,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( two_stage ) then
!
        deallocate( a, leftvec, rightvec, rla, s, d, e, tauq, taup, tauo )
!
    else
!
        deallocate( a, leftvec, rightvec, s, d, e, tauq, taup )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure1 .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix_type < 1  -> random matrix from the uniform distribution'
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type = 4  -> very slow decay of singular values'
    write (prtunit,*) ' Matrix type = 5  -> strongly clustered singular values at 1'
    write (prtunit,*) '                     spectrum with few deflations (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 6  -> geometric distribution of singular values'
    write (prtunit,*) '                     spectrum with moderate deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 7  -> arithmetic distribution of singular values'
    write (prtunit,*) '                     spectrum with complete deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type > 7  -> uniform distribution of singular values'
!
    write (prtunit,*) 
    write (prtunit,*) ' FAILURE ( from bd_singval() )  = ', failure1
    write (prtunit,*) ' FAILURE ( from bd_inviter2() ) = ', failure2
!
    if ( do_test ) then
!
        write (prtunit,*) 
!
        if ( mat_type>0_i4b ) then
!
            write (prtunit,*) 'Absolute accuracy of the computed singular values    = ', abs_err
            write (prtunit,*) 'Relative accuracy of the computed singular values    = ', rel_err
!
        end if
!
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
!
    end if
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and ', nsing, ' singular vectors of a ', &
       n, ' by ', m,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_bd_inviter2
! ==============================
!
end program ex1_bd_inviter2
