program ex1_bd_lasq1
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine BD_LASQ1
!   in module SVD_Procedures for computing all singular values of a n-by-n bidiagonal matrix.
!   The singular values are computed by the the differential quotient difference with
!   shifts (dqds) algorithm to high relative accuracy, in the absence of denormalization,
!   underflow and overflow.
!
!   Note that BD_LASQ1 is usually faster than BD_SINGVAL and BD_SINGVAL2 subroutines, which
!   are based on bisection, for computing all singular values of a real bidiagonal matrix,
!   but that BD_LASQ1 does not allow computing selected singular values of a bidiagonal matrix.
!
!
! Further Details
! ===============
!
!    The program also shows the use of subroutine BD_INVITER in module SVD_Procedures
!    fo computing singular vectors of a real bidiagonal matrix.
!
!
! LATEST REVISION : 08/02/2024
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, quarter, one, two, five, c50, c200,  &
                         bd_inviter, bd_lasq1, unit_matrix, norm, geop, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE DIMENSION OF THE BIDIAGONAL MATRIX,
! ls IS THE NUMBER OF THE LARGEST SINGULAR VECTOR COUPLETS WHICH MUST BE COMPUTED,
! maxiter IS THE NUMBER OF INVERSE ITERATIONS PERFORMED TO COMPUTE THE TOP ls SINGULAR VECTORS.
!
    integer(i4b), parameter :: prtunit=6, n=1000, ls=15, maxiter=2
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of bd_lasq1'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e, s, e2
    real(stnd), allocatable, dimension(:,:) :: id, resid, leftvec, rightvec
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: mat_type, j, i, info
!
    logical(lgl) :: failure1=false, failure2=false, bd_is_upper, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : ALL SINGULAR VALUES AND ls SINGULAR VECTORS OF A n-BY-n REAL BIDIAGONAL MATRIX
!               bd BY THE DQDS-INVERSE ITERATION METHOD (E.G. PARTIAL SVD DECOMPOSITION).
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( d(n), e(n), s(n), e2(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE AN UPPER BIDIAGONAL MATRIX bd .
!   THE DIAGONAL ELEMENTS ARE STORED IN d .
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN e .
!
    bd_is_upper = true
    e(1_i4b)    = zero
!
!   SPECIFY THE TYPE OF BIDIAGONAL MATRIX:
!
    mat_type = 9_i4b
!
!   GENERATE THE BIDIAGONAL MATRIX.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           bd IS A GRADED MATRIX.
!
            d(:n)      = geop( one, quarter, n )
            e(2_i4b:n) = d(:n-1_i4b)
!
        case( 2_i4b )
!
!           bd IS A TOEPLITZ MATRIX.
!
            d(:n)      = one
            e(2_i4b:n) = two
!
        case( 3_i4b )
!
            e(2_i4b:n) = one
            do j=1_i4b, n
	        d(j) = real( n + 1_i4b - j, stnd )
            end do
!
        case( 4_i4b )
!
            do j=1_i4b, n-1_i4b
	        d(j)       = real( n + 1_i4b - j, stnd )
                e(j+1_i4b) = d(j)/five
            end do
	    d(n) = one
!
        case( 5_i4b )
!
            do j=1_i4b, (n+1_i4b)/2_i4b
                d(j) = real( j, stnd )
            end do
!
            do j=(n+1_i4b)/2_i4b+1_i4b, n
                d(j) = real( n - j, stnd )
            end do
!
	    do j=1, n-1_i4b
                e(j+1_i4b) = d(j)/five
            end do
!
        case( 6_i4b )
!
            do j=1_i4b, n/2_i4b
                d(2_i4b*j-1_i4b) = real( n + 1_i4b - j, stnd )
                d(2_i4b*j)       = real( j, stnd )
            end do
!
	    do j=1_i4b, n-1_i4b
                e(j+1_i4b) = real( (n-j)/5_i4b, stnd )
	    end do
!
        case( 7_i4b )
!
            d(1_i4b)   = one
            d(2_i4b:n) = c200*epsilon( err )
            e(2_i4b:n) = c200*epsilon( err )
!
        case( 8_i4b )
!
            d(n) = 1e-20
            do j=n-1_i4b, 1_i4b, -1_i4b
	        d(j)       = d(j+1_i4b)*1.01_stnd
                e(j+1_i4b) = d(j)
            end do
!
        case( 9_i4b )
!
            do j=0_i4b, n/11_i4b-1_i4b
!
                do i=1_i4b, 6_i4b
                    d(11_i4b*j+i) = real( 1_i4b+10_i4b*(i-1_i4b), stnd )
                    e(11_i4b*j+i+1_i4b) = one
                end do
!
                do i=7_i4b,10_i4b
                    d(11_i4b*j+i) = real( 1_i4b+10_i4b*(11_i4b-i), stnd )
                    e(11_i4b*j+i+1_i4b) = one
                end do
!
                d(11_i4b*j+11_i4b) = one
!
                if ( 11_i4b*j+12_i4b<=n ) then
                    d(11_i4b*j+12_i4b) = one
                    e(11_i4b*j+12_i4b) = 1e-4
                end if
!
            end do
!
        case default
!
!           bd IS A UNIFORM RANDOM MATRIX.
!
            call random_number( d(:n) )
            call random_number( e(2_i4b:n) )
!
    end select
!
#ifdef _F2003
    if ( ieee_support_datatype( d ) ) then
!
        if ( .not.all( ieee_is_normal( d ) ) .and. .not.all( ieee_is_normal( e ) ) ) then
!
            call merror( name_proc//' : Exceptions occurred when generating the input bidiagonal matrix !'  )                    
!
        end if
!
    end if
#endif
!
!   MAKE A COPY OF THE BIDIAGONAL MATRIX.
!
    s(:n)  = d(:n)
    e2(:n) = e(:n)
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL n-BY-n BIDIAGONAL MATRIX bd
!   IS WRITTEN
!
!                       bd = u * s * v**(t)
!
!   WHERE s IS AN n-BY-n MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   n DIAGONAL ELEMENTS, u AND v ARE n-BY-n ORTHOGONAL MATRICES.
!   THE DIAGONAL ELEMENTS OF s ARE THE SINGULAR VALUES OF bd; THEY ARE
!   REAL AND NON-NEGATIVE.
!   THE COLUMNS OF u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF bd.
!
!   THIS PROGRAM ILLUSTRATES HOW TO COMPUTE ALL THE SINGULAR VALUES OF bd AND
!   ITS LEADING LEFT AND RIGHT SINGULAR VECTORS (E.G. A PARTIAL SVD DECOMPOSITION OF bd)
!   IN TWO STEPS:
!
!   STEP1 : COMPUTE ALL THE SINGULAR VALUES OF bd BY THE DQDS ALGORITHM
!   WITH SUBROUTINE bd_lasq1.
!
!   ON ENTRY OF bd_lasq1, s(:n) AND e2(:n) CONTAIN, RESPECTIVELY, THE DIAGONAL AND SUBDIAGONAL
!   ELEMENTS OF THE BIDIAGONAL MATRIX.
!
    call bd_lasq1( s(:n), e2(:n), failure=failure1, sort=sort, info=info )
!
!   ON EXIT OF bd_lasq1 :
!
!   failure= false :  INDICATES SUCCESSFUL EXIT IN THE DQDS ALGORITHM.
!   failure= true  :  INDICATES THAT THE DQDS ALGORITHM DID NOT CONVERGE AND
!                     THAT FULL ACCURACY WAS NOT ATTAINED IN THE
!                     COMPUTATION OF THE SINGULAR VALUES OF bd.
!
!   THE SINGULAR VALUES ARE STORED IN s IN DECREASING ORDER (IF sort='d'). HERE, THIS
!   IS REQUIRED BY SUBROUTINE bd_inviter IN THE SECOND STEP.
!
!   STEP2 : COMPUTE THE ls LEADING LEFT AND RIGHT SINGULAR VECTORS OF bd BY INVERSE
!   ITERATION WITH SUBROUTINE bd_inviter :
!
    if ( ls>0 .and. .not.failure1 ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE SINGULAR VECTORS.
!
        allocate( leftvec(n,ls),  &
                  rightvec(n,ls), &
                  stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE THE FIRST ls SINGULAR VECTORS OF bd BY maxiter INVERSE ITERATIONS WITH
!       SUBROUTINE bd_inviter .
!
        call bd_inviter( bd_is_upper, d(:n), e(:n), s(:ls), leftvec(:n,:ls), rightvec(:n,:ls),  &
                         failure=failure2, maxiter=maxiter )
!
!       ON ENTRY, PARAMETER s CONTAINS SELECTED SINGULAR VALUES OF THE BIDIAGONAL MATRIX bd .
!       THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
!       ON EXIT OF bd_inviter :
!
!          failure= false :  INDICATES SUCCESSFUL EXIT IN THE INVERSE ITERATION ALGORITHM.
!          failure= true  :  INDICATES THAT THE ALGORITHM FAILS TO CONVERGE FOR SOME SINGULAR VECTORS.
!
!       THE SINGULAR VECTORS OF THE BIDIAGONAL MATRIX bd ARE COMPUTED USING maxiter INVERSE ITERATIONS
!       FOR ALL THE SINGULAR VALUES AT ONE STEP. THE SINGULAR VECTORS OF bd ARE THEN ORTHOGONALIZED BY 
!       THE MODIFIED GRAM-SCHMIDT ALGORITHM OR A BLOCKED QR ALGORITHM IF THE SINGULAR VALUES ARE NOT WELL SEPARATED.
!       ON EXIT, leftvec AND rightvec CONTAIN THE FIRST ls LEFT AND RIGHT SINGULAR VECTORS OF bd, RESPECTIVELY.
!
!       bd_inviter MAY FAIL IF SOME THE SINGULAR VALUES SPECIFIED IN PARAMETER s ARE NEARLY IDENTICAL.
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test .and. ls>0  .and. .not.failure1 ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( id(ls,ls), resid(n,ls), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION bd*v - u*singval(:ls),
!       WHERE u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF bd .
!
        if ( bd_is_upper ) then
!
            resid(:n,:ls) = spread( d(:n),         dim=2, ncopies=ls )*rightvec                 + &
                            eoshift( spread(e(:n), dim=2,ncopies=ls)*rightvec, shift=1 )        - &
                            spread( s(:ls),     dim=1, ncopies=n )*leftvec 

        else
!
            resid(:n,:ls) = spread( d(:n),         dim=2, ncopies=ls )*leftvec                   + &
                            eoshift( spread(e(:n), dim=2,ncopies=ls)*leftvec, shift=1 )          - &
                            spread( s(:ls),     dim=1, ncopies=n )*rightvec
!
        end if
!
        id(:ls,1_i4b) = norm( resid(:n,:ls), dim=2_i4b )
        err1 = maxval( id(:ls,1_i4b) )/( sum( s(:ls) )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!       WHERE u are LEFT SINGULAR VECTORS OF THE BIDIAGONAL MATRIX bd .
!
        call unit_matrix( id )
!
        resid(:ls,:ls) = abs( id - matmul( transpose( leftvec ), leftvec ) )
        err2 = maxval( resid(:ls,:ls) )/real(n,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - rightvec**(t)*rightvec
!       WHERE rightvec ARE THE RIGHT SINGULAR VECTORS OF THE BIDIAGONAL MATRIX bd .
!
        resid(:ls,:ls) = abs( id - matmul( transpose( rightvec ), rightvec ) )
        err3 = maxval( resid(:ls,:ls) )/real(n,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( id, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( ls>0 .and. .not.failure1 ) then
        deallocate( d, e, s, e2, leftvec, rightvec )
    else
        deallocate( d, e, s, e2 )
    end if
!
    if ( err<=eps .and. .not.failure1 .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' FAILURE ( from bd_lasq1() )   = ', failure1
    write (prtunit,*) ' INFO    ( from bd_lasq1() )   = ', info
    if ( ls>0 .and. .not.failure1 ) then
        write (prtunit,*) ' FAILURE ( from bd_inviter() ) = ', failure2
    end if
!
    if ( do_test .and. ls>0 .and. .not.failure1 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Number of requested singular triplets                = ', ls
        write (prtunit,*) 'Number of computed singular triplets                 = ', ls
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*)
    write (*,'(a,i6,a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and', ls, ' singular vectors of a ', &
       n, ' by ', n,' real bidiagonal matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_bd_lasq1
! ===========================
!
end program ex1_bd_lasq1
