program ex1_comp_cor
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine COMP_COR
!   in module Mul_Stat_Procedures for computing univariate statistics and correlation
!   coefficients between a tri-dimensional array and a one-dimensional array.
!
!   All the statistics and correlations are computed with only one-pass on the data
!   in one or several steps with a very efficient algorithm for large datasets, which
!   also allows out-of-core computations.
!
!
! LATEST REVISION : 29/11/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, true, comp_cor
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT; n, m AND p ARE THE DIMENSIONS OF THE ARRAYS.
! p IS THE NUMBER OF OBSERVATIONS.
!
    integer(i4b), parameter :: prtunit=6, n=26, m=20, p=50
!
    character(len=*), parameter :: name_proc='Example 1 of comp_cor'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                      :: err_xstat, err_ystat, err_cor, xyn, eps
    real(stnd), dimension(n,m)      :: xycor1, xycor2
    real(stnd), dimension(2)        :: ystat1, ystat2
    real(stnd), dimension(n,m,2)    :: xstat1, xstat2
    real(stnd), dimension(n,m,p)    :: x
    real(stnd), dimension(p)        :: y
!
    integer(i4b) :: i
!
    logical(lgl) :: first, last
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon(eps) )
!
!   GENERATE A RANDOM TRIDIMENSIONAL OBSERVATION ARRAY x .
!
    call random_number( x(:n,:m,:p) )
!
!   GENERATE A RANDOM OBSERVATION VECTOR y .
!
    call random_number( y(:p) )
!
!   FIRST COMPUTE THE CORRELATIONS BETWEEN ARRAYS x AND y
!   FOR THE p OBSERVATIONS IN ONE STEP.
!
    first = true
    last  = true
!
    call comp_cor( x(:n,:m,:p), y(:p), first, last,                    &
                   xstat1(:n,:m,:2), ystat1(:2), xycor1(:n,:m), xyn )
!
!   ON EXIT OF COMP_COR WHEN last=true :
!
!      xstat1(i,j,1) CONTAINS THE MEAN VALUES OF THE ARRAY SECTION x(i,j,:p).
!
!      xstat1(i,j,2) CONTAINS THE VARIANCES OF THE ARRAY SECTION x(i,j,:p).
!
!      ystat1(1)     CONTAINS THE MEAN VALUE OF THE DATA VECTOR y(:p).
!
!      ystat1(2)     CONTAINS THE VARIANCE OF THE DATA VECTOR y(:p).
!
!      xycor1(i,j)   CONTAINS THE CORRELATION COEFFICIENT
!                    BETWEEN x(i,j,:p) AND y(:p).
!
!      xyn           CONTAINS THE NUMBER OF OBSERVATIONS IN THE DATA ARRAY
!                    x(:n,:m,:p) AND THE DATA VECTOR y(:p) (xyn=real(p,stnd) ).
!
!   SECOND RECOMPUTE CORRELATIONS BETWEEN ARRAYS x AND y
!   ITERATIVELY FOR THE p OBSERVATIONS.
!
    do i = 1, p
!
        first = i==1
        last  = i==p
!
        call comp_cor( x(:n,:m,i:i), y(i:i), first, last,                 &
                       xstat2(:n,:m,:2), ystat2(:2), xycor2(:n,:m), xyn )
!
    end do
!
!   CHECK THAT THE TWO SETS OF STATISTICS AGREE.
!
    err_xstat = maxval( abs( ( xstat1-xstat2)/xstat1 ) )
    err_ystat = maxval( abs( ( ystat1-ystat2)/ystat1 ) )
    err_cor   = maxval( abs( xycor1-xycor2 ) )
!
    if ( max(err_xstat, err_ystat, err_cor )<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
!
! END OF PROGRAM ex1_comp_cor
! ===========================
!
end program ex1_comp_cor
