program ex1_comp_cor_miss
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine COMP_COR_MISS
!   in module Mul_Stat_Procedures for computing univariate statistics and correlation
!   coefficients between a tri-dimensional array and a one-dimensional array, both
!   with missing values.
!
!   All the statistics and correlations are computed with only one-pass on the data
!   in one or several steps with a very efficient algorithm for large datasets, which
!   also allows out-of-core computations.
!
!
! LATEST REVISION : 29/11/2023
!
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, true, comp_cor_miss
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT; n, m AND p ARE THE DIMENSIONS OF THE ARRAYS.
! p IS THE NUMBER OF OBSERVATIONS.
!
    integer(i4b), parameter :: prtunit=6, n=26, m=20, p=50
!
! miss IS THE MISSING INDICATOR IN THE DATA.
!
    real(stnd), parameter :: miss=-999.99_stnd
!
    character(len=*), parameter :: name_proc='Example 1 of comp_cor_miss'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                      :: err_xstat, err_ystat, err_cor, eps
    real(stnd), dimension(n,m,4)    :: xycor1, xycor2
    real(stnd), dimension(4)        :: ystat1, ystat2
    real(stnd), dimension(n,m,4)    :: xstat1, xstat2
    real(stnd), dimension(n,m,p)    :: x
    real(stnd), dimension(p)        :: y
!
    integer(i4b) :: i
!
    logical(lgl) :: first, last
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon(eps) )
!
!   GENERATE A RANDOM TRI-DIMENSIONAL OBSERVATION ARRAY x WITH MISSING VALUES.
!
    call random_number( x(:n,:m,:p) )
    where ( x(:n,:m,:p)<=0.05_stnd ) x(:n,:m,:p) = miss
!
!   GENERATE A RANDOM OBSERVATION VECTOR y WITH MISSING VALUES.
!
    call random_number( y(:p) )
    where ( y(:p)<=0.05_stnd ) y(:p) = miss
!
!   FIRST COMPUTE THE CORRELATIONS BETWEEN ARRAYS x AND y
!   FOR THE p OBSERVATIONS IN ONE STEP.
!
    first = true
    last  = true
!
    call comp_cor_miss( x(:n,:m,:p), y(:p), first, last, xstat1(:n,:m,:4),   &
                        ystat1(:4), xycor1(:n,:m,:4), xymiss=miss )
!
!   ON EXIT OF COMP_COR_MISS WHEN last=true :
!
!      xstat1(i,j,1) CONTAINS THE MEAN VALUES OF THE ARRAY SECTION x(i,j,:p).
!
!      xstat1(i,j,2) CONTAINS THE VARIANCES OF THE ARRAY SECTION x(i,j,:p).
!
!      xstat1(i,j,3) CONTAINS THE NUMBER OF NON-MISSING OBSERVATIONS
!                    IN THE ARRAY SECTION x(i,j,:p).
!
!      ystat1(1)     CONTAINS THE MEAN VALUE OF THE DATA VECTOR y(:p).
!
!      ystat1(2)     CONTAINS THE VARIANCE OF THE DATA VECTOR y(:p).
!
!      ystat1(3)     CONTAINS THE NUMBER OF NON-MISSING OBSERVATIONS
!                    IN THE DATA VECTOR y(:p).
!
!      xycor1(i,j,1) CONTAINS THE CORRELATION COEFFICIENT BETWEEN x(i,j,:p) AND y(:p)
!                    COMPUTED WITH THE ABOVE UNIVARIATE STATISTICS.
!
!      xycor1(i,j,2) CONTAINS THE INCIDENCE VALUE BETWEEN x(i,j,:p) AND y(:p).
!                    xycor1(i,j,2) INDICATES THE NUMBER OF VALID PAIRS OF OBSERVATIONS
!                    WHICH WHERE USED IN THE CALCULATION OF xycor1(i,j,1) .
!
!   xstat1(:,:,4), ystat1(4) AND xycor1(:,:,3:4) ARE USED AS WORKSPACE AND CONTAIN NO USEFUL
!   INFORMATION ON OUTPUT OF comp_cor_miss.
!
!   SECOND RECOMPUTE CORRELATIONS BETWEEN ARRAYS x AND y
!   ITERATIVELY FOR THE p OBSERVATIONS.
!
    do i = 1, p
!
        first = i==1
        last  = i==p
!
        call comp_cor_miss( x(:n,:m,i:i), y(i:i), first, last, xstat2(:n,:m,:4),       &
                            ystat2(:4), xycor2(:n,:m,:4), xymiss=miss )
!
    end do
!
!   CHECK THAT THE TWO SETS OF STATISTICS AND CORRELATIONS AGREE.
!
    err_xstat = maxval( abs( ( xstat1(:n,:m,:3)-xstat2(:n,:m,:3))/xstat1(:n,:m,:3) ) )
    err_ystat = maxval( abs( ( ystat1(:3)-ystat2(:3))/ystat1(:3) ) )
    err_cor   = maxval( abs( xycor1(:n,:m,:2)-xycor2(:n,:m,:2) ) )
!
    if ( max(err_xstat, err_ystat, err_cor )<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
!
! END OF PROGRAM ex1_comp_cor_miss
! ================================
!
end program ex1_comp_cor_miss
