program ex1_comp_det
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine COMP_DET
!   in module Lin_Procedures for computing the determinant of a real square
!   matrix using a LU decomposition.
!
!
! LATEST REVISION : 29/11/2023
!
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, lgl, stnd, true, false, zero, one, comp_det, inv, merror, allocate_error
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE DIMENSION OF THE GENERATED SQUARE MATRIX.
!
    integer(i4b), parameter :: prtunit=6, n=500
!
    character(len=*), parameter :: name_proc='Example 1 of comp_det'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: adet, ainvdet, err, eps, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, ainv
!
    integer :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTE THE DETERMINANT OF A REAL MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon( err ) )
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM DATA MATRIX a .
!
    call random_number( a )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( ainv(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE THE MATRIX INVERSE OF a.
!
        ainv = inv( a ) 
!
!       COMPUTE THE DETERMINANT OF MATRIX INVERSE .
!
        call comp_det( ainv, ainvdet )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( ainv )
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE DETERMINANT OF THE DATA MATRIX WITH SUBROUTINE comp_det.
!
    call comp_det( a, adet )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK det(a**-1)*det(a)**-1 = 1.
!
        err = abs(adet*ainvdet - one) / max( abs(adet), abs(ainvdet) )
!
    end if
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( a )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*)
    write (*,'(a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the determinant of a real matrix of size ', &
       n, ' is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_comp_det
! ===========================
!
end program ex1_comp_det
