program ex1_comp_filt_rot_pc
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines COMP_EOF,
!   COMP_PC_EOF and COMP_FILT_ROT_PC in module Mul_Stat_Procedures for
!   rotating a Principal Component Analysis model towards specific frequency modes.
!
!
! LATEST REVISION : 14/12/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, true, false, one, comp_eof, comp_pc_eof, &
                         comp_filt_rot_pc, norm
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, p IS THE NUMBER OF OBSERVATIONS, m THE NUMBER OF VARIABLES
! AND nrot IS THE NUMBER OF EIGENVECTORS OR PCS TO ROTATE.
!
    integer(i4b), parameter :: prtunit=6, m=20, p=100, nrot=5
!
    character(len=*), parameter :: name_proc='Example 1 of comp_filt_rot_pc'
!
    character, parameter :: sort = 'd'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                         :: eps, err_pc, err_rot1, err_rot2, err_rot3, err_rot4, xn
    real(stnd), dimension(m,m)         :: eigvec
    real(stnd), dimension(m,nrot)      :: factor, rot_factor
    real(stnd), dimension(nrot,nrot)   :: rot
    real(stnd), dimension(m)           :: mean, std, eigval, eigvar, singval
    real(stnd), dimension(nrot)        :: std_rot_pc
    real(stnd), dimension(m,p)         :: x, x_std
    real(stnd), dimension(p,m)         :: pc
    real(stnd), dimension(p,nrot)      :: rot_pc
!
    integer(i4b) :: pl, ph
!
    logical(lgl) :: first, last, cov, failure, failure2
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PERFORM A PRINCIPAL COMPONENT (E.G., EOF) ANALYSIS AND AN
!               ORTHOGONAL ROTATION OF THE PRINCIPAL COMPONENT TIME SERIES
!               TOWARDS A SPECIFIC FREQUENCY BAND.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon(eps) )
!
!   GENERATE A RANDOM OBSERVATION MATRIX x(:m,:p) WITH m VARIABLES
!   AND p OBSERVATIONS.
!
    call random_number( x )
!
!   COMPUTE EOFs FROM THE CORRELATION MATRIX.
!
    cov = false
!
    first = true
    last  = true
!
    call comp_eof( x(:m,:p), first, last, eigval(:m), eigvec(:m,:m), xn, failure,       &
                   cov=cov, sort=sort, xmean=mean(:m), xstd=std(:m), xeigvar=eigvar(:m) )
!
!   ON EXIT OF COMP_EOF WHEN last=true :
!
!      eigval(:m)       CONTAINS THE EIGENVALUES.
!
!      eigvec(:m,:m)    CONTAINS THE EIGENVECTORS STORED COLUMNWISE IN THE ORDER
!                       OF THE EIGENVALUES STORED IN eigval.
!
!      xn               CONTAINS THE NUMBER OF OBSERVATIONS IN THE DATA ARRAY x,
!                       xn = real(p,stnd) .
!
!      failure = false  INDICATES SUCCESSFUL EXIT.
!      failure = true   INDICATES THAT FEWER THAN TWO VALID OBSERVATIONS
!                       WERE PRESENT OR THAT THE OBSERVATIONS ON SOME
!                       VARIABLE WERE CONSTANT AND THE CORRELATIONS
!                       WERE REQUESTED OR THAT MAXIMUM ACCURACY WAS NOT
!                       ACHIEVED WHEN COMPUTING THE EIGENVECTORS AND THE
!                       EIGENVALUES.
!
!      mean(:m)         CONTAINS THE MEAN VALUES OF THE m VARIABLES.
!
!      std(:m)          CONTAINS THE STANDARD-DEVIATIONS OF THE m VARIABLES.
!
!      eigvar(:m)       CONTAINS PERCENTAGES OF TOTAL VARIANCE ASSOCIATED 
!                       WITH THE EIGENVECTORS IN THE ORDER OF THE EIGENVALUES
!                       STORED IN eigval.
!
!
!   COMPUTE THE PRINCIPAL COMPONENTS FROM THE DATA AND THE EIGENVECTORS.
!
!   IN ORDER TO COMPUTE NORMALIZED PCs, SET singval(:m) = sqrt(eigval(:m)),
!   OTHERWISE SET singval(:) = one . sqrt(eigval(:m)) GIVES THE STANDARD-DEVIATIONS
!   ACCOUNTED FOR BY THE PC time series.
!
    singval(:m) = sqrt( eigval(:m) )
!
!   THE OPTIONAL PARAMETERS xmean AND xstd ARE REQUIRED IF THE EIGENVECTORS
!   HAVE BEEN COMPUTED FROM THE CORRELATION MATRIX.
!
    call comp_pc_eof( x(:m,:p), eigvec(:m,:m), singval(:m), pc(:p,:m),   &
                      xmean=mean(:m), xstd=std(:m)                       )
!
!   ON EXIT OF COMP_PC_EOF :
!
!      pc(:p,:m) CONTAINS THE (STANDARDIZED) PRINCIPAL COMPONENTS STORED COLUMNWISE.
!
!   COMPUTE THE FIRST nrot CORRESPONDING FACTORS FROM THE EIGENVECTORS AND THE SINGULAR VALUES.
!
    factor(:m,:nrot) = eigvec(:m,:nrot)*spread( singval(:nrot), dim=1, ncopies=m )
!
!   NOW ROTATE THE FIRST nrot PC TIME SERIES WITH SUBROUTINE comp_filt_rot_pc.
!
!   SPECIFY THE WINDOWED FILTER TO BE USED FOR THE ORTHOGONAL ROTATION.
!
    pl = 0
    ph = 24
!
    call comp_filt_rot_pc( pc(:p,:nrot), singval(:nrot), pl, ph, rot_pc(:p,:nrot),  &
                           rot(:nrot,:nrot), std_rot_pc(:nrot), failure2 )
!
!   NOW ROTATE THE FIRST nrot FACTORS.
!
    rot_factor(:m,:nrot) = matmul( factor(:m,:nrot), rot(:nrot,:nrot) )
!
!   RESCALE THE ORIGINAL PC TIME SERIES FOR THE TESTS.
!
    pc(:p,:m) = pc(:p,:m)*spread( singval(:m), dim=1, ncopies=p )
!
!   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION x_std**(t)*eigvec - pc
!   WHERE eigvec ARE THE EIGENVECTORS, x_std THE STANDARDIZED DATA
!   AND pc THE UNNORMALIZED PRINCIPAL COMPONENTS.
!
    x_std(:m,:p) = x(:m,:p) - spread( mean(:m) ,      dim=2, ncopies=p )
    x_std(:m,:p) = x_std(:m,:p)*spread( one/std(:m) , dim=2, ncopies=p ) 
!
    err_pc =  sum( abs( matmul(transpose(x_std),eigvec) - pc ) )
!
!   CHECK THAT THE VARIANCE EXPLAINED BY THE ORIGINAL AND ROTATED EOFS IS THE SAME.
!
    err_rot1 = abs( sum( std_rot_pc(:nrot)**2 ) - sum( eigval(:nrot) ) )
!
!   CHECK ORTHOGONALITY OF THE ROTATION MATRIX rot COMPUTED BY comp_filt_rot_eof
!   SUBROUTINE.
!
    err_rot2 = abs( sum( abs( matmul(transpose(rot),rot) ) ) - real(nrot,stnd) )
!
!   CHECK THAT THE NORMS OF THE ROTATED PCS ARE UNCHANGED.
!
    err_rot3 =  maxval( abs( norm( rot_pc(:p,:nrot), dim=2 ) - sqrt( real(p,stnd) ) ) )
!
!   CHECK THE COMPUTATION OF VARIANCES EXPLAINED BY THE ROTATED PCS.
!
    err_rot4 = maxval( abs( std_rot_pc(:nrot) - norm( rot_factor(:m,:nrot), dim=2 ) ) )
!
!   PRINT THE RESULT OF THE TESTS.
!
    if ( max(err_pc,err_rot1,err_rot2,err_rot3,err_rot4)<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
!
! END OF PROGRAM ex1_comp_filt_rot_pc
! ===================================
!
end program ex1_comp_filt_rot_pc
