program ex1_comp_sym_ginv
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine COMP_SYM_GINV
!   in module Lin_Procedures for computing a generalized inverse of a symmetric
!   semi-positive definite matrix. 
!
!
! LATEST REVISION : 28/11/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    USE Statpack, ONLY : lgl, i4b, stnd, true, false, zero, one, c10, comp_sym_ginv, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE DIMENSION OF THE GENERATED SEMI-POSITIVE DEFINITE MATRIX
! AND m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX WHICH IS USED TO DERIVE THE SEMI-POSITIVE
! DEFINITE MATRIX. IF m IS LESS THAN n, A SEMI-POSITIVE MATRIX IS GENERATED, ON THE OTHER HAND IF
! m IS GREATER OR EQUAL TO n, A POSITIVE MATRIX IS GENERATED.
!
    integer(i4b), parameter :: prtunit=6, n=1000, m=n-10_i4b
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!
    real(stnd), parameter :: fudge=c10
!
    character(len=*), parameter :: name_proc='Example 1 of comp_sym_ginv'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, tol, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, ata, atainv, res
!
    integer(i4b) :: j, krank
    integer      :: iok, istart, iend, irate
!
    logical(lgl) :: do_test, upper, failure
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : GENERALIZED INVERSE OF A REAL SYMMETRIC SEMI-POSITIVE DEFINITE MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS AND THE TOLERANCE TO BE USED
!   FOR THE TEST OF SINGULARITY IN comp_sym_ginv .
!
    tol = sqrt( epsilon( err ) )
    eps = fudge*tol
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = false
!
!   SPECIFY IF THE UPPER OR LOWER PART OF THE SYMMETRIC
!   DEFINITE POSITIVE MATRIX IS STORED.
!
    upper = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), ata(n,n), atainv(n,n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
! 
!   GENERATE A RANDOM DATA MATRIX a .
!
    call random_number( a )
!
!   GENERATE A n-BY-n SYMMETRIC POSITIVE SEMIDEFINITE MATRIX From a .
!
    ata = matmul( transpose(a), a )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count=istart, count_rate=irate )
!
!   COMPUTE A GENERALIZED INVERSE OF ata WITH SUBROUTINE comp_sym_ginv.
!   INPUT ARGUMENT ata IS NOT OVERWRITTEN.
!
    call comp_sym_ginv( ata, failure, krank, atainv, upper=upper, tol=tol )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    elapsed_time = real( iend - istart, stnd )/real( irate, stnd )
!
    if ( failure ) then
!
!       ANORMAL EXIT FROM comp_sym_ginv SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to COMP_SYM_GINV subroutine, failure=', failure
!
    else if ( do_test ) then
!
!       ALLOCATE WORK ARRAY .
!
        allocate( res(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK THE TWO, IDENTITIES ata*atainv*ata = a AND  atainv*ata*atainv = atainv,
!       WHICH DEFINE THE GENERALIZED INVERSE OF ata.
!
        res = matmul(ata, matmul(atainv,ata)) - ata
        err = sum( abs(res) ) / sum( abs(ata) )
!
        res = matmul(atainv, matmul(ata,atainv)) - atainv
        err = max( sum( abs(res) ) / sum( abs(atainv) ), err )
!
!       CHECK THE RESULTS FOR SMALL RESIDUALS.
!
        err = sum( abs(res) ) / sum( abs(ata) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( res )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, ata, atainv )
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the generalized inverse of a symmetric semi-positive definite matrix of size ', &
       n, ' is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_comp_sym_ginv
! ================================
!
end program ex1_comp_sym_ginv
