program ex1_cur_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines CUR_CMP in 
!   module Random and ORTHO_GEN_QR in module QR_Procedures for computing a
!   (randomized or deterministic) cur decomposition of a data matrix.
!
!
! LATEST REVISION : 14/12/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, seven, c30, c50, allocate_error, &
                         cur_cmp, partial_qr_cmp, ortho_gen_qr, norm, merror, gen_random_mat,     &
                         random_seed_
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!    
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT;
! m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX;
! n IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX;
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX;
! ncur IS THE TARGET RANK OF THE CUR DECOMPOSITION, WHICH IS SOUGHT.
!
    integer(i4b), parameter :: prtunit = 6, m=5000, n=3000, nsvd0=2000, ncur=100
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of cur_cmp'
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, err4, err5, err6, tmp, norma, normr, &
                                               rnorm_row, rnorm_col, eps, ulp, tol, elapsed_time
    real(stnd), allocatable, dimension(:)   :: diagr, beta, singval0
    real(stnd), allocatable, dimension(:,:) :: a, c, u, r, ur, resid
!
    integer(i4b)                            :: i, blk_size, nover, mat_type
    integer(i4b), allocatable, dimension(:) :: ip_row, ip_col
    integer                                 :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test, test_err, random_qr
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTE A (RANDOMIZED OR DETERMINISTIC) CUR DECOMPOSITION
!               OF A DATA MATRIX.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type > 3  -> VERY SLOW DECAY OF SINGULAR VALUES
!
    mat_type = 1_i4b
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE RESULTS OF THE SUBROUTINE.
!
    do_test = true
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( ulp )
    eps = fudge*ulp
!
    err = zero
    test_err = true
!
!   DETERMINE PARAMETERS OF THE CUR DECOMPOSITION ALGORITHM.
!
!   SET TOLERANCE FOR CHECKING THE RANK OF THE CUR APPROXIMATION IN THE SUBROUTINE.
!
    tol = eps
!
!   SPECIFY IF A RANDOMIZED OR DETERMINISTIC CUR ALGORITHM IS USED.
!
    random_qr = false
!
!   DETERMINE THE BLOCKSIZE PARAMETER IN THE RANDOMIZED CUR ALGORITHM.
!
    blk_size = 30_i4b
!
!   DETERMINE THE OVERSAMPLING SIZE PARAMETER IN THE RANDOMIZED CUR ALGORITHM.
!
    nover = 10_i4b
!
!   ALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
!
        i = max( m, n )
!
        allocate( a(m,i), ip_row(m), ip_col(n), singval0(nsvd0), u(ncur,ncur), c(m,ncur), &
                  r(ncur,n), ur(ncur,n), diagr(ncur), beta(ncur), resid(m,i), stat=iok    )
!
    else
!
        allocate( a(m,n), ip_row(m), ip_col(n), singval0(nsvd0), u(ncur,ncur), c(m,ncur), &
                  r(ncur,n), stat=iok )
!
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!   OF SINGULAR VALUES.
!
!   GENERATE SINGULAR VALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
            end do
!
        case default
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            norma = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i - 1_i4b, stnd )
!
                singval0(i) = exp( -tmp/norma )
!        
            end do
!
    end select
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE A m-BY-n MATRIX a OF RANK nsvd0 .
!
    call gen_random_mat( singval0(:nsvd0), a )
!
!   COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
!    norma = norm( a(:m,:n) )
    norma = sqrt(sum( singval0(:nsvd0)**2 ) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE A (RANDOMIZED OR DETERMINISTIC) CUR DECOMPOSITION OF A RANDOM
!   DATA MATRIX a  WITH SUBROUTINE cur_cmp. THE RANK OF THE CUR DECOMPOSITION IS
!   DETERMINED BY THE NUMBER OF ROWS (OR COLUMNS) OF THE ARRAY ARGUMENT u,
!   ncur = size(u,1) = size(u,2) .
!
    call cur_cmp( a(:m,:n), ip_row(:m), ip_col(:n), u(:ncur,:ncur), c=c(:m,:ncur), r=r(:ncur,:n),  &
                  rnorm_row=rnorm_row, rnorm_col=rnorm_col, tol=tol, random_qr=random_qr,          &
                  blk_size=blk_size, nover=nover )
!
!   THE ROUTINE COMPUTES A (RANDOMIZED OR DETERMINISTIC) CUR DECOMPOSITION OF a AS:
!
!                       a ≈ c * u * r
!
!   WHERE c IS A m-BY-ncur MATRIX, u IS A ncur-BY-ncur SQUARED MATRIX AND r IS A ncur-BY-n MATRIX.
!   c and r ARE SELECTED, RESPECTIVELY, AS SUBSETS OF THE COLUMNS AND ROWS OF a. U IS THEN ESTIMATED
!   TO MINIMIZE THE FROBENIUS NORM OF THE ERROR OF THE RESULTING CUR DECOMPOSITION:
!
!                    || a - c * u * r ||_F = min
!
!   SUCH CUR DECOMPOSITION CAN BE COMPUTED EFFICIENTLY WITH THE HELP OF (RANDOMIZED
!   OR DETERMINISTIC) PARTIAL QR DECOMPOSITIONS WITH COLUMN PIVOTING OF a AND a'.
!
!   MORE PRECISELY, A (RANDOMIZED OR DETERMINISTIC) PARTIAL QR DECOMPOSITION OF a IS FIRST COMPUTED AS:
!
!                     a * P ≈ Q * T = Q * [ T11  T12 ]
!
!   WHERE P IS A n-BY-n PERMUTATION MATRIX, Q IS A m-BY-ncur MATRIX WITH ORTHOGONAL COLUMNS,
!   T IS A ncur-BY-n UPPER OR TRAPEZOIDAL MATRIX AND T11 IS A ncur-BY-ncur UPPER TRIANGULAR MATRIX.
!
!   THIS LEADS TO AN ESTIMATE OF c AS:
!
!                     c ≈ a * P(:,:ncur) = Q * T11
!
!   WHERE c IS A m-BY-ncur MATRIX, WHICH CONSISTS OF A SUBSET OF ncur COLUMNS OF a.
!
!   IN A SECOND STEP, A (RANDOMIZED OR DETERMINISTIC) PARTIAL QR DECOMPOSITION OF a' IS COMPUTED AS:
!
!                     a' * N' ≈ K * L = K * [ L11  L12 ]
!
!   WHERE N IS A m-BY-m PERMUTATION MATRIX, K IS A n-BY-ncur MATRIX WITH ORTHOGONAL COLUMNS,
!   L IS A ncur-BY-m UPPER OR TRAPEZOIDAL MATRIX AND L11 IS A ncur-BY-ncur UPPER TRIANGULAR MATRIX.
!
!   THIS LEADS TO AN ESTIMATE OF r AS:
!
!                     r ≈ N(:ncur,:) * a = L11' * K'
!
!   WHERE r IS A ncur-BY-n MATRIX, WHICH CONSISTS OF A SUBSET OF ncur ROWS OF a.
!
!   FINALLY, THE SQUARED MATRIX u IS THEN COMPUTED AS:
!
!                 u(:ncur,:ncur) = pseudo-inv(c) * a * pseudo-inv(r)
!
!   WHERE pseudo-inv(c) AND pseudo-inv(r) ARE THE GENERALIZED INVERSES OF c AND r, SINCE
!   THIS CHOICE LEADS TO MINIMIZE || a - c * u * r ||_F ONCE c AND r ARE KNOWN.
!
!   ON EXIT OF cur_cmp:
!
!   - ip_col STORES THE PERMUTATION MATRIX P IN THE QR DECOMPOSITION OF a.
!     IF ip_col(j)=k, THEN THE jTH COLUMN OF a*P WAS THE kTH COLUMN OF a.
!     THE MATRIX P IS REPRESENTED IN THE ARRAY ip_col AS FOLLOWS:
!     IF ip_col(j) = i THEN THE jTH COLUMN OF P IS THE iTH CANONICAL UNIT VECTOR.
!
!   - ip_row STORES THE PERMUTATION MATRIX N IN THE QR DECOMPOSITION OF a'.
!     IF ip_row(j)=k, THEN THE jTH ROW OF N*a WAS THE kTH ROW OF a.
!     THE MATRIX N IS REPRESENTED IN THE ARRAY ip_row AS FOLLOWS:
!     IF ip_row(j) = i THEN THE jTH row OF N IS THE iTH CANONICAL UNIT VECTOR.
!                      
!   THUS, WE HAVE c = a(:,ip_col(:ncur)) AND r = a(ip_row(:ncur),:) .         
!
!   IF tol IS PRESENT AND IS IN ]0,1[, THEN :
!       CALCULATIONS TO DETERMINE THE CONDITION NUMBERS OF THE SUBMATRICES T11
!       AND L11 ARE PERFORMED. THEN, tol IS USED TO DETERMINE THE EFFECTIVE RANKS
!       OF T11 AND L11, WHICH ARE DEFINED AS THE ORDER OF THE LARGEST LEADING
!       TRIANGULAR SUBMATRICES IN THE PARTIAL QR FACTORIZATIONS WITH PIVOTING OF a AND a',
!       WHOSE ESTIMATED CONDITION NUMBER IS LESS THAN 1/tol.
!
!   IF tol IS PRESENT AND IS EQUAL TO 0, THEN :
!       THE NUMERICAL RANKS OF T11 AND L11 ARE DETERMINED, E.G., CRUDE TESTS ON T(j,j)
!       AND L(j,j) ARE DONE TO DETERMINE THE NUMERICAL RANKS OF T11 AND L11.
!
!   IF tol IS NOT SPECIFIED OR IS OUTSIDE [0,1[ :
!       THE CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF T11 AND L11 ARE NOT
!       PERFORMED AND THE RANKS OF T11 AND L11 ARE ASSUMED TO BE EQUAL TO ncur.
!
!   THE SUBROUTINE WILL EXIT WITH AN ERROR MESSAGE IF THE RANKS OF T11 OR L11 ARE LESS THAN ncur.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   COMPUTE UPPER BOUND FOR THE RELATIVE ERROR OF THE CUR APPROXIMATION.
!
    normr = sqrt( rnorm_row**2 + rnorm_col**2 )
!
    err1 = normr/norma
!                
    if ( do_test ) then
!
!       COMPUTE THE EXACT RELATIVE ERROR OF THE CUR APPROXIMATION.
!
        ur(:ncur,:n) = matmul( u(:ncur,:ncur), r(:ncur,:n) )
!
        resid(:m,:n) = a(:m,:n) - matmul( c(:m,:ncur), ur(:ncur,:n) )
!
        err2 = norm( resid(:m,:n) )/norma
!
!       CHECK COMPUTATION OF MATRIX c .
!
        resid(:m,:ncur) = a(:m,ip_col(:ncur)) - c(:m,:ncur)
!
        err3 = norm( resid(:m,:ncur) )
!
!       CHECK COMPUTATION OF MATRIX r .
!
        resid(:ncur,:n) = a(ip_row(:ncur),:n) - r(:ncur,:n)
!
        err4 = norm( resid(:ncur,:n) )
!
!       RECOMPUTE PARTIAL QR DECOMPOSITION OF MATRIX a .
!
        call partial_qr_cmp( a(:m,:n), diagr(:ncur), beta(:ncur), ip_col(:n), i )
!
!       GENERATE ORTHOGONAL MATRIX Q OF PARTIAL QR DECOMPOSITION OF MATRIX a .
!
        call ortho_gen_qr( a(:m,:m), beta(:ncur) )
!
!       HERE ortho_gen_qr GENERATES AN m-BY-m REAL ORTHOGONAL MATRIX, WHICH IS
!       DEFINED AS THE PRODUCT OF ncur ELEMENTARY REFLECTORS OF ORDER m
!
!            Q = h(1)*h(2)* ... *h(ncur)
!
!       AS RETURNED BY qr_cmp, qr_cmp2, partial_qr_cmp, partial_rqr_cmp, partial_rqr_cmp2
!       partial_rtqr_cmp, id_cmp AND ts_id_cmp SUBROUTINES.
!
!       THE SIZE OF beta DETERMINES THE NUMBER OF ELEMENTARY REFLECTORS WHOSE PRODUCT
!       DEFINES THE MATRIX Q.
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION c(:m,:ncur) - Q(:m,:ncur)*(Q(:m,:ncur)'*c(:m,:ncur)).
!
        ur(:ncur,:ncur) = matmul( transpose(a(:m,:ncur)), c(:m,:ncur) )
!
        resid(:m,:ncur) = abs( c(:m,:ncur) - matmul( a(:m,:ncur), ur(:ncur,:ncur) ) )
!
        err5 = maxval( resid(:m,:ncur) )/real(m,stnd)
!
!       CHECK ORTHOGONALITY OF c(:m,:ncur) WITH ITS ORTHOGONAL COMPLEMENT Q(:m,ncur+1:m).
!
        if ( m>ncur ) then
!
            resid(:ncur,ncur+1_i4b:m) = matmul( transpose(c(:m,:ncur)), a(:m,ncur+1_i4b:m) )
!
            err6 = maxval( abs( resid(:ncur,ncur+1_i4b:m) ) )/real(m,stnd)
!
        else
!
            err6 = zero
!
        end if
!
        test_err = err2 <= err1
!
        if ( random_qr ) then
            err = max( err3, err4 )
        else
            err = max( err3, err4, err5, err6 )
        endif
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( resid, ur, diagr, beta )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, ip_row, ip_col, singval0, c, u, r )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. test_err ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type > 3  -> very slow decay of singular values'
!
    write (prtunit,*) 
    write (prtunit,*) 'Rank of the CUR approximation              &
                       &                                       = ', ncur
!        
    write (prtunit,*) 'Upper bound of the relative error of CUR & 
                       & decomposition ||A - C*U*R||_F/||A||_F   = ', err1
!
    if ( do_test ) then
!        
        write (prtunit,*) 'Exact relative error of CUR decomposition& 
                           & ||A - C*U*R||_F/||A||_F                 = ', err2
!        
        write (prtunit,*) 'Accuracy of the range of the CUR approximation&
                           &                                    = ', err5
!
        if ( m>ncur ) then
            write (prtunit,*) 'Orthogonality of the range of the CUR approximation&
                               & and its orthogonal complement = ', err6
        end if
!
    end if
!
    write (prtunit,*)
    write (*,'(a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing a (randomized) CUR decomposition of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_cur_cmp
! ==========================
!
end program ex1_cur_cmp
