program ex1_eigvalues
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of function EIGVALUES 
!   in module Eig_Procedures for computing eigenvalues of a real symmetric matrix.
!
!
! LATEST REVISION : 22/11/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, svd_cmp, eigvalues, eigval_sort
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE DIMENSIONS OF THE SYMMETRIC MATRIX.
!
    integer(i4b), parameter :: prtunit=6, n=1000
!
    character(len=*), parameter :: name_proc='Example 1 of eigvalues'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)  :: elapsed_time
    real(stnd)  :: a(n,n), a2(n,n), d(n), s(n)
!
    integer :: istart, iend, irate, imax, itime
!
    logical(lgl) :: failure
!   
    character :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   GENERATE A RANDOM DATA MATRIX AND FROM IT
!   A SELF-ADJOINT MATRIX.
!
    call random_number( a2 )
!
    a = a2 + transpose( a2 )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE EIGENVALUES OF THE SELF-ADJOINT MATRIX.
!
    d = eigvalues( a )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   SORT ABSOLUTE VALUES OF EIGENVALUES.
!
    d = abs( d )
!
    call eigval_sort( sort, d )
!
!   FOR COMPARISON, COMPUTE THE SINGULAR VALUES OF THE SYMMETRIC MATRIX.
!
    call svd_cmp( a, s, failure, sort=sort )
!
!   CHECK THE RESULTS: MAGNITUDE OF EIGENVALUES SHOULD EQUAL THE SINGULAR VALUES.
!
    if ( sum(abs(d-s))<=sqrt(epsilon(s))*maxval(abs(s)) ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing all the eigenvalues of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_eigvalues
! ============================
!
end program ex1_eigvalues
