program ex1_fastgivens_mat_right
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine FASTGIVENS_MAT_RIGHT
!   in module Giv_Procedures and subroutine triang_solve in module Lin_Procedures for
!   solving a full rank linear least squares problem by a LQ factorization. The LQ
!   factorization is computed by the application of a series of fast Givens rotations
!   on the right of the coefficient matrix of the problem. 
!
!
! LATEST REVISION : 01/12/2023
!
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, true, false, zero, one, c50, allocate_error, merror,  &
                         fastgivens_mat_right, triang_solve
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT,
! n IS THE NUMBER OF ROWS OF THE RANDOM MATRIX AND
! m IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX.
!
    integer(i4b), parameter :: prtunit=6, n=1000, np1=n+1, m=5000
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of fastgivens_mat_right'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: ulp, eps, err, elapsed_time
    real(stnd), allocatable, dimension(:)   :: b, res, d
    real(stnd), allocatable, dimension(:,:) :: a, syst
!
    integer :: iok, istart, iend, irate, imax, itime
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : SOLVING A LINEAR LEAST SQUARES PROBLEM WITH A n-BY-m REAL COEFFICIENT
!               MATRIX USING A LQ DECOMPOSITION OF THE COEFFICIENT MATRIX.
!
!               COMPUTE SOLUTION VECTOR FOR LINEAR LEAST SQUARES PROBLEM
!
!                       Minimize || b - x*a ||_2
!
!               USING FAST GIVENS PLANE ROTATIONS APPLY FROM THE RIGHT. a IS A n-BY-m MATRIX WHICH IS ASSUMED
!               OF FULL RANK. THE RIGHT HAND SIDE b IS A m-ELEMENTS VECTOR AND THE SOLUTION VECTOR x IS A
!               n-ELEMENTS VECTOR.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( ulp )
    eps = fudge*ulp
!
    err = zero
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,m), syst(np1,m), d(m), b(m), res(m), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a(:n,:m) .
!
    call random_number( a(:n,:m) )
!
!   GENERATE A RANDOM RIGHT HAND-SIDE VECTOR b(:m) .
!
    call random_number( b(:m) )
!
!   MAKE A COPY OF COEFFICIENT MATRIX a(:n,:m) .
!
    syst(:n,:m) = a(:n,:m)
!
!   MAKE A COPY OF RIGHT HAND-SIDE VECTOR b(:m) .
!
    syst(np1,:m) = b(:m)
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE SOLUTION VECTOR FOR LINEAR LEAST SQUARES SYSTEM
!
!                  x(:n)*a(:n,:m) ≈ b(:m)
!
!   FIRST TRANSFORM THE MATRIX a TO LOWER TRAPEZOIDAL FORM BY APPLYING 
!   A SERIES OF FAST GIVENS PLANE ROTATIONS ON THE COLUMNS OF a FROM THE RIGHT
!   AND APPLY THE ROTATIONS TO b .
!
    d(:m) = one
!
    call fastgivens_mat_right( syst(:np1,:m), d(:m) )
!    
!   SOLVE THE n-BY-n LOWER TRIANGULAR SYSTEM.
!    
    call triang_solve( syst(:n,:n), syst(np1,:n), upper=false, trans=true ) 
!    
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE RANGE OF a' .
!
    res(:m) = b(:m) - matmul( syst(np1,:n), a(:n,:m) )
    err = sum(abs(matmul(a(:n,:m),res(:m))) )/ sum( abs(a(:n,:m)) )
!
!   PRINT THE RESULT OF THE TEST.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a linear least squares problem with a ', &
       n, ' by ', m,' real coefficient matrix is ', elapsed_time, ' seconds'
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, syst, d, b, res )
!
!
! END OF PROGRAM ex1_fastgivens_mat_right
! =======================================
!
end program ex1_fastgivens_mat_right
