program ex1_gen_random_mat
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine GEN_RANDOM_MAT
!   in module Random.
!                                                                            
! LATEST REVISION : 02/11/2023
!
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, lgl, stnd, true, false, zero, c50,          &
                         unit_matrix, norm, random_seed_, random_number_, &
                         gen_random_mat, singval_sort, merror,            &
                         allocate_error
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED PSEUDO-RANDOM MATRIX,
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX (WHICH MUST BE LESS OR EQUAL TO k=min(m,n)).
!
    integer(i4b), parameter :: prtunit = 6, m=105, n=50, k=min(m,n), nsvd0=50
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of gen_random_mat'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, anorm, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, leftvec, rightvec, resid
    real(stnd), dimension(:), allocatable   :: s, resnorm
!
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: hous, do_test
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : GENERATION OF A RANDOM REAL MATRIX WITH A SPECIFIED SPECTRUM OF SINGULAR VALUES.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
!   SPECIFY HOW THE SINGULAR VECTORS WILL BE GENERATED.
!
    hous = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), s(nsvd0), leftvec(m,nsvd0), rightvec(n,nsvd0), &
              resid(m,nsvd0), resnorm(nsvd0), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE SINGULAR VALUES FROM THE UNIFORM DISTRIBUTION.
!
    call random_number_( s(:nsvd0) )
!
!   SORT THE SINGULAR VALUES BY DECREASING MAGNITUDE.
!
    call singval_sort( 'D', s(:nsvd0) )
!
!   COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
    anorm = norm( s(:nsvd0) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   GENERATE A m-BY-n RANDOM REAL MATRIX a WITH THE SPECIFIED DISTRIBUTION
!   OF SINGULAR VALUES AND A RANK EQUAL TO nsvd0.
!
    call gen_random_mat( s(:nsvd0), a(:m,:n), leftvec=leftvec(:m,:nsvd0), &
                         rightvec=rightvec(:n,:nsvd0), hous=hous   )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:n)*rightvec(:n,:nsvd0) - leftvec(:m,:nsvd0)*S(:nsvd0,:nsvd0).
!
    resid(:m,:nsvd0) = matmul( a, rightvec ) - leftvec*spread(s,dim=1,ncopies=m)
    resnorm(:nsvd0) = norm( resid(:m,:nsvd0), dim=2_i4b )
!
    err1 =  maxval( resnorm(:nsvd0) )/( anorm*real(m,stnd) )
!
!   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - leftvec**(t)*leftvec.
!
    call unit_matrix( a(:nsvd0,:nsvd0) )
!
    resid(:nsvd0,:nsvd0) = abs( a(:nsvd0,:nsvd0) - matmul( transpose(leftvec), leftvec ) )
!
    err2 = maxval( resid(:nsvd0,:nsvd0) )/real(m,stnd)
!
!   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - rightvec**(t)*rightvec.
!
    resid(:nsvd0,:nsvd0) = abs( a(:nsvd0,:nsvd0) - matmul( transpose(rightvec), rightvec ) )
!
    err3 = maxval( resid(:nsvd0,:nsvd0) )/real(n,stnd)
!
    err = max( err1, err2, err3 )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, s, leftvec, rightvec, resid, resnorm )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')                           &
      'The elapsed time for generating a ', m, ' by ', n,         &
      ' pseudo-random real matrix with a specified spectrum is ', &
      elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_gen_random_mat
! =================================
!
end program ex1_gen_random_mat
