program ex1_gen_random_ortho_mat
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine GEN_RANDOM_ORTHO_MAT
!   in module Random.
!                                                                            
! LATEST REVISION : 02/11/2023
!
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, zero, c50,                         &
                         unit_matrix, random_seed_, random_number_,    &
                         gen_random_ortho_mat, merror, allocate_error
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!
#ifdef _TRANSPOSE
    use Utilities, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED ORTHOGONAL MATRIX.
!
    integer(i4b), parameter :: prtunit=6, m=105, n=50
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of gen_random_ortho_mat'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, ulp, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, resid
!
    integer      :: iok, istart, iend, irate, imax, itime
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : GENERATION OF A REAL ORTHOGONAL MATRIX USING THE HOUSEHOLDER METHOD.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
    err = zero
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), resid(n,n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   GENERATE A RANDOM m-BY-n ORTHOGONAL MATRIX H WITH FUNCTION gen_random_ortho_mat.
!
    a(:m,:n) = gen_random_ortho_mat( m, n )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - H**(t)*H.
!
    call unit_matrix( resid(:n,:n) )
!
    resid(:n,:n) = abs( resid(:n,:n) - matmul( transpose(a(:m,:n)), a(:m,:n) ) )
!
    err = maxval( resid(:n,:n) )/real(n,stnd)
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, resid )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing a ', &
       m, ' by ', n,' real orthogonal matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_gen_random_ortho_mat
! =======================================
!
end program ex1_gen_random_ortho_mat
