program ex1_gen_random_sym_mat
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine GEN_RANDOM_SYM_MAT
!   in module Random.
!                                                                            
! LATEST REVISION : 02/11/2023
!
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, lgl, stnd, true, false, zero, c50,          &
                         unit_matrix, norm, random_seed_, random_number_, &
                         gen_random_sym_mat, eigval_sort, merror,         &
                         allocate_error
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE DIMENSION OF THE GENERATED PSEUDO-RANDOM SYMMETRIC MATRIX,
! neig0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX.
!
    integer(i4b), parameter :: prtunit = 6, n=3000, neig0=3000
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of gen_random_sym_mat'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, anorm, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, eigvec, resid
    real(stnd), dimension(:), allocatable   :: s, resnorm
!
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: hous, do_test
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : GENERATION OF A RANDOM REAL SYMMETRIC MATRIX WITH A SPECIFIED SPECTRUM OF EIGENVALUES.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
!   SPECIFY HOW THE EIGENVECTORS WILL BE GENERATED.
!
    hous = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), s(neig0), eigvec(n,neig0),       &
              resid(n,neig0), resnorm(neig0), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE EIGENVALUES FROM THE UNIFORM DISTRIBUTION.
!
    call random_number_( s(:neig0) )
!
!   SORT THE EIGENVALUES BY DECREASING MAGNITUDE.
!
    call eigval_sort( 'D', s(:neig0) )
!
!   COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
    anorm = norm( s(:neig0) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   GENERATE A n-BY-n RANDOM REAL SYMMETRIC MATRIX a WITH THE SPECIFIED DISTRIBUTION
!   OF EIGENVALUES AND A RANK EQUAL TO neig0.
!
    call gen_random_sym_mat( s(:neig0), a(:n,:n), eigvec=eigvec(:n,:neig0), hous=hous )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:n,:n)*eigvec(:n,:neig0) - eigvec(:n,:neig0)*S(:neig0,:neig0).
!
    resid(:n,:neig0) = matmul( a, eigvec ) - eigvec*spread(s,dim=1,ncopies=n)
    resnorm(:neig0) = norm( resid(:n,:neig0), dim=2_i4b )
!
    err1 =  maxval( resnorm(:neig0) )/( anorm*real(n,stnd) )
!
!   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - eigvec**(t)*eigvec.
!
    call unit_matrix( a(:neig0,:neig0) )
!
    resid(:neig0,:neig0) = abs( a(:neig0,:neig0) - matmul( transpose(eigvec), eigvec ) )
!
    err2 = maxval( resid(:neig0,:neig0) )/real(n,stnd)
!
    err = max( err1, err2 )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, s, eigvec, resid, resnorm )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')                                     &
      'The elapsed time for generating a ', n, ' by ', n,                   &
      ' pseudo-random real symmetric matrix with a specified spectrum is ', &
      elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_gen_random_sym_mat
! =====================================
!
end program ex1_gen_random_sym_mat
