program ex1_id_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines ID_CMP in 
!   module Random and ORTHO_GEN_QR in module QR_Procedures for computing a
!   (deterministic or randomized) column interpolative decomposition.
!
!
! LATEST REVISION : 14/12/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, seven, c30, c50, id_cmp,  &
                         ortho_gen_qr, norm, merror, allocate_error, gen_random_mat,       &
                         random_seed_
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!    
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT;
! m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX;
! n IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX;
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX;
! nid IS THE TARGET RANK OF THE COLUMN INTERPOLATIVE DECOMPOSITION, WHICH IS SOUGHT.
!
    integer(i4b), parameter :: prtunit = 6, m=4000, n=3000, nsvd0=2000, nid=100
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of id_cmp'
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, err4, tmp, norma, normr, &
                                               eps, ulp, tol, elapsed_time
    real(stnd), allocatable, dimension(:)   :: diagr, beta, singval0
    real(stnd), allocatable, dimension(:,:) :: a, t, c, v, resid
!
    integer(i4b)                            :: i, nt, blk_size, nover, mat_type
    integer(i4b), allocatable, dimension(:) :: ip
    integer                                 :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test, random_qr
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTE A (RANDOMIZED OR DETERMINISTIC) COLUMN INTERPOLATIVE
!               DECOMPOSITION (ID) OF A DATA MATRIX.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type > 3  -> VERY SLOW DECAY OF SINGULAR VALUES
!
    mat_type = 3_i4b
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE RESULTS OF THE SUBROUTINE.
!
    do_test = true
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( ulp )
    eps = fudge*ulp
    err = zero
!
!   DETERMINE PARAMETERS OF THE COLUMN INTERPOLATIVE DECOMPOSITION (ID) ALGORITHM.
!
!   SET TOLERANCE FOR CHECKING THE RANK OF THE PARTIAL ID APPROXIMATION IN THE SUBROUTINE.
!
    tol = eps
!
!   SPECIFY IF A RANDOMIZED OR DETERMINISTIC COLUMN ID ALGORITHM IS USED.
!
    random_qr = true
!
!   DETERMINE THE BLOCKSIZE PARAMETER IN THE RANDOMIZED COLUMN ID ALGORITHM.
!
    blk_size = 20_i4b
!
!   DETERMINE THE OVERSAMPLING SIZE PARAMETER IN THE RANDOMIZED COLUMN ID ALGORITHM.
!
    nover = 10_i4b
!
!   ALLOCATE WORK ARRAYS.
!
    nt = n - nid
!
    if ( do_test ) then
!
        i = max( m, n )
!
        allocate( a(m,i), diagr(nid), beta(nid), ip(n), singval0(nsvd0),  &
                  t(nid,nt), c(m,nid), v(nid,n), resid(m,i), stat=iok )
!
    else
!
        allocate( a(m,n), diagr(nid), beta(nid), ip(n), singval0(nsvd0),  &
                  t(nid,nt), c(m,nid), v(nid,n), stat=iok )
!
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!   OF SINGULAR VALUES.
!
!   GENERATE SINGULAR VALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
            end do
!
        case default
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            norma = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i - 1_i4b, stnd )
!
                singval0(i) = exp( -tmp/norma )
!        
            end do
!
    end select
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE A m-BY-n MATRIX a OF RANK nsvd0 .
!
    call gen_random_mat( singval0(:nsvd0), a )
!
!   SAVE THE INPUT MATRIX FOR LATER USE IF REQUIRED.
!
    if ( do_test ) then
!
        resid(:m,:n) = a(:m,:n)
!
    end if
!
!   COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
!    norma = norm( a(:m,:n) )
    norma = sqrt(sum( singval0(:nsvd0)**2 ) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE A (RANDOMIZED OR DETERMINISTIC) COLUMN ID DECOMPOSITION OF A DATA MATRIX a
!   WITH SUBROUTINE id_cmp. THE RANK OF THE COLUMN ID DECOMPOSITION IS
!   DETERMINED BY THE NUMBER OF ROWS OF THE ARRAY ARGUMENT t, nid = size(t,1) .
!
    call id_cmp( a(:m,:n), ip(:n), t(:nid,:nt), c=c(:m,:nid), v=v(:nid,:n), &
                 diagr=diagr(:nid), beta=beta(:nid), rnorm=normr, tol=tol,  &
                 random_qr=random_qr, blk_size=blk_size, nover=nover )
!
!   THE ROUTINE COMPUTES A (RANDOMIZED OR DETERMINISTIC) COLUMN ID DECOMPOSITION OF a AS:
!
!                     a ≈ c * v = c * [ I  t ] * P'
!
!   WHERE c IS A m-BY-nid MATRIX, WHICH CONSISTS OF A SUBSET OF nid COLUMNS OF a,
!   v IS A nid-BY-n MATRIX, I IS THE IDENTITY MATRIX OF ORDER nid, t IS A nid-BY-(n-nid)
!   MATRIX AND P IS A n-BY-n PERMUTATION MATRIX. THE c AND v MATRICES ARE ESTIMATED
!   TO MINIMIZE THE ERROR OF THE COLUMN ID DECOMPOSITION.
!
!   SUCH COLUMN ID DECOMPOSITION CAN BE COMPUTED EFFICIENTLY WITH THE HELP OF A (RANDOMIZED
!   OR DETERMINISTIC) PARTIAL QR DECOMPOSITION WITH COLUMN PIVOTING OF a, WHICH IS DEFINED AS:
!
!                     a * P ≈ Q * R = Q * [ R11  R12 ]
!
!   WHERE Q IS A m-BY-nid MATRIX WITH ORTHOGONAL COLUMNS, R IS A nid-BY-n UPPER OR TRAPEZOIDAL MATRIX
!   AND R11 IS A nid-BY-nid UPPER TRIANGULAR MATRIX.
!
!   THE MATRIX c AND THE SUBMATRIX t IN THE COLUMN ID DECOMPOSITION OF a CAN BE COMPUTED AS:
!
!                     c = Q * R11      AND      t = inv(R11) * R12
!
!   AND THE FROBENIUS NORM OF THE ERROR OF THIS COLUMN ID DECOMPOSITION OF a IS THE SAME
!   AS THAT OF THIS PARTIAL QR DECOMPOSITION OF a.
!
!   ON EXIT OF id_cmp, a HAS BEEN OVERWRITTEN BY DETAILS OF ITS PARTIAL QR FACTORIZATION:
!                      
!   - THE ELEMENTS BELOW AND ON THE DIAGONAL OF THE ARRAY a(:m,:nid) AND THE ARRAY
!     beta(:nid) STORES Q IN FACTORED FORM.
!                      
!   - THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY a CONTAIN THE
!     CORRESPONDING ELEMENTS OF THE TRIANGULAR OR TRAPEZOIDAL MATRIX R.
!     THE ELEMENTS OF THE DIAGONAL OF R ARE STORED IN THE ARRAY diagr. 
!
!   - ip STORES THE PERMUTATION MATRIX P IN THE QR DECOMPOSITION OF a.
!     IF ip(j)=k, THEN THE jTH COLUMN OF a*P WAS THE kTH COLUMN OF a.
!     THE MATRIX p IS REPRESENTED IN THE ARRAY ip AS FOLLOWS:
!     IF ip(j) = i THEN THE jTH COLUMN OF P IS THE iTH CANONICAL UNIT VECTOR.
!                      
!   IF tol IS PRESENT AND IS IN ]0,1[, THEN :
!       CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF THE SUBMATRIX R11
!       ARE PERFORMED. THEN, tol IS USED TO DETERMINE THE EFFECTIVE RANK
!       OF R11, WHICH IS DEFINED AS THE ORDER OF THE LARGEST LEADING
!       TRIANGULAR SUBMATRIX IN THE PARTIAL QR FACTORIZATION WITH PIVOTING OF a,
!       WHOSE ESTIMATED CONDITION NUMBER IS LESS THAN 1/tol.
!       ON EXIT, THE RECIPROCAL OF THE CONDITION NUMBER IS RETURNED IN tol.
!
!   IF tol IS PRESENT AND IS EQUAL TO 0, THEN :
!       THE NUMERICAL RANK OF R IS DETERMINED, E.G., CRUDE TESTS ON R(j,j) ARE
!       DONE TO DETERMINE THE NUMERICAL RANK OF R11 AND tol IS NOT CHANGED ON EXIT.
!
!   IF tol IS NOT SPECIFIED OR IS OUTSIDE [0,1[ THE CALCULATIONS TO DETERMINE THE
!       CONDITION NUMBER OF R11 ARE NOT PERFORMED AND THE RANK OF R11 IS ASSUMED TO
!       BE EQUAL TO nid.
!
!   THE SUBROUTINE WILL EXIT WITH AN ERROR MESSAGE IF THE RANK OF R11 IS LESS THAN nid.
!
!   IT IS POSSIBLE THAT a MAY NOT BE OF FULL RANK (I.E., CERTAIN COLUMNS OF a ARE
!   LINEAR COMBINATIONS OF OTHER COLUMNS) AND THAT R11 IS SINGULAR, THEN THE LINEARLY
!   DEPENDENT COLUMNS CAN USUALLY BE EXCLUDED FROM THE QR (AND ID) APPROXIMATION AND THE
!   RANK OF R11 CAN BE DETERMINED BY USING tol=RELATIVE PRECISION OF THE ELEMENTS IN a.
!   IF EACH ELEMENT IS CORRECT TO, SAY, 5 DIGITS THEN tol=0.00001 SHOULD BE USED.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   COMPUTE RELATIVE ERROR OF THE COLUMN ID APPROXIMATION.
!
    err1 = normr/norma
!                
    if ( do_test ) then
!
!       CHECK ACCURACY OF THE FROBENIUS NORM OF THE RESIDUAL MATRIX.
!
        resid(:m,:n) = resid(:m,:n) - matmul( c(:m,:nid), v(:nid,:n) )
!
        if ( normr<=one ) then
!
            err2 = abs( norm( resid(:m,:n) ) - normr )
!
        else
!
            err2 = abs( norm( resid(:m,:n) )/normr - one )
!
        end if
!
!       GENERATE ORTHOGONAL MATRIX Q OF PARTIAL QR DECOMPOSITION OF DATA MATRIX a .
!
        call ortho_gen_qr( a(:m,:m), beta(:nid) )
!
!       HERE ortho_gen_qr GENERATES AN m-BY-m REAL ORTHOGONAL MATRIX, WHICH IS
!       DEFINED AS THE PRODUCT OF nid ELEMENTARY REFLECTORS OF ORDER m
!
!            Q = h(1)*h(2)* ... *h(nid)
!
!       AS RETURNED BY qr_cmp, qr_cmp2, partial_qr_cmp, partial_rqr_cmp, partial_rqr_cmp2
!       partial_rtqr_cmp, id_cmp AND ts_id_cmp SUBROUTINES.
!
!       THE SIZE OF beta DETERMINES THE NUMBER OF ELEMENTARY REFLECTORS WHOSE PRODUCT
!       DEFINES THE MATRIX Q.
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION c(:m,:nid) - Q(:m,:nid)*(Q(:m,:nid)'*c(:m,:nid)).
!
        t(:nid,:nid) = matmul( transpose(a(:m,:nid)), c(:m,:nid) )
!
        resid(:m,:nid) = abs( c(:m,:nid) - matmul( a(:m,:nid), t(:nid,:nid) ) )
!
        err3 = maxval( resid(:m,:nid) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q.
!
!        call unit_matrix( resid(:m,:m) )
!
!        resid(:m,:m) = abs( resid(:m,:m) - matmul( transpose(a(:m,:m)), a(:m,:m) ) )
!        err3 = maxval( resid(:m,:m) )/real(m,stnd)
!
!       CHECK ORTHOGONALITY OF c(:m,:nid) AND ITS ORTHOGONAL COMPLEMENT Q(:m,nid+1:m).
!
        if ( m>nid ) then
!
            resid(:nid,nid+1_i4b:m) = matmul( transpose(c(:m,:nid)), a(:m,nid+1_i4b:m) )
!
            err4 = maxval( abs( resid(:nid,nid+1_i4b:m) ) )/real(m,stnd)
!
        else
!
            err4 = zero
!
        end if
!
        err = max( err2, err3, err4 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, diagr, beta, ip, singval0, c, v, t )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type > 3  -> very slow decay of singular values'
!
    write (prtunit,*) 
    write (prtunit,*) 'Rank of the column ID approximation      &
                       &                                        = ', nid
!        
    write (prtunit,*) 'Relative error of the column ID decomposition  &
                       &||A - C*V||_F/||A||_F             = ', err1
!
    if ( do_test ) then
!        
        write (prtunit,*) 'Accuracy of the range of the column ID &
                          &approximation                             = ', err3
!
        if ( m>nid ) then
            write (prtunit,*) 'Orthogonality of the range of the ID approximation&
                               & and its orthogonal complement = ', err4
        end if
!
    end if
!
    write (prtunit,*)
    write (*,'(a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing a (randomized) column ID decomposition of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_id_cmp
! =========================
!
end program ex1_id_cmp
