program ex1_lapack_gesv
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of the driver subroutine GESV
!   in LAPACK software for solving a real linear system with the help of a LU decomposition
!   with partial pivoting.
!
!                                                                              
! LATEST REVISION : 06/11/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : gesv
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c100, merror, allocate_error
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE DIMENSION OF THE LINEAR SYSTEM,
! nrhs IS THE NUMBER OF RIGHT HAND SIDES OF THE LINEAR PROBLEM TO BE SOLVED.
!
    integer, parameter :: prtunit = 6, n=5000, nrhs=300
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c100
!
    character(len=*), parameter :: name_proc='Example 1 of gesv'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, b, x, res
!
    integer(i4b), dimension(:), allocatable :: ipiv
    integer                                 :: info, iok, istart, iend, irate, imax, itime
!
    logical(lgl)   :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : SOLUTION OF A LINEAR SYSTEM WITH A REAL COEFFICIENT
!               MATRIX AND SEVERAL RIGHT HAND-SIDES BY A LU DECOMPOSITION
!               WITH PARTIAL PIVOTING.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*real( n, stnd )*epsilon( err )
!    eps = sqrt( epsilon( err ) )
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), b(n,nrhs), x(n,nrhs), ipiv(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM DATA MATRIX a .
!
    call random_number( a )
!
!   GENERATE A n-by-nrhs RANDOM SOLUTION MATRIX x .
!
    call random_number( x )
!
!   COMPUTE THE MATRIX-MATRIX PRODUCT b = a*x .
!
    b = matmul( a, x )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE SOLUTION MATRIX FOR REAL LINEAR SYSTEM
!
!            a*x = b .
!
!   WHERE a IS A n-BY-n MATRIX AND b IS A n-BY-nrhs MATRIX.
!
!   THE LU DECOMPOSITION WITH PARTIAL PIVOTING AND ROW INTERCHANGES
!   IS USED TO FACTOR a AS
!
!          a = P*L*U
!
!   WHERE P IS PERMUTATION MATRIX, L IS UNIT LOWER TRIANGULAR, AND U IS
!   UPPER TRIANGULAR. THE FACTORED FORM OF a IS THEN USED TO SOLVE THE
!   SYSTEM OF EQUATIONS.
!
    call gesv( n, nrhs, a, n, ipiv, b, n, info )
!
!   ON EXIT OF gesv :
!
!      info    = 0 :  INDICATES SUCCESSFUL EXIT
!      info=-i < 0 :  INDICATES THAT the iTH ARGUMENT HAS AN ILLEGAL VALUE
!      info=i  > 0 :  INDICATES THAT THE SOLUTION CAN NOT BE COMPUTED BECAUSE
!                     THE MATRIX IS SINGULAR
!
!   ON EXIT OF gesv IF info = 0:
!
!       a IS OVERWRITTEN BY THE FACTORS L AND U OF ITS FACTORIZATION.
!       THE UNIT DIAGONAL ELEMENTS OF L ARE NOT STORED. 
!
!       ipiv STORES THE PIVOT INDICES THAT DEFINE THE PERMUTATION MATRIX P.
!       ROW i OF THE MATRIX WAS INTERCHANGED WITH ROW ipiv(i).
!
!       b IS OVERWRITTEN WITH THE SOLUTION MATRIX x.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( info/=0 ) then
!
!       ANORMAL EXIT FROM gesv SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to GESV subroutine, Info=', info
!
        if ( info<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value'')') - info
!
        end if
!
    else if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( res(n,nrhs), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK THE RESULTS FOR SMALL RESIDUALS.
!
        res(:n,:nrhs) = b(:n,:nrhs) - x(:n,:nrhs)
        err = maxval( sum( abs(res), dim=1 ) /    &
                      sum(abs(x),    dim=1 )      )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( res )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, b, x, ipiv )
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' INFO ( from gesv() ) = ', info
!
    if ( do_test .and. info==0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed solutions = ', err
    end if
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing the solutions of a linear real system of size ', &
       n, ' with ', nrhs,' right hand sides is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_gesv
! ==============================
!
end program ex1_lapack_gesv
