program ex1_partial_rqr_cmp_fixed_precision
!
!
! Purpose
! =======
!
!   This program illustrates how to compute a randomized partial QR factorization with column
!   pivoting of a matrix, which fullfills a given relative error in Frobenius norm,
!   using subroutine PARTIAL_RQR_CMP_FIXED_PRECISION in module Random.
!                                                                              
!                                                                              
! LATEST REVISION : 18/03/2022
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, two, three, seven, c30, c50,   &
                         allocate_error, merror, norm, unit_matrix, random_seed_, singval_sort, &
                         gen_random_mat, ortho_gen_qr, partial_rqr_cmp_fixed_precision
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX,
! relerr0 IS THE REQUESTED TOLERANCE FOR THE RELATIVE ERROR OF THE RANDOMIZED PARTIAL
! QR FACTORIZATION WITH COLUMN PIVOTING IN FROBENIUS NORM.
!
    integer(i4b), parameter :: prtunit=6, m=5000, n=1000, mn=min(m,n), nsvd0=500
!   
    real(stnd), parameter  :: fudge=c50, relerr0=0.2_stnd
!
    character(len=*), parameter :: name_proc='Example 1 of partial_rqr_cmp_fixed_precision'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: ulp, err, err2, eps, elapsed_time, norma, tmp, relerr, relerr2
    real(stnd), dimension(:,:), allocatable :: a, a2, id, b
    real(stnd), dimension(:),   allocatable :: singval0, diagr, beta
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: nqr, blk_size, nover, i, j, mat_type
    integer(i4b), allocatable, dimension(:) :: ip
!
    logical(lgl) :: do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL QR FACTORIZATION WITH COLUMN PIVOTING OF RANK nqr OF A m-BY-n REAL MATRIX
!               USING A RANDOMIZED ALGORITHM AS
!
!                              a(:m,:n) ≈ q(:m,:nqr)*b(:nqr,:n)
!
!               WHERE q IS A m-BY-nqr MATRIX WITH ORTHONORMAL COLUMNS, b IS A nqr-BY-n MATRIX AND
!               THE PRODUCT q*b IS A GOOD APPROXIMATION OF a ACCORDING TO THE SPECTRAL OR FROBENIUS
!               NORM. The RANK nqr IS DETERMINED SUCH THAT THE ASSOCIATED QR APPROXIMATION FULLFILLS
!               A PRESCRIBED RELATIVE ERROR IN THE FROBENIUS NORM AND IS NOT KNOWN IN ADVANCE.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type > 3  -> VERY SLOW DECAY OF SINGULAR VALUES
!
    mat_type = 2_i4b
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( ulp )
    eps = fudge*ulp
!
    err = zero
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED PARTIAL QR ALGORITHM.
!
!   THE RANK nqr OF THE COMPUTED PARTIAL QR FACTORIZATION WITH COLUMN PIVOTING
!   COMPUTED BY partial_rqr_cmp_fixed_precision IS ALWAYS A MULTIPLE OF blk_size.
!   THUS, CHOOSING blk_size INVOLVES TRADEOFFS BETWEEN SPEED AND A RANK AS SMALL
!   AS POSSIBLE FOR THE COMPUTED PARTIAL QR FACTORIZATION.
!
    blk_size = 20_i4b
!
!   CHOOSE THE OVERSAMPLING SIZE USED IN THE RANDOMIZED PARTIAL QR FACTORIZATION
!   WITH COLUMN PIVOTING.
!
    nover = 10_i4b
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ORTHOGONALITY OF THE q MATRIX
!   AND THE QUALITY OF THE APPROXIMATION.
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        allocate( a(m,n), a2(m,n), ip(n), singval0(nsvd0), diagr(mn), beta(mn), stat=iok )
    else
        allocate( a(m,n), ip(n), singval0(nsvd0), diagr(mn), beta(mn), stat=iok )
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!   OF SINGULAR VALUES.
!
!   GENERATE SINGULAR VALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
            end do
!
        case default
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            norma = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i - 1_i4b, stnd )
!
                singval0(i) = exp( -tmp/norma )
!        
            end do
!
    end select
!
!   SORT THE SINGULAR VALUES.
!
    call singval_sort( sort, singval0(:nsvd0) )
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE A m-BY-n MATRIX a OF RANK nsvd0 .
!
    call gen_random_mat( singval0(:nsvd0), a )
!
!   COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
    norma = norm( singval0(:nsvd0) )
!
    if ( do_test ) then
!
!       SAVE THE MATRIX FOR LATER USE IF REQUIRED.
!
        a2 = a
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   partial_rqr_cmp_fixed_precision COMPUTES AN APPROXIMATE PARTIAL QR DECOMPOSITION (WITH COLUMN PIVOTING) OF A REAL
!   m-BY-n MATRIX a, WITH A RANK AS SMALL AS POSSIBLE, BUT WHICH FULFILLS A PRESET TOLERANCE FOR ITS RELATIVE ERROR
!   IN THE FROBENIUS NORM:
!
!                                || A - Q*B ||_F <= ||A||_F * relerr
!
!   , WHERE Q*B IS THE COMPUTED PARTIAL QR APPROXIMATION, || ||_F IS THE FROBENIUS NORM AND
!   relerr IS A PRESCRIBED ACCURACY TOLERANCE FOR THE RELATIVE ERROR OF THE COMPUTED PARTIAL QR
!   APPROXIMATION, SPECIFIED IN THE INPUT ARGUMENT relerr.
!
!   HERE THE RANK, nqr, OF THE RANDOMIZED PARTIAL QR DECOMPOSITION IS NOT KNOWN IN ADVANCE AND
!   IS DETERMINED IN THE SUBROUTINE.
!
!   FIRST, SET THE TOLERANCE FOR THE RELATIVE ERROR OF THE PARTIAL QR FACTORIZATION IN FROBENIUS NORM.
!
    relerr = relerr0
!
    call partial_rqr_cmp_fixed_precision( a(:m,:n), relerr, diagr(:mn), beta(:mn), ip(:n), nqr, &
                                          blk_size=blk_size, nover=nover )
!
!   THE ROUTINE RETURNS THE TWO FACTORS OF THE PARTIAL QR DECOMPOSITION, WHICH FULFILLS
!   THE PRESET TOLERANCE SPECIFIED IN ARGUMENT relerr, IN FACTORED FORM IN ARRAYS a, diagr, beta AND ip.
!
!   ON EXIT OF partial_rqr_cmp_fixed_precision relerr CONTAINS THE RELATIVE ERROR IN FROBENIUS NORM
!   OF THE COMPUTED PARTIAL QR DECOMPOSITION.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   COMPUTE THE BEST RELATIVE ERROR FROM THE TRUNCATED SVD OF RANK nqr.
!
    relerr2 = norm( singval0(nqr+1_i4b:nsvd0)/norma )
!
!   TEST ACCURACY OF THE Q*B APPROXIMATION AND ORTHOGONALITY OF MATRIX Q IF REQUIRED.
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( b(nqr,n), id(nqr,nqr), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       RESTORE PERMUTED TRIANGULAR FACTOR R OF QR DECOMPOSITION OF RANDOM DATA MATRIX a
!       IN MATRIX b(:nqr,:n) .
!
        do j = 1_i4b, nqr
!
            b(1_i4b:j-1_i4b,ip(j)) = a(1_i4b:j-1_i4b,j)
            b(j,ip(j))             = diagr(j)
            b(j+1_i4b:nqr,ip(j))   = zero
!
        end do
!
        do j = nqr+1_i4b, n
!
            b(1_i4b:nqr,ip(j)) = a(1_i4b:nqr,j)
!
        end do
!
!       GENERATE ORTHOGONAL MATRIX Q OF PARTIAL QR DECOMPOSITION OF RANDOM DATA MATRIX a .
!
        call ortho_gen_qr( a(:m,:nqr), beta(:nqr) )
!
!       RECOMPUTE THE RELATIVE ERROR OF THE QR APPROXIMATION Q*B .
!
        a2(:m,:n) = a2(:m,:n) - matmul( a(:m,:nqr), b(:nqr,:n) )
!
!       CHECK ACCURACY OF THE RELATIVE ERROR COMPUTED BY partial_rqr_cmp_fixed_precision.
!
        err = abs( norm( a2(:m,:n) )/norma - relerr )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q(:m,:nqr)**(t)*q(:m,:nqr).
!
        call unit_matrix( id(:nqr,:nqr) )
!
        b(:nqr,:nqr) = abs( id(:nqr,:nqr) - matmul( transpose(a(:m,:nqr)), a(:m,:nqr) ) )
!
        err2 = maxval( b(:nqr,:nqr) )/real(m,stnd)
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( id, b, a2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, diagr, beta, singval0, ip )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( relerr<=relerr0 .and. err<=eps*norma ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type > 3  -> very slow decay of singular values'
!
    write (prtunit,*) 
    write (prtunit,*) 'Requested relative error in Frobenius norm                          = ', &
                      relerr0
    write (prtunit,*) 'Block size used in the randomized partial QR decomposition          = ', &
                      blk_size
    write (prtunit,*) 'Rank of the partial QR decomposition with column pivoting           = ', &
                      nqr
    write (prtunit,*) 'Relative error in Frobenius norm     : ||A-Q*B||_F / ||A||_F        = ', &
                      relerr
    write (prtunit,*) 'Best relative error in Frobenius norm: min( ||A-Q*B||_F / ||A||_F ) = ', &
                      relerr2
!
    if ( do_test ) then
        write (prtunit,*)
        write (prtunit,*) 'Orthogonality of the computed orthonormal matrix Q = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing a randomized partial QR approximation of rank ', nqr, ' of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_partial_rqr_cmp_fixed_precision
! ==================================================
!
end program ex1_partial_rqr_cmp_fixed_precision
