program ex1_qlp_cmp2
!
!
! Purpose
! =======
!
!   This program illustrates how to compute a partial QLP decomposition
!   using subroutine QLP_CMP2 in module SVD_Procedures.
!
!
! LATEST REVISION : 14/12/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, seven, c30, c1_e6, merror,  &
                         allocate_error, norm, unit_matrix, random_seed_, random_number_,    &
                         gen_random_mat, qlp_cmp2, singval_sort
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX (WHICH MUST BE LESS OR EQUAL TO mn),
! nqlp IS THE TARGET RANK OF THE PARTIAL QLP FACORIZATION, WHICH IS SOUGHT.
!
    integer(i4b), parameter :: prtunit=6, m=3000, n=3000, mn=min(m,n), nsvd0=300, nqlp=20
!   
! conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN CASES 5, 6 AND 7.
!   
    real(stnd), parameter  :: conda=c1_e6
!
    character(len=*), parameter :: name_proc='Example 1 of qlp_cmp2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err, err1, err2, eps, elapsed_time, anorm, lnorm, tmp, tmp2, &
                  relerr, relerr2, abs_err, rel_err
    real(stnd), dimension(:,:), allocatable :: a, qmat, lmat, pmat, res, id
    real(stnd), dimension(:),   allocatable :: singval0, lval
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: niter_qrql, blk_size, nover, i, mat_type
!
    logical(lgl) :: random_qr, truncated_qr, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL QLP DECOMPOSITION OF A m-BY-n REAL MATRIX USING A RANDOMIZED QR
!               DETERMINISTIC ALGORITHM IN THE FIRST STAGE OF THE ALGORITHM AND QR-QL ITERATIONS
!               IN A FINAL STAGE FOR IMPROVING THE ACCURACY OF THE L-VALUES.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type = 4  -> VERY SLOW DECAY OF SINGULAR VALUES
!   mat_type = 5  -> STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda
!                    SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE
!   mat_type = 6  -> GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type = 7  -> ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type > 7  -> UNIFORM DISTRIBUTION OF SINGULAR VALUES
!
    mat_type = 2_i4b
!
!   SET TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM
!   OF THE COMPUTED PARTIAL QLP FACTORIZATION COMPARED TO THE BEST SVD
!   APPROXIMATION.
!
    eps = 0.05_stnd
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE QLP DECOMPOSITION.
!
    do_test = true
!
!   CHOOSE TUNING PARAMETERS FOR THE QLP ALGORITHM.
!
!   CHOOSE THE NUMBER OF QR-QL ITERATIONS TO BE PERFORMED FOR IMPROVING THE QUALITY OF THE L-VALUES.
!
    niter_qrql = 4_i4b
!
!   DETERMINE IF A RANDOMIZED PARTIAL QR ALGORITHM IS USED IN THE FIRST PHASE OF THE QLP DECOMPOSITION.
!
    random_qr = true
!
!   DETERMINE IF A RANDOMIZED PARTIAL AND TRUNCATED QR ALGORITHM IS USED IN THE FIRST PHASE OF
!   THE QLP ALGORITHM.
!
    truncated_qr = false
!
!   DETERMINE THE BLOCK SIZE USED IN THE RANDOMIZED PARTIAL QR PHASE OF THE ALGORITHM IF
!   random_qr IS SET TO true.
!
    blk_size = 60_i4b
!    blk_size = 30_i4b
!
!   DETERMINE THE OVERSAMPLING SIZE nover FOR THE RANDOMIZED PARTIAL QR PHASE OF THE ALGORITHM IF
!   random_qr IS SET TO true.
!
    nover = 10_i4b
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), qmat(m,nqlp), pmat(nqlp,n), lmat(nqlp,nqlp), &
              singval0(nsvd0), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!   OF SINGULAR VALUES AND A RANK EQUAL TO nsvd0.
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE SINGULAR VALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
            end do
!
        case( 4_i4b )
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            tmp = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                singval0(i) = exp( -tmp2/tmp )
!        
            end do
!
        case( 5_i4b )
!
!           STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE.
!
            singval0(:nsvd0-1_i4b) = one
            singval0(nsvd0) = one/conda
!
        case( 6_i4b )
!
!           GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                singval0(i) = conda**( -tmp2/tmp )
!        
            end do
!
        case( 7_i4b )
!
!           ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
            tmp2 = one - one/conda
!
            do i = 1_i4b, nsvd0
!
                singval0(i) = one - (real( i - 1_i4b, stnd )/tmp)*tmp2
!        
            end do
!
        case default
!
!           UNIFORM DISTRIBUTION OF SINGULAR VALUES.
!
            call random_number_( singval0(:nsvd0) )
!
    end select
!
#ifdef _F2003
    if ( ieee_support_datatype( singval0 ) ) then
!
        if ( .not.all( ieee_is_normal( singval0(:nsvd0) ) ) ) then
!
            call merror( name_proc//' : Exceptions occurred when generating the input matrix !'  )                    
!
        end if
!
    end if
#endif
!
!   SORT THE SINGULAR VALUES BY DECREASING MAGNITUDE.
!
    call singval_sort( 'D', singval0(:nsvd0) )
!
!   GENERATE A m-BY-n MATRIX a OF RANK nsvd0 .
!
    call gen_random_mat( singval0(:nsvd0), a )
!
!   COMPUTE THE FROBENIUS NORM OF THE GENERATED MATRIX.
!
    anorm = norm( singval0(:nsvd0) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   qlp_cmp2 COMPUTES A PARTIAL QLP DECOMPOSITION OF A REAL m-BY-n MATRIX a. 
!   THE PARTIAL QLP IS WRITTEN
!
!                       a  ≈ Q * L * P
!
!   WHERE L IS AN nqlp-BY-nqlp LOWER TRIANGULAR MATRIX WHOSE DIAGONAL ELEMENTS (IN ABSOLUTE VALUE) ARE
!   GOOD APPROXIMATIONS OF THE nqlp LARGEST SINGULAR VALUES OF a SORTED IN DECREASING ORDER (E.G,. THE
!   SO-CALLED L-VALUES), Q IS AN m-BY-nqlp ORTHONORMAL MATRIX, AND L IS AN nqlp-BY-n ORTHONORMAL MATRIX
!   STORED ROWWISE. THE QUALITY OF L-VALUES CAN BE IMPROVED BY ADDITIONAL QR-QL ITERATIONS IF REQUIRED.
!
    call qlp_cmp2( a(:m,:n), lmat(:nqlp,:nqlp), qmat(:m,:nqlp), pmat(:nqlp,:n),           &
                   niter_qrql=niter_qrql, random_qr=random_qr, truncated_qr=truncated_qr, &
                   blk_size=blk_size, nover=nover )
!
!   THE ROUTINE RETURNS THE QLP FACTORIZATION EXPLICITLY IN THE ARRAY ARGUMENTS lmat, qmat AND pmat SPECIFIED
!   IN INPUT OF qlp_cmp2 SUBROUTINE.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   COMPUTE THE ESTIMATED RELATIVE ERROR IN FROBENIUS NORM
!   FOR THE PARTIAL QLP DECOMPOSITION OF RANK nqlp.
!
    lnorm = norm( lmat(:nqlp,:nqlp) )
!
    tmp = one -  (lnorm/anorm)**2
    relerr = sqrt( max( tmp, zero ) )
!
!   COMPUTE THE BEST RELATIVE ERROR IN FROBENIUS NORM
!   FOR A PARTIAL SVD DECOMPOSITION OF RANK nqlp.
!
    if ( nsvd0>nqlp ) then
        relerr2 = norm( singval0(nqlp+1_i4b:nsvd0)/anorm )
    else
        relerr2 = zero
    end if
!
!   COMPUTE ERROR BETWEEN THE BEST AND QLP RELATIVE ERRORS.
!
    err = abs( relerr - relerr2 )
!
!   TEST ACCURACY OF THE QLP FACTORS IF REQUIRED.
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( lval(nqlp), res(m,nqlp), id(nqlp,nqlp), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       GET THE L-VALUES.
!
        do i = 1_i4b, nqlp
            lval(i) = abs( lmat(i,i) )
        end do
!
!       COMPUTE ERRORS FOR THE L-VALUES AS ESTIMATES OF THE SINGULAR VALUES.
!
        i = min( nqlp, nsvd0 )
!
        id(:i,1_i4b) = singval0(:i)
        id(i+1_i4b:nqlp,1_i4b) = zero
!
        where( id(:nqlp,1_i4b)/=zero )
            res(:nqlp,1_i4b) = id(:nqlp,1_i4b)
        elsewhere
            res(:nqlp,1_i4b) = one
        end where
!
!       ABSOLUTE ERRORS OF L-VALUES AS ESTIMATES OF SINGULAR VALUES.
!
        abs_err = maxval( abs( lval(:nqlp) - id(:nqlp,1_i4b) ) )
!
!       RELATIVE ERRORS OF L-VALUES AS ESTIMATES OF SINGULAR VALUES.
!
        rel_err = maxval( abs( (lval(:nqlp) - id(:nqlp,1_i4b))/res(:nqlp,1_i4b) ) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q(:m,:nqlp)**(t)*q(:m,:nqlp).
!
        call unit_matrix( id(:nqlp,:nqlp) )
!
        res(:nqlp,:nqlp) = abs( id(:nqlp,:nqlp) - matmul( transpose(qmat), qmat ) )
!
        err1 = maxval( res(:nqlp,:nqlp) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - p(:nqlp,:n)*p(:nqlp,:n)**(t).
!
        res(:nqlp,:nqlp) = abs( id(:nqlp,:nqlp) - matmul( pmat, transpose(pmat) ) )
!
        err2 = maxval( res(:nqlp,:nqlp) )/real(n,stnd)
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( lval, res, id )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, singval0, lmat, qmat, pmat )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type = 4  -> very slow decay of singular values'
    write (prtunit,*) ' Matrix type = 5  -> strongly clustered singular values at 1'
    write (prtunit,*) '                     spectrum with few deflations (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 6  -> geometric distribution of singular values'
    write (prtunit,*) '                     spectrum with moderate deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 7  -> arithmetic distribution of singular values'
    write (prtunit,*) '                     spectrum with complete deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type > 7  -> uniform distribution of singular values'
!
    write (prtunit,*) 
    write (prtunit,*) 'Rank of the partial QLP approximation                                = ', &
                      nqlp
    write (prtunit,*) 'Number of QR-QL iterations performed                                 = ', &
                      niter_qrql
    write (prtunit,*) 'Relative error in Frobenius norm     : ||A-QLP||_F / ||A||_F         = ', &
                      relerr
    write (prtunit,*) 'Best relative error in Frobenius norm: min( ||A-rSVD||_F / ||A||_F ) = ', &
                      relerr2
!
    if ( do_test ) then
        write (prtunit,*)
        write (prtunit,*) 'Absolute accuracy of the computed L-values    = ', abs_err
        write (prtunit,*) 'Relative accuracy of the computed L-values    = ', rel_err
        write (prtunit,*) 'Orthogonality of the computed Q matrix        = ', err1
        write (prtunit,*) 'Orthogonality of the computed P matrix        = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing a partial QLP approximation of rank ', nqlp, ' of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_qlp_cmp2
! ===========================
!
end program ex1_qlp_cmp2
