program ex1_qr_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines QR_CMP and 
!   ORTHO_GEN_QR in module QR_Procedures for computing a QR decomposition of
!   a real matrix.
!
!
! LATEST REVISION : 15/11/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, c50, qr_cmp, ortho_gen_qr,   &
                         norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!    
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
! m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX.
! n IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX.
!
    integer(i4b), parameter :: prtunit = 6, m=4000, n=3000
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of qr_cmp'
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, ulp, elapsed_time
    real(stnd), allocatable, dimension(:)   :: diagr, beta, resid2, norma
    real(stnd), allocatable, dimension(:,:) :: a, q, r, resid
!
    integer(i4b) :: k, j, l
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTE A FULL QR DECOMPOSITION OF A DATA MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
!
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SUBROUTINE.
!
    do_test = true
!
    k = min( m, n )
    l = max( m, n )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), diagr(k), beta(k), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a(:m,:n) .
!
    call random_number( a(:m,:n) )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( r(k,n), q(m,l), resid(m,l), resid2(n), norma(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX FOR LATER USE.
!
        resid(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX WITH SUBROUTINE qr_cmp.
!
    call qr_cmp( a(:m,:n), diagr(:k), beta(:k) )
!
!   qr_cmp COMPUTES AN ORTHOGONAL FACTORIZATION OF A REAL m-BY-n MATRIX
!   BY THE HOUSEOLDER METHOD. THE MATRIX IS ASSUMED OF FULL RANK. THE ROUTINE
!   COMPUTES A QR FACTORIZATION OF a AS:
!
!                     a = Q * R
!
!   Q IS A m-BY-m ORTHOGONAL MATRIX AND R IS A m-BY-n UPPER TRIANGULAR OR
!   TRAPEZOIDAL MATRIX.
!
!   ON EXIT, a HAS BEEN OVERWRITTEN BY DETAILS OF ITS ORTHOGONAL
!   FACTORIZATION. 
!
!   THE MATRIX Q IS REPRESENTED AS A PRODUCT OF ELEMENTARY REFLECTORS
!
!            Q = h(1)*h(2)* ... *h(k), WHERE k = min( size(a,1) , size(a,2) )
!
!   EACH h(i) HAS THE FORM
!
!            h(i) = I + BETA * ( V * V' ) ,
!                      
!   WHERE BETA IS A REAL SCALAR AND V IS A REAL m-ELEMENT VECTOR WITH V(1:i-1) = 0.
!   V(i:m) IS STORED ON EXIT IN a(i:m,i) AND BETA IN beta(i).
!                      
!   THE ELEMENTS ABOVE THE DIAGONAL IN THE ARRAY a CONTAIN THE
!   CORRESPONDING ELEMENTS OF THE TRIANGULAR MATRIX R. THE ELEMENTS
!   OF THE DIAGONAL OF R ARE STORED IN THE ARRAY diagr ON EXIT.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!                
    if ( do_test ) then
!
!       NOW, RESTORE TRIANGULAR FACTOR R OF QR DECOMPOSITION OF DATA MATRIX a
!       IN MATRIX r(:k,:n) .
!
        do j = 1_i4b, k
!
            r(1_i4b:j-1_i4b,j) = a(1_i4b:j-1_i4b,j)
            r(j,j)             = diagr(j)
            r(j+1_i4b:k,j)     = zero
!
        end do
!
        do j = k+1_i4b, n
!
            r(1_i4b:k,j) = a(1_i4b:k,j)
!
        end do
!
        q(:m,:k) = a(:m,:k)
!
!       GENERATE ORTHOGONAL MATRIX Q OF THE QR DECOMPOSITION OF DATA MATRIX a
!       AND AN ORTHOGONAL BASIS FOR THE ORTHOGONAL COMPLEMENT TO THE RANGE OF a .
!       a IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( q(:m,:m), beta(:k) )
!
!       ortho_gen_qr GENERATES AN m-BY-m REAL ORTHOGONAL MATRIX, WHICH IS
!       DEFINED AS THE PRODUCT OF k ELEMENTARY REFLECTORS OF ORDER m
!
!            Q = h(1)*h(2)* ... *h(k)
!
!       AS RETURNED BY qr_cmp OR qr_cmp2.
!
!       THE SIZE OF beta DETERMINES THE NUMBER k OF ELEMENTARY REFLECTORS
!       WHOSE PRODUCT DEFINES THE MATRIX Q.
!
!       NOW q(:m,:krank) IS AN ORTHONORMAL BASIS FOR THE RANGE OF a AND q(:m,krank+1:m) IS
!       AN ORTHONORMAL BASIS FOR THE ORTHOGONAL COMPLEMENT TO THE RANGE OF a .
!
!       RESTORE THE INPUT MATRIX IN a(:m,:n) .
!
        a(:m,:n) = resid(:m,:n)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:n) - q(:m,:k)*r(:k,:n).
!
        resid(:m,:n) = a(:m,:n) - matmul( q(:m,:k), r(:k,:n) )
        resid2(:n)   = norm( resid(:m,:n), dim=2_i4b )
        norma(:n)    = norm( a(:m,:n), dim=2_i4b )
        err1         = maxval( resid2(:n) / norma(:n) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q'*Q.
!
        resid(:m,:m) = abs( resid(:m,:m) - matmul( transpose(q(:m,:m)), q(:m,:m) ) )
!
        do j = 1, m
            resid(j,j) =  resid(j,j) - one
        end do
!
        err2 = maxval( resid(:m,:m) )/real(m,stnd)
!
!       CHECK ORTHOGONALITY OF a(:m,:n) AND ITS ORTHOGONAL COMPLEMENT q(:m,n+1:m).
!
        if ( m>n ) then
!
            resid(:n,n+1_i4b:m) = matmul( transpose(a(:m,:n)), q(:m,n+1_i4b:m) )
            err3 = maxval( abs( resid(:n,n+1_i4b:m) ) )/real(m,stnd)
!
        else
!
            err3 = zero
!
        end if
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( r, q, resid, resid2, norma )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, diagr, beta )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
!
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the QR decomposition       &
                          &                                = ', err1
        write (prtunit,*) 'Orthogonality of the Q matrix          &
                          &                                = ', err2
!
        if ( m>n ) then
            write (prtunit,*) 'Orthogonality of the range of the matrix&
                              & and its orthogonal complement = ', err3
        end if
!
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing a QR decomposition of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_qr_cmp
! =========================
!
end program ex1_qr_cmp
