program ex1_rqb_solve
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine RQB_CMP in module Random
!   and RQB_SOLVE in module LLSQ_Procedures.
!                                                                              
!                                                                              
! LATEST REVISION : 05/02/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, lgl, stnd, zero, true, false, zero, one, seven, ten, c30, c100, &
                         allocate_error, merror, rqb_cmp, rqb_solve, norm, random_seed_,      &
                         random_number_ , singval_sort, gen_random_mat
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
! m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX.
! n IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX.
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED RANDOM MATRIX,
! nrhs IS THE NUMBER OF COLUMNS OF THE RIGHT HAND SIDE MATRIX.
!
    integer(i4b), parameter :: prtunit=6, m=4000, n=3000, mn=min( m, n ), nsvd0=1000, nrhs=10
!   
    real(stnd), parameter  :: fudge=c100
!
    character(len=*), parameter :: name_proc='Example 1 of rqb_solve'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: ulp, eps, tol, err1, err2, tmp, anorm, cnorm, elapsed_time
    real(stnd), allocatable, dimension(:)   :: singval0, tau
    real(stnd), allocatable, dimension(:,:) :: a, q, b, c, x, res
!
    integer(i4b)                            :: nqb, niter, mat_type, i, j, l
    integer(i4b), allocatable, dimension(:) :: ip
    integer                                 :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: ortho, comp_qr, pivoting
!   
    character :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : SOLVING LINEAR LEAST SQUARES PROBLEMS WITH A m-BY-n REAL COEFFICIENT
!               MATRIX USING A RANDOMIZED COMPLETE ORTHOGONAL DECOMPOSITION OF THE COEFFICIENT
!               MATRIX AND SEVERAL RIGHT HAND-SIDES. THE COEFFICIENT MATRIX CAN BE RANK DEFICIENT
!               AND BOTH m>=n OR m<n ARE PERMITTED.
!
!               COMPUTE APPROXIMATE SOLUTION MATRIX x FOR LINEAR LEAST SQUARES SYSTEM
!
!                              a(:m,:n)*x(:n,:nrhs) ≈ c(:m,:nrhs) .
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type > 3  -> VERY SLOW DECAY OF SINGULAR VALUES
!
    mat_type = 3_i4b
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( ulp )
    eps = fudge*ulp
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED QB OR QR ALGORITHM.
!
!   DETERMINE THE TARGET RANK OF THE RANDOMIZED PARTIAL QB/QR FACTORIZATION, WHICH IS SOUGHT.
!
    nqb = 40_i4b
!
!   DETERMINE THE NUMBER OF SUBSPACE ITERATIONS niter TO BE PERFORMED.
!
    niter = 2_i4b
!
!   DETERMINE IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE ITERATIONS, TO AVOID LOSS
!   OF ACCURACY DUE TO ROUNDING ERRORS.
!
    ortho = true
!
!   DETERMINE IF A QB OR QR FACTORIZATION IS COMPUTED.
!
    comp_qr = true
!
!   DETERMINE IF A QR FACTORIZATION WITH COLUMN PIVOTING IS COMPUTED.
!
    pivoting = true
!
!   SET TOLERANCE FOR DETERMINING THE RANK OF THE B OR R FACTOR IN THE rqb_cmp SUBROUTINE.
!
    tol = ulp
!
!   ALLOCATE WORK ARRAYS.
!
    if ( pivoting ) then
!
        allocate( a(m,n), q(m,nqb), b(nqb,n), c(m,nrhs), x(n,nrhs), singval0(nsvd0),  &
                  tau(nqb), ip(n), res(nrhs,n), stat=iok )
!
    else
!
        allocate( a(m,n), q(m,nqb), b(nqb,n), c(m,nrhs), x(n,nrhs), singval0(nsvd0),  &
                  tau(nqb), res(nrhs,n), stat=iok )
!
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!   OF SINGULAR VALUES.
!
!   GENERATE SINGULAR VALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
            end do
!
        case default
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            err1 = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i - 1_i4b, stnd )
!
                singval0(i) = exp( -tmp/err1 )
!        
            end do
!
    end select
!
!   SORT THE SINGULAR VALUES.
!
    call singval_sort( sort, singval0(:nsvd0) )
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE A m-BY-n MATRIX a OF RANK nsvd0 .
!
    call gen_random_mat( singval0(:nsvd0), a(:m,:n) )
!
!   COMPUTE THE FROBENIUS NORM OF THE COEFFICIENT MATRIX.
!
!    anorm = norm( a(:m,:n) )
    anorm = sqrt(sum( singval0(:nsvd0)**2 ) )
!
!   GENERATE A RIGHT HAND-SIDE MATRIX.
!
    call random_number_( c(:m,:nrhs) )
!
!   COMPUTE FROBENIUS NORM OF THE RIGHT HAND-SIDE MATRIX.
!
    cnorm = norm( c(:m,:nrhs) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE A RANDOMIZED COMPLETE ORTHOGONAL DECOMPOSITION OF RANDOM DATA MATRIX
!   WITH SUBROUTINE rqb_cmp.
!
!   rqb_cmp COMPUTES AN APPROXIMATE PARTIAL QB, QR OR COMPLETE ORTHOGONAL DECOMPOSITION OF A REAL m-BY-n
!   MATRIX a USING A RANDOMIZED ALGORITHM AND SUBSPACE ITERATIONS. THE PARTIAL QB OR QR DECOMPOSITION
!   IS WRITTEN
!
!                  a(:m,:n) ≈ q(:m,:nqb)*b(:nqb,:n)
!
!   WHERE q IS A m-BY-nqb MATRIX WITH ORTHONORMAL COLUMNS, b IS A nqb-BY-n (FULL OR TRAPEZOIDAL)
!   MATRIX AND THE PRODUCT q*b IS A GOOD APPROXIMATION OF a ACCORDING TO THE SPECTRAL OR FROBENIUS
!   NORM.
!
    if ( pivoting ) then
!
        call rqb_cmp( a(:m,:n), q(:m,:nqb), b(:nqb,:n), niter=niter, ortho=ortho, ip=ip(:n), &
                      tol=tol, tau=tau(:nqb) )
!
    else
!
        call rqb_cmp( a(:m,:n), q(:m,:nqb), b(:nqb,:n), niter=niter, ortho=ortho, comp_qr=comp_qr, &
                      tol=tol, tau=tau(:nqb) )
!
    end if
!
!   HERE THE ROUTINE COMPUTES AN APPROXIMATE COMPLETE ORTHOGONAL FACTORIZATION OF a USING
!   A RANDOMIZED ALGORITHM.
!                      
!   THIS APPROXIMATE COMPLETE ORTHOGONAL FACTORIZATION CAN THEN BE USED TO COMPUTE THE MINIMAL 2-NORM
!   SOLUTIONS FOR (RANK DEFICIENT) LINEAR LEAST SQUARES PROBLEMS WITH SUBROUTINE rqb_solve.
!
!   NEXT, COMPUTE THE SOLUTION MATRIX FOR LINEAR LEAST SQUARE SYSTEM
!
!                   a(:m,:n)*x(:n,:nrhs) ≈ b(:m,:nrhs) .
!
!   WITH SUBROUTINE rqb_solve AND THE APPROXIMATE COMPLETE ORTHOGONAL DECOMPOSITION OUTPUT BY rqb_cmp.
!
    call rqb_solve( q(:m,:nqb), b(:nqb,:n), c(:m,:nrhs), x(:n,:nrhs), ip(:n),  &
                    tau=tau(:nqb), comp_resid=true )
!
!   rqb_solve COMPUTES APPROXIMATES SOLUTION(S) TO REAL LINEAR LEAST SQUARES PROBLEMS 
!   OF THE FORM:
!
!                       Minimize || c - a*x ||_2
!
!   USING A RANDOMIZED QR FACTORIZATION WITH COLUMNS PIVOTING OR COMPLETE
!   ORTHOGONAL FACTORIZATION OF a AS COMPUTED BY rqb_cmp. a IS AN m-BY-n MATRIX
!   WHICH MAY BE RANK-DEFICIENT. m>=n OR n>m IS PERMITTED.
!
!   SEVERAL RIGHT HAND SIDE VECTORS c AND SOLUTION VECTORS x CAN BE
!   HANDLED IN A SINGLE CALL OF rqb_solve; THEY ARE STORED AS THE COLUMNS OF THE
!   m-BY-nrhs RIGHT HAND SIDE MATRIX c AND THE n-BY-nrhs SOLUTION
!   MATRIX x, RESPECTIVELY. IN OTHER WORDS, THE ARGUMENTS c AND x MAY
!   BE VECTORS OR MATRICES. c IS OVERWRITTEN BY rqb_solve.
!   
!   ON EXIT, IF THE OPTIONAL LOGICAL PARAMETER comp_resid IS PRESENT AND IS SET TO true,
!   THE RESIDUAL VECTOR OR MATRIX c - a*x is COMPUTED AND OVERWRITES c.
!
!   THE MINIMUN 2-NORM SOLUTION(S) ARE COMPUTED IF THE OPTIONAL ARRAY PARAMETER tau IS
!   PRESENT IN BOTH THE CALLS OF rqb_cmp AND rqb_solve SUBROUTINES. OTHERWISE, SOLUTION(S) ARE
!   COMPUTED SUCH THAT IF THE jTH COLUMN OF a IS OMITTED FROM THE BASIS, x[j] O
!   x[j,:nrhs] IS SET TO ZERO.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE ORTHONORMAL MATRIX q .
!
    res(:nrhs,:nqb) = matmul( transpose( c(:m,:nrhs) ), q(:m,:nqb) )
!
    err1 = maxval( abs(res(:nrhs,:nqb)) )
!
!   CHECK THAT THE RESIDUALS ARE ALMOST ORTHOGONAL TO THE RANGE OF THE COEFFICIENT MATRIX a .
!
    res(:nrhs,:n) = matmul( transpose( c(:m,:nrhs) ), a(:m,:n) )
!
!    err2 = maxval( abs(res(:nrhs,:n)) )/(cnorm*anorm)
    
    err2 = norm( res(:nrhs,:n) )/(cnorm*anorm)
!
!   PRINT THE RESULTS OF THE TEST.
!
    if ( err1<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type > 3  -> very slow decay of singular values'
!
    write (prtunit,*)
    write (prtunit,*) 'Orthogonality of the residual matrix to the range of the coefficient matrix = ', err2
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a linear least squares problem with a ', &
      m, ' by ', n,' real coefficient matrix and a ', m, ' by ', nrhs,       &
      ' right hand side matrix is ', elapsed_time, ' seconds'
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( pivoting ) then
        deallocate( a, q, b, c, x, res, singval0, tau, ip )
    else
        deallocate( a, q, b, c, x, res, singval0, tau )
    end if
!
!
! END OF PROGRAM ex1_rqb_solve
! ============================
!
end program ex1_rqb_solve
