program ex1_rqr_svd_cmp
!
!
! Purpose
! =======
!
!   This program illustrates how to compute an approximate partial SVD with a (randomized)
!   QR-SVD algorithm using subroutine RQR_SVD_CMP in module SVD_Procedures.
!                                                                              
!                                                                              
! LATEST REVISION : 14/12/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, seven, c30, c1_e6, merror, allocate_error,  &
                         norm, unit_matrix, random_seed_, random_number_, rqr_svd_cmp, gen_random_mat,       &
                         singval_sort
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX (WHICH MUST BE LESS OR EQUAL TO mn)
! FOR CASES GREATER THAN 0,
! nsvd IS THE TARGET RANK OF THE APPROXIMATE PARTIAL SVD, WHICH IS SOUGHT.
!
    integer(i4b), parameter :: prtunit=6, m=10000, n=10000, mn=min(m,n), nsvd0=5000, nsvd=20
!   
! conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN CASES 5, 6 AND 7.
!   
    real(stnd), parameter  :: conda=c1_e6
!
    character(len=*), parameter :: name_proc='Example 1 of rqr_svd_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err, err1, err2, err3, eps, elapsed_time, anorm, tmp, tmp2, relerr, relerr2, &
                  abs_err, rel_err
    real(stnd), dimension(:,:), allocatable :: a, leftvec, rightvec, res, id
    real(stnd), dimension(:),   allocatable :: singval, singval0
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: blk_size, nover,  nover_svd, i, mat_type
!
    logical(lgl) :: failure, random_qr, truncated_qr, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : APPROXIMATE PARTIAL SVD OF A m-BY-n REAL MATRIX USING A (RANDOMIZED) PARTIAL
!               QR-SVD DECOMPOSITION.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type < 1  -> RANDOM MATRIX FROM THE UNIFORM DISTRIBUTION
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type = 4  -> VERY SLOW DECAY OF SINGULAR VALUES
!   mat_type = 5  -> STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda
!                    SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE
!   mat_type = 6  -> GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type = 7  -> ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type > 7  -> UNIFORM DISTRIBUTION OF SINGULAR VALUES
!
    mat_type = 8_i4b
!
!   SET TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM
!   OF THE COMPUTED PARTIAL SVD.
!
    eps = 0.01_stnd
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SINGULAR TRIPLETS.
!
    do_test = true
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED QR-SVD ALGORITHM.
!
!   DETERMINE IF A RANDOMIZED PARTIAL QR ALGORITHM IS USED IN THE FIRST QR PHASE OF THE QR-SVD ALGORITHM.
!
    random_qr = true
!
!   DETERMINE IF A RANDOMIZED PARTIAL AND TRUNCATED QR ALGORITHM IS USED IN THE FIRST PHASE OF THE QR-SVD ALGORITHM.
!
    truncated_qr = false
!
!   DETERMINE THE BLOCK SIZE blk_size FOR THE RANDOMIZED PARTIAL QR PHASE OF THE QR-SVD ALGORITHM.
!
!    blk_size = 40_i4b
    blk_size = 30_i4b
!
!   DETERMINE THE OVERSAMPLING SIZE nover FOR THE RANDOMIZED PARTIAL QR PHASE OF THE QR-SVD ALGORITHM.
!
    nover = 20_i4b
!
!   DETERMINE THE OVERSAMPLING SIZE nover_svd FOR THE FINAL SVD PHASE OF THE QR-SVD ALGORITHM.
!
    nover_svd = 20_i4b
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), singval(nsvd), leftvec(m,n), rightvec(n,nsvd),   &
              singval0(nsvd0), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE A RANDOM UNIFORM MATRIX OR SINGULAR VALUES.
!
    select case( mat_type )
!
        case( :0_i4b )
!
!           RANDOM UNIFORM MATRIX.
!
            call random_number_( a )
!
!           COMPUTE THE FROBENIUS NORM OF THE RANDOM MATRIX.
!
            anorm = norm( a )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
            end do
!
        case( 4_i4b )
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            tmp = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                singval0(i) = exp( -tmp2/tmp )
!        
            end do
!
        case( 5_i4b )
!
!           STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE.
!
            singval0(:nsvd0-1_i4b) = one
            singval0(nsvd0) = one/conda
!
        case( 6_i4b )
!
!           GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                singval0(i) = conda**( -tmp2/tmp )
!        
            end do
!
        case( 7_i4b )
!
!           ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
            tmp2 = one - one/conda
!
            do i = 1_i4b, nsvd0
!
                singval0(i) = one - (real( i - 1_i4b, stnd )/tmp)*tmp2
!        
            end do
!
        case default
!
!           UNIFORM DISTRIBUTION OF SINGULAR VALUES.
!
            call random_number_( singval0(:nsvd0) )
!
    end select
!
    if ( mat_type>0_i4b ) then
!
#ifdef _F2003
        if ( ieee_support_datatype( singval0 ) ) then
!
            if ( .not.all( ieee_is_normal( singval0(:nsvd0) ) ) ) then
!
                call merror( name_proc//' : Exceptions occurred when generating the input matrix !'  )                    
!
            end if
!
        end if
#endif
!
!       SORT THE SINGULAR VALUES BY DECREASING MAGNITUDE.
!
        call singval_sort( 'D', singval0(:nsvd0) )
!
!       GENERATE A m-BY-n MATRIX a OF RANK nsvd0 .
!
        call gen_random_mat( singval0(:nsvd0), a )
!
!       COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
        anorm = norm( singval0(:nsvd0) )
!
    end if
!
!   MAKE A COPY OF THE INPUT MATRIX FOR LATER USE.
!
    leftvec(:m,:n) = a(:m,:n)
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   rqr_svd_cmp COMPUTES AN APPROXIMATE PARTIAL SINGULAR VALUE DECOMPOSITION (SVD)
!   OF A REAL m-BY-n MATRIX a. THE PARTIAL SVD IS WRITTEN
!
!                       U * S * V**(t)
!
!   WHERE S IS AN nsvd-BY-nsvd MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   DIAGONAL ELEMENTS, U IS AN m-BY-nsvd ORTHONORMAL MATRIX, AND
!   V IS AN n-BY-nsvd ORTHONORMAL MATRIX.  THE DIAGONAL ELEMENTS OF S
!   ARE THE nsvd LARGEST SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE ASSOCIATED LEFT AND RIGHT SINGULAR VECTORS OF a.
!
    call rqr_svd_cmp( leftvec(:m,:n), singval(:nsvd), failure, rightvec(:n,:nsvd),       &
                      random_qr=random_qr, truncated_qr=truncated_qr, blk_size=blk_size, &
                      nover=nover, nover_svd=nover_svd,  max_francis_steps=10_i4b        )
!
!   THE ROUTINE RETURNS APPROXIMATIONS FOR THE nsvd LARGEST SINGULAR VALUES AND THE ASSOCIATED
!   LEFT AND RIGHT SINGULAR VECTORS.
!
!   ON EXIT OF rqr_svd_cmp :
!
!       singval CONTAINS THE nsvd LARGEST SINGULAR VALUES OF a IN DECREASING ORDER.
!
!       leftvec(:m,:nsvd) CONTAINS THE ASSOCIATED nsvd LEFT SINGULAR VECTORS,
!       STORED COLUMNWISE;
!   
!       rightvec CONTAINS THE ASSOCIATED nsvd RIGHT SINGULAR VECTORS,
!       STORED COLUMNWISE;
!
!       failure= false :  INDICATES SUCCESSFUL EXIT IN THE SVD PHASE OF THE ALGORITHM.
!       failure= true  :  INDICATES THAT THE SVD IN THE SECOND PHASE OF THE ALGORITHM DID NOT
!                         CONVERGE AND THAT FULL ACCURACY WAS NOT ATTAINED IN THE (RANDOMIZED)
!                         PARTIAL QR-SVD ALGORITHM.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   COMPUTE THE ESTIMATED RELATIVE ERROR.
!
    tmp = one - sum( (singval(:nsvd)/anorm)**2 )
!
    if ( tmp>=zero ) then
!
        relerr = sqrt( tmp )
!
    else
!
        relerr = -one
!
        failure = true
!
    end if
!
!   COMPUTE THE TRUE RELATIVE ERROR.
!
    if ( mat_type>0_i4b ) then
!
        if ( nsvd0>nsvd ) then
            relerr2 = norm( singval0(nsvd+1_i4b:nsvd0)/anorm )
        else
            relerr2 = zero
        end if
!
!       COMPUTE ERROR BETWEEN THE TRUE AND ESTIMATED RELATIVE ERRORS.
!
        err = abs( relerr - relerr2 )
!
    else
!
        err = zero
!
    end if
!
!   TEST ACCURACY OF THE COMPUTED SINGULAR TRIPLETS IF REQUIRED.
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( res(m,nsvd), id(nsvd,nsvd), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        if ( mat_type>0_i4b ) then
!
!           COMPUTE ERRORS FOR SINGULAR VALUES.
!
            i = min( nsvd, nsvd0 )
!
            id(:i,1_i4b) = singval0(:i)
            id(i+1_i4b:nsvd,1_i4b) = zero
!
            where( id(:nsvd,1_i4b)/=zero )
                res(:nsvd,1_i4b) = id(:nsvd,1_i4b)
            elsewhere
                res(:nsvd,1_i4b) = one
            end where
!
!           ABSOLUTE ERRORS OF SINGULAR VALUES.
!
            abs_err = maxval( abs( singval(:nsvd) - id(:nsvd,1_i4b) ) )
!
!           RELATIVE ERRORS OF SINGULAR VALUES.
!
            rel_err = maxval( abs( (singval(:nsvd) - id(:nsvd,1_i4b))/res(:nsvd,1_i4b) ) )
!
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v(:n,:nsvd) - u(:n,:nsvd)*diag(singval(:nsvd)),
!       WHERE u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        res(:m,:nsvd) = matmul(a,rightvec) - leftvec(:m,:nsvd)*spread(singval(:nsvd),dim=1,ncopies=m)
        id(:nsvd,1_i4b) = norm( res(:m,:nsvd), dim=2_i4b )
!
        err1 = maxval( id(:nsvd,1_i4b) )/( anorm*real(mn,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u(:m,:nsvd)**(t)*u(:m,:nsvd).
!
        call unit_matrix( id(:nsvd,:nsvd) )
!
        res(:nsvd,:nsvd) = abs( id(:nsvd,:nsvd) - matmul( transpose(leftvec(:m,:nsvd)), leftvec(:m,:nsvd) ) )
!
        err2 = maxval( res(:nsvd,:nsvd) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v(:n,:nsvd)**(t)*v(:n,:nsvd).
!
        res(:nsvd,:nsvd) = abs( id(:nsvd,:nsvd) - matmul( transpose(rightvec), rightvec ) )
!
        err3 = maxval( res(:nsvd,:nsvd) )/real(n,stnd)
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( res, id )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, singval0, singval, leftvec, rightvec )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix_type < 1  -> random matrix from the uniform distribution'
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type = 4  -> very slow decay of singular values'
    write (prtunit,*) ' Matrix type = 5  -> strongly clustered singular values at 1'
    write (prtunit,*) '                     spectrum with few deflations (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 6  -> geometric distribution of singular values'
    write (prtunit,*) '                     spectrum with moderate deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 7  -> arithmetic distribution of singular values'
    write (prtunit,*) '                     spectrum with complete deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type > 7  -> uniform distribution of singular values'
!
    write (prtunit,*) 
    write (prtunit,*) 'Rank of the approximate partial SVD                                  = ', &
                      nsvd
    write (prtunit,*) 'Relative error in Frobenius norm     : ||A-rSVD||_F / ||A||_F        = ', &
                      relerr
!
    if ( mat_type>0_i4b ) then
!
        write (prtunit,*) 'Best relative error in Frobenius norm: min( ||A-rSVD||_F / ||A||_F ) = ', &
                          relerr2
!
    end if
!
    write (prtunit,*) 
    write (prtunit,*) 'FAILURE ( from rqr_svd_cmp() ) = ', failure
!
    if ( do_test ) then
!
        write (prtunit,*)
!
        if ( mat_type>0_i4b ) then
!
            write (prtunit,*) 'Absolute accuracy of the computed singular values    = ', abs_err
            write (prtunit,*) 'Relative accuracy of the computed singular values    = ', rel_err
!
        end if
!
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
!
    end if
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing approximate ', nsvd, ' singular values and vectors of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_rqr_svd_cmp
! ==============================
!
end program ex1_rqr_svd_cmp
