program ex1_select_eigval_cmp2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SELECT_EIGVAL_CMP2
!   in module Eig_Procedures for computing the largest or smallest eigenvalues of a
!   real symmetric matrix whose sum of absolute values exceeds a given (positive) threshold.
!
!   The initial tridiagonal reduction is parallelized if OpenMP is activated and a
!   rational QR method is used for computing the eigenvalues of the tridiagonal
!   form of the symmetric matrix.
!
!
! Further Details
! ===============
!
!    The program also shows the use of subroutine TRID_INVITER in module EIG_Procedures
!    for computing selected eigenvectors of a real symmetric matrix.
!
!
! LATEST REVISION : 05/12/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, c100, allocate_error,  &
                         merror, norm, get_diag, select_eigval_cmp2, trid_inviter,      &
                         norm, unit_matrix
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE DIMENSION OF THE SYMMETRIX MATRIX AND
! maxiter IS THE NUMBER OF INVERSE ITERATIONS PERFORMED TO COMPUTE THE SELECTED EIGENVECTORS.
!
    integer(i4b), parameter :: prtunit=6, n=1000, maxiter=2
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of select_eigval_cmp2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, val, elapsed_time
    real(stnd), pointer ,    dimension(:)   :: d
    real(stnd), allocatable, dimension(:)   :: res
    real(stnd), allocatable, dimension(:,:) :: eigvec, a, a2, d_e
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: m
!
    logical(lgl) :: failure, failure2, do_test, small
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTE THE LARGEST OR SMALLEST EIGENVALUES OF A n-BY-n REAL
!               SYMMETRIC MATRIX WHOSE SUM OF ABSOLUTE VALUES EXCEEDS
!               A PRESCRIBED THRESHOLD USING A RATIONAL QR ALGORITHM
!               AND SELECTED EIGENVECTORS BY INVERSE ITERATION.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   DETERMINE IF YOU WANT TO COMPUTE THE SMALLEST OR LARGEST EIGENVALUES.
!
    small = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), d_e(n,2), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM DATA MATRIX AND FROM IT
!   A SEMI-DEFINITE POSITIVE MATRIX a .
!
    call random_number( a )
!
    a = matmul( a, transpose( a ) )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), res(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM SEMI-DEFINITE POSITIVE MATRIX.
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   DETERMINE THRESHOLD FOR THE SUM OF THE EIGENVALUES.
!
    val = sum( get_diag(a) )/c100
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE LARGEST OR SMALLEST EIGENVALUES OF THE SELF-ADJOINT MATRIX a 
!   WHOSE SUM OF ABSOLUTE VALUES EXCEEDS abs( val ) AND SAVE THE INTERMEDIATE 
!   TRIDIAGONAL MATRIX IN PARAMETER d_e .
!
    call select_eigval_cmp2( a, d, small, val, failure, d_e=d_e )
!
!   ON EXIT OF select_eigval_cmp2 :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE RATIONAL
!                         QR ALGORITHM.
!
!   THE COMPUTED EIGENVALUES ARE STORED IN d AND THE
!   SYMMETRIC MATRIX IS OVERWRITTEN BY DETAILS OF ITS
!   TRIDIAGONAL DECOMPOSITION STORED IN COMPACT FORM.
!
!   DETERMINE THE NUMBER OF COMPUTED EIGENVALUES AND ALLOCATE WORK ARRAY FOR
!   COMPUTING THE ASSOCIATED EIGENVECTORS.
!
    m = size( d )
!
    if ( .not. failure .and. m>0 ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( eigvec(n,m), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE THE ASSOCIATED m EIGENVECTORS OF THE SELF-ADJOINT MATRIX a BY
!       maxiter INVERSE ITERATIONS ON THE INTERMEDIATE TRIDIAGONAL MATRIX
!       d_e AND BACK-TRANSFORMATION.
!
        call trid_inviter( d_e(:n,1), d_e(:n,2), d(:m), eigvec(:n,:m), failure2,     &
                           mat=a(:n,:n), maxiter=maxiter )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test .and. .not.failure .and. m>0 ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*eigvec - eigvec*d
!       WHERE d ARE THE EIGENVALUES AND eigvec THE EIGENVECTORS OF a.
!
        a(:n,:m) = matmul( a2(:n,:n), eigvec(:n,:m)) - eigvec(:n,:m)*spread( d(:m), dim=1, ncopies=n)
        res(:m) = norm( a(:n,:m), dim=2_i4b )
!
        err1 = maxval( res(:m) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - eigvec**(t)*eigvec.
!
        call unit_matrix( a(:m,:m) )
!
        a2(:m,:m) = abs( a(:m,:m) - matmul( transpose(eigvec(:n,:m)), eigvec(:n,:m) ) )
!
        err2 = maxval( a2(:m,:m) )/real(n,stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
!
        deallocate( a, d_e, d, a2, res )
!
    else
!
        deallocate( a, d_e, d )
!
    end if
!
    if ( allocated( eigvec ) )  deallocate( eigvec )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure  .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' FAILURE ( from select_eigval_cmp2() ) = ', failure
!
    if ( .not. failure  ) then
        write (prtunit,*) ' FAILURE ( from trid_inviter()       ) = ', failure2
    end if
!
    if ( do_test  .and. .not.failure .and. m>0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues/eigenvectors couplets = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors                 = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing ', m, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_select_eigval_cmp2
! =====================================
!
end program ex1_select_eigval_cmp2
