program ex1_select_eigval_cmp3
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SELECT_EIGVAL_CMP3
!   in module Eig_Procedures for computing the largest or smallest eigenvalues of a
!   real symmetric matrix.
!
!   The initial tridiagonal reduction is parallelized if OpenMP is activated and a
!   bisection method is used for computing the eigenvalues of the tridiagonal
!   form of the symmetric matrix.
!
!
! Further Details
! ===============
!
!    The program also shows the use of subroutine TRID_INVITER in module EIG_Procedures
!    for computing selected eigenvectors of a real symmetric (tridiagonal) matrix.
!
!
! LATEST REVISION : 05/12/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm, select_eigval_cmp3, &
                         trid_inviter, merror, allocate_error, unit_matrix
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE DIMENSION OF THE SYMMETRIX MATRIX,
! le IS THE NUMBER OF THE WANTED EIGENVALUES/EIGENVECTORS,
! maxiter IS THE NUMBER OF INVERSE ITERATIONS PERFORMED TO COMPUTE THE le ASSOCIATED EIGENVECTORS.
!
    integer(i4b), parameter :: prtunit=6, n=3000, le=100, maxiter=2
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of select_eigval_cmp3'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, elapsed_time
    real(stnd), dimension(:),   allocatable :: eigval, res
    real(stnd), dimension(:,:), allocatable :: a, a2, eigvec, d_e
!
    integer(i4b) :: neig
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: failure, failure2, do_test, small
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL EIGENVALUES/EIGENVECTORS DECOMPOSITION OF A n-BY-n REAL
!               SYMMETRIC MATRIX USING A BISECTION ALGORITHM FOR THE EIGENVALUES
!               AND THE INVERSE ITERATION TECHNIQUE FOR THE EIGENVECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   DETERMINE IF YOU WANT TO COMPUTE THE le SMALLEST OR LARGEST EIGENVALUES.
!
    small = false
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), eigval(n), d_e(n,2), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM DATA MATRIX AND FROM IT
!   A SELF-ADJOINT MATRIX a .
!
    call random_number( a )
!
    a = a + transpose( a ) 
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), res(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX .
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE FIRST le EIGENVALUES OF THE SELF-ADJOINT MATRIX a BY BISECTION AND SAVE
!   THE INTERMEDIATE TRIDIAGONAL MATRIX IN PARAMETER d_e .
!
    call select_eigval_cmp3( a, neig, eigval, small, failure, sort=sort, le=le, d_e=d_e )
!
    if ( .not. failure .and. neig>0  ) then
!
!       ALLOCATE WORK ARRAY NEEDED TO STORE THE EIGENVECTORS.
!
        allocate( eigvec(n,neig), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE THE ASSOCIATED neig EIGENVECTORS OF THE SELF-ADJOINT MATRIX a BY
!       maxiter INVERSE ITERATIONS ON THE INTERMEDIATE TRIDIAGONAL MATRIX
!       d_e AND BACK-TRANSFORMATION.
!
        call trid_inviter( d_e(:n,1), d_e(:n,2), eigval(:neig), eigvec, failure2,     &
                           mat=a, maxiter=maxiter )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test .and. .not.failure .and. neig>0 ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*eigvec - eigvec*d
!       WHERE eigval ARE THE EIGENVALUES AND eigvec THE EIGENVECTORS OF a.
!
        a(:n,:neig) = matmul( a2(:n,:n), eigvec(:n,:neig)) - eigvec(:n,:neig)*spread( eigval(:neig), dim=1, ncopies=n)
        res(:neig)  = norm( a(:n,:neig), dim=2_i4b )
!
        err1 = maxval( res(:neig) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - eigvec**(t)*eigvec.
!
        call unit_matrix( a(:neig,:neig) )
!
        a2(:neig,:neig) = abs( a(:neig,:neig) - matmul( transpose(eigvec(:n,:neig)), eigvec(:n,:neig) ) )
!
        err2 = maxval( a2(:neig,:neig) )/real(n,stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
!
        deallocate( a, d_e, eigval, a2, res )
!
    else
!
        deallocate( a, d_e, eigval )
!
    end if
!
    if ( allocated( eigvec ) )  deallocate( eigvec )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure  .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' FAILURE ( from select_eigval_cmp3() ) = ', failure
!
    if ( .not. failure  ) then
        write (prtunit,*) ' FAILURE ( from trid_inviter()       ) = ', failure2
    end if
!
    if ( do_test  .and. .not.failure .and. neig>0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues/eigenvectors couplets = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors                 = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing ', neig, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_select_eigval_cmp3
! =====================================
!
end program ex1_select_eigval_cmp3
