program ex1_svd_cmp3
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SVD_CMP3 
!   in module SVD_Procedures for computing a full SVD decomposition of a real m-by-n matrix by
!   a bidiagonal factorization and the Golub-Reinsch bidiagonal QR implicit method.
!
!   The computations are parallelized if OpenMP is used and the fast one-sided Rhala-Barlow algorithm
!   is used in the first bidiagonalization phase. Furthermore, an highly efficient variant of the
!   Golub-Reinsch bidiagonal QR implicit algorithm is used.
!
!   In many cases SVD_CMP3 can be faster than SVD_CMP or SVD_CMP2 (especially when the number of available
!   processors is reduced) at the expense of a slight loss of orthogonality for the smaller left
!   (if m>=n) or right (if m<n) singular vectors.
!
!
! LATEST REVISION : 07/02/2024
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, seven, c30, c50, c1_e6, &
                         svd_cmp3, norm, unit_matrix, random_seed_, random_number_,      &
                         gen_random_mat, singval_sort, merror, allocate_error,           &
                         ifirstloc, safmin
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX (WHICH MUST BE LESS OR EQUAL TO k)
! FOR CASES GREATER THAN 0.
!
    integer(i4b), parameter :: prtunit = 6, m=3000, n=3000, k=min(m,n), nsvd0=1000
!   
! conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN CASES 5, 6 and 7.
! tol IS A TOLERANCE FOR DETERMINING A CONSERVATIVE ESTIMATE OF THE RANK OF THE
! GENERATED MATRIX (E.G., A THRESHOLD FOR THE INVERSE OF THE CONDITION NUMBER OF
! THE TOP SINGULAR APPROXIMATION OF THE GENERATED MATRIX).
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50, conda=c1_e6, tol=0.0001_stnd
!
    character(len=*), parameter :: name_proc='Example 1 of svd_cmp3'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, tmp, tmp2, &
                                               abs_err, rel_err, anorm, elapsed_time
    real(stnd), allocatable, dimension(:)   :: s, s0
    real(stnd), allocatable, dimension(:,:) :: a2, a, c, resid
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: i, mat_type, nvec
!
    logical(lgl) :: reortho, perfect_shift, bisect, dqds, failure, failure_bd, do_test
!   
    character:: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : FULL SVD OF A REAL MATRIX USING THE ONE-SIDED
!               RALHA-BARLOW BIDIAGONAL REDUCTION ALGORITHM AND
!               THE BIDIAGONAL QR IMPLICIT METHOD WITH A WAVE-FRONT
!               ALGORITHM FOR APPLYING GIVENS ROTATIONS IN THE
!               BIDIAGONAL QR ALGORITHM AND, OPTIONALLY, A PERFECT
!               SHIFT STRATEGY FOR THE SINGULAR VECTORS.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type < 1  -> RANDOM MATRIX FROM THE UNIFORM DISTRIBUTION
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type = 4  -> VERY SLOW DECAY OF SINGULAR VALUES
!   mat_type = 5  -> STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda
!                    SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE
!   mat_type = 6  -> GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type = 7  -> ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type > 7  -> UNIFORM DISTRIBUTION OF SINGULAR VALUES
!
    mat_type = 0_i4b
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
!   SPECIFY IF REORTHOGONALIZATION IS PERFORMED IN
!   THE ONE-SIDED RHALA-BARLOW ALGORITHM.
!
    reortho = true
!
!   SPECIFY IF A PERFECT SHIFT STRATEGY MUST BE USED
!   IN THE BIDIAGONAL SVD ALGORITHM.
!
    perfect_shift = true
!
!   SPECIFY IF BISECTION IS USED FOR THE PERFECT SHIFT STRATEGY
!   IN THE BIDIAGONAL SVD ALGORITHM TO GIVE HIGHER ACCURACY.
!
    bisect = false
!
!   SPECIFY IF DQDS IS USED FOR THE PERFECT SHIFT STRATEGY
!   IN THE BIDIAGONAL SVD ALGORITHM TO GIVE HIGHER ACCURACY.
!
    dqds = false
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), c(k,k), s(k), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM UNIFORM MATRIX OR SINGULAR VALUES.
!
    select case( mat_type )
!
        case( :0_i4b )
!
!           RANDOM UNIFORM MATRIX.
!
            call random_number_( a )
!
!           COMPUTE THE FROBENIUS NORM OF THE RANDOM MATRIX.
!
            anorm = norm( a )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
            end do
!
        case( 4_i4b )
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            tmp = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                s(i) = exp( -tmp2/tmp )
!        
            end do
!
        case( 5_i4b )
!
!           STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE.
!
            s(:nsvd0-1_i4b) = one
            s(nsvd0) = one/conda
!
        case( 6_i4b )
!
!           GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                s(i) = conda**( -tmp2/tmp )
!        
            end do
!
        case( 7_i4b )
!
!           ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
            tmp2 = one - one/conda
!
            do i = 1_i4b, nsvd0
!
                s(i) = one - (real( i - 1_i4b, stnd )/tmp)*tmp2
!        
            end do
!
        case default
!
!           UNIFORM DISTRIBUTION OF SINGULAR VALUES.
!
            call random_number_( s(:nsvd0) )
!
    end select
!
    if ( mat_type>0_i4b ) then
!
#ifdef _F2003
        if ( ieee_support_datatype( s ) ) then
!
            if ( .not.all( ieee_is_normal( s(:nsvd0) ) ) ) then
!
                call merror( name_proc//' : Exceptions occurred when generating the input matrix !'  )                    
!
            end if
!
        end if
#endif
!
!       SORT THE SINGULAR VALUES BY DECREASING MAGNITUDE.
!
        call singval_sort( 'D', s(:nsvd0) )
!
!       GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!       OF SINGULAR VALUES AND A RANK EQUAL TO nsvd0.
!
        call gen_random_mat( s(:nsvd0), a )
!
!       COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
        anorm = norm( s(:nsvd0) )
!
    end if
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), resid(m,k), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX.
!
        a2(:m,:n) = a(:m,:n)
!
        if ( mat_type>0_i4b ) then
!
!           ALLOCATE WORK ARRAY.
!
            allocate( s0(k), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           SAVE THE TRUE SINGULAR VALUES.
!
            s0(:nsvd0) = s(:nsvd0)
            s0(nsvd0+1_i4b:k) = zero
!
        end if
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   svd_cmp3 COMPUTES THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL
!   m-BY-n MATRIX a. THE SVD IS WRITTEN
!
!                       a = U * S * V**(t)
!
!   WHERE s IS AN m-BY-n MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   min(m,n) DIAGONAL ELEMENTS, u IS AN m-BY-m ORTHOGONAL MATRIX, AND
!   v IS AN n-BY-n ORTHOGONAL MATRIX.  THE DIAGONAL ELEMENTS OF s
!   ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!   BY DEFAULT, svd_cmp3 USES A PERFECT SHIFT STRATEGY FOR THE SINGULAR VECTORS
!   SINCE IT IS USUALLY FASTER.
!   IF YOU DONT WANT TO USE THIS STRATEGY, YOU CAN SPECIFY THE OPTIONAL LOGICAL
!   PARAMETER perfect_shift WITH THE VALUE false.
!
    call svd_cmp3( a, s, failure, c, sort=sort, perfect_shift=perfect_shift, bisect=bisect, &
                   dqds=dqds, reortho=reortho, failure_bd=failure_bd )
!
!   THE ROUTINE RETURNS THE SINGULAR VALUES AND THE FIRST min(m,n) LEFT AND
!   RIGHT SINGULAR VECTORS. THE RIGHT SINGULAR VECTORS ARE RETURNED ROWWISE
!   IF m<n.
!
!   ON EXIT OF svd_cmp3 :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE BIDIAGONAL
!                         SVD OF AN INTERMEDIATE BIDIAGONAL FORM B OF a.
!
!       failure_bd= false :  INDICATES SUCCESSFUL EXIT IN THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION
!       failure_bd= true  :  INDICATES THAT a IS NEARLY SINGULAR AND SOME LOSS OF ORTHOGONALITY
!                            CAN BE EXPECTED IN THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION
!                            OF a FOR THE LAST LEFT (IF m>=n) OR RIGHT (IF m<n) SINGULAR VECTORS.
!
!       IF m>=n,    a IS OVERWRITTEN WITH THE FIRST n
!                   COLUMNS OF U (THE LEFT SINGULAR VECTORS,
!                   STORED COLUMNWISE);
!                   c CONTAINS THE n-BY-n ORTHOGONAL MATRIX V .
!
!       IF m<n,     a IS OVERWRITTEN WITH THE FIRST m ROWS OF
!                   V**(t) (THE RIGHT SINGULAR VECTORS,
!                   STORED ROWWISE);
!                   c CONTAINS THE m-BY-m ORTHOGONAL MATRIX U.
!
!       s IS OVERWRITTEN WITH THE SINGULAR VALUES OF a.
!
!   IF SORT = 'a' OR 'A', THE SINGULAR VALUES ARE SORTED INTO ASCENDING ORDER.
!   IF SORT = 'd' OR 'D', THE SINGULAR VALUES ARE SORTED INTO DESCENDING ORDER.
!   THE SINGULAR VECTORS ARE REARRANGED ACCORDINGLY.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!                
    if ( do_test ) then
!
        if ( mat_type>0_i4b ) then
!
!           COMPUTE ERRORS FOR SINGULAR VALUES.
!
            where( s0(:k)/=zero )
                resid(:k,1_i4b) = s0(:k)
            elsewhere
                resid(:k,1_i4b) = one
            end where
!
!           ABSOLUTE ERRORS OF SINGULAR VALUES.
!
            abs_err = maxval( abs( s(:k) - s0(:k) ) )
!
!           RELATIVE ERRORS OF SINGULAR VALUES.
!
            rel_err = maxval( abs( (s(:k) - s0(:k))/resid(:k,1_i4b) ) )
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( s0 )
!
        end if
!
!       DETERMINE THE EFFECTIVE RANK OF a .
!
        tmp = max( tol*s(1_i4b), safmin )
!
        nvec = ifirstloc( logical( s<=tmp, lgl ) ) - 1_i4b
!
        if ( m>=n ) then
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:n)*V(:n,:nvec) - U(:m,:nvec)*S(:nvec,:nvec).
!
            resid(:m,:nvec) = matmul(a2(:m,:k), c(:k,:nvec) ) - a(:m,:nvec)*spread(s(:nvec),dim=1,ncopies=m)
            a2(:nvec,1_i4b) = norm( resid(:m,:nvec), dim=2_i4b )
            err1 =  maxval( a2(:nvec,1_i4b) )/( anorm*real(m,stnd) )
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
            call unit_matrix( a2(:n,:n) )
!
            resid(:nvec,:nvec) = abs( a2(:nvec,:nvec) - matmul( transpose(a(:m,:nvec)), a(:m,:nvec) ) )
            err2 = maxval( resid(:nvec,:nvec) )/real(m,stnd)
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V**(t)*V.
!
            resid(:n,:n) = abs( a2(:n,:n) - matmul( transpose(c(:n,:n)), c(:n,:n) ) )
            err3 = maxval( resid(:n,:n) )/real(n,stnd)
!
        else
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:n)*V(:n,:nvec) - U(:m,:nvec)*S(:nvec,:nvec).
!
            resid(:m,:nvec) =  matmul(a2(:m,:n),transpose(a(:nvec,:n))) - c(:k,:nvec)*spread(s(:nvec),dim=1,ncopies=k)
            a2(:nvec,1_i4b) = norm( resid(:m,:nvec), dim=2_i4b )
            err1 =  maxval( a2(:nvec,1_i4b) )/( anorm*real(m,stnd) )
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
            call unit_matrix( a2(:m,:m) )
!
            resid(:m,:m) = abs( a2(:m,:m) - matmul( transpose(c(:m,:m )), c(:m,:m ) ) )
            err2 = maxval( resid(:m,:m) )/real(m,stnd)
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V**(t)*V.
!
            resid(:nvec,:nvec) = abs( a2(:nvec,:nvec) - matmul( a(:nvec,:n), transpose(a(:nvec,:n)) ) )
            err3 = maxval( resid(:nvec,:nvec) )/real(n,stnd)
!
        end if
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, c, s )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix_type < 1  -> random matrix from the uniform distribution'
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type = 4  -> very slow decay of singular values'
    write (prtunit,*) ' Matrix type = 5  -> strongly clustered singular values at 1'
    write (prtunit,*) '                     spectrum with few deflations (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 6  -> geometric distribution of singular values'
    write (prtunit,*) '                     spectrum with moderate deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 7  -> arithmetic distribution of singular values'
    write (prtunit,*) '                     spectrum with complete deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type > 7  -> uniform distribution of singular values'
!
    if ( do_test ) then
!
        write (prtunit,*) 
!
        if ( mat_type>0_i4b ) then
!
            write (prtunit,*) 'Absolute accuracy of the computed singular values    = ', abs_err
            write (prtunit,*) 'Relative accuracy of the computed singular values    = ', rel_err
!
        end if
!
        write (prtunit,*) 'Effective rank of the matrix                         = ', nvec
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
!
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' FAILURE_BD ( from svd_cmp3() ) = ', failure_bd
    write (prtunit,*) ' FAILURE    ( from svd_cmp3() ) = ', failure
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and vectors of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_svd_cmp3
! ===========================
!
end program ex1_svd_cmp3
