program ex1_svd_cmp4
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SVD_CMP4 in module SVD_Procedures
!   for computing all singular values and, optionally, singular vectors of a real n-by-m matrix with n>=m
!   by a bidiagonal factorization and the Golub-Reinsch bidiagonal QR implicit method.
!
!   The computations are parallelized if OpenMP is used and the fast one-sided Rhala-Barlow algorithm
!   is used in the first bidiagonalization phase. Furthermore, an highly efficient variant of the
!   Golub-Reinsch bidiagonal QR implicit algorithm is used.
!
!   In many cases SVD_CMP4 can be faster than SVD_CMP or SVD_CMP2 (especially when the number of available
!   processors is reduced) at the expense of a slight loss of orthogonality for the smaller left
!   singular vectors.
!
!   The main difference betwen SVD_CMP4 and SVD_CMP3 subroutines is that SVD_CMP4 allows the user
!   to compute only the singular values of the input matrix and to output both the bidiagonal form
!   of the input matrix and the associated orthogonal matrices in the bidiagonal decomposition of
!   this matrix. This allows the user to compute selected singular vectors in a second step with
!   other statpack subroutines like BD_INVITER2 or BD_DEFLATE2 subroutines. Another difference
!   is that, in SVD_CMP4, the input matrix must have more rows than columns.
!
!
! Further Details
! ===============
!
!    The program also shows the use of subroutine BD_INVITER2 in module SVD_Procedures.
!                                                                              
! LATEST REVISION : 07/02/2024
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c0_9, one, seven, c30, c50, c1_e6,   &
                         bd_inviter2, svd_cmp4, norm, unit_matrix, random_seed_, random_number_, &
                         gen_random_mat, singval_sort, merror, allocate_error, ifirstloc, safmin
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n AND m ARE THE DIMENSIONS OF THE GENERATED MATRIX (HERE n>=m),
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX FOR CASES GREATER THAN 0,
! nsing IS THE TARGET RANK OF THE APPROXIMATE PARTIAL SVD, WHICH IS SOUGHT,
! maxiter IS THE NUMBER OF INVERSE ITERATIONS PERFORMED TO COMPUTE THE TOP nsing SINGULAR VECTORS.
!
    integer(i4b), parameter :: prtunit = 6, n=3000, m=3000, nsvd0=1000,    &
                               nsing=3000, maxiter=2
!   
! conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN CASES 5, 6 and 7.
! tol IS A TOLERANCE FOR DETERMINING A CONSERVATIVE ESTIMATE OF THE RANK OF THE
! GENERATED MATRIX (E.G., A THRESHOLD FOR THE INVERSE OF THE CONDITION NUMBER OF
! THE TOP SINGULAR APPROXIMATION OF THE GENERATED MATRIX).
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50, conda=c1_e6, tol=0.0001_stnd
!
    character(len=*), parameter :: name_proc='Example 1 of svd_cmp4'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, tmp, tmp2, &
                                               abs_err, rel_err, anorm, elapsed_time
    real(stnd), allocatable, dimension(:)   :: s, s0, d, e
    real(stnd), allocatable, dimension(:,:) :: a, a2, p, leftvec, rightvec, resid
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: i, mat_type, mnthr_nsing, nrank, nvec
!
    logical(lgl) :: failure_bd, failure_bd_svd, failure_bd_inviter, do_test, gen_p, reortho
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : ALL SINGULAR VALUES AND nsing SINGULAR VECTORS OF A n-BY-m REAL MATRIX WITH n>=m
!               USING THE RHALA-BARLOW ONE-SIDED BIDIAGONAL REDUCTION ALGORITHM, THE DQDS
!               ALGORITHM FOR THE SINGULAR VALUES AND THE INVERSE ITERATION METHOD FOR THE
!               SINGULAR VECTORS (e.g., A PARTIAL SVD DECOMPOSITION).
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type < 1  -> RANDOM MATRIX FROM THE UNIFORM DISTRIBUTION
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type = 4  -> VERY SLOW DECAY OF SINGULAR VALUES
!   mat_type = 5  -> STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda
!                    SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE
!   mat_type = 6  -> GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type = 7  -> ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type > 7  -> UNIFORM DISTRIBUTION OF SINGULAR VALUES
!
    mat_type = 0_i4b
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
!   SPECIFY IF REORTHOGONALIZATION IS PERFORMED IN
!   THE ONE-SIDED RHALA-BARLOW ALGORITHM.
!
    reortho = true
!
!   DETERMINE CROSSOVER POINT FOR THE BACK-TRANSFORMATION
!   ALGORITHM WHEN COMPUTING RIGHT SINGULAR VECTORS. IF
!   nsing EXCEEDS mnthr_nsing, THE RIGHT ORTHOGONAL MATRIX p
!   OF THE ONE-SIDED BIDIAGONAL REDUCTION ALGORITHM IS GENERATED
!   AND BACK-TRANSFORMATION IS DONE BY A MATRIX MULTIPLICATION.
!
    mnthr_nsing = int( real( m, stnd )*c0_9, i4b )
!
    gen_p = nsing>=mnthr_nsing
!    gen_p = true
!    gen_p = false
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,m), leftvec(n,nsing), rightvec(m,nsing),  &
              p(m,m), s(m), d(m), e(m), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM UNIFORM MATRIX OR SINGULAR VALUES.
!
    select case( mat_type )
!
        case( :0_i4b )
!
!           RANDOM UNIFORM MATRIX.
!
            call random_number_( a )
!
!           COMPUTE THE FROBENIUS NORM OF THE RANDOM MATRIX.
!
            anorm = norm( a )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
            end do
!
        case( 4_i4b )
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            tmp = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                s(i) = exp( -tmp2/tmp )
!        
            end do
!
        case( 5_i4b )
!
!           STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE.
!
            s(:nsvd0-1_i4b) = one
            s(nsvd0) = one/conda
!
        case( 6_i4b )
!
!           GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                s(i) = conda**( -tmp2/tmp )
!        
            end do
!
        case( 7_i4b )
!
!           ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
            tmp2 = one - one/conda
!
            do i = 1_i4b, nsvd0
!
                s(i) = one - (real( i - 1_i4b, stnd )/tmp)*tmp2
!        
            end do
!
        case default
!
!           UNIFORM DISTRIBUTION OF SINGULAR VALUES.
!
            call random_number_( s(:nsvd0) )
!
    end select
!
    if ( mat_type>0_i4b ) then
!
#ifdef _F2003
        if ( ieee_support_datatype( s ) ) then
!
            if ( .not.all( ieee_is_normal( s(:nsvd0) ) ) ) then
!
                call merror( name_proc//' : Exceptions occurred when generating the input matrix !'  )                    
!
            end if
!
        end if
#endif
!
!       SORT THE SINGULAR VALUES BY DECREASING MAGNITUDE.
!
        call singval_sort( 'D', s(:nsvd0) )
!
!       GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!       OF SINGULAR VALUES AND A RANK EQUAL TO nsvd0.
!
        call gen_random_mat( s(:nsvd0), a )
!
!       COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
        anorm = norm( s(:nsvd0) )
!
    end if
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,m), resid(n,nsing), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX.
!
        a2(:n,:m) = a(:n,:m)
!
        if ( mat_type>0_i4b ) then
!
!           ALLOCATE WORK ARRAY.
!
            allocate( s0(nsing), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           SAVE THE TRUE SINGULAR VALUES.
!
            if ( nsvd0<nsing ) then
                s0(:nsvd0) = s(:nsvd0)
                s0(nsvd0+1_i4b:nsing) = zero
            else
                s0(:nsing) = s(:nsing)
            end if
!
        end if
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL n-BY-m MATRIX a
!   IS WRITTEN
!
!                       a = U * S * V**(t)
!
!   WHERE S IS AN n-BY-m MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   min(m,n) DIAGONAL ELEMENTS, U IS AN n-BY-n ORTHOGONAL MATRIX, AND
!   V IS AN m-BY-m ORTHOGONAL MATRIX.  THE DIAGONAL ELEMENTS OF S
!   ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a .
!
!   THIS PROGRAM ILLUSTRATES HOW TO COMPUTE ALL THE SINGULAR VALUES OF a AND
!   ONLY nsing LEFT AND RIGHT SINGULAR VECTORS OF a (E.G. A PARTIAL SVD DECOMPOSITION
!   OF a) IN TWO STEPS:
!
!   STEP1 : COMPUTE ALL SINGULAR VALUES OF a AND STORE INTERMEDIATE RESULTS WITH SUBROUTINE svd_cmp4.
!
    call svd_cmp4( a, s, failure=failure_bd_svd, v=p, sort=sort, d=d, e=e, sing_vec=false, &
                   gen_p=gen_p, reortho=reortho, failure_bd=failure_bd  )
!
!   ON EXIT OF svd_cmp4 :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE BIDIAGONAL
!                         SVD OF AN INTERMEDIATE BIDIAGONAL FORM B OF a.
!
!       failure_bd= false :  INDICATES SUCCESSFUL EXIT
!       failure_bd= true  :  INDICATES THAT a IS NEARLY SINGULAR AND SOME LOSS OF ORTHOGONALITY
!                            CAN BE EXPECTED IN THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION
!                            OF a FOR THE LAST LEFT SINGULAR VECTORS.
!
!       s IS OVERWRITTEN WITH THE SINGULAR VALUES OF a.
!
!   IF SORT = 'a' OR 'A', THE SINGULAR VALUES ARE SORTED INTO ASCENDING ORDER.
!   IF SORT = 'd' OR 'D', THE SINGULAR VALUES ARE SORTED INTO DESCENDING ORDER.
!
!   HERE, SORT = 'd' IS USED. THIS IS REQUIRED FOR THE USE OF bd_inviter2 LATER.
!                
!   IF THE PARAMETER sing_vec IS USED WITH THE VALUE false IN THE CALL OF svd_cmp4,
!   svd_cmp4 COMPUTES ONLY THE SINGULAR VALUES OF a AND, OPTIONALLY, STORES THE
!   INTERMEDIATE BIDIAGONAL FORM OF a AND THE ORTHOGONAL MATRICES USED TO REDUCE
!   a TO BIDIAGONAL FORM.
!
!   THE INTERMEDIATE BIDIAGONAL FORM OF a IS STORED IN THE OPTIONAL ARGUMENTS d AND e.
!   THE ORTHOGONAL MATRICES USED TO REDUCE a TO BIDIAGONAL FORM d AND e ARE STORED 
!   IN mat AND p ON EXIT.
!
!   STEP2 : COMPUTE THE FIRST nsing LEFT AND RIGHT SINGULAR VECTORS OF a BY maxiter
!   INVERSE ITERATIONS ON THE INTERMEDIATE BIDIAGONAL MATRIX AND BACK-TRANSFORMATION
!   WITH SUBROUTINE bd_inviter2 .
!
!   ON ENTRY, PARAMETER s CONTAINS SELECTED SINGULAR VALUES OF THE BIDIAGONAL MATRIX d_e (OR a).
!   THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
    call bd_inviter2( a, p, d, e, s(:nsing), leftvec, rightvec, failure=failure_bd_inviter, &
                      maxiter=maxiter )
!
!   ON EXIT OF bd_inviter2 :
!
!         failure= false :  INDICATES SUCCESSFUL EXIT.
!         failure= true  :  INDICATES THAT THE ALGORITHM FAILS TO CONVERGE FOR SOME SINGULAR VECTORS.
!
!   THE SINGULAR VECTORS OF THE BIDIAGONAL FORM ARE COMPUTED USING maxiter INVERSE ITERATIONS
!   FOR ALL THE SINGULAR VALUES AT ONE STEP. THE SINGULAR VECTORS OF ARE THEN ORTHOGONALIZED BY 
!   THE MODIFIED GRAM-SCHMIDT ALGORITHM IF NECESSARY. THE SINGULAR VECTORS OF a ARE FINALLY COMPUTED 
!   BY BACK TRANSFORMATIONS. ON EXIT, leftvec AND rightvec CONTAIN THE FIRST nsing LEFT AND RIGHT
!   SINGULAR VECTORS OF a, RESPECTIVELY.
!
!   NOTE THAT bd_inviter2 MAY FAIL IF SOME THE SINGULAR VALUES SPECIFIED IN PARAMETER s ARE NEARLY
!   IDENTICAL FOR SOME PATHOLOGICAL MATRICES.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
        if ( mat_type>0_i4b ) then
!
!           COMPUTE ERRORS FOR SINGULAR VALUES.
!
            where( s0(:nsing)/=zero )
                resid(:nsing,1_i4b) = s0(:nsing)
            elsewhere
                resid(:nsing,1_i4b) = one
            end where
!
!           ABSOLUTE ERRORS OF SINGULAR VALUES.
!
            abs_err = maxval( abs( s(:nsing) - s0(:nsing) ) )
!
!           RELATIVE ERRORS OF SINGULAR VALUES.
!
            rel_err = maxval( abs( (s(:nsing) - s0(:nsing))/resid(:nsing,1_i4b) ) )
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( s0 )
!
        end if
!
!       DETERMINE THE EFFECTIVE RANK OF a .
!
        tmp = max( tol*s(1_i4b), safmin )
!
        nrank = ifirstloc( logical( s<=tmp, lgl ) ) - 1_i4b
        nvec = min( nrank, nsing )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*V(:m,:nvec) - U(:n,:nvec)*S(:nvec,:nvec),
!       WHERE leftvec AND rightvec ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        resid(:n,:nvec) = matmul(a2,rightvec(:m,:nvec)) - leftvec(:n,:nvec)*spread(s(:nvec),dim=1,ncopies=n)
        a2(:nvec,1_i4b) = norm( resid(:n,:nvec), dim=2_i4b )
!
        err1 = maxval( a2(:nvec,1_i4b) )/( anorm*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U(:n,:nvec)**(t)*U(:n,:nvec).
!
        call unit_matrix( a2(:nsing,:nsing) )
!
        resid(:nvec,:nvec) = abs( a2(:nvec,:nvec) - matmul( transpose(leftvec(:n,:nvec)), leftvec(:n,:nvec) ) )
!
        err2 = maxval( resid(:nvec,:nvec) )/real(n,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V(:m,:nsing)**(t)*V(:m,:nsing).
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(rightvec), rightvec ) )
!
        err3 = maxval( resid(:nsing,:nsing) )/real(m,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, p, leftvec, rightvec, s, d, e )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure_bd_inviter ) then
!    if ( err<=eps .and. .not.failure_bd_svd .and. .not.failure_bd_inviter ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix_type < 1  -> random matrix from the uniform distribution'
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type = 4  -> very slow decay of singular values'
    write (prtunit,*) ' Matrix type = 5  -> strongly clustered singular values at 1'
    write (prtunit,*) '                     spectrum with few deflations (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 6  -> geometric distribution of singular values'
    write (prtunit,*) '                     spectrum with moderate deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 7  -> arithmetic distribution of singular values'
    write (prtunit,*) '                     spectrum with complete deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type > 7  -> uniform distribution of singular values'
!
    if ( do_test ) then
!
        write (prtunit,*) 
!
        if ( mat_type>0_i4b ) then
!
            write (prtunit,*) 'Absolute accuracy of the computed singular values    = ', abs_err
            write (prtunit,*) 'Relative accuracy of the computed singular values    = ', rel_err
!
        end if
!
        write (prtunit,*) 'Number of requested singular triplets                = ', nsing
        write (prtunit,*) 'Effective rank of the matrix                         = ', nrank
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
!
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' FAILURE_BD ( from svd_cmp4() ) = ', failure_bd
    write (prtunit,*) ' FAILURE    ( from svd_cmp4() ) = ', failure_bd_svd
    write (prtunit,*) ' FAILURE ( from bd_inviter2() ) = ', failure_bd_inviter
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and ', nsing, ' singular vectors of a ', &
       n, ' by ', m,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_svd_cmp4
! ===========================
!
end program ex1_svd_cmp4
