program ex1_svd_cmp6
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SVD_CMP6 in module SVD_Procedures
!   for computing all, or only the leading, singular values and vectors of a real m-by-n matrix by
!   a bidiagonal factorization, and the bisection and inverse iteration methods for the required singular
!   values and vectors.
!
!   The computations are parallelized if OpenMP is used and the one-sided Rhala-Barlow algorithm
!   is used in the bidiagonalization phase.                                                                        
!
!   Finally, SVD_CMP6 corrects fully the possible loss of orthogonality induced by the one-sided
!   Ralha-Barlow bidiagonal reduction algorithm if the input matrix is nearly singular contrary
!   to SVD_CMP3 and SVD_CMP4 subroutines.
!
!
! LATEST REVISION : 12/03/2024
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, seven, c30, c50, c1_e6,  &
                         svd_cmp6, norm, unit_matrix, random_seed_, random_number_,       &
                         gen_random_mat, singval_sort, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX (WHICH MUST BE LESS OR EQUAL TO k)
! FOR CASES GREATER THAN 0,
! maxiter IS THE NUMBER OF INVERSE ITERATIONS PERFORMED TO COMPUTE THE TOP SINGULAR VECTORS.
!
    integer(i4b), parameter :: prtunit = 6, m=3000, n=3000, k=min(m,n), nsvd0=3000, maxiter=2
!   
! conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN CASES 5, 6 and 7,
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50, conda=c1_e6
!
    character(len=*), parameter :: name_proc='Example 1 of svd_cmp6'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err, err1, err2, err3, eps, tmp, tmp2, ulp, &
                  abs_err, rel_err, anorm, elapsed_time
!
    real(stnd), allocatable, dimension(:)   :: singval0
    real(stnd), allocatable, dimension(:,:) :: a, a2, resid
!
    real(stnd), dimension(:),   pointer :: s
    real(stnd), dimension(:,:), pointer :: v
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: i, mat_type, nsvd
!
    logical(lgl) :: failure, failure_bd, failure_bisect, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : FULL OR PARTIAL SVD OF A REAL MATRIX USING THE ONE-SIDED
!               RALHA-BARLOW BIDIAGONAL REDUCTION ALGORITHM, A BISECTION
!               ALGORITHM FOR COMPUTING SINGULAR VALUES AND AN INVERSE
!               ITERATION ALGORITHM FOR COMPUTING SINGULAR VECTORS.
!               A FINAL BACK TRANSFORMATION AND REORTOGONALIZATION STEP
!               IS ALSO PERFORMED TO CORRECT FOR THE POSSIBLE LOSS OF
!               ORTHOGONALITY INDUCED BY THE ONE-SIDED RALHA-BARLOW
!               BIDIAGONAL REDUCTION ALGORITHM IF THE INPUT MATRIX
!               IS NEARLY SINGULAR.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type < 1  -> RANDOM MATRIX FROM THE UNIFORM DISTRIBUTION
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type = 4  -> VERY SLOW DECAY OF SINGULAR VALUES
!   mat_type = 5  -> STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda
!                    SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE
!   mat_type = 6  -> GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type = 7  -> ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type > 7  -> UNIFORM DISTRIBUTION OF SINGULAR VALUES
!
    mat_type = 8_i4b
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
!
    err = zero
!
!   SPECIFY nsvd, THE NUMBER OF THE LARGEST SINGULAR TRIPLETS WHICH MUST BE COMPUTED.
!
    nsvd = 20
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   ALLOCATE WORK ARRAYS.
!
    if ( mat_type>0_i4b ) then
!
        allocate( a(m,n), singval0(k), stat=iok )
!
    else
!
        allocate( a(m,n), stat=iok )
!
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM UNIFORM MATRIX OR SINGULAR VALUES.
!
    select case( mat_type )
!
        case( :0_i4b )
!
!           RANDOM UNIFORM MATRIX.
!
            call random_number_( a )
!
!           COMPUTE THE FROBENIUS NORM OF THE RANDOM MATRIX.
!
            anorm = norm( a )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
            end do
!
        case( 4_i4b )
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            tmp = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                singval0(i) = exp( -tmp2/tmp )
!        
            end do
!
        case( 5_i4b )
!
!           STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE.
!
            singval0(:nsvd0-1_i4b) = one
            singval0(nsvd0) = one/conda
!
        case( 6_i4b )
!
!           GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                singval0(i) = conda**( -tmp2/tmp )
!        
            end do
!
        case( 7_i4b )
!
!           ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
            tmp2 = one - one/conda
!
            do i = 1_i4b, nsvd0
!
                singval0(i) = one - (real( i - 1_i4b, stnd )/tmp)*tmp2
!        
            end do
!
        case default
!
!           UNIFORM DISTRIBUTION OF SINGULAR VALUES.
!
            call random_number_( singval0(:nsvd0) )
!
    end select
!
    if ( mat_type>0_i4b ) then
!
#ifdef _F2003
        if ( ieee_support_datatype( singval0 ) ) then
!
            if ( .not.all( ieee_is_normal( singval0(:nsvd0) ) ) ) then
!
                call merror( name_proc//' : Exceptions occurred when generating the input matrix !'  )                    
!
            end if
!
        end if
#endif
!
!       SORT THE SINGULAR VALUES BY DECREASING MAGNITUDE.
!
        call singval_sort( 'D', singval0(:nsvd0) )
!
!       GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!       OF SINGULAR VALUES AND A RANK EQUAL TO nsvd0.
!
        call gen_random_mat( singval0(:nsvd0), a )
!
!       COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
        anorm = norm( singval0(:nsvd0) )
!
    end if
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( a2(m,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX.
!
        a2(:m,:n) = a(:m,:n)
!
        if ( mat_type>0_i4b ) then
!
            singval0(nsvd0+1_i4b:k) = zero
!
        end if
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   NULLIFY THE POINTERS s AND v SO THAT THEIR STATUT CAN BE CHECKED INSIDE
!   svd_cmp6 SUBROUTINE.
!
    nullify( s, v )
!
!   svd_cmp6 COMPUTES A FULL OR PARTIAL SINGULAR VALUE DECOMPOSITION (SVD) OF
!   A REAL m-BY-n MATRIX a. THE FULL SVD IS WRITTEN
!
!                       a = U * S * V**(t)
!
!   WHERE S IS AN m-BY-n MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   min(m,n) DIAGONAL ELEMENTS, U IS AN m-BY-m ORTHOGONAL MATRIX, AND
!   V IS AN n-BY-n ORTHOGONAL MATRIX.  THE DIAGONAL ELEMENTS OF S
!   ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!   BY DEFAULT, svd_cmp6 COMPUTES min(m,n) SINGULAR TRIPLETS (E.G., A THIN SVD).
!   IF YOU WANT A PARTIAL SVD, YOU CAN SPECIFY THE OPTIONAL INTEGER
!   PARAMETER nsvd WITH THE REQUESTED RANK OF THE PARTIAL SVD.
!
    call svd_cmp6( a, s, v, failure, sort=sort, nsvd=nsvd, maxiter=maxiter, &
                   failure_bd=failure_bd, failure_bisect=failure_bisect )
!
!   THE ROUTINE RETURNS THE nsvd LARGEST SINGULAR VALUES AND THE ASSOCIATED nsvd LEFT
!   AND RIGHT SINGULAR VECTORS.
!
!   ON EXIT OF svd_cmp6 :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE INVERSE ITERATION ALGORITHM FAILS TO CONVERGE FOR
!                         SOME SINGULAR VECTORS IN maxiter ITERATIONS.
!
!       failure_bd= false :  INDICATES SUCCESSFUL EXIT AND THAT MAXIMUM ACCURACY WAS OBTAINED
!                            IN THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION OF a .
!       failure_bd= true  :  INDICATES THAT a IS NEARLY SINGULAR.
!
!       failure_bisect= false :  INDICATES SUCCESSFUL EXIT AND THAT THE BISECTION ALGORITHM CONVERGED
!                                FOR ALL THE COMPUTED SINGULAR VALUES TO THE DESIRED ACCURACY.
!       failure_bisect= true  :  INDICATES THAT SOME OR ALL OF THE SINUGULAR VALUES FAILED TO CONVERGE
!                                OR WERE NOT COMPUTED. THIS IS GENERALLY CAUSED BY INACCURATE ARITHMETIC.
!
!       nsvd IS THE NUMBER OF SINGULAR TRIPLETS WHICH HAVE BEEN
!       COMPUTED BY THE SUBROUTINE, WHICH CAN BE GREATER THAN THE REQUESTED
!       NUMBER IF MULTIPLE SINGULAR VALUES AT INDEX nsvd MAKE UNIQUE SELECTION
!       IMPOSSIBLE.
!
!       a IS OVERWRITTEN WITH THE FIRST nsvd LEFT SINGULAR VECTORS,
!       STORED COLUMNWISE.
!
!       POINTER s CONTAINS THE nsvd LARGEST SINGULAR VALUES OF a .
!   
!       POINTER v CONTAINS THE ASSOCIATED nsvd RIGHT SINGULAR VECTORS
!       STORED COLUMNWISE.
!
!   IF SORT = 'a' OR 'A', THE SINGULAR VALUES ARE SORTED INTO ASCENDING ORDER.
!   IF SORT = 'd' OR 'D', THE SINGULAR VALUES ARE SORTED INTO DESCENDING ORDER.
!   BY DEFAULT, THE SINGULAR VALUES ARE SORTED INTO INTO DESCENDING ORDER.
!   THE SINGULAR VECTORS ARE REARRANGED ACCORDINGLY.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!                
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( resid(m,nsvd), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        if ( mat_type>0_i4b ) then
!
!           COMPUTE ERRORS FOR SINGULAR VALUES.
!
            where( singval0(:nsvd)/=zero )
                resid(:nsvd,1_i4b) = singval0(:nsvd)
            elsewhere
                resid(:nsvd,1_i4b) = one
            end where
!
!           ABSOLUTE ERRORS OF SINGULAR VALUES.
!
            abs_err = maxval( abs( s(:nsvd) - singval0(:nsvd) ) )
!
!           RELATIVE ERRORS OF SINGULAR VALUES.
!
            rel_err = maxval( abs( (s(:nsvd) - singval0(:nsvd))/resid(:nsvd,1_i4b) ) )
!
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*V(:,:nsvd) - U(:,:nsvd)*S(:nsvd,:nsvd).
!
        resid(:m,:nsvd) = matmul(a2,v) - a(:m,:nsvd)*spread(s(:nsvd),dim=1,ncopies=m)
        a2(:nsvd,1_i4b) = norm( resid(:m,:nsvd), dim=2_i4b )
        err1 =  maxval( a2(:nsvd,1_i4b) )/( anorm*real(m,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( a2(:nsvd,:nsvd) )
!
        resid(:nsvd,:nsvd) = abs( a2(:nsvd,:nsvd) - matmul( transpose(a(:m,:nsvd)), a(:m,:nsvd) ) )
        err2 = maxval( resid(:nsvd,:nsvd) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V**(t)*V.
!
        resid(:nsvd,:nsvd) = abs( a2(:nsvd,:nsvd) - matmul( transpose(v(:n,:nsvd)), v(:n,:nsvd) ) )
        err3 = maxval( resid(:nsvd,:nsvd) )/real(n,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS AND POINTERS.
!
    if ( mat_type>0_i4b ) then
        deallocate( a, v, s, singval0 )
    else
        deallocate( a, v, s )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix_type < 1  -> random matrix from the uniform distribution'
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type = 4  -> very slow decay of singular values'
    write (prtunit,*) ' Matrix type = 5  -> strongly clustered singular values at 1'
    write (prtunit,*) '                     spectrum with few deflations (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 6  -> geometric distribution of singular values'
    write (prtunit,*) '                     spectrum with moderate deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 7  -> arithmetic distribution of singular values'
    write (prtunit,*) '                     spectrum with complete deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type > 7  -> uniform distribution of singular values'
!
    if ( do_test ) then
!
        write (prtunit,*) 
!
        if ( mat_type>0_i4b ) then
!
            write (prtunit,*) 'Absolute accuracy of the computed singular values    = ', abs_err
            write (prtunit,*) 'Relative accuracy of the computed singular values    = ', rel_err
!
        end if
!
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
!
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' FAILURE        ( from svd_cmp6() ) = ', failure
    write (prtunit,*) ' FAILURE_BD     ( from svd_cmp6() ) = ', failure_bd
    write (prtunit,*) ' FAILURE_BISECT ( from svd_cmp6() ) = ', failure_bisect
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing ', nsvd, ' singular values and vectors of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_svd_cmp6
! ===========================
!
end program ex1_svd_cmp6
