program ex1_symtrid_ratqri
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SYMTRID_RATQRI
!   in module Eig_Procedures for computing the largest or smallest eigenvalues of a
!   real n-ny-n symmetric tridiagonal matrix.
!
!   A rational QR method is used for computing the eigenvalues of the symmetric
!   tridiagonal matrix.
!
!
! Further Details
! ===============
!
!    The program also shows the use of subroutine TRID_INVITER in module Eig_Procedures
!    for computing selected eigenvectors of a real symmetric tridiagonal matrix.
!
!
! LATEST REVISION : 27/11/2023
!
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack,  only : lgl, i4b, stnd, true, false, zero, one, two, c50,     &
                          allocate_error, merror, trid_inviter, symtrid_ratqri, &
                          norm, unit_matrix 
#ifdef _MATMUL
    use Statpack,  only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Utilities, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE DIMENSION OF THE TRIDIAGONAL MATRIX AND neig IS THE NUMBER
! OF THE COMPUTED SMALLEST OR LARGEST EIGENVALUES/EIGENVECTORS,
! maxiter IS THE NUMBER OF INVERSE ITERATIONS PERFORMED TO COMPUTE THE neig ASSOCIATED EIGENVECTORS.
!
    integer(i4b), parameter :: prtunit=6, n=4000, neig=200, maxiter=2
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of symtrid_ratqri'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e, e2, eigval
    real(stnd), allocatable, dimension(:,:) :: a, a2, resid, eigvec
!
    integer(i4b) :: l
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: failure, failure2, do_test, small
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTE THE neig LARGEST OR SMALLEST EIGENVALUES OF
!               A SYMMETRIC TRIDIAGONAL MATRIX USING A RATIONAL QR METHOD
!               AND SELECTED EIGENVECTORS BY INVERSE ITERATION.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( eps )
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   DETERMINE IF YOU WANT TO COMPUTE THE neig SMALLEST OR LARGEST EIGENVALUES.
!
    small = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( eigvec(n,neig), d(n), e(n), e2(n), eigval(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A (1-2-1) TRIDIAGONAL MATRIX.
!   THE DIAGONAL ELEMENTS ARE STORED IN d .
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN e .
!
    d(:n) = two
    e(:n) = one
!
!    call random_number( d(:n) )
!    call random_number( e(:n) )
!
!   SAVE THE TRIDIAGONAL FORM FOR LATER USE.
!
    eigval(:n) = d(:n)
    e2(:n)     = e(:n)
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE neig LARGEST EIGENVALUES OF THE SYMMETRIC TRIDIAGONAL MATRIX.
!
    call symtrid_ratqri( eigval(:n), e2(:n), neig, failure, small=small )
!
!   ON EXIT OF symtrid_ratqri :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE RATIONAL
!                         QR ALGORITHM.
!
!   THE COMPUTED EIGENVALUES ARE STORED IN eigval(:neig)
!   AND THE SYMMETRIC TRIDIAGONAL MATRIX IS OVERWRITTEN.
!
!   NEXT, COMPUTE THE neig ASSOCIATED EIGENVECTORS OF THE TRIDIAGONAL MATRIX BY maxiter INVERSE ITERATIONS.
!
    call trid_inviter( d(:n), e(:n), eigval(:neig), eigvec(:n,:neig), failure2, maxiter=maxiter )
!
!   ON EXIT OF trid_inviter :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT
!       failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE ALGORITHM

!       eigvec CONTAINS THE neig EIGENVECTORS ASSOCIATED WITH THE EIGENVALUES eigval(:neig).
!
!       trid_inviter MAY FAIL IF SOME THE EIGENVALUES SPECIFIED IN PARAMETER eigval ARE NEARLY
!       IDENTICAL.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a(n,n), a2(neig,neig), resid(n,neig), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       FORM THE TRIDIAGONAL MATRIX.
!
        a(:n,:n) = zero
!
        do l = 1_i4b, n-1_i4b
!
            a(l,l)       = d(l)
            a(l+1_i4b,l) = e(l)
            a(l,l+1_i4b) = e(l)
!
        end do
!
        a(n,n) = d(n)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*s
!       WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX a.
!
        resid = matmul( a, eigvec ) - eigvec*spread( eigval(:neig), 1, n )
!
        err1 = norm(resid)/( norm(a)*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!       WHERE u are THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX a.
!
        call unit_matrix( a2 )
!
        resid(:neig,:neig) = a2 - matmul( transpose( eigvec ), eigvec )
!
        err2 = norm(resid(:neig,:neig))/real(n,stnd)
!
        err = max( err1, err2 )
!
!       ALLOCATE WORK ARRAYS.
!
        deallocate( a, a2, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( eigvec, d, e, e2, eigval )
!
!   PRINT THE RESULTS OF THE TESTS.
!
    if ( err<=eps .and. .not.failure  .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' FAILURE ( from symtrid_ratqri ) = ', failure
    write (prtunit,*) ' FAILURE ( from trid_inviter() ) = ', failure2
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing ', neig, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric tridiagonal matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_symtrid_ratqri
! =================================
!
end program ex1_symtrid_ratqri
