program ex1_trid_inviter
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine TRID_INVITER
!   in module Eig_Procedures for computing all or selected eigenvectors of a
!   real symmetric matrix by inverse iterations.                                                                  
!
!
! Further Details
! ===============
!
!    The program also shows the use of subroutine EIGVAL_CMP in module EIG_Procedures.
!                                                                              
! LATEST REVISION : 28/11/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, seven, c30, c50, c1_e6, &
                         eigval_cmp, trid_inviter, norm, unit_matrix, random_seed_,      &
                         random_number_, gen_random_sym_mat, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE DIMENSION OF THE GENERATED SYMMETRIC MATRIX,
! neig0 IS THE NUMERICAL RANK OF THE GENERATED SYMMETRIC MATRIX (WHICH MUST BE LESS OR EQUAL TO n) FOR CASES GREATER THAN 0,
! nvec IS THE NUMBER OF EIGENVECTORS, WHICH MUST BE COMPUTED,
! maxiter IS THE NUMBER OF INVERSE ITERATIONS PERFORMED TO COMPUTE EIGENVECTORS.
!
    integer(i4b), parameter :: prtunit=6, n=3000, neig0=3000, nvec=3000, maxiter=2
!   
! conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN CASES 5, 6 AND 7,
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50, conda=c1_e6
!
    character(len=*), parameter :: name_proc='Example 1 of trid_inviter'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, tmp, tmp2, ulp, &
                                               anorm, elapsed_time
    real(stnd), allocatable, dimension(:)   :: eigval, resid2
    real(stnd), allocatable, dimension(:,:) :: a, eigvec, a2, d_e
!
    integer       :: iok, istart, iend, irate, imax, itime
    integer(i4b)  :: i, mat_type
!
    logical(lgl)  :: failure, failure2, do_test
!   
    character     :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : EIGENVALUES AND, OPTIONALLY, SELECTED EIGENVECTORS OF
!               A REAL SYMMETRIC MATRIX USING THE INVERSE ITERATION METHOD.
!
!   SPECIFY THE TYPE OF INPUT SYMMETRIC MATRIX:
!
!   mat_type < 1  -> RANDOM SYMMETRIC MATRIX FROM THE UNIFORM DISTRIBUTION
!   mat_type = 1  -> SLOW DECAY OF EIGENVALUES
!   mat_type = 2  -> FAST DECAY OF EIGENVALUES
!   mat_type = 3  -> S-SHAPED DECAY OF EIGENVALUES
!   mat_type = 4  -> VERY SLOW DECAY OF EIGENVALUES
!   mat_type = 5  -> STRONGLY CLUSTERED EIGENVALUES AT 1 WITH CONDITION NUMBER conda
!                    SPECTRUM WITH FEW DEFLATIONS FOR LAPACK SYESVD ROUTINE
!   mat_type = 6  -> GEOMETRIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH MODERATE DEFLATION FOR LAPACK SYESVD ROUTINE
!   mat_type = 7  -> ARITHMETIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK SYESVD ROUTINE
!   mat_type > 7  -> UNIFORM DISTRIBUTION OF EIGENVALUES
!
    mat_type = 3_i4b
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), eigvec(n,nvec), eigval(n), d_e(n,2_i4b), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE A RANDOM UNIFORM SYMMETRIC MATRIX OR EIGENVALUES.
!
    select case( mat_type )
!
        case( :0_i4b )
!
!           RANDOM UNIFORM SYMMETRIC MATRIX.
!
            call random_number_( a )
!
            a = a + transpose( a )
!
!           COMPUTE THE FROBENIUS NORM OF THE RANDOM SYMMETRIC MATRIX.
!
            anorm = norm( a )
!
        case( 1_i4b )
!
!           SLOW DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                eigval(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                eigval(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                eigval(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
            end do
!
        case( 4_i4b )
!
!           VERY SLOW DECAY OF EIGENVALUES.
!
            tmp = real( neig0 - 1_i4b, stnd )
!
            do i = 1_i4b, neig0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                eigval(i) = exp( -tmp2/tmp )
!        
            end do
!
        case( 5_i4b )
!
!           STRONGLY CLUSTERED EIGENVALUES AT 1 WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH FEW DEFLATIONS FOR LAPACK SYESVD ROUTINE.
!
            eigval(:neig0-1_i4b) = one
            eigval(neig0) = one/conda
!
        case( 6_i4b )
!
!           GEOMETRIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH MODERATE DEFLATION FOR LAPACK SYESVD ROUTINE.
!
            tmp  = real( neig0 - 1_i4b, stnd )
!
            do i = 1_i4b, neig0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                eigval(i) = conda**( -tmp2/tmp )
!        
            end do
!
        case( 7_i4b )
!
!           ARITHMETIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK SYESVD ROUTINE.
!
            tmp  = real( neig0 - 1_i4b, stnd )
            tmp2 = one - one/conda
!
            do i = 1_i4b, neig0
!
                eigval(i) = one - (real( i - 1_i4b, stnd )/tmp)*tmp2
!        
            end do
!
        case default
!
!           UNIFORM DISTRIBUTION OF EIGENVALUES.
!
            call random_number_( eigval(:neig0) )
!
    end select
!
    if ( mat_type>0_i4b ) then
!
#ifdef _F2003
        if ( ieee_support_datatype( eigval ) ) then
!
            if ( .not.all( ieee_is_normal( eigval(:neig0) ) ) ) then
!
                call merror( name_proc//' : Exceptions occurred when generating the input symmetric matrix !'  )                    
!
            end if
!
        end if
#endif
!
!       GENERATE A n-BY-n RANDOM REAL SYMMETRIC MATRIX a WITH A SPECIFIED DISTRIBUTION
!       OF EIGENVALUES AND A RANK EQUAL TO neig0.
!
        call gen_random_sym_mat( eigval(:neig0), a )
!
!       COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
        anorm = norm( eigval(:neig0) )
!
    end if
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), resid2(nvec), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE RANDOM SELF-ADJOINT MATRIX a .
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE EIGENVALUE DECOMPOSITION OF A REAL n-BY-n SYMMETRIC MATRIX a
!   IS WRITTEN
!
!                       a = U * D * U**(t)
!
!   WHERE D IS AN n-BY-n DIAGONAL MATRIX AND U IS AN n-BY-n ORTHOGONAL MATRIX.
!   THE DIAGONAL ELEMENTS OF D ARE THE EIGENVALUES OF a; THEY ARE REAL.
!   THE COLUMNS OF U ARE THE EIGENVECTORS OF a.
!
!   FIRST, COMPUTE ALL EIGENVALUES OF THE SELF-ADJOINT MATRIX a AND SAVE
!   THE INTERMEDIATE TRIDIAGONAL MATRIX IN PARAMETER d_e WITH
!   SUBROUTINE eigval_cmp.
!
    call eigval_cmp( a, eigval, failure=failure, sort=sort, d_e=d_e )
!
!   ON EXIT OF eigval_cmp:
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE EIGENDECOMPOSITION 
!                         OF AN INTERMEDIATE TRIDIAGONAL FORM TRID OF a.
!
!       a IS OVERWRITTEN WITH THE HOUSEHOLDER TRANSFORMATIONS USED TO REDUCE a
!       TO TRIDIAGONAL FORM IF THE OPTIONAL PARAMETER d_e IS PRESENT, OTHERWISE
!       a IS DESTROYED.
!
!       eigval IS OVERWRITTEN WITH THE EIGENVALUES OF a.
!       IF SORT = 'a' OR 'A', THE EIGENVALUES ARE SORTED INTO ASCENDING ORDER.
!       IF SORT = 'd' OR 'D', THE EIGENVALUES ARE SORTED INTO DESCENDING ORDER.
!
!       d_e IS AN OPTIONAL ARGUMENT TO SAVE THE INTERMEDIATE TRIDIAGONAL FORM OF a.
!
!   NEXT COMPUTE THE FIRST nvec EIGENVECTORS OF THE SELF-ADJOINT MATRIX a BY
!   maxiter INVERSE ITERATIONS ON THE INTERMEDIATE TRIDIAGONAL MATRIX
!   d_e AND BACK-TRANSFORMATION.
!
!   ON ENTRY OF SUBROUTINE trid_inviter, PARAMETER eigval CONTAINS SELECTED
!   EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!   THE EIGENVALUES VALUES MUST BE GIVEN IN DECREASING ORDER.
!
    call trid_inviter( d_e(:n,1), d_e(:n,2), eigval(:nvec), eigvec(:n,:nvec), failure=failure2,   &
                       mat=a, maxiter=maxiter )
!
!   ON EXIT OF trid_inviter :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT
!       failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE ALGORITHM

!       eigvec CONTAINS THE nvec EIGENVECTORS ASSOCIATED WITH THE EIGENVALUES eigval(:nvec).
!
!       trid_inviter MAY FAIL IF SOME THE EIGENVALUES SPECIFIED IN PARAMETER eigval ARE NEARLY
!       IDENTICAL.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*U - U*D
!
        a(:n,:nvec) = matmul( a2, eigvec ) - eigvec*spread( eigval(:nvec), 1, n )
        resid2(:nvec) = norm( a(:n,:nvec), dim=2_i4b )
!
        err1 =  maxval( resid2(:nvec) )/( anorm*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( a(:nvec,:nvec) )
!
        a2(:nvec,:nvec) = abs( a(:nvec,:nvec) - matmul( transpose( eigvec ), eigvec ) )
!
        err2 = maxval(a2(:nvec,:nvec))/real(n,stnd)
!
        err = max( err1, err2 )
!
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, a2, eigvec, eigval, d_e, resid2 )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, eigvec, eigval, d_e )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix_type < 1  -> random symmetric matrix from the uniform distribution'
    write (prtunit,*) ' Matrix type = 1  -> slow decay of eigenvalues'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of eigenvalues'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of eigenvalues'
    write (prtunit,*) ' Matrix type = 4  -> very slow decay of eigenvalues'
    write (prtunit,*) ' Matrix type = 5  -> strongly clustered eigenvalues at 1'
    write (prtunit,*) '                     spectrum with few deflations (for LAPACK SYESVD routine)'
    write (prtunit,*) ' Matrix type = 6  -> geometric distribution of eigenvalues'
    write (prtunit,*) '                     spectrum with moderate deflation (for LAPACK SYESVD routine)'
    write (prtunit,*) ' Matrix type = 7  -> arithmetic distribution of eigenvalues'
    write (prtunit,*) '                     spectrum with complete deflation (for LAPACK SYESVD routine)'
    write (prtunit,*) ' Matrix type > 7  -> uniform distribution of eigenvalues'
!
    write (prtunit,*) 
    write (prtunit,*) ' FAILURE ( from eigval_cmp()   ) = ', failure
    write (prtunit,*) ' FAILURE ( from trid_inviter() ) = ', failure2
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing all eigenvalues and ', nvec, ' eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_trid_inviter
! ===============================
!
end program ex1_trid_inviter
