program ex2_bd_singval2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine BD_SINGVAL2
!   in module SVD_Procedures for computing all or selected singular values
!   of a real bidiagonal matrix by the bisection method. The singular values
!   can be computed at high relative precision at the user option.
!   BD_SINGVAL2 is faster, but less accurate than BD_SINGVAL subroutine.
!
!
! Further Details
! ===============
!
!    The program also shows the use of subroutines BD_CMP, APPLY_Q_BD, APPLY_P_BD
!    in module SVD_Procedures for computing all or selected singular vectors
!   of a real matrix by inverse iterations.
!
!
! LATEST REVISION : 07/12/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, safmin, c50, bd_inviter,  &
                         bd_cmp, bd_singval2, apply_q_bd, apply_p_bd, merror,         &
                         allocate_error, norm
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n AND m ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! ls IS THE NUMBER OF SINGULAR TRIPLETS WHICH MUST BE COMPUTED,
! maxiter IS THE NUMBER OF INVERSE ITERATIONS PERFORMED TO COMPUTE THE TOP ls SINGULAR VECTORS.
!
    integer(i4b), parameter :: prtunit=6, n=1000, m=500, mn=min(m,n), ls=20, maxiter=2
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of bd_singval2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, abstol
    real(stnd), dimension(n,m)              :: a, a2
    real(stnd), dimension(:,:), allocatable :: leftvec, rightvec
    real(stnd), dimension(mn)               :: s, d, e, tauq, taup
!
    integer(i4b) :: nsing
    integer      :: iok
!
    logical(lgl) :: failure, bd_is_upper, vector
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : PARTIAL SVD OF A n-BY-m REAL MATRIX USING THE GOLUB-REINSCH
!               BIDIAGONAL REDUCTION ALGORITHM, A BISECTION ALGORITHM FOR
!               SINGULAR VALUES AND THE INVERSE ITERATION TECHNIQUE FOR
!               SINGULAR VECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
!   CHOOSE TUNING PARAMETERS FOR THE BISECTION ALGORITHM.
!
!   SINGULAR VALUES WILL BE COMPUTED MOST ACCURATELY WHEN abstol IS SET
!   TO THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD (E.G., sqrt(safmin) ).
!
    abstol = sqrt( safmin )
!
!   COMPUTE THE SINGULAR VALUES WITH A VECTORIZED VERSION OF THE BISECTION
!   ALGORITHM.
!
    vector = true
!
!   GENERATE A RANDOM DATA MATRIX.
!
    call random_number( a )
!
!   SAVE RANDOM DATA MATRIX.
!
    a2(:n,:m) = a(:n,:m)
!
!   REDUCE a TO LOWER OR UPPER BIDIAGONAL FORM.
!
    call bd_cmp( a, d, e, tauq, taup )
!
!   COMPUTE THE FIRST ls SINGULAR VALUES OF BIDIAGONAL FORM OF a BY A BISECTION METHOD.
!
    call bd_singval2( d, e, nsing, s, failure, sort=sort, vector=vector, ls=ls, abstol=abstol )
!
    if ( .not. failure .and. nsing>0 ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE SINGULAR VECTORS.
!
        allocate( leftvec(n,nsing),    &
                  rightvec(m,nsing),   &
                  stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        bd_is_upper = n>=m
!
!       COMPUTE THE FIRST nsing SINGULAR VECTORS OF a BY maxiter
!       INVERSE ITERATIONS ON THE INTERMEDIATE BIDIAGONAL MATRIX
!       d_e AND BACK-TRANSFORMATION.
!
        call bd_inviter( bd_is_upper, d, e, s(:nsing), leftvec(:mn,:nsing), rightvec(:mn,:nsing),  &
                         failure, maxiter=maxiter )
!
!       COMPUTE SINGULAR VECTORS OF a BY BACK-TRANSFORMATION.
!
        if ( bd_is_upper ) then
            leftvec(mn+1_i4b:n,:nsing) = zero
        else
            rightvec(mn+1_i4b:m,:nsing) = zero
        end if
!
!       GENERATE LEFT SINGULAR VECTORS OF a IN leftvec BY BACK-TRANSFORMATION.
!
        call apply_q_bd( a, tauq, leftvec,  left=true, trans=false )
!
!       GENERATE RIGHT SINGULAR VECTORS OF a IN rightvec BY BACK-TRANSFORMATION.
!
        call apply_p_bd( a, taup, rightvec, left=true, trans=false )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*rightvec - leftvec*s(:nsing),
!       WHERE leftvec AND rightvec ARE THE LEFT AND RIGHT SINGULAR VECTORS.
!
        err =  norm( matmul(a2,rightvec) - leftvec*spread(s(:nsing),dim=1,ncopies=n) )/( norm( a2 )*real(mn,stnd) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( leftvec,  rightvec  )
!
    end if
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
!
! END OF PROGRAM ex2_bd_singval2
! ==============================
!
end program ex2_bd_singval2
