program ex2_eigval_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine EIGVAL_CMP
!   in module Eig_Procedures for computing all eigenvalues of a real symmetric matrix
!   stored in packed form.
!
!   The initial tridiagonal reduction is parallelized if OpenMP is used and the fast
!   Pal-Walker-Kahan variant of the QR method with implicit shift is used for
!   computing the eigenvalues.
!
!
! Further Details
! ===============
!
!    The program also shows the use of subroutine TRID_INVITER in module EIG_Procedures
!    for computing eigenvectors of a real symmetric matrix in packed form.
!
!
! LATEST REVISION : 04/12/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm, triangle, trid_inviter,  &
                         eigval_cmp, merror, allocate_error, unit_matrix
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE DIMENSION OF THE SYMMETRIX MATRIX,
! neig IS THE NUMBER OF THE WANTED EIGENVECTORS COMPUTED BY INVERSE ITERATIONS,
! maxiter IS THE NUMBER OF INVERSE ITERATIONS PERFORMED TO COMPUTE THE TOP neig EIGENVECTORS.
!
    integer(i4b), parameter :: prtunit=6, n=3000, p=(n*(n+1))/2, neig=100, maxiter=2
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of eigval_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, elapsed_time
    real(stnd), dimension(:),   allocatable :: vec, d, res
    real(stnd), dimension(:,:), allocatable :: a, eigvec, d_e, resmat
!
    integer :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: failure, failure2, do_test, upper
!   
    character :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : PARTIAL EIGENVALUES/EIGENVECTORS DECOMPOSITION OF A n-BY-n REAL
!               SYMMETRIC MATRIX STORED IN PACKED FORM USING THE FAST PAL-WALKER-KAHAN VARIANT
!               OF THE QR METHOD FOR EIGENVALUES AND THE INVERSE ITERATION TECHNIQUE FOR
!               SELECTED EIGENVECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   SPECIFY IF THE UPPER OR LOWER TRIANGLE OF THE SYMMETRIC
!   MATRIX IS STORED IN PACKED FORM.
!
    upper = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), d(n), d_e(n,2), vec(p), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM DATA MATRIX AND FROM IT
!   A SELF-ADJOINT MATRIX a .
!
    call random_number( a )
!
    a = a + transpose( a ) 
!
!   MAKE A COPY OF THE SELF-ADJOINT MATRIX IN PACKED FORM.
!
    vec(:p) = pack( a, mask=triangle( upper, n, n, extra=1_i4b) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE ALL EIGENVALUES OF THE SELF-ADJOINT MATRIX a AND SAVE
!   THE INTERMEDIATE TRIDIAGONAL MATRIX IN PARAMETER d_e .
!
    call eigval_cmp( vec, d, failure, sort=sort, d_e=d_e )
!
!   THE ROUTINE RETURNS THE EIGENVALUES OF a.
!
!   ON EXIT OF eigval_cmp:
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE EIGENDECOMPOSITION 
!                         OF AN INTERMEDIATE TRIDIAGONAL FORM TRID OF a.
!
!       vec IS OVERWRITTEN BY THE ORTHOGONAL MATRIX USED TO TRANSFORM a IN TRIDIAGONAL FORM IF
!           THE OPTIONAL ARGUMENT d_e IS SPECIFIED. THE MATRIX Q IS STORED IN FACTORED FORM.
!
!       d IS OVERWRITTEN WITH THE EIGENVALUES OF a.
!
!   IF SORT = 'a' OR 'A', THE EIGENVALUES ARE SORTED INTO ASCENDING ORDER.
!   IF SORT = 'd' OR 'D', THE EIGENVALUES ARE SORTED INTO DESCENDING ORDER.
!
    if ( .not. failure .and. neig>0  ) then
!
!       ALLOCATE WORK ARRAY NEEDED TO STORE THE EIGENVECTORS.
!
        allocate( eigvec(n,neig), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE THE FIRST neig EIGENVECTORS OF THE SELF-ADJOINT MATRIX a
!       STORED IN PACKED FORM BY maxiter INVERSE ITERATIONS ON THE INTERMEDIATE
!       TRIDIAGONAL MATRIX d_e AND BACK-TRANSFORMATION.
!
        call trid_inviter( d_e(:n,1_i4b), d_e(:n,2_i4b), d(:neig), eigvec, failure2, &
                           matp=vec, maxiter=maxiter )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test .and. .not.failure .and. neig>0 ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( resmat(n,neig), res(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*eigvec - eigvec*d
!       WHERE d ARE THE EIGENVALUES AND eigvec THE EIGENVECTORS OF a.
!
        resmat(:n,:neig) = matmul( a(:n,:n), eigvec(:n,:neig)) - eigvec(:n,:neig)*spread( d(:neig), dim=1, ncopies=n)
        res(:neig)  = norm( resmat(:n,:neig), dim=2_i4b )
!
        err1 = maxval( res(:neig) )/( norm( a )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - eigvec**(t)*eigvec.
!
        call unit_matrix( a(:neig,:neig) )
!
        resmat(:neig,:neig) = abs( a(:neig,:neig) - matmul( transpose(eigvec(:n,:neig)), eigvec(:n,:neig) ) )
!
        err2 = maxval( resmat(:neig,:neig) )/real(n,stnd)
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( resmat, res )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, d_e, d, vec )
!
    if ( allocated( eigvec ) )  deallocate( eigvec )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure  .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' FAILURE ( from eigval_cmp()   ) = ', failure
!
    if ( .not. failure .and. neig>0  ) then
        write (prtunit,*) ' FAILURE ( from trid_inviter() ) = ', failure2
    end if
!
    if ( do_test  .and. .not.failure .and. neig>0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues/eigenvectors couplets = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors                 = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i6,a,i6,a,i6,a,0pd12.4,a)')    &
      'The elapsed time for computing all the eigenvalues and ', neig, ' eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix stored in packed form is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_eigval_cmp
! =============================
!
end program ex2_eigval_cmp
