program ex2_llsq_qr_solve
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine LLSQ_QR_SOLVE
!   in module LLSQ_Procedures for solving linear least squares problems with a
!   QR decomposition (with column pivoting) or a complete orthogonal factorization
!   of the coefficient matrix.
!
!
! LATEST REVISION : 15/11/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, lgl, stnd, zero, true, false, c50, allocate_error, &
                         merror, llsq_qr_solve
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
! m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX.
! n IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX.
! nrhs IS THE NUMBER OF COLUMNS OF THE RIGHT HAND SIDE MATRIX.
!
    integer(i4b), parameter :: prtunit=6, m=4000, n=3000, nrhs=100
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of llsq_qr_solve'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: ulp, eps, tol, err, elapsed_time
    real(stnd), allocatable, dimension(:,:) :: x, resid, b, a
!
    integer(i4b) :: krank
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test, min_norm
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : SOLVING LINEAR LEAST SQUARES PROBLEMS WITH A m-BY-n REAL COEFFICIENT
!               MATRIX AND MULTIPLE RIGHT HAND SIDES USING A QR DECOMPOSITION WITH COLUMN
!               PIVOTING OR A COMPLETE ORTHOGONAL DECOMPOSITION OF THE COEFFICIENT MATRIX.
!
!               COMPUTE SOLUTION MATRIX FOR LINEAR LEAST SQUARES SYSTEM
!
!                              a(:m,:n)*x(:n,:nrhs) ≈ b(:m,:nrhs) .
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
!
    err = zero
!
!   SET TOLERANCE FOR DETERMINING THE RANK OF THE MATRIX IN THE SUBROUTINE.
!
!    tol = 0.0000001_stnd
    tol = sqrt( ulp )
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SUBROUTINE.
!
    do_test = true
!
!   DECIDE IF COLUMN PIVOTING MUST BE PERFORMED.
!
    krank = 0
!
!   DECIDE IF THE MINIMUN 2-NORM SOLUTION(S) MUST BE COMPUTED.
!
    min_norm = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), b(m,nrhs), resid(m,nrhs), x(n,nrhs), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a(:m,:n) .
!
    call random_number( a(:m,:n) )
!
!   GENERATE A RANDOM RIGHT HAND-SIDE MATRIX b(:m,:nrhs) .
!
    call random_number( b(:m,:nrhs) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   SOLVE THE LINEAR LEAST SQUARES PROBLEM BY A QR DECOMPOSITION WITH COLUMN PIVOTING
!   USING SUBROUTINE llsq_qr_solve.
!
    call llsq_qr_solve( a(:m,:n), b(:m,:nrhs), x(:n,:nrhs), resid=resid(:m,:nrhs),  &
                        krank=krank, tol=tol, min_norm=min_norm   )
!
!   llsq_qr_solve COMPUTES SOLUTION(S) TO REAL LINEAR LEAST SQUARES PROBLEMS 
!   OF THE FORM:
!
!                       Minimize || b - a*x ||_2
!
!   USING AN ORTHOGONAL FACTORIZATION WITH COLUMNS PIVOTING OF a. a IS A m-ELEMENTS
!   VECTOR OR AN m-BY-n MATRIX WHICH MAY BE RANK-DEFICIENT. IF a IS A MATRIX, m>=n OR
!   n>m IS PERMITTED.
!
!   SEVERAL RIGHT HAND SIDE VECTORS b AND SOLUTION VECTORS x CAN BE
!   HANDLED IN A SINGLE CALL; THEY ARE STORED AS THE COLUMNS OF THE
!   m-BY-nrhs RIGHT HAND SIDE MATRIX b AND THE n-BY-nrhs SOLUTION
!   MATRIX x, RESPECTIVELY. IN OTHER WORDS, THE ARGUMENTS b AND x MAY
!   BE VECTORS OR MATRICES.
!
!   a AND b ARE NOT OVERWRITTEN BY llsq_qr_solve. THE OPTIONAL ARGUMENTS krank,
!   tol AND min_norm MUST NOT BE PRESENT IF a IS A m-ELEMENTS VECTOR.
!
!   ON INPUT, IF THE OPTIONAL PARAMETER krank IS SPECIFIED AND IF krank=k,
!   THIS IMPLIES THAT THE FIRST k COLUMNS OF a ARE TO BE FORCED INTO THE BASIS.
!   PIVOTING IS PERFORMED ON THE LAST n-k COLUMNS OF a.
!   
!   WHEN krank>=min(m,n) IS USED, PIVOTING IS NOT PERFORMED. THIS IS
!   APPROPRIATE WHEN a IS KNOWN TO BE FULL RANK. krank<=0 CAN BE USED
!   WHEN PIVOTING IS DESIRED ON ALL COLUMNS OF a.
!   
!   BY DEFAULT, PIVOTING IS DONE ON ALL THE COLUMNS OF a.
!   
!   ON EXIT, IF THE OPTIONAL PARAMETER krank IS SPECIFIED,
!   krank CONTAINS THE EFFECTIVE RANK OF a. I.E., THE NUMBER
!   OF INDEPENDENT COLUMNS IN MATRIX a.
!                      
!   IF tol IS PRESENT AND IS IN [0,1[, THEN :
!       ON ENTRY, SPECIFIC CALCULATIONS TO DETERMINE THE EFFECTIVE RANK a
!       ARE PERFORMED. tol IS USED TO DETERMINE THE EFFECTIVE RANK
!       OF a, WHICH IS THEN DEFINED AS THE ORDER OF THE LARGEST LEADING
!       TRIANGULAR SUBMATRIX R11 IN THE QR FACTORIZATION WITH PIVOTING OF a,
!       WHOSE ESTIMATED CONDITION NUMBER IN THE 1-NORM IS LESS THAN 1/tol.
!       IF tol=0 IS SPECIFIED THE NUMERICAL RANK OF a IS DETERMINED.
!       ON EXIT, THE RECIPROCAL OF THE CONDITION NUMBER IS RETURNED IN tol.
!
!   IF tol IS NOT SPECIFIED OR IS OUTSIDE [0,1[ :
!       THE CALCULATIONS TO DETERMINE THE EFFECTIVE RANK OF a ARE NOT
!       PERFORMED AND CRUDE TESTS ON R(j,j) ARE DONE TO DETERMINE
!       THE NUMERICAL RANK OF a.
!
!   IF IT IS POSSIBLE THAT a MAY NOT BE FULL RANK (I.E., CERTAIN COLUMNS OF a ARE
!   LINEAR COMBINATIONS OF OTHER COLUMNS), THEN THE LINEARLY DEPENDENT COLUMNS CAN
!   USUALLY BE DETERMINED BY USING krank=0  AND tol=RELATIVE PRECISION OF THE ELEMENTS
!   IN a. IF EACH ELEMENT IS CORRECT TO, SAY, 5 DIGITS THEN tol=0.00001 SHOULD
!   BE USED. ALSO, IT MAY BE HELPFUL TO SCALE THE COLUMNS OF a SO THAT ALL ELEMENTS 
!   ARE ABOUT THE SAME ORDER OF MAGNITUDE.
!   
!   ON EXIT, IF THE OPTIONAL PARAMETER resid IS PRESENT,
!   THE RESIDUAL VECTOR OR MATRIX b - a*x is COMPUTED AND ON EXIT
!
!                               resid = b - a*x .
!
!   THE 2-NORM OF THE RESIDUAL VECTOR(S) FOR THE SOLUTION VECTOR(S) x, CAN ALSO BE COMPUTED
!   DIRECTLY BY SPECIFYING THE OPTIONAL PARAMETER rnorm IN THE CALL OF llsq_qr_solve .
!
!   THE MINIMUN 2-NORM SOLUTION(S) ARE COMPUTED IF THE OPTIONAL LOGICAL PARAMETER min_norm IS
!   PRESENT AND IS SET TO true IN THE CALL TO llsq_qr_solve. OTHERWISE, SOLUTION(S) ARE COMPUTED
!   SUCH THAT IF THE jTH COLUMN OF a IS OMITTED FROM THE BASIS, x[j] OR x[j,:nrhs] IS SET TO ZERO.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE RANGE OF THE COEFFICIENT MATRIX a .
!
        err = maxval( sum( abs( matmul( transpose(resid), a ) ), dim=2 ) )/ sum( abs(a) )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, b, resid, x )
!
!   PRINT THE RESULT OF THE TEST.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*)
    write (prtunit,*) 'Estimated rank of the coefficient matrix = ', krank
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a linear least squares problem with a ', &
      m, ' by ', n,' real coefficient matrix and a ', m, ' by ', nrhs,       &
      ' right hand side matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_llsq_qr_solve
! ================================
!
end program ex2_llsq_qr_solve
