program ex2_partial_qr_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines PARTIAL_QR_CMP
!   and ORTHO_GEN_QR in module QR_Procedures for computing a full QR decomposition
!   with column pivoting of a real matrix.
!
!
! LATEST REVISION : 12/12/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, c50, partial_qr_cmp,   &
                         ortho_gen_qr, norm, unit_matrix, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!    
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT,
! m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX,
! n IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX.
!
    integer(i4b), parameter :: prtunit = 6, m=4000, n=3000, mn=min(m,n)
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of partial_qr_cmp'
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, err1_col, eps, ulp, tol, elapsed_time
    real(stnd), allocatable, dimension(:)   :: diagr, beta, resid2, norma
    real(stnd), allocatable, dimension(:,:) :: a, a2, r, resid
!
    integer(i4b)                            :: krank, l, j, idep
    integer(i4b), allocatable, dimension(:) :: ip
    integer                                 :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test, test_lin
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : COMPUTE A FULL QR DECOMPOSITION WITH COLUMN PIVOTING
!               OF A DATA MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( ulp )
    eps = fudge*ulp
!
    err = zero
    test_lin = true
!
!   SET TOLERANCE FOR DETERMINING THE RANK OF THE MATRIX IN THE SUBROUTINE.
!
!    tol = 0.000001_stnd
    tol = sqrt( ulp )
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SUBROUTINE.
!
    do_test = true
!
    if ( do_test ) then
!
        l = max( m, n )
!
    else
!
        l = n
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,l), diagr(mn), beta(mn), ip(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a(:m,:n) .
!
    call random_number( a(:m,:n) )
!
!   GENERATE A DEFICIENT COEFFICIENT MATRIX a(:m,:n) WITH RANK min(mn-1,n) .
!
    idep = min( n, 5_i4b ) 
    a(:m,idep) = a(:m,1_i4b) + a(:m,n)
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS FOR LATER USE.
!
        allocate( a2(m,n), r(mn,n), resid(m,l), resid2(n), norma(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE DATA MATRIX FOR LATER USE.
!
        resid(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE A FULL QR DECOMPOSITION WITH COLUMN PIVOTING OF RANDOM DATA MATRIX a
!   WITH SUBROUTINE partial_qr_cmp.
!
    call partial_qr_cmp( a(:m,:n), diagr(:mn), beta(:mn), ip(:n), krank, tol=tol )
!
!    call partial_qr_cmp( a(:m,:n), diagr(:mn), beta(:mn), ip(:n), krank )
!
!   partial_qr_cmp COMPUTES A (PARTIAL OR FULL) ORTHOGONAL FACTORIZATION OF A REAL
!   m-BY-n MATRIX. THE MATRIX MAY BE RANK-DEFICIENT.
!
!   HERE THE ROUTINE FIRST COMPUTES A FULL QR FACTORIZATION WITH COLUMN PIVOTING OF a AS:
!
!                     a * P = Q * R = Q * [ R11 R12 ]
!                                         [  0  R22 ]
!
!   P IS A n-BY-n PERMUTATION MATRIX, Q IS A m-BY-m ORTHOGONAL MATRIX AND
!   R IS A m-BY-n UPPER TRIANGULAR OR TRAPEZOIDAL MATRIX.
!   
!   R11 IS DEFINED AS THE LARGEST LEADING SUBMATRIX OF R WHOSE ESTIMATED CONDITION
!   NUMBER  IS LESS THAN 1/tol IF tol IS IN ]0,1[ OR SUCH THAT ABS(R11[j,j])>0 IF
!   tol IS EQUAL TO 0 (SEE BELOW FOR DETAILS).
!   
!   THE ORDER OF R11 IS AN ESTIMATE OF THE RANK OF a AND IS STORED
!   IN THE ARGUMENT krank ON EXIT OF partial_qr_cmp.
!
!   IN OTHER WORDS, krank CONTAINS THE EFFECTIVE RANK OF a. I.E., THE NUMBER
!   OF INDEPENDENT COLUMNS IN MATRIX a.
!                      
!   IF tol IS PRESENT AND IS IN ]0,1[, THEN :
!       CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF THE SUBMATRIX R11
!       ARE PERFORMED. tol IS USED TO DETERMINE THE EFFECTIVE RANK
!       OF R (and a), WHICH IS DEFINED AS THE ORDER OF THE LARGEST LEADING
!       TRIANGULAR SUBMATRIX IN THE QR FACTORIZATION WITH PIVOTING OF a,
!       WHOSE ESTIMATED CONDITION NUMBER (IN THE 1-NORM) IS LESS THAN 1/tol.
!       ON EXIT, THE RECIPROCAL OF THE CONDITION NUMBER IS RETURNED IN tol.
!
!   IF tol IS PRESENT AND IS EQUAL TO 0, THEN :
!       THE NUMERICAL RANK OF R (and a) IS DETERMINED, E.G., CRUDE TESTS ON R(j,j) ARE
!       DONE TO DETERMINE THE RANK OF R AND tol IS NOT CHANGED ON EXIT.
!
!   IF tol IS NOT SPECIFIED OR IS OUTSIDE [0,1[ :
!       THE CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF R ARE NOT
!       PERFORMED AND THE RANK OF R IS ASSUMED TO BE EQUAL TO mn = min(m,n).
!
!   IF IT IS POSSIBLE THAT a MAY NOT BE OF FULL RANK (I.E., CERTAIN COLUMNS OF a ARE
!   LINEAR COMBINATIONS OF OTHER COLUMNS), THEN THE LINEARLY DEPENDENT COLUMNS CAN
!   USUALLY BE DETERMINED BY USING tol=RELATIVE PRECISION OF THE ELEMENTS IN a.
!   IF EACH ELEMENT IS CORRECT TO, SAY, 5 DIGITS THEN tol=0.00001 SHOULD BE USED.
!   
!   IN A SECOND STEP, IF THE OPTIONAL PARAMETER tau IS PRESENT, THEN R22 IS CONSIDERED
!   TO BE NEGLIGIBLE AND THE SUBMATRIX R12 IS ANNIHILATED BY ORTHOGONAL TRANSFORMATIONS
!   FROM THE RIGHT, ARRIVING AT THE COMPLETE ORTHOGONAL FACTORIZATION:
!  
!                     a * P ≈ Q * [ T11 0 ] * Z
!                                 [  0  0 ]
!   
!   WHERE P IS A n-BY-n PERMUTATION MATRIX, Q IS A m-BY-m ORTHOGONAL MATRIX,
!   T11 IS A krank-BY-krank UPPER TRIANGULAR MATRIX AND Z IS A n-BY-n
!   ORTHOGONAL MATRIX. P AND Q ARE ARE ALREADY COMPUTED IN THE QR FACTORIZATION
!   WITH COLUM PIVOTING.
!
!   ON EXIT, a HAS BEEN OVERWRITTEN BY DETAILS OF ITS QR OR COMPLETE 
!   ORTHOGONAL FACTORIZATION, DEPENDING IF ARGUMENT tau IS ABSENT OR PRESENT.
!    
!   IF THE OPTIONAL PARAMETER tau IS ABSENT, partial_qr_cmp COMPUTES ONLY
!   A QR FACTORIZATION WITH COLUMN PIVOTING OF a AND:
!
!   - THE ELEMENTS BELOW AND ON THE DIAGONAL OF THE ARRAY a(:m,:mn) AND THE ARRAY
!     beta(:mn) STORED Q IN FACTORED FORM.
!    
!   - THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY a CONTAIN THE
!     CORRESPONDING ELEMENTS OF THE TRIANGULAR OR TRAPEZOIDAL MATRIX R.
!     THE ELEMENTS OF THE DIAGONAL OF R ARE STORED IN THE ARRAY diagr(:mn). 
!
!   - ip STORES THE PERMUTATION MATRIX P IN THE QR DECOMPOSITION OF a.
!     IF ip(j)=k, THEN THE jTH COLUMN OF a*P WAS THE kTH COLUMN OF a.
!     THE MATRIX P IS REPRESENTED IN THE ARRAY ip AS FOLLOWS:
!     IF ip(j) = i THEN THE jTH COLUMN OF P IS THE iTH CANONICAL UNIT VECTOR.
!                      
!   - krank CONTAINS THE EFFECTIVE RANK OF a (E.G. THE RANK OF R11).
!                      
!   ON THE OTHER HAND, IF THE OPTIONAL PARAMETER tau IS PRESENT, partial_qr_cmp COMPUTES A
!   COMPLETE ORTHOGONAL FACTORIZATION FROM THE QR FACTORIZATION WITH COLUMN PIVOTING OF a
!   AND:
!                      
!   - THE ELEMENTS BELOW AND ON THE DIAGONAL OF THE ARRAY a(:m,:mn) AND THE ARRAY
!     beta(:mn) STORED Q IN FACTORED FORM.
!                      
!   - THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY SECTION a(1:krank,1:krank)
!     CONTAIN THE CORRESPONDING ELEMENTS OF THE TRIANGULAR MATRIX T11. THE ELEMENTS OF
!     THE DIAGONAL OF T11 ARE STORED IN THE ARRAY SECTION diagr(1:krank).
!
!   - ip STORES THE PERMUTATION MATRIX P IN THE COMPLETE DECOMPOSITION OF a.
!     IF ip(j)=k, THEN THE jTH COLUMN OF a*P WAS THE kTH COLUMN OF a.
!     THE MATRIX P IS REPRESENTED IN THE ARRAY ip AS FOLLOWS:
!     IF ip(j) = i THEN THE jTH COLUMN OF P IS THE iTH CANONICAL UNIT VECTOR.
!
!   - THE ORTHOGONAL MATRIX Z IS STORED IN FACTORED FORM IN THE ARRAY SECTIONS
!     a(:krank,krank+1:n) AND tau(:krank).
!                      
!   - krank CONTAINS THE EFFECTIVE RANK OF a, I.E., THE ORDER OF
!     THE SUBMATRIX T11.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   CHECK IF THE DEPENDENT COLUMNS IN MATRIX a HAVE BEEN IDENTIFIED CORRECTLY.
!
    if ( m>=n .and. idep<n ) then
!
        do l = krank+1_i4b, n
!
            j = ip(l)
!
            if ( j==idep .or. j==1_i4b .or. j==n ) exit
!
        end do
!
        test_lin = l/=n+1_i4b
!
    end if
!
    if ( do_test ) then
!
!       RESTORE TRIANGULAR FACTOR R OF QR DECOMPOSITION OF RANDOM DATA MATRIX a
!       IN MATRIX r(:mn,:n) .
!
        do j = 1_i4b, mn
!
            r(1_i4b:j-1_i4b,j) = a(1_i4b:j-1_i4b,j)
            r(j,j)             = diagr(j)
            r(j+1_i4b:mn,j)    = zero
!
        end do
!
        do j = mn+1_i4b, n
!
            r(1_i4b:mn,j) = a(1_i4b:mn,j)
!
        end do
!
!       GENERATE ORTHOGONAL MATRIX Q OF QR DECOMPOSITION OF RANDOM DATA MATRIX a
!       AND AN ORTHOGONAL BASIS FOR THE ORTHOGONAL COMPLEMENT TO THE RANGE OF a .
!       a IS NOT ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( a(:m,:m), beta(:krank) )
!
!       ortho_gen_qr GENERATES AN m-BY-m REAL ORTHOGONAL MATRIX, WHICH IS
!       DEFINED AS A PRODUCT OF k ELEMENTARY REFLECTORS OF ORDER m
!
!            Q = h(1)*h(2)* ... *h(k)
!
!       AS RETURNED BY qr_cmp, qr_cmp2, partial_qr_cmp, partial_rqr_cmp AND partial_rqr_cmp2.
!
!       THE SIZE OF beta DETERMINES THE NUMBER k OF ELEMENTARY REFLECTORS
!       WHOSE PRODUCT DEFINES THE MATRIX Q.
!
!       NOW a(:m,:krank) IS AN ORTHONORMAL BASIS FOR THE RANGE OF a AND a(:m,krank+1:m) IS
!       AN ORTHONORMAL BASIS FOR THE ORTHOGONAL COMPLEMENT TO THE RANGE OF a .
!
!       APPLY PERMUTATION TO a .
!
        do j = 1_i4b, n
!
            a2(:m,j) = resid(:m,ip(j))
!
        end do
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*P - Q(:m,:krank)*R(:krank,:n).
!
        resid(:m,:n) = a2(:m,:n) - matmul( a(:m,:krank), r(:krank,:n) )
        resid2(:n)   = norm( resid(:m,:n), dim=2_i4b )
        norma(:n)    = norm( a2(:m,:n), dim=2_i4b )
!
        err1_col     = maxval( resid2(:n) / norma(:n) )
        err1         = norm( resid2(:n) )/ norm( norma(:n) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q'*Q.
!
        call unit_matrix( resid(:m,:m) )
!
        resid(:m,:m) = abs( resid(:m,:m) - matmul( transpose(a(:m,:m)), a(:m,:m) ) )
        err2 = maxval( resid(:m,:m) )/real(m,stnd)
!
!       CHECK ORTHOGONALITY OF (a*P)(:m,:krank) AND ITS ORTHOGONAL COMPLEMENT Q(:m,krank+1:m).
!
        if ( m>krank ) then
!
            resid(:krank,krank+1:m) = matmul( transpose(a2(:m,:krank)), a(:m,krank+1:m) )
            err3 = maxval( abs( resid(:krank,krank+1:m) ) )/real(m,stnd)
!
        else
!
            err3 = zero
!
        end if
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, r, resid, resid2, norma )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. test_lin ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
!
        write (prtunit,*) 
        write (prtunit,*) 'Estimated rank of the matrix            &
                          &                                        = ', krank
!
        if ( krank/=mn ) then
            write (prtunit,*) 'Indices of linearly dependent columns   &
                              &                                        = ', ip(krank+1:n)
        end if
!        
        write (prtunit,*) 'Accuracy of the QR decomposition        &
                          &||A - Q*R||/||A||                       = ', err1
!        
        write (prtunit,*) 'Accuracy of the QR decomposition        &
                          &max( ||A(:,i) - Q*R(:,i)||/||A(:,i)|| ) = ', err1_col
        write (prtunit,*) 'Orthogonality of the Q matrix           &
                          &                                        = ', err2
!
        if ( m>krank ) then
            write (prtunit,*) 'Orthogonality of the range of the matrix&
                              & and its orthogonal complement          = ', err3
        end if
!
    end if
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing a QR decomposition with column pivoting of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, diagr, beta, ip )
!
!
! END OF PROGRAM ex2_partial_qr_cmp
! =================================
!
end program ex2_partial_qr_cmp
