program ex3_llsq_qr_solve2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine LLSQ_QR_SOLVE2
!   in module LLSQ_Procedures for solving linear least squares problems.
!
!
! LATEST REVISION : 15/11/2023
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, lgl, stnd, zero, true, false, c100, allocate_error, &
                         merror, norm, llsq_qr_solve2
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
! m IS THE NUMBER OF ELEMENTS OF THE RANDOM VECTOR.
! nrhs IS THE NUMBER OF COLUMNS OF THE RIGHT HAND SIDE MATRIX.
!
    integer(i4b), parameter :: prtunit=6, m=1000, nrhs=10
!   
! fudge IS A FUDGING FACTOR FOR THE ACCURACY TESTS.
!   
    real(stnd), parameter  :: fudge=c100
!
    character(len=*), parameter :: name_proc='Example 3 of llsq_qr_solve2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: ulp, eps, err, err1, err2, elapsed_time
    real(stnd), allocatable, dimension(:)   :: a, a2, x, bnorm, rnorm
    real(stnd), allocatable, dimension(:,:) :: b
!
    integer :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test, comp_resid
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 3 : SOLVING LINEAR LEAST SQUARES PROBLEMS WITH A m-ELEMENTS REAL COEFFICIENT
!               VECTOR AND MULTIPLE RIGHT HAND SIDES.
!
!               COMPUTE SOLUTION VECTOR FOR LINEAR LEAST SQUARES SYSTEM
!
!                              a(:m)*x(:nrhs) ≈ b(:m,:nrhs) .
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
!
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SUBROUTINE.
!
    do_test = true
!
!   SPECIFY IF RESIDUALS MUST BE COMPUTED.
!
    comp_resid = do_test
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m), b(m,nrhs), x(nrhs), bnorm(nrhs), rnorm(nrhs), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM COEFFICIENT VECTOR a(:m) .
!
    call random_number( a(:m) )
!
!   GENERATE A RANDOM RIGHT HAND-SIDE MATRIX b(:m,:nrhs) .
!
    call random_number( b(:m,:nrhs) )
!
!   COMPUTE THE NORMS OF THE nrhs DEPENDENT VARIABLES b .
!
    bnorm(:nrhs) = norm( b(:m,:nrhs), dim=2_i4b )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( a2(m), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE DATA VECTOR FOR LATER USE.
!
        a2(:m) = a(:m)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   SOLVE THE LINEAR LEAST SQUARES PROBLEM USING SUBROUTINE llsq_qr_solve2.
!
    call llsq_qr_solve2( a(:m), b(:m,:nrhs), x(:nrhs), rnorm=rnorm(:nrhs), &
                        comp_resid=comp_resid )
!
!   llsq_qr_solve2 COMPUTES SOLUTION(S) TO REAL LINEAR LEAST SQUARES PROBLEMS 
!   OF THE FORM:
!
!                       Minimize || b - a*x ||_2
!
!   USING AN ORTHOGONAL FACTORIZATION WITH COLUMNS PIVOTING OF a. a IS A m-ELEMENTS
!   VECTOR OR AN m-BY-n MATRIX WHICH MAY BE RANK-DEFICIENT. IF a IS A MATRIX, m>=n OR
!   n>m IS PERMITTED. a AND b ARE OVERWRITTEN BY llsq_qr_solve2.
!
!   SEVERAL RIGHT HAND SIDE VECTORS b CAN BE HANDLED IN A SINGLE CALL;
!   THEY ARE STORED AS THE COLUMNS OF THE m-BY-nrhs RIGHT HAND SIDE MATRIX b.
!   IN THAT CASE, x MUST BE A nrhs-ELEMENTS VECTOR.
!
!   THE 2-NORM OF THE RESIDUAL VECTOR(S) FOR THE SOLUTION VECTOR (Or SCALAR) x, CAN ALSO BE COMPUTED
!   DIRECTLY BY SPECIFYING THE OPTIONAL PARAMETER rnorm IN THE CALL OF llsq_qr_solve2 .
!   
!   IF THE OPTIONAL LOGICAL ARGUMENT comp_resid IS PRESENT WITH THE VALUE true,
!   THE RESIDUAL VECTOR OR MATRIX b - a*x is COMPUTED AND OVERWRITES b ON EXIT.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE THE COEFFICIENT VECTOR a .
!
        err1 = maxval( abs( matmul( a2, b ) ) )/ sum( abs(a2) )
!
!       CHECK THE NORMS OF THE COLUMNS OF RESIDUAL MATRIX.
!
        err2 = maxval( abs( norm( b(:m,:nrhs), dim=2_i4b ) - rnorm(:nrhs) ) )
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( a2 )
!
    end if
!
!   PRINT THE RESULT OF THE TESTS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*)
    write (prtunit,*)
    write (prtunit,*) 'Least squares solution of min of ||a(:)*x(:)-b(:,:)||**2 for vector x(:)'
    write (prtunit,*)
    write (prtunit,*) 'Residual sum of squares     ||a*x(i)-b(:,i)||**2               :',rnorm(:nrhs)**2
    write (prtunit,*) 'Residual sum of squares (%) ||a*x(i)-b(:,i)||**2/||b(:,i)||**2 :',(rnorm(:nrhs)/bnorm(:nrhs))**2
    write (prtunit,*)
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a linear least squares problem with a ', &
      m, ' real coefficient vector and a ', m, ' by ', nrhs,       &
      ' right hand side matrix is ', elapsed_time, ' seconds'
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, b, x, bnorm, rnorm )
!
!
! END OF PROGRAM ex3_llsq_qr_solve2
! =================================
!
end program ex3_llsq_qr_solve2
