program test_comp_eof2
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine COMP_EOF2 in module Mul_Stat_Procedures.
!                                                                              
! LATEST REVISION : 01/02/2022
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, false, one, seven, c30, c1_e6, comp_eof2, random_seed_,  &
                         random_number_, gen_random_mat, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, p IS THE NUMBER OF OBSERVATIONS, m THE NUMBER OF VARIABLES
! AND neig THE NUMBER OF COMPUTED EIGENVECTORS.
!
    integer(i4b), parameter :: prtunit=6, m=20, p=50, mp=min(m,p), neig=3
!   
! conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN SOME CASES.
!   
    real(stnd), parameter  :: conda=c1_e6
!
    character(len=*), parameter :: name_proc='comp_eof2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err_mean, err_std, err_eig, err, eps, xn, tmp, tmp2
    real(stnd), allocatable, dimension(:,:) :: x, xt, eigvec, cor1, cov1, res
    real(stnd), allocatable, dimension(:)   :: mean1, mean2, std1, std2, eigval, &
                                               singval0, corp
!
    integer      :: iok
    integer(i4b) :: mat_type, i, k, k2, l, inc, success, maxiter=4
!
    logical(lgl) :: first, last, cov, failure
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon(eps) )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( x(m,p), xt(m,p), eigvec(m,neig), cor1(m,m), cov1(m,m), &
              res(m,neig), mean1(m), mean2(m), std1(m), std2(m),     &
              eigval(m), singval0(mp), corp((m*(m+1))/2), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    success = 0_i4b
    l = 0_i4b
!
    cov = false
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
    do mat_type = 1_i4b, 8_i4b
!
!      GENERATE A m-BY-p RANDOM REAL MATRIX x WITH A SPECIFIED DISTRIBUTION
!      OF SINGULAR VALUES.
!
!      mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!      mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!      mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!      mat_type = 4  -> VERY SLOW DECAY OF SINGULAR VALUES
!      mat_type = 5  -> STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda
!      mat_type = 6  -> GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!      mat_type = 7  -> ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!      mat_type > 7  -> UNIFORM DISTRIBUTION OF SINGULAR VALUES
!
!      GENERATE SINGULAR VALUES.
!
       select case( mat_type )
!
           case( 1_i4b )
!
!              SLOW DECAY OF SINGULAR VALUES.
!
               do i = 1_i4b, mp
!
                   tmp = real( i, stnd )
!
                   singval0(i) = one/( tmp*tmp )
!        
               end do
!
           case( 2_i4b )
!
!               FAST DECAY OF SINGULAR VALUES.
!
                do i = 1_i4b, mp
!
                    tmp = real( i, stnd )
!
                    singval0(i) = exp( -tmp/seven )
!        
                end do
!
            case( 3_i4b )
!
!               S-SHAPED DECAY OF SINGULAR VALUES.
!
                do i = 1_i4b, mp
!
                    tmp = real( i, stnd )
!
                    singval0(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
                end do
!
            case( 4_i4b )
!
!               VERY SLOW DECAY OF SINGULAR VALUES.
!
                tmp = real( mp - 1_i4b, stnd )
!
                do i = 1_i4b, mp
!
                    tmp2 = real( i - 1_i4b, stnd )
!
                    singval0(i) = exp( -tmp2/tmp )
!        
                end do
!
            case( 5_i4b )
!
!               STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda.
!
                singval0(:mp-1_i4b) = one
                singval0(mp) = one/conda
!
            case( 6_i4b )
!
!               GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!
                tmp  = real( mp - 1_i4b, stnd )
!
                do i = 1_i4b, mp
!
                    tmp2 = real( i - 1_i4b, stnd )
!
                    singval0(i) = conda**( -tmp2/tmp )
!        
                end do
!
            case( 7_i4b )
!
!               ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!
                tmp  = real( mp - 1_i4b, stnd )
                tmp2 = one - one/conda
!
                do i = 1_i4b, mp
!
                    singval0(i) = one - (real( i - 1_i4b, stnd )/tmp)*tmp2
!        
                end do
!
            case default
!
!               UNIFORM DISTRIBUTION OF SINGULAR VALUES.
!
                call random_number_( singval0(:mp) )
!
        end select
!
!       GENERATE A m-BY-p MATRIX x OF RANK mp .
!
        call gen_random_mat( singval0(:mp), x )
!
!       COMPUTE THE MEANS AND STANDARD-DEVIATIONS OF x  .
!
        mean1 = sum( x, dim=2_i4b )/p
        xt = x - spread( mean1, dim=2_i4b, ncopies=p )
        std1 = sqrt( sum( xt**2, dim=2_i4b )/p )
!
!       COMPUTE THE COVARIANCE  MATRIX  .
!
        cov1 = matmul( xt, transpose( xt ) )/p
!
!       COMPUTE THE CORRELATION MATRIX .
!
        std2 = one/std1
        cor1 = cov1*spread( std2, dim=2_i4b, ncopies=m )
        cor1 = cor1*spread( std2, dim=1_i4b, ncopies=m )
!
        do i = 1_i4b, 4_i4b
!
            cov = .not.cov
            inc = merge( 1_i4b, p, i<=2_i4b )
!
!           COMPUTE EOFs FROM THE COVARIANCE OR CORRELATION MATRIX.
!
            k2 = 0_i4b
!
            do k = 1_i4b, p, inc
!
                k2 = min( k2+inc, p )
!
                first = k==1_i4b
                last = k2==p
!
!
                call comp_eof2( x(:,k:k2), first, last, eigval, corp, xn, failure,   &
                                cov=cov, maxiter=maxiter, xmean=mean2, xstd=std2, xeigvec=eigvec )
!
            end do
!
            if ( cov ) then
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION cov1*eigvec - eigvec*diag(eigvval)
!               WHERE eigvval ARE THE EIGENVALUES AND eigvec THE EIGENVECTORS OF cov1.
!
                res(:m,:neig) = matmul(cov1,eigvec) - eigvec(:m,:neig)*spread(eigval(:neig),1,m)
!
            else
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION cor1*eigvec - eigvec*diag(eigvval)
!               WHERE eigvval ARE THE EIGENVALUES AND eigvec THE EIGENVECTORS OF cor1.
!
                res(:m,:neig) = matmul(cor1,eigvec) - eigvec(:m,:neig)*spread(eigval(:neig),1,m)
!
            end if
!
            err_eig = sum( abs( res(:m,:neig) ) )/sum( eigval(:neig) )
!
!           CHECK THAT THE TWO SETS OF UNIVARIATE STATISTICS AGREE.
!
            err_mean = maxval( abs( (mean1-mean2)/mean2 ) )
            err_std  = maxval( abs( (std1-std2)/std2 ) )
!
!           PRINT THE RESULTS OF THE TESTS.
!
            l = l + 1_i4b
!
            err = max( err_mean, err_std, err_eig )
!
            if ( err<=eps .and. .not.failure ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',l,' is incorrect'
                write (prtunit,*) ' mat_type = ', mat_type, ', cov = ', cov, ', i = ', i
!
                if ( failure ) then
                    write (prtunit,*) '=> failure  = ', failure
                end if
!
                if ( err_mean>eps ) then
                    write (prtunit,*) '=> err_mean = ', err_mean,' > ', eps
                end if
!
                if ( err_std>eps ) then
                    write (prtunit,*) '=> err_std  =  ', err_std,' > ', eps
                end if
!
                if ( err_eig>eps ) then
                    write (prtunit,*) '=> err_eig  =  ', err_eig,' > ', eps
                end if
!
                write (prtunit,*)
!
            end if
!
        end do
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( x, xt, eigvec, cor1, cov1, res, mean1, mean2,   &
                std1, std2, eigval, singval0, corp )
!
    if ( success/=l ) then
        write (prtunit,*) l - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_comp_eof2
! =============================
!
end program test_comp_eof2
