program test_comp_ortho_rot_eof
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine COMP_EOF, COMP_PC_EOF and COMP_ORTHO_ROT_EOF
!   in module Mul_Stat_Procedures.
!                                                                              
! LATEST REVISION : 04/02/2022
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, true, false, zero, half, one, seven, ten, c30, c1_e6, &
                         comp_eof, comp_pc_eof, comp_ortho_rot_eof, random_seed_,              &
                         random_number_, gen_random_mat, norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, p IS THE NUMBER OF OBSERVATIONS, m THE NUMBER OF VARIABLES
! AND nrot IS THE NUMBER OF EIGENVECTORS TO ROTATE.
!
    integer(i4b), parameter :: prtunit=6, m=20, p=50, mp=min(m,p), nrot=6
!   
! conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN SOME CASES.
!   
    real(stnd), parameter  :: fudge=ten, conda=c1_e6
!
    character(len=*), parameter :: name_proc='comp_ortho_rot_eof'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err_pc, err_rot1, err_rot2, err_rot3, err_rot4, err, eps, xn, w, &
                  delta, tmp, tmp2
    real(stnd), allocatable, dimension(:,:) :: x, x_std, eigvec, factor, rot_factor, rot, &
                                               pc, pc2, rot_pc
    real(stnd), allocatable, dimension(:)   :: mean, std, eigval, singval, singval0, std_rot_factor
!
    integer      :: iok
    integer(i4b) :: mat_type, i, k, l, maxiter, success
!
    logical(lgl) :: first, last, cov, failure, failure2, knorm
!
    character    :: sort = 'd'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*sqrt( epsilon(eps) )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( x(m,p), x_std(m,p), eigvec(m,m), rot(nrot,nrot),         &
              factor(m,nrot), rot_factor(m,nrot), pc(p,m), pc2(p,m),   &
              rot_pc(p,nrot), mean(m), std(m), std_rot_factor(nrot),   &
              eigval(m), singval(m), singval0(mp), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   INITIALIZE COUNTERS FOR THE TESTS.
!
    success = 0_i4b
    l = 0_i4b
!
    first = true
    last  = true
!
!   EIGENVECTORS ARE COMPUTED FROM THE COVARIANCE MATRIX IF cov=true IS USED.
!
    cov = false
!
!   KAISE ROW NORMALIZATION IS PERFORMED IN THE ROTATION IF knorm=true IS USED.
!
    knorm = true
!
!   USER-SPECIFIED CONVERGENCE CRITERION FOR ROTATION.
!
    delta = 0.001_stnd
!
!   MAXIMUM NUMBER OF ITERATIONS FOR ROTATION.
!
    maxiter = 150_i4b
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
    do mat_type = 1_i4b, 8_i4b
!
!      GENERATE A m-BY-p RANDOM REAL MATRIX x WITH A SPECIFIED DISTRIBUTION
!      OF SINGULAR VALUES.
!
!      mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!      mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!      mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!      mat_type = 4  -> VERY SLOW DECAY OF SINGULAR VALUES
!      mat_type = 5  -> STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda
!      mat_type = 6  -> GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!      mat_type = 7  -> ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!      mat_type > 7  -> UNIFORM DISTRIBUTION OF SINGULAR VALUES
!
!      GENERATE SINGULAR VALUES.
!
       select case( mat_type )
!
           case( 1_i4b )
!
!              SLOW DECAY OF SINGULAR VALUES.
!
               do i = 1_i4b, mp
!
                   tmp = real( i, stnd )
!
                   singval0(i) = one/( tmp*tmp )
!        
               end do
!
           case( 2_i4b )
!
!               FAST DECAY OF SINGULAR VALUES.
!
                do i = 1_i4b, mp
!
                    tmp = real( i, stnd )
!
                    singval0(i) = exp( -tmp/seven )
!        
                end do
!
            case( 3_i4b )
!
!               S-SHAPED DECAY OF SINGULAR VALUES.
!
                do i = 1_i4b, mp
!
                    tmp = real( i, stnd )
!
                    singval0(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
                end do
!
            case( 4_i4b )
!
!               VERY SLOW DECAY OF SINGULAR VALUES.
!
                tmp = real( mp - 1_i4b, stnd )
!
                do i = 1_i4b, mp
!
                    tmp2 = real( i - 1_i4b, stnd )
!
                    singval0(i) = exp( -tmp2/tmp )
!        
                end do
!
            case( 5_i4b )
!
!               STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda.
!
                singval0(:mp-1_i4b) = one
                singval0(mp) = one/conda
!
            case( 6_i4b )
!
!               GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!
                tmp  = real( mp - 1_i4b, stnd )
!
                do i = 1_i4b, mp
!
                    tmp2 = real( i - 1_i4b, stnd )
!
                    singval0(i) = conda**( -tmp2/tmp )
!        
                end do
!
            case( 7_i4b )
!
!               ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!
                tmp  = real( mp - 1_i4b, stnd )
                tmp2 = one - one/conda
!
                do i = 1_i4b, mp
!
                    singval0(i) = one - (real( i - 1_i4b, stnd )/tmp)*tmp2
!        
                end do
!
            case default
!
!               UNIFORM DISTRIBUTION OF SINGULAR VALUES.
!
                call random_number_( singval0(:mp) )
!
        end select
!
!       GENERATE A m-BY-p MATRIX x OF RANK mp .
!
        call gen_random_mat( singval0(:mp), x )
!
        do i = 1_i4b, 2_i4b
!
            cov = .not.cov
!
!           COMPUTE EOFs FROM THE COVARIANCE OR CORRELATION MATRIX.
!
            call comp_eof( x(:m,:p), first, last, eigval(:m), eigvec(:m,:m), xn, failure, &
                           cov=cov, sort=sort, xmean=mean(:m), xstd=std(:m) )
!
!           COMPUTE THE PRINCIPAL COMPONENTS FROM THE DATA AND THE EIGENVECTORS.
!
!           IN ORDER TO COMPUTE NORMALIZED PCs, SET singval(:m) = sqrt(eigval(:m)),
!           OTHERWISE SET singval(:) = one . sqrt(eigval(:m)) GIVES THE STANDARD-DEVIATIONS
!           ACCOUNTED FOR BY THE PC time series.
!
            singval(:m) = sqrt( eigval(:m) )
!
!           THE OPTIONAL PARAMETERS xmean AND xstd ARE REQUIRED IF THE EIGENVECTORS
!           HAVE BEEN COMPUTED FROM THE COVARIANCE OR CORRELATION MATRIX.
!
!           ON EXIT OF COMP_PC_EOF :
!
!                  pc(:p,:nrot) CONTAINS THE FIRST nrot PRINCIPAL COMPONENTS STORED COLUMNWISE.
!
            if ( cov ) then
!
                call comp_pc_eof( x(:m,:p), eigvec(:m,:nrot), singval(:nrot), pc(:p,:nrot),   &
                                  xmean=mean(:m) )
!
                x_std(:m,:p) = x(:m,:p) - spread( mean(:m) , dim=2, ncopies=p )
!
            else
!
                call comp_pc_eof( x(:m,:p), eigvec(:m,:nrot), singval(:nrot), pc(:p,:nrot),   &
                                  xmean=mean(:m), xstd=std(:m) )
!
                x_std(:m,:p) = x(:m,:p) - spread( mean(:m) ,      dim=2, ncopies=p )
!
                x_std(:m,:p) = x_std(:m,:p)*spread( one/std(:m) , dim=2, ncopies=p ) 
!
            end if
!
!           COMPUTE THE FIRST nrot CORRESPONDING FACTORS FROM THE EIGENVECTORS AND EIGENVALUES.
!
            factor(:m,:nrot) = eigvec(:m,:nrot)*spread( singval(:nrot), dim=1, ncopies=m )
!
!           RESCALE THE ORIGINAL PC TIME SERIES FOR THE TESTS.
!
            pc2(:p,:nrot) = pc(:p,:nrot)*spread( singval(:nrot), dim=1, ncopies=p )
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION x_std**(t)*eigvec - pc
!           WHERE eigvec ARE THE EIGENVECTORS, x_std THE STANDARDIZED DATA
!           AND pc THE UNNORMALIZED PRINCIPAL COMPONENTS.
!
            err_pc = sum( abs( matmul(transpose(x_std),eigvec(:m,:nrot)) - pc2(:p,:nrot) ) )
!
!           NOW ROTATE THE FIRST nrot FACTORS WITH SUBROUTINE comp_ortho_rot_eof.
!
            do k = 1_i4b, 3_i4b
!
                if ( k==1_i4b ) then
!
!                   USE w=0 FOR QUARTIMAX METHOD.
!
                    w = zero
!
                else if ( k==2_i4b ) then
!
!                   USE w=1 FOR VARIMAX METHOD.
!
                    w = one
!
                else
!
!                   USE w=nrot/2 FOR EQUAMAX METHOD.
!
                    w = real( nrot, stnd)*half
!
                end if
!
!               KAISER ROW NORMALIZATION IS PERFORMED IF knorm=true IS USED.
!
                knorm = .not.knorm
!
                call comp_ortho_rot_eof( factor(:m,:nrot), rot_factor(:m,:nrot), rot(:nrot,:nrot), &
                                         std_rot_factor(:nrot), failure2, knorm=knorm, w=w,        &
                                         maxiter=maxiter, delta=delta )
!
!               NOW ROTATE THE FIRST nrot PC TIME SERIES.
!
                rot_pc(:p,:nrot) = matmul( pc(:p,:nrot), rot(:nrot,:nrot) )
!
!               CHECK THAT THE VARIANCE EXPLAINED BY THE ORIGINAL AND ROTATED EOFS IS THE SAME.
!
                err_rot1 = abs( sum( std_rot_factor(:nrot)**2 ) - sum( eigval(:nrot) ) )
!
!               CHECK ORTHOGONALITY OF THE ROTATION MATRIX rot COMPUTED BY comp_ortho_rot_eof
!               SUBROUTINE.
!
                err_rot2 = abs( sum( abs( matmul(transpose(rot),rot) ) ) - real( nrot, stnd ) )
!
!               CHECK THAT THE NORMS OF THE ROTATED PCs ARE UNCHANGED.
!
                err_rot3 = maxval( abs( norm( rot_pc(:p,:nrot), dim=2 ) - sqrt( real(p,stnd) ) ) )
!
!               CHECK COMPUTATION OF VARIANCES EXPLAINED BY THE ROTATED PCs.
!
                err_rot4 = maxval( abs( std_rot_factor(:nrot) -    &
                           norm( rot(:nrot,:nrot)*spread(singval(:nrot),dim=2,ncopies=nrot), dim=2 ) ) )
!
!               PRINT THE RESULTS OF THE TESTS.
!
                l = l + 1_i4b
!
                err = max( err_pc, err_rot1, err_rot2, err_rot3, err_rot4 )
!
                if ( err<=eps .and. .not.failure ) then
!                if ( err<=eps .and. .not.failure .and. .not.failure2 ) then
!
                    success = success + 1_i4b
!
                else
!
                    write (prtunit,*) 'Test ',l,' is incorrect'
                    write (prtunit,*) ' mat_type = ', mat_type, ', cov = ', cov, ', k = ', k
!
                    if ( failure ) then
                        write (prtunit,*) '=> failure  = ', failure
                    end if
!
                    if ( failure2 ) then
                        write (prtunit,*) '=> failure2 = ', failure2
                    end if
!
                    if ( err_pc>eps ) then
                        write (prtunit,*) '=> err_pc   = ', err_pc,' > ', eps
                    end if
!
                    if ( err_rot1>eps ) then
                        write (prtunit,*) '=> err_rot1 =  ', err_rot1,' > ', eps
                    end if
!
                    if ( err_rot2>eps ) then
                        write (prtunit,*) '=> err_rot2 =  ', err_rot2,' > ', eps
                    end if
!
                    if ( err_rot3>eps ) then
                        write (prtunit,*) '=> err_rot3 =  ', err_rot3,' > ', eps
                    end if
!
                    if ( err_rot4>eps ) then
                        write (prtunit,*) '=> err_rot4 =  ', err_rot4,' > ', eps
                    end if
!
                    write (prtunit,*)
!
                end if
!
            end do
!
        end do
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( x, x_std, eigvec, rot, factor, rot_factor, pc, pc2, rot_pc,   &
                mean, std, std_rot_factor, eigval, singval, singval0  )
!
    if ( success/=l ) then
        write (prtunit,*) l - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_comp_ortho_rot_eof
! ======================================
!
end program test_comp_ortho_rot_eof
