program test_svd_cmp4
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine SVD_CMP4
!   in module SVD_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking
!    the accuracy of subroutine SVD_CMP4.
!
!    The program assumes that subroutine BD_INVITER2 in module SVD_Procedures is correct.
!
! LATEST REVISION : 21/06/2022
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def, true, false,                    &
                         safmin, zero, half, one, two, five, seven, ten, c30, c100, c4000,       &
                         qr_cmp, ortho_gen_qr, eigval_sort, svd_cmp4, bd_inviter2, norm, arth,   &
                         ifirstloc
!   
#ifdef _MATMUL
    use Utilities,       only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Utilities,       only : transpose=>transpose2
#endif
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, maxiter IS NUMBER OF INVERSE ITERATIONS FOR COMPUTING SINGULAR VECTORS.
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, nsing_def=5, maxiter=2
!   
! def_conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN SOME CASES.
!   
    real(stnd), parameter  :: fudge=c4000, def_conda=50000._stnd
!
    character(len=*), parameter :: name_proc='svd_cmp4'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, ulp, eps_sing, &
                                               shift, thr, tol, tmp, norma, conda,        &
                                               max_sing_err, max_sing_rel_err
    real(stnd), allocatable, dimension(:)   :: eigval, singval, singval2, d, e, diagr, beta
    real(stnd), allocatable, dimension(:,:) :: a, a2, leftvec, rightvec, leftvec2, rightvec2, &
                                               rightvect, res, p
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, n, nr, m, o, nsing, nsing_old, success, notdone
!
    logical(lgl) :: gen_p, reortho, failure, failure2, failure_bd, bisect, ortho, initvec, details
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
!
!   SPECIFY A TOLERANCE FOR SELECTING THE SINGULAR TRIPLETS IN THE TESTS.
!
    tol = 0.001_stnd
!
!   SPECIFY A POSITIVE SHIFT FOR SINGULAR VALUES, EVENTUALLY.
!
    shift = zero
!
!   SPECIFY IF REORTHOGONALIZATION IS PERFORMED IN
!   THE ONE-SIDED RHALA-BARLOW ALGORITHM.
!
    reortho = true
!
!   SPECIFY IF THE RIGHT ORTHOGONAL MATRIX IS GENERATED IN
!   THE ONE-SIDED RHALA-BARLOW ALGORITHM.
!
    gen_p = true
!
!   SPECIFY IF INVERSE ITERATIONS ARE STARTED WITH RANDOM OR FERNANDO VECTORS.
!
    initvec = true
!
!   SPECIFY IF SINGULAR VALUES ARE COMPUTED ACCURATELY BY BISECTION.
!
    bisect = false
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
        gen_p = .not.gen_p
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), singval(n), singval2(n), d(n), e(n),           &
                  diagr(n), beta(n), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(singval2) )  deallocate( singval2 )
            if ( allocated(d) )         deallocate( d )
            if ( allocated(e) )         deallocate( e )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
!
            cycle
!
        end if
!
        do k = 1_i4b, 3_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: SVD OF a(:nr,:n) WITH nr=n .
!
                    nr = n
!
                case( 2_i4b )
!
!                   TEST 2: SVD OF a(:nr,:n) WITH nr>=n AND nr NOT MUCH LARGER THAN n .
!
                    nr = n + 1_i4b
!
                case( 3_i4b )
!
!                   TEST 3: SVD OF a(:nr,:n) WITH nr>=n AND nr MUCH LARGER THAN n .
!
                    nr = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(nr,n), a2(nr,n), res(nr,n),            &
                      leftvec(nr,n),   rightvec(n,n), p(n,n),  &
                      leftvec2(nr,n), rightvec2(n,n),          &
                      rightvect(n,n), stat=iok  )
!
            if ( iok/=0 ) then
!
                if ( allocated(a) )         deallocate( a )
                if ( allocated(a2) )        deallocate( a2 )
                if ( allocated(res) )       deallocate( res )
                if ( allocated(leftvec) )   deallocate( leftvec )
                if ( allocated(rightvec) )  deallocate( rightvec )
                if ( allocated(leftvec2) )  deallocate( leftvec2 )
                if ( allocated(rightvec2) ) deallocate( rightvec2 )
                if ( allocated(rightvect) ) deallocate( rightvect )
                if ( allocated(p) )         deallocate( p )
!
                cycle
!
            end if
!
!           GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:nr,:n) .
!
            call random_number( leftvec(:nr,:n) )
!
            leftvec(:nr,:n) = leftvec(:nr,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                leftvec(:nr,:n) = leftvec(:nr,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( leftvec(:nr,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( leftvec(:nr,:n), beta(:n) )
!
!           GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:n,:n) .
!
            call random_number( rightvec(:n,:n) )
!
            rightvec(:n,:n) = rightvec(:n,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                rightvec(:n,:n) = rightvec(:n,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( rightvec(:n,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( rightvec(:n,:n), beta(:n) )
!
            do j = 1_i4b, 25_i4b
!
                select case( j )
!
                    case( 1_i4b )
!
!                       TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = ulp
!
                    case( 2_i4b )
!
!                       TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = one
                        eigval(n)             = ulp
!
                    case( 3_i4b )
!
!                       TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                        eigval(n)             = one
!
                    case( 4_i4b )
!
!                       TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                        eigval(1_i4b)         = ulp
                        eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                        eigval(n)             = two
!
                    case( 5_i4b )
!
!                       TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                    case( 6_i4b )
!
!                       TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                        err = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                        end do
!
                    case( 7_i4b )
!
!                       TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                        eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                    case( 8_i4b )
!
!                       TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                        eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                    case( 9_i4b )
!
!                       TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                    case( 10_i4b )
!
!                       TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                    case( 11_i4b )
!
!                       TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                        call random_number( eigval(:n) )
!
                        eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
                    case( 12_i4b )
!
!                       TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                    case( 13_i4b )
!
!                       TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 14_i4b )
!
!                       TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 15_i4b )
!
!                       TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 16_i4b )
!
!                       TEST 16: CLUSTER OF EIGENVALUES AT 1 WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        eigval(1_i4b)   = one/conda
                        eigval(2_i4b:n) = one
!
                    case( 17_i4b )
!
!                       TEST 17: CLUSTER OF EIGENVALUES AT 1/conda WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = one/conda
!
                    case( 18_i4b )
!
!                       TEST 18: ARITHMETIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        tmp = real( n-1_i4b, stnd )
                        err1 = one - one/conda
!
                        do l = 1_i4b, n
                            eigval(l) = one - ( real(l-1_i4b,stnd)/tmp )*err1
                        end do
!
                    case( 19_i4b )
!
!                       TEST 19: GEOMETRIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        tmp = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = conda**( -real(l-1_i4b,stnd)/tmp )
                        end do
!
                    case( 20_i4b )
!
!                       TEST 20: RANDOM DISTRIBUTION OF EIGENVALUES IN THE INTERVAL 1/conda TO 1
!                                WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        call random_number( beta(:n) )
!
                        tmp = log( one/conda )

                        beta(:n) = tmp + beta(:n)*abs( tmp )
                        eigval(:n)    = exp( beta(:n) )
!
                    case( 21_i4b )
!
!                       TEST 21: VERY SLOW DECAY OF EIGENVALUES.
!
                        conda = real( n - 1_i4b, stnd )
!
                        do l = 1_i4b, n
!
                            tmp = real( l-1_i4b, stnd )
!
                            eigval(l) = exp( -tmp/conda )
!        
                        end do
!
                    case( 22_i4b )
!
!                       TEST 22: SLOW DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = one/( tmp*tmp )
!        
                        end do
!
                    case( 23_i4b )
!
!                       TEST 23: FAST DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = exp( -tmp/seven )
!        
                        end do
!
                    case( 24_i4b )
!
!                       TEST 24: S-SHAPE DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
                        end do
!
                    case default
!
!                       DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                        call random_number( eigval(:n) )
!
                        call random_number( tmp )
!
                        if ( tmp>safmin ) then
                             eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                        end if
!
                end select
!
#ifdef _F2003
                if ( ieee_support_datatype( eigval ) ) then
                    if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
                end if
#endif
!
!               COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
                singval(:n) = sqrt( eigval(:n) )
!
!               SORT THE SINGULAR VALUES.
!
                call eigval_sort( sort, singval(:n) )
!
!               APPLY POSITIVE SHIFT FOR SINGULAR VALUES, EVENTUALLY.
!
                singval(:n) = singval(:n) + shift
!
                where ( singval(:n)==zero )
!
                    eigval(:n) = one
!
                elsewhere
!
                    eigval(:n) = singval(:n)
!
                end where
!
!               FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
                rightvect(:n,:n) = transpose( rightvec(:n,:n) )
!
                do i = 1_i4b, n
!
                    rightvect(:n,i) = rightvect(:n,i)*singval(:n)
!
                end do
!
                a(:nr,:n) = matmul( leftvec(:nr,:n), rightvect(:n,:n) )
!
!               COMPUTE THE NORM OF THE MATRIX.
!
!                norma = norm( a(:nr,:n) )
                norma = norm( singval(:n) )
!
!               SAVE THE ORIGINAL MATRIX.
!
                a2(:nr,:n) = a(:nr,:n)
!
!               REDUCE THE MATRIX a TO BIDIAGONAL FORM AND
!               COMPUTE ALL SINGULAR VALUES OF THE MATRIX a TO HIGH ACCURACY.
!
                reortho = .not.reortho
!
                call svd_cmp4( a2, singval2, failure=failure, v=p, sort=sort, d=d, e=e,     &
                               bisect=bisect, sing_vec=false, gen_p=gen_p, reortho=reortho, &  
                               failure_bd=failure_bd )
!
!                if ( failure ) cycle
!
!               COMPUTE MAXIMUM OF ABSOLUTE ERRORS OF SINGULAR VALUES.
!
                max_sing_err = maxval( abs( singval2(:n)-singval(:n) ) )
!
!               COMPUTE MAXIMUM OF RELATIVE ERRORS OF SINGULAR VALUES.
!
                max_sing_rel_err =  maxval( abs( ( singval2(:n)-singval(:n) )/eigval(:n) ) )
!
!               DETERMINE THE NUMBER OF SINGULAR VECTORS TO COMPUTE FOR CHECKING ACCURACY
!               OF bd_inviter2 SUBROUTINE.
!
                thr = max( tol*singval2(1_i4b), safmin )
!
                nsing_old = ifirstloc( logical( singval2(:n)<=thr, lgl ) ) - 1_i4b
!
                eps_sing = real( n, stnd)*eps*singval(1_i4b)
!
!               COMPUTE SINGULAR VECTORS OF THE MATRIX.
!
                do o = 1_i4b, 2_i4b
!
                    if ( o==1_i4b ) then
                        nsing = min( nsing_old, nsing_def )
                    else
                        nsing = nsing_old
                    end if
!
                    ortho = true
!
                    do i = 1_i4b, 2_i4b
!
!                       COMPUTE SINGULAR VECTORS OF THE MATRIX BY AN INVERSE ITERATION METHOD.
!
!                       ALWAYS USE ORTHOGONALIZATION FOR CASE 24, WHICH IS DIFFICULT FOR INVERSE ITERATION.
!
                        if ( j/=24_i4b ) then
                            ortho = .not.ortho
                        end if
!
!                        ALWAYS USE ORTHOGONALIZATION FOR CASES 6, 20 AND 24 WHEN n IS LARGE,
!                        WHICH ARE DIFFICULT FOR INVERSE ITERATION (IN SINGLE PRECISION).
!
!                        if ( (j/=24_i4b .and. j/=20_i4b .and. j/=6_i4b) .or. n<=100_i4b ) then
!                            ortho = .not.ortho
!                        end if
!
                        if ( ortho ) then
!
                            if ( initvec ) then
!
                                call bd_inviter2( a2(:nr,:n), p(:n,:n), d(:n), e(:n), singval2(:nsing), &
                                                  leftvec2(:nr,:nsing), rightvec2(:n,:nsing), failure2, &
                                                  maxiter=maxiter, ortho=true                           )
!
                            else
!
                                call bd_inviter2( a2(:nr,:n), p(:n,:n), d(:n), e(:n), singval2(:nsing), &
                                                  leftvec2(:nr,:nsing), rightvec2(:n,:nsing), failure2, &
                                                  maxiter=maxiter, ortho=true, initvec=false            )
!
                            end if
!
                        else
!
                            if ( initvec ) then
!
                                call bd_inviter2( a2(:nr,:n), p(:n,:n), d(:n), e(:n), singval2(:nsing), &
                                                  leftvec2(:nr,:nsing), rightvec2(:n,:nsing), failure2, &
                                                  maxiter=maxiter                                       )
!
                            else
!
                                call bd_inviter2( a2(:nr,:n), p(:n,:n), d(:n), e(:n), singval2(:nsing), &
                                                  leftvec2(:nr,:nsing), rightvec2(:n,:nsing), failure2, &
                                                  maxiter=maxiter, initvec=false                        )
!
                            end if
!
                        end if
!
                        m = m + 1_i4b
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                       WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                       v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                        res(:nr,:nsing) = matmul( a(:nr,:n), rightvec2(:n,:nsing) )         &
                                         - leftvec2(:nr,:nsing)*spread(singval2(:nsing),1,nr)
!
                        if ( norma/=zero ) then
                            err1 = maxval( norm(res(:nr,:nsing), dim=2_i4b) )/( norma*real(nr,stnd) )
                        else
                            err1 = maxval( norm(res(:nr,:nsing), dim=2_i4b) )/real(nr,stnd)
                        end if
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                       WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX.
!
                        res(:nsing,:nsing) = matmul( transpose( leftvec2(:nr,:nsing) ), leftvec2(:nr,:nsing) )
!
                        do l = 1_i4b, nsing
                            res(l,l) = one - res(l,l)
                        end do
!
                        err2 = maxval( norm(res(:nsing,:nsing), dim=2_i4b) )/real(nr,stnd)
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                       WHERE v ARE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                        res(:nsing,:nsing) = matmul( transpose( rightvec2(:n,:nsing) ), rightvec2(:n,:nsing) )
!
                        do l = 1_i4b, nsing
                            res(l,l) = one - res(l,l)
                        end do
!
                        err3 = maxval( norm(res(:nsing,:nsing), dim=2_i4b) )/real(n,stnd)
!
                        err = max( err1, err2, err3 )
!
                        if ( err<=eps .and. max_sing_err<=eps_sing ) then
!
                            success = success + 1_i4b
!
                            if ( details ) then
!
                                write (prtunit,*) 'Test ',m,' is correct'
!
                                if ( ortho ) then
!
                                    write (prtunit,*) 'nr= ', nr,' , nc= ', n,' , case= ', j,             &
                                                      ' , nsing_old= ', nsing_old,' , nsing= ', nsing,    &
                                                      ' , reortho= ', reortho,' , ortho= ', true,         &
                                                      ' , failure= ', failure, ' , failure2= ', failure2, &
                                                      ' , failure_bd= ', failure_bd
!
                                else
!
                                    write (prtunit,*) 'nr= ', nr,' , nc= ', n,' , case= ', j,             &
                                                      ' , nsing_old= ', nsing_old,' , nsing= ', nsing,    &
                                                      ' , reortho= ', reortho,                            &
                                                      ' , failure= ', failure, ' , failure2= ', failure2, &
                                                      ' , failure_bd= ', failure_bd
!
                                end if
!
!                               COMPUTE ABSOLUE ERRORS OF SINGULAR VALUES.
!
                                write (prtunit,*) 'Maximum absolute error of the computed singular values = ', &
                                                  max_sing_err
!
!                               COMPUTE RELATIVE ERRORS OF SINGULAR VALUES.
!
                                write (prtunit,*) 'Maximum relative error of the computed singular values = ', &
                                                  max_sing_rel_err 
!
!                               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                               WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                               v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                                write (prtunit,*) 'Accuracy of the computed singular triplets             = ', &
                                                  err1/ulp, ' (must be less than :', fudge, ' )'
!
!                               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                               WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX.
!
                                write (prtunit,*) 'Orthogonality of the computed left singular vectors    = ', &
                                                  err2/ulp, ' (must be less than :', fudge, ' )'
!
!                               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                               WHERE v are RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                                write (prtunit,*) 'Orthogonality of the computed right singular vectors   = ', &
                                                  err3/ulp, ' (must be less than :', fudge, ' )'
!
                                write (prtunit,*)
!
                            end if
!
                        else if ( failure2 ) then
!
                            notdone = notdone + 1_i4b
!
                        else
!
                            write (prtunit,*) 'Test ',m,' is incorrect'
!
                            if ( ortho ) then
!
                                write (prtunit,*) '=> ',err,' > ',eps,' or ', max_sing_err,' > ', eps_sing, &
                                                  ' , nr= ', nr,' , nc= ', n,' , case= ', j,                &
                                                  ' , nsing_old= ', nsing_old,' , nsing= ', nsing,          &
                                                  ' , reortho= ', reortho,' , ortho= ', true,               &
                                                  ' , failure= ', failure, ' , failure2= ', failure2,       &
                                                  ' , failure_bd= ', failure_bd

!
                            else
!
                                write (prtunit,*) '=> ',err,' > ',eps,' or ', max_sing_err,' > ', eps_sing,  &
                                                  ' , nr= ', nr,' , nc= ', n,' , case= ', j,                 &
                                                  ' , nsing_old= ', nsing_old,' , nsing= ', nsing,           &
                                                  ' , reortho= ', reortho,                                   &
                                                  ' , failure= ', failure, ' , failure2= ', failure2,        &
                                                  ' , failure_bd= ', failure_bd
!
                            end if
!
!                           COMPUTE ABSOLUE ERRORS OF SINGULAR VALUES.
!
                            write (prtunit,*) 'Maximum absolute error of the computed singular values = ', &
                                              max_sing_err
!
!                           COMPUTE RELATIVE ERRORS OF SINGULAR VALUES.
!
                            write (prtunit,*) 'Maximum relative error of the computed singular values = ', &
                                              max_sing_rel_err 
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                           WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                           v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                            write (prtunit,*) 'Accuracy of the computed singular triplets             = ', &
                                              err1/ulp, ' (must be less than :', fudge, ' )'
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                           WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX.
!
                            write (prtunit,*) 'Orthogonality of the computed left singular vectors    = ', &
                                              err2/ulp, ' (must be less than :', fudge, ' )'
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                           WHERE v are RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                            write (prtunit,*) 'Orthogonality of the computed right singular vectors   = ', &
                                              err3/ulp, ' (must be less than :', fudge, ' )'
!
                            write (prtunit,*)
!
                        end if
!
                    end do
!
                end do
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, a2, leftvec, rightvec, leftvec2, rightvec2, rightvect, res, p )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, singval, singval2, d, e, diagr, beta )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    i = success + notdone
!
    if ( m/=i ) then
        write (prtunit,*) m - i, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_svd_cmp4
! ============================
!
end program test_svd_cmp4
