module Random
! ########################################################################
!
! Copyright 2024 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! Authors: Pascal Terray (LOCEAN/IPSL, Paris, France)
!
! ########################################################################
!                                                                        *
! ************************************************************************
! THIS MODULE REPLACES THE FORTRAN 90 INTRINSICS                         *
! random_number AND random_seed BY SEVERAL IMPLEMENTATIONS               *
! OF THE KISS (Keep It Simple Stupid), L'Ecuyer's LFSR113,               *
! MERSENNE TWISTER MT19937 AND MEMT19937-II RANDOM                       *
! NUMBER GENERATORS.                                                     *
!                                                                        *
! IN ADDITION TO 10 DIFFERENT UNIFORM RANDOM GENERATORS, GAUSSIAN        *
! RANDOM GENERATORS, SHUFFLING AND SAMPLING SUBROUTINES ARE ALSO         *
! PROVIDED, AS WELL AS SUBROUTINES FOR GENERATING PSEUDO-RANDOM          *
! ORTHOGONAL MATRICES FOLLOWING THE HAAR DISTRIBUTION OVER THE GROUP OF  *
! ORTHOGONAL MATRICES, PSEUDO_RANDOM SYMMETRIC MATRICES WITH A           *
! PRESCRIBED SPECTRUM OR PSEUDO_RANDOM MATRICES WITH A PRESCRIBED        *
! SINGULAR VALUE DISTRIBUTION.                                           *
! FINALLY, RANDOMIZED LINEAR ALGEBRA ROUTINES LIKE RANDOMIZED QB, QR,    *
! COLUMN INTERPOLATIVE, TWO_SIDED INTERPOLATIVE AND CUR DECOMPOSITIONS   *
! ARE ALSO INCLUDED.                                                     *
!                                                                        *
! MANY PARTS OF THIS MODULE ARE ADAPTED FROM :                           *
!                                                                        *
! Hennecke, M., 1995: A Fortran90 interface to random                    *
!                    number generation.                                  *
!                    Computer Physics Communications.                    *
!                    Volume 90, Number 1, 117-120                        *
!                                                                        *
! LATEST REVISION : 12/03/2024                                           *
!                                                                        *
! ########################################################################
!                                                                        *
! ************************************************************************
!                                                                        *
! THE C PROCESSOR MACROS USED IN THIS MODULE ARE:                        *
!                                                                        *
!  _OPENMP           FOR ACTIVATING OPENMP PARALLELIZATION               *
!  _ALLOC            FOR ALLOCATING LOCAL VARIABLES INSTEAD OF PLACING   *
!                    THEM ON THE STACK IN SOME SUBROUTINES AND FUNCTIONS *
!  _BLAS             FOR USING BLAS WHEN POSSIBLE                        *
!  _MATMUL           FOR REPLACING THE matmul INTRINSIC FUNCTION WITH    *
!                    STATPACK matmul2 FUNCTION                           *
!  _TRANSPOSE        FOR REPLACING THE transpose INTRINSIC FUNCTION WITH *
!                    STATPACK transpose2 FUNCTION                        *
!  _RANDOM_WITH0     FOR GENERATING REAL FLOATING POINT NUMBERS          *
!                    IN THE [0,1[ INTERVAL INSTEAD OF ]0,1[ INTERVAL     *
!  _RANDOM_NOINT32   FOR SIGNALING THAT 32 BIT INTEGERS ARE NOT          *
!                    AVAILABLE                                           *
!  _RANDOM_NOUNIX    FOR SIGNALING THAT THE OPERATING SYSTEM IS NOT UNIX *
!  _RANDOM_GFORTRAN  FOR SIGNALING THAT THE INTEGER Unix FUNCTION getpid *
!                    IS CONSIDERED AS AN INTRINSIC RATHER AS AN EXTERNAL *
!                    PROCEDURE AS FOR THE gfortran COMPILER              *
!  _RANDOM_NAGWARE   FOR SIGNALING THAT THE INTEGER Unix FUNCTION getpid *
!                    IS LOCATED IN THE f90_unix_env MODULE FOR THE NAG   *
!                    FORTRAN COMPILER                                    *
!  _USE_GNU          FOR SIGNALING THAT THE GNU GFORTRAN COMPILER IS     *
!                    USED. HERE, THIS INCLUDES ONLY THE ACTIVATION OF    *
!                    THE _RANDOM_GFORTRAN MACRO                          *
!  _USE_NAGWARE      FOR SIGNALING THAT THE NAG FORTRAN COMPILER IS      *
!                    USED. HERE, THIS INCLUDES ONLY THE ACTIVATION OF    *
!                    THE _RANDOM_NAGWARE MACRO AND THE DESACTIVATION     *
!                    OF THE CHECK OF ASSOCIATED STATUT OF POINTERS       *
!                    IN SOME SUBROUTINES                                *
!  _USE_PGI          FOR DESACTIVATING SOME OPENMP CONSTRUCTS FOR THE    *
!                    PORTLAND PGFORTRAN COMPILER                         *
!                                                                        *
! ************************************************************************
!
#ifdef _USE_GNU
#define _RANDOM_GFORTRAN
#endif
!
#ifdef _USE_NAGWARE
#define _RANDOM_NAGWARE
#endif
!
!
! USED MODULES
! ============
!
    use Select_Parameters,  only : lgl, i4b, stnd, extd, blksz_qr, urandom_file
    use Utilities,          only : merror, arth
    use Reals_Constants,    only : zero, half, one, two, four, twopi
    use Num_Constants,      only : machulp
    use Logical_Constants
#ifndef _BLAS
#ifdef _MATMUL
    use Utilities,          only : matmul=>matmul2
#endif
#endif
#ifdef _TRANSPOSE
    use Utilities,          only : transpose=>transpose2
#endif
#ifdef _OPENMP
    use Select_Parameters,  only : omp_limit
    use omp_lib,            only : omp_get_max_threads, omp_in_parallel, omp_get_max_active_levels
#endif
#ifdef _RANDOM_NAGWARE
    use f90_unix_env,       only : getpid
#endif
#ifdef _BLAS
    use BLAS_interfaces,    only : dot, axpy, scal, swap, trsv, trsm, gemm
#endif
!
! SELECT THE ALGORITHM FOR COMPUTING EUCLIDEAN OR FROBENIUS NORMS IN THE MODULE.   
!
!    use Utilities,       only : norm
!    use Utilities,       only : norm=>norme
    use Utilities,       only : norm=>norm2e
!    use Utilities,       only : norm=>abse
!
! OTHER MODULES USED IN SPECIFIC SUBROUTINES
! ==========================================
!
!   use Num_Constants
!   use Char_Constants
!   use Prob_Procedures
!   use Hous_Procedures
!   use Lin_Procedures
!   use QR_Procedures
!   use BLAS_interfaces
!
! STRONG TYPING IMPOSED
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::                                                            &
              rand_number,    rand_integer32,    rand_integer31,         &
              random_number_, random_integer32_, random_integer31_,      &
              random_seed_,   init_mt19937,      init_memt19937,         &
              normal_rand_number,  normal_random_number_,                &
              normal_rand_number2, normal_random_number2_,               &
              normal_rand_number3, normal_random_number3_,               &
              random_qr_cmp, ortho_gen_random_qr,                        &
              partial_rqr_cmp, partial_rqr_cmp2, partial_rtqr_cmp,       &
              partial_rqr_cmp_fixed_precision,                           &
              rqb_cmp, rqb_cmp_fixed_precision,                          &
              id_cmp, ts_id_cmp, cur_cmp,                                &
              gen_random_sym_mat, gen_random_mat, gen_random_ortho_mat,  &
              simple_shuffle, drawsample, drawbootsample
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface random_number_
        module procedure     random_r0,                        &
                             random_r1,                        &
                             random_r2,                        &
                             random_r3,                        &
                             random_r4,                        &
                             random_r5,                        &
                             random_r6,                        &
                             random_r7
    end interface
!
    interface random_integer32_
        module procedure     random_i32_0,                        &
                             random_i32_1,                        &
                             random_i32_2,                        &
                             random_i32_3,                        &
                             random_i32_4,                        &
                             random_i32_5,                        &
                             random_i32_6,                        &
                             random_i32_7
    end interface
!
    interface random_integer31_
        module procedure     random_i31_0,                        &
                             random_i31_1,                        &
                             random_i31_2,                        &
                             random_i31_3,                        &
                             random_i31_4,                        &
                             random_i31_5,                        &
                             random_i31_6,                        &
                             random_i31_7
    end interface
!
    interface init_mt19937
        module procedure     init_mt19937_r0,                     &
                             init_mt19937_r1
    end interface
!
    interface init_memt19937
        module procedure     init_memt19937_r0,                   &
                             init_memt19937_r1
    end interface
!
    interface normal_random_number_
        module procedure     normal_random_r0,                    &
                             normal_random_r1,                    &
                             normal_random_r2
    end interface
!
    interface normal_random_number2_
        module procedure     normal_random2_r0,                   &
                             normal_random2_r1,                   &
                             normal_random2_r2
    end interface
!
    interface normal_random_number3_
        module procedure     normal_random3_r0,                   &
                             normal_random3_r1,                   &
                             normal_random3_r2
    end interface
!
    interface simple_shuffle
        module procedure     simple_shuffle_rv, simple_shuffle_cv, simple_shuffle_iv
    end interface
!
!
! MODULE PARAMETERS
! =================
!
!   ki_sel IS THE KIND PARAMETER FOR 32-BIT INTEGER.
!
    integer, parameter :: ki0        = kind(0)
    integer, parameter :: ki9        = selected_int_kind(9)
    integer, parameter :: ki_sel     = max(ki9,sign(ki0,-ki9))
!
    integer, parameter :: ratio_i    = (int(bit_size(0_ki_sel)) + bit_size(0) - 1)/bit_size(0)
!
!   bit_size(0_ki_sel) MUST BE EQUAL TO fullbitsize IF 32_BIT INTEGERS ARE AVAILABLE.
!
    integer(ki_sel), parameter :: fullbitsize = 32_ki_sel
!
!   THE CONSTANTS fbs, hbs, qbs AND tbs ARE USED FOR EMULATING UNSIGNED 32-BIT INTEGER ARITHMETIC
!   WITHOUT OVERFLOWS IN INTERNAL ROUTINES. 
!
    integer, parameter :: fbs = int(fullbitsize), hbs = fbs/2, qbs = hbs/2, tbs = 3*qbs, topbit = fbs - 1
!
!   SPECIFY HERE THE LOCATION OF THE URANDOM DEVICE ON YOUR SYSTEM IF IT EXISTS.
!   FOR UNIX SYSTEM, NORMALLY IT IS /dev/urandom .
!
!    character(len=*),  parameter :: urandom_file='/dev/urandom'
!
#ifdef _RANDOM_WITH0
!  THE CONSTANT m_ran_32 IS 2^32.
!  THE CONSTANT m_ran_invm32 IS THE RECIPROCAL OF 2^32.
!
    real(stnd), parameter :: m_ran_32     = 4294967296.0_stnd,        &
                             m_ran_invm32 = one/m_ran_32
#else
!  THE CONSTANT m_ran_32 IS THE REAL NUMBER FOLLOWING 2^32.
!  THE CONSTANT m_ran_invm32 IS THE RECIPROCAL OF THIS REAL NUMBER.
!
    real(stnd), parameter :: m_ran_32     = 4294967296.0_stnd*(one + machulp ),        &
                             m_ran_invm32 = one/m_ran_32
#endif
!
!   num_bits_stnd IS THE NUMBER OF BITS IN THE MANTISSA OF REAL NUMBERS OF PRECISION stnd. WE
!   ASSUME THAT THE MANTISSA IS NORMALIZED AS IN THE IEEE STANDARD.
!
    integer,  parameter :: num_bits_stnd = min(digits(half),64)-1,       &
                           num_shift=max(min(num_bits_stnd-64,-1),-topbit)
!
!   THE FOLLOWING INTEGER PARAMETERS ARE USED BY THE 32-BIT XORSHIFT RNG WHICH IS ONE OF THE
!   COMPONENT RNGS IN THE KISS RANDOM NUMBER GENERATORS.
!
    integer,  parameter ::     kiss_shift1   = 13,  &
                               kiss_shift2   = -17, &
                               kiss_shift3   = 5
!
!   THE FOLLOWING INTEGER PARAMETERS ARE USED BY THE MERSENNE TWISTER RANDOM NUMBER GENERATORS.
!
!   PERIOD PARAMETERS FOR THE MERSENNE TWISTER MT19937 AND MEMT19937-II RNGS.
!
    integer(ki_sel),  parameter ::  mt_n    = 624_ki_sel,   &
                                    mt_m    = 397_ki_sel
!
#ifdef _RANDOM_NOINT32
    integer(ki_sel),  parameter ::  mt_upper_mask  = ishft( 1_ki_sel, topbit),                &
                                    mt_lower_mask  = ibits( not( mt_upper_mask ), 0, fbs ),   &
                                    mt_matrix_a    = ior( 419999967_ki_sel, mt_upper_mask ),  &
                                    mt_mag01(0:1)  = (/ 0_ki_sel, mt_matrix_a /)
#else
    integer(ki_sel),  parameter ::  mt_upper_mask  = ishft( 1_ki_sel, topbit),                &
                                    mt_lower_mask  = not( mt_upper_mask ),                    &
                                    mt_matrix_a    = ior( 419999967_ki_sel, mt_upper_mask ),  &
                                    mt_mag01(0:1)  = (/ 0_ki_sel, mt_matrix_a /)
#endif
!
!   THESE PARAMETERS MAY BE ALSO SPECIFIED AS FOLLOW ASSUMING 32-BIT INTEGER AND INTEGER REPRESENTATION
!   IS TWO'S COMPLEMENT. HOWEVER, THE INITIALIZATION OF mt_upper_mask TO THE NEGATIVE 32-BIT INTEGER
!   OF THE LARGEST AMPLITUDE AT COMPILATION TIME MAY CAUSE PROBLEMS ON SOME SYSTEMS (E.G. SOME VERSIONS
!   OF GFORTRAN FOR EXAMPLE). THIS IS DUE TO THE FACT THAT THE COMPILER WILL TRY TO FORM
!   THE POSITIVE INTEGER 2147483648 (WHICH IS NOT REPRESENTABLE IF TWO'S COMPLEMENT NOTATION
!   IS USED), BEFORE NEGATING IT. THIS WILL RESULT IN INTEGER OVERFLOW, WHICH IS USUALLY
!   CHECKED AT COMPILATION TIME FOR CONSTANT PARAMETERS.
!
!    integer(ki_sel),  parameter ::     mt_upper_mask  = -2147483648_ki_sel,        &
!                                       mt_lower_mask  =  2147483647_ki_sel,        &
!                                       mt_matrix_a    = -1727483681_ki_sel,        &
!                                       mt_mag01(0:1)  = (/ 0_ki_sel, mt_matrix_a /)
!
!   THESE PARAMETERS MAY BE ALSO SPECIFIED AS FOLLOW, WITH HEXADECIMAL CONSTANTS. HOWEVER, THESE HEXADECIMAL
!   CONSTANTS MAY CAUSE PROBLEMS ON SOME COMPILERS (E.G. GFORTRAN) BECAUSE THE "NEGATIVE" 32-BIT INTEGERS ARE
!   CONSIDERED AS UNSIGNED 32-BIT INTEGERS INSTEAD OF SIGNED 32-BIT INTEGERS DURING COMPILATION. THIS WILL
!   RESULT IN INTEGER OVERFLOW, WHICH IS USUALLY CHECKED AT COMPILATION TIME FOR CONSTANT PARAMETERS.
!
!    integer(ki_sel),  parameter ::     mt_upper_mask  = int( z'80000000', ki_sel),      &
!                                       mt_lower_mask  = int( z'7fffffff', ki_sel),      &
!                                       mt_matrix_a    = int( z'9908b0df', ki_sel),      &
!                                       mt_mag01(0:1)  = (/ 0_ki_sel, mt_matrix_a /)
!
!   DEFAULT SEEDS FOR THE MERSENNE TWISTER MT19937 AND MEMT19937-II RNGS.
!
    integer(ki_sel),  parameter :: mt_default_seed = 5489_ki_sel
!    integer(ki_sel),  parameter :: mt_default_seed = 21641_ki_sel
!
    integer(ki_sel),  dimension(4), parameter ::       &
           mt_default_seed_array = (/ 291_ki_sel, 564_ki_sel, 837_ki_sel, 1110_ki_sel /)
!
!   TEMPERING PARAMETERS FOR THE MERSENNE TWISTER MT19937 RNG.
!
    integer,  parameter ::     mt_t1_shift = 7,   &
                               mt_t2_shift = 15,  &
                               mt_shift0   = -11, &
                               mt_shift1   = -18
!
    integer(ki_sel),  parameter ::  mt_t1_mask     = ior( 489444992_ki_sel , mt_upper_mask),  &
                                    mt_t2_mask     = ior( 1875247104_ki_sel, mt_upper_mask)
!
!   THESE PARAMETERS MAY BE ALSO SPECIFIED AS FOLLOW ASSUMING 32-BIT INTEGER AND INTEGER REPRESENTATION
!   IS TWO'S COMPLEMENT.
!
!    integer(ki_sel),  parameter ::     mt_t1_mask     = -1658038656_ki_sel,      &
!                                       mt_t2_mask     = -272236544_ki_sel
!
!   THESE PARAMETERS MAY BE ALSO SPECIFIED AS FOLLOW, WITH HEXADECIMAL CONSTANTS. HOWEVER, THESE HEXADECIMAL
!   CONSTANTS MAY CAUSE PROBLEMS ON SOME COMPILERS (E.G. GFORTRAN) BECAUSE THE "NEGATIVE" 32-BIT INTEGERS ARE
!   CONSIDERED AS UNSIGNED 32-BIT INTEGERS INSTEAD OF SIGNED 32-BIT INTEGERS DURING COMPILATION.
!
!    integer(ki_sel),  parameter ::     mt_t1_mask     = int( z'9d2c5680', ki_sel),      &
!                                       mt_t2_mask     = int( z'efc60000', ki_sel)
!
!   TEMPERING PARAMETERS FOR THE MERSENNE TWISTER MEMT19937-II RNG.
!
    integer,  parameter ::     memt_lag1     = 151, &
                               memt_lag2     = 36,  &
                               memt_lag1over = 473, &
                               memt_lag2over = 588, &
                               memt_shift1   = 8,   &
                               memt_shift2   = 14
!
    integer(ki_sel), parameter ::  memt_mask1 = ior( 840548011_ki_sel, mt_upper_mask ),  &
                                   memt_mask2 = 1455285546_ki_sel
!
!   THESE PARAMETERS MAY BE ALSO SPECIFIED AS FOLLOW ASSUMING 32-BIT INTEGER AND INTEGER REPRESENTATION
!   IS TWO'S COMPLEMENT.
!
!    integer(ki_sel), parameter ::  memt_mask1 = -1306935637_ki_sel,  &
!                                   memt_mask2 = 1455285546_ki_sel
!
!   THESE PARAMETERS MAY BE ALSO SPECIFIED AS FOLLOW, WITH HEXADECIMAL CONSTANTS. HOWEVER, THESE HEXADECIMAL
!   CONSTANTS MAY CAUSE PROBLEMS ON SOME COMPILERS (E.G. GFORTRAN) BECAUSE THE "NEGATIVE" 32-BIT INTEGERS ARE
!   CONSIDERED AS UNSIGNED 32-BIT INTEGERS INSTEAD OF SIGNED 32-BIT INTEGERS DURING COMPILATION.
!
!    integer(ki_sel),  parameter ::     memt_mask1     = int( z'b219beab', ki_sel),      &
!                                       memt_mask2     = int( z'56bde52a', ki_sel)
!
#ifdef _RANDOM_NOINT32
!   THE FOLLOWING PARAMETER IS USED TO SET THE LAST 32 BITS OF INTEGER OF KIND ki_sel TO 1,
!   IF 32 BIT-INTEGERS ARE NOT AVAILABLE.
!
    integer(ki_sel),  parameter ::  noint32_mask  = not( ior( mt_lower_mask, mt_upper_mask ) )
!
#endif
!
!   BLOCK SIZE FOR APPLYING HOUSEHOLDER TRANSFORMATIONS
!   IN RANDOMIZED QR, ID AND CUR SUBROUTINES.
!
    integer(i4b), parameter :: blksz=blksz_qr
!
!
! MODULE VARIABLES
! ================
!
!   first     = true MEANS THAT THE random_seed_ SUBROUTINE HAS NEVER BEEN CALLED BEFORE.
!   test_kiss = true MEANS THAT THE KISS RNGS HAVE NOT BEEN YET TESTED FOR OVERFLOWS.
!
    logical(lgl), save :: first = true, test_kiss = true
!
!   THE PRIVATE INTEGER VARIABLE method SETS THE RANDOM NUMBER GENERATOR USED FOR SUBSEQUENT
!   CALLS OF THE rand_number, rand_integer32, rand_integer31, random_number_, random_integer32_,
!   random_integer31_ ROUTINES:
!
!     method = 1  : THE Marsaglia's KISS RNG IS USED.
!     method = 2  : THE FAST Marsaglia's KISS RNG, WHICH USES ONLY
!                   ADD, SHIFT, EXCLUSIVE-OR AND "AND" OPERATIONS,
!                   IS USED (SAME RESULTS AS THE C VERSION).
!     method = 3  : THE L'Ecuyer's LFSR113 RNG IS USED.
!     method = 4  : THE MERSENNE TWISTER MT19937 RNG IS USED.
!     method = 5  : THE MERSENNE TWISTER MEMT19937-II RNG IS USED.
!     method = 6  : THE EXTENDED VERSION OF THE Marsaglia's KISS RNG IS USED.
!     method = 7  : THE EXTENDED VERSION OF THE FAST Marsaglia's KISS RNG IS USED.
!     method = 8  : THE EXTENDED VERSION OF THE L'Ecuyer's LFSR113 RNG IS USED.
!     method = 9  : THE EXTENDED VERSION OF THE MERSENNE TWISTER MT19937 RNG IS USED.
!     method = 10 : THE EXTENDED VERSION OF THE MERSENNE TWISTER MEMT19937-II RNG IS USED.
!
!   IN ORDER TO CHANGE THE RNG, YOU MUST USE A CALL TO THE random_seed_ SUBROUTINE WITH
!   THE OPTIONAL alg ARGUMENT SET TO 1, 2, 3, 4, 5, 6, 7, 8, 9 OR 10. NOTE THAT FOR RANDOM INTEGER
!   GENERATION, METHODS 6, 7, 8, 9 AND 10 ARE EXACTLY IDENTICAL TO METHODS 1, 2, 3, 4 AND 5,
!   RESPECTIVELY. THEY DIFFER ONLY FOR RANDOM REAL GENERATION.
!
!   THE PRIVATE INTEGER VARIABLE method STORED THE METHOD CURRENTLY IN USE, num_seeds
!   STORED THE NUMBER OF SEEDS USED BY THIS METHOD AND seed_size IS THE CORRESPONDING SIZE
!   OF THE SEED VECTOR AS RETURNED BY A CALL TO random_seed_ SUBROUTINE WITH ARGUMENT size.
!
!   THE DEFAULT METHOD CAN BE SAFELY CHANGED HERE, BETWEEN 1, 2 AND 3.
!   NOTE THAT THE OTHER METHODS CANNOT BE USED SAFELY AS THE DEFAULT METHOD.
!   FURTHERMORE IF 32-BIT INTEGERS ARE NOT AVAILABLE THE DEFAULT METHOD MUST
!   BE 3, BECAUSE THE KISS RNGS WORK PROPERLY ONLY WITH 32-BIT INTEGERS.
!
!    integer,  save :: method  =  1
!    integer,  save :: method  =  2
    integer,  save :: method  =  3
!
!   num_seeds IS THE SIZE OF THE STATE VECTOR OF THE CURRENT RNG.
!   seed_size IS THE SIZE OF THE SEED ARRAY OF THE RANDOM GENERATOR AS
!   RETURNED BY SUBROUTINE random_seed_. NOTE THAT THE seed_size IS NOT
!   NECESSARILY EQUAL TO num_seeds DEPENDING ON THE RNG AND IF 32-BIT INTEGERS
!   ARE OR NOT THE DEFAULT INTEGERS ON YOUR SYSTEM.
!
    integer,  save :: num_seeds, seed_size
!
!   THE NEXT VARIABLES STORE THE STATE VARIABLES FOR EACH RANDOM NUMBER GENERATORS.
!
!   PRIVATE INTEGER STATE VARIABLES USED BY THE Marsaglia's KISS RNG (method = 1 or 6).
!    
    integer(ki_sel), save  :: x = 123456789_ki_sel, y = 362436069_ki_sel,             &
                              z = 521288629_ki_sel, w = 916191069_ki_sel
!
!   PRIVATE INTEGER STATE VARIABLES USED BY THE FAST Marsaglia's KISS RNG, WHICH USES ONLY
!   ADD, SHIFT, EXCLUSIVE-OR AND "AND" OPERATIONS (method = 2 or 7).
!    
    integer(ki_sel), save  :: x2 = 123456789_ki_sel, y2 = 362436069_ki_sel,            &
                              z2 = 21288629_ki_sel,  w2 = 14921776_ki_sel, c=0_ki_sel
!
!   PRIVATE INTEGER STATE VARIABLES USED BY THE L'Ecuyer's LFSR113 RNG (method = 3 or 8).
!    
    integer(ki_sel), save  :: s1 = 153587801_ki_sel,  s2 = -759022222_ki_sel,        &
                              s3 = 1288503317_ki_sel, s4 = -1718083407_ki_sel
!
!   PRIVATE INTEGER STATE VARIABLES USED BY THE THE MERSENNE TWISTER MT19937 RNG (method = 4 or 9).
!   mt_initialized = false MEANS THAT THE STATE OF MT19937 IS NOT INITIALIZED.
!   mti = mt_n             MEANS THAT THE ARRAY mt(:) IS NOT INITIALIZED.
!   THE ARRAY mt(:) IS THE STATE VECTOR OF THE MERSENNE TWISTER MT19937 RNG.
!    
    logical(lgl),    save  :: mt_initialized = false
    integer(ki_sel), save  :: mti = mt_n, mt(0_ki_sel:mt_n-1_ki_sel) = 0_ki_sel
!
!   PRIVATE INTEGER STATE VARIABLES USED BY THE THE MERSENNE TWISTER MEMT19937-II RNG (method = 5 or 10).
!   memt_initialized = false MEANS THAT THE STATE OF MEMT19937-II IS NOT INITIALIZED.
!   memti = -1_ki_sel        MEANS THAT THE ARRAY memt(:) IS NOT INITIALIZED.
!   THE ARRAY memt(:) IS THE STATE VECTOR OF THE MERSENNE TWISTER MEMT19937-II RNG.
!    
    logical(lgl),    save  :: memt_initialized = false
    integer(ki_sel), save  :: memti = -1_ki_sel, memt(0_ki_sel:mt_n-1_ki_sel) = 0_ki_sel
!
!   THE FOLLOWING PRIVATE VARIABLE m_ran_invmbits IS USED TO TRANSFORM SIGNED 32-BIT INTEGERS
!   INTO REAL FLOATING POINT NUMBERS FOR THE EXTENDED VERSIONS OF THE RNGS.
!
    real(stnd), save :: m_ran_invmbits = -one
!
#ifndef _RANDOM_NOUNIX
!    
! INTEGER Unix FUNCTION USED FOR INITIALIZING THE SEEDS OF THE RNGS ON UNIX SYSTEMS.
!
#ifdef _RANDOM_GFORTRAN
    intrinsic getpid
#else
!
#ifndef _RANDOM_NAGWARE
!
    integer :: getpid
!
    external getpid
#endif
!
#endif
!
#endif
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
! =========================================================================================
!                            RANDOM FLOATING POINT NUMBER SUBROUTINES
! =========================================================================================
!
    function rand_number() result( harvest )
!
! purpose
! _______
!
!   This function returns a uniformly distributed random number between 0 and 1,
!   exclusive of the two endpoints 0 and 1.
!
!
! Arguments
! _________
!
!   None
!
!
! Further Details
! _______________
!
!   If the CPP macro _RANDOM_WITH0 is used during compilation,
!   this routine may return 0 value.
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _RANDOM_NOINT32
    use Char_Constants,   only : random_error7, random_error9
#else
    use Char_Constants,   only : random_error1, random_error2,    &
                                 random_error7, random_error8
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(ki_sel)               :: k, j, tmp
    integer(ki_sel), dimension(2) :: tmpvec
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='rand_number'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( first ) then
#ifdef _RANDOM_NOINT32
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method/=3 ) then
            call merror( name_proc//random_error9 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ibclr(-1_ki_sel, bit_size(tmp)-1_ki_sel)
!
        if ( radix(tmp)/=2 .or. tmp/=huge(tmp) ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
#else
!
!       TEST IF 32-BIT INTEGERS ARE AVAILABLE.
!
        if ( bit_size( tmp )/=fullbitsize ) then
            call merror( name_proc//random_error1 )
        endif
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method<1 .or. method>3 ) then
            call merror( name_proc//random_error2 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ishftc( -8_ki_sel, -3)
!
        if ( radix(tmp)/=2 .or. tmp/=536870911_ki_sel ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
        if ( method/=3 ) then
!
!           TEST IF INTEGER OVERFLOWS ARE SAFE FOR THE MARSAGLIA'S KISS RNGS.
!
            tmp   = huge( tmp )
!
            if ( integer_not_safe( tmp ) ) then
!
                call merror( name_proc//random_error8 )
!
            endif
!
        endif
#endif
!
    endif
!
!$OMP CRITICAL (ran_num_)
    select case (method)
!
        case (1)
!
!           Marsaglia's KISS RNG.
!
            x = 69069_ki_sel * x + 1327217885_ki_sel
            y = ieor( y, ishft(y, kiss_shift1) )
            y = ieor( y, ishft(y, kiss_shift2) )
            y = ieor( y, ishft(y, kiss_shift3) )
!            y = m( m( m( y, kiss_shift1), kiss_shift2), kiss_shift3)
            z = 18000_ki_sel * iand( z, 65535_ki_sel) + ishft( z, -16)
            w = 30903_ki_sel * iand( w, 65535_ki_sel) + ishft( w, -16)
!
            tmp = x + y + ishft( z, 16) + w
!
        case (2)
!
!           FAST Marsaglia's KISS RNG WHICH USES ONLY ADD, SHIFT, EXCLUSIVE-OR AND "AND" OPERATIONS.
!
            x2 = x2 + 545925293_ki_sel
!            x2 = uiadd( x2, 545925293_ki_sel )
            y2 = ieor( y2, ishft(y2, kiss_shift1) )
            y2 = ieor( y2, ishft(y2, kiss_shift2) )
            y2 = ieor( y2, ishft(y2, kiss_shift3) )
!            y2 = m( m( m(y2, kiss_shift1), kiss_shift2), kiss_shift3)
            tmp = z2 + w2 + c
            z2 = w2
            c  = ishft( tmp, -topbit)
            w2 = iand( tmp, 2147483647_ki_sel)
!
            tmp = x2 + y2 + w2
!
        case (3)
!
!           L'Ecuyer's LFSR113 RNG.
!
#ifdef _RANDOM_NOINT32
            tmp = ieor( ishft(s1,6), s1)
            tmp = ishft( ibits( tmp, 0, fbs ), -13)
            tmp = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
            s1  = ibits( tmp, 0, fbs )
!
            tmp = ieor( ishft(s2,2), s2)
            tmp = ishft( ibits( tmp, 0, fbs ), -27)
            tmp = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
            s2  = ibits( tmp, 0, fbs )
!
            tmp = ieor( ishft(s3,13), s3)
            tmp = ishft( ibits( tmp, 0, fbs ), -21)
            tmp = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
            s3  = ibits( tmp, 0, fbs )
!
            tmp = ieor( ishft(s4,3), s4)
            tmp = ishft( ibits( tmp, 0, fbs ), -12)
            tmp = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
            s4  = ibits( tmp, 0, fbs )
#else
            tmp  = ishft( ieor( ishft(s1,6), s1), -13)
            s1 = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
!
            tmp  = ishft( ieor( ishft(s2,2), s2), -27)
            s2 = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
!
            tmp  = ishft( ieor( ishft(s3,13), s3), -21)
            s3 = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
!
            tmp  = ishft( ieor( ishft(s4,3), s4), -12)
            s4 = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
#endif
!
            tmp = ieor( ieor( ieor(s1,s2), s3), s4)
!
        case (4)
!
!           MERSENNE TWISTER MT19937 RNG.
!
            if ( mti>=mt_n ) then
!
!               GENERATE mt_n WORDS AT ONE TIME.
!
                do j = 0_ki_sel, mt_n - mt_m - 1_ki_sel
!
                    tmp   = ior( iand(mt(j),mt_upper_mask),            &
                                 iand(mt(j+1_ki_sel),mt_lower_mask)    )
!
                    mt(j) = ieor( ieor(mt(j+mt_m),ishft(tmp,-1)),      &
                                  mt_mag01(iand(tmp,1_ki_sel))         )
!
                end do
!
                do j = mt_n - mt_m, mt_n - 2_ki_sel
!
                    tmp   = ior( iand(mt(j),mt_upper_mask),            &
                                 iand(mt(j+1_ki_sel),mt_lower_mask)    )
!
                    mt(j) = ieor( ieor(mt(j+(mt_m-mt_n)), ishft(tmp,-1)), &
                                  mt_mag01(iand(tmp,1_ki_sel))            )
!
                end do
!
                tmp   = ior( iand(mt(j),mt_upper_mask),         &
                             iand(mt(0_ki_sel),mt_lower_mask)   )
!
                mt(j) = ieor( ieor(mt(mt_m-1_ki_sel),ishft(tmp,-1)),   &
                              mt_mag01(iand(tmp,1_ki_sel))             )
!
                mti = 0_ki_sel
!
            endif
!
            tmp = mt(mti)
            mti = mti + 1_ki_sel
!
!           TEMPERING PHASE.
!
            tmp = ieor( tmp, ishft(tmp,mt_shift0) )
            tmp = ieor( tmp, iand(ishft(tmp,mt_t1_shift),mt_t1_mask) )
            tmp = ieor( tmp, iand(ishft(tmp,mt_t2_shift),mt_t2_mask) )
            tmp = ieor( tmp, ishft(tmp,mt_shift1) )
!
        case (5)
!
!           MERSENNE TWISTER MT19937-II RNG.
!
            select case (memti)
!
                case (0_ki_sel:mt_n-mt_m-1_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+mt_m),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))          )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (mt_n-mt_m:memt_lag1over-1_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (memt_lag1over:memt_lag2over-1_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (memt_lag2over:mt_n-2_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (mt_n-1_ki_sel)
!
                    tmp = ior( iand(memt(mt_n-1_ki_sel),mt_upper_mask),  &
                               iand(memt(0_ki_sel),mt_lower_mask)        )
!
                    memt(mt_n-1_ki_sel) = ieor( ieor(memt(mt_m-1_ki_sel),ishft(tmp,-1)),   &
                                                mt_mag01(iand(tmp,1_ki_sel))               )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                    memti = 0_ki_sel
!
            end select
!
#ifdef _RANDOM_NOINT32
            tmp = ibits( tmp, 0, fbs )
#endif
!
        case (6)
!
!           EXTENDED PRECISION VERSION OF Marsaglia's KISS RNG.
!
            do k = 1_ki_sel, 2_ki_sel
!
                x = 69069_ki_sel * x + 1327217885_ki_sel
                y = ieor( y, ishft(y, kiss_shift1) )
                y = ieor( y, ishft(y, kiss_shift2) )
                y = ieor( y, ishft(y, kiss_shift3) )
!                y = m( m( m( y, kiss_shift1), kiss_shift2), kiss_shift3)
                z = 18000_ki_sel * iand( z, 65535_ki_sel) + ishft( z, -16)
                w = 30903_ki_sel * iand( w, 65535_ki_sel) + ishft( w, -16)
!
                tmpvec(k) = x + y + ishft( z, 16) + w
!
            end do
!
        case (7)
!
!           EXTENDED PRECISION VERSION OF THE FAST Marsaglia's KISS RNG,
!           WHICH USES ONLY ADD, SHIFT, EXCLUSIVE-OR AND "AND" OPERATIONS.
!
            do k = 1_ki_sel, 2_ki_sel
!
                x2 = x2 + 545925293_ki_sel
!                x2 = uiadd( x2, 545925293_ki_sel )
                y2 = ieor( y2, ishft(y2, kiss_shift1) )
                y2 = ieor( y2, ishft(y2, kiss_shift2) )
                y2 = ieor( y2, ishft(y2, kiss_shift3) )
!                y2 = m( m( m(y2, kiss_shift1), kiss_shift2), kiss_shift3)
                tmp = z2 + w2 + c
                z2 = w2
                c  = ishft( tmp, -topbit)
                w2 = iand( tmp, 2147483647_ki_sel)
!
                tmpvec(k) = x2 + y2 + w2
!
            end do
!
        case (8)
!
!           EXTENDED PRECISION VERSION OF L'Ecuyer's LFSR113 RNG.
!
            do k = 1_ki_sel, 2_ki_sel
!
#ifdef _RANDOM_NOINT32
                tmp = ieor( ishft(s1,6), s1)
                tmp = ishft( ibits( tmp, 0, fbs ), -13)
                tmp = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
                s1  = ibits( tmp, 0, fbs )
!
                tmp = ieor( ishft(s2,2), s2)
                tmp = ishft( ibits( tmp, 0, fbs ), -27)
                tmp = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
                s2  = ibits( tmp, 0, fbs )
!
                tmp = ieor( ishft(s3,13), s3)
                tmp = ishft( ibits( tmp, 0, fbs ), -21)
                tmp = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
                s3  = ibits( tmp, 0, fbs )
!
                tmp = ieor( ishft(s4,3), s4)
                tmp = ishft( ibits( tmp, 0, fbs ), -12)
                tmp = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
                s4  = ibits( tmp, 0, fbs )
#else
                tmp = ishft( ieor( ishft(s1,6), s1), -13)
                s1  = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
!
                tmp = ishft( ieor( ishft(s2,2), s2), -27)
                s2  = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
!
                tmp = ishft( ieor( ishft(s3,13), s3), -21)
                s3  = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
!
                tmp = ishft( ieor( ishft(s4,3), s4), -12)
                s4  = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
#endif
!
                tmpvec(k) = ieor( ieor( ieor(s1,s2), s3), s4)
!
            end do
!
        case (9)
!
!           EXTENDED PRECISION VERSION OF THE MERSENNE TWISTER MT19937 RNG.
!
            do k = 1_ki_sel, 2_ki_sel
!
                if ( mti>=mt_n ) then
!
!                   GENERATE mt_n WORDS AT ONE TIME.
!
                    do j = 0_ki_sel, mt_n - mt_m - 1_ki_sel
!
                        tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                     iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                        mt(j) = ieor( ieor(mt(j+mt_m),ishft(tmp,-1)),   &
                                      mt_mag01(iand(tmp,1_ki_sel))      )
!
                    end do
!
                    do j = mt_n - mt_m, mt_n - 2_ki_sel
!
                        tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                     iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                        mt(j) = ieor( ieor(mt(j+(mt_m-mt_n)), ishft(tmp,-1)),  &
                                      mt_mag01(iand(tmp,1_ki_sel))             )
!
                    end do
!
                    tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                 iand(mt(0_ki_sel),mt_lower_mask)   )
!
                    mt(j) = ieor( ieor(mt(mt_m-1_ki_sel),ishft(tmp,-1)),  &
                                  mt_mag01(iand(tmp,1_ki_sel))            )
!
                    mti = 0_ki_sel
!
                endif
!
                tmp = mt(mti)
                mti = mti + 1_ki_sel
!
!               TEMPERING PHASE.
!
                tmp = ieor( tmp, ishft(tmp,mt_shift0) )
                tmp = ieor( tmp, iand(ishft(tmp,mt_t1_shift),mt_t1_mask) )
                tmp = ieor( tmp, iand(ishft(tmp,mt_t2_shift),mt_t2_mask) )
!
                tmpvec(k) = ieor( tmp, ishft(tmp,mt_shift1) )
!
            end do
!
        case (10)
!
!           EXTENDED PRECISION VERSION OF THE MERSENNE TWISTER MEMT19937-II RNG.
!
            do k = 1_ki_sel, 2_ki_sel
!
              select case (memti)
!
                case (0_ki_sel:mt_n-mt_m-1_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+mt_m),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))          )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmpvec(k) = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (mt_n-mt_m:memt_lag1over-1_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmpvec(k) = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (memt_lag1over:memt_lag2over-1_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmpvec(k) = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (memt_lag2over:mt_n-2_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmpvec(k) = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (mt_n-1_ki_sel)
!
                    tmp = ior( iand(memt(mt_n-1_ki_sel),mt_upper_mask),  &
                               iand(memt(0_ki_sel),mt_lower_mask)        )
!
                    memt(mt_n-1_ki_sel) = ieor( ieor(memt(mt_m-1_ki_sel),ishft(tmp,-1)),   &
                                                mt_mag01(iand(tmp,1_ki_sel))               )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmpvec(k) = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                    memti = 0_ki_sel
!
              end select
!
#ifdef _RANDOM_NOINT32
              tmpvec(k) = ibits( tmpvec(k), 0, fbs )
#endif
!
            end do
!
    end select
!$OMP END CRITICAL (ran_num_)
!
    if ( method<=5 ) then
!
#ifdef _RANDOM_NOINT32
            if ( btest(tmp,topbit) ) then
                tmp = ior( tmp, noint32_mask )
            end if
#endif
!
!           CONVERT THE 32-BIT INTEGER INTO A RANDOM FLOATING POINT NUMBER IN THE
!           INTERVAL BETWEEN 0 AND 1.
!
            harvest = half + m_ran_invm32*real( tmp, kind=stnd)
!
    else
!
#ifdef _RANDOM_NOINT32
            if ( btest(tmpvec(1_ki_sel),topbit) ) then
                tmpvec(1_ki_sel) = ior( tmpvec(1_ki_sel), noint32_mask )
            end if
#endif
!
!           CONVERT THE TWO 32-BIT INTEGERS INTO A RANDOM FLOATING POINT NUMBER IN THE
!           INTERVAL BETWEEN 0 AND 1.
!
            harvest = half + m_ran_invm32*real( tmpvec(1_ki_sel),                      kind=stnd) &
                           + m_ran_invmbits*real( ishft( tmpvec(2_ki_sel), num_shift), kind=stnd)
!
    end if
!
!    harvest = max( harvest, zero)
!    if ( harvest>=one )  harvest = nearest( one, -one )
!
#ifdef _RANDOM_WITH0
    harvest = min( harvest, nearest( one, -one ) )
#endif
!
!
! END OF FUNCTION rand_number
! ___________________________
!
    end function rand_number
!
! =========================================================================================
!
    subroutine random_r0( harvest )
!
! purpose
! _______
!
!   This subroutine returns a uniformly distributed random number HARVEST between 0 and 1,
!   exclusive of the two endpoints 0 and 1.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd)
!            A uniformly distributed random real number between 0 and 1.
!
!
! Further Details
! _______________
!
!   If the CPP macro _RANDOM_WITH0 is used during compilation, this routine may return 0 value.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(1) :: harvest1
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    call random_r1( harvest1 )
!
    harvest = harvest1(1)
!
!
! END OF SUBROUTINE random_r0
! ___________________________
!
    end subroutine random_r0
!
! =========================================================================================
!
    subroutine random_r1( harvest )
!
! purpose
! _______
!
!   This subroutine returns a uniformly distributed random vector HARVEST between 0 and 1,
!   exclusive of the two endpoints 0 and 1.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd), dimension(:)
!            A uniformly distributed random real vector between 0 and 1.
!
!
! Further Details
! _______________
!
!   If the CPP macro _RANDOM_WITH0 is used during compilation,
!   this routine may return 0 value.
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _RANDOM_NOINT32
    use Char_Constants,   only : random_error7, random_error9
#else
    use Char_Constants,   only : random_error1, random_error2,    &
                                 random_error7, random_error8
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out), dimension(:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b)                  :: i, nharvest
    integer(ki_sel)               :: j, k, tmp
    integer(ki_sel), dimension(2) :: tmpvec
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='random_number_'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    nharvest = size( harvest )
!
    if ( nharvest<=0_i4b ) return
!
    if ( first ) then
#ifdef _RANDOM_NOINT32
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method/=3 ) then
            call merror( name_proc//random_error9 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ibclr(-1_ki_sel, bit_size(tmp)-1_ki_sel)
!
        if ( radix(tmp)/=2 .or. tmp/=huge(tmp) ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
#else
!
!       TEST IF 32-BIT INTEGERS ARE AVAILABLE.
!
        if ( bit_size( tmp )/=fullbitsize ) then
            call merror( name_proc//random_error1 )
        endif
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method<1 .or. method>3 ) then
            call merror( name_proc//random_error2 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ishftc( -8_ki_sel, -3)
!
        if ( radix(tmp)/=2 .or. tmp/=536870911_ki_sel ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
        if ( method/=3 ) then
!
!           TEST IF INTEGER OVERFLOWS ARE SAFE FOR THE MARSAGLIA'S KISS RNGS.
!
            tmp   = huge( tmp )
!
            if ( integer_not_safe( tmp ) ) then
!
                call merror( name_proc//random_error8 )
!
            endif
!
        endif
#endif
!
    endif
!
!$OMP CRITICAL (ran_num_)
    select case (method)
!
        case (1)
!
!           Marsaglia's KISS RNG.
!
            do i = 1_i4b, nharvest
!
                x = 69069_ki_sel * x + 1327217885_ki_sel
                y = ieor( y, ishft(y, kiss_shift1) )
                y = ieor( y, ishft(y, kiss_shift2) )
                y = ieor( y, ishft(y, kiss_shift3) )
!                y = m( m( m( y, kiss_shift1), kiss_shift2), kiss_shift3)
                z = 18000_ki_sel * iand( z, 65535_ki_sel) + ishft( z, -16)
                w = 30903_ki_sel * iand( w, 65535_ki_sel) + ishft( w, -16)
!
                tmp = x + y + ishft( z, 16) + w
!
!               CONVERT THE 32-BIT INTEGER INTO A RANDOM FLOATING POINT NUMBER IN THE
!               INTERVAL BETWEEN 0 AND 1.
!
                harvest(i) = half + m_ran_invm32*real( tmp, kind=stnd)
!
            end do
!
        case (2)
!
!           FAST Marsaglia's KISS RNG WHICH USES ONLY ADD, SHIFT, EXCLUSIVE-OR AND "AND" OPERATIONS.
!
            do i = 1_i4b, nharvest
!
!BUG: For STATPACK with some versions of the GFORTRAN compiler
!     and the optimization level -O3 the following line of code
!     must be replaced with the following commented line below
!     due to a bug in the GFORTRAN compiler.
!
                x2 = x2 + 545925293_ki_sel
!
!                x2 = uiadd( x2, 545925293_ki_sel )
!
                y2 = ieor( y2, ishft(y2, kiss_shift1) )
                y2 = ieor( y2, ishft(y2, kiss_shift2) )
                y2 = ieor( y2, ishft(y2, kiss_shift3) )
!                y2 = m( m( m(y2, kiss_shift1), kiss_shift2), kiss_shift3)
                tmp = z2 + w2 + c
                z2 = w2
                c  = ishft( tmp, -topbit)
                w2 = iand( tmp, 2147483647_ki_sel)
!
                tmp = x2 + y2 + w2
!
!               CONVERT THE 32-BIT INTEGER INTO A RANDOM FLOATING POINT NUMBER IN THE
!               INTERVAL BETWEEN 0 AND 1.
!
                harvest(i) = half + m_ran_invm32*real( tmp, kind=stnd)
!
            end do
!
        case (3)
!
!           L'Ecuyer's LFSR113 RNG.
!
            do i = 1_i4b, nharvest
!
#ifdef _RANDOM_NOINT32
                tmp = ieor( ishft(s1,6), s1)
                tmp = ishft( ibits( tmp, 0, fbs ), -13)
                tmp = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
                s1  = ibits( tmp, 0, fbs )
!
                tmp = ieor( ishft(s2,2), s2)
                tmp = ishft( ibits( tmp, 0, fbs ), -27)
                tmp = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
                s2  = ibits( tmp, 0, fbs )
!
                tmp = ieor( ishft(s3,13), s3)
                tmp = ishft( ibits( tmp, 0, fbs ), -21)
                tmp = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
                s3  = ibits( tmp, 0, fbs )
!
                tmp = ieor( ishft(s4,3), s4)
                tmp = ishft( ibits( tmp, 0, fbs ), -12)
                tmp = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
                s4  = ibits( tmp, 0, fbs )
#else
                tmp = ishft( ieor( ishft(s1,6), s1), -13)
                s1  = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
!
                tmp = ishft( ieor( ishft(s2,2), s2), -27)
                s2  = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
!
                tmp = ishft( ieor( ishft(s3,13), s3), -21)
                s3  = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
!
                tmp = ishft( ieor( ishft(s4,3), s4), -12)
                s4  = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
#endif
!
                tmp = ieor( ieor( ieor(s1,s2), s3), s4)
!
#ifdef _RANDOM_NOINT32
                if ( btest(tmp,topbit) ) then
                    tmp = ior( tmp, noint32_mask )
                end if
#endif
!
!               CONVERT THE 32-BIT INTEGER INTO A RANDOM FLOATING POINT NUMBER IN THE
!               INTERVAL BETWEEN 0 AND 1.
!
                harvest(i) = half + m_ran_invm32*real( tmp, kind=stnd)
!
            end do
!
        case (4)
!
!           MERSENNE TWISTER MT19937 RNG.
!
            do i = 1_i4b, nharvest
!
                if ( mti>=mt_n ) then
!
!                   GENERATE mt_n WORDS AT ONE TIME.
!
                    do j = 0_ki_sel, mt_n - mt_m - 1_ki_sel
!
                        tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                     iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                        mt(j) = ieor( ieor(mt(j+mt_m),ishft(tmp,-1)),   &
                                      mt_mag01(iand(tmp,1_ki_sel))      )
!
                    end do
!
                    do j = mt_n - mt_m, mt_n - 2_ki_sel
!
                        tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                     iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                        mt(j) = ieor( ieor(mt(j+(mt_m-mt_n)), ishft(tmp,-1)),  &
                                      mt_mag01(iand(tmp,1_ki_sel))             )
!
                    end do
!
                    tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                 iand(mt(0_ki_sel),mt_lower_mask)   )
!
                    mt(j) = ieor( ieor(mt(mt_m-1_ki_sel),ishft(tmp,-1)),  &
                                  mt_mag01(iand(tmp,1_ki_sel))            )
!
                    mti = 0_ki_sel
!
                endif
!
                tmp = mt(mti)
                mti = mti + 1_ki_sel
!
!               TEMPERING PHASE.
!
                tmp = ieor( tmp, ishft(tmp,mt_shift0) )
                tmp = ieor( tmp, iand(ishft(tmp,mt_t1_shift),mt_t1_mask) )
                tmp = ieor( tmp, iand(ishft(tmp,mt_t2_shift),mt_t2_mask) )
                tmp = ieor( tmp, ishft(tmp,mt_shift1) )
!
#ifdef _RANDOM_NOINT32
                if ( btest(tmp,topbit) ) then
                    tmp = ior( tmp, noint32_mask )
                end if
#endif
!
!               CONVERT THE 32-BIT INTEGER INTO A RANDOM FLOATING POINT NUMBER IN THE
!               INTERVAL BETWEEN 0 AND 1.
!
                harvest(i) = half + m_ran_invm32*real( tmp, kind=stnd)
!
            end do
!
        case (5)
!
!           MERSENNE TWISTER MEMT19937-II RNG.
!
            do i = 1_i4b, nharvest
!
                select case (memti)
!
                    case (0_ki_sel:mt_n-mt_m-1_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+mt_m),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))          )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (mt_n-mt_m:memt_lag1over-1_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (memt_lag1over:memt_lag2over-1_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (memt_lag2over:mt_n-2_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (mt_n-1_ki_sel)
!
                        tmp = ior( iand(memt(mt_n-1_ki_sel),mt_upper_mask),  &
                                   iand(memt(0_ki_sel),mt_lower_mask)        )
!
                        memt(mt_n-1_ki_sel) = ieor( ieor(memt(mt_m-1_ki_sel),ishft(tmp,-1)),   &
                                                    mt_mag01(iand(tmp,1_ki_sel))               )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                        memti = 0_ki_sel
!
                end select
!
#ifdef _RANDOM_NOINT32
                if ( btest(tmp,topbit) ) then
                    tmp = ior( tmp, noint32_mask )
                else
                    tmp = ibits( tmp, 0, fbs )
                end if
#endif
!
!               CONVERT THE 32-BIT INTEGER INTO A RANDOM FLOATING POINT NUMBER IN THE
!               INTERVAL BETWEEN 0 AND 1.
!
                harvest(i) = half + m_ran_invm32*real( tmp, kind=stnd)
!
            end do
!
        case (6)
!
!           EXTENDED PRECISION VERSION OF Marsaglia's KISS RNG.
!
            do i = 1_i4b, nharvest
!
                do k = 1_ki_sel, 2_ki_sel
!
                    x = 69069_ki_sel * x + 1327217885_ki_sel
                    y = ieor( y, ishft(y, kiss_shift1) )
                    y = ieor( y, ishft(y, kiss_shift2) )
                    y = ieor( y, ishft(y, kiss_shift3) )
!                    y = m( m( m( y, kiss_shift1), kiss_shift2), kiss_shift3)
                    z = 18000_ki_sel * iand( z, 65535_ki_sel) + ishft( z, -16)
                    w = 30903_ki_sel * iand( w, 65535_ki_sel) + ishft( w, -16)
!
                    tmpvec(k) = x + y + ishft( z, 16) + w
!
                end do
!
!               CONVERT THE TWO 32-BIT INTEGERS INTO A RANDOM FLOATING POINT NUMBER IN THE
!               INTERVAL BETWEEN 0 AND 1.
!
                harvest(i) = half + m_ran_invm32*real( tmpvec(1_ki_sel),                      kind=stnd) &
                                  + m_ran_invmbits*real( ishft( tmpvec(2_ki_sel), num_shift), kind=stnd)
!
            end do
!
        case (7)
!
!           EXTENDED PRECISION VERSION OF THE FAST Marsaglia's KISS RNG,
!           WHICH USES ONLY ADD, SHIFT, EXCLUSIVE-OR AND "AND" OPERATIONS.
!
            do i = 1_i4b, nharvest
!
                do k = 1_ki_sel, 2_ki_sel
!
!BUG: For STATPACK with some versions of the GFORTRAN compiler
!     and the optimization level -O3 the following line of code
!     must be replaced with the following commented line below
!     due to a bug in the GFORTRAN compiler.
!
                    x2 = x2 + 545925293_ki_sel
!
!                    x2 = uiadd( x2, 545925293_ki_sel )
!
                    y2 = ieor( y2, ishft(y2, kiss_shift1) )
                    y2 = ieor( y2, ishft(y2, kiss_shift2) )
                    y2 = ieor( y2, ishft(y2, kiss_shift3) )
!                    y2 = m( m( m(y2, kiss_shift1), kiss_shift2), kiss_shift3)
                    tmp = z2 + w2 + c
                    z2 = w2
                    c  = ishft( tmp, -topbit)
                    w2 = iand( tmp, 2147483647_ki_sel)
!
                    tmpvec(k) = x2 + y2 + w2
!
                end do
!
!               CONVERT THE TWO 32-BIT INTEGERS INTO A RANDOM FLOATING POINT NUMBER IN THE
!               INTERVAL BETWEEN 0 AND 1.
!
                harvest(i) = half + m_ran_invm32*real( tmpvec(1_ki_sel),                      kind=stnd) &
                                  + m_ran_invmbits*real( ishft( tmpvec(2_ki_sel), num_shift), kind=stnd)
!
            end do
!
        case (8)
!
!           EXTENDED PRECISION VERSION OF L'Ecuyer's LFSR113 RNG.
!
            do i = 1_i4b, nharvest
!
                do k = 1_ki_sel, 2_ki_sel
!
#ifdef _RANDOM_NOINT32
                    tmp = ieor( ishft(s1,6), s1)
                    tmp = ishft( ibits( tmp, 0, fbs ), -13)
                    tmp = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
                    s1  = ibits( tmp, 0, fbs )
!
                    tmp = ieor( ishft(s2,2), s2)
                    tmp = ishft( ibits( tmp, 0, fbs ), -27)
                    tmp = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
                    s2  = ibits( tmp, 0, fbs )
!
                    tmp = ieor( ishft(s3,13), s3)
                    tmp = ishft( ibits( tmp, 0, fbs ), -21)
                    tmp = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
                    s3  = ibits( tmp, 0, fbs )
!
                    tmp = ieor( ishft(s4,3), s4)
                    tmp = ishft( ibits( tmp, 0, fbs ), -12)
                    tmp = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
                    s4  = ibits( tmp, 0, fbs )
#else
                    tmp = ishft( ieor( ishft(s1,6), s1), -13)
                    s1  = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
!
                    tmp = ishft( ieor( ishft(s2,2), s2), -27)
                    s2  = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
!
                    tmp = ishft( ieor( ishft(s3,13), s3), -21)
                    s3  = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
!
                    tmp = ishft( ieor( ishft(s4,3), s4), -12)
                    s4  = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
#endif
!
                    tmpvec(k) = ieor( ieor( ieor(s1,s2), s3), s4)
!
                end do
!
#ifdef _RANDOM_NOINT32
                if ( btest(tmpvec(1_ki_sel),topbit) ) then
                    tmpvec(1_ki_sel) = ior( tmpvec(1_ki_sel), noint32_mask )
                end if
#endif
!
!               CONVERT THE TWO 32-BIT INTEGERS INTO A RANDOM FLOATING POINT NUMBER IN THE
!               INTERVAL BETWEEN 0 AND 1.
!
                harvest(i) = half + m_ran_invm32*real( tmpvec(1_ki_sel),                      kind=stnd) &
                                  + m_ran_invmbits*real( ishft( tmpvec(2_ki_sel), num_shift), kind=stnd)
!
            end do
!
        case (9)
!
!           EXTENDED PRECISION VERSION OF THE MERSENNE TWISTER MT19937 RNG.
!
            do i = 1_i4b, nharvest
!
                do k = 1_ki_sel, 2_ki_sel
!
                    if ( mti>=mt_n ) then
!
!                       GENERATE mt_n WORDS AT ONE TIME.
!
                        do j = 0_ki_sel, mt_n - mt_m - 1_ki_sel
!
                            tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                         iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                            mt(j) = ieor( ieor(mt(j+mt_m),ishft(tmp,-1)), &
                                          mt_mag01(iand(tmp,1_ki_sel))    )
!
                        end do
!
                        do j = mt_n - mt_m, mt_n - 2_ki_sel
!
                            tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                         iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                            mt(j) = ieor( ieor(mt(j+(mt_m-mt_n)),                     &
                                          ishft(tmp,-1)),mt_mag01(iand(tmp,1_ki_sel)) )
!
                        end do
!
                        tmp   = ior( iand(mt(mt_n-1_ki_sel),mt_upper_mask),    &
                                     iand(mt(0_ki_sel),mt_lower_mask)          )
!
                        mt(j) = ieor( ieor(mt(mt_m-1_ki_sel),ishft(tmp,-1)), &
                                      mt_mag01(iand(tmp,1_ki_sel))           )
!
                        mti = 0_ki_sel
!
                    endif
!
                    tmp = mt(mti)
                    mti = mti + 1_ki_sel
!
!                   TEMPERING PHASE.
!
                    tmp       = ieor( tmp, ishft(tmp,mt_shift0) )
                    tmp       = ieor( tmp, iand(ishft(tmp,mt_t1_shift),mt_t1_mask) )
                    tmp       = ieor( tmp, iand(ishft(tmp,mt_t2_shift),mt_t2_mask) )
                    tmpvec(k) = ieor( tmp, ishft(tmp,mt_shift1) )
!
                end do
!
#ifdef _RANDOM_NOINT32
                if ( btest(tmpvec(1_ki_sel),topbit) ) then
                    tmpvec(1_ki_sel) = ior( tmpvec(1_ki_sel), noint32_mask )
                end if
#endif
!
!               CONVERT THE TWO 32-BIT INTEGERS INTO A RANDOM FLOATING POINT NUMBER IN THE
!               INTERVAL BETWEEN 0 AND 1.
!
                harvest(i) = half + m_ran_invm32*real( tmpvec(1_ki_sel),                      kind=stnd) &
                                  + m_ran_invmbits*real( ishft( tmpvec(2_ki_sel), num_shift), kind=stnd)
!
            end do
!
!
        case (10)
!
!           EXTENDED PRECISION VERSION OF THE MERSENNE TWISTER MEMT19937-II RNG.
!
            do i = 1_i4b, nharvest
!
                do k = 1_ki_sel, 2_ki_sel
!
                  select case (memti)
!
                    case (0_ki_sel:mt_n-mt_m-1_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+mt_m),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))          )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmpvec(k) = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (mt_n-mt_m:memt_lag1over-1_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmpvec(k) = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (memt_lag1over:memt_lag2over-1_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmpvec(k) = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (memt_lag2over:mt_n-2_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmpvec(k) = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (mt_n-1_ki_sel)
!
                        tmp = ior( iand(memt(mt_n-1_ki_sel),mt_upper_mask),  &
                                   iand(memt(0_ki_sel),mt_lower_mask)        )
!
                        memt(mt_n-1_ki_sel) = ieor( ieor(memt(mt_m-1_ki_sel),ishft(tmp,-1)),   &
                                                    mt_mag01(iand(tmp,1_ki_sel))               )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmpvec(k) = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                        memti = 0_ki_sel
!
                  end select
!
#ifdef _RANDOM_NOINT32
                  tmpvec(k) = ibits( tmpvec(k), 0, fbs )
#endif
!
                end do
!
#ifdef _RANDOM_NOINT32
                if ( btest(tmpvec(1_ki_sel),topbit) ) then
                    tmpvec(1_ki_sel) = ior( tmpvec(1_ki_sel), noint32_mask )
                end if
#endif
!
!               CONVERT THE TWO 32-BIT INTEGERS INTO A RANDOM FLOATING POINT NUMBER IN THE
!               INTERVAL BETWEEN 0 AND 1.
!
                harvest(i) = half + m_ran_invm32*real( tmpvec(1_ki_sel),                      kind=stnd) &
                                  + m_ran_invmbits*real( ishft( tmpvec(2_ki_sel), num_shift), kind=stnd)
!
            end do
!
    end select
!$OMP END CRITICAL (ran_num_)
!
!    where( harvest(:)<zero )  harvest(:) = zero
!    where( harvest(:)>=one )  harvest(:) = nearest( one, -one )
!
#ifdef _RANDOM_WITH0
    harvest(:) = min( harvest(:), nearest( one, -one ) )
#endif
!
!
! END OF SUBROUTINE random_r1
! ___________________________
!
    end subroutine random_r1
!
! =========================================================================================
!
    subroutine random_r2( harvest )
!
! purpose
! _______
!
!   This subroutine returns a uniformly distributed random matrix HARVEST between 0 and 1,
!   exclusive of the two endpoints 0 and 1.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd), dimension(:,:)
!            A uniformly distributed random real matrix between 0 and 1.
!
!
! Further Details
! _______________
!
!   If the CPP macro _RANDOM_WITH0 is used during compilation, this routine may return 0 value.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out), dimension(:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 2 )
!
    do i = 1_i4b, nharvest
!
        call random_r1( harvest(:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_r2
! ___________________________
!
    end subroutine random_r2
!
! =========================================================================================
!
    subroutine random_r3( harvest )
!
! purpose
! _______
!
!   This subroutine returns a uniformly distributed random array of dimension 3 HARVEST
!   between 0 and 1, exclusive of the two endpoints 0 and 1.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd), dimension(:,:,:)
!            A uniformly distributed random real array of dimension 3 between 0 and 1.
!
!
! Further Details
! _______________
!
!   If the CPP macro _RANDOM_WITH0 is used during compilation, this routine may return 0 value.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out), dimension(:,:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 3 )
!
    do i = 1_i4b, nharvest
!
        call random_r2( harvest(:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_r3
! ___________________________
!
    end subroutine random_r3
!
! =========================================================================================
!
    subroutine random_r4( harvest )
!
! purpose
! _______
!
!   This subroutine returns a uniformly distributed random array of dimension 4 HARVEST
!   between 0 and 1, exclusive of the two endpoints 0 and 1.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd), dimension(:,:,:,:)
!            A uniformly distributed random real array of dimension 4 between 0 and 1.
!
!
! Further Details
! _______________
!
!   If the CPP macro _RANDOM_WITH0 is used during compilation, this routine may return 0 value.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out), dimension(:,:,:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 4 )
!
    do i = 1_i4b, nharvest
!
        call random_r3( harvest(:,:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_r4
! ___________________________
!
    end subroutine random_r4
!
! =========================================================================================
!
    subroutine random_r5( harvest )
!
! purpose
! _______
!
!   This subroutine returns a uniformly distributed random array of dimension 5 HARVEST
!   between 0 and 1, exclusive of the two endpoints 0 and 1.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd), dimension(:,:,:,:,:)
!            A uniformly distributed random real array of dimension 5 between 0 and 1.
!
!
! Further Details
! _______________
!
!   If the CPP macro _RANDOM_WITH0 is used during compilation, this routine may return 0 value.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out), dimension(:,:,:,:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 5 )
!
    do i = 1_i4b, nharvest
!
        call random_r4( harvest(:,:,:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_r5
! ___________________________
!
    end subroutine random_r5
!
! =========================================================================================
!
    subroutine random_r6( harvest )
!
! purpose
! _______
!
!   This subroutine returns a uniformly distributed random array of dimension 6 HARVEST
!   between 0 and 1, exclusive of the two endpoints 0 and 1.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd), dimension(:,:,:,:,:,:)
!            A uniformly distributed random real array of dimension 6 between 0 and 1.
!
!
! Further Details
! _______________
!
!   If the CPP macro _RANDOM_WITH0 is used during compilation, this routine may return 0 value.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out), dimension(:,:,:,:,:,:)    :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 6 )
!
    do i = 1_i4b, nharvest
!
        call random_r5( harvest(:,:,:,:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_r6
! ___________________________
!
    end subroutine random_r6
!
! =========================================================================================
!
    subroutine random_r7( harvest )
!
! purpose
! _______
!
!   This subroutine returns a uniformly distributed random array of dimension 7 HARVEST
!   between 0 and 1, exclusive of the two endpoints 0 and 1.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd), dimension(:,:,:,:,:,:,:)
!            A uniformly distributed random real array of dimension 7 between 0 and 1.
!
!
! Further Details
! _______________
!
!   If the CPP macro _RANDOM_WITH0 is used during compilation, this routine may return 0 value.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(out), dimension(:,:,:,:,:,:,:)  :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 7 )
!
    do i = 1_i4b, nharvest
!
        call random_r6( harvest(:,:,:,:,:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_r7
! ___________________________
!
    end subroutine random_r7
!
!
! =========================================================================================
!                            RANDOM 32-BIT INTEGER ROUTINES
! =========================================================================================
!
!
    function rand_integer32() result( harvest )
!
! purpose
! _______
!
!   This function returns a random integer in the interval (-2147483648,2147483647)
!   inclusive of the two endpoints. The returned integer is equivalent to
!   a signed 32-bit integer.
!
!
! Arguments
! _________
!
!   None
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _RANDOM_NOINT32
    use Char_Constants,   only : random_error7, random_error9
#else
    use Char_Constants,   only : random_error1, random_error2,    &
                                 random_error7, random_error8
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(ki_sel)  :: j, tmp
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='rand_integer32'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( first ) then
#ifdef _RANDOM_NOINT32
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method/=3 ) then
            call merror( name_proc//random_error9 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ibclr(-1_ki_sel, bit_size(tmp)-1_ki_sel)
!
        if ( radix(tmp)/=2 .or. tmp/=huge(tmp) ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
#else
!
!       TEST IF 32-BIT INTEGERS ARE AVAILABLE.
!
        if ( bit_size( tmp )/=fullbitsize ) then
            call merror( name_proc//random_error1 )
        endif
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method<1 .or. method>3 ) then
            call merror( name_proc//random_error2 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ishftc( -8_ki_sel, -3)
!
        if ( radix(tmp)/=2 .or. tmp/=536870911_ki_sel ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
        if ( method/=3 ) then
!
!           TEST IF INTEGER OVERFLOWS ARE SAFE FOR THE MARSAGLIA'S KISS RNGS.
!
            tmp   = huge( tmp )
!
            if ( integer_not_safe( tmp ) ) then
!
                call merror( name_proc//random_error8 )
!
            endif
!
        endif
#endif
!
    endif
!
!$OMP CRITICAL (ran_num_)
    select case (method)
!
        case (1,6)
!
!           Marsaglia's KISS RNG.
!
            x = 69069_ki_sel * x + 1327217885_ki_sel
            y = ieor( y, ishft(y, kiss_shift1) )
            y = ieor( y, ishft(y, kiss_shift2) )
            y = ieor( y, ishft(y, kiss_shift3) )
!            y = m( m( m( y, kiss_shift1), kiss_shift2), kiss_shift3)
            z = 18000_ki_sel * iand( z, 65535_ki_sel) + ishft( z, -16)
            w = 30903_ki_sel * iand( w, 65535_ki_sel) + ishft( w, -16)
!
            tmp = x + y + ishft( z, 16) + w
!
        case (2,7)
!
!           FAST Marsaglia's KISS RNG WHICH USES ONLY ADD, SHIFT, EXCLUSIVE-OR AND "AND" OPERATIONS.
!
            x2 = x2 + 545925293_ki_sel
!            x2 = uiadd( x2, 545925293_ki_sel )
            y2 = ieor( y2, ishft(y2, kiss_shift1) )
            y2 = ieor( y2, ishft(y2, kiss_shift2) )
            y2 = ieor( y2, ishft(y2, kiss_shift3) )
!            y2 = m( m( m(y2, kiss_shift1), kiss_shift2), kiss_shift3)
            tmp = z2 + w2 + c
            z2 = w2
            c  = ishft( tmp, -topbit)
            w2 = iand( tmp, 2147483647_ki_sel)
!
            tmp = x2 + y2 + w2
!
        case (3,8)
!
!           L'Ecuyer's LFSR113 RNG.
!
#ifdef _RANDOM_NOINT32
            tmp = ieor( ishft(s1,6), s1)
            tmp = ishft( ibits( tmp, 0, fbs ), -13)
            tmp = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
            s1  = ibits( tmp, 0, fbs )
!
            tmp = ieor( ishft(s2,2), s2)
            tmp = ishft( ibits( tmp, 0, fbs ), -27)
            tmp = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
            s2  = ibits( tmp, 0, fbs )
!
            tmp = ieor( ishft(s3,13), s3)
            tmp = ishft( ibits( tmp, 0, fbs ), -21)
            tmp = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
            s3  = ibits( tmp, 0, fbs )
!
            tmp = ieor( ishft(s4,3), s4)
            tmp = ishft( ibits( tmp, 0, fbs ), -12)
            tmp = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
            s4  = ibits( tmp, 0, fbs )
#else
            tmp  = ishft( ieor( ishft(s1,6), s1), -13)
            s1 = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
!
            tmp  = ishft( ieor( ishft(s2,2), s2), -27)
            s2 = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
!
            tmp  = ishft( ieor( ishft(s3,13), s3), -21)
            s3 = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
!
            tmp  = ishft( ieor( ishft(s4,3), s4), -12)
            s4 = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
#endif
!
            tmp = ieor( ieor( ieor(s1,s2), s3), s4)
!
        case (4,9)
!
!           MERSENNE TWISTER MT19937 RNG.
!
            if ( mti>=mt_n ) then
!
                do j = 0_ki_sel, mt_n - mt_m - 1_ki_sel
!
!                   GENERATE mt_n WORDS AT ONE TIME.
!
                    tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                 iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                    mt(j) = ieor( ieor(mt(j+mt_m),ishft(tmp,-1)),   &
                                  mt_mag01(iand(tmp,1_ki_sel))      )
!
                end do
!
                do j = mt_n - mt_m, mt_n - 2_ki_sel
!
                    tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                 iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                    mt(j) = ieor( ieor(mt(j+(mt_m-mt_n)), ishft(tmp,-1)),  &
                                  mt_mag01(iand(tmp,1_ki_sel))             )
!
                end do
!
                tmp   = ior( iand(mt(j),mt_upper_mask),        &
                             iand(mt(0_ki_sel),mt_lower_mask)  )
!
                mt(j) = ieor( ieor(mt(mt_m-1_ki_sel),ishft(tmp,-1)), &
                              mt_mag01(iand(tmp,1_ki_sel))           )
!
                mti = 0_ki_sel
!
            endif
!
            tmp = mt(mti)
            mti = mti + 1_ki_sel
!
!           TEMPERING PHASE.
!
            tmp = ieor( tmp, ishft(tmp,mt_shift0) )
            tmp = ieor( tmp, iand(ishft(tmp,mt_t1_shift),mt_t1_mask) )
            tmp = ieor( tmp, iand(ishft(tmp,mt_t2_shift),mt_t2_mask) )
            tmp = ieor( tmp, ishft(tmp,mt_shift1) )
!
        case (5,10)
!
!           MERSENNE TWISTER MEMT19937-II RNG.
!
            select case (memti)
!
                case (0_ki_sel:mt_n-mt_m-1_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+mt_m),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))          )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (mt_n-mt_m:memt_lag1over-1_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (memt_lag1over:memt_lag2over-1_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (memt_lag2over:mt_n-2_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (mt_n-1_ki_sel)
!
                    tmp = ior( iand(memt(mt_n-1_ki_sel),mt_upper_mask),  &
                               iand(memt(0_ki_sel),mt_lower_mask)        )
!
                    memt(mt_n-1_ki_sel) = ieor( ieor(memt(mt_m-1_ki_sel),ishft(tmp,-1)),   &
                                                mt_mag01(iand(tmp,1_ki_sel))               )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                    memti = 0_ki_sel
!
            end select
!
#ifdef _RANDOM_NOINT32
            tmp = ibits( tmp, 0, fbs )
#endif
!
    end select
!$OMP END CRITICAL (ran_num_)
!
#ifdef _RANDOM_NOINT32
    if ( btest(tmp,topbit) ) then
        tmp = ior( tmp, noint32_mask )
    end if
#endif
!
    harvest = tmp
!
! END OF FUNCTION rand_integer32
! ______________________________
!
    end function rand_integer32
!
! =========================================================================================
!
    subroutine random_i32_0( harvest )
!
! purpose
! _______
!
!   This subroutine returns a random integer in the interval (-2147483648,2147483647)
!   inclusive of the two endpoints. The returned integer is equivalent to
!   a signed 32-bit integer.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b)
!            A random integer in the interval (-2147483648,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b), dimension(1) :: harvest1
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    call random_i32_1( harvest1 )
!
    harvest = harvest1(1)
!
!
! END OF SUBROUTINE random_i32_0
! ______________________________
!
    end subroutine random_i32_0
!
! =========================================================================================
!
    subroutine random_i32_1( harvest )
!
! purpose
! _______
!
!   This subroutine returns a vector of random integers in the interval (-2147483648,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   signed 32-bit integers.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:)
!            A vector of random integers in the interval (-2147483648,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _RANDOM_NOINT32
    use Char_Constants,   only : random_error7, random_error9
#else
    use Char_Constants,   only : random_error1, random_error2,    &
                                 random_error7, random_error8
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out), dimension(:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b)    :: i, nharvest
    integer(ki_sel) :: j, tmp
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='random_integer32_'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    nharvest = size( harvest )
!
    if ( nharvest<=0_i4b ) return
!
    if ( first ) then
#ifdef _RANDOM_NOINT32
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method/=3 ) then
            call merror( name_proc//random_error9 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ibclr(-1_ki_sel, bit_size(tmp)-1_ki_sel)
!
        if ( radix(tmp)/=2 .or. tmp/=huge(tmp) ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
#else
!
!       TEST IF 32-BIT INTEGERS ARE AVAILABLE.
!
        if ( bit_size( tmp )/=fullbitsize ) then
            call merror( name_proc//random_error1 )
        endif
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method<1 .or. method>3 ) then
            call merror( name_proc//random_error2 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ishftc( -8_ki_sel, -3)
!
        if ( radix(tmp)/=2 .or. tmp/=536870911_ki_sel ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
        if ( method/=3 ) then
!
!           TEST IF INTEGER OVERFLOWS ARE SAFE FOR THE MARSAGLIA'S KISS RNGS.
!
            tmp   = huge( tmp )
!
            if ( integer_not_safe( tmp ) ) then
!
                call merror( name_proc//random_error8 )
!
            endif
!
        endif
#endif
!
    endif
!
!$OMP CRITICAL (ran_num_)
    select case (method)
!
        case (1,6)
!
!           Marsaglia's KISS RNG.
!
            do i = 1_i4b, nharvest
!
                x = 69069_ki_sel * x + 1327217885_ki_sel
                y = ieor( y, ishft(y, kiss_shift1) )
                y = ieor( y, ishft(y, kiss_shift2) )
                y = ieor( y, ishft(y, kiss_shift3) )
!                y = m( m( m( y, kiss_shift1), kiss_shift2), kiss_shift3)
                z = 18000_ki_sel * iand( z, 65535_ki_sel) + ishft( z, -16)
                w = 30903_ki_sel * iand( w, 65535_ki_sel) + ishft( w, -16)
!
!               GENERATE A RANDOM SIGNED 32-BIT INTEGER.
!
                harvest(i) = x + y + ishft( z, 16) + w
!
            end do
!
        case (2,7)
!
!           FAST Marsaglia's KISS RNG WHICH USES ONLY ADD, SHIFT, EXCLUSIVE-OR AND "AND" OPERATIONS.
!
            do i = 1_i4b, nharvest
!
!BUG: For STATPACK with some versions of the GFORTRAN compiler
!     and the optimization level -O3 the following line of code
!     must be replaced with the following commented line below
!     due to a bug in the GFORTRAN compiler.
!
                x2 = x2 + 545925293_ki_sel
!
!                x2 = uiadd( x2, 545925293_ki_sel )
!
                y2 = ieor( y2, ishft(y2, kiss_shift1) )
                y2 = ieor( y2, ishft(y2, kiss_shift2) )
                y2 = ieor( y2, ishft(y2, kiss_shift3) )
!                y2 = m( m( m(y2, kiss_shift1), kiss_shift2), kiss_shift3)
                tmp = z2 + w2 + c
                z2 = w2
                c  = ishft( tmp, -topbit)
                w2 = iand( tmp, 2147483647_ki_sel)
!
!               GENERATE A RANDOM SIGNED 32-BIT INTEGER.
!
                harvest(i) = x2 + y2 + w2
!
            end do
!
        case (3,8)
!
!           L'Ecuyer's LFSR113 RNG.
!
            do i = 1_i4b, nharvest
!
#ifdef _RANDOM_NOINT32
                tmp = ieor( ishft(s1,6), s1)
                tmp = ishft( ibits( tmp, 0, fbs ), -13)
                tmp = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
                s1  = ibits( tmp, 0, fbs )
!
                tmp = ieor( ishft(s2,2), s2)
                tmp = ishft( ibits( tmp, 0, fbs ), -27)
                tmp = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
                s2  = ibits( tmp, 0, fbs )
!
                tmp = ieor( ishft(s3,13), s3)
                tmp = ishft( ibits( tmp, 0, fbs ), -21)
                tmp = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
                s3  = ibits( tmp, 0, fbs )
!
                tmp = ieor( ishft(s4,3), s4)
                tmp = ishft( ibits( tmp, 0, fbs ), -12)
                tmp = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
                s4  = ibits( tmp, 0, fbs )
#else
                tmp = ishft( ieor( ishft(s1,6), s1), -13)
                s1  = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
!
                tmp = ishft( ieor( ishft(s2,2), s2), -27)
                s2  = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
!
                tmp = ishft( ieor( ishft(s3,13), s3), -21)
                s3  = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
!
                tmp = ishft( ieor( ishft(s4,3), s4), -12)
                s4  = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
#endif
!
                tmp = ieor( ieor( ieor(s1,s2), s3), s4)
!
#ifdef _RANDOM_NOINT32
                if ( btest(tmp,topbit) ) then
                    tmp = ior( tmp, noint32_mask )
                end if
#endif
!
!               GENERATE A RANDOM SIGNED 32-BIT INTEGER.
!
                harvest(i) = tmp
!
            end do
!
        case (4,9)
!
!           MERSENNE TWISTER MT19937 RNG.
!
            do i = 1_i4b, nharvest
!
                if ( mti>=mt_n ) then
!
!                   GENERATE mt_n WORDS AT ONE TIME.
!
                    do j = 0_ki_sel, mt_n - mt_m - 1_ki_sel
!
                        tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                     iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                        mt(j) = ieor( ieor(mt(j+mt_m),ishft(tmp,-1)),   &
                                      mt_mag01(iand(tmp,1_ki_sel))      )
!
                    end do
!
                    do j = mt_n - mt_m, mt_n - 2_ki_sel
!
                        tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                     iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                        mt(j) = ieor( ieor(mt(j+(mt_m-mt_n)), ishft(tmp,-1)),  &
                                      mt_mag01(iand(tmp,1_ki_sel))             )
!
                    end do
!
                    tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                 iand(mt(0_ki_sel),mt_lower_mask)   )
!
                    mt(j) = ieor( ieor(mt(mt_m-1_ki_sel),ishft(tmp,-1)),  &
                                  mt_mag01(iand(tmp,1_ki_sel))            )
!
                    mti = 0_ki_sel
!
                endif
!
                tmp = mt(mti)
                mti = mti + 1_ki_sel
!
!               TEMPERING PHASE.
!
                tmp = ieor( tmp, ishft(tmp,mt_shift0) )
                tmp = ieor( tmp, iand(ishft(tmp,mt_t1_shift),mt_t1_mask) )
                tmp = ieor( tmp, iand(ishft(tmp,mt_t2_shift),mt_t2_mask) )
                tmp = ieor( tmp, ishft(tmp,mt_shift1) )
!
#ifdef _RANDOM_NOINT32
                if ( btest(tmp,topbit) ) then
                    tmp = ior( tmp, noint32_mask )
                end if
#endif
!
!               GENERATE A RANDOM SIGNED 32-BIT INTEGER.
!
                harvest(i) = tmp
!
            end do
!
        case (5,10)
!
!           MERSENNE TWISTER MT19937-II RNG.
!
            do i = 1_i4b, nharvest
!
                select case (memti)
!
                    case (0_ki_sel:mt_n-mt_m-1_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+mt_m),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))          )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (mt_n-mt_m:memt_lag1over-1_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (memt_lag1over:memt_lag2over-1_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (memt_lag2over:mt_n-2_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (mt_n-1_ki_sel)
!
                        tmp = ior( iand(memt(mt_n-1_ki_sel),mt_upper_mask),  &
                                   iand(memt(0_ki_sel),mt_lower_mask)        )
!
                        memt(mt_n-1_ki_sel) = ieor( ieor(memt(mt_m-1_ki_sel),ishft(tmp,-1)),   &
                                                    mt_mag01(iand(tmp,1_ki_sel))               )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                        memti = 0_ki_sel
!
                end select
!
!               GENERATE A RANDOM SIGNED 32-BIT INTEGER.
!
#ifdef _RANDOM_NOINT32
                if ( btest(tmp,topbit) ) then
                    harvest(i) = ior( tmp, noint32_mask )
                else
                    harvest(i) = ibits( tmp, 0, fbs )
                end if
#else
                harvest(i) = tmp
#endif
!
            end do
!
    end select
!$OMP END CRITICAL (ran_num_)
!
!
! END OF SUBROUTINE random_i32_1
! ______________________________
!
    end subroutine random_i32_1
!
! =========================================================================================
!
    subroutine random_i32_2( harvest )
!
! purpose
! _______
!
!   This subroutine returns a matrix of random integers in the interval (-2147483648,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   signed 32-bit integers.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:,:)
!            A matrix of random integers in the interval (-2147483648,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out), dimension(:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 2 )
!
    do i = 1_i4b, nharvest
!
        call random_i32_1( harvest(:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_i32_2
! ______________________________
!
    end subroutine random_i32_2
!
! =========================================================================================
!
    subroutine random_i32_3( harvest )
!
! purpose
! _______
!
!   This subroutine returns an array of random integers in the interval (-2147483648,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   signed 32-bit integers.
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:,:,:)
!            An array of random integers in the interval (-2147483648,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out), dimension(:,:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 3 )
!
    do i = 1_i4b, nharvest
!
        call random_i32_2( harvest(:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_i32_3
! ______________________________
!
    end subroutine random_i32_3
!
! =========================================================================================
!
    subroutine random_i32_4( harvest )
!
! purpose
! _______
!
!   This subroutine returns an array of random integers in the interval (-2147483648,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   signed 32-bit integers.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:,:,:,:)
!            An array of random integers in the interval (-2147483648,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out), dimension(:,:,:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 4 )
!
    do i = 1_i4b, nharvest
!
        call random_i32_3( harvest(:,:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_i32_4
! ______________________________
!
    end subroutine random_i32_4
!
! =========================================================================================
!
    subroutine random_i32_5( harvest )
!
! purpose
! _______
!
!   This subroutine returns an array of random integers in the interval (-2147483648,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   signed 32-bit integers.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:,:,:,:,:)
!            An array of random integers in the interval (-2147483648,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out), dimension(:,:,:,:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 5 )
!
    do i = 1_i4b, nharvest
!
        call random_i32_4( harvest(:,:,:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_i32_5
! ______________________________
!
    end subroutine random_i32_5
!
! =========================================================================================
!
    subroutine random_i32_6( harvest )
!
! purpose
! _______
!
!   This subroutine returns an array of random integers in the interval (-2147483648,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   signed 32-bit integers.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:,:,:,:,:,:)
!            An array of random integers in the interval (-2147483648,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out), dimension(:,:,:,:,:,:)    :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 6 )
!
    do i = 1_i4b, nharvest
!
        call random_i32_5( harvest(:,:,:,:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_i32_6
! ______________________________
!
    end subroutine random_i32_6
!
! =========================================================================================
!
    subroutine random_i32_7( harvest )
!
! purpose
! _______
!
!   This subroutine returns an array of random integers in the interval (-2147483648,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   signed 32-bit integers.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:,:,:,:,:,:,:)
!            An array of random integers in the interval (-2147483648,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b),  intent(out), dimension(:,:,:,:,:,:,:)  :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 7 )
!
    do i = 1_i4b, nharvest
!
        call random_i32_6( harvest(:,:,:,:,:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_i32_7
! ______________________________
!
    end subroutine random_i32_7
!
!
! =========================================================================================
!                            RANDOM UNSIGNED 31-BIT INTEGER ROUTINES
! =========================================================================================
!
!
    function rand_integer31() result( harvest )
!
! purpose
! _______
!
!   This function returns a random integer in the interval (0,2147483647)
!   inclusive of the two endpoints. The returned integer is equivalent to
!   an unsigned 31-bit integer.
!
!
! Arguments
! _________
!
!   None
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _RANDOM_NOINT32
    use Char_Constants,   only : random_error7, random_error9
#else
    use Char_Constants,   only : random_error1, random_error2,    &
                                 random_error7, random_error8
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(ki_sel)  :: j, tmp
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='rand_integer31'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( first ) then
#ifdef _RANDOM_NOINT32
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method/=3 ) then
            call merror( name_proc//random_error9 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ibclr(-1_ki_sel, bit_size(tmp)-1_ki_sel)
!
        if ( radix(tmp)/=2 .or. tmp/=huge(tmp) ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
#else
!
!       TEST IF 32-BIT INTEGERS ARE AVAILABLE.
!
        if ( bit_size( tmp )/=fullbitsize ) then
            call merror( name_proc//random_error1 )
        endif
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method<1 .or. method>3 ) then
            call merror( name_proc//random_error2 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ishftc( -8_ki_sel, -3)
!
        if ( radix(tmp)/=2 .or. tmp/=536870911_ki_sel ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
        if ( method/=3 ) then
!
!           TEST IF INTEGER OVERFLOWS ARE SAFE FOR THE MARSAGLIA'S KISS RNGS.
!
            tmp   = huge( tmp )
!
            if ( integer_not_safe( tmp ) ) then
!
                call merror( name_proc//random_error8 )
!
            endif
!
        endif
#endif
!
    endif
!
!$OMP CRITICAL (ran_num_)
    select case (method)
!
        case (1,6)
!
!           Marsaglia's KISS RNG.
!
            x = 69069_ki_sel * x + 1327217885_ki_sel
            y = ieor( y, ishft(y, kiss_shift1) )
            y = ieor( y, ishft(y, kiss_shift2) )
            y = ieor( y, ishft(y, kiss_shift3) )
!            y = m( m( m( y, kiss_shift1), kiss_shift2), kiss_shift3)
            z = 18000_ki_sel * iand( z, 65535_ki_sel) + ishft( z, -16)
            w = 30903_ki_sel * iand( w, 65535_ki_sel) + ishft( w, -16)
!
            tmp = x + y + ishft( z, 16) + w
!
        case (2,7)
!
!           FAST Marsaglia's KISS RNG WHICH USES ONLY ADD, SHIFT, EXCLUSIVE-OR AND "AND" OPERATIONS.
!
            x2 = x2 + 545925293_ki_sel
!            x2 = uiadd( x2, 545925293_ki_sel )
            y2 = ieor( y2, ishft(y2, kiss_shift1) )
            y2 = ieor( y2, ishft(y2, kiss_shift2) )
            y2 = ieor( y2, ishft(y2, kiss_shift3) )
!            y2 = m( m( m(y2, kiss_shift1), kiss_shift2), kiss_shift3)
            tmp = z2 + w2 + c
            z2 = w2
            c  = ishft( tmp, -topbit)
            w2 = iand( tmp, 2147483647_ki_sel)
!
            tmp = x2 + y2 + w2
!
        case (3,8)
!
!           L'Ecuyer's LFSR113 RNG.
!
#ifdef _RANDOM_NOINT32
            tmp = ieor( ishft(s1,6), s1)
            tmp = ishft( ibits( tmp, 0, fbs ), -13)
            tmp = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
            s1  = ibits( tmp, 0, fbs )
!
            tmp = ieor( ishft(s2,2), s2)
            tmp = ishft( ibits( tmp, 0, fbs ), -27)
            tmp = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
            s2  = ibits( tmp, 0, fbs )
!
            tmp = ieor( ishft(s3,13), s3)
            tmp = ishft( ibits( tmp, 0, fbs ), -21)
            tmp = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
            s3  = ibits( tmp, 0, fbs )
!
            tmp = ieor( ishft(s4,3), s4)
            tmp = ishft( ibits( tmp, 0, fbs ), -12)
            tmp = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
            s4  = ibits( tmp, 0, fbs )
#else
            tmp = ishft( ieor( ishft(s1,6), s1), -13)
            s1  = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
!
            tmp = ishft( ieor( ishft(s2,2), s2), -27)
            s2  = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
!
            tmp = ishft( ieor( ishft(s3,13), s3), -21)
            s3  = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
!
            tmp = ishft( ieor( ishft(s4,3), s4), -12)
            s4  = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
#endif
!
            tmp = ieor( ieor( ieor(s1,s2), s3), s4)
!
        case (4,9)
!
!           MERSENNE TWISTER MT19937 RNG.
!
            if ( mti>=mt_n ) then
!
                do j = 0_ki_sel, mt_n - mt_m - 1_ki_sel
!
!                   GENERATE mt_n WORDS AT ONE TIME.
!
                    tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                 iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                    mt(j) = ieor( ieor(mt(j+mt_m),ishft(tmp,-1)),   &
                                  mt_mag01(iand(tmp,1_ki_sel))      )
!
                end do
!
                do j = mt_n - mt_m, mt_n - 2_ki_sel
!
                    tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                 iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                    mt(j) = ieor( ieor(mt(j+(mt_m-mt_n)), ishft(tmp,-1)),  &
                                  mt_mag01(iand(tmp,1_ki_sel))             )
!
                end do
!
                tmp   = ior( iand(mt(j),mt_upper_mask),        &
                             iand(mt(0_ki_sel),mt_lower_mask)  )
!
                mt(j) = ieor( ieor(mt(mt_m-1_ki_sel),ishft(tmp,-1)), &
                              mt_mag01(iand(tmp,1_ki_sel))           )
!
                mti = 0_ki_sel
!
            endif
!
            tmp = mt(mti)
            mti = mti + 1_ki_sel
!
!           TEMPERING PHASE.
!
            tmp = ieor( tmp, ishft(tmp,mt_shift0) )
            tmp = ieor( tmp, iand(ishft(tmp,mt_t1_shift),mt_t1_mask) )
            tmp = ieor( tmp, iand(ishft(tmp,mt_t2_shift),mt_t2_mask) )
            tmp = ieor( tmp, ishft(tmp,mt_shift1) )
!
        case (5,10)
!
!           MERSENNE TWISTER MEMT19937-II RNG.
!
            select case (memti)
!
                case (0_ki_sel:mt_n-mt_m-1_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+mt_m),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))          )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (mt_n-mt_m:memt_lag1over-1_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (memt_lag1over:memt_lag2over-1_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (memt_lag2over:mt_n-2_ki_sel)
!
                    tmp = ior( iand(memt(memti),mt_upper_mask),            &
                               iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                    memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                        mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                    memti = memti + 1_ki_sel
!
                case (mt_n-1_ki_sel)
!
                    tmp = ior( iand(memt(mt_n-1_ki_sel),mt_upper_mask),  &
                               iand(memt(0_ki_sel),mt_lower_mask)        )
!
                    memt(mt_n-1_ki_sel) = ieor( ieor(memt(mt_m-1_ki_sel),ishft(tmp,-1)),   &
                                                mt_mag01(iand(tmp,1_ki_sel))               )
!
!                   TEMPERING PHASE.
!
                    tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                    tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                    tmp = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                    memti = 0_ki_sel
!
            end select
!
#ifdef _RANDOM_NOINT32
            tmp = ibits( tmp, 0, fbs )
#endif
!
    end select
!$OMP END CRITICAL (ran_num_)
!
!   GENERATE A RANDOM UNSIGNED 31-BIT INTEGER.
!
    harvest = ibclr( tmp, topbit )
!
! END OF FUNCTION rand_integer31
! ______________________________
!
    end function rand_integer31
!
! =========================================================================================
!
    subroutine random_i31_0( harvest )
!
! purpose
! _______
!
!   This subroutine returns a random integer in the interval (0,2147483647)
!   inclusive of the two endpoints. The returned integer is equivalent to
!   an unsigned 31-bit integer.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b)
!            A random integer in the interval (0,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b), dimension(1) :: harvest1
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    call random_i31_1( harvest1 )
!
    harvest = harvest1(1)
!
!
! END OF SUBROUTINE random_i31_0
! ______________________________
!
    end subroutine random_i31_0
!
! =========================================================================================
!
    subroutine random_i31_1( harvest )
!
! purpose
! _______
!
!   This subroutine returns a vector of random integers in the interval (0,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   unsigned 31-bit integers.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:)
!            A vector of random integers in the interval (0,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _RANDOM_NOINT32
    use Char_Constants,   only : random_error7, random_error9
#else
    use Char_Constants,   only : random_error1, random_error2,    &
                                 random_error7, random_error8
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out), dimension(:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b)    :: i, nharvest
    integer(ki_sel) :: j, tmp
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='random_integer31_'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    nharvest = size( harvest )
!
    if ( nharvest<=0_i4b ) return
!
    if ( first ) then
#ifdef _RANDOM_NOINT32
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method/=3 ) then
            call merror( name_proc//random_error9 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ibclr(-1_ki_sel, bit_size(tmp)-1_ki_sel)
!
        if ( radix(tmp)/=2 .or. tmp/=huge(tmp) ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
#else
!
!       TEST IF 32-BIT INTEGERS ARE AVAILABLE.
!
        if ( bit_size( tmp )/=fullbitsize ) then
            call merror( name_proc//random_error1 )
        endif
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method<1 .or. method>3 ) then
            call merror( name_proc//random_error2 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ishftc( -8_ki_sel, -3)
!
        if ( radix(tmp)/=2 .or. tmp/=536870911_ki_sel ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
        if ( method/=3 ) then
!
!           TEST IF INTEGER OVERFLOWS ARE SAFE FOR THE MARSAGLIA'S KISS RNGS.
!
            tmp   = huge( tmp )
!
            if ( integer_not_safe( tmp ) ) then
!
                call merror( name_proc//random_error8 )
!
            endif
!
        endif
#endif
!
    endif
!
!$OMP CRITICAL (ran_num_)
    select case (method)
!
        case (1,6)
!
!           Marsaglia's KISS RNG.
!
            do i = 1_i4b, nharvest
!
                x = 69069_ki_sel * x + 1327217885_ki_sel
                y = ieor( y, ishft(y, kiss_shift1) )
                y = ieor( y, ishft(y, kiss_shift2) )
                y = ieor( y, ishft(y, kiss_shift3) )
!                y = m( m( m( y, kiss_shift1), kiss_shift2), kiss_shift3)
                z = 18000_ki_sel * iand( z, 65535_ki_sel) + ishft( z, -16)
                w = 30903_ki_sel * iand( w, 65535_ki_sel) + ishft( w, -16)
!
                tmp = x + y + ishft( z, 16) + w
!
!               GENERATE A RANDOM UNSIGNED 31-BIT INTEGER.
!
                harvest(i) = ibclr( tmp, topbit)
!
            end do
!
        case (2,7)
!
!           FAST Marsaglia's KISS RNG WHICH USES ONLY ADD, SHIFT, EXCLUSIVE-OR AND "AND" OPERATIONS.
!
            do i = 1_i4b, nharvest
!
!BUG: For STATPACK with some versions of the GFORTRAN compiler
!     and the optimization level -O3 the following line of code
!     must be replaced with the following commented line below
!     due to a bug in the GFORTRAN compiler.
!
                x2 = x2 + 545925293_ki_sel
!
!                x2 = uiadd( x2, 545925293_ki_sel )
!
                y2 = ieor( y2, ishft(y2, kiss_shift1) )
                y2 = ieor( y2, ishft(y2, kiss_shift2) )
                y2 = ieor( y2, ishft(y2, kiss_shift3) )
!                y2 = m( m( m(y2, kiss_shift1), kiss_shift2), kiss_shift3)
                tmp = z2 + w2 + c
                z2 = w2
                c  = ishft( tmp, -topbit)
                w2 = iand( tmp, 2147483647_ki_sel)
!
                tmp = x2 + y2 + w2
!
!               GENERATE A RANDOM UNSIGNED 31-BIT INTEGER.
!
                harvest(i) = ibclr( tmp, topbit)
!
            end do
!
        case (3,8)
!
!           L'Ecuyer's LFSR113 RNG.
!
            do i = 1_i4b, nharvest
!
#ifdef _RANDOM_NOINT32
                tmp = ieor( ishft(s1,6), s1)
                tmp = ishft( ibits( tmp, 0, fbs ), -13)
                tmp = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
                s1  = ibits( tmp, 0, fbs )
!
                tmp = ieor( ishft(s2,2), s2)
                tmp = ishft( ibits( tmp, 0, fbs ), -27)
                tmp = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
                s2  = ibits( tmp, 0, fbs )
!
                tmp = ieor( ishft(s3,13), s3)
                tmp = ishft( ibits( tmp, 0, fbs ), -21)
                tmp = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
                s3  = ibits( tmp, 0, fbs )
!
                tmp = ieor( ishft(s4,3), s4)
                tmp = ishft( ibits( tmp, 0, fbs ), -12)
                tmp = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
                s4  = ibits( tmp, 0, fbs )
#else
                tmp = ishft( ieor( ishft(s1,6), s1), -13)
                s1  = ieor( ishft( iand(s1,-2_ki_sel), 18), tmp)
!
                tmp = ishft( ieor( ishft(s2,2), s2), -27)
                s2  = ieor( ishft( iand(s2,-8_ki_sel), 2), tmp)
!
                tmp = ishft( ieor( ishft(s3,13), s3), -21)
                s3  = ieor( ishft( iand(s3,-16_ki_sel), 7), tmp)
!
                tmp = ishft( ieor( ishft(s4,3), s4), -12)
                s4  = ieor( ishft( iand(s4,-128_ki_sel), 13), tmp)
#endif
!
                tmp = ieor( ieor( ieor(s1,s2), s3), s4)
!
!               GENERATE A RANDOM UNSIGNED 31-BIT INTEGER.
!
                harvest(i) = ibclr( tmp, topbit)
!
            end do
!
        case (4,9)
!
!           MERSENNE TWISTER MT19937 RNG.
!
            do i = 1_i4b, nharvest
!
                if ( mti>=mt_n ) then
!
!                   GENERATE mt_n WORDS AT ONE TIME.
!
                    do j = 0_ki_sel, mt_n - mt_m - 1_ki_sel
!
                        tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                     iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                        mt(j) = ieor( ieor(mt(j+mt_m),ishft(tmp,-1)),   &
                                      mt_mag01(iand(tmp,1_ki_sel))      )
!
                    end do
!
                    do j = mt_n - mt_m, mt_n - 2_ki_sel
!
                        tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                     iand(mt(j+1_ki_sel),mt_lower_mask) )
!
                        mt(j) = ieor( ieor(mt(j+(mt_m-mt_n)), ishft(tmp,-1)),  &
                                      mt_mag01(iand(tmp,1_ki_sel))             )
!
                    end do
!
                    tmp   = ior( iand(mt(j),mt_upper_mask),         &
                                 iand(mt(0_ki_sel),mt_lower_mask)   )
!
                    mt(j) = ieor( ieor(mt(mt_m-1_ki_sel),ishft(tmp,-1)),  &
                                  mt_mag01(iand(tmp,1_ki_sel))            )
!
                    mti = 0_ki_sel
!
                endif
!
                tmp = mt(mti)
                mti = mti + 1_ki_sel
!
!               TEMPERING PHASE.
!
                tmp = ieor( tmp, ishft(tmp,mt_shift0) )
                tmp = ieor( tmp, iand(ishft(tmp,mt_t1_shift),mt_t1_mask) )
                tmp = ieor( tmp, iand(ishft(tmp,mt_t2_shift),mt_t2_mask) )
                tmp = ieor( tmp, ishft(tmp,mt_shift1) )
!
!               GENERATE A RANDOM UNSIGNED 31-BIT INTEGER.
!
                harvest(i) = ibclr( tmp, topbit)
!
            end do
!
        case (5,10)
!
!           MERSENNE TWISTER MT19937-II RNG.
!
            do i = 1_i4b, nharvest
!
                select case (memti)
!
                    case (0_ki_sel:mt_n-mt_m-1_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+mt_m),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))          )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (mt_n-mt_m:memt_lag1over-1_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti+memt_lag1),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (memt_lag1over:memt_lag2over-1_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti+memt_lag2),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (memt_lag2over:mt_n-2_ki_sel)
!
                        tmp = ior( iand(memt(memti),mt_upper_mask),            &
                                   iand(memt(memti+1_ki_sel),mt_lower_mask)    )
!
                        memt(memti) = ieor( ieor(memt(memti+(mt_m-mt_n)),ishft(tmp,-1)), &
                                            mt_mag01(iand(tmp,1_ki_sel))                 )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                        memti = memti + 1_ki_sel
!
                    case (mt_n-1_ki_sel)
!
                        tmp = ior( iand(memt(mt_n-1_ki_sel),mt_upper_mask),  &
                                   iand(memt(0_ki_sel),mt_lower_mask)        )
!
                        memt(mt_n-1_ki_sel) = ieor( ieor(memt(mt_m-1_ki_sel),ishft(tmp,-1)),   &
                                                    mt_mag01(iand(tmp,1_ki_sel))               )
!
!                       TEMPERING PHASE.
!
                        tmp = ieor( memt(memti), iand(memt(memti-memt_lag1over),memt_mask1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift1) )
                        tmp = ieor( tmp, ishft(tmp,memt_shift2) )
                        tmp = ieor( tmp, iand(memt(memti-memt_lag2over),memt_mask2) )
!
                        memti = 0_ki_sel
!
                end select
!
!               GENERATE A RANDOM UNSIGNED 31-BIT INTEGER.
!
#ifdef _RANDOM_NOINT32
                harvest(i) = ibits( tmp, 0, topbit )
#else
                harvest(i) = ibclr( tmp, topbit)
#endif
!
            end do
!
    end select
!$OMP END CRITICAL (ran_num_)
!
!
! END OF SUBROUTINE random_i31_1
! ______________________________
!
    end subroutine random_i31_1
!
! =========================================================================================
!
    subroutine random_i31_2( harvest )
!
! purpose
! _______
!
!   This subroutine returns a matrix of random integers in the interval (0,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   unsigned 31-bit integers.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:,:)
!            A matrix of random integers in the interval (0,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out), dimension(:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 2 )
!
    do i = 1_i4b, nharvest
!
        call random_i31_1( harvest(:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_i31_2
! ______________________________
!
    end subroutine random_i31_2
!
! =========================================================================================
!
    subroutine random_i31_3( harvest )
!
! purpose
! _______
!
!   This subroutine returns an array of random integers in the interval (0,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   unsigned 31-bit integers.
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:,:,:)
!            An array of random integers in the interval (0,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out), dimension(:,:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 3 )
!
    do i = 1_i4b, nharvest
!
        call random_i31_2( harvest(:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_i31_3
! ______________________________
!
    end subroutine random_i31_3
!
! =========================================================================================
!
    subroutine random_i31_4( harvest )
!
! purpose
! _______
!
!   This subroutine returns an array of random integers in the interval (0,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   unsigned 31-bit integers.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:,:,:,:)
!            An array of random integers in the interval (0,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out), dimension(:,:,:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 4 )
!
    do i = 1_i4b, nharvest
!
        call random_i31_3( harvest(:,:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_i31_4
! ______________________________
!
    end subroutine random_i31_4
!
! =========================================================================================
!
    subroutine random_i31_5( harvest )
!
! purpose
! _______
!
!   This subroutine returns an array of random integers in the interval (0,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   unsigned 31-bit integers.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:,:,:,:,:)
!            An array of random integers in the interval (0,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out), dimension(:,:,:,:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 5 )
!
    do i = 1_i4b, nharvest
!
        call random_i31_4( harvest(:,:,:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_i31_5
! ______________________________
!
    end subroutine random_i31_5
!
! =========================================================================================
!
    subroutine random_i31_6( harvest )
!
! purpose
! _______
!
!   This subroutine returns an array of random integers in the interval (0,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   unsigned 31-bit integers.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:,:,:,:,:,:)
!            An array of random integers in the interval (0,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(out), dimension(:,:,:,:,:,:)    :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 6 )
!
    do i = 1_i4b, nharvest
!
        call random_i31_5( harvest(:,:,:,:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_i31_6
! ______________________________
!
    end subroutine random_i31_6
!
! =========================================================================================
!
    subroutine random_i31_7( harvest )
!
! purpose
! _______
!
!   This subroutine returns an array of random integers in the interval (0,2147483647)
!   inclusive of the two endpoints. The returned integers are equivalent to
!   unsigned 31-bit integers.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) integer(i4b), dimension(:,:,:,:,:,:,:)
!            An array of random integers in the interval (0,2147483647).
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b),  intent(out), dimension(:,:,:,:,:,:,:)  :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nharvest
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nharvest = size( harvest, 7 )
!
    do i = 1_i4b, nharvest
!
        call random_i31_6( harvest(:,:,:,:,:,:,i) )
!
    end do
!
!
! END OF SUBROUTINE random_i31_7
! ______________________________
!
    end subroutine random_i31_7
!
!
! =========================================================================================
!             SUBROUTINES FOR SEEDING AND MANIPULATING THE STATE OF THE RNGS
! =========================================================================================
!
!
    subroutine init_mt19937_r0( seed )
!
! purpose
! _______
!
!   User interface subroutine for initializing the state of the MT19937 Random Number Generator (RNG)
!   with a scalar seed, directly, without using the subroutine RANDOM_SEED\ _ and its interface.
!
!
! Arguments
! _________
!
!   SEED (INPUT) integer(i4b)
!        On entry, a scalar integer that will be used to initialize the MT19937 RNG.
!
!
! Further Details
! _______________
!
!   Only the first 32 bits of the scalar SEED will be used.
!
!   For more informations on the MT19937 RNG, see:
!
!   (1)  Matsumoto, M., and Nishimura, T., 1998:
!            Mersenne Twister: A 623-dimensionally
!            equidistributed uniform pseudorandom number generator.
!            ACM Trans. on Modeling and Computer Simulation, Vol. 8, No. 1, January pp.3-30
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: seed
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(ki_sel) :: tmp
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   COPY THE FIRST 32 BITS OF seed IN tmp WITHOUT INTEGER OVERFLOW.
!
    tmp = ibits( seed, 0, topbit )
!
    if ( btest( seed,  topbit ) ) then
        tmp = ibset( tmp, topbit )
    end if
!
    call init_mt_by_scalar( tmp, mt )
!
!   ENSURE THAT INDEX mti IS INITIALIZED CORRECTLY.
!
    mti = mt_n
!
    mt_initialized = true
!
!
! END OF SUBROUTINE init_mt19937_r0
! _________________________________
!
    end subroutine init_mt19937_r0
!
! =========================================================================================
!
    subroutine init_mt19937_r1( seed )
!
! purpose
! _______
!
!   User interface subroutine for initializing the state of the MT19937 Random Number Generator (RNG)
!   with an array of seeds, directly, without using the subroutine RANDOM_SEED\ _ and its interface.
!
!
! Arguments
! _________
!
!   SEED (INPUT) integer(i4b), dimension(:)
!        On entry, a vector of integers that will be used to initialize the MT19937 RNG.
!        If size(SEED) is zero, a default scalar seed will be used instead.
!
!
! Further Details
! _______________
!
!   Only the first 32 bits of each element of the array SEED will be used.
!
!   For more informations on the MT19937 RNG, see:
!
!   (1)  Matsumoto, M., and Nishimura, T., 1998:
!            Mersenne Twister: A 623-dimensionally
!            equidistributed uniform pseudorandom number generator.
!            ACM Trans. on Modeling and Computer Simulation, Vol. 8, No. 1, January pp.3-30
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), dimension(:), intent(in)  :: seed
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: nseed
    integer(ki_sel), dimension(size(seed)) :: tmp
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nseed = size(seed)
!
    if ( nseed>=1_i4b ) then
!
!       COPY THE FIRST 32 BITS OF EACH ELEMENT OF seed IN THE CORRESPONDING ELEMENT
!       OF tmp WITHOUT INTEGER OVERFLOW.
!
        tmp(:nseed) = ibits( seed(:nseed), 0, topbit )
!
        where(  btest( seed(:nseed),  topbit ) ) tmp(:nseed) = ibset( tmp(:nseed), topbit )
!
        call init_mt_by_array( tmp(:nseed), mt )
!
    else
!
!       IF THE ARRAY seed IS OF ZERO SIZE, USE
!       A DEFAULT SEED.
!
        call init_mt_by_scalar( mt_default_seed, mt )
!
    end if
!
!   ENSURE THAT INDEX mti IS INITIALIZED CORRECTLY.
!
    mti = mt_n
!
    mt_initialized = true
!
!
! END OF SUBROUTINE init_mt19937_r1
! _________________________________
!
    end subroutine init_mt19937_r1
!
! =========================================================================================
!
    subroutine init_memt19937_r0( seed )
!
! purpose
! _______
!
!   User interface subroutine for initializing the state of the MEMT19937-II Random Number Generator (RNG)
!   with a seed, directly, without using the subroutine RANDOM_SEED\ _ and its interface.
!
!
! Arguments
! _________
!
!   SEED (INPUT) integer(i4b)
!        On entry, a scalar integer that will be used to initialize the MEMT19937-II RNG.
!
!
! Further Details
! _______________
!
!   Only the first 32 bits of the scalar SEED will be used.
!
!   For more informations on the MEMT19937-II RNG, see:
!
!   (1)  Harase, S., 2014:
!            On the F2-linear relations of Mersenne Twister pseudorandom number generators.
!            Mathematics and Computers in Simulation, Volume 100, Pages 103-113.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: seed
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(ki_sel) :: tmp
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   COPY THE FIRST 32 BITS OF seed IN tmp WITHOUT INTEGER OVERFLOW.
!
    tmp = ibits( seed, 0, topbit )
!
    if ( btest( seed,  topbit ) ) then
        tmp = ibset( tmp, topbit )
    end if
!
    call init_mt_by_scalar( tmp, memt )
!
!   ENSURE THAT INDEX memti IS INITIALIZED CORRECTLY.
!
    memti = 0_ki_sel
!
    memt_initialized = true
!
!
! END OF SUBROUTINE init_memt19937_r0
! ___________________________________
!
    end subroutine init_memt19937_r0
!
! =========================================================================================
!
    subroutine init_memt19937_r1( seed )
!
! purpose
! _______
!
!   User interface subroutine for initializing the state of the MEMT19937-II Random Number Generator (RNG)
!   with an array of seeds, directly, without using the subroutine RANDOM_SEED\ _ and its interface.
!
!
! Arguments
! _________
!
!   SEED (INPUT) integer(i4b), dimension(:)
!        On entry, a vector of integers that will be used to initialize the MT19937 RNG.
!        If size(SEED) is zero, a default scalar seed will be used instead.
!
!
! Further Details
! _______________
!
!   Only the first 32 bits of each element of the array SEED will be used.
!
!   For more informations on the MEMT19937-II RNG, see:
!
!   (1)  Harase, S., 2014:
!            On the F2-linear relations of Mersenne Twister pseudorandom number generators.
!            Mathematics and Computers in Simulation, Volume 100, Pages 103-113.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), dimension(:), intent(in)  :: seed
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: nseed
    integer(ki_sel), dimension(size(seed)) :: tmp
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nseed = size(seed)
!
    if ( nseed>=1_i4b ) then
!
!       COPY THE FIRST 32 BITS OF EACH ELEMENT OF seed IN THE CORRESPONDING ELEMENT
!       OF tmp WITHOUT INTEGER OVERFLOW.
!
        tmp(:nseed) = ibits( seed(:nseed), 0, topbit )
!
        where(  btest( seed(:nseed),  topbit ) ) tmp(:nseed) = ibset( tmp(:nseed), topbit )
!
        call init_mt_by_array( tmp(:nseed), memt )
!
    else
!
!       IF THE ARRAY seed IS OF ZERO SIZE, USE
!       A DEFAULT SEED.
!
        call init_mt_by_scalar( mt_default_seed, memt )
!
    end if
!
!   ENSURE THAT INDEX memti IS INITIALIZED CORRECTLY.
!
    memti = 0_ki_sel
!
    memt_initialized = true
!
!
! END OF SUBROUTINE init_memt19937_r1
! ___________________________________
!
    end subroutine init_memt19937_r1
!
! =========================================================================================
!
    subroutine random_seed_( alg, size, put, get )
!
! purpose
! _______
!
!   User interface for seeding the random number routines in module RANDOM.
!
!   Syntax is like RANDOM_SEED intrinsic subroutine and a call to RANDOM_SEED\ _()
!   without arguments initiates a non-repeatable reset of the seeds used by the
!   random number subroutines in module RANDOM.
!
!   As for RANDOM_SEED intrinsic subroutine, no more than one argument may be specified
!   in a call to RANDOM_SEED\ _ .
!
!
! Arguments
! _________
!
!   ALG  (INPUT, OPTIONAL) integer
!        On entry, a scalar default integer to select the random number generator used in
!        subsequent calls to subroutines RANDOM_NUMBER\ _ , RANDOM_INTEGER32\ _ , RANDOM_INTEGER31\ _
!        and functions RAND_NUMBER, RAND_INTEGER32 and RAND_INTEGER31. The possible values are:
!
!        - ALG=1  : selects the Marsaglia's KISS random number generator;
!        - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!        - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!        - ALG=4  : selects the Mersenne Twister random number generator;
!        - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!        - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!        - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!        - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!        - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!        - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!          random number generator;
!
!        For other values the random number generator is not changed.
!        The default value is the L'Ecuyer's LFSR113 random number generator (e.g. ALG=3).
!
!   SIZE (OUTPUT, OPTIONAL) integer
!        On exit, the size of the seed array used by the random number generators.
!
!   PUT  (INPUT, OPTIONAL) integer, dimension(:)
!        On entry, a scalar default integer vector of size at least equal to the size of
!        the seed array (as returned by a call to RANDOM_SEED\ _ with argument SIZE) that
!        will be used to reset the seed for subsequent calls to subroutine RANDOM_NUMBER\ _.
!
!   GET  (OUTPUT, OPTIONAL) integer, dimension(:)
!        On exit, a scalar default integer vector, which is the current value of the seed array.
!        Argument GET must be of size at least equal to the size of the seed array (as
!        returned by a call to RANDOM_SEED\ _ with argument SIZE).
!
!
! Further Details
! _______________
!
!   This subroutine is not thread-safe and must not be called in parallel when OPENMP
!   is used. On the other hand, the associated routines RAND_NUMBER, RANDOM_NUMBER\ _,
!   RAND_INTEGER32, RANDOM_INTEGER32\ _, RAND_INTEGER31 and RANDOM_INTEGER31\ _ are
!   thread-safe if used with OpenMP directives and their states will be consistent
!   while called from multiple OpenMP threads.
!
!   The Marsaglia's KISS (Keep It Simple Stupid) random number generator combines:
!
!     1) The congruential generator x(n) = 69069 \cdot x(n-1) + 1327217885 with
!        a period of 2^32;
!     2) A 3-shift shift-register generator with a period of 2^32 - 1;
!     3) Two 16-bit multiply-with-carry generators with a period of
!        597273182964842497 > 2^59.
!
!   The overall period of this KISS random number generator exceeds 2^123. More details
!   on this Marsaglia's KISS random number generator are available in the references (3) and (4).
!   This generator is also the one used by the intrinsic subroutine RANDOM_NUMBER as implemented
!   in the GNU gfortran compiler.
!
!   The "fast" version of the Marsaglia's KISS random number generator uses only
!   add, shift, exclusive-or and 'and' operations to produces exactly the same 32-bit integer
!   output, which C views as unsigned and Fortran views as signed integers. This version avoids
!   multiplication and is probably faster. More details are available in the reference (5).
!
!   The LFSR113 random number generator is described in the reference (2). This
!   random number generator has a period length of about 2^113.
!
!   The MT19937 Mersenne Twister random number generator is described in the reference (7).
!   This random number generator has a period length of about 2^19937-1, and 623-dimensional
!   equidistribution property is assured.
!   
!   The MEMT19937-II Mersenne Twister random number generator is described in the reference (8).
!   This random number generator has also a period length of about 2^19937-1, and a new set of
!   parameters is introduced in the tempering phase of MT19937, which gives a maximally
!   equidistributed Mersenne Twister random number generator.
!
!   Note that the size of the seed array varies according to the selected random
!   number generator.
!
!   For all the random number generators described above, extended precision versions are
!   also available to generate full precision random real numbers of kind STND (up to 63-bit precision),
!   using the method described in the reference (6).
!
!   The FORTRAN versions of these random number generators as implemented here require that
!   32-bits integer type is available on your computer and that 32-bits integers are represented
!   in base 2 with two's complement notation.
!
!   However, the LFSR113, MT19937 and MEMT19937-II Mersenne Twister random number generators will
!   also work if only 64-bits integer type is available on your system, but in that case you must
!   specify the CPP macro _RANDOM_NOINT32 at compilation. Note, however that the other random number
!   generators will not work properly with 64-bits integer type so they cannot be used on such system.
!
!   The KISS random number generators also assumed that integer overflows do not cause crashes.
!   These assumptions are checked before using these random number generators.
!
!   On the other hand, the LFSR113, MT19937 and MEMT19937-II random number generators do not
!   use integer arithmetic and are free of such assumptions.
!
!   This subroutine is adapted from:
!
!   (1)  Hennecke, M., 1995:
!            A Fortran90 interface to random number generation.
!            Computer Physics Communications, Volume 90, Number 1, 117-120
!
!   (2)  L'Ecuyer, P., 1999:
!            Tables of Maximally-Equidistributed Combined LFSR Generators.
!            Mathematics of Computation, 68, 225, 261-269.
!
!   (3)  Marsaglia, G., 1999:
!            Random number generators for Fortran.
!            Posted to the computer-programming-forum. See:
!            http://computer-programming-forum.com/49-fortran/b89977aa62f72ee8.htm
!
!   (4)  Marsaglia, G., 2005:
!            Double precision RNGs.
!            Posted to the electronic billboard sci.math.num-analysis. See:
!            http://sci.tech-archive.net/Archive/sci.math.num-analysis/2005-11/msg00352.html
!
!   (5)  Marsaglia, G., 2007:
!            Fortran and C: United with a KISS.
!            Posted to the Google comp.lang.forum . See:
!            http://groups.google.co.uk/group/comp.lang.fortran/msg/6edb8ad6ec5421a5
!
!   (6)  Doornik, J.A, 2007:
!            Conversion of high-period random number to floating point.
!            ACM Transactions on Modeling and Computer Simulation, 
!            Volume 17, Issue 1, Article No. 3.
!
!   (7)  Matsumoto, M., and Nishimura, T., 1998:
!            Mersenne Twister: A 623-dimensionally
!            equidistributed uniform pseudorandom number generator.
!            ACM Trans. on Modeling and Computer Simulation, Vol. 8, No. 1, January pp.3-30
!
!   (8)  Harase, S., 2014:
!            On the F2-linear relations of Mersenne Twister pseudorandom number generators.
!            Mathematics and Computers in Simulation, Volume 100, Pages 103-113.
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _RANDOM_NOINT32
    use Char_Constants,   only : random_error3, random_error4, random_error7,   &
                                 random_error9, random_error10
#else
    use Char_Constants,   only : random_error1, random_error2, random_error3,   &
                                 random_error4, random_error7, random_error8
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer, optional, intent(in)                :: alg
    integer, optional, intent(out)               :: size
    integer, optional, intent(in),  dimension(:) :: put
    integer, optional, intent(out), dimension(:) :: get
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(ki_sel) :: tmp
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='random_seed_'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( ( present(size) .and. present(put) ) .or.                    &
         ( present(size) .and. present(get) ) .or.                    &
         ( present(put)  .and. present(get) ) .or.                    &
         ( present(alg)  .and. present(get) ) .or.                    &
         ( present(alg)  .and. present(put) ) .or.                    &
         ( present(alg ) .and. present(size) )      ) then
!
         call merror( name_proc//random_error3 )
!
    endif
!
    if ( first ) then
#ifdef _RANDOM_NOINT32
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method/=3 ) then
            call merror( name_proc//random_error9 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ibclr(-1_ki_sel, bit_size(tmp)-1_ki_sel)
!
        if ( radix(tmp)/=2 .or. tmp/=huge(tmp) ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
#else
!
!       TEST IF 32-BIT INTEGERS ARE AVAILABLE.
!
        if ( bit_size( tmp )/=fullbitsize ) then
            call merror( name_proc//random_error1 )
        endif
!
!       TEST IF THE DEFAULT RNG IS SET CORRECTLY.
!
        if ( method<1 .or. method>3 ) then
            call merror( name_proc//random_error2 )
        endif
!
!       TEST IF THE BASE OF THE INTEGER SYSTEM IS 2 AND
!       IF INTEGER REPRESENTATION IS TWO'S COMPLEMENT NOTATION.
!
        tmp = ishftc( -8_ki_sel, -3)
!
        if ( radix(tmp)/=2 .or. tmp/=536870911_ki_sel ) then
!
            call merror( name_proc//random_error7 )
!
        endif
!
        if ( method/=3 ) then
!
!           TEST IF INTEGER OVERFLOWS ARE SAFE FOR THE MARSAGLIA'S KISS RNGS.
!
            tmp   = huge( tmp )
!
            if ( integer_not_safe( tmp ) ) then
!
                call merror( name_proc//random_error8 )
!
            endif
!
            test_kiss = false
!
        endif
#endif
!
!       SAVE THE NUMBER OF SEEDS AND SIZE OF THE SEED ARRAY OF
!       THE DEFAULT RNG FOR LATER USE.
!
        if ( method==2 ) then
            num_seeds = 5
        else
            num_seeds = 4
        end if
!
        seed_size = num_seeds*ratio_i
!
        first = false
!
    endif
!
    if ( present(alg ) ) then
!
        if ( alg>0 .and. alg<=10 ) then
!
            method = alg
!
!           REVERT TO THE STANDARD VERSION OF THE RNGS IF THE NUMBER OF
!           BITS IN THE MANTISSA IS LESS THAN 32.
!
            if ( method>5 .and. num_bits_stnd<=int(fullbitsize) ) method = method - 5
!
            if ( method>5 .and. m_ran_invmbits==-one ) then
!
!               COMPUTE THE CONSTANTS NEEDED TO CONVERT RANDOM INTEGERS
!               INTO RANDOM FLOATING POINT NUMBERS IN THE EXTENDED
!               VERSIONS OF THE RNGS.
!
                m_ran_invmbits = half**(num_bits_stnd)
!
            end if
!
            if ( test_kiss ) then
!
!               TEST IF INTEGER OVERFLOWS ARE SAFE FOR THE MARSAGLIA'S KISS RNGS
!               IF THEY ARE SELECTED.
!
                select case (method)
!
                    case(1,2,6,7)
!
#ifdef _RANDOM_NOINT32
                        call merror( name_proc//random_error10 )
#else
                        tmp  = huge( tmp )
!
                        if ( integer_not_safe( tmp ) ) then
!
                            call merror( name_proc//random_error8 )
!
                        end if
!
                        test_kiss = false
#endif
!
                end select
!
            endif
!
!           SAVE THE NUMBER OF SEEDS AND SIZE OF THE SEED ARRAY OF
!           THE CURRENT RNG FOR LATER USE.
!
            select case (method)
!
                case(1,3,6,8)
!
                    num_seeds = 4
                    seed_size = num_seeds*ratio_i
!
                case(2,7)
!
                    num_seeds = 5
                    seed_size = num_seeds*ratio_i
!
                case(4,5,9,10)
!
                    num_seeds = int( mt_n )
                    seed_size = (num_seeds+1)*ratio_i
!
!                   INITIALIZE THE MERSENNE TWISTER RNGS.
!
                    if ( method==4 .or. method==9 ) then
!
                        if ( .not.mt_initialized ) then
!
                            call init_mt_by_array( mt_default_seed_array, mt )
!                            call init_mt_by_scalar( mt_default_seed, mt )
!
!                           ENSURE THAT INDEX mti IS INITIALIZED CORRECTLY.
!
                            mti = mt_n
!
                            mt_initialized = true
!
                        end if
!
                    else
!
                        if ( .not.memt_initialized ) then
!
                            call init_mt_by_array( mt_default_seed_array, memt )
!                            call init_mt_by_scalar( mt_default_seed, memt )
!
!                           ENSURE THAT INDEX memti IS INITIALIZED CORRECTLY.
!
                            memti = 0_ki_sel
!
                            memt_initialized = true
!
                        end if
!
                    end if
!
            end select
!
        end if
!
    else if ( present(size) ) then
!
        size = seed_size
!
    else if ( present(get ) ) then
!
        call seed_get( get )
!
    else if ( present(put ) ) then
!
        call seed_put( put )
!
    else
!
        call seed_set ( )
!
    endif
!
!----------------------------------------------------------------------
                         contains
!----------------------------------------------------------------------
!
        subroutine seed_set ( )
!
! purpose
! _______
!
!   Internal routine used for initializing the state of the random number generators.
!
! ___________________________________________________________________________________
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
        integer               :: count, rate, i, unit, istat
        integer, dimension(8) :: dt
!
        integer(ki_sel)                       :: jsr, jsr_old
#ifdef _USE_PGI
        integer(ki_sel), dimension(mt_n) :: iseed
#else
        integer(ki_sel), dimension(num_seeds) :: iseed
#endif
!        integer(ki_sel), dimension(5) :: iseed
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!       FIRST TRY IF THE OS PROVIDES A RNG.
!
        open( unit=unit, file=urandom_file, access="stream",                &
              form="unformatted", action="read", status="old", iostat=istat )
!
        if ( istat==0 ) then
!
            read(unit) iseed(:num_seeds)
            close(unit)
!
        else
!
!           OTHERWISE USE THE CURRENT TIME AND PID. THE PID IS USEFUL
!           IN CASE ONE LAUNCHES MULTIPLE INSTANCES OF THE SAME PROGRAM
!           IN PARALLEL.
!
            call system_clock( count=count, count_rate=rate )
!
            if ( rate==0 ) then
!
                call merror( name_proc//random_error4 )
!
            else
!
                if ( count==0 ) then
!
                    call date_and_time( values=dt )
!
                    count = ( dt(1) - 2010 )*365*24*60*60*1000  &
                            + dt(2)         *31*24*60*60*1000   &
                            + dt(3)         *24*60*60*1000      &
                            + dt(5)         *60*60*1000         &
                            + dt(6)         *60*1000            &
                            + dt(7)         *1000               &
                            + dt(8)
                end if
!
#ifndef _RANDOM_NOUNIX
                i = getpid()
                count = ieor( count, i )
#endif
!
                jsr = int( count, ki_sel )
!
!                jsr = ieor( 777755555_ki_sel, jsr )
!
!               NOW USE A FAST BASIC 32-BIT XORSHIFT RNG TO GENERATE
!               THE SEEDS FOR THE SELECTED RNG.
!
#ifdef _RANDOM_NOINT32
                jsr = ibits( jsr, 0, fbs )
!
                do i = 1, num_seeds
!
                    jsr_old  = jsr
!
                    jsr      = ibits( ieor( jsr, ishft(jsr, 5) ), 0, fbs )
                    jsr      = ieor( jsr, ishft(jsr, -7) )
                    jsr      = ibits( ieor( jsr, ishft(jsr, 22) ), 0, fbs )
                    iseed(i) = uiadd( jsr, jsr_old )
!
                end do
#else
                do i = 1, num_seeds
!
                    jsr_old  = jsr
!
                    jsr      = ieor( jsr, ishft(jsr, 5) )
                    jsr      = ieor( jsr, ishft(jsr, -7) )
                    jsr      = ieor( jsr, ishft(jsr, 22) )
!                    jsr      = m( m( m(jsr, 5), -7), 22)
                    iseed(i) = uiadd( jsr, jsr_old )
!                    iseed(i) = jsr + jsr_old
!
                end do
#endif
!
            end if
!
        end if
!
        select case (method)
!
            case(1,6)
!
!               SET THE SEED FOR THE Marsaglia's KISS RNG.
!
                x = iseed(1)
                y = iseed(2)
                z = iseed(3)
                w = iseed(4)
!
!               y MUST NOT BE ZERO.
!
                y = ieor( 777755555_ki_sel, y )
!
            case(2,7)
!
!               SET THE SEEDS FOR THE FAST VERSION OF Marsaglia's KISS RNG.
!
                x2 = iseed(1)
                y2 = iseed(2)
                z2 = ibclr( iseed(3), topbit )
                w2 = ibclr( iseed(4), topbit )
                c  = ishft( iseed(5), -topbit)
!
!               THE FOLLOWING INSTRUCTIONS ENSURE THAT THE RNG WILL HAVE
!               A FULL PERIOD.
!
                if ( y2==0_ki_sel) y2 = 777755555_ki_sel
                if ( mod(z2,7559_ki_sel)==0_ki_sel ) z2 = z2 + 1_ki_sel
                if ( mod(w2,7559_ki_sel)==0_ki_sel ) w2 = w2 + 7558_ki_sel
!
            case(3,8)
!
!               SET THE SEEDS FOR THE L'Ecuyer's LFSR113 RNG.
!
#ifdef _RANDOM_NOINT32
                s1 = ibits( iseed(1), 0, fbs )
                s2 = ibits( iseed(2), 0, fbs )
                s3 = ibits( iseed(3), 0, fbs )
                s4 = ibits( iseed(4), 0, fbs )
!
!               THE INITIAL SEEDS s1, s2, s3 ANS s4 MUST BE LARGER THAN 1, 7, 15 AND 127,
!               RESPECTIVELY, WHEN CONSIDERED AS UNSIGNED 32-BIT INTEGERS.
!               THE FOLLOWING INSTRUCTIONS CHECK THESE CONDITIONS ASSUMING THAT
!               THE BASE OF THE INTEGER SYSTEM IS 2 AND INTEGER REPRESENTATION IS
!               TWO'S COMPLEMENT.
!
                if ( iand(s1,  -2_ki_sel)==0_ki_sel ) s1 = uisub( s1, 1023_ki_sel )
                if ( iand(s2,  -8_ki_sel)==0_ki_sel ) s2 = uisub( s2, 1023_ki_sel )
                if ( iand(s3, -16_ki_sel)==0_ki_sel ) s3 = uisub( s3, 1023_ki_sel )
                if ( iand(s4,-128_ki_sel)==0_ki_sel ) s4 = uisub( s4, 1023_ki_sel )
#else
                s1 = iseed(1)
                s2 = iseed(2)
                s3 = iseed(3)
                s4 = iseed(4)
!
!               THE INITIAL SEEDS s1, s2, s3 ANS s4 MUST BE LARGER THAN 1, 7, 15 AND 127,
!               RESPECTIVELY, WHEN CONSIDERED AS UNSIGNED 32-BIT INTEGERS.
!               THE FOLLOWING INSTRUCTIONS CHECK THESE CONDITIONS ASSUMING THAT
!               THE BASE OF THE INTEGER SYSTEM IS 2 AND INTEGER REPRESENTATION IS
!               TWO'S COMPLEMENT.
!
                if ( iand(s1,  -2_ki_sel)==0_ki_sel ) s1 = s1 - 1023_ki_sel
                if ( iand(s2,  -8_ki_sel)==0_ki_sel ) s2 = s2 - 1023_ki_sel
                if ( iand(s3, -16_ki_sel)==0_ki_sel ) s3 = s3 - 1023_ki_sel
                if ( iand(s4,-128_ki_sel)==0_ki_sel ) s4 = s4 - 1023_ki_sel
#endif
!
            case(4,9)
!
!               SET THE SEEDS FOR THE MT19937 Mersenne Twister RNG.
!
                call init_mt_by_array( iseed, mt )
!
!               ENSURE THAT INDEX mti IS REINITIALIZED CORRECTLY.
!
                mti = mt_n
!
            case(5,10)
!
!               SET THE SEEDS FOR THE MEMT19937-II Mersenne Twister RNG.
!
                call init_mt_by_array( iseed, memt )
!
!               ENSURE THAT INDEX memti IS REINITIALIZED CORRECTLY.
!
                memti = 0_ki_sel
!
        end select
!
!
!    END OF subroutine seed_set
! _____________________________
!
        end subroutine seed_set
!
!----------------------------------------------------------------------
!
!
! END OF SUBROUTINE random_seed_
! ______________________________
!
    end subroutine random_seed_
!
!
! =========================================================================================
!                                   INTERNAL PRIVATE SUBROUTINES
! =========================================================================================
!
!
!    function m(k, n)
!
! purpose
! _______
!
!   Internal routine used by the Marsaglia Shift Register RNGs.
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
!    integer(ki_sel) :: m, k
!
!    integer         :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!    m = ieor( k, ishft(k, n) )
!
!
! END OF function m
! _________________
!
!    end function m
!
! =========================================================================================
!
    function integer_not_safe( ival )
!
! purpose
! _______
!
!   Internal routine used for testing integer arithmetic. This function must be called
!   with IVAL set to huge(1_ki_sel) to perform properly.
!   This function returns .true. if integer overflows don't work as expected
!   and are unsafe. The function returns .false. otherwise.
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(ki_sel), intent(in) :: ival
!
    logical :: integer_not_safe
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(ki_sel) :: tmp
!
    logical :: test1, test2, test3
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    tmp = ival + 1_ki_sel
!
    test1 = tmp /= -ival - 1_ki_sel
    test2 = ival - tmp + 1_ki_sel /= 0_ki_sel
    test3 = ival + tmp + 1_ki_sel /= 0_ki_sel
!
    integer_not_safe = test1 .or. test2 .or. test3
!
!
! END OF function integer_not_safe
! ________________________________
!
    end function integer_not_safe
!
! =========================================================================================
!
    subroutine init_mt_by_array( seed, state )
!
! purpose
! _______
!
!   Internal routine used for initializing the states of the MT19937 or MEMT19937-II 
!   RNGs with an array of seeds.
!
!   This is done without integer overflows.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(ki_sel), intent(in)    :: seed(0_ki_sel:)
    integer(ki_sel), intent(inout) :: state(0_ki_sel:mt_n-1_ki_sel)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(ki_sel) :: nseed, i, im1, j, k, tmp
!
!
! PARAMETERS
! __________
!
    integer(ki_sel), parameter :: seed_d =    19650218_ki_sel !z'12BD6AA'
    integer(ki_sel), parameter :: mult_a =     1664525_ki_sel !z'19660D'
    integer(ki_sel), parameter :: mult_b =  1566083941_ki_sel !z'5D588B65'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nseed = size( seed )
!
    call init_mt_by_scalar( seed_d, state )
!
    i = 1_ki_sel
    j = 0_ki_sel
!
    do k = max(mt_n,nseed), 1_ki_sel, -1_ki_sel
!
        im1 = i - 1_ki_sel
!
        tmp      = ieor( state(im1), ishft(state(im1),-30) )
        tmp      = uimlt( tmp, mult_a )
        state(i) = ieor( state(i), tmp )
        state(i) = uiadd( state(i), uiadd( seed(j), j ) )
         
!        state(i) = ieor( state(i),ieor(state(im1),ishft(state(im1),-30))*mult_a ) + seed(j) + j
!
#ifdef _RANDOM_NOINT32
        state(i) = ibits( state(i), 0, fbs )
#endif
!
        i = i + 1_ki_sel
        j = j + 1_ki_sel
!
        if ( i>=mt_n ) then
          state(0_ki_sel) = state(mt_n-1_ki_sel)
          i = 1_ki_sel
        endif
!
        if ( j>=nseed ) then
          j = 0_ki_sel
        endif
!
      end do
!
      do k = mt_n-1_ki_sel, 1_ki_sel, -1_ki_sel
!
        im1 = i - 1_ki_sel
!
        tmp = ieor( state(im1), ishft(state(im1),-30) ) 
        tmp = uimlt( tmp, mult_b )
        state(i) = ieor( state(i), tmp )
        state(i) = uisub( state(i), i )
         
!        state(i) = ieor( state(i), ieor(state(im1),ishft(state(im1),-30))*mult_b ) - i
!
#ifdef _RANDOM_NOINT32
        state(i) = ibits( state(i), 0, fbs )
#endif
!
        i = i + 1_ki_sel
!
        if ( i>=mt_n ) then
          state(0_ki_sel) = state(mt_n-1_ki_sel)
          i = 1_ki_sel
        endif
!
      end do
!
      state(0_ki_sel) = mt_upper_mask
!
!
! END OF SUBROUTINE init_mt_by_array
! __________________________________
!
    end subroutine init_mt_by_array
!
! =========================================================================================
!
    subroutine init_mt_by_scalar( seed, state )
!
! purpose
! _______
!
!   Internal routine used for initializing the states of the MT19937 or MEMT19937-II 
!   RNGs with a seed.
!
!   This is done without integer overflows.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(ki_sel), intent(in)    :: seed
    integer(ki_sel), intent(inout) :: state(0_ki_sel:mt_n-1_ki_sel)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(ki_sel) :: i, tmp
!
!
! PARAMETERS
! __________
!
    integer(ki_sel), parameter :: mult_a = 1812433253_ki_sel  !z'6C078965'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    state(0_ki_sel) = seed
#ifdef _RANDOM_NOINT32
    state(0_ki_sel) = ibits( seed, 0, fbs )
#endif
!
    do  i = 1_ki_sel, mt_n-1_ki_sel
!
        tmp      = ieor( state(i-1_ki_sel), ishft(state(i-1_ki_sel),-30) )
        tmp      = uimlt( tmp, mult_a )
        state(i) = uiadd( tmp, i )
!
!        state(i) = mult_a*ieor(state(i-1_ki_sel),ishft(state(i-1_ki_sel),-30)) + i
!
#ifdef _RANDOM_NOINT32
        state(i) = ibits( state(i), 0, fbs )
#endif
!
     end do
!
!
! END OF SUBROUTINE init_mt_by_scalar
! ___________________________________
!
    end subroutine init_mt_by_scalar
!
! =========================================================================================
!
    elemental function uiadd( a, b ) result( c )
!
! purpose
! _______
!
!   Internal routine used to simulate addition for unsigned 32-bit integers.
!   This is done without overflows
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(ki_sel), intent(in) :: a, b
!
    integer(ki_sel) :: c
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(ki_sel) :: a1, a2, b1, b2, s1, s2
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    a1 = ibits( a, 0, hbs )
    a2 = ibits( a, hbs, hbs )
!
    b1 = ibits( b, 0, hbs )
    b2 = ibits( b, hbs, hbs )
!
    s1 = a1 + b1
    s2 = a2 + b2 + ibits( s1, hbs, hbs )
!
    c  = ior( ishft( s2, hbs ), ibits( s1, 0, hbs ) )
!
!
! END OF FUNCTION uiadd
! _____________________
!
    end function uiadd
!
! =========================================================================================
!
    elemental function uisub( a, b ) result( c )
!
! purpose
! _______
!
!   Internal routine used to simulate substraction for unsigned 32-bit integers.
!   This is done without underflows or overflows.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(ki_sel), intent(in) :: a, b
!
    integer(ki_sel) :: c
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(ki_sel) :: a1, a2, b1, b2, s1, s2
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    a1 = ibits( a, 0, hbs )
    a2 = ibits( a, hbs, hbs )
!
    b1 = ibits( b, 0, hbs )
    b2 = ibits( b, hbs, hbs )
!
    s1 = a1 - b1
    s2 = a2 - b2 + ibits( s1, hbs, hbs )
!
    c  = ior( ishft( s2, hbs ), ibits( s1, 0, hbs ) )
!
!
! END OF FUNCTION uisub
! _____________________
!
    end function uisub
!
! =========================================================================================
!
    elemental function uimlt( a, b ) result( c )
!
! purpose
! _______
!
!   Internal routine used to simulate multiplication for unsigned 32-bit integers.
!   This is done without underflows or overflows.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(ki_sel), intent(in) :: a, b
!
    integer(ki_sel) :: c
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(ki_sel) :: a0, a1, a2, a3,  &
                       b0, b1, b2, b3,  &
                       p0, p1, p2, p3
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    a0 = ibits( a, 0, qbs )
    a1 = ibits( a, qbs, qbs )
    a2 = ibits( a, hbs, qbs )
    a3 = ibits( a, tbs, qbs )
!
    b0 = ibits( b, 0, qbs )
    b1 = ibits( b, qbs, qbs )
    b2 = ibits( b, hbs, qbs )
    b3 = ibits( b, tbs, qbs )
!
    p0 = a0 * b0
    p1 = a1 * b0 + a0 * b1 + ibits( p0, qbs, tbs )
    p2 = a2 * b0 + a1 * b1 + a0 * b2 + ibits( p1, qbs, tbs )
    p3 = a3 * b0 + a2 * b1 + a1 * b2 + a0 * b3 + ibits( p2, qbs, tbs )
!
    c  = ior( ishft( p1, qbs ), ibits( p0, 0, qbs ) )
    c  = ior( ishft( p2, hbs ), ibits( c, 0, hbs ) )
    c  = ior( ishft( p3, tbs ), ibits( c, 0, tbs ) )
!
!
! END OF FUNCTION uimlt
! _____________________
!
    end function uimlt
!
! =========================================================================================
!
    subroutine seed_put( put )
!
! purpose
! _______
!
!   Internal subroutine for the STATPACK RANDOM_SEED_ subroutine.
!
! Further Details
! _______________
!
!   This subroutine is adapted from:
!
!   (1)  Hennecke, M., 1995:
!             A Fortran90 interface to random number generation.
!             Computer Physics Communications, Volume 90, Number 1, 117-120
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : random_error5
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer, intent(in), dimension(:) :: put
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer                               :: i, i1, i2
    integer,         dimension(ratio_i)   :: put2
    integer(ki_sel), dimension(num_seeds) :: iseed
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='random_seed_'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( size(put) >= seed_size ) then
!
        do i = 1, num_seeds
!
            i1 = 1+((i-1)*ratio_i)
            i2 = i*ratio_i
            put2(:ratio_i) = put(i1:i2)
            iseed(i)       = transfer(put2(:ratio_i), s1)
!
!BUG: For STATPACK on MACOS10 machines with the G95 compiler,
!     replace the following line with the four lines of codes
!     above due to a bug in the TRANSFER intrinsic function.
!
!            iseed(i) = transfer(put(1+((i-1)*ratio_i):i*ratio_i), s1)
!
#ifdef _RANDOM_NOINT32
            iseed(i) = ibits( iseed(i), 0, fbs )
#endif
!
        end do
!
        select case (method)
!
            case(1,6)
!
!               SET THE SEED FOR THE Marsaglia's KISS RANDOM NUMBER GENERATOR.
!
                x = iseed(1)
                y = iseed(2)
                z = iseed(3)
                w = iseed(4)
!
            case(2,7)
!
!               SET THE SEED FOR THE Marsaglia's KISS RANDOM NUMBER GENERATOR.
!
                x2 = iseed(1)
                y2 = iseed(2)
                z2 = iseed(3)
                w2 = iseed(4)
                c  = iseed(5)
!
            case(3,8)
!
!               SET THE SEED FOR THE L'Ecuyer's LFSR113 RANDOM NUMBER GENERATOR.
!
                s1 = iseed(1)
                s2 = iseed(2)
                s3 = iseed(3)
                s4 = iseed(4)
!
            case(4,9)
!
!               SET THE SEEDS FOR THE MT19937 Mersenne Twister RNG.
!
                mt(0:num_seeds-1) = iseed(:num_seeds)
!
                i1 = 1+((i-1)*ratio_i)
                i2 = i*ratio_i
!
                put2(:ratio_i) = put(i1:i2)
                mti = transfer(put2(:ratio_i), s1)
#ifdef _RANDOM_NOINT32
                mti = ibits( mti, 0, fbs )
#endif
!
            case(5,10)
!
!               SET THE SEEDS FOR THE MEMT19937-II Mersenne Twister RNG.
!
                memt(0:num_seeds-1) = iseed(:num_seeds)
!
                i1 = 1+((i-1)*ratio_i)
                i2 = i*ratio_i
!
                put2(:ratio_i) = put(i1:i2)
                memti = transfer(put2(:ratio_i), s1)
#ifdef _RANDOM_NOINT32
                memti = ibits( memti, 0, fbs )
#endif
!
        end select
!
    else
!
        call merror( name_proc//random_error5 )
!
    endif
!
!
! END OF SUBROUTINE seed_put
! __________________________
!
    end subroutine seed_put
!
! =========================================================================================
!
    subroutine seed_get ( get )
!
! purpose
! _______
!
!   Internal subroutine for the STATPACK RANDOM_SEED_ subroutine.
!
! Further Details
! _______________
!
!   This subroutine is adapted from:
!
!   (1)  Hennecke, M., 1995:
!             A Fortran90 interface to random number generation.
!             Computer Physics Communications, Volume 90, Number 1, 117-120
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : random_error6
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer, intent(out), dimension(:) :: get
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer                                 :: i, nseeds
    integer(ki_sel), dimension(num_seeds+1) :: iseed
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='random_seed_'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( size(get) >= seed_size ) then
!
        nseeds = num_seeds
!
        select case (method)
!
            case(1,6)
!
!               GET THE SEED FOR THE Marsaglia's KISS RANDOM NUMBER GENERATOR.
!
                iseed(:4) = (/ x, y, z, w /)
!
            case(2,7)
!
!               GET THE SEED FOR THE Marsaglia's KISS RANDOM NUMBER GENERATOR.
!
                iseed(:5) = (/ x2, y2, z2, w2, c /)
!
            case(3,8)
!
!               GET THE SEED FOR THE L'Ecuyer's LFSR113 RANDOM NUMBER GENERATOR.
!
                iseed(:4) = (/ s1, s2, s3, s4 /)
!
            case(4,9)
!
!               GET THE SEEDS FOR THE MT19937 Mersenne Twister RNG.
!
                iseed(:nseeds) = mt(0:nseeds-1)
!
                nseeds = nseeds + 1
                iseed(nseeds) = mti

!
            case(5,10)
!
!               GET THE SEEDS FOR THE MEMT19937-II Mersenne Twister RNG.
!
                iseed(:nseeds) = memt(0:nseeds-1)
!
                nseeds = nseeds + 1
                iseed(nseeds) = memti
!
        end select
!
        do i = 1, nseeds
!
            get(1+((i-1)*ratio_i):i*ratio_i) = transfer(iseed(i), get, ratio_i)
!
        end do
!
    else
!
        call merror( name_proc//random_error6 )
!
    endif
!
!
! END OF SUBROUTINE seed_get
! __________________________
!
    end subroutine seed_get
!
!
! =========================================================================================
!                            GAUSSIAN RANDOM FLOATING POINT NUMBER SUBROUTINES
! =========================================================================================
!
!
    function normal_rand_number() result( harvest )
!
! purpose
! _______
!
!   This function returns a Gaussian distributed random real number.
!
!
! Arguments
! _________
!
!   None
!
!
! Further Details
! _______________
!
!   This function uses the Cumulative Density Function (CDF) inversion method to generate
!   a Gaussian random real number. Starting with a random number produced by the STATPACK
!   uniform random number generator that can produce random numbers with the
!   uniform distribution over the continuous range (0, 1) (denoted U(0, 1)), the CDF
!   method simply inverts the CDF of a standard Gaussian distribution to produce a
!   standard Gaussian  (e.g. a Gaussian distribution with mean zero and standard
!   deviation one) random real number.
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 2 and 3) by the subroutine PPND7 described in the reference (1).
!   This method gives 7 decimal digits of accuracy in the range [10**(-316), 1-10**(-316)]
!   if computations are done in double or higher precision.
!
!   For more details on Uniform and Gaussian random number generators or the approximation
!   of the inverse Gaussian CDF used here, see:
!
!   (1) Devroye, L., 1986:
!           Non-Uniform Random Variate Generation. Springer-Verlag,
!           http://cg.scs.carleton.ca/~luc/rnbookindex.html, New York.
!
!   (2) Thomas, D.B., Luk, W., Leong, P.H.W., and Villasenor, J.D., 2007:
!           Gaussian random number generators.
!           ACM Comput. Surv. 39, 4, Article 11 (October 2007), 38 pages,
!           DOI = 10.1145/1287620.1287622 (http://doi.acm.org/10.1145/1287620.1287622)
!
!   (3) Wichura, M.J., 1988:
!           Algorithm AS 241: The percentage points of the normal distribution.
!           Appl. Statis. 37, 3, 477-484.
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Prob_Procedures, only : pinvn
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: harvest1
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    harvest1 = rand_number()
!
#ifdef _RANDOM_WITH0
    if ( harvest1<=zero )  harvest1 = nearest( zero, one )
#endif
!
    harvest = pinvn( harvest1 )
!
!
! END OF FUNCTION normal_rand_number
! __________________________________
!
    end function normal_rand_number
!
! =========================================================================================
!
    subroutine normal_random_r0( harvest )
!
! purpose
! _______
!
!   This subroutine returns a random real number HARVEST following the standard
!   Gaussian distribution.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd)
!            A Gaussian distributed random real number.
!
!
! Further Details
! _______________
!
!   This subroutine uses the Cumulative Density Function (CDF) inversion method to generate
!   a Gaussian random real number. Starting with a random number produced by the STATPACK
!   uniform random number generator that can produce random numbers with the
!   uniform distribution over the continuous range (0, 1) (denoted U(0, 1)), the CDF
!   method simply inverts the CDF of a standard Gaussian distribution to produce a
!   standard Gaussian  (e.g. a Gaussian distribution with mean zero and standard
!   deviation one) random real number.
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 2 and 3) by the subroutine PPND7 described in the reference (1).
!   This method gives 7 decimal digits of accuracy in the range [10**(-316), 1-10**(-316)]
!   if computations are done in double or higher precision.
!
!   For more details on Uniform and Gaussian random number generators or the approximation
!   of the inverse Gaussian CDF used here, see:
!
!   (1) Devroye, L., 1986:
!           Non-Uniform Random Variate Generation. Springer-Verlag,
!           http://cg.scs.carleton.ca/~luc/rnbookindex.html, New York.
!
!   (2) Thomas, D.B., Luk, W., Leong, P.H.W., and Villasenor, J.D., 2007:
!           Gaussian random number generators.
!           ACM Comput. Surv. 39, 4, Article 11 (October 2007), 38 pages,
!           DOI = 10.1145/1287620.1287622 (http://doi.acm.org/10.1145/1287620.1287622)
!
!   (3) Wichura, M.J., 1988:
!           Algorithm AS 241: The percentage points of the normal distribution.
!           Appl. Statis. 37, 3, 477-484.
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Prob_Procedures, only : pinvn
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(1) :: harvest1
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    call random_r1( harvest1 )
!
#ifdef _RANDOM_WITH0
    if ( harvest1(1)<=zero )  harvest1(1) = nearest( zero, one )
#endif
!
    harvest = pinvn( harvest1(1) )
!
!
! END OF SUBROUTINE normal_random_r0
! __________________________________
!
    end subroutine normal_random_r0
!
! =========================================================================================
!
    subroutine normal_random_r1( harvest )
!
! purpose
! _______
!
!   This subroutine returns a random vector HARVEST following the standard
!   normal (Gaussian) distribution.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd), dimension(:)
!            A Gaussian distributed random real vector.
!
!
! Further Details
! _______________
!
!   This subroutines uses the Cumulative Density Function (CDF) inversion method to generate
!   Gaussian random numbers. Starting with random numbers produced by the STATPACK
!   uniform random number generator that can produce random numbers with the
!   uniform distribution over the continuous range (0, 1) (denoted U(0, 1)), the CDF
!   method simply inverts the CDF of a standard Gaussian distribution to produce
!   standard Gaussian  (e.g. a Gaussian distribution with mean zero and standard
!   deviation one) random numbers.
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 2 and 3) by the subroutine PPND7 described in the reference (1).
!   This method gives 7 decimal digits of accuracy in the range [10**(-316), 1-10**(-316)]
!   if computations are done in double or higher precision.
!
!   For more details on Uniform and Gaussian random number generators or the approximation
!   of the inverse Gaussian CDF used here, see:
!
!   (1) Devroye, L., 1986:
!           Non-Uniform Random Variate Generation. Springer-Verlag,
!           http://cg.scs.carleton.ca/~luc/rnbookindex.html, New York.
!
!   (2) Thomas, D.B., Luk, W., Leong, P.H.W., and Villasenor, J.D., 2007:
!           Gaussian random number generators.
!           ACM Comput. Surv. 39, 4, Article 11 (October 2007), 38 pages,
!           DOI = 10.1145/1287620.1287622 (http://doi.acm.org/10.1145/1287620.1287622)
!
!   (3) Wichura, M.J., 1988:
!           Algorithm AS 241: The percentage points of the normal distribution.
!           Appl. Statis. 37, 3, 477-484.
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _ALLOC
    use Char_Constants,  only : allocate_error
#endif
    use Prob_Procedures, only : pinvn
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out), dimension(:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _ALLOC
    real(stnd), dimension(:), allocatable :: harvest1
!
    integer      :: iok
!
#else
    real(stnd), dimension(size(harvest)) :: harvest1
#endif
!
    integer(i4b) :: m
!
!
! PARAMETERS
! __________
!
#ifdef _ALLOC
    character(len=*),  parameter :: name_proc='normal_random_number_'
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( harvest )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b )  return
!
#ifdef _ALLOC
!
!   ALLOCATE WORK VARIABLE.
!
    allocate( harvest1(m), stat = iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
#endif
!
    call random_r1( harvest1(:m) )
!
#ifdef _RANDOM_WITH0
    where ( harvest1(:m)<=zero )  harvest1(:m) = nearest( zero, one )
#endif
!
    harvest(:m) = pinvn( harvest1(:m) )
!
#ifdef _ALLOC
!
!   DEALLOCATE WORK VARIABLE.
!
    deallocate( harvest1 )
#endif
!
!
! END OF SUBROUTINE normal_random_r1
! __________________________________
!
    end subroutine normal_random_r1
!
! =========================================================================================
!
    subroutine normal_random_r2( harvest )
!
! purpose
! _______
!
!   This subroutine returns a random matrix HARVEST following the standard
!   normal (Gaussian) distribution.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd), dimension(:,:)
!            A Gaussian distributed random real matrix.
!
!
! Further Details
! _______________
!
!   This subroutines uses the Cumulative Density Function (CDF) inversion method to generate
!   Gaussian random real numbers. Starting with random numbers produced by the STATPACK
!   uniform random number generator that can produce random numbers with the
!   uniform distribution over the continuous range (0, 1) (denoted U(0, 1)), the CDF
!   method simply inverts the CDF of a standard Gaussian distribution to produce
!   standard Gaussian  (e.g. a Gaussian distribution with mean zero and standard
!   deviation one) random real numbers.
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 2 and 3) by the subroutine PPND7 described in the reference (1).
!   This method gives 7 decimal digits of accuracy in the range [10**(-316), 1-10**(-316)]
!   if computations are done in double or higher precision.
!
!   For more details on Uniform and Gaussian random number generators or the approximation
!   of the inverse Gaussian CDF used here, see :
!
!   (1) Devroye, L., 1986:
!           Non-Uniform Random Variate Generation. Springer-Verlag,
!           http://cg.scs.carleton.ca/~luc/rnbookindex.html, New York.
!
!   (2) Thomas, D.B., Luk, W., Leong, P.H.W., and Villasenor, J.D., 2007:
!           Gaussian random number generators.
!           ACM Comput. Surv. 39, 4, Article 11 (October 2007), 38 pages,
!           DOI = 10.1145/1287620.1287622 (http://doi.acm.org/10.1145/1287620.1287622)
!
!   (3) Wichura, M.J., 1988:
!           Algorithm AS 241: The percentage points of the normal distribution.
!           Appl. Statis. 37, 3, 477-484.
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _ALLOC
    use Char_Constants,  only : allocate_error
#endif
    use Prob_Procedures, only : pinvn
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out), dimension(:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _ALLOC
    real(stnd), dimension(:), allocatable :: harvest1
#else
    real(stnd), dimension(size(harvest,1)) :: harvest1
#endif
!
    integer(i4b) :: m, n, i
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
#ifdef _ALLOC
    character(len=*),  parameter :: name_proc='normal_random_number_'
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( harvest, 1 )
    n = size( harvest, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b )  return
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m*n>=omp_limit                      .and.       &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par),PRIVATE(i,harvest1)
!
#ifdef _ALLOC
!
!   ALLOCATE WORK VARIABLE.
!
    allocate( harvest1(m), stat = i )
!
    if ( i/=0_i4b ) then
!
        call merror( name_proc//allocate_error )
!
    end if
#endif
!
!$OMP DO SCHEDULE(STATIC)
!
    do i = 1_i4b, n
!
        call random_r1( harvest1(:m) )
!
#ifdef _RANDOM_WITH0
        where ( harvest1(:m)<=zero )  harvest1(:m) = nearest( zero, one )
#endif
!
        harvest(:m,i) = pinvn( harvest1(:m) )
!
    end do
!
!$OMP END DO
!
#ifdef _ALLOC
!
!   DEALLOCATE WORK VARIABLE.
!
    deallocate( harvest1 )
#endif
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE normal_random_r2
! __________________________________
!
    end subroutine normal_random_r2
!
! =========================================================================================
!
    function normal_rand_number2() result( harvest )
!
! purpose
! _______
!
!   This function returns a Gaussian distributed real random number of kind extd.
!
!
! Arguments
! _________
!
!   None
!
!
! Further Details
! _______________
!
!   This function uses the Cumulative Density Function (CDF) inversion method to generate
!   a Gaussian random real number of kind extd. Starting with a random number produced by
!   the STATPACK uniform random number generator that can produce random numbers with the
!   uniform distribution over the continuous range (0, 1) (denoted U(0, 1)), the CDF
!   method simply inverts the CDF of a standard Gaussian distribution to produce a
!   standard Gaussian  (e.g. a Gaussian distribution with mean zero and standard
!   deviation one) random real number of kind extd.
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 7) by the subroutine PPND16 described in the reference (1).
!   This method gives about 16 decimal digits of accuracy in the range [10**(-316), 1-10**(-316)]
!   if computations are done in double or higher precision.
!
!   This function is more accurate than NORMAL_RAND_NUMBER function, but it is slower.
!
!   For more details on Uniform and Gaussian random number generators or the approximation
!   of the inverse Gaussian CDF used here, see :
!
!   (1) Devroye, L., 1986:
!           Non-Uniform Random Variate Generation. Springer-Verlag,
!           http://cg.scs.carleton.ca/~luc/rnbookindex.html, New York.
!
!   (2) Thomas, D.B., Luk, W., Leong, P.H.W., and Villasenor, J.D., 2007:
!           Gaussian random number generators.
!           ACM Comput. Surv. 39, 4, Article 11 (October 2007), 38 pages,
!           DOI = 10.1145/1287620.1287622 (http://doi.acm.org/10.1145/1287620.1287622)
!
!   (3) Wichura, M.J., 1988:
!           Algorithm AS 241: The percentage points of the normal distribution.
!           Appl. Statis. 37, 3, 477-484.
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Prob_Procedures, only : pinvn2
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(extd) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: harvest1
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    harvest1 = rand_number()
!
#ifdef _RANDOM_WITH0
    if ( harvest1<=zero )  harvest1 = nearest( zero, one )
#endif
!
    harvest = pinvn2( real( harvest1, extd ) )
!
!
! END OF FUNCTION normal_rand_number2
! ___________________________________
!
    end function normal_rand_number2
!
! =========================================================================================
!
    subroutine normal_random2_r0( harvest )
!
! purpose
! _______
!
!   This subroutine  returns a Gaussian distributed real random number of kind extd.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(extd)
!            A Gaussian distributed random real number of kind extd.
!
!
! Further Details
! _______________
!
!   This subroutine uses the Cumulative Density Function (CDF) inversion method to generate
!   a Gaussian random ral number of kind extd. Starting with a random number produced by
!   the STATPACK uniform random number generator that can produce random numbers with the
!   uniform distribution over the continuous range (0, 1) (denoted U(0, 1)), the CDF
!   method simply inverts the CDF of a standard Gaussian distribution to produce a
!   standard Gaussian  (e.g. a Gaussian distribution with mean zero and standard
!   deviation one) random real number of kind extd.
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 7) by the subroutine PPND16 described in the reference (1).
!   This method gives about 16 decimal digits of accuracy in the range [10**(-316), 1-10**(-316)]
!   if computations are done in double or higher precision.
!
!   For more details on Uniform and Gaussian random number generators or the approximation
!   of the inverse Gaussian CDF used here, see :
!
!   (1) Devroye, L., 1986:
!           Non-Uniform Random Variate Generation. Springer-Verlag,
!           http://cg.scs.carleton.ca/~luc/rnbookindex.html, New York.
!
!   (2) Thomas, D.B., Luk, W., Leong, P.H.W., and Villasenor, J.D., 2007:
!           Gaussian random number generators.
!           ACM Comput. Surv. 39, 4, Article 11 (October 2007), 38 pages,
!           DOI = 10.1145/1287620.1287622 (http://doi.acm.org/10.1145/1287620.1287622)
!
!   (3) Wichura, M.J., 1988:
!           Algorithm AS 241: The percentage points of the normal distribution.
!           Appl. Statis. 37, 3, 477-484.
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Prob_Procedures, only : pinvn2
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(extd), intent(out) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(1) :: harvest1
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    call random_r1( harvest1 )
!
#ifdef _RANDOM_WITH0
    if ( harvest1(1)<=zero )  harvest1(1) = nearest( zero, one )
#endif
!
    harvest = pinvn2( real( harvest1(1), extd) )
!
!
! END OF SUBROUTINE normal_random2_r0
! ___________________________________
!
    end subroutine normal_random2_r0
!
! =========================================================================================
!
    subroutine normal_random2_r1( harvest )
!
! purpose
! _______
!
!   This subroutine returns a random real vector of kind extd following the standard
!   normal (Gaussian) distribution.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(extd), dimension(:)
!            A Gaussian distributed random real vector of kind extd.
!
!
! Further Details
! _______________
!
!   This subroutines uses the Cumulative Density Function (CDF) inversion method to generate
!   Gaussian random real numbers of kind extd. Starting with random numbers produced by the
!   STATPACK uniform random number generator that can produce random numbers with the
!   uniform distribution over the continuous range (0, 1) (denoted U(0, 1)), the CDF
!   method simply inverts the CDF of a standard Gaussian distribution to produce
!   standard Gaussian  (e.g. a Gaussian distribution with mean zero and standard
!   deviation one) random real numbers of kind extd.
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 7) by the subroutine PPND16 described in the reference (1).
!   This method gives about 16 decimal digits of accuracy in the range [10**(-316), 1-10**(-316)]
!   if computations are done in double or higher precision.
!
!   For more details on Uniform and Gaussian random number generators or the approximation
!   of the inverse Gaussian CDF used here, see :
!
!   (1) Devroye, L., 1986:
!           Non-Uniform Random Variate Generation. Springer-Verlag,
!           http://cg.scs.carleton.ca/~luc/rnbookindex.html, New York.
!
!   (2) Thomas, D.B., Luk, W., Leong, P.H.W., and Villasenor, J.D., 2007:
!           Gaussian random number generators.
!           ACM Comput. Surv. 39, 4, Article 11 (October 2007), 38 pages,
!           DOI = 10.1145/1287620.1287622 (http://doi.acm.org/10.1145/1287620.1287622)
!
!   (3) Wichura, M.J., 1988:
!           Algorithm AS 241: The percentage points of the normal distribution.
!           Appl. Statis. 37, 3, 477-484.
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _ALLOC
    use Char_Constants,  only : allocate_error
#endif
    use Prob_Procedures, only : pinvn2
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(extd), intent(out), dimension(:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _ALLOC
    real(stnd), dimension(:), allocatable :: harvest1
!
    integer      :: iok
!
#else
    real(stnd), dimension(size(harvest)) :: harvest1
#endif
!
    integer(i4b) :: m
!
!
! PARAMETERS
! __________
!
#ifdef _ALLOC
    character(len=*),  parameter :: name_proc='normal_random_number2_'
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( harvest )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b )  return
!
#ifdef _ALLOC
!
!   ALLOCATE WORK VARIABLE.
!
    allocate( harvest1(m), stat = iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
#endif
!
    call random_r1( harvest1(:m) )
!
#ifdef _RANDOM_WITH0
    where ( harvest1(:m)<=zero )  harvest1(:m) = nearest( zero, one )
#endif
!
    harvest(:m) = pinvn2( real( harvest1(:m), extd ) )
!
#ifdef _ALLOC
!
!   DEALLOCATE WORK VARIABLE.
!
    deallocate( harvest1 )
#endif
!
!
! END OF SUBROUTINE normal_random2_r1
! __________________________________
!
    end subroutine normal_random2_r1
!
! =========================================================================================
!
    subroutine normal_random2_r2( harvest )
!
! purpose
! _______
!
!   This subroutine returns a random real matrix of kind extd following the standard
!   normal (Gaussian) distribution.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(extd), dimension(:,:)
!            A Gaussian distributed random real matrix of kind extd.
!
!
! Further Details
! _______________
!
!   This subroutines uses the Cumulative Density Function (CDF) inversion method to generate
!   Gaussian random real numbers of kind extd. Starting with random numbers produced by the
!   STATPACK uniform random number generator that can produce random numbers with the
!   uniform distribution over the continuous range (0, 1) (denoted U(0, 1)), the CDF
!   method simply inverts the CDF of a standard Gaussian distribution to produce
!   standard Gaussian  (e.g. a Gaussian distribution with mean zero and standard
!   deviation one) random real numbers of kind extd.
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 7) by the subroutine PPND16 described in the reference (1).
!   This method gives about 16 decimal digits of accuracy in the range [10**(-316), 1-10**(-316)]
!   if computations are done in double or higher precision.
!
!   For more details on Uniform and Gaussian random number generators or the approximation
!   of the inverse Gaussian CDF used here, see :
!
!   (1) Devroye, L., 1986:
!           Non-Uniform Random Variate Generation. Springer-Verlag,
!           http://cg.scs.carleton.ca/~luc/rnbookindex.html, New York.
!
!   (2) Thomas, D.B., Luk, W., Leong, P.H.W., and Villasenor, J.D., 2007:
!           Gaussian random number generators.
!           ACM Comput. Surv. 39, 4, Article 11 (October 2007), 38 pages,
!           DOI = 10.1145/1287620.1287622 (http://doi.acm.org/10.1145/1287620.1287622)
!
!   (3) Wichura, M.J., 1988:
!           Algorithm AS 241: The percentage points of the normal distribution.
!           Appl. Statis. 37, 3, 477-484.
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _ALLOC
    use Char_Constants,  only : allocate_error
#endif
    use Prob_Procedures, only : pinvn2
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(extd), intent(out), dimension(:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _ALLOC
    real(stnd), dimension(:), allocatable :: harvest1
#else
    real(stnd), dimension(size(harvest,1)) :: harvest1
#endif
!
    integer(i4b) :: m, n, i
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
#ifdef _ALLOC
    character(len=*),  parameter :: name_proc='normal_random_number2_'
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( harvest, 1 )
    n = size( harvest, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b )  return
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m*n>=omp_limit                      .and.       &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par),PRIVATE(i,harvest1)
!
#ifdef _ALLOC
!
!   ALLOCATE WORK VARIABLE.
!
    allocate( harvest1(m), stat = i )
!
    if ( i/=0_i4b ) then
!
        call merror( name_proc//allocate_error )
!
    end if
#endif
!
!$OMP DO SCHEDULE(STATIC)
!
    do i = 1_i4b, n
!
        call random_r1( harvest1(:m) )
!
#ifdef _RANDOM_WITH0
        where ( harvest1(:m)<=zero )  harvest1(:m) = nearest( zero, one )
#endif
!
        harvest(:m,i) = pinvn2( real( harvest1(:m), extd ) )
!
    end do
!
!$OMP END DO
!
#ifdef _ALLOC
!
!   DEALLOCATE WORK VARIABLE.
!
    deallocate( harvest1 )
#endif
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE normal_random2_r2
! ___________________________________
!
    end subroutine normal_random2_r2
!
! =========================================================================================
!
    function normal_rand_number3() result( harvest )
!
! purpose
! _______
!
!   This function returns a Gaussian distributed random real number.
!
!
! Arguments
! _________
!
!   None
!
!
! Further Details
! _______________
!
!   This function uses the classical Box-Muller method to generate a Gaussian random real number.
!
!   For more details on Uniform and Gaussian random number generators or the Box-Muller method
!   used here, see :
!
!   (1) Devroye, L., 1986:
!           Non-Uniform Random Variate Generation. Springer-Verlag,
!           http://cg.scs.carleton.ca/~luc/rnbookindex.html, New York.
!
!   (2) Thomas, D.B., Luk, W., Leong, P.H.W., and Villasenor, J.D., 2007:
!           Gaussian random number generators.
!           ACM Comput. Surv. 39, 4, Article 11 (October 2007), 38 pages,
!           DOI = 10.1145/1287620.1287622 (http://doi.acm.org/10.1145/1287620.1287622)
!
!   (3) Brent, R.P., 1993:
!           Fast Normal Random Generators for vector processors.
!           Report TR-CS-93-04, Computer Sciences Laboratory, Australian National University.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)               :: r, theta
    real(stnd), dimension(2) :: harvest1
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    call random_r1( harvest1(:2) )
!
#ifdef _RANDOM_WITH0
    r = sqrt( -two*log( one-harvest1(1) ) )
#else
    r = sqrt( -two*log( harvest1(1) ) )
#endif
    theta   = twopi*harvest1(2)
    harvest = r*sin( theta )
!
!
! END OF FUNCTION normal_rand_number3
! ___________________________________
!
    end function normal_rand_number3
!
! =========================================================================================
!
    subroutine normal_random3_r0( harvest )
!
! purpose
! _______
!
!   This subroutine returns a random real number HARVEST following the standard
!   Gaussian distribution.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd)
!            A Gaussian distributed random real number.
!
!
! Further Details
! _______________
!
!   This subroutine uses the classical Box-Muller method to generate a Gaussian random real number.
!
!   For more details on Uniform and Gaussian random number generators or the Box-Muller method
!   used here, see :
!
!   (1) Devroye, L., 1986:
!           Non-Uniform Random Variate Generation. Springer-Verlag,
!           http://cg.scs.carleton.ca/~luc/rnbookindex.html, New York.
!
!   (2) Thomas, D.B., Luk, W., Leong, P.H.W., and Villasenor, J.D., 2007:
!           Gaussian random number generators.
!           ACM Comput. Surv. 39, 4, Article 11 (October 2007), 38 pages,
!           DOI = 10.1145/1287620.1287622 (http://doi.acm.org/10.1145/1287620.1287622)
!
!   (3) Brent, R.P., 1993:
!           Fast Normal Random Generators for vector processors.
!           Report TR-CS-93-04, Computer Sciences Laboratory, Australian National University.
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)               :: r, theta
    real(stnd), dimension(2) :: harvest1
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    call random_r1( harvest1(:2) )
!
#ifdef _RANDOM_WITH0
    r = sqrt( -two*log( one-harvest1(1) ) )
#else
    r = sqrt( -two*log( harvest1(1) ) )
#endif
    theta   = twopi*harvest1(2)
    harvest = r*sin( theta )
!
!
! END OF SUBROUTINE normal_random3_r0
! ___________________________________
!
    end subroutine normal_random3_r0
!
! =========================================================================================
!
    subroutine normal_random3_r1( harvest )
!
! purpose
! _______
!
!   This subroutine returns a random real vector HARVEST following the standard
!   normal (Gaussian) distribution.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd), dimension(:)
!            A Gaussian distributed random real vector.
!
!
! Further Details
! _______________
!
!   This subroutine uses the classical Box-Muller method to generate Gaussian random real numbers.
!   The computations are parallelized if OPENMP is used.
!
!   For more details on Uniform and Gaussian random number generators or the Box-Muller method
!   used here, see :
!
!   (1) Devroye, L., 1986:
!           Non-Uniform Random Variate Generation. Springer-Verlag,
!           http://cg.scs.carleton.ca/~luc/rnbookindex.html, New York.
!
!   (2) Thomas, D.B., Luk, W., Leong, P.H.W., and Villasenor, J.D., 2007:
!           Gaussian random number generators.
!           ACM Comput. Surv. 39, 4, Article 11 (October 2007), 38 pages,
!           DOI = 10.1145/1287620.1287622 (http://doi.acm.org/10.1145/1287620.1287622)
!
!   (3) Brent, R.P., 1993:
!           Fast Normal Random Generators for vector processors.
!           Report TR-CS-93-04, Computer Sciences Laboratory, Australian National University.
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _ALLOC
    use Char_Constants,  only : allocate_error
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out), dimension(:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: r, theta
!
#ifdef _ALLOC
    real(stnd), dimension(:), allocatable :: harvest1
!
#else
    real(stnd), dimension(size(harvest)+1) :: harvest1
#endif
!
    integer(i4b) :: m, j
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
#ifdef _ALLOC
    character(len=*),  parameter :: name_proc='normal_random_number3_'
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( harvest )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b )  return
!
#ifdef _ALLOC
!
!   ALLOCATE WORK VARIABLE.
!
    allocate( harvest1(m+1_i4b), stat = j )
!
    if ( j/=0_i4b ) then
!
        call merror( name_proc//allocate_error )
!
    end if
#endif
!
    call random_r1( harvest1(:m+1_i4b) )
!
#ifdef _OPENMP
    j           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m>=omp_limit                        .and.       &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)     &
!$OMP            ,PRIVATE(j,r,theta)
!
    do j = 1_i4b, m-1_i4b, 2_i4b
!
#ifdef _RANDOM_WITH0
        r = sqrt( -two*log( one-harvest1(j) ) )
#else
        r = sqrt( -two*log( harvest1(j) ) )
#endif
!
        theta            = twopi*harvest1(j+1_i4b)
        harvest(j)       = r*sin( theta )
        harvest(j+1_i4b) = r*cos( theta )
!
    end do
!
!$OMP END PARALLEL DO
!
    if ( m/=2_i4b*(m/2_i4b) ) then
!
#ifdef _RANDOM_WITH0
        r = sqrt( -two*log( one-harvest1(m) ) )
#else
        r = sqrt( -two*log( harvest1(m) ) )
#endif
!
        theta      = twopi*harvest1(m+1_i4b)
        harvest(m) = r*sin( theta )
!
    end if
!
#ifdef _ALLOC
!
!   DEALLOCATE WORK VARIABLE.
!
    deallocate( harvest1 )
#endif
!
!
! END OF SUBROUTINE normal_random3_r1
! ___________________________________
!
    end subroutine normal_random3_r1
!
! =========================================================================================
!
    subroutine normal_random3_r2( harvest )
!
! purpose
! _______
!
!   This subroutine returns a random matrix HARVEST following the standard
!   normal (Gaussian) distribution.
!
!
! Arguments
! _________
!
!   HARVEST  (OUTPUT) real(stnd), dimension(:,:)
!            A Gaussian distributed random real matrix.
!
!
! Further Details
! _______________
!
!   This subroutine uses the classical Box-Muller method to generate Gaussian random real numbers.
!   The computations are parallelized if OPENMP is used.
!
!   For more details on Uniform and Gaussian random number generators or the Box-Muller method
!   used here, see :
!
!   (1) Devroye, L., 1986:
!           Non-Uniform Random Variate Generation. Springer-Verlag,
!           http://cg.scs.carleton.ca/~luc/rnbookindex.html, New York.
!
!   (2) Thomas, D.B., Luk, W., Leong, P.H.W., and Villasenor, J.D., 2007:
!           Gaussian random number generators.
!           ACM Comput. Surv. 39, 4, Article 11 (October 2007), 38 pages,
!           DOI = 10.1145/1287620.1287622 (http://doi.acm.org/10.1145/1287620.1287622)
!
!   (3) Brent, R.P., 1993:
!           Fast Normal Random Generators for vector processors.
!           Report TR-CS-93-04, Computer Sciences Laboratory, Australian National University.
!
!
! __________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _ALLOC
    use Char_Constants,  only : allocate_error
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out), dimension(:,:) :: harvest
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: r, theta
!
#ifdef _ALLOC
    real(stnd), dimension(:), allocatable :: harvest1
!
#else
    real(stnd), dimension(size(harvest,1)+1) :: harvest1
#endif
!
    integer(i4b) :: m, n, j, i
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
#ifdef _ALLOC
    character(len=*),  parameter :: name_proc='normal_random_number3_'
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( harvest, 1 )
    n = size( harvest, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b )  return
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m*n>=omp_limit                      .and.       &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par),PRIVATE(i,j,r,theta,harvest1)
!
#ifdef _ALLOC
!
!   ALLOCATE WORK VARIABLE.
!
    allocate( harvest1(m+1_i4b), stat = i )
!
    if ( i/=0_i4b ) then
!
        call merror( name_proc//allocate_error )
!
    end if
#endif
!
!$OMP DO SCHEDULE(STATIC)
!
    do i = 1_i4b, n
!
        call random_r1( harvest1(:m+1_i4b) )
!
        do j = 1_i4b, m-1_i4b, 2_i4b
!
#ifdef _RANDOM_WITH0
            r = sqrt( -two*log( one-harvest1(j) ) )
#else
            r = sqrt( -two*log( harvest1(j) ) )
#endif
!
            theta              = twopi*harvest1(j+1_i4b)
            harvest(j,i)       = r*sin( theta )
            harvest(j+1_i4b,i) = r*cos( theta )
!
        end do
!
        if ( m/=2_i4b*(m/2_i4b) ) then
!
#ifdef _RANDOM_WITH0
            r = sqrt( -two*log( one-harvest1(m) ) )
#else
            r = sqrt( -two*log( harvest1(m) ) )
#endif
!
            theta        = twopi*harvest1(m+1_i4b)
            harvest(m,i) = r*sin( theta )
!
        end if
!
    end do
!
!$OMP END DO
!
#ifdef _ALLOC
!
!   DEALLOCATE WORK VARIABLE.
!
    deallocate( harvest1 )
#endif
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE normal_random3_r2
! ___________________________________
!
    end subroutine normal_random3_r2
!
!
! =========================================================================================
!               PSEUDO-RANDOM ORTHOGONAL MATRICES FROM THE HAAR DISTRIBUTION
!               PSEUDO-RANDOM SYMMETRIC MATRICES WITH A PRESCRIBED SPECTRUM
!               PSEUDO-RANDOM MATRICES WITH A PRESCRIBED SINGULAR VALUE DISTRIBUTION
! =========================================================================================
!
!
    subroutine random_qr_cmp( mat, diagr, beta, fillr, initseed )
!
! Purpose
! _______
!
!   RANDOM_QR_CMP generates the first k columns of a pseudo-random QR factorization of a
!   hypothetical real n-by-n matrix MAT, whose elements follow independently a Laplace_Gauss(0;1)
!   distribution (e.g. the standard normal distribution):
!
!                              MAT = Q * R
!
!   where Q is a pseudo-random orthogonal matrix following the Haar distribution from the group
!   of orthogonal matrices and R is upper triangular. The upper-diagonal elements of R follow a
!   Laplace_Gauss(0;1) distribution (e.g. the standard normal distribution) and the squares
!   of the diagonal elements of R, (R(i,i))**(2) follow a chi-squared distribution with n-i+1
!   degrees of freedom.
!
!
! Arguments
! _________
!
!   MAT      (OUTPUT) real(stnd), dimension(:,:)
!            On exit, the elements above the diagonal of the array MAT contain  
!            the corresponding elements of R if FILLR is present and set to true;
!            otherwise the upper-diagonal elements of MAT are not modified.
!            The elements on and below the diagonal, with the arrays BETA and DIAGR, 
!            represent the first k columns (with k=size(MAT,2) and k<=n=size(MAT,1) )
!            of a pseudo-random orthogonal matrix Q following the Haar distribution
!            as a product of elementary reflectors and a diagonal matrix,
!            see Further Details.
!
!            The shape of MAT must verify:
!
!            - size( MAT, 2 ) <= size( MAT, 1 ).
!
!   DIAGR    (OUTPUT) real(stnd), dimension(:) 
!            On exit, the diagonal elements of the matrix R,
!            see Further Details.
!
!            The size of DIAGR must verify:
!
!            - size( DIAGR ) = size( MAT, 2 ) <= size( MAT, 1 ).
!
!   BETA     (OUTPUT) real(stnd), dimension(:)
!            On exit, the scalars factors of the elementary reflectors,
!            see Further Details.
!
!            The size of BETA must verify:
!
!            - size( BETA ) = size( DIAGR ) = size( MAT, 2 ) <= size( MAT, 1 ).
!
!   FILLR    (INPUT, OPTIONAL) logical(lgl)
!            on entry, if FILLR is set to true, the super-diagonal elements of R
!            are generated in MAT on exit.
!            If FILLR is set to false, the super-diagonal elements of MAT are not
!            modified or used.
!
!            The default is FILLR = false.
!
!   INITSEED (INPUT, OPTIONAL) logical(lgl)
!            On entry, if INITSEED=true, a call to RANDOM_SEED_() without arguments
!            is done in the subroutine, in order to initiate a non-repeatable reset
!            of the seed used by the STATPACK random generator.
!
!            The default is INITSEED = false.
!
!
! Further Details
! _______________
!
!   RANDOM_QR_CMP uses the method described in the reference (1), based on Householder
!   transformations, for generating the first k columns of a n-by-n pseudo-random orthogonal
!   matrices Q distributed according to the Haar measure over the orthogonal group of order n,
!   in a factored form.
!
!   The pseudo-random orthogonal matrix Q is represented as a product of n elementary reflectors
!   and of a diagonal matrix
!
!            Q = H(1) * H(2) * ... * H(n) * diag(sign(DIAGR))
!
!   Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!   where beta is a real scalar and v is a real n-element vector with v(1:i-1) = 0.
!   v(i:n) is stored on exit in MAT(i:n,i) and beta in BETA(i).
!
!   diag(sign(DIAGR)) is the n-by-n diagonal matrix with diagonal elements equal to
!   sign( one, DIAGR) (e.g., its ith diagonal element equals to one if DIAGR(i) is positive
!   and -one otherwise).
!
!   It is possible to compute only the first k columns of Q and R by restricting the number
!   of columns of MAT and the sizes of DIAGR and BETA on entry of the subroutine.
!
!   Finally, note that the computations are parallelized if OPENMP is used.
!
!   Q can be generated with the help of subroutine ORTHO_GEN_RANDOM_QR or can be applied
!   to a vector or a matrix with the help of subroutine APPLY_Q_QR in module QR_Procedures.
!
!   For further details on the QR factorization and its use or pseudo-random orthogonal
!   matrices distributed according to the Haar measure over the orthogonal group, see:
!
!   (1) Stewart, G.W., 1980:
!          The efficient generation of random orthogonal matrices with
!          an application to condition estimators.
!          SIAM J. Numer. Anal., 17, 403-409
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!          Solving least square problems.
!          Prentice-Hall.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,          only : assert, assert_eq
    use Hous_Procedures,    only : hous1
    use Prob_Procedures,    only : pinvn
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(out), dimension(:,:) :: mat
    real(stnd),   intent(out), dimension(:)   :: diagr, beta
!
    logical(lgl), intent(in),  optional :: fillr, initseed
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension( size(mat,1) ) :: tmp
!
    integer(i4b) :: n, k, j
!
    logical(lgl) :: fill_r, init_seed
!
#ifndef _USE_PGI
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='random_qr_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = size( mat, 1 )
!
    k = assert_eq( int(size(beta),i4b),   &
                   int(size(diagr),i4b),  &
                   int(size(mat,2),i4b),  &
                   name_proc              )
!
    call assert( logical( k<=n, lgl ), name_proc )
!
    init_seed = false
!
    if ( present(initseed) ) then
!
        init_seed = initseed
!
    end if
!
    fill_r = false
!
    if ( present(fillr) ) then
!
        fill_r = fillr
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. k<=0_i4b ) return
!
#ifndef _USE_PGI
#ifdef _OPENMP
    j           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  n*k>=omp_limit                      .and.       &
                  j>1_i4b
#endif
#endif
!
!   INITIALIZE THE RANDOM GENERATOR IF REQUIRED.
!
    if ( init_seed ) then
!
#ifndef _USE_PGI
!$OMP SINGLE
#endif
        call random_seed_()
#ifndef _USE_PGI
!$OMP END SINGLE
#endif
!
    end if
!
    if ( fill_r ) then
!
!       GENERATE A RANDOM ORTHOGONAL MATRIX AND AN UPPER
!       TRIANGULAR MATRIX IN FACTORED FORM.
!
#ifndef _USE_PGI
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)     &
!$OMP            ,PRIVATE(j,tmp)
#endif
!
        do j = 1_i4b, k
!
            call random_r1( tmp(1_i4b:n) )
!
#ifdef _RANDOM_WITH0
            tmp(1_i4b:n) = max( tmp(1_i4b:n), nearest( zero, one ) )
#endif
!    
!           GENERATE UPPER DIAGONAL ELEMENTS OF TRIANGULAR MATRIX.
!
            mat(i4b:n,j) = pinvn( tmp(i4b:n) )
!    
!           GENERATE ELEMENTARY REFLECTOR H(j) TO ANNIHILATE mat(j+1:n,j) .
!    
            call hous1( mat(j:n,j), beta(j), diagr(j) )
!
        end do
!
#ifndef _USE_PGI
!$OMP END PARALLEL DO
#endif
!
    else
!
!       GENERATE A RANDOM ORTHOGONAL MATRIX IN FACTORED FORM.
!
#ifndef _USE_PGI
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)     &
!$OMP            ,PRIVATE(j,tmp)
#endif
!
        do j = 1_i4b, k
!
            call random_r1( tmp(j:n) )
!
#ifdef _RANDOM_WITH0
            tmp(j:n) =  max( tmp(j:n), nearest( zero, one ) )
#endif
!
            mat(j:n,j) = pinvn( tmp(j:n) )
!    
!           GENERATE ELEMENTARY REFLECTOR H(j) TO ANNIHILATE mat(j+1:n,j) .
!    
            call hous1( mat(j:n,j), beta(j), diagr(j) )
!
        end do
!
#ifndef _USE_PGI
!$OMP END PARALLEL DO
#endif
!
    end if
!
!
! END OF SUBROUTINE random_qr_cmp
! _______________________________
!
    end subroutine random_qr_cmp
!
! =========================================================================================
!
    subroutine ortho_gen_random_qr( mat, diagr, beta )
!
! Purpose
! _______
!
!   ORTHO_GEN_RANDOM_QR generates a n-by-n real pseudo-random orthogonal matrix following
!   the Haar distribution, which is defined as the product of n elementary reflectors of
!   order n and of a n-by-n diagonal matrix with diagonal elements equal to sign( one, DIAGR):
!
!            Q = H(1) * H(2) * ... * H(n) * diag(sign(DIAGR))
!
!   as returned by RANDOM_QR_CMP.
!
!   Optionally, it is possible to generate only the first k columns of Q by restricting
!   arguments MAT, BETA and DIAGR to the first k columns or elements of the corresponding
!   arguments as returned by RANDOM_QR_CMP.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the i-th column must contain the vector which
!            defines the elementary reflector H(i), for i = 1,2,...,k,
!            ( with k<=n=size(MAT,1) and k=size(MAT,2) ) as returned by RANDOM_QR_CMP
!            in its array argument MAT.
!            On exit, the first k columns of the pseudo-random n-by-n orthogonal
!            matrix Q.
!
!            The shape of MAT must verify: size( MAT, 2 ) <= size( MAT, 1 ).
!
!   DIAGR    (INPUT) real(stnd), dimension(:) 
!            On entry, the diagonal elements of the matrix R, as returned
!            by RANDOM_QR_CMP in its argument DIAGR.
!
!            The size of DIAGR must verify: size( DIAGR ) = size( MAT, 2 ).
!
!   BETA     (INPUT) real(stnd), dimension(:)
!            On entry, BETA(i) must contain the scalar factor of the elementary
!            reflector H(i), as returned by RANDOM_QR_CMP in its argument BETA.
!
!            The size of BETA must verify: size( BETA ) = size( DIAGR ) = size( MAT, 2 ) .
!
!
! Further Details
! _______________
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in the lower triangle
!   of MAT and generating the pseudo-random orthogonal matrix Q of the QR factorization
!   returned by subroutine RANDOM_QR_CMP.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the QR factorization and its use or the blocked
!   algorithm used here, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,     only : assert, assert_eq
    use QR_Procedures, only : ortho_gen_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: diagr, beta
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, k, j
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='ortho_gen_random_qr'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = size( mat, 1 )
!
    k = assert_eq( int(size(beta),i4b),   &
                   int(size(diagr),i4b),  &
                   int(size(mat,2),i4b),  &
                   name_proc              )
!
    call assert( logical( k<=n, lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. k<=0_i4b ) return
!
!   GENERATE THE RANDOM ORTHOGONAL MATRIX.
!
    call ortho_gen_qr( mat(:n,:k), beta(:k) )
!
    do j = 1_i4b, k
!
        if ( diagr(j) > zero ) cycle
!
        mat(:n,j) = -mat(:n,j)
!
    end do
!
!
! END OF SUBROUTINE ortho_gen_random_qr
! _____________________________________
!
    end subroutine ortho_gen_random_qr
!
! =========================================================================================
!
    subroutine gen_random_sym_mat( eigval, mat, eigvec, initseed, hous )
!
! Purpose
! _______
!
!   GEN_RANDOM_SYM_MAT generates a pseudo-random n-by-n real symmetric matrix
!   with prescribed eigenvalues.
!
!   Optionally, the corresponding eigenvectors of the generated
!   pseudo-random n-by-n real symmetric matrix can be output if required.
!
!
! Arguments
! _________
!
!   EIGVAL   (INPUT) real(stnd), dimension(:)
!            On entry, the prescribed eigenvalues of the pseudo-random n-by-n
!            real symmetric matrix.
!            IF size(EIGVAL)<n, the remaining eigenvalues are assumed to be zero.
!
!            The size of EIGVAL must verify:
!
!            - size( EIGVAL ) <= size( MAT, 1 ) = size( MAT, 2 ) = n . 
!
!   MAT      (OUTPUT) real(stnd), dimension(:,:)
!            On exit, the pseudo-random n-by-n real symmetric matrix.
!
!            The shape of MAT must verify:
!
!            - size( MAT, 2 ) = size( MAT, 1 ) = n .
!
!   EIGVEC   (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, the pseudo-random eigenvectors corresponding
!            to the eigenvalues prescribed in EIGVAL. The eigenvectors
!            are returned columnwise.
!
!            The shape of EIGVEC must verify:
!
!            - size( EIGVEC, 1 ) = size( MAT, 1 ) = size( MAT, 2 ) = n ;
!            - size( EIGVEC, 2 ) = size( EIGVAL ) .
!
!   INITSEED (INPUT, OPTIONAL) logical(lgl)
!            On entry, if INITSEED=true, a call to RANDOM_SEED_() without arguments
!            is done in the subroutine, in order to initiate a non-repeatable reset
!            of the seed used by the STATPACK random generator.
!
!            The default is INITSEED = false.
!
!   HOUS     (INPUT, OPTIONAL) logical(lgl)
!            On entry, if:
!
!            - HOUS=true, the pseudo-random eigenvectors are
!              generated as Householder transformation matrices ; 
!            - HOUS=false, the pseudo-random eigenvectors are
!              generated as pseudo-random orthogonal matrices following
!              the Haar distribution from the group of orthogonal matrices.
!
!            The default is HOUS = true.
!
!
! Further Details
! _______________
!
!   Pseudo-random eigenvectors are generated as a pseudo-random orthogonal matrix
!   following the Haar distribution from the group of orthogonal matrices and are
!   computed with the help of subroutines RANDOM_QR_CMP and ORTHO_GEN_RANDOM_QR if
!   the logical optional argument HOUS has the value false. If the logical optional argument
!   HOUS has the value true, they are generated as Householder transformation matrices
!   with the help of function GEN_RANDOM_ORTHO_MAT.
!
!   These computed pseudo-random eigenvectors and the corresponding prescribed
!   eigenvalues are then used to generate a pseudo-random n-by-n symmetric matrix
!   whose eigenvalues are the prescribed eigenvalues.
!
!   These computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   For further details on Householder matrices and pseudo-random orthogonal matrices
!   following the Haar distribution, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!          Solving least square problems.
!          Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Stewart, G.W., 1980:
!          The efficient generation of random orthogonal matrices with
!          an application to condition estimators.
!          SIAM J. Numer. Anal., 17, 403-409
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),            dimension(:)   :: eigval
    real(stnd),   intent(out),           dimension(:,:) :: mat
    real(stnd),   intent(out), optional, dimension(:,:) :: eigvec
!
    logical(lgl), intent(in),  optional :: initseed, hous
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),  allocatable, dimension(:)   :: beta, diagr
    real(stnd),  allocatable, dimension(:,:) :: eigvec0, eigvec00
!
    integer      :: iok
    integer(i4b) :: n, k, i
!
    logical(lgl)  :: use_hous, test_eigvec
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='gen_random_sym_mat'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    k = size( eigval )
!
    n = assert_eq( int(size(mat,1),i4b),  &
                   int(size(mat,2),i4b),  &
                   name_proc              )
!
    call assert( logical( k<=n, lgl ), name_proc )
!
    test_eigvec = false
!
    if ( present(eigvec) ) then
!
        call assert( logical( int(size(eigvec,1),i4b)==n, lgl ),  &
                     logical( int(size(eigvec,2),i4b)==k, lgl ),  &
                     name_proc )
!
        test_eigvec = true
!
    end if
!
    use_hous = true
!
    if ( present(hous) ) then
!
        use_hous = hous
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. k<=0_i4b ) return
!
    if ( use_hous ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( eigvec00(n,k), stat=iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
!       GENERATE THE PSEUDO-RANDOM EIGENVECTORS AS AN HOUSEHOLDER MATRIX.
!
        eigvec00(:n,:k) = gen_random_ortho_mat( n, k, initseed=initseed )
!
    else
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigvec00(n,k), beta(k), diagr(k), stat=iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
!       GENERATE THE PSEUDO-RANDOM EIGENVECTORS FROM THE HAAR DISTRIBUTION.
!
        call random_qr_cmp( eigvec00(:n,:k), diagr(:k), beta(:k), initseed=initseed )
!
        call ortho_gen_random_qr( eigvec00(:n,:k), diagr(:k), beta(:k) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( beta, diagr )
!
    end if
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  n*k>=omp_limit                      .and.       &
                  i>1_i4b
#endif
!
    if ( test_eigvec ) then
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)     &
!$OMP            ,PRIVATE(i)
!
        do i = 1_i4b, k
!
            eigvec(:n,i) = eigvec00(:n,i)
!
#ifdef _BLAS
            call scal( n, eigval(i), eigvec00(:n,i), 1_i4b )
#else
            eigvec00(:n,i) = eigval(i)*eigvec00(:n,i)
#endif
!
        end do
!
!$OMP END PARALLEL DO
!
!       FORM THE SYMMETRIC MATRIX FROM THE SPECIFIED EIGENVALUE DISTRIBUTION AND RANDOM EIGENVECTORS.
!
#ifdef _BLAS
        call gemm( 'N', 'T', n, n, k, one, eigvec(1_i4b:n,1_i4b:k), n,          &
                   eigvec00(1_i4b:n,1_i4b:k), n, zero, mat(1_i4b:n,1_i4b:n), n  )
#else
        mat(:n,:n) = matmul( eigvec(:n,:k), transpose(eigvec00(:n,:k)) )
#endif
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( eigvec00 )
!
    else
!
!       ALLOCATE WORK ARRAY.
!
        allocate( eigvec0(n,k), stat=iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)     &
!$OMP            ,PRIVATE(i)
!
        do i = 1_i4b, k
!
            eigvec0(:n,i) = eigval(i)*eigvec00(:n,i)
!
        end do
!
!$OMP END PARALLEL DO
!
!       FORM THE SYMMETRIC MATRIX FROM THE SPECIFIED EIGENVALUE DISTRIBUTION AND RANDOM EIGENVECTORS.
!
#ifdef _BLAS
        call gemm( 'N', 'T', n, n, k, one, eigvec0(1_i4b:n,1_i4b:k), n,         &
                   eigvec00(1_i4b:n,1_i4b:k), n, zero, mat(1_i4b:n,1_i4b:n), n  )
#else
        mat(:n,:n) = matmul( eigvec0(:n,:k), transpose(eigvec00(:n,:k)) )
#endif
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigvec0, eigvec00 )
!
    end if
!
!
! END OF SUBROUTINE gen_random_sym_mat
! ____________________________________
!
    end subroutine gen_random_sym_mat
!
! =========================================================================================
!
    subroutine gen_random_mat( singval, mat, leftvec, rightvec, initseed, hous )
!
! Purpose
! _______
!
!   GEN_RANDOM_MAT generates a pseudo-random m-by-n real matrix
!   with prescribed singular values.
!
!   Optionally, the corresponding singular vectors of the generated
!   pseudo-random m-by-n real matrix can be output if required.
!
!
! Arguments
! _________
!
!   SINGVAL  (INPUT) real(stnd), dimension(:)
!            On entry , the prescribed singular values of the pseudo-random m-by-n
!            real matrix.
!
!            The prescribed singular values must be greater or equal to zero. Furthermore,
!            iF size(SINGVAL)<min(m,n), the remaining singular values are assumed to be zero.
!
!            The size of SINGVAL must verify:
!
!            - size( SINGVAL ) <= min( size( MAT, 1 ), size( MAT, 2 ) ) = min( m, n) . 
!
!   MAT      (OUTPUT) real(stnd), dimension(:,:)
!            On exit, the pseudo-random m-by-n real matrix.
!
!   LEFTVEC  (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, the pseudo-random left singular vectors corresponding
!            to the singular values prescribed in SINGVAL. The left singular vectors
!            are returned columnwise.
!
!            The shape of LEFTVEC must verify:
!
!            - size( LEFTVEC, 1 ) = size( MAT, 1 ) =  m ;
!            - size( LEFTVEC, 2 ) = size( SINGVAL ) .
!
!   RIGHTVEC (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, the pseudo-random right singular vectors corresponding
!            to the singular values prescribed in SINGVAL. The right singular vectors
!            are returned columnwise.
!
!            The shape of RIGHTVEC must verify:
!
!            - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ;
!            - size( RIGHTVEC, 2 ) = size( SINGVAL ) .
!
!   INITSEED (INPUT, OPTIONAL) logical(lgl)
!            On entry, if INITSEED=true, a call to RANDOM_SEED_() without arguments
!            is done in the subroutine, in order to initiate a non-repeatable reset
!            of the seed used by the STATPACK random generator.
!
!            The default is INITSEED = false.
!
!   HOUS     (INPUT, OPTIONAL) logical(lgl)
!            On entry, if:
!
!            - HOUS=true, the pseudo-random singular vectors are
!              generated as Householder transformation matrices ; 
!            - HOUS=false, the pseudo-random singular vectors are
!              generated as pseudo-random orthogonal matrices following
!              the Haar distribution from the group of orthogonal matrices.
!
!            The default is HOUS = true.
!
!
! Further Details
! _______________
!
!   Pseudo-random singular vectors are generated as pseudo-random orthogonal matrices
!   following the Haar distribution from the group of orthogonal matrices and are
!   computed with the help of subroutines RANDOM_QR_CMP and ORTHO_GEN_RANDOM_QR if
!   the logical optional argument HOUS has the value false. If the logical optional argument
!   HOUS has the value true, they are generated as Householder transformation matrices
!   with the help of function GEN_RANDOM_ORTHO_MAT.
!
!   These computed pseudo-random singular vectors and the corresponding prescribed
!   singular values are then used to generate a pseudo-random m-by-n real matrix
!   whose singular values are the prescribed singular values.
!
!   These computations are parallelized if OPENMP is used.
!
!   For further details on Householder matrices and pseudo-random orthogonal matrices
!   following the Haar distribution, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!          Solving least square problems.
!          Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Stewart, G.W., 1980:
!          The efficient generation of random orthogonal matrices with
!          an application to condition estimators.
!          SIAM J. Numer. Anal., 17, 403-409
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error, random_error16
    use Utilities,       only : assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),            dimension(:)   :: singval
    real(stnd),   intent(out),           dimension(:,:) :: mat
    real(stnd),   intent(out), optional, dimension(:,:) :: leftvec, rightvec
!
    logical(lgl), intent(in),  optional :: initseed, hous
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), allocatable, dimension(:)   :: beta, diagr
    real(stnd), allocatable, dimension(:,:) :: leftvec0, rightvec0
!
    integer      :: iok
    integer(i4b) :: m, n, mn, k, i
!
    logical(lgl)  :: use_hous, test_leftvec, test_rightvec
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='gen_random_mat'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m  = size( mat, 1 )
    n  = size( mat, 2 )
    mn = min( m, n )
!
    k = size( singval )
!
    call assert( logical( k<=mn, lgl ), name_proc )
!
    if ( any( singval<zero ) ) then
!
        call merror( name_proc//random_error16 )
!
    end if
!
    test_leftvec = false
!
    if ( present(leftvec) ) then
!
        call assert( logical( int(size(leftvec,1),i4b)==m, lgl ),  &
                     logical( int(size(leftvec,2),i4b)==k, lgl ),  &
                     name_proc )
!
        test_leftvec = true
!
    end if
!
    test_rightvec = false
!
    if ( present(rightvec) ) then
!
        call assert( logical( int(size(rightvec,1),i4b)==n, lgl ),  &
                     logical( int(size(rightvec,2),i4b)==k, lgl ),  &
                     name_proc )
!
        test_rightvec = true
!
    end if
!
    use_hous = true
!
    if ( present(hous) ) then
!
        use_hous = hous
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( mn<=0_i4b .or. k<=0_i4b ) return
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( leftvec0(m,k), rightvec0(n,k), stat=iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
!
    if ( use_hous ) then
!
!       GENERATE THE PSEUDO-RANDOM LEFT SINGULAR VECTORS AS AN HOUSEHOLDER MATRIX.
!
        leftvec0(:m,:k) = gen_random_ortho_mat( m, k, initseed=initseed )
!
!       GENERATE THE PSEUDO-RANDOM RIGHT SINGULAR VECTORS AS AN HOUSEHOLDER MATRIX.
!
        rightvec0(:n,:k) = gen_random_ortho_mat( n, k, initseed=initseed )
!
    else
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( beta(k), diagr(k), stat=iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
!       GENERATE THE PSEUDO-RANDOM LEFT SINGULAR VECTORS FROM THE HAAR DISTRIBUTION.
!
        call random_qr_cmp( leftvec0(:m,:k), diagr(:k), beta(:k), initseed=initseed )
!
        call ortho_gen_random_qr( leftvec0(:m,:k), diagr(:k), beta(:k) )
!
!       GENERATE THE PSEUDO-RANDOM RIGHT SINGULAR VECTORS FROM THE HAAR DISTRIBUTION.
!
        call random_qr_cmp( rightvec0(:n,:k), diagr(:k), beta(:k), initseed=initseed )
!
        call ortho_gen_random_qr( rightvec0(:n,:k), diagr(:k), beta(:k) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( beta, diagr )
!
    end if
!
    if ( test_leftvec ) then
!
        leftvec(:m,:k) = leftvec0(:m,:k)
!
    end if
!
    if ( test_rightvec ) then
!
        rightvec(:n,:k) = rightvec0(:n,:k)
!
    end if
!
!   FORM THE m-BY-n REAL MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  n*k>=omp_limit                      .and.       &
                  i>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)     &
!$OMP            ,PRIVATE(i)
!
    do i = 1_i4b, k
!
#ifdef _BLAS
        call scal( n, singval(i), rightvec0(:n,i), 1_i4b )
#else
        rightvec0(:n,i) = singval(i)*rightvec0(:n,i)
#endif
!
    end do
!
!$OMP END PARALLEL DO
!
#ifdef _BLAS
    call gemm( 'N', 'T', m, n, k, one, leftvec0(1_i4b:m,1_i4b:k), m,          &
                rightvec0(1_i4b:n,1_i4b:k), n, zero, mat(1_i4b:m,1_i4b:n), m  )
#else
    mat(:m,:n) = matmul( leftvec0(:m,:k), transpose(rightvec0(:n,:k)) )
#endif
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( leftvec0, rightvec0 )
!
!
! END OF SUBROUTINE gen_random_mat
! ________________________________
!
    end subroutine gen_random_mat
!
! =========================================================================================
!
    function gen_random_ortho_mat( m, n, initseed ) result( mat )
!
! Purpose
! _______
!
!   GEN_RANDOM_ORTHO_MAT generates a pseudo-random m-by-n real matrix
!   with orthonormal columns.
!
!
! Arguments
! _________
!
!   M        (INPUT) integer(i4b)
!            On entry, the number of rows of the pseudo-random m-by-n real orthonormal matrix.
!
!   N        (INPUT) integer(i4b)
!            On entry, the number of columns of the pseudo-random m-by-n real orthonormal matrix.
!
!            N must verify:
!
!            - N <= M .
!
!   INITSEED (INPUT, OPTIONAL) logical(lgl)
!            On entry, if INITSEED=true, a call to RANDOM_SEED_() without arguments
!            is done in the subroutine, in order to initiate a non-repeatable reset
!            of the seed used by the STATPACK random generator.
!
!            The default is INITSEED = false.
!
!
! Further Details
! _______________
!
!   The pseudo-random m-by-n real matrix with orthonormal columns is generated from
!   the first n columns of an m-by-m Householder matrix H (e.g. a Householder transformation),
!   which is an orthogonal matrix defined as
!
!           H = I + tau * ( v * v' ) ,
!                      
!   where tau is a real scalar and v is a real m-element vector.
!   The scalar tau and vector v are computed to annilhate all the elements of a m-element
!   pseudo-random uniform vector, excepted the first element.      
!
!   These computations are parallelized if OPENMP is used.
!
!   For further details, on Householder matrices, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!          Solving least square problems.
!          Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, merror
    use Hous_Procedures, only : h1
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in) :: m, n
!
    real(stnd), dimension(m,n) :: mat
!
    logical(lgl), intent(in),  optional :: initseed
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: beta, tau
    real(stnd), allocatable, dimension(:)   :: u
    real(stnd), allocatable, dimension(:,:) :: mat2
!
    integer      :: iok
    integer(i4b) :: i
!
    logical(lgl) :: init_seed
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='gen_random_ortho_mat'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    call assert( logical( 0_i4b<=n, lgl ), name_proc )
    call assert( logical( n<=m, lgl ),     name_proc )
!
    init_seed = false
!
    if ( present(initseed) ) then
!
        init_seed = initseed
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n==0_i4b ) return
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( u(m), mat2(m,n), stat=iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
!
!   INITIALIZE THE RANDOM GENERATOR IF REQUIRED.
!
    if ( init_seed ) then
!
#ifndef _USE_PGI
!$OMP SINGLE
#endif
        call random_seed_()
#ifndef _USE_PGI
!$OMP END SINGLE
#endif
!
    end if
!
!   GENERATE A PSEUDO-RANDOM UNIFORM VECTOR.
!
    call random_r1( u(1_i4b:m) )
!    
!   GENERATE ELEMENTARY REFLECTOR H TO ANNIHILATE u(2:m) .
!    
    call h1( u(1_i4b:m), beta, tau )
!
!   GENERATE THE PSEUDO-RANDOM ORTHONORMAL MATRIX.
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m*n>=omp_limit                      .and.       &
                  i>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)     &
!$OMP            ,PRIVATE(i,beta)
!
!
    do i = 1_i4b, n
!
        beta = tau*u(i)
!
        mat2(1_i4b:m,i) = beta*u(1_i4b:m)
!
        mat2(i,i) = one + mat2(i,i)
!
    end do
!
!$OMP END PARALLEL DO
!
    mat(:m,:n) = mat2(:m,:n)
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( u, mat2 )
!
!
! END OF FUNCTION gen_random_ortho_mat
! ____________________________________
!
    end function gen_random_ortho_mat
!
!
! =========================================================================================
!                             RANDOMIZED LINEAR ALGEBRA
!             RANDOMIZED (PARTIAL) QR, COMPLETE ORTHOGONAL AND QB FACTORIZATIONS
!             RANDOMIZED (PARTIAL) ID, TWO-SIDED AND CUR DECOMPOSITIONS
! =========================================================================================
!
!
    subroutine partial_rqr_cmp( mat, diagr, beta, ip, krank, tol, tau, rng_alg, blk_size, nover )
!
! Purpose
! _______
!
!   PARTIAL_RQR_CMP computes a randomized (partial or complete) QR factorization with column
!   pivoting or orthogonal factorization of a m-by-n matrix MAT:
!
!                     MAT * P = Q * R
!
!   , where P is a n-by-n permutation matrix, R is an upper triangular or trapezoidal
!   (i.e., if n>m) m-by-n matrix and Q is a m-by-m orthogonal matrix.
!
!   At the user option, the randomized QR factorization can be only partial, e.g., the subroutine
!   ends when the numbers of columns of Q is equal to a predefined value equals to
!   kpartial = size( DIAGR ) = size( BETA ).
!
!   This leads implicitly to the following partition of Q:
!
!   [  Q1  Q2  ]
!
!   where Q1 is a m-by-kpartial orthonormal matrix and Q2 is a m-by-(m-kpartial) orthonormal
!   matrix orthogonal to Q1, and to the following corresponding partition of R:
!
!   [  R11  R12  ]
!
!   [  R21  R22  ]
!
!   where R11 is a kpartial-by-kpartial triangular matrix, R21 is zero by construction,
!   R12 is a full kpartial-by-(n-kpartial) matrix and R22 is a full (m-kpartial)-by-(n-kpartial)
!   matrix.
!
!   Then, if the optional scalar argument TOL is present and:
!
!   - is in ]0,1[, the rank of R11 is determined by finding the submatrix of R11
!     which is defined as the largest leading submatrix whose estimated condition
!     number, in the 1-norm, is less than 1/TOL.
!   - is equal to 0, the rank of R11, krank, is determined by finding the largest
!     submatrix of R11 such that abs(R11[j,j])>0. 
!
!   In both cases, the order of this submatrix, krank, is the effective rank of R11
!   (and MAT if krank is less than kpartial or if krank=kpartial=min(m,n)).
!
!   If TOL is absent or outside [0,1[, the rank of R11 is not checked and is assumed to be
!   equal to kpartial.
!
!   If krank is less than kpartial, then MAT is not of full rank (i.e., certain columns of
!   MAT(:m,:kpartial) are linear combinations of other columns of MAT(:m,:kpartial)) and
!   krank is also an estimate of the rank of MAT.
!
!   This leads to a redefinition of the partition of Q = [  Q1  Q2  ], where
!   Q1 and Q2 are now m-by-krank and m-by-(m-krank) orthonormal matrices, and a corresponding 
!   redefinition of the associated partition of R, where R11 is now a krank-by-krank 
!   triangular matrix, R21 is again zero by construction, R12 is a full krank-by-(n-krank)
!   matrix and R22 is a (m-krank)-by-(n-krank) matrix.
!
!   In a final step, if TAU is present, R22 is considered to be negligible
!   and R12 is annihilated by orthogonal transformations from the right, arriving at
!   the partial or complete orthogonal factorization:
!
!                       MAT * P = Q * T * Z
!
!   , where P is a n-by-n permutation matrix, Q is a m-by-m orthogonal matrix,
!   Z is a n-by-n orthogonal matrix and T is a m-by-n matrix and has the form:
!
!   [  T11  T12  ]
!
!   [  T21  T22  ]
!
!   Here T21 (=R21) and T12 are all zero, T22 (=R22) is considered to be negligible and
!   T11 is a krank-by-krank upper triangular matrix.
!   
!   On exit, P is stored compactly in the integer vector argument IP and if:
!   
!   - TAU is absent, PARTIAL_RQR_CMP computes Q and submatrices R11, R12 and R22. Submatrices R11,
!     R12 and R22 are stored in the array arguments MAT and DIAGR.
!     Q is stored compactly in factored form in the array arguments MAT and BETA.
!   
!   - TAU is present, PARTIAL_RQR_CMP computes Q, Z and submatrices T11 and T22 (=R22). Submatrices T11
!     and T22 are stored in the array arguments MAT and DIAGR.
!     Q is stored compactly in factored form in the array arguments MAT and BETA.
!     Z is stored compactly in factored form in the array arguments MAT and TAU.
!   
!   See Further Details for more information on how the (partial or complete) QR or orthogonal
!   decomposition is stored in MAT on exit.
!   
!   PARTIAL_RQR_CMP performs the same task as subroutine PARTIAL_QR_CMP in module QR_Procedures,
!   but is much faster on large matrices because of the use of a randomized and blocked "BLAS3"
!   algorithm instead of a standard "BLAS2" algorithm. Note, however, that PARTIAL_RQR_CMP is
!   an effective and efficient way for computing a low-rank approximation of MAT, but is less effective
!   to find the rank of MAT because of the use of randomization. As an illustration, the diagonal elements
!   of R11 are not necessarily of decreasing magnitude when computed by PARTIAL_RQR_CMP, while this property
!   is enforced with PARTIAL_QR_CMP.
!   
!   On the other hand, PARTIAL_QR_CMP can be used for both tasks, but is much slower than PARTIAL_RQR_CMP.
!   
!   Note that PARTIAL_RQR_CMP performs also exactly the same task as subroutine PARTIAL_RQR_CMP2 in module Random.
!   The main difference between the two routines is that PARTIAL_RQR_CMP2 recomputes the Gaussian and
!   compression matrices at each iteration of the randomized (partial or complete) QR algorithm, while PARTIAL_RQR_CMP uses
!   an efficient updating formulae to recompute the compression matrix at each iteration. In other words, PARTIAL_RQR_CMP
!   is usually faster than PARTIAL_RQR_CMP2, but may be sligthly less robust. See references (3), (4), (5) and (6)
!   for further information.
!   
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the real m-by-n matrix to be decomposed.
!
!            On exit, MAT has been overwritten by details of its (partial)
!            QR or orthogonal factorization.
!
!            See Further Details.
!
!   DIAGR    (OUTPUT) real(stnd), dimension(:) 
!            On exit, the diagonal elements of the matrix R11 or T11.
!
!            See Further Details.
!
!            The size of DIAGR must verify:
!
!            - size( DIAGR ) = kpartial <= min( size(MAT,1) , size(MAT,2) ).
!
!   BETA     (OUTPUT) real(stnd), dimension(:)
!            On exit, the scalars factors of the elementary reflectors defining Q.
!
!            See Further Details.
!
!            The size of BETA must verify:
!
!            - size( BETA ) = kpartial <= min( size(MAT,1) , size(MAT,2) ).
!
!   IP       (OUTPUT) integer(i4b), dimension(:) 
!            On exit, if IP(j)=k, then the j-th column of MAT*P was 
!            the k-th column of MAT.
!
!            See Further Details.
!
!            The size of IP must be equal to size( MAT, 2 ) = n.
!
!   KRANK    (OUTPUT) integer(i4b)
!            On exit, KRANK contains the effective rank of R11, i.e., krank, which is
!            the order of this submatrix R11. This is the same as the order of
!            the submatrix T11 in the complete orthogonal factorization of MAT and is also
!            the rank of MAT if krank is less than kpartial = size( BETA ).
!
!            If the computed pseudo-rank, krank, is less than kpartial = size( BETA ),
!            BETA(krank+1:kpartial) and, eventually, TAU(krank+1:kpartial) are set to zero
!            and MAT(krank+1:m,krank+1:n) (e.g., R22=T22) is updated on exit.
!
!            In other words, the subroutine outputs a partial QR factorization of rank krank
!            instead of rank kpartial.
!
!            In all cases, norm(MAT(krank+1:m,krank+1:n)) gives the error of the associated
!            matrix approximation in the Frobenius norm, on exit.
!
!   TOL      (INPUT/OUTPUT, OPTIONAL) real(stnd)
!            On entry, if TOL is present  then:
!
!            - if TOL is in ]0,1[, the calculations to determine the condition number of R11
!              are performed. Then, TOL is used to determine the effective pseudo-rank
!              of R11, which is defined as the order of the largest leading triangular
!              submatrix in the partial QR factorization with column pivoting of MAT,
!              whose estimated condition number in the 1-norm is less than 1/TOL.
!              On exit, the reciprocal of the condition number is returned in TOL.
!            - if TOL=0 is specified, the calculations to determine the condition number
!              of R11 are not performed and crude tests on R(j,j) are done to determine
!              the numerical pseudo-rank of R11. On exit, TOL is not changed.
!
!            If TOL is not specified or is outside [0,1[, the calculations to determine the
!            rank of R11 are not performed and this rank is assumed to be equal to kpartial.
!
!   TAU      (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if TAU is present, a (partial) complete orthogonal factorization of
!            MAT is computed. Otherwise, a simple QR factorization with column pivoting
!            of MAT is computed.
!
!            On exit, the scalars factors of the elementary reflectors defining the
!            orthogonal matrix Z in the (partial) orthogonal factorization of MAT.
!
!            See Further Details.
!
!            The size of TAU must verify:
!
!            - size( TAU ) = kpartial <= min( size(MAT,1) , size(MAT,2) ).
!
!   RNG_ALG  (INPUT, OPTIONAL) integer(i4b)
!            On entry, a scalar integer to select the random (uniform) number generator used to build
!            the random gaussian matrix in the randomized partial QR algorithm.
!
!            The possible values are:
!
!            - ALG=1  : selects the Marsaglia's KISS random number generator;
!            - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!            - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!            - ALG=4  : selects the Mersenne Twister random number generator;
!            - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!            - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!            - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!            - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!            - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!            - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!              random number generator;
!
!            For other values, the current random number generator and its current state are not changed.
!            Note further, that, on exit, the current random number generator is not reset to its previous
!            value before the call to PARTIAL_RQR_CMP.
!
!            See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   BLK_SIZE (INPUT, OPTIONAL) integer(i4b)
!            On entry, the block size used in the randomized partial QR factorization.
!
!            BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!            smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!            By default, BLK_SIZE is set to min( BLKSZ_QR, min(m,n) ), where parameter BLKSZ_QR is the default
!            block size for QR related algorithms specified in module Select_Parameters.
!
!   NOVER    (INPUT, OPTIONAL) integer(i4b)
!            The oversampling size used in the randomized partial QR algorithm.
!
!            NOVER must be positive or null and verifies the relationship:
!
!                 NOVER + BLK_SIZE <= size(MAT,1)
!
!            and is adjusted if necessary to verify this relationship in all cases.
!
!            See Further Details and the cited references for the meaning and usefulness
!            of the oversampling size in the partial randomized QR algorithm.
!
!            By default, the oversampling size is set to 10.
!
!
! Further Details
! _______________
!
!   The matrix Q is represented as a product of elementary reflectors
!
!            Q = H(1) * H(2) * ... * H(krank), where krank = size( BETA ) <= min( m , n ).
!
!   Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!   where beta is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in MAT(i:m,i) and beta in BETA(i).
!
!   On exit of PARTIAL_RQR_CMP, the orthonormal matrix Q (or its first n columns) can be
!   computed explicitly by a call to subroutine ORTHO_GEN_QR with arguments MAT and BETA.
!
!   The matrix P is represented in the array IP as follows:
!   If IP(j) = i then the jth column of P is the ith canonical unit vector.
!
!   On exit, if the optional argument TAU is absent:
!
!   - The elements above the diagonal of the array MAT(:krank,:krank) contain the 
!     corresponding elements of the triangular matrix R11.
!   - The elements of the diagonal of R11 are stored in the array DIAGR.
!   - The submatrix R12 is stored in MAT(:krank,krank+1:n).
!   - The submatrix R22 is stored in MAT(krank+1:m,krank+1:n).
!   - krank is stored in the real argument KRANK.
!
!   If TAU is present, a partial or complete orthogonal factorization of MAT is computed.
!   The factorization is obtained by Householder's method. The kth transformation
!   matrix, Z(k), which is used to introduce zeros into the kth row of R (e.g., in
!   R12), is given in the form
!
!   [  I  0    ]
!
!   [  0  T(k) ]
!
!   where
!
!       T(k) = I + tau * ( u(k) * u(k)' ) and u(k)' = (  1  0  z(k)  )
!
!   tau is a scalar, u(k) is a n-k+1 vector and z(k) is an (n-krank) element vector.
!   tau and z(k) are chosen to annihilate the elements of the kth row of R12.
!
!   The Z n-by-n orthogonal matrix is given by
!
!        Z = Z(1) * Z(2) * ... * Z(krank)
!
!   On exit, if the optional argument TAU is present:
!
!   - The scalar tau defining T(k) is returned in the kth element of TAU and the vector u(k)
!     in the kth row of MAT, such that the elements of z(k) are in MAT(k,krank+1:n).
!     The other elements of u(k) are not stored.
!   - The elements above the diagonal of the array section MAT(:krank,:krank)
!     contain the corresponding elements of the triangular matrix T11. The elements of
!     the diagonal of T11 are stored in the array DIAGR.
!   - The submatrix T22 (=R22) is stored in MAT(krank+1:m,krank+1:n).
!   - krank is stored in the real argument KRANK.
!
!   In both cases, norm(MAT(krank+1:m,krank+1:n)) gives the error of the associated
!   matrix approximation in the Frobenius norm, on exit.
!
!   If it is possible that MAT may not be of full rank (i.e., certain columns of MAT
!   are linear combinations of other columns), then the eventual linearly dependent columns
!   in the (partial or complete) QR decomposition of MAT, which is sought, can be determined
!   by using TOL=relative precision of the elements in MAT.
!   If each element is correct to, say, 5 digits then TOL=0.00001 should
!   be used. Also, it may be helpful to scale the columns of MAT so that all elements 
!   are about the same order of magnitude.
!
!   The computations are parallelized if OPENMP is used. Note also that PARTIAL_RQR_CMP
!   uses a randomized "BLAS3" algorithm described in the references (3), (4), (5) and (6), which
!   has about the same efficiency of a "BLAS3" QR algorithm without column pivoting and is thus
!   highly efficient on large matrices.
!
!   For further details, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Duersch, J.A., and Gu, M., 2017:
!           Randomized QR with column pivoting.
!           SIAM J. Sci. Comput., Volume 39, Issue 4, C263-C291.
!
!   (4) Martinsson, P.G., Quintana-Orti, G., Heavner, N., and Van de Geijn, R., 2017:
!           Householder QR factorization with randomization for column pivoting (HQRRP).
!           SIAM J. Sci. Comput., Volume 39, Issue 2, C96-C115.
!
!   (5) Xiao, J., Gu, M., and Langou, J., 2017:
!           Fast parallel randomized QR with column pivoting algorithms
!           for reliable low-rank matrix approximations.
!           IEEE 24th International Conference
!           on High Performance Computing (HiPC), IEEE, 2017, 233-242.
!
!   (6) Duersch, J.A., and Gu, M., 2020:
!           Randomized projection for rank-revealing matrix factorizations
!           and low-rank approximations.
!           SIAM Review, Volume 62, Issue 3, 661-682.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Num_Constants,   only : machhuge, safmin
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert_eq, assert, merror, arth
    use Hous_Procedures, only : hous1, hous2, apply_hous2
    use QR_Procedures,   only : partial_qr_cmp
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: diagr, beta
    real(stnd),   intent(inout),                 optional :: tol
    real(stnd),   intent(out),   dimension(:),   optional :: tau
!
    integer(i4b), intent(out)               :: krank
    integer(i4b), intent(out), dimension(:) :: ip
    integer(i4b), intent(in), optional :: rng_alg, blk_size, nover
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: anorm, xnorm, rcond, prod, beta1
    real(stnd), dimension(:),   allocatable :: diagr_b, beta_b
    real(stnd), dimension(:,:), allocatable :: b, lmat, umat, mat2, omega
#ifdef _BLAS
    real(stnd), dimension(size(mat,2))      :: tmp
    real(stnd), dimension(:,:), allocatable :: x
#else
    real(stnd), dimension(max(size(mat,1),size(mat,2))) :: tmp
    real(stnd), dimension(:), allocatable               :: x
#endif
!
    integer      :: iok, alg
    integer(i4b) :: m, n, ndiag, p, nover2, blk_size2, nbr, in, nblk, &
                    i, i1, i2, i3, itmp, j, j2, l, l1, l2
    integer(i4b), dimension(:), allocatable :: ip_b
!
    logical(lgl) :: comp_min_norm, test
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='partial_rqr_cmp'
!
!   nover0 IS THE DEFAULT OVERSAMPLING SIZE.
!
    integer(i4b),  parameter :: nover0=10_i4b
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b) ,   &
                   int(size(ip),i4b)    ,   &
                   name_proc )
!
    ndiag = min( m, n )
!
    comp_min_norm = present( tau )
!
    if ( comp_min_norm ) then
!
        krank =  assert_eq( int(size(beta),i4b)  ,   &
                            int(size(diagr),i4b) ,   &
                            int(size(tau),i4b)   ,   &
                            name_proc )
!
        tau(:krank) = zero
!
    else
!
        krank =  assert_eq( int(size(beta),i4b)  ,   &
                            int(size(diagr),i4b) ,   &
                            name_proc )
!
    end if
!
    call assert( logical( krank<=ndiag, lgl ), name_proc )
!
!   INITIALIZE PERMUTATION VECTOR FOR COLUMN PIVOTING.
!
    ip(:n) = arth( 1_i4b, 1_i4b, n )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( krank<=0_i4b ) then
!
        if ( present(tol) ) tol = machhuge
!
        return
!
    end if
!
!   SELECT AND SET THE RANDOM NUMBER GENERATOR TO BE USED IF REQUIRED.
!
    if ( present(rng_alg) ) then
!
        if ( rng_alg>=1_i4b .and. rng_alg<=10_i4b ) then
!
            alg = rng_alg
!
!           SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
            call random_seed_( alg=alg )
!
!           RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
            call random_seed_()
!
        end if
!
    end if
!
!   DETERMINE THE OVERSAMPLING SIZE nover2 .
!
    if ( present(nover) ) then
        nover2 = max( nover, 0_i4b )
    else
        nover2 = nover0
    end if
!
!   DETERMINE THE BLOCKSIZE blk_size2 .
!
    if ( present(blk_size) ) then
        blk_size2 = max( blk_size, 1_i4b )
    else
        blk_size2 = blksz
    end if
!    
    blk_size2 = min( blk_size2, ndiag )
!    
!   COMPUTE RANDOMIZED PARTIAL QR FACTORIZATION WITH COLUMN PIVOTING OF mat(:m,:n) .
!
!   nbr IS THE NUMBER OF ROWS IN MATRICES b and omega .
!
    nbr = min( nover2 + blk_size2, m )
!
!   ALLOCATE WORK ARRAYS, b(:nbr,:n) IS THE COMPRESSION MATRIX,
!   omega(:nbr,:m) IS A GAUSSIAN RANDOM MATRIX.
!
    allocate( b(nbr,n), omega(nbr,m), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A nbr-BY-m IID GAUSSIAN MATRIX.
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
    call normal_random_number3_( omega(:nbr,:m) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
!    do i = 1_i4b, m
!
!        call normal_random_number3_( omega(:nbr,i) )
!
!    end do
!
!   COMPUTE COMPRESSION MATRIX.
!
#ifdef _BLAS
    call gemm( 'N', 'N', nbr, n, m, one, omega(1_i4b:nbr,1_i4b:m), nbr,  &
               mat(1_i4b:m,1_i4b:n), m, zero, b(1_i4b:nbr,1_i4b:n), nbr  )
#else
    b(:nbr,:n) = matmul( omega(:nbr,:m), mat(:m,:n) )
#endif
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( omega )
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  n*m>=omp_limit                      .and.       &
                  n/blk_size2>1_i4b                   .and.       &
                  i>1_i4b
#endif
!
    if ( krank<=blk_size2 ) then
!
!       krank LESS OR EQUAL TO blk_size2, USE A FAST RANDOMIZED QR FACTORIZATION OF mat WITHOUT
!       UPDATES OF THE COMPRESSION MATRIX.
!
!       DO PARTIAL QR WITH COLUMN PIVOTING ON COMPRESSION MATRIX b(:nbr,:n) TO FIND krank PIVOTS.
!           
        call partial_qr_cmp( b(:nbr,:n), diagr(:krank), beta(:krank), ip(:n), i )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( b )
!
!       SWAP CORRESPONDING COLUMNS OF mat(:m,:n) .
!
        ip(:n) = -ip(:n)
!
        do i = 1_i4b, n
!
            if ( ip(i)>0_i4b ) cycle
!
            l     = i
            ip(l) = -ip(l)
            in    = ip(l)
!
            do
!
                if ( ip(in)>0_i4b ) exit
!
                if ( l/=in ) then
#ifdef _BLAS
                    call swap( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,in), 1_i4b )
#else
                    tmp(1_i4b:m)    = mat(1_i4b:m,l)
                    mat(1_i4b:m,l)  = mat(1_i4b:m,in)
                    mat(1_i4b:m,in) = tmp(1_i4b:m)
#endif
!
                end if
!
                ip(in) = -ip(in)
                l      = in
                in     = ip(in)
!
            end do
!
        end do
!
!       COMPUTE H(1), ... , H(krank) AND DO QR DECOMPOSITION 
!       OF THE FIRST BLOCK OF mat .
!
        do i = 1_i4b, krank
!    
!           GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+1:m,i) .
!    
            call hous1( mat(i:m,i), beta(i), diagr(i) )
!    
!           APPLY H(i) TO mat(i:m,i+1:krank) FROM THE LEFT.
!
            beta1 = beta(i)
!
            do l = i + 1_i4b, krank
!
#ifdef _BLAS
                prod = beta1*dot( m-i+1_i4b, mat(i:m,i), 1_i4b, mat(i:m,l), 1_i4b )
                call axpy( m-i+1_i4b, prod, mat(i:m,i), 1_i4b, mat(i:m,l), 1_i4b )
#else
                prod = beta1*dot_product( mat(i:m,i), mat(i:m,l) )
                mat(i:m,l) = mat(i:m,l) + prod*mat(i:m,i)
#endif
!    
            end do
!
        end do
!    
        if ( krank<n ) then
!
!           BLOCK UPDATE OF THE TRAILING MATRIX mat(:m,krank+1:n) IF REQUIRED.
!
!           ALLOCATE WORK ARRAY.
!
            allocate( lmat(krank,krank), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           INITIALIZE THE UNIT LOWER TRIANGULAR MATRIX lmat(:krank,:krank)
!           USED IN THE BLOCK UPDATE ALGORITHM AND SAVE TRIANGULAR FACTOR
!           R OF QR FACTORIZATION OF THE FIRST BLOCK OF mat IN UPPER TRIANGLE
!           OF lmat IF BLAS IS USED.
!
#ifdef _BLAS
            do i = 1_i4b, krank
!
                lmat(1_i4b:i,i)       = mat(1_i4b:i,i)
                lmat(i+1_i4b:krank,i) = zero
!
                mat(1_i4b:i-1_i4b,i)  = zero
!
            end do
#else
            lmat(:krank,:krank) = zero
!
            do i = 1_i4b, krank
!
                lmat(i,i) = one
!
            end do
#endif
!
!$OMP PARALLEL IF(test_par)                 &
!$OMP         ,PRIVATE(iok,i,i1,i2,i3,l,l2,x)
!
!           ALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
            allocate( x(krank,blk_size2), stat = iok )
#else
            allocate( x(krank), stat = iok )
#endif
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!    
!           APPLY H(krank)* ...*H(1) TO mat(:m,krank+1:n) FROM THE LEFT
!           WITH A BLOCKED ALGORITHM AS
!
!            ( I + U(:m,:krank)*inv( lmat(:krank,:krank) )*U(:m,:krank)' )*mat(:m,krank+1:n) .
!
!           FIRST COMPUTE THE krank BY krank UNIT LOWER TRIANGULAR MATRIX lmat(:krank,:krank)
!           USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do l = 1_i4b, krank
!
                do i = l + 1_i4b, krank
!
#ifdef _BLAS
                    lmat(i,l) = -beta(i)*dot( m-i+1_i4b, mat(i:m,l), 1_i4b, mat(i:m,i), 1_i4b )
#else
                    lmat(i,l) = -beta(i)*dot_product( mat(i:m,l), mat(i:m,i) )
#endif
!
                end do
!                    
            end do
!
!$OMP END DO
!
!           BLOCK UPDATE OF THE TRAILING MATRIX mat(:m,krank+1:n) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do i = krank, n-1_i4b, blk_size2
!
                i1 = i + 1_i4b
                i2 = min( i+blk_size2, n )
                i3 = i2 - i
!
#ifdef _BLAS
!
!               COMPUTE U(:m,:krank)'*mat(:m,i1:i2) .
!
                call gemm( 'T', 'N', krank, i3, m, one, mat, m, mat(1_i4b:m,i1:i2), m, zero, x, krank )
!
                do l = 1_i4b, i3
!
                    x(:krank,l) = beta(:krank)*x(:krank,l)
!                    
                end do
!
!               SOLVE lmat(:krank,:krank)*x(:krank,:i3) = U(:m,:krank)'*mat(:m,i1:i2) .
!
                call trsm( 'L', 'L', 'N', 'U', krank, i3, one, lmat, krank, x, krank )
!
!               FULL VERSION...
!
!               COMPUTE mat(:m,i1:i2) = mat(:m,i1:i2) + U(:m,:krank)*x(:krank,:i3) .
!
                call gemm( 'N', 'N', m, i3, krank, one, mat, m, x, krank, one, mat(1_i4b:m,i1:i2), m )
!
!               TRUNCATED VERSION...
!
!               COMPUTE mat(:krank,i1:i2) = mat(:krank,i1:i2) + U(:krank,:krank)*x(:krank,:i3) .
!
!                call gemm( 'N', 'N', krank, i3, krank, one, mat, m, x, krank, one, mat(1_i4b:m,i1:i2), m )
#else
                do l = i1, i2
!
!                   COMPUTE U(:m,:krank)'*mat(:m,i1:i2) .
!
                    do l2 = 1_i4b, krank
!
                        x(l2) = beta(l2)*dot_product( mat(l2:m,l2), mat(l2:m,l) )
!                    
                    end do
!
!                   SOLVE lmat(:krank,:krank)*x(:krank,:i3) = U(:m,:krank)'*mat(:m,i1:i2) .
!                   COMPUTE mat(:m,i1:i2) = mat(:m,i1:i2) + U(:m,:krank)*x(:krank,:i3) .
!
                    do l2 = 1_i4b, krank
!
                        x(l2+1_i4b:krank) = x(l2+1_i4b:krank) - x(l2)*lmat(l2+1_i4b:krank,l2)
!
!                       FULL VERSION...
!
                        mat(l2:m,l) = mat(l2:m,l) + x(l2)*mat(l2:m,l2)
!
!                       TRUNCATED VERSION...
!
!                        mat(l2:krank,l) = mat(l2:krank,l) + x(l2)*mat(l2:krank,l2)
!
                    end do
!
                end do
#endif
!                    
            end do
!
!$OMP END DO
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( x )
!
!$OMP END PARALLEL
!
!           RESTORE TRIANGULAR FACTOR R OF QR FACTORIZATION
!           OF THE FIRST BLOCK OF mat IF BLAS IS USED.
!
#ifdef _BLAS
            do i = 2_i4b, krank
!
                mat(1_i4b:i-1_i4b,i)  = lmat(1_i4b:i-1_i4b,i)
!
            end do
#endif
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( lmat )
!
        end if
!
    else
!
!       krank GREATER THAN blk_size2, USE RANDOMIZED QR FACTORIZATION OF mat WITH
!       UPDATES OF THE COMPRESSION MATRIX.
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( ip_b(n), diagr_b(blk_size2), beta_b(blk_size2),        &
                  lmat(blk_size2,blk_size2), umat(m,blk_size2), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       INITIALIZE THE UNIT LOWER TRIANGULAR MATRIX lmat(:blk_size2,:blk_size2)
!       USED IN THE BLOCKED QR ALGORITHM.
!
        lmat(1_i4b:blk_size2,1_i4b:blk_size2) = zero
!
        do i = 1_i4b, blk_size2
            lmat(i,i) = one
        end do
!
!       INITIALIZE ARRAY FOR STORING HOUSEHOLDER VECTORS USED IN THE BLOCKED QR
!       ALGORITHM.
!
        umat(:m,:blk_size2) = zero
!
!$OMP PARALLEL IF(test_par)                                         &
!$OMP         ,PRIVATE(test,iok,i,i1,i2,i3,j,j2,l,l1,l2,in,itmp,nblk,p,beta1,prod,x,mat2,tmp)
!
!       ALLOCATE WORK VARIABLES.
!
#ifdef _BLAS
        allocate( x(blk_size2,blk_size2), mat2(m,blk_size2), stat = iok )
#else
        allocate( x(blk_size2), mat2(m,blk_size2), stat = iok )
#endif
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       RANDOMIZED BLOCKED PARTIAL QR DECOMPOSITION OF mat WITH COLUMN PIVOTING.
!
        do j = 1_i4b, krank, blk_size2
!
            nblk = min( krank-j+1_i4b, blk_size2 )
!
            j2 = j + nblk - 1_i4b
            p  = m - j + 1_i4b
!
            test = nbr < m - j2 .and. j2 < krank
!
!$OMP MASTER
!
!           DO PARTIAL QR WITH COLUMN PIVOTING ON COMPRESSION MATRIX b(:nbr,j:n) TO FIND nblk NEW PIVOTS.
!           
            call partial_qr_cmp( b(:nbr,j:n), diagr_b(:nblk), beta_b(:nblk), ip_b(j:n), i )
!
!           SWAP CORRESPONDING COLUMNS OF mat(:m,j:n) .
!
            ip_b(j:n) = -ip_b(j:n)
!
            do i = j, n
!
                if ( ip_b(i)>0_i4b ) cycle
!
                l       = i
                ip_b(l) = -ip_b(l)
                in      = ip_b(l) + j - 1_i4b
!
                do
!
                    if ( ip_b(in)>0_i4b ) exit
!
                    if ( l/=in ) then
#ifdef _BLAS
                        call swap( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,in), 1_i4b )
#else
                        tmp(1_i4b:m)    = mat(1_i4b:m,l)
                        mat(1_i4b:m,l)  = mat(1_i4b:m,in)
                        mat(1_i4b:m,in) = tmp(1_i4b:m)
#endif
!
                        itmp   = ip(l)
                        ip(l)  = ip(in)
                        ip(in) = itmp
!
                    end if
!
                    ip_b(in) = -ip_b(in)
                    l        = in
                    in       = ip_b(in) + j - 1_i4b
!
                end do
!
            end do
!
!           COMPUTE H(j), ... , H(j+nblk-1_i4b) AND DO QR DECOMPOSITION 
!           OF THE CURRENT BLOCK OF mat .
!
            do i = j, j2
!    
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+1:m,i) .
!    
                call hous1( mat(i:m,i), beta(i), diagr(i) )
!    
!               APPLY H(i) TO mat(i:m,i+1:j+nblk-1) FROM THE LEFT.
!
                beta1 = beta(i)
!
                do l = i + 1_i4b, j2
!
#ifdef _BLAS
                    prod = beta1*dot( m-i+1_i4b, mat(i:m,i), 1_i4b, mat(i:m,l), 1_i4b )
                    call axpy( m-i+1_i4b, prod, mat(i:m,i), 1_i4b, mat(i:m,l), 1_i4b )
#else
                    prod = beta1*dot_product( mat(i:m,i), mat(i:m,l) )
                    mat(i:m,l) = mat(i:m,l) + prod*mat(i:m,i)
#endif
!    
                end do
!
                if ( test ) then
!    
!                   RESTORE UPPER TRIANGULAR FACTOR IN QR FACTORIZATION OF CURRENT BLOCK
!                   OF COMPRESSION MATRIX IN b(:nblk,j:j2) FOR LATER USE.
!
                    i1 = i - j + 1_i4b
!
                    b(i1,i) = diagr_b(i1)                
!
                    b(i1+1:nblk,i) = zero
!    
!                   START UPDATE OF THE COMPRESSION MATRIX FOR LATER USE BY FORMING
!    
!                       b(:nblk,j:j2)*inv( R(:nblk,:nblk) )
!    
!                   WHERE R(:nblk,:nblk) IS THE UPPER TRIANGULAR FACTOR IN QR FACTORIZATION
!                   OF CURRENT BLOCK OF mat, STORING THE RESULT IN b(:nblk,j:j2) .
!    
                    do l = j, i-1_i4b
!    
                        l1 = l - j + 1_i4b
!
                        b(1_i4b:l1,i) = b(1_i4b:l1,i) - mat(l,i)*b(1_i4b:l1,l)
!
                    end do
!
                    beta1 = sign( max( abs(diagr(i)), safmin ), diagr(i) )
!
                    prod = one/beta1
!
                    b(1_i4b:i1,i) = prod*b(1_i4b:i1,i)
!
                end if
!
            end do
!
!$OMP END MASTER
!
!$OMP BARRIER
!    
            if ( j2<n ) then
!    
!               APPLY H(j+nblk-1_i4b)* ...*H(j) TO mat(j:m,j+nblk:n) FROM THE LEFT
!               WITH A BLOCKED ALGORITHM AS
!
!            ( I + U(j:m,:nblk)*inv( lmat(:nblk,:nblk) )*U(j:m,:nblk)' )*mat(j:m,j+nblk:n) .
!
!               FIRST SAVE H(j) ...H(j+nblk-1_i4b) FOR LATER USE AND COMPUTE THE nblk BY nblk UNIT
!               LOWER TRIANGULAR MATRIX lmat(:nblk,:nblk) USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, nblk
!
                    l1 = j + l - 1_i4b
!
!                   SAVE SHIFTED HOUSEHOLDER VECTOR mat(j+l-1:m,j+l-1) .
!
                    umat(l:p,l) = mat(l1:m,l1)
!
                    do i = l + 1_i4b, nblk
!
                        l2 = j + i - 1_i4b
!
#ifdef _BLAS
                        lmat(i,l) = -beta(l2)*dot( p-i+1_i4b, umat(i:p,l), 1_i4b, mat(l2:m,l2), 1_i4b )
#else
                        lmat(i,l) = -beta(l2)*dot_product( umat(i:p,l), mat(l2:m,l2) )
#endif
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!               BLOCK UPDATE OF THE TRAILING MATRIX mat(j:m,j+nblk:n) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do i = j2, n-1_i4b, nblk
!
                    i1 = i + 1_i4b
                    i2 = min( i+nblk, n )
                    i3 = i2 - i
!
!                   MAKE A LOCAL COPY OF mat(j:m,i1:i2) .
!
                    mat2(:p,:i3) = mat(j:m,i1:i2)
!
#ifdef _BLAS
!
!                   COMPUTE U(j:m,:nblk)'*mat(j:m,i1:i2) .
!
                    call gemm( 'T', 'N', nblk, i3, p, one, umat, m, mat2, m, zero, x, blk_size2 )
!
                    do l = 1_i4b, i3
!
                        x(:nblk,l) = beta(j:j2)*x(:nblk,l)
!                    
                    end do
!
!                   SOLVE lmat(:nblk,:nblk)*x(:nblk,:i3) = U(j:m,:nblk)'*mat(j:m,i1:i2) .
!
                    call trsm( 'L', 'L', 'N', 'U', nblk, i3, one, lmat, blk_size2, x, blk_size2 )
!
!                   COMPUTE mat(j:m,i1:i2) = mat(j:m,i1:i2) + U(j:m,:nblk)*x(:nblk,:i3) .
!
                    call gemm( 'N', 'N', p, i3, nblk, one, umat, m, x, blk_size2, one, mat2, m )
#else
                    do l = 1_i4b, i3
!
!                       COMPUTE U(j:m,:nblk)'*mat(j:m,i1:i2) .
!
                        do l2 = 1_i4b, nblk
!
                           x(l2) = beta(j+l2-1_i4b)*dot_product( umat(l2:p,l2), mat2(l2:p,l) )
!                    
                        end do
!
!                       SOLVE lmat(:nblk,:nblk)*x(:nblk,:i3) = U(j:m,:nblk)'*mat(j:m,i1:i2) .
!                       COMPUTE mat(j:m,i1:i2) = mat(j:m,i1:i2) + U(j:m,:nblk)*x(:nblk,:i3) .
!
                        do l2 = 1_i4b, nblk
!
                            x(l2+1_i4b:nblk) = x(l2+1_i4b:nblk) - x(l2)*lmat(l2+1_i4b:nblk,l2)
!
                            mat2(l2:p,l) = mat2(l2:p,l) + x(l2)*umat(l2:p,l2)
!
                        end do
!
                    end do
#endif
!
!                   SAVE UPDATE OF mat(j:m,i1:i2) .
!
                    mat(j:m,i1:i2) = mat2(:p,:i3)           
!
!                   UPDATE COMPRESSION MATRIX FOR NEXT PANEL QR ITERATION.
!
                    if ( test ) then
!
#ifdef _BLAS
                        call gemm( 'N', 'N', nblk, i3, nblk, -one, b(1_i4b:nbr,j:j2), nbr,  &
                                   mat2, m, one, b(1_i4b:nbr,i1:i2), nbr  )
#else
                        b(:nblk,i1:i2) = b(:nblk,i1:i2) - matmul( b(:nblk,j:j2), mat2(:nblk,:i3) )
#endif
!                    
                    else if ( j2 < krank ) then
!
                        l = m - j2
!                    
                        b(:l,i1:i2)          = mat2(nblk+1_i4b:p,:i3) 
                        b(l+1_i4b:nbr,i1:i2) = zero 
!                    
                    end if
!                    
                end do
!
!$OMP END DO
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( x, mat2 )
!
!$OMP END PARALLEL
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( b, ip_b, diagr_b, beta_b, lmat, umat )
!    
    end if
!    
!   TEST TOLERANCE.
!    
    if ( present(tol) ) then
        test = zero<=tol .and. tol<one
    else
        test = false
    end if
!    
    if ( test ) then
!    
        tmp(:krank) = abs( diagr(:krank) )
!    
        if ( tol/=zero ) then
!    
!           DETERMINE THE PSEUDO-RANK USING THE TOLERANCE, tol, AND
!           THE CONDITION NUMBER IN THE 1-NORM OF THE TRIANGULAR FACTOR R
!           OF THE QR FACTORIZATION.
!    
            anorm = zero
            xnorm = zero
            rcond = zero
!
            do j = 1_i4b, krank
!
!               CHECK IF abs( R(j,j) ) IS SMALL AND IF R(:j,:j) IS SINGULAR.
!               IF R(:j,:j) IS SINGULAR, EXIT AND THE PSEUDO-RANK IS j-1 .
!
                if ( tmp(j)<safmin ) exit
!
!               COMPUTE 1-NORM OF R(:j,:j).
!
                anorm  = max( anorm, sum(abs(mat(1_i4b:j-1_i4b,j)))+tmp(j) )
!
!               COMPUTE 1-NORM OF inv( R(:j,:j) ).
!
                tmp(j)             = one
                tmp(1_i4b:j-1_i4b) = zero
!
#ifdef _BLAS
                mat(j,j) = diagr(j)
!
                call trsv( 'U', 'N', 'N', j, mat, m, tmp(1_i4b:j), 1_i4b )    
#else
                do i = j, 1_i4b, -1_i4b
                    tmp(i)             = tmp(i)/diagr(i)
                    tmp(1_i4b:i-1_i4b) = tmp(1_i4b:i-1_i4b) - tmp(i)*mat(1_i4b:i-1_i4b,i)
                end do
#endif
!
                xnorm = max( xnorm, sum(abs(tmp(1_i4b:j))) )
!
!               COMPUTE INVERSE OF CONDITION NUMBER OF R(:j,:j).
!
                prod = one/( anorm*xnorm )
!
!               EXIT IF INVERSE OF CONDITION NUMBER IS LESS THAN THE TOLERANCE.
!
                if ( tol>=prod ) exit
!
                rcond = prod
!
            end do
!
            tol = rcond
!
#ifdef _BLAS
!
!           RESTORE FIRST ELEMENT OF HOUSEHOLDER VECTORS IF BLAS IS USED.
!
            do i = 1_i4b, min(j,krank)
                mat(i,i) = one
            end do
#endif
!
        else
!    
!           DETERMINE THE NUMERICAL PSEUDO-RANK.
!    
            do j = 1_i4b, krank
                if ( tmp(j)<safmin ) exit
            end do
!
        end if
!    
!       UPDATE RANK AND PARTIAL QR FACTORIZATION OF mat ON EXIT IF mat IS RANK DEFICIENT.
!    
        beta(j:krank) = zero
!    
        do i = j, krank
!    
            mat(i,i)         = diagr(i)
            mat(i+1_i4b:m,i) = zero
!    
        end do
!    
        krank = j - 1_i4b
!    
    end if
!    
!   IF THE PSEUDORANK krank IS LESS THAN n AND THE MINIMUM_NORM SOLUTION IS REQUIRED,
!   COMPUTE HOUSEHOLDER DECOMPOSITION OF mat(1:krank,:n) .
!    
    if ( krank/=n .and. comp_min_norm ) then
!    
        j = krank + 1_i4b
        l = n - krank
!
!       ALLOCATE WORK VARIABLES FOR STORING HOUSEHOLDER VECTORS.
!
        allocate( umat(l,krank), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        umat(:l,:krank) = transpose( mat(:krank,j:n) )
!    
        do i = krank, 1_i4b, -1_i4b
!    
!           COMPUTE THE I-TH HOUSEHOLDER TRANSFORMATION.
!    
            call hous2( diagr(i), umat(:l,i), tau(i) )
!    
!           APPLY THE I-TH HOUSEHOLDER TRANSFORMATION FROM THE RIGHT.
!    
            call apply_hous2( umat(:l,i), tau(i),                                &
                              mat(1_i4b:i-1_i4b,i), umat(:l,1_i4b:i-1_i4b), true )
!    
        end do
!
        mat(:krank,j:n) = transpose( umat(:l,:krank) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( umat )
!
    end if
!
!
! END OF SUBROUTINE partial_rqr_cmp
! _________________________________
!
    end subroutine partial_rqr_cmp
!
! =========================================================================================
!
    subroutine partial_rqr_cmp2( mat, diagr, beta, ip, krank, tol, tau, rng_alg, blk_size, nover )
!
! Purpose
! _______
!
!   PARTIAL_RQR_CMP2 computes a randomized (partial or complete) QR factorization with column
!   pivoting or orthogonal factorization of a m-by-n matrix MAT:
!
!                     MAT * P = Q * R
!
!   , where P is a n-by-n permutation matrix, R is an upper triangular or trapezoidal
!   (i.e., if n>m) m-by-n matrix and Q is a m-by-m orthogonal matrix.
!
!   At the user option, the randomized QR factorization can be only partial, e.g., the subroutine
!   ends when the numbers of columns of Q is equal to a predefined value equals to
!   kpartial = size( DIAGR ) = size( BETA ).
!
!   This leads implicitly to the following partition of Q:
!
!   [  Q1  Q2  ]
!
!   where Q1 is a m-by-kpartial orthonormal matrix and Q2 is a m-by-(m-kpartial) orthonormal matrix
!   orthogonal to Q1, and to the following corresponding partition of R:
!
!   [  R11  R12  ]
!
!   [  R21  R22  ]
!
!   where R11 is a kpartial-by-kpartial triangular matrix, R21 is zero by construction,
!   R12 is a full kpartial-by-(n-kpartial) matrix and R22 is a full (m-kpartial)-by-(n-kpartial)
!   matrix.
!
!   Then, if the optional scalar argument TOL is present and:
!
!   - is in ]0,1[, the rank of R11 is determined by finding the submatrix of R11
!     which is defined as the largest leading submatrix whose estimated condition
!     number, in the 1-norm, is less than 1/TOL.
!   - is equal to 0, the rank of R11, krank, is determined by finding the largest
!     submatrix of R11 such that abs(R11[j,j])>0. 
!
!   In both cases, the order of this submatrix, krank, is the effective rank of R11
!   (and MAT if krank is less than kpartial or if krank=kpartial=min(m,n)).
!
!   If TOL is absent or outside [0,1[, the rank of R11 is not checked and is assumed to be
!   equal to kpartial.
!
!   If krank is less than kpartial, then MAT is not of full rank (i.e., certain columns of
!   MAT(:m,:kpartial) are linear combinations of other columns of MAT(:m,:kpartial)) and
!   krank is also an estimate of the rank of MAT.
!
!   This leads to a redefinition of the partition of Q = [  Q1  Q2  ], where
!   Q1 and Q2 are now m-by-krank and m-by-(m-krank) orthonormal matrices, and a corresponding 
!   redefinition of the associated partition of R, where R11 is now a krank-by-krank 
!   triangular matrix, R21 is again zero by construction, R12 is a full krank-by-(n-krank)
!   matrix and R22 is a (m-krank)-by-(n-krank) matrix.
!
!   In a final step, if TAU is present, R22 is considered to be negligible
!   and R12 is annihilated by orthogonal transformations from the right, arriving at
!   the partial or complete orthogonal factorization:
!
!                       MAT * P = Q * T * Z
!
!   , where P is a n-by-n permutation matrix, Q is a m-by-m orthogonal matrix,
!   Z is a n-by-n orthogonal matrix and T is a m-by-n matrix and has the form:
!
!   [  T11  T12  ]
!
!   [  T21  T22  ]
!
!   Here T21 (=R21) and T12 are all zero, T22 (=R22) is considered to be negligible and
!   T11 is a krank-by-krank upper triangular matrix.
!   
!   On exit, P is stored compactly in the vector argument IP and if:
!   
!   - TAU is absent, PARTIAL_RQR_CMP2 computes Q and submatrices R11, R12 and R22. Submatrices R11,
!     R12 and R22 are stored in the array arguments MAT and DIAGR.
!     Q is stored compactly in factored form in the array arguments MAT and BETA.
!   
!   - TAU is present, PARTIAL_RQR_CMP2 computes Q, Z and submatrices T11 and T22 (=R22). Submatrices T11
!     and T22 are stored in the array arguments MAT and DIAGR.
!     Q is stored compactly in factored form in the array arguments MAT and BETA.
!     Z is stored compactly in factored form in the array arguments MAT and TAU.
!   
!   See Further Details for more information on how the (partial or complete) QR or orthogonal
!   decomposition is stored in MAT on exit.
!   
!   PARTIAL_RQR_CMP2 performs the same task as subroutine PARTIAL_QR_CMP in module QR_Procedures,
!   but is much faster on large matrices because of the use of a randomized and blocked "BLAS3"
!   algorithm instead of a standard "BLAS2" algorithm. Note, however, that PARTIAL_RQR_CMP2 is
!   an effective and efficient way for computing a low-rank approximation of MAT, but is less effective
!   to find the rank of MAT because of the use of randomization. As an illustration, the diagonal elements
!   of R11 are not necessarily of decreasing magnitude when computed by PARTIAL_RQR_CMP2, while this property
!   is enforced with PARTIAL_QR_CMP.
!   
!   On the other hand, PARTIAL_QR_CMP can be used for both tasks, but is much slower than PARTIAL_RQR_CMP2.
!   
!   Note that PARTIAL_RQR_CMP2 performs also exactly the same task as subroutine PARTIAL_RQR_CMP in module Random.
!   The main difference between the two routines is that PARTIAL_RQR_CMP2 recomputes the Gaussian and
!   compression matrices at each iteration of the randomized (partial or complete) QR algorithm, while PARTIAL_RQR_CMP uses
!   an efficient updating formulae to recompute the compression matrix at each iteration. In other words, PARTIAL_RQR_CMP
!   is usually faster than PARTIAL_RQR_CMP2, but may be sligthly less robust. See the references (3), (4), (5) and (6)
!   for further information.
!   
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the real m-by-n matrix to be decomposed.
!
!            On exit, MAT has been overwritten by details of its (partial)
!            QR or orthogonal factorization.
!
!            See Further Details.
!
!   DIAGR    (OUTPUT) real(stnd), dimension(:) 
!            On exit, the diagonal elements of the matrix R11 or T11.
!
!            See Further Details.
!
!            The size of DIAGR must verify:
!
!            - size( DIAGR ) = kpartial <= min( size(MAT,1) , size(MAT,2) ).
!
!   BETA     (OUTPUT) real(stnd), dimension(:)
!            On exit, the scalars factors of the elementary reflectors defining Q.
!
!            See Further Details.
!
!            The size of BETA must verify:
!
!            - size( BETA ) = kpartial <= min( size(MAT,1) , size(MAT,2) ).
!
!   IP       (OUTPUT) integer(i4b), dimension(:) 
!            On exit, if IP(j)=k, then the j-th column of MAT*P was 
!            the k-th column of MAT.
!
!            See Further Details.
!
!            The size of IP must be equal to size( MAT, 2 ) = n.
!
!   KRANK    (OUTPUT) integer(i4b)
!            On exit, KRANK contains the effective rank of R11, i.e., krank, which is
!            the order of this submatrix R11. This is the same as the order of
!            the submatrix T11 in the complete orthogonal factorization of MAT and is also
!            the rank of MAT if krank is less than kpartial = size( BETA ).
!
!            If the computed pseudo-rank, krank, is less than kpartial = size( BETA ),
!            BETA(krank+1:kpartial) and, eventually, TAU(krank+1:kpartial) are set to zero
!            and MAT(krank+1:m,krank+1:n) (e.g., R22=T22) is updated on exit.
!            In other words, the subroutine outputs a partial QR factorization of rank krank
!            instead of rank kpartial.
!
!            In all cases, norm(MAT(krank+1:m,krank+1:n)) gives the error of the associated
!            matrix approximation in the Frobenius norm, on exit.
!
!   TOL      (INPUT/OUTPUT, OPTIONAL) real(stnd)
!            On entry, if TOL is present  then:
!
!            - if TOL is in ]0,1[, the calculations to determine the condition number of R11
!              are performed. Then, TOL is used to determine the effective pseudo-rank
!              of R11, which is defined as the order of the largest leading triangular
!              submatrix in the partial QR factorization with column pivoting of MAT,
!              whose estimated condition number in the 1-norm is less than 1/TOL.
!              On exit, the reciprocal of the condition number is returned in TOL.
!            - if TOL=0 is specified, the calculations to determine the condition number
!              of R11 are not performed and crude tests on R(j,j) are done to determine
!              the numerical pseudo-rank of R11. On exit, TOL is not changed.
!
!            If TOL is not specified or is outside [0,1[, the calculations to determine the
!            rank of R11 are not performed and this rank is assumed to be equal to kpartial.
!
!   TAU      (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if TAU is present, a (partial) orthogonal factorization of
!            MAT is computed. Otherwise, a simple QR factorization with column pivoting
!            of MAT is computed.
!
!            On exit, the scalars factors of the elementary reflectors defining the
!            orthogonal matrix Z in the (partial) orthogonal factorization of MAT.
!
!            See Further Details.
!
!            The size of TAU must verify:
!
!            - size( TAU ) = kpartial <= min( size(MAT,1) , size(MAT,2) ).
!
!   RNG_ALG  (INPUT, OPTIONAL) integer(i4b)
!            On entry, a scalar integer to select the random (uniform) number generator used to build
!            the random gaussian matrix in the randomized partial QR algorithm.
!
!            The possible values are:
!
!            - ALG=1  : selects the Marsaglia's KISS random number generator;
!            - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!            - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!            - ALG=4  : selects the Mersenne Twister random number generator;
!            - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!            - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!            - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!            - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!            - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!            - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!              random number generator;
!
!            For other values, the current random number generator and its current state are not changed.
!            Note further, that, on exit, the current random number generator is not reset to its previous
!            value before the call to PARTIAL_RQR_CMP2.
!
!            See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   BLK_SIZE (INPUT, OPTIONAL) integer(i4b)
!            On entry, the block size used in the randomized partial QR factorization.
!
!            BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!            smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!            By default, BLK_SIZE is set to min( BLKSZ_QR, min(m,n) ), where parameter BLKSZ_QR is the default
!            block size for QR related algorithms specified in module Select_Parameters.
!
!   NOVER    (INPUT, OPTIONAL) integer(i4b)
!            The oversampling size used in the randomized partial QR algorithm.
!
!            NOVER must be positive or null and verifies the relationship:
!
!                 NOVER + BLK_SIZE <= size(MAT,1)
!
!            and is adjusted if necessary to verify this relationship in all cases.
!
!            See Further Details and the cited references for the meaning and usefulness
!            of the oversampling size in the partial randomized QR algorithm.
!
!            By default, the oversampling size is set to 10.
!
!
! Further Details
! _______________
!
!   The matrix Q is represented as a product of elementary reflectors
!
!            Q = H(1) * H(2) * ... * H(krank), where krank = size( BETA ) <= min( m , n ).
!
!   Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!   where beta is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in MAT(i:m,i) and beta in BETA(i).
!
!   On exit of PARTIAL_RQR_CMP2, the orthonormal matrix Q (or its first n columns) can be computed
!   explicitly by a call to subroutine ORTHO_GEN_QR with arguments MAT and BETA.
!
!   The matrix P is represented in the array IP as follows:
!   If IP(j) = i then the jth column of P is the ith canonical unit vector.
!
!   On exit, if the optional argument TAU is absent:
!
!   - The elements above the diagonal of the array MAT(:krank,:krank) contain the 
!     corresponding elements of the triangular matrix R11.
!   - The elements of the diagonal of R11 are stored in the array DIAGR.
!   - The submatrix R12 is stored in MAT(:krank,krank+1:n).
!   - The submatrix R22 is stored in MAT(krank+1:m,krank+1:n).
!   - krank is stored in the real argument KRANK.
!
!   If TAU is present, a partial or complete orthogonal factorization of MAT is computed.
!   The factorization is obtained by Householder's method. The kth transformation
!   matrix, Z(k), which is used to introduce zeros into the kth row of R (e.g., in
!   R12), is given in the form
!
!   [  I  0    ]
!
!   [  0  T(k) ]
!
!   where
!
!       T(k) = I + tau * ( u(k) * u(k)' ) and u(k)' = (  1  0  z(k)  )
!
!   tau is a scalar, u(k) is a n-k+1 vector and z(k) is an (n-krank) element vector.
!   tau and z(k) are chosen to annihilate the elements of the kth row of R12.
!
!   The Z n-by-n orthogonal matrix is given by
!
!        Z = Z(1) * Z(2) * ... * Z(krank)
!
!   On exit, if the optional argument TAU is present:
!
!   - The scalar tau defining T(k) is returned in the kth element of TAU and the vector u(k)
!     in the kth row of MAT, such that the elements of z(k) are in MAT(k,krank+1:n).
!     The other elements of u(k) are not stored.
!   - The elements above the diagonal of the array section MAT(:krank,:krank)
!     contain the corresponding elements of the triangular matrix T11. The elements of
!     the diagonal of T11 are stored in the array DIAGR.
!   - The submatrix T22 (=R22) is stored in MAT(krank+1:m,krank+1:n).
!   - krank is stored in the real argument KRANK.
!
!   In both cases, norm(MAT(krank+1:m,krank+1:n)) gives the error of the associated
!   matrix approximation in the Frobenius norm, on exit.
!
!   If it is possible that MAT may not be of full rank (i.e., certain columns of MAT
!   are linear combinations of other columns), then the eventual linearly dependent columns
!   in the partial QR decomposition of MAT, which is sought, can be determined by using
!   TOL=relative precision of the elements in MAT.
!   If each element is correct to, say, 5 digits then TOL=0.00001 should
!   be used. Also, it may be helpful to scale the columns of MAT so that all elements 
!   are about the same order of magnitude.
!
!   The computations are parallelized if OPENMP is used. Note also that PARTIAL_RQR_CMP2
!   uses a randomized "BLAS3" algorithm described in the references (3), (4), (5) and (6), which
!   has about the same efficiency of a "BLAS3" QR algorithm without column pivoting and is thus
!   highly efficient on large matrices.
!
!   For further details, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Duersch, J.A., and Gu, M., 2017:
!           Randomized QR with column pivoting.
!           SIAM J. Sci. Comput., Volume 39, Issue 4, C263-C291.
!
!   (4) Martinsson, P.G., Quintana-Orti, G., Heavner, N., and Van de Geijn, R., 2017:
!           Householder QR factorization with randomization for column pivoting (HQRRP).
!           SIAM J. Sci. Comput., Volume 39, Issue 2, C96-C115.
!
!   (5) Xiao, J., Gu, M., and Langou, J., 2017:
!           Fast parallel randomized QR with column pivoting algorithms
!           for reliable low-rank matrix approximations.
!           IEEE 24th International Conference
!           on High Performance Computing (HiPC), IEEE, 2017, 233-242.
!
!   (6) Duersch, J.A., and Gu, M., 2020:
!           Randomized projection for rank-revealing matrix factorizations
!           and low-rank approximations.
!           SIAM Review, Volume 62, Issue 3, 661-682.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Num_Constants,   only : machhuge, safmin
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert_eq, assert, merror, arth
    use Hous_Procedures, only : hous1, hous2, apply_hous2
    use QR_Procedures,   only : partial_qr_cmp
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: diagr, beta
    real(stnd),   intent(inout),                 optional   :: tol
    real(stnd),   intent(out),   dimension(:),   optional   :: tau
!
    integer(i4b), intent(out)               :: krank
    integer(i4b), intent(out), dimension(:) :: ip
    integer(i4b), intent(in), optional :: rng_alg, blk_size, nover
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: anorm, xnorm, rcond, prod, beta1
    real(stnd), dimension(:),   allocatable :: diagr_b, beta_b
    real(stnd), dimension(:,:), allocatable :: b, omega, lmat, umat, mat2
#ifdef _BLAS
    real(stnd), dimension(size(mat,2))      :: tmp
    real(stnd), dimension(:,:), allocatable :: x
#else
    real(stnd), dimension(max(size(mat,1),size(mat,2))) :: tmp
    real(stnd), dimension(:), allocatable               :: x
#endif
!
    integer      :: iok, alg
    integer(i4b) :: m, n, ndiag, p, nover2, blk_size2, nbr, in, nblk, &
                    i, i1, i2, i3, itmp, j, j2, l, l1, l2
    integer(i4b), dimension(:), allocatable :: ip_b
!
    logical(lgl) :: comp_min_norm, test
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='partial_rqr_cmp2'
!
!   nover0 IS THE DEFAULT OVERSAMPLING SIZE.
!
    integer(i4b),  parameter :: nover0=10_i4b
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b) ,   &
                   int(size(ip),i4b)    ,   &
                   name_proc )
!
    ndiag = min( m, n )
!
    comp_min_norm = present( tau )
!
    if ( comp_min_norm ) then
!
        krank =  assert_eq( int(size(beta),i4b)  ,   &
                            int(size(diagr),i4b) ,   &
                            int(size(tau),i4b)   ,   &
                            name_proc )
!
        tau(:krank) = zero
!
    else
!
        krank =  assert_eq( int(size(beta),i4b)  ,   &
                            int(size(diagr),i4b) ,   &
                            name_proc )
!
    end if
!
    call assert( logical( krank<=ndiag, lgl ), name_proc )
!
!   INITIALIZE PERMUTATION VECTOR FOR COLUMN PIVOTING.
!
    ip(:n) = arth( 1_i4b, 1_i4b, n )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( krank<=0_i4b ) then
!
        if ( present(tol) ) tol = machhuge
!
        return
!
    end if
!
!   SELECT AND SET THE RANDOM NUMBER GENERATOR TO BE USED IF REQUIRED.
!
    if ( present(rng_alg) ) then
!
        if ( rng_alg>=1_i4b .and. rng_alg<=10_i4b ) then
!
            alg = rng_alg
!
!           SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
            call random_seed_( alg=alg )
!
!           RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
            call random_seed_()
!
        end if
!
    end if
!
!   DETERMINE THE OVERSAMPLING SIZE nover2 .
!
    if ( present(nover) ) then
        nover2 = max( nover, 0_i4b )
    else
        nover2 = nover0
    end if
!
!   DETERMINE THE BLOCKSIZE blk_size2 .
!
    if ( present(blk_size) ) then
        blk_size2 = max( blk_size, 1_i4b )
    else
        blk_size2 = blksz
    end if
!    
    blk_size2 = min( blk_size2, ndiag )
!    
!   COMPUTE RANDOMIZED PARTIAL QR FACTORIZATION WITH COLUMN PIVOTING OF mat(:m,:n) .
!
!   nbr IS THE NUMBER OF ROWS IN MATRICES b and omega .
!
    nbr = min( nover2 + blk_size2, m )
!
!   ALLOCATE WORK ARRAYS, b IS THE COMPRESSION MATRIX
!   AND omega IS A GAUSSIAN RANDOM MATRIX.
!
    allocate( b(nbr,n), omega(nbr,m), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A nbr-BY-m IID GAUSSIAN MATRIX.
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
    call normal_random_number3_( omega(:nbr,:m) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
!    do i = 1_i4b, m
!
!        call normal_random_number3_( omega(:nbr,i) )
!
!    end do
!
!   COMPUTE COMPRESSION MATRIX.
!
#ifdef _BLAS
    call gemm( 'N', 'N', nbr, n, m, one, omega(1_i4b:nbr,1_i4b:m), nbr,  &
               mat(1_i4b:m,1_i4b:n), m, zero, b(1_i4b:nbr,1_i4b:n), nbr  )
#else
    b(:nbr,:n) = matmul( omega(:nbr,:m), mat(:m,:n) )
#endif
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  n*m>=omp_limit                      .and.       &
                  n/blk_size2>1_i4b                   .and.       &
                  i>1_i4b
#endif
!
    if ( krank<=blk_size2 ) then
!
!       krank LESS OR EQUAL TO blk_size2, USE A FAST RANDOMIZED QR FACTORIZATION OF mat WITHOUT
!       UPDATES OF THE COMPRESSION MATRIX.
!
!       DO PARTIAL QR WITH COLUMN PIVOTING ON COMPRESSION MATRIX b(:nbr,:n) TO FIND krank NEW PIVOTS.
!           
        call partial_qr_cmp( b(:nbr,:n), diagr(:krank), beta(:krank), ip(:n), i )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( b, omega )
!
!       SWAP CORRESPONDING COLUMNS OF mat(:m,:n) .
!
        ip(:n) = -ip(:n)
!
        do i = 1_i4b, n
!
            if ( ip(i)>0_i4b ) cycle
!
            l     = i
            ip(l) = -ip(l)
            in    = ip(l)
!
            do
!
                if ( ip(in)>0_i4b ) exit
!
                if ( l/=in ) then
#ifdef _BLAS
                    call swap( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,in), 1_i4b )
#else
                    tmp(1_i4b:m)    = mat(1_i4b:m,l)
                    mat(1_i4b:m,l)  = mat(1_i4b:m,in)
                    mat(1_i4b:m,in) = tmp(1_i4b:m)
#endif
!
                end if
!
                ip(in) = -ip(in)
                l      = in
                in     = ip(in)
!
            end do
!
        end do
!
!       COMPUTE H(1), ... , H(krank) AND DO QR DECOMPOSITION 
!       OF THE FIRST BLOCK OF mat .
!
        do i = 1_i4b, krank
!    
!           GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+1:m,i) .
!    
            call hous1( mat(i:m,i), beta(i), diagr(i) )
!    
!           APPLY H(i) TO mat(i:m,i+1:krank) FROM THE LEFT.
!
            beta1 = beta(i)
!
            do l = i + 1_i4b, krank
!
#ifdef _BLAS
                prod = beta1*dot( m-i+1_i4b, mat(i:m,i), 1_i4b, mat(i:m,l), 1_i4b )
                call axpy( m-i+1_i4b, prod, mat(i:m,i), 1_i4b, mat(i:m,l), 1_i4b )
#else
                prod = beta1*dot_product( mat(i:m,i), mat(i:m,l) )
                mat(i:m,l) = mat(i:m,l) + prod*mat(i:m,i)
#endif
!    
            end do
!
        end do
!    
        if ( krank<n ) then
!
!           BLOCK UPDATE OF THE TRAILING MATRIX mat(:m,krank+1:n) IF REQUIRED.
!
!           ALLOCATE WORK ARRAY.
!
            allocate( lmat(krank,krank), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           INITIALIZE THE UNIT LOWER TRIANGULAR MATRIX lmat(:krank,:krank)
!           USED IN THE BLOCK UPDATE ALGORITHM AND SAVE TRIANGULAR FACTOR
!           R OF QR FACTORIZATION OF THE FIRST BLOCK OF mat IN UPPER TRIANGLE
!           OF lmat IF BLAS IS USED.
!
#ifdef _BLAS
            do i = 1_i4b, krank
!
                lmat(1_i4b:i,i)       = mat(1_i4b:i,i)
                lmat(i+1_i4b:krank,i) = zero
!
                mat(1_i4b:i-1_i4b,i)  = zero
!
            end do
#else
            lmat(:krank,:krank) = zero
!
            do i = 1_i4b, krank
!
                lmat(i,i) = one
!
            end do
#endif
!
!$OMP PARALLEL IF(test_par)                 &
!$OMP         ,PRIVATE(iok,i,i1,i2,i3,l,l2,x)
!
!           ALLOCATE WORK VARIABLES.
!
#ifdef _BLAS
            allocate( x(krank,blk_size2), stat = iok )
#else
            allocate( x(krank), stat = iok )
#endif
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!    
!           APPLY H(krank)* ...*H(1) TO mat(:m,krank+1:n) FROM THE LEFT
!           WITH A BLOCKED ALGORITHM AS
!
!            ( I + U(:m,:krank)*inv( lmat(:krank,:krank) )*U(:m,:krank)' )*mat(:m,krank+1:n) .
!
!           FIRST COMPUTE THE krank BY krank UNIT LOWER TRIANGULAR MATRIX lmat(:krank,:krank)
!           USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do l = 1_i4b, krank
!
                do i = l + 1_i4b, krank
!
#ifdef _BLAS
                    lmat(i,l) = -beta(i)*dot( m-i+1_i4b, mat(i:m,l), 1_i4b, mat(i:m,i), 1_i4b )
#else
                    lmat(i,l) = -beta(i)*dot_product( mat(i:m,l), mat(i:m,i) )
#endif
!
                end do
!                    
            end do
!
!$OMP END DO
!
!           BLOCK UPDATE OF THE TRAILING MATRIX mat(:m,krank+1:n) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do i = krank, n-1_i4b, blk_size2
!
                i1 = i + 1_i4b
                i2 = min( i+blk_size2, n )
                i3 = i2 - i
!
#ifdef _BLAS
!
!               COMPUTE U(:m,:krank)'*mat(:m,i1:i2) .
!
                call gemm( 'T', 'N', krank, i3, m, one, mat, m, mat(1_i4b:m,i1:i2), m, zero, x, krank )
!
                do l = 1_i4b, i3
!
                    x(:krank,l) = beta(:krank)*x(:krank,l)
!                    
                end do
!
!               SOLVE lmat(:krank,:krank)*x(:krank,:i3) = U(:m,:krank)'*mat(:m,i1:i2) .
!
                call trsm( 'L', 'L', 'N', 'U', krank, i3, one, lmat, krank, x, krank )
!
!               FULL VERSION...
!
!               COMPUTE mat(:m,i1:i2) = mat(:m,i1:i2) + U(:m,:krank)*x(:krank,:i3) .
!
                call gemm( 'N', 'N', m, i3, krank, one, mat, m, x, krank, one, mat(1_i4b:m,i1:i2), m )
!
!               TRUNCATED VERSION...
!
!               COMPUTE mat(:krank,i1:i2) = mat(:krank,i1:i2) + U(:krank,:krank)*x(:krank,:i3) .
!
!                call gemm( 'N', 'N', krank, i3, krank, one, mat, m, x, krank, one, mat(1_i4b:m,i1:i2), m )
#else
                do l = i1, i2
!
!                   COMPUTE U(:m,:krank)'*mat(:m,i1:i2) .
!
                    do l2 = 1_i4b, krank
!
                        x(l2) = beta(l2)*dot_product( mat(l2:m,l2), mat(l2:m,l) )
!                    
                    end do
!
!                   SOLVE lmat(:krank,:krank)*x(:krank,:i3) = U(:m,:krank)'*mat(:m,i1:i2) .
!                   COMPUTE mat(:m,i1:i2) = mat(:m,i1:i2) + U(:m,:krank)*x(:krank,:i3) .
!
                    do l2 = 1_i4b, krank
!
                        x(l2+1_i4b:krank) = x(l2+1_i4b:krank) - x(l2)*lmat(l2+1_i4b:krank,l2)
!
!                       FULL VERSION...
!
                        mat(l2:m,l) = mat(l2:m,l) + x(l2)*mat(l2:m,l2)
!
!                       TRUNCATED VERSION...
!
!                        mat(l2:krank,l) = mat(l2:krank,l) + x(l2)*mat(l2:krank,l2)
!
                    end do
!
                end do
#endif
!                    
            end do
!
!$OMP END DO
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( x )
!
!$OMP END PARALLEL
!
!           RESTORE TRIANGULAR FACTOR R OF QR FACTORIZATION
!           OF THE FIRST BLOCK OF mat IF BLAS IS USED.
!
#ifdef _BLAS
            do i = 2_i4b, krank
!
                mat(1_i4b:i-1_i4b,i)  = lmat(1_i4b:i-1_i4b,i)
!
            end do
#endif
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( lmat )
!
        end if
!
    else
!
!       krank GREATER THAN blk_size2, USE RANDOMIZED QR FACTORIZATION OF mat WITH
!       UPDATES OF THE COMPRESSION MATRIX.
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( ip_b(n), diagr_b(blk_size2), beta_b(blk_size2),        &
                  lmat(blk_size2,blk_size2), umat(m,blk_size2), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       INITIALIZE THE UNIT LOWER TRIANGULAR MATRIX lmat(:blk_size2,:blk_size2)
!       USED IN THE BLOCKED QR ALGORITHM.
!
        lmat(1_i4b:blk_size2,1_i4b:blk_size2) = zero
!
        do i = 1_i4b, blk_size2
            lmat(i,i) = one
        end do
!
!       INITIALIZE ARRAY FOR STORING HOUSEHOLDER VECTORS USED IN THE BLOCKED QR
!       ALGORITHM.
!
        umat(:m,:blk_size2) = zero
!
!$OMP PARALLEL IF(test_par)                                         &
!$OMP         ,PRIVATE(test,iok,i,i1,i2,i3,j,j2,l,l1,l2,in,itmp,nblk,p,beta1,prod,x,mat2,tmp)
!
!       ALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
        allocate( x(blk_size2,blk_size2), mat2(m,blk_size2), stat = iok )
#else
        allocate( x(blk_size2), mat2(m,blk_size2), stat = iok )
#endif
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       RANDOMIZED BLOCKED PARTIAL QR DECOMPOSITION OF mat WITH COLUMN PIVOTING.
!
        do j = 1_i4b, krank, blk_size2
!
            nblk = min( krank-j+1_i4b, blk_size2 )
!
            j2 = j + nblk - 1_i4b
            p  = m - j + 1_i4b
!
            test = nbr < m - j2 .and. j2 < krank
!
!$OMP SECTIONS
!
!$OMP SECTION
!
!           GENERATE A nbr-BY-m IID GAUSSIAN MATRIX.
!
!           GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!           FORM OF SUBROUTINE normal_random_number3_.
!
            call normal_random_number3_( omega(:nbr,j2+1_i4b:m) )
!
!           GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!           FORM OF SUBROUTINE normal_random_number_.
!
!            do i = j2+1_i4b, m
!
!                call normal_random_number3_( omega(:nbr,i) )
!
!            end do
!
!$OMP SECTION
!
!           DO PARTIAL QR WITH COLUMN PIVOTING ON COMPRESSION MATRIX b(:nbr,j:n) TO FIND nblk NEW PIVOTS.
!           
            call partial_qr_cmp( b(:nbr,j:n), diagr_b(:nblk), beta_b(:nblk), ip_b(j:n), i )
!
!           SWAP CORRESPONDING COLUMNS OF mat(:m,j:n) .
!
            ip_b(j:n) = -ip_b(j:n)
!
            do i = j, n
!
                if ( ip_b(i)>0_i4b ) cycle
!
                l       = i
                ip_b(l) = -ip_b(l)
                in      = ip_b(l) + j - 1_i4b
!
                do
!
                    if ( ip_b(in)>0_i4b ) exit
!
                    if ( l/=in ) then
#ifdef _BLAS
                        call swap( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,in), 1_i4b )
#else
                        tmp(1_i4b:m)    = mat(1_i4b:m,l)
                        mat(1_i4b:m,l)  = mat(1_i4b:m,in)
                        mat(1_i4b:m,in) = tmp(1_i4b:m)
#endif
!
                        itmp   = ip(l)
                        ip(l)  = ip(in)
                        ip(in) = itmp
!
                    end if
!
                    ip_b(in) = -ip_b(in)
                    l        = in
                    in       = ip_b(in) + j - 1_i4b
!
                end do
!
            end do
!
!           COMPUTE H(j), ... , H(j+nblk-1_i4b) AND DO QR DECOMPOSITION 
!           OF THE CURRENT BLOCK OF mat .
!
            do i = j, j2
!    
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+1:m,i) .
!    
                call hous1( mat(i:m,i), beta(i), diagr(i) )
!    
!               APPLY H(i) TO mat(i:m,i+1:j+nblk-1) FROM THE LEFT.
!
                beta1 = beta(i)
!
                do l = i + 1_i4b, j2
!
#ifdef _BLAS
                    prod = beta1*dot( m-i+1_i4b, mat(i:m,i), 1_i4b, mat(i:m,l), 1_i4b )
                    call axpy( m-i+1_i4b, prod, mat(i:m,i), 1_i4b, mat(i:m,l), 1_i4b )
#else
                    prod = beta1*dot_product( mat(i:m,i), mat(i:m,l) )
                    mat(i:m,l) = mat(i:m,l) + prod*mat(i:m,i)
#endif
!    
                end do
!    
            end do
!
!$OMP END SECTIONS
!    
            if ( j2<n ) then
!    
!               APPLY H(j+nblk-1_i4b)* ...*H(j) TO mat(j:m,j+nblk:n) FROM THE LEFT
!               WITH A BLOCKED ALGORITHM AS
!
!            ( I + U(j:m,:nblk)*inv( lmat(:nblk,:nblk) )*U(j:m,:nblk)' )*mat(j:m,j+nblk:n) .
!
!               FIRST SAVE H(j) ...H(j+nblk-1_i4b) FOR LATER USE AND COMPUTE THE nblk BY nblk UNIT
!               LOWER TRIANGULAR MATRIX lmat(:nblk,:nblk) USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, nblk
!
                    l1 = j + l - 1_i4b
!
!                   SAVE SHIFTED HOUSEHOLDER VECTOR mat(j+l-1:m,j+l-1) .
!
                    umat(l:p,l) = mat(l1:m,l1)
!
                    do i = l + 1_i4b, nblk
!
                        l2 = j + i - 1_i4b
!
#ifdef _BLAS
                        lmat(i,l) = -beta(l2)*dot( p-i+1_i4b, umat(i:p,l), 1_i4b, mat(l2:m,l2), 1_i4b )
#else
                        lmat(i,l) = -beta(l2)*dot_product( umat(i:p,l), mat(l2:m,l2) )
#endif
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!               BLOCK UPDATE OF THE TRAILING MATRIX mat(j:m,j+nblk:n) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do i = j2, n-1_i4b, nblk
!
                    i1 = i + 1_i4b
                    i2 = min( i+nblk, n )
                    i3 = i2 - i
!
!                   MAKE A LOCAL COPY OF mat(j:m,i1:i2) .
!
                    mat2(:p,:i3) = mat(j:m,i1:i2)
!
#ifdef _BLAS
!
!                   COMPUTE U(j:m,:nblk)'*mat(j:m,i1:i2) .
!
                    call gemm( 'T', 'N', nblk, i3, p, one, umat, m, mat2, m, zero, x, blk_size2 )
!
                    do l = 1_i4b, i3
!
                        x(:nblk,l) = beta(j:j2)*x(:nblk,l)
!                    
                    end do
!
!                   SOLVE lmat(:nblk,:nblk)*x(:nblk,:i3) = U(j:m,:nblk)'*mat(j:m,i1:i2) .
!
                    call trsm( 'L', 'L', 'N', 'U', nblk, i3, one, lmat, blk_size2, x, blk_size2 )
!
!                   COMPUTE mat(j:m,i1:i2) = mat(j:m,i1:i2) + U(j:m,:nblk)*x(:nblk,:i3) .
!
                    call gemm( 'N', 'N', p, i3, nblk, one, umat, m, x, blk_size2, one, mat2, m )
#else
                    do l = 1_i4b, i3
!
!                       COMPUTE U(j:m,:nblk)'*mat(j:m,i1:i2) .
!
                        do l2 = 1_i4b, nblk
!
                            x(l2) = beta(j+l2-1_i4b)*dot_product( umat(l2:p,l2), mat2(l2:p,l) )
!                    
                        end do
!
!                       SOLVE lmat(:nblk,:nblk)*x(:nblk,:i3) = U(j:m,:nblk)'*mat(j:m,i1:i2) .
!                       COMPUTE mat(j:m,i1:i2) = mat(j:m,i1:i2) + U(j:m,:nblk)*x(:nblk,:i3) .
!
                        do l2 = 1_i4b, nblk
!
                            x(l2+1_i4b:nblk) = x(l2+1_i4b:nblk) - x(l2)*lmat(l2+1_i4b:nblk,l2)
!
                            mat2(l2:p,l) = mat2(l2:p,l) + x(l2)*umat(l2:p,l2)
!
                        end do
!
                    end do
#endif
!
!                   SAVE UPDATE OF mat(j:m,i1:i2) .
!
                    mat(j:m,i1:i2) = mat2(:p,:i3)           
!
!                   COMPUTE COMPRESSION MATRIX FOR NEXT PANEL QR ITERATION.
!
                    if ( test ) then
!
#ifdef _BLAS
                        mat2(:nblk,:i3) = zero
!
                        call gemm( 'N', 'N', nbr, i3, p, one, omega(1_i4b:nbr,j:m), nbr,  &
                                   mat2, m, zero, b(1_i4b:nbr,i1:i2), nbr  )
#else
                        b(:nbr,i1:i2) = matmul( omega(:nbr,j2+1_i4b:m), mat2(nblk+1_i4b:p,:i3) )
#endif
!                    
                    else if ( j2 < krank ) then
!
                        l = m - j2
!                    
                        b(:l,i1:i2)          = mat2(nblk+1_i4b:p,:i3) 
                        b(l+1_i4b:nbr,i1:i2) = zero 
!                    
                    end if
!                    
                end do
!
!$OMP END DO
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( x, mat2 )
!
!$OMP END PARALLEL
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( b, ip_b, diagr_b, beta_b, omega, umat, lmat )
!    
    end if
!    
!   TEST TOLERANCE.
!    
    if ( present(tol) ) then
        test = zero<=tol .and. tol<one
    else
        test = false
    end if
!    
    if ( test ) then
!    
        tmp(:krank) = abs( diagr(:krank) )
!    
        if ( tol/=zero ) then
!    
!           DETERMINE THE PSEUDO-RANK USING THE TOLERANCE, tol, AND
!           THE CONDITION NUMBER IN THE 1-NORM OF THE TRIANGULAR FACTOR R
!           OF THE QR FACTORIZATION.
!    
            anorm = zero
            xnorm = zero
            rcond = zero
!
            do j = 1_i4b, krank
!
!               CHECK IF abs( R(j,j) ) IS SMALL AND IF R(:j,:j) IS SINGULAR.
!               IF R(:j,:j) IS SINGULAR, EXIT AND THE PSEUDO-RANK IS j-1 .
!
                if ( tmp(j)<safmin ) exit
!
!               COMPUTE 1-NORM OF R(:j,:j).
!
                anorm  = max( anorm, sum(abs(mat(1_i4b:j-1_i4b,j)))+tmp(j) )
!
!               COMPUTE 1-NORM OF inv( R(:j,:j) ).
!
                tmp(j)             = one
                tmp(1_i4b:j-1_i4b) = zero
!
#ifdef _BLAS
                mat(j,j) = diagr(j)
!
                call trsv( 'U', 'N', 'N', j, mat, m, tmp(1_i4b:j), 1_i4b )    
#else
                do i = j, 1_i4b, -1_i4b
                    tmp(i)             = tmp(i)/diagr(i)
                    tmp(1_i4b:i-1_i4b) = tmp(1_i4b:i-1_i4b) - tmp(i)*mat(1_i4b:i-1_i4b,i)
                end do
#endif
!
                xnorm = max( xnorm, sum(abs(tmp(1_i4b:j))) )
!
!               COMPUTE INVERSE OF CONDITION NUMBER OF R(:j,:j).
!
                prod = one/( anorm*xnorm )
!
!               EXIT IF INVERSE OF CONDITION NUMBER IS LESS THAN THE TOLERANCE.
!
                if ( tol>=prod ) exit
!
                rcond = prod
!
            end do
!
            tol = rcond
!
#ifdef _BLAS
!
!           RESTORE FIRST ELEMENT OF HOUSEHOLDER VECTORS IF BLAS IS USED.
!
            do i = 1_i4b, min(j,krank)
                mat(i,i) = one
            end do
#endif
!
        else
!    
!           DETERMINE THE NUMERICAL PSEUDO-RANK.
!    
            do j = 1_i4b, krank
                if ( tmp(j)<safmin ) exit
            end do
!
        end if
!    
!       UPDATE RANK AND PARTIAL QR FACTORIZATION OF mat ON EXIT IF mat IS RANK DEFICIENT.
!    
        beta(j:krank) = zero
!    
        do i = j, krank
!    
            mat(i,i)         = diagr(i)
            mat(i+1_i4b:m,i) = zero
!    
        end do
!    
        krank = j - 1_i4b
!    
    end if
!    
!   IF THE PSEUDORANK krank IS LESS THAN n AND THE MINIMUM_NORM SOLUTION IS REQUIRED,
!   COMPUTE HOUSEHOLDER DECOMPOSITION OF mat(1:krank,:n) .
!
    if ( krank/=n .and. comp_min_norm ) then
!    
        j = krank + 1_i4b
        l = n - krank
!
!       ALLOCATE WORK VARIABLES FOR STORING HOUSEHOLDER VECTORS.
!
        allocate( umat(l,krank), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        umat(:l,:krank) = transpose( mat(:krank,j:n) )
!    
        do i = krank, 1_i4b, -1_i4b
!    
!           COMPUTE THE I-TH HOUSEHOLDER TRANSFORMATION.
!    
            call hous2( diagr(i), umat(:l,i), tau(i) )
!    
!           APPLY THE I-TH HOUSEHOLDER TRANSFORMATION FROM THE RIGHT.
!    
            call apply_hous2( umat(:l,i), tau(i),                                &
                              mat(1_i4b:i-1_i4b,i), umat(:l,1_i4b:i-1_i4b), true )
!    
        end do
!
        mat(:krank,j:n) = transpose( umat(:l,:krank) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( umat )
!
    end if
!
!
! END OF SUBROUTINE partial_rqr_cmp2
! __________________________________
!
    end subroutine partial_rqr_cmp2
!
! =========================================================================================
!
    subroutine partial_rtqr_cmp( mat, diagr, beta, ip, krank, tol, tau, rng_alg, niter, nover )
!
! Purpose
! _______
!
!   PARTIAL_RTQR_CMP computes a randomized partial and truncated QR factorization with
!   column pivoting, or an orthogonal factorization, of a m-by-n matrix MAT:
!
!                     MAT * P ≈ Q * R
!
!   , where P is a n-by-n permutation matrix, R is an upper triangular or trapezoidal
!   kpartial-by-n matrix and Q is a m-by-kpartial matrix with orthogonal columns.
!
!   The randomized QR factorization is only partial, e.g., the subroutine
!   ends when the numbers of columns of Q is equal to a predefined value equals to
!   kpartial = size( DIAGR ) = size( BETA ) <= min(m,n).
!
!   This leads implicitly to the following partition of R:
!
!   [  R11  R12  ]
!
!   where R11 is a kpartial-by-kpartial triangular matrix and R12 is a full kpartial-by-(n-kpartial)
!   matrix.
!
!   Then, if the optional scalar argument TOL is present and:
!
!   - is in ]0,1[, the rank of R11 is determined by finding the submatrix of R11
!     which is defined as the largest leading submatrix whose estimated condition
!     number, in the 1-norm, is less than 1/TOL.
!   - is equal to 0, the rank of R11, krank, is determined by finding the largest
!     submatrix of R11 such that abs(R11[j,j])>0. 
!
!   In both cases, the order of this submatrix, krank, is the effective rank of R11
!   (and MAT if krank is less than kpartial or if krank=kpartial=min(m,n)).
!
!   If TOL is absent or outside [0,1[, the rank of R11 is not checked and is assumed to be
!   equal to kpartial.
!
!   If krank is less than kpartial, then MAT is not of full rank (i.e., certain columns of
!   MAT(:m,:kpartial) are linear combinations of other columns of MAT(:m,:kpartial)) and
!   krank is also an estimate of the rank of MAT.
!
!   This leads to a redefinition of the partition of Q = [  Q1  Q2  ], where
!   Q1 and Q2 are now m-by-krank and m-by-(kpartial-krank) orthonormal matrices, and a corresponding 
!   redefinition of the associated partition of R, where R11 is now a krank-by-krank 
!   triangular matrix and R12 is a full krank-by-(n-krank) matrix.
!
!   In a final step, if TAU is present, R12 is annihilated by orthogonal transformations from the right,
!   arriving at the partial orthogonal factorization:
!
!                       MAT * P ≈ Q * T * Z
!
!   , where P is a n-by-n permutation matrix, Q is a m-by-krank orthogonal matrix,
!   Z is a n-by-n orthogonal matrix and T is a krank-by-n matrix and has the form:
!
!   [  T11  T12  ]
!
!   Here T12 is all zero and T11 is a krank-by-krank upper triangular matrix.
!   
!   On exit, P is stored compactly in the vector argument IP and if:
!   
!   - TAU is absent, PARTIAL_RTQR_CMP computes Q and submatrices R11 and R12. Submatrices R11,
!     and R12 are stored in the array arguments MAT and DIAGR.
!     Q is stored compactly in factored form in the array arguments MAT and BETA.
!   
!   - TAU is present, PARTIAL_RTQR_CMP computes Q, Z and submatrix T11. Submatrix T11
!     is stored in the array arguments MAT and DIAGR.
!     Q is stored compactly in factored form in the array arguments MAT and BETA.
!     Z is stored compactly in factored form in the array arguments MAT and TAU.
!   
!   See Further Details for more information on how the partial QR or orthogonal decomposition
!   is stored in MAT on exit.
!   
!   PARTIAL_RTQR_CMP performs the same task as subroutines PARTIAL_RQR_CMP and PARTIAL_RQR_CMP2
!   subroutines in module Random, but is significantly faster when krank is relatively small and MAT is a
!   very large matrix. This is due to the fact that PARTIAL_RTQR_CMP computes Q (in factored form), R11 and R12,
!   but not R22 (where R22 is the bottom left (m-krank)-by-(n-krank) submatrix in the QR factorization of MAT)
!   as PARTIAL_RQR_CMP and PARTIAL_RQR_CMP2 subroutines. Furthermore, only an estimate of R12 is computed by
!   PARTIAL_RTQR_CMP, using a randomization algorithm described in the reference (7), while the computation of
!   R12 is exact in PARTIAL_RQR_CMP and PARTIAL_RQR_CMP2 subroutines.
!   
!   In other words, PARTIAL_RTQR_CMP avoids trailing updates of MAT (e.g., R22) during the randomized QR
!   factorization, which reduces significantly the CPU time compared to PARTIAL_RQR_CMP and PARTIAL_RQR_CMP2
!   subroutines, if krank is small and MAT is a very large matrix, as these subroutines perform these trailing updates.
!   
!   Note that PARTIAL_RTQR_CMP also does not recompute or update the compression matrix at each iteration
!   as PARTIAL_RQR_CMP and PARTIAL_RQR_CMP2 subroutines. See references (3), (4), (5), (6) and (7) for
!   further information.
!   
!   In summary, PARTIAL_RTQR_CMP subroutine is usually faster than PARTIAL_RQR_CMP and PARTIAL_RQR_CMP2 subroutines
!   but is less accurate, especially for matrices with a slow decay of their singular values. See reference (7) for details.
!   
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the real m-by-n matrix to be decomposed.
!
!            On exit, MAT has been overwritten by details of its partial
!            and truncated QR or orthogonal factorization.
!
!            See Further Details.
!
!   DIAGR    (OUTPUT) real(stnd), dimension(:) 
!            On exit, the diagonal elements of the matrix R11 or T11.
!
!            See Further Details.
!
!            The size of DIAGR must verify:
!
!            - size( DIAGR ) = kpartial <= min( size(MAT,1) , size(MAT,2) ).
!
!   BETA     (OUTPUT) real(stnd), dimension(:)
!            On exit, the scalars factors of the elementary reflectors defining Q.
!
!            See Further Details.
!
!            The size of BETA must verify:
!
!            - size( BETA ) = kpartial <= min( size(MAT,1) , size(MAT,2) ).
!
!   IP       (OUTPUT) integer(i4b), dimension(:) 
!            On exit, if IP(j)=k, then the j-th column of MAT*P was 
!            the k-th column of MAT.
!
!            See Further Details.
!
!            The size of IP must be equal to size( MAT, 2 ) = n.
!
!   KRANK    (OUTPUT) integer(i4b)
!            On exit, KRANK contains the effective rank of R11, i.e., krank, which is
!            the order of this submatrix R11. This is the same as the order of
!            the submatrix T11 in the complete orthogonal factorization of MAT and is also
!            the rank of MAT if krank is less than kpartial = size( BETA ).
!
!   TOL      (INPUT/OUTPUT, OPTIONAL) real(stnd)
!            On entry, if TOL is present  then:
!
!            - if TOL is in ]0,1[, the calculations to determine the condition number of R11
!              are performed. Then, TOL is used to determine the effective pseudo-rank
!              of R11, which is defined as the order of the largest leading triangular
!              submatrix in the partial QR factorization with column pivoting of MAT,
!              whose estimated condition number in the 1-norm is less than 1/TOL.
!              On exit, the reciprocal of the condition number is returned in TOL.
!            - if TOL=0 is specified, the calculations to determine the condition number
!              of R11 are not performed and crude tests on R(j,j) are done to determine
!              the numerical pseudo-rank of R11. On exit, TOL is not changed.
!
!            If TOL is not specified or is outside [0,1[, the calculations to determine the
!            rank of R11 are not performed and this rank is assumed to be equal to kpartial.
!
!   TAU      (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if TAU is present, a (partial) complete orthogonal factorization of
!            MAT is computed. Otherwise, a simple QR factorization with column pivoting
!            of MAT is computed.
!
!            On exit, the scalars factors of the elementary reflectors defining the
!            orthogonal matrix Z in the partial orthogonal factorization of MAT.
!
!            See Further Details.
!
!            The size of TAU must verify:
!
!            - size( TAU ) = kpartial <= min( size(MAT,1) , size(MAT,2) ).
!
!   RNG_ALG  (INPUT, OPTIONAL) integer(i4b)
!            On entry, a scalar integer to select the random (uniform) number generator used to build
!            the random gaussian matrix in the randomized partial QR algorithm.
!
!            The possible values are:
!
!            - ALG=1  : selects the Marsaglia's KISS random number generator;
!            - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!            - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!            - ALG=4  : selects the Mersenne Twister random number generator;
!            - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!            - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!            - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!            - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!            - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!            - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!              random number generator;
!
!            For other values, the current random number generator and its current state are not changed.
!            Note further, that, on exit, the current random number generator is not reset to its previous
!            value before the call to PARTIAL_RTQR_CMP.
!
!            See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   NITER    (INPUT, OPTIONAL) integer(i4b)
!            The number of randomized subspace iterations performed in the subroutine for improving
!            the accuracy of the compression matrix before computing its partial QR factorization
!            with column pivoting.
!
!            NITER must be positive or null.
!
!            By default, 0 randomized subspace iterations are performed.
!
!   NOVER    (INPUT, OPTIONAL) integer(i4b)
!            The oversampling size used in the randomized partial QR algorithm.
!
!            NOVER must be positive or null and verifies the relationship:
!
!                 NOVER + kpartial <= size(MAT,1)
!
!            and is adjusted if necessary to verify this relationship in all cases.
!
!            See Further Details and the cited references for the meaning and usefulness
!            of the oversampling size in the randomized partial and truncated QR algorithm.
!
!            By default, the oversampling size is set to max( kpartial/2_i4b, 10 ).
!
!
! Further Details
! _______________
!
!   The matrix Q is represented as a product of elementary reflectors
!
!            Q = H(1) * H(2) * ... * H(krank), where krank <= min( m , n ).
!
!   Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!   where beta is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in MAT(i:m,i) and beta in BETA(i).
!
!   On exit of PARTIAL_RTQR_CMP, the orthonormal matrix Q (or its first krank columns) can be computed
!   explicitly by a call to subroutine ORTHO_GEN_QR with arguments MAT(:m,:krank) and BETA(:krank).
!
!   The matrix P is represented in the array IP as follows:
!   If IP(j) = i then the jth column of P is the ith canonical unit vector.
!
!   On exit, if the optional argument TAU is absent:
!
!   - The elements above the diagonal of the array MAT(:krank,:krank) contain the 
!     corresponding elements of the triangular matrix R11.
!   - The elements of the diagonal of R11 are stored in the array DIAGR.
!   - The approximation of the submatrix R12 is stored in MAT(:krank,krank+1:n).
!   - krank is stored in the real argument KRANK.
!
!   If TAU is present, a partial complete orthogonal factorization of MAT is computed.
!   The factorization is obtained by Householder's method. The kth transformation
!   matrix, Z(k), which is used to introduce zeros into the kth row of R (e.g., in
!   R12), is given in the form
!
!   [  I  0    ]
!
!   [  0  T(k) ]
!
!   where
!
!       T(k) = I + tau * ( u(k) * u(k)' ) and u(k)' = (  1  0  z(k)  )
!
!   tau is a scalar, u(k) is a n-k+1 vector and z(k) is an (n-krank) element vector.
!   tau and z(k) are chosen to annihilate the elements of the kth row of R12.
!
!   The Z n-by-n orthogonal matrix is given by
!
!        Z = Z(1) * Z(2) * ... * Z(krank)
!
!   On exit, if the optional argument TAU is present:
!
!   - The scalar tau defining T(k) is returned in the kth element of TAU and the vector u(k)
!     in the kth row of MAT, such that the elements of z(k) are in MAT(k,krank+1:n).
!     The other elements of u(k) are not stored.
!   - The elements above the diagonal of the array section MAT(:krank,:krank)
!     contain the corresponding elements of the triangular matrix T11. The elements of
!     the diagonal of T11 are stored in the array DIAGR.
!   - krank is stored in the real argument KRANK.
!
!   If it is possible that MAT may not be of full rank (i.e., certain columns of MAT
!   are linear combinations of other columns), then the eventual linearly dependent columns
!   in the partial QR decomposition of MAT, which is sought, can be determined by using
!   TOL=relative precision of the elements in MAT.
!   If each element is correct to, say, 5 digits then TOL=0.00001 should
!   be used. Also, it may be helpful to scale the columns of MAT so that all elements 
!   are about the same order of magnitude.
!
!   The computations are parallelized if OPENMP is used. Note also that PARTIAL_RTQR_CMP
!   uses a randomized "BLAS3" algorithm described in the reference (7), which has about
!   the same efficiency of a "BLAS3" QR algorithm without column pivoting and is thus
!   highly efficient on large matrices.
!
!   For further details, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Duersch, J.A., and Gu, M., 2017:
!           Randomized QR with column pivoting.
!           SIAM J. Sci. Comput., Volume 39, Issue 4, C263-C291.
!
!   (4) Martinsson, P.G., Quintana-Orti, G., Heavner, N., and Van de Geijn, R., 2017:
!           Householder QR factorization with randomization for column pivoting (HQRRP).
!           SIAM J. Sci. Comput., Volume 39, Issue 2, C96-C115.
!
!   (5) Xiao, J., Gu, M., and Langou, J., 2017:
!           Fast parallel randomized QR with column pivoting algorithms
!           for reliable low-rank matrix approximations.
!           IEEE 24th International Conference
!           on High Performance Computing (HiPC), IEEE, 2017, 233-242.
!
!   (6) Duersch, J.A., and Gu, M., 2020:
!           Randomized projection for rank-revealing matrix factorizations
!           and low-rank approximations.
!           SIAM Review, Volume 62, Issue 3, 661-682.
!
!   (7) Mary, T., Yamazaki, I., Kurzak, J., Luszczek, P., Tomov, S., and Dongarra, J., 2015:
!           Performance of Random Sampling for Computing Low-rank Approximations of a
!           Dense Matrix on GPUs.
!           International Conference for High Performance Computing, Networking, Storage and Analysis (SC'15).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Num_Constants,   only : machhuge, safmin
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert_eq, assert, merror, arth
    use Hous_Procedures, only : hous1, hous2, apply_hous2
    use QR_Procedures,   only : partial_qr_cmp, qr_cmp, ortho_gen_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: diagr, beta
    real(stnd),   intent(inout),                 optional   :: tol
    real(stnd),   intent(out),   dimension(:),   optional   :: tau
!
    integer(i4b), intent(out)               :: krank
    integer(i4b), intent(out), dimension(:) :: ip
    integer(i4b), intent(in), optional :: rng_alg, niter, nover
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: anorm, xnorm, rcond, prod, beta1
    real(stnd), dimension(:,:), allocatable :: b, bt, q, omega
    real(stnd), dimension(:),   allocatable :: diag, alpha
#ifdef _BLAS
    real(stnd), dimension(size(mat,2)) :: tmp
#else
    real(stnd), dimension(max(size(mat,1),size(mat,2))) :: tmp
#endif
!
    integer      :: iok, alg
    integer(i4b) :: m, n, ndiag, niter2, nover2, nbr, in, i, j, l
!
    logical(lgl) :: comp_min_norm, test
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='partial_rtqr_cmp'
!
!   niter0 IS THE DEFAULT NUMBER OF RANDOMIZED SUBSPACE ITERATIONS PERFORMED.
!   nover0 IS THE DEFAULT OVERSAMPLING SIZE.
!
    integer(i4b),  parameter :: niter0=0_i4b, nover0=10_i4b
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b) ,   &
                   int(size(ip),i4b)    ,   &
                   name_proc )
!
    ndiag = min( m, n )
!
    comp_min_norm = present( tau )
!
    if ( comp_min_norm ) then
!
        krank =  assert_eq( int(size(beta),i4b)  ,   &
                            int(size(diagr),i4b) ,   &
                            int(size(tau),i4b)   ,   &
                            name_proc )
!
        tau(:krank) = zero
!
    else
!
        krank =  assert_eq( int(size(beta),i4b)  ,   &
                            int(size(diagr),i4b) ,   &
                            name_proc )
!
    end if
!
    call assert( logical( krank<=ndiag, lgl ), name_proc )
!
!   INITIALIZE PERMUTATION VECTOR FOR COLUMN PIVOTING.
!
    ip(:n) = arth( 1_i4b, 1_i4b, n )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( krank<=0_i4b ) then
!
        if ( present(tol) ) tol = machhuge
!
        return
!
    end if
!
!   SET OPTIONAL ARGUMENTS.
!
!   SELECT AND SET THE RANDOM NUMBER GENERATOR TO BE USED IF REQUIRED.
!
    if ( present(rng_alg) ) then
!
        if ( rng_alg>=1_i4b .and. rng_alg<=10_i4b ) then
!
            alg = rng_alg
!
!           SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
            call random_seed_( alg=alg )
!
!           RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
            call random_seed_()
!
        end if
!
    end if
!
!   DETERMINE THE NUMBER OF SUBSPACE ITERATIONS niter2 TO BE PERFORMED.
!
    if ( present(niter) ) then
        niter2 = max( niter, 0_i4b )
    else
        niter2 = niter0
    end if
!
!   DETERMINE THE OVERSAMPLING SIZE nover2 .
!
    if ( present(nover) ) then
        nover2 = max( nover, 0_i4b )
    else
        nover2 = max( krank/2_i4b, nover0 )
    end if
!    
!   COMPUTE RANDOMIZED PARTIAL QR FACTORIZATION WITH COLUMN PIVOTING OF mat(:m,:n) .
!
!   ALLOCATE WORK ARRAYS, b(:nbr,:n) IS THE COMPRESSION MATRIX,
!   omega(:nbr,:m) IS A GAUSSIAN RANDOM MATRIX.
!
    if ( niter2>0_i4b ) then
!
!       nbr IS THE NUMBER OF ROWS IN MATRICES b and omega .
!
        nbr = min( nover2 + krank, ndiag )
!
        allocate( b(nbr,n), bt(n,nbr), q(m,nbr), omega(nbr,m), &
                  diag(nbr), alpha(nbr), stat=iok )
!
    else 
!
!       nbr IS THE NUMBER OF ROWS IN MATRICES b and omega .
!
        nbr = min( nover2 + krank, m )
!
        allocate( b(nbr,n), omega(nbr,m), stat=iok )
!
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A nbr-BY-m IID GAUSSIAN MATRIX.
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
    call normal_random_number3_( omega(:nbr,:m) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
!    do i = 1_i4b, m
!
!        call normal_random_number3_( omega(:nbr,i) )
!
!    end do
!
!   COMPUTE COMPRESSION MATRIX.
!
#ifdef _BLAS
    call gemm( 'N', 'N', nbr, n, m, one, omega(1_i4b:nbr,1_i4b:m), nbr,  &
               mat(1_i4b:m,1_i4b:n), m, zero, b(1_i4b:nbr,1_i4b:n), nbr  )
#else
    b(:nbr,:n) = matmul( omega(:nbr,:m), mat(:m,:n) )
#endif
!
!   DO SUBSPACE ITERATIONS FOR IMPROVING ACCURACY OF COMPRESSION MATRIX.
!
    do i = 1_i4b, niter2
!
        bt(:n,:nbr) = transpose( b(1_i4b:nbr,1_i4b:n) )
!
!       COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
        call qr_cmp( bt(:n,:nbr), diag(:nbr), alpha(:nbr) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX
!       TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM SUBSPACE.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( bt(:n,:nbr), alpha(:nbr) )
!
!       COMPUTE RANDOM SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nbr, n, one, mat(1_i4b:m,1_i4b:n), m,      &
                   bt(1_i4b:n,1_i4b:nbr), n, zero, q(1_i4b:m,1_i4b:nbr), m )
#else
        q(:m,:nbr) = matmul( mat(:m,:n), bt(:n,:nbr) )
#endif
!
!       COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
        call qr_cmp( q(:m,:nbr), diag(:nbr), alpha(:nbr) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX
!       TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM SUBSPACE.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( q(:m,:nbr), alpha(:nbr) )
!
!       COMPUTE RANDOM SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'T', 'N', nbr, n, m, one, q(1_i4b:m,1_i4b:nbr), m,       &
                   mat(1_i4b:m,1_i4b:n), m, zero, b(1_i4b:nbr,1_i4b:n), nbr )
#else
        omega(:nbr,:m) = transpose( q(:m,:nbr) )
!
        b(:nbr,:n) = matmul( omega(:nbr,:m), mat(:m,:n) )
#endif
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( niter2>0_i4b ) then
        deallocate( bt, q, omega, diag, alpha )
    else 
        deallocate( omega )
    end if
!
!   USE A FAST RANDOMIZED QR FACTORIZATION OF mat WITHOUT UPDATES OF THE COMPRESSION MATRIX AND
!   THE TRAILING MATRIX.
!
!   DO PARTIAL QR WITH COLUMN PIVOTING ON COMPRESSION MATRIX b(:nbr,:n) TO FIND krank PIVOTS.
!           
    call partial_qr_cmp( b(:nbr,:n), diagr(:krank), beta(:krank), ip(:n), i )
!
!   SWAP CORRESPONDING COLUMNS OF mat(:m,:n) .
!
    ip(:n) = -ip(:n)
!
    do i = 1_i4b, n
!
        if ( ip(i)>0_i4b ) cycle
!
        l     = i
        ip(l) = -ip(l)
        in    = ip(l)
!
        do
!
            if ( ip(in)>0_i4b ) exit
!
            if ( l/=in ) then
#ifdef _BLAS
                call swap( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,in), 1_i4b )
#else
                tmp(1_i4b:m)    = mat(1_i4b:m,l)
                mat(1_i4b:m,l)  = mat(1_i4b:m,in)
                mat(1_i4b:m,in) = tmp(1_i4b:m)
#endif
!
            end if
!
            ip(in) = -ip(in)
            l      = in
            in     = ip(in)
!
        end do
!
    end do
!
!   COMPUTE H(1), ... , H(krank) AND DO QR DECOMPOSITION 
!   OF THE FIRST BLOCK OF mat .
!
    do i = 1_i4b, krank
!    
!       RESTORE UPPER TRIANGULAR FACTOR IN THE QR FACTORIZATION OF COMPRESSION
!       MATRIX IN b(:krank,:krank) FOR LATER USE.
!
        b(i,i) = diagr(i)                
!    
!       GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+1:m,i) .
!    
        call hous1( mat(i:m,i), beta(i), diagr(i) )
!    
!       APPLY H(i) TO mat(i:m,i+1:krank) FROM THE LEFT.
!
        beta1 = beta(i)
!
        do l = i + 1_i4b, krank
!
#ifdef _BLAS
            prod = beta1*dot( m-i+1_i4b, mat(i:m,i), 1_i4b, mat(i:m,l), 1_i4b )
            call axpy( m-i+1_i4b, prod, mat(i:m,i), 1_i4b, mat(i:m,l), 1_i4b )
#else
            prod = beta1*dot_product( mat(i:m,i), mat(i:m,l) )
            mat(i:m,l) = mat(i:m,l) + prod*mat(i:m,i)
#endif
!    
        end do
!    
!       START UPDATE OF THE SUBMATRIX R12 STORED IN mat(:krank,krank+1:n) BY FORMING
!    
!            R(:krank,:krank)*inv( b(:krank,:krank) )
!    
!       WHERE R(:krank,:krank) AND b(:krank,:krank) ARE, RESPECTIVELY, THE UPPER TRIANGULAR FACTORS
!       (e.g., R11) IN THE QR FACTORIZATIONS OF FIRST BLOCK OF mat AND COMPRESSION MATRICES, STORING
!       THE RESULT IN b(:krank,:krank) .
!    
        tmp(:i-1_i4b) = mat(:i-1_i4b,i)
        tmp(i)        = diagr(i)
!    
        do l = 1_i4b, i-1_i4b
!
            tmp(1_i4b:l) = tmp(1_i4b:l) - b(l,i)*b(1_i4b:l,l)
!
        end do
!
        beta1 = sign( max( abs(b(i,i)), safmin ), b(i,i) )
!
        prod = one/beta1
!
        b(1_i4b:i,i) = prod*tmp(1_i4b:i)
!
        b(i+1:krank,i) = zero
!
    end do
!    
!   FINISH BLOCK UPDATE OF THE SUB MATRIX R12 STORED IN mat(:krank,krank+1:n).
!
    i = krank + 1_i4b
    l = n - krank
!
#ifdef _BLAS
    call gemm( 'N', 'N', krank, l, krank, one, b(1_i4b:nbr,1_i4b:krank), nbr,  &
               b(1_i4b:nbr,i:n), nbr, zero, mat(1_i4b:m,i:n), m  )
#else
    mat(:krank,i:n) = matmul( b(:krank,:krank), b(:krank,i:n) )
#endif
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( b )
!    
!   TEST TOLERANCE.
!    
    if ( present(tol) ) then
        test = zero<=tol .and. tol<one
    else
        test = false
    end if
!    
    if ( test ) then
!    
        tmp(:krank) = abs( diagr(:krank) )
!    
        if ( tol/=zero ) then
!    
!           DETERMINE THE PSEUDO-RANK USING THE TOLERANCE, tol, AND
!           THE CONDITION NUMBER IN THE 1-NORM OF THE TRIANGULAR FACTOR R11
!           OF THE QR FACTORIZATION.
!    
            anorm = zero
            xnorm = zero
            rcond = zero
!
            do j = 1_i4b, krank
!
!               CHECK IF abs( R(j,j) ) IS SMALL AND IF R(:j,:j) IS SINGULAR.
!               IF R(:j,:j) IS SINGULAR, EXIT AND THE PSEUDO-RANK IS j-1 .
!
                if ( tmp(j)<safmin ) exit
!
!               COMPUTE 1-NORM OF R(:j,:j).
!
                anorm  = max( anorm, sum(abs(mat(1_i4b:j-1_i4b,j)))+tmp(j) )
!
!               COMPUTE 1-NORM OF inv( R(:j,:j) ).
!
                tmp(j)             = one
                tmp(1_i4b:j-1_i4b) = zero
!
#ifdef _BLAS
                mat(j,j) = diagr(j)
!
                call trsv( 'U', 'N', 'N', j, mat, m, tmp(1_i4b:j), 1_i4b )    
#else
                do i = j, 1_i4b, -1_i4b
                    tmp(i)             = tmp(i)/diagr(i)
                    tmp(1_i4b:i-1_i4b) = tmp(1_i4b:i-1_i4b) - tmp(i)*mat(1_i4b:i-1_i4b,i)
                end do
#endif
!
                xnorm = max( xnorm, sum(abs(tmp(1_i4b:j))) )
!
!               COMPUTE INVERSE OF CONDITION NUMBER OF R(:j,:j).
!
                prod = one/( anorm*xnorm )
!
!               EXIT IF INVERSE OF CONDITION NUMBER IS LESS THAN THE TOLERANCE.
!
                if ( tol>=prod ) exit
!
                rcond = prod
!
            end do
!
            tol = rcond
!
#ifdef _BLAS
!
!           RESTORE FIRST ELEMENT OF HOUSEHOLDER VECTORS IF BLAS IS USED.
!
            do i = 1_i4b, min(j,krank)
                mat(i,i) = one
            end do
#endif
!
        else
!    
!           DETERMINE THE NUMERICAL PSEUDO-RANK.
!    
            do j = 1_i4b, krank
                if ( tmp(j)<safmin ) exit
            end do
!
        end if
!    
        krank = j - 1_i4b
!    
    end if
!    
!   IF THE PSEUDORANK krank IS LESS THAN n AND THE MINIMUM_NORM SOLUTION IS REQUIRED,
!   COMPUTE HOUSEHOLDER DECOMPOSITION OF mat(1:krank,:n) .
!    
    if ( krank/=n .and. comp_min_norm ) then
!    
        j = krank + 1_i4b
        l = n - krank
!
!       ALLOCATE WORK VARIABLES FOR STORING HOUSEHOLDER VECTORS.
!
        allocate( b(l,krank), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        b(:l,:krank) = transpose( mat(:krank,j:n) )
!    
        do i = krank, 1_i4b, -1_i4b
!    
!           COMPUTE THE I-TH HOUSEHOLDER TRANSFORMATION.
!    
            call hous2( diagr(i), b(:l,i), tau(i) )
!    
!           APPLY THE I-TH HOUSEHOLDER TRANSFORMATION FROM THE RIGHT.
!    
            call apply_hous2( b(:l,i), tau(i),                                &
                              mat(1_i4b:i-1_i4b,i), b(:l,1_i4b:i-1_i4b), true )
!    
        end do
!
        mat(:krank,j:n) = transpose( b(:l,:krank) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( b )
!    
    end if
!
!
! END OF SUBROUTINE partial_rtqr_cmp
! __________________________________
!
    end subroutine partial_rtqr_cmp
!
! =========================================================================================
!
    subroutine partial_rqr_cmp_fixed_precision( mat, relerr, diagr, beta, ip, krank, tau, &
                                                rng_alg, blk_size, nover )
!
! Purpose
! _______
!
!   PARTIAL_RQR_CMP_FIXED_PRECISION computes a randomized partial QR factorization with column
!   pivoting or orthogonal factorization of a m-by-n matrix MAT:
!
!                     MAT * P ≈ Q * R
!
!   , where P is a n-by-n permutation matrix, R is a krank-by-n (upper trapezoidal) matrix
!   and Q is a m-by-krank matrix with orthonormal columns. This leads to the following matrix
!   approximation of MAT of rank krank:
!
!                     MAT ≈ Q * (R * P')
!
!   krank is the target rank of the matrix approximation, which is sought, 
!   and this partial factorization  must have an approximation error which fulfills:
!
!            || MAT - Q * ( R * P' ) ||_F <= ||MAT||_F * relerr
!
!   || ||_F is the Frobenius norm and relerr is a prescribed accuracy tolerance for the
!   relative error of the computed matrix approximation, specified in the input argument RELERR.
!
!   PARTIAL_RQR_CMP_FIXED_PRECISION searches incrementally the best (e.g., of smallest rank) Q * (R * P')
!   approximation, which fulfills the prescribed accuracy tolerance for the relative error. More precisely,
!   the rank of the matrix approximation is increased progressively of BLK_SIZE by BLK_SIZE until the prescribed
!   accuracy tolerance is satisfied.
!
!   In other words, the rank, krank, of the matrix approximation is not known in advance and is
!   determined in the subroutine. krank is stored in the argument KRANK and the relative error of
!   the computed matrix approximation is output in argument RELERR on exit.
!
!   The computed matrix approximation leads implicitly to the following partition of R:
!
!   [  R1  R2  ]
!
!   where R1 is a krank-by-krank triangular matrix and R2 is a full krank-by-(n-krank)
!   matrix.
!  
!   In a final step, if TAU is present, R2 is annihilated by orthogonal transformations
!   from the right, arriving at the partial orthogonal factorization:
!
!                       MAT * P ≈ Q * T1 * Z
!
!   , where P is a n-by-n permutation matrix, Q is a m-by-krank matrix with orthonormal columns,
!   Z is a krank-by-n matrix with orthonormal rows and T1 is a krank-by-krank upper triangular matrix.
!   
!   Note, however, that this final step does not change the matrix approximation and its relative
!   error, only the output format of this matrix approximation, which is now composed of four factors instead
!   of three.
!   
!   On exit, P is stored compactly in the vector argument IP, krank is stored in the scalar
!   argument KRANK and if:
!   
!   - TAU is absent, PARTIAL_RQR_CMP_FIXED_PRECISION computes Q and submatrices R1 and R2. Submatrices
!     R1 and R2 are stored in the array arguments MAT and DIAGR(:KRANK).
!     Q is stored compactly in factored form in the array arguments MAT and BETA(:KRANK).
!   
!   - TAU is present, PARTIAL_RQR_CMP_FIXED_PRECISION computes Q, Z and submatrice T1. Submatrice T1
!     is stored in the array arguments MAT and DIAGR.
!     Q is stored compactly in factored form in the array arguments MAT and BETA(:KRANK).
!     Z is stored compactly in factored form in the array arguments MAT and TAU(:KRANK).
!
!   In all cases, the relative error of the computed matrix approximation is output in argument RELERR.
!   
!   See Further Details for more information.
!   
!   PARTIAL_RQR_CMP_FIXED_PRECISION performs the same task as subroutine PARTIAL_QR_CMP_FIXED_PRECISION
!   in module QR_Procedures, but is much faster on large matrices because of the use of a randomized and
!   blocked "BLAS3" algorithm instead of a standard "BLAS2" algorithm. Another difference is that, in
!   PARTIAL_RQR_CMP_FIXED_PRECISION, the rank of the matrix approximation is increased progressively
!   of BLK_SIZE by BLK_SIZE until the prescribed tolerance for the relative error is satisfied while
!   in PARTIAL_QR_CMP_FIXED_PRECISION subroutine, the rank of the matrix approximation is increased
!   one by one until the prescribed tolerance for the relative error is satisfied. In other words,
!   the rank of the matrix approximation found by PARTIAL_RQR_CMP_FIXED_PRECISION is always larger
!   than the one found by PARTIAL_QR_CMP_FIXED_PRECISION and is a multiple of BLK_SIZE.
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the real m-by-n matrix to be decomposed.
!
!            On exit, MAT has been overwritten by details of its partial
!            QR or orthogonal factorization.
!
!            See Further Details.
!
!   RELERR   (INPUT/OUTPUT) real(stnd)
!            On entry, the requested accuracy tolerance for the relative error of
!            the computed partial matrix approximation.
!
!            The preset value for RELERR must be greater than 4*epsilon( RELERR ) and less than one.
!
!            On exit, RELERR contains the relative error of the computed partial matrix
!            approximation:
!
!            - RELERR = || MAT - Q * ( R * P' ) ||_F / ||MAT||_F
!
!   DIAGR    (OUTPUT) real(stnd), dimension(:) 
!            On exit, the diagonal elements of the matrix R1 or T1 are
!            stored in the array section DIAGR(:KRANK). Other elements
!            of DIAGR are set to zero on exit.
!
!            See Further Details.
!
!            The size of DIAGR must verify:
!
!            - size( DIAGR ) = min( size(MAT,1) , size(MAT,2) ).
!
!   BETA     (OUTPUT) real(stnd), dimension(:)
!            On exit, the scalars factors of the elementary reflectors defining Q
!            are stored in the array section BETA(:KRANK). Other elements
!            of BETA are set to zero on exit.
!
!            See Further Details.
!
!            The size of BETA must verify:
!
!            - size( BETA ) = min( size(MAT,1) , size(MAT,2) ).
!
!   IP       (OUTPUT) integer(i4b), dimension(:) 
!            On exit, if IP(j)=k, then the j-th column of MAT*P was 
!            the k-th column of MAT.
!
!            See Further Details.
!
!            The size of IP must be equal to size( MAT, 2 ) = n.
!
!   KRANK    (OUTPUT) integer(i4b)
!            On exit, KRANK contains the rank of R1, i.e., krank, which is
!            the order of this submatrix R1. This is the same as the order of
!            the submatrix T1 in the "partial" complete orthogonal factorization of MAT
!            and is also the rank of the computed matrix approximation.
!
!            In all cases, norm(MAT(KRANK+1:m,KRANK+1:n)) gives the error of the associated
!            matrix approximation in the Frobenius norm, on exit.
!
!   TAU      (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if TAU is present, a partial complete orthogonal factorization of
!            MAT is computed. Otherwise, a simple QR factorization with column pivoting
!            of MAT is computed.
!
!            On exit, the scalars factors of the elementary reflectors defining the
!            orthogonal matrix Z in the partial complete orthogonal factorization of MAT
!            are stored in the array section TAU(:KRANK). Other elements
!            of TAU are set to zero on exit.
!
!            See Further Details.
!
!            The size of TAU must verify:
!
!            - size( TAU ) = min( size(MAT,1) , size(MAT,2) ).
!
!   RNG_ALG  (INPUT, OPTIONAL) integer(i4b)
!            On entry, a scalar integer to select the random (uniform) number generator used to build
!            the random gaussian matrix in the randomized partial QR algorithm.
!
!            The possible values are:
!
!            - ALG=1  : selects the Marsaglia's KISS random number generator;
!            - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!            - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!            - ALG=4  : selects the Mersenne Twister random number generator;
!            - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!            - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!            - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!            - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!            - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!            - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!              random number generator;
!
!            For other values, the current random number generator and its current state are not changed.
!            Note further, that, on exit, the current random number generator is not reset to its previous
!            value before the call to PARTIAL_RQR_CMP_FIXED_PRECISION.
!
!            See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   BLK_SIZE (INPUT, OPTIONAL) integer(i4b)
!            On entry, the block size used in the randomized partial QR factorization.
!
!            BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!            smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!            By default, BLK_SIZE is set to min( BLKSZ_QR, min(m,n) ), where parameter BLKSZ_QR is the default
!            block size for QR related algorithms specified in module Select_Parameters.
!
!   NOVER    (INPUT, OPTIONAL) integer(i4b)
!            The oversampling size used in the randomized partial QR algorithm.
!
!            NOVER must be positive or null and verifies the relationship:
!
!                 NOVER + BLK_SIZE <= size(MAT,1)
!
!            and is adjusted if necessary to verify this relationship in all cases.
!
!            See Further Details and the cited references for the meaning and usefulness
!            of the oversampling size in the partial randomized QR algorithm.
!
!            By default, the oversampling size is set to 10.
!
!
! Further Details
! _______________
!
!   The matrix Q is represented as a product of elementary reflectors
!
!            Q = H(1) * H(2) * ... * H(krank), where krank <= min( m , n ).
!
!   Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!   where beta is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in MAT(i:m,i) and beta in BETA(i).
!
!   On exit of PARTIAL_RQR_CMP_FIXED_PRECISION, the matrix Q can be computed
!   explicitly by a call to subroutine ORTHO_GEN_QR with arguments MAT and BETA(:KRANK).
!
!   The matrix P is represented in the array IP as follows:
!   If IP(j) = i then the jth column of P is the ith canonical unit vector.
!
!   On exit, if the optional argument TAU is absent:
!
!   - The elements above the diagonal of the array MAT(:krank,:krank) contain the 
!     corresponding elements of the triangular matrix R1.
!   - The elements of the diagonal of R1 are stored in the array DIAGR.
!   - The submatrix R2 is stored in MAT(:krank,krank+1:n).
!   - krank is stored in the real argument KRANK.
!
!   If TAU is present, a "partial" complete orthogonal factorization of MAT is computed.
!   The factorization is obtained by Householder's method. The kth transformation
!   matrix, Z(k), which is used to introduce zeros into the kth row of R (e.g., in
!   R2), is given in the form
!
!   [  I  0    ]
!
!   [  0  T(k) ]
!
!   where
!
!       T(k) = I + tau * ( u(k) * u(k)' ) and u(k)' = (  1  0  z(k)  )
!
!   tau is a scalar, u(k) is a n-k+1 vector and z(k) is an (n-krank) element vector.
!   tau and z(k) are chosen to annihilate the elements of the kth row of R2.
!
!   The Z n-by-n orthogonal matrix is given by
!
!        Z = Z(1) * Z(2) * ... * Z(krank)
!
!   On exit, if the optional argument TAU is present:
!
!   - The scalar tau defining T(k) is returned in the kth element of TAU and the vector u(k)
!     in the kth row of MAT, such that the elements of z(k) are in MAT(k,krank+1:n).
!     The other elements of u(k) are not stored.
!   - The elements above the diagonal of the array section MAT(:krank,:krank)
!     contain the corresponding elements of the triangular matrix T1. The elements of
!     the diagonal of T1 are stored in the array DIAGR.
!   - krank is stored in the real argument KRANK.
!
!   In both cases, norm(MAT(KRANK+1:m,KRANK+1:n)) gives the error of the associated partial
!   matrix approximation in the Frobenius norm, and argument RELERR stores the relative error
!   in the Frobenius norm of the matrix approximation on exit.
!
!   The computations are parallelized if OPENMP is used. Note also that PARTIAL_RQR_CMP_FIXED_PRECISION
!   uses a randomized "BLAS3" algorithm described in the references (3), (4), (5) and (6), which
!   has about the same efficiency of a "BLAS3" QR algorithm without column pivoting and is thus
!   highly efficient on large matrices.
!
!   For further details, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Duersch, J.A., and Gu, M., 2017:
!           Randomized QR with column pivoting.
!           SIAM J. Sci. Comput., Volume 39, Issue 4, C263-C291.
!
!   (4) Martinsson, P.G., Quintana-Orti, G., Heavner, N., and Van de Geijn, R., 2017:
!           Householder QR factorization with randomization for column pivoting (HQRRP).
!           SIAM J. Sci. Comput., Volume 39, Issue 2, C96-C115.
!
!   (5) Xiao, J., Gu, M., and Langou, J., 2017:
!           Fast parallel randomized QR with column pivoting algorithms
!           for reliable low-rank matrix approximations.
!           IEEE 24th International Conference
!           on High Performance Computing (HiPC), IEEE, 2017, 233-242.
!
!   (6) Duersch, J.A., and Gu, M., 2020:
!           Randomized projection for rank-revealing matrix factorizations
!           and low-rank approximations.
!           SIAM Review, Volume 62, Issue 3, 661-682.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Num_Constants,   only : safmin
    use Char_Constants,  only : allocate_error, random_error12
    use Utilities,       only : assert_eq, assert, merror, arth, lassq
    use Hous_Procedures, only : hous1, hous2, apply_hous2
    use QR_Procedures,   only : partial_qr_cmp
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)                 :: relerr
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: diagr, beta
    real(stnd),   intent(out),   dimension(:), optional :: tau
!
    integer(i4b), intent(out)               :: krank
    integer(i4b), intent(out), dimension(:) :: ip
    integer(i4b), intent(in), optional :: rng_alg, blk_size, nover
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: prod, beta1, threshold, norm_mat, scal, ssq, scal0, ssq0
    real(stnd), dimension(:),   allocatable :: diagr_b, beta_b
    real(stnd), dimension(:,:), allocatable :: b, lmat, umat, mat2, omega
#ifdef _BLAS
    real(stnd), dimension(:,:), allocatable :: x
#else
    real(stnd), dimension(size(mat,1))    :: tmp
    real(stnd), dimension(:), allocatable :: x
#endif
!
    integer      :: iok, alg
    integer(i4b) :: m, n, ndiag, p, nover2, blk_size2, nbr, in, nblk, &
                    i, i1, i2, i3, itmp, j, j2, l, l1, l2
    integer(i4b), dimension(size(ip)) :: ip_b
!
    logical(lgl) :: comp_min_norm, test
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='partial_rqr_cmp_fixed_precision'
!
!   nover0 IS THE DEFAULT OVERSAMPLING SIZE.
!
    integer(i4b),  parameter :: nover0=10_i4b
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b) ,   &
                   int(size(ip),i4b)    ,   &
                   name_proc )
!
    ndiag = min( m, n )
!
    comp_min_norm = present( tau )
!
    if ( comp_min_norm ) then
!
        i = assert_eq( int(size(beta),i4b)  ,   &
                       int(size(diagr),i4b) ,   &
                       int(size(tau),i4b)   ,   &
                       name_proc )
!
        tau(:i) = zero
!
    else
!
        i = assert_eq( int(size(beta),i4b)  ,   &
                       int(size(diagr),i4b) ,   &
                       name_proc )
!
    end if
!
    call assert( logical( i==ndiag, lgl ), name_proc )
!
!   INITIALIZE PERMUTATION VECTOR FOR COLUMN PIVOTING.
!
    ip(:n) = arth( 1_i4b, 1_i4b, n )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( ndiag<=0_i4b ) then
!
        krank = 0_i4b
!
        relerr = one
!
        return
!
    end if
!
    diagr(:ndiag) = zero
    beta(:ndiag)  = zero
!
!   CHECK VALIDITY OF THE REQUESTED TOLERANCE FOR THE RELATIVE
!   ERROR OF THE COMPUTED MATRIX APPROXIMATION.
!
    prod = four*machulp
!
    if ( relerr<=prod .or. relerr>=one ) then
!
        call merror( name_proc//random_error12 )
!
    end if
!    
    threshold = relerr
    relerr    = one
!
!   SELECT AND SET THE RANDOM NUMBER GENERATOR TO BE USED IF REQUIRED.
!
    if ( present(rng_alg) ) then
!
        if ( rng_alg>=1_i4b .and. rng_alg<=10_i4b ) then
!
            alg = rng_alg
!
!           SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
            call random_seed_( alg=alg )
!
!           RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
            call random_seed_()
!
        end if
!
    end if
!
!   DETERMINE THE OVERSAMPLING SIZE nover2 .
!
    if ( present(nover) ) then
        nover2 = max( nover, 0_i4b )
    else
        nover2 = nover0
    end if
!
!   DETERMINE THE BLOCKSIZE blk_size2 .
!
    if ( present(blk_size) ) then
        blk_size2 = max( blk_size, 1_i4b )
    else
        blk_size2 = blksz
    end if
!    
    blk_size2 = min( blk_size2, ndiag )
!    
!   COMPUTE RANDOMIZED PARTIAL QR FACTORIZATION WITH COLUMN PIVOTING OF mat(:m,:n) .
!    
!   COMPUTE FROBENIUS NORM OF mat .
!
    norm_mat = norm( mat(:m,:n) )
!
!   nbr IS THE NUMBER OF ROWS IN MATRICES b and omega .
!
    nbr = min( nover2 + blk_size2, m )
!
!   ALLOCATE WORK ARRAYS, b(:nbr,:n) IS THE COMPRESSION MATRIX
!   AND omega(:nbr,:m) IS A GAUSSIAN RANDOM MATRIX.
!
    allocate( b(nbr,n), omega(nbr,m), diagr_b(blk_size2), beta_b(blk_size2),   &
              umat(m,blk_size2), lmat(blk_size2,blk_size2), stat=iok )
!
    if ( iok/=0 ) then
         call merror( name_proc//allocate_error )
    end if
!
!   INITIALIZE THE UNIT LOWER TRIANGULAR MATRIX lmat(:blk_size2,:blk_size2)
!   USED IN THE BLOCKED QR ALGORITHM.
!
    lmat(1_i4b:blk_size2,1_i4b:blk_size2) = zero
!
    do i = 1_i4b, blk_size2
        lmat(i,i) = one
    end do
!
!   INITIALIZE ARRAY FOR STORING HOUSEHOLDER VECTORS USED IN THE BLOCKED QR
!   ALGORITHM.
!
    umat(:m,:blk_size2) = zero
!
!   GENERATE A nbr-BY-m IID GAUSSIAN MATRIX.
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
    call normal_random_number3_( omega(:nbr,:m) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
!    do i = 1_i4b, m
!
!        call normal_random_number3_( omega(:nbr,i) )
!
!    end do
!
!   COMPUTE COMPRESSION MATRIX.
!
#ifdef _BLAS
    call gemm( 'N', 'N', nbr, n, m, one, omega(1_i4b:nbr,1_i4b:m), nbr,  &
               mat(1_i4b:m,1_i4b:n), m, zero, b(1_i4b:nbr,1_i4b:n), nbr  )
#else
    b(:nbr,:n) = matmul( omega(:nbr,:m), mat(:m,:n) )
#endif
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( omega )
!
#ifdef _OPENMP
!
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  n*m>=omp_limit                      .and.       &
                  n/blk_size2>1_i4b                   .and.       &
                  i>1_i4b
!
#endif
!
#ifdef _BLAS
!$OMP PARALLEL IF(test_par)                                        &
!$OMP         ,PRIVATE(test,iok,i,i1,i2,i3,j,j2,l,l1,l2,in,itmp,   &
!$OMP                  nblk,p,beta1,prod,scal,ssq,x,mat2)
#else
!$OMP PARALLEL IF(test_par)                                        &
!$OMP         ,PRIVATE(test,iok,i,i1,i2,i3,j,j2,l,l1,l2,in,itmp,   &
!$OMP                  nblk,p,beta1,prod,scal,ssq,x,mat2,tmp)
#endif
!
!   ALLOCATE WORK VARIABLES.
!
#ifdef _BLAS
    allocate( x(blk_size2,blk_size2), mat2(m,blk_size2), stat = iok )
#else
    allocate( x(blk_size2), mat2(m,blk_size2), stat = iok )
#endif
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   RANDOMIZED BLOCKED PARTIAL QR DECOMPOSITION OF mat WITH COLUMN PIVOTING.
!
    do j = 1_i4b, ndiag, blk_size2
!
        nblk = min( ndiag-j+1_i4b, blk_size2 )
!
        j2 = j + nblk - 1_i4b
        p  = m - j + 1_i4b
!
        test = nbr < m - j2 .and. j2 < n
!
!$OMP MASTER
!
!       DO PARTIAL QR WITH COLUMN PIVOTING ON COMPRESSION MATRIX b(:nbr,j:n) TO FIND nblk NEW PIVOTS.
!           
        call partial_qr_cmp( b(:nbr,j:n), diagr_b(:nblk), beta_b(:nblk), ip_b(j:n), i )
!
!       SWAP CORRESPONDING COLUMNS OF mat(:m,j:n) .
!
        ip_b(j:n) = -ip_b(j:n)
!
        do i = j, n
!
            if ( ip_b(i)>0_i4b ) cycle
!
            l       = i
            ip_b(l) = -ip_b(l)
            in      = ip_b(l) + j - 1_i4b
!
            do
!
                if ( ip_b(in)>0_i4b ) exit
!
                if ( l/=in ) then
#ifdef _BLAS
                    call swap( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,in), 1_i4b )
#else
                    tmp(1_i4b:m)    = mat(1_i4b:m,l)
                    mat(1_i4b:m,l)  = mat(1_i4b:m,in)
                    mat(1_i4b:m,in) = tmp(1_i4b:m)
#endif
!
                    itmp   = ip(l)
                    ip(l)  = ip(in)
                    ip(in) = itmp
!
                end if
!
                ip_b(in) = -ip_b(in)
                l        = in
                in       = ip_b(in) + j - 1_i4b
!
            end do
!
        end do
!
!       COMPUTE H(j), ... , H(j+nblk-1_i4b) AND DO QR DECOMPOSITION 
!       OF THE CURRENT BLOCK OF mat .
!
        do i = j, j2
!    
!           GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+1:m,i) .
!    
            call hous1( mat(i:m,i), beta(i), diagr(i) )
!    
!           APPLY H(i) TO mat(i:m,i+1:j+nblk-1) FROM THE LEFT.
!
            beta1 = beta(i)
!
            do l = i + 1_i4b, j2
!
#ifdef _BLAS
                prod = beta1*dot( m-i+1_i4b, mat(i:m,i), 1_i4b, mat(i:m,l), 1_i4b )
                call axpy( m-i+1_i4b, prod, mat(i:m,i), 1_i4b, mat(i:m,l), 1_i4b )
#else
                prod = beta1*dot_product( mat(i:m,i), mat(i:m,l) )
                mat(i:m,l) = mat(i:m,l) + prod*mat(i:m,i)
#endif
!    
            end do
!
            if ( test ) then
!    
!               RESTORE UPPER TRIANGULAR FACTOR IN QR FACTORIZATION OF CURRENT BLOCK
!               OF COMPRESSION MATRIX IN b(:nblk,j:j2) FOR LATER USE.
!
                i1 = i - j + 1_i4b
!
                b(i1,i) = diagr_b(i1)                
!
                b(i1+1:nblk,i) = zero
!    
!               START UPDATE OF THE COMPRESSION MATRIX FOR LATER USE BY FORMING
!    
!                   b(:nblk,j:j2)*inv( R(:nblk,:nblk) )
!    
!               WHERE R(:nblk,:nblk) IS THE UPPER TRIANGULAR FACTOR IN QR FACTORIZATION
!               OF CURRENT BLOCK OF mat, STORING THE RESULT IN b(:nblk,j:j2) .
!    
                do l = j, i-1_i4b
!    
                    l1 = l - j + 1_i4b
!
                    b(1_i4b:l1,i) = b(1_i4b:l1,i) - mat(l,i)*b(1_i4b:l1,l)
!
                end do
!
                beta1 = sign( max( abs(diagr(i)), safmin ), diagr(i) )
!
                prod = one/beta1
!
                b(1_i4b:i1,i) = prod*b(1_i4b:i1,i)
!
            end if
!
        end do
!
!       INITIALIZE GLOBAL VARIABLES FOR COMPUTING THE FROBENIUS NORM OF THE RESIDUAL MATRIX.
!
        scal0 = one
        ssq0  = zero
!
!$OMP END MASTER
!    
!$OMP BARRIER
!    
        if ( j2<n ) then
!    
!           APPLY H(j+nblk-1_i4b)* ...*H(j) TO mat(j:m,j+nblk:n) FROM THE LEFT
!           WITH A BLOCKED ALGORITHM AS
!
!        ( I + U(j:m,:nblk)*inv( lmat(:nblk,:nblk) )*U(j:m,:nblk)' )*mat(j:m,j+nblk:n) .
!
!           FIRST SAVE H(j) ...H(j+nblk-1_i4b) FOR LATER USE AND COMPUTE THE nblk BY nblk UNIT
!           LOWER TRIANGULAR MATRIX lmat(:nblk,:nblk) USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do l = 1_i4b, nblk
!
                l1 = j + l - 1_i4b
!
!               SAVE SHIFTED HOUSEHOLDER VECTOR mat(j+l-1:m,j+l-1) .
!
                umat(l:p,l) = mat(l1:m,l1)
!
                do i = l + 1_i4b, nblk
!
                    l2 = j + i - 1_i4b
!
#ifdef _BLAS
                    lmat(i,l) = -beta(l2)*dot( p-i+1_i4b, umat(i:p,l), 1_i4b, mat(l2:m,l2), 1_i4b )
#else
                    lmat(i,l) = -beta(l2)*dot_product( umat(i:p,l), mat(l2:m,l2) )
#endif
!
                end do
!                    
            end do
!
!$OMP END DO
!
!           INITIALIZE LOCAL VARIABLES FOR COMPUTING THE FROBENIUS NORM OF THE RESIDUAL MATRIX.
!
            scal = one
            ssq  = zero
!
!           BLOCK UPDATE OF THE TRAILING MATRIX mat(j:m,j+nblk:n) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do i = j2, n-1_i4b, nblk
!
                i1 = i + 1_i4b
                i2 = min( i+nblk, n )
                i3 = i2 - i
!
!               MAKE A LOCAL COPY OF mat(j:m,i1:i2) .
!
                mat2(:p,:i3) = mat(j:m,i1:i2)
!
#ifdef _BLAS
!
!               COMPUTE U(j:m,:nblk)'*mat(j:m,i1:i2) .
!
                call gemm( 'T', 'N', nblk, i3, p, one, umat, m, mat2, m, zero, x, blk_size2 )
!
                do l = 1_i4b, i3
!
                    x(:nblk,l) = beta(j:j2)*x(:nblk,l)
!                    
                end do
!
!               SOLVE lmat(:nblk,:nblk)*x(:nblk,:i3) = U(j:m,:nblk)'*mat(j:m,i1:i2) .
!
                call trsm( 'L', 'L', 'N', 'U', nblk, i3, one, lmat, blk_size2, x, blk_size2 )
!
!               COMPUTE mat(j:m,i1:i2) = mat(j:m,i1:i2) + U(j:m,:nblk)*x(:nblk,:i3) .
!
                call gemm( 'N', 'N', p, i3, nblk, one, umat, m, x, blk_size2, one, mat2, m )
#else
                do l = 1_i4b, i3
!
!                   COMPUTE U(j:m,:nblk)'*mat(j:m,i1:i2) .
!
                    do l2 = 1_i4b, nblk
!
                       x(l2) = beta(j+l2-1_i4b)*dot_product( umat(l2:p,l2), mat2(l2:p,l) )
!                    
                    end do
!
!                   SOLVE lmat(:nblk,:nblk)*x(:nblk,:i3) = U(j:m,:nblk)'*mat(j:m,i1:i2) .
!                   COMPUTE mat(j:m,i1:i2) = mat(j:m,i1:i2) + U(j:m,:nblk)*x(:nblk,:i3) .
!
                    do l2 = 1_i4b, nblk
!
                        x(l2+1_i4b:nblk) = x(l2+1_i4b:nblk) - x(l2)*lmat(l2+1_i4b:nblk,l2)
!
                        mat2(l2:p,l) = mat2(l2:p,l) + x(l2)*umat(l2:p,l2)
!
                    end do
!
                end do
#endif
!
!               SAVE UPDATE OF mat(j:m,i1:i2) .
!
                mat(j:m,i1:i2) = mat2(:p,:i3)           
!
!               UPDATE COMPRESSION MATRIX FOR NEXT PANEL QR ITERATION.
!
                if ( test ) then
!
#ifdef _BLAS
                    call gemm( 'N', 'N', nblk, i3, nblk, -one, b(1_i4b:nbr,j:j2), nbr,  &
                               mat2, m, one, b(1_i4b:nbr,i1:i2), nbr  )
#else
                    b(:nblk,i1:i2) = b(:nblk,i1:i2) - matmul( b(:nblk,j:j2), mat2(:nblk,:i3) )
#endif
!                    
                else
!
                    l = m - j2
!                    
                    b(:l,i1:i2)          = mat2(nblk+1_i4b:p,:i3) 
                    b(l+1_i4b:nbr,i1:i2) = zero 
!                    
                end if
!    
!               COMPUTE FROBENIUS NORM OF THE RESIDUAL MATRIX.
!
                call lassq( mat2(nblk+1_i4b:p,:i3), scal, ssq )
!                    
            end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compssq)
!
            prod = max( scal0, scal )
!
            ssq0 = ((scal/prod)**2)*ssq + ((scal0/prod)**2)*ssq0
!
            scal0 = prod
!
!$OMP END CRITICAL (compssq)
!
!$OMP BARRIER
!
        end if
!
!$OMP MASTER
!
        relerr = ( scal0 *sqrt(ssq0) )/norm_mat
!
!$OMP END MASTER
!
!$OMP BARRIER
!
!       CHECK IF THRESHOLD IS SATISFIED.
!
        if ( relerr<=threshold ) exit
!
    end do
!
!$OMP MASTER
!
    krank = min( j2, ndiag )
!
!$OMP END MASTER
!
!$OMP BARRIER
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( x, mat2 )
!
!$OMP END PARALLEL
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( b, diagr_b, beta_b, umat, lmat )
!    
!   IF THE PSEUDORANK krank IS LESS THAN n AND THE MINIMUM_NORM SOLUTION IS REQUIRED,
!   COMPUTE HOUSEHOLDER DECOMPOSITION OF mat(1:krank,:n) .
!
    if ( krank/=n .and. comp_min_norm ) then
!    
        j = krank + 1_i4b
        l = n - krank
!
!       ALLOCATE WORK VARIABLES FOR STORING HOUSEHOLDER VECTORS.
!
        allocate( umat(l,krank), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        umat(:l,:krank) = transpose( mat(:krank,j:n) )
!    
        do i = krank, 1_i4b, -1_i4b
!    
!           COMPUTE THE I-TH HOUSEHOLDER TRANSFORMATION.
!    
            call hous2( diagr(i), umat(:l,i), tau(i) )
!    
!           APPLY THE I-TH HOUSEHOLDER TRANSFORMATION FROM THE RIGHT.
!    
            call apply_hous2( umat(:l,i), tau(i),                                &
                              mat(1_i4b:i-1_i4b,i), umat(:l,1_i4b:i-1_i4b), true )
!    
        end do
!
        mat(:krank,j:n) = transpose( umat(:l,:krank) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( umat )
!    
    end if
!
!
! END OF SUBROUTINE partial_rqr_cmp_fixed_precision
! _________________________________________________
!
    end subroutine partial_rqr_cmp_fixed_precision
!
! =========================================================================================
!
    subroutine rqb_cmp( mat, q, b, niter, rng_alg, ortho, comp_qr, ip, tol, tau )
!
! Purpose
! _______
!
!   RQB_CMP computes a partial QB, QR (eventually with column pivoting) or complete orthogonal
!   factorization of a full m-by-n real matrix MAT using randomized power or subspace iterations.
!
!   nqb is the target rank of the partial QB, QR or complete orthogonal decomposition, which is sought,
!   and is equal to the number of columns of the output real matrix argument Q, i.e., nqb = size( Q, 2 ).
!
!   The routine first computes a partial QB factorization of MAT with the help of a randomized algorithm:
!
!                     MAT ≈ Q * B
!
!   , where Q is a m-by-nqb orthonormal matrix, B is a nqb-by-n matrix and the product Q*B is a good
!   approximation of MAT according to the spectral or Frobenius norm.
!
!   In a second step:
!
!   - if the optional logical argument COMP_QR is used with the value true and the optional array arguments
!     IP and TAU are absent, a QR factorization of B is computed to obtain an approximate QR factorization of MAT:
!
!             MAT ≈ Q * B = Q * ( O * R ) = (Q * O ) * R
!
!     , where O is an nqb-by-nqb orthogonal matrix and R is an nqb-by-n upper trapezoidal matrix.
!
!   - if the optional array argument IP is present, a QR factorization of B with column pivoting is performed
!     to obtain an approximate QR factorization with column pivoting of MAT :
!
!             MAT * P ≈ Q * ( O * R ) = (Q * O ) * R
!
!     , where P is an n-by-n permutation matrix, O is an nqb-by-nqb orthogonal matrix and R is an nqb-by-n upper
!     trapezoidal matrix.
!
!   - if the optional array argument TAU is present, a complete orthogonal factorization of B is performed
!     to obtain an approximate complete orthogonal factorization of MAT:
!
!             MAT ≈ Q * ( O * T * Z ) = (Q * O ) * T * Z
!
!     , where O is an nqb-by-nqb orthogonal matrix, Z is a n-by-n orthogonal matrix and
!     T is a nqb-by-n matrix, which has the form:
!
!     [  T1  0  ]
!
!     , where T1 is a nqb-by-nqb upper triangular matrix.
!
!   - if, finally, both the optional array arguments IP and TAU are present, a complete orthogonal factorization
!     of B with column pivoting is performed to obtain an approximate complete orthogonal factorization with column 
!     pivoting of MAT :
!
!             MAT * P  ≈ Q * ( O * T * Z ) = (Q * O ) * T * Z
!
!     , where P is a n-by-n permutation matrix, O is an nqb-by-nqb orthogonal matrix,
!     Z is a n-by-n orthogonal matrix and T a nqb-by-n matrix, which has the form:
!
!     [  T1  0  ]
!
!     , where T1 is a nqb-by-nqb upper triangular matrix.
!   
!   If a partial QR or complete orthogonal factorization is computed, we have the following partition of R:
!
!                                   R = [  R1  R2  ]
!
!   where R1 is a nqb-by-nqb upper triangular matrix and R2 is a full nqb-by-(n-nqb) matrix. Then, if:
!
!   - the optional scalar argument TOL is present and is in ]0,1[, the rank of R1 is determined by finding
!     the submatrix of R1 which is defined as the largest leading submatrix whose estimated condition number,
!     in the 1-norm, is less than 1/TOL. The order of this submatrix, krank, is the effective rank of R1.
!   - the optional scalar argument TOL is present and is equal to 0, the numerical rank of R1, krank, is determined.
!   - the optional scalar argument TOL is absent or outside [0,1[, the numerical rank of R1, krank, is
!     determined by finding the largest leading submatrix of R1 such that abs(R1[j,j])>0.
!
!   In all cases, if krank < nqb, this indicates that column pivoting must be used or, if column pivoting has been already
!   specified, that the rank of MAT is probably less than nqb and nqb = size( Q, 2 ) has been set to a too large value.
!   In such cases, the subroutine will exit with an error message.
!   
!   
! Arguments
! _________
!
!   MAT       (INPUT) real(stnd), dimension(:,:)
!             On entry, the m-by-n matrix MAT.
!
!             MAT is not modified by the routine.
!
!   Q         (OUTPUT) real(stnd), dimension(:,:)
!             On exit, the computed m-by-nqb orthonormal matrix of the partial QB, QR or complete
!             orthogonal factorization of MAT.
!
!             See Further Details.
!
!             The shape of Q must verify:
!
!             - size( Q, 1 ) = m    = size( MAT, 1 ),
!             - size( Q, 2 ) = nqb <= min( size(MAT,1) , size(MAT,2) ) .
!
!   B         (OUTPUT) real(stnd), dimension(:,:)
!             On exit:
!
!             - the computed B matrix of the partial QB factorization of MAT if the optional arguments
!               COMP_QR, IP and TAU are absent or if COMP_QR is used with the value false;
!             - the upper trapezoidal matrix R of the partial QR factorization if the optional logical
!               argument COMP_QR is used with the value true or if the optional array argument IP is present;
!             - the computed matrices T1 and Z of the partial complete orthogonal
!               factorization of MAT if the optional argument TAU is present.
!
!             See Further Details.
!
!             The shape of B must verify:  
!
!             - size( B, 1 ) = size( Q, 2 )   = nqb ,
!             - size( B, 2 ) = size( MAT, 2 ) = n .
!
!   NITER     (INPUT, OPTIONAL) integer(i4b)
!             The number of randomized power or subspace iterations performed
!             in the first phase of the randomized algorithm for computing the preliminary
!             QB factorization.
!
!             NITER must be positive or null.
!
!             By default, 5 randomized power or subspace iterations are performed.
!
!   RNG_ALG   (INPUT, OPTIONAL) integer(i4b)
!             On entry, a scalar integer to select the random (uniform) number generator used to build
!             the random gaussian test matrix in the randomized partial QB or QR algorithm.
!
!             The possible values are:
!
!             - ALG=1  : selects the Marsaglia's KISS random number generator;
!             - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!             - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!             - ALG=4  : selects the Mersenne Twister random number generator;
!             - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!             - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!             - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!             - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!             - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!             - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!               random number generator;
!
!             For other values, the current random number generator and its current state are not changed.
!             Note further, that, on exit, the current random number generator is not reset to its previous
!             value before the call to RQB_CMP.
!
!             See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   ORTHO     (INPUT, OPTIONAL) logical(lgl)
!             On entry, if:
!
!             - ORTHO=true, orthonormalization is carried out between each step of
!               the power iterations, to avoid loss of accuracy due to rounding errors.
!               This means that subspace iterations are used instead of power iterations
!               in the QB phase of the algorithm,
!             - ORTHO=false, orthonormalization is not performed. 
!
!             The default is to use orthonormalization, e.g., ORTHO=true.
!
!   COMP_QR   (INPUT, OPTIONAL) logical(lgl)
!             The optional logical argument COMP_QR determines if a partial QB or QR
!             factorization is computed.
!
!             On entry, if:
!
!             - COMP_QR=true, a partial QR factorization is computed;
!             - COMP_QR=false, a partial QB factorization is computed. 
!
!             The default is to compute a partial QB factorization, e.g., COMP_QR=false.
!
!   IP        (OUTPUT, OPTIONAL) integer(i4b), dimension(:) 
!             On entry, if IP is present a partial QR factorization with column pivoting of MAT
!             is performed instead of a QB factorization (e.g., this implies COMP_QR=true).
!
!             On exit, if IP(j)=k, then the j-th column of MAT*P was the k-th column of MAT.
!
!             See Further Details.
!
!             The size of IP must verify: size( MAT, 2 ) = n.
!
!   TOL       (INPUT/OUTPUT, OPTIONAL) real(stnd)
!             If COMP_QR=true and TOL is present and is in [0,1[, then:
!
!             - the calculations to determine the condition number of
!               of R1 in the 1-norm are performed. Then, TOL is used to determine
!               the effective rank of R1, which is defined as the order of 
!               the largest leading triangular submatrix of R1, whose estimated
!               condition number is less than 1/TOL.
!             - if TOL=0 is specified the numerical rank of R1 is determined.
!             - on exit, the reciprocal of the condition number is returned in TOL.
!
!             If COMP_QR=true, but TOL is not specified or is outside [0,1[ :
!
!             - the calculations to determine the condition number of R1 are not
!               performed and crude tests on R1(j,j) are done to determine
!               the rank of R1. If TOL is present, it is not changed.
!
!             If each element of MAT is correct to, say, 5 digits then TOL=0.00001 should
!             be used on entry.
!
!   TAU       (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!             On entry, if TAU is present, a complete orthogonal factorization of
!             MAT is computed instead of a QB factorization (e.g., this implies COMP_QR=true).
!
!             On exit, the scalars factors of the elementary reflectors defining Z.
!
!             See Further Details.
!
!             The size of TAU must verify: size( TAU ) = nqb = size( Q, 2 ).
!
!
! Further Details
! _______________
!
!   For a good introduction to randomized linear algebra, see the references (1) and (2).
!
!   The randomized power or subspace iteration was proposed in (3; see Algorithm 4.4) to compute
!   an orthonormal matrix whose range approximates the range of MAT. An approximate partial QB,
!   QR or complete orthogonal factorization can then be computed using the aforementioned
!   orthonormal matrix, see the references (1) and (3) for details.
!
!   The orthonormal m-by-nqb matrix Q of the partial QB, QR or complete orthogonal factorization of MAT
!   is computed explicitly and stored on exit in the real array argument Q.
!
!   The nqb-by-n matrix B of the partial QB factorization or the upper trapezoidal matrix R of the partial
!   QR or orthogonal factorization of MAT is stored on exit in the real array argument B.
!
!   If the integer array argument IP is present, a (partial) QR factorization of MAT with column pivoting is
!   computed, as described above, and on exit the permutation matrix P is represented in the array IP as follows:
!   If IP(j) = i then the jth column of P is the ith canonical unit vector.
!
!   Next, if the real array argument TAU is present, a (partial) complete orthogonal factorization of MAT is
!   computed from the partial QR factorization of MAT. The factorization is obtained by Householder's method.
!   The kth transformation matrix, Z(k), which is used to introduce zeros into the kth row of R of the (partial)
!   QR factorization of MAT, is given in the form
!
!   [  I  0    ]
!
!   [  0  T(k) ]
!
!   where
!
!         T(k) = I + tau * ( u(k) * u(k)' ) and u(k)' = (  1  0  z(k)  )
!
!   tau is a scalar, u(k) is a n-k+1 vector and z(k) is an (n-nqb) element vector.
!   tau and z(k) are chosen to annihilate the elements of the kth row of R12.
!
!   The Z n-by-n orthogonal matrix is finally given by the product
!
!         Z = Z(1) * Z(2) * ... * Z(nqb)
!
!   On exit:
!
!   - the orthogonal matrix Z is stored in factored form. The scalar tau,
!     which defines Z(k) is returned in the kth element of TAU and the vector u(k)
!     in the kth row of B, such that the elements of z(k) are in B(k,nqb+1:n).
!   - the upper triangular nqb-by-nqb matrix T1, which defines the T factor of the
!     (partial) complete orthogonal factorization of MAT (see above) is stored in the upper
!     triangle of the real array argument B.
!
!   Finally, if the optional arguments IP and TAU are both present, a (partial) complete orthogonal
!   factorization with column pivoting of MAT is computed and on exit this factorization is stored
!   with the same conventions as described above.
!
!   For further details on randomized linear algebra, computing low-rank matrix approximations
!   using randomized power or subspace iterations, see:
!
!   (1) Martinsson, P.G., 2019:
!            Randomized methods for matrix computations.
!            arXiv.1607.01649
!
!   (2) Erichson, N.B., Voronin, S., Brunton, S.L., and Kutz, J.N., 2019:
!            Randomized matrix decompositions using R.
!            arXiv.1608.02148
!
!   (3) Halko, N., Martinsson, P.G., and Tropp, J.A., 2011:
!            Finding structure with randomness:
!            probabilistic algorithms for constructing approximate matrix decompositions.
!            SIAM Rev., 53, 217-288.
!
!   (4) Gu, M., 2015:
!            Subspace iteration randomization and singular value problems.
!            SIAM J. Sci. Comput., 37, A1139-A1173.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error, random_error13
    use Utilities,       only : assert, assert_eq, merror
    use QR_Procedures,   only : qr_cmp, ortho_gen_qr, apply_q_qr, qrfac
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:,:)           :: mat
    real(stnd),  intent(out), dimension(:,:)           :: q, b
    real(stnd),  intent(out), dimension(:),   optional :: tau
    real(stnd),  intent(inout),               optional :: tol
!
    integer(i4b), intent(in),                optional :: niter, rng_alg
    integer(i4b), intent(out), dimension(:), optional :: ip
!
    logical(lgl), intent(in),  optional :: ortho, comp_qr
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _BLAS
    real(stnd), dimension(:,:), allocatable :: bt
#else
    real(stnd), dimension(:,:), allocatable :: qt, bt
#endif
    real(stnd), dimension(:),   allocatable :: diagr, beta
!
    integer      :: iok, alg
    integer(i4b) :: m, n, mn, nqb, i, niter2, kfix
!
    logical(lgl) :: do_ortho, do_qr, comp_min_norm

!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='rqb_cmp'
!
!   niter0 IS THE DEFAULT NUMBER OF RANDOMIZED POWER OR SUBSPACE ITERATIONS PERFORMED.
!
    integer(i4b),  parameter :: niter0=5_i4b
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
!
    m = assert_eq( int(size( mat, 1 ),i4b)  ,   &
                   int(size( q,   1 ),i4b)  ,   &
                   name_proc )
!
    n = assert_eq( int(size( mat, 2 ),i4b)  ,  &
                   int(size( b,   2 ),i4b)  ,  &
                   name_proc )
!
    nqb = assert_eq( int(size( q, 2 ),i4b) ,  &
                     int(size( b, 1 ),i4b) ,  &
                     name_proc )
!
    mn = min( m, n )
!!
!   QUICK RETURN IF POSSIBLE.
!
    if ( mn<=0_i4b  .or. nqb<=0_i4b ) then
!
        return
!
    end if
!
!   CHECK THE RANK OF THE QB OR QR FACTORIZATION.
!
    call assert( logical( nqb>0_i4b .and. nqb<=mn, lgl ), name_proc )
!
!   SET OPTIONAL ARGUMENTS.
!
!   DETERMINE THE NUMBER OF POWER OR SUBSPACE ITERATIONS
!   niter2 TO BE PERFORMED.
!
    if ( present(niter) ) then
        niter2 = max( niter, 0_i4b )
    else
        niter2 = niter0
    end if
!
!   DETERMINE IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE POWER (SUBSPACE) ITERATIONS, TO AVOID LOSS OF ACCURACY
!   DUE TO ROUNDING ERRORS.
!
    if ( present(ortho) ) then
        do_ortho = ortho
    else
        do_ortho = true
    end if
!
!   DETERMINE IF A QB OR QR FACTORIZATION MUST BE COMPUTED.
!
    if ( present(comp_qr) ) then
        do_qr = comp_qr
    else
        do_qr = false
    end if
!
!   DETERMINE IF A QR FACTORIZATION WITH COLUMN PIVOTING MUST BE COMPUTED.
!
    if ( present(ip) ) then
!
        i = size( ip )
!
        call assert( logical( i==n, lgl ), name_proc )
!
        do_qr = true
!
        kfix = 0_i4b
!
    else
!
        kfix = n
!
    end if
!
!   DETERMINE IF A COMPLETE ORTHOGONAL FACTORIZATION MUST BE COMPUTED.
!
    comp_min_norm = present( tau )
!
    if ( comp_min_norm ) then
!
        i = size( tau )
!
        call assert( logical( i==nqb, lgl ), name_proc )
!
        tau(:nqb) = zero
!
        do_qr = true
!
    end if
!
!   SELECT AND SET THE RANDOM NUMBER GENERATOR TO BE USED IF REQUIRED.
!
    if ( present(rng_alg) ) then
!
        if ( rng_alg>=1_i4b .and. rng_alg<=10_i4b ) then
!
            alg = rng_alg
!
!           SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
            call random_seed_( alg=alg )
!
!           RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
            call random_seed_()
!
        end if
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    allocate( bt(n,nqb), diagr(nqb), beta(nqb), stat=iok )
#else
    allocate( bt(n,nqb), qt(nqb,m), diagr(nqb), beta(nqb), stat=iok )
#endif
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-BY-nqb IID GAUSSIAN MATRIX.
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
    call normal_random_number3_( bt(:n,:nqb) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
!    do i = 1_i4b, nqb
!
!        call normal_random_number3_( bt(:n,i) )
!
!    end do
!
!   COMPUTE RANDOM SAMPLE MATRIX.
!
#ifdef _BLAS
    call gemm( 'N', 'N', m, nqb, n, one, mat(1_i4b:m,1_i4b:n), m,       &
               bt(1_i4b:n,1_i4b:nqb), n, zero, q(1_i4b:m,1_i4b:nqb), m  )
#else
    q(:m,:nqb) = matmul( mat(:m,:n), bt(:n,:nqb) )
#endif
!
!   DO POWER OR SUBSPACE ITERATIONS.
!
    do i = 1_i4b, niter2
!
        if ( do_ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
            call qr_cmp( q(:m,:nqb), diagr(:nqb), beta(:nqb) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX
!           TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM COLUMN SUBSPACE.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( q(:m,:nqb), beta(:nqb) )
!
        end if
!
!       COMPUTE RANDOM COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'T', 'N', n, nqb, m, one, mat(1_i4b:m,1_i4b:n), m,       &
                   q(1_i4b:m,1_i4b:nqb), m, zero, bt(1_i4b:n,1_i4b:nqb), n  )
#else
        qt(:nqb,:m) = transpose( q(:m,:nqb) )
!
        b(:nqb,:n) = matmul( qt(:nqb,:m), mat(:m,:n) )
!
        bt(:n,:nqb) = transpose( b(:nqb,:n) )
#endif
!
        if ( do_ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
            call qr_cmp( bt(:n,:nqb), diagr(:nqb), beta(:nqb) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX
!           TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM ROW SUBSPACE.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( bt(:n,:nqb), beta(:nqb) )
!
        end if
!
!       COMPUTE RANDOM ROW SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nqb, n, one, mat(1_i4b:m,1_i4b:n), m,       &
                   bt(1_i4b:n,1_i4b:nqb), n, zero, q(1_i4b:m,1_i4b:nqb), m  )
#else
        q(:m,:nqb) = matmul( mat(:m,:n), bt(:n,:nqb) )
#endif
!
    end do
!
!   COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
    call qr_cmp( q(:m,:nqb), diagr(:nqb), beta(:nqb) )
!
!   GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX
!   TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM COLUMN SUBSPACE.
!   q IS ASSUMED OF FULL RANK.
!
    call ortho_gen_qr( q(:m,:nqb), beta(:nqb) )
!
!   COMPUTE FINAL RANDOM COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
    call gemm( 'T', 'N', nqb, n, m, one, q(1_i4b:m,1_i4b:nqb), m,        &
               mat(1_i4b:m,1_i4b:n), m, zero, b(1_i4b:nqb,1_i4b:n), nqb  )
#else
    qt(:nqb,:m) = transpose( q(:m,:nqb) )
!
    b(:nqb,:n) = matmul( qt(:nqb,:m), mat(:m,:n) )
#endif
!
!   COMPUTE A PARTIAL QR OR COMPLETE ORTHOGONAL FACTORIZATION OF mat IF REQUIRED.
!
    if ( do_qr ) then
!
        i = n
!
        call qrfac( name_proc, b(:nqb,:n), kfix, i, comp_min_norm, diagr(:nqb), &
                    beta(:nqb), bt(:n,1_i4b), tol=tol, ip=ip )
!
        if ( i<nqb ) then
!
            call merror( name_proc//random_error13 )
!
        end if
!
!       OUTPUT OPTIONAL ARGUMENT tau, IF REQUIRED.
!
        if ( comp_min_norm ) then
!
            tau(:i) = bt(:i,1_i4b)
!
        end if
!
!        call qr_cmp( b(:nqb,:n), diagr(:nqb), beta(:nqb) )
!
!       UPDATE THE ORTHONORMAL MATRIX q.
!
        call apply_q_qr( b(:nqb,:nqb), beta(:nqb), q(:m,:nqb), left=false, trans=false )
!
!       RESTORE THE UPPER TRAPEZOIDAL R OR TRIANGULAR MATRIX T1.
!
        do i = 1_i4b, nqb
!
            b(i,i) = diagr(i)
!
            b(i+1_i4b:nqb,i) = zero
!
        end do
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    deallocate( bt, diagr, beta )
#else
    deallocate( bt, qt, diagr, beta )
#endif
!
!
! END OF SUBROUTINE rqb_cmp
! _________________________
!
    end subroutine rqb_cmp
!
! =========================================================================================
!
    subroutine rqb_cmp_fixed_precision( mat, relerr, q, b, failure_relerr, niter, rng_alg, &
                                        blk_size, maxiter_qb, ortho, reortho, niter_qb,    &
                                        comp_qr, ip, tol, tau                              )
!
! Purpose
! _______
!
!   RQB_CMP_FIXED_PRECISION computes a partial QB, QR (eventually with column pivoting) or complete orthogonal
!   factorization of a full m-by-n real matrix MAT using randomized power or subspace iterations.
!
!   nqb is the target rank of the partial QB, QR or complete orthogonal factorization, which is sought, 
!   and this partial factorization  must have an approximation error which fulfills:
!
!            || MAT - Q*B ||_F <= ||MAT||_F * relerr
!
!   , where Q is a m-by-nqb orthonormal matrix, B is a nqb-by-n matrix and the matrix product Q*B
!   is the computed matrix approximation. || ||_F is the Frobenius norm and relerr is a prescribed
!   accuracy tolerance for the relative error of the computed matrix approximation, specified in
!   the input argument RELERR.
!   
!   In other words, the rank, nqb, of the matrix approximation is not known in advance and is
!   determined in the subroutine. This explains why the output array arguments Q and B, which contain
!   the computed factors of the matrix approximation, must be declared in the calling program as pointers.
!   
!   On exit, nqb is equal to the numbers of columns of the output array pointer argument Q, which is also equal
!   to the numbers of rows of the output array pointer argument B. In other words, nqb = size( Q, 2 ) = size( B, 1 )
!   and the relative error, in the Frobenius norm, of the computed matrix approximation Q * B is output in argument
!   RELERR.
!
!   RQB_CMP_FIXED_PRECISION searches incrementally the best (e.g., of smallest rank) Q * B  approximation,
!   which fulfills the prescribed accuracy tolerance for the relative error. More precisely, the rank of
!   the matrix approximation is increased progressively of BLK_SIZE by BLK_SIZE until the prescribed
!   accuracy tolerance is satisfied and than adjusted precisely to obtain the Q * B matrix
!   approximation of smallest rank, which satisfies the prescribed tolerance.
!
!   Note that the product of the two integer arguments BLK_SIZE and MAXITER_QB (see below for their precise meaning),
!   BLK_SIZE*MAXITER_QB, determines the maximum allowable rank of the matrix approximation,
!   which is sought. In other words, the subroutine will stop the search for the best (e.g., smallest)
!   matrix approximation, which fulfills the requested tolerance, if the rank of this matrix
!   approximation exceeds BLK_SIZE*MAXITER_QB. In that case, the subroutine will return the current
!   matrix approximation (with a rank equal to BLK_SIZE*MAXITER_QB).
!
!   In all cases the relative error of the computed matrix approximation is output in argument RELERR.
!
!   If, finally, the optional logical argument FAILURE_RELERR is used, it will be set
!   to true if the computed matrix approximation does not fulfill the requested relative error specified
!   on entry in the argument RELERR and to false otherwise.
!
!   In a second step:
!
!   - if the optional logical argument COMP_QR is used with the value true and the optional array arguments
!     IP and TAU are absent, a QR factorization of B is computed to obtain an approximate QR factorization of MAT:
!
!             MAT ≈ Q * B = Q * ( O * R ) = (Q * O ) * R
!
!     , where O is an nqb-by-nqb orthogonal matrix and R is an nqb-by-n upper trapezoidal matrix.
!
!   - if the optional array argument IP is present, a QR factorization of B with column pivoting is performed
!     to obtain an approximate QR factorization with column pivoting of MAT :
!
!             MAT * P ≈ Q * ( O * R ) = (Q * O ) * R
!
!     , where P is an n-by-n permutation matrix, O is an nqb-by-nqb orthogonal matrix and R is an nqb-by-n upper
!     trapezoidal matrix.
!
!   - if the optional array argument TAU is present, a complete orthogonal factorization of B is performed
!     to obtain an approximate complete orthogonal factorization of MAT:
!
!             MAT ≈ Q * ( O * T * Z ) = (Q * O ) * T * Z
!
!     , where O is an nqb-by-nqb orthogonal matrix, Z is a n-by-n orthogonal matrix and
!     T is a nqb-by-n matrix, which has the form:
!
!     [  T1  0  ]
!
!     , where T1 is a nqb-by-nqb upper triangular matrix.
!
!   - if, finally, both the optional array arguments IP and TAU are present, a complete orthogonal factorization
!     of B with column pivoting is performed to obtain an approximate complete orthogonal factorization with column 
!     pivoting of MAT :
!
!             MAT * P  ≈ Q * ( O * T * Z ) = (Q * O ) * T * Z
!
!     , where P is a n-by-n permutation matrix, O is an nqb-by-nqb orthogonal matrix,
!     Z is a n-by-n orthogonal matrix and T a nqb-by-n matrix, which has the form:
!
!     [  T1  0  ]
!
!     , where T1 is a nqb-by-nqb upper triangular matrix.
!   
!   If a partial QR or complete orthogonal factorization is computed, we have the following partition of R:
!
!                                   R = [  R1  R2  ]
!
!   where R1 is a nqb-by-nqb upper triangular matrix and R2 is a full nqb-by-(n-nqb) matrix. Then, if:
!
!   - the optional scalar argument TOL is present and is in ]0,1[, the rank of R1 is determined by finding
!     the submatrix of R1 which is defined as the largest leading submatrix whose estimated condition number,
!     in the 1-norm, is less than 1/TOL. The order of this submatrix, krank, is the effective rank of R1.
!   - the optional scalar argument TOL is present and is equal to 0, the numerical rank of R1, krank, is determined.
!   - the optional scalar argument TOL is absent or outside [0,1[, the numerical rank of R1, krank, is
!     determined by finding the largest leading submatrix of R1 such that abs(R1[j,j])>0.
!
!   In all cases, if krank < nqb, this indicates that column pivoting must be used or, if column pivoting has been already
!   specified, that the rank of MAT is probably less than nqb and nqb = size( Q, 2 ) has been set to a too large value.
!   In such cases, the subroutine will exit with an error message.
!   
!   
! Arguments
! _________
!
!   MAT                 (INPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       MAT is not modified by the routine.
!
!   RELERR              (INPUT/OUTPUT) real(stnd)
!                       On entry, the requested accuracy tolerance for the relative error of
!                       the computed matrix approximation.
!
!                       The preset value for RELERR must be greater than 4*epsilon( RELERR ), less than one
!                       and verifies:
!
!                       - RELERR >= 2 * sqrt( epsilon( RELERR )/RELERR )
!
!                       and is forced to be greater than 2*sqrt( epsilon( RELERR )/RELERR )
!                       if this is not the case to avoid loss of accuracy in the algorithm.
!                       See reference (6) for more details.
!
!                       On exit, RELERR contains the relative error of the computed matrix
!                       approximation:
!
!                       - RELERR = ||MAT-Q*B||_F / ||MAT||_F
!
!   Q                   (OUTPUT) real(stnd), dimension(:,:), pointer
!                       On exit, the computed m-by-nqb orthonormal matrix of the partial QB, QR or complete
!                       orthogonal factorization of MAT.
!
!                       The statut of the pointer Q must not be undefined on entry.
!                       If, on entry, the pointer Q is already allocated, it will be first deallocated
!                       and then reallocated with the correct shape.
!
!                       On exit, the shape of the pointer Q will verify:
!
!                       - size( Q, 1 ) = m    = size( MAT, 1 ),
!                       - size( Q, 2 ) = nqb <= min( size(MAT,1) , size(MAT,2) ) .
!
!   B                   (OUTPUT) real(stnd), dimension(:,:), pointer
!                       On exit:
!
!                       - the computed B matrix of the partial QB factorization of MAT if the optional arguments
!                         COMP_QR, IP and TAU are absent or if COMP_QR is used with the value false;
!                       - the upper trapezoidal matrix R of the partial QR factorization if the optional logical
!                         argument COMP_QR is used with the value true or if the optional array argument IP is present;
!                       - the computed matrices T1 and Z of the partial complete orthogonal
!                         factorization of MAT if the optional argument TAU is present.
!
!                       The statut of the pointer B must not be undefined on entry.
!                       If, on entry, the pointer B is already allocated, it will be first deallocated
!                       and then reallocated with the correct shape.
!
!                       On exit, the shape of the pointer B will verify:
!
!                       - size( B, 1 ) = size( Q, 2 )   = nqb ,
!                       - size( B, 2 ) = size( MAT, 2 ) = n .
!
!   FAILURE_RELERR      (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit, if the optional logical argument FAILURE_RELERR is present, it is
!                       set as follows:
!
!                       - FAILURE_RELERR = false : indicates successful exit and the computed matrix approximation
!                         fulfills the requested relative error specified on entry in the argument RELERR,
!                       - FAILURE_RELERR = true  : indicates that the computed matrix approximation has a relative error
!                         larger than the requested relative error. This means that the requested accuracy tolerance
!                         for the relative error is too small (i.e., RELERR < 2 * sqrt( epsilon( RELERR )/RELERR )
!                         or that the input parameters BLK_SIZE and/or MAXITER_QB have a too small value, given the
!                         distribution of the singular values of MAT, and must be increased to fullfill the preset
!                         accuracy tolerance for the relative error of the matrix approximation.
!
!   NITER               (INPUT, OPTIONAL) integer(i4b)
!                       The number of randomized power or subspace iterations performed
!                       in the first phase of the randomized algorithm for computing the preliminary
!                       QB factorization.
!
!                       NITER must be positive or null.
!
!                       By default, 1 randomized power or subspace iteration is performed.
!
!   RNG_ALG             (INPUT, OPTIONAL) integer(i4b)
!                       On entry, a scalar integer to select the random (uniform) number generator used to build
!                       the random gaussian test matrix in the randomized partial QB or QR algorithm.
!
!                       The possible values are:
!
!                       - ALG=1  : selects the Marsaglia's KISS random number generator;
!                       - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!                       - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!                       - ALG=4  : selects the Mersenne Twister random number generator;
!                       - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!                       - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!                       - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!                       - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!                       - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!                       - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                         random number generator;
!
!                       For other values, the current random number generator and its current state are not changed.
!                       Note further, that, on exit, the current random number generator is not reset to its previous
!                       value before the call to RQB_CMP_FIXED_PRECISION.
!
!                       See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   BLK_SIZE            (INPUT, OPTIONAL) integer(i4b)
!                       On entry, the block size used in the randomized QB factorization, which is used in
!                       the first phase of the subroutine.
!
!                       BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!                       smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!                       By default, BLK_SIZE is set to min( 10, min(m,n) ).
!
!   MAXITER_QB          (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER_QB controls the maximum number of allowed iterations in the randomized
!                       QB algorithm, which is used in the first phase of the subroutine.
!
!                       MAXITER_QB must be set greater or equal to one and less than int( min(m,n)/BLK_SIZE ).
!
!                       By default, MAXITER_QB is set to max( 1, int( min(m,n)/(4*BLK_SIZE) ) ).
!
!   ORTHO               (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if:
!
!                       - ORTHO=true, orthonormalization is carried out between each step of
!                         the power iterations, to avoid loss of accuracy due to rounding errors.
!                         This means that subspace iterations are used instead of power iterations
!                         in the QB phase of the algorithm,
!                       - ORTHO=false, orthonormalization is not performed. 
!
!                       The default is to use orthonormalization, e.g., ORTHO=true.
!
!   REORTHO             (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if:
!
!                       - REORTHO=true, a reorthogonalization step is performed to avoid the loss
!                         of orthogonality in the Gram-Schmidt procedure, which is used in the
!                         randomized QB factorization;
!                       - REORTHO=false, a reorthogonalization step is not performed in
!                         the Gram-Schmidt procedure.
!
!                       The default is to use a reorthogonalization step, e.g., REORTHO=true.
!
!   NITER_QB            (INPUT, OPTIONAL) integer(i4b)
!                       The number of subspace iterations performed in the last phase of the QB algorithm
!                       for improving the initial QB factorization of MAT.
!
!                       NITER_QB must be greater or equal to 0.
!
!                       By default, 2 final subspace iterations are performed.
!
!   COMP_QR             (INPUT, OPTIONAL) logical(lgl)
!                       The optional logical argument COMP_QR determines if a partial QB or QR
!                       factorization is computed.
!
!                       On entry, if:
!
!                       - COMP_QR=true, a partial QR factorization is computed;
!                       - COMP_QR=false, a partial QB factorization is computed. 
!
!                       The default is to compute a partial QB factorization, e.g., COMP_QR=false.
!
!   IP                  (OUTPUT, OPTIONAL) integer(i4b), dimension(:) 
!                       On entry, if IP is present a partial QR factorization with column pivoting of MAT
!                       is performed instead of a QB factorization (e.g., this implies COMP_QR=true).
!
!                       On exit, if IP(j)=k, then the j-th column of MAT*P was the k-th column of MAT.
!
!                       See Further Details.
!
!                       The size of IP must verify: size( MAT, 2 ) = n.
!
!   TOL                 (INPUT/OUTPUT, OPTIONAL) real(stnd)
!                       If COMP_QR=true and TOL is present and is in [0,1[, then:
!
!                       - the calculations to determine the condition number of
!                         of R1 in the 1-norm are performed. Then, TOL is used to determine
!                         the effective rank of R1, which is defined as the order of 
!                         the largest leading triangular submatrix of R1, whose estimated
!                         condition number is less than 1/TOL.
!                       - if TOL=0 is specified the numerical rank of R1 is determined.
!                       - on exit, the reciprocal of the condition number is returned in TOL.
!
!                       If COMP_QR=true, but TOL is not specified or is outside [0,1[ :
!
!                       - the calculations to determine the condition number of R1 are not
!                         performed and crude tests on R1(j,j) are done to determine
!                         the rank of R1. If TOL is present, it is not changed.
!
!                       If each element of MAT is correct to, say, 5 digits then TOL=0.00001 should
!                       be used on entry.
!
!   TAU                 (OUTPUT, OPTIONAL) real(stnd), dimension(:), pointer
!                       On entry, if TAU is present, a complete orthogonal factorization of
!                       MAT is computed instead of a QB factorization (e.g., this implies COMP_QR=true).
!
!                       On exit, the scalars factors of the elementary reflectors defining Z.
!
!                       See Further Details.
!
!                       The statut of the pointer TAU must not be undefined on entry.
!                       If, on entry, the pointer TAU is already allocated, it will be first deallocated
!                       and then reallocated with the correct size.
!
!                       On exit, the size of the pointer TAU will verify:
!
!                       - size( TAU ) = nqb = size( Q, 2 ) = size( B, 1 ).
!
!
! Further Details
! _______________
!
!   For a good introduction to randomized linear algebra , see the references (1) and (2).
!
!   The randomized subspace iteration was proposed in (3; see Algorithm 4.4) to compute
!   an orthonormal matrix whose range approximates the range of MAT. An approximate partial QB,
!   QR or complete orthogonal factorization can then be computed using the aforementioned
!   orthonormal matrix, see the references (1) and (3) for details.
!
!   Usually, the problem of low-rank matrix approximation falls into two categories:
!
!   - the fixed-rank problem, where the rank parameter nqb is given;
!   - the fixed-precision problem, where we seek a partial matrix factorization, Q * B, of rank
!     as small as possible such that
!     
!     || MAT - Q*B ||_F <= eps
!
!     , where eps is a given accuracy tolerance.
!
!   RQB_CMP_FIXED_PRECISION is dedicated to solve the fixed-precision problem. The fixed-rank problem
!   can be solved by subroutine RQB_CMP.
!
!   RQB_CMP_FIXED_PRECISION uses an improved version of the "randQB_FP" algorithm described in the
!   reference (6) to solve the fixed-precision problem.
!
!   The orthonormal m-by-nqb matrix Q of the partial QB, QR or complete orthogonal factorization of MAT
!   is computed explicitly and stored on exit in the real array pointer Q.
!
!   The nqb-by-n matrix B of the partial QB factorization or the upper trapezoidal matrix R of the partial
!   QR or orthogonal factorization of MAT is stored on exit in the real array pointer B.
!
!   If the integer array argument IP is present, a (partial) QR factorization of MAT with column pivoting is
!   computed, as described above, and on exit the permutation matrix P is represented in the array IP as follows:
!   If IP(j) = i then the jth column of P is the ith canonical unit vector.
!
!   Next, if the real array pointer TAU is present, a (partial) complete orthogonal factorization of MAT is
!   computed from the partial QR factorization of MAT. The factorization is obtained by Householder's method.
!   The kth transformation matrix, Z(k), which is used to introduce zeros into the kth row of R of the (partial)
!   QR factorization of MAT, is given in the form
!
!   [  I  0    ]
!
!   [  0  T(k) ]
!
!   where
!
!         T(k) = I + tau * ( u(k) * u(k)' ) and u(k)' = (  1  0  z(k)  )
!
!   tau is a scalar, u(k) is a n-k+1 vector and z(k) is an (n-nqb) element vector.
!   tau and z(k) are chosen to annihilate the elements of the kth row of R12.
!
!   The Z n-by-n orthogonal matrix is finally given by the product
!
!         Z = Z(1) * Z(2) * ... * Z(nqb)
!
!   On exit:
!
!   - the orthogonal matrix Z is stored in factored form. The scalar tau,
!     which defines Z(k) is returned in the kth element of TAU and the vector u(k)
!     in the kth row of B, such that the elements of z(k) are in B(k,nqb+1:n).
!   - the upper triangular nqb-by-nqb matrix T1, which defines the T factor of the
!     (partial) complete orthogonal factorization of MAT (see above) is stored in the upper
!     triangle of the real array argument B.
!
!   Finally, if the optional arguments IP and TAU are both present, a (partial) complete orthogonal
!   factorization with column pivoting of MAT is computed and on exit this factorization is stored
!   with the same conventions as described above.
!
!   For further details, on randomized linear algebra, computing low-rank matrix approximations
!   using randomized power or subspace iterations or solving the fixed-precision problem, see:
!
!   (1) Martinsson, P.G., 2019:
!            Randomized methods for matrix computations.
!            arXiv.1607.01649
!
!   (2) Erichson, N.B., Voronin, S., Brunton, S.L., and Kutz, J.N., 2019:
!            Randomized matrix decompositions using R.
!            arXiv.1608.02148
!
!   (3) Halko, N., Martinsson, P.G., and Tropp, J.A., 2011:
!            Finding structure with randomness:
!            probabilistic algorithms for constructing approximate matrix decompositions.
!            SIAM Rev., 53, 217-288.
!
!   (4) GU, M., 2015:
!            Subspace iteration randomization and singular value problems.
!            SIAM J. Sci. Comput., 37, A1139-A1173.
!
!   (5) Martinsson, P.-G., and Voronin, S., 2016:
!            A randomized blocked algorithm for efficiently computing
!            rank-revealing factorizations of matrices.
!            SIAM J. Sci. Comput., 38:5, S485-S507.
!
!   (6) Yu, W.,  Gu, Y., and Li, Y., 2018:
!            Efficient randomized algorithms for the fixed-precision low-rank
!            matrix approximation.
!            SIAM J. Mat. Ana. Appl., 39:3, 1339-1359.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error, random_error12, random_error13
    use Utilities,       only : assert, merror
    use QR_Procedures,   only : qrfac, qr_cmp, ortho_gen_qr, apply_q_qr
    use Lin_Procedures,  only : triang_solve
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:,:)                    :: mat
    real(stnd),  intent(out), dimension(:,:), pointer           :: q, b
    real(stnd),  intent(out), dimension(:),   pointer, optional :: tau
    real(stnd),  intent(inout)                                  :: relerr
    real(stnd),  intent(inout),                        optional :: tol
!
    integer(i4b), intent(in),                optional :: niter, blk_size, maxiter_qb, niter_qb, rng_alg
    integer(i4b), intent(out), dimension(:), optional :: ip
!
    logical(lgl), intent(in),  optional :: ortho, reortho, comp_qr
    logical(lgl), intent(out), optional :: failure_relerr
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: err, threshold, norm_mat, tmp, tmp2
!
#ifdef _BLAS
    real(stnd), dimension(:,:), allocatable :: q0, b0, h, r, r2, ti, ti2, yi, bi
#else
    real(stnd), dimension(:,:), allocatable :: q0, q0t, b0, b0t, r, r2, ti, ti2, yi, bi
#endif
    real(stnd), dimension(:),   allocatable :: diagr, beta
!
    integer      :: iok, alg
    integer(i4b) :: m, n, mn, nqb, np, i, i0, i1, i2, j, kfix, niter2, blk_size2, maxiter_qb2, niter_qb2
!
    logical(lgl) :: do_test_relerr, do_ortho, do_reortho, do_qr, comp_min_norm
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='rqb_cmp_fixed_precision'
!
!   niter0 IS THE DEFAULT NUMBER OF RANDOMIZED POWER OR SUBSPACE ITERATIONS PERFORMED FOR COMPUTING
!   THE PRELIMINARY QB FACTORIZATION IN THE FIRST PHASE OF THE ALGORITHM.
!   blk_size0 IS THE DEFAULT BLOCK SIZE USED IN THE RANDOMIZED QB FACTORIZATION.
!   niter_qb0 IS THE DEFAULT NUMBER OF SUBSPACE ITERATIONS PERFORMED FOR IMPROVING THE INITIAL QB FACTORIZATION
!   BEFORE COMPUTING THE PARTIAL MATRIX FACTORIZATION IN THE SECOND PHASE OF THE ALGORITHM.
!
    integer(i4b),  parameter :: niter0=1_i4b, blk_size0=10_i4b, niter_qb0=2_i4b
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
!
    m = size( mat, 1 )
!
    n = size( mat, 2 )
!
    mn = min( m, n )
!
!   DEALLOCATE POINTERS ON ENTRY.
!
#ifndef _USE_NAGWARE
    if ( associated( q ) ) then
!
        deallocate( q )
!
    end if
!
    if ( associated( b ) ) then
!
        deallocate( b )
!
    end if
#endif
!
!   DETERMINE IF A TEST IS REQUIRED.
!
    do_test_relerr = present( failure_relerr )
!
    if ( do_test_relerr ) then
        failure_relerr = true
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( mn<=0_i4b  ) then
!
        relerr = one
!
        return
!
    end if
!
!   CHECK VALIDITY OF THE REQUESTED TOLERANCE FOR THE RELATIVE
!   ERROR OF THE COMPUTED MATRIX APPROXIMATION.
!
    tmp = four*machulp
!
    if ( relerr<=tmp .or. relerr>=one ) then
!
        call merror( name_proc//random_error12 )
!
    end if
!
!   CHECK OPTIONAL ARGUMENTS.
!
!   DETERMINE THE NUMBER OF POWER OR SUBSPACE ITERATIONS
!   niter2 TO BE PERFORMED IN THE FIRST PHASE OF THE QB FACTORIZATION.
!
    if ( present(niter) ) then
        niter2 = max( niter, 0_i4b )
    else
        niter2 = niter0
    end if
!
!   DETERMINE THE BLOCK SIZE USED IN THE RANDOMIZED QB FACTORIZATION.
!
    if ( present(blk_size) ) then
        blk_size2 = max( blk_size, 1_i4b )
    else
        blk_size2 = blk_size0
    end if
!
    blk_size2 = min( mn, blk_size2 )
!
!   DETERMINE THE MAXIMUM NUMBER OF ALLOWED ITERATIONS
!   IN THE RANDOMIZED QB FACTORIZATION.
!
    if ( present(maxiter_qb) ) then
!
        maxiter_qb2 = max( maxiter_qb, 1_i4b )
!
        tmp = real( mn, stnd )/real( blk_size2, stnd )
        maxiter_qb2 = min( maxiter_qb2, int(tmp,i4b) )
!
    else
!
        tmp = real( mn, stnd )/(four*real( blk_size2, stnd ))
        maxiter_qb2 = max( int(tmp,i4b), 1_i4b )
!
    end if
!
!   np SETS THE MAXIMUM ALLOWABLE RANK OF THE QB FACTORIZATION AND MATRIX APPROXIMATION, WHICH ARE SOUGHT.
!
    np = blk_size2*maxiter_qb2
!
!   DETERMINE IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE POWER ITERATIONS (I.E. SHIFTING TO SUBSPACE ITERATIONS),
!   TO AVOID LOSS OF ACCURACY DUE TO ROUNDING ERRORS.
!
    if ( present(ortho) ) then
        do_ortho = ortho
    else
        do_ortho = true
    end if
!
!   DETERMINE IF REORTHOGONALIZATION IS CARRIED OUT AFTER EACH STEP
!   OF THE GRAM-SCHMIDT PROCEDURE TO AVOID LOSS OF ORTHOGONALITY IN
!   THE QB FACTORIZATION.
!
    if ( present(reortho) ) then
        do_reortho = reortho
    else
        do_reortho = true
    end if
!
!   DETERMINE THE NUMBER OF SUBSPACE ITERATIONS
!   niter_qb2 TO BE PERFORMED IN THE LAST PHASE OF THE
!   ALGORITHM TO IMPROVE ESTIMATION OF THE QB FACTORIZATION
!   AND OF THE SINGULAR TRIPLETS.
!
    if ( present(niter_qb) ) then
        niter_qb2 = max( niter_qb, 1_i4b )
    else
        niter_qb2 = niter_qb0
    end if
!
!   DETERMINE IF A QB OR QR FACTORIZATION MUST BE COMPUTED.
!
    if ( present(comp_qr) ) then
        do_qr = comp_qr
    else
        do_qr = false
    end if
!
!   DETERMINE IF A QR FACTORIZATION WITH COLUMN PIVOTING MUST BE COMPUTED.
!
    if ( present(ip) ) then
!
        i = size( ip )
!
        call assert( logical( i==n, lgl ), name_proc )
!
        do_qr = true
!
        kfix = 0_i4b
!
    else
!
        kfix = n
!
    end if
!
!   DETERMINE IF A COMPLETE ORTHOGONAL FACTORIZATION MUST BE COMPUTED.
!
    if ( present(tau) ) then
!
#ifndef _USE_NAGWARE
        if ( associated( tau ) ) then
!
            deallocate( tau )
!
        end if
#endif
!
        comp_min_norm = true
        do_qr         = true
!
    else
!
        comp_min_norm = false
!
    end if
!
!   SELECT AND SET THE RANDOM NUMBER GENERATOR TO BE USED IF REQUIRED.
!
    if ( present(rng_alg) ) then
!
        if ( rng_alg>=1_i4b .and. rng_alg<=10_i4b ) then
!
            alg = rng_alg
!
!           SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
            call random_seed_( alg=alg )
!
!           RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
            call random_seed_()
!
        end if
!
    end if
!
!   TEST VALIDITY OF THE PRESET RELATIVE ACCURACY TOLERANCE relerr AND
!   COMPUTE THE REQUIRED THRESHOLD.
!
    tmp = two*sqrt( machulp/relerr )
    threshold = max( tmp, relerr )
!
!   COMPUTE THE FROBENIUS NORM OF THE INPUT MATRIX.
!
    norm_mat = norm( mat(1_i4b:m,1_i4b:n) )
!
    err = one
!
!   ALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    allocate( q0(m,np), b0(n,np), diagr(np), beta(np), h(n,np),  &
              r(blk_size2,blk_size2), yi(m,blk_size2),           &
              ti(np,blk_size2), bi(blk_size2,n), stat=iok        )
#else
    allocate( q0(m,np), q0t(np,m), b0(n,np), b0t(np,n) ,diagr(np), &
              beta(np), r(blk_size2,blk_size2), yi(m,blk_size2),   &
              ti(np,blk_size2), bi(blk_size2,n), stat=iok          )
#endif
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    if ( do_reortho ) then
!
        allocate( r2(blk_size2,blk_size2), ti2(np,blk_size2), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
    end if
!
!   GENERATE A n-BY-np IID GAUSSIAN MATRIX.
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
    call normal_random_number3_( b0(:n,:np) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
!    do i = 1_i4b, np
!
!        call normal_random_number3_( b0(:n,i) )
!
!    end do
!
!   COMPUTE RANDOM SAMPLE MATRIX.
!
#ifdef _BLAS
    call gemm( 'N', 'N', m, np, n, one, mat(1_i4b:m,1_i4b:n), m,       &
               b0(1_i4b:n,1_i4b:np), n, zero, q0(1_i4b:m,1_i4b:np), m  )
#else
    q0(:m,:np) = matmul( mat(:m,:n), b0(:n,:np) )
#endif
!
!   DO POWER OR SUBSPACE ITERATIONS.
!
    do i = 1_i4b, niter2
!
        if ( do_ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
            call qr_cmp( q0(:m,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX Q OF QR DECOMPOSITION OF RANDOM MATRIX
!           TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM COLUMN SUBSPACE.
!           Q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( q0(:m,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'T', 'N', n, np, m, one, mat(1_i4b:m,1_i4b:n), m,       &
                   q0(1_i4b:m,1_i4b:np), m, zero, b0(1_i4b:n,1_i4b:np), n  )
#else
        q0t(:np,:m) = transpose( q0(:m,:np) )
!
        b0t(:np,:n) = matmul( q0t(:np,:m), mat(:m,:n) )
!
        b0(:n,:np) = transpose( b0t(:np,:n) )
#endif
!
        if ( do_ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
            call qr_cmp( b0(:n,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX Q OF QR DECOMPOSITION OF RANDOM MATRIX.
!           TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM ROW SUBSPACE.
!           Q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( b0(:n,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM ROW SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, np, n, one, mat(1_i4b:m,1_i4b:n), m,       &
                   b0(1_i4b:n,1_i4b:np), n, zero, q0(1_i4b:m,1_i4b:np), m  )
#else
        q0(:m,:np) = matmul( mat(:m,:n), b0(:n,:np) )
#endif
!
    end do
!
!   FINAL COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
    call gemm( 'T', 'N', n, np, m, one, mat(1_i4b:m,1_i4b:n), m,     &
               q0(1_i4b:m,1_i4b:np), m, zero, h(1_i4b:n,1_i4b:np), n  )
#else
    q0t(:np,:m) = transpose( q0(:m,:np) )
!
    b0t(:np,:n) = matmul( q0t(:np,:m), mat(:m,:n) )
#endif
!
    i0  = 0_i4b
    nqb = np
!
    r(1_i4b:blk_size2,1_i4b:blk_size2) = zero
    ti(1_i4b:np,1_i4b:blk_size2)       = zero
!
    if ( do_reortho ) then
!
        r2(1_i4b:blk_size2,1_i4b:blk_size2) = zero
        ti2(1_i4b:np,1_i4b:blk_size2)       = zero
!
    end if
!
!   COMPUTE THE QB FACTORIZATION.
!
    iter : do i = 1_i4b, maxiter_qb2
!
        i1 = i0 + 1_i4b
        i2 = i0 + blk_size2
!
        if ( i==1_i4b ) then
!
            yi(:m,:blk_size2) = q0(:m,i1:i2)
!
        else
!
!           PERFORM BLOCK GRAM-SCHMIDT ORTHOGONALISATION STEP.
!
#ifdef _BLAS
            call gemm( 'T', 'N', i0, blk_size2, n, one, b0(1_i4b:n,1_i4b:i0), n,     &
                       b0(1_i4b:n,i1:i2), n, zero, ti(1_i4b:np,1_i4b:blk_size2), np  )
!
            call gemm( 'N', 'N', m, blk_size2, i0, -one, q0(1_i4b:m,1_i4b:i0), m,   &
                       ti(1_i4b:np,1_i4b:blk_size2), np, one, q0(1_i4b:m,i1:i2), m  )
#else
            ti(:i0,:blk_size2) = matmul( b0t(:i0,:n), b0(:n,i1:i2) )
!
            q0(:m,i1:i2) = q0(:m,i1:i2) - matmul( q0(:m,:i0), ti(:i0,:blk_size2) )
#endif
!            
            yi(:m,:blk_size2) = q0(:m,i1:i2)
!
        end if
!
        call qr_cmp( q0(:m,i1:i2), diagr(:blk_size2), beta(:blk_size2) )
!
!       SAVE UPPER TRIANGULAR MATRIX FOR LATER USE.
!
        if ( i==1_i4b .or. .not.do_reortho ) then
!
            do j = 1, blk_size2
!
                r(1_i4b:j-1_i4b,j) = q0(1_i4b:j-1_i4b,i0+j)
                r(j,j)             = diagr(j)
!                r(j+1_i4b:blk_size2,j) = zero
!
            end do
!
        else
!
            do j = 1, blk_size2
!
                r2(1_i4b:j-1_i4b,j) = q0(1_i4b:j-1_i4b,i0+j)
                r2(j,j)             = diagr(j)
!                r2(j+1_i4b:blk_size2,j) = zero
!
            end do
!
        end if
!
!       GENERATE COLUMNS i1 TO i2 OF THE ORTHOGONAL MATRIX OF THE QB FACTORIZATION.
!
        call ortho_gen_qr( q0(:m,i1:i2), beta(:blk_size2) )
!
        if ( i/=1_i4b ) then
!
            if ( do_reortho ) then
!
!               REORTHONORMALIZATION IS CARRIED OUT TO AVOID LOSS OF ORTHOGONALITY
!               IN THE BLOCK GRAM-SCHMIDT ORTHOGONALISATION SCHEME USED TO BUILD
!               THE ORTHOGONAL MATRIX OF THE QB DECOMPOSITION.
!
#ifdef _BLAS
                call gemm( 'T', 'N', i0, blk_size2, m, one, q0(1_i4b:m,1_i4b:i0), m,     &
                           q0(1_i4b:m,i1:i2), m, zero, ti2(1_i4b:np,1_i4b:blk_size2), np )
!
                call gemm( 'N', 'N', m, blk_size2, i0, -one, q0(1_i4b:m,1_i4b:i0), m,    &
                           ti2(1_i4b:np,1_i4b:blk_size2), np, one, q0(1_i4b:m,i1:i2), m  )
#else
                ti2(:i0,:blk_size2) = matmul( q0t(:i0,:m), q0(:m,i1:i2) )
!
                q0(:m,i1:i2) = q0(:m,i1:i2) - matmul( q0(:m,:i0), ti2(:i0,:blk_size2) )
#endif
!
!               IMPROVE ESTIMATES OF COLUMNS i1 TO i2 OF THE ORTHOGONAL MATRIX.
!
                call qr_cmp( q0(:m,i1:i2), diagr(:blk_size2), beta(:blk_size2) )
!
!               SAVE UPPER TRIANGULAR MATRIX FOR LATER USE.
!
                do j = 1, blk_size2
!
                    bi(1_i4b:j-1_i4b,j)     = q0(1_i4b:j-1_i4b,i0+j)
                    bi(j,j)                 = diagr(j)
                    bi(j+1_i4b:blk_size2,j) = zero
!
                end do
!
!               UPDATE COLUMNS i1 TO i2 OF THE ORTHOGONAL MATRIX.
!
                call ortho_gen_qr( q0(:m,i1:i2), beta(:blk_size2) )
!
#ifdef _BLAS
                call gemm( 'N', 'N', blk_size2, blk_size2, blk_size2, one,       &
                           bi(1_i4b:blk_size2,1_i4b:blk_size2), blk_size2,       &
                           r2(1_i4b:blk_size2,1_i4b:blk_size2), blk_size2, zero, &
                           r(1_i4b:blk_size2,1_i4b:blk_size2), blk_size2 )
!
                call gemm( 'T', 'N', i0, blk_size2, m, one, q0(1_i4b:m,1_i4b:i0), m,             &
                           yi(1_i4b:m,1_i4b:blk_size2), m, one, ti(1_i4b:np,1_i4b:blk_size2), np )
#else
                r(:blk_size2,:blk_size2) = matmul( bi(:blk_size2,:blk_size2), r2(:blk_size2,:blk_size2) )
!
                ti(:i0,:blk_size2) = ti(:i0,:blk_size2) + matmul( q0t(:i0,:m), yi(:m,:blk_size2) )
#endif
!
            end if
!
#ifdef _BLAS
!
            call gemm( 'N', 'N', n, blk_size2, i0, -one, b0(1_i4b:n,1_i4b:i0), n,  &
                       ti(1_i4b:np,1_i4b:blk_size2), np, one, h(1_i4b:n,i1:i2), n  )
#else
            bi(:blk_size2,:i0) = transpose( ti(:i0,:blk_size2) )
            b0t(i1:i2,:n)      = b0t(i1:i2,:n) - matmul( bi(:blk_size2,:i0), b0t(:i0,:n) )
#endif
!
        end if
!
!       COMPUTE ROWS i1 TO i2 OF THE B FACTOR OF QB DECOMPOSITION.
!
#ifdef _BLAS
        bi(:blk_size2,:n) = transpose( h(:n,i1:i2) )
!
        call triang_solve( r(:blk_size2,:blk_size2), bi(:blk_size2,:n), upper=true, trans=true )
!
        b0(:n,i1:i2) = transpose( bi(:blk_size2,:n) )
#else
        bi(:blk_size2,:n) = b0t(i1:i2,:n)
!
        call triang_solve( r(:blk_size2,:blk_size2), bi(:blk_size2,:n), upper=true, trans=true )
!
        b0t(i1:i2,:n) = bi(:blk_size2,:n)
!
        if ( do_reortho ) then
!
            q0t(i1:i2,:m) = transpose( q0(:m,i1:i2) )
!
        end if
#endif
!
!       COMPUTE RELATIVE ERROR OF THE CURRENT QB FACTORIZATION.
!
        tmp  = norm( bi(:blk_size2,:n) )/norm_mat
        tmp2 = ( err - tmp )*( err + tmp )
        tmp  = sqrt( max( zero, tmp2 ) )
!
!       CHECK IF THE THRESHOLD IS SATISFIED WITH INCLUSION OF THE CURRENT BLOCK.
!
        if ( tmp<threshold  ) then
!
!           DETREMINE THE PRECISE RANK OF THE ORTHOGONAL MATRIX OF THE QB DECOMPOSITION.
!
            do j = 1, blk_size2
!
#ifdef _BLAS
                tmp = norm( b0(:n,i0+j) )/norm_mat
#else
                tmp = norm( bi(j,:n) )/norm_mat
#endif
                tmp2 = ( err - tmp )*( err + tmp )
                err  = sqrt( max( zero, tmp2 ) )
!
                if ( err<threshold  ) then
!
!                   nqb IS THE NUMBER OF COLUMNS OF THE ORTHOGONAL MATRIX OF THE QB DECOMPOSITION.
!
                    nqb = i0 + j
!
                    exit iter
!
                end if
!
            end do
!
        else
!
            err = tmp
!
        end if
!
        i0 = i0 + blk_size2
!
    end do iter
!
!   DEALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    deallocate( h, r, yi, ti, bi )
#else
    deallocate( r, yi, ti, bi )
#endif
!
    if ( do_reortho ) then
!
        deallocate(  r2, ti2 )
!
    end if
!
!   DO FINAL SUBSPACE ITERATIONS FOR IMPROVING THE QB FACTORIZATION.
!
    do i = 1_i4b, niter_qb2
!
#ifndef _BLAS
        b0(:n,:nqb) = transpose( b0t(:nqb,:n) )
#endif
!
!       COMPUTE QR DECOMPOSITION OF COLUMN PROJECTION MATRIX.
!
        call qr_cmp( b0(:n,:nqb), diagr(:nqb), beta(:nqb) )
!
!       GENERATE ORTHOGONAL MATRIX Q OF QR DECOMPOSITION OF COLUMN PROJECTION
!       MATRIX TO OBTAIN AN ORTHONORMAL BASIS OF ROW SUBSPACE.
!       Q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( b0(:n,:nqb), beta(:nqb) )
!
!       COMPUTE NEXT ROW SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nqb, n, one, mat(1_i4b:m,1_i4b:n), m,        &
                   b0(1_i4b:n,1_i4b:nqb), n, zero, q0(1_i4b:m,1_i4b:nqb), m  )
#else
        q0(:m,:nqb) = matmul( mat(:m,:n), b0(:n,:nqb) )
#endif
!
!       COMPUTE QR DECOMPOSITION OF ROW PROJECTION MATRIX.
!
        call qr_cmp( q0(:m,:nqb), diagr(:nqb), beta(:nqb) )
!
!       GENERATE ORTHOGONAL MATRIX Q OF QR DECOMPOSITION OF ROW PROJECTION
!       MATRIX TO OBTAIN AN ORTHONORMAL BASIS OF COLUMN SUBSPACE.
!       Q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( q0(:m,:nqb), beta(:nqb) )
!
!       COMPUTE NEXT COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'T', 'N', n, nqb, m, one, mat(1_i4b:m,1_i4b:n), m,        &
                   q0(1_i4b:m,1_i4b:nqb), m, zero, b0(1_i4b:n,1_i4b:nqb), n  )
#else
        q0t(:nqb,:m) = transpose( q0(:m,:nqb) )
!
        b0t(:nqb,:n) = matmul( q0t(:nqb,:m), mat(:m,:n) )
!
!        b0(:n,:nqb) = transpose( b0t(:nqb,:n) )
#endif
!
    end do
!
!   ALLOCATE OUTPUT POINTER ARGUMENTS.
!
    allocate( q(m,nqb), b(nqb,n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SAVE THE QB FACTORIZATION.
!
    q(:m,:nqb) = q0(:m,:nqb)
!
#ifdef _BLAS
    b(:nqb,:n) = transpose( b0(:n,:nqb) )
#else
    b(:nqb,:n) = b0t(:nqb,:n)
#endif
!
!   COMPUTE A PARTIAL QR OR COMPLETE ORTHOGONAL FACTORIZATION OF mat IF REQUIRED.
!
    if ( do_qr ) then
!
        i = n
!
        call qrfac( name_proc, b(:nqb,:n), kfix, i, comp_min_norm, diagr(:nqb), &
                    beta(:nqb), b0(:n,1_i4b), tol=tol, ip=ip )
!
        if ( i<nqb ) then
!
            call merror( name_proc//random_error13 )
!
        end if
!
!       OUTPUT OPTIONAL ARGUMENT tau, IF REQUIRED.
!
        if ( comp_min_norm ) then
!
!           ALLOCATE INPUT POINTER ARGUMENT tau.
!
            allocate( tau(nqb), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            tau(:i) = b0(:i,1_i4b)
            tau(i+1_i4b:nqb) = zero
!
        end if
!
!       UPDATE THE ORTHONORMAL MATRIX q.
!
        call apply_q_qr( b(:nqb,:nqb), beta(:nqb), q(:m,:nqb), left=false, trans=false )
!
!       RESTORE THE UPPER TRAPEZOIDAL R OR TRIANGULAR MATRIX T1.
!
        do i = 1_i4b, nqb
!
            b(i,i) = diagr(i)
!
            b(i+1_i4b:nqb,i) = zero
!
        end do
!
    end if
!
!   UPDATE RELATIVE ERROR OF THE COMPUTED MATRIX APPROXIMATION IF REQUIRED.
!
    if ( niter_qb2>0_i4b ) then
!
        if ( comp_min_norm ) then
!
            tmp  = norm( b(:nqb,:nqb) )/norm_mat
!
        else
!
            tmp  = norm( b(:nqb,:n) )/norm_mat
!
        end if
!
        tmp2 = ( one - tmp )*( one + tmp )
        err  = sqrt( max( tmp2, zero ) )
!
    end if
!
!   CHECK ACCURACY OF THE APPROXIMATE QB FACTORIZATION IF REQUIRED.
!
    if ( do_test_relerr ) then
!
        failure_relerr = threshold/=relerr .or. err>=threshold
!
    end if
!
!   OUTPUT RELATIVE ERROR OF THE COMPUTED QB FACTORIZATION
!   IN FROBENIUS NORM.
!
    relerr = err
!
!   DEALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    deallocate( q0, b0, beta, diagr )
#else
    deallocate( q0, q0t, b0, b0t, beta, diagr )
#endif
!
!
! END OF SUBROUTINE rqb_cmp_fixed_precision
! _________________________________________
!
    end subroutine rqb_cmp_fixed_precision
!
! =========================================================================================
!
    subroutine id_cmp( mat, ip, t, c, v, diagr, beta, rnorm, tol, random_qr, rng_alg, blk_size, nover )
!
! Purpose
! _______
!
!   ID_CMP computes a (partial) column Interpolative Decomposition (ID) of a m-by-n real matrix MAT.
!
!   A column ID factorization of rank krank approximates MAT as:
!
!                     MAT ≈ C * V
!
!   where C is an m-by-krank matrix, which consists of a subset of krank columns of MAT and V is a krank-by-n matrix,
!   which contains a krank-by-krank identity matrix as a submatrix.
!
!   Such column ID factorization can be computed with the help of a (randomized) (partial) QR factorization with
!   column pivoting of MAT (see the references (1), (2) and (4)-(7) for details), which is defined as
!
!                     MAT * P = Q * R
!
!   , where P is a n-by-n permutation matrix, R is an upper triangular or trapezoidal
!   (i.e., if n>m) matrix and Q is a m-by-m orthogonal matrix if the QR factorization is complete.
!
!   For computing a column ID decomposition of rank, krank, a partial QR factorization with column pivoting
!   of MAT of rank, krank, is sufficient, e.g., the QR decomposition can be stopped when the numbers of computed
!   columns of Q is equal to krank.
!
!   This leads implicitly to the following partition of Q:
!
!                     Q = [  Q1  Q2  ]
!
!   where Q1 is an m-by-krank orthonormal matrix and Q2 is m-by-(m-krank) orthonormal matrix
!   orthogonal to Q1, and to the following corresponding partition of R:
!
!   [  R11  R12  ]
!
!   [  R21  R22  ]
!
!   where R11 is a krank-by-krank triangular matrix, R21 is zero by construction,
!   R12 is a full krank-by-(n-krank) matrix and R22 is a full (m-krank)-by-(n-krank)
!   matrix. This leads to the following approximation of MAT:
!
!                     MAT ≈ Q1 * [  R11  R12  ] * P'
!
!   Here, we see that Q1 * R11 equals the first krank columns of A * P, and so we can define the matrix
!   C in the partial ID factorization of MAT as
!
!                     C = Q1 * R11 = MAT * P(:,:krank)
!
!   , then the dominant term Q1 * [  R11  R12  ] in the partial QR decomposition of MAT * P can be written as
!
!                     Q1 * [  R11  R12  ] = Q1 * R11 * [ I  T ] = C * [ I  T ]
!
!   where I is the identity matrix of order krank and T a krank-by-(n-krank) matrix and a solution to
!   the matrix equation:
!
!                     R11 * T = R12   ,e.g.,  T = inv(R11) * R12
!
!   Thus, we can obtain a column ID factorization of MAT from its partial QR
!   factorization with column pivoting as
!
!                     MAT ≈ (Q1 * R11) * [ I  T ] * P' = C * V
!
!   where V = [ I  T ] * P' is a krank-by-n matrix.
!
!   Finally, observe that the error matrix associated with the column ID
!   decomposition of MAT is given by
!
!                     MAT - C * V = [ 0  Q2*R22] * P'
!
!   and that the Frobenius norm of this error matrix can be computed efficiently as
!
!                  || MAT - C * V ||_F = || Q2 * R22 ||_F = || R22 ||_F
!
!
!   krank is the target rank of the column ID, which is sought, and is set to the number of rows of the
!   output array argument T.
!
!   If the optional logical argument RANDOM_QR is used with the value true, a fast randomized partial QR
!   factorization with column pivoting is used in the first phase of the ID algorithm.
!
!   
! Arguments
! _________
!
!   MAT        (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, the m-by-n matrix MAT.
!
!              On exit, MAT is overwritten by details of its partial QR factorization
!              with column pivoting.
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures or of 
!              PARTIAL_RQR_CMP subroutine in module Random for further details on how
!              the partial QR decomposition is stored compactly in MAT (and arguments IP,
!              DIAGR and BETA) on exit.
!
!              Note that, on exit, the Frobenius norm of the error associated with the computed
!              column ID is equal to norm( mat(krank+1:m,krank+1:n) ), which is the same as the Frobenius
!              norm of the error associated with the partial QR decomposition with column pivoting (of rank
!              krank) of MAT.
!
!   IP         (OUTPUT) integer(i4b), dimension(:) 
!              On exit, if IP(j)=k, then the j-th column of MAT*P was 
!              the k-th column of MAT.
!
!              The matrix C in the (column) ID of MAT corresponds to the subset of the columns
!              of MAT with the indices IP(:krank).
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures
!              and PARTIAL_RQR_CMP subroutine in module Random for further details.
!
!              The size of IP must be equal to size( MAT, 2 ) = n
!
!   T          (OUTPUT) real(stnd), dimension(:,:)
!              On exit, the krank-by-(n-krank) submatrix T = inv(R11) * R12 in the column
!              ID factorization of MAT.
!
!              The shape of T must verify:
!
!              - size( T, 1 ) = krank <= min( size(MAT,1) , size(MAT,2) ) = min(m,n)
!              - size( T, 2 ) = size( MAT, 2 ) - size( T, 1 ) = n - krank
!
!   C          (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!              On exit, the m-by-krank matrix C in the ID factorization of MAT.
!              Note that C is computed as C = Q * R11 as MAT is overwritten by
!              its partial QR factorization with column pivoting before C can be
!              estimated.
!
!              The shape of C must verify:
!
!              - size( C, 1 ) = size( MAT, 1 ) = m
!              - size( C, 2 ) = size( T, 1 )   = krank
!
!   V          (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!              On exit, the krank-by-n matrix V in the ID factorization of MAT.
!
!              The shape of V must verify:
!
!              - size( V, 1 ) = size( T, 1 )   = krank
!              - size( V, 2 ) = size( MAT, 2 ) = n
!
!   RNORM      (OUTPUT, OPTIONAL) real(stnd)
!              On exit, the Frobenius norm of the error matrix associated with the column
!              ID computed as || R22 ||_F .
!
!   DIAGR      (OUTPUT, OPTIONAL) real(stnd), dimension(:) 
!              On exit, the diagonal elements of the matrix R11 in the partial QR
!              factorization with column pivoting of MAT.
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures
!              and PARTIAL_RQR_CMP subroutine in module Random for further details.
!
!              The size of DIAGR must verify:
!
!              - size( DIAGR ) = krank <= min( size(MAT,1) , size(MAT,2) )
!
!   BETA       (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!              On exit, the scalars factors of the elementary reflectors defining Q
!              in the partial QR factorization with column pivoting of MAT.
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures
!              and PARTIAL_RQR_CMP subroutine in module Random for further details.
!
!              The size of BETA must verify:
!
!              - size( BETA ) = krank <= min( size(MAT,1) , size(MAT,2) ).
!
!   TOL        (INPUT/OUTPUT, OPTIONAL) real(stnd)
!              On entry, if TOL is present  then:
!
!              - if TOL is in ]0,1[, calculations to determine the condition number of
!                submatrix R11 in the partial QR factorization of MAT are performed.
!                TOL is used to determine the effective pseudo-rank of R11,
!                which is defined as the order of the largest leading triangular
!                submatrix in the partial QR factorization with column pivoting of MAT,
!                whose estimated condition number in the 1-norm is less than 1/TOL.
!                On exit, the reciprocal of the condition number is returned in TOL.
!              - if TOL=0 is specified, the calculations to determine the condition number
!                of R11 are not performed and crude tests on R(j,j) are done to determine
!                the numerical pseudo-rank of R11. On exit, TOL is not changed.
!
!              The TOL argument is useful to check if the matrix R11 in the QR decomposition of MAT,
!              which must be inverted to compute T (in the column ID of MAT), is sufficiently well
!              conditioned to obtain a stable and robust column ID of MAT.
!
!              If the computed pseudo-rank of R11 is less than krank = size( T, 1 ),
!              the subroutine will exit with an error message as a stable solution to the matrix
!              equation R11 * T = R12 cannot be computed.
!
!              On the other hand, if TOL is not specified or is outside [0,1[, the calculations
!              to determine the rank of R11 are not performed and this rank is assumed to be equal
!              to krank = size( T, 1 ).
!
!              If it is possible that MAT may not be of full rank then the stability of the column ID
!              decomposition of MAT, which is sought, can be checked by using TOL=relative precision
!              of the elements in MAT. If each element of MAT is correct to, say, 5 digits then TOL=0.00001
!              should be used.
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures
!              and PARTIAL_RQR_CMP subroutine in module Random for further details.
!
!   RANDOM_QR  (INPUT, OPTIONAL) logical(lgl)
!              On entry, if RANDOM_QR is used with the value true, a fast randomized partial QR
!              factorization with column pivoting is used in the first phase of the ID algorithm.
!
!              By default, RANDOM_QR = false, i.e., a standard (partial) QR factorization with
!              column pivoting is used in the first phase of the ID algorithm.
!
!   RNG_ALG    (INPUT, OPTIONAL) integer(i4b)
!              On entry, a scalar integer to select the random (uniform) number generator used to build
!              the random gaussian matrix in the randomized (partial) QR phase of the ID algorithm
!              if RANDOM_QR = true.
!
!              The possible values are:
!
!              - ALG=1  : selects the Marsaglia's KISS random number generator;
!              - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!              - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!              - ALG=4  : selects the Mersenne Twister random number generator;
!              - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!              - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!              - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!              - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!              - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!              - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                random number generator;
!
!              For other values, the current random number generator and its current state are not changed.
!              Note further, that, on exit, the current random number generator is not reset to its previous
!              value before the call to ID_CMP subroutine.
!
!              See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!              This optional argument has no effect if logical argument RANDOM_QR is set to false or is absent
!              in the call to ID_CMP subroutine.
!
!   BLK_SIZE   (INPUT, OPTIONAL) integer(i4b)
!              On entry, the block size used in the randomized partial QR phase of the ID algorithm
!              if RANDOM_QR = true.
!
!              BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!              smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!              By default, BLK_SIZE is set to min( BLKSZ_QR, min(m,n) ), where parameter BLKSZ_QR is the default
!              block size for QR related algorithms specified in module Select_Parameters.
!
!              See description of subroutine PARTIAL_RQR_CMP in module Random for the meaning
!              of the the block size in the randomized (partial) QR algorithm.
!
!              This optional argument has no effect if logical argument RANDOM_QR is set to false or is absent
!              in the call to ID_CMP.
!
!   NOVER      (INPUT, OPTIONAL) integer(i4b)
!              The oversampling size used in the randomized partial QR phase of the ID algorithm
!              if RANDOM_QR = true.
!
!              NOVER must be positive or null and verifies the relationship:
!
!                       NOVER + BLK_SIZE <= size(MAT,1)
!
!              and is adjusted if necessary to verify this relationship in all cases.
!
!              See description of subroutine PARTIAL_RQR_CMP in module Random for the meaning
!              and usefulness of the oversampling size in the randomized (partial) QR algorithm.
!
!              By default, the oversampling size is set to 10.
!
!              This optional argument has no effect if logical argument RANDOM_QR is set to false or is absent
!              in the call to ID_CMP.
!
!
! Further Details
! _______________
!
!   For further details on the ID decomposition and computing a column ID from a (randomized) partial
!   QR factorization with column pivoting, see:
!
!   (1) Martinsson, P.G., 2019:
!            Randomized methods for matrix computations.
!            arXiv.1607.01649
!
!   (2) Erichson, N.B., Voronin, S., Brunton, S.L., and Kutz, J.N., 2019:
!            Randomized matrix decompositions using R.
!            arXiv.1608.02148
!
!   (3) Voronin, S., Martinsson, P.G., 2015:
!            Rsvdpack: Subroutines for computing partial singular value decompositions
!            via randomized sampling on single core, multi core, and gpu architectures.
!            arXiv.1502.05366
!
!   (4) Voronin, S., Martinsson, P.G., 2017:
!            Efficient algorithms for cur and interpolative matrix decompositions
!            Adv Comput Math, Volume 43, 495-516.
!
!   (5) Xiao, J., Gu, M., and Langou, J., 2017:
!            Fast parallel randomized QR with column pivoting algorithms
!            for reliable low-rank matrix approximations.
!            IEEE 24th International Conference
!            on High Performance Computing (HiPC), IEEE, 2017, 233-242.
!
!   (6) Stewart, G.W., 1999:
!            Four algorithms for the the efficient computation of truncated pivoted qr
!            approximations to a sparse matrix.
!            Numerische Mathematik, Volume 83, 313-323.
!
!   (7) Berry, M.W., Pulatova, S.A., and Stewart, G.W., 2005:
!            Algorithm 844: Computing sparse reduced-rank approximations to sparse matrices.
!            ACM Transactions on Mathematical Software, Volume 31, No. 2, 252-269.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error, random_error14
    use Utilities,       only : assert, assert_eq, merror, arth
    use QR_Procedures,   only : partial_qr_cmp, ortho_gen_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout), dimension(:,:) :: mat
    real(stnd),  intent(out),   dimension(:,:) :: t
    real(stnd),  intent(out),   dimension(:,:), optional :: c, v
    real(stnd),  intent(out),   dimension(:),   optional :: diagr, beta
    real(stnd),  intent(out),                   optional :: rnorm
    real(stnd),  intent(inout),                 optional :: tol
!
    integer(i4b), intent(out), dimension(:) :: ip
    integer(i4b), intent(in),      optional :: rng_alg, blk_size, nover
!
    logical(lgl), intent(in),  optional :: random_qr
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(t,1))        :: diagr_tmp, beta_tmp
    real(stnd), dimension(:,:), allocatable :: qmat, rmat
!
    integer      :: iok
    integer(i4b) :: m, n, ndiag, krank, nct, i, j
!
    logical(lgl) :: comp_c, comp_v, store_diagr, store_beta, use_random_qr
!
#ifndef _BLAS
    real(stnd) :: cl
#endif
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='id_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b) ,   &
                   int(size(ip),i4b)    ,   &
                   name_proc )
!
    ndiag = min( m, n )
!
    krank = size( t, 1 )
    nct   = size( t, 2 )
!
    call assert( logical( krank<=ndiag, lgl ), &
                 logical( nct==n-krank, lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( ndiag<=0_i4b  .or. krank<=0_i4b ) then
!
        ip(:n) = arth( 1_i4b, 1_i4b, n )
!
        if ( present(rnorm) ) rnorm = zero
!
        return
!
    end if
!
!   SET OPTIONAL ARGUMENTS.
!
!   DETERMINE IF A RANDOMIZED QR ALGORITHM MUST BE USED.
!
    if ( present(random_qr) ) then
        use_random_qr = random_qr
    else
        use_random_qr = false
    end if
!
    comp_c = present( c )
!
    if ( comp_c ) then
!
        i = size( c, 1 )
        j = size( c, 2 )
!
        call assert( logical( i==m, lgl ),               &
                     logical( j==krank, lgl ), name_proc )
!
    end if
!
    comp_v = present( v )
!
    if ( comp_v ) then
!
        i = size( v, 1 )
        j = size( v, 2 )
!
        call assert( logical( i==krank, lgl ),       &
                     logical( j==n, lgl ), name_proc )
!
    end if
!
    store_diagr = present( diagr )
!
    if ( store_diagr ) then
!
        i = size( diagr )
!
        call assert( logical( i==krank, lgl ), name_proc )
!
    end if
!
    store_beta = present( beta )
!
    if ( store_beta ) then
!
        i = size( beta )
!
        call assert( logical( i==krank, lgl ), name_proc )
!
    end if
!
!   FIRST STEP, DO PARTIAL QR WITH COLUMN PIVOTING OF mat AS
!           
!                   mat * P ≈ Q1 * [  R11  R12  ]
!           
!   WHERE P IS A n-BY-n PERMUTATION MATRIX, Q1 AN m-BY-krank ORTHONORMAL
!   MATRIX, R11 A krank-BY-krank UPPER TRIANGULAR MATRIX AND R12 A
!   krank-BY-(n-krank) MATRIX.
!           
    if ( use_random_qr ) then
!           
!       USE A RANDOMIZED QR ALGORITHM.
!           
        call partial_rqr_cmp( mat(:m,:n), diagr_tmp(:krank), beta_tmp(:krank), ip(:n), i,    &
                              tol=tol, rng_alg=rng_alg, blk_size=blk_size, nover=nover )
!           
!        call partial_rqr_cmp2( mat(:m,:n), diagr_tmp(:krank), beta_tmp(:krank), ip(:n), i,   &
!                               tol=tol, rng_alg=rng_alg, blk_size=blk_size, nover=nover )
!           
    else
!           
!       USE A STANDARD QR ALGORITHM.
!           
        call partial_qr_cmp( mat(:m,:n), diagr_tmp(:krank), beta_tmp(:krank), ip(:n), i, tol=tol )
!
    end if
!
    if ( i<krank ) then
!
        call merror( name_proc//random_error14 )
!
    end if
!
!   COMPUTE SUBMATRIX t BY SOLVING THE MATRIX EQUATION (E.G., krank BY krank TRIANGULAR SYSTEMS).
!
!             R11 * t = R12
!
    t(:krank,:nct) = mat(:krank,krank+1_i4b:n)
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. nct>=2_i4b .and. i>1_i4b
#endif
!
#ifdef _BLAS
    do i = 1_i4b, krank
!
        mat(i,i) = diagr_tmp(i)
!
    end do
!
    call trsm( 'L', 'U', 'N', 'N', krank, nct, one, mat, m, t, krank )
!
    do i = 1_i4b, krank
!
        mat(i,i) = one
!
    end do
!
#else
!    
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC,1), PRIVATE(i,j,cl)
!
    do j = 1_i4b, nct
!
        do i = krank, 1_i4b, -1_i4b
!
            cl = t(i,j)/diagr_tmp(i)
            t(1_i4b:i-1_i4b,j) = t(1_i4b:i-1_i4b,j) - cl*mat(1_i4b:i-1_i4b,i)
            t(i,j) = cl
!
        end do
!
    end do
!    
!$OMP END PARALLEL DO
#endif
!
!   COMPUTE AND OUTPUT OPTIONAL ARGUMENTS.
!
!   STORE DIAGONAL OF R11 IN THE PARTIAL QR FACTORIZATION OF mat IF REQUIRED.
!
    if ( store_diagr ) then
!
       diagr(:krank) = diagr_tmp(:krank)
!
    end if
!
!   STORE SCALAR FACTORS DEFINING HOUSEHOLDER TRANSFORMATIONS IN THE
!   PARTIAL QR FACTORIZATION OF mat IF REQUIRED.
!
    if ( store_beta ) then
!
        beta(:krank) = beta_tmp(:krank)
!
    end if
!
!   COMPUTE AND STORE THE FROBENIUS NORM OF THE RESIDAL MATRIX IF REQUIRED.
!
    if ( present(rnorm) ) then
!
        rnorm = norm( mat(krank+1_i4b:m,krank+1_i4b:n) )
!
    end if
!
!   COMPUTE AND STORE c MATRIX IN THE ID OF mat IF REQUIRED.
!
    if ( comp_c ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( qmat(m,krank), rmat(krank,krank), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        qmat(:m,:krank) = mat(:m,:krank)
!
!       FIRST GENERATE FIRST krank COLUMNS OF ORTHOGONAL MATRIX Q OF QR DECOMPOSITION OF mat .
!
        call ortho_gen_qr( qmat(:m,:krank), beta_tmp(:krank) )
!
!       RESTORE THE UPPER TRIANGULAR FACTOR OF THE QR FACTORIZATION OF mat ,
!       STORING RESULTS IN rmat.
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC), PRIVATE(j)
!
        do j = 1_i4b, krank
!
            rmat(:j-1_i4b,j)      = mat(:j-1_i4b,j)
            rmat(j,j)             = diagr_tmp(j)
            rmat(j+1_i4b:krank,j) = zero
!
        end do
!
!$OMP END PARALLEL DO
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, krank, krank, one, qmat, m, rmat, &
                   krank, zero, c, m  )
#else
        c(:m,:krank) = matmul( qmat(:m,:krank), rmat(:krank,:krank) ) 
#endif
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( qmat, rmat )
!
    end if
!
!   COMPUTE AND STORE v MATRIX IN THE ID OF mat IF REQUIRED.
!
    if ( comp_v ) then
!
!       RESTORE AND APPLY TRANSPOSE PERMUTATION TO [ I  T ] FROM THE RIGHT, STORING RESULTS IN c
!           
!                   c = [ I  T ] * P'
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC), PRIVATE(j,i)
!
        do j = 1_i4b, n
!
            i = ip(j)
!
            if ( j<=krank ) then
!
                v(:krank,i) = zero
                v(j,i)      = one
!
            else
!
                v(:krank,i) = t(:krank,j-krank)
!
            end if
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!
! END OF SUBROUTINE id_cmp
! ________________________
!
    end subroutine id_cmp
!
! =========================================================================================
!
    subroutine ts_id_cmp( mat, ip_row, ip_col, w, v, skelmat, diagr, beta, rnorm, tol, &
                          random_qr, rng_alg, blk_size, nover )
!
! Purpose
! _______
!
!   TS_ID_CMP computes a (partial) two-sided Interpolative Decomposition (tsID) of a full
!   m-by-n real matrix MAT.
!
!   A tsID factorization of rank krank approximates MAT as the matrix product
!
!                   MAT ≈ W * MAT_skel * V
!
!   where W is a m-by-krank matrix, V is a krank-by-n matrix and MAT_skel consists of a squared
!   krank-by-krank submatrix of MAT, which defines the so-called skeleton of MAT.
!
!   The tsID factorization can be computed with the help of (randomized) (partial) QR factorizations with
!   column pivoting of MAT and of a matrix derived from its partial QR decomposition, more precisely with a
!   column ID of MAT and a row ID of a subset of the columns of MAT.
!
!   The first step is thus to compute a partial column ID decomposition of MAT as
!           
!                   MAT ≈ C * V
!
!   where C is a m-by-krank subset of the columns of MAT and V is a krank-by-n matrix. See description of
!   subroutine ID_CMP for more details about the ID decomposition and how such decomposition can be computed
!   from a QR decomposition with column pivoting of MAT.
!
!   In a second step, a complete column ID decomposition of C' (e.g., a row ID decomposition of C) is
!   computed as
!
!                   C' = MAT_skel' * W'
!   
!   where MAT_skel is a squared krank-by-krank matrix, which is a submatrix of MAT, and W is a m-by-krank
!   matrix. This gives the desired tsID decomposition of MAT as
!
!                    MAT ≈ W * MAT_skel * V
!
!   See the references (1) and (4) and also description of ID_CMP subroutine in module Random for more details
!   on the ID and tsID decompositions.
!
!   krank is the target rank of the tsID decomposition, which is sought, and is equal to the number of
!   rows or columns of the output array argument SKELMAT, which stores the skeleton of MAT, e.g.,
!   the matrix MAT_skel.
!
!   If the optional logical argument RANDOM_QR is used with the value true, fast randomized (partial) QR
!   factorizations with column pivoting are used for computing the column and row ID decompositions in the
!   two phases of the tsID algorithm.
!
!   
! Arguments
! _________
!
!   MAT        (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, the m-by-n matrix MAT.
!
!              On exit, MAT is overwritten by details of its partial QR factorization
!              with column pivoting.
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures or of 
!              PARTIAL_RQR_CMP subroutine in module Random for further details on how
!              the partial QR decomposition is stored compactly in MAT (and arguments IP,
!              DIAGR and BETA) on exit.
!
!              Note that, the Frobenius norm of the error associated with the computed
!              tsID is the same as the one associated with this partial QR factorization with
!              column pivoting of MAT or its column ID decomposition and is given by
!              norm( mat(krank+1:m,krank+1:n) ) on exit. See description of ID_CMP
!              subroutine for further details.
!
!   IP_ROW     (OUTPUT) integer(i4b), dimension(:) 
!              On exit, IP_ROW stores the permutation matrix N in the partial
!              LQ decomposition with row pivoting of C, which is computed
!              to obtain the row ID decomposition of C.
!
!              On exit, if IP_ROW(j)=k, then the j-th row of N*C was 
!              the k-th row of C, where N = I(IP_ROW,:) and I is the identity
!              matrix of order m.
!
!              The matrix MAT_skel in the tsID decomposition of MAT is the submatrix defined as
!              the intersection of the rows IP_ROW(:krank) and the columns IP_COL(:krank) of MAT.
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures
!              and PARTIAL_RQR_CMP and ID_CMP subroutines in module Random for further details.
!
!              The size of IP_ROW must be equal to size( MAT, 1 ) = m.
!
!   IP_COL     (OUTPUT) integer(i4b), dimension(:) 
!              On exit, IP_COL stores the permutation matrix P in the partial
!              QR decomposition with column pivoting of MAT, which is computed
!              to obtain the column ID decomposition of MAT.
!
!              If IP_COL(j)=k, then the j-th column of MAT*P was the k-th column
!              of MAT, where P = I(:,IP_COL) and I is the identity matrix of order n.
!
!              The matrix C in the (column) ID of MAT corresponds to the subset of the columns
!              of MAT with the indices IP_COL(:krank). Furthermore, the matrix MAT_skel in the tsID
!              decomposition of MAT is the submatrix defined as the intersection of the rows
!              IP_ROW(:krank) and the columns IP_COL(:krank) of MAT.
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures
!              and PARTIAL_RQR_CMP and ID_CMP subroutines in module Random for further details.
!
!              The size of IP_COL must be equal to size( MAT, 2 ) = n.
!
!   W          (OUTPUT) real(stnd), dimension(:,:)
!              On exit, the m-by-krank matrix W in the tsID factorization of MAT.
!
!              The shape of W must verify:
!
!              - size( W, 1 ) = size( MAT, 1 )     = m
!              - size( W, 2 ) = size( SKELMAT, 1 ) = krank
!
!   V          (OUTPUT) real(stnd), dimension(:,:)
!              On exit, the krank-by-n matrix V in the tsID factorization of MAT.
!
!              The shape of V must verify:
!
!              - size( V, 1 ) = size( SKELMAT, 1 ) = krank
!              - size( V, 2 ) = size( MAT, 2 )     = n
!
!   SKELMAT    (OUTPUT) real(stnd), dimension(:,:)
!              On exit, the squared krank-by-krank matrix MAT_skel (e.g., the skeleton of MAT)
!              in the tsID factorization of MAT.
!
!              MAT_skel is the submatrix of MAT defines by the rows IP_ROW(:krank) and
!              the columns IP_COL(:krank) of MAT on entry. However, in the routine, MAT_skel
!              is recomputed from the column ID of MAT and row ID of C to save space as MAT
!              is overwritten by details of its partial QR factorization when computing the
!              column ID of MAT.
!
!              The shape of SKELMAT must verify:
!
!              - size( SKELMAT, 1 ) = size( SKELMAT, 2 ) = krank <= min(m,n)
!
!   RNORM      (OUTPUT, OPTIONAL) real(stnd)
!              On exit, the Frobenius norm of the error matrix associated with the tsID,
!              which is defined as || MAT - W * MAT_skel * V ||_F . Note that this error matrix
!              is the same as the one associated with the partial QR decomposition with column
!              pivoting of MAT or its column ID decomposition.
!
!   DIAGR      (OUTPUT, OPTIONAL) real(stnd), dimension(:) 
!              On exit, the diagonal elements of the matrix R11 in the partial QR
!              factorization with column pivoting of MAT.
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures
!              and PARTIAL_RQR_CMP subroutine in module Random for further details.
!
!              The size of DIAGR must verify:
!
!              - size( DIAGR ) = krank <= min( size(MAT,1) , size(MAT,2) ) = min(m,n)
!
!   BETA       (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!              On exit, the scalars factors of the elementary reflectors defining Q
!              in the partial QR factorization with column pivoting of MAT.
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures
!              and PARTIAL_RQR_CMP subroutine in module Random for further details.
!
!              The size of BETA must verify:
!
!              - size( BETA ) = krank <= min( size(MAT,1) , size(MAT,2) ) = min(m,n)
!
!   TOL        (INPUT/OUTPUT, OPTIONAL) real(stnd)
!              On entry, if TOL is present  then:
!
!              - if TOL is in ]0,1[, calculations to determine the condition number of
!                submatrix R11 in the partial QR factorization of MAT are performed.
!                TOL is used to determine the effective pseudo-rank of R11,
!                which is defined as the order of the largest leading triangular
!                submatrix in the partial QR factorization with column pivoting of MAT,
!                whose estimated condition number in the 1-norm is less than 1/TOL.
!                On exit, the reciprocal of the condition number is returned in TOL.
!              - if TOL=0 is specified, the calculations to determine the condition number
!                of R11 are not performed and crude tests on R(j,j) are done to determine
!                the numerical pseudo-rank of R11. On exit, TOL is not changed.
!
!              The TOL argument is useful to check if the matrix R11 in the QR decomposition of MAT,
!              which must be inverted to compute T (in the column ID of MAT), is sufficiently well
!              conditioned to obtain a stable and robust column ID of MAT and, consequently, a robust
!              tsID decomposition.
!
!              If the computed pseudo-rank of R11 is less than krank = size( SKELMAT, 1 ),
!              the subroutine will exit with an error message.
!
!              On the other hand, if TOL is not specified or is outside [0,1[, the calculations
!              to determine the rank of R11 are not performed and this rank is assumed to be equal
!              to krank = size( SKELMAT, 1 ).
!
!              If it is possible that MAT may not be of full rank then the stability of the tsID
!              decomposition of MAT, which is sought, can be checked by using TOL=relative precision
!              of the elements in MAT. If each element of MAT is correct to, say, 5 digits then TOL=0.00001
!              should be used.
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures
!              and PARTIAL_RQR_CMP subroutine in module Random for further details.
!
!   RANDOM_QR  (INPUT, OPTIONAL) logical(lgl)
!              On entry, if RANDOM_QR is used with the value true, fast randomized (partial) QR
!              factorizations are used in the two phases of the tsID algorithm.
!
!              By default, RANDOM_QR = false, i.e., a standard (partial) QR factorization with
!              column pivoting is used in the two phases of the tsID algorithm.
!
!   RNG_ALG    (INPUT, OPTIONAL) integer(i4b)
!              On entry, a scalar integer to select the random (uniform) number generator used to build
!              the random gaussian matrix in the two randomized (partial) QR phases of the tsID algorithm
!              if RANDOM_QR = true.
!
!              The possible values are:
!
!              - ALG=1  : selects the Marsaglia's KISS random number generator;
!              - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!              - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!              - ALG=4  : selects the Mersenne Twister random number generator;
!              - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!              - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!              - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!              - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!              - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!              - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                random number generator;
!
!              For other values, the current random number generator and its current state are not changed.
!              Note further, that, on exit, the current random number generator is not reset to its previous
!              value before the call to TS_ID_CMP subroutine.
!
!              See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!              This optional argument has no effect if logical argument RANDOM_QR is set to false or is absent
!              in the call to TS_ID_CMP subroutine.
!
!   BLK_SIZE   (INPUT, OPTIONAL) integer(i4b)
!              On entry, the block size used in the randomized partial QR phases of the tsID algorithm
!              if RANDOM_QR = true.
!
!              BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!              smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!              By default, BLK_SIZE is set to min( BLKSZ_QR, min(m,n) ), where parameter BLKSZ_QR is the default
!              block size for QR related algorithms specified in module Select_Parameters.
!
!              See description of subroutine PARTIAL_RQR_CMP in module Random for the meaning
!              of the the block size in the randomized (partial) QR algorithm.
!
!              This optional argument has no effect if logical argument RANDOM_QR is set to false or is absent
!              in the call to TS_ID_CMP subroutine.
!
!   NOVER      (INPUT, OPTIONAL) integer(i4b)
!              The oversampling size used in the randomized partial QR phases of the tsID algorithm
!              if RANDOM_QR = true.
!
!              NOVER must be positive or null and verifies the relationship:
!
!                       NOVER + BLK_SIZE <= size(MAT,1)
!
!              and is adjusted if necessary to verify this relationship in all cases.
!
!              See description of subroutine PARTIAL_RQR_CMP in module Random for the meaning
!              and usefulness of the oversampling size in the randomized (partial) QR algorithm.
!
!              By default, the oversampling size is set to 10.
!
!              This optional argument has no effect if logical argument RANDOM_QR is set to false or is absent
!              in the call to TS_ID_CMP subroutine.
!
!
! Further Details
! _______________
!
!   For further details on the tsID and computing the tsID decomposition from the (randomized)
!   partial QR factorizations with column pivoting of a matrix, see:
!
!   (1) Martinsson, P.G., 2019:
!            Randomized methods for matrix computations.
!            arXiv.1607.01649
!
!   (2) Erichson, N.B., Voronin, S., Brunton, S.L., and Kutz, J.N., 2019:
!            Randomized matrix decompositions using R.
!            arXiv.1608.02148
!
!   (3) Voronin, S., Martinsson, P.G., 2015:
!            Rsvdpack: Subroutines for computing partial singular value decompositions
!            via randomized sampling on single core, multi core, and gpu architectures.
!            arXiv.1502.05366
!
!   (4) Voronin, S., Martinsson, P.G., 2017:
!            Efficient algorithms for cur and interpolative matrix decompositions
!            Adv Comput Math, volume 43, 495-516.
!
!   (5) Stewart, G.W., 1999:
!            Four algorithms for the the efficient computation of truncated pivoted qr
!            approximations to a sparse matrix.
!            Numerische Mathematik, Volume 83, 313-323.
!
!   (6) Berry, M.W., Pulatova, S.A., and Stewart, G.W., 2005:
!            Algorithm 844: Computing sparse reduced-rank approximations to sparse matrices.
!            ACM Transactions on Mathematical Software, Volume 31, No. 2, 252-269.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, assert_eq, merror, arth
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout), dimension(:,:) :: mat
    real(stnd),  intent(out),   dimension(:,:) :: w, v, skelmat
    real(stnd),  intent(out),   dimension(:),   optional :: diagr, beta
    real(stnd),  intent(out),                   optional :: rnorm
    real(stnd),  intent(inout),                 optional :: tol
!
    integer(i4b), intent(out), dimension(:) :: ip_row, ip_col
    integer(i4b), intent(in),      optional :: rng_alg, blk_size, nover
!
    logical(lgl), intent(in),  optional :: random_qr
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(:,:), allocatable :: t, ct, wt, skelt
!
    integer      :: iok
    integer(i4b) :: m, n, ndiag, krank, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='ts_id_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size(mat,1),i4b),    &
                   int(size(w,1),i4b),      &
                   int(size(ip_row),i4b),   &
                   name_proc )
!
    n = assert_eq( int(size(mat,2),i4b) ,   &
                   int(size(v,2),i4b),      &
                   int(size(ip_col),i4b),   &
                   name_proc )
!
    ndiag = min( m, n )
!
    krank = assert_eq( int(size(skelmat,1),i4b), &
                       int(size(skelmat,2),i4b), &
                       int(size(w,2),i4b),       &
                       int(size(v,1),i4b),       &
                       name_proc )
!
!
    call assert( logical( krank<=ndiag, lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( ndiag<=0_i4b  .or. krank<=0_i4b ) then
!
        ip_row(:m) = arth( 1_i4b, 1_i4b, m )
        ip_col(:n) = arth( 1_i4b, 1_i4b, n )
!
        if ( present(rnorm) ) rnorm = zero
!
        return
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    i = max( m, n ) - krank
!
    allocate( t(krank,i), ct(krank,m), wt(krank,m), skelt(krank,krank), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   FIRST STEP, COMPUTE A PARTIAL COLUMN ID DECOMPOSITION OF mat AS
!           
!                   mat = C * V
!           
!   WHERE C IS A m-BY-krank SUBSET OF THE COLUMS OF mat AND V IS A krank-BY-n MATRIX.
!           
    call id_cmp( mat(:m,:n), ip_col(:n), t(:krank,:n-krank), c=w, v=v, diagr=diagr,      &
                 beta=beta, rnorm=rnorm, tol=tol, random_qr=random_qr, rng_alg=rng_alg,  &
                 blk_size=blk_size, nover=nover )
!           
!   TRANSPOSE C.
!           
    ct(:krank,:m) = transpose( w(:m,:krank) )
!
!   SECOND STEP, COMPUTE A COMPLETE COLUMN ID DECOMPOSITION OF C' AS
!           
!                   C' = MAT_skel' * W'
!           
!   WHERE MAT_skel IS A krank-BY-krank MATRIX, WHICH IS A SQUARED SUBMATRIX OF mat,
!   AND W IS A m-by-krank MATRIX.
!           
    call id_cmp( ct(:krank,:m), ip_row(:m), t(:krank,:m-krank), c=skelt, v=wt,        &
                 random_qr=random_qr, rng_alg=rng_alg, blk_size=blk_size, nover=nover )
!           
!   FINALLY, TRANSPOSE RESULTS TO GET THE DESIRED FACTORS IN THE tsID OF mat .
!           
    skelmat(:krank,:krank) = transpose( skelt(:krank,:krank) )
!           
    w(:m,:krank) = transpose( wt(:krank,:m) )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( t, ct, wt, skelt )
!
!
! END OF SUBROUTINE ts_id_cmp
! ___________________________
!
    end subroutine ts_id_cmp
!
! =========================================================================================
!
    subroutine cur_cmp( mat, ip_row, ip_col, u, c, r, rnorm_row, rnorm_col, tol, &
                        random_qr, rng_alg, blk_size, nover )
!
! Purpose
! _______
!
!   CUR_CMP computes a (partial) CUR decomposition of a m-by-n real matrix MAT. A CUR
!   decomposition provides a reduced-rank approximate decomposition of a full m-by-n real matrix
!   MAT of the form:
!
!                    MAT ≈ C * U * R
!
!   where C and R are m-by-krank and krank-by-n matrices, which are, respectively, subsets of the
!   columns and rows of MAT, and U is a krank-by-krank matrix, which is estimated to make the matrix
!   product C * U * R a good approximation of MAT according to the Frobenius norm. The CUR factorization
!   is an important tool for handling large-scale data sets, offering several advantages over the
!   Singular Value Decomposition (SVD): the columns and rows that comprise C and R are representative
!   of the data and they are sparse if MAT is sparse. See references (1) and (2) for a discussion.
!
!   Computing an approximate CUR decomposition is generally a three steps process.
!
!   The C and R submatrices in the CUR factorization can be first estimated with the help of (randomized)
!   partial QR factorizations with column pivoting of MAT and MAT', respectively.
!
!   The initial step is thus to compute a (partial) QR factorization with column pivoting of MAT, which is
!   defined as
!           
!                     MAT * P = Q * T
!
!   , where P is an n-by-n permutation matrix, T is an upper triangular or trapezoidal
!   (i.e., if n>m) matrix and Q is a m-by-m orthogonal matrix if the QR factorization is complete.
!
!   For computing a CUR decomposition of rank, krank, a partial QR factorization with column pivoting
!   of MAT of rank, krank, is sufficient, e.g., the QR decomposition can be stopped when the numbers
!   of columns of Q is equal to krank .
!
!   This leads implicitly to the following partition of Q:
!
!                     Q = [  Q1  Q2  ]
!
!   where Q1 is an m-by-krank orthonormal matrix and Q2 is m-by-(m-krank) orthonormal matrix
!   orthogonal to Q1, and to the following corresponding partition of T:
!
!   [  T11  T12  ]
!
!   [  T21  T22  ]
!
!   where T11 is a krank-by-krank triangular matrix, T21 is zero by construction,
!   T12 is a full krank-by-(n-krank) matrix and T22 is a full (m-krank)-by-(n-krank)
!   matrix. This leads to the following (partial QR) approximation of MAT:
!
!                     MAT ≈ Q1 * [  T11  T12  ] * P'
!
!   Here, we see that Q1 * T11 equals the first krank columns of A * P, and we can define the matrix
!   C in the CUR factorization of MAT as
!
!                     C = Q1 * T11 = MAT * P(:,:krank)
!
!   This choice leads to a subset of the columns of MAT, which will minimize the Frobenius norm,
!   RNORM_COL, of the error associated with the reduced-rank QR approximation or column ID of MAT and,
!   also, of the resulting CUR approximation of MAT as we will illustrate below.
!
!   Furthermore, remember that C is also the selected subset of the columns of MAT, which defines the
!   left factor in the column ID of MAT of rank krank:
!
!                     MAT ≈ (Q1 * T11) * V = C * V
!
!   where V = [ I  Z ] * P' is a krank-by-n matrix, I is a krank-by-krank identity matrix and
!   Z is a krank-by-(n-krank) matrix, which is the solution to the matrix equation:
!
!                     T11 * Z = T12   ,e.g.,  Z = inv(T11) * T12
!
!   see description of ID_CMP subroutine for details.
!
!   In a second step, the same partial QR algorithm is applied to MAT' to get a representative subset of
!   the rows of MAT, R, such that
!
!                     R = L11 * K1 = N(:krank,:) * MAT
!
!   where R is a krank-by-n matrix, which is a subset of the rows of MAT, L11 is a krank-by-krank lower
!   triangular matrix, K1 is a krank-by-n matrix with orthonormal rows and N is a m-by-m permutation matrix.
!   Again, this choice leads to a subset of the rows of MAT, R, which will minimize the Frobenius norm, RNORM_ROW,
!   of the error associated with the reduced-rank LQ approximation of MAT (and also of the resulting CUR approximation
!   of MAT).
!
!   In a final step, we then seek a krank-by-krank matrix U such that 
!
!                     || MAT - C * U * R ||_F = min
!
!   Once C and U are fixed, it can be shown that the minimum is attained for a matrix U, which is defined as
!
!                     U = pseudo-inv(C) * MAT * pseudo-inv(R) = inv(T11) * Q1' * MAT * K1' * inv( L11 )
!
!   where pseudo-inv(C) is the pseudo-inverse of C, which is equal to inv(T11) * Q1' if T11 is of full rank and with
!   similar results for the matrix pseudo-inv(R). See reference (5) for details.
!
!   However, such direct computation of U involves the explicit inversion of two triangular matrices,
!   which can be severely affected by ill-conditioning of the matrices T11 and L11. Using the column ID of MAT,
!   we have
!
!                     Q1' * MAT = T11 * V
!
!   , we deduce that
!
!                     U = inv(T11) * Q1' * MAT * K1' * inv( L11 ) = V * K1' * inv( L11 ) = V * pseudo-inv(R)
!
!   and we observe that we can also determine U as the solution of the least squares problem:
!
!                     U * R ≈ V
!
!   Such a least squares problem must have an accurate solution as both the rows of R and V should span roughly the same space,
!   namely, the space spanned by the krank leading right singular vectors of MAT (see reference (4) for discussion). If it is
!   the case, the Frobenius norm of the error associated with the resulting CUR decomposition will be about of the same order
!   of the error associated with the column ID of MAT, only slightly larger.
!
!   Using this approach, the matrix U can thus be estimated by solving the linear least squares problem:
!
!                     U * R ≈ V
!
!   and this computation never involves the explicit inversion of the matrices T11 and L11, but
!   rather solving two linear least squares problems associated with these two triangular matrices.
!
!   Moreover, we have:
!
!                  || MAT - C * U * R ||_F <= sqrt( RNORM_COL**2 + RNORM_ROW**2 )
!
!   See references (5) and (6) for further details.
!
!   Note that for matrices whose singular values experience a fast decay, the accuracy of CUR
!   factorization, as computed by CUR_CMP, can deteriorate due to ill-conditioning and the need
!   to solve linear (least squares) problems associated with the triangular matrices T11 and L11
!   in order to estimate the matrix U. See reference (4) for a more detailed discussion of this problem.
!
!   The condition numbers (in the 1-norm) and the ranks of the matrices T11 and L11 are
!   estimated if the optional argument TOL is present in order to check the robustness of the
!   computed CUR decomposition.
!
!   krank is the target rank of the partial CUR decomposition, which is sought, and is equal to the
!   number of rows or columns of the output array argument U, which stores the factor U in the CUR of MAT.
!
!   If the optional logical argument RANDOM_QR is used with the value true, fast randomized (partial) QR
!   factorizations with column pivoting are used for computing the partial column QR and row LQ decompositions of
!   MAT in the first two phases of the CUR algorithm.
!
!   
! Arguments
! _________
!
!   MAT        (INPUT) real(stnd), dimension(:,:)
!              On entry, the m-by-n matrix MAT.
!
!   IP_ROW     (OUTPUT) integer(i4b), dimension(:) 
!              On exit, IP_ROW stores the permutation matrix N in the partial
!              QR decomposition with column pivoting of MAT', which is computed
!              to obtain the row LQ decomposition of MAT.
!
!              On exit, if IP_ROW(j)=k, then the j-th row of N*MAT was 
!              the k-th row of MAT, where N = I(IP_ROW,:) and I is the identity
!              matrix of order m.
!
!              The matrix R in the CUR of MAT corresponds to the subset of the rows
!              of MAT with the indices IP_ROW(:krank).
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures
!              and PARTIAL_RQR_CMP subroutine in module Random for further details.
!
!              The size of IP_ROW must be equal to size( MAT, 1 ) = m.
!
!   IP_COL     (OUTPUT) integer(i4b), dimension(:) 
!              On exit, IP_COL stores the permutation matrix P in the partial
!              QR decomposition with column pivoting of MAT, which is computed
!              to obtain the column QR decomposition of MAT.
!
!              If IP_COL(j)=k, then the j-th column of MAT*P was the k-th column
!              of MAT, where P = I(:,IP_COL) and I is the identity matrix of order n.
!
!              The matrix C in the CUR of MAT corresponds to the subset of the columns
!              of MAT with the indices IP_COL(:krank).
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures
!              and PARTIAL_RQR_CMP subroutine in module Random for further details.
!
!              The size of IP_COL must be equal to size( MAT, 2 ) = n.
!
!   U          (OUTPUT) real(stnd), dimension(:,:)
!              On exit, the krank-by-krank matrix U in the approximate CUR factorization of MAT.
!
!              The shape of U must verify:
!
!              - size( U, 1 ) = size( U, 2 ) = krank <= min( size(MAT,1) , size(MAT,2) ) = min(m,n)
!
!   C          (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!              On exit, the m-by-krank matrix C in the approximate CUR factorization of MAT.
!
!              C consists of the subset of the columns of MAT defined by the indices IP_COL(:krank).
!
!              The shape of C must verify:
!
!              - size( C, 1 ) = size( MAT, 1 ) = m
!              - size( C, 2 ) = size( U, 2 )   = krank
!
!   R          (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!              On exit, the krank-by-n matrix R in the approximate CUR factorization of MAT.
!
!              R consists of the subset of the rows of MAT defined by the indices IP_ROW(:krank).
!
!              The shape of R must verify:
!
!              - size( R, 1 ) = size( U, 1 )   = krank
!              - size( R, 2 ) = size( MAT, 2 ) = n
!
!   RNORM_ROW  (OUTPUT, OPTIONAL) real(stnd)
!              On exit, the Frobenius norm of the error matrix associated with the row
!              partial LQ decomposition of MAT (e.g., the QR decomposition of MAT'),
!              computed as || L22 ||_F .
!
!   RNORM_COL  (OUTPUT, OPTIONAL) real(stnd)
!              On exit, the Frobenius norm of the error matrix associated with the column
!              partial QR decomposition of MAT, computed as || T22 ||_F .
!
!   TOL        (INPUT, OPTIONAL) real(stnd)
!              If TOL is present and is in [0,1[, then calculations to determine the condition number of
!              submatrices T11 and L11 in the partial QR factorizations of MAT and MAT' are performed.
!
!              TOL is used to determine the effective pseudo-ranks of T11 and L11, which are defined as the
!              order of the largest leading triangular submatrices in the partial QR factorizations with column
!              pivoting of MAT and MAT' whose estimated condition numbers in the 1-norm are less than 1/TOL.
!              If TOL=0 is specified, the calculations to determine the condition numbers of T11 and L11
!              are not performed and crude tests on T(j,j) and L(j,j) are done to determine the numerical
!              pseudo-ranks of T11 and L11.
!
!              The TOL argument is useful to check if the matrices T11 and L11 are sufficiently well
!              conditioned to obtain a stable and robust CUR decomposition.
!
!              If the computed pseudo-ranks of T11 or L11 are less than krank = size( U, 1 ),
!              the subroutine will exit with an error message.
!
!              On the other hand, if TOL is not specified or is outside [0,1[, the calculations
!              to determine the ranks of T11 and L11 are not performed and these ranks are assumed
!              to be equal to krank = size( U, 1 ).
!
!              If it is possible that MAT, T11 and L11 may not be of full rank, then the stability of
!              the CUR decomposition of MAT, which is sought, can be checked by using TOL=relative precision
!              of the elements in MAT. If each element of MAT is correct to, say, 5 digits then TOL=0.00001
!              should be used.
!
!              See description of PARTIAL_QR_CMP subroutine in module QR_Procedures
!              and PARTIAL_RQR_CMP subroutine in module Random for further details.
!
!   RANDOM_QR  (INPUT, OPTIONAL) logical(lgl)
!              On entry, if RANDOM_QR is used with the value true, fast randomized (partial) QR
!              factorizations are used in the first two phases of the CUR algorithm.
!
!              By default, RANDOM_QR = false, i.e., a standard (partial) QR factorization with
!              column pivoting is used in the first two phases of the CUR algorithm.
!
!   RNG_ALG    (INPUT, OPTIONAL) integer(i4b)
!              On entry, a scalar integer to select the random (uniform) number generator used to build
!              the random gaussian matrix in the two randomized (partial) QR phases of the CUR algorithm
!              if RANDOM_QR = true.
!
!              The possible values are:
!
!              - ALG=1  : selects the Marsaglia's KISS random number generator;
!              - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!              - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!              - ALG=4  : selects the Mersenne Twister random number generator;
!              - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!              - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!              - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!              - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!              - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!              - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                random number generator;
!
!              For other values, the current random number generator and its current state are not changed.
!              Note further, that, on exit, the current random number generator is not reset to its previous
!              value before the call to CUR_CMP subroutine.
!
!              See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!              This optional argument has no effect if logical argument RANDOM_QR is set to false or is absent
!              in the call to CUR_CMP subroutine.
!
!   BLK_SIZE   (INPUT, OPTIONAL) integer(i4b)
!              On entry, the block size used in the randomized partial QR phases of the CUR algorithm
!              if RANDOM_QR = true.
!
!              BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!              smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!              By default, BLK_SIZE is set to min( BLKSZ_QR, min(m,n) ), where parameter BLKSZ_QR is the default
!              block size for QR related algorithms specified in module Select_Parameters.
!
!              See description of subroutine PARTIAL_RQR_CMP in module Random for the meaning
!              of the the block size in the randomized (partial) QR algorithm.
!
!              This optional argument has no effect if logical argument RANDOM_QR is set to false or is absent
!              in the call to CUR_CMP subroutine.
!
!   NOVER      (INPUT, OPTIONAL) integer(i4b)
!              The oversampling size used in the randomized partial QR phases of the CUR algorithm
!              if RANDOM_QR = true.
!
!              NOVER must be positive or null and verifies the relationship:
!
!                       NOVER + BLK_SIZE <= size(MAT,1)
!
!              and is adjusted if necessary to verify this relationship in all cases.
!
!              See description of subroutine PARTIAL_RQR_CMP in module Random for the meaning
!              and usefulness of the oversampling size in the randomized (partial) QR algorithm.
!
!              By default, the oversampling size is set to 10.
!
!              This optional argument has no effect if logical argument RANDOM_QR is set to false or is absent
!              in the call to CUR_CMP subroutine.
!
!
! Further Details
! _______________
!
!   For further details on the CUR and computing the CUR decomposition from (randomized)
!   partial QR factorizations with column pivoting of a matrix and its transpose, see:
!
!   (1) Mahoney, M.W., and Drineas, P., 2009:
!            CUR matrix decompositions for improved data analysis.
!            PNAS, Volume 106, No. 3, 697-702.
!
!   (2) Martinsson, P.G., 2019:
!            Randomized methods for matrix computations.
!            arXiv.1607.01649
!
!   (3) Voronin, S., Martinsson, P.G., 2015:
!            Rsvdpack: Subroutines for computing partial singular value decompositions
!            via randomized sampling on single core, multi core, and gpu architectures.
!            arXiv.1502.05366
!
!   (4) Voronin, S., Martinsson, P.G., 2017:
!            Efficient algorithms for cur and interpolative matrix decompositions
!            Adv Comput Math, Volume 43, 495-516.
!
!   (5) Stewart, G.W., 1999:
!            Four algorithms for the the efficient computation of truncated pivoted qr
!            approximations to a sparse matrix.
!            Numerische Mathematik, Volume 83, 313-323.
!
!   (6) Berry, M.W., Pulatova, S.A., and Stewart, G.W., 2005:
!            Algorithm 844: Computing sparse reduced-rank approximations to sparse matrices.
!            ACM Transactions on Mathematical Software, Volume 31, No. 2, 252-269.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error, random_error15
    use Utilities,       only : assert, assert_eq, merror, arth
    use QR_Procedures,   only : partial_qr_cmp, apply_q_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:,:)           :: mat
    real(stnd),  intent(out), dimension(:,:)           :: u
    real(stnd),  intent(out), dimension(:,:), optional :: c, r
    real(stnd),  intent(out),                 optional :: rnorm_row, rnorm_col
    real(stnd),  intent(inout),               optional :: tol
!
    integer(i4b), intent(out), dimension(:) :: ip_row, ip_col
    integer(i4b), intent(in),      optional :: rng_alg, blk_size, nover
!
    logical(lgl), intent(in),  optional :: random_qr
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tol2, cl
    real(stnd), dimension(size(u,1))        :: diagl, beta
    real(stnd), dimension(:,:), allocatable :: tmp, v, vt
!
    integer      :: iok
    integer(i4b) :: m, n, ndiag, krank, i, j
!
    logical(lgl) :: comp_c, comp_r, use_random_qr
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='cur_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size(mat,1),i4b),    &
                   int(size(ip_row),i4b),   &
                   name_proc )
!
    n = assert_eq( int(size(mat,2),i4b) ,   &
                   int(size(ip_col),i4b),   &
                   name_proc )
!
    ndiag = min( m, n )
!
    krank = assert_eq( int(size(u,1),i4b), &
                       int(size(u,2),i4b), &
                       name_proc )
!
    call assert( logical( krank<=ndiag, lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( ndiag<=0_i4b  .or. krank<=0_i4b ) then
!
        ip_row(:m) = arth( 1_i4b, 1_i4b, m )
        ip_col(:n) = arth( 1_i4b, 1_i4b, n )
!
        if ( present(rnorm_row) ) rnorm_row = zero
        if ( present(rnorm_col) ) rnorm_col = zero
!
        return
!
    end if
!
!   SET OPTIONAL ARGUMENTS.
!    
!   TEST IF TOLERANCE IS SPECIFIED.
!    
    if ( present(tol) ) then
        tol2 = tol
    else
        tol2 = -one
    end if
!
!   DETERMINE IF A RANDOMIZED QR ALGORITHM MUST BE USED.
!
    if ( present(random_qr) ) then
        use_random_qr = random_qr
    else
        use_random_qr = false
    end if
!
    comp_c = present( c )
!
    if ( comp_c ) then
!
        i = size( c, 1 )
        j = size( c, 2 )
!
        call assert( logical( i==m, lgl ),               &
                     logical( j==krank, lgl ), name_proc )
!
    end if
!
    comp_r = present( r )
!
    if ( comp_r ) then
!
        i = size( r, 1 )
        j = size( r, 2 )
!
        call assert( logical( i==krank, lgl ),       &
                     logical( j==n, lgl ), name_proc )
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    i = n - krank
!
    allocate( tmp(m,n), vt(krank,i), v(krank,n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   MAKE A COPY OF mat .
!
    tmp(:m,:n) = mat(:m,:n)
!
    cl = tol2
!
!   FIRST STEP, COMPUTE A PARTIAL COLUMN ID DECOMPOSITION OF mat AS
!           
!                   mat = C * V
!           
!   WHERE C IS A m-BY-krank SUBSET OF THE COLUMS OF mat AND V IS A krank-BY-n MATRIX.
!           
    call id_cmp( tmp(:m,:n), ip_col(:n), vt(:krank,:i), v=v, tol=cl, random_qr=random_qr, &
                 rng_alg=rng_alg, blk_size=blk_size, nover=nover )
!
!   COMPUTE AND STORE THE FROBENIUS NORM OF THE RESIDAL MATRIX IF REQUIRED.
!
    if ( present(rnorm_col) ) then
!
        rnorm_col = norm( tmp(krank+1_i4b:m,krank+1_i4b:n) )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!           
    deallocate( tmp, vt )
!
!   REALLOCATE WORK ARRAYS.
!
    allocate( tmp(n,m), vt(n,krank), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   MAKE A TRANSPOSE COPY OF mat .
!
    tmp(:n,:m) = transpose( mat(:m,:n) )
!
!   MAKE A TRANSPOSE COPY OF v .
!
    vt(:n,:krank) = transpose( v(:krank,:n) )
!
    cl = tol2
!
!   NEXT STEP, DO PARTIAL QR WITH COLUMN PIVOTING OF mat' (e.g., A PARTIAL LQ
!   DECOMPOSITION WITH ROW PIVOTING OF MAT) AS
!           
!                   mat' * N' ≈ K1' * [  L11'  L21' ]
!           
!   WHERE N IS A m-BY-m PERMUTATION MATRIX, K1 A krank-BY-n MATRIX WITH
!   ORTHONORMAL ROWS, L11 A krank-BY-krank LOWER TRIANGULAR MATRIX AND L21 A
!   (m-krank)-BY-krank MATRIX.
!           
    if ( use_random_qr ) then
!           
!       USE A RANDOMIZED QR ALGORITHM.
!           
        call partial_rqr_cmp( tmp(:n,:m), diagl(:krank), beta(:krank), ip_row(:m), i,   &
                              tol=cl, rng_alg=rng_alg, blk_size=blk_size, nover=nover   )
!           
!        call partial_rqr_cmp2( tmp(:n,:m), diagl(:krank), beta(:krank), ip_row(:m), i,  &
!                               tol=cl, rng_alg=rng_alg, blk_size=blk_size, nover=nover  )
!           
    else
!           
!       USE A STANDARD QR ALGORITHM.
!           
        call partial_qr_cmp( tmp(:n,:m), diagl(:krank), beta(:krank), ip_row(:m), i, tol=cl )
!
    end if
!
    if ( i<krank ) then
!
       call merror( name_proc//random_error15 )
!
    end if
!
!   COMPUTE AND STORE THE FROBENIUS NORM OF THE RESIDAL MATRIX IF REQUIRED.
!
    if ( present(rnorm_row) ) then
!
        rnorm_row = norm( tmp(krank+1_i4b:n,krank+1_i4b:m) )
!
    end if
!
!   COMPUTE MATRIX U BY SOLVING THE LINEAR LEAST SQUARES PROBLEM U * ( L11 * K1 ) ≈ V FOR U.
!
!   APPLY THE HOUSEHOLDER TRANSFORMATIONS DEFINING K1 IN PARTIAL QR WITH COLUMN PIVOTING OF mat' TO V' .
!
    call apply_q_qr( tmp(:n,:krank), beta(:krank), vt(:n,:krank), left=true, trans=true )
!
!   SOLVE THE MATRIX EQUATION L11' * U' = K1 * V' FOR U WHERE L11' IS UPPER TRIANGULAR.
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. krank>=2_i4b .and. i>1_i4b
#endif
!
#ifdef _BLAS
    do i = 1_i4b, krank
!
        tmp(i,i) = diagl(i)
!
    end do
!
    call trsm( 'L', 'U', 'N', 'N', krank, krank, one, tmp, n, vt, n )
!
#else
!    
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC,1), PRIVATE(i,j,cl)
!
    do j = 1_i4b, krank
!
        do i = krank, 1_i4b, -1_i4b
!
            cl = vt(i,j)/diagl(i)
            vt(1_i4b:i-1_i4b,j) = vt(1_i4b:i-1_i4b,j) - cl*tmp(1_i4b:i-1_i4b,i)
            vt(i,j) = cl
!
        end do
!
    end do
!    
!$OMP END PARALLEL DO
#endif
!           
!   TRANSPOSE RESULTS TO GET THE DESIRED U MATRIX IN THE CUR OF mat .
!    
    u(:krank,:krank) = transpose( vt(:krank,:krank) )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( tmp, v, vt )
!
!   COMPUTE AND STORE C MATRIX IN THE CUR DECOMPOSITION OF mat IF REQUIRED.
!
    if ( comp_c ) then
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC), PRIVATE(i,j)
!
        do j = 1_i4b, krank
!
            i = ip_col(j)
!
            c(:m,j) = mat(:m,i)
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!   COMPUTE AND STORE R MATRIX IN THE CUR DECOMPOSITION OF mat IF REQUIRED.
!
    if ( comp_r ) then
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC), PRIVATE(j)
!
        do j = 1_i4b, n
!
            r(:krank,j) = mat(ip_row(:krank),j)
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!
! END OF SUBROUTINE cur_cmp
! _________________________
!
    end subroutine cur_cmp
!
!
! =========================================================================================
!                                   BOOTSTRAP AND PERMUTATION PROCEDURES
! =========================================================================================
!
!
    subroutine simple_shuffle_rv( vec )
!
! purpose
! _______
!
!   This subroutine shuffles all the elements of the real vector VEC.
!
!
! Arguments
! _________
!
!   VEC      (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, the real vector to be shuffled.
!
!            On exit, the permuted real vector.
!
!
! Further Details
! _______________
!
!   For more details and algorithm, see:
!
!   (1) Noreen, E.W., 1989:
!           Computer-intensive methods for testing hypotheses: an introduction.
!           Wiley and Sons, New York, USA, ISBN:978-0-471-61136-3
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:) :: vec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: temp
    real(stnd), dimension(size(vec)) :: tempvec
!
    integer(i4b) :: i, k, nvec
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nvec = size( vec )
!
    if ( nvec<=0_i4b ) return
!
    call random_r1( tempvec )
!
    do i = 1_i4b, nvec-1_i4b
!
        k = i + int( tempvec(i)*real( nvec-i+1_i4b, stnd ) )
        temp   = vec(k)
        vec(k) = vec(i)
        vec(i) = temp
!
    end do
!
!
! END OF SUBROUTINE simple_shuffle_rv
! ___________________________________
!
    end subroutine simple_shuffle_rv
!
! =========================================================================================
!
    subroutine simple_shuffle_cv( vec )
!
! purpose
! _______
!
!   This subroutine shuffles all the elements of the complex vector VEC.
!
!
! Arguments
! _________
!
!   VEC      (INPUT/OUTPUT) complex(stnd), dimension(:)
!            On entry, the complex vector to be shuffled.
!
!            On exit, the permuted complex vector.
!
!
! Further Details
! _______________
!
!   For more details and algorithm, see:
!
!   (1) Noreen, E.W., 1989:
!           Computer-intensive methods for testing hypotheses: an introduction.
!           Wiley and Sons, New York, USA, ISBN:978-0-471-61136-3
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(inout), dimension(:) :: vec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    complex(stnd)                    :: temp
    real(stnd), dimension(size(vec)) :: tempvec
!
    integer(i4b) :: i, k, nvec
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nvec = size( vec )
!
    if ( nvec<=0_i4b ) return
!
    call random_r1( tempvec )
!
    do i = 1_i4b, nvec-1_i4b
!
        k = i + int( tempvec(i)*real( nvec-i+1_i4b, stnd ) )
        temp   = vec(k)
        vec(k) = vec(i)
        vec(i) = temp
!
    end do
!
!
! END OF SUBROUTINE simple_shuffle_cv
! ___________________________________
!
    end subroutine simple_shuffle_cv
!
! =========================================================================================
!
    subroutine simple_shuffle_iv( vec )
!
! purpose
! _______
!
!   This subroutine shuffles all the elements of the integer vector VEC.
!
!
! Arguments
! _________
!
!   VEC      (INPUT/OUTPUT) intger(i4b), dimension(:)
!            On entry, the integer vector to be shuffled.
!
!            On exit, the permuted integer vector.
!
!
! Further Details
! _______________
!
!   For more details and algorithm, see:
!
!   (1) Noreen, E.W., 1989:
!           Computer-intensive methods for testing hypotheses: an introduction.
!           Wiley and Sons, New York, USA, ISBN:978-0-471-61136-3
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(inout), dimension(:) :: vec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: temp
!
    real(stnd), dimension(size(vec)) :: tempvec
!
    integer(i4b)  :: i, k, nvec
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nvec = size( vec )
!
    if ( nvec<=0_i4b ) return
!
    call random_r1( tempvec )
!
    do i = 1_i4b, nvec-1_i4b
!
        k = i + int( tempvec(i)*real( nvec-i+1_i4b, stnd ) )
        temp   = vec(k)
        vec(k) = vec(i)
        vec(i) = temp
!
    end do
!
!
! END OF SUBROUTINE simple_shuffle_iv
! ___________________________________
!
    end subroutine simple_shuffle_iv
!
! =========================================================================================
!
    subroutine drawsample( nsample, pop )
!
! purpose
! _______
!
!   This subroutine may be used to draw a sample, without replacement of size NSAMPLE
!   from a population of size SIZE(POP). On output, the integer vector POP(1:NSAMPLE)
!   indicates which observations are included in the sample.
!   
!   The integer vector POP must be dimensioned at least as large as NSAMPLE
!   in the calling program.
!
!
! Arguments
! _________
!
!   NSAMPLE  (INPUT) intger(i4b)
!            On entry, the size of the sample.
!
!   POP      (OUTPUT) integer(i4b), dimension(:)
!            On exit, the indices of the observations belonging to the sample
!            are in POP(1:NSAMPLE) and the indices of the observations, which
!            are not in the sample are in POP(NSAMPLE+1:).
!
!            The size of POP must greater or equal to NSAMPLE. If this condition is not meet
!            POP(:) is set to -1.
!
!
! Further Details
! _______________
!
!   For more details and algorithm, see:
!
!   (1) Noreen, E.W., 1989:
!           Computer-intensive methods for testing hypotheses: an introduction.
!           Wiley and Sons, New York, USA, ISBN:978-0-471-61136-3
!
!
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)                :: nsample
    integer(i4b), intent(out), dimension(:) :: pop
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: temp, npop
!
    real(stnd), dimension(nsample) :: tempvec
!
    integer(i4b) :: i, k 
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    npop = size( pop )
!
    if ( npop<=0_i4b .or. nsample<=0_i4b ) return
!
    if ( nsample<=npop ) then
!
        pop(1_i4b:npop) = arth( 1_i4b, 1_i4b, npop )
!
        call random_r1( tempvec )
!
        do i = 1_i4b, nsample
!
            k = i + int( tempvec(i)*real( npop-i+1_i4b, stnd ) )
            temp   = pop(k)
            pop(k) = pop(i)
            pop(i) = temp
!
        end do
!
    else
!
        pop(1_i4b:npop) = -1_i4b
!
    end if
!
!
! END OF SUBROUTINE drawsample
! ____________________________
!
    end subroutine drawsample
!
! =========================================================================================
!
    subroutine drawbootsample( npop, sample )
!
! purpose
! _______
!
!   This subroutine may be used to draw a bootstrap random sample of size SIZE(SAMPLE)
!   from a population of size NPOP. On output, the integer vector SAMPLE indicates which
!   observations are included in the bootstrap sample.
!
!
! Arguments
! _________
!
!   NPOP     (INPUT) intger(i4b)
!            On entry, the size of the population.
!
!   SAMPLE   (OUTPUT) integer(i4b), dimension(:)
!            On exit, the indices of the observations belonging to the sample.
!
!
! Further Details
! _______________
!
!   The sampling is done with replacement, meaning that the sample may contain
!   duplicate observations.
!
!   For more details and algorithm, see:
!
!   (1) Noreen, E.W., 1989:
!           Computer-intensive methods for testing hypotheses: an introduction.
!           Wiley and Sons, New York, USA, ISBN:978-0-471-61136-3
!
!   
! __________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)                :: npop
    integer(i4b), intent(out), dimension(:) :: sample
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: nsample
!
    real(stnd), dimension(size(sample)) :: tempvec
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nsample = size( sample )
!
    if ( npop<=0_i4b .or. nsample<=0_i4b ) return
!
    call random_r1( tempvec )
!
    sample(:) = int( tempvec(:)*real( npop, stnd ), i4b ) + 1_i4b
!
!
! END OF SUBROUTINE drawbootsample
! ________________________________
!
    end subroutine drawbootsample
!
! =========================================================================================
!
! **********************
! END OF MODULE Random *
! **********************
!    
end module Random
