module SVD_Procedures
! #############################################################################
!
! Copyright 2024 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! Authors: Pascal Terray (LOCEAN/IPSL, Paris, France)
!
! ##############################################################################
!                                                                              *
! ******************************************************************************
! MODULE EXPORTING SUBROUTINES AND FUNCTIONS FOR COMPUTING FULL, PARTIAL SVD   *
! OR QLP DECOMPOSITIONS AND GENERALIZED INVERSE OF A MATRIX.                   *
!                                                                              *
! SUBROUTINES FOR COMPUTING PARTIAL EIGENVALUE, SVD OR QLP DECOMPOSITIONS      *
! BASED ON RANDOMIZED ALGORITHMS ARE ALSO PROVIDED.                            *
!                                                                              *
! LATEST REVISION : 12/03/2024                                                 *
!                                                                              *
! ##############################################################################
!                                                                              *
! ******************************************************************************
!                                                                              *
! SOME OPENMP DIRECTIVES IN THIS MODULE ARE ACTIVATED WITH THE C PROCESSOR     *
! MACRO _OPENMP2 AND, BY DEFAULT, THESE DIRECTIVES ARE ACTIVATED IF OPENMP IS  *
! USED. THESE OPENMP DIRECTIVES ACTIVATE OPENMP PARALLELIZATION IN THE         *
! GENERIC BIDIAGONAL SVD SUBROUTINES bd_svd AND bd_svd2.                       *
! HOWEVER, IN CASE OF PROBLEMS, THESE OPENMP DIRECTIVES MAY BE DESACTIVATED    *
! WITH THE C PROCESSOR MACRO _NOOPENMP2 DURING COMPILATION.                    *
!                                                                              *
! MORE GENERALLY, THE C PROCESSOR MACROS USED IN THIS MODULE ARE:              *
!                                                                              *
!  _OPENMP       FOR ACTIVATING OPENMP PARALLELIZATION                         *
!  _OPENMP2      FOR ACTIVATING OPENMP PARALLELIZATION IN bd_svd AND bd_svd2   *
!                SUBROUTINES                                                   *
!  _F2003        FOR ACTIVATING FORTRAN 2003 CONSTRUCTS                        *
!  _SCALE        FOR USING THE INTRINSIC FUNCTION scale() FOR COMPUTING        *
!                POWERS OF TWO AND NUMERICAL SCALING CONSTANTS                 *
!  _BLAS         FOR USING BLAS WHEN POSSIBLE                                  *
!  _DOT_PRODUCT  FOR REPLACING THE dot_product INTRINSIC FUNCTION WITH         *
!                STATPACK dot_product2 FUNCTION                                *
!  _MATMUL       FOR REPLACING THE matmul INTRINSIC FUNCTION WITH STATPACK     *
!                matmul2 FUNCTION                                              *
!  _TRANSPOSE    FOR REPLACING THE transpose INTRINSIC FUNCTION WITH STATPACK  *
!                transpose2 FUNCTION                                           *
!  _USE_NAGWARE  FOR SIGNALING THAT THE NAG FORTRAN COMPILER IS USED. HERE,    *
!                THIS INCLUDES ONLY THE DESACTIVATION OF THE CHECK OF THE      *
!                ASSOCIATED STATUT OF POINTERS IN SOME SUBROUTINES             *
!                                                                              *
! ******************************************************************************
!
#ifdef _OPENMP
!
#ifndef _NOOPENMP2
#define _OPENMP2
#endif
!
#else
!
#undef _OPENMP2
!
#endif
!
! USED MODULES
! ============
!   
    use Select_Parameters,     only : lgl, i4b, stnd, blksz_qr, blksz2_svd, max_francis_steps_svd
    use Reals_Constants,       only : zero, eighth, quarter, half, one, two, three, four, five, six, eight, &
                                      ten, c1_m4, c0_1, c0_6, c0_9, c0_95, c1_5, c20, c50, c100, c1000
    use Num_Constants,         only : safmin, machbase, maxexp, macheps, ulp=>machulp, smlnum=>rtmin, bignum=>rtmax, nan
    use Logical_Constants
#ifdef _MATMUL
    use Utilities,             only : matmul=>matmul2
#endif
#ifdef _DOT_PRODUCT
    use Utilities,             only : dot_product=>dot_product2
#endif
#ifdef _TRANSPOSE
    use Utilities,             only : transpose=>transpose2
#endif
#ifdef _OPENMP
    use Select_Parameters,     only : omp_limit, omp_limit2, max_num_threads_bd_cmp, &
                                      max_num_threads_bd_cmp2
    use omp_lib,               only : omp_get_max_threads, omp_in_parallel, omp_get_max_active_levels, &
                                      omp_get_num_threads, omp_get_thread_num
#endif
!
! SELECT THE ALGORITHM FOR COMPUTING EUCLIDEAN OR FROBENIUS NORMS IN THE MODULE.   
!
!    use Utilities,       only : norm
!    use Utilities,       only : norm=>norme
    use Utilities,       only : norm=>norm2e
!    use Utilities,       only : norm=>abse

!
! USED MODULES IN SPECIFIC PROCEDURES
! ===================================
!   
!   use Num_Constants
!   use Char_Constants
!   use Utilities
!   use Giv_Procedures
!   use Hous_Procedures
!   use Sort_Procedures
!   use Lin_Procedures
!   use QR_Procedures
!   use Random
!   use Eig_Procedures
!   use BLAS_interfaces
!   use ieee_arithmetic  (if the CPP macro _F2003 is activated)
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::   bd_cmp, bd_cmp2, bd_cmp3, apply_q_bd, apply_p_bd,          &
                ortho_gen_bd, ortho_gen_bd2,                               &
                ortho_gen_q_bd, ortho_gen_p_bd,                            &
                singvalues, singval_sort, singvec_sort,                    &
                bd_svd, svd_sort, bd_svd2, svd_sort2,                      &
                svd_cmp, svd_cmp2, svd_cmp3, svd_cmp4,                     &
                svd_cmp5, svd_cmp6, svd_cmp7, svd_cmp8,                    &
                rsvd_cmp, rsvd_cmp_fixed_precision, reig_pos_cmp,          &
                rqr_svd_cmp, rqr_svd_cmp_fixed_precision,                  &
                rqlp_svd_cmp, rqlp_svd_cmp2, rqlp_svd_cmp_fixed_precision, &
                qlp_cmp, qlp_cmp2, rqlp_cmp,                               &
                bd_singval, select_singval_cmp, select_singval_cmp3,       &
                bd_singval2, select_singval_cmp2, select_singval_cmp4,     &
                bd_max_singval, bd_lasq1, bd_dqds, las2,                   &
                maxdiag_gkinv_qr, maxdiag_gkinv_ldu, gk_qr_cmp,            &
                bd_inviter, bd_inviter2,                                   &
                upper_bd_dsqd, upper_bd_dpqd,                              &
                upper_bd_dsqd2, upper_bd_dpqd2,                            &
                dflgen_bd, dflgen2_bd, dflapp_bd,                          &
                qrstep_bd, qrstep_zero_bd,                                 &
                upper_bd_deflate, bd_deflate, bd_deflate2,                 &
                product_svd_cmp, ginv, comp_ginv,                          &
                gen_bd_mat
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface bd_cmp
        module procedure    bd_cmp_a, bd_cmp_b, bd_cmp_c, bd_cmp_t
    end interface
!
    interface bd_cmp2
        module procedure    bd_cmp2_a, bd_cmp2_b
    end interface
!
    interface bd_svd
        module procedure    bd_svd_a, bd_svd_b, bd_svd_t
    end interface
!
    interface svd_cmp
        module procedure    svd_cmp_a, svd_cmp_b
    end interface
!
    interface svd_sort
        module procedure    lr_singvec_sort, singvec_sort, singval_sort
    end interface
!
    interface bd_svd2
        module procedure    bd_svd2_a, bd_svd_b, bd_svd_t
    end interface
!
    interface svd_cmp2
        module procedure    svd_cmp2_a, svd_cmp_b
    end interface
!
    interface svd_cmp3
        module procedure    svd_cmp3_a, svd_cmp3_b
    end interface
!
    interface svd_cmp4
        module procedure    svd_cmp4_a, svd_cmp3_b
    end interface
!
    interface svd_cmp5
        module procedure    svd_cmp5_a, svd_cmp3_b
    end interface
!
    interface svd_sort2
        module procedure    lr_singvec_sort2, singvec_sort, singval_sort
    end interface
!
    interface gk_qr_cmp
        module procedure    gk_qr_cmp_rv
    end interface
!
    interface select_singval_cmp
        module procedure    select_singval_cmp_a, select_singval_cmp_b
    end interface
!
    interface select_singval_cmp2
        module procedure    select_singval_cmp2_a, select_singval_cmp2_b
    end interface
!
    interface select_singval_cmp3
        module procedure    select_singval_cmp3_a, select_singval_cmp3_b
    end interface
!
    interface select_singval_cmp4
        module procedure    select_singval_cmp4_a, select_singval_cmp4_b
    end interface
!
    interface bd_inviter
        module procedure    bd_inviter_rv, bd_inviter_rm
    end interface
!
    interface bd_inviter2
        module procedure    bd_inviter2_a, bd_inviter2_b, bd_inviter2_c, bd_inviter2_d
    end interface
!
    interface upper_bd_dsqd
        module procedure    upper_bd_dsqd_a, upper_bd_dsqd_b, upper_bd_dsqd_c
    end interface
!
    interface upper_bd_dpqd
        module procedure    upper_bd_dpqd_a, upper_bd_dpqd_b, upper_bd_dpqd_c
    end interface
!
    interface upper_bd_dsqd2
        module procedure    upper_bd_dsqd2_a, upper_bd_dsqd2_b
    end interface
!
    interface upper_bd_dpqd2
        module procedure    upper_bd_dpqd2_a, upper_bd_dpqd2_b
    end interface
!
    interface upper_bd_deflate
        module procedure    upper_bd_deflate_rv, upper_bd_deflate_rm
    end interface
!
    interface bd_deflate2
        module procedure    bd_deflate2_a, bd_deflate2_b, bd_deflate2_c, bd_deflate2_d
    end interface
!
    interface comp_ginv
        module procedure    comp_ginv_a, comp_ginv_b
    end interface
!    
!
! MODULE PARAMETERS
! =================
!
!   BLOCK SIZE FOR REDUCTION OF A RECTANGULAR MATRIX TO BIDIAGONAL FORM
!   IN bd_cmp2_a, bd_cmp2_b AND bd_cmp3 SUBROUTINES.
!
    integer(i4b), parameter :: blksz2 = blksz2_svd
!
!   CROSSOVER POINT WHEN REDUCING A m-BY-n MATRIX TO BIDIAGONAL FORM IN
!   singvalues, select_singval_cmp_a, select_singval_cmp2_a, svd_cmp_a,
!   svd_cmp2_a AND svd_cmp_b ROUTINES. IF max(m,n) EXCEEDS thr_bd*min(m,n), 
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE MATRIX TO
!   TRIANGULAR FORM.
!    
    real(stnd), parameter :: thr_bd = c1_5
!
!   CROSSOVER POINT WHEN REDUCING A m-BY-n MATRIX TO BIDIAGONAL FORM IN
!   svd_cmp3_a, svd_cmp4_a AND svd_cmp3_b ROUTINES. IF max(m,n) EXCEEDS
!   thr_bd2*min(m,n),  A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE
!   THE MATRIX TO TRIANGULAR FORM.
!    
    real(stnd), parameter :: thr_bd2 = c1_5
!
!   CROSSOVER POINT WHEN APPLYING HOUSEHOLDER TRANSFORMATIONS IN
!   A m-BY-n MATRIX TO A SET OF SINGULAR VECTORS IN A n-BY-nv MATRIX
!   IN THE BACK-TRANSFORMATION ALGORITHM OF THE bd_inviter2_a,
!   bd_inviter2_c, bd_inviter2_d, bd_deflate2_a, bd_deflate2_c AND
!   bd_deflate2_d SUBROUTINES. IF nv EXCEEDS thr_bt*n (E.G. nv IS ALMOST
!   EQUAL TO n), THE FIRST n COLUMNS OF THE ORTHOGONAL MATRIX ARE
!   GENERATED AND A MATRIX MULTIPLICATION IS PERFORMED INSTEAD OF
!   APPLYING DIRECTLY THE HOUSEHOLDER TRANSFORMATIONS TO THE SINGULAR
!   VECTORS.
!    
    real(stnd), parameter :: thr_bt = c0_9
!
!   CROSSOVER POINT FOR USING A BLAS3 ALGORITHM WHEN APPLYING A SET OF GIVENS ROTATIONS TO
!   UPDATE THE SINGULAR VECTORS IN THE BIDIAGONAL SVD ALGORITHM OF bd_svd_a, bd_svd2_a AND
!   bd_svd_b SUBROUTINES.
!   IF THE NUMBER OF ELEMENTS IN THE SINGULAR VECTORS EXCEEDS thr_bd_svd_blas3*2*nqrs
!   WHERE nqrs IS THE NUMBER OF QR SWEEPS, WHICH HAVE BEEN DELAYED BEFORE THE UPDATE
!   OF THE SINGULAR VECTORS, A BLAS3 ALGORITHM IS USED INSTEAD OF THE STANDARD BLAS1
!   ALGORITHM.
!
    integer(i4b), parameter :: thr_bd_svd_blas3 = 2_i4b
!
!   DEFAULT BLOCK SIZE FOR THE WAVE FRONT ALGORITHM USED TO APPLY GIVENS ROTATIONS
!   IN THE BIDIAGONAL SVD ALGORITHM IN bd_svd_a, bd_svd2_a and bd_svd_b SUBROUTINES.
!
    integer(i4b), parameter :: max_francis_steps_default = max_francis_steps_svd
!
!   A "FUDGE FACTOR" TO DETERMINE WHEN A DQDS STEP IS ACCEPTED IN bd_lasq1, bd_lasq2,
!   bd_lasq3, bd_lasq2_aggdef2, bd_aggdef2, bd_one_singval_lasq1, bd_one_singval_lasq2
!   and bd_one_singval_lasq3 SUBROUTINES. NOTE THAT THIS PARAMETER HAS EFFECT ON ACCURACY
!   OF THE SOLUTION AND SPEED OF COMPUTATION. A SMALLER VALUE WILL GIVE HIGHER ACCURACY
!   AT THE EXPENSE OF THE SPEED OF THE COMPUTATION.
!
    real(stnd),  parameter :: fudge_lasq = c100
!
!   A "FUDGE FACTOR" TO DETERMINE WHEN A DQDS STEP IS ACCEPTED IN bd_dqds AND
!   bd_one_singval_dqds SUBROUTINES. NOTE THAT THIS PARAMETER HAS EFFECT ON ACCURACY
!   OF THE SOLUTION AND SPEED OF COMPUTATION. A SMALLER VALUE WILL GIVE HIGHER ACCURACY
!   AT THE EXPENSE OF THE SPEED OF THE COMPUTATION.
!
    real(stnd),  parameter :: fudge_dqds = c1000
!
!   A "FUDGE FACTOR" TO DETERMINE WHEN A DEFLATION STEP IS ACCEPTED IN dflgen2_bd AND
!   dflapp_bd SUBROUTINES. IDEALLY, A VALUE OF 1 SHOULD WORK, BUT A LARGER OR LOWER VALUE
!   CAN BE USED. NOTE THAT THIS PARAMETER HAS EFFECT ON ACCURACY OF THE SOLUTION AND SPEED
!   OF COMPUTATION. A SMALLER VALUE WILL GIVE HIGHER ACCURACY AT THE EXPENSE OF THE SPEED
!   OF THE COMPUTATION.
!
    real(stnd),  parameter :: fudge_deflate = c1000
!
!   A "FUDGE FACTOR" TO DETERMINE WHEN ORTHOGONALIZATION MUST BE PERFORMED IN bd_inviter
!   AND bd_inviter2 SUBROUTINES OR WHEN SUCCESSIVE DEFLATION MUST BE PERFORMED IN bd_deflate
!   AND bd_deflate2 SUBROUTINES. IDEALLY, A VALUE OF 1 SHOULD WORK, BUT A LARGER VALUE
!   CAN BE USED FOR IMPROVED ORTHOGONALITY AT THE EXPENSE OF SPEED OF COMPUTATION.
!   NOTE THAT THIS PARAMETER HAS EFFECT ON ACCURACY AND ORTHOGONALITY OF SINGULAR VECTORS
!   PRODUCED BY bd_inviter, bd_inviter2, bd_deflate AND bd_deflate2 SUBROUTINES.
!
    real(stnd),  parameter :: fudge_ortho = c50
!
!   A "FUDGE FACTOR" TO DETERMINE WHEN REORTHOGONALIZATION OF ORTHOGONAL VECTORS MUST
!   BE PERFORMED IN bd_cmp2 AND bd_inviter SUBROUTINES. A VALUE BETWEEN 0 AND 1 MUST BE SELECTED.
!   IDEALLY, A VERY SMALL VALUE SHOULD WORK, BUT A LARGER VALUE CAN BE USED FOR IMPROVED ORTHOGONALITY
!   AT THE EXPENSE OF SPEED OF COMPUTATION. NOTE THAT THIS PARAMETER HAS EFFECT ON ACCURACY
!   AND ORTHOGONALITY OF THE ORTHOGONAL VECTORS PRODUCED BY bd_cmp2 AND bd_inviter SUBROUTINES AS WELL
!   AS SINGULAR VECTORS PRODUCED BY svd_cmp3 AND bd_inviter2 SUBROUTINES.
!
    real(stnd),  parameter :: fudge_ortho2 = c0_6
!
!   A TOLERANCE FOR THE TEST OF SINGULARITY IN THE bd_cmp2 SUBROUTINES.
!
    real(stnd),  parameter :: eps_bd_cmp2 = min( c0_1, fudge_ortho2 )
!
!   A TOLERANCE FOR THE TEST OF SINGULARITY IN THE bd_cmp3 SUBROUTINE.
!
    real(stnd),  parameter :: eps_bd_cmp3 = min( c0_1, fudge_ortho2 )
!
!   A "FUDGE FACTOR" FOR THE TESTS OF ACCURACY OF SINGULAR VECTORS COMPUTED BY bd_inviter, bd_inviter2,
!   bd_deflate AND bd_deflate2 SUBROUTINES. A LARGE VALUE RELAXES THE TESTS OF ACCURACY IN THESE
!   SUBROUTINES. THIS FUDGE FACTOR ALSO MODIFIES THE AMPLITUDE OF THE RELATIVE PERTURBATION ADDED
!   TO THE SINGULAR VALUES IF THESE SINGULAR VALUES ARE TOO CLOSED IN bd_inviter AND bd_inviter2
!   SUBROUTINES.
!
    real(stnd),  parameter :: fudge_prec = c20
!
!   A "FUDGE FACTOR" FOR THE TOLERANCE IN THE bd_svd AND bd_svd2 BIDIAGONAL SVD SUBROUTINES.
!   A LARGE VALUE RELAXES THE TESTS OF ACCURACY IN THESE SUBROUTINES.
!
    real(stnd),  parameter :: fudge_eps = one
!
!   TWO TOLERANCES TO DETERMINE WHEN LEFT SINGULAR VECTORS COMPUTED BY bd_inviter2 AND bd_deflate2
!   SUBROUTINES MUST BE REORTOGONALIZED TO CORRECT FOR THE LOSS OF ORTHOGONALITY IN THE RALHA-BARLOW
!   ONE-SIDED BIDIAGONAL REDUCTION ALGORITHM.
!
    real(stnd),  parameter :: tol_reortho_def = sqrt( ulp ), tol_reortho_partial_def = c1_m4
!
!   A THRESHOLD TO DETERMINE SMALL AND BIG CLUSTERS OF SINGULAR VALUES IN bd_inviter, bd_inviter2
!   bd_deflate AND bd_deflate2 SUBROUTINES. NOTE THAT THIS PARAMETER HAS EFFECT ON SPEED AND
!   PARALLELIZATION WHEN OPENMP IS USED, BUT NOT ON ACCURACY OF SINGULAR VECTORS PRODUCED BY
!   bd_inviter, bd_inviter2, bd_deflate AND bd_deflate2. nsing_cluster MUST BE GREATER OR EQUAL
!   TO 1.
!
    integer(i4b), parameter :: nsing_cluster = blksz_qr+20_i4b
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
!
! =========================================================================================
!                         BIDIAGONAL DECOMPOSITION OF A MATRIX
! =========================================================================================
!
!
    subroutine bd_cmp_a( mat, d, e, tauq, taup )
!
! Purpose
! _______
!
!   BD_CMP reduces a general m-by-n matrix MAT to upper or lower bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal. If:
!
!   - m >= n, BD is upper bidiagonal;
!   - m < n, BD is lower bidiagonal.
!
!   BD_CMP computes BD, Q and P.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the general m-by-n matrix to be reduced.
!
!            On exit, if:
!
!            - m >= n, the elements on and below the diagonal,
!              with the array TAUQ, represent the orthogonal
!              matrix Q as a product of elementary reflectors, 
!              and the elements above the diagonal, with the
!              array TAUP, represent the orthogonal matrix P
!              as a product of elementary reflectors;
!            - m < n, the elements below the diagonal, with
!              the array TAUQ, represent the orthogonal
!              matrix Q as a product of elementary reflectors,
!              and the elements on and above the diagonal,
!              with the array TAUP, represent the orthogonal
!              matrix P as a product of elementary reflectors.
!
!            See Further Details.
!
!   D        (OUTPUT) real(stnd), dimension(:)
!            The diagonal elements of the bidiagonal matrix BD
!
!            The size of D must be min( size(MAT,1) , size(MAT,2) ).
!
!   E        (OUTPUT) real(stnd), dimension(:)
!            The off-diagonal elements of the bidiagonal matrix BD:
!
!            - if m >= n, E(i) = BD(i-1,i) for i = 2,3,...,n;
!            - if m < n,  E(i) = BD(i,i-1) for i = 2,3,...,m.
!
!            The size of E must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUQ     (OUTPUT) real(stnd), dimension(:)
!            The scalar factors of the elementary reflectors which
!            represent the orthogonal matrix Q. See Further Details.
!
!            The size of TAUQ must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUP     (OUTPUT) real(stnd), dimension(:)
!            The scalar factors of the elementary reflectors which
!            represent the orthogonal matrix P. See Further Details.
!
!            The size of TAUP must be min( size(MAT,1) , size(MAT,2) ).
!
!
! Further Details
! _______________
!
!   The matrices Q and P are represented as products of elementary
!   reflectors:
!
!   If m >= n,
!
!       Q = H(1) * H(2) * . . . * H(n)  and  P = G(1) * G(2) * . . . * G(n-1)
!
!   Each H(i) and G(i) has the form:
!
!       H(i) = I + tauq * u * u'  and G(i) = I + taup * v * v'
!
!   where tauq and taup are real scalars, and u and v are real vectors;
!   u(1:i-1) = 0 and u(i:m) is stored on exit in MAT(i:m,i);
!   v(1:i) = 0 and v(i+1:n) is stored on exit in MAT(i,i+1:n);
!   tauq is stored in TAUQ(i) and taup in TAUP(i).
!
!   If m < n,
!
!       Q = H(1) * H(2) * . . . * H(m-1)  and  P = G(1) * G(2) * . . . * G(m)
!
!   Each H(i) and G(i) has the form:
!
!       H(i) = I + tauq * u * u'  and G(i) = I + taup * v * v'
!
!   where tauq and taup are real scalars, and u and v are real vectors;
!   u(1:i) = 0 and u(i+1:m) is stored on exit in MAT(i+1:m,i);
!   v(1:i-1) = 0 and v(i:n) is stored on exit in MAT(i,i:n);
!   tauq is stored in TAUQ(i) and taup in TAUP(i).
!
!   The contents of MAT on exit are illustrated by the following examples:
!
!   m = 6 and n = 5 (m >= n):
!
!        (  u1  v1  v1  v1  v1 )
!
!        (  u1  u2  v2  v2  v2 )
!
!        (  u1  u2  u3  v3  v3 )
!
!        (  u1  u2  u3  u4  v4 )
!
!        (  u1  u2  u3  u4  u5 )
!
!        (  u1  u2  u3  u4  u5 )
!
!   m = 5 and n = 6 (m < n):
!
!        (  v1  v1  v1  v1  v1  v1 )
!
!        (  u1  v2  v2  v2  v2  v2 )
!
!        (  u1  u2  v3  v3  v3  v3 )
!
!        (  u1  u2  u3  v4  v4  v4 )
!
!        (  u1  u2  u3  u4  v5  v5 )
!
!   where ui denotes an element of the vector defining H(i), and vi an element 
!   of the vector defining G(i).
!
!   This subroutine is adapted from the routine DGEBD2 in LAPACK.
!   An efficient variant of the classic Golub and Kahan Householder bidiagonalization
!   algorithm is used. This variant reduces the traffic on the data bus from four reads
!   and two writes per column-row elimination of the bidiagonalization process to one
!   read and one write. Furthermore, the algorithm is parallelized if OPENMP is used.
!
!   For further details on the bidiagonal reduction algorithm and its use
!   or the efficient variant used here, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Howell, G.W., Demmel, J., Fulton, C.T., Hammarling, S., and Marmol, K., 2008:
!           Cache efficient bidiagonalization using BLAS 2.5 operators.
!           ACM Transactions on Mathematical Software (TOMS) Volume 34, Issue 3.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
    use Hous_Procedures, only : hous1
#ifdef _BLAS
    use BLAS_interfaces, only : dot, axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: d, e, tauq, taup
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: prod, tau, beta, alpha
    real(stnd), dimension(size(mat,1))   :: u, u2, tmp
    real(stnd), dimension(size(mat,2))   :: v, v2
!
    integer(i4b) :: m, n, k, i, j
!
#ifdef _OPENMP
!
    real(stnd), dimension(size(mat,1))   :: w
!
    integer(i4b) :: nn, max_num_threads, num_threads, thread_num, jinc, jrem, joffset, j1, j2
!
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k =  assert_eq( (/ min(m,n)            ,   &
                       int(size(tauq),i4b) ,   &
                       int(size(taup),i4b) ,   &
                       int(size(d),i4b)    ,   &
                       int(size(e),i4b) /) ,   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
    e(1_i4b) = zero
!
    if ( k==1_i4b ) then
!
!       SPECIAL CASES IF mat(:,:) IS A ROW OR COLUMN VECTOR.
!
        if ( n==1_i4b ) then
!
!           GENERATE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(2:m,1).
!
            call hous1( mat(1_i4b:m,1_i4b), tauq(1_i4b), d(1_i4b) )
!
            taup(1_i4b) = zero         
!
        else
!
!           GENERATE ELEMENTARY REFLECTOR G(1) TO ANNIHILATE mat(1,2:n).
!
            call hous1( mat(1_i4b,1_i4b:n), taup(1_i4b), d(1_i4b) )
!
            tauq(1_i4b) = zero         
!
        end if
!
        return
!
    end if
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m*n>=omp_limit                      .and.       &
                  i>1_i4b
!
    max_num_threads = min( i, max_num_threads_bd_cmp )
#endif
!
    if ( m>=n ) then
!
!       REDUCE mat TO UPPER BIDIAGONAL FORM.
!
!       GENERATE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(2:m,1).
!
        call hous1( mat(1_i4b:m,1_i4b), tau, d(1_i4b) )
!
        tauq(1_i4b) = tau
!
#ifdef _OPENMP
!
        if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(i,k,nn,num_threads,thread_num,jinc,jrem,joffset,j,j1,j2,tmp,u,w,alpha),  &
!$OMP          NUM_THREADS(max_num_threads)
!
            num_threads = omp_get_num_threads()
            thread_num  = omp_get_thread_num()
!
!           MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(1).
!
            u(1_i4b:m) = mat(1_i4b:m,1_i4b)
!
!           DISTRIBUTE THE WORK AMONG THE THREADS FOR THE FIRST
!           RANK-2 COLUMN-ROW UPDATE.
!
            nn   = n - 1_i4b
            jrem = mod( nn, num_threads )
            jinc = ( nn - jrem )/num_threads
!
            if ( thread_num<jrem ) then
                jinc    = jinc + 1_i4b
                joffset = 2_i4b
            else
                joffset = jrem + 2_i4b
            end if
!
            j1 = thread_num*jinc + joffset
            j2 = j1 + jinc - 1_i4b
!
!           START THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            tmp(2_i4b:m) = zero
!
            do j = j1, j2
!
#ifdef _BLAS
                v2(j) = tau*dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,j), 1_i4b )
#else
                v2(j) = tau*dot_product( u(1_i4b:m), mat(1_i4b:m,j) )
#endif
                v(j)  = mat(1_i4b,j) + v2(j)
#ifdef _BLAS
                call axpy( m-1_i4b, v(j), mat(2_i4b:m,j), 1_i4b, tmp(2_i4b:m), 1_i4b )
#else
                tmp(2_i4b:m) = tmp(2_i4b:m) + v(j)*mat(2_i4b:m,j)
#endif
!
            end do
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!           GENERATE ELEMENTARY REFLECTOR G(1) TO ANNIHILATE mat(1,3:n).
!
            alpha = v(2_i4b)
!
            call hous1( v(2_i4b:n), tau, prod )
!
            taup(1_i4b) = tau
            e(2_i4b)    = prod
!
!           CONTINUE THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            if ( tau/=zero ) then
                beta        = tau/(alpha-prod)
                u2(2_i4b:m) = - mat(2_i4b:m,2_i4b)
            else
                beta        = zero
                u2(2_i4b:m) = zero
            end if
!
            prod = zero
!
!$OMP END SINGLE
!
            do i = 2_i4b, n-1_i4b
!
                k = i + 1_i4b
!
!               SAVE ELEMENTARY REFLECTOR G(i-1).
!
                mat(i-1_i4b,j1:j2) = v(j1:j2)
!
!               CONTINUE THE i-1 RANK-2 COLUMN-ROW UPDATE.
!
                alpha = dot_product( v2(j1:j2), v(j1:j2) )*tau
!
!$OMP CRITICAL (updateua)
!
#ifdef _BLAS
                call axpy( m-i+1_i4b, beta, tmp(i:m), 1_i4b, u2(i:m), 1_i4b )
#else
                u2(i:m) = u2(i:m) + beta*tmp(i:m) 
#endif
!
                prod = prod + alpha
!
!$OMP END CRITICAL (updateua)
!
!$OMP BARRIER
!
                v2(j1:j2) = v2(j1:j2) + prod*v(j1:j2)
!
!$OMP MASTER
!
!               APPLY H(i-1) TO mat(i:m,i) FROM THE LEFT AND 
!               APPLY G(i-1)   TO mat(i:m,i) FROM THE RIGHT.
!
#ifdef _BLAS
                w(i:m) = mat(i:m,i)
                call axpy( m-i+1_i4b, v2(i), u(i:m), 1_i4b, w(i:m), 1_i4b )
                call axpy( m-i+1_i4b, v(i), u2(i:m), 1_i4b, w(i:m), 1_i4b )
#else
                w(i:m) = mat(i:m,i) + v2(i)*u(i:m) + v(i)*u2(i:m)
#endif
!
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+1:m,i).
!
                call hous1( w(i:m), tau, d(i) )
!
!               SAVE ELEMENTARY REFLECTOR H(i).
!
                mat(i:m,i) = w(i:m)
                tauq(i)    = tau
!
!$OMP END MASTER
!
!$OMP BARRIER
!
                if ( thread_num/=0_i4b ) then
!
!                   MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(i).
!
                    w(i:m) = mat(i:m,i)
!
                end if
!
!               DISTRIBUTE THE WORK AMONG THE THREADS FOR THE i
!               RANK-2 COLUMN-ROW UPDATE.
!
                nn   = nn - 1_i4b
                jrem = mod( nn, num_threads )
                jinc = ( nn - jrem )/num_threads
!
                if ( thread_num<jrem ) then
                    jinc    = jinc + 1_i4b
                    joffset = i + 1_i4b
                else
                    joffset = i + jrem + 1_i4b
                end if
!
                j1 = thread_num*jinc + joffset
                j2 = j1 + jinc - 1_i4b
!
!               FINISH THE i-1 RANK-2 COLUMN-ROW UPDATE AND
!               START THE i RANK-2 COLUMN-ROW UPDATE.
!
                tmp(k:m) = zero
!
                do j = j1, j2
!
!                   APPLY H(i-1) TO mat(i:m,i+1:n) FROM THE LEFT AND 
!                   APPLY G(i-1) TO mat(i:m,i+1:n) FROM THE RIGHT.
!
#ifdef _BLAS
                    call axpy( m-i+1_i4b, v2(j), u(i:m), 1_i4b, mat(i:m,j), 1_i4b )
                    call axpy( m-i+1_i4b, v(j), u2(i:m), 1_i4b, mat(i:m,j), 1_i4b )
#else
                    mat(i:m,j) = mat(i:m,j) + v2(j)*u(i:m) + v(j)*u2(i:m)
#endif
!
!                   START THE i RANK-2 COLUMN-ROW UPDATE.
!
#ifdef _BLAS
                    v2(j) = tau*dot( m-i+1_i4b, w(i:m), 1_i4b, mat(i:m,j), 1_i4b )
#else
                    v2(j) = tau*dot_product( w(i:m), mat(i:m,j) )
#endif
                    v(j)  = mat(i,j) + v2(j)
#ifdef _BLAS
                    call axpy( m-k+1_i4b, v(j), mat(k:m,j), 1_i4b, tmp(k:m), 1_i4b )
#else
                    tmp(k:m) = tmp(k:m) + v(j)*mat(k:m,j)
#endif
!
                end do
!
!               SAVE A LOCAL COPY OF ELEMENTARY REFLECTOR H(i).
!
                u(i:m) = w(i:m)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!               GENERATE ELEMENTARY REFLECTOR G(i) TO ANNIHILATE mat(i,i+2:n)
!
                alpha = v(k)
!
                call hous1( v(k:n), tau, prod )
!
                taup(i) = tau
                e(k)    = prod
!
!               CONTINUE THE i RANK-2 COLUMN-ROW UPDATE.
!
                if ( tau/=zero ) then
                    beta    = tau/(alpha-prod)
                    u2(k:m) = - mat(k:m,k)
                else
                    beta    = zero
                    u2(k:m) = zero
                end if
!
                prod = zero
!
!$OMP END SINGLE
!
            end do
!
!$OMP SINGLE
!
!           CONTINUE THE n-1 RANK-2 COLUMN-ROW UPDATE.
!
            prod = v2(n)*tau
!
            v2(n) = v2(n) + prod*v(n)
!
!           SAVE ELEMENTARY REFLECTOR G(n-1).
!
            mat(n-1_i4b,n) = one
!
!$OMP END SINGLE
!
!$OMP END PARALLEL
!
        else
#endif
!
!           MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(1).
!
            u(1_i4b:m) = mat(1_i4b:m,1_i4b)
!
!           START THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            tmp(2_i4b:m) = zero
!
            do j = 2_i4b, n
!
#ifdef _BLAS
                v2(j) = tau*dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,j), 1_i4b )
#else
                v2(j) = tau*dot_product( u(1_i4b:m), mat(1_i4b:m,j) )
#endif
                v(j)  = mat(1_i4b,j) + v2(j)
#ifdef _BLAS
                call axpy( m-1_i4b, v(j), mat(2_i4b:m,j), 1_i4b, tmp(2_i4b:m), 1_i4b )
#else
                tmp(2_i4b:m) = tmp(2_i4b:m) + v(j)*mat(2_i4b:m,j)
#endif
!
            end do
!
!           GENERATE ELEMENTARY REFLECTOR G(1) TO ANNIHILATE mat(1,3:n).
!
            alpha = v(2_i4b)
!
            call hous1( v(2_i4b:n), tau, prod )
!
!           SAVE ELEMENTARY REFLECTOR G(1).
!
            mat(1_i4b,2_i4b:n) = v(2_i4b:n)
            taup(1_i4b)        = tau
            e(2_i4b)           = prod
!
!           CONTINUE THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            if ( tau/=zero ) then
!
                beta        = tau/(alpha-prod)
                u2(2_i4b:m) = beta*tmp(2_i4b:m) - mat(2_i4b:m,2_i4b)
!
#ifdef _BLAS
                prod = dot( n-1_i4b, v2(2_i4b:n), 1_i4b, v(2_i4b:n), 1_i4b )*tau
!
                call axpy( n-1_i4b, prod, v(2_i4b:n), 1_i4b, v2(2_i4b:n), 1_i4b )
#else
                prod = dot_product( v2(2_i4b:n), v(2_i4b:n) )*tau
!
                v2(2_i4b:n) = v2(2_i4b:n) + prod*v(2_i4b:n)
#endif
!
            else
!
                u2(2_i4b:m) = zero
!
            end if
!
            do i = 2_i4b, n-1_i4b
!
                k = i + 1_i4b
!
!               APPLY H(i-1) TO mat(i:m,i) FROM THE LEFT AND 
!               APPLY G(i-1)   TO mat(i:m,i) FROM THE RIGHT.
!
#ifdef _BLAS
                tmp(i:m) = mat(i:m,i)
                call axpy( m-i+1_i4b, v2(i), u(i:m), 1_i4b, tmp(i:m), 1_i4b )
                call axpy( m-i+1_i4b, v(i), u2(i:m), 1_i4b, tmp(i:m), 1_i4b )
#else
                tmp(i:m) = mat(i:m,i) + v2(i)*u(i:m) + v(i)*u2(i:m)
#endif
!
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+1:m,i).
!
                call hous1( tmp(i:m), tau, d(i) )
!
!               SAVE ELEMENTARY REFLECTOR H(i).
!
                mat(i:m,i) = tmp(i:m)
                tauq(i)    = tau
!
!               FINISH THE i-1 RANK-2 COLUMN-ROW UPDATE AND
!               START THE i RANK-2 COLUMN-ROW UPDATE.
!
                tmp(k:m) = zero
!
                do j = k, n
!
!                   APPLY H(i-1) TO mat(i:m,i+1:n) FROM THE LEFT AND 
!                   APPLY G(i-1) TO mat(i:m,i+1:n) FROM THE RIGHT.
!
#ifdef _BLAS
                    call axpy( m-i+1_i4b, v2(j), u(i:m), 1_i4b, mat(i:m,j), 1_i4b )
                    call axpy( m-i+1_i4b, v(j), u2(i:m), 1_i4b, mat(i:m,j), 1_i4b )
#else
                    mat(i:m,j) = mat(i:m,j) + v2(j)*u(i:m) + v(j)*u2(i:m)
#endif
!
!                   START THE i RANK-2 COLUMN-ROW UPDATE.
!
#ifdef _BLAS
                    v2(j) = tau*dot( m-i+1_i4b, mat(i:m,i), 1_i4b, mat(i:m,j), 1_i4b )
#else
                    v2(j) = tau*dot_product( mat(i:m,i), mat(i:m,j) )
#endif
                    v(j)  = mat(i,j) + v2(j)
#ifdef _BLAS
                    call axpy( m-k+1_i4b, v(j), mat(k:m,j), 1_i4b, tmp(k:m), 1_i4b )
#else
                    tmp(k:m) = tmp(k:m) + v(j)*mat(k:m,j)
#endif
!
                end do
!
!               MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(i).
!
                u(i:m) = mat(i:m,i)
!
!               GENERATE ELEMENTARY REFLECTOR G(i) TO ANNIHILATE mat(i,i+2:n)
!
                alpha = v(k)
!
                call hous1( v(k:n), tau, prod )
!
!               SAVE ELEMENTARY REFLECTOR G(i).
!
                mat(i,k:n) = v(k:n)
                taup(i)    = tau
                e(k)       = prod
!
!               CONTINUE THE i RANK-2 COLUMN-ROW UPDATE.
!
                if ( tau/=zero ) then
!
                    beta    = tau/(alpha-prod)
                    u2(k:m) = beta*tmp(k:m) - mat(k:m,k)
!
#ifdef _BLAS
                    prod = dot( n-k+1_i4b, v2(k:n), 1_i4b, v(k:n), 1_i4b )*tau
!
                    call axpy( n-k+1_i4b, prod, v(k:n), 1_i4b, v2(k:n), 1_i4b )
#else
                    prod    = dot_product( v2(k:n), v(k:n) )*tau
!
                    v2(k:n) = v2(k:n) + prod*v(k:n)
#endif
!
                else
!
                    u2(k:m) = zero
!
                end if
!
            end do
!
#ifdef _OPENMP
        end if
#endif
!
!       FINISH THE n-1 RANK-2 COLUMN-ROW UPDATE.
!
!       APPLY H(n-1) TO mat(n:m,n) FROM THE LEFT AND 
!       APPLY G(n-1) TO mat(n:m,n) FROM THE RIGHT.
!
#ifdef _BLAS
        call axpy( m-n+1_i4b, v2(n), mat(n:m,n-1_i4b), 1_i4b, mat(n:m,n), 1_i4b )
#else
        mat(n:m,n) = mat(n:m,n) + v2(n)*mat(n:m,n-1_i4b)
#endif
!
!       GENERATE ELEMENTARY REFLECTOR H(n).
!
        call hous1( mat(n:m,n), tauq(n), d(n) )
!
        taup(n) = zero
!
    else
!
!       REDUCE mat TO LOWER BIDIAGONAL FORM.
!
        v(1_i4b:n) = mat(1_i4b,1_i4b:n)
!
!       GENERATE ELEMENTARY REFLECTOR G(1) TO ANNIHILATE mat(1,2:n).
!
        call hous1( v(1_i4b:n), tau, d(1_i4b) )
!
!       SAVE ELEMENTARY REFLECTOR G(1).
!
        mat(1_i4b,1_i4b:n) = v(1_i4b:n)
        taup(1_i4b)        = tau
!
#ifdef _OPENMP
!
        if ( test_par ) then
!
            u2(2_i4b:m)  = mat(2_i4b:m,1_i4b)
!
!$OMP PARALLEL PRIVATE(i,k,nn,num_threads,thread_num,jinc,jrem,joffset,j,j1,j2,tmp,u,alpha),  &
!$OMP          NUM_THREADS(max_num_threads)
!
            num_threads = omp_get_num_threads()
            thread_num  = omp_get_thread_num()
!
!           DISTRIBUTE THE WORK AMONG THE THREADS FOR THE FIRST
!           RANK-2 COLUMN-ROW UPDATE.
!
            nn   = n - 1_i4b
            jrem = mod( nn, num_threads )
            jinc = ( nn - jrem )/num_threads
!
            if ( thread_num<jrem ) then
                jinc    = jinc + 1_i4b
                joffset = 2_i4b
            else
                joffset = jrem + 2_i4b
            end if
!
            j1 = thread_num*jinc + joffset
            j2 = j1 + jinc - 1_i4b
!
!           START THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            if ( j1<=j2 ) then
!
                tmp(2_i4b:m) = matmul( mat(2_i4b:m,j1:j2), v(j1:j2) )
!
!$OMP CRITICAL (updateua)
!
#ifdef _BLAS
                call axpy( m-1_i4b, one, tmp(2_i4b:m), 1_i4b, u2(2_i4b:m), 1_i4b )
#else
                u2(2_i4b:m) = u2(2_i4b:m) + tmp(2_i4b:m)  
#endif
!
!$OMP END CRITICAL (updateua)
!
            end if
!
!$OMP BARRIER
!
!$OMP MASTER
!
            u2(2_i4b:m) = tau*u2(2_i4b:m) 
            u(2_i4b:m)  = mat(2_i4b:m,1_i4b) + u2(2_i4b:m) 
!
!           GENERATE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(3:m,1).
!
            call hous1( u(2_i4b:m), tau, e(2_i4b) )
!
!           SAVE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(3:m,1).
!
            mat(2_i4b:m,1_i4b ) = u(2_i4b:m)
            tauq(1_i4b)         = tau
!
#ifdef _BLAS
            prod = dot( m-1_i4b, u(2_i4b:m), 1_i4b, u2(2_i4b:m), 1_i4b )
#else
            prod = dot_product( u(2_i4b:m), u2(2_i4b:m) )
#endif
!
            w(2_i4b:m) = zero
!
!$OMP END MASTER
!
!$OMP BARRIER
!
            if ( thread_num/=0_i4b ) then
!
!               MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(1).
!
                u(2_i4b:m)  = mat(2_i4b:m,1_i4b)
!
            end if
!
!           FINISH THE FIRST RANK-2 COLUMN-ROW UPDATE AND
!           START THE SECOND RANK-2 COLUMN-ROW UPDATE.
!
            tmp(2_i4b:m) = zero
!
            do j = j1, j2
!
#ifdef _BLAS
                alpha = tau*( dot( m-1_i4b, u(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b ) + prod*v(j) )
#else
                alpha = tau*( dot_product( u(2_i4b:m), mat(2_i4b:m,j) ) + prod*v(j) )
#endif
!
!               APPLY G(1) TO mat(2:m,2:n) FROM THE RIGHT AND
!               APPLY H(1) TO mat(2:m,2:n) FROM THE LEFT.
!
#ifdef _BLAS
                call axpy( m-1_i4b, alpha, u(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b )
                call axpy( m-1_i4b, v(j), u2(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b )
#else
                mat(2_i4b:m,j) = mat(2_i4b:m,j) + alpha*u(2_i4b:m) + v(j)*u2(2_i4b:m)
#endif
!
!               START THE SECOND RANK-2 COLUMN-ROW UPDATE.
!
                v(j) = mat(2_i4b,j)
#ifdef _BLAS
                call axpy( m-2_i4b, v(j), mat(3_i4b:m,j), 1_i4b, tmp(3_i4b:m), 1_i4b )
#else
                tmp(3_i4b:m) = tmp(3_i4b:m) + v(j)*mat(3_i4b:m,j)
#endif
!                    
            end do
!
!$OMP CRITICAL (updateua)
!
#ifdef _BLAS
            call axpy( m-2_i4b, one, tmp(3_i4b:m), 1_i4b, w(3_i4b:m), 1_i4b )
#else
            w(3_i4b:m) = w(3_i4b:m) + tmp(3_i4b:m)  
#endif
!
!$OMP END CRITICAL (updateua)
!
!$OMP BARRIER
!
            do i = 2_i4b, m-1_i4b
!
                k = i + 1_i4b
!
!               DISTRIBUTE THE WORK AMONG THE THREADS FOR THE i
!               RANK-2 COLUMN-ROW UPDATE.
!
                nn   = nn - 1_i4b
                jrem = mod( nn, num_threads )
                jinc = ( nn - jrem )/num_threads
!
                if ( thread_num<jrem ) then
                    jinc    = jinc + 1_i4b
                    joffset = i + 1_i4b
                else
                    joffset = i + jrem + 1_i4b
                end if
!
                j1 = thread_num*jinc + joffset
                j2 = j1 + jinc - 1_i4b
!
!$OMP MASTER
!
                alpha = v(i)
!
!               GENERATE ELEMENTARY REFLECTOR G(i) TO ANNIHILATE mat(i,i+1:n).
!
                call hous1( v(i:n), tau, prod )
!
!               SAVE ELEMENTARY REFLECTOR G(i).
!
                mat(i,i) = one
                d(i)     = prod
                taup(i)  = tau
!
!               CONTINUE THE i RANK-2 COLUMN-ROW UPDATE.
!
                if ( tau/=zero ) then
                    beta    = tau/(alpha-prod)
                    u2(k:m) = beta*( w(k:m) - prod*mat(k:m,i) )
                    u(k:m)  = mat(k:m,i) + u2(k:m) 
                else
                    u2(k:m) = zero
                    u(k:m)  = mat(k:m,i)
                end if
!
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+2:m,i).
!
                call hous1( u(k:m), tau, e(k) )
!
!               SAVE ELEMENTARY REFLECTOR H(i).
!
                mat(k:m,i) = u(k:m)
                tauq(i)    = tau
!
#ifdef _BLAS
                prod = dot( m-k+1_i4b, u(k:m), 1_i4b, u2(k:m), 1_i4b )
#else
                prod = dot_product( u(k:m), u2(k:m) )
#endif
!
                w(k:m) = zero
!
!$OMP END MASTER
!
!$OMP BARRIER
!
                if ( thread_num/=0_i4b ) then
!
!                   MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(i).
!
                    u(k:m)  = mat(k:m,i)
!
                end if
!
!               SAVE ELEMENTARY REFLECTOR G(i).
!
                mat(i,j1:j2) = v(j1:j2)
!
!               FINISH THE i RANK-2 COLUMN-ROW UPDATE AND
!               START THE i+1 RANK-2 COLUMN-ROW UPDATE.
!
                tmp(k:m) = zero
!
                do j = j1, j2
!
#ifdef _BLAS
                    alpha= tau*( dot( m-k+1_i4b, u(k:m), 1_i4b, mat(k:m,j), 1_i4b ) + prod*v(j) )
#else
                    alpha= tau*( dot_product( u(k:m), mat(k:m,j) ) + prod*v(j) )
#endif
!
!                   APPLY G(i) TO mat(i+1:m,i+1:n) FROM THE RIGHT AND
!                   APPLY H(i) TO mat(i+1:m,i+1:n) FROM THE LEFT.
!
#ifdef _BLAS
                    call axpy( m-k+1_i4b, alpha, u(k:m), 1_i4b, mat(k:m,j), 1_i4b )
                    call axpy( m-k+1_i4b, v(j), u2(k:m), 1_i4b, mat(k:m,j), 1_i4b )
#else
                    mat(k:m,j) = mat(k:m,j) + alpha*u(k:m) + v(j)*u2(k:m)
#endif
!
!                   START THE i+1 RANK-2 COLUMN-ROW UPDATE.
!
                    v(j) = mat(k,j)
#ifdef _BLAS
                    call axpy( m-k, v(j), mat(k+1_i4b:m,j), 1_i4b, tmp(k+1_i4b:m), 1_i4b )
#else
                    tmp(k+1_i4b:m) = tmp(k+1_i4b:m) + v(j)*mat(k+1_i4b:m,j)
#endif
!                    
                end do
!
!$OMP CRITICAL (updateua)
!
#ifdef _BLAS
                call axpy( m-k, one, tmp(k+1_i4b:m), 1_i4b, w(k+1_i4b:m), 1_i4b )
#else
                w(k+1_i4b:m) = w(k+1_i4b:m) + tmp(k+1_i4b:m)  
#endif
!
!$OMP END CRITICAL (updateua)
!
!$OMP BARRIER
!
            end do
!
!$OMP END PARALLEL
!
        else
#endif
!
!           START THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            u2(2_i4b:m) = tau*matmul( mat(2_i4b:m,1_i4b:n), v(1_i4b:n) )
!
            u(2_i4b:m) = mat(2_i4b:m,1_i4b) + u2(2_i4b:m) 
!
!           GENERATE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(3:m,1).
!
            call hous1( u(2_i4b:m), tau, e(2_i4b) )
!
!           SAVE ELEMENTARY REFLECTOR H(1)
!
            mat(2_i4b:m,1_i4b) = u(2_i4b:m)
            tauq(1_i4b)        = tau
!
#ifdef _BLAS
            prod = dot( m-1_i4b, u(2_i4b:m), 1_i4b, u2(2_i4b:m), 1_i4b )
#else
            prod = dot_product( u(2_i4b:m), u2(2_i4b:m) )
#endif
!
!           FINISH THE FIRST RANK-2 COLUMN-ROW UPDATE AND
!           START THE SECOND RANK-2 COLUMN-ROW UPDATE.
!
            tmp(2_i4b:m) = zero
!
            do j = 2_i4b, n
!
#ifdef _BLAS
                alpha = tau*( dot( m-1_i4b, u(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b ) + prod*v(j) )
#else
                alpha = tau*( dot_product( u(2_i4b:m), mat(2_i4b:m,j) ) + prod*v(j) )
#endif
!
!               APPLY G(1) TO mat(2:m,2:n) FROM THE RIGHT AND
!               APPLY H(1) TO mat(2:m,2:n) FROM THE LEFT.
!
#ifdef _BLAS
                call axpy( m-1_i4b, alpha, u(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b )
                call axpy( m-1_i4b, v(j), u2(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b )
#else
                mat(2_i4b:m,j) = mat(2_i4b:m,j) + alpha*u(2_i4b:m) + v(j)*u2(2_i4b:m)
#endif
!
!               START THE SECOND RANK-2 COLUMN-ROW UPDATE.
!
                v(j) = mat(2_i4b,j)
#ifdef _BLAS
                call axpy( m-2_i4b, v(j), mat(3_i4b:m,j), 1_i4b, tmp(3_i4b:m), 1_i4b )
#else
                tmp(3_i4b:m) = tmp(3_i4b:m) + v(j)*mat(3_i4b:m,j)
#endif
!                    
            end do
! 
            do i = 2_i4b, m-1_i4b
!
                k = i + 1_i4b
!
                alpha = v(i)
!
!               GENERATE ELEMENTARY REFLECTOR G(i) TO ANNIHILATE mat(i,i+1:n).
!
                call hous1( v(i:n), tau, prod )
!
!               SAVE ELEMENTARY REFLECTOR G(i).
!
                mat(i,i:n) = v(i:n)
                d(i)       = prod
                taup(i)    = tau
!
!               CONTINUE THE i RANK-2 COLUMN-ROW UPDATE.
!
                if ( tau/=zero ) then
                    beta    = tau/(alpha-prod)
                    u2(k:m) = beta*( tmp(k:m) - prod*mat(k:m,i) )
                    u(k:m)  = mat(k:m,i) + u2(k:m) 
                else
                    u2(k:m) = zero
                    u(k:m)  = mat(k:m,i)
                end if
!
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+2:m,i).
!
                call hous1( u(k:m), tau, e(k) )
!
!               SAVE ELEMENTARY REFLECTOR H(i).
!
                mat(k:m,i) = u(k:m)
                tauq(i)    = tau
!
#ifdef _BLAS
                prod = dot( m-k+1_i4b, u(k:m), 1_i4b, u2(k:m), 1_i4b )
#else
                prod = dot_product( u(k:m), u2(k:m) )
#endif
!
!               FINISH THE i RANK-2 COLUMN-ROW UPDATE AND
!               START THE i+1 RANK-2 COLUMN-ROW UPDATE.
!
                tmp(k:m) = zero
!
                do j = k, n
!
#ifdef _BLAS
                    alpha= tau*( dot( m-k+1_i4b, u(k:m), 1_i4b, mat(k:m,j), 1_i4b ) + prod*v(j) )
#else
                    alpha= tau*( dot_product( u(k:m), mat(k:m,j) ) + prod*v(j) )
#endif
!
!                   APPLY G(i) TO mat(i+1:m,i+1:n) FROM THE RIGHT AND
!                   APPLY H(i) TO mat(i+1:m,i+1:n) FROM THE LEFT.
!
#ifdef _BLAS
                    call axpy( m-k+1_i4b, alpha, u(k:m), 1_i4b, mat(k:m,j), 1_i4b )
                    call axpy( m-k+1_i4b, v(j), u2(k:m), 1_i4b, mat(k:m,j), 1_i4b )
#else
                    mat(k:m,j) = mat(k:m,j) + alpha*u(k:m) + v(j)*u2(k:m)
#endif
!
!                   START THE i+1 RANK-2 COLUMN-ROW UPDATE.
!
                    v(j) = mat(k,j)
#ifdef _BLAS
                    call axpy( m-k, v(j), mat(k+1_i4b:m,j), 1_i4b, tmp(k+1_i4b:m), 1_i4b )
#else
                    tmp(k+1_i4b:m) = tmp(k+1_i4b:m) + v(j)*mat(k+1_i4b:m,j)
#endif
!                    
                end do
!                    
            end do
!
#ifdef _OPENMP
        end if
#endif
!
!       GENERATE ELEMENTARY REFLECTOR G(m) TO ANNIHILATE mat(m,m+1:n).
!
        call hous1( v(m:n), taup(m), d(m) )
!
!       SAVE ELEMENTARY REFLECTOR G(m).
!
        mat(m,m:n) = v(m:n)
!
        tauq(m)    = zero
!
    end if
!
!
! END OF SUBROUTINE bd_cmp_a
! __________________________
!
    end subroutine bd_cmp_a
!
! =========================================================================================
!
    subroutine bd_cmp_b( mat, d, e, tauq )
!
! Purpose
! _______
!
!   BD_CMP reduces a general m-by-n matrix MAT to upper or lower bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal. If:
!
!   - m >= n, BD is upper bidiagonal;
!   - m < n, BD is lower bidiagonal.
!
!   BD_CMP computes only BD and Q.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the general m-by-n matrix to be reduced.
!
!            On exit, if:
!
!            - m >= n, the elements on and below the diagonal,
!              with the array TAUQ, represent the orthogonal
!              matrix Q as a product of elementary reflectors, 
!              and the elements above the diagonal are destroyed;
!            - m < n, the elements below the diagonal, with
!              the array TAUQ, represent the orthogonal
!              matrix Q as a product of elementary reflectors,
!              and the elements on and above the diagonal are
!              destroyed.
!
!            See Further Details.
!
!   D        (OUTPUT) real(stnd), dimension(:)
!            The diagonal elements of the bidiagonal matrix BD
!
!            The size of D must be min( size(MAT,1) , size(MAT,2) ).
!
!   E        (OUTPUT) real(stnd), dimension(:)
!            The off-diagonal elements of the bidiagonal matrix BD:
!
!            - if m >= n, E(i) = BD(i-1,i) for i = 2,3,...,n;
!            - if m < n,  E(i) = BD(i,i-1) for i = 2,3,...,m.
!
!            The size of E must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUQ     (OUTPUT) real(stnd), dimension(:)
!            The scalar factors of the elementary reflectors which
!            represent the orthogonal matrix Q. See Further Details.
!
!            The size of TAUQ must be min( size(MAT,1) , size(MAT,2) ).
!
!
! Further Details
! _______________
!
!   The matrice Q is represented as products of elementary reflectors:
!
!   If m >= n,
!
!       Q = H(1) * H(2) * . . . * H(n)
!
!   Each H(i) has the form:
!
!       H(i) = I + tauq * u * u' 
!
!   where tauq is a real scalar and u is a real vector;
!   u(1:i-1) = 0 and u(i:m) is stored on exit in MAT(i:m,i);
!   tauq is stored in TAUQ(i).
!
!   If m < n,
!
!       Q = H(1) * H(2) * . . . * H(m-1)
!
!   Each H(i) has the form:
!
!       H(i) = I + tauq * u * u'
!
!   where tauq is a real scalar and u is a real vector;
!   u(1:i) = 0 and u(i+1:m) is stored on exit in MAT(i+1:m,i);
!   tauq is stored in TAUQ(i).
!
!   The contents of MAT on exit are illustrated by the following examples:
!
!   m = 6 and n = 5 (m > n):
!
!        (  u1  xx  xx  xx  xx )
!
!        (  u1  u2  xx  xx  xx )
!
!        (  u1  u2  u3  xx  xx )
!
!        (  u1  u2  u3  u4  xx )
!
!        (  u1  u2  u3  u4  u5 )
!
!        (  u1  u2  u3  u4  u5 )
!
!   m = 5 and n = 6 (m < n):
!
!        (  xx  xx  xx  xx  xx  xx )
!
!        (  u1  xx  xx  xx  xx  xx )
!
!        (  u1  u2  xx  xx  xx  xx )
!
!        (  u1  u2  u3  xx  xx  xx )
!
!        (  u1  u2  u3  u4  xx  xx )
!
!   where ui denotes an element of the vector defining H(i). The upper triangular
!   part of MAT is destroyed on exit.
!
!   This subroutine is adapted from the routine DGEBD2 in LAPACK.
!   An efficient variant of the classic Golub and Kahan Householder bidiagonalization
!   algorithm is used. This variant reduces the traffic on the data bus from four reads
!   and two writes per column-row elimination of the bidiagonalization process to one
!   read and one write. Furthermore, the algorithm is parallelized if OPENMP is used.
!
!   For further details on the bidiagonal reduction algorithm and its use
!   or the efficient variant used here, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Howell, G.W., Demmel, J., Fulton, C.T., Hammarling, S., and Marmol, K., 2008:
!          Cache efficient bidiagonalization using BLAS 2.5 operators.
!          ACM Transactions on Mathematical Software (TOMS) Volume 34, Issue 3.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
    use Hous_Procedures, only : hous1
#ifdef _BLAS
    use BLAS_interfaces, only : dot, axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: d, e, tauq
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: prod, tau, beta, alpha
    real(stnd), dimension(size(mat,1))   :: u, u2, tmp
    real(stnd), dimension(size(mat,2))   :: v, v2
!
    integer(i4b) :: m, n, k, i, j
!
#ifdef _OPENMP
!
    real(stnd), dimension(size(mat,1))   :: w
!
    integer(i4b) :: nn, max_num_threads, num_threads, thread_num, jinc, jrem, joffset, j1, j2
!
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k =  assert_eq( (/ min(m,n)            ,   &
                       int(size(tauq),i4b) ,   &
                       int(size(d),i4b)    ,   &
                       int(size(e),i4b) /) ,   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
    e(1_i4b) = zero
!
    if ( k==1_i4b ) then
!
!       SPECIAL CASES IF mat(:,:) IS A ROW OR COLUMN VECTOR.
!
        if ( n==1_i4b ) then
!
!           GENERATE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(2:m,1).
!
            call hous1( mat(1_i4b:m,1_i4b), tauq(1_i4b), d(1_i4b) )
!
        else
!
!           GENERATE ELEMENTARY REFLECTOR G(1) TO ANNIHILATE mat(1,2:n).
!
            call hous1( mat(1_i4b,1_i4b:n), tau, d(1_i4b) )
!
            tauq(1_i4b) = zero         
!
        end if
!
        return
!
    end if
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m*n>=omp_limit                      .and.       &
                  i>1_i4b
!
    max_num_threads = min( i, max_num_threads_bd_cmp )
#endif
!
    if ( m>=n ) then
!
!       REDUCE mat TO UPPER BIDIAGONAL FORM.
!
!       GENERATE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(2:m,1).
!
        call hous1( mat(1_i4b:m,1_i4b), tau, d(1_i4b) )
        tauq(1_i4b) = tau
!
#ifdef _OPENMP
        if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(i,k,nn,num_threads,thread_num,jinc,jrem,joffset,j,j1,j2,tmp,u,w,alpha),  &
!$OMP          NUM_THREADS(max_num_threads)
!
            num_threads = omp_get_num_threads()
            thread_num = omp_get_thread_num()
!
!           MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(1).
!
            u(1_i4b:m) = mat(1_i4b:m,1_i4b)
!
!           DISTRIBUTE THE WORK AMONG THE THREADS FOR THE FIRST
!           RANK-2 COLUMN-ROW UPDATE.
!
            nn = n - 1_i4b
            jrem = mod( nn, num_threads )
            jinc = ( nn - jrem )/num_threads
!
            if ( thread_num<jrem ) then
                jinc    = jinc + 1_i4b
                joffset = 2_i4b
            else
                joffset = jrem + 2_i4b
            end if
!
            j1 = thread_num*jinc + joffset
            j2 = j1 + jinc - 1_i4b
!
!           START THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            tmp(2_i4b:m) = zero
!
            do j = j1, j2
!
#ifdef _BLAS
                v2(j) = tau*dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,j), 1_i4b )
#else
                v2(j) = tau*dot_product( u(1_i4b:m), mat(1_i4b:m,j) )
#endif
                v(j)         = mat(1_i4b,j) + v2(j)
#ifdef _BLAS
                call axpy( m-1_i4b, v(j), mat(2_i4b:m,j), 1_i4b, tmp(2_i4b:m), 1_i4b )
#else
                tmp(2_i4b:m) = tmp(2_i4b:m) + v(j)*mat(2_i4b:m,j)
#endif
!
            end do
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!           GENERATE ELEMENTARY REFLECTOR G(1) TO ANNIHILATE mat(1,3:n).
!
            alpha = v(2_i4b)
!
            call hous1( v(2_i4b:n), tau, prod )
            e(2_i4b)    = prod
!
!           CONTINUE THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            if ( tau/=zero ) then
                beta        = tau/(alpha-prod)
                u2(2_i4b:m) = - mat(2_i4b:m,2_i4b)
            else
                beta        = zero
                u2(2_i4b:m) = zero
            end if
!
            prod = zero
!
!$OMP END SINGLE
!
            do i = 2_i4b, n-1_i4b
!
                k = i + 1_i4b
!
!               CONTINUE THE i-1 RANK-2 COLUMN-ROW UPDATE.
!
                alpha    = dot_product( v2(j1:j2), v(j1:j2) )*tau
!
!$OMP CRITICAL (updateua)
!
#ifdef _BLAS
                call axpy( m-i+1_i4b, beta, tmp(i:m), 1_i4b, u2(i:m), 1_i4b )
#else
                u2(i:m) = u2(i:m) + beta*tmp(i:m) 
#endif
!
                prod = prod + alpha
!
!$OMP END CRITICAL (updateua)
!
!$OMP BARRIER
!
                v2(j1:j2) = v2(j1:j2) + prod*v(j1:j2)
!
!$OMP MASTER
!
!               APPLY H(i-1) TO mat(i:m,i) FROM THE LEFT AND 
!               APPLY G(i-1)   TO mat(i:m,i) FROM THE RIGHT.
!
#ifdef _BLAS
                w(i:m) = mat(i:m,i)
                call axpy( m-i+1_i4b, v2(i), u(i:m), 1_i4b, w(i:m), 1_i4b )
                call axpy( m-i+1_i4b, v(i), u2(i:m), 1_i4b, w(i:m), 1_i4b )
#else
                w(i:m) = mat(i:m,i) + v2(i)*u(i:m) + v(i)*u2(i:m)
#endif
!
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+1:m,i).
!
                call hous1( w(i:m), tau, d(i) )
!
!               SAVE ELEMENTARY REFLECTOR H(i).
!
                mat(i:m,i) = w(i:m)
                tauq(i)    = tau
!
!$OMP END MASTER
!
!$OMP BARRIER
!
                if ( thread_num/=0_i4b ) then
!
!                   MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(i).
!
                    w(i:m) = mat(i:m,i)
!
                end if
!
!               DISTRIBUTE THE WORK AMONG THE THREADS FOR THE i
!               RANK-2 COLUMN-ROW UPDATE.
!
                nn   = nn - 1_i4b
                jrem = mod( nn, num_threads )
                jinc = ( nn - jrem )/num_threads
!
                if ( thread_num<jrem ) then
                    jinc    = jinc + 1_i4b
                    joffset = i + 1_i4b
                else
                    joffset = i + jrem + 1_i4b
                end if
!
                j1 = thread_num*jinc + joffset
                j2 = j1 + jinc - 1_i4b
!
!               FINISH THE i-1 RANK-2 COLUMN-ROW UPDATE AND
!               START THE i RANK-2 COLUMN-ROW UPDATE.
!
                tmp(k:m) = zero
!
                do j = j1, j2
!
!                   APPLY H(i-1) TO mat(i:m,i+1:n) FROM THE LEFT AND 
!                   APPLY G(i-1) TO mat(i:m,i+1:n) FROM THE RIGHT.
!
#ifdef _BLAS
                    call axpy( m-i+1_i4b, v2(j), u(i:m), 1_i4b, mat(i:m,j), 1_i4b )
                    call axpy( m-i+1_i4b, v(j), u2(i:m), 1_i4b, mat(i:m,j), 1_i4b )
#else
                    mat(i:m,j) = mat(i:m,j) + v2(j)*u(i:m) + v(j)*u2(i:m)
#endif
!
!                   START THE i RANK-2 COLUMN-ROW UPDATE.
!
#ifdef _BLAS
                    v2(j) = tau*dot( m-i+1_i4b, w(i:m), 1_i4b, mat(i:m,j), 1_i4b )
#else
                    v2(j) = tau*dot_product( w(i:m), mat(i:m,j) )
#endif
                    v(j)  = mat(i,j) + v2(j)
#ifdef _BLAS
                    call axpy( m-k+1_i4b, v(j), mat(k:m,j), 1_i4b, tmp(k:m), 1_i4b )
#else
                    tmp(k:m) = tmp(k:m) + v(j)*mat(k:m,j)
#endif
!
                end do
!
!               SAVE A LOCAL COPY OF ELEMENTARY REFLECTOR H(i).
!
                u(i:m) = w(i:m)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!               GENERATE ELEMENTARY REFLECTOR G(i) TO ANNIHILATE mat(i,i+2:n)
!
                alpha = v(k)
!
                call hous1( v(k:n), tau, prod )
                e(k)    = prod
!
!               CONTINUE THE i RANK-2 COLUMN-ROW UPDATE.
!
                if ( tau/=zero ) then
                    beta    = tau/(alpha-prod)
                    u2(k:m) = - mat(k:m,k)
                else
                    beta    = zero
                    u2(k:m) = zero
                end if
!
                prod = zero
!
!$OMP END SINGLE
!
            end do
!
!$OMP SINGLE
!
!           CONTINUE THE n-1 RANK-2 COLUMN-ROW UPDATE.
!
            prod = v2(n)*tau
!
            v2(n) = v2(n) + prod*v(n)
!
!$OMP END SINGLE
!
!$OMP END PARALLEL
!
        else
#endif
!
!           MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(1).
!
            u(1_i4b:m) = mat(1_i4b:m,1_i4b)
!
!           START THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            tmp(2_i4b:m) = zero
!
            do j = 2_i4b, n
!
#ifdef _BLAS
                v2(j) = tau*dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,j), 1_i4b )
#else
                v2(j) = tau*dot_product( u(1_i4b:m), mat(1_i4b:m,j) )
#endif
                v(j)  = mat(1_i4b,j) + v2(j)
#ifdef _BLAS
                call axpy( m-1_i4b, v(j), mat(2_i4b:m,j), 1_i4b, tmp(2_i4b:m), 1_i4b )
#else
                tmp(2_i4b:m) = tmp(2_i4b:m) + v(j)*mat(2_i4b:m,j)
#endif
!
            end do
!
!           GENERATE ELEMENTARY REFLECTOR G(1) TO ANNIHILATE mat(1,3:n).
!
            alpha = v(2_i4b)
!
            call hous1( v(2_i4b:n), tau, prod )
!
            e(2_i4b) = prod
!
!           CONTINUE THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            if ( tau/=zero ) then
!
                beta        = tau/(alpha-prod)
                u2(2_i4b:m) = beta*tmp(2_i4b:m) - mat(2_i4b:m,2_i4b)
!
#ifdef _BLAS
                prod = dot( n-1_i4b, v2(2_i4b:n), 1_i4b, v(2_i4b:n), 1_i4b )*tau
!
                call axpy( n-1_i4b, prod, v(2_i4b:n), 1_i4b, v2(2_i4b:n), 1_i4b )
#else
                prod = dot_product( v2(2_i4b:n), v(2_i4b:n) )*tau
!
                v2(2_i4b:n) = v2(2_i4b:n) + prod*v(2_i4b:n)
#endif
!
            else
!
                u2(2_i4b:m) = zero
!
            end if
!
            do i = 2_i4b, n-1_i4b
!
                k = i + 1_i4b
!
!               APPLY H(i-1) TO mat(i:m,i) FROM THE LEFT AND 
!               APPLY G(i-1)   TO mat(i:m,i) FROM THE RIGHT.
!
#ifdef _BLAS
                tmp(i:m) = mat(i:m,i)
                call axpy( m-i+1_i4b, v2(i), u(i:m), 1_i4b, tmp(i:m), 1_i4b )
                call axpy( m-i+1_i4b, v(i), u2(i:m), 1_i4b, tmp(i:m), 1_i4b )
#else
                tmp(i:m) = mat(i:m,i) + v2(i)*u(i:m) + v(i)*u2(i:m)
#endif
!
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+1:m,i).
!
                call hous1( tmp(i:m), tau, d(i) )
!
!               SAVE ELEMENTARY REFLECTOR H(i).
!
                mat(i:m,i) = tmp(i:m)
                tauq(i)    = tau
!
!               FINISH THE i-1 RANK-2 COLUMN-ROW UPDATE AND
!               START THE i RANK-2 COLUMN-ROW UPDATE.
!
                tmp(k:m) = zero
!
                do j = k, n
!
!                   APPLY H(i-1) TO mat(i:m,i+1:n) FROM THE LEFT AND 
!                   APPLY G(i-1) TO mat(i:m,i+1:n) FROM THE RIGHT.
!
#ifdef _BLAS
                    call axpy( m-i+1_i4b, v2(j), u(i:m), 1_i4b, mat(i:m,j), 1_i4b )
                    call axpy( m-i+1_i4b, v(j), u2(i:m), 1_i4b, mat(i:m,j), 1_i4b )
#else
                    mat(i:m,j) = mat(i:m,j) + v2(j)*u(i:m) + v(j)*u2(i:m)
#endif
!
!                   START THE i RANK-2 COLUMN-ROW UPDATE.
!
#ifdef _BLAS
                    v2(j) = tau*dot( m-i+1_i4b,  mat(i:m,i), 1_i4b, mat(i:m,j), 1_i4b )
#else
                    v2(j) = tau*dot_product( mat(i:m,i), mat(i:m,j) )
#endif
                    v(j)  = mat(i,j) + v2(j)
#ifdef _BLAS
                    call axpy( m-k+1_i4b, v(j), mat(k:m,j), 1_i4b, tmp(k:m), 1_i4b )
#else
                    tmp(k:m) = tmp(k:m) + v(j)*mat(k:m,j)
#endif
!
                end do
!
!               MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(i).
!
                u(i:m) = mat(i:m,i)
!
!               GENERATE ELEMENTARY REFLECTOR G(i) TO ANNIHILATE mat(i,i+2:n)
!
                alpha = v(k)
!
                call hous1( v(k:n), tau, prod )
!
                e(k) = prod
!
!               CONTINUE THE i RANK-2 COLUMN-ROW UPDATE.
!
                if ( tau/=zero ) then
!
                    beta    = tau/(alpha-prod)
                    u2(k:m) = beta*tmp(k:m) - mat(k:m,k)
!
#ifdef _BLAS
                    prod = dot( n-k+1_i4b, v2(k:n), 1_i4b, v(k:n), 1_i4b )*tau
!
                    call axpy( n-k+1_i4b, prod, v(k:n), 1_i4b, v2(k:n), 1_i4b )
#else
                    prod    = dot_product( v2(k:n), v(k:n) )*tau
!
                    v2(k:n) = v2(k:n) + prod*v(k:n)
#endif
!
                else
!
                    u2(k:m) = zero
!
                end if
!
            end do
!
#ifdef _OPENMP
        end if
#endif
!
!       FINISH THE n-1 RANK-2 COLUMN-ROW UPDATE.
!
!       APPLY H(n-1) TO mat(n:m,n) FROM THE LEFT AND 
!       APPLY G(n-1) TO mat(n:m,n) FROM THE RIGHT.
!
#ifdef _BLAS
        call axpy( m-n+1_i4b, v2(n), mat(n:m,n-1_i4b), 1_i4b, mat(n:m,n), 1_i4b )
#else
        mat(n:m,n) = mat(n:m,n) + v2(n)*mat(n:m,n-1_i4b)
#endif
!
!       GENERATE ELEMENTARY REFLECTOR H(n).
!
        call hous1( mat(n:m,n), tauq(n), d(n) )
!
    else
!
!       REDUCE mat TO LOWER BIDIAGONAL FORM.
!
        v(1_i4b:n) = mat(1_i4b,1_i4b:n)
!
!       GENERATE ELEMENTARY REFLECTOR G(1) TO ANNIHILATE mat(1,2:n).
!
        call hous1( v(1_i4b:n), tau, d(1_i4b) )
!
#ifdef _OPENMP
!
        if ( test_par ) then
!
            u2(2_i4b:m)  = mat(2_i4b:m,1_i4b)
!
!$OMP PARALLEL PRIVATE(i,k,nn,num_threads,thread_num,jinc,jrem,joffset,j,j1,j2,tmp,u,alpha),  &
!$OMP          NUM_THREADS(max_num_threads)
!
            num_threads = omp_get_num_threads()
            thread_num  = omp_get_thread_num()
!
!           DISTRIBUTE THE WORK AMONG THE THREADS FOR THE FIRST
!           RANK-2 COLUMN-ROW UPDATE.
!
            nn   = n - 1_i4b
            jrem = mod( nn, num_threads )
            jinc = ( nn - jrem )/num_threads
!
            if ( thread_num<jrem ) then
                jinc    = jinc + 1_i4b
                joffset = 2_i4b
            else
                joffset = jrem + 2_i4b
            end if
!
            j1 = thread_num*jinc + joffset
            j2 = j1 + jinc - 1_i4b
!
!           START THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            if ( j1<=j2 ) then
!
                tmp(2_i4b:m) = matmul( mat(2_i4b:m,j1:j2), v(j1:j2) )
!
!$OMP CRITICAL (updateua)
!
#ifdef _BLAS
                call axpy( m-1_i4b, one, tmp(2_i4b:m), 1_i4b, u2(2_i4b:m), 1_i4b )
#else
                u2(2_i4b:m) = u2(2_i4b:m) + tmp(2_i4b:m)  
#endif
!
!$OMP END CRITICAL (updateua)
!
            end if
!
!$OMP BARRIER
!
!$OMP MASTER
!
            u2(2_i4b:m) = tau*u2(2_i4b:m) 
            u(2_i4b:m)  = mat(2_i4b:m,1_i4b) + u2(2_i4b:m) 
!
!           GENERATE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(3:m,1).
!
            call hous1( u(2_i4b:m), tau, e(2_i4b) )
!
!           SAVE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(3:m,1).
!
            mat(2_i4b:m,1_i4b ) = u(2_i4b:m)
            tauq(1_i4b)         = tau
!
#ifdef _BLAS
            prod = dot( m-1_i4b, u(2_i4b:m), 1_i4b, u2(2_i4b:m), 1_i4b )
#else
            prod = dot_product( u(2_i4b:m), u2(2_i4b:m) )
#endif
!
            w(2_i4b:m) = zero
!
!$OMP END MASTER
!
!$OMP BARRIER
!
            if ( thread_num/=0_i4b ) then
!
!              MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(1).
!
               u(2_i4b:m)  = mat(2_i4b:m,1_i4b)
!
            end if
!
!           FINISH THE FIRST RANK-2 COLUMN-ROW UPDATE AND
!           START THE SECOND RANK-2 COLUMN-ROW UPDATE.
!
            tmp(2_i4b:m) = zero
!
            do j = j1, j2
!
#ifdef _BLAS
                alpha = tau*( dot( m-1_i4b, u(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b ) + prod*v(j) )
#else
                alpha = tau*( dot_product( u(2_i4b:m), mat(2_i4b:m,j) ) + prod*v(j) )
#endif
!
!               APPLY G(1) TO mat(2:m,2:n) FROM THE RIGHT AND
!               APPLY H(1) TO mat(2:m,2:n) FROM THE LEFT.
!
#ifdef _BLAS
                call axpy( m-1_i4b, alpha, u(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b )
                call axpy( m-1_i4b, v(j), u2(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b )
#else
                mat(2_i4b:m,j) = mat(2_i4b:m,j) + alpha*u(2_i4b:m) + v(j)*u2(2_i4b:m)
#endif
!
!               START THE SECOND RANK-2 COLUMN-ROW UPDATE.
!
                v(j) = mat(2_i4b,j)
#ifdef _BLAS
                call axpy( m-2_i4b, v(j), mat(3_i4b:m,j), 1_i4b, tmp(3_i4b:m), 1_i4b )
#else
                tmp(3_i4b:m) = tmp(3_i4b:m) + v(j)*mat(3_i4b:m,j)
#endif
!                    
            end do
!
!$OMP CRITICAL (updateua)
!
#ifdef _BLAS
            call axpy( m-2_i4b, one, tmp(3_i4b:m), 1_i4b, w(3_i4b:m), 1_i4b )
#else
            w(3_i4b:m) = w(3_i4b:m) + tmp(3_i4b:m)  
#endif
!
!$OMP END CRITICAL (updateua)
!
!$OMP BARRIER
!
            do i = 2_i4b, m-1_i4b
!
                k = i + 1_i4b
!
!               DISTRIBUTE THE WORK AMONG THE THREADS FOR THE i
!               RANK-2 COLUMN-ROW UPDATE.
!
                nn   = nn - 1_i4b
                jrem = mod( nn, num_threads )
                jinc = ( nn - jrem )/num_threads
!
                if ( thread_num<jrem ) then
                    jinc    = jinc + 1_i4b
                    joffset = i + 1_i4b
                else
                    joffset = i + jrem + 1_i4b
                end if
!
                j1 = thread_num*jinc + joffset
                j2 = j1 + jinc - 1_i4b
!
!$OMP MASTER
!
                alpha = v(i)
!
!               GENERATE ELEMENTARY REFLECTOR G(i) TO ANNIHILATE mat(i,i+1:n).
!
                call hous1( v(i:n), tau, prod )
!
                d(i) = prod
!
!               CONTINUE THE i RANK-2 COLUMN-ROW UPDATE.
!
                if ( tau/=zero ) then
                    beta    = tau/(alpha-prod)
                    u2(k:m) = beta*( w(k:m) - prod*mat(k:m,i) )
                    u(k:m)  = mat(k:m,i) + u2(k:m) 
                else
                    u2(k:m) = zero
                    u(k:m)  = mat(k:m,i)
                end if
!
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+2:m,i).
!
                call hous1( u(k:m), tau, e(k) )
!
!               SAVE ELEMENTARY REFLECTOR H(i).
!
                mat(k:m,i) = u(k:m)
                tauq(i)    = tau
!
#ifdef _BLAS
                prod = dot( m-k+1_i4b, u(k:m), 1_i4b, u2(k:m), 1_i4b )
#else
                prod = dot_product( u(k:m), u2(k:m) )
#endif
!
                w(k:m) = zero
!
!$OMP END MASTER
!
!$OMP BARRIER
!
                if ( thread_num/=0_i4b ) then
!
!                  MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(i).
!
                   u(k:m)  = mat(k:m,i)
!
                end if
!
!               FINISH THE i RANK-2 COLUMN-ROW UPDATE AND
!               START THE i+1 RANK-2 COLUMN-ROW UPDATE.
!
                tmp(k:m) = zero
!
                do j = j1, j2
!
#ifdef _BLAS
                    alpha= tau*( dot( m-k+1_i4b, u(k:m), 1_i4b, mat(k:m,j), 1_i4b ) + prod*v(j) )
#else
                    alpha= tau*( dot_product( u(k:m), mat(k:m,j) ) + prod*v(j) )
#endif
!
!                   APPLY G(i) TO mat(i+1:m,i+1:n) FROM THE RIGHT AND
!                   APPLY H(i) TO mat(i+1:m,i+1:n) FROM THE LEFT.
!
#ifdef _BLAS
                    call axpy( m-k+1_i4b, alpha, u(k:m), 1_i4b, mat(k:m,j), 1_i4b )
                    call axpy( m-k+1_i4b, v(j), u2(k:m), 1_i4b, mat(k:m,j), 1_i4b )
#else
                    mat(k:m,j) = mat(k:m,j) + alpha*u(k:m) + v(j)*u2(k:m)
#endif
!
!                   START THE i+1 RANK-2 COLUMN-ROW UPDATE.
!
                    v(j) = mat(k,j)
#ifdef _BLAS
                    call axpy( m-k, v(j), mat(k+1_i4b:m,j), 1_i4b, tmp(k+1_i4b:m), 1_i4b )
#else
                    tmp(k+1_i4b:m) = tmp(k+1_i4b:m) + v(j)*mat(k+1_i4b:m,j)
#endif
!                    
                end do
!
!$OMP CRITICAL (updateua)
!
#ifdef _BLAS
                call axpy( m-k, one, tmp(k+1_i4b:m), 1_i4b, w(k+1_i4b:m), 1_i4b )
#else
                w(k+1_i4b:m) = w(k+1_i4b:m) + tmp(k+1_i4b:m)  
#endif
!
!$OMP END CRITICAL (updateua)
!
!$OMP BARRIER
!
            end do
!
!$OMP END PARALLEL
!
        else
#endif
!
!           START THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            u2(2_i4b:m) = tau*matmul( mat(2_i4b:m,1_i4b:n), v(1_i4b:n) )
!
            u(2_i4b:m) = mat(2_i4b:m,1_i4b) + u2(2_i4b:m) 
!
!           GENERATE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(3:m,1).
!
            call hous1( u(2_i4b:m), tau, e(2_i4b) )
!
!           SAVE ELEMENTARY REFLECTOR H(1)
!
            mat(2_i4b:m,1_i4b) = u(2_i4b:m)
            tauq(1_i4b)        = tau
!
#ifdef _BLAS
            prod = dot( m-1_i4b, u(2_i4b:m), 1_i4b, u2(2_i4b:m), 1_i4b )
#else
            prod = dot_product( u(2_i4b:m), u2(2_i4b:m) )
#endif
!
!           FINISH THE FIRST RANK-2 COLUMN-ROW UPDATE AND
!           START THE SECOND RANK-2 COLUMN-ROW UPDATE.
!
            tmp(2_i4b:m) = zero
!
            do j = 2_i4b, n
!
#ifdef _BLAS
                alpha = tau*( dot( m-1_i4b, u(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b ) + prod*v(j) )
#else
                alpha = tau*( dot_product( u(2_i4b:m), mat(2_i4b:m,j) ) + prod*v(j) )
#endif
!
!               APPLY G(1) TO mat(2:m,2:n) FROM THE RIGHT AND
!               APPLY H(1) TO mat(2:m,2:n) FROM THE LEFT.
!
#ifdef _BLAS
                call axpy( m-1_i4b, alpha, u(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b )
                call axpy( m-1_i4b, v(j), u2(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b )
#else
                mat(2_i4b:m,j) = mat(2_i4b:m,j) + alpha*u(2_i4b:m) + v(j)*u2(2_i4b:m)
#endif
!
!               START THE SECOND RANK-2 COLUMN-ROW UPDATE.
!
                v(j) = mat(2_i4b,j)
#ifdef _BLAS
                call axpy( m-2_i4b, v(j), mat(3_i4b:m,j), 1_i4b, tmp(3_i4b:m), 1_i4b )
#else
                tmp(3_i4b:m) = tmp(3_i4b:m) + v(j)*mat(3_i4b:m,j)
#endif
!                    
            end do
! 
            do i = 2_i4b, m-1_i4b
!
                k = i + 1_i4b
!
                alpha = v(i)
!
!               GENERATE ELEMENTARY REFLECTOR G(i) TO ANNIHILATE mat(i,i+1:n).
!
                call hous1( v(i:n), tau, prod )
!
!               SAVE ELEMENTARY REFLECTOR G(i).
!
                d(i) = prod
!
!               CONTINUE THE i RANK-2 COLUMN-ROW UPDATE.
!
                if ( tau/=zero ) then
                    beta    = tau/(alpha-prod)
                    u2(k:m) = beta*( tmp(k:m) - prod*mat(k:m,i) )
                    u(k:m)  = mat(k:m,i) + u2(k:m) 
                else
                    u2(k:m) = zero
                    u(k:m)  = mat(k:m,i)
                end if
!
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+2:m,i).
!
                call hous1( u(k:m), tau, e(k) )
!
!               SAVE ELEMENTARY REFLECTOR H(i).
!
                mat(k:m,i) = u(k:m)
                tauq(i)    = tau
!
#ifdef _BLAS
                prod = dot( m-k+1_i4b, u(k:m), 1_i4b, u2(k:m), 1_i4b )
#else
                prod = dot_product( u(k:m), u2(k:m) )
#endif
!
!               FINISH THE i RANK-2 COLUMN-ROW UPDATE AND
!               START THE i+1 RANK-2 COLUMN-ROW UPDATE.
!
                tmp(k:m) = zero
!
                do j = k, n
!
#ifdef _BLAS
                    alpha= tau*( dot( m-k+1_i4b, u(k:m), 1_i4b, mat(k:m,j), 1_i4b ) + prod*v(j) )
#else
                    alpha= tau*( dot_product( u(k:m), mat(k:m,j) ) + prod*v(j) )
#endif
!
!                   APPLY G(i) TO mat(i+1:m,i+1:n) FROM THE RIGHT AND
!                   APPLY H(i) TO mat(i+1:m,i+1:n) FROM THE LEFT.
!
#ifdef _BLAS
                    call axpy( m-k+1_i4b, alpha, u(k:m), 1_i4b, mat(k:m,j), 1_i4b )
                    call axpy( m-k+1_i4b, v(j), u2(k:m), 1_i4b, mat(k:m,j), 1_i4b )
#else
                    mat(k:m,j) = mat(k:m,j) + alpha*u(k:m) + v(j)*u2(k:m)
#endif
!
!                   START THE i+1 RANK-2 COLUMN-ROW UPDATE.
!
                    v(j) = mat(k,j)
#ifdef _BLAS
                    call axpy( m-k, v(j), mat(k+1_i4b:m,j), 1_i4b, tmp(k+1_i4b:m), 1_i4b )
#else
                    tmp(k+1_i4b:m) = tmp(k+1_i4b:m) + v(j)*mat(k+1_i4b:m,j)
#endif
!                    
                end do
!                    
            end do
!
#ifdef _OPENMP
        end if
#endif
!
        d(m) = -sign( norm(v(m:n)), v(m) )
!
        tauq(m) = zero
!
    end if
!
!
! END OF SUBROUTINE bd_cmp_b
! __________________________
!
    end subroutine bd_cmp_b
!
! =========================================================================================
!
    subroutine bd_cmp_c( mat, d, e, tauq, taup, rlmat, tauo )
!
! Purpose
! _______
!
!   BD_CMP reduces a general m-by-n matrix MAT to upper bidiagonal form BD
!   by a two-step algorithm:
!
!   - If m >= n, a QR factorization of the real m-by-n matrix MAT is first computed
!
!                        MAT = O * R
!
!   where O is orthogonal and R is upper triangular. In a second step, the n-by-n upper
!   triangular matrix R is reduced to upper bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * R * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix.
!
!   - If m < n, an LQ factorization of the real m-by-n matrix MAT is first computed
!
!                       MAT = L * O
!
!   where O is orthogonal and L is lower triangular. In a second step, the m-by-m lower
!   triangular matrix L is reduced to upper bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * L * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix.
!
!   BD_CMP computes O, BD, Q and P. The matrix O is stored in factored form
!   if the optional argument TAUO is present or explicitly computed if this
!   argument is absent.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the general m-by-n matrix to be reduced.
!
!            On exit, if:
!
!            - m >= n, the elements on and below the diagonal,
!              with the array TAUO, represent the orthogonal
!              matrix O of the QR factorization of MAT,
!              as a product of elementary reflectors,
!              if the argument TAUO is present. Otherwise,
!              the argument MAT contains the first n columns
!              of O on output.
!            - m < n, the elements on and above the diagonal,
!              with the array TAUO, represent the orthogonal
!              matrix O of the LQ factorization of MAT,
!              as a product of elementary reflectors,
!              if the argument TAUO is present. Otherwise,
!              the argument MAT contains the first m rows
!              of O on output.
!
!            See Further Details.
!
!   D        (OUTPUT) real(stnd), dimension(:)
!            The diagonal elements of the bidiagonal matrix BD
!
!            The size of D must be min( size(MAT,1) , size(MAT,2) ).
!
!   E        (OUTPUT) real(stnd), dimension(:)
!            The off-diagonal elements of the bidiagonal matrix BD:
!
!                       E(i) = BD(i-1,i) for i = 2,3,...,n;
!
!            The size of E must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUQ     (OUTPUT) real(stnd), dimension(:)
!            The scalar factors of the elementary reflectors which
!            represent the orthogonal matrix Q. See Further Details.
!
!            The size of TAUQ must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUP     (OUTPUT) real(stnd), dimension(:)
!            The scalar factors of the elementary reflectors which
!            represent the orthogonal matrix P. See Further Details.
!
!            The size of TAUP must be min( size(MAT,1) , size(MAT,2) ).
!
!   RLMAT    (OUTPUT) real(stnd), dimension(:,:)
!
!            On exit, the elements on and below the diagonal,
!            with the array TAUQ, represent the orthogonal
!            matrix Q as a product of elementary reflectors, 
!            and the elements above the diagonal, with the
!            array TAUP, represent the orthogonal matrix P
!            as a product of elementary reflectors;
!
!            See Further Details.
!
!            The shape of RLMAT must verify:
!
!            - size( RLMAT, 1 ) = size( RLMAT, 2 ) = min( size(MAT,1) , size(MAT,2) ).
!
!   TAUO     (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            The scalar factors of the elementary reflectors which
!            represent the orthogonal matrix O of the QR or LQ decomposition of MAT.
!
!            If the optional argument TAUO is present, the orthogonal matrix O is stored
!            in factored form, as a product of elementary reflectors, in the argument MAT on exit.
!
!            If the optional argument TAUO is absent, the orthogonal matrix O is explicitly generated
!            and stored in the argument MAT on exit.
!
!            See description of the argument MAT above and Further Details below.
!
!            The size of TAUO must be min( size(MAT,1) , size(MAT,2) ).
!            
!
! Further Details
! _______________
!
!   If m >= n, the matrix O of the QR factorization of MAT is represented as a product
!   of elementary reflectors
!
!            O = W(1) * W(2) * ... * W(n)
!
!   Each W(i) has the form
!
!            W(i) = I + tauo * ( v * v' ) ,
!                      
!   where tauo is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in MAT(i:m,i) and tauo in TAUO(i). If the optional argument
!   TAUO is absent, the first n columns of O are generated and stored in the argument MAT.
!
!   If m < n, The matrix O of the LQ factorization of MAT is represented as a product
!   of elementary reflectors
!
!           O = W(m) * . . . * W(2) * W(1)
!
!   Each W(i) has the form
!
!            W(i) = I + tauo * ( v * v' ) ,
!                      
!   where tauo is a real scalar and v is a real n-element vector with v(1:i-1) = 0.
!   v(i:n) is stored on exit in MAT(i,i:n) and tauo in TAUO(i).
!
!   A blocked algorithm is used for computing the QR or LQ factorization of MAT. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   After, the initial QR or LQ factorization of MAT, the (upper or lower) triangular
!   matrix is reduced to upper bidiagonal form BD.
!
!   The matrices Q and P of the bidiagonal factorization of the triangular matrix R or L
!   are represented as products of elementary reflectors:
!
!       Q = H(1) * H(2) * . . . * H(k)  and  P = G(1) * G(2) * . . . * G(k-1)
!
!   , where k = min( size(MAT,1) , size(MAT,2) ). Each H(i) and G(i) has the form:
!
!       H(i) = I + tauq * u * u'  and G(i) = I + taup * v * v'
!
!   where tauq and taup are real scalars, and u and v are real vectors;
!   u(1:i-1) = 0 and u(i:min(m,n)) is stored on exit in RLMAT(i:min(m,n),i);
!   v(1:i) = 0 and v(i+1:min(m,n)) is stored on exit in RLMAT(i,i+1:min(m,n));
!   tauq is stored in TAUQ(i) and taup in TAUP(i).
!
!   The contents of RLMAT on exit are illustrated by the following example:
!
!   m = 6 and n = 5 (m >= n):
!
!        (  u1  v1  v1  v1  v1 )
!
!        (  u1  u2  v2  v2  v2 )
!
!        (  u1  u2  u3  v3  v3 )
!
!        (  u1  u2  u3  u4  v4 )
!
!        (  u1  u2  u3  u4  u5 )
!
!   where ui denotes an element of the vector defining H(i), and vi an element 
!   of the vector defining G(i).
!
!   An efficient variant of the classic Golub and Kahan Householder bidiagonalization
!   algorithm is used. This variant reduces the traffic on the data bus from four reads
!   and two writes per column-row elimination of the bidiagonalization process to one
!   read and one write. Furthermore, the algorithm is parallelized if OPENMP is used.
!
!   For further details on the bidiagonal reduction algorithm and its use
!   or the variant used here, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Howell, G.W., Demmel, J., Fulton, C.T., Hammarling, S., and Marmol, K., 2008:
!           Cache efficient bidiagonalization using BLAS 2.5 operators.
!           ACM Transactions on Mathematical Software (TOMS) Volume 34, Issue 3.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,     only : assert, assert_eq
    use QR_Procedures, only : qr_cmp, lq_cmp, ortho_gen_lq, ortho_gen_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout),           dimension(:,:) :: mat
    real(stnd),   intent(out),             dimension(:,:) :: rlmat
    real(stnd),   intent(out),             dimension(:)   :: d, e, tauq, taup
    real(stnd),   intent(out), optional,   dimension(:)   :: tauo
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(d)) :: tau
!
    integer(i4b) :: m, n, k, i
!
    logical(lgl) :: do_gen_o
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k =  assert_eq( (/ min(m,n)            ,    &
                       int(size(tauq),i4b) ,    &
                       int(size(taup),i4b) ,    &
                       int(size(rlmat,1),i4b) , &
                       int(size(rlmat,2),i4b) , &
                       int(size(d),i4b)    ,    &
                       int(size(e),i4b) /) ,    &
                       name_proc )
!
    do_gen_o = true
!
    if ( present( tauo ) ) then
!
        call assert( logical( k==size(tauo), lgl ), name_proc )
!
        do_gen_o = false
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
    if ( m>=n ) then
!
!       PATH 1 : m EQUAL OR LARGER THAN n .
!
!       FIRST REDUCE mat USING THE QR DECOMPOSITION.
!
!       COMPUTE mat = O * R .
!
        call qr_cmp( mat, d, tau )
!
!       COPY R TO rlmat , ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
        do i = 1_i4b, n
!
            rlmat(1_i4b:i-1_i4b,i) = mat(1_i4b:i-1_i4b,i)
            rlmat(i,i) = d(i)
            rlmat(i+1_i4b:n,i) = zero
!
        end do
!
!$OMP END PARALLEL DO
!
        if ( do_gen_o ) then
!
!           GENERATE O IN mat, IF REQUIRED.
!
            call ortho_gen_qr( mat, tau )
!
        end if
!
    else
!
!       PATH 2 : n LARGER THAN m .
!
!       FIRST REDUCE mat USING THE LQ DECOMPOSITION.
!
!       COMPUTE mat = L * O .
!
        call lq_cmp( mat, d, tau )
!
!       COPY L TO rlmat , ZEROING OUT ABOVE IT.
!
!$OMP PARALLEL DO IF(m*m>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
        do i = 1_i4b, m
!
            rlmat(1_i4b:i-1_i4b,i) = zero
            rlmat(i,i) = d(i)
            rlmat(i+1_i4b:m,i) = mat(i+1_i4b:m,i)
!
        end do
!
!$OMP END PARALLEL DO
!
        if ( do_gen_o ) then
!
!           GENERATE O IN mat, IF REQUIRED.
!
            call ortho_gen_lq( mat, tau )
!
        end if
!
    end if
!
!   SECOND REDUCE R OR L TO UPPER BIDIAGONAL FORM.
!
    call bd_cmp_a( rlmat, d, e, tauq, taup )
!
    if ( .not.do_gen_o ) then
!
!       OUTPUT SCALAR FACTORS OF THE ELEMENTARY REFLECTORS
!       WHICH DEFINE O, IF REQUIRED.
!
        tauo(:k) = tau(:k)
!
    end if
!
!
! END OF SUBROUTINE bd_cmp_c
! __________________________
!
    end subroutine bd_cmp_c
!
! =========================================================================================
!
    subroutine bd_cmp_t( mat, d, e )
!
! Purpose
! _______
!
!   BD_CMP reduces a general m-by-n matrix MAT to upper or lower bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal. If:
!
!   - m >= n, BD is upper bidiagonal;
!   - m < n, BD is lower bidiagonal.
!
!   BD_CMP computes only BD and the matrices Q and P are not saved.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the general m-by-n matrix to be reduced.
!
!            On exit,  the general m-by-n matrix is destroyed.
!
!   D        (OUTPUT) real(stnd), dimension(:)
!            The diagonal elements of the bidiagonal matrix BD
!
!            The size of D must be min( size(MAT,1) , size(MAT,2) ).
!
!   E        (OUTPUT) real(stnd), dimension(:)
!            The off-diagonal elements of the bidiagonal matrix BD:
!
!            - if m >= n, E(i) = BD(i-1,i) for i = 2,3,...,n;
!            - if m < n,  E(i) = BD(i,i-1) for i = 2,3,...,m.
!
!            The size of E must be min( size(MAT,1) , size(MAT,2) ).
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine DGEBD2 in LAPACK.
!   An efficient variant of the classic Golub and Kahan Householder bidiagonalization
!   algorithm is used. This variant reduces the traffic on the data bus from four reads
!   and two writes per column-row elimination of the bidiagonalization process to one
!   read and one write. Furthermore, the algorithm is parallelized if OPENMP is used.
!
!   For further details on the bidiagonal reduction algorithm and its use
!   or the efficient variant used here, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Howell, G.W., Demmel, J., Fulton, C.T., Hammarling, S., and Marmol, K., 2008:
!           Cache efficient bidiagonalization using BLAS 2.5 operators.
!           ACM Transactions on Mathematical Software (TOMS) Volume 34, Issue 3.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
    use Hous_Procedures, only : hous1
#ifdef _BLAS
    use BLAS_interfaces, only : dot, axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: d, e
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: prod, tau, beta, alpha
    real(stnd), dimension(size(mat,1))   :: u, u2, tmp, w
    real(stnd), dimension(size(mat,2))   :: v, v2
!
    integer(i4b) :: m, n, k, i, j
!
#ifdef _OPENMP
!
    integer(i4b) :: nn, max_num_threads, num_threads, thread_num, jinc, jrem, joffset, j1, j2
!
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k =  assert_eq( (/ min(m,n)            ,   &
                       int(size(d),i4b)    ,   &
                       int(size(e),i4b) /) ,   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
    e(1_i4b) = zero
!
    if ( k==1_i4b ) then
!
!       SPECIAL CASES IF mat(:,:) IS A ROW OR COLUMN VECTOR.
!
        if ( n==1_i4b ) then
!
!           GENERATE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(2:m,1).
!
            call hous1( mat(1_i4b:m,1_i4b), tau, d(1_i4b) )
!
        else
!
!           GENERATE ELEMENTARY REFLECTOR G(1) TO ANNIHILATE mat(1,2:n).
!
            call hous1( mat(1_i4b,1_i4b:n), tau, d(1_i4b) )
!
        end if
!
        return
!
    end if
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m*n>=omp_limit                      .and.       &
                  i>1_i4b
!
    max_num_threads = min( i, max_num_threads_bd_cmp )
#endif
!
    if ( m>=n ) then
!
!       REDUCE mat TO UPPER BIDIAGONAL FORM.
!
!       GENERATE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(2:m,1).
!
        call hous1( mat(1_i4b:m,1_i4b), tau, d(1_i4b) )
!
#ifdef _OPENMP
        if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(i,k,nn,num_threads,thread_num,jinc,jrem,joffset,j,j1,j2,tmp,u,alpha),  &
!$OMP          NUM_THREADS(max_num_threads)
!
            num_threads = omp_get_num_threads()
            thread_num = omp_get_thread_num()
!
!           MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(1).
!
            u(1_i4b:m) = mat(1_i4b:m,1_i4b)
!
!           DISTRIBUTE THE WORK AMONG THE THREADS FOR THE FIRST
!           RANK-2 COLUMN-ROW UPDATE.
!
            nn   = n - 1_i4b
            jrem = mod( nn, num_threads )
            jinc = ( nn - jrem )/num_threads
!
            if ( thread_num<jrem ) then
                jinc    = jinc + 1_i4b
                joffset = 2_i4b
            else
                joffset = jrem + 2_i4b
            end if
!
            j1 = thread_num*jinc + joffset
            j2 = j1 + jinc - 1_i4b
!
!           START THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            tmp(2_i4b:m) = zero
!
            do j = j1, j2
!
#ifdef _BLAS
                v2(j) = tau*dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,j), 1_i4b )
#else
                v2(j) = tau*dot_product( u(1_i4b:m), mat(1_i4b:m,j) )
#endif
                v(j)  = mat(1_i4b,j) + v2(j)
#ifdef _BLAS
                call axpy( m-1_i4b, v(j), mat(2_i4b:m,j), 1_i4b, tmp(2_i4b:m), 1_i4b )
#else
                tmp(2_i4b:m) = tmp(2_i4b:m) + v(j)*mat(2_i4b:m,j)
#endif
!
            end do
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!           GENERATE ELEMENTARY REFLECTOR G(1) TO ANNIHILATE mat(1,3:n).
!
            alpha = v(2_i4b)
!
            call hous1( v(2_i4b:n), tau, prod )
            e(2_i4b)    = prod
!
!           CONTINUE THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            if ( tau/=zero ) then
                beta        = tau/(alpha-prod)
                u2(2_i4b:m) = - mat(2_i4b:m,2_i4b)
            else
                beta        = zero
                u2(2_i4b:m) = zero
            end if
!
            prod = zero
!
!$OMP END SINGLE
!
            do i = 2_i4b, n-1_i4b
!
                k = i + 1_i4b
!
!               CONTINUE THE i-1 RANK-2 COLUMN-ROW UPDATE.
!
                alpha    = dot_product( v2(j1:j2), v(j1:j2) )*tau
!
!$OMP CRITICAL (updateua)
!
#ifdef _BLAS
                call axpy( m-i+1_i4b, beta, tmp(i:m), 1_i4b, u2(i:m), 1_i4b )
#else
                u2(i:m) = u2(i:m) + beta*tmp(i:m) 
#endif
!
                prod = prod + alpha
!
!$OMP END CRITICAL (updateua)
!
!$OMP BARRIER
!
                v2(j1:j2) = v2(j1:j2) + prod*v(j1:j2)
!
!$OMP MASTER
!
!               APPLY H(i-1) TO mat(i:m,i) FROM THE LEFT AND 
!               APPLY G(i-1)   TO mat(i:m,i) FROM THE RIGHT.
!
#ifdef _BLAS
                w(i:m) = mat(i:m,i)
                call axpy( m-i+1_i4b, v2(i), u(i:m), 1_i4b, w(i:m), 1_i4b )
                call axpy( m-i+1_i4b, v(i), u2(i:m), 1_i4b, w(i:m), 1_i4b )
#else
                w(i:m) = mat(i:m,i) + v2(i)*u(i:m) + v(i)*u2(i:m)
#endif
!
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+1:m,i).
!
                call hous1( w(i:m), tau, d(i) )
!
!$OMP END MASTER
!
!$OMP BARRIER
!
!               DISTRIBUTE THE WORK AMONG THE THREADS FOR THE i
!               RANK-2 COLUMN-ROW UPDATE.
!
                nn = nn - 1_i4b
                jrem = mod( nn, num_threads )
                jinc = ( nn - jrem )/num_threads
!
                if ( thread_num<jrem ) then
                    jinc    = jinc + 1_i4b
                    joffset = i + 1_i4b
                else
                    joffset = i + jrem + 1_i4b
                end if
!
                j1 = thread_num*jinc + joffset
                j2 = j1 + jinc - 1_i4b
!
!               FINISH THE i-1 RANK-2 COLUMN-ROW UPDATE AND
!               START THE i RANK-2 COLUMN-ROW UPDATE.
!
                tmp(k:m) = zero
!
                do j = j1, j2
!
!                   APPLY H(i-1) TO mat(i:m,i+1:n) FROM THE LEFT AND 
!                   APPLY G(i-1) TO mat(i:m,i+1:n) FROM THE RIGHT.
!
#ifdef _BLAS
                    call axpy( m-i+1_i4b, v2(j), u(i:m), 1_i4b, mat(i:m,j), 1_i4b )
                    call axpy( m-i+1_i4b, v(j), u2(i:m), 1_i4b, mat(i:m,j), 1_i4b )
#else
                    mat(i:m,j) = mat(i:m,j) + v2(j)*u(i:m) + v(j)*u2(i:m)
#endif
!
!                   START THE i RANK-2 COLUMN-ROW UPDATE.
!
#ifdef _BLAS
                    v2(j) = tau*dot( m-i+1_i4b, w(i:m), 1_i4b, mat(i:m,j), 1_i4b )
#else
                    v2(j) = tau*dot_product( w(i:m), mat(i:m,j) )
#endif
                    v(j)  = mat(i,j) + v2(j)
#ifdef _BLAS
                    call axpy( m-k+1_i4b, v(j), mat(k:m,j), 1_i4b, tmp(k:m), 1_i4b )
#else
                    tmp(k:m) = tmp(k:m) + v(j)*mat(k:m,j)
#endif
!
                end do
!
!               SAVE A LOCAL COPY OF ELEMENTARY REFLECTOR H(i).
!
                u(i:m) = w(i:m)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!               GENERATE ELEMENTARY REFLECTOR G(i) TO ANNIHILATE mat(i,i+2:n)
!
                alpha = v(k)
!
                call hous1( v(k:n), tau, prod )
                e(k)    = prod
!
!               CONTINUE THE i RANK-2 COLUMN-ROW UPDATE.
!
                if ( tau/=zero ) then
                    beta    = tau/(alpha-prod)
                    u2(k:m) = - mat(k:m,k)
                else
                    beta    = zero
                    u2(k:m) = zero
                end if
!
                prod = zero
!
!$OMP END SINGLE
!
            end do
!
!$OMP SINGLE
!
!           CONTINUE THE n-1 RANK-2 COLUMN-ROW UPDATE.
!
            prod = v2(n)*tau
!
            v2(n) = v2(n) + prod*v(n)
!
!$OMP END SINGLE
!
!$OMP END PARALLEL
!
        else
#endif
!
!           MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(1).
!
            u(1_i4b:m) = mat(1_i4b:m,1_i4b)
!
!           START THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            tmp(2_i4b:m) = zero
!
            do j = 2_i4b, n
!
#ifdef _BLAS
                v2(j) = tau*dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,j), 1_i4b )
#else
                v2(j) = tau*dot_product( u(1_i4b:m), mat(1_i4b:m,j) )
#endif
                v(j)  = mat(1_i4b,j) + v2(j)
#ifdef _BLAS
                call axpy( m-1_i4b, v(j), mat(2_i4b:m,j), 1_i4b, tmp(2_i4b:m), 1_i4b )
#else
                tmp(2_i4b:m) = tmp(2_i4b:m) + v(j)*mat(2_i4b:m,j)
#endif
!
            end do
!
!           GENERATE ELEMENTARY REFLECTOR G(1) TO ANNIHILATE mat(1,3:n).
!
            alpha = v(2_i4b)
!
            call hous1( v(2_i4b:n), tau, prod )
!
            e(2_i4b) = prod
!
!           CONTINUE THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            if ( tau/=zero ) then
!
                beta        = tau/(alpha-prod)
                u2(2_i4b:m) = beta*tmp(2_i4b:m) - mat(2_i4b:m,2_i4b)
!
#ifdef _BLAS
                prod = dot( n-1_i4b, v2(2_i4b:n), 1_i4b, v(2_i4b:n), 1_i4b )*tau
!
                call axpy( n-1_i4b, prod, v(2_i4b:n), 1_i4b, v2(2_i4b:n), 1_i4b )
#else
                prod = dot_product( v2(2_i4b:n), v(2_i4b:n) )*tau
!
                v2(2_i4b:n) = v2(2_i4b:n) + prod*v(2_i4b:n)
#endif
!
            else
!
                u2(2_i4b:m) = zero
!
            end if
!
            do i = 2_i4b, n-1_i4b
!
                k = i + 1_i4b
!
!               APPLY H(i-1) TO mat(i:m,i) FROM THE LEFT AND 
!               APPLY G(i-1)   TO mat(i:m,i) FROM THE RIGHT.
!
#ifdef _BLAS
                w(i:m) = mat(i:m,i)
                call axpy( m-i+1_i4b, v2(i), u(i:m), 1_i4b, w(i:m), 1_i4b )
                call axpy( m-i+1_i4b, v(i), u2(i:m), 1_i4b, w(i:m), 1_i4b )
#else
                w(i:m) = mat(i:m,i) + v2(i)*u(i:m) + v(i)*u2(i:m)
#endif
!
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+1:m,i).
!
                call hous1( w(i:m), tau, d(i) )
!
!               FINISH THE i-1 RANK-2 COLUMN-ROW UPDATE AND
!               START THE i RANK-2 COLUMN-ROW UPDATE.
!
                tmp(k:m) = zero
!
                do j = k, n
!
!                   APPLY H(i-1) TO mat(i:m,i+1:n) FROM THE LEFT AND 
!                   APPLY G(i-1) TO mat(i:m,i+1:n) FROM THE RIGHT.
!
#ifdef _BLAS
                    call axpy( m-i+1_i4b, v2(j), u(i:m), 1_i4b, mat(i:m,j), 1_i4b )
                    call axpy( m-i+1_i4b, v(j), u2(i:m), 1_i4b, mat(i:m,j), 1_i4b )
#else
                    mat(i:m,j) = mat(i:m,j) + v2(j)*u(i:m) + v(j)*u2(i:m)
#endif
!
!                   START THE i RANK-2 COLUMN-ROW UPDATE.
!
#ifdef _BLAS
                    v2(j) = tau*dot( m-i+1_i4b, w(i:m), 1_i4b, mat(i:m,j), 1_i4b )
#else
                    v2(j) = tau*dot_product( w(i:m), mat(i:m,j) )
#endif
                    v(j)  = mat(i,j) + v2(j)
#ifdef _BLAS
                    call axpy( m-k+1_i4b, v(j), mat(k:m,j), 1_i4b, tmp(k:m), 1_i4b )
#else
                    tmp(k:m) = tmp(k:m) + v(j)*mat(k:m,j)
#endif
!
                end do
!
!               MAKE A LOCAL COPY OF ELEMENTARY REFLECTOR H(i).
!
                u(i:m) = w(i:m)
!
!               GENERATE ELEMENTARY REFLECTOR G(i) TO ANNIHILATE mat(i,i+2:n)
!
                alpha = v(k)
!
                call hous1( v(k:n), tau, prod )
!
                e(k) = prod
!
!               CONTINUE THE i RANK-2 COLUMN-ROW UPDATE.
!
                if ( tau/=zero ) then
!
                    beta    = tau/(alpha-prod)
                    u2(k:m) = beta*tmp(k:m) - mat(k:m,k)
!
#ifdef _BLAS
                    prod = dot( n-k+1_i4b, v2(k:n), 1_i4b, v(k:n), 1_i4b )*tau
!
                    call axpy( n-k+1_i4b, prod, v(k:n), 1_i4b, v2(k:n), 1_i4b )
#else
                    prod    = dot_product( v2(k:n), v(k:n) )*tau
!
                    v2(k:n) = v2(k:n) + prod*v(k:n)
#endif
!
                else
!
                    u2(k:m) = zero
!
                end if
!
            end do
!
#ifdef _OPENMP
        end if
#endif
!
!       FINISH THE n-1 RANK-2 COLUMN-ROW UPDATE.
!
!       APPLY H(n-1) TO mat(n:m,n) FROM THE LEFT AND 
!       APPLY G(n-1) TO mat(n:m,n) FROM THE RIGHT.
!
#ifdef _BLAS
        call axpy( m-n+1_i4b, v2(n), w(n:m), 1_i4b, mat(n:m,n), 1_i4b )
#else
        mat(n:m,n) = mat(n:m,n) + v2(n)*w(n:m)
#endif
!
        d(n) = -sign( norm(mat(n:m,n)), mat(n,n) )
!
    else
!
!       REDUCE mat TO LOWER BIDIAGONAL FORM.
!
        v(1_i4b:n) = mat(1_i4b,1_i4b:n)
!
!       GENERATE ELEMENTARY REFLECTOR G(1) TO ANNIHILATE mat(1,2:n).
!
        call hous1( v(1_i4b:n), tau, d(1_i4b) )
!
#ifdef _OPENMP
!
        if ( test_par ) then
!
            u2(2_i4b:m)  = mat(2_i4b:m,1_i4b)
!
!$OMP PARALLEL PRIVATE(i,k,nn,num_threads,thread_num,jinc,jrem,joffset,j,j1,j2,tmp,alpha),  &
!$OMP          NUM_THREADS(max_num_threads)
!
            num_threads = omp_get_num_threads()
            thread_num  = omp_get_thread_num()
!
!           DISTRIBUTE THE WORK AMONG THE THREADS FOR THE FIRST
!           RANK-2 COLUMN-ROW UPDATE.
!
            nn   = n - 1_i4b
            jrem = mod( nn, num_threads )
            jinc = ( nn - jrem )/num_threads
!
            if ( thread_num<jrem ) then
                jinc    = jinc + 1_i4b
                joffset = 2_i4b
            else
                joffset = jrem + 2_i4b
            end if
!
            j1 = thread_num*jinc + joffset
            j2 = j1 + jinc - 1_i4b
!
!           START THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            if ( j1<=j2 ) then
!
                tmp(2_i4b:m) = matmul( mat(2_i4b:m,j1:j2), v(j1:j2) )
!
!$OMP CRITICAL (updateua)
!
#ifdef _BLAS
                call axpy( m-1_i4b, one, tmp(2_i4b:m), 1_i4b, u2(2_i4b:m), 1_i4b )
#else
                u2(2_i4b:m) = u2(2_i4b:m) + tmp(2_i4b:m)  
#endif
!
!$OMP END CRITICAL (updateua)
!
            end if
!
!$OMP BARRIER
!
!$OMP MASTER
!
            u2(2_i4b:m) = tau*u2(2_i4b:m) 
            u(2_i4b:m)  = mat(2_i4b:m,1_i4b) + u2(2_i4b:m) 
!
!           GENERATE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(3:m,1).
!
            call hous1( u(2_i4b:m), tau, e(2_i4b) )
!
#ifdef _BLAS
            prod = dot( m-1_i4b, u(2_i4b:m), 1_i4b, u2(2_i4b:m), 1_i4b )
#else
            prod = dot_product( u(2_i4b:m), u2(2_i4b:m) )
#endif
!
            w(2_i4b:m) = zero
!
!$OMP END MASTER
!
!$OMP BARRIER
!
!           FINISH THE FIRST RANK-2 COLUMN-ROW UPDATE AND
!           START THE SECOND RANK-2 COLUMN-ROW UPDATE.
!
            tmp(2_i4b:m) = zero
!
            do j = j1, j2
!
#ifdef _BLAS
                alpha = tau*( dot( m-1_i4b, u(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b ) + prod*v(j) )
#else
                alpha = tau*( dot_product( u(2_i4b:m), mat(2_i4b:m,j) ) + prod*v(j) )
#endif
!
!               APPLY G(1) TO mat(2:m,2:n) FROM THE RIGHT AND
!               APPLY H(1) TO mat(2:m,2:n) FROM THE LEFT.
!
#ifdef _BLAS
                call axpy( m-1_i4b, alpha, u(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b )
                call axpy( m-1_i4b, v(j), u2(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b )
#else
                mat(2_i4b:m,j) = mat(2_i4b:m,j) + alpha*u(2_i4b:m) + v(j)*u2(2_i4b:m)
#endif
!
!               START THE SECOND RANK-2 COLUMN-ROW UPDATE.
!
                v(j) = mat(2_i4b,j)
#ifdef _BLAS
                call axpy( m-2_i4b, v(j), mat(3_i4b:m,j), 1_i4b, tmp(3_i4b:m), 1_i4b )
#else
                tmp(3_i4b:m) = tmp(3_i4b:m) + v(j)*mat(3_i4b:m,j)
#endif
!                    
            end do
!
!$OMP CRITICAL (updateua)
!
#ifdef _BLAS
            call axpy( m-2_i4b, one, tmp(3_i4b:m), 1_i4b, w(3_i4b:m), 1_i4b )
#else
            w(3_i4b:m) = w(3_i4b:m) + tmp(3_i4b:m)  
#endif
!
!$OMP END CRITICAL (updateua)
!
!$OMP BARRIER
!
            do i = 2_i4b, m-1_i4b
!
                k = i + 1_i4b
!
!               DISTRIBUTE THE WORK AMONG THE THREADS FOR THE i
!               RANK-2 COLUMN-ROW UPDATE.
!
                nn   = nn - 1_i4b
                jrem = mod( nn, num_threads )
                jinc = ( nn - jrem )/num_threads
!
                if ( thread_num<jrem ) then
                    jinc    = jinc + 1_i4b
                    joffset = i + 1_i4b
                else
                    joffset = i + jrem + 1_i4b
                end if
!
                j1 = thread_num*jinc + joffset
                j2 = j1 + jinc - 1_i4b
!
!$OMP MASTER
!
                alpha = v(i)
!
!               GENERATE ELEMENTARY REFLECTOR G(i) TO ANNIHILATE mat(i,i+1:n).
!
                call hous1( v(i:n), tau, prod )
!
                d(i) = prod
!
!               CONTINUE THE i RANK-2 COLUMN-ROW UPDATE.
!
                if ( tau/=zero ) then
                    beta    = tau/(alpha-prod)
                    u2(k:m) = beta*( w(k:m) - prod*mat(k:m,i) )
                    u(k:m)  = mat(k:m,i) + u2(k:m) 
                else
                    u2(k:m) = zero
                    u(k:m)  = mat(k:m,i)
                end if
!
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+2:m,i).
!
                call hous1( u(k:m), tau, e(k) )
!
#ifdef _BLAS
                prod = dot( m-k+1_i4b, u(k:m), 1_i4b, u2(k:m), 1_i4b )
#else
                prod = dot_product( u(k:m), u2(k:m) )
#endif
!
                w(k:m) = zero
!
!$OMP END MASTER
!
!$OMP BARRIER
!
!               FINISH THE i RANK-2 COLUMN-ROW UPDATE AND
!               START THE i+1 RANK-2 COLUMN-ROW UPDATE.
!
                tmp(k:m) = zero
!
                do j = j1, j2
!
#ifdef _BLAS
                    alpha= tau*( dot( m-k+1_i4b, u(k:m), 1_i4b, mat(k:m,j), 1_i4b ) + prod*v(j) )
#else
                    alpha= tau*( dot_product( u(k:m), mat(k:m,j) ) + prod*v(j) )
#endif
!
!                   APPLY G(i) TO mat(i+1:m,i+1:n) FROM THE RIGHT AND
!                   APPLY H(i) TO mat(i+1:m,i+1:n) FROM THE LEFT.
!
#ifdef _BLAS
                    call axpy( m-k+1_i4b, alpha, u(k:m), 1_i4b, mat(k:m,j), 1_i4b )
                    call axpy( m-k+1_i4b, v(j), u2(k:m), 1_i4b, mat(k:m,j), 1_i4b )
#else
                    mat(k:m,j) = mat(k:m,j) + alpha*u(k:m) + v(j)*u2(k:m)
#endif
!
!                   START THE i+1 RANK-2 COLUMN-ROW UPDATE.
!
                    v(j) = mat(k,j)
#ifdef _BLAS
                    call axpy( m-k, v(j), mat(k+1_i4b:m,j), 1_i4b, tmp(k+1_i4b:m), 1_i4b )
#else
                    tmp(k+1_i4b:m) = tmp(k+1_i4b:m) + v(j)*mat(k+1_i4b:m,j)
#endif
!                    
                end do
!
!$OMP CRITICAL (updateua)
!
#ifdef _BLAS
                call axpy( m-k, one, tmp(k+1_i4b:m), 1_i4b, w(k+1_i4b:m), 1_i4b )
#else
                w(k+1_i4b:m) = w(k+1_i4b:m) + tmp(k+1_i4b:m)  
#endif
!
!$OMP END CRITICAL (updateua)
!
!$OMP BARRIER
!
            end do
!
!$OMP END PARALLEL
!
        else
#endif
!
!           START THE FIRST RANK-2 COLUMN-ROW UPDATE.
!
            u2(2_i4b:m) = tau*matmul( mat(2_i4b:m,1_i4b:n), v(1_i4b:n) )
!
            u(2_i4b:m) = mat(2_i4b:m,1_i4b) + u2(2_i4b:m) 
!
!           GENERATE ELEMENTARY REFLECTOR H(1) TO ANNIHILATE mat(3:m,1).
!
            call hous1( u(2_i4b:m), tau, e(2_i4b) )
!
#ifdef _BLAS
            prod = dot( m-1_i4b, u(2_i4b:m), 1_i4b, u2(2_i4b:m), 1_i4b )
#else
            prod = dot_product( u(2_i4b:m), u2(2_i4b:m) )
#endif
!
!           FINISH THE FIRST RANK-2 COLUMN-ROW UPDATE AND
!           START THE SECOND RANK-2 COLUMN-ROW UPDATE.
!
            tmp(2_i4b:m) = zero
!
            do j = 2_i4b, n
!
#ifdef _BLAS
                alpha = tau*( dot( m-1_i4b, u(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b ) + prod*v(j) )
#else
                alpha = tau*( dot_product( u(2_i4b:m), mat(2_i4b:m,j) ) + prod*v(j) )
#endif
!
!               APPLY G(1) TO mat(2:m,2:n) FROM THE RIGHT AND
!               APPLY H(1) TO mat(2:m,2:n) FROM THE LEFT.
!
#ifdef _BLAS
                call axpy( m-1_i4b, alpha, u(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b )
                call axpy( m-1_i4b, v(j), u2(2_i4b:m), 1_i4b, mat(2_i4b:m,j), 1_i4b )
#else
                mat(2_i4b:m,j) = mat(2_i4b:m,j) + alpha*u(2_i4b:m) + v(j)*u2(2_i4b:m)
#endif
!
!               START THE SECOND RANK-2 COLUMN-ROW UPDATE.
!
                v(j) = mat(2_i4b,j)
#ifdef _BLAS
                call axpy( m-2_i4b, v(j), mat(3_i4b:m,j), 1_i4b, tmp(3_i4b:m), 1_i4b )
#else
                tmp(3_i4b:m) = tmp(3_i4b:m) + v(j)*mat(3_i4b:m,j)
#endif
!                    
            end do
! 
            do i = 2_i4b, m-1_i4b
!
                k = i + 1_i4b
!
                alpha = v(i)
!
!               GENERATE ELEMENTARY REFLECTOR G(i) TO ANNIHILATE mat(i,i+1:n).
!
                call hous1( v(i:n), tau, prod )
!
!               SAVE ELEMENTARY REFLECTOR G(i).
!
                d(i) = prod
!
!               CONTINUE THE i RANK-2 COLUMN-ROW UPDATE.
!
                if ( tau/=zero ) then
                    beta    = tau/(alpha-prod)
                    u2(k:m) = beta*( tmp(k:m) - prod*mat(k:m,i) )
                    u(k:m)  = mat(k:m,i) + u2(k:m) 
                else
                    u2(k:m) = zero
                    u(k:m)  = mat(k:m,i)
                end if
!
!               GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+2:m,i).
!
                call hous1( u(k:m), tau, e(k) )
!
#ifdef _BLAS
                prod = dot( m-k+1_i4b, u(k:m), 1_i4b, u2(k:m), 1_i4b )
#else
                prod = dot_product( u(k:m), u2(k:m) )
#endif
!
!               FINISH THE i RANK-2 COLUMN-ROW UPDATE AND
!               START THE i+1 RANK-2 COLUMN-ROW UPDATE.
!
                tmp(k:m) = zero
!
                do j = k, n
!
#ifdef _BLAS
                    alpha= tau*( dot( m-k+1_i4b, u(k:m), 1_i4b, mat(k:m,j), 1_i4b ) + prod*v(j) )
#else
                    alpha= tau*( dot_product( u(k:m), mat(k:m,j) ) + prod*v(j) )
#endif
!
!                   APPLY G(i) TO mat(i+1:m,i+1:n) FROM THE RIGHT AND
!                   APPLY H(i) TO mat(i+1:m,i+1:n) FROM THE LEFT.
!
#ifdef _BLAS
                    call axpy( m-k+1_i4b, alpha, u(k:m), 1_i4b, mat(k:m,j), 1_i4b )
                    call axpy( m-k+1_i4b, v(j), u2(k:m), 1_i4b, mat(k:m,j), 1_i4b )
#else
                    mat(k:m,j) = mat(k:m,j) + alpha*u(k:m) + v(j)*u2(k:m)
#endif
!
!                   START THE i+1 RANK-2 COLUMN-ROW UPDATE.
!
                    v(j) = mat(k,j)
#ifdef _BLAS
                    call axpy( m-k, v(j), mat(k+1_i4b:m,j), 1_i4b, tmp(k+1_i4b:m), 1_i4b )
#else
                    tmp(k+1_i4b:m) = tmp(k+1_i4b:m) + v(j)*mat(k+1_i4b:m,j)
#endif
!                    
                end do
!                    
            end do
!
#ifdef _OPENMP
        end if
#endif
!
        d(m) = -sign( norm(v(m:n)), v(m) )
!
    end if
!
!
! END OF SUBROUTINE bd_cmp_t
! __________________________
!
    end subroutine bd_cmp_t
!
! =========================================================================================
!
    subroutine bd_cmp2_a( mat, d, e, p, failure, gen_p, reortho )
!
! Purpose
! _______
!
!   BD_CMP2 reduces a m-by-n matrix MAT with m >= n to upper bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal.
!
!   BD_CMP2 computes BD, Q and P using the one-sided Ralha-Barlow
!   bidiagonal reduction algorithm.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the general m-by-n matrix to be reduced.
!
!            On exit, the first n columns of Q are stored in
!            in MAT(1:m,1:n).
!
!            The shape of MAT must verify: size( MAT, 1 ) >= size( MAT, 2 ) = n .
!
!   D        (OUTPUT) real(stnd), dimension(:)
!            The diagonal elements of the bidiagonal matrix BD
!
!            The size of D must be size( MAT, 2) = n .
!
!   E        (OUTPUT) real(stnd), dimension(:)
!            The off-diagonal elements of the bidiagonal matrix BD:
!
!                E(i) = BD(i-1,i) for i = 2,3,...,n;
!
!            The size of E must be size( MAT, 2 ) = n .
!
!   P        (OUTPUT) real(stnd), dimension(:,:)
!            On exit, the n-by-n matrix P.
!
!            The shape of P must verify: size( P, 1 ) = size( P, 2 ) = n .
!
!   FAILURE  (OUTPUT, OPTIONAL) logical(lgl)
!            On exit:
!
!            - FAILURE = false :  indicates successful exit ;
!            - FAILURE = true  :  indicates that MAT is nearly singular and some loss of
!              orthogonality of Q can be expected in the Ralha-Barlow one-sided
!              algorithm.
!
!            See further details.
!
!   GEN_P    (INPUT, OPTIONAL) logical(lgl)
!            If the optional argument GEN_P is used and is set to true, the orthogonal
!            matrix P is generated on output of the subroutine. If this argument is set
!            to false, the orthogonal matrix is stored in factored form as products of
!            elementary reflectors in the lower triangle of the array P.
!
!            See further details.
!
!            The default is GEN_P = true.
!
!   REORTHO  (INPUT, OPTIONAL) logical(lgl)
!            If the optional argument REORTHO is set to true, reorthogonalization of the
!            matrix Q is performed (when needed).
!            If this optional logical argument is set to false, reorthogonalization of Q is never
!            performed and the columns of the  matrix Q can be far from orthogonal to each other
!            for input matrices MAT with a large condition number.
!
!            See further details.
!
!            The default is REORTHO = true.
!
!
! Further Details
! _______________
!
!   This subroutine is an implementation of the Ralha-Barlow one-sided method
!   to reduce a rectangular matrix MAT to bidiagonal form BD. Q is computed by
!   a recurrence relationship and P as a product of n-1 elementary reflectors
!   (e.g. Householder transformations):
!
!       P = G(1) * G(2) * . . . * G(n-1)
!
!   Each G(i) has the form:
!
!       G(i) = I + taup * v * v'
!
!   where taup is a real scalar, and v is a real vector. IF GEN_P is used and set to false,
!   the n-1 G(i) elementary reflectors are stored in the lower triangle of the array P.
!   For the G(i) reflector, taup is stored in P(i+1,1) and v is stored in P(i+1:n,i+1).
!   IF GEN_P is set to true, P is generated in P(:n,:n).
!        
!   In addition, P(1,1) is set to -1 if GEN_P=false and is equal to 1 if GEN_P=true.
!   In other words, the value of P(1,1) indicates if the orthogonal matrix P is stored
!   in factored form or not. Note that if n is equal to 1, no elementary reflectors are
!   needed and consequently P(1,1) is set to 1, independently of the value of GEN_P.
!
!   This is the blocked version of the algorithm. See the references (1), (2) and (3)
!   for further details. Note also that the blocked algorithm implemented here is
!   more efficient than the version described in the reference (3). Furthermore the
!   algorithm is parallelized if OPENMP is used.
!
!   Since Q is computed by a recurrence relationship, a loss of orthogonality of Q can be
!   observed when the rectangular matrix MAT is singular or nearly singular or has a large
!   condition number, see the reference (2) for details.
!
!   To correct partly this deficiency, partial reorthogonalization can be performed to ensure
!   orthogonality at the expense of speed of computation. The reorthogonalization uses the Gram-Schmidt
!   method described in the reference (4) and is activated if the optional logical argument REORTHO is
!   set to true.
!
!   The reference (2) also explains how to handle the case of an exactly singular
!   matrix MAT (a very rare event). However, in this subroutine, the partial reorthogonalization
!   described in the reference (4), which is used when REORTHO is set true, corrects automatically
!   this problem.
!
!   For further details, see:
!
!   (1) Ralha, R.M.S., 2003:
!           One-sided reduction to bidiagonal form.
!           Linear Algebra Appl., No 358, pp. 219-238.
!
!   (2) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!           A new stable bidiagonal reduction algorithm.
!           Linear Algebra Appl., No 397, pp. 35-84.
!
!   (3) Bosner, N., and Barlow, J.L., 2007:
!           Block and Parallel versions of one-sided bidiagonalization.
!           SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!   (4) Stewart, G.W., 2007:
!           Block Gram-Schmidt Orthogonalization.
!           Report TR-4823, Department of Computer Science, College Park, University of Maryland.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error5
    use Utilities,       only : assert_eq, merror, pythag
    use Random,          only : random_number_
    use Hous_Procedures, only : hous1
    use QR_Procedures,   only : ortho_gen_qr
#ifdef _BLAS
    use BLAS_interfaces, only : dot, axpy, gemv, ger, gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: d, e
    real(stnd),   intent(out),   dimension(:,:) :: p
!
    logical(lgl), intent(in),  optional :: gen_p, reortho
    logical(lgl), intent(out), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                :: tau, beta, alpha, scal, normc
    real(stnd), dimension(size(mat,1))        :: u, u_tmp
    real(stnd), dimension(size(mat,2))        :: v, taup
    real(stnd), dimension(blksz2)             :: prodvec
    real(stnd), dimension(size(mat,1),blksz2) :: umat
    real(stnd), dimension(blksz2,size(mat,2)) :: vmat
!
    integer(i4b) :: m, n, i, i1, i2, i3, l, j, k, blksz
!
    logical(lgl) :: failure2, do_gen_p, do_reortho
!
#ifdef _OPENMP
    real(stnd)                         :: beta2
    real(stnd), dimension(size(mat,1)) :: tmp
!
    integer(i4b) :: max_num_threads, num_threads, thread_num, iinc, irem, ioffset
!
    logical(lgl) :: apply_hous
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n =  assert_eq( (/ int(size(mat,2),i4b),     &
                       int(size(d),i4b)      ,   &
                       int(size(e),i4b)      ,   &
                       int(size(p,1),i4b)    ,   &
                       int(size(p,2),i4b) /) ,   &
                       name_proc )
!
!   CHECK DIMENSIONS OF THE INPUT MATRIX.
!
    if ( m<n ) then
!
        call merror( name_proc//svd_error5 )
!
    end if
!
    do_gen_p = true
!
    if ( present( gen_p ) ) then
!
        do_gen_p = gen_p
!
    end if
!
    do_reortho = true
!
    if ( present( reortho ) ) then
!
        do_reortho = reortho
!
    end if
!
    failure2 = false
!
    if ( present(failure) ) then
!
        failure = failure2
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
!   ADJUST BLOCK SIZE IF NUMBER OF COLUMNS IS TOO SMALL.
!
    blksz = min( blksz2, n )
!
    e(1_i4b) = zero
!
!   REDUCE mat TO UPPER BIDIAGONAL FORM.
!
!   GENERATE THE FIRST COLUM OF Q AND STORE IT IN mat(1:m,1)
!
    d(1_i4b) = norm( mat(1_i4b:m,1_i4b) )
!
    if ( d(1_i4b)/=zero ) then
!
!       NORMALIZATION IS ENOUGH.
!
        mat(1_i4b:m,1_i4b) = mat(1_i4b:m,1_i4b)/d(1_i4b)
!
    else
!
!       IF THE NORM OF THE FIRST COLUMN OF mat IS ZERO,
!       SET IT TO A NONZERO NORMALIZED RANDOM VECTOR
!       AND PROCEED.
!
        call random_number_( u(:m) ) 
!
        u(:m) = u(:m) - half
        scal  = norm( u(:m) )
        mat(1_i4b:m,1_i4b) = u(:m)/scal
!
        failure2 = true
!
    end if
!
!   SPECIAL CASE IF mat(:,:) IS A COLUMN VECTOR.
!
    if ( n==1_i4b ) then
!
        p(1_i4b,1_i4b) = one
!
        if ( present(failure) ) then
!
            failure = failure2
!
        end if
!
        return
!
    end if
!
    u_tmp(1_i4b:m) = zero
!
#ifdef _OPENMP
!
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m*n>=omp_limit                      .and.       &
                  n>blksz                             .and.       &
                  i>1_i4b
!
    if ( test_par ) then
!
        max_num_threads = min( i, max_num_threads_bd_cmp2 )
!
!$OMP PARALLEL PRIVATE(num_threads,thread_num,iinc,irem,ioffset,           &
!$OMP                  i,i1,i2,i3,l,j,k,alpha,beta,tau,scal,normc,u,tmp),  &
!$OMP          NUM_THREADS(max_num_threads)
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
!       PROCESS THE FIRST BLOCK OF COLUMNS OF mat .
!
!       START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(1)
!       HOUSEHOLDER TRANSFORMATION.
!
        u(1_i4b:m)   = mat(1_i4b:m,1_i4b)
        tmp(1_i4b:m) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
        do i = 2_i4b, n
!
#ifdef _BLAS
            v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b )
!
            call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
            v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )
!
            tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
        end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compu0)
!
#ifdef _BLAS
        call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
        u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu0)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!       COMPUTE P IN FACTORED FORM, STORING THE G(1) HOUSEHOLDER TRANSFORMATION
!       IN taup(1) AND p(2:n,2).
!
        alpha = v(2_i4b)        
!
        call hous1( v(2_i4b:n), tau, beta )         
!
        p(2_i4b:n,2_i4b) = v(2_i4b:n)
        taup(1_i4b)      = tau
!
!       INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
        if ( tau/=zero ) then
!
            scal  = tau/( alpha - beta )
!
#ifdef _BLAS
            call axpy( m, -beta, mat(1_i4b:m,2_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
            umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
            umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,2_i4b) )
#endif
!
        else
!
            umat(1_i4b:m,1_i4b) = zero
!
        end if
!
        umat(1_i4b:m,blksz) = zero
!
        vmat(1_i4b:blksz,1_i4b:n) = zero
        vmat(1_i4b,2_i4b:n)       = v(2_i4b:n)
!
!       UPDATE mat(1:m,2) WITH G(1) AND ORTHOGONALIZE mat(1:m,2) AGAINST THE FIRST COLUMN OF Q.
!
#ifdef _BLAS
        call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,2_i4b), 1_i4b )
        call axpy( m, -beta, mat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,2_i4b), 1_i4b )
#else
        mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b) + umat(1_i4b:m,1_i4b) - beta*mat(1_i4b:m,1_i4b)
#endif
!
!       GENERATE THE SECOND COLUM OF Q AND STORE IT IN mat(1:m,2)
!
        scal  = norm( mat(1_i4b:m,2_i4b) )
        normc = pythag( scal, beta )
!
        if ( scal<=fudge_ortho2*normc ) then
!
!           REORTHOGONALIZATION IS REQUIRED.
!
            call ortho_vec( mat(1_i4b:m,1_i4b:1_i4b), mat(1_i4b:m,2_i4b),   &
                            scal, normvec_orig=normc )
!
            failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
        else
!
!           NORMALIZATION IS ENOUGH.
!
            mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b)/scal
!
        end if
!
!       COMPUTE SECOND ROW OF THE BIDIAGONAL MATRIX.
!
        d(2_i4b) = scal
        e(2_i4b) = beta
!
        u_tmp(1_i4b:m) = zero  
!
!$OMP END SINGLE
!
        do j = 2_i4b, blksz-1_i4b
!
!           START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(j)
!           HOUSEHOLDER TRANSFORMATION.
!
            u(1_i4b:m) = mat(1_i4b:m,j)
!
!$OMP DO SCHEDULE(STATIC)
!
            do i = 1_i4b, j-1_i4b
!
#ifdef _BLAS
                prodvec(i) = dot( m, u(1_i4b:m), 1_i4b, umat(1_i4b:m,i), 1_i4b )
#else
                prodvec(i) = dot_product( u(1_i4b:m), umat(1_i4b:m,i) )
#endif
!
            end do
!
!$OMP END DO
!
            tmp(1_i4b:m) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
            do i = j+1_i4b, n
!
#ifdef _BLAS
                v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b ) +                      &
                       dot( j-1_i4b, prodvec(1_i4b:j-1_i4b), 1_i4b, vmat(1_i4b:j-1_i4b,i), 1_i4b )
!
                call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) ) +                &
                       dot_product( prodvec(1_i4b:j-1_i4b), vmat(1_i4b:j-1_i4b,i) )
!
                tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
            end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compu1)
!
#ifdef _BLAS
            call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu1)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!           COMPUTE P IN FACTORED FORM, STORING THE G(j) HOUSEHOLDER TRANSFORMATION
!           IN taup(j) AND p(j+1:n,j+1).
!
            alpha = v(j+1_i4b)        
!
            call hous1( v(j+1_i4b:n), tau, beta )
!
            p(j+1:n,j+1) = v(j+1_i4b:n)
            taup(j)      = tau
!
!           SAVE G(j) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
            vmat(j,j+1_i4b:n) = v(j+1_i4b:n)
!
#ifdef _BLAS
            call gemv( 'N', j-1_i4b, n-j, one, vmat(1_i4b:blksz2,j+1_i4b:n), blksz2, &
                       v(j+1_i4b:n), 1_i4b, zero, prodvec, 1_i4b )
#else
            prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,j+1_i4b:n), v(j+1_i4b:n) )
#endif
!
            if ( tau/=zero ) then
!
                scal = tau/( alpha - beta )
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,j+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                umat(1_i4b:m,j) = scal*u_tmp(1_i4b:m)
#else
                umat(1_i4b:m,j) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,j+1_i4b) )
#endif
            else
!
                umat(1_i4b:m,j) = zero
!
            end if
!
#ifdef _BLAS
            call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                        1_i4b, one, umat(1_i4b:m,j), 1_i4b )
#else
            umat(1_i4b:m,j) = umat(1_i4b:m,j) + tau*matmul( umat(1_i4b:m,1_i4b:j-1_i4b), prodvec(1_i4b:j-1_i4b) )
#endif
!
!           UPDATE mat(1:m,j+1) WITH G(1)* ... *G(j) AND ORTHOGONALIZE mat(1:m,j+1) AGAINST THE j COLUMN OF Q.
!
#ifdef _BLAS
            call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,j+1_i4b), 1_i4b,     &
                       one, mat(1_i4b:m,j+1_i4b), 1_i4b )
            call axpy( m, -beta, mat(1_i4b:m,j), 1_i4b, mat(1_i4b:m,j+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,j+1_i4b) ) &
                                   - beta*mat(1_i4b:m,j)
#endif
!
!           GENERATE THE j+1 COLUM OF Q AND STORE IT IN mat(1:m,j+1).
!
            scal  = norm( mat(1_i4b:m,j+1_i4b) )
            normc = pythag( scal, beta )
!
            if ( scal<=fudge_ortho2*normc ) then
!
                if ( do_reortho ) then
!
!                   REORTHOGONALIZATION IS REQUIRED.
!
                    call ortho_vec( mat(1_i4b:m,1_i4b:j), mat(1_i4b:m,j+1_i4b),   &
                                    scal, normvec_orig=normc )
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                else
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    scal = max( scal, safmin )
!
                    mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b)/scal
!
                end if
!
            else
!
!               NORMALIZATION IS ENOUGH.
!
                mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b)/scal
!
            end if
!
!           COMPUTE THE j+1 ROW OF THE BIDIAGONAL MATRIX.
!
            d(j+1_i4b) = scal
            e(j+1_i4b) = beta
!
            u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
        end do
!
!       PROCESS THE SECOND AND FOLLOWING BLOCKS OF COLUMNS OF mat .
!
        do l = blksz, n-blksz, blksz
!
!           PERFORM A BLOCK UPDATE OF THE TRAILING MATRIX AND START COMPUTATIONS FOR
!           GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION.
!
            tmp(1_i4b:m) = zero
!
!           DISTRIBUTE THE WORK AMONG THE THREADS FOR THE BLOCK UPDATE OF THE TRAILING MATRIX.
!
            i  = n - l
            irem = mod( i, num_threads )
            iinc = ( i - irem )/num_threads
!
            if ( thread_num<irem ) then
                iinc    = iinc + 1_i4b
                ioffset = l + 1_i4b
            else
                ioffset = irem + l + 1_i4b
            end if
!
            i1 = thread_num*iinc + ioffset
            i2 = i1 + iinc - 1_i4b
            i3 = i2 - i1 + 1_i4b
!
            if ( i3>0_i4b ) then
!
#ifdef _BLAS
!
!               COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) AS A BLOCK UPDATE.
!
                call gemm( 'N', 'N', m, i3, blksz, one, umat(1_i4b:m,1_i4b:blksz), m,   &
                           vmat(1_i4b:blksz2,i1:i2), blksz2, one, mat(1_i4b:m,i1:i2), m )
!
!               START GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION
!               WITH A BLAS2.5 GEMVT OPERATOR.
!
!                call gemv( 'T', m, i3, one,  mat(1_i4b:m,i1:i2), m, mat(1_i4b:m,l), &
!                           1_i4b, zero, v(i1:i2), 1_i4b )
!
!                call gemv( 'N', m, i3, one,  mat(1_i4b:m,i1:i2), m, v(i1:i2), &
!                           1_i4b, one, tmp(1_i4b:m), 1_i4b )
!
                do j = i1, i2
!
                    v(j) = dot( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,j), 1_i4b )
!
                    call axpy( m, v(j), mat(1_i4b:m,j), 1_i4b, tmp(1_i4b:m), 1_i4b )
!
                end do
#else
!
!               COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) AS A BLOCK UPDATE.
!
                mat(1_i4b:m,i1:i2) = mat(1_i4b:m,i1:i2) +                                       &
                                     matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i1:i2) )
!
!               START GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION
!               WITH A BLAS2.5 GEMVT OPERATOR.
!
!                v(i1:i2) = matmul( mat(1_i4b:m,l), mat(1_i4b:m,i1:i2) )
!
!                tmp(1_i4b:m) = tmp(1_i4b:m) +  matmul( mat(1_i4b:m,i1:i2), v(i1:i2) )
!
                do j = i1, i2
!
                    v(j) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,j) )
!
                    tmp(1_i4b:m) = tmp(1_i4b:m) +  v(j)*mat(1_i4b:m,j)
!
                end do
#endif
!
            end if
!
!$OMP CRITICAL (compu2)
!
#ifdef _BLAS
            call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu2)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!           COMPUTE P IN FACTORED FORM, STORING THE G(l) HOUSEHOLDER TRANSFORMATION
!           IN taup(l) AND p(l+1:n,l+1).
!
            alpha = v(l+1_i4b)        
!
            call hous1( v(l+1_i4b:n), tau, beta )
!
            p(l+1_i4b:n,l+1_i4b) = v(l+1_i4b:n)
            taup(l)              = tau
!
!           INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
            if ( tau/=zero ) then
!
                scal = tau/( alpha - beta )
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,l+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
                umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+1_i4b) )
#endif
!
            else
!
                umat(1_i4b:m,1_i4b) = zero
!
            end if
!
            vmat(:blksz,:n)       = zero
            vmat(1_i4b,l+1_i4b:n) = v(l+1_i4b:n)
!
!           UPDATE mat(1:m,l+1) WITH G(l) AND ORTHOGONALIZE mat(1:m,l+1) AGAINST THE l COLUMN OF Q.
!
#ifdef _BLAS
            call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
            call axpy( m, -beta, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) + umat(1_i4b:m,1_i4b) - beta*mat(1_i4b:m,l)
#endif
!
!           GENERATE THE l+1 COLUM OF Q AND STORE IT IN mat(1:m,l+1).
!
            scal  = norm( mat(1_i4b:m,l+1_i4b) )
            normc = pythag( scal, beta )
!
            if ( scal<=fudge_ortho2*normc ) then
!
                if ( do_reortho ) then
!
!                   REORTHOGONALIZATION IS REQUIRED.
!
                    call ortho_vec( mat(1_i4b:m,1_i4b:l), mat(1_i4b:m,l+1_i4b),   &
                                    scal, normvec_orig=normc )
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                else
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    scal = max( scal, safmin )
!
                    mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
                end if
!
            else
!
!               NORMALIZATION IS ENOUGH.
!
                mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
            end if
!
!           COMPUTE THE l+1 ROW OF THE BIDIAGONAL MATRIX.
!
            d(l+1_i4b) = scal
            e(l+1_i4b) = beta
!
            u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
            do j = 2_i4b, blksz
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(l+j-1)
!               HOUSEHOLDER TRANSFORMATION.
!
                u(1_i4b:m) = mat(1_i4b:m,l+j-1_i4b)
!
!$OMP DO SCHEDULE(STATIC)
!
                do i = 1_i4b, j-1_i4b
!
#ifdef _BLAS
                    prodvec(i) = dot( m, u(1_i4b:m), 1_i4b, umat(1_i4b:m,i), 1_i4b )
#else
                    prodvec(i) = dot_product( u(1_i4b:m), umat(1_i4b:m,i) )
#endif
!
                end do
!
!$OMP END DO
!
                tmp(1_i4b:m) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
                do i = l+j, n
!
#ifdef _BLAS
                    v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b ) +                      &
                           dot( j-1_i4b, prodvec(1_i4b:j-1_i4b), 1_i4b, vmat(1_i4b:j-1_i4b,i), 1_i4b )
!
                    call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                    v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )                  + &
                           dot_product( prodvec(1_i4b:j-1_i4b), vmat(1_i4b:j-1_i4b,i) )
!
                    tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compu3)
!
#ifdef _BLAS
                call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu3)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!               COMPUTE P IN FACTORED FORM, STORING THE G(l+j-1) HOUSEHOLDER TRANSFORMATION
!               IN taup(l+j-1) AND p(l+j:n,l+j).
!
                alpha = v(l+j)        
!
                call hous1( v(l+j:n), tau, beta )
!
                p(l+j:n,l+j) = v(l+j:n)
                taup(l+j-1)  = tau
!
!               SAVE G(l+j-1) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
                vmat(j,l+j:n) = v(l+j:n)
!
#ifdef _BLAS
                call gemv( 'N', j-1_i4b, n-j-l+1_i4b, one, vmat(1_i4b:blksz2,l+j:n), blksz2, &
                           v(l+j:n), 1_i4b, zero, prodvec, 1_i4b )
#else
                prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,l+j:n), v(l+j:n) )
#endif
!
                if ( tau/=zero ) then
!
                    scal = tau/( alpha - beta )
!
#ifdef _BLAS
                    call axpy( m, -beta, mat(1_i4b:m,l+j), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                    umat(1_i4b:m,j) = scal*u_tmp(1_i4b:m)
#else
                    umat(1_i4b:m,j) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+j) )
#endif
!
                else
!
                    umat(1_i4b:m,j) = zero
!
                end if
!
#ifdef _BLAS
                call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                           1_i4b, one, umat(1_i4b:m,j), 1_i4b )
#else
                umat(1_i4b:m,j) = umat(1_i4b:m,j) + tau*matmul( umat(1_i4b:m,:j-1_i4b), prodvec(1_i4b:j-1_i4b) )
#endif
!
!               UPDATE mat(1:m,l+j) WITH G(l)* ... *G(l+j-1) AND ORTHOGONALIZE mat(1:m,l+j)
!               AGAINST THE l+j-1 COLUMN OF Q.
!
#ifdef _BLAS
                call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,l+j), 1_i4b, &
                           one, mat(1_i4b:m,l+j), 1_i4b )
                call axpy( m, -beta, mat(1_i4b:m,l+j-1_i4b), 1_i4b, mat(1_i4b:m,l+j), 1_i4b )
#else
                mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,l+j) ) &
                                   - beta*mat(1_i4b:m,l+j-1_i4b)
#endif
!
!               GENERATE THE l+j COLUM OF Q AND STORE IT IN mat(1:m,l+j).
!
                scal  = norm( mat(1_i4b:m,l+j) )
                normc = pythag( scal, beta )
!
                if ( scal<=fudge_ortho2*normc ) then
!
                    if ( do_reortho ) then
!
!                       REORTHOGONALIZATION IS REQUIRED.
!
                        call ortho_vec( mat(1_i4b:m,1_i4b:l+j-1_i4b), mat(1_i4b:m,l+j), &
                                        scal, normvec_orig=normc )
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    else
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                        scal = max( scal, safmin )
!
                        mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j)/scal
!
                    end if
!
                else
!
!                   NORMALIZATION IS ENOUGH.
!
                    mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j)/scal
!
                end if
!
!               COMPUTE THE l+j ROW OF THE BIDIAGONAL MATRIX.
!
                d(l+j) = scal
                e(l+j) = beta
!
                u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
            end do
!
        end do
!
        k = blksz*(n/blksz)
!
        if ( n/=k ) then
!
!           PROCESS THE LAST BLOCK OF COLUMNS OF mat .
!
            do l = k, n-1_i4b
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(l)
!               HOUSEHOLDER TRANSFORMATION.
!
                tmp(1_i4b:m) = zero
!
                if ( l==k ) then
!
!$OMP DO SCHEDULE(STATIC)
!
                    do i = l+1_i4b, n
!
!                       COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) WITH A LAST BLOCK UPDATE.
!
                        u(1_i4b:m) = mat(1_i4b:m,i)
!
#ifdef _BLAS
                        call gemv( 'N', m, blksz, one, umat(1_i4b:m,1_i4b:blksz), m, vmat(1_i4b:blksz,i), &
                                   1_i4b, one, u(1_i4b:m), 1_i4b )
!
                        v(i) = dot( m, mat(1_i4b:m,l), 1_i4b, u(1_i4b:m), 1_i4b )
!
                        call axpy( m, v(i), u(1_i4b:m), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                        u(1_i4b:m) = u(1_i4b:m) + matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i) )
!
                        v(i) = dot_product( mat(1_i4b:m,l), u(1_i4b:m) )
!
                        tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*u(1_i4b:m)
#endif
!
                        mat(1_i4b:m,i) = u(1_i4b:m)
!
                    end do
!
!$OMP END DO NOWAIT
!
                else
!
!$OMP DO SCHEDULE(STATIC)
!
                    do i = l+1_i4b, n
!
#ifdef _BLAS
                        v(i) = dot( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,i), 1_i4b )
!
                        call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                        v(i) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,i) )
!
                        tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                    end do
!
!$OMP END DO NOWAIT
!
                end if
!
!$OMP CRITICAL (compu4)
!
#ifdef _BLAS
                call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu4)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!               COMPUTE P IN FACTORED FORM, STORING THE G(l) HOUSEHOLDER TRANSFORMATION
!               IN taup(l) AND p(l+1:n,l+1).
!
                alpha = v(l+1_i4b)        
!
                call hous1( v(l+1_i4b:n), tau, beta2 )
!
                p(l+1_i4b:n,l+1_i4b) = v(l+1_i4b:n)
                taup(l) = tau
!
                apply_hous = tau/=zero
!
                if ( apply_hous ) then
!
                    scal = tau/( alpha - beta2 )
!
#ifdef _BLAS
                    call axpy( m, -beta2, mat(1_i4b:m,l+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                    umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
                    umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta2*mat(1_i4b:m,l+1_i4b) )
#endif
!
!                   APPLY THE G(l) HOUSEHOLDER TRANSFORMATION TO mat(:m,l+1) .
!
#ifdef _BLAS
                    call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
                    mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) + umat(1_i4b:m,1_i4b)
#endif
!
                end if
!
!$OMP END SINGLE
!
                if ( apply_hous ) then
!
!                   APPLY THE G(l) HOUSEHOLDER TRANSFORMATION TO mat(:m,l+2:n) .
!
!$OMP DO SCHEDULE(STATIC)
!
                    do i = l+2_i4b, n
!
#ifdef _BLAS
                        call axpy( m, v(i), umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,i), 1_i4b )
#else
                        mat(1_i4b:m,i) = mat(1_i4b:m,i) + umat(1_i4b:m,1_i4b)*v(i)
#endif
!
                    end do
!
!$OMP END DO NOWAIT
!
                end if
!
!$OMP SINGLE
!
!               ORTHOGONALIZE mat(1:m,l+1) AGAINST THE l COLUMN OF Q.
!
#ifdef _BLAS
                call axpy( m, -beta2, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
                mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) - beta2*mat(1_i4b:m,l)
#endif
!
!               GENERATE THE l+1 COLUM OF Q AND STORE IT IN mat(1:m,l+1).
!
                scal  = norm( mat(1_i4b:m,l+1_i4b) )
                normc = pythag( scal, beta2 )
!
                if ( scal<=fudge_ortho2*normc ) then
!
                    if ( do_reortho ) then
!
!                       REORTHOGONALIZATION IS REQUIRED.
!
                        call ortho_vec( mat(1_i4b:m,1_i4b:l), mat(1_i4b:m,l+1_i4b),   &
                                        scal, normvec_orig=normc )
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    else
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                        scal = max( scal, safmin )
!
                        mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
                    end if
!
                else
!
!                   NORMALIZATION IS ENOUGH.
!
                    mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
                end if
!
!               COMPUTE THE l+1 ROW OF THE BIDIAGONAL MATRIX.
!
                d(l+1_i4b) = scal
                e(l+1_i4b) = beta2
!
                u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
            end do
!
        end if
!
!$OMP END PARALLEL
!
    else
#endif
!
!       PROCESS THE FIRST BLOCK OF COLUMNS OF mat .
!
!       START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(1)
!       HOUSEHOLDER TRANSFORMATION.
!
        u(1_i4b:m) = mat(1_i4b:m,1_i4b)
!
        do i = 2_i4b, n
!
#ifdef _BLAS
            v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b )
!
            call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
            v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )
!
            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
        end do
!
!       COMPUTE P IN FACTORED FORM, STORING THE G(1) HOUSEHOLDER TRANSFORMATION
!       IN taup(1) AND p(2:n,2).
!
        alpha = v(2_i4b)        
!
        call hous1( v(2_i4b:n), tau, beta )         
!
        p(2_i4b:n,2_i4b) = v(2_i4b:n)
        taup(1_i4b)      = tau
!
!       INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
        if ( tau/=zero ) then
!
            scal = tau/( alpha - beta )
!
#ifdef _BLAS
            call axpy( m, -beta, mat(1_i4b:m,2_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
            umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
            umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,2_i4b) )
#endif
!
        else
!
            umat(1_i4b:m,1_i4b) = zero
!
        end if
!
        umat(1_i4b:m,blksz) = zero
!
        vmat(1_i4b:blksz,1_i4b:n) = zero
        vmat(1_i4b,2_i4b:n)       = v(2_i4b:n)
!
!       UPDATE mat(1:m,2) WITH G(1) AND ORTHOGONALIZE mat(1:m,2) AGAINST THE FIRST COLUMN OF Q.
!
#ifdef _BLAS
        call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,2_i4b), 1_i4b )
        call axpy( m, -beta, mat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,2_i4b), 1_i4b )
#else
        mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b) + umat(1_i4b:m,1_i4b) - beta*mat(1_i4b:m,1_i4b)
#endif
!
!       GENERATE THE SECOND COLUM OF Q AND STORE IT IN mat(1:m,2)
!
        scal  = norm( mat(1_i4b:m,2_i4b) )
        normc = pythag( scal, beta )
!
        if ( scal<=fudge_ortho2*normc ) then
!
!           REORTHOGONALIZATION IS REQUIRED.
!
            call ortho_vec( mat(1_i4b:m,1_i4b:1_i4b), mat(1_i4b:m,2_i4b),   &
                            scal, normvec_orig=normc )
!
            failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
        else
!
!           NORMALIZATION IS ENOUGH.
!
            mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b)/scal
!
        end if
!
!       COMPUTE SECOND ROW OF THE BIDIAGONAL MATRIX.
!
        d(2_i4b) = scal
        e(2_i4b) = beta
!
        do j = 2_i4b, blksz-1_i4b
!
!           START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(j)
!           HOUSEHOLDER TRANSFORMATION.
!
            u(1_i4b:m) = mat(1_i4b:m,j)
!
#ifdef _BLAS
            call gemv( 'T', m, j-1_i4b, one,  umat(1_i4b:m,1_i4b:j-1_i4b), m, &
                       u(1_i4b:m), 1_i4b, zero, prodvec(1_i4b:j-1_i4b), 1_i4b )
#else
            prodvec(1_i4b:j-1_i4b) = matmul( u(1_i4b:m), umat(1_i4b:m,1_i4b:j-1_i4b) )
#endif
!
            u_tmp(1_i4b:m) = zero
!
            do i = j+1_i4b, n
!
#ifdef _BLAS
                v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b ) +                      &
                       dot( j-1_i4b, prodvec(1_i4b:j-1_i4b), 1_i4b, vmat(1_i4b:j-1_i4b,i), 1_i4b )
!
                call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) ) +                &
                       dot_product( prodvec(1_i4b:j-1_i4b), vmat(1_i4b:j-1_i4b,i) )
!
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
            end do
!
!           COMPUTE P IN FACTORED FORM, STORING THE G(j) HOUSEHOLDER TRANSFORMATION
!           IN taup(j) AND p(j+1:n,j+1).
!
            alpha = v(j+1_i4b)        
!
            call hous1( v(j+1_i4b:n), tau, beta )
!
            p(j+1:n,j+1) = v(j+1_i4b:n)
            taup(j)      = tau
!
!           SAVE G(j) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
            vmat(j,j+1_i4b:n) = v(j+1_i4b:n)
!
#ifdef _BLAS
            call gemv( 'N', j-1_i4b, n-j, one, vmat(1_i4b:blksz2,j+1_i4b:n), blksz2, &
                       v(j+1_i4b:n), 1_i4b, zero, prodvec, 1_i4b )
#else
            prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,j+1_i4b:n), v(j+1_i4b:n) )
#endif
!
            if ( tau/=zero ) then
!
                scal = tau/( alpha - beta )
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,j+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                umat(1_i4b:m,j) = scal*u_tmp(1_i4b:m)
#else
                umat(1_i4b:m,j) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,j+1_i4b) )
#endif
!
            else
!
                umat(1_i4b:m,j) = zero
!
            end if
!
#ifdef _BLAS
            call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                        1_i4b, one, umat(1_i4b:m,j), 1_i4b )
#else
            umat(1_i4b:m,j) = umat(1_i4b:m,j) + tau*matmul( umat(1_i4b:m,1_i4b:j-1_i4b), prodvec(1_i4b:j-1_i4b) )
#endif
!
!           UPDATE mat(1:m,j+1) WITH G(1)* ... *G(j) AND ORTHOGONALIZE mat(1:m,j+1) AGAINST THE j COLUMN OF Q.
!
#ifdef _BLAS
            call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,j+1_i4b), 1_i4b,     &
                       one, mat(1_i4b:m,j+1_i4b), 1_i4b )
            call axpy( m, -beta, mat(1_i4b:m,j), 1_i4b, mat(1_i4b:m,j+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,j+1_i4b) ) &
                                   - beta*mat(1_i4b:m,j)
#endif
!
!           GENERATE THE j+1 COLUM OF Q AND STORE IT IN mat(1:m,j+1).
!
            scal  = norm( mat(1_i4b:m,j+1_i4b) )
            normc = pythag( scal, beta )
!
            if ( scal<=fudge_ortho2*normc ) then
!
                if ( do_reortho ) then
!
!                   REORTHOGONALIZATION IS REQUIRED.
!
                    call ortho_vec( mat(1_i4b:m,1_i4b:j), mat(1_i4b:m,j+1_i4b),   &
                                    scal, normvec_orig=normc )
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                else
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    scal = max( scal, safmin )
!
                    mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b)/scal
!
                end if
!
            else
!
!               NORMALIZATION IS ENOUGH.
!
                mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b)/scal
!
            end if
!
!           COMPUTE THE j+1 ROW OF THE BIDIAGONAL MATRIX.
!
            d(j+1_i4b) = scal
            e(j+1_i4b) = beta
!
        end do
!
!       PROCESS THE SECOND AND FOLLOWING BLOCKS OF COLUMNS OF mat .
!
        do l = blksz, n-blksz, blksz
!
!           PERFORM A BLOCK UPDATE OF THE TRAILING MATRIX AND START COMPUTATIONS FOR
!           GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION.
!
            u_tmp(1_i4b:m) = zero
!
            i1 = l + 1_i4b
            i2 = n
            i3 = i2 - l
!
#ifdef _BLAS
!
!           COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) AS A BLOCK UPDATE.
!
            call gemm( 'N', 'N', m, i3, blksz, one, umat(1_i4b:m,1_i4b:blksz), m,    &
                       vmat(1_i4b:blksz2,i1:i2), blksz2, one, mat(1_i4b:m,i1:i2), m  )
!
!           START GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION
!           WITH A BLAS2.5 GEMVT OPERATOR.
!
!            call gemv( 'T', m, i3, one,  mat(1_i4b:m,i1:i2), m, mat(1_i4b:m,l), &
!                       1_i4b, zero, v(i1:i2), 1_i4b )
!
!            call gemv( 'N', m, i3, one,  mat(1_i4b:m,i1:i2), m, v(i1:i2), &
!                       1_i4b, one, u_tmp(1_i4b:m), 1_i4b )
!
            do j = i1, i2
!
                v(j) = dot( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,j), 1_i4b )
!
                call axpy( m, v(j), mat(1_i4b:m,j), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
!
            end do
#else
!
!           COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) AS A BLOCK UPDATE.
!
            mat(1_i4b:m,i1:i2) = mat(1_i4b:m,i1:i2) +                                       &
                                 matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i1:i2) )
!
!           START GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION
!           WITH A BLAS2.5 GEMVT OPERATOR.
!
!            v(i1:i2) = matmul( mat(1_i4b:m,l), mat(1_i4b:m,i1:i2) )
!
!            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) +  matmul( mat(1_i4b:m,i1:i2), v(i1:i2) )
!
            do j = i1, i2
!
                v(j) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,j) )
!
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) +  v(j)*mat(1_i4b:m,j)
!
            end do
#endif
!
!           COMPUTE P IN FACTORED FORM, STORING THE G(l) HOUSEHOLDER TRANSFORMATION
!           IN taup(l) AND p(l+1:n,l+1).
!
            alpha = v(l+1_i4b)        
!
            call hous1( v(l+1_i4b:n), tau, beta )
!
            p(l+1_i4b:n,l+1_i4b) = v(l+1_i4b:n)
            taup(l)              = tau
!
!           INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
            if ( tau/=zero ) then
!
                scal = tau/( alpha - beta )
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,l+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
                umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+1_i4b) )
#endif
!
            else
!
                umat(1_i4b:m,1_i4b) = zero
!
            end if
!
            vmat(:blksz,:n)       = zero
            vmat(1_i4b,l+1_i4b:n) = v(l+1_i4b:n)
!
!           UPDATE mat(1:m,l+1) WITH G(l) AND ORTHOGONALIZE mat(1:m,l+1) AGAINST THE l COLUMN OF Q.
!
#ifdef _BLAS
            call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
            call axpy( m, -beta, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) + umat(1_i4b:m,1_i4b) - beta*mat(1_i4b:m,l)
#endif
!
!           GENERATE THE l+1 COLUM OF Q AND STORE IT IN mat(1:m,l+1).
!
            scal  = norm( mat(1_i4b:m,l+1_i4b) )
            normc = pythag( scal, beta )
!
            if ( scal<=fudge_ortho2*normc ) then
!
                if ( do_reortho ) then
!
!                   REORTHOGONALIZATION IS REQUIRED.
!
                    call ortho_vec( mat(1_i4b:m,1_i4b:l), mat(1_i4b:m,l+1_i4b),   &
                                    scal, normvec_orig=normc )
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                else
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    scal = max( scal, safmin )
!
                    mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
                end if
!
            else
!
!               NORMALIZATION IS ENOUGH.
!
                mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
            end if
!
!           COMPUTE THE l+1 ROW OF THE BIDIAGONAL MATRIX.
!
            d(l+1_i4b) = scal
            e(l+1_i4b) = beta
!
            do j = 2_i4b, blksz
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(l+j-1)
!               HOUSEHOLDER TRANSFORMATION.
!
                u(1_i4b:m) = mat(1_i4b:m,l+j-1_i4b)
!
#ifdef _BLAS
                call gemv( 'T', m, j-1_i4b, one,  umat(1_i4b:m,1_i4b:j-1_i4b), m, &
                           u(1_i4b:m), 1_i4b, zero, prodvec(1_i4b:j-1_i4b), 1_i4b )
#else
                prodvec(1_i4b:j-1_i4b) = matmul( u(1_i4b:m), umat(1_i4b:m,1_i4b:j-1_i4b) )
#endif
!
                u_tmp(1_i4b:m) = zero
!
                do i = l+j, n
!
#ifdef _BLAS
                    v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b ) +                      &
                           dot( j-1_i4b, prodvec(1_i4b:j-1_i4b), 1_i4b, vmat(1_i4b:j-1_i4b,i), 1_i4b )
!
                    call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                    v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )                  + &
                           dot_product( prodvec(1_i4b:j-1_i4b), vmat(1_i4b:j-1_i4b,i) )
!
                    u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                end do
!
!               COMPUTE P IN FACTORED FORM, STORING THE G(l+j-1) HOUSEHOLDER TRANSFORMATION
!               IN taup(l+j-1) AND p(l+j:n,l+j).
!
                alpha = v(l+j)        
!
                call hous1( v(l+j:n), tau, beta )
!
                p(l+j:n,l+j) = v(l+j:n)
                taup(l+j-1)  = tau
!
!               SAVE G(l+j-1) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
                vmat(j,l+j:n) = v(l+j:n)
!
#ifdef _BLAS
                call gemv( 'N', j-1_i4b, n-j-l+1_i4b, one, vmat(1_i4b:blksz2,l+j:n), blksz2, &
                           v(l+j:n), 1_i4b, zero, prodvec, 1_i4b )
#else
                prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,l+j:n), v(l+j:n) )
#endif
!
                if ( tau/=zero ) then
!
                    scal = tau/( alpha - beta )
!
#ifdef _BLAS
                    call axpy( m, -beta, mat(1_i4b:m,l+j), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                    umat(1_i4b:m,j) = scal*u_tmp(1_i4b:m)
#else
                    umat(1_i4b:m,j) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+j) )
#endif
!
                else
!
                    umat(1_i4b:m,j) = zero
!
                end if
!
#ifdef _BLAS
                call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                           1_i4b, one, umat(1_i4b:m,j), 1_i4b )
#else
                umat(1_i4b:m,j) = umat(1_i4b:m,j) + tau*matmul( umat(1_i4b:m,:j-1_i4b), prodvec(1_i4b:j-1_i4b) )
#endif
!
!               UPDATE mat(1:m,l+j) WITH G(l)* ... *G(l+j-1) AND ORTHOGONALIZE mat(1:m,l+j)
!               AGAINST THE l+j-1 COLUMN OF Q.
!
#ifdef _BLAS
                call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,l+j), 1_i4b, &
                           one, mat(1_i4b:m,l+j), 1_i4b )
                call axpy( m, -beta, mat(1_i4b:m,l+j-1_i4b), 1_i4b, mat(1_i4b:m,l+j), 1_i4b )
#else
                mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,l+j) ) &
                                   - beta*mat(1_i4b:m,l+j-1_i4b)
#endif
!
!               GENERATE THE l+j COLUM OF Q AND STORE IT IN mat(1:m,l+j).
!
                scal  = norm( mat(1_i4b:m,l+j) )
                normc = pythag( scal, beta )
!
                if ( scal<=fudge_ortho2*normc ) then
!
                    if ( do_reortho ) then
!
!                       REORTHOGONALIZATION IS REQUIRED.
!
                        call ortho_vec( mat(1_i4b:m,1_i4b:l+j-1_i4b), mat(1_i4b:m,l+j), &
                                        scal, normvec_orig=normc )
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    else
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                        scal = max( scal, safmin )
!
                        mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j)/scal
!
                    end if
!
                else
!
!                   NORMALIZATION IS ENOUGH.
!
                    mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j)/scal
!
                end if
!
!               COMPUTE THE l+j ROW OF THE BIDIAGONAL MATRIX.
!
                d(l+j) = scal
                e(l+j) = beta
!
            end do
!
        end do
!
        k = blksz*(n/blksz)
!
        if ( n/=k ) then
!
!           PROCESS THE LAST BLOCK OF COLUMNS OF mat .
!
            do l = k, n-1_i4b
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(l)
!               HOUSEHOLDER TRANSFORMATION.
!
                u_tmp(1_i4b:m) = zero
!
                if ( l==k ) then
!
                    do i = l+1_i4b, n
!
!                       COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) WITH A LAST BLOCK UPDATE.
!
                        u(1_i4b:m) = mat(1_i4b:m,i)
!
#ifdef _BLAS
                        call gemv( 'N', m, blksz, one, umat(1_i4b:m,1_i4b:blksz), m, vmat(1_i4b:blksz,i), &
                                   1_i4b, one, u(1_i4b:m), 1_i4b )
!
                        v(i) = dot( m, mat(1_i4b:m,l), 1_i4b, u(1_i4b:m), 1_i4b )
!
                        call axpy( m, v(i), u(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                        u(1_i4b:m) = u(1_i4b:m) + matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i) )
!
                        v(i) = dot_product( mat(1_i4b:m,l), u(1_i4b:m) )
!
                        u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*u(1_i4b:m)
#endif
!
                        mat(1_i4b:m,i) = u(1_i4b:m)
!
                    end do
!
                else
!
                    do i = l+1_i4b, n
!
#ifdef _BLAS
                        v(i) = dot( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,i), 1_i4b )
!
                        call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                        v(i) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,i) )
!
                        u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                    end do
!
                end if
!
!               COMPUTE P IN FACTORED FORM, STORING THE G(l) HOUSEHOLDER TRANSFORMATION
!               IN taup(l) AND p(l+1:n,l+1).
!
                alpha = v(l+1_i4b)        
!
                call hous1( v(l+1_i4b:n), tau, beta )
!
                p(l+1_i4b:n,l+1_i4b) = v(l+1_i4b:n)
                taup(l) = tau
!
                if ( tau/=zero ) then
!
!                   APPLY THE G(j) HOUSEHOLDER TRANSFORMATION TO mat(:m,l+1:n) .
!
                    scal = tau/( alpha - beta )
!
#ifdef _BLAS
                    call axpy( m, -beta, mat(1_i4b:m,l+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
!
                    call ger( m, n-l, scal, u_tmp(1_i4b:m), 1_i4b, v(l+1_i4b:n), 1_i4b, mat(1_i4b:m,l+1_i4b:n), m )
#else
                    u(1_i4b:m) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+1_i4b) )
!
                    do i = l+1_i4b, n
!
                        mat(1_i4b:m,i) = mat(1_i4b:m,i) + u(1_i4b:m)*v(i)
!
                    end do
#endif
!
                end if
!
!               ORTHOGONALIZE mat(1:m,l+1) AGAINST THE l COLUMN OF Q.
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
                mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) - beta*mat(1_i4b:m,l)
#endif
!
!               GENERATE THE l+1 COLUM OF Q AND STORE IT IN mat(1:m,l+1).
!
                scal  = norm( mat(1_i4b:m,l+1_i4b) )
                normc = pythag( scal, beta )
!
                if ( scal<=fudge_ortho2*normc ) then
!
                    if ( do_reortho ) then
!
!                       REORTHOGONALIZATION IS REQUIRED.
!
                        call ortho_vec( mat(1_i4b:m,1_i4b:l), mat(1_i4b:m,l+1_i4b),   &
                                        scal, normvec_orig=normc )
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    else
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                        scal = max( scal, safmin )
!
                        mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
                    end if
!
                else
!
!                   NORMALIZATION IS ENOUGH.
!
                    mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
                end if
!
!               COMPUTE THE l+1 ROW OF THE BIDIAGONAL MATRIX.
!
                d(l+1_i4b) = scal
                e(l+1_i4b) = beta
!
            end do
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
    if ( do_gen_p ) then
!
!       SET THE FIRST COLUMN AND ROW OF p TO THAT OF THE UNIT MATRIX.
!
        p(1_i4b,1_i4b)   = one
        p(2_i4b:n,1_i4b) = zero
        p(1_i4b,2_i4b:n) = zero
!
!       GENERATE RIGHT VECTORS BIDIAGONALIZING mat IN p .
!
        call ortho_gen_qr( p(2_i4b:n,2_i4b:n), taup(:n-1_i4b) )
!
    else
!
!       STORE THE ELEMENTARY REFLECTORS IN THE ARRAY p.
!
        p(1_i4b,1_i4b)   = -one
        p(2_i4b:n,1_i4b) = taup(1_i4b:n-1_i4b)
        p(1_i4b,2_i4b:n) = zero
!
    end if
!
!   SET failure INDICATOR IF REQUIRED.
!
    if ( present(failure) ) then
!
        failure = failure2
!
    end if
!
!
! END OF SUBROUTINE bd_cmp2_a
! ___________________________
!
    end subroutine bd_cmp2_a
!
! =========================================================================================
!
    subroutine bd_cmp2_b( mat, d, e, failure, reortho )
!
! Purpose
! _______
!
!   BD_CMP2 reduces a m-by-n matrix MAT with m >= n to upper bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal.
!
!   BD_CMP2 computes BD and Q using the one-sided Ralha-Barlow
!   bidiagonal reduction algorithm.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the general m-by-n matrix to be reduced.
!
!            On exit, the first n columns of Q are stored in
!            in MAT(1:m,1:n).
!
!            The shape of MAT must verify: size( MAT, 1 ) >= size( MAT, 2 ) = n .
!
!   D        (OUTPUT) real(stnd), dimension(:)
!            The diagonal elements of the bidiagonal matrix BD
!
!            The size of D must be size( MAT, 2 ) = n .
!
!   E        (OUTPUT) real(stnd), dimension(:)
!            The off-diagonal elements of the bidiagonal matrix BD:
!
!                E(i) = BD(i-1,i) for i = 2,3,...,n;
!
!            The size of E must be size( MAT, 2 ) = n .
!
!   FAILURE  (OUTPUT, OPTIONAL) logical(lgl)
!            On exit:
!
!            - FAILURE = false :  indicates successful exit ;
!            - FAILURE = true  :  indicates that MAT is nearly singular and some loss of
!              orthogonality of Q can be expected in the Ralha-Barlow one-sided
!              algorithm.
!
!            See further details.
!
!   REORTHO  (INPUT, OPTIONAL) logical(lgl)
!            If the optional argument REORTHO is set to true, reorthogonalization of the
!            matrix Q is performed (when needed).
!            If this optional logical argument is set to false, reorthogonalization of Q is never
!            performed and the columns of the  matrix Q can be far from orthogonal to each other
!            for input matrices MAT with a large condition number.
!
!            See further details.
!
!            The default is REORTHO = true.
!
!
! Further Details
! _______________
!
!   This subroutine is an implementation of the Ralha-Barlow one-sided method
!   to reduce a rectangular matrix MAT to bidiagonal form BD. Q is computed by
!   a recurrence relationship.
!
!   This is the blocked version of the algorithm. See the references (1), (2) and (3)
!   for further details. Note also that the blocked algorithm implemented here is
!   more efficient than the version described in the reference (3). Furthermore the
!   algorithm is parallelized if OPENMP is used.
!
!   Since Q is computed by a recurrence relationship, a loss of orthogonality of Q can be
!   observed when the rectangular matrix MAT is singular or nearly singular or has a large
!   condition number, see the reference (2) for details.
!
!   To correct partly this deficiency, partial reorthogonalization can be performed to ensure
!   orthogonality at the expense of speed of computation. The reorthogonalization uses the Gram-Schmidt
!   method described in the reference (4) and is activated if the optional logical argument REORTHO is
!   set to true.
!
!   The reference (2) also explains how to handle the case of an exactly singular
!   matrix MAT (a very rare event). However, in this subroutine, the partial reorthogonalization
!   described in the reference (4), which is used when REORTHO is set true, corrects automatically
!   this problem.
!
!   For further details, see:
!
!   (1) Ralha, R.M.S., 2003:
!           One-sided reduction to bidiagonal form.
!           Linear Algebra Appl., No 358, pp. 219-238.
!
!   (2) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!           A new stable bidiagonal reduction algorithm.
!           Linear Algebra Appl., No 397, pp. 35-84.
!
!   (3) Bosner, N., and Barlow, J.L., 2007:
!           Block and Parallel versions of one-sided bidiagonalization.
!           SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!   (4) Stewart, G.W., 2007:
!           Block Gram-Schmidt Orthogonalization.
!           Report TR-4823, Department of Computer Science, College Park, University of Maryland.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error5
    use Utilities,       only : assert_eq, merror, pythag
    use Random,          only : random_number_
    use Hous_Procedures, only : hous1
#ifdef _BLAS
    use BLAS_interfaces, only : dot, axpy, gemv, ger, gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: d, e
!
    logical(lgl), intent(in),  optional :: reortho
    logical(lgl), intent(out), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                :: tau, beta, alpha, scal, normc
    real(stnd), dimension(size(mat,1))        :: u, u_tmp
    real(stnd), dimension(size(mat,2))        :: v
    real(stnd), dimension(blksz2)             :: prodvec
    real(stnd), dimension(size(mat,1),blksz2) :: umat
    real(stnd), dimension(blksz2,size(mat,2)) :: vmat
!
    integer(i4b) :: m, n, i, i1, i2, i3, l, j, k, blksz
!
    logical(lgl) :: failure2, do_reortho
!
#ifdef _OPENMP
    real(stnd)                         :: beta2
    real(stnd), dimension(size(mat,1)) :: tmp
!
    integer(i4b) :: max_num_threads, num_threads, thread_num, iinc, irem, ioffset
!
    logical(lgl) :: apply_hous
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n =  assert_eq( int(size(mat,2),i4b),     &
                    int(size(d),i4b)      ,   &
                    int(size(e),i4b)      ,   &
                    name_proc )
!
!   CHECK DIMENSIONS OF THE INPUT MATRIX.
!
    if ( m<n ) then
!
        call merror( name_proc//svd_error5 )
!
    end if
!
    do_reortho = true
!
    if ( present( reortho ) ) then
!
        do_reortho = reortho
!
    end if
!
    failure2 = false
!
    if ( present(failure) ) then
!
        failure = failure2
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
!   ADJUST BLOCK SIZE IF NUMBER OF COLUMNS IS TOO SMALL.
!
    blksz = min( blksz2, n )
!
    e(1_i4b) = zero
!
!   REDUCE mat TO UPPER BIDIAGONAL FORM.
!
!   GENERATE THE FIRST COLUM OF Q AND STORE IT IN mat(1:m,1)
!
    d(1_i4b) = norm( mat(1_i4b:m,1_i4b) )
!
    if ( d(1_i4b)/=zero ) then
!
!       NORMALIZATION IS ENOUGH.
!
        mat(1_i4b:m,1_i4b) = mat(1_i4b:m,1_i4b)/d(1_i4b)
!
    else
!
!       IF THE NORM OF THE FIRST COLUMN OF mat IS ZERO,
!       SET IT TO A NONZERO NORMALIZED RANDOM VECTOR
!       AND PROCEED.
!
        call random_number_( u(:m) ) 
!
        u(:m) = u(:m) - half
        scal  = norm( u(:m) )
        mat(1_i4b:m,1_i4b) = u(:m)/scal
!
        failure2 = true
!
    end if
!
!   SPECIAL CASE IF mat(:,:) IS A COLUMN VECTOR.
!
    if ( n==1_i4b ) then
!
        if ( present(failure) ) then
!
            failure = failure2
!
        end if
!
        return
!
    end if
!
    u_tmp(1_i4b:m) = zero
!
#ifdef _OPENMP
!
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m*n>=omp_limit                      .and.       &
                  n>blksz                             .and.       &
                  i>1_i4b
!
    if ( test_par ) then
!
        max_num_threads = min( i, max_num_threads_bd_cmp2 )
!
!$OMP PARALLEL PRIVATE(num_threads,thread_num,iinc,irem,ioffset,           &
!$OMP                  i,i1,i2,i3,l,j,k,alpha,beta,tau,scal,normc,u,tmp),  &
!$OMP          NUM_THREADS(max_num_threads)
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
!       PROCESS THE FIRST BLOCK OF COLUMNS OF mat .
!
!       START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(1)
!       HOUSEHOLDER TRANSFORMATION.
!
        u(1_i4b:m)   = mat(1_i4b:m,1_i4b)
        tmp(1_i4b:m) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
        do i = 2_i4b, n
!
#ifdef _BLAS
            v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b )
!
            call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
            v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )
!
            tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
        end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compu0)
!
#ifdef _BLAS
        call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
        u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu0)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!       COMPUTE P IN FACTORED FORM, STORING THE G(1) HOUSEHOLDER TRANSFORMATION
!       IN tau AND v2(2:n).
!
        alpha = v(2_i4b)        
!
        call hous1( v(2_i4b:n), tau, beta )         
!
!       INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
        if ( tau/=zero ) then
!
            scal = tau/( alpha - beta )
!
#ifdef _BLAS
            call axpy( m, -beta, mat(1_i4b:m,2_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
            umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
            umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,2_i4b) )
#endif
        else
!
            umat(1_i4b:m,1_i4b) = zero
!
        end if
!
        umat(1_i4b:m,blksz) = zero
!
        vmat(1_i4b:blksz,1_i4b:n) = zero
        vmat(1_i4b,2_i4b:n)       = v(2_i4b:n)
!
!       UPDATE mat(1:m,2) WITH G(1) AND ORTHOGONALIZE mat(1:m,2) AGAINST THE FIRST COLUMN OF Q.
!
#ifdef _BLAS
        call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,2_i4b), 1_i4b )
        call axpy( m, -beta, mat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,2_i4b), 1_i4b )
#else
        mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b) + umat(1_i4b:m,1_i4b) - beta*mat(1_i4b:m,1_i4b)
#endif
!
!       GENERATE THE SECOND COLUM OF Q AND STORE IT IN mat(1:m,2)
!
        scal  = norm( mat(1_i4b:m,2_i4b) )
        normc = pythag( scal, beta )
!
        if ( scal<=fudge_ortho2*normc ) then
!
!           REORTHOGONALIZATION IS REQUIRED.
!
            call ortho_vec( mat(1_i4b:m,1_i4b:1_i4b), mat(1_i4b:m,2_i4b),   &
                            scal, normvec_orig=normc )
!
            failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
        else
!
!           NORMALIZATION IS ENOUGH.
!
            mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b)/scal
!
        end if
!
!       COMPUTE SECOND ROW OF THE BIDIAGONAL MATRIX.
!
        d(2_i4b) = scal
        e(2_i4b) = beta
!
        u_tmp(1_i4b:m) = zero  
!
!$OMP END SINGLE
!
        do j = 2_i4b, blksz-1_i4b
!
!           START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(j)
!           HOUSEHOLDER TRANSFORMATION.
!
            u(1_i4b:m) = mat(1_i4b:m,j)
!
!$OMP DO SCHEDULE(STATIC)
!
            do i = 1_i4b, j-1_i4b
!
#ifdef _BLAS
                prodvec(i) = dot( m, u(1_i4b:m), 1_i4b, umat(1_i4b:m,i), 1_i4b )
#else
                prodvec(i) = dot_product( u(1_i4b:m), umat(1_i4b:m,i) )
#endif
!
            end do
!
!$OMP END DO
!
            tmp(1_i4b:m) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
            do i = j+1_i4b, n
!
#ifdef _BLAS
                v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b ) +                      &
                       dot( j-1_i4b, prodvec(1_i4b:j-1_i4b), 1_i4b, vmat(1_i4b:j-1_i4b,i), 1_i4b )
!
                call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) ) +                &
                       dot_product( prodvec(1_i4b:j-1_i4b), vmat(1_i4b:j-1_i4b,i) )
!
                tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
            end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compu1)
!
#ifdef _BLAS
            call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu1)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!           COMPUTE P IN FACTORED FORM, STORING THE G(j) HOUSEHOLDER TRANSFORMATION
!           IN tau AND v(j+1:n).
!
            alpha = v(j+1_i4b)        
!
            call hous1( v(j+1_i4b:n), tau, beta )
!
!           SAVE G(j) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
            vmat(j,j+1_i4b:n) = v(j+1_i4b:n)
!
#ifdef _BLAS
            call gemv( 'N', j-1_i4b, n-j, one, vmat(1_i4b:blksz2,j+1_i4b:n), blksz2, &
                       v(j+1_i4b:n), 1_i4b, zero, prodvec, 1_i4b )
#else
            prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,j+1_i4b:n), v(j+1_i4b:n) )
#endif
!
            if ( tau/=zero ) then
!
                scal = tau/( alpha - beta )
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,j+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                umat(1_i4b:m,j) = scal*u_tmp(1_i4b:m)
#else
                umat(1_i4b:m,j) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,j+1_i4b) )
#endif
!
            else
!
                umat(1_i4b:m,j) = zero
!
            end if
!
#ifdef _BLAS
            call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                        1_i4b, one, umat(1_i4b:m,j), 1_i4b )
#else
            umat(1_i4b:m,j) = umat(1_i4b:m,j) + tau*matmul( umat(1_i4b:m,1_i4b:j-1_i4b), prodvec(1_i4b:j-1_i4b) )
#endif
!
!           UPDATE mat(1:m,j+1) WITH G(1)* ... *G(j) AND ORTHOGONALIZE mat(1:m,j+1) AGAINST THE j COLUMN OF Q.
!
#ifdef _BLAS
            call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,j+1_i4b), 1_i4b,     &
                       one, mat(1_i4b:m,j+1_i4b), 1_i4b )
            call axpy( m, -beta, mat(1_i4b:m,j), 1_i4b, mat(1_i4b:m,j+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,j+1_i4b) ) &
                                   - beta*mat(1_i4b:m,j)
#endif
!
!           GENERATE THE j+1 COLUM OF Q AND STORE IT IN mat(1:m,j+1).
!
            scal  = norm( mat(1_i4b:m,j+1_i4b) )
            normc = pythag( scal, beta )
!
            if ( scal<=fudge_ortho2*normc ) then
!
                if ( do_reortho ) then
!
!                   REORTHOGONALIZATION IS REQUIRED.
!
                    call ortho_vec( mat(1_i4b:m,1_i4b:j), mat(1_i4b:m,j+1_i4b),   &
                                    scal, normvec_orig=normc )
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                else
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    scal = max( scal, safmin )
!
                    mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b)/scal
!
                end if
!
            else
!
!               NORMALIZATION IS ENOUGH.
!
                mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b)/scal
!
            end if
!
!           COMPUTE THE j+1 ROW OF THE BIDIAGONAL MATRIX.
!
            d(j+1_i4b) = scal
            e(j+1_i4b) = beta
!
            u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
        end do
!
!       PROCESS THE SECOND AND FOLLOWING BLOCKS OF COLUMNS OF mat .
!
        do l = blksz, n-blksz, blksz
!
!           PERFORM A BLOCK UPDATE OF THE TRAILING MATRIX AND START COMPUTATIONS FOR
!           GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION.
!
            tmp(1_i4b:m) = zero
!
!           DISTRIBUTE THE WORK AMONG THE THREADS FOR THE BLOCK UPDATE OF THE TRAILING MATRIX.
!
            i  = n - l
            irem = mod( i, num_threads )
            iinc = ( i - irem )/num_threads
!
            if ( thread_num<irem ) then
                iinc    = iinc + 1_i4b
                ioffset = l + 1_i4b
            else
                ioffset = irem + l + 1_i4b
            end if
!
            i1 = thread_num*iinc + ioffset
            i2 = i1 + iinc - 1_i4b
            i3 = i2 - i1 + 1_i4b
!
            if ( i3>0_i4b ) then
!
#ifdef _BLAS
!
!               COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) AS A BLOCK UPDATE.
!
                call gemm( 'N', 'N', m, i3, blksz, one, umat(1_i4b:m,1_i4b:blksz), m,   &
                           vmat(1_i4b:blksz2,i1:i2), blksz2, one, mat(1_i4b:m,i1:i2), m )
!
!               START GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION
!               WITH A BLAS2.5 GEMVT OPERATOR.
!
!                call gemv( 'T', m, i3, one,  mat(1_i4b:m,i1:i2), m, mat(1_i4b:m,l), &
!                           1_i4b, zero, v(i1:i2), 1_i4b )
!
!                call gemv( 'N', m, i3, one,  mat(1_i4b:m,i1:i2), m, v(i1:i2), &
!                           1_i4b, one, tmp(1_i4b:m), 1_i4b )
!
                do j = i1, i2
!
                    v(j) = dot( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,j), 1_i4b )
!
                    call axpy( m, v(j), mat(1_i4b:m,j), 1_i4b, tmp(1_i4b:m), 1_i4b )
!
                end do
#else
!
!               COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) AS A BLOCK UPDATE.
!
                mat(1_i4b:m,i1:i2) = mat(1_i4b:m,i1:i2) +                                       &
                                     matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i1:i2) )
!
!               START GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION
!               WITH A BLAS2.5 GEMVT OPERATOR.
!
!                v(i1:i2) = matmul( mat(1_i4b:m,l), mat(1_i4b:m,i1:i2) )
!
!                tmp(1_i4b:m) = tmp(1_i4b:m) +  matmul( mat(1_i4b:m,i1:i2), v(i1:i2) )
!
                do j = i1, i2
!
                    v(j) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,j) )
!
                    tmp(1_i4b:m) = tmp(1_i4b:m) +  v(j)*mat(1_i4b:m,j)
!
                end do
#endif
!
            end if
!
!$OMP CRITICAL (compu2)
!
#ifdef _BLAS
            call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu2)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!           COMPUTE P IN FACTORED FORM, STORING THE G(l) HOUSEHOLDER TRANSFORMATION
!           IN tau AND v(l+1_i4b:n).
!
            alpha = v(l+1_i4b)        
!
            call hous1( v(l+1_i4b:n), tau, beta )
!
!           INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
            if ( tau/=zero ) then
!
                scal = tau/( alpha - beta )
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,l+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
                umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+1_i4b) )
#endif
!
            else
!
                umat(1_i4b:m,1_i4b) = zero
!
            end if
!
            vmat(:blksz,:n)       = zero
            vmat(1_i4b,l+1_i4b:n) = v(l+1_i4b:n)
!
!           UPDATE mat(1:m,l+1) WITH G(l) AND ORTHOGONALIZE mat(1:m,l+1) AGAINST THE l COLUMN OF Q.
!
#ifdef _BLAS
            call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
            call axpy( m, -beta, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) + umat(1_i4b:m,1_i4b) - beta*mat(1_i4b:m,l)
#endif
!
!           GENERATE THE l+1 COLUM OF Q AND STORE IT IN mat(1:m,l+1).
!
            scal  = norm( mat(1_i4b:m,l+1_i4b) )
            normc = pythag( scal, beta )
!
            if ( scal<=fudge_ortho2*normc ) then
!
                if ( do_reortho ) then
!
!                   REORTHOGONALIZATION IS REQUIRED.
!
                    call ortho_vec( mat(1_i4b:m,1_i4b:l), mat(1_i4b:m,l+1_i4b),   &
                                    scal, normvec_orig=normc )
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                else
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    scal = max( scal, safmin )
!
                    mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
                end if
!
            else
!
!               NORMALIZATION IS ENOUGH.
!
                mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
            end if
!
!           COMPUTE THE l+1 ROW OF THE BIDIAGONAL MATRIX.
!
            d(l+1_i4b) = scal
            e(l+1_i4b) = beta
!
            u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
            do j = 2_i4b, blksz
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(l+j-1)
!               HOUSEHOLDER TRANSFORMATION.
!
                u(1_i4b:m) = mat(1_i4b:m,l+j-1_i4b)
!
!$OMP DO SCHEDULE(STATIC)
!
                do i = 1_i4b, j-1_i4b
!
#ifdef _BLAS
                    prodvec(i) = dot( m, u(1_i4b:m), 1_i4b, umat(1_i4b:m,i), 1_i4b )
#else
                    prodvec(i) = dot_product( u(1_i4b:m), umat(1_i4b:m,i) )
#endif
!
                end do
!
!$OMP END DO
!
                tmp(1_i4b:m) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
                do i = l+j, n
!
#ifdef _BLAS
                    v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b ) +                      &
                           dot( j-1_i4b, prodvec(1_i4b:j-1_i4b), 1_i4b, vmat(1_i4b:j-1_i4b,i), 1_i4b )
!
                    call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                    v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )                  + &
                           dot_product( prodvec(1_i4b:j-1_i4b), vmat(1_i4b:j-1_i4b,i) )
!
                    tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compu3)
!
#ifdef _BLAS
                call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu3)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!               COMPUTE P IN FACTORED FORM, STORING THE G(l+j-1) HOUSEHOLDER TRANSFORMATION
!               IN tau AND v(l+j:n).
!
                alpha = v(l+j)        
!
                call hous1( v(l+j:n), tau, beta )
!
!               SAVE G(l+j-1) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
                vmat(j,l+j:n) = v(l+j:n)
!
#ifdef _BLAS
                call gemv( 'N', j-1_i4b, n-j-l+1_i4b, one, vmat(1_i4b:blksz2,l+j:n), blksz2, &
                           v(l+j:n), 1_i4b, zero, prodvec, 1_i4b )
#else
                prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,l+j:n), v(l+j:n) )
#endif
!
                if ( tau/=zero ) then
!
                    scal = tau/( alpha - beta )
!
#ifdef _BLAS
                    call axpy( m, -beta, mat(1_i4b:m,l+j), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                    umat(1_i4b:m,j) = scal*u_tmp(1_i4b:m)
#else
                    umat(1_i4b:m,j) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+j) )
#endif
!
                else
!
                    umat(1_i4b:m,j) = zero
!
                end if
!
#ifdef _BLAS
                call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                           1_i4b, one, umat(1_i4b:m,j), 1_i4b )
#else
                umat(1_i4b:m,j) = umat(1_i4b:m,j) + tau*matmul( umat(1_i4b:m,:j-1_i4b), prodvec(1_i4b:j-1_i4b) )
#endif
!
!               UPDATE mat(1:m,l+j) WITH G(l)* ... *G(l+j-1) AND ORTHOGONALIZE mat(1:m,l+j)
!               AGAINST THE l+j-1 COLUMN OF Q.
!
#ifdef _BLAS
                call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,l+j), 1_i4b, &
                           one, mat(1_i4b:m,l+j), 1_i4b )
                call axpy( m, -beta, mat(1_i4b:m,l+j-1_i4b), 1_i4b, mat(1_i4b:m,l+j), 1_i4b )
#else
                mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,l+j) ) &
                                   - beta*mat(1_i4b:m,l+j-1_i4b)
#endif
!
!               GENERATE THE l+j COLUM OF Q AND STORE IT IN mat(1:m,l+j).
!
                scal  = norm( mat(1_i4b:m,l+j) )
                normc = pythag( scal, beta )
!
                if ( scal<=fudge_ortho2*normc ) then
!
                    if ( do_reortho ) then
!
!                       REORTHOGONALIZATION IS REQUIRED.
!
                        call ortho_vec( mat(1_i4b:m,1_i4b:l+j-1_i4b), mat(1_i4b:m,l+j), &
                                        scal, normvec_orig=normc )
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    else
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                        scal = max( scal, safmin )
!
                        mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j)/scal
!
                    end if
!
                else
!
!                   NORMALIZATION IS ENOUGH.
!
                    mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j)/scal
!
                end if
!
!               COMPUTE THE l+j ROW OF THE BIDIAGONAL MATRIX.
!
                d(l+j) = scal
                e(l+j) = beta
!
                u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
            end do
!
        end do
!
        k = blksz*(n/blksz)
!
        if ( n/=k ) then
!
!           PROCESS THE LAST BLOCK OF COLUMNS OF mat .
!
            do l = k, n-1_i4b
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(l)
!               HOUSEHOLDER TRANSFORMATION.
!
                tmp(1_i4b:m) = zero
!
                if ( l==k ) then
!
!$OMP DO SCHEDULE(STATIC)
!
                    do i = l+1_i4b, n
!
!                       COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) WITH A LAST BLOCK UPDATE.
!
                        u(1_i4b:m) = mat(1_i4b:m,i)
!
#ifdef _BLAS
                        call gemv( 'N', m, blksz, one, umat(1_i4b:m,1_i4b:blksz), m, vmat(1_i4b:blksz,i), &
                                   1_i4b, one, u(1_i4b:m), 1_i4b )
!
                        v(i) = dot( m, mat(1_i4b:m,l), 1_i4b, u(1_i4b:m), 1_i4b )
!
                        call axpy( m, v(i), u(1_i4b:m), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                        u(1_i4b:m) = u(1_i4b:m) + matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i) )
!
                        v(i) = dot_product( mat(1_i4b:m,l), u(1_i4b:m) )
!
                        tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*u(1_i4b:m)
#endif
!
                        mat(1_i4b:m,i) = u(1_i4b:m)
!
                    end do
!
!$OMP END DO NOWAIT
!
                else
!
!$OMP DO SCHEDULE(STATIC)
!
                    do i = l+1_i4b, n
!
#ifdef _BLAS
                        v(i) = dot( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,i), 1_i4b )
!
                        call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                        v(i) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,i) )
!
                        tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                    end do
!
!$OMP END DO NOWAIT
!
                end if
!
!$OMP CRITICAL (compu4)
!
#ifdef _BLAS
                call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu4)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!               COMPUTE P IN FACTORED FORM, STORING THE G(l) HOUSEHOLDER TRANSFORMATION
!               IN tau AND v(l+1_i4b:n).
!
                alpha = v(l+1_i4b)        
!
                call hous1( v(l+1_i4b:n), tau, beta2 )
!
                apply_hous = tau/=zero
!
                if ( apply_hous ) then
!
                    scal = tau/( alpha - beta2 )
!
#ifdef _BLAS
                    call axpy( m, -beta2, mat(1_i4b:m,l+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                    umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
                    umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta2*mat(1_i4b:m,l+1_i4b) )
#endif
!
!                   APPLY THE G(l) HOUSEHOLDER TRANSFORMATION TO mat(:m,l+1) .
!
#ifdef _BLAS
                    call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
                    mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) + umat(1_i4b:m,1_i4b)
#endif
!
                end if
!
!$OMP END SINGLE
!
                if ( apply_hous ) then
!
!                   APPLY THE G(l) HOUSEHOLDER TRANSFORMATION TO mat(:m,l+2:n) .
!
!$OMP DO SCHEDULE(STATIC)
!
                    do i = l+2_i4b, n
!
#ifdef _BLAS
                        call axpy( m, v(i), umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,i), 1_i4b )
#else
                        mat(1_i4b:m,i) = mat(1_i4b:m,i) + umat(1_i4b:m,1_i4b)*v(i)
#endif
!
                    end do
!
!$OMP END DO NOWAIT
!
                end if
!
!$OMP SINGLE
!
!               ORTHOGONALIZE mat(1:m,l+1) AGAINST THE l COLUMN OF Q.
!
#ifdef _BLAS
                call axpy( m, -beta2, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
                mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) - beta2*mat(1_i4b:m,l)
#endif
!
!               GENERATE THE l+1 COLUM OF Q AND STORE IT IN mat(1:m,l+1).
!
                scal  = norm( mat(1_i4b:m,l+1_i4b) )
                normc = pythag( scal, beta2 )
!
                if ( scal<=fudge_ortho2*normc ) then
!
                    if ( do_reortho ) then
!
!                       REORTHOGONALIZATION IS REQUIRED.
!
                        call ortho_vec( mat(1_i4b:m,1_i4b:l), mat(1_i4b:m,l+1_i4b),   &
                                        scal, normvec_orig=normc )
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    else
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                        scal = max( scal, safmin )
!
                        mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
                    end if
!
                else
!
!                   NORMALIZATION IS ENOUGH.
!
                    mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
                end if
!
!               COMPUTE THE l+1 ROW OF THE BIDIAGONAL MATRIX.
!
                d(l+1_i4b) = scal
                e(l+1_i4b) = beta2
!
                u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
            end do
!
        end if
!
!$OMP END PARALLEL
!
    else
#endif
!
!       PROCESS THE FIRST BLOCK OF COLUMNS OF mat .
!
!       START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(1)
!       HOUSEHOLDER TRANSFORMATION.
!
        u(1_i4b:m) = mat(1_i4b:m,1_i4b)
!
        do i = 2_i4b, n
!
#ifdef _BLAS
            v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b )
!
            call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
            v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )
!
            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
        end do
!
!       COMPUTE P IN FACTORED FORM, STORING THE G(1) HOUSEHOLDER TRANSFORMATION
!       IN tau AND v(2_i4b:n).
!
        alpha = v(2_i4b)        
!
        call hous1( v(2_i4b:n), tau, beta )         
!
!       INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
        if ( tau/=zero ) then
!
            scal = tau/( alpha - beta )
!
#ifdef _BLAS
            call axpy( m, -beta, mat(1_i4b:m,2_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
            umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
            umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,2_i4b) )
#endif
!
        else
!
            umat(1_i4b:m,1_i4b) = zero
!
        end if
!
        umat(1_i4b:m,blksz) = zero
!
        vmat(1_i4b:blksz,1_i4b:n) = zero
        vmat(1_i4b,2_i4b:n)       = v(2_i4b:n)
!
!       UPDATE mat(1:m,2) WITH G(1) AND ORTHOGONALIZE mat(1:m,2) AGAINST THE FIRST COLUMN OF Q.
!
#ifdef _BLAS
        call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,2_i4b), 1_i4b )
        call axpy( m, -beta, mat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,2_i4b), 1_i4b )
#else
        mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b) + umat(1_i4b:m,1_i4b) - beta*mat(1_i4b:m,1_i4b)
#endif
!
!       GENERATE THE SECOND COLUM OF Q AND STORE IT IN mat(1:m,2)
!
        scal  = norm( mat(1_i4b:m,2_i4b) )
        normc = pythag( scal, beta )
!
        if ( scal<=fudge_ortho2*normc ) then
!
!           REORTHOGONALIZATION IS REQUIRED.
!
            call ortho_vec( mat(1_i4b:m,1_i4b:1_i4b), mat(1_i4b:m,2_i4b),   &
                            scal, normvec_orig=normc )
!
            failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
        else
!
!           NORMALIZATION IS ENOUGH.
!
            mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b)/scal
!
        end if
!
!       COMPUTE SECOND ROW OF THE BIDIAGONAL MATRIX.
!
        d(2_i4b) = scal
        e(2_i4b) = beta
!
        do j = 2_i4b, blksz-1_i4b
!
!           START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(j)
!           HOUSEHOLDER TRANSFORMATION.
!
            u(1_i4b:m) = mat(1_i4b:m,j)
!
#ifdef _BLAS
            call gemv( 'T', m, j-1_i4b, one,  umat(1_i4b:m,1_i4b:j-1_i4b), m, &
                       u(1_i4b:m), 1_i4b, zero, prodvec(1_i4b:j-1_i4b), 1_i4b )
#else
            prodvec(1_i4b:j-1_i4b) = matmul( u(1_i4b:m), umat(1_i4b:m,1_i4b:j-1_i4b) )
#endif
!
            u_tmp(1_i4b:m) = zero
!
            do i = j+1_i4b, n
!
#ifdef _BLAS
                v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b ) +                      &
                       dot( j-1_i4b, prodvec(1_i4b:j-1_i4b), 1_i4b, vmat(1_i4b:j-1_i4b,i), 1_i4b )
!
                call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) ) +                &
                       dot_product( prodvec(1_i4b:j-1_i4b), vmat(1_i4b:j-1_i4b,i) )
!
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
            end do
!
!           COMPUTE P IN FACTORED FORM, STORING THE G(j) HOUSEHOLDER TRANSFORMATION
!           IN tau AND v(j+1_i4b:n).
!
            alpha = v(j+1_i4b)        
!
            call hous1( v(j+1_i4b:n), tau, beta )
!
!           SAVE G(j) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
            vmat(j,j+1_i4b:n) = v(j+1_i4b:n)
!
#ifdef _BLAS
            call gemv( 'N', j-1_i4b, n-j, one, vmat(1_i4b:blksz2,j+1_i4b:n), blksz2, &
                       v(j+1_i4b:n), 1_i4b, zero, prodvec, 1_i4b )
#else
            prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,j+1_i4b:n), v(j+1_i4b:n) )
#endif
!
            if ( tau/=zero ) then
!
                scal = tau/( alpha - beta )
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,j+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                umat(1_i4b:m,j) = scal*u_tmp(1_i4b:m)
#else
                umat(1_i4b:m,j) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,j+1_i4b) )
#endif
!
            else
!
                umat(1_i4b:m,j) = zero
!
            end if
!
#ifdef _BLAS
            call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                        1_i4b, one, umat(1_i4b:m,j), 1_i4b )
#else
            umat(1_i4b:m,j) = umat(1_i4b:m,j) + tau*matmul( umat(1_i4b:m,1_i4b:j-1_i4b), prodvec(1_i4b:j-1_i4b) )
#endif
!
!           UPDATE mat(1:m,j+1) WITH G(1)* ... *G(j) AND ORTHOGONALIZE mat(1:m,j+1) AGAINST THE j COLUMN OF Q.
!
#ifdef _BLAS
            call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,j+1_i4b), 1_i4b,     &
                       one, mat(1_i4b:m,j+1_i4b), 1_i4b )
            call axpy( m, -beta, mat(1_i4b:m,j), 1_i4b, mat(1_i4b:m,j+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,j+1_i4b) ) &
                                   - beta*mat(1_i4b:m,j)
#endif
!
!           GENERATE THE j+1 COLUM OF Q AND STORE IT IN mat(1:m,j+1).
!
            scal  = norm( mat(1_i4b:m,j+1_i4b) )
            normc = pythag( scal, beta )
!
            if ( scal<=fudge_ortho2*normc ) then
!
                if ( do_reortho ) then
!
!                   REORTHOGONALIZATION IS REQUIRED.
!
                    call ortho_vec( mat(1_i4b:m,1_i4b:j), mat(1_i4b:m,j+1_i4b),   &
                                    scal, normvec_orig=normc )
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                else
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    scal = max( scal, safmin )
!
                    mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b)/scal
!
                end if
!
            else
!
!               NORMALIZATION IS ENOUGH.
!
                mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b)/scal
!
            end if
!
!           COMPUTE THE j+1 ROW OF THE BIDIAGONAL MATRIX.
!
            d(j+1_i4b) = scal
            e(j+1_i4b) = beta
!
        end do
!
!       PROCESS THE SECOND AND FOLLOWING BLOCKS OF COLUMNS OF mat .
!
        do l = blksz, n-blksz, blksz
!
!           PERFORM A BLOCK UPDATE OF THE TRAILING MATRIX AND START COMPUTATIONS FOR
!           GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION.
!
            u_tmp(1_i4b:m) = zero
!
            i1 = l + 1_i4b
            i2 = n
            i3 = i2 - l
!
#ifdef _BLAS
!
!           COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) AS A BLOCK UPDATE.
!
            call gemm( 'N', 'N', m, i3, blksz, one, umat(1_i4b:m,1_i4b:blksz), m,    &
                       vmat(1_i4b:blksz2,i1:i2), blksz2, one, mat(1_i4b:m,i1:i2), m  )
!
!           START GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION
!           WITH A BLAS2.5 GEMVT OPERATOR.
!
!            call gemv( 'T', m, i3, one,  mat(1_i4b:m,i1:i2), m, mat(1_i4b:m,l), &
!                       1_i4b, zero, v(i1:i2), 1_i4b )
!
!            call gemv( 'N', m, i3, one,  mat(1_i4b:m,i1:i2), m, v(i1:i2), &
!                       1_i4b, one, u_tmp(1_i4b:m), 1_i4b )
!
            do j = i1, i2
!
                v(j) = dot( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,j), 1_i4b )
!
                call axpy( m, v(j), mat(1_i4b:m,j), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
!
            end do
#else
!
!           COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) AS A BLOCK UPDATE.
!
            mat(1_i4b:m,i1:i2) = mat(1_i4b:m,i1:i2) +                                       &
                                 matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i1:i2) )
!
!           START GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION
!           WITH A BLAS2.5 GEMVT OPERATOR.
!
!            v(i1:i2) = matmul( mat(1_i4b:m,l), mat(1_i4b:m,i1:i2) )
!
!            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) +  matmul( mat(1_i4b:m,i1:i2), v(i1:i2) )
!
            do j = i1, i2
!
                v(j) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,j) )
!
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) +  v(j)*mat(1_i4b:m,j)
!
            end do
#endif
!
!           COMPUTE P IN FACTORED FORM, STORING THE G(l) HOUSEHOLDER TRANSFORMATION
!           IN tau AND v(l+1_i4b:n).
!
            alpha = v(l+1_i4b)        
!
            call hous1( v(l+1_i4b:n), tau, beta )
!
!           INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
            if ( tau/=zero ) then
!
                scal = tau/( alpha - beta )
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,l+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
                umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+1_i4b) )
#endif
!
            else
!
                umat(1_i4b:m,1_i4b) = zero
!
            end if
!
            vmat(:blksz,:n)       = zero
            vmat(1_i4b,l+1_i4b:n) = v(l+1_i4b:n)
!
!           UPDATE mat(1:m,l+1) WITH G(l) AND ORTHOGONALIZE mat(1:m,l+1) AGAINST THE l COLUMN OF Q.
!
#ifdef _BLAS
            call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
            call axpy( m, -beta, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) + umat(1_i4b:m,1_i4b) - beta*mat(1_i4b:m,l)
#endif
!
!           GENERATE THE l+1 COLUM OF Q AND STORE IT IN mat(1:m,l+1).
!
            scal  = norm( mat(1_i4b:m,l+1_i4b) )
            normc = pythag( scal, beta )
!
            if ( scal<=fudge_ortho2*normc ) then
!
                if ( do_reortho ) then
!
!                   REORTHOGONALIZATION IS REQUIRED.
!
                    call ortho_vec( mat(1_i4b:m,1_i4b:l), mat(1_i4b:m,l+1_i4b),   &
                                    scal, normvec_orig=normc )
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                else
!
                    failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    scal = max( scal, safmin )
!
                    mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
                end if
!
            else
!
!               NORMALIZATION IS ENOUGH.
!
                mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
            end if
!
!           COMPUTE THE l+1 ROW OF THE BIDIAGONAL MATRIX.
!
            d(l+1_i4b) = scal
            e(l+1_i4b) = beta
!
            do j = 2_i4b, blksz
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(l+j-1)
!               HOUSEHOLDER TRANSFORMATION.
!
                u(1_i4b:m) = mat(1_i4b:m,l+j-1_i4b)
!
#ifdef _BLAS
                call gemv( 'T', m, j-1_i4b, one,  umat(1_i4b:m,1_i4b:j-1_i4b), m, &
                           u(1_i4b:m), 1_i4b, zero, prodvec(1_i4b:j-1_i4b), 1_i4b )
#else
                prodvec(1_i4b:j-1_i4b) = matmul( u(1_i4b:m), umat(1_i4b:m,1_i4b:j-1_i4b) )
#endif
!
                u_tmp(1_i4b:m) = zero
!
                do i = l+j, n
!
#ifdef _BLAS
                    v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b ) +                      &
                           dot( j-1_i4b, prodvec(1_i4b:j-1_i4b), 1_i4b, vmat(1_i4b:j-1_i4b,i), 1_i4b )
!
                    call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                    v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )                  + &
                           dot_product( prodvec(1_i4b:j-1_i4b), vmat(1_i4b:j-1_i4b,i) )
!
                    u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                end do
!
!               COMPUTE P IN FACTORED FORM, STORING THE G(l+j-1) HOUSEHOLDER TRANSFORMATION
!               IN tau AND v(l+j:n).
!
                alpha = v(l+j)        
!
                call hous1( v(l+j:n), tau, beta )
!
!               SAVE G(l+j-1) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
                vmat(j,l+j:n) = v(l+j:n)
!
#ifdef _BLAS
                call gemv( 'N', j-1_i4b, n-j-l+1_i4b, one, vmat(1_i4b:blksz2,l+j:n), blksz2, &
                           v(l+j:n), 1_i4b, zero, prodvec, 1_i4b )
#else
                prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,l+j:n), v(l+j:n) )
#endif
!
                if ( tau/=zero ) then
!
                    scal = tau/( alpha - beta )
!
#ifdef _BLAS
                    call axpy( m, -beta, mat(1_i4b:m,l+j), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                    umat(1_i4b:m,j) = scal*u_tmp(1_i4b:m)
#else
                    umat(1_i4b:m,j) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+j) )
#endif
!
                else
!
                    umat(1_i4b:m,j) = zero
!
                end if
!
#ifdef _BLAS
                call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                           1_i4b, one, umat(1_i4b:m,j), 1_i4b )
#else
                umat(1_i4b:m,j) = umat(1_i4b:m,j) + tau*matmul( umat(1_i4b:m,:j-1_i4b), prodvec(1_i4b:j-1_i4b) )
#endif
!
!               UPDATE mat(1:m,l+j) WITH G(l)* ... *G(l+j-1) AND ORTHOGONALIZE mat(1:m,l+j)
!               AGAINST THE l+j-1 COLUMN OF Q.
!
#ifdef _BLAS
                call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,l+j), 1_i4b, &
                           one, mat(1_i4b:m,l+j), 1_i4b )
                call axpy( m, -beta, mat(1_i4b:m,l+j-1_i4b), 1_i4b, mat(1_i4b:m,l+j), 1_i4b )
#else
                mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,l+j) ) &
                                   - beta*mat(1_i4b:m,l+j-1_i4b)
#endif
!
!               GENERATE THE l+j COLUM OF Q AND STORE IT IN mat(1:m,l+j).
!
                scal  = norm( mat(1_i4b:m,l+j) )
                normc = pythag( scal, beta )
!
                if ( scal<=fudge_ortho2*normc ) then
!
                    if ( do_reortho ) then
!
!                       REORTHOGONALIZATION IS REQUIRED.
!
                        call ortho_vec( mat(1_i4b:m,1_i4b:l+j-1_i4b), mat(1_i4b:m,l+j), &
                                        scal, normvec_orig=normc )
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    else
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                        scal = max( scal, safmin )
!
                        mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j)/scal
!
                    end if
!
                else
!
!                   NORMALIZATION IS ENOUGH.
!
                    mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j)/scal
!
                end if
!
!               COMPUTE THE l+j ROW OF THE BIDIAGONAL MATRIX.
!
                d(l+j) = scal
                e(l+j) = beta
!
            end do
!
        end do
!
        k = blksz*(n/blksz)
!
        if ( n/=k ) then
!
!           PROCESS THE LAST BLOCK OF COLUMNS OF mat .
!
            do l = k, n-1_i4b
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(l)
!               HOUSEHOLDER TRANSFORMATION.
!
                u_tmp(1_i4b:m) = zero
!
                if ( l==k ) then
!
                    do i = l+1_i4b, n
!
!                       COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) WITH A LAST BLOCK UPDATE.
!
                        u(1_i4b:m) = mat(1_i4b:m,i)
!
#ifdef _BLAS
                        call gemv( 'N', m, blksz, one, umat(1_i4b:m,1_i4b:blksz), m, vmat(1_i4b:blksz,i), &
                                   1_i4b, one, u(1_i4b:m), 1_i4b )
!
                        v(i) = dot( m, mat(1_i4b:m,l), 1_i4b, u(1_i4b:m), 1_i4b )
!
                        call axpy( m, v(i), u(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                        u(1_i4b:m) = u(1_i4b:m) + matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i) )
!
                        v(i) = dot_product( mat(1_i4b:m,l), u(1_i4b:m) )
!
                        u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*u(1_i4b:m)
#endif
!
                        mat(1_i4b:m,i) = u(1_i4b:m)
!
                    end do
!
                else
!
                    do i = l+1_i4b, n
!
#ifdef _BLAS
                        v(i) = dot( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,i), 1_i4b )
!
                        call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                        v(i) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,i) )
!
                        u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                    end do
!
                end if
!
!               COMPUTE P IN FACTORED FORM, STORING THE G(l) HOUSEHOLDER TRANSFORMATION
!               IN tau AND v(l+1_i4b:n).
!
                alpha = v(l+1_i4b)        
!
                call hous1( v(l+1_i4b:n), tau, beta )
!
                if ( tau/=zero ) then
!
!                   APPLY THE G(j) HOUSEHOLDER TRANSFORMATION TO mat(:m,l+1:n) .
!
                    scal = tau/( alpha - beta )
!
#ifdef _BLAS
                    call axpy( m, -beta, mat(1_i4b:m,l+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
!
                    call ger( m, n-l, scal, u_tmp(1_i4b:m), 1_i4b, v(l+1_i4b:n), 1_i4b, mat(1_i4b:m,l+1_i4b:n), m )
#else
                    u(1_i4b:m) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+1_i4b) )
!
                    do i = l+1_i4b, n
!
                        mat(1_i4b:m,i) = mat(1_i4b:m,i) + u(1_i4b:m)*v(i)
!
                    end do
#endif
!
                end if
!
!               ORTHOGONALIZE mat(1:m,l+1) AGAINST THE l COLUMN OF Q.
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
                mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) - beta*mat(1_i4b:m,l)
#endif
!
!               GENERATE THE l+1 COLUM OF Q AND STORE IT IN mat(1:m,l+1).
!
                scal  = norm( mat(1_i4b:m,l+1_i4b) )
                normc = pythag( scal, beta )
!
                if ( scal<=fudge_ortho2*normc ) then
!
                    if ( do_reortho ) then
!
!                       REORTHOGONALIZATION IS REQUIRED.
!
                        call ortho_vec( mat(1_i4b:m,1_i4b:l), mat(1_i4b:m,l+1_i4b),   &
                                        scal, normvec_orig=normc )
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                    else
!
                        failure2 = failure2 .or. scal<=eps_bd_cmp2*normc
!
                        scal = max( scal, safmin )
!
                        mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
                    end if
!
                else
!
!                   NORMALIZATION IS ENOUGH.
!
                    mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
                end if
!
!               COMPUTE THE l+1 ROW OF THE BIDIAGONAL MATRIX.
!
                d(l+1_i4b) = scal
                e(l+1_i4b) = beta
!
            end do
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
!   SET failure INDICATOR IF REQUIRED.
!
    if ( present(failure) ) then
!
        failure = failure2
!
    end if
!
!
! END OF SUBROUTINE bd_cmp2_b
! ___________________________
!
    end subroutine bd_cmp2_b
!
! =========================================================================================
!
    subroutine bd_cmp3( mat, d, e, gen_p, failure )
!
! Purpose
! _______
!
!   BD_CMP3 reduces a m-by-n matrix MAT with m >= n to upper bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal.
!
!   BD_CMP3 computes BD and P using the one-sided Ralha-Barlow
!   bidiagonal reduction algorithm.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the general m-by-n matrix to be reduced.
!
!            On exit, P is stored in MAT(1:n,1:n). See Further Details.
!
!            The shape of MAT must verify: size( MAT, 1 ) >= size( MAT, 2 ) = n .
!
!   D        (OUTPUT) real(stnd), dimension(:)
!            The diagonal elements of the bidiagonal matrix BD
!
!            The size of D must be size( MAT, 2) = n .
!
!   E        (OUTPUT) real(stnd), dimension(:)
!            The off-diagonal elements of the bidiagonal matrix BD:
!
!                E(i) = BD(i-1,i) for i = 2,3,...,n;
!
!            The size of E must be size( MAT, 2 ) = n .
!
!   GEN_P    (INPUT) logical(lgl)
!            If:
!
!            - GEN_P = true : the orthogonal matrix P is generated in MAT(1:n,1:n)
!              on output of the subroutine.
!            - GEN_P = false : the orthogonal matrix is stored in factored form as products of
!              elementary reflectors in the lower triangle of the array MAT(1:n,1:n).
!
!            See further details.
!
!   FAILURE  (OUTPUT, OPTIONAL) logical(lgl)
!            On exit:
!
!            - FAILURE = false :  indicates successful exit ;
!            - FAILURE = true  :  indicates that MAT is nearly singular.
!
!
! Further Details
! _______________
!
!   This subroutine is an implementation of the Ralha-Barlow one-sided method
!   to reduce a rectangular matrix MAT to bidiagonal form BD. Q is computed by
!   a recurrence relationship (but is not stored) and P as a product of n-1
!   elementary reflectors (e.g., Householder transformations):
!
!       P = G(1) * G(2) * . . . * G(n-1)
!
!   Each G(i) has the form:
!
!       G(i) = I + taup * v * v'
!
!   where taup is a real scalar, and v is a real vector. IF GEN_P is set to false,
!   the n-1 G(i) elementary reflectors are stored in the lower triangle of the array MAT.
!   For the G(i) reflector, taup is stored in MAT(i+1,1) and v is stored in MAT(i+1:n,i+1).
!   IF GEN_P is set to true, P is generated in MAT(:n,:n).
!
!   In addition, MAT(1,1) is set to -1 if GEN_P=false and is equal to 1 if GEN_P=true.
!   In other words, the value of MAT(1,1) indicates if the orthogonal matrix P is stored
!   in factored form or not in MAT. Note that if n is equal to 1, no elementary reflectors are needed
!   and consequently MAT(1,1) (e.g., P(1,1)) is set to 1, independently of the value of GEN_P.
!
!   This is the blocked version of the algorithm. See the references (1), (2) and (3)
!   for further details. Note also that the blocked algorithm implemented here is
!   more efficient than the version described in the reference (3). Furthermore the
!   algorithm is parallelized if OPENMP is used.
!
!   The reference (2) also explains how to handle the case of an exactly singular
!   matrix MAT (a very rare event). However, this case is not implemented here as
!   this subroutine outputs only BD and P.
!
!   For further details, see:
!
!   (1) Ralha, R.M.S., 2003:
!           One-sided reduction to bidiagonal form.
!           Linear Algebra Appl., No 358, pp. 219-238.
!
!   (2) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!           A new stable bidiagonal reduction algorithm.
!           Linear Algebra Appl., No 397, pp. 35-84.
!
!   (3) Bosner, N., and Barlow, J.L., 2007:
!           Block and Parallel versions of one-sided bidiagonalization.
!           SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error5
    use Utilities,       only : assert_eq, merror, pythag
    use Random,          only : random_number_
    use Hous_Procedures, only : hous1
    use QR_Procedures,   only : ortho_gen_qr
#ifdef _BLAS
    use BLAS_interfaces, only : dot, axpy, gemv, ger, gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: d, e
!
    logical(lgl), intent(in)            :: gen_p
    logical(lgl), intent(out), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                :: tau, beta, alpha, scal, normc
    real(stnd), dimension(size(mat,1))        :: u, u_tmp
    real(stnd), dimension(size(mat,2))        :: v, taup
    real(stnd), dimension(blksz2)             :: prodvec
    real(stnd), dimension(size(mat,1),blksz2) :: umat
    real(stnd), dimension(blksz2,size(mat,2)) :: vmat
!
    integer(i4b) :: m, n, i, i1, i2, i3, l, j, k, blksz
!
    logical(lgl) :: failure2
!
#ifdef _OPENMP
    real(stnd)                         :: beta2
    real(stnd), dimension(size(mat,1)) :: tmp
!
    integer(i4b) :: max_num_threads, num_threads, thread_num, iinc, irem, ioffset
!
    logical(lgl) :: apply_hous
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n =  assert_eq( int(size(mat,2),i4b)  ,   &
                    int(size(d),i4b)      ,   &
                    int(size(e),i4b)      ,   &
                    name_proc )
!
!   CHECK DIMENSIONS OF THE INPUT MATRIX.
!
    if ( m<n ) then
!
        call merror( name_proc//svd_error5 )
!
    end if
!
    failure2 = false
!
    if ( present(failure) ) then
!
        failure = failure2
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
!   ADJUST BLOCK SIZE IF NUMBER OF COLUMNS IS TOO SMALL.
!
    blksz = min( blksz2, n )
!
    e(1_i4b) = zero
!
!   REDUCE mat TO UPPER BIDIAGONAL FORM.
!
!   GENERATE THE FIRST COLUM OF Q AND STORE IT IN mat(1:m,1)
!
    d(1_i4b) = norm( mat(1_i4b:m,1_i4b) )
!
    if ( d(1_i4b)/=zero ) then
!
!       NORMALIZATION IS ENOUGH.
!
        mat(1_i4b:m,1_i4b) = mat(1_i4b:m,1_i4b)/d(1_i4b)
!
    else
!
!       IF THE NORM OF THE FIRST COLUMN OF mat IS ZERO,
!       SET IT TO A NONZERO NORMALIZED RANDOM VECTOR
!       AND PROCEED.
!
        call random_number_( u(:m) ) 
!
        u(:m) = u(:m) - half
        scal  = norm( u(:m) )
        mat(1_i4b:m,1_i4b) = u(:m)/scal
!
        failure2 = true
!
    end if
!
!   SPECIAL CASE IF mat(:,:) IS A COLUMN VECTOR.
!
    if ( n==1_i4b ) then
!
        mat(1_i4b,1_i4b) = one
!
        if ( present(failure) ) then
!
            failure = failure2
!
        end if
!
        return
!
    end if
!
    u_tmp(1_i4b:m) = zero
!
#ifdef _OPENMP
!
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m*n>=omp_limit                      .and.       &
                  n>blksz                             .and.       &
                  i>1_i4b
!
    if ( test_par ) then
!
        max_num_threads = min( i, max_num_threads_bd_cmp2 )
!
!$OMP PARALLEL PRIVATE(num_threads,thread_num,iinc,irem,ioffset,           &
!$OMP                  i,i1,i2,i3,l,j,k,alpha,beta,tau,scal,normc,u,tmp),  &
!$OMP          NUM_THREADS(max_num_threads)
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
!       PROCESS THE FIRST BLOCK OF COLUMNS OF mat .
!
!       START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(1)
!       HOUSEHOLDER TRANSFORMATION.
!
        u(1_i4b:m)   = mat(1_i4b:m,1_i4b)
        tmp(1_i4b:m) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
        do i = 2_i4b, n
!
#ifdef _BLAS
            v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b )
!
            call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
            v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )
!
            tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
        end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compu0)
!
#ifdef _BLAS
        call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
        u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu0)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!       COMPUTE P IN FACTORED FORM, STORING THE G(1) HOUSEHOLDER TRANSFORMATION
!       IN taup(1) AND mat(2:n,1).
!
        alpha = v(2_i4b)        
!
        call hous1( v(2_i4b:n), tau, beta )         
!
        taup(1_i4b) = tau
!
!       INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
        if ( tau/=zero ) then
!
            scal = tau/( alpha - beta )
!
#ifdef _BLAS
            call axpy( m, -beta, mat(1_i4b:m,2_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
            umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
            umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,2_i4b) )
#endif
!
        else
!
            umat(1_i4b:m,1_i4b) = zero
!
        end if
!
        umat(1_i4b:m,blksz) = zero
!
        vmat(1_i4b:blksz,1_i4b:n) = zero
        vmat(1_i4b,2_i4b:n)       = v(2_i4b:n)
!
!       UPDATE mat(1:m,2) WITH G(1) AND ORTHOGONALIZE mat(1:m,2) AGAINST THE FIRST COLUMN OF Q.
!
#ifdef _BLAS
        call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,2_i4b), 1_i4b )
        call axpy( m, -beta, mat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,2_i4b), 1_i4b )
#else
        mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b) + umat(1_i4b:m,1_i4b) - beta*mat(1_i4b:m,1_i4b)
#endif
!
!       GENERATE THE SECOND COLUM OF Q AND STORE IT IN mat(1:m,2)
!
        scal  = norm( mat(1_i4b:m,2_i4b) )
        normc = pythag( scal, beta )
!
        if ( scal<=fudge_ortho2*normc ) then
!
!           REORTHOGONALIZATION IS REQUIRED.
!
            call ortho_vec( mat(1_i4b:m,1_i4b:1_i4b), mat(1_i4b:m,2_i4b),   &
                            scal, normvec_orig=normc )
!
            failure2 = failure2 .or. scal<=eps_bd_cmp3*normc
!
        else
!
!           NORMALIZATION IS ENOUGH.
!
            mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b)/scal
!
        end if
!
!       COMPUTE SECOND ROW OF THE BIDIAGONAL MATRIX.
!
        d(2_i4b) = scal
        e(2_i4b) = beta
!
        mat(2_i4b:n,1_i4b) = v(2_i4b:n)
!
        u_tmp(1_i4b:m) = zero  
!
!$OMP END SINGLE
!
        do j = 2_i4b, blksz-1_i4b
!
!           START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(j)
!           HOUSEHOLDER TRANSFORMATION.
!
            u(1_i4b:m) = mat(1_i4b:m,j)
!
!$OMP DO SCHEDULE(STATIC)
!
            do i = 1_i4b, j-1_i4b
!
#ifdef _BLAS
                prodvec(i) = dot( m, u(1_i4b:m), 1_i4b, umat(1_i4b:m,i), 1_i4b )
#else
                prodvec(i) = dot_product( u(1_i4b:m), umat(1_i4b:m,i) )
#endif
!
            end do
!
!$OMP END DO
!
            tmp(1_i4b:m) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
            do i = j+1_i4b, n
!
#ifdef _BLAS
                v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b ) +                      &
                       dot( j-1_i4b, prodvec(1_i4b:j-1_i4b), 1_i4b, vmat(1_i4b:j-1_i4b,i), 1_i4b )
!
                call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) ) +                &
                       dot_product( prodvec(1_i4b:j-1_i4b), vmat(1_i4b:j-1_i4b,i) )
!
                tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
            end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compu1)
!
#ifdef _BLAS
            call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu1)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!           COMPUTE P IN FACTORED FORM, STORING THE G(j) HOUSEHOLDER TRANSFORMATION
!           IN taup(j) AND mat(j+1:n,j).
!
            alpha = v(j+1_i4b)        
!
            call hous1( v(j+1_i4b:n), tau, beta )
!
            taup(j) = tau
!
!           SAVE G(j) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
            vmat(j,j+1_i4b:n) = v(j+1_i4b:n)
!
#ifdef _BLAS
            call gemv( 'N', j-1_i4b, n-j, one, vmat(1_i4b:blksz2,j+1_i4b:n), blksz2, &
                       v(j+1_i4b:n), 1_i4b, zero, prodvec, 1_i4b )
#else
            prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,j+1_i4b:n), v(j+1_i4b:n) )
#endif
!
            if ( tau/=zero ) then
!
                scal = tau/( alpha - beta )
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,j+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                umat(1_i4b:m,j) = scal*u_tmp(1_i4b:m)
#else
                umat(1_i4b:m,j) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,j+1_i4b) )
#endif
!
            else
!
                umat(1_i4b:m,j) = zero
!
            end if
!
#ifdef _BLAS
            call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                        1_i4b, one, umat(1_i4b:m,j), 1_i4b )
#else
            umat(1_i4b:m,j) = umat(1_i4b:m,j) + tau*matmul( umat(1_i4b:m,1_i4b:j-1_i4b), prodvec(1_i4b:j-1_i4b) )
#endif
!
!           UPDATE mat(1:m,j+1) WITH G(1)* ... *G(j) AND ORTHOGONALIZE mat(1:m,j+1) AGAINST THE j COLUMN OF Q.
!
#ifdef _BLAS
            call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,j+1_i4b), 1_i4b,     &
                       one, mat(1_i4b:m,j+1_i4b), 1_i4b )
            call axpy( m, -beta, mat(1_i4b:m,j), 1_i4b, mat(1_i4b:m,j+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,j+1_i4b) ) &
                                   - beta*mat(1_i4b:m,j)
#endif
!
!           GENERATE THE j+1 COLUM OF Q AND STORE IT IN mat(1:m,j+1).
!
            scal  = norm( mat(1_i4b:m,j+1_i4b) )
            normc = pythag( scal, beta )
!
            if ( scal<=eps_bd_cmp3*normc ) then
!
                failure2 = true
!
                scal = max( scal, safmin )
!
            end if
!
!           NORMALIZATION.
!
            mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b)/scal
!
!           COMPUTE THE j+1 ROW OF THE BIDIAGONAL MATRIX.
!
            d(j+1_i4b) = scal
            e(j+1_i4b) = beta
!
            mat(j+1_i4b:n,j) = v(j+1_i4b:n)
!
            u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
        end do
!
!       PROCESS THE SECOND AND FOLLOWING BLOCKS OF COLUMNS OF mat .
!
        do l = blksz, n-blksz, blksz
!
!           PERFORM A BLOCK UPDATE OF THE TRAILING MATRIX AND START COMPUTATIONS FOR
!           GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION.
!
            tmp(1_i4b:m) = zero
!
!           DISTRIBUTE THE WORK AMONG THE THREADS FOR THE BLOCK UPDATE OF THE TRAILING MATRIX.
!
            i  = n - l
            irem = mod( i, num_threads )
            iinc = ( i - irem )/num_threads
!
            if ( thread_num<irem ) then
                iinc    = iinc + 1_i4b
                ioffset = l + 1_i4b
            else
                ioffset = irem + l + 1_i4b
            end if
!
            i1 = thread_num*iinc + ioffset
            i2 = i1 + iinc - 1_i4b
            i3 = i2 - i1 + 1_i4b
!
            if ( i3>0_i4b ) then
!
#ifdef _BLAS
!
!               COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) AS A BLOCK UPDATE.
!
                call gemm( 'N', 'N', m, i3, blksz, one, umat(1_i4b:m,1_i4b:blksz), m,   &
                           vmat(1_i4b:blksz2,i1:i2), blksz2, one, mat(1_i4b:m,i1:i2), m )
!
!               START GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION
!               WITH A BLAS2.5 GEMVT OPERATOR.
!
!                call gemv( 'T', m, i3, one,  mat(1_i4b:m,i1:i2), m, mat(1_i4b:m,l), &
!                           1_i4b, zero, v(i1:i2), 1_i4b )
!
!                call gemv( 'N', m, i3, one,  mat(1_i4b:m,i1:i2), m, v(i1:i2), &
!                           1_i4b, one, tmp(1_i4b:m), 1_i4b )
!
                do j = i1, i2
!
                    v(j) = dot( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,j), 1_i4b )
!
                    call axpy( m, v(j), mat(1_i4b:m,j), 1_i4b, tmp(1_i4b:m), 1_i4b )
!
                end do
#else
!
!               COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) AS A BLOCK UPDATE.
!
                mat(1_i4b:m,i1:i2) = mat(1_i4b:m,i1:i2) +                                       &
                                     matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i1:i2) )
!
!               START GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION
!               WITH A BLAS2.5 GEMVT OPERATOR.
!
!                v(i1:i2) = matmul( mat(1_i4b:m,l), mat(1_i4b:m,i1:i2) )
!
!                tmp(1_i4b:m) = tmp(1_i4b:m) +  matmul( mat(1_i4b:m,i1:i2), v(i1:i2) )
!
                do j = i1, i2
!
                    v(j) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,j) )
!
                    tmp(1_i4b:m) = tmp(1_i4b:m) +  v(j)*mat(1_i4b:m,j)
!
                end do
#endif
!
            end if
!
!$OMP CRITICAL (compu2)
!
#ifdef _BLAS
            call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu2)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!           COMPUTE P IN FACTORED FORM, STORING THE G(l) HOUSEHOLDER TRANSFORMATION
!           IN taup(l) AND mat(l+1:n,l).
!
            alpha = v(l+1_i4b)        
!
            call hous1( v(l+1_i4b:n), tau, beta )
!
            taup(l) = tau
!
!           INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
            if ( tau/=zero ) then
!
                scal = tau/( alpha - beta )
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,l+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
                umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+1_i4b) )
#endif
!
            else
!
                umat(1_i4b:m,1_i4b) = zero
!
            end if
!
            vmat(:blksz,:n)      = zero
            vmat(1_i4b,l+1_i4b:n) = v(l+1_i4b:n)
!
!           UPDATE mat(1:m,l+1) WITH G(l) AND ORTHOGONALIZE mat(1:m,l+1) AGAINST THE l COLUMN OF Q.
!
#ifdef _BLAS
            call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
            call axpy( m, -beta, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) + umat(1_i4b:m,1_i4b) - beta*mat(1_i4b:m,l)
#endif
!
!           GENERATE THE l+1 COLUM OF Q AND STORE IT IN mat(1:m,l+1).
!
            scal  = norm( mat(1_i4b:m,l+1_i4b) )
            normc = pythag( scal, beta )
!
            if ( scal<=eps_bd_cmp3*normc ) then
!
                failure2 = true
!
                scal = max( scal, safmin )
!
            end if
!
!           NORMALIZATION.
!
            mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
!           COMPUTE THE l+1 ROW OF THE BIDIAGONAL MATRIX.
!
            d(l+1_i4b) = scal
            e(l+1_i4b) = beta
!
            mat(l+1_i4b:n,l) = v(l+1_i4b:n)
!
            u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
            do j = 2_i4b, blksz
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(l+j-1)
!               HOUSEHOLDER TRANSFORMATION.
!
                u(1_i4b:m) = mat(1_i4b:m,l+j-1_i4b)
!
!$OMP DO SCHEDULE(STATIC)
!
                do i = 1_i4b, j-1_i4b
!
#ifdef _BLAS
                    prodvec(i) = dot( m, u(1_i4b:m), 1_i4b, umat(1_i4b:m,i), 1_i4b )
#else
                    prodvec(i) = dot_product( u(1_i4b:m), umat(1_i4b:m,i) )
#endif
!
                end do
!
!$OMP END DO
!
                tmp(1_i4b:m) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
                do i = l+j, n
!
#ifdef _BLAS
                    v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b ) +                      &
                           dot( j-1_i4b, prodvec(1_i4b:j-1_i4b), 1_i4b, vmat(1_i4b:j-1_i4b,i), 1_i4b )
!
                    call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                    v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )                  + &
                           dot_product( prodvec(1_i4b:j-1_i4b), vmat(1_i4b:j-1_i4b,i) )
!
                    tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compu3)
!
#ifdef _BLAS
                call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu3)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!               COMPUTE P IN FACTORED FORM, STORING THE G(l+j-1) HOUSEHOLDER TRANSFORMATION
!               IN taup(l+j-1) AND mat(l+j:n,l+j-1).
!
                alpha = v(l+j)        
!
                call hous1( v(l+j:n), tau, beta )
!
                taup(l+j-1) = tau
!
!               SAVE G(l+j-1) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
                vmat(j,l+j:n) = v(l+j:n)
!
#ifdef _BLAS
                call gemv( 'N', j-1_i4b, n-j-l+1_i4b, one, vmat(1_i4b:blksz2,l+j:n), blksz2, &
                           v(l+j:n), 1_i4b, zero, prodvec, 1_i4b )
#else
                prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,l+j:n), v(l+j:n) )
#endif
!
                if ( tau/=zero ) then
!
                    scal = tau/( alpha - beta )
#ifdef _BLAS
                    call axpy( m, -beta, mat(1_i4b:m,l+j), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                    umat(1_i4b:m,j) = scal*u_tmp(1_i4b:m)
#else
                    umat(1_i4b:m,j) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+j) )
#endif
!
                else
!
                    umat(1_i4b:m,j) = zero
!
                end if
!
#ifdef _BLAS
                call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                           1_i4b, one, umat(1_i4b:m,j), 1_i4b )
#else
                umat(1_i4b:m,j) = umat(1_i4b:m,j) + tau*matmul( umat(1_i4b:m,:j-1_i4b), prodvec(1_i4b:j-1_i4b) )
#endif
!
!               UPDATE mat(1:m,l+j) WITH G(l)* ... *G(l+j-1) AND ORTHOGONALIZE mat(1:m,l+j)
!               AGAINST THE l+j-1 COLUMN OF Q.
!
#ifdef _BLAS
                call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,l+j), 1_i4b, &
                           one, mat(1_i4b:m,l+j), 1_i4b )
                call axpy( m, -beta, mat(1_i4b:m,l+j-1_i4b), 1_i4b, mat(1_i4b:m,l+j), 1_i4b )
#else
                mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,l+j) ) &
                                   - beta*mat(1_i4b:m,l+j-1_i4b)
#endif
!
!               GENERATE THE l+j COLUM OF Q AND STORE IT IN mat(1:m,l+j).
!
                scal  = norm( mat(1_i4b:m,l+j) )
                normc = pythag( scal, beta )
!
                if ( scal<=eps_bd_cmp3*normc ) then
!
                    failure2 = true
!
                    scal = max( scal, safmin )
!
                end if
!
!               NORMALIZATION.
!
                mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j)/scal
!
!               COMPUTE THE l+j ROW OF THE BIDIAGONAL MATRIX.
!
                d(l+j) = scal
                e(l+j) = beta
!
                mat(l+j:n,l+j-1) = v(l+j:n)
!
                u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
            end do
!
        end do
!
        k = blksz*(n/blksz)
!
        if ( n/=k ) then
!
!           PROCESS THE LAST BLOCK OF COLUMNS OF mat .
!
            do l = k, n-1_i4b
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(l)
!               HOUSEHOLDER TRANSFORMATION.
!
                tmp(1_i4b:m) = zero
!
                if ( l==k ) then
!
!$OMP DO SCHEDULE(STATIC)
!
                    do i = l+1_i4b, n
!
!                       COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) WITH A LAST BLOCK UPDATE.
!
                        u(1_i4b:m) = mat(1_i4b:m,i)
!
#ifdef _BLAS
                        call gemv( 'N', m, blksz, one, umat(1_i4b:m,1_i4b:blksz), m, vmat(1_i4b:blksz,i), &
                                   1_i4b, one, u(1_i4b:m), 1_i4b )
!
                        v(i) = dot( m, mat(1_i4b:m,l), 1_i4b, u(1_i4b:m), 1_i4b )
!
                        call axpy( m, v(i), u(1_i4b:m), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                        u(1_i4b:m) = u(1_i4b:m) + matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i) )
!
                        v(i) = dot_product( mat(1_i4b:m,l), u(1_i4b:m) )
!
                        tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*u(1_i4b:m)
#endif
!
                        mat(1_i4b:m,i) = u(1_i4b:m)
!
                    end do
!
!$OMP END DO NOWAIT
!
                else
!
!$OMP DO SCHEDULE(STATIC)
!
                    do i = l+1_i4b, n
!
#ifdef _BLAS
                        v(i) = dot( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,i), 1_i4b )
!
                        call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                        v(i) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,i) )
!
                        tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                    end do
!
!$OMP END DO NOWAIT
!
                end if
!
!$OMP CRITICAL (compu4)
!
#ifdef _BLAS
                call axpy( m, one, tmp(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
#endif
!
!$OMP END CRITICAL (compu4)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!               COMPUTE P IN FACTORED FORM, STORING THE G(l) HOUSEHOLDER TRANSFORMATION
!               IN taup(l) AND mat(l+1:n,l).
!
                alpha = v(l+1_i4b)        
!
                call hous1( v(l+1_i4b:n), tau, beta2 )
!
                taup(l) = tau
!
                apply_hous = tau/=zero
!
                if ( apply_hous ) then
!
                    scal = tau/( alpha - beta2 )
!
#ifdef _BLAS
                    call axpy( m, -beta2, mat(1_i4b:m,l+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                    umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
                    umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta2*mat(1_i4b:m,l+1_i4b) )
#endif
!
!                   APPLY THE G(l) HOUSEHOLDER TRANSFORMATION TO mat(:m,l+1) .
!
#ifdef _BLAS
                    call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
                    mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) + umat(1_i4b:m,1_i4b)
#endif
!
                end if
!
!$OMP END SINGLE
!
                if ( apply_hous ) then
!
!                   APPLY THE G(l) HOUSEHOLDER TRANSFORMATION TO mat(:m,l+2:n) .
!
!$OMP DO SCHEDULE(STATIC)
!
                    do i = l+2_i4b, n
!
#ifdef _BLAS
                        call axpy( m, v(i), umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,i), 1_i4b )
#else
                        mat(1_i4b:m,i) = mat(1_i4b:m,i) + umat(1_i4b:m,1_i4b)*v(i)
#endif
!
                    end do
!
!$OMP END DO NOWAIT
!
                end if
!
!$OMP SINGLE
!
!               ORTHOGONALIZE mat(1:m,l+1) AGAINST THE l COLUMN OF Q.
!
#ifdef _BLAS
                call axpy( m, -beta2, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
                mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) - beta2*mat(1_i4b:m,l)
#endif
!
!               GENERATE THE l+1 COLUM OF Q AND STORE IT IN mat(1:m,l+1).
!
                scal  = norm( mat(1_i4b:m,l+1_i4b) )
                normc = pythag( scal, beta2 )
!
                if ( scal<=eps_bd_cmp3*normc ) then
!
                    failure2 = true
!
                    scal = max( scal, safmin )
!
                end if
!
!               NORMALIZATION.
!
                mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
!               COMPUTE THE l+1 ROW OF THE BIDIAGONAL MATRIX.
!
                d(l+1_i4b) = scal
                e(l+1_i4b) = beta2
!
                mat(l+1:n,l) = v(l+1_i4b:n)
!
                u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
            end do
!
        end if
!
!$OMP END PARALLEL
!
    else
#endif
!
!       PROCESS THE FIRST BLOCK OF COLUMNS OF mat .
!
!       START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(1)
!       HOUSEHOLDER TRANSFORMATION.
!
        u(1_i4b:m) = mat(1_i4b:m,1_i4b)
!
        do i = 2_i4b, n
!
#ifdef _BLAS
            v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b )
!
            call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
            v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )
!
            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
        end do
!
!       COMPUTE P IN FACTORED FORM, STORING THE G(1) HOUSEHOLDER TRANSFORMATION
!       IN taup(1) AND mat(2:n,1).
!
        alpha = v(2_i4b)        
!
        call hous1( v(2_i4b:n), tau, beta )         
!
        taup(1_i4b) = tau
!
!       INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
        if ( tau/=zero ) then
!
            scal = tau/( alpha - beta )
!
#ifdef _BLAS
            call axpy( m, -beta, mat(1_i4b:m,2_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
            umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
            umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,2_i4b) )
#endif
!
        else
!
            umat(1_i4b:m,1_i4b) = zero
!
        end if
!
        umat(1_i4b:m,blksz) = zero
!
        vmat(1_i4b:blksz,1_i4b:n) = zero
        vmat(1_i4b,2_i4b:n)       = v(2_i4b:n)
!
!       UPDATE mat(1:m,2) WITH G(1) AND ORTHOGONALIZE mat(1:m,2) AGAINST THE FIRST COLUMN OF Q.
!
#ifdef _BLAS
        call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,2_i4b), 1_i4b )
        call axpy( m, -beta, mat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,2_i4b), 1_i4b )
#else
        mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b) + umat(1_i4b:m,1_i4b) - beta*mat(1_i4b:m,1_i4b)
#endif
!
!       GENERATE THE SECOND COLUM OF Q AND STORE IT IN mat(1:m,2)
!
        scal  = norm( mat(1_i4b:m,2_i4b) )
        normc = pythag( scal, beta )
!
        if ( scal<=fudge_ortho2*normc ) then
!
!           REORTHOGONALIZATION IS REQUIRED.
!
            call ortho_vec( mat(1_i4b:m,1_i4b:1_i4b), mat(1_i4b:m,2_i4b),   &
                            scal, normvec_orig=normc )
!
            failure2 = failure2 .or. scal<=eps_bd_cmp3*normc
!
        else
!
!           NORMALIZATION IS ENOUGH.
!
            mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b)/scal
!
        end if
!
!       COMPUTE SECOND ROW OF THE BIDIAGONAL MATRIX.
!
        d(2_i4b) = scal
        e(2_i4b) = beta
!
        mat(2_i4b:n,1_i4b) = v(2_i4b:n)
!
        do j = 2_i4b, blksz-1_i4b
!
!           START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(j)
!           HOUSEHOLDER TRANSFORMATION.
!
            u(1_i4b:m) = mat(1_i4b:m,j)
!
#ifdef _BLAS
            call gemv( 'T', m, j-1_i4b, one,  umat(1_i4b:m,1_i4b:j-1_i4b), m, &
                       u(1_i4b:m), 1_i4b, zero, prodvec(1_i4b:j-1_i4b), 1_i4b )
#else
            prodvec(1_i4b:j-1_i4b) = matmul( u(1_i4b:m), umat(1_i4b:m,1_i4b:j-1_i4b) )
#endif
!
            u_tmp(1_i4b:m) = zero
!
            do i = j+1_i4b, n
!
#ifdef _BLAS
                v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b ) +                      &
                       dot( j-1_i4b, prodvec(1_i4b:j-1_i4b), 1_i4b, vmat(1_i4b:j-1_i4b,i), 1_i4b )
!
                call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) ) +                &
                       dot_product( prodvec(1_i4b:j-1_i4b), vmat(1_i4b:j-1_i4b,i) )
!
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
            end do
!
!           COMPUTE P IN FACTORED FORM, STORING THE G(j) HOUSEHOLDER TRANSFORMATION
!           IN taup(j) AND mat(j+1:n,j).
!
            alpha = v(j+1_i4b)        
!
            call hous1( v(j+1_i4b:n), tau, beta )
!
            taup(j) = tau
!
!           SAVE G(j) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
            vmat(j,j+1_i4b:n) = v(j+1_i4b:n)
!
#ifdef _BLAS
            call gemv( 'N', j-1_i4b, n-j, one, vmat(1_i4b:blksz2,j+1_i4b:n), blksz2, &
                       v(j+1_i4b:n), 1_i4b, zero, prodvec, 1_i4b )
#else
            prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,j+1_i4b:n), v(j+1_i4b:n) )
#endif
!
            if ( tau/=zero ) then
!
                scal = tau/( alpha - beta )
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,j+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                umat(1_i4b:m,j) = scal*u_tmp(1_i4b:m)
#else
                umat(1_i4b:m,j) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,j+1_i4b) )
#endif
!
            else
!
                umat(1_i4b:m,j) = zero
!
            end if
!
#ifdef _BLAS
            call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                        1_i4b, one, umat(1_i4b:m,j), 1_i4b )
#else
            umat(1_i4b:m,j) = umat(1_i4b:m,j) + tau*matmul( umat(1_i4b:m,1_i4b:j-1_i4b), prodvec(1_i4b:j-1_i4b) )
#endif
!
!           UPDATE mat(1:m,j+1) WITH G(1)* ... *G(j) AND ORTHOGONALIZE mat(1:m,j+1) AGAINST THE j COLUMN OF Q.
!
#ifdef _BLAS
            call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,j+1_i4b), 1_i4b,     &
                       one, mat(1_i4b:m,j+1_i4b), 1_i4b )
            call axpy( m, -beta, mat(1_i4b:m,j), 1_i4b, mat(1_i4b:m,j+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,j+1_i4b) ) &
                                   - beta*mat(1_i4b:m,j)
#endif
!
!           GENERATE THE j+1 COLUM OF Q AND STORE IT IN mat(1:m,j+1).
!
            scal  = norm( mat(1_i4b:m,j+1_i4b) )
            normc = pythag( scal, beta )
!
            if ( scal<=eps_bd_cmp3*normc ) then
!
                failure2 = true
!
                scal = max( scal, safmin )
!
            end if
!
!           NORMALIZATION.
!
            mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b)/scal
!
!           COMPUTE THE j+1 ROW OF THE BIDIAGONAL MATRIX.
!
            d(j+1_i4b) = scal
            e(j+1_i4b) = beta
!
            mat(j+1_i4b:n,j) = v(j+1_i4b:n)
!
        end do
!
!       PROCESS THE SECOND AND FOLLOWING BLOCKS OF COLUMNS OF mat .
!
        do l = blksz, n-blksz, blksz
!
!           PERFORM A BLOCK UPDATE OF THE TRAILING MATRIX AND START COMPUTATIONS FOR
!           GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION.
!
            u_tmp(1_i4b:m) = zero
!
            i1 = l + 1_i4b
            i2 = n
            i3 = i2 - l
!
#ifdef _BLAS
!
!           COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) AS A BLOCK UPDATE.
!
            call gemm( 'N', 'N', m, i3, blksz, one, umat(1_i4b:m,1_i4b:blksz), m,    &
                       vmat(1_i4b:blksz2,i1:i2), blksz2, one, mat(1_i4b:m,i1:i2), m  )
!
!           START GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION
!           WITH A BLAS2.5 GEMVT OPERATOR.
!
!            call gemv( 'T', m, i3, one,  mat(1_i4b:m,i1:i2), m, mat(1_i4b:m,l), &
!                       1_i4b, zero, v(i1:i2), 1_i4b )
!
!            call gemv( 'N', m, i3, one,  mat(1_i4b:m,i1:i2), m, v(i1:i2), &
!                       1_i4b, one, u_tmp(1_i4b:m), 1_i4b )
!
            do j = i1, i2
!
                v(j) = dot( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,j), 1_i4b )
!
                call axpy( m, v(j), mat(1_i4b:m,j), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
!
            end do
#else
!
!           COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) AS A BLOCK UPDATE.
!
            mat(1_i4b:m,i1:i2) = mat(1_i4b:m,i1:i2) +                                       &
                                 matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i1:i2) )
!
!           START GENERATION AND APPLICATION OF THE G(l) HOUSEHOLDER TRANSFORMATION
!           WITH A BLAS2.5 GEMVT OPERATOR.
!
!            v(i1:i2) = matmul( mat(1_i4b:m,l), mat(1_i4b:m,i1:i2) )
!
!            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) +  matmul( mat(1_i4b:m,i1:i2), v(i1:i2) )
!
            do j = i1, i2
!
                v(j) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,j) )
!
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) +  v(j)*mat(1_i4b:m,j)
!
            end do
#endif
!
!           COMPUTE P IN FACTORED FORM, STORING THE G(l) HOUSEHOLDER TRANSFORMATION
!           IN taup(l) AND mat(l+1:n,l).
!
            alpha = v(l+1_i4b)        
!
            call hous1( v(l+1_i4b:n), tau, beta )
!
            taup(l) = tau
!
!           INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
            if ( tau/=zero ) then
!
                scal = tau/( alpha - beta )
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,l+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                umat(1_i4b:m,1_i4b) = scal*u_tmp(1_i4b:m)
#else
                umat(1_i4b:m,1_i4b) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+1_i4b) )
#endif
!
            else
!
                umat(1_i4b:m,1_i4b) = zero
!
            end if
!
            vmat(:blksz,:n)       = zero
            vmat(1_i4b,l+1_i4b:n) = v(l+1_i4b:n)
!
!           UPDATE mat(1:m,l+1) WITH G(l) AND ORTHOGONALIZE mat(1:m,l+1) AGAINST THE l COLUMN OF Q.
!
#ifdef _BLAS
            call axpy( m, one, umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
            call axpy( m, -beta, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) + umat(1_i4b:m,1_i4b) - beta*mat(1_i4b:m,l)
#endif
!
!           GENERATE THE l+1 COLUM OF Q AND STORE IT IN mat(1:m,l+1).
!
            scal  = norm( mat(1_i4b:m,l+1_i4b) )
            normc = pythag( scal, beta )
!
            if ( scal<=eps_bd_cmp3*normc ) then
!
                failure2 = true
!
                scal = max( scal, safmin )
!
            end if
!
!           NORMALIZATION.
!
            mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
!           COMPUTE THE l+1 ROW OF THE BIDIAGONAL MATRIX.
!
            d(l+1_i4b) = scal
            e(l+1_i4b) = beta
!
            mat(l+1_i4b:n,l) = v(l+1_i4b:n)
!
            do j = 2_i4b, blksz
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(l+j-1)
!               HOUSEHOLDER TRANSFORMATION.
!
                u(1_i4b:m) = mat(1_i4b:m,l+j-1_i4b)
!
#ifdef _BLAS
                call gemv( 'T', m, j-1_i4b, one,  umat(1_i4b:m,1_i4b:j-1_i4b), m, &
                           u(1_i4b:m), 1_i4b, zero, prodvec(1_i4b:j-1_i4b), 1_i4b )
#else
                prodvec(1_i4b:j-1_i4b) = matmul( u(1_i4b:m), umat(1_i4b:m,1_i4b:j-1_i4b) )
#endif
!
                u_tmp(1_i4b:m) = zero
!
                do i = l+j, n
!
#ifdef _BLAS
                    v(i) = dot( m, u(1_i4b:m), 1_i4b, mat(1_i4b:m,i), 1_i4b ) +                      &
                           dot( j-1_i4b, prodvec(1_i4b:j-1_i4b), 1_i4b, vmat(1_i4b:j-1_i4b,i), 1_i4b )
!
                    call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                    v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )                  + &
                           dot_product( prodvec(1_i4b:j-1_i4b), vmat(1_i4b:j-1_i4b,i) )
!
                    u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                end do
!
!               COMPUTE P IN FACTORED FORM, STORING THE G(l+j-1) HOUSEHOLDER TRANSFORMATION
!               IN taup(l+j-1) AND mat(l+j:n,l+j-1).
!
                alpha = v(l+j)        
!
                call hous1( v(l+j:n), tau, beta )
!
                taup(l+j-1) = tau
!
!               SAVE G(l+j-1) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
                vmat(j,l+j:n) = v(l+j:n)
!
#ifdef _BLAS
                call gemv( 'N', j-1_i4b, n-j-l+1_i4b, one, vmat(1_i4b:blksz2,l+j:n), blksz2, &
                           v(l+j:n), 1_i4b, zero, prodvec, 1_i4b )
#else
                prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,l+j:n), v(l+j:n) )
#endif
!
                if ( tau/=zero ) then
!
                    scal = tau/( alpha - beta )
#ifdef _BLAS
                    call axpy( m, -beta, mat(1_i4b:m,l+j), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
                    umat(1_i4b:m,j) = scal*u_tmp(1_i4b:m)
#else
                    umat(1_i4b:m,j) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+j) )
#endif
!
                else
!
                    umat(1_i4b:m,j) = zero
!
                end if
!
#ifdef _BLAS
                call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                           1_i4b, one, umat(1_i4b:m,j), 1_i4b )
#else
                umat(1_i4b:m,j) = umat(1_i4b:m,j) + tau*matmul( umat(1_i4b:m,:j-1_i4b), prodvec(1_i4b:j-1_i4b) )
#endif
!
!               UPDATE mat(1:m,l+j) WITH G(l)* ... *G(l+j-1) AND ORTHOGONALIZE mat(1:m,l+j)
!               AGAINST THE l+j-1 COLUMN OF Q.
!
#ifdef _BLAS
                call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,l+j), 1_i4b, &
                           one, mat(1_i4b:m,l+j), 1_i4b )
                call axpy( m, -beta, mat(1_i4b:m,l+j-1_i4b), 1_i4b, mat(1_i4b:m,l+j), 1_i4b )
#else
                mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,l+j) ) &
                                   - beta*mat(1_i4b:m,l+j-1_i4b)
#endif
!
!               GENERATE THE l+j COLUM OF Q AND STORE IT IN mat(1:m,l+j).
!
                scal  = norm( mat(1_i4b:m,l+j) )
                normc = pythag( scal, beta )
!
                if ( scal<=eps_bd_cmp3*normc ) then
!
                    failure2 = true
!
                    scal = max( scal, safmin )
!
                end if
!
!               NORMALIZATION.
!
                mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j)/scal
!
!               COMPUTE THE l+j ROW OF THE BIDIAGONAL MATRIX.
!
                d(l+j) = scal
                e(l+j) = beta
!
                mat(l+j:n,l+j-1) = v(l+j:n)
!
            end do
!
        end do
!
        k = blksz*(n/blksz)
!
        if ( n/=k ) then
!
!           PROCESS THE LAST BLOCK OF COLUMNS OF mat .
!
            do l = k, n-1_i4b
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE G(l)
!               HOUSEHOLDER TRANSFORMATION.
!
                u_tmp(1_i4b:m) = zero
!
                if ( l==k ) then
!
                    do i = l+1_i4b, n
!
!                       COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) WITH A LAST BLOCK UPDATE.
!
                        u(1_i4b:m) = mat(1_i4b:m,i)
!
#ifdef _BLAS
                        call gemv( 'N', m, blksz, one, umat(1_i4b:m,1_i4b:blksz), m, vmat(1_i4b:blksz,i), &
                                   1_i4b, one, u(1_i4b:m), 1_i4b )
!
                        v(i) = dot( m, mat(1_i4b:m,l), 1_i4b, u(1_i4b:m), 1_i4b )
!
                        call axpy( m, v(i), u(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                        u(1_i4b:m) = u(1_i4b:m) + matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i) )
!
                        v(i) = dot_product( mat(1_i4b:m,l), u(1_i4b:m) )
!
                        u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*u(1_i4b:m)
#endif
!
                        mat(1_i4b:m,i) = u(1_i4b:m)
!
                    end do
!
                else
!
                    do i = l+1_i4b, n
!
#ifdef _BLAS
                        v(i) = dot( m, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,i), 1_i4b )
!
                        call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                        v(i) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,i) )
!
                        u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                    end do
!
                end if
!
!               COMPUTE P IN FACTORED FORM, STORING THE G(l) HOUSEHOLDER TRANSFORMATION
!               IN taup(l) AND mat(l+1:n,l).
!
                alpha = v(l+1_i4b)        
!
                call hous1( v(l+1_i4b:n), tau, beta )
!
                taup(l) = tau
!
                if ( tau/=zero ) then
!
!                   APPLY THE G(j) HOUSEHOLDER TRANSFORMATION TO mat(:m,l+1:n) .
!
                    scal = tau/( alpha - beta )
!
#ifdef _BLAS
                    call axpy( m, -beta, mat(1_i4b:m,l+1_i4b), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
!
                    call ger( m, n-l, scal, u_tmp(1_i4b:m), 1_i4b, v(l+1_i4b:n), 1_i4b, mat(1_i4b:m,l+1_i4b:n), m )
#else
                    u(1_i4b:m) = scal*( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+1_i4b) )
!
                    do i = l+1_i4b, n
!
                        mat(1_i4b:m,i) = mat(1_i4b:m,i) + u(1_i4b:m)*v(i)
!
                    end do
#endif
!
                end if
!
!               ORTHOGONALIZE mat(1:m,l+1) AGAINST THE l COLUMN OF Q.
!
#ifdef _BLAS
                call axpy( m, -beta, mat(1_i4b:m,l), 1_i4b, mat(1_i4b:m,l+1_i4b), 1_i4b )
#else
                mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) - beta*mat(1_i4b:m,l)
#endif
!
!               GENERATE THE l+1 COLUM OF Q AND STORE IT IN mat(1:m,l+1).
!
                scal  = norm( mat(1_i4b:m,l+1_i4b) )
                normc = pythag( scal, beta )
!
                if ( scal<=eps_bd_cmp3*normc ) then
!
                    failure2 = true
!
                    scal = max( scal, safmin )
!
                end if
!
!               NORMALIZATION.
!
                mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b)/scal
!
!               COMPUTE THE l+1 ROW OF THE BIDIAGONAL MATRIX.
!
                d(l+1_i4b) = scal
                e(l+1_i4b) = beta
!
                mat(l+1:n,l) = v(l+1_i4b:n)
!
            end do
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
!   SHIFT ONE COLUM AHEAD THE ELEMENTARY REFLECTORS DEFINING P IN THE ARRAY mat.
!
    do i = n, 2_i4b, -1_i4b
!
        mat(i:n,i) = mat(i:n,i-1_i4b)
!
    end do
!
    if ( gen_p ) then
!
!       SET THE FIRST COLUMN AND ROW OF mat TO THAT OF THE UNIT MATRIX.
!
        mat(1_i4b,1_i4b)   = one
        mat(2_i4b:n,1_i4b) = zero
        mat(1_i4b,2_i4b:n) = zero
!
!       GENERATE RIGHT VECTORS BIDIAGONALIZING mat IN mat .
!
        call ortho_gen_qr( mat(2_i4b:n,2_i4b:n), taup(:n-1_i4b) )
!
    else
!
!       STORE THE ELEMENTARY REFLECTORS DEFINING P IN THE ARRAY mat.
!
        mat(1_i4b,1_i4b)   = -one
        mat(2_i4b:n,1_i4b) = taup(1_i4b:n-1_i4b)
        mat(1_i4b,2_i4b:n) = zero
!
    end if
!
!   SET failure INDICATOR IF REQUIRED.
!
    if ( present(failure) ) then
!
        failure = failure2
!
    end if
!
!
! END OF SUBROUTINE bd_cmp3
! _________________________
!
    end subroutine bd_cmp3
!
! =========================================================================================
!
    subroutine ortho_gen_bd( mat, tauq, taup, p )
!
! Purpose
! _______
!
!   ORTHO_GEN_BD generates the real orthogonal matrices Q and P determined by BD_CMP
!   when reducing a m-by-n real matrix MAT to bidiagonal form: 
!
!                              MAT = Q * BD * P'.
!
!   Q and P are defined as products of elementary reflectors H(i) and G(i),
!   respectively, determined by BD_CMP and stored in its array arguments MAT, 
!   TAUQ and TAUP.
!
!   If m >= n:
!
!   - Q = H(1) * H(2) * . . . * H(n) and ORTHO_GEN_BD returns the first n
!     columns of Q in MAT;
!   - P = G(1) * G(2) * . . . * G(n-1) and ORTHO_GEN_BD returns P as
!     an n-by-n matrix in P.
!
!   If m < n:
!
!   - Q = H(1) * H(2) * . . . * H(m-1) and ORTHO_GEN_BD returns Q as an
!     m-by-m matrix in MAT(1:m,1:m);
!   - P = G(1) * G(2) * . . . * G(m) and ORTHO_GEN_BD returns the first m
!     columns of P, in P.
!
!
! Arguments
! _________
!
!   MAT     (INPUT/OUTPUT) real(stnd), dimension(:,:)
!           On entry, the vectors which define the elementary reflectors
!           H(i) and G(i), as returned by BD_CMP in its array argument MAT.
!
!           On exit, the first min(m,n) columns of Q are stored 
!           in MAT(1:m,1:min(m,n)).
!
!   TAUQ    (INPUT) real(stnd), dimension(:)
!           TAUQ(i) must contain the scalar factor of the elementary
!           reflector H(i), which determines Q, as returned by BD_CMP
!           in its array argument TAUQ.
!
!           The size of TAUQ must verify: size( TAUQ ) = min(m,n) .
!
!   TAUP    (INPUT) real(stnd), dimension(:)
!           TAUP(i) must contain the scalar factor of the elementary
!           reflector G(i), which determines P, as returned by BD_CMP
!           in its array argument TAUP.
!
!           The size of TAUP must verify: size( TAUP ) = min(m,n) .
!
!   P       (OUTPUT) real(stnd), dimension(:,:)
!           On exit, the first min(m,n) columns of the n-by-n matrix P
!
!           The shape of p must verify: 
!
!           - size( P, 1 ) = n ,
!           - size( P, 2 ) = min(m,n).
!
!
! Further Details
! _______________
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in MAT and
!   generating the orthogonal matrices Q  and P of the bidiagonal
!   decomposition of MAT.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the bidiagonal reduction algorithm and its use
!   or the blocked algorithm used here, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,     only : assert_eq
    use QR_Procedures, only : ortho_gen_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: tauq, taup
    real(stnd),   intent(out),   dimension(:,:) :: p
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: m, n, k, l, i
!
#ifdef _OPENMP
    logical      :: test_par, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='ortho_gen_bd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b), &
                   int(size(p,1),i4b),   &
                   name_proc )
!
    k = assert_eq( min(m,n),             &
                   int(size(taup),i4b),  &
                   int(size(tauq),i4b),  &
                   int(size(p,2),i4b),   &
                   name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompnested   = omp_get_max_active_levels()>1
!
    test_par    = ompnested .and. i>1_i4b
!
    if ( test_par ) then
!
        if ( m>=n ) then
!
!           mat WAS AN m-BY-n MATRIX WITH m>=n .
!
!$OMP PARALLEL NUM_THREADS( 2 )
!
!           SHIFT THE VECTORS, IN mat, WHICH DEFINE THE ELEMENTARY 
!           REFLECTORS NEEDED TO GENERATE RIGHT VECTORS OF mat ONE 
!           ROW DOWNWARD, TRANSPOSE THEM AND COPY THEM IN p .
!
!$OMP DO SCHEDULE(STATIC)
!
            do i = 2_i4b, n
                p(i,2_i4b:i) = mat(1_i4b:i-1_i4b,i)
            end do
!
!$OMP END DO
!
!$OMP SECTIONS
!
!$OMP SECTION
!
!           SET THE FIRST COLUMN AND ROW OF p TO THAT OF THE UNIT MATRIX.
!
            p(1_i4b,1_i4b) = one
            p(2_i4b:n,1_i4b) = zero
            p(1_i4b,2_i4b:n) = zero
!
!           GENERATE RIGHT VECTORS BIDIAGONALIZING mat IN p .
!
            call ortho_gen_qr( p(2_i4b:n,2_i4b:n), taup(:n-1_i4b) )        
!
!$OMP SECTION
!
!           GENERATE LEFT VECTORS BIDIAGONALIZING mat IN mat .
!
            call ortho_gen_qr( mat, tauq )
!
!$OMP END SECTIONS
!
!$OMP END PARALLEL
!
        else
!
!           mat WAS AN m-BY-n MATRIX WITH m<n .
!
!           TRANSPOSE THE VECTORS, IN mat, WHICH DEFINE THE ELEMENTARY 
!           REFLECTORS NEEDED TO GENERATE RIGHT VECTORS BIDIAGONALIZING
!           mat AND COPY THEM IN p.
!
            do i = n, 1_i4b, -1_i4b
!
                l = min( i, m )
                p(i,1_i4b:l) = mat(1_i4b:l,i)
!
                if ( i<m ) then
!
!                   SHIFT THE VECTORS, IN mat, WHICH DEFINE THE ELEMENTARY 
!                   REFLECTORS NEEDED TO GENERATE LEFT VECTORS OF mat ONE 
!                   COLUMN RIGHTWARD.
!
                    mat(i+1_i4b:m,i+1_i4b) = mat(i+1_i4b:m,i)
!
                    mat(1_i4b,i+1_i4b) = zero
!
                end if
!
            end do
!
!$OMP PARALLEL NUM_THREADS( 2 )
!
!$OMP SECTIONS
!
!$OMP SECTION
!
!           GENERATE RIGHT VECTORS BIDIAGONALIZING mat IN p .
!
            call ortho_gen_qr( p, taup )
!
!$OMP SECTION
!
!           GENERATE LEFT VECTORS BIDIAGONALIZING mat IN mat(:m,:m) .
!
            call ortho_gen_qr( mat(2_i4b:m,2_i4b:m), tauq(:m-1_i4b) )        
!
!           SET THE FIRST COLUMN OF mat TO THAT OF THE UNIT MATRIX.
!
            mat(1_i4b,1_i4b) = one
            mat(2_i4b:m,1_i4b) = zero
!
!$OMP END SECTIONS
!
!$OMP END PARALLEL
!
        end if
!
    else
#endif
!
        if ( m>=n ) then
!
!           mat WAS AN m-BY-n MATRIX WITH m>=n .
!
!           SHIFT THE VECTORS, IN mat, WHICH DEFINE THE ELEMENTARY 
!           REFLECTORS NEEDED TO GENERATE RIGHT VECTORS OF mat ONE 
!           ROW DOWNWARD, TRANSPOSE THEM AND COPY THEM IN p .
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC,1)
!
            do i = 2_i4b, n
                p(i,2_i4b:i) = mat(1_i4b:i-1_i4b,i)
            end do
!
!$OMP END PARALLEL DO
!
!           SET THE FIRST COLUMN AND ROW OF p TO THAT OF THE UNIT MATRIX.
!
            p(1_i4b,1_i4b) = one
            p(2_i4b:n,1_i4b) = zero
            p(1_i4b,2_i4b:n) = zero
!
!           GENERATE RIGHT VECTORS BIDIAGONALIZING mat IN p .
!
            call ortho_gen_qr( p(2_i4b:n,2_i4b:n), taup(:n-1_i4b) )        
!
!           GENERATE LEFT VECTORS BIDIAGONALIZING mat IN mat .
!
            call ortho_gen_qr( mat, tauq )
!
        else
!
!           mat WAS AN m-BY-n MATRIX WITH m<n .
!
!           TRANSPOSE THE VECTORS, IN mat, WHICH DEFINE THE ELEMENTARY 
!           REFLECTORS NEEDED TO GENERATE RIGHT VECTORS BIDIAGONALIZING
!           mat AND COPY THEM IN p.
!
            do i = n, 1_i4b, -1_i4b
!
                l = min( i, m )
                p(i,1_i4b:l) = mat(1_i4b:l,i)
!
                if ( i<m ) then
!
!                   SHIFT THE VECTORS, IN mat, WHICH DEFINE THE ELEMENTARY 
!                   REFLECTORS NEEDED TO GENERATE LEFT VECTORS OF mat ONE 
!                   COLUMN RIGHTWARD.
!
                    mat(i+1_i4b:m,i+1_i4b) = mat(i+1_i4b:m,i)
!
                    mat(1_i4b,i+1_i4b) = zero
!
                end if
!
            end do
!
!           GENERATE RIGHT VECTORS BIDIAGONALIZING mat IN p .
!
            call ortho_gen_qr( p, taup )
!
!           GENERATE LEFT VECTORS BIDIAGONALIZING mat IN mat(:m,:m) .
!
            call ortho_gen_qr( mat(2_i4b:m,2_i4b:m), tauq(:m-1_i4b) )        
!
!           SET THE FIRST COLUMN OF mat TO THAT OF THE UNIT MATRIX.
!
            mat(1_i4b,1_i4b) = one
            mat(2_i4b:m,1_i4b) = zero
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
!
! END OF SUBROUTINE ortho_gen_bd
! ______________________________
!
    end subroutine ortho_gen_bd
!
! =========================================================================================
!
    subroutine ortho_gen_bd2( mat, tauq, taup, q_pt )
!
! Purpose
! _______
!
!   ORTHO_GEN_BD2 generates the real orthogonal matrices Q and P' determined by BD_CMP
!   when reducing a m-by-n real matrix MAT to bidiagonal form: 
!
!                              MAT = Q * BD * P'
!
!   Q and P' are defined as products of elementary reflectors H(i) and G(i),
!   respectively, determined by BD_CMP and stored in its array arguments MAT, 
!   TAUQ and TAUP.
!
!   If m >= n:
!
!   - Q = H(1) * H(2) * . . . * H(n) and ORTHO_GEN_BD2 returns the first n
!     columns of Q in MAT;
!   - P' = G(n-1) * . . . * G(2) * G(1) and ORTHO_GEN_BD2 returns P' as
!     an n-by-n matrix in Q_PT.
!
!   If m < n:
!
!   - Q = H(1) * H(2) * . . . * H(m-1) and ORTHO_GEN_BD2 returns Q as an
!     m-by-m matrix in Q_PT;
!   - P' = G(m) * . . . * G(2) * G(1) and ORTHO_GEN_BD2 returns the first m
!     rows of P', in MAT.
!
!
! Arguments
! _________
!
!   MAT     (INPUT/OUTPUT) real(stnd), dimension(:,:)
!           On entry, the vectors which define the elementary reflectors
!           H(i) and G(i), as returned by BD_CMP in its array argument MAT.
!
!           On exit:
!
!           - the first n columns of Q if m >= n ;
!           - the first m rows of P'   if m <  n .
!
!   TAUQ    (INPUT) real(stnd), dimension(:)
!           TAUQ(i) must contain the scalar factor of the elementary
!           reflector H(i), which determines Q, as returned by BD_CMP
!           in its array argument TAUQ.
!
!           The size of TAUQ must verify: size( TAUQ ) = min(m,n) .
!
!   TAUP    (INPUT) real(stnd), dimension(:)
!           TAUP(i) must contain the scalar factor of the elementary
!           reflector G(i), which determines P', as returned by BD_CMP 
!           in its array argument TAUP.
!
!           The size of TAUP must verify: size( TAUP ) = min(m,n) .
!
!   Q_PT    (OUTPUT) real(stnd), dimension(:,:)
!           On exit:
!
!           - the n-by-n matrix P'  if m >= n ;
!           - the m-by-m matrix Q   if m <  n .
!
!           The shape of Q_PT must verify: size( Q_PT, 1 ) = size( Q_PT, 2 ) = min(m,n).
!
!
! Further Details
! _______________
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in MAT and
!   generating the orthogonal matrices Q  and P of the bidiagonal
!   decomposition of MAT.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the bidiagonal reduction algorithm and its use
!   or the blocked algorithm, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,     only : assert_eq
    use QR_Procedures, only : ortho_gen_lq, ortho_gen_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: tauq, taup
    real(stnd),   intent(out),   dimension(:,:) :: q_pt
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b)   :: m, n, k, i
!
#ifdef _OPENMP
    logical      :: test_par, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='ortho_gen_bd2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = assert_eq( (/ min(m,n),                   &
                      int(size(taup),i4b),        &
                      int(size(tauq),i4b),        &
                      int(size(q_pt,1),i4b),      &
                      int(size(q_pt,2),i4b)  /),  &
                      name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompnested   = omp_get_max_active_levels()>1
!
    test_par    = ompnested .and. i>1_i4b
!
    if ( test_par ) then
!
        if ( m>=n ) then
!
!           mat WAS AN m-BY-n MATRIX WITH m>=n .
!
!
!$OMP PARALLEL NUM_THREADS( 2 )
!
!           SHIFT ONE ROW DOWNWARD AND COPY IN q_pt THE VECTORS, STORED IN mat, 
!           WHICH DEFINE THE ELEMENTARY REFLECTORS NEEDED TO GENERATE P' .
!
!$OMP DO SCHEDULE(STATIC)
!
            do i = 2_i4b, n
                q_pt(2_i4b:i,i) = mat(1_i4b:i-1_i4b,i)
            end do
!
!$OMP END DO
!
!$OMP SECTIONS
!
!$OMP SECTION
!
!           SET THE FIRST COLUMN AND ROW OF q_pt TO THAT OF THE UNIT MATRIX.
!
            q_pt(1_i4b,1_i4b) = one
            q_pt(1_i4b,2_i4b:n) = zero
            q_pt(2_i4b:n,1) = zero
!
!           FORM P', DETERMINED BY A CALL TO bd_cmp_a TO REDUCE mat, IN q_pt.
!
            call ortho_gen_lq( q_pt(2_i4b:n,2_i4b:n), taup(:n-1_i4b) )
!
!$OMP SECTION
!
!           FORM Q, DETERMINED BY A CALL TO bd_cmp_a TO REDUCE mat, IN mat.
!
            call ortho_gen_qr( mat, tauq )        
!
!$OMP END SECTIONS
!
!$OMP END PARALLEL
!
        else
!
!           mat WAS AN m-BY-n MATRIX WITH m<n .
!
!
!$OMP PARALLEL NUM_THREADS( 2 )
!
!           SHIFT ONE COLUMN AHEAD AND COPY IN q_pt THE VECTORS, STORED IN mat, 
!           WHICH DEFINE THE ELEMENTARY REFLECTORS NEEDED TO GENERATE Q .
!
!$OMP DO SCHEDULE(STATIC)
!
            do i = 1_i4b, m-1_i4b
                q_pt(i+1_i4b:m,i+1_i4b) = mat(i+1_i4b:m,i)
            end do
!
!$OMP END DO
!
!$OMP SECTIONS
!
!$OMP SECTION
!
!           SET THE FIRST COLUMN AND ROW OF q_pt TO THAT OF THE UNIT MATRIX.
!
            q_pt(1_i4b,1_i4b) = one
            q_pt(1_i4b,2_i4b:m) = zero
            q_pt(2_i4b:m,1_i4b) = zero
!
!           FORM Q, DETERMINED BY A CALL TO bd_cmp_a TO REDUCE mat, IN q_pt.
!
            call ortho_gen_qr( q_pt(2_i4b:m,2_i4b:m), tauq(:m-1_i4b) )
!
!$OMP SECTION
!
!           FORM P', DETERMINED BY A CALL TO bd_cmp_a TO REDUCE mat, IN mat.
!
            call ortho_gen_lq( mat, taup )        
!
!$OMP END SECTIONS
!
!$OMP END PARALLEL
!
        end if
!
    else
#endif
!
        if ( m>=n ) then
!
!           mat WAS AN m-BY-n MATRIX WITH m>=n .
!
!
!           SHIFT ONE ROW DOWNWARD AND COPY IN q_pt THE VECTORS, STORED IN mat, 
!           WHICH DEFINE THE ELEMENTARY REFLECTORS NEEDED TO GENERATE P' .
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC)
!
            do i = 2_i4b, n
                q_pt(2_i4b:i,i) = mat(1_i4b:i-1_i4b,i)
            end do
!
!$OMP END PARALLEL DO
!
!           SET THE FIRST COLUMN AND ROW OF q_pt TO THAT OF THE UNIT MATRIX.
!
            q_pt(1_i4b,1_i4b) = one
            q_pt(1_i4b,2_i4b:n) = zero
            q_pt(2_i4b:n,1) = zero
!
!           FORM P', DETERMINED BY A CALL TO bd_cmp_a TO REDUCE mat, IN q_pt.
!
            call ortho_gen_lq( q_pt(2_i4b:n,2_i4b:n), taup(:n-1_i4b) )
!
!           FORM Q, DETERMINED BY A CALL TO bd_cmp_a TO REDUCE mat, IN mat.
!
            call ortho_gen_qr( mat, tauq )        
!
        else
!
!           mat WAS AN m-BY-n MATRIX WITH m<n .
!
!
!           SHIFT ONE COLUMN AHEAD AND COPY IN q_pt THE VECTORS, STORED IN mat, 
!           WHICH DEFINE THE ELEMENTARY REFLECTORS NEEDED TO GENERATE Q .
!
!$OMP PARALLEL DO IF(m*m>=omp_limit), SCHEDULE(STATIC)
!
            do i = 1_i4b, m-1_i4b
                q_pt(i+1_i4b:m,i+1_i4b) = mat(i+1_i4b:m,i)
            end do
!
!$OMP END PARALLEL DO
!
!           SET THE FIRST COLUMN AND ROW OF q_pt TO THAT OF THE UNIT MATRIX.
!
            q_pt(1_i4b,1_i4b) = one
            q_pt(1_i4b,2_i4b:m) = zero
            q_pt(2_i4b:m,1_i4b) = zero
!
!           FORM Q, DETERMINED BY A CALL TO bd_cmp_a TO REDUCE mat, IN q_pt.
!
            call ortho_gen_qr( q_pt(2_i4b:m,2_i4b:m), tauq(:m-1_i4b) )
!
!           FORM P', DETERMINED BY A CALL TO bd_cmp_a TO REDUCE mat, IN mat.
!
            call ortho_gen_lq( mat, taup )        
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
!
! END OF SUBROUTINE ortho_gen_bd2
! _______________________________
!
    end subroutine ortho_gen_bd2
!
! =========================================================================================
!
    subroutine ortho_gen_q_bd( mat, tauq )
!
! Purpose
! _______
!
!   ORTHO_GEN_Q_BD generates the real orthogonal matrix Q determined by BD_CMP
!   when reducing a m-by-n real matrix MAT to bidiagonal form:
!
!                               MAT = Q * BD * P'
!
!   Q is defined as products of elementary reflectors H(i) determined by BD_CMP
!   and stored in its array arguments MAT and TAUQ.
!
!   If m >= n:
!
!   - Q = H(1) * H(2) * . . . * H(n) and ORTHO_GEN_Q_BD returns the first n
!     columns of Q in MAT.
!
!   If m < n:
!
!   - Q = H(1) * H(2) * . . . * H(m-1) and ORTHO_GEN_Q_BD returns Q as an
!     m-by-m matrix in MAT(:m,:m).
!
!
! Arguments
! _________
!
!   MAT     (INPUT/OUTPUT) real(stnd), dimension(:,:)
!           On entry, the vectors which define the elementary reflectors
!           H(i), as returned by BD_CMP.
!
!           On exit, the first min(m,n) columns of Q.
!
!   TAUQ    (INPUT) real(stnd), dimension(:)
!           TAUQ(i) must contain the scalar factor of the elementary
!           reflector H(i), which determines Q, as returned by BD_CMP
!           in its array argument TAUQ.
!
!           The size of TAUQ must verify: size( TAUQ ) = min(m,n) .
!
!
! Further Details
! _______________
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in MAT and
!   generating the orthogonal matrix Q of the bidiagonal decomposition
!   of MAT.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the bidiagonal reduction algorithm and its use
!   or the blocked algorithm, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,     only : assert_eq
    use QR_Procedures, only : ortho_gen_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: tauq
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b)   :: m, n, k, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='ortho_gen_q_bd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = assert_eq( min(m,n),                   &
                   int(size(tauq),i4b),        &
                   name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   FORM Q, DETERMINED BY A CALL TO bd_cmp_a OR bd_cmp_b TO REDUCE mat, IN mat.
!
    if ( m>=n ) then
!
!       mat WAS AN m-BY-n MATRIX WITH m>=n .
!
        call ortho_gen_qr( mat, tauq )        
!
    else
!
!       mat WAS AN m-BY-n MATRIX WITH m<n .
!
!
!       SHIFT THE VECTORS, IN mat, WHICH DEFINE THE ELEMENTARY 
!       REFLECTORS NEEDED TO GENERATE LEFT VECTORS OF mat ONE 
!       COLUMN RIGHTWARD.
!
        do i = m-1_i4b, 1_i4b, -1_i4b
            mat(i+1_i4b:m,i+1_i4b) = mat(i+1_i4b:m,i)
            mat(1_i4b,i+1_i4b) = zero
        end do
!
!       GENERATE LEFT VECTORS BIDIAGONALIZING mat IN mat(:m,:m) .
!
        call ortho_gen_qr( mat(2_i4b:m,2_i4b:m), tauq(:m-1_i4b) )        
!
!       SET THE FIRST COLUMN OF mat TO THAT OF THE UNIT MATRIX.
!
        mat(1_i4b,1_i4b) = one
        mat(2_i4b:m,1_i4b) = zero
!
    end if
!
!
! END OF SUBROUTINE ortho_gen_q_bd
! ________________________________
!
    end subroutine ortho_gen_q_bd
!
! =========================================================================================
!
    subroutine ortho_gen_p_bd( mat, taup, p )
!
! Purpose
! _______
!
!   ORTHO_GEN_P_BD generates the real orthogonal matrix P determined by BD_CMP
!   when reducing a m-by-n real matrix MAT to bidiagonal form:
!
!                               MAT = Q * BD * P'
!
!   P is defined as products of elementary reflectors G(i) determined by BD_CMP
!   and stored in its array arguments MAT and TAUP.
!
!   If m >= n:
!
!   - P = G(1) * G(2) * . . . * G(n-1) and ORTHO_GEN_P_BD returns P as
!     an n-by-n matrix in P.
!
!   If m < n:
!
!   - P = G(1) * G(2) * . . . * G(m) and ORTHO_GEN_P_BD returns the first m
!     columns of P, in P.
!
!
! Arguments
! _________
!
!   MAT     (INPUT) real(stnd), dimension(:,:)
!           On entry, the vectors which define the elementary reflectors
!           G(i), as returned by BD_CMP in its array argument MAT.
!
!   TAUP    (INPUT) real(stnd), dimension(:)
!           TAUP(i) must contain the scalar factor of the elementary
!           reflector G(i), which determines P, as returned by BD_CMP 
!           in its array argument TAUP.
!
!           The size of TAUP must verify: size( TAUP ) = min(m,n) .
!
!   P       (OUTPUT) real(stnd), dimension(:,:)
!           On exit, the first min(m,n) columns of the n-by-n matrix P
!
!           The shape of p must verify: 
!
!           - size( P, 1 ) = n ,
!           - size( P, 2 ) = min(m,n).
!
!
! Further Details
! _______________
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in MAT and
!   generating the orthogonal matrix P of the bidiagonal decomposition
!   of MAT.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the bidiagonal reduction algorithm and its use
!   or the blocked algorithm, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,     only : assert_eq
    use QR_Procedures, only : ortho_gen_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),  dimension(:,:) :: mat
    real(stnd),   intent(in),  dimension(:)   :: taup
    real(stnd),   intent(out), dimension(:,:) :: p
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b)   :: m, n, k, l, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='ortho_gen_p_bd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b),       &
                   int(size(p,1),i4b),         &
                   name_proc )
!
    k = assert_eq( min(m,n),                &
                   int(size(taup),i4b),     &
                   int(size(p,2),i4b),      &
                   name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   FORM P, DETERMINED BY A CALL TO bd_cmp_a TO REDUCE mat, IN p.
!
    if ( m>=n ) then
!
!       mat WAS AN m-BY-n MATRIX WITH m>=n .
!
!
!       SHIFT THE VECTORS, IN mat, WHICH DEFINE THE ELEMENTARY 
!       REFLECTORS NEEDED TO GENERATE RIGHT VECTORS OF mat ONE ROW
!       DOWNWARD, TRANSPOSE THEM AND COPY THEM IN p .
!
!$OMP PARALLEL DO IF(m*n>=omp_limit), SCHEDULE(STATIC,1_i4b), PRIVATE(i)
!
        do i = 2_i4b, n
            p(i,2_i4b:i) = mat(1_i4b:i-1_i4b,i)
        end do
!
!$OMP END PARALLEL DO
!
!       SET THE FIRST COLUMN AND ROW OF p TO THAT OF THE UNIT MATRIX.
!
        p(1_i4b,1_i4b) = one
        p(2_i4b:n,1_i4b) = zero
        p(1_i4b,2_i4b:n) = zero
!
!       GENERATE RIGHT VECTORS BIDIAGONALIZING mat IN p .
!
        call ortho_gen_qr( p(2_i4b:n,2_i4b:n), taup(:n-1_i4b) )        
!
    else
!
!       mat WAS AN m-BY-n MATRIX WITH m<n .
!
!
!       TRANSPOSE THE VECTORS, IN mat, WHICH DEFINE THE ELEMENTARY 
!       REFLECTORS NEEDED TO GENERATE RIGHT VECTORS BIDIAGONALIZING
!       mat AND COPY THEM IN p.
!
!$OMP PARALLEL DO IF(n*m>=omp_limit), SCHEDULE(STATIC), PRIVATE(i,l)
!
        do i = 1_i4b, n
            l = min( i, m )
            p(i,1_i4b:l) = mat(1_i4b:l,i)
        end do
!
!$OMP END PARALLEL DO
!
!
!       GENERATE RIGHT VECTORS BIDIAGONALIZING mat IN p .
!
        call ortho_gen_qr( p, taup )
!
    end if
!
!
! END OF SUBROUTINE ortho_gen_p_bd
! ________________________________
!
    end subroutine ortho_gen_p_bd
!
! =========================================================================================
!
    subroutine apply_q_bd( mat, tauq, c, left, trans )
!
! Purpose
! _______
!
!   APPLY_Q_BD overwrites the general real m-by-n matrix C with:
!
!   - Q  * C  if LEFT = true  and TRANS = false ;
!   - Q' * C  if LEFT = true  and TRANS = true  ;
!   - C  * Q  if LEFT = false and TRANS = false ;
!   - C  * Q' if LEFT = false and TRANS = true  .
!
!   Here Q is the orthogonal matrix determined by BD_CMP when reducing
!   a real matrix MAT to bidiagonal form:
!
!                       MAT = Q * BD * P'
!
!   and Q is defined as products of elementary reflectors H(i).
!
!   Let nq = m if LEFT = true and nq = n if LEFT = false. Thus nq is the
!   order of the orthogonal matrix Q that is applied. MAT is assumed to
!   have been an nq-by-k matrix and
!
!         Q = H(1) * H(2) * . . . * H(k)      , if nq >= k ; 
!
!   or
!
!         Q = H(1) * H(2) * . . . * H(nq-1)   , if nq < k  .
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                The vectors which define the elementary reflectors H(i),
!                whose products determine the matrix Q, as returned by
!                BD_CMP. MAT must be specified as in BD_CMP and is not
!                modified by the routine.
!
!                The shape of MAT must verify:
!
!                - if LEFT = true   : size( C, 1 ) = size( MAT, 1 ) = nq ;
!                - if LEFT = false  : size( C, 2 ) = size( MAT, 1 ) = nq .
!
!   TAUQ         (INPUT) real(stnd), dimension(:)
!                TAUQ(i) must contain the scalar factor of the elementary
!                reflector H(i) which determines Q, as returned by BD_CMP
!                in the array argument TAUQ.
!
!                The size of TAUQ must verify:
!                size( TAUQ ) = min( size(MAT,1) , size(MAT,2) ) .
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m by n matrix C.
!
!                On exit, C is overwritten by Q * C or Q' * C or C * Q' or C * Q .
!
!                The shape of C must verify:
!
!                - if LEFT = true   : size( C, 1 ) = size( MAT, 1 ) = nq ;
!                - if LEFT = false  : size( C, 2 ) = size( MAT, 1 ) = nq .
!
!   LEFT         (INPUT) logical(lgl)
!                On entry, if:
!
!                - LEFT= true  : apply Q or Q' from the left
!                - LEFT= false : apply Q or Q' from the right
!
!   TRANS        (INPUT) logical(lgl)
!                On entry, if:
!
!                - TRANS = false : apply Q  (no transpose)
!                - TRANS = true  : apply Q' (transpose)
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  DORMBR in LAPACK.
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in the lower triangle
!   of MAT and applying the orthogonal matrix Q of the bidiagonal factorization
!   to the real m-by-n matrix C.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the bidiagonal reduction algorithm and the blocked
!   version of the algorithm used here, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,     only : assert
    use QR_Procedures, only : apply_q_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: tauq
    real(stnd),   intent(inout), dimension(:,:) :: c
!
    logical(lgl), intent(in)    :: left, trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: m, n, mmat, nmat, nq, k, i1, i2
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='apply_q_bd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    mmat = size( mat, 1 )
    nmat = size( mat, 2 )
!
    m = size( c, 1 )
    n = size( c, 2 )
!
!   nq IS THE ORDER OF Q.
!
    if (  left ) then
        nq = m
    else
        nq = n
    end if
!
    k = size( tauq )
!
    call assert( logical( nq==mmat,           lgl ),    &
                 logical( k==min(mmat, nmat), lgl ),    &
                 name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b .or.  m<=0_i4b .or. n<=0_i4b ) return
!
!   APPLY Q TO c.
!
    if ( mmat>=nmat ) then
!
!       Q WAS DETERMINED BY A CALL TO bd_cmp_a WITH mmat >= nmat .
!
        call apply_q_qr( mat, tauq, c, left, trans )
!
    else if ( mmat>1_i4b ) then
!
!       Q WAS DETERMINED BY A CALL TO bd_cmp_a WITH mmat < nmat .
!
        if ( left ) then
            i1 = 2_i4b
            i2 = 1_i4b
        else
            i1 = 1_i4b
            i2 = 2_i4b
        end if
!
        call apply_q_qr( mat(2_i4b:,:), tauq(:k-1_i4b), c(i1:,i2:), left, trans )
!
    end if
!
!
! END OF SUBROUTINE apply_q_bd
! ____________________________
!
    end subroutine apply_q_bd
!
! =========================================================================================
!
    subroutine apply_p_bd( mat, taup, c, left, trans )
!
! Purpose
! _______
!
!   APPLY_P_BD overwrites the general real m-by-n matrix C with
!
!   - P  * C  if LEFT = true  and TRANS = false ;
!   - P' * C  if LEFT = true  and TRANS = true  ;
!   - C  * P  if LEFT = false and TRANS = false ;
!   - C  * P' if LEFT = false and TRANS = true  .
!
!   Here P is the orthogonal matrix determined by BD_CMP when reducing
!   a real matrix MAT to bidiagonal form:
!
!                        MAT = Q * BD * P'
!
!   and P is defined as products of elementary reflectors G(i).
!
!   Let np = m if LEFT = true and np = n if LEFT = false. Thus np is the
!   order of the orthogonal matrix P that is applied. MAT is assumed to
!   have been an k-by-np matrix and
!
!        P = G(1) * G(2) * . . . * G(k)       ,   if k < np  ;
!
!   or
!
!        P = G(1) * G(2) * . . . * G(np-1)    ,   if k >= np .
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                The vectors which define the elementary reflectors G(i),
!                whose products determine the matrix P, as returned by
!                BD_CMP. MAT must be specified as in BD_CMP and is not
!                modified by the routine.
!
!                The shape of MAT must verify:
!
!                - if LEFT = true  : size( C, 1 ) = size( MAT, 2 ) = np ;
!                - if LEFT = false : size( C, 2 ) = size( MAT, 2 ) = np .
!
!   TAUP         (INPUT) real(stnd), dimension(:)
!                TAUP(i) must contain the scalar factor of the elementary
!                reflector G(i) which determines P, as returned
!                by BD_CMP in the array argument TAUP.
!
!                The size of TAUP must verify:
!                size( TAUP ) = min( size(MAT,1) , size(MAT,2) ) .
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m by n matrix C.
!
!                On exit, C is overwritten by P * C or P' * C or C * P or C * P'.
!
!                The shape of C must verify:
!
!                - if LEFT = true  : size( C, 1 ) = size( MAT, 2 ) = np ;
!                - if LEFT = false : size( C, 2 ) = size( MAT, 2 ) = np .
!
!   LEFT         (INPUT) logical(lgl)
!                On entry, if:
!
!                - LEFT= true  : apply P or P' from the left
!                - LEFT= false : apply P or P' from the right
!
!   TRANS        (INPUT) logical(lgl)
!                On entry, if:
!
!                - TRANS = false : apply P  (no transpose)
!                - TRANS = true  : apply P' (transpose)
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  DORMBR in LAPACK.
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in the upper triangle
!   of MAT and applying the orthogonal matrix P of the bidiagonal factorization
!   to the real m-by-n matrix C.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the bidiagonal reduction algorithm and the blocked
!   version of the algorithm used here, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,     only : assert
    use QR_Procedures, only : apply_q_lq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: taup
    real(stnd),   intent(inout), dimension(:,:) :: c
!
    logical(lgl), intent(in)    :: left, trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: m, n, mmat, nmat, np, k, i1, i2
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='apply_p_bd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    mmat = size( mat, 1 )
    nmat = size( mat, 2 )
!
    m = size( c, 1 )
    n = size( c, 2 )
!
!   np IS THE ORDER P.
!
    if (  left ) then
        np = m
    else
        np = n
    end if
!
    k = size( taup )
!
    call assert( logical( np==nmat,           lgl ),    &
                 logical( k==min(mmat, nmat), lgl ),    &
                 name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b .or.  m<=0_i4b .or. n<=0_i4b ) return
!
!   APPLY P.
!
    if ( mmat<nmat ) then
!
!       P WAS DETERMINED BY A CALL TO bd_cmp_a WITH mmat < nmat .
!
        call apply_q_lq( mat, taup, c, left, .not.trans )
!
    else if ( nmat>1_i4b ) then
!
!       P WAS DETERMINED BY A CALL TO bd_cmp_a WITH mmat >= nmat .
!
        if ( left ) then
            i1 = 2_i4b
            i2 = 1_i4b
        else
            i1 = 1_i4b
            i2 = 2_i4b
        end if
!
        call apply_q_lq( mat(:,2_i4b:), taup(:k-1_i4b), c(i1:,i2:), left, .not.trans )
!
    end if
!
!
! END OF SUBROUTINE apply_p_bd
! ____________________________
!
    end subroutine apply_p_bd
!
!
! =========================================================================================
!                             SVD OF A BIDIAGONAL MATRIX
! =========================================================================================
!
!
    subroutine bd_svd_a( upper, d, e, failure, u, v, sort, maxiter,     &
                         max_francis_steps, perfect_shift, bisect, dqds )
!
! Purpose
! _______
!
!   BD_SVD computes the singular value decomposition (SVD) of a real
!   n-by-n (upper or lower) bidiagonal matrix B:
!
!                            B = Q * S * P'
!
!   , where S is a diagonal matrix with non-negative diagonal
!   elements (the singular values of B), and, Q and P are
!   orthogonal matrices (P' denotes the transpose of P).
!
!   The routine computes S, U * Q, and V * P, for given real input 
!   matrices U, V.
!
!
! Arguments
! _________
!
!   UPPER               (INPUT) logical(lgl)
!                       On entry, if:
!
!                       - UPPER = true  : B is upper bidiagonal ;
!                       - UPPER = false : B is lower bidiagonal.
!
!   D                   (INPUT/OUTPUT) real(stnd), dimension(:)
!                       On entry, D contains the diagonal elements of the bidiagonal
!                       matrix B.
!
!                       On exit,  D contains the singular values of B.
!
!   E                   (INPUT/OUTPUT) real(stnd), dimension(:)
!                       On entry, E contains the off-diagonal elements of the bidiagonal
!                       matrix whose SVD is desired. E(1) is arbitrary.
!
!                       On exit, E is destroyed.
!
!                       The size of E must verify: size( E ) = size( D ) = n . 
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit ;
!                       - FAILURE = true  :   indicates that the algorithm did not converge and
!                         that full accuracy was not attained in the bidiagonal
!                         SVD of B.
!
!   U                   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the matrix U.
!
!                       On exit, U is overwritten by U * Q.
!
!                       The shape of U must verify:
!
!                       - size( U, 1 ) > 0 ;
!                       - size( U, 2 ) = size( D ) = n .
!
!   V                   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the matrix V.
!
!                       On exit, V is overwritten by V * P.
!
!                       The shape of V must verify:
!
!                       - size( V, 1 ) > 0 ;
!                       - size( V, 2 ) = size( D ) = n .
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the singular values into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd'.
!                       The singular vectors are rearranged accordingly.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       algorithm. The algorithm fails to converge if the number 
!                       of QR sweeps exceeds MAXITER * n. Convergence usually 
!                       occurs in about 2 * n QR sweeps.
!
!                       The default is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the implicit QR algorithm. MAX_FRANCIS_STEPS
!                       is a strictly positive integer, otherwise the default value is used.
!
!                       The default is the minimum of n and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm applied to the associated n-by-n symmetric tridiagonal
!                       matrix B * B'  whose eigenvalues are the squares of the singular values of B.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm applied to the associated n-by-n symmetric tridiagonal
!                       matrix B * B'  whose eigenvalues are the squares of the singular values of B.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   If, on entry, arguments U and V are n-by-n identity matrices,
!   on exit they are replaced by Q and P, respectively.
!
!   This subroutine is adapted from subroutine QRBD given in the reference (1),
!   with modifications suggested in the references (2) and (3) for the application
!   of a set of Givens rotations to the singular vectors, and extensions to the bidiagonal
!   case of the perfect shift strategy presented in the references (4) and (5) for the
!   tridiagonal case.
!
!   Furthermore, the computation of the singular vectors is parallelized 
!   if OPENMP is used.
!
!   Note, finally, that the bidiagonal matrix is not scaled before computing the singular
!   values and vectors. If some of the elements of the bidiagonal matrix are very small or
!   large, it may be appropriate to scale the bidiagonal matrix before calling BD_SVD. 
!
!   For further details, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Lang, B., 1998:
!           Using level 3 BLAS in rotation-based algorithms.
!           Siam J. Sci. Comput., Vol. 19, 626-634.
!
!   (3) Van Zee, F.G., Van de Geijn, R., and Quintana-Orti, G., 2011:
!           Restructuring the QR Algorithm for High-Performance Application of Givens Rotations.
!           FLAME Working Note 60. The University of Texas at Austin, Department of Computer
!           Sciences. Technical Report TR-11-36.
!
!   (4) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, 1013-1034.
!
!   (5) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,        only : allocate_error, svd_error10
    use Utilities,             only : assert_eq, merror, iminloc
#ifdef _F2003
    use Utilities,             only : realloc => mvalloc
#else
    use Utilities_With_Pnter,  only : realloc
#endif
#ifdef _BLAS
    use BLAS_interfaces,       only : rot, gemm
!    use BLAS_interfaces,       only : rot, gemm, trmm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)    :: d, e
    real(stnd), intent(inout), dimension(:,:)  :: u, v
!
    integer(i4b), intent(in), optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: perfect_shift, bisect, dqds
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: eps, cs, sn, f, z, x, y, g, h, t
#ifndef _BLAS
    real(stnd), dimension( size(u,1) )      :: u_tmp
    real(stnd), dimension( size(v,1) )      :: v_tmp
#endif
    real(stnd), dimension( size(d) )        :: cs2, sn2
    real(stnd), dimension(:,:), allocatable :: uv, uvq
#ifdef _F2003
    real(stnd), dimension(:,:), allocatable :: csv, snv, csu, snu
#else
    real(stnd), dimension(:,:), pointer     :: csv, snv, csu, snu
#endif
!
#ifdef _STAT
    integer                                 :: iter_deflate_singval
#endif
    integer                                 :: iok
    integer(i4b)                            :: n, nru, nrv, maxqrs, nqrs, i, j, k,           &
                                               nfrancis, maxqrs2v ,maxqrs2u, nqrs2v, nqrs2u, &
                                               blkuv, nuv, blk, l, l1, l2, l3,               &
                                               i1, i2, ii1, ii2, q, qstart, qend
#ifdef _F2003
    integer(i4b), dimension(:), allocatable :: debv, debu, finv, finu
#else
    integer(i4b), dimension(:), pointer     :: debv, debu, finv, finu
#endif
!
    logical(lgl)                            :: deficient, small, deflate, fail, use_perfect_shift, &
                                               use_bisect, use_dqds
!
#ifdef _OPENMP2
    real(stnd), dimension(:),   allocatable :: d2, e2
    real(stnd), dimension(:,:), allocatable :: w
!
    integer(i4b)                            :: j1, j2, nrw, itv, itu, jrem, joffset,                &
                                               num_threads, thread_num, num_threads_v, num_threads_u
!
    logical                                 :: test_par, ompparallel, ompnested, comp_v
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_svd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b)  ,    &
                   int(size(e),i4b),      &
                   int(size(u,2),i4b),    &
                   int(size(v,2),i4b),    &
                   name_proc )
!
    nru = size( u, 1 )
    nrv = size( v, 1 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( min(nru,nrv)<=0_i4b ) then
!
        call merror( name_proc//svd_error10 )
!
    end if
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( d(1_i4b)<zero ) then
!
                d = -d
                u(:,:) = -u(:,:)
!
            end if
!
        end if
!
        return
!
    end if
!
    e(1_i4b) = zero
!
!   DETERMINE THE TOLERANCES FOR CONVERGENCE IN QR ITERATIONS.
!
    eps = fudge_eps*macheps*maxval( abs(d(:n)) + abs(e(:n)) )
!
!   SET OPTIONAL ARGUMENTS.
!
    if ( present(maxiter) ) then
        maxqrs = n*merge( maxiter, 10_i4b, maxiter>=1_i4b )
    else
        maxqrs = n*10_i4b
    end if
!    
    if ( present(max_francis_steps) ) then
        nfrancis = merge( max_francis_steps,                                  &
                          max_francis_steps_default,                          &
                          max_francis_steps>=1_i4b .and. max_francis_steps<=n )
    else
        nfrancis = max_francis_steps_default
    end if
!
    use_perfect_shift = true
!
    if ( present(perfect_shift) ) then
        use_perfect_shift = perfect_shift
    end if
!
    use_bisect = false
!
    if ( present(bisect) ) then
        use_bisect = bisect
    end if
!
    use_dqds = false
!
    if ( present(dqds) ) then
        use_dqds = dqds
    end if
!
!   INITIALIZE COUNTERS AND THRESHOLDS FOR QR ITERATIONS.
!
    nqrs  = 0_i4b
!
    nqrs2v    = 0_i4b
    nqrs2u    = 0_i4b
    maxqrs2v  = nfrancis + 5_i4b
    maxqrs2u  = nfrancis + 5_i4b
!
!   ALLOCATE WORK VARIABLES NEEDED TO STORE THE GIVENS ROTATIONS.
!
    allocate( csv(n,maxqrs2v),      &
              snv(n,maxqrs2v),      &
              csu(n,maxqrs2u),      &
              snu(n,maxqrs2u),      &
              debv(maxqrs2v),       &
              finv(maxqrs2v),       &
              debu(maxqrs2u),       &
              finu(maxqrs2u),       &
              stat = iok      )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    debu(:) = 1_i4b
    finu(:) = -1_i4b
!
    debv(:) = 1_i4b
    finv(:) = -1_i4b
!
#ifdef _STAT
    iter_deflate_singval  = 0
#endif
!
!   IF MATRIX IS LOWER BIDIAGONAL, ROTATE TO BE UPPER BIDIAGONAL
!   BY APPLYING GIVENS ROTATIONS ON THE LEFT.
!
    if ( .not.upper ) then

        do l = 2_i4b, n
!
            i = l - 1_i4b
!
!           DEFINE THE GIVENS ROTATION.
!
            if ( abs(d(i))>abs(e(l)) ) then
!
                t    = e(l)/d(i)
                sn   = sign( sqrt(one+t*t), d(i) )
                d(i) = d(i)*sn
                cs   = one/sn
                sn   = t*cs
!
            else if ( e(l)/=zero ) then
!
                t    = d(i)/e(l)
                cs   = sign( sqrt(one+t*t), e(l) )
                d(i) = e(l)*cs
                sn   = one/cs
                cs   = t*sn
!
            else
!
                cs = zero
                sn = one
!
            end if
!
!           APPLY THE GIVENS ROTATION.
!
            e(l) = sn*d(l)
            d(l) = cs*d(l)
!
!           STORE THE GIVENS ROTATION.
!
            csu(i,1_i4b) = cs
            snu(i,1_i4b) = sn
!                           
        end do
!
        nqrs2u = 1_i4b
!
        debu(1_i4b) = 1_i4b
        finu(1_i4b) = n - 1_i4b
!
    end if
!
#ifdef _OPENMP2
!
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel  .or. ompnested         ) .and.      &
                  ( (nrv+nru)*n>=omp_limit                   ) .and.      &
                  ( (nrv+nru)/i>=omp_limit2 )                  .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
#ifdef _BLAS
!$OMP PARALLEL PRIVATE(deflate,fail,comp_v,k,i,i1,i2,itv,itu,j,j1,j2,q,                    &
!$OMP                  qstart,qend,jrem,joffset,nrw,blkuv,nuv,blk,l,l1,l2,l3,iok,          &
!$OMP                  num_threads,thread_num,num_threads_v,num_threads_u,                 &
!$OMP                  cs,sn,f,t,z,x,y,h,g,d2,e2,w,uv,uvq),                                &
!$OMP          SHARED(upper,failure,deficient,small,use_perfect_shift,use_bisect,use_dqds, &
!$OMP                 ii1,ii2,nqrs,nqrs2u,maxqrs2u,nqrs2v,maxqrs2v,nrv,nru,                &
!$OMP                 debu,finu,debv,finv,csv,snv,csu,snu,cs2,sn2,d,e)
#else
!$OMP PARALLEL PRIVATE(deflate,fail,comp_v,k,i,i1,i2,itv,itu,j,j1,j2,q,                    &
!$OMP                  qstart,qend,jrem,joffset,nrw,blkuv,nuv,blk,l,l1,l2,l3,iok,          &
!$OMP                  num_threads,thread_num,num_threads_v,num_threads_u,                 &
!$OMP                  cs,sn,f,t,z,x,y,h,g,d2,e2,v_tmp,u_tmp,w,uv,uvq),                    &
!$OMP          SHARED(upper,failure,deficient,small,use_perfect_shift,use_bisect,use_dqds, &
!$OMP                 ii1,ii2,nqrs,nqrs2u,maxqrs2u,nqrs2v,maxqrs2v,nrv,nru,                &
!$OMP                 debu,finu,debv,finv,csv,snv,csu,snu,cs2,sn2,d,e)
#endif
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        thread_num  = omp_get_thread_num()
        num_threads = omp_get_num_threads()
!
        num_threads_u = ( nru*num_threads)/( nru + nrv )
!
        if ( num_threads_u==0_i4b ) num_threads_u = 1_i4b
!
        num_threads_v = num_threads - num_threads_u
!
        comp_v = thread_num<=num_threads_v-1_i4b
!
        if ( comp_v ) then        
!
            k = num_threads_v
            j = thread_num
            l = nrv
!
        else
!
            k = num_threads_u
            j = thread_num - num_threads_v
            l = nru
!
        end if
!
        jrem = mod( l, k )
        nrw  = ( l - jrem )/k
!
        if ( j<jrem ) then
            nrw     = nrw + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = j*nrw + joffset
        j2 = j1 + nrw - 1_i4b
!
!       ALLOCATE WORK VARIABLE TO STORE A LOCAL COPY OF THE
!       LEFT AND RIGHT SINGULAR MATRICES.
!
        allocate( w(nrw,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A LOCAL COPY OF THE INPUT MATRICES.
!
        if ( comp_v ) then        
!
            w(:nrw,:n) = v(j1:j2,:n)
!
        else
!
            w(:nrw,:n) = u(j1:j2,:n)
!
        end if
!
        nuv   = nrw
        blkuv = 0_i4b
!
        if ( thread_num==0_i4b ) then
!
            allocate( d2(n), e2(n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           MAKE A LOCAL COPY OF THE BIDIAGONAL MATRIX.
!
            d2(:n) = d(:n)
            e2(:n) = e(:n)
!
        end if
!
!       DIAGONALISATION OF THE BIDIAGONAL FORM : LOOP OVER SINGULAR VALUES
!       AND OVER ALLOWED ITERATIONS maxqrs .
!
        k = n
!
        do
!
            if ( k<1_i4b ) exit
!
            if ( thread_num==0_i4b ) then
!
                itv = nqrs2v
                itu = nqrs2u
!
                deflate = use_perfect_shift
!
                do
!
                    small     = false
                    deficient = false
!
!                   FIRST MAKE TEST FOR LAST DIAGONAL TERM d(k) BEING SMALL.
!
                    if ( abs( d2(k) )<=eps ) then
!
!                       SINCE d(k) IS SMALL WE WILL MAKE A SPECIAL PASS TO
!                       TRANSFORM e(k) TO ZERO.
!    
                        f     = e2(k)
                        e2(k) = zero
!
                        do i = k - 1_i4b, 1_i4b, -1_i4b
!
!                           CONSTRUCT AND APPLY TRANSFORMATION TO ZERO POSITION (i,k).
!
                            if ( abs(d2(i))>abs(f) ) then
!
                                t     = f/d2(i)
                                sn    = sign( sqrt(one+t*t), d2(i) )
                                d2(i) = d2(i)*sn
                                cs    = one/sn
                                sn    = t*cs
!
                            else if ( f/=zero ) then
!
                                t     = d2(i)/f
                                cs    = sign( sqrt(one+t*t), f )
                                d2(i) = f*cs
                                sn    = one/cs
                                cs    = t*sn
!
                            else
!
                                cs = zero
                                sn = one
!
                            end if
!
!                           APPLY THE GIVENS ROTATION.
!
                            f     = -sn*e2(i)
                            e2(i) =  cs*e2(i)
!
!                           STORE THE GIVENS ROTATION.
!
                            cs2(i) = cs
                            sn2(i) = sn                                
!
                        end do
!
                        small = true
!
!                       THE MATRIX IS NOW BIDIAGONAL, AND OF LOWER ORDER SINCE e(k) = zero .
!
                        exit
!
                    end if    
!
!                   TEST FOR SPLITTING OR RANK DEFICIENCIES.
!
                    do l = k, 1_i4b, -1_i4b
!
                        if ( abs( e2(l) )<=eps ) exit
!
!                       NOTE THAT THIS LOOP CANT COMPLETE SINCE e(1) = zero .
!
                        j = l - 1_i4b
!
                        if ( abs( d2(j) )<=eps ) then
!
                            deficient = true
!
!                           CANCELLATION OF e(l) , IF l > 1 .
!
                            cs = zero
                            sn = -one
!
                            do i = l, k
!
                                f     = -sn*e2(i)
                                e2(i) =  cs*e2(i)
!
                                if ( abs(f)<=eps ) exit
!
!                               DEFINE THE GIVENS ROTATION.
!
                                if ( abs(d2(i))>abs(f) ) then
!
                                    t     = f/d2(i)
                                    sn    = sign( sqrt(one+t*t), d2(i) )
                                    d2(i) = d2(i)*sn
                                    cs    = one/sn
                                    sn    = t*cs
!
                                else if ( f/=zero ) then
!
                                    t     = d2(i)/f
                                    cs    = sign( sqrt(one+t*t), f )
                                    d2(i) = f*cs
                                    sn    = one/cs
                                    cs    = t*sn
!
                                else
!
                                    cs = zero
                                    sn = one
!
                                end if
!
!                               STORE THE GIVENS ROTATION.
!
                                cs2(i) = cs
                                sn2(i) = sn                                
!
                            end do
!
                            ii1 = l
                            ii2 = i - 1_i4b
!
                            exit
!
                        end if
!
                    end do
!
!                   TEST FOR RANK DEFICIENCIES.
!
                    if ( deficient ) exit
!
!                   CUTOFF FOR CONVERGENCE FAILURE.
!                   nqrs WILL BE AT MOST 2*n USUALLY .
!
                    if ( nqrs>=maxqrs ) then
!
!                       IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM    
!                       IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
                        if ( l/=k ) then
!
!                           SET FAILURE INDICATOR.
!
                            failure = true
!
                            l     = iminloc( abs(e2(l+1_i4b:k))) + l
                            e2(l) = zero
                            nqrs  = 0_i4b
!
                        end if
!
                    end if
!
!                   TEST FOR CONVERGENCE.
!
                    if ( l==k ) exit
!
                    nqrs = nqrs + 1_i4b
!
!                   REALLOCATE WORKSPACE TO STORE THE FRANCIS SETS
!                   OF GIVENS ROTATIONS FOR THE RIGHT SINGULAR VECTORS
!                   IF NEEDED.
!
                    itv = itv + 1_i4b
!
                    if ( itv>maxqrs2v ) then
!
                        maxqrs2v = maxqrs2v + 5_i4b
!
                        call realloc( debv, maxqrs2v, iok )
!
                        if ( iok==0 ) then
!
                            call realloc( finv, maxqrs2v, iok  )
!
                            if ( iok==0 ) then
!
                                call realloc( csv, n, maxqrs2v, iok  )
!
                                if ( iok==0 ) then
                                    call realloc( snv, n, maxqrs2v, iok  )
                                end if
!
                            end if
!
                        end if
!
                        if ( iok/=0 ) then
                            call merror( name_proc//allocate_error )
                        end if
!
                    end if
!
!                   REALLOCATE WORKSPACE TO STORE THE FRANCIS SETS
!                   OF GIVENS ROTATIONS FOR THE LEFT SINGULAR VECTORS
!                   IF NEEDED.
!
                    itu = itu + 1_i4b
!
                    if ( itu>maxqrs2u ) then
!
                        maxqrs2u = maxqrs2u + 5_i4b
!
                        call realloc( debu, maxqrs2u, iok )
!
                        if ( iok==0 ) then
!
                            call realloc( finu, maxqrs2u, iok  )
!
                            if ( iok==0 ) then
!
                                call realloc( csu, n, maxqrs2u, iok  )
!
                                if ( iok==0 ) then
                                    call realloc( snu, n, maxqrs2u, iok  )
                                end if
!
                            end if
!
                        end if
!
                        if ( iok/=0 ) then
                            call merror( name_proc//allocate_error )
                        end if
!
                    end if
!
                    if ( deflate ) then
!
                        if ( use_bisect ) then
!
!                           COMPUTE THE NEXT SINGULAR VALUE BY A BISECTION ALGORITHM.
!
                            call bd_one_singval_bisect( d2(l:k), e2(l:k), g, fail )
!
                        else if ( use_dqds ) then
!
!                           COMPUTE THE NEXT SINGULAR VALUE BY THE DQDS ALGORITHM.
!
                            call bd_one_singval_dqds( d2(l:k), e2(l:k), g, fail )
!
!                            call bd_one_singval_lasq1( d2(l:k), e2(l:k), g, fail )
!
                        else
!
!                           COMPUTE THE NEXT SINGULAR VALUE BY THE PAL-WALKER-KAHAN ALGORITHM.
!
                            call bd_one_singval_pwk( d2(l:k), e2(l:k), g, fail )
!
                        end if
!
                        deflate = .not.fail
!
!                       NOW COMPUTE THE SINGULAR VECTORS.
!
                        if ( deflate ) then
!
!                           ONE SINGULAR VALUE FOUND.
!
!                           FIRST TRY FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION.
!
!                            call dflgen_bd( d2(l:k), e2(l+1_i4b:k), g,                &
!                                            csu(l:k-1_i4b,itu), snu(l:k-1_i4b,itu),   &
!                                            csv(l:k-1_i4b,itv), snv(l:k-1_i4b,itv)    )
!
!                            call dflapp_bd( d2(l:k), e2(l+1_i4b:k),                         &
!                                            csu(l:k-1_i4b,itu), snu(l:k-1_i4b,itu),         &
!                                            csv(l:k-1_i4b,itv), snv(l:k-1_i4b,itv), deflate )
!
                            call dflgen2_bd( d2(l:k), e2(l+1_i4b:k), g,                      &
                                             csu(l:k-1_i4b,itu), snu(l:k-1_i4b,itu),         &
                                             csv(l:k-1_i4b,itv), snv(l:k-1_i4b,itv), deflate )
!
                            if ( deflate ) then
!
#ifdef _STAT
                                iter_deflate_singval  = iter_deflate_singval + 1
#endif
!
                                e2(l) = zero
!
                                debv(itv) = l
                                finv(itv) = k - 1_i4b
!
                                debu(itu) = l
                                finu(itu) = k - 1_i4b
!
!                               GODUNOV DEFLATION SUCCESSFUL, EXIT.
!
                                exit
!
                            end if
!
                        end if
!
                    end if
!
!                   IF GODUNOV DEFLATION IS NOT SUCCESSFUL, USE IMPLICIT QR STEPS.
!
!                   COMPUTE WILKINSON SHIFT FROM BOTTOM 2 BY 2 MINOR OF B'*B .
!
                    x = d2(l)
                    z = d2(k)
                    y = d2(k-1_i4b)
                    g = e2(k-1_i4b)
                    h = e2(k)
                    f = ( (y-z)*(y+z) + (g-h)*(g+h) )/(two*h*y)
!
                    if ( abs(f)>one ) then
!
                        t = one/f
                        g = f*sqrt( one + t*t )
!
                    else
!
                        g = sign( sqrt( one + f*f ), f )
!
                    end if
!
                    f = ( (x-z)*(x+z) + h*( (y/( f + g )) - h ) )/x
!
!                   NEXT QR SWEEP.
!
                    cs = one
                    sn = one
!
                    do i = l + 1_i4b, k
!
                        j = i - 1_i4b
                        g = e2(i)
                        y = d2(i)
                        h = sn*g
                        g = cs*g
!
!                       DEFINE THE GIVENS ROTATION.
!
                        if ( abs(f)>abs(h) ) then
!
                            t     = h/f
                            sn    = sign( sqrt(one+t*t), f )
                            e2(j) = f*sn
                            cs    = one/sn
                            sn    = t*cs
!
                        else if ( h/=zero ) then
!
                            t     = f/h
                            cs    = sign( sqrt(one+t*t), h )
                            e2(j) = h*cs
                            sn    = one/cs
                            cs    = t*sn
!
                        else
!
                            cs    = zero
                            sn    = one
                            e2(j) = zero
!
                        end if
!
!                       APPLY THE GIVENS ROTATION.
!
                        f = x*cs + g*sn
                        g = -x*sn + g*cs
                        h = y*sn
                        y = y*cs
!
!                       STORE THE GIVENS ROTATION.
!
                        csv(j,itv) = cs
                        snv(j,itv) = sn
!
!                       DEFINE THE GIVENS ROTATION.
!
                        if ( abs(f)>abs(h) ) then
!
                            t     = h/f
                            sn    = sign( sqrt(one+t*t), f )
                            d2(j) = f*sn
                            cs    = one/sn
                            sn    = t*cs
!
                        else if ( h/=zero ) then
!
                            t     = f/h
                            cs    = sign( sqrt(one+t*t), h )
                            d2(j) = h*cs
                            sn    = one/cs
                            cs    = t*sn
!
                        else
!
                            cs    = zero
                            sn    = one
                            d2(j) = zero
!
                        end if
!
!                       APPLY THE GIVENS ROTATION.
!
                        f =  cs*g + sn*y
                        x = -sn*g + cs*y
!
!                       STORE THE GIVENS ROTATION.
!
                        csu(j,itu) = cs
                        snu(j,itu) = sn
!
                    end do
!
                    e2(l) = zero
                    e2(k) = f
                    d2(k) = x
!
                    debv(itv) = l
                    finv(itv) = k - 1_i4b
!
                    debu(itu) = l
                    finu(itu) = k - 1_i4b
!
!                   RETURN TO TEST FOR RANK DEFICIENCIES AND SPLITTING.    
!
                end do
!
                nqrs2v = itv
                nqrs2u = itu
!
            end if
!
!$OMP BARRIER
!
            if ( comp_v .and. ( nqrs2v>=nfrancis .or. k==1_i4b .or. small )  ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS
!               TO THE RIGHT SINGULAR VECTORS.
!
                qstart = maxval( debv(:nqrs2v) ) + nqrs2v - 1_i4b
                qend   = minval( finv(:nqrs2v) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2v==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2v
!
                        i1 = debv(j)
                        i2 = finv(j)
!
#ifdef _BLAS
                        do i = i1, i2
!
                            call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                        end do
#else
                        v_tmp(:nrw) = w(:nrw,i1)
!
                        do i = i1, i2
!
                            cs = csv(i,j)
                            sn = snv(i,j)
!
                            w(:nrw,i)   =  cs*v_tmp(:nrw) + sn*w(:nrw,i+1_i4b)
                            v_tmp(:nrw) = -sn*v_tmp(:nrw) + cs*w(:nrw,i+1_i4b)
!
                        end do
!
                        w(:nrw,i) = v_tmp(:nrw)
#endif
!
                    end do
!
                else if ( nqrs2v>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    l  = 0_i4b
                    i2 = qstart - nqrs2v + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        l = l + 1_i4b
!
                        i1 = debv(l)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
!
                                call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csv(i,l), snv(i,l) )
!
                            end do
#else
                            v_tmp(:nrw) = w(:nrw,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                cs = csv(i,l)
                                sn = snv(i,l)
!
                                w(:nrw,i)   =  cs*v_tmp(:nrw) + sn*w(:nrw,i+1_i4b)
                                v_tmp(:nrw) = -sn*v_tmp(:nrw) + cs*w(:nrw,i+1_i4b)
!
                            end do
!
                            w(:nrw,i) = v_tmp(:nrw)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, l
!
                            call rot( nrw, w(:nrw,i+1_i4b), 1_i4b, w(:nrw,i), 1_i4b, csv(i,j), -snv(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        v_tmp(:nrw) = w(:nrw,i+1_i4b)
!
                        do j = 1_i4b, l
!
                            cs = csv(i,j)
                            sn = snv(i,j)
!
                            w(:nrw,i+1_i4b) = -sn*w(:nrw,i) + cs*v_tmp(:nrw)
                            v_tmp(:nrw)     =  cs*w(:nrw,i) + sn*v_tmp(:nrw)
!
                            i = i - 1_i4b
!
                        end do
!
                        w(:nrw,i+1_i4b) = v_tmp(:nrw)
#endif
!
                    end do
!
!                   PIPELINE PHASE.
!
                    blk = 2_i4b*nqrs2v
!
                    if ( nrw>thr_bd_svd_blas3*blk ) then
!
!                       USE A BLAS3 ALGORITHM IF THE NUMBER OF ELEMENTS IN RIGHT
!                       SINGULAR VECTORS IS SIGNIFICANTLY LARGER THAN THE BLOCKSIZE.
!
                        if ( blk>blkuv ) then
!
!                           REALLOCATE WORKSPACE TO STORE THE ACCUMULATED GIVENS
!                           ROTATIONS AND UPDATE THE RIGHT SINGULAR VECTORS.
!
                            blkuv = 2_i4b*maxqrs2v
!
                            if ( allocated(uv) ) then
!
                                deallocate( uv, uvq )
!
                            end if
!
                            allocate( uv(nuv,blkuv), uvq(blkuv,blkuv), stat = iok )
!
                            if ( iok/=0 ) then
                                call merror( name_proc//allocate_error )
                            end if
!
                        end if
!
                        l   = qend - qstart + 1_i4b
                        l3  = l/nqrs2v
!
                        if ( l/=l3*nqrs2v ) l3 = l3 + 1_i4b
!
                        l1  = qstart - blk + 1_i4b
                        l2  = qstart
!
                        do l = 1_i4b, l3
!
                            l1  = l1 + nqrs2v
                            l2  = min( l2+nqrs2v, qend+1_i4b )
!
                            blk = l2 - l1 + 1_i4b
!
!                           ACCUMULATE GIVENS ROTATIONS OF THE CURRENT SET
!                           IN MATRIX uvq .
!
                            do j = 1_i4b, blk
!
                                uvq(:blkuv,j) = zero
                                uvq(j,j)      = one
!
                            end do
!
                            i1 = nqrs2v - 1_i4b
!
#ifdef _BLAS
                            do q = l1+nqrs2v-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2v
!
                                    call rot( blk, uvq(:blk,i2), 1_i4b, uvq(:blk,i2+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE RIGHT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
!                            if ( blk==2_i4b*nqrs2v ) then
!
!                               PROCESS THE REGULAR SETS.
!
!                                uv(:nrw,nqrs2v+1_i4b:blk) = w(:nrw,l1:l2-nqrs2v)
!
!                                call trmm( 'R', 'L', 'N', 'N', nrw, nqrs2v, one, uvq(:blkuv,nqrs2v+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,nqrs2v+1_i4b:blk), nuv )
!
!                                uvq(:nqrs2v,nqrs2v+1_i4b:blk) = uvq(nqrs2v+1_i4b:blk,nqrs2v+1_i4b:blk)
!
!                                call gemm( 'N', 'N', nrw, nqrs2v, nqrs2v, one, w(:nrw,l1+nqrs2v:l2), nrw,            &
!                                           uvq(:blkuv,nqrs2v+1_i4b:blk), blkuv, one, uv(:nuv,nqrs2v+1_i4b:blk), nuv  )
!
!                                uv(:nrw,:nqrs2v) = w(:nrw,l1+nqrs2v:l2)
!
!                                uvq(:nqrs2v,nqrs2v+1_i4b:blk) = uvq(nqrs2v+1_i4b:blk,:nqrs2v)
!
!                                call trmm( 'R', 'U', 'N', 'N', nrw, nqrs2v, one, uvq(:blkuv,nqrs2v+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,:nqrs2v), nuv )
!
!                                call gemm( 'N', 'N', nrw,  nqrs2v, nqrs2v, one, w(:nrw,l1:l2-nqrs2v), nrw, &
!                                           uvq(:blkuv,:nqrs2v), blkuv, one, uv(:nuv,:nqrs2v), nuv )
!
!                            else
!
!                               PROCESS THE RESIDUAL SET IF REQUIRED.
!
                                call gemm( 'N', 'N', nrw, blk, blk, one, w(:nrw,l1:l2), nrw,  &
                                           uvq(:blkuv,:blk), blkuv, zero, uv(:nuv,:blk), nuv  )
!
!                            end if
!
#else
                            do q = l1+nqrs2v-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2v
!
                                    cs = csv(i,j)
                                    sn = snv(i,j)
!
                                    v_tmp(:blk)        = cs*uvq(:blk,i2) + sn*uvq(:blk,i2+1_i4b)
                                    uvq(:blk,i2+1_i4b) = cs*uvq(:blk,i2+1_i4b) - sn*uvq(:blk,i2)
                                    uvq(:blk,i2)       = v_tmp(:blk)
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE RIGHT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
                            uv(:nrw,:blk) = matmul( w(:nrw,l1:l2), uvq(:blk,:blk) )
#endif
!
                            w(:nrw,l1:l2) = uv(:nrw,:blk)
!
                        end do
!
                    else
!
!                       USE A BLAS1 ALGORITHM OTHERWISE.
!
                        do q = qstart, qend
!
                            i = q
!
#ifdef _BLAS
                            do j = 1_i4b, nqrs2v
!
                                call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                                i = i - 1_i4b
!
                            end do
#else
                            v_tmp(:nrw) = w(:nrw,i+1_i4b)
!
                            do j = 1_i4b, nqrs2v
!
                                cs = csv(i,j)
                                sn = snv(i,j)
!
                                w(:nrw,i+1_i4b) = -sn*w(:nrw,i) + cs*v_tmp(:nrw)
                                v_tmp(:nrw)     =  cs*w(:nrw,i) + sn*v_tmp(:nrw)
!
                                i = i - 1_i4b
!
                            end do
!
                            w(:nrw,i+1_i4b) = v_tmp(:nrw)
#endif
!
                        end do
!
                    end if
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2v
!
                        i2 = finv(j)
!
                        if ( i1<=i2 ) then
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                            end do
#else
                            v_tmp(:nrw) = w(:nrw,i1)
!
                            do i = i1, i2
!
                                cs = csv(i,j)
                                sn = snv(i,j)
!
                                w(:nrw,i)   =  cs*v_tmp(:nrw) + sn*w(:nrw,i+1_i4b)
                                v_tmp(:nrw) = -sn*v_tmp(:nrw) + cs*w(:nrw,i+1_i4b)
!
                            end do
!
                            w(:nrw,i) = v_tmp(:nrw)
#endif
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
            else if ( .not.comp_v .and. ( nqrs2u>=nfrancis .or. k==1_i4b  .or. deficient ) ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS
!               TO THE LEFT SINGULAR VECTORS.
!
                qstart = maxval( debu(:nqrs2u) ) + nqrs2u - 1_i4b
                qend   = minval( finu(:nqrs2u) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2u==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2u
!
                        i1 = debu(j)
                        i2 = finu(j)
!
#ifdef _BLAS
                        do i = i1, i2
!
                            call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                        end do
#else
                        u_tmp(:nrw) = w(:nrw,i1)
!
                        do i = i1, i2
!
                            cs = csu(i,j)
                            sn = snu(i,j)
!
                            w(:nrw,i)   =  cs*u_tmp(:nrw) + sn*w(:nrw,i+1_i4b)
                            u_tmp(:nrw) = -sn*u_tmp(:nrw) + cs*w(:nrw,i+1_i4b)
!
                        end do
!
                        w(:nrw,i) = u_tmp(:nrw)
#endif
!
                    end do
!
                else if ( nqrs2u>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    l  = 0_i4b
                    i2 = qstart - nqrs2u + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        l = l + 1_i4b
!
                        i1 = debu(l)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
!
                                call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csu(i,l), snu(i,l) )
!
                            end do
#else
                            u_tmp(:nrw) = w(:nrw,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                cs = csu(i,l)
                                sn = snu(i,l)
!
                                w(:nrw,i)   =  cs*u_tmp(:nrw) + sn*w(:nrw,i+1_i4b)
                                u_tmp(:nrw) = -sn*u_tmp(:nrw) + cs*w(:nrw,i+1_i4b)
!
                            end do
!
                            w(:nrw,i) = u_tmp(:nrw)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, l
!
                            call rot( nrw, w(:nrw,i+1_i4b), 1_i4b, w(:nrw,i), 1_i4b, csu(i,j), -snu(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        u_tmp(:nrw) = w(:nrw,i+1_i4b)
!
                        do j = 1_i4b, l
!
                            cs = csu(i,j)
                            sn = snu(i,j)
!
                            w(:nrw,i+1_i4b) = -sn*w(:nrw,i) + cs*u_tmp(:nrw)
                            u_tmp(:nrw)     =  cs*w(:nrw,i) + sn*u_tmp(:nrw)
!
                            i = i - 1_i4b
!
                        end do
!
                        w(:nrw,i+1_i4b) = u_tmp(:nrw)
#endif
!
                    end do
!
!                   PIPELINE PHASE.
!
                    blk = 2_i4b*nqrs2u
!
                    if ( nrw>thr_bd_svd_blas3*blk ) then
!
!                       USE A BLAS3 ALGORITHM IF THE NUMBER OF ELEMENTS IN LEFT
!                       SINGULAR VECTORS IS SIGNIFICANTLY LARGER THAN THE BLOCKSIZE.
!
                        if ( blk>blkuv ) then
!
!                           REALLOCATE WORKSPACE TO STORE THE ACCUMULATED GIVENS
!                           ROTATIONS AND UPDATE THE LEFT SINGULAR VECTORS.
!
                            blkuv = 2_i4b*maxqrs2u
!
                            if ( allocated(uv) ) then
!
                                deallocate( uv, uvq )
!
                            end if
!
                            allocate( uv(nuv,blkuv), uvq(blkuv,blkuv), stat = iok )
!
                            if ( iok/=0 ) then
                                call merror( name_proc//allocate_error )
                            end if
!
                        end if
!
                        l   = qend - qstart + 1_i4b
                        l3  = l/nqrs2u
!
                        if ( l/=l3*nqrs2u ) l3 = l3 + 1_i4b
!
                        l1  = qstart - blk + 1_i4b
                        l2  = qstart
!
                        do l = 1_i4b, l3
!
                            l1  = l1 + nqrs2u
                            l2  = min( l2+nqrs2u, qend+1_i4b )
!
                            blk = l2 - l1 + 1_i4b
!
!                           ACCUMULATE GIVENS ROTATIONS OF THE CURRENT SET
!                           IN MATRIX uvq .
!
                            do j = 1_i4b, blk
!
                                uvq(:blkuv,j) = zero
                                uvq(j,j)      = one
!
                            end do
!
                            i1 = nqrs2u - 1_i4b
!
#ifdef _BLAS
                            do q = l1+nqrs2u-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2u
!
                                    call rot( blk, uvq(:blk,i2), 1_i4b, uvq(:blk,i2+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE LEFT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
!                            if ( blk==2_i4b*nqrs2u ) then
!
!                               PROCESS THE REGULAR SETS.
!
!                                uv(:nrw,nqrs2u+1_i4b:blk) = w(:nrw,l1:l2-nqrs2u)
!
!                                call trmm( 'R', 'L', 'N', 'N', nrw, nqrs2u, one, uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,nqrs2u+1_i4b:blk), nuv )
!
!                                uvq(:nqrs2u,nqrs2u+1_i4b:blk) = uvq(nqrs2u+1_i4b:blk,nqrs2u+1_i4b:blk)
!
!                                call gemm( 'N', 'N', nrw, nqrs2u, nqrs2u, one, w(:nrw,l1+nqrs2u:l2), nrw,            &
!                                           uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv, one, uv(:nuv,nqrs2u+1_i4b:blk), nuv  )
!
!                                uv(:nrw,:nqrs2u) = w(:nrw,l1+nqrs2u:l2)
!
!                                uvq(:nqrs2u,nqrs2u+1_i4b:blk) = uvq(nqrs2u+1_i4b:blk,:nqrs2u)
!    
!                                call trmm( 'R', 'U', 'N', 'N', nrw, nqrs2u, one, uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,:nqrs2u), nuv )
!
!                                call gemm( 'N', 'N', nrw,  nqrs2u, nqrs2u, one, w(:nrw,l1:l2-nqrs2u), nrw, &
!                                           uvq(:blkuv,:nqrs2u), blkuv, one, uv(:nuv,:nqrs2u), nuv )
!
!                            else
!
!                               PROCESS THE RESIDUAL SET IF REQUIRED.
!
                                call gemm( 'N', 'N', nrw, blk, blk, one, w(:nrw,l1:l2), nrw,  &
                                           uvq(:blkuv,:blk), blkuv, zero, uv(:nuv,:blk), nuv  )
!
!                            end if
#else
                            do q = l1+nqrs2u-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2u
!
                                    cs = csu(i,j)
                                    sn = snu(i,j)
!
                                    u_tmp(:blk)        = cs*uvq(:blk,i2) + sn*uvq(:blk,i2+1_i4b)
                                    uvq(:blk,i2+1_i4b) = cs*uvq(:blk,i2+1_i4b) - sn*uvq(:blk,i2)
                                    uvq(:blk,i2)       = u_tmp(:blk)
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE LEFT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
                            uv(:nrw,:blk) = matmul( w(:nrw,l1:l2), uvq(:blk,:blk) )
#endif
!
                            w(:nrw,l1:l2) = uv(:nrw,:blk)
!
                        end do
!
                    else
!
!                       USE A BLAS1 ALGORITHM OTHERWISE.
!
                        do q = qstart, qend
!
                            i = q
!
#ifdef _BLAS
                            do j = 1_i4b, nqrs2u
!
                                call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                                i = i - 1_i4b
!
                            end do
#else
                            u_tmp(:nrw) = w(:nrw,i+1_i4b)
!
                            do j = 1_i4b, nqrs2u
!
                                cs = csu(i,j)
                                sn = snu(i,j)
!
                                w(:nrw,i+1_i4b) = -sn*w(:nrw,i) + cs*u_tmp(:nrw)
                                u_tmp(:nrw)     =  cs*w(:nrw,i) + sn*u_tmp(:nrw)
!
                                i = i - 1_i4b
!
                            end do
!
                            w(:nrw,i+1_i4b) = u_tmp(:nrw)
#endif
!
                        end do
!
                    end if
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2u
!
                        i2 = finu(j)
!
                        if ( i1<=i2 ) then
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                            end do
#else
                            u_tmp(:nrw) = w(:nrw,i1)
!
                            do i = i1, i2
!
                                cs = csu(i,j)
                                sn = snu(i,j)
!
                                w(:nrw,i)   =  cs*u_tmp(:nrw) + sn*w(:nrw,i+1_i4b)
                                u_tmp(:nrw) = -sn*u_tmp(:nrw) + cs*w(:nrw,i+1_i4b)
!
                            end do
!
                            w(:nrw,i) = u_tmp(:nrw)
#endif
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
            end if
!
!$OMP BARRIER
!
!$OMP MASTER
            if ( nqrs2v>=nfrancis .or. k==1_i4b .or. small  )    nqrs2v = 0_i4b
            if ( nqrs2u>=nfrancis .or. k==1_i4b .or. deficient ) nqrs2u = 0_i4b
!$OMP END MASTER
!
            if ( small .and. comp_v ) then        
!
                do i = k - 1_i4b, 1_i4b, -1_i4b
!
!                   APPLY PREVIOUS GIVENS ROTATIONS TO TRANSFORM e(k) TO ZERO
!                   TO THE RIGHT SINGULAR VECTORS.
!
#ifdef _BLAS
                    call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,k), 1_i4b, cs2(i), sn2(i) )
#else
                    cs = cs2(i)
                    sn = sn2(i)                             
!
                    v_tmp(:nrw) = -sn*w(:nrw,i)
                    w(:nrw,i)   =  cs*w(:nrw,i) + sn*w(:nrw,k)
                    w(:nrw,k)   =  v_tmp(:nrw)  + cs*w(:nrw,k)
#endif
!
                end do
!
            end if
!
            if ( deficient ) then
!
                if ( .not.comp_v ) then        
!
!                   APPLY THE GIVENS ROTATIONS TO COPE WITH
!                   RANK DEFICIENCIES TO THE LEFT SINGULAR VECTORS.
!
                    j = ii1 - 1_i4b
!
                    do i = ii1, ii2
!
#ifdef _BLAS
                        call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,j), 1_i4b, cs2(i), sn2(i) )
#else
                        cs = cs2(i)
                        sn = sn2(i)
!
                        u_tmp(:nrw) = -sn*w(:nrw,i)
                        w(:nrw,i)   =  cs*w(:nrw,i) + sn*w(:nrw,j)
                        w(:nrw,j)   =  u_tmp(:nrw)  + cs*w(:nrw,j)
#endif
!
                    end do
!
                end if
!
            else
!
                k = k - 1_i4b
!
            end if
!
!$OMP BARRIER
!
        end do
!
        if ( thread_num==0_i4b ) then
!
            d(:n) = d2(:n)
            e(:n) = e2(:n)
!
            deallocate( d2, e2 )
!
        end if
!
!$OMP BARRIER
!
!       SAVE THE COMPUTED SINGULAR VECTORS.
!
        if ( comp_v ) then        
!
            v(j1:j2,:n) = w(:nrw,:n)
!
        else
!
!           CONVERGENCE, MAKE SINGULAR VALUES POSITIVE AND
!           CHANGE SIGN OF LEFT SINGULAR VECTORS.
!
            do i = 1_i4b, n
!
                if ( d(i)>=zero ) cycle
!
                w(:nrw,i) = -w(:nrw,i)
!
            end do
!
            u(j1:j2,:n) = w(:nrw,:n)
!
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( w )
!
        if ( allocated(uv) ) then
!
            deallocate( uv, uvq )
!
        end if
!
!$OMP END PARALLEL
!
!       CONVERGENCE, MAKE SINGULAR VALUES POSITIVE.
!
        where ( d(1_i4b:n)<zero ) d(1_i4b:n) = -d(1_i4b:n)
!
    else
#endif
!
        blkuv = 0_i4b
        nuv   = max( nru, nrv ) 
!
!       DIAGONALISATION OF THE BIDIAGONAL FORM : LOOP OVER SINGULAR VALUES
!       AND OVER ALLOWED ITERATIONS maxqrs .
!
        k = n
!
        do
!
            if ( k<1_i4b ) exit
!
            deflate = use_perfect_shift
!
            do
!
                small     = false
                deficient = false
!
!               FIRST MAKE TEST FOR LAST DIAGONAL TERM d(k) BEING SMALL.
!
                if ( abs( d(k) )<=eps ) then
!
!                   SINCE d(k) IS SMALL WE WILL MAKE A SPECIAL PASS TO
!                   TRANSFORM e(k) TO ZERO.
!    
                    f    = e(k)
                    e(k) = zero
!
                    do i = k - 1_i4b, 1_i4b, -1_i4b
!
!                       CONSTRUCT AND APPLY TRANSFORMATION TO ZERO POSITION (i,k).
!
                        if ( abs(d(i))>abs(f) ) then
!
                            t    = f/d(i)
                            sn   = sign( sqrt(one+t*t), d(i) )
                            d(i) = d(i)*sn
                            cs   = one/sn
                            sn   = t*cs
!
                        else if ( f/=zero ) then
!
                            t    = d(i)/f
                            cs   = sign( sqrt(one+t*t), f )
                            d(i) = f*cs
                            sn   = one/cs
                            cs   = t*sn
!
                        else
!
                            cs = zero
                            sn = one
!
                        end if
!
!                       APPLY THE GIVENS ROTATION.
!
                        f    = -sn*e(i)
                        e(i) =  cs*e(i)
!
!                       STORE THE GIVENS ROTATION.
!
                        cs2(i) = cs
                        sn2(i) = sn                                
!
                    end do
!
                    small = true
!
!                   THE MATRIX IS NOW BIDIAGONAL, AND OF LOWER ORDER SINCE e(k) = zero .
!
                    exit
!
                end if    
!
!               TEST FOR SPLITTING OR RANK DEFICIENCIES.
!
                do l = k, 1_i4b, -1_i4b
!
                    if ( abs( e(l) )<=eps ) exit
!
!                   NOTE THAT THIS LOOP CANT COMPLETE SINCE e(1) = zero .
!
                    j = l - 1_i4b
                    if ( abs( d(j) )<=eps ) then
!
                        deficient = true
!
!                       CANCELLATION OF e(l) , IF l > 1 .
!
                        cs = zero
                        sn = -one
!
                        do i = l, k
!
                            f    = -sn*e(i)
                            e(i) =  cs*e(i)
!
                            if ( abs(f)<=eps ) exit
!
!                           DEFINE THE GIVENS ROTATION.
!
                            if ( abs(d(i))>abs(f) ) then
!
                                t    = f/d(i)
                                sn   = sign( sqrt(one+t*t), d(i) )
                                d(i) = d(i)*sn
                                cs   = one/sn
                                sn   = t*cs
!
                            else if ( f/=zero ) then
!
                                t    = d(i)/f
                                cs   = sign( sqrt(one+t*t), f )
                                d(i) = f*cs
                                sn   = one/cs
                                cs   = t*sn
!
                            else
!
                                cs = zero
                                sn = one
!
                            end if
!
!                           STORE THE GIVENS ROTATION.
!
                            cs2(i) = cs
                            sn2(i) = sn                                
!
                        end do
!
                        ii1 = l
                        ii2 = i - 1_i4b
!
                        exit
!
                    end if
!
                end do
!
!               TEST FOR RANK DEFICIENCIES.
!
                if ( deficient ) exit
!
!               CUTOFF FOR CONVERGENCE FAILURE.
!               nqrs WILL BE AT MOST 2*n USUALLY .
!
                if ( nqrs>=maxqrs ) then
!
!                   IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM    
!                   IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
                    if ( l/=k ) then
!
!                       SET FAILURE INDICATOR.
!
                        failure = true
!
                        l    = iminloc( abs(e(l+1_i4b:k))) + l
                        e(l) = zero
                        nqrs = 0_i4b
!
                    end if
!
                end if
!
!               TEST FOR CONVERGENCE.
!
                if ( l==k ) exit
!
                nqrs = nqrs + 1_i4b
!
!               REALLOCATE WORKSPACE TO STORE THE FRANCIS SETS
!               OF GIVENS ROTATIONS FOR THE RIGHT SINGULAR VECTORS
!               IF NEEDED.
!
                nqrs2v = nqrs2v + 1_i4b
!
                if ( nqrs2v>maxqrs2v ) then
!
                    maxqrs2v = maxqrs2v + 5_i4b
!
                    call realloc( debv, maxqrs2v, iok )
!
                    if ( iok==0 ) then
!
                        call realloc( finv, maxqrs2v, iok  )
!
                        if ( iok==0 ) then
!
                            call realloc( csv, n, maxqrs2v, iok  )
!
                            if ( iok==0 ) then
                                call realloc( snv, n, maxqrs2v, iok  )
                            end if
!
                        end if
!
                    end if
!
                    if ( iok/=0 ) then
                        call merror( name_proc//allocate_error )
                    end if
!
                end if
!
!               REALLOCATE WORKSPACE TO STORE THE FRANCIS SETS
!               OF GIVENS ROTATIONS FOR THE LEFT SINGULAR VECTORS
!               IF NEEDED.
!
                nqrs2u = nqrs2u + 1_i4b
!
                if ( nqrs2u>maxqrs2u ) then
!
                    maxqrs2u = maxqrs2u + 5_i4b
!
                    call realloc( debu, maxqrs2u, iok )
!
                    if ( iok==0 ) then
!
                        call realloc( finu, maxqrs2u, iok  )
!
                        if ( iok==0 ) then
!
                            call realloc( csu, n, maxqrs2u, iok  )
!
                            if ( iok==0 ) then
                                call realloc( snu, n, maxqrs2u, iok  )
                            end if
!
                        end if
!
                    end if
!
                    if ( iok/=0 ) then
                        call merror( name_proc//allocate_error )
                    end if
!
                end if
!
                if ( deflate ) then
!
                    if ( use_bisect ) then
!
!                       COMPUTE THE NEXT SINGULAR VALUE BY A BISECTION ALGORITHM.
!
                        call bd_one_singval_bisect( d(l:k), e(l:k), g, fail )
!
                    else if ( use_dqds ) then
!
!                       COMPUTE THE NEXT SINGULAR VALUE BY THE DQDS ALGORITHM.
!
                        call bd_one_singval_dqds( d(l:k), e(l:k), g, fail )
!
!                        call bd_one_singval_lasq1( d2(l:k), e2(l:k), g, fail )
!
                    else
!
!                       COMPUTE THE NEXT SINGULAR VALUE BY THE PAL-WALKER-KAHAN ALGORITHM.
!
                        call bd_one_singval_pwk( d(l:k), e(l:k), g, fail )
!
                    end if
!
                    deflate = .not.fail
!
!                   NOW COMPUTE THE SINGULAR VECTORS.
!
                    if ( deflate ) then
!
!                       ONE SINGULAR VALUE FOUND.
!
!                       TRY FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION.
!
!                        call dflgen_bd( d(l:k), e(l+1_i4b:k), g,                        &
!                                        csu(l:k-1_i4b,nqrs2u), snu(l:k-1_i4b,nqrs2u),   &
!                                        csv(l:k-1_i4b,nqrs2v), snv(l:k-1_i4b,nqrs2v)    )
!
!                        call dflapp_bd( d(l:k), e(l+1_i4b:k),                                 &
!                                        csu(l:k-1_i4b,nqrs2u), snu(l:k-1_i4b,nqrs2u),         &
!                                        csv(l:k-1_i4b,nqrs2v), snv(l:k-1_i4b,nqrs2v), deflate )
!
                        call dflgen2_bd( d(l:k), e(l+1_i4b:k), g,                              &
                                         csu(l:k-1_i4b,nqrs2u), snu(l:k-1_i4b,nqrs2u),         &
                                         csv(l:k-1_i4b,nqrs2v), snv(l:k-1_i4b,nqrs2v), deflate )
!
                        if ( deflate ) then
!
#ifdef _STAT
                            iter_deflate_singval  = iter_deflate_singval + 1
#endif
!
                            e(l) = zero
!
                            debv(nqrs2v) = l
                            finv(nqrs2v) = k - 1_i4b
!
                            debu(nqrs2u) = l
                            finu(nqrs2u) = k - 1_i4b
!
!                           GODUNOV DEFLATION SUCCESSFUL, EXIT.
!
                            exit
!
                        end if
!
                    end if
!
                end if
!
!               IF GODUNOV DEFLATION IS NOT SUCCESSFUL, USE IMPLICIT QR STEPS.
!
!               COMPUTE WILKINSON SHIFT FROM BOTTOM 2 BY 2 MINOR OF B'*B .
!
                x = d(l)
                z = d(k)
                y = d(k-1_i4b)
                g = e(k-1_i4b)
                h = e(k)
                f = ( (y-z)*(y+z) + (g-h)*(g+h) )/(two*h*y)
!
                if ( abs(f)>one ) then
!
                    t = one/f
                    g = f*sqrt( one + t*t )
!
                else
!
                    g = sign( sqrt( one + f*f ), f )
!
                end if
!
                f = ( (x-z)*(x+z) + h*( (y/( f + g )) - h ) )/x
!
!               NEXT QR SWEEP.
!
                cs = one
                sn = one
!
                do i = l + 1_i4b, k
!
                    j = i - 1_i4b
                    g = e(i)
                    y = d(i)
                    h = sn*g
                    g = cs*g
!
!                   DEFINE THE GIVENS ROTATION.
!
                    if ( abs(f)>abs(h) ) then
!
                        t    = h/f
                        sn   = sign( sqrt(one+t*t), f )
                        e(j) = f*sn
                        cs   = one/sn
                        sn   = t*cs
!
                    else if ( h/=zero ) then
!
                        t    = f/h
                        cs   = sign( sqrt(one+t*t), h )
                        e(j) = h*cs
                        sn   = one/cs
                        cs   = t*sn
!
                    else
!
                        cs   = zero
                        sn   = one
                        e(j) = zero
!
                    end if
!
!                   APPLY THE GIVENS ROTATION.
!
                    f = x*cs + g*sn
                    g = -x*sn + g*cs
                    h = y*sn
                    y = y*cs
!
!                   STORE THE GIVENS ROTATION.
!
                    csv(j,nqrs2v) = cs
                    snv(j,nqrs2v) = sn
!
!                   DEFINE THE GIVENS ROTATION.
!
                    if ( abs(f)>abs(h) ) then
!
                        t    = h/f
                        sn   = sign( sqrt(one+t*t), f )
                        d(j) = f*sn
                        cs   = one/sn
                        sn   = t*cs
!
                    else if ( h/=zero ) then
!
                        t    = f/h
                        cs   = sign( sqrt(one+t*t), h )
                        d(j) = h*cs
                        sn   = one/cs
                        cs   = t*sn
!
                    else
!
                        cs   = zero
                        sn   = one
                        d(j) = zero
!
                    end if
!
!                   APPLY THE GIVENS ROTATION.
!
                    f =  cs*g + sn*y
                    x = -sn*g + cs*y
!
!                   STORE THE GIVENS ROTATION.
!
                    csu(j,nqrs2u) = cs
                    snu(j,nqrs2u) = sn
!
                end do
!
                e(l) = zero
                e(k) = f
                d(k) = x
!
                debv(nqrs2v) = l
                finv(nqrs2v) = k - 1_i4b
!
                debu(nqrs2u) = l
                finu(nqrs2u) = k - 1_i4b
!
!               RETURN TO TEST FOR RANK DEFICIENCIES AND SPLITTING.    
!
            end do
!
            if ( nqrs2v>=nfrancis .or. k==1_i4b .or. small  ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS
!               TO THE RIGHT SINGULAR VECTORS.
!
                qstart = maxval( debv(:nqrs2v) ) + nqrs2v - 1_i4b
                qend   = minval( finv(:nqrs2v) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2v==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2v
!
                        i1 = debv(j)
                        i2 = finv(j)
!
#ifdef _BLAS
                        do i = i1, i2
!
                            call rot( nrv, v(:nrv,i), 1_i4b, v(:nrv,i+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                        end do
#else
                        v_tmp(:nrv) = v(:nrv,i1)
!
                        do i = i1, i2
!
                            cs = csv(i,j)
                            sn = snv(i,j)
!
                            v(:nrv,i)   =  cs*v_tmp(:nrv) + sn*v(:nrv,i+1_i4b)
                            v_tmp(:nrv) = -sn*v_tmp(:nrv) + cs*v(:nrv,i+1_i4b)
!
                        end do
!
                        v(:nrv,i) = v_tmp(:nrv)
#endif
!
                    end do
!
                else if ( nqrs2v>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    l  = 0_i4b
                    i2 = qstart - nqrs2v + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        l = l + 1_i4b
!
                        i1 = debv(l)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
!
                                call rot( nrv, v(:nrv,i), 1_i4b, v(:nrv,i+1_i4b), 1_i4b, csv(i,l), snv(i,l) )
!
                            end do
#else
                            v_tmp(:nrv) = v(:nrv,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                cs = csv(i,l)
                                sn = snv(i,l)
!
                                v(:nrv,i)   =  cs*v_tmp(:nrv) + sn*v(:nrv,i+1_i4b)
                                v_tmp(:nrv) = -sn*v_tmp(:nrv) + cs*v(:nrv,i+1_i4b)
!
                            end do
!
                            v(:nrv,i) = v_tmp(:nrv)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, l
!
                            call rot( nrv, v(:nrv,i+1_i4b), 1_i4b, v(:nrv,i), 1_i4b, csv(i,j), -snv(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        v_tmp(:nrv) = v(:nrv,i+1_i4b)
!
                        do j = 1_i4b, l
!
                            cs = csv(i,j)
                            sn = snv(i,j)
!
                            v(:nrv,i+1_i4b) = -sn*v(:nrv,i) + cs*v_tmp(:nrv)
                            v_tmp(:nrv)     =  cs*v(:nrv,i) + sn*v_tmp(:nrv)
!
                            i = i - 1_i4b
!
                        end do
!
                        v(:nrv,i+1_i4b) = v_tmp(:nrv)
#endif
!
                    end do
!
!                   PIPELINE PHASE.
!
                    blk = 2_i4b*nqrs2v
!
                    if ( nrv>thr_bd_svd_blas3*blk ) then
!
!                       USE A BLAS3 ALGORITHM IF THE NUMBER OF ELEMENTS IN RIGHT
!                       SINGULAR VECTORS IS SIGNIFICANTLY LARGER THAN THE BLOCKSIZE.
!
                        if ( blk>blkuv ) then
!
!                           REALLOCATE WORKSPACE TO STORE THE ACCUMULATED GIVENS
!                           ROTATIONS AND UPDATE THE RIGHT SINGULAR VECTORS.
!
                            blkuv = 2_i4b*maxqrs2v
!
                            if ( allocated(uv) ) then
!
                                deallocate( uv, uvq )
!
                            end if
!
                            allocate( uv(nuv,blkuv), uvq(blkuv,blkuv), stat = iok )
!
                            if ( iok/=0 ) then
                                call merror( name_proc//allocate_error )
                            end if
!
                        end if
!
                        l   = qend - qstart + 1_i4b
                        l3  = l/nqrs2v
!
                        if ( l/=l3*nqrs2v ) l3 = l3 + 1_i4b
!
                        l1  = qstart - blk + 1_i4b
                        l2  = qstart
!
                        do l = 1_i4b, l3
!
                            l1  = l1 + nqrs2v
                            l2  = min( l2+nqrs2v, qend+1_i4b )
!
                            blk = l2 - l1 + 1_i4b
!
!                           ACCUMULATE GIVENS ROTATIONS OF THE CURRENT SET
!                           IN MATRIX uvq .
!
                            do j = 1_i4b, blk
!
                                uvq(:blkuv,j) = zero
                                uvq(j,j)      = one
!
                            end do
!
                            i1 = nqrs2v - 1_i4b
!
#ifdef _BLAS
                            do q = l1+nqrs2v-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2v
!
                                    call rot( blk, uvq(:blk,i2), 1_i4b, uvq(:blk,i2+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE RIGHT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
!                            if ( blk==2_i4b*nqrs2v ) then
!
!                               PROCESS THE REGULAR SETS.
!
!                                uv(:nrv,nqrs2v+1_i4b:blk) = v(:nrv,l1:l2-nqrs2v)
!
!                                call trmm( 'R', 'L', 'N', 'N', nrv, nqrs2v, one, uvq(:blkuv,nqrs2v+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,nqrs2v+1_i4b:blk), nuv )
!
!                                uvq(:nqrs2v,nqrs2v+1_i4b:blk) = uvq(nqrs2v+1_i4b:blk,nqrs2v+1_i4b:blk)
!
!                                call gemm( 'N', 'N', nrv, nqrs2v, nqrs2v, one, v(:nrv,l1+nqrs2v:l2), nrv,            &
!                                           uvq(:blkuv,nqrs2v+1_i4b:blk), blkuv, one, uv(:nuv,nqrs2v+1_i4b:blk), nuv  )
!
!                                uv(:nrv,:nqrs2v) = v(:nrv,l1+nqrs2v:l2)
!
!                                uvq(:nqrs2v,nqrs2v+1_i4b:blk) = uvq(nqrs2v+1_i4b:blk,:nqrs2v)
!
!                                call trmm( 'R', 'U', 'N', 'N', nrv, nqrs2v, one, uvq(:blkuv,nqrs2v+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,:nqrs2v), nuv )
!
!                                call gemm( 'N', 'N', nrv,  nqrs2v, nqrs2v, one, v(:nrv,l1:l2-nqrs2v), nrv, &
!                                           uvq(:blkuv,:nqrs2v), blkuv, one, uv(:nuv,:nqrs2v), nuv )
!
!                            else
!
!                               PROCESS THE RESIDUAL SET IF REQUIRED.
!
                                call gemm( 'N', 'N', nrv, blk, blk, one, v(:nrv,l1:l2), nrv,  &
                                           uvq(:blkuv,:blk), blkuv, zero, uv(:nuv,:blk), nuv  )
!
!                            end if
#else
                            do q = l1+nqrs2v-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2v
!
                                    cs = csv(i,j)
                                    sn = snv(i,j)
!
                                    v_tmp(:blk)        = cs*uvq(:blk,i2) + sn*uvq(:blk,i2+1_i4b)
                                    uvq(:blk,i2+1_i4b) = cs*uvq(:blk,i2+1_i4b) - sn*uvq(:blk,i2)
                                    uvq(:blk,i2)       = v_tmp(:blk)
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE RIGHT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
                            uv(:nrv,:blk) = matmul( v(:nrv,l1:l2), uvq(:blk,:blk) )
#endif
!
                            v(:nrv,l1:l2) = uv(:nrv,:blk)
!
                        end do
!
                    else
!
!                       USE A BLAS1 ALGORITHM OTHERWISE.
!
                        do q = qstart, qend
!
                            i = q
!
#ifdef _BLAS
                            do j = 1_i4b, nqrs2v
!
                                call rot( nrv, v(:nrv,i), 1_i4b, v(:nrv,i+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                                i = i - 1_i4b
!
                            end do
#else
                            v_tmp(:nrv) = v(:nrv,i+1_i4b)
!
                            do j = 1_i4b, nqrs2v
!
                                cs = csv(i,j)
                                sn = snv(i,j)
!
                                v(:nrv,i+1_i4b) = -sn*v(:nrv,i) + cs*v_tmp(:nrv)
                                v_tmp(:nrv)     =  cs*v(:nrv,i) + sn*v_tmp(:nrv)
!
                                i = i - 1_i4b
!
                            end do
!
                            v(:nrv,i+1_i4b) = v_tmp(:nrv)
#endif
!
                        end do
!
                    end if
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2v
!
                        i2 = finv(j)
!
                        if ( i1<=i2 ) then
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( nrv, v(:nrv,i), 1_i4b, v(:nrv,i+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                            end do
#else
                            v_tmp(:nrv) = v(:nrv,i1)
!
                            do i = i1, i2
!
                                cs = csv(i,j)
                                sn = snv(i,j)
!
                                v(:nrv,i)   =  cs*v_tmp(:nrv) + sn*v(:nrv,i+1_i4b)
                                v_tmp(:nrv) = -sn*v_tmp(:nrv) + cs*v(:nrv,i+1_i4b)
!
                            end do
!
                            v(:nrv,i) = v_tmp(:nrv)
#endif
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
                nqrs2v = 0_i4b
!
                if ( small ) then
!
                    do i = k - 1_i4b, 1_i4b, -1_i4b
!
!                       APPLY PREVIOUS GIVENS ROTATIONS TO TRANSFORM e(k) TO ZERO
!                       TO THE RIGHT SINGULAR VECTORS.
!
#ifdef _BLAS
                        call rot( nrv, v(:nrv,i), 1_i4b, v(:nrv,k), 1_i4b, cs2(i), sn2(i) )
#else
                        cs = cs2(i)
                        sn = sn2(i)                             
!
                        v_tmp(:nrv) = -sn*v(:nrv,i)
                        v(:nrv,i)   =  cs*v(:nrv,i)  + sn*v(:nrv,k)
                        v(:nrv,k)   =  v_tmp(:nrv)   + cs*v(:nrv,k)
#endif
!
                    end do
!
                end if
!
            end if
!
            if ( nqrs2u>=nfrancis .or. k==1_i4b  .or. deficient ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS
!               TO THE LEFT SINGULAR VECTORS.
!
                qstart = maxval( debu(:nqrs2u) ) + nqrs2u - 1_i4b
                qend   = minval( finu(:nqrs2u) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2u==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2u
!
                        i1 = debu(j)
                        i2 = finu(j)
!
#ifdef _BLAS
                        do i = i1, i2
!
                            call rot( nru, u(:nru,i), 1_i4b, u(:nru,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                        end do
#else
                        u_tmp(:nru) = u(:nru,i1)
!
                        do i = i1, i2
!
                            cs = csu(i,j)
                            sn = snu(i,j)
!
                            u(:nru,i)   =  cs*u_tmp(:nru) + sn*u(:nru,i+1_i4b)
                            u_tmp(:nru) = -sn*u_tmp(:nru) + cs*u(:nru,i+1_i4b)
!
                        end do
!
                        u(:nru,i) = u_tmp(:nru)
#endif
!
                    end do
!
                else if ( nqrs2u>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    l  = 0_i4b
                    i2 = qstart - nqrs2u + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        l = l + 1_i4b
!
                        i1 = debu(l)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
!
                                call rot( nru, u(:nru,i), 1_i4b, u(:nru,i+1_i4b), 1_i4b, csu(i,l), snu(i,l) )
!
                            end do
#else
                            u_tmp(:nru) = u(:nru,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                cs = csu(i,l)
                                sn = snu(i,l)
!
                                u(:nru,i)   =  cs*u_tmp(:nru) + sn*u(:nru,i+1_i4b)
                                u_tmp(:nru) = -sn*u_tmp(:nru) + cs*u(:nru,i+1_i4b)
!
                            end do
!
                            u(:nru,i) = u_tmp(:nru)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, l
!
                            call rot( nru, u(:nru,i+1_i4b), 1_i4b, u(:nru,i), 1_i4b, csu(i,j), -snu(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        u_tmp(:nru) = u(:nru,i+1_i4b)
!
                        do j = 1_i4b, l
!
                            cs = csu(i,j)
                            sn = snu(i,j)
!
                            u(:nru,i+1_i4b) = -sn*u(:nru,i) + cs*u_tmp(:nru)
                            u_tmp(:nru)     =  cs*u(:nru,i) + sn*u_tmp(:nru)
!
                            i = i - 1_i4b
!
                        end do
!
                        u(:nru,i+1_i4b) = u_tmp(:nru)
#endif
!
                    end do
!
!                   PIPELINE PHASE.
!
                    blk = 2_i4b*nqrs2u
!
                    if ( nru>thr_bd_svd_blas3*blk ) then
!
!                       USE A BLAS3 ALGORITHM IF THE NUMBER OF ELEMENTS IN LEFT
!                       SINGULAR VECTORS IS SIGNIFICANTLY LARGER THAN THE BLOCKSIZE.
!
                        if ( blk>blkuv ) then
!
!                           REALLOCATE WORKSPACE TO STORE THE ACCUMULATED GIVENS
!                           ROTATIONS AND UPDATE THE LEFT SINGULAR VECTORS.
!
                            blkuv = 2_i4b*maxqrs2u
!
                            if ( allocated(uv) ) then
!
                                deallocate( uv, uvq )
!
                            end if
!
                            allocate( uv(nuv,blkuv), uvq(blkuv,blkuv), stat = iok )
!
                            if ( iok/=0 ) then
                                call merror( name_proc//allocate_error )
                            end if
!
                        end if
!
                        l   = qend - qstart + 1_i4b
                        l3  = l/nqrs2u
!
                        if ( l/=l3*nqrs2u ) l3 = l3 + 1_i4b
!
                        l1  = qstart - blk + 1_i4b
                        l2  = qstart
!
                        do l = 1_i4b, l3
!
                            l1  = l1 + nqrs2u
                            l2  = min( l2+nqrs2u, qend+1_i4b )
!
                            blk = l2 - l1 + 1_i4b
!
!                           ACCUMULATE GIVENS ROTATIONS OF THE CURRENT SET
!                           IN MATRIX uvq .
!
                            do j = 1_i4b, blk
!
                                uvq(:blkuv,j) = zero
                                uvq(j,j)      = one
!
                            end do
!
                            i1 = nqrs2u - 1_i4b
!
#ifdef _BLAS
                            do q = l1+nqrs2u-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2u
!
                                    call rot( blk, uvq(:blk,i2), 1_i4b, uvq(:blk,i2+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE LEFT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
!                            if ( blk==2_i4b*nqrs2u ) then
!
!                               PROCESS THE REGULAR SETS.
!
!                                uv(:nru,nqrs2u+1_i4b:blk) = u(:nru,l1:l2-nqrs2u)
!
!                                call trmm( 'R', 'L', 'N', 'N', nru, nqrs2u, one, uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,nqrs2u+1_i4b:blk), nuv )
!
!                                uvq(:nqrs2u,nqrs2u+1_i4b:blk) = uvq(nqrs2u+1_i4b:blk,nqrs2u+1_i4b:blk)
!
!                                call gemm( 'N', 'N', nru, nqrs2u, nqrs2u, one, u(:nru,l1+nqrs2u:l2), nru,        &
!                                           uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv, one, uv(:nuv,nqrs2u+1_i4b:blk), nuv  )
!
!                                uv(:nru,:nqrs2u) = u(:nru,l1+nqrs2u:l2)
!
!                                uvq(:nqrs2u,nqrs2u+1_i4b:blk) = uvq(nqrs2u+1_i4b:blk,:nqrs2u)
!
!                                call trmm( 'R', 'U', 'N', 'N', nru, nqrs2u, one, uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,:nqrs2u), nuv )
!
!                                call gemm( 'N', 'N', nru,  nqrs2u, nqrs2u, one, u(:nru,l1:l2-nqrs2u), nru, &
!                                           uvq(:blkuv,:nqrs2u), blkuv, one, uv(:nuv,:nqrs2u), nuv )
!
!                            else
!
!                               PROCESS THE RESIDUAL SET IF REQUIRED.
!
!
                                call gemm( 'N', 'N', nru, blk, blk, one, u(:nru,l1:l2), nru,   &
                                           uvq(:blkuv,:blk), blkuv, zero, uv(:nuv,:blk), nuv   )
!
!                            end if
#else
                            do q = l1+nqrs2u-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2u
!
                                    cs = csu(i,j)
                                    sn = snu(i,j)
!
                                    u_tmp(:blk)        = cs*uvq(:blk,i2) + sn*uvq(:blk,i2+1_i4b)
                                    uvq(:blk,i2+1_i4b) = cs*uvq(:blk,i2+1_i4b) - sn*uvq(:blk,i2)
                                    uvq(:blk,i2)       = u_tmp(:blk)
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE LEFT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
                            uv(:nru,:blk) = matmul( u(:nru,l1:l2), uvq(:blk,:blk) )
#endif
!
                            u(:nru,l1:l2) = uv(:nru,:blk)
!
                        end do
!
                    else
!
!                       USE A BLAS1 ALGORITHM OTHERWISE.
!
                        do q = qstart, qend
!
                            i = q
!
#ifdef _BLAS
                            do j = 1_i4b, nqrs2u
!
                                call rot( nru, u(:nru,i), 1_i4b, u(:nru,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                                i = i - 1_i4b
!
                            end do
#else
                            u_tmp(:nru) = u(:nru,i+1_i4b)
!
                            do j = 1_i4b, nqrs2u
!
                                cs = csu(i,j)
                                sn = snu(i,j)
!
                                u(:nru,i+1_i4b)  = -sn*u(:nru,i) + cs*u_tmp(:nru)
                                u_tmp(:nru)      =  cs*u(:nru,i) + sn*u_tmp(:nru)
!
                                i = i - 1_i4b
!
                            end do
!
                            u(:nru,i+1_i4b) = u_tmp(:nru)
#endif
!
                        end do
!
                    end if
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2u
!
                        i2 = finu(j)
!
                        if ( i1<=i2 ) then
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( nru, u(:nru,i), 1_i4b, u(:nru,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                            end do
#else
                            u_tmp(:nru) = u(:nru,i1)
!
                            do i = i1, i2
!
                                cs = csu(i,j)
                                sn = snu(i,j)
!
                                u(:nru,i)   =  cs*u_tmp(:nru) + sn*u(:nru,i+1_i4b)
                                u_tmp(:nru) = -sn*u_tmp(:nru) + cs*u(:nru,i+1_i4b)
!
                            end do
!
                            u(:nru,i) = u_tmp(:nru)
#endif
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
                nqrs2u = 0_i4b
!
            end if
!
            if ( deficient ) then
!
!               APPLY THE GIVENS ROTATIONS TO COPE WITH
!               RANK DEFICIENCIES.
!
                j = ii1 - 1_i4b
!
                do i = ii1, ii2
!
#ifdef _BLAS
                    call rot( nru, u(:nru,i), 1_i4b, u(:nru,j), 1_i4b, cs2(i), sn2(i) )
#else
                    cs = cs2(i)
                    sn = sn2(i)
!
                    u_tmp(:nru) = -sn*u(:nru,i)
                    u(:nru,i)   =  cs*u(:nru,i) + sn*u(:nru,j)
                    u(:nru,j)   =  u_tmp(:nru)  + cs*u(:nru,j)
#endif
!
                end do
!
            else
!
                k = k - 1_i4b
!
            end if
!
        end do
!
!       CONVERGENCE, MAKE SINGULAR VALUES POSITIVE AND
!       CHANGE SIGN OF LEFT SINGULAR VECTORS.
!
        do i = 1_i4b, n
!
            if ( d(i)>=zero ) cycle
!
            d(i)      = -d(i)
            u(:nru,i) = -u(:nru,i)
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        if ( allocated(uv) ) then
!
            deallocate( uv, uvq )
!
        end if
!
#ifdef _OPENMP2
    end if
#endif
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( csv, snv, csu, snu, debv, finv, debu, finu )
!
!   ORDER SINGULAR VALUES AND VECTORS IF REQUIRED.
!
    if ( present(sort) )  call lr_singvec_sort( sort, d, u, v )
!
#ifdef _STAT
    write(*,*)
    write(*,*) name_proc//" :"
    write(*,*) repeat('*', len(name_proc) )
    write(*,*)
    write(*,*) "perfect_shift = ", use_perfect_shift
    write(*,*) "bisect        = ", use_bisect
    write(*,*) "dqds          = ", use_dqds
    write(*,*)
    write(*,*) "iter_deflate_singval  = ", iter_deflate_singval
    write(*,*)
#endif
!
!
! END OF SUBROUTINE bd_svd_a
! __________________________
!
    end subroutine bd_svd_a
!
! =========================================================================================
!
    subroutine bd_svd2_a( upper, d, e, failure, u, vt, sort, maxiter,    &
                          max_francis_steps, perfect_shift, bisect, dqds )
!
! Purpose
! _______
!
!   BD_SVD2 computes the singular value decomposition (SVD) of a real
!   n-by-n (upper or lower) bidiagonal matrix B:
!
!                            B = Q * S * P'
!
!   , where S is a diagonal matrix with non-negative diagonal
!   elements (the singular values of B), and, Q and P are
!   orthogonal matrices (P' denotes the transpose of P).
!
!   The routine computes S, U * Q, and P' * VT, for given real input 
!   matrices U, VT.
!
!
! Arguments
! _________
!
!   UPPER               (INPUT) logical(lgl)
!                       On entry, if:
!
!                       - UPPER = true  : B is upper bidiagonal ;
!                       - UPPER = false : B is lower bidiagonal.
!
!   D                   (INPUT/OUTPUT) real(stnd), dimension(:)
!                       On entry, D contains the diagonal elements of the bidiagonal
!                       matrix B.
!
!                       On exit,  D contains the singular values of B.
!
!   E                   (INPUT/OUTPUT) real(stnd), dimension(:)
!                       On entry, E contains the off-diagonal elements of the bidiagonal
!                       matrix whose SVD is desired. E(1) is arbitrary.
!
!                       On exit, E is destroyed.
!
!                       The size of E must verify: size( E ) = size( D ) = n . 
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit ;
!                       - FAILURE = true  :   indicates that the algorithm did not converge and
!                         that full accuracy was not attained in the bidiagonal
!                         SVD of B.
!
!   U                   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the matrix U.
!
!                       On exit, U is overwritten by U * Q.
!
!                       The shape of U must verify:
!
!                       - size( U, 1 ) > 0 ;
!                       - size( U, 2 ) = size( D ) = n .
!
!   VT                  (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the matrix VT.
!
!                       On exit, VT is overwritten by P' * VT.
!
!                       The shape of VT must verify:
!
!                       - size( VT, 1 ) = size( D ) = n ;
!                       - size( VT, 2 ) > 0 .
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the singular values into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd'.
!                       The singular vectors are rearranged accordingly.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       algorithm. The algorithm fails to converge if the number 
!                       of QR sweeps exceeds MAXITER * n. Convergence usually 
!                       occurs in about 2 * n QR sweeps.
!
!                       The default is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the implicit QR algorithm. MAX_FRANCIS_STEPS
!                       is a strictly positive integer, otherwise the default value is used.
!
!                       The default is the minimum of n and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm applied to the associated n-by-n symmetric tridiagonal
!                       matrix B * B'  whose eigenvalues are the squares of the singular values of B.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm applied to the associated n-by-n symmetric tridiagonal
!                       matrix B * B'  whose eigenvalues are the squares of the singular values of B.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   If arguments U and VT are n-by-n identity matrices, on exit they are replaced
!   by Q and  P', respectively.
!
!   This subroutine is adapted from subroutine QRBD given in the reference (1),
!   with modifications suggested in the references (2) and (3) for the application
!   of a set of Givens rotations to the singular vectors, and extensions to the bidiagonal
!   case of the perfect shift strategy presented in the references (4) and (5) for the
!   tridiagonal case.
!
!   Furthermore, the computation of the singular vectors is parallelized 
!   if OPENMP is used.
!
!   Note, finally, that the bidiagonal matrix is not scaled before computing the singular
!   values and vectors. If some of the elements of the bidiagonal matrix are very small or
!   large, it may be appropriate to scale the bidiagonal matrix before calling BD_SVD2. 
!
!   For further details, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Lang, B., 1998:
!           Using level 3 BLAS in rotation-based algorithms.
!           Siam J. Sci. Comput., Vol. 19, 626-634.
!
!   (3) Van Zee, F.G., Van de Geijn, R., and Quintana-Orti, G., 2011:
!           Restructuring the QR Algorithm for High-Performance Application of Givens Rotations.
!           FLAME Working Note 60. The University of Texas at Austin, Department of Computer
!           Sciences. Technical Report TR-11-36.
!
!   (4) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, 1013-1034.
!
!   (5) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,        only : allocate_error, svd_error11
    use Utilities,             only : assert_eq, merror, iminloc
#ifdef _F2003
    use Utilities,             only : realloc => mvalloc
#else
    use Utilities_With_Pnter,  only : realloc
#endif
#ifdef _BLAS
    use BLAS_interfaces,       only : rot, gemm
!    use BLAS_interfaces,       only : rot, gemm, trmm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)    :: d, e
    real(stnd), intent(inout), dimension(:,:)  :: u, vt
!
    integer(i4b), intent(in), optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: perfect_shift, bisect, dqds
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: eps, cs, sn, f, z, x, y, g, h, t
#ifndef _BLAS
    real(stnd), dimension( size(u,1) )      :: u_tmp
    real(stnd), dimension( size(vt,2) )     :: v_tmp
#endif
    real(stnd), dimension( size(d) )        :: cs2, sn2
    real(stnd), dimension(:,:), allocatable :: v, uv, uvq
#ifdef _F2003
    real(stnd), dimension(:,:), allocatable :: csv, snv, csu, snu
#else
    real(stnd), dimension(:,:), pointer     :: csv, snv, csu, snu
#endif
!
#ifdef _STAT
    integer                                 :: iter_deflate_singval
#endif
    integer                                 :: iok
    integer(i4b)                            :: n, nru, ncvt, maxqrs, nqrs, i, j, k,          &
                                               nfrancis, maxqrs2v ,maxqrs2u, nqrs2v, nqrs2u, &
                                               blkuv, nuv, blk, l, l1, l2, l3,               &
                                               i1, i2, ii1, ii2, q, qstart, qend
#ifdef _F2003
    integer(i4b), dimension(:), allocatable :: debv, debu, finv, finu
#else
    integer(i4b), dimension(:), pointer     :: debv, debu, finv, finu
#endif
!
    logical                                 :: deficient, small, deflate, fail, use_perfect_shift, &
                                               use_bisect, use_dqds
!
#ifdef _OPENMP2
    real(stnd), dimension(:),   allocatable :: d2, e2
    real(stnd), dimension(:,:), allocatable :: w
!
    integer(i4b)                            :: j1, j2, nrw, itv, itu, jrem, joffset,                 &
                                               num_threads, thread_num, num_threads_v, num_threads_u
!
    logical                                 :: test_par, ompparallel, ompnested, comp_v
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_svd2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b)  ,    &
                   int(size(e),i4b),      &
                   int(size(u,2),i4b),    &
                   int(size(vt,1),i4b),   &
                   name_proc )
!
    nru  = size(  u, 1 )
    ncvt = size( vt, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( min(nru,ncvt)<=0_i4b ) then
!
        call merror( name_proc//svd_error11 )
!
    end if
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( d(1_i4b)<zero ) then
!
                d = -d
                u(:,:) = -u(:,:)
!
            end if
!
        end if
!
        return
!
    end if
!
    e(1_i4b) = zero
!
!   DETERMINE THE TOLERANCES FOR CONVERGENCE IN QR ITERATIONS.
!
    eps = fudge_eps*macheps*maxval( abs(d(:n)) + abs(e(:n)) )
!
!   SET OPTIONAL ARGUMENTS.
!
    if ( present(maxiter) ) then
        maxqrs = n*merge( maxiter, 10_i4b, maxiter>=1_i4b )
    else
        maxqrs = n*10_i4b
    end if
!    
    if ( present(max_francis_steps) ) then
        nfrancis = merge( max_francis_steps,                                  &
                          max_francis_steps_default,                          &
                          max_francis_steps>=1_i4b .and. max_francis_steps<=n )
    else
        nfrancis = max_francis_steps_default
    end if
!
    use_perfect_shift = true
!
    if ( present(perfect_shift) ) then
        use_perfect_shift = perfect_shift
    end if
!
    use_bisect = false
!
    if ( present(bisect) ) then
        use_bisect = bisect
    end if
!
    use_dqds = false
!
    if ( present(dqds) ) then
        use_dqds = dqds
    end if
!
!   INITIALIZE COUNTERS AND THRESHOLDS FOR QR ITERATIONS.
!
    nqrs  = 0_i4b
!
    nqrs2v    = 0_i4b
    nqrs2u    = 0_i4b
    maxqrs2v  = nfrancis + 5_i4b
    maxqrs2u  = nfrancis + 5_i4b
!
!   ALLOCATE WORK VARIABLES NEEDED TO STORE THE GIVENS ROTATIONS.
!
    allocate( csv(n,maxqrs2v),      &
              snv(n,maxqrs2v),      &
              csu(n,maxqrs2u),      &
              snu(n,maxqrs2u),      &
              debv(maxqrs2v),       &
              finv(maxqrs2v),       &
              debu(maxqrs2u),       &
              finu(maxqrs2u),       &
              stat = iok      )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    debu(:) = 1_i4b
    finu(:) = -1_i4b
!
    debv(:) = 1_i4b
    finv(:) = -1_i4b
!
#ifdef _STAT
    iter_deflate_singval  = 0
#endif
!
!   IF MATRIX IS LOWER BIDIAGONAL, ROTATE TO BE UPPER BIDIAGONAL
!   BY APPLYING GIVENS ROTATIONS ON THE LEFT.
!
    if ( .not.upper ) then

        do l = 2_i4b, n
!
            i = l - 1_i4b
!
!           DEFINE THE GIVENS ROTATION.
!
            if ( abs(d(i))>abs(e(l)) ) then
!
                t    = e(l)/d(i)
                sn   = sign( sqrt(one+t*t), d(i) )
                d(i) = d(i)*sn
                cs   = one/sn
                sn   = t*cs
!
            else if ( e(l)/=zero ) then
!
                t    = d(i)/e(l)
                cs   = sign( sqrt(one+t*t), e(l) )
                d(i) = e(l)*cs
                sn   = one/cs
                cs   = t*sn
!
            else
!
                cs = zero
                sn = one
!
            end if
!
!           APPLY THE GIVENS ROTATION.
!
            e(l) = sn*d(l)
            d(l) = cs*d(l)
!
!           STORE THE GIVENS ROTATION.
!
            csu(i,1_i4b) = cs
            snu(i,1_i4b) = sn
!                           
        end do
!
        nqrs2u = 1_i4b
!
        debu(1_i4b) = 1_i4b
        finu(1_i4b) = n - 1_i4b
!
    end if
!
#ifdef _OPENMP2
!
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel   .or. ompnested         ) .and.      &
                  ( (ncvt+nru)*n>=omp_limit   )                 .and.      &
                  ( (ncvt+nru)/i>=omp_limit2  )                 .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
#ifdef _BLAS
!$OMP PARALLEL PRIVATE(deflate,fail,comp_v,k,i,i1,i2,itv,itu,j,j1,j2,q,                    &
!$OMP                  qstart,qend,jrem,joffset,nrw,blkuv,nuv,blk,iok,l,l1,l2,l3,          &
!$OMP                  num_threads,thread_num,num_threads_v,num_threads_u,                 &
!$OMP                  cs,sn,f,t,z,x,y,h,g,d2,e2,w,uv,uvq),                                &
!$OMP          SHARED(upper,failure,deficient,small,use_perfect_shift,use_bisect,use_dqds, &
!$OMP                 ii1,ii2,nqrs,nqrs2u,maxqrs2u,nqrs2v,maxqrs2v,ncvt,nru,               &
!$OMP                 debu,finu,debv,finv,csv,snv,csu,snu,d,e)
#else
!$OMP PARALLEL PRIVATE(deflate,fail,comp_v,k,i,i1,i2,itv,itu,j,j1,j2,q,                    &
!$OMP                  qstart,qend,jrem,joffset,nrw,blkuv,nuv,blk,iok,l,l1,l2,l3,          &
!$OMP                  num_threads,thread_num,num_threads_v,num_threads_u,                 &
!$OMP                  cs,sn,f,t,z,x,y,h,g,d2,e2,v_tmp,u_tmp,w,uv,uvq),                    &
!$OMP          SHARED(upper,failure,deficient,small,use_perfect_shift,use_bisect,use_dqds, &
!$OMP                 ii1,ii2,nqrs,nqrs2u,maxqrs2u,nqrs2v,maxqrs2v,ncvt,nru,               &
!$OMP                 debu,finu,debv,finv,csv,snv,csu,snu,d,e)
#endif
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        thread_num  = omp_get_thread_num()
        num_threads = omp_get_num_threads()
!
        num_threads_u = ( nru*num_threads)/( nru + ncvt )
!
        if ( num_threads_u==0_i4b ) num_threads_u = 1_i4b
!
        num_threads_v = num_threads - num_threads_u
!
        comp_v = thread_num<=num_threads_v-1_i4b
!
!       HALF OF THE THREADS WILL COMPUTE THE LEFT SINGULAR VECTORS
!       AND THE OTHER HALF WILL COMPUTE THE RIGHT SINGULAR VECTORS.
!
        if ( comp_v ) then        
!
            k = num_threads_v
            j = thread_num
            l = ncvt
!
        else
!
            k = num_threads_u
            j = thread_num - num_threads_v
            l = nru
!
        end if
!
        jrem = mod( l, k )
        nrw  = ( l - jrem )/k
!
        if ( j<jrem ) then
            nrw     = nrw + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = j*nrw + joffset
        j2 = j1 + nrw - 1_i4b
!
!       ALLOCATE WORK VARIABLE TO STORE A LOCAL COPY OF THE
!       LEFT AND RIGHT SINGULAR MATRICES.
!
        allocate( w(nrw,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A LOCAL COPY OF THE INPUT MATRICES.
!
        if ( comp_v ) then        
!
            w(:nrw,:n) = transpose( vt(:n,j1:j2) )
!
        else
!
            w(:nrw,:n) = u(j1:j2,:n)
!
        end if
!
        nuv   = nrw
        blkuv = 0_i4b
!
        if ( thread_num==0_i4b ) then
!
            allocate( d2(n), e2(n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           MAKE A LOCAL COPY OF THE BIDIAGONAL MATRIX.
!
            d2(:n) = d(:n)
            e2(:n) = e(:n)
!
        end if
!
!       DIAGONALISATION OF THE BIDIAGONAL FORM : LOOP OVER SINGULAR VALUES
!       AND OVER ALLOWED ITERATIONS maxqrs .
!
        k = n
!
        do
!
            if ( k<1_i4b ) exit
!
            if ( thread_num==0_i4b ) then
!
                itv = nqrs2v
                itu = nqrs2u
!
                deflate = use_perfect_shift
!
                do
!
                    small     = false
                    deficient = false
!
!                   FIRST MAKE TEST FOR LAST DIAGONAL TERM d(k) BEING SMALL.
!
                    if ( abs( d2(k) )<=eps ) then
!
!                       SINCE d(k) IS SMALL WE WILL MAKE A SPECIAL PASS TO
!                       TRANSFORM e(k) TO ZERO.
!    
                        f     = e2(k)
                        e2(k) = zero
!
                        do i = k - 1_i4b, 1_i4b, -1_i4b
!
!                           CONSTRUCT AND APPLY TRANSFORMATION TO ZERO POSITION (i,k).
!
                            if ( abs(d2(i))>abs(f) ) then
!
                                t     = f/d2(i)
                                sn    = sign( sqrt(one+t*t), d2(i) )
                                d2(i) = d2(i)*sn
                                cs    = one/sn
                                sn    = t*cs
!
                            else if ( f/=zero ) then
!
                                t     = d2(i)/f
                                cs    = sign( sqrt(one+t*t), f )
                                d2(i) = f*cs
                                sn    = one/cs
                                cs    = t*sn
!
                            else
!
                                cs = zero
                                sn = one
!
                            end if
!
!                           APPLY THE GIVENS ROTATION.
!
                            f     = -sn*e2(i)
                            e2(i) =  cs*e2(i)
!
!                           STORE THE GIVENS ROTATION.
!
                            cs2(i) = cs
                            sn2(i) = sn                                
!
                        end do
!
                        small = true
!
!                       THE MATRIX IS NOW BIDIAGONAL, AND OF LOWER ORDER SINCE e(k) = zero .
!
                        exit
!
                    end if    
!
!                   TEST FOR SPLITTING OR RANK DEFICIENCIES.
!
                    do l = k, 1_i4b, -1_i4b
!
                        if ( abs( e2(l) )<=eps ) exit
!
!                       NOTE THAT THIS LOOP CANT COMPLETE SINCE e(1) = zero .
!
                        j = l - 1_i4b
!
                        if ( abs( d2(j) )<=eps ) then
!
                            deficient = true
!
!                           CANCELLATION OF e(l) , IF l > 1 .
!
                            cs = zero
                            sn = -one
!
                            do i = l, k
!
                                f     = -sn*e2(i)
                                e2(i) =  cs*e2(i)
!
                                if ( abs(f)<=eps ) exit
!
!                               DEFINE THE GIVENS ROTATION.
!
                                if ( abs(d2(i))>abs(f) ) then
!
                                    t     = f/d2(i)
                                    sn    = sign( sqrt(one+t*t), d2(i) )
                                    d2(i) = d2(i)*sn
                                    cs    = one/sn
                                    sn    = t*cs
!
                                else if ( f/=zero ) then
!
                                    t     = d2(i)/f
                                    cs    = sign( sqrt(one+t*t), f )
                                    d2(i) = f*cs
                                    sn    = one/cs
                                    cs    = t*sn
!
                                else
!
                                    cs = zero
                                    sn = one
!
                                end if
!
!                               STORE THE GIVENS ROTATION.
!
                                cs2(i) = cs
                                sn2(i) = sn                                
!
                            end do
!
                            ii1 = l
                            ii2 = i - 1_i4b
!
                            exit
!
                        end if
!
                    end do
!
!                   TEST FOR RANK DEFICIENCIES.
!
                    if ( deficient ) exit
!
!                   CUTOFF FOR CONVERGENCE FAILURE.
!                   nqrs WILL BE AT MOST 2*n USUALLY .
!
                    if ( nqrs>=maxqrs ) then
!
!                       IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM    
!                       IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
                        if ( l/=k ) then
!
!                           SET FAILURE INDICATOR.
!
                            failure = true
!
                            l     = iminloc( abs(e2(l+1_i4b:k))) + l
                            e2(l) = zero
                            nqrs  = 0_i4b
!
                        end if
!
                    end if
!
!                   TEST FOR CONVERGENCE.
!
                    if ( l==k ) exit
!
                    nqrs = nqrs + 1_i4b
!
!                   REALLOCATE WORKSPACE TO STORE THE FRANCIS SETS
!                   OF GIVENS ROTATIONS FOR THE RIGHT SINGULAR VECTORS
!                   IF NEEDED.
!
                    itv = itv + 1_i4b
!
                    if ( itv>maxqrs2v ) then
!
                        maxqrs2v = maxqrs2v + 5_i4b
!
                        call realloc( debv, maxqrs2v, iok )
!
                        if ( iok==0 ) then
!
                            call realloc( finv, maxqrs2v, iok  )
!
                            if ( iok==0 ) then
!
                                call realloc( csv, n, maxqrs2v, iok  )
!
                                if ( iok==0 ) then
                                    call realloc( snv, n, maxqrs2v, iok  )
                                end if
!
                            end if
!
                        end if
!
                        if ( iok/=0 ) then
                            call merror( name_proc//allocate_error )
                        end if
!
                    end if
!
!                   REALLOCATE WORKSPACE TO STORE THE FRANCIS SETS
!                   OF GIVENS ROTATIONS FOR THE LEFT SINGULAR VECTORS
!                   IF NEEDED.
!
                    itu = itu + 1_i4b
!
                    if ( itu>maxqrs2u ) then
!
                        maxqrs2u = maxqrs2u + 5_i4b
!
                        call realloc( debu, maxqrs2u, iok )
!
                        if ( iok==0 ) then
!
                            call realloc( finu, maxqrs2u, iok  )
!
                            if ( iok==0 ) then
!
                                call realloc( csu, n, maxqrs2u, iok  )
!
                                if ( iok==0 ) then
                                    call realloc( snu, n, maxqrs2u, iok  )
                                end if
!
                            end if
!
                        end if
!
                        if ( iok/=0 ) then
                            call merror( name_proc//allocate_error )
                        end if
!
                    end if
!
                    if ( deflate ) then
!
                        if ( use_bisect ) then
!
!                           COMPUTE THE NEXT SINGULAR VALUE BY A BISECTION ALGORITHM.
!
                            call bd_one_singval_bisect( d2(l:k), e2(l:k), g, fail )
!
                        else if ( use_dqds ) then
!
!                           COMPUTE THE NEXT SINGULAR VALUE BY THE DQDS ALGORITHM.
!
                            call bd_one_singval_dqds( d2(l:k), e2(l:k), g, fail )
!
!                            call bd_one_singval_lasq1( d2(l:k), e2(l:k), g, fail )
!
                        else
!
!                           COMPUTE THE NEXT SINGULAR VALUE BY THE PAL-WALKER-KAHAN ALGORITHM.
!
                            call bd_one_singval_pwk( d2(l:k), e2(l:k), g, fail )
!
                        end if
!
                        deflate = .not.fail
!
!                       NOW COMPUTE THE SINGULAR VECTORS.
!
                        if ( deflate ) then
!
!                           ONE SINGULAR VALUE FOUND.
!
!                           TRY FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION.
!
!                            call dflgen_bd( d2(l:k), e2(l+1_i4b:k), g,                &
!                                            csu(l:k-1_i4b,itu), snu(l:k-1_i4b,itu),   &
!                                            csv(l:k-1_i4b,itv), snv(l:k-1_i4b,itv)    )
!
!                            call dflapp_bd( d2(l:k), e2(l+1_i4b:k),                         &
!                                            csu(l:k-1_i4b,itu), snu(l:k-1_i4b,itu),         &
!                                            csv(l:k-1_i4b,itv), snv(l:k-1_i4b,itv), deflate )
!
                            call dflgen2_bd( d2(l:k), e2(l+1_i4b:k), g,                      &
                                             csu(l:k-1_i4b,itu), snu(l:k-1_i4b,itu),         &
                                             csv(l:k-1_i4b,itv), snv(l:k-1_i4b,itv), deflate )
!
                            if ( deflate ) then
!
#ifdef _STAT
                                iter_deflate_singval  = iter_deflate_singval + 1
#endif
!
                                e2(l) = zero
!
                                debv(itv) = l
                                finv(itv) = k - 1_i4b
!
                                debu(itu) = l
                                finu(itu) = k - 1_i4b
!
!                               GODUNOV DEFLATION SUCCESSFUL, EXIT.
!
                                exit
!
                            end if
!
                        end if
!
                    end if
!
!                   IF GODUNOV DEFLATION IS NOT SUCCESSFUL, USE IMPLICIT QR STEPS.
!
!                   COMPUTE WILKINSON SHIFT FROM BOTTOM 2 BY 2 MINOR OF B'*B .
!
                    z = d2(k)
                    x = d2(l)
                    y = d2(k-1_i4b)
                    g = e2(k-1_i4b)
                    h = e2(k)
                    f = ( (y-z)*(y+z) + (g-h)*(g+h) )/(two*h*y)
!
                    if ( abs(f)>one ) then
!
                       t = one/f
                       g = f*sqrt( one + t*t )
!
                    else
!
                       g = sign( sqrt( one + f*f ), f )
!
                    end if
!
                    f = ( (x-z)*(x+z) + h*( (y/( f + g )) - h ) )/x
!
!                   NEXT QR SWEEP.
!
                    cs = one
                    sn = one
!
                    do i = l + 1_i4b, k
!
                        j = i - 1_i4b
                        g = e2(i)
                        y = d2(i)
                        h = sn*g
                        g = cs*g
!
!                       DEFINE THE GIVENS ROTATION.
!
                        if ( abs(f)>abs(h) ) then
!
                            t     = h/f
                            sn    = sign( sqrt(one+t*t), f )
                            e2(j) = f*sn
                            cs    = one/sn
                            sn    = t*cs
!
                        else if ( h/=zero ) then
!
                            t     = f/h
                            cs    = sign( sqrt(one+t*t), h )
                            e2(j) = h*cs
                            sn    = one/cs
                            cs    = t*sn
!
                        else
!
                            cs    = zero
                            sn    = one
                            e2(j) = zero
!
                        end if
!
!                       APPLY THE GIVENS ROTATION.
!
                        f = x*cs + g*sn
                        g = -x*sn + g*cs
                        h = y*sn
                        y = y*cs
!
!                       STORE THE GIVENS ROTATION.
!
                        csv(j,itv) = cs
                        snv(j,itv) = sn
!
!                       DEFINE THE GIVENS ROTATION.
!
                        if ( abs(f)>abs(h) ) then
!
                            t     = h/f
                            sn    = sign( sqrt(one+t*t), f )
                            d2(j) = f*sn
                            cs    = one/sn
                            sn    = t*cs
!
                        else if ( h/=zero ) then
!
                            t     = f/h
                            cs    = sign( sqrt(one+t*t), h )
                            d2(j) = h*cs
                            sn    = one/cs
                            cs    = t*sn
!
                        else
!
                            cs    = zero
                            sn    = one
                            d2(j) = zero
!
                        end if
!
!                       APPLY THE GIVENS ROTATION.
!
                        f =  cs*g + sn*y
                        x = -sn*g + cs*y
!
!                       STORE THE GIVENS ROTATION.
!
                        csu(j,itu) = cs
                        snu(j,itu) = sn
!
                    end do
!
                    e2(l) = zero
                    e2(k) = f
                    d2(k) = x
!
                    debv(itv) = l
                    finv(itv) = k - 1_i4b
!
                    debu(itu) = l
                    finu(itu) = k - 1_i4b
!
!                   RETURN TO TEST FOR RANK DEFICIENCIES AND SPLITTING.    
!
                end do
!
                nqrs2v = itv
                nqrs2u = itu
!
            end if
!
!$OMP BARRIER
!
            if ( comp_v .and. ( nqrs2v>=nfrancis .or. k==1_i4b .or. small )  ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS
!               TO THE RIGHT SINGULAR VECTORS.
!
                qstart = maxval( debv(:nqrs2v) ) + nqrs2v - 1_i4b
                qend   = minval( finv(:nqrs2v) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2v==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2v
!
                        i1 = debv(j)
                        i2 = finv(j)
!
#ifdef _BLAS
                        do i = i1, i2
!
                            call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                        end do
#else
                        v_tmp(:nrw) = w(:nrw,i1)
!
                        do i = i1, i2
!
                            cs = csv(i,j)
                            sn = snv(i,j)
!
                            w(:nrw,i)   =  cs*v_tmp(:nrw) + sn*w(:nrw,i+1_i4b)
                            v_tmp(:nrw) = -sn*v_tmp(:nrw) + cs*w(:nrw,i+1_i4b)
!
                        end do
!
                        w(:nrw,i) = v_tmp(:nrw)
#endif
!
                    end do
!
                else if ( nqrs2v>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    l  = 0_i4b
                    i2 = qstart - nqrs2v + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        l = l + 1_i4b
!
                        i1 = debv(l)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
!
                                call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csv(i,l), snv(i,l) )
!
                            end do
#else
                            v_tmp(:nrw) = w(:nrw,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                cs = csv(i,l)
                                sn = snv(i,l)
!
                                w(:nrw,i)   =  cs*v_tmp(:nrw) + sn*w(:nrw,i+1_i4b)
                                v_tmp(:nrw) = -sn*v_tmp(:nrw) + cs*w(:nrw,i+1_i4b)
!
                            end do
!
                            w(:nrw,i) = v_tmp(:nrw)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, l
!
                            call rot( nrw, w(:nrw,i+1_i4b), 1_i4b, w(:nrw,i), 1_i4b, csv(i,j), -snv(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        v_tmp(:nrw) = w(:nrw,i+1_i4b)
!
                        do j = 1_i4b, l
!
                            cs = csv(i,j)
                            sn = snv(i,j)
!
                            w(:nrw,i+1_i4b) = -sn*w(:nrw,i) + cs*v_tmp(:nrw)
                            v_tmp(:nrw)     =  cs*w(:nrw,i) + sn*v_tmp(:nrw)
!
                            i = i - 1_i4b
!
                        end do
!
                        w(:nrw,i+1_i4b) = v_tmp(:nrw)
#endif
!
                    end do
!
!                   PIPELINE PHASE.
!
                    blk = 2_i4b*nqrs2v
!
                    if ( nrw>thr_bd_svd_blas3*blk ) then
!
!                       USE A BLAS3 ALGORITHM IF THE NUMBER OF ELEMENTS IN RIGHT
!                       SINGULAR VECTORS IS SIGNIFICANTLY LARGER THAN THE BLOCKSIZE.
!
                        if ( blk>blkuv ) then
!
!                           REALLOCATE WORKSPACE TO STORE THE ACCUMULATED GIVENS
!                           ROTATIONS AND UPDATE THE RIGHT SINGULAR VECTORS.
!
                            blkuv = 2_i4b*maxqrs2v
!
                            if ( allocated(uv) ) then
!
                                deallocate( uv, uvq )
!
                            end if
!
                            allocate( uv(nuv,blkuv), uvq(blkuv,blkuv), stat = iok )
!
                            if ( iok/=0 ) then
                                call merror( name_proc//allocate_error )
                            end if
!
                        end if
!
                        l   = qend - qstart + 1_i4b
                        l3  = l/nqrs2v
!
                        if ( l/=l3*nqrs2v ) l3 = l3 + 1_i4b
!
                        l1  = qstart - blk + 1_i4b
                        l2  = qstart
!
                        do l = 1_i4b, l3
!
                            l1  = l1 + nqrs2v
                            l2  = min( l2+nqrs2v, qend+1_i4b )
!
                            blk = l2 - l1 + 1_i4b
!
!                           ACCUMULATE GIVENS ROTATIONS OF THE CURRENT SET
!                           IN MATRIX uvq .
!
                            do j = 1_i4b, blk
!
                                uvq(:blkuv,j) = zero
                                uvq(j,j)      = one
!
                            end do
!
                            i1 = nqrs2v - 1_i4b
!
#ifdef _BLAS
                            do q = l1+nqrs2v-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2v
!
                                    call rot( blk, uvq(:blk,i2), 1_i4b, uvq(:blk,i2+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE RIGHT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
!                            if ( blk==2_i4b*nqrs2v ) then
!
!                               PROCESS THE REGULAR SETS.
!
!                                uv(:nrw,nqrs2v+1_i4b:blk) = w(:nrw,l1:l2-nqrs2v)
!
!                                call trmm( 'R', 'L', 'N', 'N', nrw, nqrs2v, one, uvq(:blkuv,nqrs2v+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,nqrs2v+1_i4b:blk), nuv )
!
!                                uvq(:nqrs2v,nqrs2v+1_i4b:blk) = uvq(nqrs2v+1_i4b:blk,nqrs2v+1_i4b:blk)
!
!                                call gemm( 'N', 'N', nrw, nqrs2v, nqrs2v, one, w(:nrw,l1+nqrs2v:l2), nrw,            &
!                                           uvq(:blkuv,nqrs2v+1_i4b:blk), blkuv, one, uv(:nuv,nqrs2v+1_i4b:blk), nuv  )
!
!                                uv(:nrw,:nqrs2v) = w(:nrw,l1+nqrs2v:l2)
!
!                                uvq(:nqrs2v,nqrs2v+1_i4b:blk) = uvq(nqrs2v+1_i4b:blk,:nqrs2v)
!
!                                call trmm( 'R', 'U', 'N', 'N', nrw, nqrs2v, one, uvq(:blkuv,nqrs2v+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,:nqrs2v), nuv )
!
!                                call gemm( 'N', 'N', nrw,  nqrs2v, nqrs2v, one, w(:nrw,l1:l2-nqrs2v), nrw, &
!                                           uvq(:blkuv,:nqrs2v), blkuv, one, uv(:nuv,:nqrs2v), nuv )
!
!                            else
!
!                               PROCESS THE RESIDUAL SET IF REQUIRED.
!
                                call gemm( 'N', 'N', nrw, blk, blk, one, w(:nrw,l1:l2), nrw,  &
                                           uvq(:blkuv,:blk), blkuv, zero, uv(:nuv,:blk), nuv  )
!
!                            end if
!
#else
                            do q = l1+nqrs2v-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2v
!
                                    cs = csv(i,j)
                                    sn = snv(i,j)
!
                                    v_tmp(:blk)        = cs*uvq(:blk,i2) + sn*uvq(:blk,i2+1_i4b)
                                    uvq(:blk,i2+1_i4b) = cs*uvq(:blk,i2+1_i4b) - sn*uvq(:blk,i2)
                                    uvq(:blk,i2)       = v_tmp(:blk)
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE RIGHT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
                            uv(:nrw,:blk) = matmul( w(:nrw,l1:l2), uvq(:blk,:blk) )
#endif
!
                            w(:nrw,l1:l2) = uv(:nrw,:blk)
!
                        end do
!
                    else
!
!                       USE A BLAS1 ALGORITHM OTHERWISE.
!
                        do q = qstart, qend
!
                            i = q
!
#ifdef _BLAS
                            do j = 1_i4b, nqrs2v
!
                                call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                                i = i - 1_i4b
!
                            end do
#else
                            v_tmp(:nrw) = w(:nrw,i+1_i4b)
!
                            do j = 1_i4b, nqrs2v
!
                                cs = csv(i,j)
                                sn = snv(i,j)
!
                                w(:nrw,i+1_i4b) = -sn*w(:nrw,i) + cs*v_tmp(:nrw)
                                v_tmp(:nrw)     =  cs*w(:nrw,i) + sn*v_tmp(:nrw)
!
                                i = i - 1_i4b
!
                            end do
!
                            w(:nrw,i+1_i4b) = v_tmp(:nrw)
#endif
!
                        end do
!
                    end if
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2v
!
                        i2 = finv(j)
!
                        if ( i1<=i2 ) then
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                            end do
#else
                            v_tmp(:nrw) = w(:nrw,i1)
!
                            do i = i1, i2
!
                                cs = csv(i,j)
                                sn = snv(i,j)
!
                                w(:nrw,i)   =  cs*v_tmp(:nrw) + sn*w(:nrw,i+1_i4b)
                                v_tmp(:nrw) = -sn*v_tmp(:nrw) + cs*w(:nrw,i+1_i4b)
!
                            end do
!
                            w(:nrw,i) = v_tmp(:nrw)
#endif
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
            else if ( .not.comp_v .and. ( nqrs2u>=nfrancis .or. k==1_i4b  .or. deficient ) ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS
!               TO THE LEFT SINGULAR VECTORS.
!
                qstart = maxval( debu(:nqrs2u) ) + nqrs2u - 1_i4b
                qend   = minval( finu(:nqrs2u) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2u==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2u
!
                        i1 = debu(j)
                        i2 = finu(j)
!
#ifdef _BLAS
                        do i = i1, i2
!
                            call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                        end do
#else
                        u_tmp(:nrw) = w(:nrw,i1)
!
                        do i = i1, i2
!
                            cs = csu(i,j)
                            sn = snu(i,j)
!
                            w(:nrw,i)   =  cs*u_tmp(:nrw) + sn*w(:nrw,i+1_i4b)
                            u_tmp(:nrw) = -sn*u_tmp(:nrw) + cs*w(:nrw,i+1_i4b)
!
                        end do
!
                        w(:nrw,i) = u_tmp(:nrw)
#endif
!
                    end do
!
                else if ( nqrs2u>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    l  = 0_i4b
                    i2 = qstart - nqrs2u + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        l = l + 1_i4b
!
                        i1 = debu(l)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
!
                                call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csu(i,l), snu(i,l) )
!
                            end do
#else
                            u_tmp(:nrw) = w(:nrw,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                cs = csu(i,l)
                                sn = snu(i,l)
!
                                w(:nrw,i)   =  cs*u_tmp(:nrw) + sn*w(:nrw,i+1_i4b)
                                u_tmp(:nrw) = -sn*u_tmp(:nrw) + cs*w(:nrw,i+1_i4b)
!
                            end do
!
                            w(:nrw,i) = u_tmp(:nrw)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, l
!
                            call rot( nrw, w(:nrw,i+1_i4b), 1_i4b, w(:nrw,i), 1_i4b, csu(i,j), -snu(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        u_tmp(:nrw) = w(:nrw,i+1_i4b)
!
                        do j = 1_i4b, l
!
                            cs = csu(i,j)
                            sn = snu(i,j)
!
                            w(:nrw,i+1_i4b) = -sn*w(:nrw,i) + cs*u_tmp(:nrw)
                            u_tmp(:nrw)     =  cs*w(:nrw,i) + sn*u_tmp(:nrw)
!
                            i = i - 1_i4b
!
                        end do
!
                        w(:nrw,i+1_i4b) = u_tmp(:nrw)
#endif
!
                    end do
!
!                   PIPELINE PHASE.
!
                    blk = 2_i4b*nqrs2u
!
                    if ( nrw>thr_bd_svd_blas3*blk ) then
!
!                       USE A BLAS3 ALGORITHM IF THE NUMBER OF ELEMENTS IN LEFT
!                       SINGULAR VECTORS IS SIGNIFICANTLY LARGER THAN THE BLOCKSIZE.
!
                        if ( blk>blkuv ) then
!
!                           REALLOCATE WORKSPACE TO STORE THE ACCUMULATED GIVENS
!                           ROTATIONS AND UPDATE THE LEFT SINGULAR VECTORS.
!
                            blkuv = 2_i4b*maxqrs2u
!
                            if ( allocated(uv) ) then
!
                                deallocate( uv, uvq )
!
                            end if
!
                            allocate( uv(nuv,blkuv), uvq(blkuv,blkuv), stat = iok )
!
                            if ( iok/=0 ) then
                                call merror( name_proc//allocate_error )
                            end if
!
                        end if
!
                        l   = qend - qstart + 1_i4b
                        l3  = l/nqrs2u
!
                        if ( l/=l3*nqrs2u ) l3 = l3 + 1_i4b
!
                        l1  = qstart - blk + 1_i4b
                        l2  = qstart
!
                        do l = 1_i4b, l3
!
                            l1  = l1 + nqrs2u
                            l2  = min( l2+nqrs2u, qend+1_i4b )
!
                            blk = l2 - l1 + 1_i4b
!
!                           ACCUMULATE GIVENS ROTATIONS OF THE CURRENT SET
!                           IN MATRIX uvq .
!
                            do j = 1_i4b, blk
!
                                uvq(:blkuv,j) = zero
                                uvq(j,j)      = one
!
                            end do
!
                            i1 = nqrs2u - 1_i4b
!
#ifdef _BLAS
                            do q = l1+nqrs2u-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2u
!
                                    call rot( blk, uvq(:blk,i2), 1_i4b, uvq(:blk,i2+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE LEFT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
!                            if ( blk==2_i4b*nqrs2u ) then
!
!                               PROCESS THE REGULAR SETS.
!
!                                uv(:nrw,nqrs2u+1_i4b:blk) = w(:nrw,l1:l2-nqrs2u)
!
!                                call trmm( 'R', 'L', 'N', 'N', nrw, nqrs2u, one, uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,nqrs2u+1_i4b:blk), nuv )
!
!                                uvq(:nqrs2u,nqrs2u+1_i4b:blk) = uvq(nqrs2u+1_i4b:blk,nqrs2u+1_i4b:blk)
!
!                                call gemm( 'N', 'N', nrw, nqrs2u, nqrs2u, one, w(:nrw,l1+nqrs2u:l2), nrw,            &
!                                           uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv, one, uv(:nuv,nqrs2u+1_i4b:blk), nuv  )
!
!                                uv(:nrw,:nqrs2u) = w(:nrw,l1+nqrs2u:l2)
!
!                                uvq(:nqrs2u,nqrs2u+1_i4b:blk) = uvq(nqrs2u+1_i4b:blk,:nqrs2u)
!    
!                                call trmm( 'R', 'U', 'N', 'N', nrw, nqrs2u, one, uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,:nqrs2u), nuv )
!
!                                call gemm( 'N', 'N', nrw,  nqrs2u, nqrs2u, one, w(:nrw,l1:l2-nqrs2u), nrw, &
!                                           uvq(:blkuv,:nqrs2u), blkuv, one, uv(:nuv,:nqrs2u), nuv )
!
!                            else
!
!                               PROCESS THE RESIDUAL SET IF REQUIRED.
!
                                call gemm( 'N', 'N', nrw, blk, blk, one, w(:nrw,l1:l2), nrw,  &
                                           uvq(:blkuv,:blk), blkuv, zero, uv(:nuv,:blk), nuv  )
!
!                            end if
#else
                            do q = l1+nqrs2u-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2u
!
                                    cs = csu(i,j)
                                    sn = snu(i,j)
!
                                    u_tmp(:blk)        = cs*uvq(:blk,i2) + sn*uvq(:blk,i2+1_i4b)
                                    uvq(:blk,i2+1_i4b) = cs*uvq(:blk,i2+1_i4b) - sn*uvq(:blk,i2)
                                    uvq(:blk,i2)       = u_tmp(:blk)
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE LEFT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
                            uv(:nrw,:blk) = matmul( w(:nrw,l1:l2), uvq(:blk,:blk) )
#endif
!
                            w(:nrw,l1:l2) = uv(:nrw,:blk)
!
                        end do
!
                    else
!
!                       USE A BLAS1 ALGORITHM OTHERWISE.
!
                        do q = qstart, qend
!
                            i = q
!
#ifdef _BLAS
                            do j = 1_i4b, nqrs2u
!
                                call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                                i = i - 1_i4b
!
                            end do
#else
                            u_tmp(:nrw) = w(:nrw,i+1_i4b)
!
                            do j = 1_i4b, nqrs2u
!
                                cs = csu(i,j)
                                sn = snu(i,j)
!
                                w(:nrw,i+1_i4b) = -sn*w(:nrw,i) + cs*u_tmp(:nrw)
                                u_tmp(:nrw)     =  cs*w(:nrw,i) + sn*u_tmp(:nrw)
!
                                i = i - 1_i4b
!
                            end do
!
                            w(:nrw,i+1_i4b) = u_tmp(:nrw)
#endif
!
                        end do
!
                    end if
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2u
!
                        i2 = finu(j)
!
                        if ( i1<=i2 ) then
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                            end do
#else
                            u_tmp(:nrw) = w(:nrw,i1)
!
                            do i = i1, i2
!
                                cs = csu(i,j)
                                sn = snu(i,j)
!
                                w(:nrw,i)   =  cs*u_tmp(:nrw) + sn*w(:nrw,i+1_i4b)
                                u_tmp(:nrw) = -sn*u_tmp(:nrw) + cs*w(:nrw,i+1_i4b)
!
                            end do
!
                            w(:nrw,i) = u_tmp(:nrw)
#endif
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
            end if
!
!$OMP BARRIER
!
!$OMP MASTER
            if ( nqrs2v>=nfrancis .or. k==1_i4b .or. small  )    nqrs2v = 0_i4b
            if ( nqrs2u>=nfrancis .or. k==1_i4b .or. deficient ) nqrs2u = 0_i4b
!$OMP END MASTER
!
            if ( small .and. comp_v ) then        
!
                do i = k - 1_i4b, 1_i4b, -1_i4b
!
!                   APPLY PREVIOUS GIVENS ROTATIONS TO TRANSFORM e(k) TO ZERO
!                   TO THE RIGHT SINGULAR VECTORS.
!
#ifdef _BLAS
                    call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,k), 1_i4b, cs2(i), sn2(i) )
#else
                    cs = cs2(i)
                    sn = sn2(i)                             
!
                    v_tmp(:nrw) = -sn*w(:nrw,i)
                    w(:nrw,i)   =  cs*w(:nrw,i) + sn*w(:nrw,k)
                    w(:nrw,k)   =  v_tmp(:nrw)  + cs*w(:nrw,k)
#endif
!
                end do
!
            end if
!
            if ( deficient ) then
!
                if ( .not.comp_v ) then        
!
!                   APPLY THE GIVENS ROTATIONS TO COPE WITH
!                   RANK DEFICIENCIES TO THE LEFT SINGULAR VECTORS.
!
                    j = ii1 - 1_i4b
!
                    do i = ii1, ii2
!
#ifdef _BLAS
                        call rot( nrw, w(:nrw,i), 1_i4b, w(:nrw,j), 1_i4b, cs2(i), sn2(i) )
#else
                        cs = cs2(i)
                        sn = sn2(i)
!
                        u_tmp(:nrw) = -sn*w(:nrw,i)
                        w(:nrw,i)   =  cs*w(:nrw,i) + sn*w(:nrw,j)
                        w(:nrw,j)   =  u_tmp(:nrw)  + cs*w(:nrw,j)
#endif
!
                    end do
!
                end if
!
            else
!
                k = k - 1_i4b
!
            end if
!
!$OMP BARRIER
!
        end do
!
        if ( thread_num==0_i4b ) then
!
            d(:n) = d2(:n)
            e(:n) = e2(:n)
!
            deallocate( d2, e2 )
!
        end if
!
!$OMP BARRIER
!
!       SAVE THE COMPUTED SINGULAR VECTORS.
!
        if ( comp_v ) then        
!
            vt(:n,j1:j2) = transpose( w(:nrw,:n) )
!
        else
!
!           CONVERGENCE, MAKE SINGULAR VALUES POSITIVE AND
!           CHANGE SIGN OF LEFT SINGULAR VECTORS.
!
            do i = 1_i4b, n
!
                if ( d(i)>=zero ) cycle
!
                w(:nrw,i) = -w(:nrw,i)
!
            end do
!
            u(j1:j2,:n) = w(:nrw,:n)
!
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( w )
!
        if ( allocated(uv) ) then
!
            deallocate( uv, uvq )
!
        end if
!
!$OMP END PARALLEL
!
!       CONVERGENCE, MAKE SINGULAR VALUES POSITIVE.
!
        where ( d(1_i4b:n)<zero ) d(1_i4b:n) = -d(1_i4b:n)
!
    else
#endif
!
!       ALLOCATE WORK ARRAY TO STORE V .
!
        allocate( v(ncvt,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        v(:ncvt,:n) = transpose( vt(:n,:ncvt) )
!
        blkuv = 0_i4b
        nuv   = max( nru, ncvt ) 
!
!       DIAGONALISATION OF THE BIDIAGONAL FORM : LOOP OVER SINGULAR VALUES
!       AND OVER ALLOWED ITERATIONS maxqrs .
!
        k = n
!
        do
!
            if ( k<1_i4b ) exit
!
            deflate = use_perfect_shift
!
            do
!
                small     = false
                deficient = false
!
!               FIRST MAKE TEST FOR LAST DIAGONAL TERM d(k) BEING SMALL.
!
                if ( abs( d(k) )<=eps ) then
!
!                   SINCE d(k) IS SMALL WE WILL MAKE A SPECIAL PASS TO
!                   TRANSFORM e(k) TO ZERO.
!    
                    f    = e(k)
                    e(k) = zero
!
                    do i = k - 1_i4b, 1_i4b, -1_i4b
!
!                       CONSTRUCT AND APPLY TRANSFORMATION TO ZERO POSITION (i,k).
!
                        if ( abs(d(i))>abs(f) ) then
!
                            t    = f/d(i)
                            sn   = sign( sqrt(one+t*t), d(i) )
                            d(i) = d(i)*sn
                            cs   = one/sn
                            sn   = t*cs
!
                        else if ( f/=zero ) then
!
                            t    = d(i)/f
                            cs   = sign( sqrt(one+t*t), f )
                            d(i) = f*cs
                            sn   = one/cs
                            cs   = t*sn
!
                        else
!
                            cs = zero
                            sn = one
!
                        end if
!
!                       APPLY THE GIVENS ROTATION.
!
                        f    = -sn*e(i)
                        e(i) =  cs*e(i)
!
!                       STORE THE GIVENS ROTATION.
!
                        cs2(i) = cs
                        sn2(i) = sn                                
!
                    end do
!
                    small = true
!
!                   THE MATRIX IS NOW BIDIAGONAL, AND OF LOWER ORDER SINCE e(k) = zero .
!
                    exit
!
                end if    
!
!               TEST FOR SPLITTING OR RANK DEFICIENCIES.
!
                do l = k, 1_i4b, -1_i4b
!
                    if ( abs( e(l) )<=eps ) exit
!
!                   NOTE THAT THIS LOOP CANT COMPLETE SINCE e(1) = zero .
!
                    j = l - 1_i4b
                    if ( abs( d(j) )<=eps ) then
!
                        deficient = true
!
!                       CANCELLATION OF e(l) , IF l > 1 .
!
                        cs = zero
                        sn = -one
!
                        do i = l, k
!
                            f    = -sn*e(i)
                            e(i) =  cs*e(i)
!
                            if ( abs(f)<=eps ) exit
!
!                           DEFINE THE GIVENS ROTATION.
!
                            if ( abs(d(i))>abs(f) ) then
!
                                t    = f/d(i)
                                sn   = sign( sqrt(one+t*t), d(i) )
                                d(i) = d(i)*sn
                                cs   = one/sn
                                sn   = t*cs
!
                            else if ( f/=zero ) then
!
                                t    = d(i)/f
                                cs   = sign( sqrt(one+t*t), f )
                                d(i) = f*cs
                                sn   = one/cs
                                cs   = t*sn
!
                            else
!
                                cs = zero
                                sn = one
!
                            end if
!
!                           STORE THE GIVENS ROTATION.
!
                            cs2(i) = cs
                            sn2(i) = sn                                
!
                        end do
!
                        ii1 = l
                        ii2 = i - 1_i4b
!
                        exit
!
                    end if
!
                end do
!
!               TEST FOR RANK DEFICIENCIES.
!
                if ( deficient ) exit
!
!               CUTOFF FOR CONVERGENCE FAILURE.
!               nqrs WILL BE AT MOST 2*n USUALLY .
!
                if ( nqrs>=maxqrs ) then
!
!                   IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM    
!                   IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
                    if ( l/=k ) then
!
!                       SET FAILURE INDICATOR.
!
                        failure = true
!
                        l    = iminloc( abs(e(l+1_i4b:k))) + l
                        e(l) = zero
                        nqrs = 0_i4b
!
                    end if
!
                end if
!
!               TEST FOR CONVERGENCE.
!
                if ( l==k ) exit
!
                nqrs = nqrs + 1_i4b
!
!               REALLOCATE WORKSPACE TO STORE THE FRANCIS SETS
!               OF GIVENS ROTATIONS FOR THE RIGHT SINGULAR VECTORS
!               IF NEEDED.
!
                nqrs2v = nqrs2v + 1_i4b
!
                if ( nqrs2v>maxqrs2v ) then
!
                    maxqrs2v = maxqrs2v + 5_i4b
!
                    call realloc( debv, maxqrs2v, iok )
!
                    if ( iok==0 ) then
!
                        call realloc( finv, maxqrs2v, iok  )
!
                        if ( iok==0 ) then
!
                            call realloc( csv, n, maxqrs2v, iok  )
!
                            if ( iok==0 ) then
                                call realloc( snv, n, maxqrs2v, iok  )
                            end if
!
                        end if
!
                    end if
!
                    if ( iok/=0 ) then
                        call merror( name_proc//allocate_error )
                    end if
!
                end if
!
!               REALLOCATE WORKSPACE TO STORE THE FRANCIS SETS
!               OF GIVENS ROTATIONS FOR THE LEFT SINGULAR VECTORS
!               IF NEEDED.
!
                nqrs2u = nqrs2u + 1_i4b
!
                if ( nqrs2u>maxqrs2u ) then
!
                    maxqrs2u = maxqrs2u + 5_i4b
!
                    call realloc( debu, maxqrs2u, iok )
!
                    if ( iok==0 ) then
!
                        call realloc( finu, maxqrs2u, iok  )
!
                        if ( iok==0 ) then
!
                            call realloc( csu, n, maxqrs2u, iok  )
!
                            if ( iok==0 ) then
                                call realloc( snu, n, maxqrs2u, iok  )
                            end if
!
                        end if
!
                    end if
!
                    if ( iok/=0 ) then
                        call merror( name_proc//allocate_error )
                    end if
!
                end if
!
                if ( deflate ) then
!
                    if ( use_bisect ) then
!
!                       COMPUTE THE NEXT SINGULAR VALUE BY A BISECTION ALGORITHM.
!
                        call bd_one_singval_bisect( d(l:k), e(l:k), g, fail )
!
                    else if ( use_dqds ) then
!
!                       COMPUTE THE NEXT SINGULAR VALUE BY THE DQDS ALGORITHM.
!
                        call bd_one_singval_dqds( d(l:k), e(l:k), g, fail )
!
!                        call bd_one_singval_lasq1( d2(l:k), e2(l:k), g, fail )
!
                    else
!
!                       COMPUTE THE NEXT SINGULAR VALUE BY THE PAL-WALKER-KAHAN ALGORITHM.
!
                        call bd_one_singval_pwk( d(l:k), e(l:k), g, fail )
!
                    end if
!
                    deflate = .not.fail
!
!                   NOW COMPUTE THE SINGULAR VECTORS.
!
                    if ( deflate ) then
!
!                       ONE SINGULAR VALUE FOUND.
!
!                       TRY FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION.
!
!                        call dflgen_bd( d(l:k), e(l+1_i4b:k), g,                        &
!                                        csu(l:k-1_i4b,nqrs2u), snu(l:k-1_i4b,nqrs2u),   &
!                                        csv(l:k-1_i4b,nqrs2v), snv(l:k-1_i4b,nqrs2v)    )
!
!                        call dflapp_bd( d(l:k), e(l+1_i4b:k),                                 &
!                                        csu(l:k-1_i4b,nqrs2u), snu(l:k-1_i4b,nqrs2u),         &
!                                        csv(l:k-1_i4b,nqrs2v), snv(l:k-1_i4b,nqrs2v), deflate )
!
                        call dflgen2_bd( d(l:k), e(l+1_i4b:k), g,                              &
                                         csu(l:k-1_i4b,nqrs2u), snu(l:k-1_i4b,nqrs2u),         &
                                         csv(l:k-1_i4b,nqrs2v), snv(l:k-1_i4b,nqrs2v), deflate )
!
                        if ( deflate ) then
!
#ifdef _STAT
                            iter_deflate_singval  = iter_deflate_singval + 1
#endif
!
                            e(l) = zero
!
                            debv(nqrs2v) = l
                            finv(nqrs2v) = k - 1_i4b
!
                            debu(nqrs2u) = l
                            finu(nqrs2u) = k - 1_i4b
!
!                           GODUNOV DEFLATION SUCCESSFUL, EXIT.
!
                            exit
!
                        end if
!
                    end if
!
                end if
!
!               IF GODUNOV DEFLATION IS NOT SUCCESSFUL, USE IMPLICIT QR STEPS.
!
!               COMPUTE WILKINSON SHIFT FROM BOTTOM 2 BY 2 MINOR OF B'*B .
!
                z = d(k)
                x = d(l)
                y = d(k-1_i4b)
                g = e(k-1_i4b)
                h = e(k)
                f = ( (y-z)*(y+z) + (g-h)*(g+h) )/(two*h*y)
!
                if ( abs(f)>one ) then
!
                    t = one/f
                    g = f*sqrt( one + t*t )
!
                else
!
                    g = sign( sqrt( one + f*f ), f )
!
                end if
!
                f = ( (x-z)*(x+z) + h*( (y/( f + g )) - h ) )/x
!
!               NEXT QR SWEEP.
!
                cs = one
                sn = one
!
                do i = l + 1_i4b, k
!
                    j = i - 1_i4b
                    g = e(i)
                    y = d(i)
                    h = sn*g
                    g = cs*g
!
!                   DEFINE THE GIVENS ROTATION.
!
                    if ( abs(f)>abs(h) ) then
!
                        t    = h/f
                        sn   = sign( sqrt(one+t*t), f )
                        e(j) = f*sn
                        cs   = one/sn
                        sn   = t*cs
!
                    else if ( h/=zero ) then
!
                        t    = f/h
                        cs   = sign( sqrt(one+t*t), h )
                        e(j) = h*cs
                        sn   = one/cs
                        cs   = t*sn
!
                    else
!
                        cs   = zero
                        sn   = one
                        e(j) = zero
!
                    end if
!
!                   APPLY THE GIVENS ROTATION.
!
                    f = x*cs + g*sn
                    g = -x*sn + g*cs
                    h = y*sn
                    y = y*cs
!
!                   STORE THE GIVENS ROTATION.
!
                    csv(j,nqrs2v) = cs
                    snv(j,nqrs2v) = sn
!
!                   DEFINE THE GIVENS ROTATION.
!
                    if ( abs(f)>abs(h) ) then
!
                        t    = h/f
                        sn   = sign( sqrt(one+t*t), f )
                        d(j) = f*sn
                        cs   = one/sn
                        sn   = t*cs
!
                    else if ( h/=zero ) then
!
                        t    = f/h
                        cs   = sign( sqrt(one+t*t), h )
                        d(j) = h*cs
                        sn   = one/cs
                        cs   = t*sn
!
                    else
!
                        cs   = zero
                        sn   = one
                        d(j) = zero
!
                    end if
!
!                   APPLY THE GIVENS ROTATION.
!
                    f =  cs*g + sn*y
                    x = -sn*g + cs*y
!
!                   STORE THE GIVENS ROTATION.
!
                    csu(j,nqrs2u) = cs
                    snu(j,nqrs2u) = sn
!
                end do
!
                e(l) = zero
                e(k) = f
                d(k) = x
!
                debv(nqrs2v) = l
                finv(nqrs2v) = k - 1_i4b
!
                debu(nqrs2u) = l
                finu(nqrs2u) = k - 1_i4b
!
!               RETURN TO TEST FOR RANK DEFICIENCIES AND SPLITTING.    
!
            end do
!
            if ( nqrs2v>=nfrancis .or. k==1_i4b .or. small  ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS
!               TO THE RIGHT SINGULAR VECTORS.
!
                qstart = maxval( debv(:nqrs2v) ) + nqrs2v - 1_i4b
                qend   = minval( finv(:nqrs2v) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2v==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2v
!
                        i1 = debv(j)
                        i2 = finv(j)
!
#ifdef _BLAS
                        do i = i1, i2
                            call rot( ncvt, v(:ncvt,i), 1_i4b, v(:ncvt,i+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
                        end do
#else
                        v_tmp(:ncvt) = v(:ncvt,i1)
!
                        do i = i1, i2
!
                            cs = csv(i,j)
                            sn = snv(i,j)
!
                            v(:ncvt,i)   =  cs*v_tmp(:ncvt) + sn*v(:ncvt,i+1_i4b)
                            v_tmp(:ncvt) = -sn*v_tmp(:ncvt) + cs*v(:ncvt,i+1_i4b)
!
                        end do
!
                        v(:ncvt,i) = v_tmp(:ncvt)
#endif
!
                    end do
!
                else if ( nqrs2v>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    l  = 0_i4b
                    i2 = qstart - nqrs2v + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        l = l + 1_i4b
!
                        i1 = debv(l)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
                                call rot( ncvt, v(:ncvt,i), 1_i4b, v(:ncvt,i+1_i4b), 1_i4b, csv(i,l), snv(i,l) )
                            end do
#else
                            v_tmp(:ncvt) = v(:ncvt,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                cs = csv(i,l)
                                sn = snv(i,l)
!
                                v(:ncvt,i)   =  cs*v_tmp(:ncvt) + sn*v(:ncvt,i+1_i4b)
                                v_tmp(:ncvt) = -sn*v_tmp(:ncvt) + cs*v(:ncvt,i+1_i4b)
!
                            end do
!
                            v(:ncvt,i) = v_tmp(:ncvt)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, l
!
                            call rot( ncvt, v(:ncvt,i+1_i4b), 1_i4b, v(:ncvt,i), 1_i4b, csv(i,j), -snv(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        v_tmp(:ncvt) = v(:ncvt,i+1_i4b)
!
                        do j = 1_i4b, l
!
                            cs = csv(i,j)
                            sn = snv(i,j)
!
                            v(:ncvt,i+1_i4b) = -sn*v(:ncvt,i) + cs*v_tmp(:ncvt)
                            v_tmp(:ncvt)     =  cs*v(:ncvt,i) + sn*v_tmp(:ncvt)
!
                            i = i - 1_i4b
!
                        end do
!
                        v(:ncvt,i+1_i4b) = v_tmp(:ncvt)
#endif
!
                    end do
!
!                   PIPELINE PHASE.
!
                    blk = 2_i4b*nqrs2v
!
                    if ( ncvt>thr_bd_svd_blas3*blk ) then
!
!                       USE A BLAS3 ALGORITHM IF THE NUMBER OF ELEMENTS IN RIGHT
!                       SINGULAR VECTORS IS SIGNIFICANTLY LARGER THAN THE BLOCKSIZE.
!
                        if ( blk>blkuv ) then
!
!                           REALLOCATE WORKSPACE TO STORE THE ACCUMULATED GIVENS
!                           ROTATIONS AND UPDATE THE RIGHT SINGULAR VECTORS.
!
                            blkuv = 2_i4b*maxqrs2v
!
                            if ( allocated(uv) ) then
!
                                deallocate( uv, uvq )
!
                            end if
!
                            allocate( uv(nuv,blkuv), uvq(blkuv,blkuv), stat = iok )
!
                            if ( iok/=0 ) then
                                call merror( name_proc//allocate_error )
                            end if
!
                        end if
!
                        l   = qend - qstart + 1_i4b
                        l3  = l/nqrs2v
!
                        if ( l/=l3*nqrs2v ) l3 = l3 + 1_i4b
!
                        l1  = qstart - blk + 1_i4b
                        l2  = qstart
!
                        do l = 1_i4b, l3
!
                            l1  = l1 + nqrs2v
                            l2  = min( l2+nqrs2v, qend+1_i4b )
!
                            blk = l2 - l1 + 1_i4b
!
!                           ACCUMULATE GIVENS ROTATIONS OF THE CURRENT SET
!                           IN MATRIX uvq .
!
                            do j = 1_i4b, blk
!
                                uvq(:blkuv,j) = zero
                                uvq(j,j)      = one
!
                            end do
!
                            i1 = nqrs2v - 1_i4b
!
#ifdef _BLAS
                            do q = l1+nqrs2v-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2v
!
                                    call rot( blk, uvq(:blk,i2), 1_i4b, uvq(:blk,i2+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE RIGHT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
!                            if ( blk==2_i4b*nqrs2v ) then
!
!                               PROCESS THE REGULAR SETS.
!
!                                uv(:ncvt,nqrs2v+1_i4b:blk) = v(:ncvt,l1:l2-nqrs2v)
!
!                                call trmm( 'R', 'L', 'N', 'N', ncvt, nqrs2v, one, uvq(:blkuv,nqrs2v+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,nqrs2v+1_i4b:blk), nuv )
!
!                                uvq(:nqrs2v,nqrs2v+1_i4b:blk) = uvq(nqrs2v+1_i4b:blk,nqrs2v+1_i4b:blk)
!
!                                call gemm( 'N', 'N', ncvt, nqrs2v, nqrs2v, one, v(:ncvt,l1+nqrs2v:l2), ncvt,         &
!                                           uvq(:blkuv,nqrs2v+1_i4b:blk), blkuv, one, uv(:nuv,nqrs2v+1_i4b:blk), nuv  )
!
!                                uv(:ncvt,:nqrs2v) = v(:ncvt,l1+nqrs2v:l2)
!
!                                uvq(:nqrs2v,nqrs2v+1_i4b:blk) = uvq(nqrs2v+1_i4b:blk,:nqrs2v)
!
!                                call trmm( 'R', 'U', 'N', 'N', ncvt, nqrs2v, one, uvq(:blkuv,nqrs2v+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,:nqrs2v), nuv )
!
!                                call gemm( 'N', 'N', ncvt,  nqrs2v, nqrs2v, one, v(:ncvt,l1:l2-nqrs2v), ncvt,  &
!                                           uvq(:blkuv,:nqrs2v), blkuv, one, uv(:nuv,:nqrs2v), nuv )
!
!                            else
!
!                               PROCESS THE RESIDUAL SET IF REQUIRED.
!
                                call gemm( 'N', 'N', ncvt, blk, blk, one, v(:ncvt,l1:l2), ncvt, &
                                           uvq(:blkuv,:blk), blkuv, zero, uv(:nuv,:blk), nuv    )
!
!                            end if
#else
                            do q = l1+nqrs2v-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2v
!
                                    cs = csv(i,j)
                                    sn = snv(i,j)
!
                                    v_tmp(:blk)        = cs*uvq(:blk,i2) + sn*uvq(:blk,i2+1_i4b)
                                    uvq(:blk,i2+1_i4b) = cs*uvq(:blk,i2+1_i4b) - sn*uvq(:blk,i2)
                                    uvq(:blk,i2)       = v_tmp(:blk)
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE RIGHT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
                            uv(:ncvt,:blk) = matmul( v(:ncvt,l1:l2), uvq(:blk,:blk) )
#endif
!
                            v(:ncvt,l1:l2) = uv(:ncvt,:blk)
!
                        end do
!
                    else
!
!                       USE A BLAS1 ALGORITHM OTHERWISE.
!
                        do q = qstart, qend
!
                            i = q
!
#ifdef _BLAS
                            do j = 1_i4b, nqrs2v
!
                                call rot( ncvt, v(:ncvt,i), 1_i4b, v(:ncvt,i+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                                i = i - 1_i4b
!
                            end do
#else
                            v_tmp(:ncvt) = v(:ncvt,i+1_i4b)
!
                            do j = 1_i4b, nqrs2v
!
                                cs = csv(i,j)
                                sn = snv(i,j)
!
                                v(:ncvt,i+1_i4b) = -sn*v(:ncvt,i) + cs*v_tmp(:ncvt)
                                v_tmp(:ncvt)     =  cs*v(:ncvt,i) + sn*v_tmp(:ncvt)
!
                                i = i - 1_i4b
!
                            end do
!
                            v(:ncvt,i+1_i4b) = v_tmp(:ncvt)
#endif
!
                        end do
!
                    end if
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2v
!
                        i2 = finv(j)
!
                        if ( i1<=i2 ) then
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( ncvt, v(:ncvt,i), 1_i4b, v(:ncvt,i+1_i4b), 1_i4b, csv(i,j), snv(i,j) )
!
                            end do
#else
                            v_tmp(:ncvt) = v(:ncvt,i1)
!
                            do i = i1, i2
!
                                cs = csv(i,j)
                                sn = snv(i,j)
!
                                v(:ncvt,i)   =  cs*v_tmp(:ncvt) + sn*v(:ncvt,i+1_i4b)
                                v_tmp(:ncvt) = -sn*v_tmp(:ncvt) + cs*v(:ncvt,i+1_i4b)
!
                            end do
!
                            v(:ncvt,i) = v_tmp(:ncvt)
#endif
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
                nqrs2v = 0_i4b
!
                if ( small ) then
!
                    do i = k - 1_i4b, 1_i4b, -1_i4b
!
!                       APPLY PREVIOUS GIVENS ROTATIONS TO TRANSFORM e(k) TO ZERO
!                       TO THE RIGHT SINGULAR VECTORS.
!
#ifdef _BLAS
                        call rot( ncvt, v(:ncvt,i), 1_i4b, v(:ncvt,k), 1_i4b, cs2(i), sn2(i) )
#else
                        cs = cs2(i)
                        sn = sn2(i)                             
!
                        v_tmp(:ncvt) = -sn*v(:ncvt,i)
                        v(:ncvt,i)   =  cs*v(:ncvt,i) + sn*v(:ncvt,k)
                        v(:ncvt,k)   =  v_tmp(:ncvt)  + cs*v(:ncvt,k)
#endif
!
                    end do
!
                end if
!
            end if
!
            if ( nqrs2u>=nfrancis .or. k==1_i4b  .or. deficient ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS
!               TO THE LEFT SINGULAR VECTORS.
!
                qstart = maxval( debu(:nqrs2u) ) + nqrs2u - 1_i4b
                qend   = minval( finu(:nqrs2u) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2u==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2u
!
                        i1 = debu(j)
                        i2 = finu(j)
!
#ifdef _BLAS
                        do i = i1, i2
!
                            call rot( nru, u(:nru,i), 1_i4b, u(:nru,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                        end do
#else
                        u_tmp(:nru) = u(:nru,i1)
!
                        do i = i1, i2
!
                            cs = csu(i,j)
                            sn = snu(i,j)
!
                            u(:nru,i)   =  cs*u_tmp(:nru) + sn*u(:nru,i+1_i4b)
                            u_tmp(:nru) = -sn*u_tmp(:nru) + cs*u(:nru,i+1_i4b)
!
                        end do
!
                        u(:nru,i) = u_tmp(:nru)
#endif
!
                    end do
!
                else if ( nqrs2u>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    l  = 0_i4b
                    i2 = qstart - nqrs2u + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        l = l + 1_i4b
!
                        i1 = debu(l)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
!
                                call rot( nru, u(:nru,i), 1_i4b, u(:nru,i+1_i4b), 1_i4b, csu(i,l), snu(i,l) )
!
                            end do
#else
                            u_tmp(:nru) = u(:nru,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                cs = csu(i,l)
                                sn = snu(i,l)
!
                                u(:nru,i)   =  cs*u_tmp(:nru) + sn*u(:nru,i+1_i4b)
                                u_tmp(:nru) = -sn*u_tmp(:nru) + cs*u(:nru,i+1_i4b)
!
                            end do
!
                            u(:nru,i) = u_tmp(:nru)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, l
!
                            call rot( nru, u(:nru,i+1_i4b), 1_i4b, u(:nru,i), 1_i4b, csu(i,j), -snu(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        u_tmp(:nru) = u(:nru,i+1_i4b)
!
                        do j = 1_i4b, l
!
                            cs = csu(i,j)
                            sn = snu(i,j)
!
                            u(:nru,i+1_i4b) = -sn*u(:nru,i) + cs*u_tmp(:nru)
                            u_tmp(:nru)     =  cs*u(:nru,i) + sn*u_tmp(:nru)
!
                            i = i - 1_i4b
!
                        end do
!
                        u(:nru,i+1_i4b) = u_tmp(:nru)
#endif
!
                    end do
!
!                   PIPELINE PHASE.
!
                    blk = 2_i4b*nqrs2u
!
                    if ( nru>thr_bd_svd_blas3*blk ) then
!
!                       USE A BLAS3 ALGORITHM IF THE NUMBER OF ELEMENTS IN LEFT
!                       SINGULAR VECTORS IS SIGNIFICANTLY LARGER THAN THE BLOCKSIZE.
!
                        if ( blk>blkuv ) then
!
!                           REALLOCATE WORKSPACE TO STORE THE ACCUMULATED GIVENS
!                           ROTATIONS AND UPDATE THE LEFT SINGULAR VECTORS.
!
                            blkuv = 2_i4b*maxqrs2u
!
                            if ( allocated(uv) ) then
!
                                deallocate( uv, uvq )
!
                            end if
!
                            allocate( uv(nuv,blkuv), uvq(blkuv,blkuv), stat = iok )
!
                            if ( iok/=0 ) then
                                call merror( name_proc//allocate_error )
                            end if
!
                        end if
!
                        l   = qend - qstart + 1_i4b
                        l3  = l/nqrs2u
!
                        if ( l/=l3*nqrs2u ) l3 = l3 + 1_i4b
!
                        l1  = qstart - blk + 1_i4b
                        l2  = qstart
!
                        do l = 1_i4b, l3
!
                            l1  = l1 + nqrs2u
                            l2  = min( l2+nqrs2u, qend+1_i4b )
!
                            blk = l2 - l1 + 1_i4b
!
!                           ACCUMULATE GIVENS ROTATIONS OF THE CURRENT SET
!                           IN MATRIX uvq .
!
                            do j = 1_i4b, blk
!
                                uvq(:blkuv,j) = zero
                                uvq(j,j)      = one
!
                            end do
!
                            i1 = nqrs2u - 1_i4b
!
#ifdef _BLAS
                            do q = l1+nqrs2u-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2u
!
                                    call rot( blk, uvq(:blk,i2), 1_i4b, uvq(:blk,i2+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE LEFT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
!                            if ( blk==2_i4b*nqrs2u ) then
!
!                               PROCESS THE REGULAR SETS.
!
!                                uv(:nru,nqrs2u+1_i4b:blk) = u(:nru,l1:l2-nqrs2u)
!
!                                call trmm( 'R', 'L', 'N', 'N', nru, nqrs2u, one, uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,nqrs2u+1_i4b:blk), nuv )
!
!                                uvq(:nqrs2u,nqrs2u+1_i4b:blk) = uvq(nqrs2u+1_i4b:blk,nqrs2u+1_i4b:blk)
!
!                                call gemm( 'N', 'N', nru, nqrs2u, nqrs2u, one, u(:nru,l1+nqrs2u:l2), nru,        &
!                                           uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv, one, uv(:nuv,nqrs2u+1_i4b:blk), nuv  )
!
!                                uv(:nru,:nqrs2u) = u(:nru,l1+nqrs2u:l2)
!
!                                uvq(:nqrs2u,nqrs2u+1_i4b:blk) = uvq(nqrs2u+1_i4b:blk,:nqrs2u)
!
!                                call trmm( 'R', 'U', 'N', 'N', nru, nqrs2u, one, uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv,  &
!                                           uv(:nuv,:nqrs2u), nuv )
!
!                                call gemm( 'N', 'N', nru,  nqrs2u, nqrs2u, one, u(:nru,l1:l2-nqrs2u), nru, &
!                                           uvq(:blkuv,:nqrs2u), blkuv, one, uv(:nuv,:nqrs2u), nuv )
!
!                            else
!
!                               PROCESS THE RESIDUAL SET IF REQUIRED.
!
!
                                call gemm( 'N', 'N', nru, blk, blk, one, u(:nru,l1:l2), nru,   &
                                           uvq(:blkuv,:blk), blkuv, zero, uv(:nuv,:blk), nuv   )
!
!                            end if
#else
                            do q = l1+nqrs2u-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2u
!
                                    cs = csu(i,j)
                                    sn = snu(i,j)
!
                                    u_tmp(:blk)        = cs*uvq(:blk,i2) + sn*uvq(:blk,i2+1_i4b)
                                    uvq(:blk,i2+1_i4b) = cs*uvq(:blk,i2+1_i4b) - sn*uvq(:blk,i2)
                                    uvq(:blk,i2)       = u_tmp(:blk)
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE LEFT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
                            uv(:nru,:blk) = matmul( u(:nru,l1:l2), uvq(:blk,:blk) )
#endif
!
                            u(:nru,l1:l2) = uv(:nru,:blk)
!
                        end do
!
                    else
!
!                       USE A BLAS1 ALGORITHM OTHERWISE.
!
                        do q = qstart, qend
!
                            i = q
!
#ifdef _BLAS
                            do j = 1_i4b, nqrs2u
!
                                call rot( nru, u(:nru,i), 1_i4b, u(:nru,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                                i = i - 1_i4b
!
                            end do
#else
                            u_tmp(:nru) = u(:nru,i+1_i4b)
!
                            do j = 1_i4b, nqrs2u
!
                                cs = csu(i,j)
                                sn = snu(i,j)
!
                                u(:nru,i+1_i4b)  = -sn*u(:nru,i) + cs*u_tmp(:nru)
                                u_tmp(:nru)      =  cs*u(:nru,i) + sn*u_tmp(:nru)
!
                                i = i - 1_i4b
!
                            end do
!
                            u(:nru,i+1_i4b) = u_tmp(:nru)
#endif
!
                        end do
!
                    end if
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2u
!
                        i2 = finu(j)
!
                        if ( i1<=i2 ) then
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( nru, u(:nru,i), 1_i4b, u(:nru,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                            end do
#else
                            u_tmp(:nru) = u(:nru,i1)
!
                            do i = i1, i2
!
                                cs = csu(i,j)
                                sn = snu(i,j)
!
                                u(:nru,i)   =  cs*u_tmp(:nru) + sn*u(:nru,i+1_i4b)
                                u_tmp(:nru) = -sn*u_tmp(:nru) + cs*u(:nru,i+1_i4b)
!
                            end do
!
                            u(:nru,i) = u_tmp(:nru)
#endif
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
                nqrs2u = 0_i4b
!
            end if
!
            if ( deficient ) then
!
!               APPLY THE GIVENS ROTATIONS TO COPE WITH
!               RANK DEFICIENCIES.
!
                j = ii1 - 1_i4b
!
                do i = ii1, ii2
!
#ifdef _BLAS
                    call rot( nru, u(:nru,i), 1_i4b, u(:nru,j), 1_i4b, cs2(i), sn2(i) )
#else
                    cs = cs2(i)
                    sn = sn2(i)
!
                    u_tmp(:nru) = -sn*u(:nru,i)
                    u(:nru,i)   =  cs*u(:nru,i) + sn*u(:nru,j)
                    u(:nru,j)   =  u_tmp(:nru)  + cs*u(:nru,j)
#endif
!
                end do
!
            else
!
                k = k - 1_i4b
!
            end if
!
        end do
!
!       CONVERGENCE, MAKE SINGULAR VALUES POSITIVE AND
!       CHANGE SIGN OF LEFT SINGULAR VECTORS.
!
        do i = 1_i4b, n
!
            if ( d(i)>=zero ) cycle
!
            d(i)      = -d(i)
            u(:nru,i) = -u(:nru,i)
!
        end do
!
!       TRANSPOSE RIGHT SINGULAR VECTORS.
!
        vt(:n,:ncvt) = transpose( v(:ncvt,:n) )        
!
!       DEALLOCATE WORK ARRAYS.
!
        if ( allocated(uv) ) then
!
            deallocate( uv, uvq )
!
        end if
!
        deallocate( v )
!
#ifdef _OPENMP2
    end if
#endif
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( csv, snv, csu, snu, debv, finv, debu, finu )
!
!   ORDER SINGULAR VALUES AND VECTORS IF REQUIRED.
!
    if ( present(sort) )  call lr_singvec_sort2( sort, d, u, vt )
!
#ifdef _STAT
    write(*,*)
    write(*,*) name_proc//" :"
    write(*,*) repeat('*', len(name_proc) )
    write(*,*)
    write(*,*) "perfect_shift = ", use_perfect_shift
    write(*,*) "bisect        = ", use_bisect
    write(*,*) "dqds          = ", use_dqds
    write(*,*)
    write(*,*) "iter_deflate_singval  = ", iter_deflate_singval
    write(*,*)
#endif
!
!
! END OF SUBROUTINE bd_svd2_a
! ___________________________
!
    end subroutine bd_svd2_a
!
! =========================================================================================
!
    subroutine bd_svd_b( upper, d, e, failure, u, sort, maxiter,        &
                         max_francis_steps, perfect_shift, bisect, dqds )
!
! Purpose
! _______
!
!   BD_SVD computes the singular value decomposition (SVD) of a real
!   n-by-n (upper or lower) bidiagonal matrix B:
!
!                            B = Q * S * P'
!
!   , where S is a diagonal matrix with non-negative diagonal
!   elements (the singular values of B), and, Q and P are
!   orthogonal matrices (P' denotes the transpose of P).
!
!   The routine computes S and U * Q for a given real input matrix U.
!
!
! Arguments
! _________
!
!   UPPER               (INPUT) logical(lgl)
!                       On entry, if:
!
!                       - UPPER = true  : B is upper bidiagonal ;
!                       - UPPER = false : B is lower bidiagonal.
!
!   D                   (INPUT/OUTPUT) real(stnd), dimension(:)
!                       On entry, D contains the diagonal elements of the bidiagonal
!                       matrix B.
!
!                       On exit,  D contains the singular values of B.
!
!   E                   (INPUT/OUTPUT) real(stnd), dimension(:)
!                       On entry, E contains the off-diagonal elements of the bidiagonal
!                       matrix whose SVD is desired. E(1) is arbitrary.
!
!                       On exit, E is destroyed.
!
!                       The size of E must verify: size( E ) = size( D ) = n . 
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit ;
!                       - FAILURE = true  :   indicates that the algorithm did not converge and
!                         that full accuracy was not attained in the bidiagonal
!                         SVD of B.
!
!   U                   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the matrix U.
!
!                       On exit, U is overwritten by U * Q.
!
!                       The shape of U must verify:
!
!                       - size( U, 1 ) > 0 ;
!                       - size( U, 2 ) = size( D ) = n .
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the singular values into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd'.
!                       The singular vectors U are rearranged accordingly.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       algorithm. The algorithm fails to converge if the number 
!                       of QR sweeps exceeds MAXITER * n. Convergence usually 
!                       occurs in about 2 * n QR sweeps.
!
!                       The default is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the implicit QR algorithm. MAX_FRANCIS_STEPS
!                       is a strictly positive integer, otherwise the default value is used.
!
!                       The default is the minimum of n and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm applied to the associated n-by-n symmetric tridiagonal
!                       matrix B * B'  whose eigenvalues are the squares of the singular values of B.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm applied to the associated n-by-n symmetric tridiagonal
!                       matrix B * B'  whose eigenvalues are the squares of the singular values of B.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   If argument U is a n-by-n identity matrix, on exit it is replaced by Q.
!
!   This subroutine is adapted from subroutine QRBD given in the reference (1),
!   with modifications suggested in the references (2) and (3) for the application
!   of a set of Givens rotations to the singular vectors, and extensions to the bidiagonal
!   case of the perfect shift strategy presented in the references (4) and (5) for the
!   tridiagonal case.
!
!   Furthermore, the computation of the singular vectors is parallelized 
!   if OPENMP is used.
!
!   Note, finally, that the bidiagonal matrix is not scaled before computing the singular
!   values and vectors. If some of the elements of the bidiagonal matrix are very small or
!   large, it may be appropriate to scale the bidiagonal matrix before calling BD_SVD. 
!
!   For further details, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Lang, B., 1998:
!           Using level 3 BLAS in rotation-based algorithms.
!           Siam J. Sci. Comput., Vol. 19, 626-634.
!
!   (3) Van Zee, F.G., Van de Geijn, R., and Quintana-Orti, G., 2011:
!           Restructuring the QR Algorithm for High-Performance Application of Givens Rotations.
!           FLAME Working Note 60. The University of Texas at Austin, Department of Computer
!           Sciences. Technical Report TR-11-36.
!
!   (4) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, 1013-1034.
!
!   (5) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,        only : allocate_error, svd_error12
    use Utilities,             only : assert_eq, merror, iminloc
#ifdef _F2003
    use Utilities,             only : realloc => mvalloc
#else
    use Utilities_With_Pnter,  only : realloc
#endif
#ifdef _BLAS
    use BLAS_interfaces,       only : rot, gemm
!    use BLAS_interfaces,       only : rot, gemm, trmm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)    :: d, e
    real(stnd), intent(inout), dimension(:,:)  :: u
!
    integer(i4b), intent(in), optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: perfect_shift, bisect, dqds
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: eps, cs, sn, f, z, x, y, g, h, t
#ifndef _BLAS
    real(stnd), dimension( size(u,1) )      :: u_tmp
#endif
    real(stnd), dimension( size(d) )        :: cs2, sn2
    real(stnd), dimension(:), allocatable   :: diag, sub
    real(stnd), dimension(:,:), allocatable :: uv, uvq
#ifdef _F2003
    real(stnd), dimension(:,:), allocatable :: csu, snu
#else
    real(stnd), dimension(:,:), pointer     :: csu, snu
#endif
!
#ifdef _STAT
    integer                                 :: iter_deflate_singval
#endif
    integer                                 :: iok
    integer(i4b)                            :: n, nru, maxqrs, nqrs, i, j, k,                 &
                                               nfrancis, maxqrs2u, nqrs2u, i1, i2, ii1, ii2,  &
                                               blkuv, blk, l, l1, l2, l3, q, qstart, qend
#ifdef _F2003
    integer(i4b), dimension(:), allocatable :: debu, finu
#else
    integer(i4b), dimension(:), pointer     :: debu, finu
#endif
!
    logical                                 :: deficient, deflate, fail, use_perfect_shift, &
                                               use_bisect, use_dqds
!
#ifdef _OPENMP2
    real(stnd), dimension(:),   allocatable :: d2, e2
    real(stnd), dimension(:,:), allocatable :: u2
!
    integer(i4b)                            :: k1, k2, nru2, itu,                     &
                                               num_threads, thread_num, jrem, joffset
!
    logical                                 :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_svd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b)  ,    &
                   int(size(e),i4b),      &
                   int(size(u,2),i4b),    &
                   name_proc )
!
    nru = size( u, 1 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( nru<=0_i4b ) then
!
        call merror( name_proc//svd_error12 )
!
    end if
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( d(1_i4b)<zero ) then
!
                d = -d
                u(:,:) = -u(:,:)
!
            end if
!
        end if
!
        return
!
    end if
!
    e(1_i4b) = zero
!
!   DETERMINE THE TOLERANCES FOR CONVERGENCE IN QR ITERATIONS.
!
    eps = fudge_eps*macheps*maxval( abs(d(:n)) + abs(e(:n)) )
!
!   SET OPTIONAL ARGUMENTS.
!
    if ( present(maxiter) ) then
        maxqrs = n*merge( maxiter, 10_i4b, maxiter>=1_i4b )
    else
        maxqrs = n*10_i4b
    end if
!    
    if ( present(max_francis_steps) ) then
        nfrancis = merge( max_francis_steps,                                  &
                          max_francis_steps_default,                          &
                          max_francis_steps>=1_i4b .and. max_francis_steps<=n )
    else
        nfrancis = max_francis_steps_default
    end if
!
    use_perfect_shift = true
!
    if ( present(perfect_shift) ) then
        use_perfect_shift = perfect_shift
    end if
!
    use_bisect = false
!
    if ( present(bisect) ) then
        use_bisect = bisect
    end if
!
    use_dqds = false
!
    if ( present(dqds) ) then
        use_dqds = dqds
    end if
!
!   INITIALIZE COUNTERS AND THRESHOLDS FOR QR ITERATIONS.
!
    nqrs  = 0_i4b
!
    nqrs2u    = 0_i4b
    maxqrs2u  = nfrancis + 5_i4b
!
!   ALLOCATE WORK VARIABLES NEEDED TO STORE THE GIVENS ROTATIONS .
!
    allocate( csu(n,maxqrs2u),      &
              snu(n,maxqrs2u),      &
              debu(maxqrs2u),       &
              finu(maxqrs2u),       &
              stat = iok      )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    debu(:) = 1_i4b
    finu(:) = -1_i4b
!
#ifdef _STAT
    iter_deflate_singval  = 0
#endif
!
!   IF MATRIX IS LOWER BIDIAGONAL, ROTATE TO BE UPPER BIDIAGONAL
!   BY APPLYING GIVENS ROTATIONS ON THE LEFT.
!
    if ( .not.upper ) then

        do l = 2_i4b, n
!
            i = l - 1_i4b
!
!           DEFINE THE GIVENS ROTATION.
!
            if ( abs(d(i))>abs(e(l)) ) then
!
                t    = e(l)/d(i)
                sn   = sign( sqrt(one+t*t), d(i) )
                d(i) = d(i)*sn
                cs   = one/sn
                sn   = t*cs
!
            else if ( e(l)/=zero ) then
!
                t    = d(i)/e(l)
                cs   = sign( sqrt(one+t*t), e(l) )
                d(i) = e(l)*cs
                sn   = one/cs
                cs   = t*sn
!
            else
!
                cs = zero
                sn = one
!
            end if
!
!           APPLY THE GIVENS ROTATION.
!
            e(l) = sn*d(l)
            d(l) = cs*d(l)
!
!           STORE THE GIVENS ROTATION.
!
            csu(i,1_i4b) = cs
            snu(i,1_i4b) = sn
!                           
        end do
!
        nqrs2u = 1_i4b
!
        debu(1_i4b) = 1_i4b
        finu(1_i4b) = n - 1_i4b
!
    end if
!
#ifdef _OPENMP2
!
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested         ) .and.      &
                  nru*n>=omp_limit                            .and.      &
                  nru/i>=omp_limit2                           .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
#ifdef _BLAS
!$OMP PARALLEL PRIVATE(deflate,fail,k,k1,k2,i,i1,i2,itu,j,num_threads,thread_num,q,  &
!$OMP                  qstart,qend,jrem,joffset,nru2,blkuv,blk,iok,l,l1,l2,l3,       &
!$OMP                  cs,sn,f,t,z,x,y,h,g,d2,e2,diag,sub,u2,uv,uvq),                &
!$OMP          SHARED(upper,failure,deficient,use_perfect_shift,use_bisect,use_dqds, &
!$OMP                 ii1,ii2,nqrs,nqrs2u,maxqrs2u,nru,debu,finu,csu,snu,d,e)
#else
!$OMP PARALLEL PRIVATE(deflate,fail,k,k1,k2,i,i1,i2,itu,j,num_threads,thread_num,q,  &
!$OMP                  qstart,qend,jrem,joffset,nru2,blkuv,blk,iok,l,l1,l2,l3,       &
!$OMP                  cs,sn,f,t,z,x,y,h,g,d2,e2,diag,sub,u_tmp,u2,uv,uvq),          &
!$OMP          SHARED(upper,failure,deficient,use_perfect_shift,use_bisect,use_dqds, &
!$OMP                 ii1,ii2,nqrs,nqrs2u,maxqrs2u,nru,debu,finu,csu,snu,d,e)
#endif
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        thread_num  = omp_get_thread_num()
        num_threads = omp_get_num_threads()
!
        jrem = mod( nru, num_threads )
        nru2 = ( nru - jrem )/num_threads
!
        if ( thread_num<jrem ) then
            nru2    = nru2 + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        k1 = thread_num*nru2 + joffset
        k2 = k1 + nru2 - 1_i4b
!
        blkuv = 0_i4b
!
!       ALLOCATE WORK VARIABLES TO STORE A LOCAL COPY OF THE
!       INPUT MATRIX.
!
        allocate( u2(nru2,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A LOCAL COPY OF THE INPUT MATRIX.
!
        u2(:nru2,:n) = u(k1:k2,:n)
!
        if ( thread_num==0_i4b ) then
!
            allocate( d2(n), e2(n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           MAKE A LOCAL COPY OF THE BIDIAGONAL MATRIX.
!
            d2(:n) = d(:n)
            e2(:n) = e(:n)
!
            if ( use_perfect_shift ) then
!
!               ALLOCATE WORK ARRAYS.
!
                allocate( diag(n), sub(n), stat = iok )
!
                if ( iok/=0 ) then
                    call merror( name_proc//allocate_error )
                end if
!
            end if
!
        end if
!
!       DIAGONALISATION OF THE BIDIAGONAL FORM : LOOP OVER SINGULAR VALUES
!       AND OVER ALLOWED ITERATIONS maxqrs .
!
        k = n
!
        do
!
            if ( k<1_i4b ) exit
!
            if ( thread_num==0_i4b ) then
!
                itu = nqrs2u
!
                deflate = use_perfect_shift
!
                do
!
                    deficient = false
!
!                   FIRST MAKE TEST FOR LAST DIAGONAL TERM d(k) BEING SMALL.
!
                    if ( abs( d2(k) )<=eps ) then
!
!                       SINCE d(k) IS SMALL WE WILL MAKE A SPECIAL PASS TO
!                       TRANSFORM e(k) TO ZERO.
!    
                        f    = e2(k)
                        e2(k) = zero
!
                        do i = k - 1_i4b, 1_i4b, -1_i4b
!
!                           CONSTRUCT AND APPLY TRANSFORMATION TO ZERO POSITION (i,k).
!
                            if ( abs(d2(i))>abs(f) ) then
!
                                t     = f/d2(i)
                                sn    = sign( sqrt(one+t*t), d2(i) )
                                d2(i) = d2(i)*sn
                                cs    = one/sn
                                sn    = t*cs
!
                            else if ( f/=zero ) then
!
                                t     = d2(i)/f
                                cs    = sign( sqrt(one+t*t), f )
                                d2(i) = f*cs
                                sn    = one/cs
                                cs    = t*sn
!
                            else
!
                                cs = zero
                                sn = one
!
                            end if
!
!                           APPLY THE GIVENS ROTATION.
!
                            f    = -sn*e2(i)
                            e2(i) =  cs*e2(i)
!
                        end do
!
!                       THE MATRIX IS NOW BIDIAGONAL, AND OF LOWER ORDER SINCE e(k) = zero .
!
                        exit
!
                    end if    
!
!                   TEST FOR SPLITTING OR RANK DEFICIENCIES.
!
                    do l = k, 1_i4b, -1_i4b
!
                        if ( abs( e2(l) )<=eps ) exit
!
!                       NOTE THAT THIS LOOP CANT COMPLETE SINCE e(1) = zero .
!
                        j = l - 1_i4b
                        if ( abs( d2(j) )<=eps ) then
!
                            deficient = true
!
!                           CANCELLATION OF e(l) , IF l > 1 .
!
                            cs = zero
                            sn = -one
!
                            do i = l, k
!
                                f     = -sn*e2(i)
                                e2(i) =  cs*e2(i)
!
                                if ( abs(f)<=eps ) exit
!
!                               DEFINE THE GIVENS ROTATION.
!
                                if ( abs(d2(i))>abs(f) ) then
!
                                    t     = f/d2(i)
                                    sn    = sign( sqrt(one+t*t), d2(i) )
                                    d2(i) = d2(i)*sn
                                    cs    = one/sn
                                    sn    = t*cs
!
                                else if ( f/=zero ) then
!
                                    t     = d2(i)/f
                                    cs    = sign( sqrt(one+t*t), f )
                                    d2(i) = f*cs
                                    sn    = one/cs
                                    cs    = t*sn
!
                                else
!
                                    cs = zero
                                    sn = one
!
                                end if
!
!                               STORE THE GIVENS ROTATION.
!
                                cs2(i) = cs
                                sn2(i) = sn                                
!
                            end do
!
                            ii1 = l
                            ii2 = i - 1_i4b
!
                            exit
!
                        end if
!
                    end do
!
!                   TEST FOR RANK DEFICIENCIES.
!
                    if ( deficient ) exit
!
!                   CUTOFF FOR CONVERGENCE FAILURE.
!                   nqrs WILL BE AT MOST 2*n USUALLY .
!
                    if ( nqrs>=maxqrs ) then
!
!                       IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM    
!                       IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
                        if ( l/=k ) then
!
!                           SET FAILURE INDICATOR.
!
                            failure = true
!
                            l     = iminloc( abs(e2(l+1_i4b:k))) + l
                            e2(l) = zero
                            nqrs  = 0_i4b
!
                        end if
!
                    end if
!
!                   TEST FOR CONVERGENCE.
!
                    if ( l==k ) exit
!
                    nqrs = nqrs + 1_i4b
!
!                   REALLOCATE WORKSPACE TO STORE THE FRANCIS SETS
!                   OF GIVENS ROTATIONS FOR THE LEFT SINGULAR VECTORS
!                   IF NEEDED.
!
                    itu = itu + 1_i4b
!
                    if ( itu>maxqrs2u ) then
!
                        maxqrs2u = maxqrs2u + 5_i4b
!
                        call realloc( debu, maxqrs2u, iok )
!
                        if ( iok==0 ) then
!
                            call realloc( finu, maxqrs2u, iok  )
!
                            if ( iok==0 ) then
!
                                call realloc( csu, n, maxqrs2u, iok  )
!
                                if ( iok==0 ) then
                                    call realloc( snu, n, maxqrs2u, iok  )
                                end if
!
                            end if
!
                        end if
!
                        if ( iok/=0 ) then
                            call merror( name_proc//allocate_error )
                        end if
!
                    end if
!
                    if ( deflate ) then
!
                        if ( use_bisect ) then
!
!                           COMPUTE THE NEXT SINGULAR VALUE BY A BISECTION ALGORITHM.
!
                            call bd_one_singval_bisect( d2(l:k), e2(l:k), g, fail )
!
                        else if ( use_dqds ) then
!
!                           COMPUTE THE NEXT SINGULAR VALUE BY THE DQDS ALGORITHM.
!
                            call bd_one_singval_dqds( d2(l:k), e2(l:k), g, fail )
!
!                            call bd_one_singval_lasq1( d2(l:k), e2(l:k), g, fail )
!
                        else
!
!                           COMPUTE THE NEXT SINGULAR VALUE BY THE PAL-WALKER-KAHAN ALGORITHM.
!
                            call bd_one_singval_pwk( d2(l:k), e2(l:k), g, fail )
!
                        end if
!
                        deflate = .not.fail
!
!                       NOW COMPUTE THE SINGULAR VECTORS.
!
                        if ( deflate ) then
!
!                           ONE SINGULAR VALUE FOUND.
!
!                           FIRST TRY FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION.
!
!                            call dflgen_bd( d2(l:k), e2(l+1_i4b:k), g,                &
!                                            csu(l:k-1_i4b,itu), snu(l:k-1_i4b,itu),   &
!                                            diag(l:k-1_i4b), sub(l:k-1_i4b)           )
!
!                            call dflapp_bd( d2(l:k), e2(l+1_i4b:k),                  &
!                                            csu(l:k-1_i4b,itu), snu(l:k-1_i4b,itu),  &
!                                            diag(l:k-1_i4b), sub(l:k-1_i4b), deflate )
!
                            call dflgen2_bd( d2(l:k), e2(l+1_i4b:k), g,                &
                                             csu(l:k-1_i4b,itu), snu(l:k-1_i4b,itu),   &
                                             diag(l:k-1_i4b), sub(l:k-1_i4b), deflate  )
!
                            if ( deflate ) then
!
#ifdef _STAT
                                iter_deflate_singval  = iter_deflate_singval + 1
#endif
!
                                e2(l) = zero
!
                                debu(itu) = l
                                finu(itu) = k - 1_i4b
!
!                               GODUNOV DEFLATION SUCCESSFUL, EXIT.
!
                                exit
!
                            end if
!
                        end if
!
                    end if
!
!                   IF GODUNOV DEFLATION IS NOT SUCCESSFUL, USE IMPLICIT QR STEPS.
!
!                   COMPUTE WILKINSON SHIFT FROM BOTTOM 2 BY 2 MINOR OF B'*B .
!
                    z = d2(k)
                    x = d2(l)
                    y = d2(k-1_i4b)
                    g = e2(k-1_i4b)
                    h = e2(k)
                    f = ( (y-z)*(y+z) + (g-h)*(g+h) )/(two*h*y)
!
                    if ( abs(f)>one ) then
!
                       t = one/f
                       g = f*sqrt( one + t*t )
!
                    else
!
                       g = sign( sqrt( one + f*f ), f )
!
                    end if
!
                    f = ( (x-z)*(x+z) + h*( (y/( f + g )) - h ) )/x
!
!                   NEXT QR SWEEP.
!
                    cs = one
                    sn = one
!
                    do i = l + 1_i4b, k
!
                        j = i - 1_i4b
                        g = e2(i)
                        y = d2(i)
                        h = sn*g
                        g = cs*g
!
!                       DEFINE THE GIVENS ROTATION.
!
                        if ( abs(f)>abs(h) ) then
!
                            t     = h/f
                            sn    = sign( sqrt(one+t*t), f )
                            e2(j) = f*sn
                            cs    = one/sn
                            sn    = t*cs
!
                        else if ( h/=zero ) then
!
                            t     = f/h
                            cs    = sign( sqrt(one+t*t), h )
                            e2(j) = h*cs
                            sn    = one/cs
                            cs    = t*sn
!
                        else
!
                            cs    = zero
                            sn    = one
                            e2(j) = zero
!
                        end if
!
!                       APPLY THE GIVENS ROTATION.
!
                        f = x*cs + g*sn
                        g = -x*sn + g*cs
                        h = y*sn
                        y = y*cs
!
!                       DEFINE THE GIVENS ROTATION.
!
                        if ( abs(f)>abs(h) ) then
!
                            t     = h/f
                            sn    = sign( sqrt(one+t*t), f )
                            d2(j) = f*sn
                            cs    = one/sn
                            sn    = t*cs
!
                        else if ( h/=zero ) then
!
                            t     = f/h
                            cs    = sign( sqrt(one+t*t), h )
                            d2(j) = h*cs
                            sn    = one/cs
                            cs    = t*sn
!
                        else
!
                            cs    = zero
                            sn    = one
                            d2(j) = zero
!
                        end if
!
!                       APPLY THE GIVENS ROTATION.
!
                        f =  cs*g + sn*y
                        x = -sn*g + cs*y
!
!                       STORE THE GIVENS ROTATION.
!
                        csu(j,itu) = cs
                        snu(j,itu) = sn
!
                    end do
!
                    e2(l) = zero
                    e2(k) = f
                    d2(k) = x
!
                    debu(itu) = l
                    finu(itu) = k - 1_i4b
!
!                   RETURN TO TEST FOR RANK DEFICIENCIES AND SPLITTING.    
!
                end do
!
                nqrs2u = itu
!
            end if
!
!$OMP BARRIER
!
            if ( nqrs2u>=nfrancis .or. k==1_i4b  .or. deficient ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS
!               TO THE LEFT SINGULAR VECTORS.
!
                qstart = maxval( debu(:nqrs2u) ) + nqrs2u - 1_i4b
                qend   = minval( finu(:nqrs2u) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2u==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2u
!
                        i1 = debu(j)
                        i2 = finu(j)
!
#ifdef _BLAS
                        do i = i1, i2
!
                            call rot( nru2, u2(:nru2,i), 1_i4b, u2(:nru2,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                        end do
#else
                        u_tmp(:nru2) = u2(:nru2,i1)
!
                        do i = i1, i2
!
                            cs = csu(i,j)
                            sn = snu(i,j)
!
                            u2(:nru2,i)  =  cs*u_tmp(:nru2) + sn*u2(:nru2,i+1_i4b)
                            u_tmp(:nru2) = -sn*u_tmp(:nru2) + cs*u2(:nru2,i+1_i4b)
!
                        end do
!
                        u2(:nru2,i) = u_tmp(:nru2)
#endif
!
                    end do
!
                else if ( nqrs2u>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    l  = 0_i4b
                    i2 = qstart - nqrs2u + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        l = l + 1_i4b
!
                        i1 = debu(l)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
!
                                call rot( nru2, u2(:nru2,i), 1_i4b, u2(:nru2,i+1_i4b), 1_i4b, csu(i,l), snu(i,l) )
!
                            end do
#else
                            u_tmp(:nru2) = u2(:nru2,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                cs = csu(i,l)
                                sn = snu(i,l)
!
                                u2(:nru2,i)  =  cs*u_tmp(:nru2) + sn*u2(:nru2,i+1_i4b)
                                u_tmp(:nru2) = -sn*u_tmp(:nru2) + cs*u2(:nru2,i+1_i4b)
!
                            end do
!
                            u2(:nru2,i) = u_tmp(:nru2)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, l
!
                            call rot( nru2, u2(:nru2,i+1_i4b), 1_i4b, u2(:nru2,i), 1_i4b, csu(i,j), -snu(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        u_tmp(:nru2) = u2(:nru2,i+1_i4b)
!
                        do j = 1_i4b, l
!
                            cs = csu(i,j)
                            sn = snu(i,j)
!
                            u2(:nru2,i+1_i4b) = -sn*u2(:nru2,i) + cs*u_tmp(:nru2)
                            u_tmp(:nru2)      =  cs*u2(:nru2,i) + sn*u_tmp(:nru2)
!
                            i = i - 1_i4b
!
                        end do
!
                        u2(:nru2,i+1_i4b) = u_tmp(:nru2)
#endif
!
                    end do
!
!                   PIPELINE PHASE.
!
                    blk = 2_i4b*nqrs2u
!
                    if ( nru2>thr_bd_svd_blas3*blk ) then
!
!                       USE A BLAS3 ALGORITHM IF THE NUMBER OF ELEMENTS IN LEFT
!                       SINGULAR VECTORS IS SIGNIFICANTLY LARGER THAN THE BLOCKSIZE.
!
                        if ( blk>blkuv ) then
!
!                           REALLOCATE WORKSPACE TO STORE THE ACCUMULATED GIVENS
!                           ROTATIONS AND UPDATE THE LEFT SINGULAR VECTORS.
!
                            blkuv = 2_i4b*maxqrs2u
!
                            if ( allocated(uv) ) then
                                deallocate( uv, uvq )
                            end if
!
                            allocate( uv(nru2,blkuv), uvq(blkuv,blkuv), stat = iok )
!
                            if ( iok/=0 ) then
                                call merror( name_proc//allocate_error )
                            end if
!
                        end if
!
                        l   = qend - qstart + 1_i4b
                        l3  = l/nqrs2u
!
                        if ( l/=l3*nqrs2u ) l3 = l3 + 1_i4b
!
                        l1  = qstart - blk + 1_i4b
                        l2  = qstart
!
                        do l = 1_i4b, l3
!
                            l1  = l1 + nqrs2u
                            l2  = min( l2+nqrs2u, qend+1_i4b )
!
                            blk = l2 - l1 + 1_i4b
!
!                           ACCUMULATE GIVENS ROTATIONS OF THE CURRENT SET
!                           IN MATRIX uvq .
!
                            do j = 1_i4b, blk
!
                                uvq(:blkuv,j) = zero
                                uvq(j,j)      = one
!
                            end do
!
                            i1 = nqrs2u - 1_i4b
!
#ifdef _BLAS
                            do q = l1+nqrs2u-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2u
!
                                    call rot( blk, uvq(:blk,i2), 1_i4b, uvq(:blk,i2+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE LEFT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
!                            if ( blk==2_i4b*nqrs2u ) then
!
!                               PROCESS THE REGULAR SETS.
!
!                                uv(:nru2,nqrs2u+1_i4b:blk) = u2(:nru2,l1:l2-nqrs2u)
!
!                                call trmm( 'R', 'L', 'N', 'N', nru2, nqrs2u, one, uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv,  &
!                                           uv(:nru2,nqrs2u+1_i4b:blk), nru2 )
!
!                                uvq(:nqrs2u,nqrs2u+1_i4b:blk) = uvq(nqrs2u+1_i4b:blk,nqrs2u+1_i4b:blk)
!
!                                call gemm( 'N', 'N', nru2, nqrs2u, nqrs2u, one, u2(:nru2,l1+nqrs2u:l2), nru2,          &
!                                           uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv, one, uv(:nru2,nqrs2u+1_i4b:blk), nru2  )
!
!                                uv(:nru2,:nqrs2u) = u2(:nru2,l1+nqrs2u:l2)
!
!                                uvq(:nqrs2u,nqrs2u+1_i4b:blk) = uvq(nqrs2u+1_i4b:blk,:nqrs2u)
!    
!                                call trmm( 'R', 'U', 'N', 'N', nru2, nqrs2u, one, uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv,  &
!                                           uv(:nru2,:nqrs2u), nru2 )
!
!                                call gemm( 'N', 'N', nru2,  nqrs2u, nqrs2u, one, u2(:nru2,l1:l2-nqrs2u), nru2, &
!                                           uvq(:blkuv,:nqrs2u), blkuv, one, uv(:nru2,:nqrs2u), nru2 )
!
!                            else
!
!                               PROCESS THE RESIDUAL SET IF REQUIRED.
!
                                call gemm( 'N', 'N', nru2, blk, blk, one, u2(:nru2,l1:l2), nru2,  &
                                           uvq(:blkuv,:blk), blkuv, zero, uv(:nru2,:blk), nru2    )
!
!                            end if
#else
                            do q = l1+nqrs2u-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2u
!
                                    cs = csu(i,j)
                                    sn = snu(i,j)
!
                                    u_tmp(:blk)        = cs*uvq(:blk,i2) + sn*uvq(:blk,i2+1_i4b)
                                    uvq(:blk,i2+1_i4b) = cs*uvq(:blk,i2+1_i4b) - sn*uvq(:blk,i2)
                                    uvq(:blk,i2)       = u_tmp(:blk)
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE LEFT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
                            uv(:nru2,:blk) = matmul( u2(:nru2,l1:l2), uvq(:blk,:blk) )
#endif
!
                            u2(:nru2,l1:l2) = uv(:nru2,:blk)
!
                        end do
!
                    else
!
!                       USE A BLAS1 ALGORITHM OTHERWISE.
!
                        do q = qstart, qend
!
                            i = q
!
#ifdef _BLAS
                            do j = 1_i4b, nqrs2u
!
                                call rot( nru2, u2(:nru2,i), 1_i4b, u2(:nru2,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                                i = i - 1_i4b
!
                            end do
#else
                            u_tmp(:nru2) = u2(:nru2,i+1_i4b)
!
                            do j = 1_i4b, nqrs2u
!
                                cs = csu(i,j)
                                sn = snu(i,j)
!
                                u2(:nru2,i+1_i4b) = -sn*u2(:nru2,i) + cs*u_tmp(:nru2)
                                u_tmp(:nru2)      =  cs*u2(:nru2,i) + sn*u_tmp(:nru2)
!
                                i = i - 1_i4b
!
                            end do
!
                            u2(:nru2,i+1_i4b) = u_tmp(:nru2)
#endif
!
                        end do
!
                    end if
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2u
!
                        i2 = finu(j)
!
                        if ( i1<=i2 ) then
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( nru2, u2(:nru2,i), 1_i4b, u2(:nru2,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                            end do
#else
                            u_tmp(:nru2) = u2(:nru2,i1)
!
                            do i = i1, i2
!
                                cs = csu(i,j)
                                sn = snu(i,j)
!
                                u2(:nru2,i)  =  cs*u_tmp(:nru2) + sn*u2(:nru2,i+1_i4b)
                                u_tmp(:nru2) = -sn*u_tmp(:nru2) + cs*u2(:nru2,i+1_i4b)
!
                            end do
!
                            u2(:nru2,i) = u_tmp(:nru2)
#endif
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
!$OMP BARRIER
!
!$OMP MASTER
                nqrs2u = 0_i4b
!$OMP END MASTER
!
            end if
!
            if ( deficient ) then
!
!               APPLY THE GIVENS ROTATIONS TO COPE WITH
!               RANK DEFICIENCIES TO THE LEFT SINGULAR VECTORS.
!
                j = ii1 - 1_i4b
!
                do i = ii1, ii2
!
#ifdef _BLAS
                    call rot( nru2, u2(:nru2,i), 1_i4b, u2(:nru2,j), 1_i4b, cs2(i), sn2(i) )
#else
                    cs = cs2(i)
                    sn = sn2(i)
!
                    u_tmp(:nru2) = -sn*u2(:nru2,i)
                    u2(:nru2,i)  =  cs*u2(:nru2,i) + sn*u2(:nru2,j)
                    u2(:nru2,j)  =  u_tmp(:nru2)   + cs*u2(:nru2,j)
#endif
!
                end do
!
            else
!
                k = k - 1_i4b
!
            end if
!
!$OMP BARRIER
!
        end do
!
        if ( thread_num==0_i4b ) then
!
            d(:n) = d2(:n)
            e(:n) = e2(:n)
!
            deallocate( d2, e2 )
!
            if ( use_perfect_shift ) then
!
                deallocate( diag, sub )
!
            end if
!
        end if
!
!$OMP BARRIER
!
!       CONVERGENCE, MAKE SINGULAR VALUES POSITIVE AND
!       CHANGE SIGN OF LEFT SINGULAR VECTORS.
!
        do i = 1_i4b, n
!
            if ( d(i)>=zero ) cycle
!
            u2(:nru2,i) = -u2(:nru2,i)
!
        end do
!
        u(k1:k2,:n) = u2(:nru2,:n)
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( u2 )
!
        if ( allocated(uv) ) then
!
            deallocate( uv, uvq )
!
        end if
!
!$OMP END PARALLEL
!
!       CONVERGENCE, MAKE SINGULAR VALUES POSITIVE.
!
        where ( d(1_i4b:n)<zero ) d(1_i4b:n) = -d(1_i4b:n)
!
    else
#endif
!
        if ( use_perfect_shift ) then
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( diag(n), sub(n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
        end if
!
        blkuv = 0_i4b
!
!       DIAGONALISATION OF THE BIDIAGONAL FORM : LOOP OVER SINGULAR VALUES
!       AND OVER ALLOWED ITERATIONS maxqrs .
!
        k = n
!
        do
!
            if ( k<1_i4b ) exit
!
            deflate = use_perfect_shift
!
            do
!
                deficient = false
!
!               FIRST MAKE TEST FOR LAST DIAGONAL TERM d(k) BEING SMALL.
!
                if ( abs( d(k) )<=eps ) then
!
!                   SINCE d(k) IS SMALL WE WILL MAKE A SPECIAL PASS TO
!                   TRANSFORM e(k) TO ZERO.
!    
                    f    = e(k)
                    e(k) = zero
!
                    do i = k - 1_i4b, 1_i4b, -1_i4b
!
!                       CONSTRUCT AND APPLY TRANSFORMATION TO ZERO POSITION (i,k).
!
                        if ( abs(d(i))>abs(f) ) then
!
                            t    = f/d(i)
                            sn   = sign( sqrt(one+t*t), d(i) )
                            d(i) = d(i)*sn
                            cs   = one/sn
                            sn   = t*cs
!
                        else if ( f/=zero ) then
!
                            t    = d(i)/f
                            cs   = sign( sqrt(one+t*t), f )
                            d(i) = f*cs
                            sn   = one/cs
                            cs   = t*sn
!
                        else
!
                            cs = zero
                            sn = one
!
                        end if
!
!                       APPLY THE GIVENS ROTATION.
!
                        f    = -sn*e(i)
                        e(i) =  cs*e(i)
!
                    end do
!
!                   THE MATRIX IS NOW BIDIAGONAL, AND OF LOWER ORDER SINCE e(k) = zero .
!
                    exit
!
                end if    
!
!               TEST FOR SPLITTING OR RANK DEFICIENCIES.
!
                do l = k, 1_i4b, -1_i4b
!
                    if ( abs( e(l) )<=eps ) exit
!
!                   NOTE THAT THIS LOOP CANT COMPLETE SINCE e(1) = zero .
!
                    j = l - 1_i4b
!
                    if ( abs( d(j) )<=eps ) then
!
                        deficient = true
!
!                       CANCELLATION OF e(l) , IF l > 1 .
!
                        cs = zero
                        sn = -one
!
                        do i = l, k
!
                            f    = -sn*e(i)
                            e(i) =  cs*e(i)
!
                            if ( abs(f)<=eps ) exit
!
!                           DEFINE THE GIVENS ROTATION.
!
                            if ( abs(d(i))>abs(f) ) then
!
                                t    = f/d(i)
                                sn   = sign( sqrt(one+t*t), d(i) )
                                d(i) = d(i)*sn
                                cs   = one/sn
                                sn   = t*cs
!
                            else if ( f/=zero ) then
!
                                t    = d(i)/f
                                cs   = sign( sqrt(one+t*t), f )
                                d(i) = f*cs
                                sn   = one/cs
                                cs   = t*sn
!
                            else
!
                                cs = zero
                                sn = one
!
                            end if
!
!                           STORE THE GIVENS ROTATION.
!
                            cs2(i) = cs
                            sn2(i) = sn                                
!
                        end do
!
                        ii1 = l
                        ii2 = i - 1_i4b
!
                        exit
!
                    end if
!
                end do
!
!               TEST FOR RANK DEFICIENCIES.
!
                if ( deficient ) exit
!
!               CUTOFF FOR CONVERGENCE FAILURE.
!               nqrs WILL BE AT MOST 2*n USUALLY .
!
                if ( nqrs>=maxqrs ) then
!
!                   IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM    
!                   IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
                    if ( l/=k ) then
!
!                       SET FAILURE INDICATOR.
!
                        failure = true
!
                        l    = iminloc( abs(e(l+1_i4b:k))) + l
                        e(l) = zero
                        nqrs = 0_i4b
!
                    end if
!
                end if
!
!               TEST FOR CONVERGENCE.
!
                if ( l==k ) exit
!
                nqrs = nqrs + 1_i4b
!
!               REALLOCATE WORKSPACE TO STORE THE FRANCIS SETS
!               OF GIVENS ROTATIONS FOR THE LEFT SINGULAR VECTORS
!               IF NEEDED.
!
                nqrs2u = nqrs2u + 1_i4b
!
                if ( nqrs2u>maxqrs2u ) then
!
                    maxqrs2u = maxqrs2u + 5_i4b
!
                    call realloc( debu, maxqrs2u, iok )
!
                    if ( iok==0 ) then
!
                        call realloc( finu, maxqrs2u, iok  )
!
                        if ( iok==0 ) then
!
                            call realloc( csu, n, maxqrs2u, iok  )
!
                            if ( iok==0 ) then
                                call realloc( snu, n, maxqrs2u, iok  )
                            end if
!
                        end if
!
                    end if
!
                    if ( iok/=0 ) then
                        call merror( name_proc//allocate_error )
                    end if
!
                end if
!
                if ( deflate ) then
!
                    if ( use_bisect ) then
!
!                       COMPUTE THE NEXT SINGULAR VALUE BY A BISECTION ALGORITHM.
!
                        call bd_one_singval_bisect( d(l:k), e(l:k), g, fail )
!
                    else if ( use_dqds ) then
!
!                       COMPUTE THE NEXT SINGULAR VALUE BY THE DQDS ALGORITHM.
!
                        call bd_one_singval_dqds( d(l:k), e(l:k), g, fail )
!
!                        call bd_one_singval_lasq1( d2(l:k), e2(l:k), g, fail )
!
                    else
!
!                       COMPUTE THE NEXT SINGULAR VALUE BY THE PAL-WALKER-KAHAN ALGORITHM.
!
                        call bd_one_singval_pwk( d(l:k), e(l:k), g, fail )
!
                    end if
!
                    deflate = .not.fail
!
!                   NOW COMPUTE THE SINGULAR VECTORS.
!
                    if ( deflate ) then
!
!                       ONE SINGULAR VALUE FOUND.
!
!                       TRY FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION.
!
!                        call dflgen_bd( d(l:k), e(l+1_i4b:k), g,                        &
!                                        csu(l:k-1_i4b,nqrs2u), snu(l:k-1_i4b,nqrs2u),   &
!                                        diag(l:k-1_i4b), sub(l:k-1_i4b)                 )
!
!                        call dflapp_bd( d(l:k), e(l+1_i4b:k),                          &
!                                        csu(l:k-1_i4b,nqrs2u), snu(l:k-1_i4b,nqrs2u),  &
!                                        diag(l:k-1_i4b), sub(l:k-1_i4b), deflate       )
!
                        call dflgen2_bd( d(l:k), e(l+1_i4b:k), g,                       &
                                         csu(l:k-1_i4b,nqrs2u), snu(l:k-1_i4b,nqrs2u),  &
                                         diag(l:k-1_i4b), sub(l:k-1_i4b), deflate       )
!
                        if ( deflate ) then
!
#ifdef _STAT
                            iter_deflate_singval  = iter_deflate_singval + 1
#endif
!
                            e(l) = zero
!
                            debu(nqrs2u) = l
                            finu(nqrs2u) = k - 1_i4b
!
!                           GODUNOV DEFLATION SUCCESSFUL, EXIT.
!
                            exit
!
                        end if
!
                    end if
!
                end if
!
!               IF GODUNOV DEFLATION IS NOT SUCCESSFUL, USE IMPLICIT QR STEPS.
!
!               COMPUTE WILKINSON SHIFT FROM BOTTOM 2 BY 2 MINOR OF B'*B .
!
                z = d(k)
                x = d(l)
                y = d(k-1_i4b)
                g = e(k-1_i4b)
                h = e(k)
                f = ( (y-z)*(y+z) + (g-h)*(g+h) )/(two*h*y)
!
                if ( abs(f)>one ) then
!
                    t = one/f
                    g = f*sqrt( one + t*t )
!
                else
!
                    g = sign( sqrt( one + f*f ), f )
!
                end if
!
                f = ( (x-z)*(x+z) + h*( (y/( f + g )) - h ) )/x
!
!               NEXT QR SWEEP.
!
                cs = one
                sn = one
!
                do i = l + 1_i4b, k
!
                    j = i - 1_i4b
                    g = e(i)
                    y = d(i)
                    h = sn*g
                    g = cs*g
!
!                   DEFINE THE GIVENS ROTATION.
!
                    if ( abs(f)>abs(h) ) then
!
                        t    = h/f
                        sn   = sign( sqrt(one+t*t), f )
                        e(j) = f*sn
                        cs   = one/sn
                        sn   = t*cs
!
                    else if ( h/=zero ) then
!
                        t    = f/h
                        cs   = sign( sqrt(one+t*t), h )
                        e(j) = h*cs
                        sn   = one/cs
                        cs   = t*sn
!
                    else
!
                        cs   = zero
                        sn   = one
                        e(j) = zero
!
                    end if
!
!                   APPLY THE GIVENS ROTATION.
!
                    f = x*cs + g*sn
                    g = -x*sn + g*cs
                    h = y*sn
                    y = y*cs
!
!                   DEFINE THE GIVENS ROTATION.
!
                    if ( abs(f)>abs(h) ) then
!
                        t    = h/f
                        sn   = sign( sqrt(one+t*t), f )
                        d(j) = f*sn
                        cs   = one/sn
                        sn   = t*cs
!
                    else if ( h/=zero ) then
!
                        t    = f/h
                        cs   = sign( sqrt(one+t*t), h )
                        d(j) = h*cs
                        sn   = one/cs
                        cs   = t*sn
!
                    else
!
                        cs   = zero
                        sn   = one
                        d(j) = zero
!
                    end if
!
!                   APPLY THE GIVENS ROTATION.
!
                    f =  cs*g + sn*y
                    x = -sn*g + cs*y
!
!                   STORE THE GIVENS ROTATION.
!
                    csu(j,nqrs2u) = cs
                    snu(j,nqrs2u) = sn
!
                end do
!
                e(l) = zero
                e(k) = f
                d(k) = x
!
                debu(nqrs2u) = l
                finu(nqrs2u) = k - 1_i4b
!
!               RETURN TO TEST FOR RANK DEFICIENCIES AND SPLITTING.    
!
            end do
!
            if ( nqrs2u>=nfrancis .or. k==1_i4b  .or. deficient ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS
!               TO THE LEFT SINGULAR VECTORS.
!
                qstart = maxval( debu(:nqrs2u) ) + nqrs2u - 1_i4b
                qend   = minval( finu(:nqrs2u) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2u==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2u
!
                        i1 = debu(j)
                        i2 = finu(j)
!
#ifdef _BLAS
                        do i = i1, i2
!
                            call rot( nru, u(:nru,i), 1_i4b, u(:nru,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                        end do
#else
                        u_tmp(:nru) = u(:nru,i1)
!
                        do i = i1, i2
!
                            cs = csu(i,j)
                            sn = snu(i,j)
!
                            u(:nru,i)   =  cs*u_tmp(:nru) + sn*u(:nru,i+1_i4b)
                            u_tmp(:nru) = -sn*u_tmp(:nru) + cs*u(:nru,i+1_i4b)
!
                        end do
!
                        u(:nru,i) = u_tmp(:nru)
#endif
!
                    end do
!
                else if ( nqrs2u>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    l  = 0_i4b
                    i2 = qstart - nqrs2u + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        l = l + 1_i4b
!
                        i1 = debu(l)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
!
                                call rot( nru, u(:nru,i), 1_i4b, u(:nru,i+1_i4b), 1_i4b, csu(i,l), snu(i,l) )
!
                            end do
#else
                            u_tmp(:nru) = u(:nru,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                cs = csu(i,l)
                                sn = snu(i,l)
!
                                u(:nru,i)   =  cs*u_tmp(:nru) + sn*u(:nru,i+1_i4b)
                                u_tmp(:nru) = -sn*u_tmp(:nru) + cs*u(:nru,i+1_i4b)
!
                            end do
!
                            u(:nru,i) = u_tmp(:nru)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, l
!
                            call rot( nru, u(:nru,i+1_i4b), 1_i4b, u(:nru,i), 1_i4b, csu(i,j), -snu(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        u_tmp(:nru) = u(:nru,i+1_i4b)
!
                        do j = 1_i4b, l
!
                            cs = csu(i,j)
                            sn = snu(i,j)
!
                            u(:nru,i+1_i4b) = -sn*u(:nru,i) + cs*u_tmp(:nru)
                            u_tmp(:nru)     =  cs*u(:nru,i) + sn*u_tmp(:nru)
!
                            i = i - 1_i4b
!
                        end do
!
                        u(:nru,i+1_i4b) = u_tmp(:nru)
#endif
!
                    end do
!
!                   PIPELINE PHASE.
!
                    blk = 2_i4b*nqrs2u
!
                    if ( nru>thr_bd_svd_blas3*blk ) then
!
!                       USE A BLAS3 ALGORITHM IF THE NUMBER OF ELEMENTS IN LEFT
!                       SINGULAR VECTORS IS SIGNIFICANTLY LARGER THAN THE BLOCKSIZE.
!
                        if ( blk>blkuv ) then
!
!                           REALLOCATE WORKSPACE TO STORE THE ACCUMULATED GIVENS
!                           ROTATIONS AND UPDATE THE LEFT SINGULAR VECTORS.
!
                            blkuv = 2_i4b*maxqrs2u
!
                            if ( allocated(uv) ) then
                                deallocate( uv, uvq )
                            end if
!
                            allocate( uv(nru,blkuv), uvq(blkuv,blkuv), stat = iok )
!
                            if ( iok/=0 ) then
                                call merror( name_proc//allocate_error )
                            end if
!
                        end if
!
                        l   = qend - qstart + 1_i4b
                        l3  = l/nqrs2u
!
                        if ( l/=l3*nqrs2u ) l3 = l3 + 1_i4b
!
                        l1  = qstart - blk + 1_i4b
                        l2  = qstart
!
                        do l = 1_i4b, l3
!
                            l1  = l1 + nqrs2u
                            l2  = min( l2+nqrs2u, qend+1_i4b )
!
                            blk = l2 - l1 + 1_i4b
!
!                           ACCUMULATE GIVENS ROTATIONS OF THE CURRENT SET
!                           IN MATRIX uvq .
!
                            do j = 1_i4b, blk
!
                                uvq(:blkuv,j) = zero
                                uvq(j,j)      = one
!
                            end do
!
                            i1 = nqrs2u - 1_i4b
!
#ifdef _BLAS
                            do q = l1+nqrs2u-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2u
!
                                    call rot( blk, uvq(:blk,i2), 1_i4b, uvq(:blk,i2+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE LEFT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
!                            if ( blk==2_i4b*nqrs2u ) then
!
!                               PROCESS THE REGULAR SETS.
!
!                                uv(:nru,nqrs2u+1_i4b:blk) = u(:nru,l1:l2-nqrs2u)
!
!                                call trmm( 'R', 'L', 'N', 'N', nru, nqrs2u, one, uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv,  &
!                                           uv(:nru,nqrs2u+1_i4b:blk), nru )
!
!                                uvq(:nqrs2u,nqrs2u+1_i4b:blk) = uvq(nqrs2u+1_i4b:blk,nqrs2u+1_i4b:blk)
!
!                                call gemm( 'N', 'N', nru, nqrs2u, nqrs2u, one, u(:nru,l1+nqrs2u:l2), nru,            &
!                                           uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv, one, uv(:nru,nqrs2u+1_i4b:blk), nru  )
!
!                                uv(:nru,:nqrs2u) = u(:nru,l1+nqrs2u:l2)
!
!                                uvq(:nqrs2u,nqrs2u+1_i4b:blk) = uvq(nqrs2u+1_i4b:blk,:nqrs2u)
!
!                                call trmm( 'R', 'U', 'N', 'N', nru, nqrs2u, one, uvq(:blkuv,nqrs2u+1_i4b:blk), blkuv,  &
!                                           uv(:nru,:nqrs2u), nru )
!
!                                call gemm( 'N', 'N', nru,  nqrs2u, nqrs2u, one, u(:nru,l1:l2-nqrs2u), nru, &
!                                           uvq(:blkuv,:nqrs2u), blkuv, one, uv(:nru,:nqrs2u), nru )
!
!                            else
!
!                               PROCESS THE RESIDUAL SET IF REQUIRED.
!
!
                                call gemm( 'N', 'N', nru, blk, blk, one, u(:nru,l1:l2), nru,   &
                                           uvq(:blkuv,:blk), blkuv, zero, uv(:nru,:blk), nru   )
!
!                            end if
#else
                            do q = l1+nqrs2u-1_i4b, l2-1_i4b
!
                                i = q
!
                                i1 = i1 + 1_i4b
                                i2 = i1
!
                                do j = 1_i4b, nqrs2u
!
                                    cs = csu(i,j)
                                    sn = snu(i,j)
!
                                    u_tmp(:blk)        = cs*uvq(:blk,i2) + sn*uvq(:blk,i2+1_i4b)
                                    uvq(:blk,i2+1_i4b) = cs*uvq(:blk,i2+1_i4b) - sn*uvq(:blk,i2)
                                    uvq(:blk,i2)       = u_tmp(:blk)
!
                                    i  = i  - 1_i4b
                                    i2 = i2 - 1_i4b
!
                                end do
!
                            end do
!
!                           NOW APPLY THE ACCUMULATED GIVENS ROTATIONS IN MATRIX uvq TO THE LEFT SINGULAR VECTORS
!                           WITH A BLAS3 ALGORITHM.
!
                            uv(:nru,:blk) = matmul( u(:nru,l1:l2), uvq(:blk,:blk) )
#endif
!
                            u(:nru,l1:l2) = uv(:nru,:blk)
!
                        end do
!
                    else
!
!                       USE A BLAS1 ALGORITHM OTHERWISE.
!
                        do q = qstart, qend
!
                            i = q
!
#ifdef _BLAS
                            do j = 1_i4b, nqrs2u
!
                                call rot( nru, u(:nru,i), 1_i4b, u(:nru,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                                i = i - 1_i4b
!
                            end do
#else
                            u_tmp(:nru) = u(:nru,i+1_i4b)
!
                            do j = 1_i4b, nqrs2u
!
                                cs = csu(i,j)
                                sn = snu(i,j)
!
                                u(:nru,i+1_i4b)  = -sn*u(:nru,i) + cs*u_tmp(:nru)
                                u_tmp(:nru)      =  cs*u(:nru,i) + sn*u_tmp(:nru)
!
                                i = i - 1_i4b
!
                            end do
!
                            u(:nru,i+1_i4b) = u_tmp(:nru)
#endif
!
                        end do
!
                    end if
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2u
!
                        i2 = finu(j)
!
                        if ( i1<=i2 ) then
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( nru, u(:nru,i), 1_i4b, u(:nru,i+1_i4b), 1_i4b, csu(i,j), snu(i,j) )
!
                            end do
#else
                            u_tmp(:nru) = u(:nru,i1)
!
                            do i = i1, i2
!
                                cs = csu(i,j)
                                sn = snu(i,j)
!
                                u(:nru,i)   =  cs*u_tmp(:nru) + sn*u(:nru,i+1_i4b)
                                u_tmp(:nru) = -sn*u_tmp(:nru) + cs*u(:nru,i+1_i4b)
!
                            end do
!
                            u(:nru,i) = u_tmp(:nru)
#endif
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
                nqrs2u = 0_i4b
!
            end if
!
            if ( deficient ) then
!
!               APPLY THE GIVENS ROTATIONS TO COPE WITH
!               RANK DEFICIENCIES.
!
                j = ii1 - 1_i4b
!
                do i = ii1, ii2
!
#ifdef _BLAS
                    call rot( nru, u(:nru,i), 1_i4b, u(:nru,j), 1_i4b, cs2(i), sn2(i) )
#else
                    cs =  cs2(i)
                    sn =  sn2(i)
!
                    u_tmp(:nru) = -sn*u(:nru,i)
                    u(:nru,i)   =  cs*u(:nru,i) + sn*u(:nru,j)
                    u(:nru,j)   =  u_tmp(:nru)  + cs*u(:nru,j)
#endif
!
                end do
!
            else
!
                k = k - 1_i4b
!
            end if
!
        end do
!
!       CONVERGENCE, MAKE SINGULAR VALUES POSITIVE AND
!       CHANGE SIGN OF LEFT SINGULAR VECTORS.
!
        do i = 1_i4b, n
!
            if ( d(i)>=zero ) cycle
!
            d(i)      = -d(i)
            u(:nru,i) = -u(:nru,i)
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        if ( use_perfect_shift ) then
!
            deallocate( diag, sub )
!
        end if
!
        if ( allocated(uv) ) then
!
            deallocate( uv, uvq )
!
        end if
!
#ifdef _OPENMP2
    end if
#endif
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( csu, snu, debu, finu )
!
!   ORDER SINGULAR VALUES AND VECTORS IF REQUIRED.
!
    if ( present(sort) )  call singvec_sort( sort, d, u )
!
#ifdef _STAT
    write(*,*)
    write(*,*) name_proc//" :"
    write(*,*) repeat('*', len(name_proc) )
    write(*,*)
    write(*,*) "perfect_shift = ", use_perfect_shift
    write(*,*) "bisect        = ", use_bisect
    write(*,*) "dqds          = ", use_dqds
    write(*,*)
    write(*,*) "iter_deflate_singval  = ", iter_deflate_singval
    write(*,*)
#endif
!
!
! END OF SUBROUTINE bd_svd_b
! __________________________
!
    end subroutine bd_svd_b
!
! =========================================================================================
!
    subroutine bd_svd_t( upper, d, e, failure, sort, maxiter )
!
! Purpose
! _______
!
!   BD_SVD computes the singular values, S, of a real n-by-n 
!   (upper or lower) bidiagonal matrix B:
!
!                            B = Q * S * P'
!
!   , where S is a diagonal matrix with non-negative diagonal
!   elements (the singular values of B), and, Q and P are
!   orthogonal matrices (P' denotes the transpose of P).
!
!   The singular values are computed by the bidiagonal
!   implicit QR method. See the references (1) and (2)
!   for details.
!
!
! Arguments
! _________
!
!   UPPER        (INPUT) logical(lgl)
!                On entry, if:
!
!                - UPPER = true  : B is upper bidiagonal ;
!                - UPPER = false : B is lower bidiagonal.
!
!   D            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, D contains the diagonal elements of the bidiagonal
!                matrix B.
!
!                On exit,  D contains the singular values of B.
!
!   E            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, E contains the off-diagonal elements of the bidiagonal
!                matrix whose singular values are desired. E(1) is arbitrary.
!
!                On exit, E is destroyed.
!
!                The size of E must verify: size( E ) = size( D ) = n . 
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit ;
!                - FAILURE = true  :   indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the bidiagonal
!                  SVD of B.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the
!                algorithm. The algorithm fails to converge if the number 
!                of QR sweeps exceeds MAXITER * n. Convergence usually 
!                occurs in about 2 * n QR sweeps.
!
!                The default is 10.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from subroutine QRBD in the reference (1).
!
!   For further details, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,     only : assert_eq, iminloc
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:) :: d, e
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)     :: eps, cs, sn, f, z, x, y, g, h, t
!
    integer(i4b)   :: n, maxqrs, nqrs, i, j, l, k
!
    logical(lgl)   :: bidon
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_svd'
!
    character , parameter ::  a_maj=achar(65),  a_min=achar(97),  &
                              d_maj=achar(68),  d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
    bidon = upper
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b),    &
                   int(size(e),i4b),    & 
                   name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=2_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( d(1_i4b)<zero ) d = -d
!
        else if ( n==2_i4b ) then
!
            call las2( d(1_i4b), e(2_i4b), d(2_i4b), z, x )
!
            if ( present(sort) ) then
!
                select case( sort )
!
                    case( a_maj, a_min ) 
!
!                       SORT SINGULAR VALUES IN ASCENDING ORDER.
!
                        d(1_i4b) = z
                        d(2_i4b) = x
!
                    case( d_maj, d_min ) 
!
!                       SORT SINGULAR VALUES IN DESCENDING ORDER.
!
                        d(1_i4b) = x
                        d(2_i4b) = z
!
                end select
!
            else
!
                d(1_i4b) = x
                d(2_i4b) = z
!
            end if
!
        end if
!
        return
!
    end if
!
    e(1_i4b) = zero
!
!   INITIALIZE COUNTERS AND THRESHOLDS FOR QR ITERATIONS.
!
    nqrs = 0_i4b
!
    if ( present(maxiter) ) then
        maxqrs = n*merge( maxiter, 10_i4b, maxiter>=1_i4b )
    else
        maxqrs = n*10_i4b
    end if
!
!   STORE THE ELEMENTS OF THE BIDIAGONAL MATRIX IN REVERSED ORDER IF
!   THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL IN ORDER
!   TO MAINTAIN ACCURACY OF THE QR ITERATIONS.
!
!    if ( reverse( d(:n) ) ) then
!        d(1_i4b:n)  = d(n:1_i4b:-1_i4b)
!        e(2_i4b:n)  = e(n:2_i4b:-1_i4b)
!    end if
!
    if ( c1_5*d(1_i4b)<d(n) ) then
        d(1_i4b:n)  = d(n:1_i4b:-1_i4b)
        e(2_i4b:n)  = e(n:2_i4b:-1_i4b)
    end if
!
!   DETERMINE THE TOLERANCE FOR CONVERGENCE IN QR ITERATIONS.
!
    eps = fudge_eps*macheps*maxval( abs(d(:n)) + abs(e(:n)) )
!
!   DIAGONALISATION OF THE BIDIAGONAL FORM : LOOP OVER SINGULAR VALUES
!   AND OVER ALLOWED ITERATIONS maxqrs .
!
loop_ext :    do k = n, 1_i4b, -1_i4b
!
        do
!
!           FIRST MAKE TEST FOR LAST DIAGONAL TERM d(k) BEING SMALL.
!
            if ( abs( d(k) )<=eps ) then
!
!               SINCE d(k) IS SMALL WE WILL MAKE A SPECIAL PASS TO
!               TRANSFORM e(k) TO ZERO.
!
                f    = e(k)
                e(k) = zero
!
                do l = k - 1_i4b, 1_i4b, -1_i4b
!
!                   CONSTRUCT AND APPLY TRANSFORMATION TO ZERO POSITION (l,k).
!
                    if ( abs(d(l))>abs(f) ) then
!
                        t    = f/d(l)
                        sn   = sign( sqrt(one+t*t), d(l) )
                        d(l) = d(l)*sn
                        cs   = one/sn
                        sn   = t*cs
!
                    else if ( f/=zero ) then
!
                        t    = d(l)/f
                        cs   = sign( sqrt(one+t*t), f )
                        d(l) = f*cs
                        sn   = one/cs
                        cs   = t*sn
!
                    else
!
                        cs = zero
                        sn = one
!
                    end if
!
                    f    = -sn*e(l)
                    e(l) = cs*e(l)
!
                end do
!
!               THE MATRIX IS NOW BIDIAGONAL, AND OF LOWER ORDER SINCE e(k) = zero .
!
                cycle loop_ext
!
            end if    
!
!           TEST FOR SPLITTING OR RANK DEFICIENCIES.
!
            do l = k, 1_i4b, -1_i4b
!
                if ( abs( e(l) )<=eps ) exit
!
!               NOTE THAT THIS LOOP CANT COMPLETE SINCE e(1) = zero .
!
                j = l - 1_i4b
!
                if ( abs( d(j) )<=eps ) then
!
!                   CANCELLATION OF e(l) , IF l > 1 .
!
                    cs = zero
                    sn = -one
!
                    do i = l, k
!
                        f    = -sn*e(i)
                        e(i) = cs*e(i)
!
                        if ( abs(f)<=eps ) exit
!
!                       DEFINE THE GIVENS ROTATION.
!
                        if ( abs(d(i))>abs(f) ) then
!
                            t    = f/d(i)
                            sn   = sign( sqrt(one+t*t), d(i) )
                            d(i) = d(i)*sn
                            cs   = one/sn
                            sn   = t*cs
!
                        else if ( f/=zero ) then
!
                            t    = d(i)/f
                            cs   = sign( sqrt(one+t*t), f )
                            d(i) = f*cs
                            sn   = one/cs
                            cs   = t*sn
!
                        else
!
                            cs = zero
                            sn = one
!
                        end if
!
                    end do
!
                    exit
!
                end if
!
            end do
!
!           CUTOFF FOR CONVERGENCE FAILURE.
!           nqrs WILL BE AT MOST 2*n USUALLY .
!
            if ( nqrs>=maxqrs ) then
!
!               IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM    
!               IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
                if ( l/=k ) then
!
!                   SET FAILURE INDICATOR.
!
                    failure = true
!
                    l = iminloc( abs(e(l+1_i4b:k))) + l
                    e(l) = zero
                    nqrs = 0_i4b
!
                end if
!
            end if
!
!           TEST FOR CONVERGENCE.
!
            if ( l==k ) cycle loop_ext
!
            nqrs = nqrs + 1_i4b
!
!           COMPUTE WILKINSON SHIFT FROM BOTTOM 2 BY 2 MINOR OF B'*B .
!
            z = d(k)
            x = d(l)
            y = d(k-1_i4b)
            g = e(k-1_i4b)
            h = e(k)
            f = ( (y-z)*(y+z) + (g-h)*(g+h) )/(two*h*y)
!
            if ( abs(f)>one ) then
!
               t = one/f
               g = f*sqrt( one + t*t )
!
            else
!
               g = sign( sqrt( one + f*f ), f )
!
            end if
!
            f = ( (x-z)*(x+z) + h*( (y/( f + g )) - h ) )/x
!
!           NEXT QR SWEEP.
!
            cs = one
            sn = one
!
            do i = l + 1_i4b, k
!
                j = i - 1_i4b
                g = e(i)
                y = d(i)
                h = sn*g
                g = cs*g
!
!               DEFINE THE GIVENS ROTATION.
!
                if ( abs(f)>abs(h) ) then
!
                    t    = h/f
                    sn   = sign( sqrt(one+t*t), f )
                    e(j) = f*sn
                    cs   = one/sn
                    sn   = t*cs
!
                else if ( h/=zero ) then
!
                    t    = f/h
                    cs   = sign( sqrt(one+t*t), h )
                    e(j) = h*cs
                    sn   = one/cs
                    cs   = t*sn
!
                else
!
                    cs   = zero
                    sn   = one
                    e(j) = zero
!
                end if
!
!               APPLY THE GIVENS ROTATION.
!
                f = x*cs + g*sn
                g = -x*sn + g*cs
                h = y*sn
                y = y*cs
!
!               DEFINE THE GIVENS ROTATION.
!
                if ( abs(f)>abs(h) ) then
!
                    t    = h/f
                    sn   = sign( sqrt(one+t*t), f )
                    d(j) = f*sn
                    cs   = one/sn
                    sn   = t*cs
!
                else if ( h/=zero ) then
!
                    t    = f/h
                    cs   = sign( sqrt(one+t*t), h )
                    d(j) = h*cs
                    sn   = one/cs
                    cs   = t*sn
!
                else
!
                    cs   = zero
                    sn   = one
                    d(j) = zero
!
                end if
!
!               APPLY THE GIVENS ROTATION.
!
                f = cs*g + sn*y
                x = -sn*g + cs*y
!
            end do
!
            e(l) = zero
            e(k) = f
            d(k) = x
!
!           RETURN TO TEST FOR RANK DEFICIENCIES AND SPLITTING.    
!
        end do
!
    end do loop_ext
!
!   CONVERGENCE, MAKE SINGULAR VALUES POSITIVE.
!
    where( d(:n)<zero ) d(:n) = -d(:n)
!
!   ORDER SINGULAR VALUES IF REQUIRED.
!
    if ( present(sort) ) call singval_sort( sort, d )
!
!
! END OF SUBROUTINE bd_svd_t
! __________________________
!
    end subroutine bd_svd_t
!
! =========================================================================================
!
    subroutine bd_singval( d, e, nsing, s, failure, sort, vector, abstol, &
                           ls, theta, scaling, init )
!
! Purpose
! _______
!
!   BD_SINGVAL computes all or some of the greatest singular values of a real n-by-n
!   (upper or lower) bidiagonal matrix B by a bisection algorithm.
!
!   The Singular Value Decomposition of B is:
!
!                           B = Q * S * P'
!
!   where S is a diagonal matrix with non-negative diagonal elements (the singular
!   values of B), and, Q and P are orthogonal matrices (P' denotes the transpose of P).
!
!
! Arguments
! _________
!
!   D            (INPUT) real(stnd), dimension(:)
!                On entry, D contains the diagonal elements of the bidiagonal
!                matrix B.
!
!   E            (INPUT) real(stnd), dimension(:)
!                On entry, E contains the off-diagonal elements of the bidiagonal
!                matrix whose singular values are desired. E(1) is arbitrary.
!
!                The size of E must verify: size( E ) = size( D ) . 
!
!   NSING        (OUTPUT) integer(i4b)
!                On output, NSING specifies the number of singular values which have
!                been computed. Note that NSING may be greater than the optional argument
!                LS, if multiple singular values at index LS make unique selection
!                impossible.
!
!                If none of the optional arguments LS and THETA are used, NSING is set to size(D)
!                and all the singular values are computed.
!
!   S            (OUTPUT) real(stnd), dimension(:)
!                On exit, S(1:NSING)  contains the NSING greatest singular values of B.
!                The other values in S ( S(NSING+1:size(D)) ) are flagged  by a quiet NAN.
!
!                The size of S must verify: size( S ) = size( D ) . 
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection algorithm converged for all
!                  the computed singular values to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the singular values failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic. The sign of the incorrect singular values is set to negative.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'. For other values of SORT
!                nothing is done and S(:nsing) may  not be sorted in decreasing order of
!                of magnitude.
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to TRUE, a vectorized version of the bisection algorithm
!                is used.
!
!                The default is VECTOR=false.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the singular values. A singular value
!                (or cluster) is considered to be located if it has been determined to lie in
!                an interval whose width is ABSTOL or less.
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | T(GK) | will be used,
!                where | T(GK) | means the 1-norm of the GOLUB-KAHAN tridiagonal form of the bidiagonal matrix B
!                and ULP is the machine precision (distance from 1 to the next larger floating point number).
!
!                Singular values will be computed most accurately when ABSTOL is set to the square root of
!                the underflow threshold, sqrt(LAMCH('S')), not zero. 
!
!   LS           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LS specifies the number of singular values which must be
!                computed by the subroutine. On output, NSING may be different than LS 
!                if multiple singular values at index LS make unique selection
!                impossible.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LS must be greater than 0 and less or equal to size( D ) .
!
!                The default is LS = size(D).
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry, THETA specifies that the singular values which are greater or equal to
!                THETA must be computed. If none of the singular values are greater or equal to
!                THETA, NSING is set to zero and S(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                The default is THETA = 0.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix B is scaled before computing
!                the singular values.
!
!                The default is to scale the bidiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps are
!                computed from estimates of the eigenvalues of the associated B' * B tridiagonal
!                matrix obtained from the Pal-Walker-Kahan algorithm.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!
! Further Details
! _______________
!
!   Let S(i), i=1,...,N=size(D), be the N singular values of the bidiagonal matrix B in
!   decreasing order of magnitude. BD_SINGVAL then computes the LS largest singular values
!   ( or the singular values which are greater or equal to THETA) of B by a bisection method
!   (see the reference (1) below, Sec.8.5 ). The bisection method is applied to an associated 
!   2N by 2N symmetric tridiagonal matrix T (the so-called GOLUB-KAHAN form of B) whose 
!   eigenvalues are the singular values of B and their negatives (see the reference (2) 
!   below, Sec.3.3 ).
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error, svd_error2
    use Utilities,       only : assert_eq, assert, merror, arth
    use Eig_Procedures,  only : symtrid_qri
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#else
    use Num_Constants,   only : is_nan
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),  dimension(:) :: d, e
    real(stnd), intent(out), dimension(:) :: s
    real(stnd), intent(in), optional      :: abstol, theta
!
    integer(i4b), intent(out)          :: nsing
    integer(i4b), intent(in), optional :: ls
!
    logical(lgl), intent(in), optional :: vector, scaling, init
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: gamma, omega, tau, scalfac, maxbd, tnorm, atol1, rtol1, tmp1, tmp2, y, z, th, gu, gl
    real(stnd),  dimension(2*size(d))             :: z2
    real(stnd),  dimension(:),        allocatable :: lb, ub, yvec, zvec, thvec, tmpvec
!
    integer(i4b)                             :: n, n2, n2m1, num, nth, nz, ny, ngu, ngl, i, j, k
    integer(i4b),  dimension(:), allocatable :: nlb, nub, nyvec, nzvec, nthvec, nsvec
!
    integer  :: iok
!
    logical(lgl) :: use_theta, use_vectorization, do_scaling, use_init
!
#ifdef _OPENMP
!
    integer(i4b) :: num_threads, thread_num, jinc, jrem, joffset, j1, j2
!
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
!   relfac IS THE RELATIVE TOLERANCE.  AN INTERVAL (a,b] LIES WITHIN "RELATIVE TOLERANCE" IF
!   b-a < relfac*ulp*max(|a|,|b|), WHERE "ulp" IS THE MACHINE PRECISION.
!
    real(stnd),  parameter :: relfac=1.95_stnd
!
!   A "FUDGE FACTOR" TO WIDEN THE UPPER BOUND OF THE GERSHGORIN INTERVAL. IDEALLY, A VALUE OF 1 SHOULD WORK,
!   BUT ON MACHINES WITH SLOPPY ARITHMETIC, THIS NEEDS TO BE LARGER. 
!   THE DEFAULT FOR PUBLICLY RELEASED VERSIONS SHOULD BE LARGE ENOUGH TO HANDLE THE WORST MACHINE AROUND.
!   NOTE THAT THIS PARAMETER HAS NO EFFECT ON ACCURACY OF THE SOLUTION.
!
    real(stnd),  parameter :: fudge=two
!
    character(len=*),  parameter :: name_proc='bd_singval'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure           = false
    use_vectorization = false
    do_scaling        = true
    use_init          = false
    th                = zero
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b),    &
                   int(size(e),i4b),    & 
                   int(size(s),i4b),    & 
                   name_proc )
!
    n2    = 2_i4b*n
    n2m1  = n2 - 1_i4b
    nsing = n
!
    if ( present(ls) ) then
!
        call assert( logical( ls<=n, lgl ),     &
                     logical( 1_i4b<=ls, lgl ), &
                     name_proc )
        nsing = ls
!
    end if
!
    use_theta = false
!
    if ( present(theta) ) then
!
        if ( present(ls)  )     &
        call merror( name_proc//svd_error2 )
!
        th = max( theta, zero )
        if ( th>zero) use_theta = true
!
    end if
!
    if ( present(abstol) ) then
        atol1 = abstol
    else
        atol1 = -one
    end if
!
    if ( present(vector) ) then
        use_vectorization = vector
    end if
!
    if ( present(scaling) ) then
        do_scaling = scaling
    end if
!
    if ( present(init) ) then
        use_init = init
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
            if ( abs( d(1_i4b) )>=th ) then
                nsing    = 1_i4b
                s(1_i4b) = abs( d(1_i4b) )
            else
                nsing    = 0_i4b
                s(1_i4b) = nan()
            end if
        else
            nsing = 0_i4b
        end if
!
        return
!
    end if
!
!   FORM THE OFF DIAGONAL ELEMENTS OF THE GOLUB-KAHAN TRIDIAGONAL MATRIX.
!
    z2(1_i4b:n2:2_i4b)   = d(1_i4b:n)
    z2(2_i4b:n2m1:2_i4b) = e(2_i4b:n)
    z2(n2)               = zero
!
    maxbd = maxval( abs(z2(1_i4b:n2)) )
!
    if ( maxbd<=zero ) then
!
        if ( th<=zero ) then
            nsing      = n
            s(1_i4b:n) = zero
        else
            nsing      = 0_i4b
            s(1_i4b:n) = nan()
        end if
!
       return
!
    end if
!
!   GET MACHINE PARAMETERS.
!
    rtol1 = ulp*relfac
!
    gamma  = sqrt( safmin )
!
    if ( do_scaling ) then
!
#ifdef _SCALE
        omega  = scale(one,maxexp-1_i4b)
#else
        omega  = machbase**(maxexp-1_i4b)
#endif
        tau    = sqrt( gamma*omega )
!
!       COMPUTE SCALING FACTOR TO ESTIMATE THE (TINY) SINGULAR VALUES
!       TO RELATIVELY HIGH ACCURACY AND TO AVOID OVERFLOW.
!
        j       = exponent( tau ) - exponent( maxbd ) -1_i4b
        scalfac = machbase**j
!
    else
!
        scalfac = one
!
    end if
!
!   SCALE AND SQUARE THE OFF DIAGONAL ELEMENTS OF THE GOLUB-KAHAN TRIDIAGONAL MATRIX.
!   IF ANY ELEMENT OF abs(z2(:)) IS LESS THAN gamma/scalfac THAN THE CORRESPONDING SCALED
!   AND SQUARED ELEMENTS WILL UNDERFLOW TO zero AND HIGH RELATIVE ACCURACY WILL BE LOST.
!
    tmp1 = gamma/scalfac
!
    where( abs(z2(1_i4b:n2m1))>=tmp1 )
        z2(1_i4b:n2m1) = (scalfac*z2(1_i4b:n2m1))**2
    elsewhere
        z2(1_i4b:n2m1) = safmin
    end where
!
!   COMPUTE THE UPPER GERSCHGORIN BOUND OF THE GOLUB-KAHAN TRIDIAGONAL MATRIX.
!
    gu   = zero
    tmp1 = zero
!
    do i=1_i4b, n2
!
        tmp2 = sqrt( z2(i) )
        gu   = max( gu, tmp1+tmp2 )
        tmp1 = tmp2
!
        if ( z2(i)/=zero ) z2(i) = -z2(i)
!
    end do
!
    tnorm = gu
    gu    = gu + (real( n2, stnd)*fudge + eight)*gu*ulp
    ngu   = 0_i4b
!
!   CHECK ABSOLUTE TOLERANCE AND REPLACE IT BY DEFAULT VALUE IF NECESSARY.
!
    if ( atol1<=zero ) then
        atol1 = ulp*tnorm
    end if
!
    atol1 = max( gamma, atol1 )
!
!   DEFAULT LOWER BOUNDS FOR ALL THE SINGULAR VALUES.
!
    gl  = zero
    ngl = n
!
    if ( use_theta ) then
!
        gl = th*scalfac
!
!       DETERMINE nsing THE NUMBER OF SINGULAR VALUES WHICH ARE EQUAL
!       OR GREATER THAN th.
!
        ngl  = 0_i4b
        tmp2 = -max( gl, gamma )
!
        do j=1, n2m1
!
            tmp2 = z2(j)/tmp2 - gl
!
            if ( tmp2>=gamma ) then
                ngl = ngl + 1_i4b
            else
                tmp2 = min( tmp2, -gamma )
            end if
!
        end do
!
        nsing = ngl
!
    else if ( nsing/=n ) then
!
!       DETERMINE AN INTERVAL CONTAINING EXACTLY THE DESIRED SINGULAR VALUES.
!
!
!       FIRST GUESS FOR THE LOWER BOUND gl IF SINGULAR VALUES ARE WELL-SEPARATED.
!
        if ( nsing==n-1_i4b ) then
            gl = minval( abs(d(1_i4b:n)) )
        else
            gl = abs( d(nsing+1_i4b) )
        end if
!
        gl = gl*scalfac
!
!       CHECK QUALITY OF INITIAL ESTIMATE OF gl.
!
        if ( gl>zero ) then
!
            ngl  = 0_i4b
            tmp2 = -max( gl, gamma )
!
            do j=1_i4b, n2m1
!
                tmp2 = z2(j)/tmp2  - gl
!
                if ( tmp2>=gamma ) then
                    ngl = ngl + 1_i4b
                else
                    tmp2 = min( tmp2, -gamma )
                end if
!
            end do
!
        else
!
            ngl = n
!
        end if
!
        if ( ngl/=nsing ) then
!
            if ( ngl>nsing ) then
                y  = gl
                ny = ngl
                z  = gu
                nz = 0_i4b
            else
                y  = zero
                ny = n
                z  = gl
                nz = ngl
            end if
!
!           BISECTION METHOD FOR FINDING THE LOWER BOUND gl ON THE nsing
!           LARGEST SINGULAR VALUES OF THE BIDIAGONAL MATRIX.
!
            do
!
                tmp1 = abs( z - y )
                tmp2 = max( abs(z), abs(y) )
!
                if ( ngl==nsing .or. tmp1<max(gamma,rtol1*tmp2) .or. nz>=ny ) exit
!
                gl = ( z + y )*half
!
                if ( gl>zero ) then
!
                    num  = 0_i4b
                    tmp2 = -max( gl, gamma )
!
                    do j=1, n2m1
!
                        tmp2 = z2(j)/tmp2 - gl
!
                        if ( tmp2>=gamma ) then
                            num = num + 1_i4b
                        else
                            tmp2 = min( tmp2, -gamma )
                        end if
!
                    end do
!
                else
!
                    num = n
!
                end if
!
                ngl = min( max(num,nz), ny ) 
!
                if ( ngl<nsing ) then
                    z  = gl
                    nz = ngl
                else
                    y  = gl
                    ny = ngl
                end if
!
            end do
!
            if ( ngl/=nsing ) then
!
                gl  = y
                ngl = ny
!
                nsing = ny
!
            end if
!
        end if
!
    end if
!
    if ( nsing>0_i4b ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE UPPER AND LOWER BOUNDS
!       AND SINGULAR VALUES COUNTS ASSOCIATED WITH THE BOUNDS.
!
        allocate( lb(nsing), nlb(nsing), ub(nsing), nub(nsing), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        if ( use_init ) then
!
!           IF THE NUMBER OF REQUESTED SINGULAR VALUES IS GREATER THAN N/10,
!           COMPUTE THE INITIAL INTERVALS FROM THE EIGENVALUES OF THE GOLUB-KAHAN
!           TRIDIAGONAL MATRIX ESTIMATED FROM THE PAL-WALKER-KAHAN ALGORITHM.
!
            allocate( yvec(n2), zvec(n2), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           FORM THE ELEMENTS OF THE GOLUB-KAHAN TRIDIAGONAL MATRIX.
!
            yvec(1_i4b:n2)         = zero
!
            zvec(1_i4b:n2:2_i4b)   = d(1_i4b:n)
            zvec(2_i4b:n2m1:2_i4b) = e(2_i4b:n)
            zvec(n2)               = zero
!
!           COMPUTE EIGENVALUES OF THE GOLUB-KAHAN TRIDIAGONAL MATRIX.
!
            call symtrid_qri( yvec(:n2), zvec(:n2), failure, sort='d' )
!
            if ( failure ) then
!
                lb(:nsing) = gl
                ub(:nsing) = gu
!
            else
!
                where( yvec(:nsing)>zero )
                    yvec(:nsing) = scalfac*yvec(:nsing)
                elsewhere
                    yvec(:nsing) = zero
                end where
!
!               COMPUTE LOWER AND UPPER BOUNDS FOR THE EIGENVALUES OF THE GOLUB-KAHAN TRIDIAGONAL MATRIX.
!
                tmp1 = two*gu*ulp
!
                lb(:nsing) = yvec(:nsing) - tmp1
                ub(:nsing) = yvec(:nsing) + tmp1
!
                where( lb(:nsing)<gl )
                    ub(:nsing) = min( ub(:nsing) + gl - lb(:nsing), gu )
                    lb(:nsing) = gl
                end where
!
                where( ub(:nsing)>gu )
                    lb(:nsing) = max( lb(:nsing) + gu - ub(:nsing), gl )
                    ub(:nsing) = gu
                end where
!
            end if
!
            deallocate( yvec, zvec )
!
        else
!
            lb(:nsing) = gl
            ub(:nsing) = gu
!
        end if
!
        if ( use_vectorization ) then  
!
!           VECTORIZED VERSION OF THE BISECTION ALGORITHM.
!
!           ALLOCATE WORK VARIABLES NEEDED TO STORE THE UPPER AND LOWER BOUNDS
!           FOR THE SINGULAR VALUES.
!
            allocate( zvec(nsing), nzvec(nsing), yvec(nsing), nyvec(nsing),       &
                      thvec(nsing), nthvec(nsing), tmpvec(nsing), nsvec(nsing),   &
                      stat = iok    )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           INITIALIZE THE LOWER AND THE UPPER BOUNDS FOR THE REQUESTED SINGULAR VALUES.
!
            zvec(:nsing)  = ub(:nsing)
            nzvec(:nsing) = ngu
!
            yvec(:nsing)  = lb(:nsing)
            nyvec(:nsing) = ngl
!
            k    = 0_i4b
            num  = nsing
            nsvec(:nsing) = arth( 1_i4b, 1_i4b, nsing )
!
            do
!
                i = 0_i4b
!
                do j=1_i4b, num
!
                    y    = yvec(j)
                    ny   = nyvec(j)
                    z    = zvec(j)
                    nz   = nzvec(j)
!
                    tmp1 = abs( z - y )
                    tmp2 = max( abs(z), abs(y) )
!
!                   CHECK CONVERGENCE OF THE nsvec(j) SINGULAR VALUE.
!
                    if ( tmp1<max(rtol1*tmp2,atol1) .or. nz>=ny ) then
!
                        k = k + 1_i4b
!
                        s(k) = min( max(( y + z )*half,y), z )
!
                        lb(k)  = y
                        nlb(k) = ny
                        ub(k)  = z
                        nub(k) = nz
!
                    else
!
                        i = i + 1_i4b
!
                        if ( i/=j ) then
!
                            yvec(i)   = y
                            nyvec(i)  = ny
                            zvec(i)   = z
                            nzvec(i)  = nz
!
                            nsvec(i)  = nsvec(j)
!
                        end if
!
                    end if
!
                end do
!
!               CHECK IF ALL THE REQUESTED SINGULAR VALUES HAVE CONVERGED.
!
                if ( k==nsing ) exit
!
                num = i
!
!               NEXT BISECTION STEP FOR THE REMAINING i SINGULAR VALUES.
!
                thvec(:i) = ( yvec(:i) + zvec(:i) )*half
!
!               COMPUTE nthvec(:i) = COUNT(thvec(:i))  THE NUMBER OF SINGULAR VALUES GREATER
!               OR EQUAL TO thvec(:i) BY STURM SEQUENCE.
!
                nthvec(:i) = 0_i4b
                tmpvec(:i) = -max( thvec(:i), gamma )
!
                do j=1_i4b, n2m1
!
                    tmpvec(:i) =  z2(j)/tmpvec(:i) - thvec(:i)
!
                    where ( tmpvec(:i)>=gamma )
                        nthvec(:i) = nthvec(:i) + 1_i4b
                    elsewhere
                        tmpvec(:i) = min( tmpvec(:i), -gamma )
                    end where
!
                end do
!
!               INSURE THAT COUNT(thvec(:i)) IS MONOTONE.
!
                nthvec(:i) = min( max(nthvec(:i),nzvec(:i)),nyvec(:i) ) 
!
!               KEEP ONLY THE INTERVAL CONTAINING w(:i) S.T. COUNT(w(:i))=nsvec(:i)
!
                where ( nthvec(:i)<nsvec(:i) )
                    zvec(:i)  = thvec(:i)
                    nzvec(:i) = nthvec(:i)
                elsewhere
                    yvec(:i)  = thvec(:i)
                    nyvec(:i) = nthvec(:i)
                end where
!
            end do        
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( yvec, nyvec, zvec, nzvec, thvec, nthvec, tmpvec, nsvec )
!
        else
!
!           SCALAR VERSION OF THE BISECTION ALGORITHM, BUT MAY BE PARALLELIZED
!           if OPENMP IS USED.
!
#ifdef _OPENMP
!
            i           = omp_get_max_threads()
            ompparallel = omp_in_parallel()
            ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
            test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                          nsing>1_i4b                         .and.       &
                          i>1_i4b
!
            if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(num_threads,thread_num,jinc,jrem,joffset,k,i,j,j1,j2,iok,num,tmp1,tmp2,z,nz,y,ny,th,nth)
!
                num_threads = omp_get_num_threads()
                thread_num  = omp_get_thread_num()
!
!               DISTRIBUTE THE WORK AMONG THE THREADS.
!
                jrem = mod( nsing, num_threads )
                jinc = ( nsing - jrem )/num_threads
!
                if ( thread_num<jrem ) then
                    jinc    = jinc + 1_i4b
                    joffset = 1_i4b
                else
                    joffset = jrem + 1_i4b
                end if
!
                j1 = thread_num*jinc + joffset
                j2 = j1 + jinc - 1_i4b
!
                if ( j1<=j2 ) then
!
!                   INITIALIZE THE UPPER BOUNDS FOR THE REQUESTED SINGULAR VALUES.
!
                    z  = gu
                    nz = ngu
!
                    do k=j1, j2
!
!                       UPDATE THE LOWER AND UPPER BOUNDS FOR THE kTH SINGULAR VALUE.
!
                        y  = gl
                        ny = ngl
!
                        do i=k, j2
!
                            if ( lb(i)>gl ) then
!
                                y  =  lb(i)
!
                                exit
!
                            end if
!
                        end do
!
                        tmp1 = min( z, ub(k) )
!
                        if ( tmp1<y ) then
                            y  = gl
                            z  = gu
                        else
                            z  = tmp1
                        end if
!
                        do
!
                            tmp1 = abs( z - y )
                            tmp2 = max( abs(z), abs(y) )
!
!                           CHECK FOR CONVERGENCE.
!
                            if ( tmp1<max(rtol1*tmp2,atol1) .or. nz>=ny ) exit
!
!                           NEXT BISECTION STEP.
!
                            th = ( y + z )*half
!
!                           COMPUTE nth=COUNT(th)  THE NUMBER OF SINGULAR VALUES GREATER OR EQUAL TO th
!                           BY STURM SEQUENCE.
!
                            if ( th>zero ) then
!
                                num  = 0_i4b
                                tmp2 = -max( th, gamma )
!
                                do j=1_i4b, n2m1
!
                                    tmp2 = z2(j)/tmp2 - th
!
                                    if ( tmp2>=gamma ) then
                                        num = num + 1_i4b
                                    else
                                        tmp2 = min( tmp2, -gamma )
                                    end if
!
                                end do
!
                            else
!
                                num = n
!
                            end if
!
!                           INSURE THAT COUNT(th) IS MONOTONE.
!
                            nth = min( max(num,nz),ny ) 
!
!                           KEEP ONLY THE INTERVAL CONTAINING w S.T. COUNT(w)=k
!
                            if ( nth<k ) then
!
                                z  = th
                                nz = nth
!
                            else
!
                                y  = th
                                ny = nth
!
!                               REFINE THE REMINDING INTERVALS.
!
                                if ( j2<=nth ) then
!
                                    lb(j2) = max( th, lb(j2) )
!
                                else
!
                                    lb(nth) = max( th, lb(nth) )
!
                                    ub(nth+1_i4b) = min( th, ub(nth+1_i4b) )
!
                                end if
!
                            end if
!
                        end do        
!
!                       SINGULAR VALUE FOUND.
!
                        s(k) = min( max(( y + z )*half,y), z )
!
                        lb(k)  = y
                        nlb(k) = ny
!
                        ub(k)  = z
                        nub(k) = nz
!
                    end do
!
                end if
!
!$OMP END PARALLEL
!
            else
!
#endif
!
!               INITIALIZE THE UPPER BOUNDS FOR THE REQUESTED SINGULAR VALUES.
!
                z  = gu
                nz = ngu
!
                do k=1_i4b, nsing
!
!                   UPDATE THE LOWER AND UPPER BOUNDS FOR THE kTH SINGULAR VALUE.
!
                    y  = gl
                    ny = ngl
!
                    do i=k, nsing
!
                        if ( lb(i)>gl ) then
!
                            y  =  lb(i)
!
                            exit
!
                        end if
!
                    end do
!
                    tmp1 = min( z, ub(k) )
!
                    if ( tmp1<y ) then
                        y  = gl
                        z  = gu
                    else
                        z  = tmp1
                    end if
!
                    do
!
                        tmp1 = abs( z - y )
                        tmp2 = max( abs(z), abs(y) )
!
!                       CHECK FOR CONVERGENCE.
!
                        if ( tmp1<max(rtol1*tmp2,atol1) .or. nz>=ny ) exit
!
!                       NEXT BISECTION STEP.
!
                        th = ( y + z )*half
!
!                       COMPUTE nth=COUNT(th)  THE NUMBER OF SINGULAR VALUES GREATER OR EQUAL TO th
!                       BY STURM SEQUENCE.
!
                        if ( th>zero ) then
!
                            num  = 0_i4b
                            tmp2 = -max( th, gamma )
!
                            do j=1_i4b, n2m1
!
                                tmp2 = z2(j)/tmp2 - th
!
                                if ( tmp2>=gamma ) then
                                    num = num + 1_i4b
                                else
                                    tmp2 = min( tmp2, -gamma )
                                end if
!
                            end do
!
                        else
!
                            num = n
!
                        end if
!
!                       INSURE THAT COUNT(th) IS MONOTONE.
!
                        nth = min( max(num,nz),ny ) 
!
!                       KEEP ONLY THE INTERVAL CONTAINING w S.T. COUNT(w)=k
!
                        if ( nth<k ) then
!
                            z  = th
                            nz = nth
!
                        else
!
                            y  = th
                            ny = nth
!
!                           REFINE THE REMINDING INTERVALS.
!
                            if ( nsing<=nth ) then
!
                                lb(nsing) = max( th, lb(nsing) )
!
                            else
!
                                lb(nth) = max( th, lb(nth) )
!
                                ub(nth+1_i4b) = min( th, ub(nth+1_i4b) )
!
                            end if
!
                        end if
!
                    end do        
!
!                   SINGULAR VALUE FOUND.
!
                    s(k) = min( max(( y + z )*half,y), z )
!
                    lb(k)  = y
                    nlb(k) = ny
!
                    ub(k)  = z
                    nub(k) = nz
!
                end do
!
#ifdef _OPENMP
!
            end if
!
#endif
!
        end if
!
!       CHECK FOR CONVERGENCE.
!
        failure = false
        th = one/scalfac
!
        do k=1_i4b, nsing
!
            tmp1 = s(k)
!
            if ( lb(k)>ub(k)   .or.  &
                nub(k)>=nlb(k) .or.  & 
                tmp1>=gu       .or.  & 
                tmp1<gl  )   then
!
!               FLAG THE INCORRECT SINGULAR VALUES.
!
                tmp1    = - abs( tmp1 )
                failure = true
!
            end if
!
!           UNSCALED THE SINGULAR VALUES.
!
            s(k) = th*tmp1
!
#ifdef _F2003
            if ( ieee_support_datatype( s(k) ) ) then
!
                if ( .not.ieee_is_normal( s(k) ) )  failure = true
!
            end if
#else
            if ( is_nan( s(k) ) ) then
!
                failure = true
!
            end if
#endif
!
        end do
!
!       ORDER THE nsing SINGULAR VALUES.
!
        if ( present(sort) ) then
            call singval_sort( sort, s(:nsing) )
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( lb, nlb, ub, nub )
!
    end if
!
!   FLAG THE OTHER SINGULAR VALUES IF NECESSARY.
!
    if ( nsing/=n ) then
        th = nan()
        s(nsing+1_i4b:n) = th
    end if
!
!
! END OF SUBROUTINE bd_singval
! ____________________________
!
    end subroutine bd_singval
!
! =========================================================================================
!
    subroutine bd_singval2( d, e, nsing, s, failure, sort, vector, abstol, &
                            ls, theta, scaling, init )
!
! Purpose
! _______
!
!   BD_SINGVAL2 computes all or some of the greatest singular values of a real n-by-n
!   (upper or lower) bidiagonal matrix B by a bisection algorithm.
!
!   The Singular Value Decomposition of B is:
!
!                           B = Q * S * P'
!
!   where S is a diagonal matrix with non-negative diagonal elements (the singular
!   values of B), and, Q and P are orthogonal matrices (P' denotes the transpose of P).
!
!
! Arguments
! _________
!
!   D            (INPUT) real(stnd), dimension(:)
!                On entry, D contains the diagonal elements of the bidiagonal
!                matrix B.
!
!   E            (INPUT) real(stnd), dimension(:)
!                On entry, E contains the off-diagonal elements of the bidiagonal
!                matrix whose singular values are desired. E(1) is arbitrary.
!
!                The size of E must verify: size( E ) = size( D ) . 
!
!   NSING        (OUTPUT) integer(i4b)
!                On output, NSING specifies the number of singular values which have
!                been computed. Note that NSING may be greater than the optional argument
!                LS, if multiple singular values at index LS make unique selection
!                impossible.
!
!                If none of the optional arguments LS and THETA are used, NSING is set to size(D)
!                and all the singular values are computed.
!
!   S            (OUTPUT) real(stnd), dimension(:)
!                On exit, S(1:NSING)  contains the NSING greatest singular values of B.
!                The other values in S ( S(NSING+1:size(D)) ) are flagged  by a quiet NAN.
!
!                The size of S must verify: size( S ) = size( D ) . 
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection algorithm converged for all
!                  the computed singular values to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the singular values failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic. The sign of the incorrect singular values is set to negative.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'. For other values of SORT
!                nothing is done and S(:nsing) may  not be sorted in decreasing order of
!                of magnitude.
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to TRUE, a vectorized version of the bisection algorithm
!                is used.
!
!                The default is VECTOR=false.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the singular values. A singular value
!                (or cluster) is considered to be located if its square has been determined to lie in
!                an interval whose width is ABSTOL or less.
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | B' * B | will be used,
!                where | B' * B | means the 1-norm of the tridiagonal matrix B' * B ( B' means the transpose of B) 
!                and ULP is the machine precision (distance from 1 to the next larger floating point number).
!
!                Singular values will be computed most accurately when ABSTOL is set to the square root of the
!                underflow threshold, sqrt(LAMCH('S')), not zero. 
!
!   LS           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LS specifies the number of singular values which must be
!                computed by the subroutine. On output, NSING may be different than LS 
!                if multiple singular values at index LS make unique selection
!                impossible.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LS must be greater than 0 and less or equal to size( D ) .
!
!                The default is LS = size( D ).
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry, THETA specifies that the singular values which are greater or equal to
!                THETA must be computed. If none of the singular values are greater or equal to
!                THETA, NSING is set to zero and S(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                The default is THETA = 0.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix B is scaled before computing
!                the singular values.
!
!                The default is to scale the bidiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps are
!                computed from estimates of the eigenvalues of the associated B' * B tridiagonal
!                matrix obtained from the Pal-Walker-Kahan algorithm.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!
! Further Details
! _______________
!
!   Let S(i), i=1,...,N=size(D), be the N singular values of the bidiagonal matrix B in
!   decreasing order of magnitude. BD_SINGVAL2 then computes the LS largest singular values
!   ( or the singular values which are greater or equal to THETA) of B by a bisection method
!   (see the reference (1) below, Sec.8.5 ). The bisection method is applied (implicitly) to
!   the associated N by N symmetric tridiagonal matrix B' * B  whose eigenvalues are the squares
!   of the singular values of B by using the differential stationary form of the qd algorithm
!   of Rutishauser (see the reference (2) below, Sec.3.1 ).
!
!   BD_SINGVAL2 is faster than BD_SINGVAL, however if relative accuracy for small singular values
!   is required, BD_SINGVAL (which is based on the Golub-Kahan form of the bidiagonal matrix) is 
!   the best choice.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error, svd_error2
    use Utilities,       only : assert_eq, assert, merror, arth
    use Eig_Procedures,  only : symtrid_qri
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#else
    use Num_Constants,   only : is_nan
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),  dimension(:) :: d, e
    real(stnd), intent(out), dimension(:) :: s
    real(stnd), intent(in), optional      :: abstol, theta
!
    integer(i4b), intent(out)          :: nsing
    integer(i4b), intent(in), optional :: ls
!
    logical(lgl), intent(in), optional :: vector, scaling, init
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: gamma, omega, tau, scalfac, maxd, maxe, maxde, maxbd, tnorm, atol1, rtol1, tmp1, tmp2, y, z, th, gu, gl
    real(stnd),  dimension(size(d))               :: d2, e2
    real(stnd),  dimension(:),        allocatable :: lb, ub, yvec, zvec, thvec, tmp1vec, tmp2vec
!
    integer(i4b)                             :: n, num, nth, nz, ny, ngu, ngl, i, j, j1, j2, k
    integer(i4b),  dimension(:), allocatable :: nlb, nub, nyvec, nzvec, nthvec, nsvec
!
    integer  :: iok
!
    logical(lgl) :: use_theta, use_vectorization, do_scaling, use_init
!
#ifdef _OPENMP
!
    integer(i4b) :: num_threads, thread_num, jinc, jrem, joffset
!
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
!   relfac IS THE RELATIVE TOLERANCE.  AN INTERVAL (a,b] LIES WITHIN "RELATIVE TOLERANCE" IF
!   b-a < relfac*ulp*max(|a|,|b|), WHERE "ulp" IS THE MACHINE PRECISION.
!
    real(stnd),  parameter :: relfac=1.95_stnd
!
!   A "FUDGE FACTOR" TO WIDEN THE UPPER BOUND OF THE GERSHGORIN INTERVAL. IDEALLY, A VALUE OF 1 SHOULD WORK,
!   BUT ON MACHINES WITH SLOPPY ARITHMETIC, THIS NEEDS TO BE LARGER. 
!   THE DEFAULT FOR PUBLICLY RELEASED VERSIONS SHOULD BE LARGE ENOUGH TO HANDLE THE WORST MACHINE AROUND.
!   NOTE THAT THIS PARAMETER HAS NO EFFECT ON ACCURACY OF THE SOLUTION.
!
    real(stnd),  parameter :: fudge=two
!
    character(len=*),  parameter :: name_proc='bd_singval2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure           = false
    use_vectorization = false
    do_scaling        = true
    use_init          = false
    th                = zero
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b),    &
                   int(size(e),i4b),    & 
                   int(size(s),i4b),    & 
                   name_proc )
!
    nsing = n
!
    if ( present(ls) ) then
        call assert( logical( ls<=n, lgl ),     &
                     logical( 1_i4b<=ls, lgl ), &
                     name_proc )
        nsing = ls
    end if
!
    use_theta = false
!
    if ( present(theta) ) then
!
        if ( present(ls)  )     &
        call merror( name_proc//svd_error2 )
!
        th = max( theta, zero )
!
        if ( th>zero) use_theta = true
!
    end if
!
    if ( present(abstol) ) then
        atol1 = abstol
    else
        atol1 = -one
    end if
!
    if ( present(vector) ) then
        use_vectorization = vector
    end if
!
    if ( present(scaling) ) then
        do_scaling = scaling
    end if
!
    if ( present(init) ) then
        use_init = init
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
            if ( abs( d(1_i4b) )>=th ) then
                nsing    = 1_i4b
                s(1_i4b) = abs( d(1_i4b) )
            else
                nsing    = 0_i4b
                s(1_i4b) = nan()
            end if
        else
            nsing = 0_i4b
        end if
!
        return
!
    end if
!
    d2(1_i4b:n) =  abs( d(1_i4b:n) )
    maxd = maxval( d2(1_i4b:n) )
!
    e2(1_i4b)   = zero
    e2(2_i4b:n) =  abs( e(2_i4b:n) )
    maxe = maxval( e2(2_i4b:n) )
!
    maxde = maxval( d2(1_i4b:n-1_i4b)*e2(2_i4b:n) )
!
    maxbd = max( maxd, maxe )
!
    if ( maxbd<=zero ) then
!
        if ( th<=zero ) then
            nsing      = n
            s(1_i4b:n) = zero
        else
            nsing      = 0_i4b
            s(1_i4b:n) = nan()
        end if
!
       return
!
    end if
!
!   GET MACHINE PARAMETERS.
!
    rtol1 = ulp*relfac
!
    gamma  = sqrt( safmin )
!
    if ( do_scaling ) then
!
#ifdef _SCALE
        omega  = scale(one,maxexp-1_i4b)
#else
        omega  = machbase**(maxexp-1_i4b)
#endif
        tau    = sqrt( omega*gamma )
!
!       COMPUTE SCALING FACTOR TO ESTIMATE THE (TINY) SINGULAR VALUES
!       TO RELATIVELY HIGH ACCURACY AND TO AVOID OVERFLOW.
!
        j  = exponent( tau )
        j1 = j - exponent( maxbd ) - 1_i4b
        j2 = ( j - exponent( maxde ) - 1_i4b )/2_i4b
        j  = min( j1, j2 )
        scalfac = machbase**j
!
    else
!
        scalfac = one
!
    end if
!
!   SCALE AND SQUARE THE OFF DIAGONAL ELEMENTS OF THE GOLUB-KAHAN TRIDIAGONAL MATRIX.
!   IF ANY ELEMENT OF abs(d2(:)) OR abs(e2(:)) IS LESS THAN gamma/scalfac THAN THE CORRESPONDING
!   SCALED AND SQUARED ELEMENTS WILL UNDERFLOW TO zero AND HIGH RELATIVE ACCURACY WILL BE LOST.
!
    tmp1 = gamma/scalfac
!
    where( d2(1_i4b:n)>=tmp1 )
        d2(1_i4b:n) = (scalfac*d2(1_i4b:n))**2
    elsewhere
        d2(1_i4b:n) = safmin
    end where
!
    where( e2(2_i4b:n)>=tmp1 )
        e2(2_i4b:n) = (scalfac*e2(2_i4b:n))**2
    elsewhere
        e2(2_i4b:n) = safmin
    end where
!
!   COMPUTE THE UPPER GERSCHGORIN BOUND FOR EIGENVALUES OF B(t)B.
!
    gu   = zero
    tmp1 = zero
!
    do i=1_i4b, n-1_i4b
!
        y    = d2(i) + e2(i)
        tmp2 = sqrt( d2(i) )*sqrt( e2(i+1_i4b) )
        gu   = max( gu, y+tmp1+tmp2 )
        tmp1 = tmp2
!
    end do
!
    y    = d2(n) + e2(n)
    gu   = max( gu, y+tmp1 )
!
    tnorm = gu
    gu    = gu + (real( n, stnd)*fudge + eight)*gu*ulp
    ngu   = 0_i4b
!
!   CHECK ABSOLUTE TOLERANCE AND REPLACE IT BY DEFAULT VALUE IF NECESSARY.
!
    if ( atol1<=zero ) then
        atol1 = ulp*tnorm
    end if
!
    atol1 = max( gamma, atol1 )
!
!   DEFAULT LOWER BOUNDS FOR ALL THE SINGULAR VALUES.
!
    gl  = zero
    ngl = n
!
    if ( use_theta ) then
!
        gl = (th*scalfac)**2
!
!       DETERMINE nsing THE NUMBER OF SINGULAR VALUES WHICH ARE EQUAL
!       OR GREATER THAN th BY USING THE DIFFERENTIAL STATIONARY QD ALGORITHM.
!
        ngl  = 0_i4b
        tmp1 = one
        tmp2 = one
!
        do j=1_i4b, n
!
            tmp2 = gl + e2(j)*(tmp2/tmp1)
            tmp1 = d2(j) - tmp2
!
            if ( tmp1>=gamma ) then
                ngl = ngl + 1_i4b
            else
                tmp1 = min( tmp1, -gamma )
            end if
!
        end do
!
!       TO USE THE MODIFIED DIFFERENTIAL STATIONARY QD ALGORITHM
!       REPLACE THE PRECEDING LOOP WITH THIS ONE.
!
!        do j=1_i4b, n
!            tau = d2(j)           
!            tmp1 = tau - gl - e2(j)*(tmp2/tmp1 - one)
!            if ( tmp1>=gamma ) then
!                ngl = ngl + 1_i4b
!            else
!                tmp1 = min( tmp1, -gamma )
!            end if
!            tmp2 = tau
!        end do
!
        nsing = ngl
!
    else if ( nsing/=n ) then
!
!       DETERMINE AN INTERVAL CONTAINING EXACTLY THE DESIRED SINGULAR VALUES.
!
!
!       FIRST GUESS FOR THE LOWER BOUND gl IF SINGULAR VALUES ARE WELL-SEPARATED.
!
        if ( nsing==n-1_i4b ) then
            gl = minval( abs(d(1_i4b:n)) )
        else
            gl = abs( d(nsing+1_i4b) )
        end if
!
        gl = (gl*scalfac)**2
!
!       CHECK QUALITY OF INITIAL ESTIMATE OF gl.
!
        if ( gl>zero ) then
!
            ngl  = 0_i4b
            tmp1 = one
            tmp2 = one
!
            do j=1_i4b, n
!
                tmp2 = gl + e2(j)*(tmp2/tmp1)
                tmp1 = d2(j) - tmp2
!
                if ( tmp1>=gamma ) then
                    ngl = ngl + 1_i4b
                else
                    tmp1 = min( tmp1, -gamma )
                end if
!
            end do
!
!            TO USE THE MODIFIED DIFFERENTIAL STATIONARY QD ALGORITHM
!            REPLACE THE PRECEDING LOOP WITH THIS ONE.
!
!             do j=1_i4b, n
!                 tau = d2(j)           
!                 tmp1 = tau - gl - e2(j)*(tmp2/tmp1 - one)
!                 if ( tmp1>=gamma ) then
!                     ngl = ngl + 1_i4b
!                 else
!                     tmp1 = min( tmp1, -gamma )
!                 end if
!                 tmp2 = tau
!             end do
        else
!
            ngl = n
!
        end if
!
        if ( ngl/=nsing ) then
!
            if ( ngl>nsing ) then
                y  = gl
                ny = ngl
                z  = gu
                nz = 0_i4b
            else
                y  = zero
                ny = n
                z  = gl
                nz = ngl
            end if
!
!           BISECTION METHOD FOR FINDING THE LOWER BOUND gl ON THE SQUARES OF THE nsing
!           LARGEST SINGULAR VALUES OF THE BIDIAGONAL MATRIX.
!
            do
!
                tmp1 = abs( z - y )
                tmp2 = max( abs(z), abs(y) )
!
                if ( ngl==nsing .or. tmp1<max(gamma,rtol1*tmp2) .or. nz>=ny ) exit
!
                gl = ( z + y )*half
!
                if ( gl>zero ) then
!
                    num  = 0_i4b
                    tmp1 = one
                    tmp2 = one
!
                    do j=1_i4b, n
!
                        tmp2 = gl + e2(j)*(tmp2/tmp1)
                        tmp1 = d2(j) - tmp2
!
                        if ( tmp1>=gamma ) then
                            num = num + 1_i4b
                        else
                            tmp1 = min( tmp1, -gamma )
                        end if
!
                    end do
!
!                    TO USE THE MODIFIED DIFFERENTIAL STATIONARY QD ALGORITHM
!                    REPLACE THE PRECEDING LOOP WITH THIS ONE.
!
!                    do j=1_i4b, n
!                        tau = d2(j)           
!                        tmp1 = tau - gl - e2(j)*(tmp2/tmp1 - one)
!                        if ( tmp1>=gamma ) then
!                            num = num + 1_i4b
!                        else
!                            tmp1 = min( tmp1, -gamma )
!                        end if
!                        tmp2 = tau
!                    end do
!
                else
!
                    num = n
!
                end if
!
                ngl = min( max(num,nz), ny ) 
!
                if ( ngl<nsing ) then
                    z  = gl
                    nz = ngl
                else
                    y  = gl
                    ny = ngl
                end if
!
            end do
!
            if ( ngl/=nsing ) then
!
                gl  = y
                ngl = ny
!
                nsing = ny
!
            end if
!
        end if
!
    end if
!
    if ( nsing>0_i4b ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE UPPER AND LOWER BOUNDS
!       AND SINGULAR VALUES COUNTS ASSOCIATED WITH THE BOUNDS.
!
        allocate( lb(nsing), nlb(nsing), ub(nsing), nub(nsing), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        if ( use_init ) then
!
!           COMPUTE THE INITIAL INTERVALS FROM THE EIGENVALUES OF THE B(t)B
!           TRIDIAGONAL MATRIX ESTIMATED FROM THE PAL-WALKER-KAHAN ALGORITHM.
!
            allocate( yvec(n), zvec(n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            yvec(1_i4b:n) = d(1_i4b:n)/maxbd
            zvec(1_i4b:n) = e(1_i4b:n)/maxbd
!
!           COMPUTE ELEMENTS OF THE TRIDIAGONAL MATRIX B(t)B.
!
            tmp2 = zero
!
            do i=1_i4b, n-1_i4b
!
                tmp1    = yvec(i)
                yvec(i) = tmp1*tmp1 + tmp2*tmp2
                tmp2    = zvec(i+1_i4b)
                zvec(i) = abs( tmp1*tmp2 )
!
            end do
!
            yvec(n) = yvec(n)*yvec(n) + tmp2*tmp2
            zvec(n) = zero            
!
!           COMPUTE EIGENVALUES OF THE TRIDIAGONAL MATRIX B(t)B.
!
            call symtrid_qri( yvec(:n), zvec(:n), failure, sort='d' )
!
!           COMPUTE LOWER AND UPPER BOUNDS FOR THE EIGENVALUES OF THE TRIDIAGONAL MATRIX B(t)B.
!
            if ( failure ) then
!
                lb(:nsing) = gl
                ub(:nsing) = gu
!
            else
!
                tmp1 = (scalfac*maxbd)**2
!
                where( yvec(:nsing)>zero )
                    yvec(:nsing) = tmp1*yvec(:nsing)
                elsewhere
                    yvec(:nsing) = zero
                end where
!
                tmp1 = two*gu*ulp
!
                lb(:nsing) = yvec(:nsing) - tmp1
                ub(:nsing) = yvec(:nsing) + tmp1
!
                where( lb(:nsing)<gl )
                    ub(:nsing) = min( ub(:nsing) + gl - lb(:nsing), gu )
                    lb(:nsing) = gl
                end where
!
                where( ub(:nsing)>gu )
                    lb(:nsing) = max( lb(:nsing) + gu - ub(:nsing), gl )
                    ub(:nsing) = gu
                end where
!
            end if
!
            deallocate( yvec, zvec )
!
        else
!
            lb(:nsing) = gl
            ub(:nsing) = gu
!
        end if
!
        if ( use_vectorization ) then  
!
!           VECTORIZED VERSION OF THE BISECTION ALGORITHM.
!
!           ALLOCATE WORK VARIABLES NEEDED TO STORE THE UPPER AND LOWER BOUNDS
!           FOR THE SINGULAR VALUES.
!
            allocate( zvec(nsing), nzvec(nsing), yvec(nsing), nyvec(nsing),        &
                      thvec(nsing), nthvec(nsing), tmp1vec(nsing), tmp2vec(nsing), &
                      nsvec(nsing),  stat = iok    )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           INITIALIZE THE LOWER AND THE UPPER BOUNDS FOR THE SQUARES
!           OF THE REQUESTED SINGULAR VALUES.
!
            zvec(:nsing)  = ub(:nsing)
            nzvec(:nsing) = ngu
!
            yvec(:nsing)  = lb(:nsing)
            nyvec(:nsing) = ngl
!
            k    = 0_i4b
            num  = nsing
            nsvec(:nsing) = arth( 1_i4b, 1_i4b, nsing )
!
            do
                i = 0_i4b
!
                do j=1_i4b, num
!
                    y    = yvec(j)
                    ny   = nyvec(j)
                    z    = zvec(j)
                    nz   = nzvec(j)
!
                    tmp1 = abs( z - y )
                    tmp2 = max( abs(z), abs(y) )
!
!                   CHECK CONVERGENCE OF THE nsvec(j) SINGULAR VALUE.
!
                    if ( tmp1<max(rtol1*tmp2,atol1) .or. nz>=ny ) then
!
                        k = k + 1_i4b
!
                        s(k) = min( max(( y + z )*half,y), z )
!
                        lb(k)  = y
                        nlb(k) = ny
                        ub(k)  = z
                        nub(k) = nz
!
                    else
!
                        i = i + 1_i4b
!
                        if ( i/=j ) then
!
                            yvec(i)   = y
                            nyvec(i)  = ny
                            zvec(i)   = z
                            nzvec(i)  = nz
!
                            nsvec(i)  = nsvec(j)
!
                        end if
!
                    end if
!
                end do
!
!               CHECK IF ALL THE REQUESTED SINGULAR VALUES HAVE CONVERGED.
!
                if ( k==nsing ) exit
!
                num = i
!
!               NEXT BISECTION STEP FOR THE REMAINING i SINGULAR VALUES.
!
                thvec(:i) = ( yvec(:i) + zvec(:i) )*half
!
!               COMPUTE nthvec(:i) = COUNT(thvec(:i))  THE NUMBER OF SINGULAR VALUES GREATER
!               OR EQUAL TO sqrt( thvec(:i) ) BY STURM SEQUENCE.
!
                nthvec(:i)  = 0_i4b
                tmp1vec(:i) = one
                tmp2vec(:i) = one
!
                do j=1_i4b, n
!
                    tmp2vec(:i) = thvec(:i) + e2(j)*(tmp2vec(:i)/tmp1vec(:i))
                    tmp1vec(:i) = d2(j) - tmp2vec(:i)
!
                    where ( tmp1vec(:i)>=gamma )
                        nthvec(:i) = nthvec(:i) + 1_i4b
                    elsewhere
                        tmp1vec(:i) = min( tmp1vec(:i), -gamma )
                    end where
!
                end do
!
!                    TO USE THE MODIFIED DIFFERENTIAL STATIONARY QD ALGORITHM
!                    REPLACE THE PRECEDING LOOP WITH THIS ONE.
!
!                    do j=1_i4b, n
!                        tau = d2(j)           
!                        tmp1vec(:i) = tau - thvec(:i) - e2(j)*(tmp2/tmp1vec(:i) - one)
!                        where ( tmp1vec(:i)>=gamma )
!                            nthvec(:i) = nthvec(:i) + 1_i4b
!                        elsewhere
!                            tmp1vec(:i) = min( tmp1vec(:i), -gamma )
!                        end where
!                        tmp2 = tau
!                    end do
!
!               INSURE THAT COUNT(thvec(:i)) IS MONOTONE.
!
                nthvec(:i) = min( max(nthvec(:i),nzvec(:i)),nyvec(:i) ) 
!
!               KEEP ONLY THE INTERVAL CONTAINING w(:i) S.T. COUNT(w(:i))=nsvec(:i)
!
                where ( nthvec(:i)<nsvec(:i) )
                    zvec(:i)  = thvec(:i)
                    nzvec(:i) = nthvec(:i)
                elsewhere
                    yvec(:i)  = thvec(:i)
                    nyvec(:i) = nthvec(:i)
                end where
!
            end do        
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( yvec, nyvec, zvec, nzvec, thvec, nthvec, tmp1vec, tmp2vec, nsvec )
!
        else
!
!           SCALAR VERSION OF THE BISECTION ALGORITHM, BUT MAY BE PARALLELIZED
!           if OPENMP IS USED.
!
#ifdef _OPENMP
!
            i           = omp_get_max_threads()
            ompparallel = omp_in_parallel()
            ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
            test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                          nsing>1_i4b                         .and.       &
                          i>1_i4b
!
            if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(num_threads,thread_num,jinc,jrem,joffset,k,i,j,j1,j2,iok,num,tmp1,tmp2,z,nz,y,ny,th,nth)
!
                num_threads = omp_get_num_threads()
                thread_num  = omp_get_thread_num()
!
!               DISTRIBUTE THE WORK AMONG THE THREADS.
!
                jrem = mod( nsing, num_threads )
                jinc = ( nsing - jrem )/num_threads
!
                if ( thread_num<jrem ) then
                    jinc    = jinc + 1_i4b
                    joffset = 1_i4b
                else
                    joffset = jrem + 1_i4b
                end if
!
                j1 = thread_num*jinc + joffset
                j2 = j1 + jinc - 1_i4b
!
                if ( j1<=j2 ) then
!
!                   INITIALIZE THE UPPER BOUNDS FOR THE REQUESTED SINGULAR VALUES.
!
                    z  = gu
                    nz = ngu
!
                    do k=j1, j2
!
!                       UPDATE THE LOWER AND UPPER BOUNDS FOR THE SQUARE OF THE kTH SINGULAR VALUE.
!
                        y  = gl
                        ny = ngl
!
                        do i=k, j2
!
                            if ( lb(i)>gl ) then
                                y  =  lb(i)
!
                                exit
!
                            end if
!
                        end do
!
                        tmp1 = min( z, ub(k) )
!
                        if ( tmp1<y ) then
                            y  = gl
                            z  = gu
                        else
                            z  = tmp1
                        end if
!
                        do
!
                            tmp1 = abs( z - y )
                            tmp2 = max( abs(z), abs(y) )
!
!                           CHECK FOR CONVERGENCE.
!
                            if ( tmp1<max(rtol1*tmp2,atol1) .or. nz>=ny ) exit
!
!                           NEXT BISECTION STEP.
!
                            th = ( y + z )*half
!
!                           COMPUTE nth=COUNT(th)  THE NUMBER OF SINGULAR VALUES GREATER OR EQUAL
!                           TO sqrt( th ) BY STURM SEQUENCE.
!
                            if ( th>zero ) then
!
                                num  = 0_i4b
                                tmp1 = one
                                tmp2 = one
!
                                do j=1_i4b, n
!
                                    tmp2 = th + e2(j)*(tmp2/tmp1)
                                    tmp1 = d2(j) - tmp2
!
                                    if ( tmp1>=gamma ) then
                                        num = num + 1_i4b
                                    else
                                        tmp1 = min( tmp1, -gamma )
                                    end if
!
                                end do
!
!                                TO USE THE MODIFIED DIFFERENTIAL STATIONARY QD ALGORITHM
!                                REPLACE THE PRECEDING LOOP WITH THIS ONE.
!
!                                do j=1_i4b, n
!                                    tau = d2(j)           
!                                    tmp1 = tau - th - e2(j)*(tmp2/tmp1 - one)
!                                    if ( tmp1>=gamma ) then
!                                        num = num + 1_i4b
!                                        if ( (num>=j2) .or. (n-j+num<k) ) exit
!                                    else
!                                        tmp1 = min( tmp1, -gamma )
!                                    end if
!                                    tmp2 = tau
!                                end do
!
                            else
!
                                num = n
!
                            end if
!
!                           INSURE THAT COUNT(th) IS MONOTONE.
!
                            nth = min( max(num,nz),ny ) 
!
!                           KEEP ONLY THE INTERVAL CONTAINING w S.T. COUNT(w)=k
!
                            if ( nth<k ) then
!
                                z  = th
                                nz = nth
!
                            else
!
                                y  = th
                                ny = nth
!
!                               REFINE THE REMINDING INTERVALS.
!
                                if ( j2<=nth ) then
!
                                    lb(j2) = max( th, lb(j2) )
!
                                else
!
                                    lb(nth) = max( th, lb(nth) )
!
                                    ub(nth+1_i4b) = min( th, ub(nth+1_i4b) )
!
                                end if
!
                            end if
!
                        end do        
!
!                       SINGULAR VALUE FOUND.
!
                        s(k) = min( max(( y + z )*half,y), z )
!
                        lb(k)  = y
                        nlb(k) = ny
!
                        ub(k)  = z
                        nub(k) = nz
!
                    end do
!
                end if
!
!$OMP END PARALLEL
!
            else
!
#endif
!
!               INITIALIZE THE UPPER BOUNDS FOR THE REQUESTED SINGULAR VALUES.
!
                z  = gu
                nz = ngu
!
                do k=1_i4b, nsing
!
!                   UPDATE THE LOWER AND UPPER BOUNDS FOR THE SQUARE OF THE kTH SINGULAR VALUE.
!
                    y  = gl
                    ny = ngl
!
                    do i=k, nsing
!
                        if ( lb(i)>gl ) then
!
                            y  =  lb(i)
!
                            exit
!
                        end if
!
                    end do
!
                    tmp1 = min( z, ub(k) )
!
                    if ( tmp1<y ) then
                        y  = gl
                        z  = gu
                    else
                        z  = tmp1
                    end if
!
                    do
!
                        tmp1 = abs( z - y )
                        tmp2 = max( abs(z), abs(y) )
!
!                       CHECK FOR CONVERGENCE.
!
                        if ( tmp1<max(rtol1*tmp2,atol1) .or. nz>=ny ) exit
!
!                       NEXT BISECTION STEP.
!
                        th = ( y + z )*half
!
!                       COMPUTE nth=COUNT(th)  THE NUMBER OF SINGULAR VALUES GREATER OR EQUAL
!                       TO sqrt(th) BY STURM SEQUENCE.
!
                        if ( th>zero ) then
!
                            num  = 0_i4b
                            tmp1 = one
                            tmp2 = one
!
                            do j=1_i4b, n
!
                                tmp2 = th + e2(j)*(tmp2/tmp1)
                                tmp1 = d2(j) - tmp2
!
                                if ( tmp1>=gamma ) then
                                    num = num + 1_i4b
!                                    if ( (num>=nsing) .or. (n-j+num<k) ) exit
                                else
                                    tmp1 = min( tmp1, -gamma )
                                end if
!
                            end do
!
!                             TO USE THE MODIFIED DIFFERENTIAL STATIONARY QD ALGORITHM
!                             REPLACE THE PRECEDING LOOP WITH THIS ONE.
!
!                             do j=1_i4b, n
!                                 tau = d2(j)           
!                                 tmp1 = tau - th - e2(j)*(tmp2/tmp1 - one)
!                                 if ( tmp1>=gamma ) then
!                                     num = num + 1_i4b
!                                     if ( (num>=nsing) .or. (n-j+num<k) ) exit
!                                 else
!                                     tmp1 = min( tmp1, -gamma )
!                                 end if
!                                 tmp2 = tau
!                             end do
!
                        else
!
                            num = n
!
                        end if
!
!                       INSURE THAT COUNT(th) IS MONOTONE.
!
                        nth = min( max(num,nz),ny ) 
!
!                       KEEP ONLY THE INTERVAL CONTAINING w S.T. COUNT(w)=k
!
                        if ( nth<k ) then
!
                            z  = th
                            nz = nth
!
                        else
!
                            y  = th
                            ny = nth
!
!                           REFINE THE REMINDING INTERVALS.
!
                            if ( nsing<=nth ) then
!
                                lb(nsing) = max( th, lb(nsing) )
!
                            else
!
                                lb(nth) = max( th, lb(nth) )
!
                                ub(nth+1_i4b) = min( th, ub(nth+1_i4b) )
!
                            end if
!
                        end if
!
                    end do        
!
!                   SINGULAR VALUE FOUND.
!
                    s(k) = min( max(( y + z )*half,y), z )
!
                    lb(k)  = y
                    nlb(k) = ny
!
                    ub(k)  = z
                    nub(k) = nz
!
                end do
!
#ifdef _OPENMP
!
            end if
!
#endif
!
        end if
!
!       CHECK FOR CONVERGENCE.
!
        failure = false
        th = one/scalfac
!
        do k=1_i4b, nsing
!
            tmp1 = s(k)
!
            if ( lb(k)>ub(k)   .or.  &
                nub(k)>=nlb(k) .or.  & 
                tmp1>=gu       .or.  & 
                tmp1<gl  )   then
!
!               FLAG THE INCORRECT SINGULAR VALUES.
!
                tmp2    = - sqrt( abs( tmp1 ) )
!
                failure = true
!
            else
!
                tmp2 = sqrt( tmp1 )
!
            end if
!
!           UNSCALED THE SINGULAR VALUES.
!
            s(k) = th*tmp2
!
#ifdef _F2003
            if ( ieee_support_datatype( s(k) ) ) then
!
                if ( .not.ieee_is_normal( s(k) ) )  failure = true
!
            end if
#else
            if ( is_nan( s(k) ) ) then
!
                failure = true
!
            end if
#endif
!
        end do
!
!       ORDER THE nsing SINGULAR VALUES.
!
        if ( present(sort) ) then
            call singval_sort( sort, s(:nsing) )
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( lb, nlb, ub, nub )
!
    end if
!
!   FLAG THE OTHER SINGULAR VALUES IF NECESSARY.
!
    if ( nsing/=n ) then
        th = nan()
        s(nsing+1_i4b:n) = th
    end if
!
!
! END OF SUBROUTINE bd_singval2
! _____________________________
!
    end subroutine bd_singval2
!
! =========================================================================================
!
    subroutine bd_max_singval( d, e, nsing, s, failure, abstol, scaling )
!
! Purpose
! _______
!
!   BD_MAX_SINGVAL computes the greatest singular value of a real n-by-n
!   (upper or lower) bidiagonal matrix B by a bisection algorithm.
!
!   The Singular Value Decomposition of a bidiagonal matrix B is:
!
!                           B = Q * S * P'
!
!   where S is a diagonal matrix with non-negative diagonal elements (the singular
!   values of B), and, Q and P are orthogonal matrices (P' denotes the transpose of P).
!
!
! Arguments
! _________
!
!   D            (INPUT) real(stnd), dimension(:)
!                On entry, D contains the diagonal elements of the bidiagonal
!                matrix B.
!
!   E            (INPUT) real(stnd), dimension(:)
!                On entry, E contains the off-diagonal elements of the bidiagonal
!                matrix whose singular values are desired. E(1) is arbitrary.
!
!                The size of E must verify: size( E ) = size( D ) . 
!
!   NSING        (OUTPUT) integer(i4b)
!                On output, NSING specifies the number of singular values which have
!                been computed. Note that NSING may be greater than 1 if multiple singular values 
!                make unique selection of the greatest singular value impossible.
!
!   S            (OUTPUT) real(stnd), dimension(:)
!                On exit, S(1:NSING)  contains the first NSING  greatest singular values of B.
!                The other values in S ( S(NSING+1:size(D)) ) are flagged  by a quiet NAN.
!
!                The size of S must verify: size( S ) = size( D ) . 
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection algorithm converged for all
!                  the computed singular values to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the singular values failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic. The sign of the incorrect singular values is set to negative.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the singular values. A singular value
!                (or cluster) is considered to be located if its square has been determined to lie in
!                an interval whose width is ABSTOL or less.
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | B' * B | will be used,
!                where | B' * B | means the 1-norm of the tridiagonal matrix B' * B ( B' means the transpose of B) 
!                and ULP is the machine precision (distance from 1 to the next larger floating point number).
!
!                Singular values will be computed most accurately when ABSTOL is set to the square root of the
!                underflow threshold, sqrt(LAMCH('S')), not zero. 
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix B is scaled before computing
!                the greatest singular values.
!
!                The default is to scale the bidiagonal matrix.
!
!
! Further Details
! _______________
!
!   BD_MAX_SINGVAL computes the largest singular value of B by a bisection method (see the reference (1)
!   below, Sec.8.5 ).
!
!   The bisection method is applied (implicitly) to the associated n-by-n symmetric tridiagonal matrix
!   B' * B  whose eigenvalues are the squares of the singular values of B by using the differential stationary
!   form of the qd algorithm of Rutishauser (see the reference (2) below, Sec.3.1 ).
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert_eq, merror
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#else
    use Num_Constants,   only : is_nan
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),  dimension(:) :: d, e
    real(stnd), intent(out), dimension(:) :: s
    real(stnd), intent(in), optional      :: abstol
!
    integer(i4b), intent(out)          :: nsing
!
    logical(lgl), intent(in), optional :: scaling
    logical(lgl), intent(out)          :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: gamma, omega, tau, scalfac, maxd, maxe, maxde, maxbd, tnorm, atol1, rtol1, &
                  tmp1, tmp2, y, z, th, gu, gl
    real(stnd),  dimension(size(d))             :: d2, e2
    real(stnd),  dimension(:),      allocatable :: lb, ub
!
    integer(i4b)                             :: n, num, nth, nz, ny, ngu, ngl, i, j, j1, j2, k
    integer(i4b),  dimension(:), allocatable :: nlb, nub
!
    integer  :: iok
!
    logical(lgl) :: do_scaling
!
!
! PARAMETERS
! __________
!
!   relfac IS THE RELATIVE TOLERANCE.  AN INTERVAL (a,b] LIES WITHIN "RELATIVE TOLERANCE" IF
!   b-a < relfac*ulp*max(|a|,|b|), WHERE "ulp" IS THE MACHINE PRECISION.
!
    real(stnd),  parameter :: relfac=1.95_stnd
!
!   A "FUDGE FACTOR" TO WIDEN THE UPPER BOUND OF THE GERSHGORIN INTERVAL. IDEALLY, A VALUE OF 1 SHOULD WORK,
!   BUT ON MACHINES WITH SLOPPY ARITHMETIC, THIS NEEDS TO BE LARGER. 
!   THE DEFAULT FOR PUBLICLY RELEASED VERSIONS SHOULD BE LARGE ENOUGH TO HANDLE THE WORST MACHINE AROUND.
!   NOTE THAT THIS PARAMETER HAS NO EFFECT ON ACCURACY OF THE SOLUTION.
!
    real(stnd),  parameter :: fudge=two
!
    character,   parameter :: sort='d'
!
    character(len=*),  parameter :: name_proc='bd_max_singval'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure    = false
    do_scaling = true
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b),    &
                   int(size(e),i4b),    & 
                   int(size(s),i4b),    & 
                   name_proc )
!
    nsing = 1_i4b
!
    if ( present(abstol) ) then
        atol1 = abstol
    else
        atol1 = -one
    end if
!
    if ( present(scaling) ) then
        do_scaling = scaling
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
            nsing    = 1_i4b
            s(1_i4b) = abs( d(1_i4b) )
        else
            nsing = 0_i4b
        end if
!
        return
!
    end if
!
    d2(1_i4b:n) =  abs( d(1_i4b:n) )
    maxd = maxval( d2(1_i4b:n) )
!
    e2(1_i4b)   = zero
    e2(2_i4b:n) =  abs( e(2_i4b:n) )
    maxe = maxval( e2(2_i4b:n) )
!
    maxde = maxval( d2(1_i4b:n-1_i4b)*e2(2_i4b:n) )
!
    maxbd = max( maxd, maxe )
!
    if ( maxbd<=zero ) then
!
        nsing      = n
        s(1_i4b:n) = zero
!
       return
!
    end if
!
!   GET MACHINE PARAMETERS.
!
    rtol1 = ulp*relfac
!
    gamma  = sqrt( safmin )
!
    if ( do_scaling ) then
!
#ifdef _SCALE
        omega  = scale(one,maxexp-1_i4b)
#else
        omega  = machbase**(maxexp-1_i4b)
#endif
        tau    = sqrt( omega*gamma )
!
!       COMPUTE SCALING FACTOR TO ESTIMATE THE (TINY) SINGULAR VALUES
!       TO RELATIVELY HIGH ACCURACY AND TO AVOID OVERFLOW.
!
        j  = exponent( tau )
        j1 = j - exponent( maxbd ) - 1_i4b
        j2 = ( j - exponent( maxde ) - 1_i4b )/2_i4b
        j  = min( j1, j2 )
        scalfac = machbase**j
!
    else
!
        scalfac = one
!
    end if
!
!   SCALE AND SQUARE THE OFF DIAGONAL ELEMENTS OF THE GOLUB-KAHAN TRIDIAGONAL MATRIX.
!   IF ANY ELEMENT OF abs(d2(:)) OR abs(e2(:)) IS LESS THAN gamma/scalfac THAN THE CORRESPONDING
!   SCALED AND SQUARED ELEMENTS WILL UNDERFLOW TO zero AND HIGH RELATIVE ACCURACY WILL BE LOST.
!
    tmp1 = gamma/scalfac
!
    where( d2(1_i4b:n)>=tmp1 )
        d2(1_i4b:n) = (scalfac*d2(1_i4b:n))**2
    elsewhere
        d2(1_i4b:n) = safmin
    end where
!
    where( e2(2_i4b:n)>=tmp1 )
        e2(2_i4b:n) = (scalfac*e2(2_i4b:n))**2
    elsewhere
        e2(2_i4b:n) = safmin
    end where
!
!   COMPUTE THE UPPER GERSCHGORIN BOUND FOR EIGENVALUES OF B' * B.
!
    gu   = zero
    tmp1 = zero
!
    do i=1_i4b, n-1_i4b
!
        y    = d2(i) + e2(i)
        tmp2 = sqrt( d2(i) )*sqrt( e2(i+1_i4b) )
        gu   = max( gu, y+tmp1+tmp2 )
        tmp1 = tmp2
!
    end do
!
    y    = d2(n) + e2(n)
    gu   = max( gu, y+tmp1 )
!
    tnorm = gu
    gu    = gu + (real( n, stnd)*fudge + eight)*gu*ulp
    ngu   = 0_i4b
!
!   CHECK ABSOLUTE TOLERANCE AND REPLACE IT BY DEFAULT VALUE IF NECESSARY.
!
    if ( atol1<=zero ) then
        atol1 = ulp*tnorm
    end if
!
    atol1 = max( gamma, atol1 )
!
!   DEFAULT LOWER BOUND FOR THE GREATEST SINGULAR VALUE.
!
!   DETERMINE AN INTERVAL CONTAINING EXACTLY THE DESIRED SINGULAR VALUE.
!
!   FIRST GUESS FOR THE LOWER BOUND gl IF SINGULAR VALUES ARE WELL-SEPARATED.
!
    if ( n==2_i4b ) then
        gl = minval( abs(d(1_i4b:2_i4b)) )
    else
        gl = abs( d(2_i4b) )
    end if
!
    gl = (gl*scalfac)**2
!
!   CHECK QUALITY OF INITIAL ESTIMATE OF gl.
!
    if ( gl>zero ) then
!
        ngl  = 0_i4b
        tmp1 = one
        tmp2 = one
!
        do j=1_i4b, n
!
            tmp2 = gl + e2(j)*(tmp2/tmp1)
            tmp1 = d2(j) - tmp2
!
            if ( tmp1>=gamma ) then
                ngl = ngl + 1_i4b
            else
                tmp1 = min( tmp1, -gamma )
            end if
!
        end do
!
    else
!
        ngl = n
!
    end if
!
    if ( ngl/=1_i4b ) then
!
        if ( ngl>1_i4b ) then
            y  = gl
            ny = ngl
            z  = gu
            nz = 0_i4b
        else
            y  = zero
            ny = n
            z  = gl
            nz = ngl
        end if
!
!       BISECTION METHOD FOR FINDING THE LOWER BOUND gl ON THE SQUARES OF THE
!       LARGEST SINGULAR VALUE OF THE BIDIAGONAL MATRIX.
!
        do
!
            tmp1 = abs( z - y )
            tmp2 = max( abs(z), abs(y) )
!
            if ( ngl==1_i4b .or. tmp1<max(gamma,rtol1*tmp2) .or. nz>=ny ) exit
!
            gl = ( z + y )*half
!
            if ( gl>zero ) then
!
                num  = 0_i4b
                tmp1 = one
                tmp2 = one
!
                do j=1_i4b, n
!
                    tmp2 = gl + e2(j)*(tmp2/tmp1)
                    tmp1 = d2(j) - tmp2
!
                    if ( tmp1>=gamma ) then
                        num = num + 1_i4b
                    else
                        tmp1 = min( tmp1, -gamma )
                    end if
!
                end do
!
            else
!
                num = n
!
            end if
!
            ngl = min( max(num,nz), ny ) 
!
            if ( ngl<1_i4b ) then
                z  = gl
                nz = ngl
            else
                y  = gl
                ny = ngl
            end if
!
        end do
!
        if ( ngl/=1_i4b ) then
!
            gl  = y
            ngl = ny
!
            nsing = ny
!
        end if
!
    end if
!
    if ( nsing>0_i4b ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE UPPER AND LOWER BOUNDS
!       AND SINGULAR VALUES COUNTS ASSOCIATED WITH THE BOUNDS.
!
        allocate( lb(nsing), nlb(nsing), ub(nsing), nub(nsing), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        lb(:nsing) = gl
        ub(:nsing) = gu
!
!       INITIALIZE THE UPPER BOUNDS FOR THE REQUESTED SINGULAR VALUES.
!
        z  = gu
        nz = ngu
!
        do k=1_i4b, nsing
!
!           UPDATE THE LOWER AND UPPER BOUNDS FOR THE SQUARE OF THE kTH SINGULAR VALUE.
!
            y  = gl
            ny = ngl
!
            do i=k, nsing
                if ( lb(i)>gl ) then
                    y  =  lb(i)
                    exit
                end if
            end do
!
            tmp1 = min( z, ub(k) )
!
            if ( tmp1<y ) then
                y  = gl
                z  = gu
            else
                z  = tmp1
            end if
!
            do
!
                tmp1 = abs( z - y )
                tmp2 = max( abs(z), abs(y) )
!
!               CHECK FOR CONVERGENCE.
!
                if ( tmp1<max(rtol1*tmp2,atol1) .or. nz>=ny ) exit
!
!               NEXT BISECTION STEP.
!
                th = ( y + z )*half
!
!               COMPUTE nth=COUNT(th)  THE NUMBER OF SINGULAR VALUES GREATER OR EQUAL
!               TO sqrt(th) BY STURM SEQUENCE.
!
                if ( th>zero ) then
!
                    num  = 0_i4b
                    tmp1 = one
                    tmp2 = one
!
                    do j=1_i4b, n
!
                        tmp2 = th + e2(j)*(tmp2/tmp1)
                        tmp1 = d2(j) - tmp2
!
                        if ( tmp1>=gamma ) then
                            num = num + 1_i4b
                        else
                            tmp1 = min( tmp1, -gamma )
                        end if
!
                    end do
!
                else
!
                    num = n
!
                end if
!
!               INSURE THAT COUNT(th) IS MONOTONE.
!
                nth = min( max(num,nz),ny ) 
!
!               KEEP ONLY THE INTERVAL CONTAINING w S.T. COUNT(w)=k
!
                if ( nth<k ) then
!
                    z  = th
                    nz = nth
!
                else
!
                    y  = th
                    ny = nth
!
!                   REFINE THE REMINDING INTERVALS.
!
                    if ( nsing<=nth ) then
                        lb(nsing)  = max( th, lb(nsing) )
                    else
                        lb(nth)  = max( th, lb(nth) )
                        ub(nth+1_i4b)  = min( th, ub(nth+1_i4b) )
                    end if
!
                end if
!
            end do        
!
!           SINGULAR VALUE FOUND.
!
            s(k) = min( max(( y + z )*half,y), z )
!
            lb(k)  = y
            nlb(k) = ny
!
            ub(k)  = z
            nub(k) = nz
!
        end do
!
!       CHECK FOR CONVERGENCE.
!
        th = one/scalfac
!
        do k=1_i4b, nsing
!
            tmp1 = s(k)
!
            if ( lb(k)>ub(k)   .or.  &
                nub(k)>=nlb(k) .or.  & 
                tmp1>=gu       .or.  & 
                tmp1<gl  )   then
!
!               FLAG THE INCORRECT SINGULAR VALUES.
!
                tmp2    = - sqrt( abs( tmp1 ) )
                failure = true
!
            else
!
                tmp2 = sqrt( tmp1 )
!
            end if
!
!           UNSCALED THE SINGULAR VALUES.
!
            s(k) = th*tmp2
!
#ifdef _F2003
            if ( ieee_support_datatype( s(k) ) ) then
!
                if ( .not.ieee_is_normal( s(k) ) )  failure = true
!
            end if
#else
            if ( is_nan( s(k) ) ) then
!
                failure = true
!
            end if
#endif
!
        end do
!
!       ORDER THE nsing SINGULAR VALUES.
!
        if ( nsing>1_i4b ) then
            call singval_sort( sort, s(:nsing) )
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( lb, nlb, ub, nub )
!
    else
!
        failure = true
!
    end if
!
!   FLAG THE OTHER SINGULAR VALUES IF NECESSARY.
!
    if ( nsing/=n ) then
        th = nan()
        s(nsing+1_i4b:n) = th
    end if
!
!
! END OF SUBROUTINE bd_max_singval
! ________________________________
!
    end subroutine bd_max_singval
!
! =========================================================================================
!
    subroutine bd_lasq1( d, e, failure, maxiter, sort, scaling, ieee, aggdef2, max_win, freq, info )
!
! Purpose
! _______
!
!   BD_LASQ1 computes the singular values of a real n-by-n bidiagonal
!   matrix BD (with diagonal D and off-diagonal E) by the differential quotient
!   difference with shifts (dqds) algorithm and, optionally, with an early
!   aggressive deflation strategy.
!
!   The singular values are computed to high relative accuracy, in the absence of
!   denormalization, underflow and overflow. The algorithm was first
!   presented in the reference (1) and the present implementation is
!   described in the references (2), (3) and (4). The early aggressive deflation
!   strategy is described in the reference (4).
!
!
! Arguments
! _________
!
!   D            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, D contain the diagonal elements of the bidiagonal
!                matrix BD whose SVD is desired.
!
!                On normal exit, D contains the singular values of BD (e.g., if FAILURE is false).
!                On abnormal exit (e.g. if FAILURE is true), D is overwritten only if INFO is
!                equal to 2.
!
!                See description of FAILURE and INFO arguments for further details.
!
!   E            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, E(2:) contain the off-diagonal elements of the bidiagonal
!                matrix BD whose singular values are desired. E(1) is arbitrary.
!
!                On normal exit, E is unchanged, but, on abnormal exit, E(2:) is
!                overwritten if INFO is equal to 2.
!
!                See description of FAILURE and INFO arguments for further details.
!
!                The size of E must verify: size( E ) = size( D ) = n . 
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit;
!                - FAILURE = true  :  indicates that the algorithm did not converge.
!                  The specific reasons for the failure are given by the optional argument
!                  INFO described below if this argument is specified in the call to BD_LASQ1.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of iterations in the inner loop
!                of the dqds algorithm (see BD_LASQ2 and BD_LASQ2_AGGDEF2 subroutines).
!                MAXITER must be greater than 10, otherwise the default value is used.
!
!                The default is 30.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true, the bidiagonal matrix BD is scaled before computing
!                the singular values.
!
!                The default is to scale the bidiagonal matrix.
!
!   IEEE         (INPUT, OPTIONAL) logical(lgl)
!                On entry, logical flag for IEEE or non IEEE arithmetic.
!                When IEEE is set to true, the division in the dqds loop is not protected from
!                incurring an exception (division by zero or overflow) since the powerful feature of
!                arithmetic units conforming to IEEE floating point standard 754 is that computation
!                is not held up by an exception. At the end of the loop, the code tests whether an
!                infinite or an NaN (not a number) occurred and acts appropriately.
!
!                The default is to test inside the subroutine if IEEE arithmetic is supported or not.
!
!   AGGDEF2      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if AGGDEF2=true, early aggressive deflation is used in conjunction with
!                the dqds algorithm.
!
!                The default is to use early aggressive deflation only if the dimension of the input
!                bidiagonal matrix is sufficiently large (e.g., if size(D)>=20000).
!
!   MAX_WIN      (INPUT, OPTIONAL) integer(i4b)
!                On entry, the maximum window size to be used in early aggressive deflation.
!                MAX_WIN must be greater or equal to 5.
!
!                This optional argument has an effect only if AGGDEF2 is equal to true.
!
!                The default value is max( int(sqrt(real(size(D)))), 5).
!
!   FREQ         (INPUT, OPTIONAL) integer(i4b)
!                On entry, the frequency with which early aggressive deflation will be invoked.
!                FREQ must be greater or equal to 1.
!
!                This optional argument has an effect only if AGGDEF2 is equal to true.
!
!                The default value is 16.
!
!   INFO         (OUTPUT, OPTIONAL) integer(i4b)
!                On exit:
!
!                - INFO = 0 :  indicates successful exit;
!                - INFO = 1 :  indicates a failure because
!                  a split was marked by a positive value in E;
!                - INFO = 2 :  indicates that the algorithm did not converge
!                  and that the maximum number of iterations in the inner loop
!                  of the dqds algorithm is exceeded;
!                  On exit, D and E represent a matrix with the same singular values
!                  which the calling subroutine could use to finish the computation,
!                  or even feed back into BD_LASQ1;
!                - INFO = 3 :  indicates that the termination criterion of the outer while
!                  loop of the algorithm (in BD_LASQ2 or BD_LASQ2_AGGDEF2 subroutines) was
!                  not met (e.g., the program created more than n=size(D) unreduced blocks);
!                - INFO = 4 :  indicates that the maximum number of iterations in the inner
!                  loop of the dqds algorithm is exceeded (as if INFO=2), but that a bidiagonal
!                  matrix with the same singular values as the original one cannot be recovered.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the DLASQ1 subroutine in LAPACK version 3.12.0
!   and the codes (e.g., DLASQDD1 subroutine) for performing early
!   aggressive deflation given in the reference (4).
!
!   This subroutine is a variant of the dqds algorithm described
!   in the references (2), (3) and (4). Early aggressive deflation
!   strategy, as described in the reference (4), is included in this version
!   of the dqds algorithm when the optional logical argument AGGDEF2 is used
!   with the value true.
!
!   For further details on the dqds algorithm, see:
!
!   (1) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (2) Parlett, B.N., and Marques, O.A., 2000:
!           An implementation of the dqds algorithm (positive case).
!           Linear Algebra Appl., Volume 309, pp. 217-259.
!
!   (3) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!   (4) Nakatsukasa, Y., Aishima, K., and Yamazaki, I., 2012:
!           dqds with aggressive early deflation.
!           SIAM Journal on Matrix Analysis Applications, Volume 33, No 1, pp. 22-51.
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
    use Utilities,     only : assert_eq, lascl
    use Num_Constants, only : test_ieee
!   
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:) :: d, e
!
    integer(i4b), intent(in),  optional :: max_win, freq, maxiter
    integer(i4b), intent(out), optional :: info
!
    logical(lgl), intent(out) :: failure
    logical(lgl), intent(in), optional :: scaling, ieee, aggdef2
!
    character,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: scale, sigmn, sigmx, maxbd
    real(stnd),  dimension(4*size(d)) :: work
!
    integer(i4b) :: n, max_win2, freq2, maxiter2, info2
!
    logical(lgl) :: do_scaling, use_ieee, do_aggdef2
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_lasq1'
!
    character , parameter ::  a_maj=achar(65),  a_min=achar(97),  &
                              d_maj=achar(68),  d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure    = false
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b)  ,    &
                   int(size(e),i4b),      &
                   name_proc )
!
    if ( present(info) ) then
        info = 0_i4b
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=2_i4b ) then
!
        if ( n==1_i4b ) then
!
            d(1_i4b) = abs( d(1_i4b) )
!
        else if ( n==2_i4b ) then
!
            call las2( d(1_i4b), e(2_i4b), d(2_i4b), sigmn, sigmx )
!
            if ( present(sort) ) then
!
                select case( sort )
!
                    case( a_maj, a_min ) 
!
!                       SORT SINGULAR VALUES IN ASCENDING ORDER.
!
                        d(1_i4b) = sigmn
                        d(2_i4b) = sigmx
!
                    case( d_maj, d_min ) 
!
!                       SORT SINGULAR VALUES IN DESCENDING ORDER.
!
                        d(1_i4b) = sigmx
                        d(2_i4b) = sigmn
!
                end select
!
            else
!
                d(1_i4b) = sigmx
                d(2_i4b) = sigmn
!
            end if
!
        end if
!
        return
!
    end if
!
!   TEST OTHER OPTIONAL ARGUMENTS.
!
!   DETERMINE MAXIMUM NUMBER OF ITERATIONS IN THE DQDS LOOP.
!
    if ( present(maxiter) ) then
        maxiter2 = merge( maxiter, 30_i4b, maxiter>=10_i4b )
    else
        maxiter2 = 30_i4b
    end if
!
!   DETERMINE IF THE BIDIAGONAL MATRIX MUST BE SCALED.
!
    if ( present(scaling) ) then
        do_scaling = scaling
    else
        do_scaling = true
    end if
!
    if ( present(ieee) ) then
!
!       DECIDE IF IEEE ARITHMETIC IS USED.
!
        use_ieee = ieee
!
    else
!
!       TEST IF IEEE ARITHMETIC IS SUPPORTED.
!
        use_ieee = test_ieee()
!
    end if
!
!   OPTIONAL ARGUMENTS RELATED TO AGGRESSIVE DEFLATION.
!
    if ( present(aggdef2) ) then
        do_aggdef2 = aggdef2
    else
        do_aggdef2 = n>=20000_i4b
    end if
!
    if ( do_aggdef2 ) then
!
!       DETERMINE WINDOW SIZE.
!
        if ( present(max_win) ) then
            max_win2 = max_win
        else if ( n>=20000_i4b ) then
            max_win2 = 50_i4b
        else
            max_win2 = int( sqrt(real(n,stnd)), i4b)
        end if
!
        max_win2 = max( max_win2, 5_i4b)
!
!       DETERMINE FREQUENCY.
!
        if ( present(freq) ) then
            freq2 = max( freq, 1_i4b)
        else
            freq2 = 16_i4b
        end if
!
    end if
!
!   MAKE A COPY OF THE BIDIAGONAL MATRIX.
!
    work(2_i4b*n+1_i4b:3_i4b*n) = abs( d(1_i4b:n) )
    work(3_i4b*n+1_i4b:4_i4b*n-1_i4b) = abs( e(2_i4b:n) )
!
    sigmx = maxval( work(3_i4b*n+1_i4b:4_i4b*n-1_i4b) )
!
    if ( sigmx==zero ) then
!
!       EARLY RETURN IF sigmx IS ZERO AS MATRIX IS ALREADY DIAGONAL.
!
        if ( present(sort) ) then
            call singval_sort( sort, work(2_i4b*n+1_i4b:3_i4b*n) )
        end if
!
        d(1_i4b:n) = work(2_i4b*n+1_i4b:3_i4b*n)
!
        return
!
    end if
!
!   SCALE THE BIDIAGONAL MATRIX IF REQUIRED.
!   NOTE THAT SQUARING THE INPUT DATA MAKES
!   SCALING BY A POWER OF THE RADIX POINTLESS.
!  
    if ( do_scaling ) then
!
!       ESTIMATE THE LARGEST SINGULAR VALUE.
!
        sigmn = maxval( work(2_i4b*n+1_i4b:3_i4b*n) )
        maxbd = max( sigmn, sigmx )
!
        scale = sqrt( ulp/safmin )
!
        call lascl(  work(2_i4b*n+1_i4b:4_i4b*n-1_i4b), maxbd, scale )
!  
    end if
!  
!   COPY d AND e INTO work(1:2*n-1) IN THE Z FORMAT.
!  
    work(1_i4b:2_i4b*n-1_i4b:2_i4b) = work(2_i4b*n+1_i4b:3_i4b*n)
    work(2_i4b:2_i4b*n-1_i4b:2_i4b) = work(3_i4b*n+1_i4b:4_i4b*n-1_i4b) 
!  
!   SQUARE THE ELEMENTS OF THE BIDIAGONAL MATRIX (E.G., COMPUTE THE Q'S AND E'S).
!  
    work(1_i4b:2_i4b*n-1_i4b) = work(1_i4b:2_i4b*n-1_i4b)**2
    work(2_i4b*n:4_i4b*n) = zero
!  
!   PERFORM DQDS ITERATIONS.
!  
    if ( do_aggdef2 ) then
!
        call bd_lasq2_aggdef2( n, max_win2, freq2, maxiter2, use_ieee, work, info2 )
!  
    else
!  
        call bd_lasq2( n, maxiter2, use_ieee, work, info2 )
!  
    end if
!  
    if ( info2==0_i4b ) then
!  
!       COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!  
        d(1_i4b:n) = sqrt( work(1_i4b:n) )
!
!       ORDER THE SINGULAR VALUES IF REQUIRED.
!
        if ( present(sort) ) then
            call singval_sort( sort, d(1_i4b:n) )
        end if
!      
!       UNSCALED THEM IF REQUIRED.
!      
        if ( do_scaling ) then
            call lascl( d(1_i4b:n), scale, maxbd )
        end if
!      
    else
!      
!       SET FAILURE INDICATOR.
!      
        failure = true
!      
        if ( info2==2_i4b ) then
!      
!           MAXIMUM NUMBER OF ITERATIONS EXCEEDED. MOVE DATA FROM work
!           INTO d AND e SO THE CALLING SUBROUTINE CAN TRY TO FINISH THE
!           COMPUTATIONS.
!      
            d(1_i4b:n) = sqrt( work(1_i4b:2_i4b*n-1_i4b:2_i4b) )
            e(2_i4b:n) = sqrt( work(2_i4b:2_i4b*n-1_i4b:2_i4b) )
!      
            if ( do_scaling ) then
                call lascl( d(1_i4b:n), scale, maxbd )
                call lascl( e(2_i4b:n), scale, maxbd )
            end if
!      
        end if
!      
    end if
!
!   OUTPUT OPTIONAL ARGUMENT IF REQUIRED.
!
    if ( present(info) ) then
        info = info2
    end if
!
!
! END OF SUBROUTINE bd_lasq1
! __________________________
!      
    end subroutine bd_lasq1
!
! ===========================================================================================
!
    subroutine bd_dqds( d, e, failure, maxiter, sort, scaling )
!
! Purpose
! _______
!
!   BD_DQDS computes all the singular values, S, of a real n-by-n 
!   (upper or lower) bidiagonal matrix BD:
!
!                            BD = Q * S * P'
!
!   , where S is a diagonal matrix with non-negative diagonal
!   elements (the singular values of BD), and, Q and P are
!   orthogonal matrices (P' denotes the transpose of P).
!
!   The singular values are computed to high relative precision in the absence
!   of denormalization, underflow and overflow with a variant of the
!   differential quotient difference with shifts (dqds) algorithm described 
!   in the reference (3).
!
!
! Arguments
! _________
!
!   D            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, D contains the diagonal elements of the bidiagonal
!                matrix BD.
!
!                On exit, D contains the singular values of BD if FAILURE is false,
!                but is unchanged if FAILURE is true.
!
!                See description of FAILURE argument for further details.
!
!   E            (INPUT) real(stnd), dimension(:)
!                On entry, E contains the off-diagonal elements of the bidiagonal
!                matrix BD whose singular values are desired. E(1) is arbitrary.
!
!                On exit, E is unchanged.
!
!                The size of E must verify: size( E ) = size( D ) = n . 
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit ;
!                - FAILURE = true  :  indicates that the algorithm did not converge and that
!                  the maximum number of iterations in the inner loop of the algorithm is exceeded.
!                  In that case, D is unchanged on exit.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of iterations in the inner loop
!                of the dqds algorithm. MAXITER must be greater than 5, otherwise the
!                default value is used.
!
!                The default is 30.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix BD is scaled before computing
!                the singular values.
!
!                The default is to scale the bidiagonal matrix.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the variant of the dqds algorithm described
!   in the reference (3), which uses a new shift strategy in the dqds algorithm.
!
!   BD_DQDS is a Fortran95/2003 translation and improved version of a Fortran77
!   subroutine available at:
!
!   - http://syskiso.fuee.u-fukui.ac.jp/~kkimur/LAPROGNC/LAPROGNC.html
!
!   For further details on the dqds algorithm and its implementation, see:
!
!   (1) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (2) Parlett, B.N., and Marques, O.A., 2000:
!           An implementation of the dqds algorithm (positive case).
!           Linear Algebra Appl., Volume 309, pp. 217-259.
!
!   (3) Yamashita, T., Kimura, K., Takata, M., and Nakamura, Y., 2013:
!           An application of the Kato-Temple inequality on matrix
!           eigenvalues to the dqds algorithm for singular values.
!           JSIAM  Letters, Volume 5, pp. 21-24.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
#ifdef _F2003
    use ieee_arithmetic,   only : ieee_support_datatype, ieee_is_normal
#endif
    use Num_Constants,     only : is_nan
    use Utilities,         only : assert_eq, lascl
!   
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:) :: d
    real(stnd),   intent(in),   dimension(:)  :: e
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out) :: failure
    logical(lgl), intent(in), optional :: scaling
!
    character,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7, tmp8, tmp9,      &
                  tau, tau1, tau2, tau3, s, t, scale, maxbd, tol, tol2,      &
                  sigma, sigma1, desig, desig0, scale1, dmin2, dn0, dn1, dn2
    real(stnd),  dimension(4*size(d)) :: work
    real(stnd),  dimension(size(d))   :: work5, work6, work7, work8
!
    integer(i4b) :: n0, n, m, i, j, k, l, p, maxiter2, maxit,  &
                    indrv1, indrv2, indrv3, indrv4
!
    logical(lgl) :: do_scaling, test
#ifdef _F2003
    logical(lgl) :: use_ieee
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_dqds'
!
    real(stnd),  parameter :: const=0.75_stnd
!
    character , parameter ::  a_maj=achar(65),  a_min=achar(97),       &
                              d_maj=achar(68),  d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure    = false
    do_scaling = true
!
!   TEST THE ARGUMENTS.
!
    n0 = assert_eq( int(size(d),i4b)  ,    &
                    int(size(e),i4b),      &
                    name_proc )
!
    if ( present(scaling) ) then
        do_scaling = scaling
    end if
!
    if ( present(maxiter) ) then
        maxiter2 = merge( maxiter, 30_i4b, maxiter>=5_i4b )
    else
        maxiter2 = 30_i4b
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n0<=2_i4b ) then
!
        if ( n0==1_i4b ) then
!
            d(1_i4b) = abs( d(1_i4b) )
!
        else if ( n0==2_i4b ) then
!
            call las2( d(1_i4b), e(2_i4b), d(2_i4b), tmp2, tmp1 )
!
            if ( present(sort) ) then
!
                select case( sort )
!
                    case( a_maj, a_min ) 
!
!                       SORT SINGULAR VALUES IN ASCENDING ORDER.
!
                        d(1_i4b) = tmp2
                        d(2_i4b) = tmp1
!
                    case( d_maj, d_min ) 
!
!                       SORT SINGULAR VALUES IN DESCENDING ORDER.
!
                        d(1_i4b) = tmp1
                        d(2_i4b) = tmp2
!
                end select
!
            else
!
                d(1_i4b) = tmp1
                d(2_i4b) = tmp2
!
            end if
!
        end if
!
        return
!
    end if
!
!   MAKE A COPY OF THE BIDIAGONAL MATRIX.
!
    work(1_i4b:n0) = abs( d(1_i4b:n0) )
    work(n0+1_i4b:2_i4b*n0-1_i4b) = abs( e(2_i4b:n0) )
!
    tmp2 = maxval( work(n0+1_i4b:2_i4b*n0-1_i4b) )
!
    if ( tmp2==zero ) then
!
!       EARLY RETURN IF tmp2 IS ZERO AS MATRIX IS ALREADY DIAGONAL.
!
        if ( present(sort) ) then
            call singval_sort( sort, work(1_i4b:n0) )
        end if
!
        d(1_i4b:n0) = work(1_i4b:n0)
!
        return
!
    end if
!
!   ESTIMATE THE LARGEST SINGULAR VALUE.
!
    tmp1  = maxval( work(1_i4b:n0) )
    maxbd = max( tmp1, tmp2 )
!
!   DETERMINE TOLERANCES AND SCALING FACTORS.
!
    tol = fudge_dqds*ulp
    tol2 = tol**2
!
    scale = sqrt( half/safmin )
    scale1 = one/scale
!
!   TEST IF IEEE ARITHMETIC IS SUPPORTED.
!
#ifdef _F2003
    use_ieee = ieee_support_datatype( dn0 )
#endif
!
!   SCALE THE BIDIAGONAL MATRIX IF REQUIRED.
!  
    if ( do_scaling ) then
!
        call lascl(  work(1_i4b:2_i4b*n0-1_i4b), maxbd, scale )
!  
    end if
!  
!   SQUARE THE ELEMENTS OF THE BIDIAGONAL MATRIX (E.G., COMPUTE THE Q'S AND E'S).
!  
    work(1_i4b:2_i4b*n0-1_i4b) = work(1_i4b:2_i4b*n0-1_i4b)**2
!
!   SET INDICES FOR DQDS ITERATIONS.
!
    indrv1 = 0_i4b
    indrv2 = indrv1 + n0
    indrv3 = indrv2 + n0
    indrv4 = indrv3 + n0
!
    m = 1_i4b
    n = n0
!
!   STORE THE ELEMENTS OF THE BIDIAGONAL MATRIX IN REVERSED ORDER IF
!   THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL IN ORDER
!   TO BOOST OF THE DQDS ITERATIONS.
!
    if ( c1_5*work(1_i4b)<work(n) ) then
!    if ( reverse( work(1_i4b:n) ) ) then
!
!        work8(1_i4b:n) = work(n:1_i4b:-1_i4b)
!        work(1_i4b:n)  = work8(1_i4b:n)
!
!        work8(1_i4b:n-1_i4b)            = work(indrv3-1_i4b:indrv2+1_i4b:-1_i4b)
!        work(indrv2+1_i4b:indrv3-1_i4b) = work8(1_i4b:n-1_i4b)
!
        k=n/2_i4b
!
        do j=1_i4b, k
!
            tmp1            = work(j)
            work(j)         = work(n+1_i4b-j)
            work(n+1_i4b-j) = tmp1
!
            tmp1             = work(indrv2+j)
            work(indrv2+j)   = work(indrv2+n-j)
            work(indrv2+n-j) = tmp1
!
        end do
!
    end if
!
    tmp1 = work(1_i4b)
!
    do j=1_i4b, n-3_i4b
!
        if ( work(indrv2+j)<=tol2*tmp1 ) then
!
            work(indrv2+j) = -zero
            work(indrv4+j) = -zero
!
            m = j + 1_i4b
!
            tmp1 = work(j+1_i4b)
!
        else
!
            tmp1 = work(j+1_i4b)*( tmp1/( tmp1 + work(indrv2+j) ) )
!
        end if
!
    end do
!
    if ( work(indrv2+n-2_i4b)<=tol2*tmp1 ) then
         work(indrv2+n-2_i4b) = zero
         tmp1                 = work(n-1_i4b)
    else
         tmp1 = work(n-1_i4b)*( tmp1/( tmp1 + work(indrv2+n-2_i4b) ) )
    end if
!
    if ( work(indrv2+n-1_i4b)<=tol2*tmp1 ) then
         work(indrv2+n-1_i4b) = zero
    end if
!
    work(indrv2+n) = zero
    work(indrv4+n) = zero
!
!   DQDS ITERATIONS.
!
    dqds_iter : do
!
        sigma  = -work(indrv2+n)
        sigma1 = tol2*sigma
        desig  = -work(indrv4+n)
!
        maxit = maxiter2*( n - m + 1_i4b )
!
        do i=1_i4b, maxit
!
            do
!
                if ( n==m .or. n==m+1_i4b ) then
!
                    if ( n==m+1_i4b ) then
!
                        call sub_2( indrv1, indrv2 )
!
                    else
!
                        call dqds2( sigma, desig, work(n), t, desig0 )
!
                        work(n) = t
!
                    end if
!
                    n = m - 1_i4b
!
                    if ( n<=0_i4b ) then
                        exit dqds_iter
                    end if
!  
                    do j=m-2_i4b, 1_i4b, -1_i4b
!  
                        if ( work(indrv2+j)<=zero ) exit
!  
                    end do
!  
                    m = j + 1_i4b
!
                    cycle dqds_iter
!  
                end if
!  
                if ( work(indrv2+n-2_i4b)<=sigma1 ) then
!
                    call sub_2( indrv1, indrv2 )
!
                    n = n - 2_i4b
!
                    cycle     
!
                end if
! 
                if ( work(indrv2+n-1_i4b)<=sigma1 ) then
!            
                    call dqds2( sigma, desig, work(n), t, desig0 )
!
                    work(n) = t
!
                    n = n - 1_i4b
!
                    cycle     
!
                end if
!
                exit
!
            end do
!
!           STORE THE ELEMENTS OF THE BIDIAGONAL MATRIX IN REVERSED ORDER IF
!           THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL IN ORDER
!           TO BOOST OF THE DQDS ITERATIONS.
!         
!            if ( reverse( work(m:n) ) ) then
            if ( c1_5*work(m)<work(n) ) then
!
!                work8(m:n) = work(n:m:-1_i4b)
!                work(m:n)  = work8(m:n)
!
!                work8(m:n-1_i4b)              = work(indrv2+n-1_i4b:indrv2+m:-1_i4b)
!                work(indrv2+m:indrv2+n-1_i4b) = work8(m:n-1_i4b)
!         
                k = ( n - m + 1_i4b )/2_i4b + m - 1_i4b
!         
                do j=m, k
!         
                    tmp1        = work(j)
                    work(j)     = work(n+m-j)
                    work(n+m-j) = tmp1
!         
                    tmp1                     = work(indrv2+j)
                    work(indrv2+j)           = work(indrv2+n-1_i4b+m-j)
                    work(indrv2+n-1_i4b+m-j) = tmp1
!         
                end do
!         
            end if
!
            tau  = work(n)
            tmp3 = work(n-1_i4b)
!         
            if ( tau>tmp3 ) then
!         
                s    = tau
                tau  = tmp3
                tmp3 = s
!         
            end if
!         
            t = half*( ( tmp3 - tau ) + work(indrv2+n-1_i4b) )
!         
            if ( work(indrv2+n-1_i4b)>tau*tol2 .and. t/=zero ) then
!         
                s = tau*( work(indrv2+n-1_i4b)/t )
!         
                if( s<=t ) then
                   s = tau*( work(indrv2+n-1_i4b)/( t*( one + sqrt( one + s/t ) ) ) )
                else
                   s = tau*( work(indrv2+n-1_i4b)/( t + sqrt( t )*sqrt( t + s ) ) )
                end if
!         
                t    = tmp3 + ( s + work(indrv2+n-1_i4b) )
                tau  = tau*( tmp3/t )
                tmp3 = t
!         
            end if
!         
            do         
!
                tau  = min( tau, work(n) )
                tau2 = minval( work(m:n-1_i4b) )
!         
                if ( min(tau,tau2)==zero ) then
                    call sub_1()
                    exit
                end if
!         
                test = true
!         
                do         
!         
                    if ( tau2<=tau ) then
!         
                        tau1 = tau2
!         
                        exit
!         
                    else
                        tau1 = tau
                    end if
!
                    dn2 = work(m) - tau
!         
                    if ( dn2<=zero ) exit
!         
                    do j=m, n-3_i4b
!         
                        dn2 = dn2*( work(j+1_i4b)/( dn2 + work(indrv2+j) ) ) - tau
!
                        if ( dn2<=zero ) exit
!
                    end do
!
                    if ( j>n-3_i4b ) exit
!
                    tmp1 = dn2 + work(indrv2+n-2_i4b)
!
                    if ( safmin*tmp1<work(n-1_i4b) .and. safmin*work(n-1)<tmp1 ) then
                        dn1 = dn2*( work(n-1_i4b)/tmp1 ) - tau
                    else
                        dn1 = work(n-1_i4b)*( dn2/tmp1 ) - tau
                    end if
!
                    if ( dn1<=zero ) exit
!
                    tmp1 = dn1 + work(indrv2+n-1_i4b)
!
                    if ( safmin*tmp1<work(n) .and. safmin*work(n)<tmp1 ) then
                        dn0 = dn1*( work(n)/tmp1 ) - tau
                    else
                        dn0 = work(n)*( dn1/tmp1 ) - tau
                    end if
!
                    if( dn0<zero .and. tau+dn0==tau ) dn0 = zero
!
                    if ( dn0<zero ) then
!
                        tau2 = max( dn0+tau, zero )
                        if ( tau2==tau ) tau2 = zero
!
                        tau3 = const*tau
                        if ( tau3==tau ) tau3 = half*tau
!
                        tau = max( tau2, tau3 )
!
                    end if
!
                    test = tau<=zero
!
                    exit
!         
                end do
!
                if ( test ) then
!
                    tmp7  = scale/work(m)
                    tmp1  = tmp7
                    tmp5  = tmp7**2
                    tmp3  = tmp5
                    tmp2  = tmp3
                    dmin2 = tmp3
!
                    do j=m+1_i4b, n
!
                        tmp4  = scale/work(j)
                        tmp8  = ( work(indrv2+j-1_i4b)*scale1 )*tmp4
                        tmp7  = tmp8*tmp7 + tmp4
                        tmp1  = tmp1 + tmp7
                        tmp6  = tmp7**2
                        tmp3  = tmp8*( tmp3 + tmp5) + tmp6
                        tmp2  = tmp2 + tmp3
                        dmin2 = max( dmin2, tmp3 )
                        tmp5  = tmp6
!
                    end do
!
                    tau  = max( zero, scale/tmp1, scale/sqrt(tmp2) )
                    tmp8 = real( n-m+1_i4b, stnd )
                    tmp9 = scale*( tmp8/( tmp1 + sqrt( tmp8-one )*sqrt( tmp8*tmp2 - tmp1**2 ) ) )
!
                    if ( tmp9==tmp9 ) then
                        tau = max( tau, tmp9 )
                    end if
!
                    tau2 = tau1
                    tmp8 = ceiling( (tmp1/tmp2)*tmp1 )
!
                    if ( tmp8>one ) then
!
                        tmp9 = scale*( tmp8/( tmp1 + sqrt( tmp8*tmp2 - tmp1**2 )/sqrt( tmp8 - one ) ) )
!
                        if ( tmp9==tmp9 ) then
                            tau2 = min( tmp9, scale/sqrt( dmin2 ), tau2 )
                        else
                            tau2 = min( scale/sqrt( dmin2 ), tau2 )
                        end if
!
                    else
!
                        tau2 = min( scale/sqrt( dmin2 ), tau2 )
!
                    end if
!
                    tau = min( tau, tau2 )
!
                    if ( tau2>=two*tau ) then
!
                        work7(m:n)       = one/sqrt( work(m:n) )
                        work8(m:n-1_i4b) = sqrt( work(indrv2+m:indrv2+n-1_i4b) )
!         
                        work5(n) = work7(n)
                        tmp3     = work5(n)
!
                        do j=n-1_i4b, m, -1_i4b
!
                            work(indrv3+j) = work8(j)*work7(j)
                            work5(j)       = work7(j) + work(indrv3+j)*work5(j+1_i4b)
                            tmp3           = max( tmp3, work5(j) )
!
                        end do
!
                        tmp3     = one/tmp3
                        work5(m) = (work5(m)*tmp3)*work7(m)
                        tmp1     = work5(m)
!
                        do j=m+1_i4b, n
!
                            work(indrv4+j) = work8(j-1_i4b)*work7(j)
                            work5(j)       = ( work5(j)*tmp3 )*work7(j) + work(indrv4+j)*work5(j-1_i4b)
                            tmp1           = max( tmp1, work5(j) )
!
                        end do
!
                        tau = zero
!
                        if ( tmp1>zero ) then
                            tau = max( tau, tmp3/tmp1 )
                        end if
!
                        tmp1     = one/tmp1
                        work5(n) = work5(n)*tmp1
                        work6(n) = work5(n)*work7(n)
                        tmp3     = work6(n)
!
                        do j=n-1_i4b, m, -1_i4b
!
                            work5(j) = work5(j)*tmp1
                            work6(j) = work5(j)*work7(j) + work(indrv3+j)*work6(j+1_i4b)
                            tmp3     = max( tmp3, work6(j) )
!
                        end do
!
                        tmp3 = one/tmp3
                        tmp2 = ( work6(m)*tmp3 )*work7(m)
                        tmp1 = tmp2/work5(m)
!
                        do j=m+1_i4b, n
!
                            tmp2 = ( work6(j)*tmp3 )*work7(j) + work(indrv4+j)*tmp2
                            tmp1 = max( tmp1, tmp2/work5(j) )
!
                        end do
!
                        if ( tmp1>zero ) then
                            tau = max( tau, tmp3/tmp1 )
                        end if
!
                        tau = min( tau, tau1 )
!
                        if ( tau<=zero ) then
!
                            tau = sqrt( work(n) ) - half*work8(n-1_i4b)
!
                            if ( tau<=zero ) then
                                call sub_1()
                                exit
                            end if
!
                            do j=n-1_i4b, m+1_i4b, -1_i4b
!
                                tmp2 = sqrt( work(j) ) - half*( work8(j) + work8(j-1_i4b) )
!
                                if ( tmp2<=zero ) then
                                    call sub_1()
                                    exit
                                end if
!
                                tau = min( tau, tmp2 )
!
                            end do
!
                            if ( j>m ) exit
!
                            tmp2 = sqrt( work(m) ) - half*work8(m)
!
                            if ( tmp2<=zero )  then
                                call sub_1()
                                exit
                            end if
!
                            tau = min( tau, tmp2 )**2
!
                        end if
!
                    end if
!
                end if
!
                do
!
                    if ( tau==zero ) then
                        call sub_1()
                        exit
                    end if
!
                    dn2   = work(m) - tau
                    dmin2 = dn2
!
                    do j=m, n-3_i4b
!
                        work(indrv3+j) = dn2 + work(indrv2+j)
                        tmp3           = work(j+1_i4b)/work(indrv3+j)
                        work(indrv4+j) = work(indrv2+j)*tmp3
                        dn2            = dn2*tmp3 - tau
                        dmin2          = min( dmin2, dn2 )
!
                    end do
!
                    if ( dmin2<=zero ) then 
!
                        tau3 = const*tau
                        if ( tau3==tau ) tau3 = half*tau
                        tau = tau3
!
                        cycle
!
                    end if
!
                    work(indrv3+n-2_i4b) = dn2 + work(indrv2+n-2_i4b)
!
                    if ( safmin*work(indrv3+n-2_i4b)<work(n-1_i4b) .and. safmin*work(n-1_i4b)<work(indrv3+n-2_i4b) ) then
                        tmp3                 = work(n-1_i4b)/work(indrv3+n-2_i4b)
                        work(indrv4+n-2_i4b) = work(indrv2+n-2_i4b)*tmp3
                        dn1                  = dn2*tmp3 - tau
                    else
                        work(indrv4+n-2_i4b) = work(n-1_i4b)*( work(indrv2+n-2_i4b)/work(indrv3+n-2_i4b) )
                        dn1                  = work(n-1_i4b)*( dn2/work(indrv3+n-2_i4b) ) - tau
                    end if
!
                    if ( dn1<=zero ) then
!
                        tau2 = max( dn1+tau, zero )
                        if ( tau2==tau ) tau2 = zero
!
                        tau3 = const*tau
                        if ( tau3==tau ) tau3 = half*tau
!
                        tau = max( tau2, tau3 )
!
                        cycle
!
                    end if
!
                    work(indrv3+n-1_i4b) = dn1 + work(indrv2+n-1_i4b)
!
                    if ( safmin*work(indrv3+n-1_i4b)<work(n) .and.  safmin*work(n)<work(indrv3+n-1_i4b) ) then
                        tmp3                 = work(n)/work(indrv3+n-1_i4b)
                        work(indrv4+n-1_i4b) = work(indrv2+n-1_i4b)*tmp3
                        dn0                  = dn1*tmp3 - tau
                    else
                        work(indrv4+n-1_i4b) = work(n)*( work(indrv2+n-1_i4b)/work(indrv3+n-1_i4b) )
                        dn0                  = work(n)*( dn1/work(indrv3+n-1_i4b) ) - tau
                    end if
!
                    if ( dn0<zero .and. tau+dn0==tau ) dn0 = zero
!
                    if ( dn0<zero ) then
!
                        tau2 = max( dn0+tau, zero )
                        if ( tau2==tau ) tau2 = zero
!
                        tau3 = const*tau
                        if ( tau3==tau ) tau3 = half*tau
!
                        tau = max( tau2, tau3 )
!
                        cycle
!
                    end if
!
                    if ( dn0>=work(n) ) then
!
                        tmp3 = dn1/work(indrv3+n-1_i4b) - tau/work(n)
!
                        if ( tmp3>=zero ) then
                            dn0 = work(n)*tmp3
                        end if
!
                    end if
!        
                    work(indrv3+n)=dn0
!
#ifdef _F2003
                    if ( use_ieee ) then
                        test = .not.ieee_is_normal( dn0 )
                    else
#endif
                        test = is_nan( dn0 )
#ifdef _F2003
                    end if
#endif
!
                    if ( test ) then
!
                        tau3 = const*tau
                        if ( tau3==tau ) tau3 = half*tau
                        tau = tau3
!
                        cycle
!
                    end if
!
                    call dqds2( sigma, desig, tau, t, desig0 )
!
                    sigma = t
                    sigma1 = tol2*sigma
                    desig = desig0
!
                    exit
!
                end do
!
                exit
!
            end do
!
            if ( work(indrv4+n-2_i4b)<=sigma1 .or. work(indrv2+n-2_i4b)<=tol2*work(indrv3+n-2_i4b) ) then
!
                call sub_2( indrv3, indrv4 )
!            
                n = n - 2_i4b
!            
                if ( n==m .or. n==m+1_i4b ) then
!            
                    if ( n==m+1_i4b ) then
!
                        call sub_2( indrv3, indrv4 )
!
                    else
!
                        call dqds2( sigma, desig, work(indrv3+n), t, desig0 )
!
                        work(n) = t
!
                    end if
!
                    n = m - 1_i4b
!
                    if ( n<=0_i4b ) then
                        exit dqds_iter
                    end if
!  
                    do j=m-2_i4b, 1_i4b, -1_i4b
!  
                        if ( work(indrv2+j)<=zero ) exit
!  
                    end do
!  
                    m = j + 1_i4b
!
                    cycle dqds_iter
!
                end if
!            
            end if
!
            if ( work(indrv4+n-1_i4b)<=sigma1 .or. work(indrv2+n-1_i4b)<=tol2*work(indrv3+n-1_i4b) ) then
!
                call dqds2( sigma, desig, work(indrv3+n), t, desig0 )
!
                work(n) = t
!
                n = n - 1_i4b
!
                if ( n==m .or. n==m+1_i4b ) then
!
                    if ( n==m+1_i4b ) then
!
                        call sub_2( indrv3, indrv4 )
!
                    else
!
                        call dqds2( sigma, desig ,work(indrv3+n), t, desig0 )
!
                        work(n) = t
!
                    end if
!
                    n = m - 1_i4b
!
                    if ( n<=0_i4b ) then
                        exit dqds_iter
                    end if
!
                    do j=m-2_i4b, 1_i4b, -1_i4b
!
                        if ( work(indrv2+j)<=zero ) exit
!
                    end do
!
                    m = j + 1_i4b
!
                    cycle dqds_iter
!
                end if
!
            end if
!
            dn2   = work(indrv3+m)
            dmin2 = dn2
!
            do j=m, n-3_i4b
!
                work(j)        = dn2 + work(indrv4+j)
                tmp2           = work(indrv3+j+1_i4b)/work(j)
                work(indrv2+j) = work(indrv4+j)*tmp2
                dn2            = dn2*tmp2
                dmin2          = min( dmin2, dn2 )
!
            end do
!
            call sub_3( indrv1, indrv2, indrv3, indrv4 )
         
            do p=1_i4b, 2_i4b
!
#ifdef _F2003
                if ( use_ieee ) then
                    test = .not.ieee_is_normal( dn2 )
                else
#endif
                    test = is_nan( dn2 )
#ifdef _F2003
                end if
#endif
                if ( test ) then
!
                    dn2 = work(indrv3+m)
!
                    if ( sigma==dn2+sigma ) dn2 = zero
!
                    do j=m, n-3_i4b
!
                        work(j) = dn2 + work(indrv4+j)
!
                        if ( work(j)==zero ) then
                            work(indrv2+j) = zero
                            dn2            = work(indrv3+j+1_i4b)
                        else if ( safmin*work(j)<work(indrv3+j+1_i4b) .and. safmin*work(indrv3+j+1_i4b)<work(j) ) then
                            tmp2           = work(indrv3+j+1_i4b)/work(j)
                            work(indrv2+j) = work(indrv4+j)*tmp2
                            dn2             = dn2*tmp2
                        else
                            work(indrv2+j) = work(indrv3+j+1_i4b)*( work(indrv4+j)/work(j) )
                            dn2            = work(indrv3+j+1_i4b)*( dn2/work(j) )
                        end if
!
                        if ( sigma==dn2+sigma ) dn2 = zero
!
                    end do
!
                    call sub_3( indrv1, indrv2, indrv3, indrv4 )
!
                    exit       
!
                end if
!
                if ( dmin2+sigma/=sigma .or. p==2_i4b ) exit
!
                dn2 = work(indrv3+m)
!
                if ( sigma==dn2+sigma ) dn2 = zero
!
                do j=m, n-3_i4b
!
                    work(j)        = dn2 + work(indrv4+j)
                    tmp2           = work(indrv3+j+1_i4b)/work(j)
                    work(indrv2+j) = work(indrv4+j)*tmp2
                    dn2            = dn2*tmp2
!
                    if ( sigma==dn2+sigma ) dn2 = zero
!
                end do
!         
                call sub_3( indrv1, indrv2, indrv3, indrv4 )
!             
            end do
!
            do j=m, n-3_i4b
!
                if ( work(indrv2+j)<=sigma1 .or. work(indrv4+j)<=tol2*work(j) ) then
!
                    work(indrv2+j) = -sigma
                    work(indrv4+j) = -desig
!
                    m = j + 1_i4b
!
                end if
!
            end do
!
            if ( work(indrv4+n-2_i4b)<=tol2*work(n-2_i4b) ) then
                work(indrv2+n-2_i4b) = zero
            end if
!
            if ( work(indrv4+n-1_i4b)<=tol2*work(n-1_i4b) ) then
                work(indrv2+n-1_i4b) = zero
            end if
!
        end do
!      
!       MAXIMUM NUMBER OF ITERATIONS EXCEEDED, SET FAILURE INDICATOR AND EXIT.
!      
        failure = true
!      
        exit dqds_iter
!
    end do dqds_iter
!  
    if ( .not.failure ) then
!  
        d(1_i4b:n0) = sqrt( work(1_i4b:n0) )
!
!       ORDER THE SINGULAR VALUES IF REQUIRED.
!
        if ( present(sort) ) then
            call singval_sort( sort, d(1_i4b:n0) )
        end if
!      
        if ( do_scaling ) then
            call lascl( d(1_i4b:n0), scale, maxbd )
        end if
!      
    end if
!                                                                                    
! ___________________________________________________________________________________________
!
                             contains
!                            ________
!
! ___________________________________________________________________________________________
!
        subroutine sub_1(  )
!
!
!   EXECUTABLE STATEMENTS
!   ---------------------
!
!
        dn2 = work(m)
!
        if ( sigma==dn2+sigma ) dn2 = zero
!
        do l=m, n-3_i4b
!
            work(indrv3+l) = dn2 + work(indrv2+l)
            tmp3           = work(l+1_i4b)/work(indrv3+l)
            work(indrv4+l) = work(indrv2+l)*tmp3
            dn2            = dn2*tmp3
!
            if ( sigma==dn2+sigma ) dn2 = zero
!
        end do
!
        call sub_3( indrv3, indrv4, indrv1, indrv2 )
!
#ifdef _F2003
        if ( use_ieee ) then
            test = .not.ieee_is_normal( dn2 )
        else
#endif
            test = is_nan( dn2 )
#ifdef _F2003
        end if
#endif
        if ( test ) then
!
            dn2 = work(m)
!
            if ( sigma==dn2+sigma ) dn2 = zero
!
            do l=m, n-3_i4b
!
                work(indrv3+l) = dn2 + work(indrv2+l)
!
                if ( safmin*work(indrv3+l)<work(l+1_i4b) .and. safmin*work(l+1_i4b)<work(indrv3+l) ) then
                    tmp2           = work(l+1_i4b)/work(indrv3+l)
                    work(indrv4+l) = work(indrv2+l)*tmp2
                    dn2            = dn2*tmp2
                else
                    work(indrv4+l) = work(l+1_i4b)*( work(indrv2+l)/work(indrv3+l) )
                    dn2            = work(l+1_i4b)*( dn2/work(indrv3+l) )
                end if
!
                if ( sigma==dn2+sigma ) dn2 = zero
!
            end do
!
            call sub_3( indrv3, indrv4, indrv1, indrv2 )
!
        end if
!
!
!   END OF SUBROUTINE sub_1
!   -----------------------
!
        end subroutine sub_1
!
! ___________________________________________________________________________________________
!
        subroutine sub_2( i1, i2 )
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ----------------------------
!
        integer(i4b), intent(in) :: i1, i2
!
!
!   EXECUTABLE STATEMENTS
!   ---------------------
!
!
        if ( work(i1+n)>work(i1+n-1_i4b) ) then
            s                = work(i1+n)
            work(i1+n)       = work(i1+n-1_i4b)
            work(i1+n-1_i4b) = s
        end if
!
        t = half*( ( work(i1+n-1_i4b) - work(i1+n) ) + work(i2+n-1_i4b) )
!
        if ( work(i2+n-1_i4b)>work(i1+n)*tol2 .and. t/=zero ) then
!
            s = work(i1+n)*( work(i2+n-1_i4b)/t )
!
            if ( s<=t ) then
                s = work(i1+n)*( work(i2+n-1_i4b)/( t*( one + sqrt( one + s/t ) ) ) )
            else
                s = work(i1+n)*( work(i2+n-1_i4b) / ( t + sqrt( t )*sqrt( t + s ) ) )
            end if
!
            t                = work(i1+n-1_i4b) + ( s + work(i2+n-1_i4b) )
            work(i1+n)       = work(i1+n)*( work(i1+n-1_i4b)/t )
            work(i1+n-1_i4b) = t
!
        end if
!
        call dqds2( sigma, desig, work(i1+n-1_i4b), t, desig0 )
!
        work(n-1_i4b) = t
!
        call dqds2( sigma, desig, work(i1+n), t, desig0 )
!
        work(n) = t
!
!
!   END OF SUBROUTINE sub_2
!   -----------------------
!
        end subroutine sub_2
!
! ___________________________________________________________________________________________
!
        subroutine sub_3( j1, j2, j3, j4 )
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ----------------------------
!
        integer(i4b), intent(in) :: j1, j2, j3, j4
!
!
!   EXECUTABLE STATEMENTS
!   ---------------------
!
!
        do l=2_i4b, 1_i4b, -1_i4b
!         
            work(j1+n-l) = dn2 + work(j4+n-l)
!
            if ( safmin*work(j1+n-l)<work(j3+n+1_i4b-l) .and. safmin*work(j3+n+1_i4b-l)<work(j1+n-l) ) then
                tmp2         = work(j3+n+1-l)/work(j1+n-l)
                work(j2+n-l) = work(j4+n-l)*tmp2
                dn2          = dn2*tmp2
            else
                work(j2+n-l) = work(j3+n+1_i4b-l)*( work(j4+n-l)/work(j1+n-l) )
                dn2          = work(j3+n+1_i4b-l)*( dn2/work(j1+n-l) )
            end if
!
            if ( sigma==dn2+sigma ) dn2 = zero
!
        end do
!
        work(j1+n) = dn2
!
!
!   END OF SUBROUTINE sub_3
!   -----------------------
!
        end subroutine sub_3
!
! ___________________________________________________________________________________________
!
!
! END OF SUBROUTINE bd_dqds
! _________________________
!      
    end subroutine bd_dqds
!
! ===========================================================================================
!
!
    subroutine las2( f, g, h, ssmin, ssmax )
!
! Purpose
! _______
!
!   LAS2 computes singular values of a 2-by-2 triangular matrix.
!
!
! Arguments
! _________
!
!   F            (INPUT) real(stnd)
!                On entry, the (1,1) element of the 2-by-2 matrix.
!
!   G            (INPUT) real(stnd)
!                On entry, the (1,2) element of the 2-by-2 matrix.
!
!   H            (INPUT) real(stnd)
!                On entry, the (2,2) element of the 2-by-2 matrix.
!
!   SSMIN        (OUTPUT) real(stnd)
!                On exit, the smaller singular value.
!
!   SSMAX        (OUTPUT) real(stnd)
!                On exit, the larger singular value.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the DLAS2 subroutine from LAPACK version 3.12.0.
!
!   Barring over/underflow, all output quantities are correct to within
!   a few units in the last place (ulps), even in the absence of a guard
!   digit in addition/subtraction.
!
!   In IEEE arithmetic, the code works also correctly if one matrix element is infinite.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)  :: f, g, h
    real(stnd), intent(out) :: ssmin, ssmax
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
!
    real(stnd) :: as, at, au, c, fa, fhmn, fhmx, ga, ha
!
! EXECUTABLE STATEMENTS
! _____________________
!
!      
      fa = abs( f )
      ga = abs( g )
      ha = abs( h )
!      
      fhmn = min( fa, ha )
      fhmx = max( fa, ha )
!      
      if ( fhmn==zero ) then
!      
         ssmin = zero
!      
         if ( fhmx==zero ) then
            ssmax = ga
         else
            ssmax = max( fhmx, ga )*sqrt( one + ( min( fhmx, ga )/max( fhmx, ga ) )**2 )
         end if
!      
      else
!      
         if ( ga<fhmx ) then
!      
            as = one + fhmn/fhmx
            at = ( fhmx - fhmn )/fhmx
            au = ( ga/fhmx )**2
            c = two/( sqrt( as*as+au ) + sqrt( at*at+au ) )
            ssmin = fhmn*c
            ssmax = fhmx/c
!      
         else
!
            au = fhmx / ga
!
            if ( au==zero ) then
!
!              AVOID POSSIBLE HARMFUL UNDERFLOW IF EXPONENT RANGE
!              ASYMMETRIC (TRUE ssmin MAY NOT UNDERFLOW EVEN IF
!              au UNDERFLOWS).
!
               ssmin = ( fhmn*fhmx )/ga
               ssmax = ga
!
            else
!
               as = one + fhmn/fhmx
               at = ( fhmx - fhmn )/fhmx
               c = one/( sqrt( one+( as*au )**2 )+ sqrt( one+( at*au )**2 ) )
               ssmin = ( fhmn*c )*au
               ssmin = ssmin + ssmin
               ssmax = ga/( c + c )
!
            end if
!
         end if
!
      end if
!
!
! END OF SUBROUTINE las2
! ______________________
!
    end subroutine las2
!
!
! =========================================================================================
!                             SVD DECOMPOSITION OF A RECTANGULAR MATRIX
! =========================================================================================
!
!
    function singvalues( mat, sort, mul_size, maxiter, dqds ) result( s )
!
! Purpose
! _______
!
!   Function SINGVALUES computes the singular values of a real m-by-n matrix MAT.
!   The Singular Value Decomposition (SVD) is written
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix.  The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!
!   The original matrix MAT is first reduced to upper or lower
!   bidiagonal form BD by an orthogonal transformation:
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal. The singular values SIGMA of the bidiagonal
!   matrix BD are then computed by the bidiagonal implicit QR algorithm (if DQDS=false)
!   or the dqds algorithm (if DQDS=true). If these algorithms fail to converge, function
!   SINGVALUES returns a min(m,n)-vector filled with a quiet NaN.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n matrix MAT.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MUL_SIZE     (INPUT, OPTIONAL) integer(i4b)
!                Internal parameter. MUL_SIZE must verify: 1 <= MUL_SIZE <= max(m,n),
!                otherwise a default value is used. For better performance, 
!                at the expense of more workspace, a large value can be used.
!
!                The default is 32.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the
!                bidiagonal SVD phase of the SVD algorithm.
!
!                The bidiagonal SVD algorithm of an intermediate bidiagonal 
!                form BD of MAT fails to converge if the number of QR sweeps 
!                exceeds MAXITER*min(m,n). Convergence usually occurs in 
!                about 2*min(m,n) QR sweeps.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is 10.
!
!   DQDS         (INPUT, OPTIONAL) logical(lgl)
!                DQDS determines how the singular values of the intermediate min(m,n)-by-min(m,n)
!                bidiagonal matrix BD are computed.
!
!                If DQDS is set to true, singular values are computed with a more accurate
!                dqds algorithm delivering improved accuracy in the final computed SVD
!                decomposition. Moreover, the dqds algorithm is usually also faster than the
!                bidiagonal implicit QR algorithm.
!
!                If DQDS is set to false, singular values are computed with the bidiagonal implicit
!                QR algorithm applied to the associated min(m,n)-by-min(m,n) bidiagonal matrix BD.
!
!                The default is true.
!
!
! Further Details
! _______________
!
!   Computing the singular values of a rectangular matrix in function SINGVALUES consists of two steps:
!
!       1) reduction of the rectangular matrix to bidiagonal form BD, see the references (1) and (2);
!       2) computation of the singular values of the  min(m,n)-by-min(m,n) bidiagonal matrix BD by the
!          bidiagonal implicit QR or dqds algorithms, see the references(1), (2), (3) and (4).
!
!   Note that if max(m,n) is much larger than min(m,n) the rectangular matrix is first reduced to upper
!   or lower triangular form by a QR or LQ factorization and the bdiagonal reduction algorithm is applied
!   to the resulting triangular factor. The singular values of the rectangular rmatrix are then obtained
!   from those of the triangular factor.
!
!   If the SVD algorithm did not converge and full accuracy was not attained in
!   the bidiagonal SVD of an intermediate bidiagonal form BD of MAT or in the dqds algorithm
!   applied to BD, function SINGVALUES returns a min(m,n)-vector filled with NAN() function.
!
!   For further details, on the SVD of a rectangular matrix and the algorithms to
!   compute it or simply its singular values, see the references (1), (2), (3) or (4).
!   In SINGVALUES function, the reduction to bidiagonal form by orthogonal transformations
!   is parallelized if OPENMP is used, but not the computation of the singular values.
!
!   For more informations, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (4) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : lascl, merror
    use QR_Procedures,   only : qr_cmp
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in), dimension(:,:) :: mat
!
    real(stnd), dimension(min(size(mat,1),size(mat,2))) :: s
!
    integer(i4b), intent(in), optional :: mul_size, maxiter
!
    logical(lgl), intent(in), optional :: dqds
!
    character , intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                          :: matnorm, tmp
    real(stnd), dimension(min(size(mat,1),size(mat,2))) :: d, e
    real(stnd), dimension(:,:), allocatable             :: mat2
!
    integer(i4b) :: i, j, k, l, m, n, inc, mnthr, i1, i2, info
    integer      ::  iok
!
    logical(lgl) :: failure, mat_is_scaled, use_dqds
!
#ifdef _OPENMP
    logical :: test_par
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='singvalues'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = min( m, n )
!
    if ( present(mul_size) ) then
        inc = merge( mul_size, 32_i4b, 1_i4b<=mul_size .and. mul_size<=max(m,n) )
    else
        inc = 32_i4b
    end if
!
    use_dqds = true
!
    if ( present(dqds) ) then
        use_dqds = dqds
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   ALLOCATE WORKSPACE NEEDED FOR STORING mat OR ITS TRANSPOSE.
!
    i = max( m, n )
!
    allocate( mat2(i,k), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   MAKE A COPY OF mat AND COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
#ifdef _OPENMP
    test_par = n*m>=omp_limit .and. inc+inc<l
#endif
!
    if ( m>=n ) then
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)               &
!$OMP            ,PRIVATE(i,i1,i2,tmp), REDUCTION(max:matnorm)
!
        do i = inc, l, inc
!
            i1 = i + j
            i2 = min( i, n )
            tmp = maxval( abs( mat(:m,i1:i2) ) ) 
            matnorm = max( matnorm, tmp )
            mat2(:m,i1:i2) = mat(:m,i1:i2)
!
        end do
!
!$OMP END PARALLEL DO
!
    else
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)               &
!$OMP            ,PRIVATE(i,i1,i2,tmp), REDUCTION(max:matnorm)
!
        do i = inc, l, inc
!
            i1 = i + j
            i2 = min( i, n )
            tmp = maxval( abs( mat(:m,i1:i2) ) )
            matnorm = max( matnorm, tmp )
            mat2(i1:i2,:m) = transpose( mat(:m,i1:i2) )
!
        end do
!
!$OMP END PARALLEL DO
!
        m = n
        n = k
!
    end if
!
!   SCALE mat2 IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
        s(:k) = zero
!
        return
!
    else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
        call lascl( mat2, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
        call lascl( mat2, matnorm, bignum )
!
    end if
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat2  TO BIDIAGONAL FORM. IF m EXCEEDS mnthr ,
!   A QR FACTORIZATION IS USED FIRST TO REDUCE THE MATRIX
!   mat2 TO A TRIANGULAR FORM.
!    
    mnthr = int( real( k, stnd )*thr_bd, i4b )
!
    if (  m>=mnthr  ) then
!
!       PATH 1 : m MUCH LARGER THAN n .
!
!       FIRST REDUCE mat2 USING THE QR DECOMPOSITION.
!       SECOND REDUCE R TO UPPER BIDIAGONAL FORM.
!
!       COMPUTE mat2 = Q * R .
!
        call qr_cmp( mat2, d, e )
!
!       RECONSTRUCT R IN mat2, ZEROING OUT BELOW IT.
!
!$OMP  PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)    &
!$OMP             ,SCHEDULE(STATIC,1), PRIVATE(i)
!
        do i = 1_i4b, n
!
            mat2(i,i) = d(i)
            mat2(i+1_i4b:n,i) = zero
!
        end do
!
!$OMP END PARALLEL DO
!
!       UPPER BIDIAGONALIZE R IN mat .
!
        call bd_cmp_t( mat2(:n,:n), d, e )
!
    else
!
!       PATH 2 : m AT LEAST n, BUT NOT MUCH LARGER;
!
!       REDUCE mat2 TO UPPER BIDIAGONAL FORM WITHOUT 
!       QR DECOMPOSITION.
!
!       UPPER BIDIAGONALIZE mat .
!
        call bd_cmp_t( mat2(:m,:n), d, e )
!
    end if
!
    if ( use_dqds ) then
!
!       COMPUTE SINGULAR VALUES BY THE DQDS ALGORITHM.
!
        call bd_lasq1( d, e, failure, maxiter=60_i4b, sort=sort, info=info )
!
        if ( info==2_i4b ) then
!
            call bd_dqds( d, e, failure, maxiter=60_i4b, sort=sort )
!
        end if
!
    else
!
!       PERFORM BIDIAGONAL QR ITERATIONS FOR COMPUTING SINGULAR VALUES OTHERWISE.
!
        call bd_svd_t( true, d, e, failure, sort=sort, maxiter=maxiter  )
!
    end if
!
!   DEALLOCATE WORKSPACE.
!
    deallocate( mat2 )
!    
!   VERIFY FAILURE INDICATOR.
!    
    if ( failure ) then
!
!       FAILURE : BIDIAGONAL SVD OR DQDS ALGORITHM DID NOT CONVERGE.
!
        s(:k) = nan()
!
    else if ( mat_is_scaled ) then
!
!       IF MATRIX mat WAS SCALED, THEN RESCALE SINGULAR VALUES APPROPRIATELY.
!
        if ( matnorm<smlnum ) then
            tmp = smlnum
        else
            tmp = bignum
        end if
!
        call lascl( d, tmp, matnorm )
!
        s(:k) = d(:k)
!
    else
!
        s(:k) = d(:k)
!
    end if
!
!
! END OF FUNCTION singvalues
! __________________________
!
    end function singvalues
!
! =========================================================================================
!
    subroutine select_singval_cmp_a( mat, nsing, s, failure, sort, mul_size, vector, abstol, &
                                     ls, theta, d, e, tauq, taup, scaling, init, dqds )
!
! Purpose
! _______
!
!   SELECT_SINGVAL_CMP computes all or some of the greatest singular values of
!   a real m-by-n matrix MAT.
!
!   The Singular Value Decomposition (SVD) is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix.  The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!
!   The original matrix MAT is first reduced to upper or lower
!   bidiagonal form BD by an orthogonal transformation:
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal (see the reference (1) below).
!
!   The singular values SIGMA of the bidiagonal matrix BD, which are also the singular values
!   of MAT, are then computed by a bisection algorithm applied to the Tridiagonal Golub-Kahan (TGK)
!   form of the bidiagonal matrix BD (see the reference (2) below, Sec.3.3) or the differential
!   quotient difference with shifts (dqds) algorithm (see the references (3) and (4)).
!
!   The routine outputs (parts of) SIGMA and optionally Q and P (in packed form), and
!   BD for a given matrix MAT. SIGMA, Q, P and BD may then be used to obtain selected or all
!   singular vectors of MAT with subroutines BD_INVITER2 or BD_DEFLATE2.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n matrix MAT.
!
!                On exit, MAT is destroyed and if TAUQ or TAUP are present
!                MAT is overwritten as follows:
!
!                - if m >= n, the elements on and below the diagonal,
!                  with the array TAUQ, represent the orthogonal
!                  matrix Q as a product of elementary reflectors, 
!                  and the elements above the diagonal, with the
!                  array TAUP, represent the orthogonal matrix P
!                  as a product of elementary reflectors;
!                - if m < n, the elements below the diagonal, with
!                  the array TAUQ, represent the orthogonal
!                  matrix Q as a product of elementary reflectors,
!                  and the elements on and above the diagonal,
!                  with the array TAUP, represent the orthogonal
!                  matrix P as a product of elementary reflectors.
!
!                See Further Details.
!
!   NSING        (OUTPUT) integer(i4b)
!                On output, NSING specifies the number of singular values which have
!                been computed. Note that NSING may be greater than the optional argument
!                LS, if multiple singular values at index LS make unique selection
!                impossible and bisection is used.
!
!                If none of the optional arguments LS and THETA are used, NSING is set to
!                min( size(MAT,1) , size(MAT,2) ) and all the singular values are computed.
!
!   S            (OUTPUT) real(stnd), dimension(:)
!                On exit, S(1:NSING)  contains the first NSING  singular values of MAT.
!                The other values in S ( S(NSING+1:) ) are flagged  by a quiet NAN.
!
!                The size of S must be min( size(MAT,1) , size(MAT,2) ).
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection or dqds algorithm
!                  converged for all the computed singular values to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the singular values failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic. The sign of the incorrect singular values is set to negative.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MUL_SIZE     (INPUT, OPTIONAL) integer(i4b)
!                Internal parameter. MUL_SIZE must verify: 1 <= MUL_SIZE <= n,
!                otherwise a default value is used. For better performance, 
!                at the expense of more workspace, a large value can be used.
!
!                The default is min( 32, n ) .
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to TRUE, a vectorized version of the bisection algorithm
!                is used to compute the singular values SIGMA of the bidiagonal matrix BD.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is VECTOR = false.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the singular values. A singular value
!                (or cluster) is considered to be located in the bisection algorithm if it has
!                has been determined to lie in an interval whose width is ABSTOL or less.
!
!                Singular values will be computed most accurately when ABSTOL is set to the square root of
!                the underflow threshold (e.g., sqrt(safmin)=sqrt(LAMCH('S')), not zero. 
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | T(GK) | will be used,
!                where | T(GK) | means the 1-norm of the GOLUB-KAHAN tridiagonal form of the bidiagonal matrix BD
!                and ULP is the machine precision (distance from 1 to the next larger floating point number).
!
!                This argument has no effect if DQDS is equal to true.
!
!   LS           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LS specifies the number of singular values which must be
!                computed by the subroutine. On output, NSING may be different than LS 
!                if multiple singular values at index LS make unique selection
!                impossible.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LS must be greater than 0 and less or equal to min( size(MAT,1) , size(MAT,2) ).
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is LS = min( size(MAT,1) , size(MAT,2) ).
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry, THETA specifies that the singular values which are greater or equal to
!                THETA must be computed. If none of the singular values are greater or equal to
!                THETA, NSING is set to zero and S(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is THETA = 0.
!
!   D            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The diagonal elements of the intermediate bidiagonal matrix BD
!
!                The size of D must be min( size(MAT,1) , size(MAT,2) ).
!
!   E            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The off-diagonal elements of the intermediate bidiagonal matrix BD:
!
!                - if m >= n, E(i) = BD(i-1,i) for i = 2,3,...,n;
!                - if m < n,  E(i) = BD(i,i-1) for i = 2,3,...,m.
!
!                The size of E must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUQ         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix Q. See Further Details.
!
!                The size of TAUQ must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUP         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix P. See Further Details.
!
!                The size of TAUP must be min( size(MAT,1) , size(MAT,2) ).
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix BD is scaled
!                before computing the singular values.
!
!                The default is to scale the bidiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps are
!                computed from estimates of the eigenvalues of the associated BD' * BD tridiagonal
!                matrix obtained from the Pal-Walker-Kahan algorithm.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!   DQDS         (INPUT, OPTIONAL) logical(lgl)
!                DQDS determines how the singular values are computed.
!
!                If DQDS is set to true, all singular values are computed with the dqds algorithm
!                The dqds algorithm is much faster than bisection when all singular values are
!                requested in many cases and delivers about the same accuracy than bisection.
!
!                If DQDS is set to false, the requested singular values are computed by bisection.
!
!                The default is to use the dqds algorithm if all the singular values are requested
!                and to use bisection if selected singular values are requested (e.g., if the optional
!                arguments LS or THETA are present).
!
!
! Further Details
! _______________
!
!   The matrices Q and P are represented as products of elementary
!   reflectors:
!
!   - If m >= n,
!
!       Q = H(1) * H(2) * . . . * H(n)  and  P = G(1) * G(2) * . . . * G(n-1)
!
!     Each H(i) and G(i) has the form:
!
!       H(i) = I + tauq * u * u'  and G(i) = I + taup * v * v'
!
!     where tauq and taup are real scalars, and u and v are real vectors.
!     Moreover, u(1:i-1) = 0 and v(1:i) = 0.
!
!     If TAUQ or TAUP are present:
!
!     - u(i:m)   is stored on exit in MAT(i:m,i);
!     - v(i+1:n) is stored on exit in MAT(i,i+1:n).
!
!     If TAUQ is present : tauq is stored in TAUQ(i).
!
!     If TAUP is present : taup is stored in TAUP(i).
!
!
!   - If m < n,
!
!         Q = H(1) * H(2) * . . . * H(m-1)  and  P = G(1) * G(2) * . . . * G(m)
!
!     Each H(i) and G(i) has the form:
!
!         H(i) = I + tauq * u * u'  and G(i) = I + taup * v * v'
!
!     where tauq and taup are real scalars, and u and v are real vectors.
!     Moreover, u(1:i) = 0 and v(1:i-1) = 0.
!
!     If TAUQ or TAUP are present:
!
!     -  u(i+1:m) is stored on exit in MAT(i+1:m,i);
!     -  v(i:n)   is stored on exit in MAT(i,i:n).
!
!     If TAUQ is present : tauq is stored in TAUQ(i).
!
!     If TAUP is present : taup is stored in TAUP(i).
!
!   The contents of MAT on exit, if TAUQ or TAUP are present,
!   are illustrated by the following examples:
!
!   - m = 6 and n = 5 (m >= n):
!
!        (  u1  v1  v1  v1  v1 )
!
!        (  u1  u2  v2  v2  v2 )
!
!        (  u1  u2  u3  v3  v3 )
!
!        (  u1  u2  u3  u4  v4 )
!
!        (  u1  u2  u3  u4  u5 )
!
!        (  u1  u2  u3  u4  u5 )
!
!   - m = 5 and n = 6 (m < n):
!
!        (  v1  v1  v1  v1  v1  v1 )
!
!        (  u1  v2  v2  v2  v2  v2 )
!
!        (  u1  u2  v3  v3  v3  v3 )
!
!        (  u1  u2  u3  v4  v4  v4 )
!
!        (  u1  u2  u3  u4  v5  v5 )
!
!   where ui denotes an element of the vector defining H(i), and vi an element 
!   of the vector defining G(i).
!
!   Now, let SIGMA(i), i=1,...,N=min(m,n), be the singular values of the intermediate bidiagonal
!   matrix BD in decreasing order of magnitude.
!
!   If DQDS is equal to false, the subroutine computes the LS largest singular
!   values (or the singular values which are greater or equal to THETA) of BD by a bisection
!   method (see the reference (1) below, Sec.8.5). The bisection method is applied to an associated 
!   2.N-by-2.N symmetric tridiagonal matrix T (the so-called GOLUB-KAHAN form of BD) whose 
!   eigenvalues are the singular values of BD and their negatives (see the reference (2) below).
!
!   On the other hand, if DQDS is equal to true, all the singular values of BD are computed by the dqds
!   algorithm (see the references (3) and (4).
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!   (3) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (4) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error2
    use Utilities,       only : assert, assert_eq, merror, lascl
    use QR_Procedures,   only : qr_cmp, lq_cmp
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout),  dimension(:,:)         :: mat
    real(stnd),  intent(out),    dimension(:)           :: s
    real(stnd),  intent(in),                   optional :: abstol, theta
    real(stnd),  intent(out),    dimension(:), optional :: tauq, taup, d, e
!
    integer(i4b), intent(out)          :: nsing
    integer(i4b), intent(in), optional :: mul_size, ls
!
    logical(lgl), intent(out)          :: failure
    logical(lgl), intent(in), optional :: vector, scaling, init, dqds
!
    character , intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: matnorm, tmp
    real(stnd), dimension(size(s)) :: d2, e2, tauq2, taup2
!
    integer(i4b) :: i, j, k, l, m, m2, n, n2, inc, mnthr, i1, i2, info
!
    logical(lgl) :: mat_is_scaled, use_dqds
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_singval_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = assert_eq( min(m,n),         &
                   int(size(s),i4b), &
                   name_proc  )
!
    if ( present(ls) ) then
!
        call assert( logical( ls<=k, lgl ),     &
                     logical( 1_i4b<=ls, lgl ), &
                     name_proc )
!
        if ( present(theta) ) then
            call merror( name_proc//svd_error2 )
        end if
!
    end if
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat  TO BIDIAGONAL FORM. IF max(m,n) EXCEEDS mnthr ,
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE MATRIX
!   mat TO A TRIANGULAR FORM.
!    
    mnthr = int( real( k, stnd )*thr_bd, i4b )
!
!   TEST OPTIONAL ARGUMENTS.
!
    if ( present(d) ) then
        call assert( logical( k==size(d), lgl ), name_proc )
    end if
!
    if ( present(e) ) then
        call assert( logical( k==size(e), lgl ), name_proc )
    end if
!
    if ( present(tauq) ) then
        call assert( logical( k==size(tauq), lgl ), name_proc )
        mnthr = huge( i )
    end if
!
    if ( present(taup) ) then
        call assert( logical( k==size(taup), lgl ), name_proc )
        mnthr = huge( i )
    end if
!
    if ( present(mul_size) ) then
        inc = merge( mul_size, 32_i4b, 1_i4b<=mul_size .and. mul_size<=n )
    else
        inc = 32_i4b
    end if
!
    if ( present(ls) .or. present(theta) ) then
        use_dqds = false
    else
        use_dqds = true
    end if
!
    if ( present(dqds) ) then
        use_dqds = dqds
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) then
!
        nsing = 0_i4b
!
        return
!
    end if
!
!   COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, n )
        tmp = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       mat IS ZERO, GENERATE Q AND P IN PACKED FORM.
!
        if ( m>=n ) then
!
            do i = 1_i4b, k-1_i4b
                mat(i,i:i+1_i4b) = one
            end do
!
        else
!
            do i = 1_i4b, k-1_i4b
                mat(i:i+1_i4b,i) = one
            end do
!
        end if
!
        mat(k,k) = one
!
        nsing = k
        s(:k) = zero
!
        if ( present(d) )    d(:k) = zero
        if ( present(e) )    e(:k) = zero
        if ( present(tauq) ) tauq(:k) = zero
        if ( present(taup) ) taup(:k) = zero
!
        return
!
    else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, bignum )
!
    end if
!
    m2 = m
    n2 = n
!
    if (  m>mnthr  ) then
!
!       m MUCH LARGER THAN n .
!
!       FIRST REDUCE mat USING THE QR DECOMPOSITION.
!
!       COMPUTE mat = Q * R .
!
        call qr_cmp( mat, tauq2, taup2 )
!
!       RECONSTRUCT R IN mat, ZEROING OUT BELOW IT.
!
!$OMP  PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)         &
!$OMP             ,SCHEDULE(STATIC,1), PRIVATE(i)
!
        do i = 1_i4b, n
            mat(i,i) = tauq2(i)
            mat(i+1_i4b:n,i) = zero
        end do
!
!$OMP END PARALLEL DO
!
        m2 = n
!
    else if (  n>mnthr ) then
!
!       n MUCH LARGER THAN m.
!
!       FIRST REDUCE mat USING THE LQ DECOMPOSITION.
!
!       COMPUTE mat = L * Q .
!
        call lq_cmp( mat, tauq2, taup2 )
!
!       RECONSTRUCT L IN mat , ZEROING ABOUT ABOVE IT.
!
!$OMP  PARALLEL DO IF((m*(m+1_i4b))/2_i4b>=omp_limit)         &
!$OMP             ,SCHEDULE(STATIC,1), PRIVATE(i)
!
        do i = 1_i4b, m
            mat(1_i4b:i-1_i4b,i) = zero
            mat(i,i) = tauq2(i)
        end do
!
!$OMP END PARALLEL DO
!
        n2 = m
!
    end if
!
!   REDUCE mat TO LOWER OR UPPER BIDIAGONAL FORM.
!
    call bd_cmp_a( mat(:m2,:n2), d2, e2, tauq2, taup2 )
!
!   SAVE INTERMEDIATE QUANTITIES, IF REQUIRED.
!
    if ( present(d) ) then
        d = d2
    end if
!
    if ( present(e) ) then
        e = e2
    end if
!
    if ( present(tauq) ) then
        tauq = tauq2
    end if
!
    if ( present(taup) ) then
        taup = taup2
    end if
!
    if ( use_dqds ) then
!
!       COMPUTE SINGULAR VALUES BY THE DQDS ALGORITHM.
!
        call bd_lasq1( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling, info=info )
!
        if ( info==2_i4b ) then
!
            call bd_dqds( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling )
!
        end if
!
        if ( .not.failure ) then
!
            s(:k) = d2(:k)
!
        else
!
            s(:k) = -one
!
        end if
!
        nsing = k
!
    else
!
!       COMPUTE THE SINGULAR VALUES BY THE BISECTION METHOD APPLIED TO THE GOLUB-KAHAN
!       FORM OF THE BIDIAGONAL MATRIX.
!
        call bd_singval( d2, e2, nsing, s, failure, sort=sort,    &
                         vector=vector, abstol=abstol, ls=ls,     &
                         theta=theta, scaling=scaling, init=init  )
!
    end if
!
!       UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled .and. nsing>0_i4b ) then
!
        call lascl( s(:nsing), merge(smlnum,bignum,matnorm<smlnum) , matnorm )
!
    end if
!
!
! END OF SUBROUTINE select_singval_cmp_a
! ______________________________________
!
    end subroutine select_singval_cmp_a
!
! =========================================================================================
!
    subroutine select_singval_cmp_b( mat, rlmat, nsing, s, failure, sort, mul_size, vector, abstol, &
                                     ls, theta, d, e, tauo, tauq, taup, scaling, init, dqds )
!
! Purpose
! _______
!
!   SELECT_SINGVAL_CMP computes all or some of the greatest singular values of
!   a real m-by-n matrix MAT.
!
!   The Singular Value Decomposition (SVD) is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix.  The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!
!   The original matrix MAT is first reduced to upper bidiagonal form
!   by a two-step algorithm :
!
!   - If m >= n, a QR factorization of the real m-by-n matrix MAT is first computed
!
!                        MAT = O * R
!
!   where O is orthogonal and R is upper triangular. In a second step, the n-by-n upper
!   triangular matrix R is reduced to upper bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * R * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix.
!
!   - If m < n, an LQ factorization of the real m-by-n matrix MAT is first computed
!
!                       MAT = L * O
!
!   where O is orthogonal and L is lower triangular. In a second step, the m-by-m lower
!   triangular matrix L is reduced to upper bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * L * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix.
!
!   SELECT_SINGVAL_CMP computes O, BD, Q and P.
!
!   The singular values SIGMA of the bidiagonal matrix BD, which are also the singular values
!   of MAT, are then computed by a bisection algorithm applied to the Tridiagonal Golub-Kahan (TGK)
!   form of the bidiagonal matrix BD (see the reference (2) below, Sec.3.3) or the differential
!   quotient difference with shifts (dqds) algorithm (see the references (3) and (4)).
!
!   The routine outputs (parts of) SIGMA, and optionally O, Q and P (in packed form), and
!   BD for a given matrix MAT. The matrix O is stored in factored form in the argument MAT
!   if the optional argument TAUO is present or explicitly computed if this argument is absent.
!   SIGMA, O, Q, P and BD may then be used to obtain selected singular vectors of
!   MAT with subroutines BD_INVITER2 or BD_DEFLATE2.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n matrix MAT.
!
!                On exit, if:
!
!                - m >= n, the elements on and below the diagonal,
!                  with the array TAUO, represent the orthogonal
!                  matrix O of the QR factorization of MAT,
!                  as a product of elementary reflectors,
!                  if the argument TAUO is present. Otherwise,
!                  the argument MAT contains the first n columns
!                  of O on output.
!                - m < n, the elements on and above the diagonal,
!                  with the array TAUO, represent the orthogonal
!                  matrix O of the LQ factorization of MAT,
!                  as a product of elementary reflectors,
!                  if the argument TAUO is present. Otherwise,
!                  the argument MAT contains the first m rows
!                  of O on output.
!
!                See Further Details.
!
!   RLMAT        (OUTPUT) real(stnd), dimension(:,:)
!
!                On exit, the elements on and below the diagonal,
!                with the array TAUQ, represent the orthogonal
!                matrix Q as a product of elementary reflectors, 
!                and the elements above the diagonal, with the
!                array TAUP, represent the orthogonal matrix P
!                as a product of elementary reflectors;
!
!                See Further Details.
!
!                The shape of RLMAT must verify:
!                size( RLMAT, 1 ) = size( RLMAT, 2 ) = min( size(MAT,1) , size(MAT,2) ).
!
!   NSING        (OUTPUT) integer(i4b)
!                On output, NSING specifies the number of singular values which have
!                been computed. Note that NSING may be greater than the optional argument
!                LS, if multiple singular values at index LS make unique selection
!                impossible.
!
!                If none of the optional arguments LS and THETA are used, NSING is set to
!                min( size(MAT,1) , size(MAT,2) ) and all the singular values are computed.
!
!   S            (OUTPUT) real(stnd), dimension(:)
!                On exit, S(1:NSING)  contains the first NSING  singular values of MAT.
!                The other values in S ( S(NSING+1:) ) are flagged  by a quiet NAN.
!
!                The size of S must be min( size(MAT,1) , size(MAT,2) ).
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection or dqds  algorithm
!                  converged for all the computed singular values to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the singular values failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic. The sign of the incorrect singular values is set to negative.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MUL_SIZE     (INPUT, OPTIONAL) integer(i4b)
!                Internal parameter. MUL_SIZE must verify: 1 <= MUL_SIZE <= n,
!                otherwise a default value is used. For better performance, 
!                at the expense of more workspace, a large value can be used.
!
!                The default is min( 32, n ) .
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to TRUE, a vectorized version of the bisection algorithm
!                is used to compute the singular values SIGMA of the bidiagonal matrix BD.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is VECTOR = false.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the singular values. A singular value
!                (or cluster) is considered to be located in the bisection algorithm if it has
!                has been determined to lie in an interval whose width is ABSTOL or less.
!
!                Singular values will be computed most accurately when ABSTOL is set to the square root of
!                the underflow threshold (e.g., sqrt(safmin)=sqrt(LAMCH('S')), not zero. 
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | T(GK) | will be used,
!                where | T(GK) | means the 1-norm of the GOLUB-KAHAN tridiagonal form of the bidiagonal matrix BD
!                and ULP is the machine precision (distance from 1 to the next larger floating point number).
!
!                This argument has no effect if DQDS is equal to true.
!
!   LS           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LS specifies the number of singular values which must be
!                computed by the subroutine. On output, NSING may be different than LS 
!                if multiple singular values at index LS make unique selection
!                impossible.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LS must be greater than 0 and less or equal to min( size(MAT,1) , size(MAT,2) ).
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is LS = min( size(MAT,1) , size(MAT,2) ).
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry, THETA specifies that the singular values which are greater or equal to
!                THETA must be computed. If none of the singular values are greater or equal to
!                THETA, NSING is set to zero and S(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is THETA = 0.
!
!   D            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The diagonal elements of the intermediate bidiagonal matrix BD
!
!                The size of D must be min( size(MAT,1) , size(MAT,2) ).
!
!   E            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The off-diagonal elements of the intermediate upper bidiagonal matrix BD:
!
!                     E(i) = BD(i-1,i) for i = 2,3,...,k;
!
!                The size of E must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUO         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix O of the QR or LQ decomposition of MAT.
!
!                If the optional argument TAUO is present, the orthogonal matrix O is stored
!                in factored form, as a product of elementary reflectors, in the argument MAT on exit.
!
!                If the optional argument TAUO is absent, the orthogonal matrix O is explicitly generated
!                and stored in the argument MAT on exit.
!
!                See description of the argument MAT above and Further Details below.
!
!                The size of TAUO must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUQ         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix Q. See Further Details.
!
!                The size of TAUQ must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUP         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix P. See Further Details.
!
!                The size of TAUP must be min( size(MAT,1) , size(MAT,2) ).
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix BD is scaled
!                before computing the singular values.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is to scale the bidiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps are
!                computed from estimates of the eigenvalues of the associated BD' * BD tridiagonal
!                matrix obtained from the Pal-Walker-Kahan algorithm.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!   DQDS         (INPUT, OPTIONAL) logical(lgl)
!                DQDS determines how the singular values are computed.
!
!                If DQDS is set to true, all singular values are computed with the dqds algorithm
!                The dqds algorithm is much faster than bisection when all singular values are
!                requested in many cases and delivers about the same accuracy than bisection.
!
!                If DQDS is set to false, the requested singular values are computed by bisection.
!
!                The default is to use the dqds algorithm if all the singular values are requested
!                and to use bisection if selected singular values are requested (e.g., if the optional
!                arguments LS or THETA are present).
!
!
! Further Details
! _______________
!
!   If m >= n, the matrix O of the QR factorization of MAT is represented as a product
!   of elementary reflectors
!
!            O = W(1) * W(2) * ... * W(n)
!
!   Each W(i) has the form
!
!            W(i) = I + tauo * ( v * v' ) ,
!                      
!   where tauo is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in MAT(i:m,i) and tauo in TAUO(i). If the optional argument
!   TAUO is absent, the first n columns of O are generated and stored in the argument MAT.
!
!   If m < n, The matrix O of the LQ factorization of MAT is represented as a product
!   of elementary reflectors
!
!           O = W(m) * . . . * W(2) * W(1)
!
!   Each W(i) has the form
!
!            W(i) = I + tauo * ( v * v' ) ,
!                      
!   where tauo is a real scalar and v is a real n-element vector with v(1:i-1) = 0.
!   v(i:n) is stored on exit in MAT(i,i:n) and tauo in TAUO(i). If the optional argument
!   TAUO is absent, the first m rows of O are generated and stored in the argument MAT.
!
!   The matrix O is stored in factored form if the optional argument TAUO is present
!   or explicitly computed if this argument is absent.
!
!   A blocked algorithm is used for computing the QR or LQ factorization of MAT. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   After, the initial QR or LQ factorization of MAT, the (upper or lower) triangular
!   matrix is reduced to upper bidiagonal form BD.
!
!   The matrices Q and P of the bidiagonal factorization of the triangular matrix R or L
!   are represented as products of elementary reflectors:
!
!       Q = H(1) * H(2) * . . . * H(k)  and  P = G(1) * G(2) * . . . * G(k-1)
!
!   , where k = min( size(MAT,1) , size(MAT,2) ). Each H(i) and G(i) has the form:
!
!       H(i) = I + tauq * u * u'  and G(i) = I + taup * v * v'
!
!   where tauq and taup are real scalars, and u and v are real vectors;
!   u(1:i-1) = 0 and u(i:min(m,n)) is stored on exit in RLMAT(i:min(m,n),i);
!   v(1:i) = 0 and v(i+1:min(m,n)) is stored on exit in RLMAT(i,i+1:min(m,n));
!   tauq is stored in TAUQ(i) and taup in TAUP(i).
!
!   The contents of RLMAT on exit are illustrated by the following example:
!
!   m = 6 and n = 5 (m >= n):
!
!        (  u1  v1  v1  v1  v1 )
!
!        (  u1  u2  v2  v2  v2 )
!
!        (  u1  u2  u3  v3  v3 )
!
!        (  u1  u2  u3  u4  v4 )
!
!        (  u1  u2  u3  u4  u5 )
!
!   where ui denotes an element of the vector defining H(i), and vi an element 
!   of the vector defining G(i).
!
!   Now, let SIGMA(i), i=1,...,N=min(m,n), be the singular values of the intermediate bidiagonal
!   matrix BD in decreasing order of magnitude.
!
!   If DQDS is equal to false, the subroutine computes the LS largest singular
!   values (or the singular values which are greater or equal to THETA) of BD by a bisection
!   method (see the reference (1) below, Sec.8.5). The bisection method is applied to an associated 
!   2.N-by-2.N symmetric tridiagonal matrix T (the so-called GOLUB-KAHAN form of BD) whose 
!   eigenvalues are the singular values of BD and their negatives (see the reference (2) below).
!
!   On the other hand, if DQDS is equal to true, all the singular values of BD are computed by the dqds
!   algorithm (see the references (3) and (4).
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!   (3) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (4) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error2
    use Utilities,       only : assert, assert_eq, merror, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout),  dimension(:,:)         :: mat
    real(stnd),  intent(out),    dimension(:,:)         :: rlmat
    real(stnd),  intent(out),    dimension(:)           :: s
    real(stnd),  intent(in),                   optional :: abstol, theta
    real(stnd),  intent(out),    dimension(:), optional :: tauo, tauq, taup, d, e
!
    integer(i4b), intent(out)          :: nsing
    integer(i4b), intent(in), optional :: mul_size, ls
!
    logical(lgl), intent(out)          :: failure
    logical(lgl), intent(in), optional :: vector, scaling, init, dqds
!
    character , intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: matnorm, tmp
    real(stnd), dimension(size(s)) :: d2, e2, tauq2, taup2
!
    integer(i4b) :: i, j, k, l, m, n, inc, i1, i2, info
!
    logical(lgl) :: mat_is_scaled, use_dqds
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_singval_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = assert_eq( min(m,n),                &
                   int(size(s),i4b),        &
                   int(size(rlmat,1),i4b) , &
                   int(size(rlmat,2),i4b) , &
                   name_proc  )
!
    if ( present(ls) ) then
!
        call assert( logical( ls<=k, lgl ),     &
                     logical( 1_i4b<=ls, lgl ), &
                     name_proc )
!
        if ( present(theta) ) then
            call merror( name_proc//svd_error2 )
        end if
!
    end if
!
!   TEST OPTIONAL ARGUMENTS.
!
    if ( present(d) ) then
        call assert( logical( k==size(d), lgl ), name_proc )
    end if
!
    if ( present(e) ) then
        call assert( logical( k==size(e), lgl ), name_proc )
    end if
!
    if ( present( tauo ) ) then
        call assert( logical( k==size(tauo), lgl ), name_proc )
    end if
!
    if ( present(tauq) ) then
        call assert( logical( k==size(tauq), lgl ), name_proc )
    end if
!
    if ( present(taup) ) then
        call assert( logical( k==size(taup), lgl ), name_proc )
    end if
!
    if ( present(mul_size) ) then
        inc = merge( mul_size, 32_i4b, 1_i4b<=mul_size .and. mul_size<=n )
    else
        inc = 32_i4b
    end if
!
    if ( present(ls) .or. present(theta) ) then
        use_dqds = false
    else
        use_dqds = true
    end if
!
    if ( present(dqds) ) then
        use_dqds = dqds
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) then
!
        nsing = 0_i4b
!
        return
!
    end if
!
!   COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, n )
        tmp = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       mat IS ZERO, GENERATE O, Q AND P IN PACKED FORM.
!
        do i = 1_i4b, k
            mat(i,i) = one
        end do
!
        rlmat(:k,:k) = zero
!
        do i = 1_i4b, k-1_i4b
            rlmat(i,i:i+1_i4b) = one
        end do
!
        rlmat(k,k) = one
!
        nsing = k
        s(:k) = zero
!
        if ( present(d) )    d(:k) = zero
        if ( present(e) )    e(:k) = zero
        if ( present(tauo) ) tauo(:k) = zero
        if ( present(tauq) ) tauq(:k) = zero
        if ( present(taup) ) taup(:k) = zero
!
        return
!
    else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   REDUCE mat TO UPPER BIDIAGONAL FORM BY A TWO-STEP ALGORITHM.
!
    call bd_cmp_c( mat(:m,:n), d2, e2, tauq2, taup2, rlmat(:k,:k), tauo=tauo )
!
!   SAVE INTERMEDIATE QUANTITIES, IF REQUIRED.
!
    if ( present(d) ) then
        d = d2
    end if
!
    if ( present(e) ) then
        e = e2
    end if
!
    if ( present(tauq) ) then
        tauq = tauq2
    end if
!
    if ( present(taup) ) then
        taup = taup2
    end if
!
    if ( use_dqds ) then
!
!       COMPUTE SINGULAR VALUES BY THE DQDS ALGORITHM.
!
        call bd_lasq1( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling, info=info )
!
        if ( info==2_i4b ) then
!
            call bd_dqds( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling )
!
        end if
!
        if ( .not.failure ) then
!
            s(:k) = d2(:k)
!
        else
!
            s(:k) = -one
!
        end if
!
        nsing = k
!
    else
!
!       COMPUTE THE SINGULAR VALUES BY THE BISECTION METHOD APPLIED TO THE GOLUB-KAHAN
!       FORM OF THE BIDIAGONAL MATRIX.
!
        call bd_singval( d2, e2, nsing, s, failure, sort=sort,    &
                         vector=vector, abstol=abstol, ls=ls,     &
                         theta=theta, scaling=scaling, init=init  )
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled .and. nsing>0_i4b ) then
!
        call lascl( s(:nsing), merge(smlnum,bignum,matnorm<smlnum) , matnorm )
!
    end if
!
!
! END OF SUBROUTINE select_singval_cmp_b
! ______________________________________
!
    end subroutine select_singval_cmp_b
!
! =========================================================================================
!    
    subroutine select_singval_cmp2_a( mat, nsing, s, failure, sort, mul_size, vector, abstol, &
                                      ls, theta, d, e, tauq, taup, scaling, init, dqds )
!
! Purpose
! _______
!
!   SELECT_SINGVAL_CMP2 computes all or some of the greatest singular values of
!   a real m-by-n matrix MAT.
!
!   The Singular Value Decomposition (SVD) is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix.  The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!
!   The original matrix MAT is first reduced to upper or lower
!   bidiagonal form BD by an orthogonal transformation:
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal (see the reference (1) below).
!
!   The singular values SIGMA of the bidiagonal matrix BD, which are also the singular values of MAT, 
!   are then computed by a bisection algorithm (see the reference (1) below, Sec.8.5 ). The bisection 
!   method is applied (implicitly) to the associated min(m,n)-by-min(m,n) symmetric tridiagonal matrix BD' * BD  
!   whose eigenvalues are the squares of the singular values of BD by using the differential stationary
!   form of the qd algorithm of Rutishauser (see the reference (2) below, Sec.3.1 ). Alternatively, at
!   the user option, all singular values can be computed by the differential quotient difference with
!   shifts (dqds) algorithm (see the references (3) and (4)).
!
!   The routine outputs (parts of) SIGMA and optionally Q and P (in packed form), and
!   BD for a given matrix MAT. SIGMA, Q, P and BD may then be used to obtain selected or all
!   singular vectors of MAT with subroutines BD_INVITER2 or BD_DEFLATE2.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n matrix MAT.
!
!                On exit, MAT is destroyed and if TAUQ or TAUP are present
!                MAT is overwritten as follows:
!
!                - if m >= n, the elements on and below the diagonal,
!                  with the array TAUQ, represent the orthogonal
!                  matrix Q as a product of elementary reflectors, 
!                  and the elements above the diagonal, with the
!                  array TAUP, represent the orthogonal matrix P
!                  as a product of elementary reflectors;
!                - if m < n, the elements below the diagonal, with
!                  the array TAUQ, represent the orthogonal
!                  matrix Q as a product of elementary reflectors,
!                  and the elements on and above the diagonal,
!                  with the array TAUP, represent the orthogonal
!                  matrix P as a product of elementary reflectors.
!
!                See Further Details.
!
!   NSING        (OUTPUT) integer(i4b)
!                On output, NSING specifies the number of singular values which have
!                been computed. Note that NSING may be greater than the optional argument
!                LS, if multiple singular values at index LS make unique selection
!                impossible.
!
!                If none of the optional arguments LS and THETA are used, NSING is set to
!                min( size(MAT,1) , size(MAT,2) ) and all the singular values are computed.
!
!   S            (OUTPUT) real(stnd), dimension(:)
!                On exit, S(1:NSING)  contains the first NSING  singular values of MAT.
!                The other values in S ( S(NSING+1:) ) are flagged  by a quiet NAN.
!
!                The size of S must be min( size(MAT,1) , size(MAT,2) ).
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection or dqds algorithm
!                  converged for all the computed singular values to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the singular values failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic. The sign of the incorrect singular values is set to negative.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MUL_SIZE     (INPUT, OPTIONAL) integer(i4b)
!                Internal parameter. MUL_SIZE must verify: 1 <= MUL_SIZE <= n,
!                otherwise a default value is used. For better performance, 
!                at the expense of more workspace, a large value can be used.
!
!                The default is min( 32, n ) .
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to TRUE, a vectorized version of the bisection algorithm
!                is used to compute the singular values SIGMA of the bidiagonal matrix BD.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is VECTOR = false.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the singular values. A singular value
!                (or cluster) is considered to be located in the bisection algorithm if it has
!                has been determined to lie in an interval whose width is ABSTOL or less.
!
!                Singular values will be computed most accurately when ABSTOL is set to the square root of
!                the underflow threshold (e.g., sqrt(safmin)=sqrt(LAMCH('S')), not zero. 
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | BD' * BD | will be used,
!                where | BD' * BD | means the 1-norm of the tridiagonal matrix BD' * BD ( BD' means the transpose of BD) 
!                and ULP is the machine precision (distance from 1 to the next larger floating point number).
!
!                This argument has no effect if DQDS is equal to true.
!
!   LS           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LS specifies the number of singular values which must be
!                computed by the subroutine. On output, NSING may be different than LS 
!                if multiple singular values at index LS make unique selection
!                impossible.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LS must be greater than 0 and less or equal to min( size(MAT,1) , size(MAT,2) ).
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is LS = min( size(MAT,1) , size(MAT,2) ).
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry, THETA specifies that the singular values which are greater or equal to
!                THETA must be computed. If none of the singular values are greater or equal to
!                THETA, NSING is set to zero and S(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is THETA = 0.
!
!   D            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The diagonal elements of the intermediate bidiagonal matrix BD
!
!                The size of D must be min( size(MAT,1) , size(MAT,2) ).
!
!   E            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The off-diagonal elements of the intermediate bidiagonal matrix BD:
!
!                - if m >= n, E(i) = BD(i-1,i) for i = 2,3,...,n;
!                - if m < n,  E(i) = BD(i,i-1) for i = 2,3,...,m.
!
!                The size of E must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUQ         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix Q. See Further Details.
!
!                The size of TAUQ must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUP         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix P. See Further Details.
!
!                The size of TAUP must be min( size(MAT,1) , size(MAT,2) ).
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix BD is scaled
!                before computing the singular values.
!
!                The default is to scale the bidiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps are
!                computed from estimates of the eigenvalues of the associated BD' * BD tridiagonal
!                matrix obtained from the Pal-Walker-Kahan algorithm.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!   DQDS         (INPUT, OPTIONAL) logical(lgl)
!                DQDS determines how the singular values are computed.
!
!                If DQDS is set to true, all singular values are computed with the dqds algorithm
!                The dqds algorithm is much faster than bisection when all singular values are
!                requested in many cases and delivers about the same accuracy than bisection.
!
!                If DQDS is set to false, the requested singular values are computed by bisection.
!
!                The default is to use the dqds algorithm if all the singular values are requested
!                and to use bisection if selected singular values are requested (e.g., if the optional
!                arguments LS or THETA are present).
!
!
! Further Details
! _______________
!
!   The matrices Q and P are represented as products of elementary
!   reflectors:
!
!   - If m >= n,
!
!       Q = H(1) * H(2) * . . . * H(n)  and  P = G(1) * G(2) * . . . * G(n-1)
!
!     Each H(i) and G(i) has the form:
!
!       H(i) = I + tauq * u * u'  and G(i) = I + taup * v * v'
!
!     where tauq and taup are real scalars, and u and v are real vectors.
!     Moreover, u(1:i-1) = 0 and v(1:i) = 0.
!
!     If TAUQ or TAUP are present:
!
!     - u(i:m)   is stored on exit in MAT(i:m,i);
!     - v(i+1:n) is stored on exit in MAT(i,i+1:n).
!
!     If TAUQ is present : tauq is stored in TAUQ(i).
!     If TAUP is present : taup is stored in TAUP(i).
!
!
!   - If m < n,
!
!         Q = H(1) * H(2) * . . . * H(m-1)  and  P = G(1) * G(2) * . . . * G(m)
!
!     Each H(i) and G(i) has the form:
!
!         H(i) = I + tauq * u * u'  and G(i) = I + taup * v * v'
!
!     where tauq and taup are real scalars, and u and v are real vectors.
!     Moreover, u(1:i) = 0 and v(1:i-1) = 0.
!
!     If TAUQ or TAUP are present:
!
!     - u(i+1:m) is stored on exit in MAT(i+1:m,i);
!     - v(i:n)   is stored on exit in MAT(i,i:n).
!
!     If TAUQ is present : tauq is stored in TAUQ(i).
!
!     If TAUP is present : taup is stored in TAUP(i).
!
!   The contents of MAT on exit, if TAUQ or TAUP are present,
!   are illustrated by the following examples:
!
!   - m = 6 and n = 5 (m >= n):
!
!        (  u1  v1  v1  v1  v1 )
!
!        (  u1  u2  v2  v2  v2 )
!
!        (  u1  u2  u3  v3  v3 )
!
!        (  u1  u2  u3  u4  v4 )
!
!        (  u1  u2  u3  u4  u5 )
!
!        (  u1  u2  u3  u4  u5 )
!
!   - m = 5 and n = 6 (m < n):
!
!        (  v1  v1  v1  v1  v1  v1 )
!
!        (  u1  v2  v2  v2  v2  v2 )
!
!        (  u1  u2  v3  v3  v3  v3 )
!
!        (  u1  u2  u3  v4  v4  v4 )
!
!        (  u1  u2  u3  u4  v5  v5 )
!
!   where ui denotes an element of the vector defining H(i), and vi an element 
!   of the vector defining G(i).
!
!   Now, let SIGMA(i), i=1,...,N=min(m,n), be the singular values of the intermediate bidiagonal
!   matrix BD in decreasing order of magnitude.
!
!   If DQDS is equal to false, the subroutine computes the LS largest singular
!   values (or the singular values which are greater or equal to THETA) of BD by a bisection
!   method (see the reference (1) below, Sec.8.5 ). The bisection method is applied (implicitly) to
!   the associated N-by-N symmetric tridiagonal matrix BD' * BD  whose eigenvalues are the squares
!   of the singular values of BD by using the differential stationary form of the qd algorithm
!   of Rutishauser (see the reference (2) below, Sec.3.1 ).
!
!   On the other hand, if DQDS is equal to true, all the singular values of BD are computed by the dqds
!   algorithm (see the references (3) and (4).
!
!   When bisection is used, SELECT_SINGVAL_CMP2 subroutine is less accurate, but faster, than SELECT_SINGVAL_CMP
!   subroutine since SELECT_SINGVAL_CMP works on the 2.N-by-2.N symmetric tridiagonal GOLUB-KAHAN form of BD,
!   while SELECT_SINGVAL_CMP2 works implicitly on the associated N-by-N symmetric tridiagonal matrix BD' * BD
!   whose eigenvalues are the squares of the singular values of BD.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!   (3) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (4) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error2
    use Utilities,       only : assert, assert_eq, merror, lascl
    use QR_Procedures,   only : qr_cmp, lq_cmp
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout),  dimension(:,:)         :: mat
    real(stnd),  intent(out),    dimension(:)           :: s
    real(stnd),  intent(in),                   optional :: abstol, theta
    real(stnd),  intent(out),    dimension(:), optional :: tauq, taup, d, e
!
    integer(i4b), intent(out)          :: nsing
    integer(i4b), intent(in), optional :: mul_size, ls
!
    logical(lgl), intent(out)          :: failure
    logical(lgl), intent(in), optional :: vector, scaling, init, dqds
!
    character , intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: matnorm, tmp
    real(stnd), dimension(size(s)) :: d2, e2, tauq2, taup2
!
    integer(i4b) :: i, j, k, l, m, m2, n, n2, inc, mnthr, i1, i2, info
!
    logical(lgl) :: mat_is_scaled, use_dqds
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_singval_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = assert_eq( min(m,n),         &
                   int(size(s),i4b), &
                   name_proc  )
!
    if ( present(ls) ) then
!
        call assert( logical( ls<=k, lgl ),     &
                     logical( 1_i4b<=ls, lgl ), &
                     name_proc )
!
        if ( present(theta) ) then
            call merror( name_proc//svd_error2 )
        end if
!
    end if
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat  TO BIDIAGONAL FORM. IF max(m,n) EXCEEDS mnthr ,
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE MATRIX
!   mat TO A TRIANGULAR FORM.
!    
    mnthr = int( real( k, stnd )*thr_bd, i4b )
!
!   TEST OPTIONAL ARGUMENTS.
!
    if ( present(d) ) then
        call assert( logical( k==size(d), lgl ), name_proc )
    end if
!
    if ( present(e) ) then
        call assert( logical( k==size(e), lgl ), name_proc )
    end if
!
    if ( present(tauq) ) then
        call assert( logical( k==size(tauq), lgl ), name_proc )
        mnthr = huge( i )
    end if
!
    if ( present(taup) ) then
        call assert( logical( k==size(taup), lgl ), name_proc )
        mnthr = huge( i )
    end if
!
    if ( present(mul_size) ) then
        inc = merge( mul_size, 32_i4b, 1_i4b<=mul_size .and. mul_size<=n )
    else
        inc = 32_i4b
    end if
!
    if ( present(ls) .or. present(theta) ) then
        use_dqds = false
    else
        use_dqds = true
    end if
!
    if ( present(dqds) ) then
        use_dqds = dqds
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) then
!
        nsing = 0_i4b
!
        return
!
    end if
!
!   COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, n )
        tmp = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       mat IS ZERO, GENERATE Q AND P IN PACKED FORM.
!
        if ( m>=n ) then
!
            do i = 1_i4b, k-1_i4b
                mat(i,i:i+1_i4b) = one
            end do
!
        else
!
            do i = 1_i4b, k-1_i4b
                mat(i:i+1_i4b,i) = one
            end do
!
        end if
!
        mat(k,k) = one
!
        nsing = k
        s(:k) = zero
!
        if ( present(d) )    d(:k) = zero
        if ( present(e) )    e(:k) = zero
        if ( present(tauq) ) tauq(:k) = zero
        if ( present(taup) ) taup(:k) = zero
!
        return
!
   else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, bignum )
!
    end if
!
    m2 = m
    n2 = n
!
    if (  m>mnthr  ) then
!
!       m MUCH LARGER THAN n .
!
!       FIRST REDUCE mat USING THE QR DECOMPOSITION.
!
!       COMPUTE mat = Q * R .
!
        call qr_cmp( mat, tauq2, taup2 )
!
!       RECONSTRUCT R IN mat, ZEROING OUT BELOW IT.
!
!$OMP  PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)         &
!$OMP             ,SCHEDULE(STATIC,1), PRIVATE(i)
!
        do i = 1_i4b, n
            mat(i,i) = tauq2(i)
            mat(i+1_i4b:n,i) = zero
        end do
!
!$OMP END PARALLEL DO
!
        m2 = n
!
    else if (  n>mnthr ) then
!
!       n MUCH LARGER THAN m.
!
!       FIRST REDUCE mat USING THE LQ DECOMPOSITION.
!
!       COMPUTE mat = L * Q .
!
        call lq_cmp( mat, tauq2, taup2 )
!
!       RECONSTRUCT L IN mat , ZEROING ABOUT ABOVE IT.
!
!$OMP  PARALLEL DO IF((m*(m+1_i4b))/2_i4b>=omp_limit)         &
!$OMP             ,SCHEDULE(STATIC,1), PRIVATE(i)
!
        do i = 1_i4b, m
            mat(1_i4b:i-1_i4b,i) = zero
            mat(i,i) = tauq2(i)
        end do
!
!$OMP END PARALLEL DO
!
        n2 = m
!
    end if
!
!   REDUCE mat TO LOWER OR UPPER BIDIAGONAL FORM.
!
    call bd_cmp_a( mat(:m2,:n2), d2, e2, tauq2, taup2 )
!
!   SAVE INTERMEDIATE QUANTITIES, IF REQUIRED.
!
    if ( present(d) ) then
        d = d2
    end if
!
    if ( present(e) ) then
        e = e2
    end if
!
    if ( present(tauq) ) then
        tauq = tauq2
    end if
!
    if ( present(taup) ) then
        taup = taup2
    end if
!
    if ( use_dqds ) then
!
!       COMPUTE SINGULAR VALUES BY THE DQDS ALGORITHM.
!
        call bd_lasq1( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling, info=info )
!
        if ( info==2_i4b ) then
!
            call bd_dqds( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling )
!
        end if
!
        if ( .not.failure ) then
!
            s(:k) = d2(:k)
!
        else
!
            s(:k) = -one
!
        end if
!
        nsing = k
!
    else
!
!       COMPUTE THE SINGULAR VALUES BY THE BISECTION METHOD APPLIED (IMPLICITLY)
!       TO THE PRODUCT OF THE TRANSPOSE OF THE BIDIAGONAL MATRIX WITH ITSELF.
!
        call bd_singval2( d2, e2, nsing, s, failure, sort=sort,               &
                          vector=vector, abstol=abstol, ls=ls, theta=theta,   &
                          scaling=scaling, init=init                          )
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled .and. nsing>0_i4b ) then
!
        call lascl( s(:nsing), merge(smlnum,bignum,matnorm<smlnum) , matnorm )
!
    end if
!
!
! END OF SUBROUTINE select_singval_cmp2_a
! _______________________________________
!
    end subroutine select_singval_cmp2_a
!
! =========================================================================================
!
    subroutine select_singval_cmp2_b( mat, rlmat, nsing, s, failure, sort, mul_size, vector, abstol, &
                                      ls, theta, d, e, tauo, tauq, taup, scaling, init, dqds )
!
! Purpose
! _______
!
!   SELECT_SINGVAL_CMP2 computes all or some of the greatest singular values of
!   a real m-by-n matrix MAT.
!
!   The Singular Value Decomposition (SVD) is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix.  The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!
!   The original matrix MAT is first reduced to upper bidiagonal form
!   by a two-step algorithm :
!
!   - If m >= n, a QR factorization of the real m-by-n matrix MAT is first computed
!
!                        MAT = O * R
!
!   where O is orthogonal and R is upper triangular. In a second step, the n-by-n upper
!   triangular matrix R is reduced to upper bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * R * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix.
!
!   - If m < n, an LQ factorization of the real m-by-n matrix MAT is first computed
!
!                       MAT = L * O
!
!   where O is orthogonal and L is lower triangular. In a second step, the m-by-m lower
!   triangular matrix L is reduced to upper bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * L * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix.
!
!   SELECT_SINGVAL_CMP2 computes O, BD, Q and P.
!
!   The singular values SIGMA of the bidiagonal matrix BD, which are also the singular values of MAT, 
!   are then computed by a bisection algorithm (see the reference (1) below, Sec.8.5 ). The bisection 
!   method is applied (implicitly) to the associated min(m,n)-by-min(m,n) symmetric tridiagonal matrix BD' * BD  
!   whose eigenvalues are the squares of the singular values of BD by using the differential stationary
!   form of the qd algorithm of Rutishauser (see the reference (2) below, Sec.3.1 ). Alternatively, at
!   the user option, all singular values can be computed by the differential quotient difference with
!   shifts (dqds) algorithm (see the references (3) and (4)).
!
!   The routine outputs (parts of) SIGMA, and optionally O, Q and P (in packed form), and
!   BD for a given matrix MAT. The matrix O is stored in factored form in the argument MAT
!   if the optional argument TAUO is present or explicitly computed if this argument is absent.
!   SIGMA, O, Q, P and BD may then be used to obtain selected singular vectors of 
!   MAT with subroutines BD_INVITER2 or BD_DEFLATE2.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n matrix MAT.
!
!                On exit, if:
!
!                - m >= n, the elements on and below the diagonal,
!                  with the array TAUO, represent the orthogonal
!                  matrix O of the QR factorization of MAT,
!                  as a product of elementary reflectors,
!                  if the argument TAUO is present. Otherwise,
!                  the argument MAT contains the first n columns
!                  of O on output.
!                - m < n, the elements on and above the diagonal,
!                  with the array TAUO, represent the orthogonal
!                  matrix O of the LQ factorization of MAT,
!                  as a product of elementary reflectors,
!                  if the argument TAUO is present. Otherwise,
!                  the argument MAT contains the first m rows
!                  of O on output.
!
!                See Further Details.
!
!   RLMAT        (OUTPUT) real(stnd), dimension(:,:)
!
!                On exit, the elements on and below the diagonal,
!                with the array TAUQ, represent the orthogonal
!                matrix Q as a product of elementary reflectors, 
!                and the elements above the diagonal, with the
!                array TAUP, represent the orthogonal matrix P
!                as a product of elementary reflectors;
!
!                See Further Details.
!
!                The shape of RLMAT must verify:
!                size( RLMAT, 1 ) = size( RLMAT, 2 ) = min( size(MAT,1) , size(MAT,2) ).
!
!   NSING        (OUTPUT) integer(i4b)
!                On output, NSING specifies the number of singular values which have
!                been computed. Note that NSING may be greater than the optional argument
!                LS, if multiple singular values at index LS make unique selection
!                impossible.
!
!                If none of the optional arguments LS and THETA are used, NSING is set to
!                min( size(MAT,1) , size(MAT,2) ) and all the singular values are computed.
!
!   S            (OUTPUT) real(stnd), dimension(:)
!                On exit, S(1:NSING)  contains the first NSING  singular values of MAT.
!                The other values in S ( S(NSING+1:) ) are flagged  by a quiet NAN.
!
!                The size of S must be min( size(MAT,1) , size(MAT,2) ).
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection algorithm converged for all
!                  the computed singular values to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the singular values failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic. The sign of the incorrect singular values is set to negative.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MUL_SIZE     (INPUT, OPTIONAL) integer(i4b)
!                Internal parameter. MUL_SIZE must verify: 1 <= MUL_SIZE <= n,
!                otherwise a default value is used. For better performance, 
!                at the expense of more workspace, a large value can be used.
!
!                The default is min( 32, n ) .
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to TRUE, a vectorized version of the bisection algorithm
!                is used to compute the singular values SIGMA of the bidiagonal matrix BD.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is VECTOR = false.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the singular values. A singular value
!                (or cluster) is considered to be located in the bisection algorithm if it has
!                has been determined to lie in an interval whose width is ABSTOL or less.
!
!                Singular values will be computed most accurately when ABSTOL is set to the square root of the
!                underflow threshold, sqrt(LAMCH('S')), not zero. 
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | BD' * BD | will be used,
!                where | BD' * BD | means the 1-norm of the tridiagonal matrix BD' * BD ( BD' means the transpose of BD) 
!                and ULP is the machine precision (distance from 1 to the next larger floating point number).
!
!   LS           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LS specifies the number of singular values which must be
!                computed by the subroutine. On output, NSING may be different than LS 
!                if multiple singular values at index LS make unique selection
!                impossible.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LS must be greater than 0 and less or equal to min( size(MAT,1) , size(MAT,2) ).
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is LS = min( size(MAT,1) , size(MAT,2) ).
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry, THETA specifies that the singular values which are greater or equal to
!                THETA must be computed. If none of the singular values are greater or equal to
!                THETA, NSING is set to zero and S(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is THETA = 0.
!
!   D            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The diagonal elements of the intermediate bidiagonal matrix BD
!
!                The size of D must be min( size(MAT,1) , size(MAT,2) ).
!
!   E            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The off-diagonal elements of the intermediate upper bidiagonal matrix BD:
!
!                     E(i) = BD(i-1,i) for i = 2,3,...,k;
!
!                The size of E must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUO         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix O of the QR or LQ decomposition of MAT.
!
!                If the optional argument TAUO is present, the orthogonal matrix O is stored
!                in factored form, as a product of elementary reflectors, in the argument MAT on exit.
!
!                If the optional argument TAUO is absent, the orthogonal matrix O is explicitly generated
!                and stored in the argument MAT on exit.
!
!                See description of the argument MAT above and Further Details below.
!
!                The size of TAUO must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUQ         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix Q. See Further Details.
!
!                The size of TAUQ must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAUP         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix P. See Further Details.
!
!                The size of TAUP must be min( size(MAT,1) , size(MAT,2) ).
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix BD is scaled
!                before computing the singular values.
!
!                The default is to scale the bidiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps are
!                computed from estimates of the eigenvalues of the associated BD' * BD tridiagonal
!                matrix obtained from the Pal-Walker-Kahan algorithm.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!   DQDS         (INPUT, OPTIONAL) logical(lgl)
!                DQDS determines how the singular values are computed.
!
!                If DQDS is set to true, all singular values are computed with the dqds algorithm
!                The dqds algorithm is much faster than bisection when all singular values are
!                requested in many cases and delivers about the same accuracy than bisection.
!
!                If DQDS is set to false, the requested singular values are computed by bisection.
!
!                The default is to use the dqds algorithm if all the singular values are requested
!                and to use bisection if selected singular values are requested (e.g., if the optional
!                arguments LS or THETA are present).
!
!
! Further Details
! _______________
!
!   If m >= n, the matrix O of the QR factorization of MAT is represented as a product
!   of elementary reflectors
!
!            O = W(1) * W(2) * ... * W(n)
!
!   Each W(i) has the form
!
!            W(i) = I + tauo * ( v * v' ) ,
!                      
!   where tauo is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in MAT(i:m,i) and tauo in TAUO(i). If the optional argument
!   TAUO is absent, the first n columns of O are generated and stored in the argument MAT.
!
!   If m < n, The matrix O of the LQ factorization of MAT is represented as a product
!   of elementary reflectors
!
!           O = W(m) * . . . * W(2) * W(1)
!
!   Each W(i) has the form
!
!            W(i) = I + tauo * ( v * v' ) ,
!                      
!   where tauo is a real scalar and v is a real n-element vector with v(1:i-1) = 0.
!   v(i:n) is stored on exit in MAT(i,i:n) and tauo in TAUO(i). If the optional argument
!   TAUO is absent, the first m rows of O are generated and stored in the argument MAT.
!
!   The matrix O is stored in factored form if the optional argument TAUO is present
!   or explicitly computed if this argument is absent.
!
!   A blocked algorithm is used for computing the QR or LQ factorization of MAT. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   After, the initial QR or LQ factorization of MAT, the (upper or lower) triangular
!   matrix is reduced to upper bidiagonal form BD.
!
!   The matrices Q and P of the bidiagonal factorization of the triangular matrix R or L
!   are represented as products of elementary reflectors:
!
!       Q = H(1) * H(2) * . . . * H(k)  and  P = G(1) * G(2) * . . . * G(k-1)
!
!   , where k = min( size(MAT,1) , size(MAT,2) ). Each H(i) and G(i) has the form:
!
!       H(i) = I + tauq * u * u'  and G(i) = I + taup * v * v'
!
!   where tauq and taup are real scalars, and u and v are real vectors;
!   u(1:i-1) = 0 and u(i:min(m,n)) is stored on exit in RLMAT(i:min(m,n),i);
!   v(1:i) = 0 and v(i+1:min(m,n)) is stored on exit in RLMAT(i,i+1:min(m,n));
!   tauq is stored in TAUQ(i) and taup in TAUP(i).
!
!   The contents of RLMAT on exit are illustrated by the following example:
!
!   m = 6 and n = 5 (m >= n):
!
!        (  u1  v1  v1  v1  v1 )
!
!        (  u1  u2  v2  v2  v2 )
!
!        (  u1  u2  u3  v3  v3 )
!
!        (  u1  u2  u3  u4  v4 )
!
!        (  u1  u2  u3  u4  u5 )
!
!   where ui denotes an element of the vector defining H(i), and vi an element 
!   of the vector defining G(i).
!
!   Now, let SIGMA(i), i=1,...,N=min(m,n), be the singular values of the intermediate bidiagonal
!   matrix BD in decreasing order of magnitude.
!
!   If DQDS is equal to false, the subroutine computes the LS largest singular
!   values (or the singular values which are greater or equal to THETA) of BD by a bisection
!   method (see the reference (1) below, Sec.8.5 ). The bisection method is applied (implicitly) to
!   the associated N-by-N symmetric tridiagonal matrix BD' * BD  whose eigenvalues are the squares
!   of the singular values of BD by using the differential stationary form of the qd algorithm
!   of Rutishauser (see the reference (2) below, Sec.3.1 ).
!
!   On the other hand, if DQDS is equal to true, all the singular values of BD are computed by the dqds
!   algorithm (see the references (3) and (4).
!
!   When bisection is used, SELECT_SINGVAL_CMP2 subroutine is less accurate, but faster, than SELECT_SINGVAL_CMP
!   subroutine since SELECT_SINGVAL_CMP works on the 2.N-by-2.N symmetric tridiagonal GOLUB-KAHAN form of BD,
!   while SELECT_SINGVAL_CMP2 works implicitly on the associated N-by-N symmetric tridiagonal matrix BD' * BD
!   whose eigenvalues are the squares of the singular values of BD.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!   (3) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (4) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error2
    use Utilities,       only : assert, assert_eq, merror, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout),  dimension(:,:)         :: mat
    real(stnd),  intent(out),    dimension(:,:)         :: rlmat
    real(stnd),  intent(out),    dimension(:)           :: s
    real(stnd),  intent(in),                   optional :: abstol, theta
    real(stnd),  intent(out),    dimension(:), optional :: tauo, tauq, taup, d, e
!
    integer(i4b), intent(out)          :: nsing
    integer(i4b), intent(in), optional :: mul_size, ls
!
    logical(lgl), intent(out)          :: failure
    logical(lgl), intent(in), optional :: vector, scaling, init, dqds
!
    character , intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: matnorm, tmp
    real(stnd), dimension(size(s)) :: d2, e2, tauq2, taup2
!
    integer(i4b) :: i, j, k, l, m, n, inc, i1, i2, info
!
    logical(lgl) :: mat_is_scaled, use_dqds
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_singval_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = assert_eq( min(m,n),                &
                   int(size(s),i4b),        &
                   int(size(rlmat,1),i4b) , &
                   int(size(rlmat,2),i4b) , &
                   name_proc  )
!
    if ( present(ls) ) then
!
        call assert( logical( ls<=k, lgl ),     &
                     logical( 1_i4b<=ls, lgl ), &
                     name_proc )
!
        if ( present(theta) ) then
            call merror( name_proc//svd_error2 )
        end if
!
    end if
!
!   TEST OPTIONAL ARGUMENTS.
!
    if ( present(d) ) then
        call assert( logical( k==size(d), lgl ), name_proc )
    end if
!
    if ( present(e) ) then
        call assert( logical( k==size(e), lgl ), name_proc )
    end if
!
    if ( present( tauo ) ) then
        call assert( logical( k==size(tauo), lgl ), name_proc )
    end if
!
    if ( present(tauq) ) then
        call assert( logical( k==size(tauq), lgl ), name_proc )
    end if
!
    if ( present(taup) ) then
        call assert( logical( k==size(taup), lgl ), name_proc )
    end if
!
    if ( present(mul_size) ) then
        inc = merge( mul_size, 32_i4b, 1_i4b<=mul_size .and. mul_size<=n )
    else
        inc = 32_i4b
    end if
!
    if ( present(ls) .or. present(theta) ) then
        use_dqds = false
    else
        use_dqds = true
    end if
!
    if ( present(dqds) ) then
        use_dqds = dqds
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) then
!
        nsing = 0_i4b
!
        return
!
    end if
!
!   COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, n )
        tmp = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       mat IS ZERO, GENERATE O, Q AND P IN PACKED FORM.
!
        do i = 1_i4b, k
            mat(i,i) = one
        end do
!
        rlmat(:k,:k) = zero
!
        do i = 1_i4b, k-1_i4b
            rlmat(i,i:i+1_i4b) = one
        end do
!
        rlmat(k,k) = one
!
        nsing = k
        s(:k) = zero
!
        if ( present(d) )    d(:k) = zero
        if ( present(e) )    e(:k) = zero
        if ( present(tauo) ) tauo(:k) = zero
        if ( present(tauq) ) tauq(:k) = zero
        if ( present(taup) ) taup(:k) = zero
!
        return
!
   else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   REDUCE mat TO UPPER BIDIAGONAL FORM BY A TWO-STEP ALGORITHM.
!
    call bd_cmp_c( mat(:m,:n), d2, e2, tauq2, taup2, rlmat(:k,:k), tauo=tauo )
!
!   SAVE INTERMEDIATE QUANTITIES, IF REQUIRED.
!
    if ( present(d) ) then
        d = d2
    end if
!
    if ( present(e) ) then
        e = e2
    end if
!
    if ( present(tauq) ) then
        tauq = tauq2
    end if
!
    if ( present(taup) ) then
        taup = taup2
    end if
!
    if ( use_dqds ) then
!
!       COMPUTE SINGULAR VALUES BY THE DQDS ALGORITHM.
!
        call bd_lasq1( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling, info=info )
!
        if ( info==2_i4b ) then
!
            call bd_dqds( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling )
!
        end if
!
        if ( .not.failure ) then
!
            s(:k) = d2(:k)
!
        else
!
            s(:k) = -one
!
        end if
!
        nsing = k
!
    else
!
!       COMPUTE THE SINGULAR VALUES BY THE BISECTION METHOD APPLIED (IMPLICITLY)
!       TO THE PRODUCT OF THE TRANSPOSE OF THE BIDIAGONAL MATRIX WITH ITSELF.
!
        call bd_singval2( d2, e2, nsing, s, failure, sort=sort,               &
                          vector=vector, abstol=abstol, ls=ls, theta=theta,   &
                          scaling=scaling, init=init                          )
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled .and. nsing>0_i4b ) then
!
        call lascl( s(:nsing), merge(smlnum,bignum,matnorm<smlnum) , matnorm )
!
    end if
!
!
! END OF SUBROUTINE select_singval_cmp2_b
! _______________________________________
!
    end subroutine select_singval_cmp2_b
!
! =========================================================================================
!   
    subroutine select_singval_cmp3_a( mat, nsing, s, failure, sort, mul_size, vector,   &
                                      abstol, ls, theta, d, e, p, gen_p, reortho,       &
                                      scaling, init, dqds, failure_bd )
!
! Purpose
! _______
!
!   SELECT_SINGVAL_CMP3 computes all or some of the greatest singular values of
!   a real m-by-n matrix MAT with m>=n.
!
!   The Singular Value Decomposition (SVD) is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix.  The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!
!   The original matrix MAT is first reduced to upper bidiagonal form BD
!   by an orthogonal transformation:
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal (see the reference (5) below). The Ralha-Barlow
!   one-sided method is used for this purpose (see the references (1) to (3) below).
!
!   The singular values SIGMA of the bidiagonal matrix BD, which are also the singular values
!   of MAT, are then computed by a bisection algorithm applied to the Tridiagonal Golub-Kahan (TGK)
!   of the bidiagonal matrix BD (see the reference (6) below, Sec.3.3) or the differential
!   quotient difference with shifts (dqds) algorithm (see the references (7) and (8)).
!
!   The routine outputs (parts of) SIGMA, Q and optionally P (in packed form) and
!   BD for a given matrix MAT. SIGMA, Q, P and BD may then be used to obtain selected
!   singular vectors of MAT with subroutines BD_INVITER2 or BD_DEFLATE2.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n matrix MAT.
!
!                On exit, MAT is overwritten with the first n columns
!                of Q (stored column-wise), the orthogonal matrix
!                used to reduce MAT to bidiagonal form as returned
!                by subroutine BD_CMP2 in its argument MAT.
!
!                The shape of MAT must verify: 
!                size( MAT, 1 ) >= size( MAT, 2 ) = n .
!
!   NSING        (OUTPUT) integer(i4b)
!                On output, NSING specifies the number of singular values which have
!                been computed. Note that NSING may be greater than the optional argument
!                LS, if multiple singular values at index LS make unique selection
!                impossible.
!
!                If none of the optional arguments LS and THETA are used, NSING is set to
!                size(MAT,2) and all the singular values are computed.
!
!   S            (OUTPUT) real(stnd), dimension(:)
!                On exit, S(1:NSING)  contains the first NSING  singular values of MAT.
!                The other values in S ( S(NSING+1:) ) are flagged  by a quiet NAN.
!
!                The size of S must be equal to size( MAT, 2 ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection or dqds algorithm
!                  converged for all the computed singular values to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the singular values failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic. The sign of the incorrect singular values is set to negative.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MUL_SIZE     (INPUT, OPTIONAL) integer(i4b)
!                Internal parameter. MUL_SIZE must verify: 1 <= MUL_SIZE <= n,
!                otherwise a default value is used. For better performance, 
!                at the expense of more workspace, a large value can be used.
!
!                The default is min( 32, n ).
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, the absolute tolerance for the singular values. A singular value
!                (or cluster) is considered to be located in the bisection algorithm if it has
!                has been determined to lie in an interval whose width is ABSTOL or less.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is VECTOR = false.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the singular values. A singular value
!                (or cluster) is considered to be located in the bisection algorithm if it
!                been determined to lie in an interval whose width is ABSTOL or less.
!
!                Singular values will be computed most accurately when ABSTOL is set to the square root of
!                the underflow threshold (e.g., sqrt(safmin)=sqrt(LAMCH('S')), not zero. 
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | T(GK) | will be used,
!                where | T(GK) | means the 1-norm of the GOLUB-KAHAN tridiagonal form of the bidiagonal matrix BD
!                and ULP is the machine precision (distance from 1 to the next larger floating point number).
!
!                This argument has no effect if DQDS is equal to true.
!
!   LS           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LS specifies the number of singular values which must be
!                computed by the subroutine. On output, NSING may be different than LS 
!                if multiple singular values at index LS make unique selection
!                impossible.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LS must be greater than 0 and less or equal to size( MAT, 2 ) = n .
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is LS = size( MAT, 2 ) = n .
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry, THETA specifies that the singular values which are greater or equal to
!                THETA must be computed. If none of the singular values are greater or equal to
!                THETA, NSING is set to zero and S(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is THETA = 0 .
!
!   D            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The diagonal elements of the intermediate bidiagonal matrix BD
!
!                The size of D must be equal to size( MAT, 2 ) = n .
!
!   E            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The off-diagonal elements of the intermediate upper bidiagonal matrix BD:
!
!                            E(i) = B(i-1,i) for i = 2,3,...,n;
!
!                E(1) is arbitrary.
!
!                The size of E must be equal to size( MAT, 2 ) = n .
!
!   P            (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!                On exit, P is overwritten with the n-by-n orthogonal matrix P
!                (stored column-wise or in packed form), the orthogonal matrix
!                used to reduce MAT to bidiagonal form as returned
!                by subroutine BD_CMP2 in its argument P.
!
!                The shape of P must verify: 
!                size( P, 1 ) = size( P, 2 ) = n .
!
!   GEN_P        (INPUT, OPTIONAL) logical(lgl)
!                On entry, this optional argument has an effect only if the optional argument
!                P is also used.
!
!                In this case, if the optional argument GEN_P is used and is set to true, the
!                orthogonal matrix P used to reduce MAT to bidiagonal form is generated on output
!                of the subroutine in its argument P.
!
!                If GEN_P is set to false, the orthogonal matrix P is stored in factored
!                form as products of elementary reflectors in the lower triangle of the array P.
!
!                See the description of BD_CMP2 subroutine for more details.
!
!                The default is true.
!
!   REORTHO      (INPUT, OPTIONAL) logical(lgl)
!                If the optional argument REORTHO is set to true, reorthogonalization of the
!                matrix Q, used to reduce MAT to bidiagonal form, is performed (when needed).
!                If this optional logical argument is set to false, reorthogonalization of Q is never
!                performed and the columns of the  matrix Q can be far from orthogonal to each other
!                for input matrices MAT with a large condition number.
!
!                See the description of BD_CMP2 subroutine for more details.
!
!                The default is REORTHO = false.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix BD is scaled
!                before computing the singular values.
!
!                The default is to scale the bidiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps are
!                computed from estimates of the eigenvalues of the associated BD' * BD tridiagonal
!                matrix obtained from the Pal-Walker-Kahan algorithm.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!   DQDS         (INPUT, OPTIONAL) logical(lgl)
!                DQDS determines how the singular values are computed.
!
!                If DQDS is set to true, all singular values are computed with the dqds algorithm
!                The dqds algorithm is much faster than bisection when all singular values are
!                requested in many cases and delivers about the same accuracy than bisection.
!
!                If DQDS is set to false, the requested singular values are computed by bisection.
!
!                The default is to use the dqds algorithm if all the singular values are requested
!                and to use bisection if selected singular values are requested (e.g., if the optional
!                arguments LS or THETA are present).
!
!   FAILURE_BD   (OUTPUT, OPTIONAL) logical(lgl)
!                On exit:
!
!                - FAILURE_BD = false :  indicates that maximum accuracy was obtained in the
!                  Ralha-Barlow one-sided bidiagonalization of MAT.
!                - FAILURE_BD = true  :  indicates that MAT is nearly singular and some
!                  loss of orthogonality can be expected in the Ralha-Barlow
!                  bidiagonalization algorithm.
!
!                See description of the BD_CMP2 subroutine for further details.
!
!
! Further Details
! _______________
!
!   The matrices Q, P and BD are computed with the help of the Ralha-Barlow one-sided method.
!   Q is computed by a recurrence relationship and the first n columns of Q are stored in the
!   argument MAT on exit. P is computed as a product of n-1 elementary reflectors
!   (e.g. Householder transformations):
!
!       P = G(1) * G(2) * . . . * G(n-1)
!
!   Each G(i) has the form:
!
!       G(i) = I + taup * v * v'
!
!   where taup is a real scalar, and v is a real vector. IF GEN_P is used and set to false,
!   the n-1 G(i) elementary reflectors are stored in the lower triangle of the array P.
!
!   For the G(i) reflector, taup is stored in P(i+1,1) and v is stored in P(i+1:n,i+1).
!   In addition, P(1,1) is set to -1 if GEN_P=false and is equal to 1 if GEN_P=true.
!
!   In other words, the value of P(1,1) indicates if the orthogonal matrix P is stored
!   in factored form or not. Note that if n is equal to 1, elementary reflectors are not needed
!   and consequently P(1,1) is set to 1, independently of the value of GEN_P.
!
!   This is the blocked version of the Ralha-Barlow one-sided algorithm. See the references (1),
!   (2) and (3) for further details. Furthermore, the algorithm is parallelized if OPENMP is used.
!
!   Since Q is computed by a recurrence relationship, a loss of orthogonality of Q can be
!   observed when the rectangular matrix MAT is singular or nearly singular or has a large
!   condition number, see the reference (2) for details.
!
!   To correct partly this deficiency, partial reorthogonalization can be performed to ensure
!   orthogonality at the expense of speed of computation. The reorthogonalization uses the Gram-Schmidt
!   method described in the reference (4) and is activated if the optional logical argument REORTHO is
!   set to true.
!
!   The reference (2) also explains how to handle the case of an exactly singular
!   matrix MAT (a very rare event). However, in this subroutine, the partial reorthogonalization
!   described in the reference (4), which is used when REORTHO is set true, corrects automatically
!   this problem.
!
!   Now, let SIGMA(i), i=1,...,n, be the singular values of the intermediate bidiagonal
!   matrix BD in decreasing order of magnitude.
!
!   If DQDS is equal to false, the subroutine computes the LS largest singular
!   values (or the singular values which are greater or equal to THETA) of BD by a bisection
!   method (see the reference (5) below, Sec.8.5 ). The bisection method is applied to an associated 
!   2.n-by-2.n symmetric tridiagonal matrix T (the so-called GOLUB-KAHAN form of BD) whose 
!   eigenvalues are the singular values of BD and their negatives (see the reference (6) below).
!
!   On the other hand, if DQDS is equal to true, all the singular values of BD are computed by the dqds
!   algorithm (see the references (7) and (8).
!
!   For further details, see:
!
!   (1) Ralha, R.M.S., 2003:
!           One-sided reduction to bidiagonal form.
!           Linear Algebra Appl., No 358, pp. 219-238.
!
!   (2) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!           A new stable bidiagonal reduction algorithm.
!           Linear Algebra Appl., No 397, pp. 35-84.
!
!   (3) Bosner, N., and Barlow, J.L., 2007:
!           Block and Parallel versions of one-sided bidiagonalization.
!           SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!   (4) Stewart, G.W., 2007:
!           Block Gram-Schmidt Orthogonalization.
!           Report TR-4823, Department of Computer Science, College Park, University of Maryland.
!
!   (5) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (6) Fernando, K.V., 1998:
!           Accurately counting singular values of bidiagonal matrices and
!           eigenvalues of skew-symmetric tridiagonal matrices.
!           SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!   (7) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (8) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error2, svd_error5
    use Utilities,       only : assert, assert_eq, merror, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout),  dimension(:,:)           :: mat
    real(stnd),  intent(out),    dimension(:)             :: s
    real(stnd),  intent(in),                     optional :: abstol, theta
    real(stnd),  intent(out),    dimension(:),   optional :: d, e
    real(stnd),  intent(out),    dimension(:,:), optional :: p
!
    integer(i4b), intent(out)          :: nsing
    integer(i4b), intent(in), optional :: mul_size, ls
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: gen_p, reortho, vector, scaling, init, dqds
    logical(lgl), intent(out), optional :: failure_bd
!
    character , intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: matnorm, tmp
    real(stnd), dimension(size(s)) :: d2, e2
!
    integer(i4b) :: i, j, l, m, n, inc, i1, i2, info
!
    logical(lgl) :: mat_is_scaled, use_dqds, do_comp_p, do_gen_p, do_reortho
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_singval_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
    if ( present( failure_bd ) ) then
!
        failure_bd = false
!
    end if
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b),  &
                   int(size(s),i4b),      &
                   name_proc  )
!
    if ( m<n ) then
!
        call merror( name_proc//svd_error5 )
!
    end if
!
!   TEST OPTIONAL ARGUMENTS.
!
    if ( present(ls) ) then
!
        call assert( logical( ls<=n, lgl ),     &
                     logical( 1_i4b<=ls, lgl ), &
                     name_proc )
!
        if ( present(theta) ) then
            call merror( name_proc//svd_error2 )
        end if
!
    end if
!
    do_gen_p = true
!
    if ( present(p) ) then
!
        call assert( logical( int(size(p,1),i4b)==n, lgl ),   &
                     logical( int(size(p,2),i4b)==n, lgl ),   &
                     name_proc )
!
        do_comp_p = true
!
        if ( present( gen_p ) ) then
!
            do_gen_p = gen_p
!
        end if
!
    else
!
        do_comp_p = false
!
    end if
!
    do_reortho = false
!
    if ( present( reortho ) ) then
!
        do_reortho = reortho
!
    end if
!
    if ( present(ls) .or. present(theta) ) then
!
        use_dqds = false
!
    else
!
        use_dqds = true
!
    end if
!
    if ( present(dqds) ) then
!
        use_dqds = dqds
!
    end if
!
    if ( present(d) ) then
        call assert( logical( n==size(d), lgl ), name_proc )
    end if
!
    if ( present(e) ) then
        call assert( logical( n==size(e), lgl ), name_proc )
    end if
!
    if ( present(mul_size) ) then
        inc = merge( mul_size, 32_i4b, 1_i4b<=mul_size .and. mul_size<=n )
    else
        inc = 32_i4b
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
!
        nsing = 0_i4b
!
        return
!
    end if
!
!   COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, n )
        tmp = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       mat IS THE ZERO MATRIX, GENERATE Q AND P.
!
        nsing = n
        s(:n) = zero
!
        do i = 1_i4b, n
            mat(i,i) = one
        end do
!
        if ( present(d) )    d(:n) = zero
        if ( present(e) )    e(:n) = zero
!
        if ( do_comp_p ) then
!
            do i = 1_i4b, n
                p(1_i4b:n,i) = zero
                p(i,i)       = one
            end do
!
            if ( n/=1_i4b .and. .not.do_gen_p ) then
                p(1_i4b,1_i4b) = -one
            end if
!
        end if
!
        return
!
    else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   mat HAS AT LEAST AS MANY ROWS AS COLUMNS.
!   FIRST BIDIAGONALIZE mat WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
    if (  do_comp_p ) then
!
!       P IS REQUIRED, USE bd_cmp2_a.
!
        call bd_cmp2_a( mat(:m,:n), d2(:n), e2(:n), p(:n,:n), failure=failure_bd, gen_p=do_gen_p, reortho=do_reortho )
!
    else
!
!       P IS NOT REQUIRED, USE bd_cmp2_b.
!
        call bd_cmp2_b( mat(:m,:n), d2(:n), e2(:n), failure=failure_bd, reortho=do_reortho )
!
    end if
!
!   SAVE INTERMEDIATE QUANTITIES, IF REQUIRED.
!
    if ( present(d) ) then
        d(:n) = d2(:n)
    end if
!
    if ( present(e) ) then
        e(:n) = e2(:n)
    end if
!
    if ( use_dqds ) then
!
!       COMPUTE SINGULAR VALUES BY THE DQDS ALGORITHM.
!
        call bd_lasq1( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling, info=info )
!
        if ( info==2_i4b ) then
!
            call bd_dqds( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling )
!
        end if
!
        if ( .not.failure ) then
!
            s(:n) = d2(:n)
!
        else
!
            s(:n) = -one
!
        end if
!
        nsing = n
!
    else
!
!       COMPUTE THE SINGULAR VALUES BY THE BISECTION METHOD APPLIED TO THE GOLUB-KAHAN
!       FORM OF THE BIDIAGONAL MATRIX.
!
        call bd_singval( d2, e2, nsing, s, failure, sort=sort,    &
                         vector=vector, abstol=abstol, ls=ls,     &
                         theta=theta, scaling=scaling, init=init  )
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled .and. nsing>0_i4b ) then
!
        call lascl( s(:nsing), merge(smlnum,bignum,matnorm<smlnum) , matnorm )
!
    end if
!
!
! END OF SUBROUTINE select_singval_cmp3_a
! _______________________________________
!
    end subroutine select_singval_cmp3_a
!
! =========================================================================================
!
    subroutine select_singval_cmp3_b( mat, rmat, nsing, s, failure, sort, mul_size, vector,   &
                                      abstol, ls, theta, d, e, tauo, p, gen_p, reortho,       &
                                      scaling, init, dqds, failure_bd )
!
! Purpose
! _______
!
!   SELECT_SINGVAL_CMP3 computes all or some of the greatest singular values of
!   a real m-by-n matrix MAT with m>=n.
!
!   The Singular Value Decomposition (SVD) is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix.  The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!
!   The original matrix MAT is first reduced to upper bidiagonal form
!   by a two-step algorithm :
!
!   A QR factorization of the real m-by-n matrix MAT is first computed
!
!                        MAT = O * R
!
!   where O is orthogonal and R is upper triangular.
!
!   In a second step, the n-by-n upper triangular matrix R is reduced to upper
!   bidiagonal form BD by an orthogonal transformation :
!
!                       Q' * R * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix (see the reference (5) below).
!   The Ralha-Barlow one-sided method is used in this second step (see the references (1) to (3) below).
!
!   SELECT_SINGVAL_CMP3 computes O, BD, Q and P.
!
!   The singular values SIGMA of the bidiagonal matrix BD, which are also the singular values
!   of MAT, are then computed by a bisection algorithm applied to the Tridiagonal Golub-Kahan (TGK)
!   of the bidiagonal matrix BD (see the reference (6) below, Sec.3.3) or the differential
!   quotient difference with shifts (dqds) algorithm (see the references (7) and (8)).
!
!   The routine outputs (parts of) SIGMA, and optionally O, Q and P and BD for a given matrix MAT. 
!   The matrix O is stored in factored form in the argument MAT if the optional argument TAUO is present
!   or explicitly computed if this argument is absent. The first n columns of Q are stored in the argument
!   RMAT. Finally, P is stored in the optional argument P. P is stored in factored form or explicitly generated
!   depending on the value of the optional logical argument GEN_P.
!
!   SIGMA, O, Q, P and BD may then be used to obtain selected singular vectors of 
!   MAT with subroutines BD_INVITER2 or BD_DEFLATE2.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n matrix MAT.
!
!                On exit, the elements on and below the diagonal,
!                with the array TAUO, represent the orthogonal
!                matrix O of the QR factorization of MAT,
!                as a product of elementary reflectors,
!                if the argument TAUO is present. Otherwise,
!                the argument MAT contains the first n columns
!                of O (stored column-wise) on output.
!
!                See Further Details.
!
!                The shape of MAT must verify: 
!                size( MAT, 1 ) >= size( MAT, 2 ) = n .
!
!   RMAT         (OUTPUT) real(stnd), dimension(:,:)
!
!                On exit, RMAT contains the first n columns
!                of Q (stored column-wise), the orthogonal matrix
!                used to reduce R to bidiagonal form as returned
!                by subroutine BD_CMP2 in its argument MAT.
!
!                See Further Details.
!
!                The shape of RMAT must verify:
!                size( RMAT, 1 ) = size( RMAT, 2 ) = size( MAT, 2 ) = n .
!
!   NSING        (OUTPUT) integer(i4b)
!                On output, NSING specifies the number of singular values which have
!                been computed. Note that NSING may be greater than the optional argument
!                LS, if multiple singular values at index LS make unique selection
!                impossible.
!
!                If none of the optional arguments LS and THETA are used, NSING is set to
!                size(MAT,2) and all the singular values are computed.
!
!   S            (OUTPUT) real(stnd), dimension(:)
!                On exit, S(1:NSING)  contains the first NSING  singular values of MAT.
!                The other values in S ( S(NSING+1:) ) are flagged  by a quiet NAN.
!
!                The size of S must be equal to size( MAT, 2 ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection or dqds algorithm
!                  converged for all the computed singular values to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the singular values failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic. The sign of the incorrect singular values is set to negative.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MUL_SIZE     (INPUT, OPTIONAL) integer(i4b)
!                Internal parameter. MUL_SIZE must verify: 1 <= MUL_SIZE <= n,
!                otherwise a default value is used. For better performance, 
!                at the expense of more workspace, a large value can be used.
!
!                The default is min( 32, n ).
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to TRUE, a vectorized version of the bisection algorithm
!                is used to compute the singular values SIGMA of the bidiagonal matrix BD.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is VECTOR = false.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the singular values. A singular value
!                (or cluster) is considered to be located in the bisection algorithm if it
!                been determined to lie in an interval whose width is ABSTOL or less.
!
!                Singular values will be computed most accurately when ABSTOL is set to the square root of
!                the underflow threshold (e.g., sqrt(safmin)=sqrt(LAMCH('S')), not zero. 
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | T(GK) | will be used,
!                where | T(GK) | means the 1-norm of the GOLUB-KAHAN tridiagonal form of the bidiagonal matrix BD
!                and ULP is the machine precision (distance from 1 to the next larger floating point number).
!
!                This argument has no effect if DQDS is equal to true.
!
!   LS           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LS specifies the number of singular values which must be
!                computed by the subroutine. On output, NSING may be different than LS 
!                if multiple singular values at index LS make unique selection
!                impossible.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LS must be greater than 0 and less or equal to size( MAT, 2 ) = n .
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is LS = size( MAT, 2 ) = n .
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry, THETA specifies that the singular values which are greater or equal to
!                THETA must be computed. If none of the singular values are greater or equal to
!                THETA, NSING is set to zero and S(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is THETA = 0 .
!
!   D            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The diagonal elements of the intermediate bidiagonal matrix BD
!
!                The size of D must be equal to size( MAT, 2 ) = n .
!
!   E            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The off-diagonal elements of the intermediate upper bidiagonal matrix BD:
!
!                            E(i) = B(i-1,i) for i = 2,3,...,n;
!
!                E(1) is arbitrary.
!
!                The size of E must be equal to size( MAT, 2 ) = n .
!
!   TAUO         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix O of the QR decomposition of MAT.
!
!                If the optional argument TAUO is present, the orthogonal matrix O is stored
!                in factored form, as a product of elementary reflectors, in the argument MAT on exit.
!
!                If the optional argument TAUO is absent, the first n columns of the orthogonal matrix O
!                are explicitly generated and stored in the argument MAT on exit.
!
!                See description of the argument MAT above and Further Details below.
!
!                The size of TAUO must be equal to size( MAT, 2 ) = n .
!
!   P            (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!                On exit, P is overwritten with the n-by-n orthogonal matrix P
!                (stored column-wise or in packed form), the orthogonal matrix
!                used to reduce MAT to bidiagonal form as returned
!                by subroutine BD_CMP2 in its argument P.
!
!                The shape of P must verify: 
!                size( P, 1 ) = size( P, 2 ) = n .
!
!   GEN_P        (INPUT, OPTIONAL) logical(lgl)
!                On entry, this optional argument has an effect only if the optional argument
!                P is also used.
!
!                In this case, if the optional argument GEN_P is used and is set to true, the
!                orthogonal matrix P used to reduce MAT to bidiagonal form is generated on output
!                of the subroutine in its argument P.
!
!                If GEN_P is set to false, the orthogonal matrix P is stored in factored
!                form as products of elementary reflectors in the lower triangle of the array P.
!
!                See the description of BD_CMP2 subroutine for more details.
!
!                The default is true.
!
!   REORTHO      (INPUT, OPTIONAL) logical(lgl)
!                If the optional argument REORTHO is set to true, reorthogonalization of the
!                matrix Q, used to reduce MAT to bidiagonal form, is performed (when needed).
!                If this optional logical argument is set to false, reorthogonalization of Q is never
!                performed and the columns of the  matrix Q can be far from orthogonal to each other
!                for input matrices MAT with a large condition number.
!
!                See the description of BD_CMP2 subroutine for more details.
!
!                The default is REORTHO = false.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix BD is scaled
!                before computing the singular values.
!
!                The default is to scale the bidiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps are
!                computed from estimates of the eigenvalues of the associated BD' * BD tridiagonal
!                matrix obtained from the Pal-Walker-Kahan algorithm.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!   DQDS         (INPUT, OPTIONAL) logical(lgl)
!                DQDS determines how the singular values are computed.
!
!                If DQDS is set to true, all singular values are computed with the dqds algorithm
!                The dqds algorithm is much faster than bisection when all singular values are
!                requested in many cases and delivers about the same accuracy than bisection.
!
!                If DQDS is set to false, the requested singular values are computed by bisection.
!
!                The default is to use the dqds algorithm if all the singular values are requested
!                and to use bisection if selected singular values are requested (e.g., if the optional
!                arguments LS or THETA are present).
!
!   FAILURE_BD   (OUTPUT, OPTIONAL) logical(lgl)
!                On exit:
!
!                - FAILURE_BD = false :  indicates that maximum accuracy was obtained in the
!                  Ralha-Barlow one-sided bidiagonalization of MAT.
!                - FAILURE_BD = true  :  indicates that MAT is nearly singular and some
!                  loss of orthogonality can be expected in the Ralha-Barlow
!                  bidiagonalization algorithm.
!
!                See description of the BD_CMP2 subroutine for further details.
!
!
! Further Details
! _______________
!
!   the matrix O of the QR factorization of MAT is represented as a product
!   of elementary reflectors
!
!            O = W(1) * W(2) * ... * W(n)
!
!   Each W(i) has the form
!
!            W(i) = I + tauo * ( v * v' ) ,
!                      
!   where tauo is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in MAT(i:m,i) and tauo in TAUO(i). If the optional argument
!   TAUO is absent, the first n columns of O are generated and stored in the argument MAT.
!
!   The matrix O is stored in factored form if the optional argument TAUO is present
!   or explicitly computed if this argument is absent.
!
!   A blocked algorithm is used for computing the QR factorization of MAT. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   After, the initial QR factorization of MAT, the upper triangular
!   matrix R is reduced to upper bidiagonal form BD:
!
!                       Q' * R * P = BD
!
!   The matrices Q, P and BD are computed with the help of the Ralha-Barlow one-sided method.
!   Q is computed by a recurrence relationship and the first n columns of Q are stored in the
!   argument RMAT on exit. P is computed as a product of n-1 elementary reflectors
!   (e.g. Householder transformations):
!
!       P = G(1) * G(2) * . . . * G(n-1)
!
!   Each G(i) has the form:
!
!       G(i) = I + taup * v * v'
!
!   where taup is a real scalar, and v is a real vector. IF GEN_P is used and set to false,
!   the n-1 G(i) elementary reflectors are stored in the lower triangle of the array P.
!
!   For the G(i) reflector, taup is stored in P(i+1,1) and v is stored in P(i+1:n,i+1).
!   In addition, P(1,1) is set to -1 if GEN_P=false and is equal to 1 if GEN_P=true.
!
!   In other words, the value of P(1,1) indicates if the orthogonal matrix P is stored
!   in factored form or not. Note that if n is equal to 1, elementary reflectors are not needed
!   and consequently P(1,1) is set to 1, independently of the value of GEN_P.
!
!   This is the blocked version of the Ralha-Barlow one-sided algorithm. See the references (1),
!   (2) and (3) for further details. Furthermore, the algorithm is parallelized if OPENMP is used.
!
!   Since Q is computed by a recurrence relationship, a loss of orthogonality of Q can be
!   observed when the rectangular matrix MAT is singular or nearly singular or has a large
!   condition number, see the reference (2) for details.
!
!   To correct partly this deficiency, partial reorthogonalization can be performed to ensure
!   orthogonality at the expense of speed of computation. The reorthogonalization uses the Gram-Schmidt
!   method described in the reference (4) and is activated if the optional logical argument REORTHO is
!   set to true.
!
!   The reference (2) also explains how to handle the case of an exactly singular
!   matrix MAT (a very rare event). However, in this subroutine, the partial reorthogonalization
!   described in the reference (4), which is used when REORTHO is set true, corrects automatically
!   this problem.
!
!   Now, let SIGMA(i), i=1,...,n, be the singular values of the intermediate bidiagonal
!   matrix BD in decreasing order of magnitude.
!
!   If DQDS is equal to false, the subroutine computes the LS largest singular
!   values (or the singular values which are greater or equal to THETA) of BD by a bisection
!   method (see the reference (5) below, Sec.8.5 ). The bisection method is applied to an associated 
!   2.n-by-2.n symmetric tridiagonal matrix T (the so-called GOLUB-KAHAN form of BD) whose 
!   eigenvalues are the singular values of BD and their negatives (see the reference (6) below).
!
!   On the other hand, if DQDS is equal to true, all the singular values of BD are computed by the dqds
!   algorithm (see the references (7) and (8).
!
!   For further details, see:
!
!   (1) Ralha, R.M.S., 2003:
!           One-sided reduction to bidiagonal form.
!           Linear Algebra Appl., No 358, pp. 219-238.
!
!   (2) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!           A new stable bidiagonal reduction algorithm.
!           Linear Algebra Appl., No 397, pp. 35-84.
!
!   (3) Bosner, N., and Barlow, J.L., 2007:
!           Block and Parallel versions of one-sided bidiagonalization.
!           SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!   (4) Stewart, G.W., 2007:
!           Block Gram-Schmidt Orthogonalization.
!           Report TR-4823, Department of Computer Science, College Park, University of Maryland.
!
!   (5) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (6) Fernando, K.V., 1998:
!           Accurately counting singular values of bidiagonal matrices and
!           eigenvalues of skew-symmetric tridiagonal matrices.
!           SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!   (7) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (8) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error2, svd_error5
    use Utilities,       only : assert, assert_eq, merror, lascl
    use QR_Procedures,   only : qr_cmp, ortho_gen_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout),  dimension(:,:)           :: mat
    real(stnd),  intent(out),    dimension(:,:)           :: rmat
    real(stnd),  intent(out),    dimension(:)             :: s
    real(stnd),  intent(in),                     optional :: abstol, theta
    real(stnd),  intent(out),    dimension(:),   optional :: d, e, tauo
    real(stnd),  intent(out),    dimension(:,:), optional :: p
!
    integer(i4b), intent(out)          :: nsing
    integer(i4b), intent(in), optional :: mul_size, ls
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: gen_p, reortho, vector, scaling, init, dqds
    logical(lgl), intent(out), optional :: failure_bd
!
    character , intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: matnorm, tmp
    real(stnd), dimension(size(s)) :: d2, e2
!
    integer(i4b) :: i, j, l, m, n, inc, i1, i2, info
!
    logical(lgl) :: mat_is_scaled, use_dqds, do_comp_p, do_gen_p, do_gen_o, do_reortho
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_singval_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
    if ( present( failure_bd ) ) then
!
        failure_bd = false
!
    end if
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b),   &
                   int(size(rmat,1),i4b) , &
                   int(size(rmat,2),i4b) , &
                   int(size(s),i4b),       &
                   name_proc  )
!
    if ( m<n ) then
!
        call merror( name_proc//svd_error5 )
!
    end if
!
!   TEST OPTIONAL ARGUMENTS.
!
    if ( present(ls) ) then
!
        call assert( logical( ls<=n, lgl ),     &
                     logical( 1_i4b<=ls, lgl ), &
                     name_proc )
!
        if ( present(theta) ) then
            call merror( name_proc//svd_error2 )
        end if
!
    end if
!
    do_gen_p = true
!
    if ( present(p) ) then
!
        call assert( logical( int(size(p,1),i4b)==n, lgl ),   &
                     logical( int(size(p,2),i4b)==n, lgl ),   &
                     name_proc )
!
        do_comp_p = true
!
        if ( present( gen_p ) ) then
!
            do_gen_p = gen_p
!
        end if
!
    else
!
        do_comp_p = false
!
    end if
!
    do_reortho = false
!
    if ( present( reortho ) ) then
!
        do_reortho = reortho
!
    end if
!
    if ( present(ls) .or. present(theta) ) then
!
        use_dqds = false
!
    else
!
        use_dqds = true
!
    end if
!
    if ( present(dqds) ) then
!
        use_dqds = dqds
!
    end if
!
    if ( present(d) ) then
        call assert( logical( n==size(d), lgl ), name_proc )
    end if
!
    if ( present(e) ) then
        call assert( logical( n==size(e), lgl ), name_proc )
    end if
!
    do_gen_o = true
!
    if ( present( tauo ) ) then
!
        call assert( logical( n==size(tauo), lgl ), name_proc )
!
        do_gen_o = false
!
    end if
!
    if ( present(mul_size) ) then
        inc = merge( mul_size, 32_i4b, 1_i4b<=mul_size .and. mul_size<=n )
    else
        inc = 32_i4b
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
!
        nsing = 0_i4b
!
        return
!
    end if
!
!   COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
    do i = inc, l, inc
        i1 = i + j
        i2 = min( i, n )
        tmp = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, tmp )
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       mat IS THE ZERO MATRIX, GENERATE O, Q AND P.
!
        do i = 1_i4b, n
            mat(i,i) = one
        end do
!
        rmat(:n,:n) = zero
!
        do i = 1_i4b, n
            rmat(i,i) = one
        end do
!
        nsing = n
        s(:n) = zero
!
        if ( present(d) )    d(:n) = zero
        if ( present(e) )    e(:n) = zero
        if ( present(tauo) ) tauo(:n) = zero
!
        if ( do_comp_p ) then
!
            do i = 1_i4b, n
                p(1_i4b:n,i) = zero
                p(i,i)       = one
            end do
!
            if ( n/=1_i4b .and. .not.do_gen_p ) then
                p(1_i4b,1_i4b) = -one
            end if
!
        end if
!
        return
!
    else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   FIRST REDUCE mat USING THE QR DECOMPOSITION.
!   SECOND REDUCE R TO UPPER BIDIAGONAL FORM WITH
!   THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!   COMPUTE mat = Q * R .
!
    call qr_cmp( mat(:m,:n), d2(:n), e2(:n) )
!
!   COPY R TO rmat , ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
    do i = 1_i4b, n
        rmat(1_i4b:i-1_i4b,i) = mat(1_i4b:i-1_i4b,i)
        rmat(i,i) = d2(i)
        rmat(i+1_i4b:n,i) = zero
    end do
!
!$OMP END PARALLEL DO
!
!   GENERATE Q IN mat IF REQUIRED.
!
    if ( do_gen_o ) then
!
        call ortho_gen_qr( mat(:m,:n), e2(:n) )
!
    else
!
        tauo(:n) = e2(:n)
!
    end if
!
!   BIDIAGONALIZE R IN rmat WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
    if (  do_comp_p ) then
!
!       P IS REQUIRED, USE bd_cmp2_a.
!
        call bd_cmp2_a( rmat(:n,:n), d2(:n), e2(:n), p(:n,:n), failure=failure_bd, gen_p=do_gen_p, reortho=do_reortho )
!
    else
!
!       P IS NOT REQUIRED, USE bd_cmp2_b.
!
        call bd_cmp2_b( rmat(:n,:n), d2(:n), e2(:n), failure=failure_bd, reortho=do_reortho )
!
    end if
!
!   SAVE INTERMEDIATE QUANTITIES, IF REQUIRED.
!
    if ( present(d) ) then
        d(:n) = d2(:n)
    end if
!
    if ( present(e) ) then
        e(:n) = e2(:n)
    end if
!
    if ( use_dqds ) then
!
!       COMPUTE SINGULAR VALUES BY THE DQDS ALGORITHM.
!
        call bd_lasq1( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling, info=info )
!
        if ( info==2_i4b ) then
!
            call bd_dqds( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling )
!
        end if
!
        if ( .not.failure ) then
!
            s(:n) = d2(:n)
!
        else
!
            s(:n) = -one
!
        end if
!
        nsing = n
!
    else
!
!       COMPUTE THE SINGULAR VALUES BY THE BISECTION METHOD APPLIED TO THE GOLUB-KAHAN
!       FORM OF THE BIDIAGONAL MATRIX.
!
        call bd_singval( d2, e2, nsing, s, failure, sort=sort,    &
                         vector=vector, abstol=abstol, ls=ls,     &
                         theta=theta, scaling=scaling, init=init  )
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled .and. nsing>0_i4b ) then
!
        call lascl( s(:nsing), merge(smlnum,bignum,matnorm<smlnum) , matnorm )
!
    end if
!
!
! END OF SUBROUTINE select_singval_cmp3_b
! _______________________________________
!
    end subroutine select_singval_cmp3_b
!
! =========================================================================================
!
    subroutine select_singval_cmp4_a( mat, nsing, s, failure, sort, mul_size, vector,    &
                                      abstol, ls, theta, d, e, p, gen_p, reortho,        &
                                      scaling, init, dqds, failure_bd )
!
! Purpose
! _______
!
!   SELECT_SINGVAL_CMP4 computes all or some of the greatest singular values of
!   a real m-by-n matrix MAT with m>=n.
!
!   The Singular Value Decomposition (SVD) is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix.  The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!
!   The original matrix MAT is first reduced to upper bidiagonal form BD
!   by an orthogonal transformation:
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal (see the reference (5) below). The Ralha-Barlow
!   one-sided method is used for this purpose (see the references (1) to (3) below).
!
!   The singular values SIGMA of the bidiagonal matrix BD, which are also the singular values of MAT, 
!   are then computed by a bisection algorithm (see the reference (5) below, Sec.8.5 ). The bisection 
!   method is applied (implicitly) to the associated n-by-n symmetric tridiagonal matrix BD' * BD  
!   whose eigenvalues are the squares of the singular values of BD by using the differential stationary
!   form of the qd algorithm of Rutishauser (see the reference (6) below, Sec.3.1 ). Alternatively, at
!   the user option, all singular values can be computed by the differential quotient difference with
!   shifts (dqds) algorithm (see the references (7) and (8)).
!
!   The routine outputs (parts of) SIGMA, Q and optionally P (in packed form) and
!   BD for a given matrix MAT. SIGMA, Q, P and BD may then be used to obtain selected
!   singular vectors of MAT with subroutines BD_INVITER2 or BD_DEFLATE2.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n matrix MAT.
!
!                On exit, MAT is overwritten with the first n columns
!                of Q (stored column-wise), the orthogonal matrix
!                used to reduce MAT to bidiagonal form as returned
!                by subroutine BD_CMP2 in its argument MAT.
!
!                The shape of MAT must verify: 
!                size( MAT, 1 ) >= size( MAT, 2 ) = n .
!
!   NSING        (OUTPUT) integer(i4b)
!                On output, NSING specifies the number of singular values which have
!                been computed. Note that NSING may be greater than the optional argument
!                LS, if multiple singular values at index LS make unique selection
!                impossible.
!
!                If none of the optional arguments LS and THETA are used, NSING is set to
!                size(MAT,2) and all the singular values are computed.
!
!   S            (OUTPUT) real(stnd), dimension(:)
!                On exit, S(1:NSING)  contains the first NSING  singular values of MAT.
!                The other values in S ( S(NSING+1:) ) are flagged  by a quiet NAN.
!
!                The size of S must be equal to size( MAT, 2 ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection or dqds algorithm
!                  converged for all the computed singular values to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the singular values failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic. The sign of the incorrect singular values is set to negative.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MUL_SIZE     (INPUT, OPTIONAL) integer(i4b)
!                Internal parameter. MUL_SIZE must verify: 1 <= MUL_SIZE <= n,
!                otherwise a default value is used. For better performance, 
!                at the expense of more workspace, a large value can be used.
!
!                The default is min( 32, n ).
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to TRUE, a vectorized version of the bisection algorithm
!                is used to compute the singular values SIGMA of the bidiagonal matrix BD.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is VECTOR = false.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the singular values. A singular value
!                (or cluster) is considered to be located in the bisection algorithm if it
!                been determined to lie in an interval whose width is ABSTOL or less.
!
!                Singular values will be computed most accurately when ABSTOL is set to the square root of
!                the underflow threshold (e.g., sqrt(safmin)=sqrt(LAMCH('S')), not zero. 
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | BD' * BD | will be used,
!                where | BD' * BD | means the 1-norm of the tridiagonal matrix BD' * BD ( BD' means the transpose of BD) 
!                and ULP is the machine precision (distance from 1 to the next larger floating point number).
!
!                This argument has no effect if DQDS is equal to true.
!
!   LS           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LS specifies the number of singular values which must be
!                computed by the subroutine. On output, NSING may be different than LS 
!                if multiple singular values at index LS make unique selection
!                impossible.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LS must be greater than 0 and less or equal to size( MAT, 2 ) = n .
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is LS = size( MAT, 2 ) = n .
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry, THETA specifies that the singular values which are greater or equal to
!                THETA must be computed. If none of the singular values are greater or equal to
!                THETA, NSING is set to zero and S(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is THETA = 0.
!
!   D            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The diagonal elements of the intermediate bidiagonal matrix BD
!
!                The size of D must be equal to size( MAT, 2 ) = n .
!
!   E            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The off-diagonal elements of the intermediate upper bidiagonal matrix BD:
!
!                            E(i) = B(i-1,i) for i = 2,3,...,n;
!
!                E(1) is arbitrary.
!
!                The size of E must be equal to size( MAT, 2 ) = n .
!
!   P            (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!                On exit, P is overwritten with the n-by-n orthogonal matrix P
!                (stored column-wise or in packed form), the orthogonal matrix
!                used to reduce MAT to bidiagonal form as returned
!                by subroutine BD_CMP2 in its argument P.
!
!                The shape of P must verify: size( P, 1 ) = size( P, 2 ) = n .
!
!   GEN_P        (INPUT, OPTIONAL) logical(lgl)
!                On entry, this optional argument has an effect only if the optional argument
!                P is also used.
!
!                In this case, if the optional argument GEN_P is used and is set to true, the
!                orthogonal matrix P used to reduce MAT to bidiagonal form is generated on output
!                of the subroutine in its argument P..
!
!                If GEN_P is set to false, the orthogonal matrix P is stored in factored
!                form as products of elementary reflectors in the lower triangle of the array P.
!
!                See the description of BD_CMP2 subroutine for more details.
!
!                The default is true.
!
!   REORTHO      (INPUT, OPTIONAL) logical(lgl)
!                If the optional argument REORTHO is set to true, reorthogonalization of the
!                matrix Q, used to reduce MAT to bidiagonal form, is performed (when needed).
!                If this optional logical argument is set to false, reorthogonalization of Q is never
!                performed and the columns of the  matrix Q can be far from orthogonal to each other
!                for input matrices MAT with a large condition number.
!
!                See the description of BD_CMP2 subroutine for more details.
!
!                The default is REORTHO = false.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix BD is scaled
!                before computing the singular values.
!
!                The default is to scale the bidiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps are
!                computed from estimates of the eigenvalues of the associated BD' * BD tridiagonal
!                matrix obtained from the Pal-Walker-Kahan algorithm.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!   DQDS         (INPUT, OPTIONAL) logical(lgl)
!                DQDS determines how the singular values are computed.
!
!                If DQDS is set to true, all singular values are computed with the dqds algorithm
!                The dqds algorithm is much faster than bisection when all singular values are
!                requested in many cases and delivers about the same accuracy than bisection.
!
!                If DQDS is set to false, the requested singular values are computed by bisection.
!
!                The default is to use the dqds algorithm if all the singular values are requested
!                and to use bisection if selected singular values are requested (e.g., if the optional
!                arguments LS or THETA are present).
!
!   FAILURE_BD   (OUTPUT, OPTIONAL) logical(lgl)
!                On exit:
!
!                - FAILURE_BD = false :  indicates that maximum accuracy was obtained in the
!                  Ralha-Barlow one-sided bidiagonalization of MAT.
!                - FAILURE_BD = true  :  indicates that MAT is nearly singular and some
!                  loss of orthogonality can be expected in the Ralha-Barlow
!                  bidiagonalization algorithm.
!
!                See description of the BD_CMP2 subroutine for further details.
!
!
! Further Details
! _______________
!
!   The matrices Q, P and BD are computed with the help of the Ralha-Barlow one-sided method.
!   Q is computed by a recurrence relationship and the first n columns of Q are stored in the
!   argument MAT on exit. P is computed as a product of n-1 elementary reflectors
!   (e.g. Householder transformations):
!
!       P = G(1) * G(2) * . . . * G(n-1)
!
!   Each G(i) has the form:
!
!       G(i) = I + taup * v * v'
!
!   where taup is a real scalar, and v is a real vector. IF GEN_P is used and set to false,
!   the n-1 G(i) elementary reflectors are stored in the lower triangle of the array P.
!
!   For the G(i) reflector, taup is stored in P(i+1,1) and v is stored in P(i+1:n,i+1).
!   In addition, P(1,1) is set to -1 if GEN_P=false and is equal to 1 if GEN_P=true.
!
!   In other words, the value of P(1,1) indicates if the orthogonal matrix P is stored
!   in factored form or not. Note that if n is equal to 1, elementary reflectors are not needed
!   and consequently P(1,1) is set to 1, independently of the value of GEN_P.
!
!   This is the blocked version of the Ralha-Barlow one-sided algorithm. See the references (1),
!   (2) and (3) for further details. Furthermore, the algorithm is parallelized if OPENMP is used.
!
!   Since Q is computed by a recurrence relationship, a loss of orthogonality of Q can be
!   observed when the rectangular matrix MAT is singular or nearly singular or has a large
!   condition number, see the reference (2) for details.
!
!   To correct partly this deficiency, partial reorthogonalization can be performed to ensure
!   orthogonality at the expense of speed of computation. The reorthogonalization uses the Gram-Schmidt
!   method described in the reference (4) and is activated if the optional logical argument REORTHO is
!   set to true.
!
!   The reference (2) also explains how to handle the case of an exactly singular
!   matrix MAT (a very rare event). However, in this subroutine, the partial reorthogonalization
!   described in the reference (4), which is used when REORTHO is set true, corrects automatically
!   this problem.
!
!   Now, let SIGMA(i), i=1,...,n, be the singular values of the intermediate bidiagonal
!   matrix BD in decreasing order of magnitude.
!
!   If DQDS is equal to false, the subroutine computes the LS largest singular
!   values (or the singular values which are greater or equal to THETA) of BD by a bisection
!   method (see the reference (5) below, Sec.8.5 ). The bisection method is applied (implicitly) to
!   the associated n-by-n symmetric tridiagonal matrix BD' * BD  whose eigenvalues are the squares
!   of the singular values of BD by using the differential stationary form of the qd algorithm
!   of Rutishauser (see the reference (6) below, Sec.3.1 ).
!
!   On the other hand, if DQDS is equal to true, all the singular values of BD are computed by the dqds
!   algorithm (see the references (7) and (8).
!
!   If bisection is used, SELECT_SINGVAL_CMP4 subroutine is less accurate, but faster, than SELECT_SINGVAL_CMP3
!   subroutine since SELECT_SINGVAL_CMP3 works on the 2.n-by-2.n symmetric tridiagonal GOLUB-KAHAN form of BD, while
!   SELECT_SINGVAL_CMP4 works implicitly on the associated n-by-n symmetric tridiagonal matrix BD' * BD
!   whose eigenvalues are the squares of the singular values of BD.
!
!   For further details, see:
!
!   (1) Ralha, R.M.S., 2003:
!           One-sided reduction to bidiagonal form.
!           Linear Algebra Appl., No 358, pp. 219-238.
!
!   (2) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!           A new stable bidiagonal reduction algorithm.
!           Linear Algebra Appl., No 397, pp. 35-84.
!
!   (3) Bosner, N., and Barlow, J.L., 2007:
!           Block and Parallel versions of one-sided bidiagonalization.
!           SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!   (4) Stewart, G.W., 2007:
!           Block Gram-Schmidt Orthogonalization.
!           Report TR-4823, Department of Computer Science, College Park, University of Maryland.
!
!   (5) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (6) Fernando, K.V., 1998:
!           Accurately counting singular values of bidiagonal matrices and
!           eigenvalues of skew-symmetric tridiagonal matrices.
!           SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!   (7) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (8) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error2, svd_error5
    use Utilities,       only : assert, assert_eq, merror, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout),  dimension(:,:)           :: mat
    real(stnd),  intent(out),    dimension(:)             :: s
    real(stnd),  intent(in),                     optional :: abstol, theta
    real(stnd),  intent(out),    dimension(:),   optional :: d, e
    real(stnd),  intent(out),    dimension(:,:), optional :: p
!
    integer(i4b), intent(out)          :: nsing
    integer(i4b), intent(in), optional :: mul_size, ls
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: gen_p, reortho, vector, scaling, init, dqds
    logical(lgl), intent(out), optional :: failure_bd
!
    character , intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: matnorm, tmp
    real(stnd), dimension(size(s)) :: d2, e2
!
    integer(i4b) :: i, j, l, m, n, inc, i1, i2, info
!
    logical(lgl) :: mat_is_scaled, use_dqds, do_comp_p, do_gen_p, do_reortho
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_singval_cmp4'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
    if ( present( failure_bd ) ) then
!
        failure_bd = false
!
    end if
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b),  &
                   int(size(s),i4b),      &
                   name_proc  )
!
    if ( m<n ) then
!
        call merror( name_proc//svd_error5 )
!
    end if
!
!   TEST OPTIONAL ARGUMENTS.
!
    if ( present(ls) ) then
!
        call assert( logical( ls<=n, lgl ),     &
                     logical( 1_i4b<=ls, lgl ), &
                     name_proc )
!
        if ( present(theta) ) then
            call merror( name_proc//svd_error2 )
        end if
!
    end if
!
    do_gen_p = true
!
    if ( present(p) ) then
!
        call assert( logical( int(size(p,1),i4b)==n, lgl ),   &
                     logical( int(size(p,2),i4b)==n, lgl ),   &
                     name_proc )
!
        do_comp_p = true
!
        if ( present( gen_p ) ) then
!
            do_gen_p = gen_p
!
        end if
!
    else
!
        do_comp_p = false
!
    end if
!
    do_reortho = false
!
    if ( present( reortho ) ) then
!
        do_reortho = reortho
!
    end if
!
    if ( present(ls) .or. present(theta) ) then
!
        use_dqds = false
!
    else
!
        use_dqds = true
!
    end if
!
    if ( present(dqds) ) then
!
        use_dqds = dqds
!
    end if
!
    if ( present(d) ) then
        call assert( logical( n==size(d), lgl ), name_proc )
    end if
!
    if ( present(e) ) then
        call assert( logical( n==size(e), lgl ), name_proc )
    end if
!
    if ( present(mul_size) ) then
        inc = merge( mul_size, 32_i4b, 1_i4b<=mul_size .and. mul_size<=n )
    else
        inc = 32_i4b
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
!
        nsing = 0_i4b
!
        return
!
    end if
!
!   COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, n )
        tmp = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       mat IS THE ZERO MATRIX, GENERATE Q AND P.
!
        nsing = n
        s(:n) = zero
!
        do i = 1_i4b, n
            mat(i,i) = one
        end do
!
        if ( present(d) )    d(:n) = zero
        if ( present(e) )    e(:n) = zero
!
        if ( do_comp_p ) then
!
            do i = 1_i4b, n
                p(1_i4b:n,i) = zero
                p(i,i)       = one
            end do
!
            if ( n/=1_i4b .and. .not.do_gen_p ) then
                p(1_i4b,1_i4b) = -one
            end if
!
        end if
!
        return
!
   else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   mat HAS AT LEAST AS MANY ROWS AS COLUMNS.
!   FIRST BIDIAGONALIZE mat WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
    if (  do_comp_p ) then
!
!       P IS REQUIRED, USE bd_cmp2_a.
!
        call bd_cmp2_a( mat(:m,:n), d2(:n), e2(:n), p(:n,:n), failure=failure_bd, gen_p=do_gen_p, reortho=do_reortho )
!
    else
!
!       P IS NOT REQUIRED, USE bd_cmp2_b.
!
        call bd_cmp2_b( mat(:m,:n), d2(:n), e2(:n), failure=failure_bd, reortho=do_reortho )
!
    end if
!
!   SAVE INTERMEDIATE QUANTITIES, IF REQUIRED.
!
    if ( present(d) ) then
        d(:n) = d2(:n)
    end if
!
    if ( present(e) ) then
        e(:n) = e2(:n)
    end if
!
    if ( use_dqds ) then
!
!       COMPUTE SINGULAR VALUES BY THE DQDS ALGORITHM.
!
        call bd_lasq1( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling, info=info )
!
        if ( info==2_i4b ) then
!
            call bd_dqds( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling )
!
        end if
!
        if ( .not.failure ) then
!
            s(:n) = d2(:n)
!
        else
!
            s(:n) = -one
!
        end if
!
        nsing = n
!
    else
!
!       COMPUTE THE SINGULAR VALUES BY THE BISECTION METHOD APPLIED (IMPLICITLY)
!       TO THE PRODUCT OF THE TRANSPOSE OF THE BIDIAGONAL MATRIX WITH ITSELF.
!
        call bd_singval2( d2, e2, nsing, s, failure, sort=sort,               &
                          vector=vector, abstol=abstol, ls=ls, theta=theta,   &
                          scaling=scaling, init=init                          )
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled .and. nsing>0_i4b ) then
!
        call lascl( s(:nsing), merge(smlnum,bignum,matnorm<smlnum) , matnorm )
!
    end if
!
!
! END OF SUBROUTINE select_singval_cmp4_a
! _______________________________________
!
    end subroutine select_singval_cmp4_a
!
! =========================================================================================
!
    subroutine select_singval_cmp4_b( mat, rmat, nsing, s, failure, sort, mul_size, vector,  &
                                      abstol, ls, theta, d, e, tauo, p, gen_p, reortho,      &
                                      scaling, init, dqds, failure_bd )
!
! Purpose
! _______
!
!   SELECT_SINGVAL_CMP4 computes all or some of the greatest singular values of
!   a real m-by-n matrix MAT with m>=n.
!
!   The Singular Value Decomposition (SVD) is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix.  The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!
!   The original matrix MAT is first reduced to upper bidiagonal form
!   by a two-step algorithm :
!
!   A QR factorization of the real m-by-n matrix MAT is first computed
!
!                        MAT = O * R
!
!   where O is orthogonal and R is upper triangular.
!
!   In a second step, the n-by-n upper triangular matrix R is reduced to upper
!   bidiagonal form BD by an orthogonal transformation :
!
!                       Q' * R * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix (see the reference (5) below).
!   The Ralha-Barlow one-sided method is used in this second step (see the references (1) to (3) below).
!
!   SELECT_SINGVAL_CMP4 computes O, BD, Q and P.
!
!   The singular values SIGMA of the bidiagonal matrix BD, which are also the singular values of MAT, 
!   are then computed by a bisection algorithm (see the reference (5) below, Sec.8.5 ). The bisection 
!   method is applied (implicitly) to the associated n-by-n symmetric tridiagonal matrix BD' * BD  
!   whose eigenvalues are the squares of the singular values of BD by using the differential stationary
!   form of the qd algorithm of Rutishauser (see the reference (6) below, Sec.3.1 ). Alternatively, at
!   the user option, all singular values can be computed by the differential quotient difference with
!   shifts (dqds) algorithm (see the references (7) and (8)).
!
!   The routine outputs (parts of) SIGMA, and optionally O, Q and P and BD for a given matrix MAT. 
!   The matrix O is stored in factored form in the argument MAT if the optional argument TAUO is present
!   or explicitly computed if this argument is absent. The first n columns of Q are stored in the argument
!   RMAT. Finally, P is stored in the optional argument P. P is stored in factored form or explicitly generated
!   depending on the value of the optional logical argument GEN_P.
!
!   SIGMA, O, Q, P and BD may then be used to obtain selected singular vectors of 
!   MAT with subroutines BD_INVITER2 or BD_DEFLATE2.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n matrix MAT.
!
!                On exit, the elements on and below the diagonal,
!                with the array TAUO, represent the orthogonal
!                matrix O of the QR factorization of MAT,
!                as a product of elementary reflectors,
!                if the argument TAUO is present. Otherwise,
!                the argument MAT contains the first n columns
!                of O (stored column-wise) on output.
!
!                See Further Details.
!
!                The shape of MAT must verify: size( MAT, 1 ) >= size( MAT, 2 ) = n .
!
!   RMAT         (OUTPUT) real(stnd), dimension(:,:)
!
!                On exit, RMAT contains the first n columns
!                of Q (stored column-wise), the orthogonal matrix
!                used to reduce R to bidiagonal form as returned
!                by subroutine BD_CMP2 in its argument MAT.
!
!                See Further Details.
!
!                The shape of RMAT must verify:
!                size( RMAT, 1 ) = size( RMAT, 2 ) = size( MAT, 2 ) = n .
!
!   NSING        (OUTPUT) integer(i4b)
!                On output, NSING specifies the number of singular values which have
!                been computed. Note that NSING may be greater than the optional argument
!                LS, if multiple singular values at index LS make unique selection
!                impossible.
!
!                If none of the optional arguments LS and THETA are used, NSING is set to
!                size(MAT,2) and all the singular values are computed.
!
!   S            (OUTPUT) real(stnd), dimension(:)
!                On exit, S(1:NSING)  contains the first NSING  singular values of MAT.
!                The other values in S ( S(NSING+1:) ) are flagged  by a quiet NAN.
!
!                The size of S must be equal to size( MAT, 2 ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection or dqds algorithm
!                  converged for all the computed singular values to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the singular values failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic. The sign of the incorrect singular values is set to negative.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MUL_SIZE     (INPUT, OPTIONAL) integer(i4b)
!                Internal parameter. MUL_SIZE must verify: 1 <= MUL_SIZE <= n,
!                otherwise a default value is used. For better performance, 
!                at the expense of more workspace, a large value can be used.
!
!                The default is min( 32, n ).
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to TRUE, a vectorized version of the bisection algorithm
!                is used to compute the singular values SIGMA of the bidiagonal matrix BD.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is VECTOR = false.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the singular values. A singular value
!                (or cluster) is considered to be located in the bisection algorithm if it
!                been determined to lie in an interval whose width is ABSTOL or less.
!
!                Singular values will be computed most accurately when ABSTOL is set to the square root of
!                the underflow threshold (e.g., sqrt(safmin)=sqrt(LAMCH('S')), not zero. 
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | BD' * BD | will be used,
!                where | BD' * BD | means the 1-norm of the tridiagonal matrix BD' * BD ( BD' means the transpose of BD) 
!                and ULP is the machine precision (distance from 1 to the next larger floating point number).
!
!                This argument has no effect if DQDS is equal to true.
!
!   LS           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LS specifies the number of singular values which must be
!                computed by the subroutine. On output, NSING may be different than LS 
!                if multiple singular values at index LS make unique selection
!                impossible.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LS must be greater than 0 and less or equal to size( MAT, 2 ) = n .
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is LS = size( MAT, 2 ) = n .
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry, THETA specifies that the singular values which are greater or equal to
!                THETA must be computed. If none of the singular values are greater or equal to
!                THETA, NSING is set to zero and S(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is THETA = 0.
!
!   D            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The diagonal elements of the intermediate bidiagonal matrix BD
!
!                The size of D must be equal to size( MAT, 2 ) = n .
!
!   E            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The off-diagonal elements of the intermediate upper bidiagonal matrix BD:
!
!                            E(i) = B(i-1,i) for i = 2,3,...,n;
!
!                E(1) is arbitrary.
!
!                The size of E must be equal to size( MAT, 2 ) = n .
!
!   TAUO         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix O of the QR decomposition of MAT.
!
!                If the optional argument TAUO is present, the orthogonal matrix O is stored
!                in factored form, as a product of elementary reflectors, in the argument MAT on exit.
!
!                If the optional argument TAUO is absent, the first n columns of the orthogonal matrix O
!                are explicitly generated and stored in the argument MAT on exit.
!
!                See description of the argument MAT above and Further Details below.
!
!                The size of TAUO must be equal to size( MAT, 2 ) = n .
!
!   P            (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!                On exit, P is overwritten with the n-by-n orthogonal matrix P
!                (stored column-wise or in packed form), the orthogonal matrix
!                used to reduce MAT to bidiagonal form as returned
!                by subroutine BD_CMP2 in its argument P.
!
!                The shape of P must verify: size( P, 1 ) = size( P, 2 ) = n .
!
!   GEN_P        (INPUT, OPTIONAL) logical(lgl)
!                On entry, this optional argument has an effect only if the optional argument
!                P is also used.
!
!                In this case, if the optional argument GEN_P is used and is set to true, the
!                orthogonal matrix P used to reduce MAT to bidiagonal form is generated on output
!                of the subroutine in its argument P..
!
!                If GEN_P is set to false, the orthogonal matrix P is stored in factored
!                form as products of elementary reflectors in the lower triangle of the array P.
!
!                See the description of BD_CMP2 subroutine for more details.
!
!                The default is true.
!
!   REORTHO      (INPUT, OPTIONAL) logical(lgl)
!                If the optional argument REORTHO is set to true, reorthogonalization of the
!                matrix Q, used to reduce MAT to bidiagonal form, is performed (when needed).
!                If this optional logical argument is set to false, reorthogonalization of Q is never
!                performed and the columns of the  matrix Q can be far from orthogonal to each other
!                for input matrices MAT with a large condition number.
!
!                See the description of BD_CMP2 subroutine for more details.
!
!                The default is REORTHO = false.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix BD is scaled
!                before computing the singular values.
!
!                The default is to scale the bidiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps are
!                computed from estimates of the eigenvalues of the associated BD' * BD tridiagonal
!                matrix obtained from the Pal-Walker-Kahan algorithm.
!
!                This argument has no effect if DQDS is equal to true.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!   DQDS         (INPUT, OPTIONAL) logical(lgl)
!                DQDS determines how the singular values are computed.
!
!                If DQDS is set to true, all singular values are computed with the dqds algorithm
!                The dqds algorithm is much faster than bisection when all singular values are
!                requested in many cases and delivers about the same accuracy than bisection.
!
!                If DQDS is set to false, the requested singular values are computed by bisection.
!
!                The default is to use the dqds algorithm if all the singular values are requested
!                and to use bisection if selected singular values are requested (e.g., if the optional
!                arguments LS or THETA are present).
!
!   FAILURE_BD   (OUTPUT, OPTIONAL) logical(lgl)
!                On exit:
!
!                - FAILURE_BD = false :  indicates that maximum accuracy was obtained in the
!                  Ralha-Barlow one-sided bidiagonalization of MAT.
!                - FAILURE_BD = true  :  indicates that MAT is nearly singular and some
!                  loss of orthogonality can be expected in the Ralha-Barlow
!                  bidiagonalization algorithm.
!
!                See description of the BD_CMP2 subroutine for further details.
!
!
! Further Details
! _______________
!
!   the matrix O of the QR factorization of MAT is represented as a product
!   of elementary reflectors
!
!            O = W(1) * W(2) * ... * W(n)
!
!   Each W(i) has the form
!
!            W(i) = I + tauo * ( v * v' ) ,
!                      
!   where tauo is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in MAT(i:m,i) and tauo in TAUO(i). If the optional argument
!   TAUO is absent, the first n columns of O are generated and stored in the argument MAT.
!
!   The matrix O is stored in factored form if the optional argument TAUO is present
!   or explicitly computed if this argument is absent.
!
!   A blocked algorithm is used for computing the QR factorization of MAT. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   After, the initial QR factorization of MAT, the upper triangular
!   matrix R is reduced to upper bidiagonal form BD:
!
!                       Q' * R * P = BD
!
!   The matrices Q, P and BD are computed with the help of the Ralha-Barlow one-sided method.
!   Q is computed by a recurrence relationship and the first n columns of Q are stored in the
!   argument RMAT on exit. P is computed as a product of n-1 elementary reflectors
!   (e.g. Householder transformations):
!
!       P = G(1) * G(2) * . . . * G(n-1)
!
!   Each G(i) has the form:
!
!       G(i) = I + taup * v * v'
!
!   where taup is a real scalar, and v is a real vector. IF GEN_P is used and set to false,
!   the n-1 G(i) elementary reflectors are stored in the lower triangle of the array P.
!
!   For the G(i) reflector, taup is stored in P(i+1,1) and v is stored in P(i+1:n,i+1).
!   In addition, P(1,1) is set to -1 if GEN_P=false and is equal to 1 if GEN_P=true.
!
!   In other words, the value of P(1,1) indicates if the orthogonal matrix P is stored
!   in factored form or not. Note that if n is equal to 1, elementary reflectors are not needed
!   and consequently P(1,1) is set to 1, independently of the value of GEN_P.
!
!   This is the blocked version of the Ralha-Barlow one-sided algorithm. See the references (1),
!   (2) and (3) for further details. Furthermore, the algorithm is parallelized if OPENMP is used.
!
!   Since Q is computed by a recurrence relationship, a loss of orthogonality of Q can be
!   observed when the rectangular matrix MAT is singular or nearly singular or has a large
!   condition number, see the reference (2) for details.
!
!   To correct partly this deficiency, partial reorthogonalization can be performed to ensure
!   orthogonality at the expense of speed of computation. The reorthogonalization uses the Gram-Schmidt
!   method described in the reference (4) and is activated if the optional logical argument REORTHO is
!   set to true.
!
!   The reference (2) also explains how to handle the case of an exactly singular
!   matrix MAT (a very rare event). However, in this subroutine, the partial reorthogonalization
!   described in the reference (4), which is used when REORTHO is set true, corrects automatically
!   this problem.
!
!   Now, let SIGMA(i), i=1,...,n, be the singular values of the intermediate bidiagonal
!   matrix BD in decreasing order of magnitude.
!
!   If DQDS is equal to false, the subroutine computes the LS largest singular
!   values (or the singular values which are greater or equal to THETA) of BD by a bisection
!   method (see the reference (5) below, Sec.8.5 ). The bisection method is applied (implicitly) to
!   the associated n-by-n symmetric tridiagonal matrix BD' * BD  whose eigenvalues are the squares
!   of the singular values of BD by using the differential stationary form of the qd algorithm
!   of Rutishauser (see the reference (6) below, Sec.3.1 ).
!
!   On the other hand, if DQDS is equal to true, all the singular values of BD are computed by the dqds
!   algorithm (see the references (7) and (8).
!
!   If bisection is used, SELECT_SINGVAL_CMP4 subroutine is less accurate, but faster, than SELECT_SINGVAL_CMP3
!   subroutine since SELECT_SINGVAL_CMP3 works on the 2n-by-2n symmetric tridiagonal GOLUB-KAHAN form of BD, while
!   SELECT_SINGVAL_CMP4 works implicitly on the associated n-by-n symmetric tridiagonal matrix BD' * BD
!   whose eigenvalues are the squares of the singular values of BD.
!
!   For further details, see:
!
!   (1) Ralha, R.M.S., 2003:
!           One-sided reduction to bidiagonal form.
!           Linear Algebra Appl., No 358, pp. 219-238.
!
!   (2) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!           A new stable bidiagonal reduction algorithm.
!           Linear Algebra Appl., No 397, pp. 35-84.
!
!   (3) Bosner, N., and Barlow, J.L., 2007:
!           Block and Parallel versions of one-sided bidiagonalization.
!           SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!   (4) Stewart, G.W., 2007:
!           Block Gram-Schmidt Orthogonalization.
!           Report TR-4823, Department of Computer Science, College Park, University of Maryland.
!
!   (5) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (6) Fernando, K.V., 1998:
!           Accurately counting singular values of bidiagonal matrices and
!           eigenvalues of skew-symmetric tridiagonal matrices.
!           SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!   (7) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (8) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error2, svd_error5
    use Utilities,       only : assert, assert_eq, merror, lascl
    use QR_Procedures,   only : qr_cmp, ortho_gen_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout),  dimension(:,:)           :: mat
    real(stnd),  intent(out),    dimension(:,:)           :: rmat
    real(stnd),  intent(out),    dimension(:)             :: s
    real(stnd),  intent(in),                     optional :: abstol, theta
    real(stnd),  intent(out),    dimension(:),   optional :: d, e, tauo
    real(stnd),  intent(out),    dimension(:,:), optional :: p
!
    integer(i4b), intent(out)          :: nsing
    integer(i4b), intent(in), optional :: mul_size, ls
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: gen_p, reortho, vector, scaling, init, dqds
    logical(lgl), intent(out), optional :: failure_bd
!
    character , intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: matnorm, tmp
    real(stnd), dimension(size(s)) :: d2, e2
!
    integer(i4b) :: i, j, l, m, n, inc, i1, i2, info
!
    logical(lgl) :: mat_is_scaled, use_dqds, do_comp_p, do_gen_p, do_gen_o, do_reortho
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_singval_cmp4'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
    if ( present( failure_bd ) ) then
!
        failure_bd = false
!
    end if
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b),   &
                   int(size(rmat,1),i4b) , &
                   int(size(rmat,2),i4b) , &
                   int(size(s),i4b),       &
                   name_proc  )
!
    if ( m<n ) then
!
        call merror( name_proc//svd_error5 )
!
    end if
!
!   TEST OPTIONAL ARGUMENTS.
!
    if ( present(ls) ) then
!
        call assert( logical( ls<=n, lgl ),     &
                     logical( 1_i4b<=ls, lgl ), &
                     name_proc )
!
        if ( present(theta) ) then
            call merror( name_proc//svd_error2 )
        end if
!
    end if
!
    do_gen_p = true
!
    if ( present(p) ) then
!
        call assert( logical( int(size(p,1),i4b)==n, lgl ),   &
                     logical( int(size(p,2),i4b)==n, lgl ),   &
                     name_proc )
!
        do_comp_p = true
!
        if ( present( gen_p ) ) then
!
            do_gen_p = gen_p
!
        end if
!
    else
!
        do_comp_p = false
!
    end if
!
    do_reortho = false
!
    if ( present( reortho ) ) then
!
        do_reortho = reortho
!
    end if
!
    if ( present(ls) .or. present(theta) ) then
!
        use_dqds = false
!
    else
!
        use_dqds = true
!
    end if
!
    if ( present(dqds) ) then
!
        use_dqds = dqds
!
    end if
!
    if ( present(d) ) then
        call assert( logical( n==size(d), lgl ), name_proc )
    end if
!
    if ( present(e) ) then
        call assert( logical( n==size(e), lgl ), name_proc )
    end if
!
    do_gen_o = true
!
    if ( present( tauo ) ) then
!
        call assert( logical( n==size(tauo), lgl ), name_proc )
!
        do_gen_o = false
!
    end if
!
    if ( present(mul_size) ) then
        inc = merge( mul_size, 32_i4b, 1_i4b<=mul_size .and. mul_size<=n )
    else
        inc = 32_i4b
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
!
        nsing = 0_i4b
!
        return
!
    end if
!
!   COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, n )
        tmp = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       mat IS THE ZERO MATRIX, GENERATE O, Q AND P.
!
        do i = 1_i4b, n
            mat(i,i) = one
        end do
!
        rmat(:n,:n) = zero
!
        do i = 1_i4b, n
            rmat(i,i) = one
        end do
!
        nsing = n
        s(:n) = zero
!
        if ( present(d) )    d(:n) = zero
        if ( present(e) )    e(:n) = zero
        if ( present(tauo) ) tauo(:n) = zero
!
        if ( do_comp_p ) then
!
            do i = 1_i4b, n
                p(1_i4b:n,i) = zero
                p(i,i)       = one
            end do
!
            if ( n/=1_i4b .and. .not.do_gen_p ) then
                p(1_i4b,1_i4b) = -one
            end if
!
        end if
!
        return
!
   else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   FIRST REDUCE mat USING THE QR DECOMPOSITION.
!   SECOND REDUCE R TO UPPER BIDIAGONAL FORM WITH
!   THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!   COMPUTE mat = Q * R .
!
    call qr_cmp( mat(:m,:n), d2(:n), e2(:n) )
!
!   COPY R TO rmat , ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
    do i = 1_i4b, n
        rmat(1_i4b:i-1_i4b,i) = mat(1_i4b:i-1_i4b,i)
        rmat(i,i) = d2(i)
        rmat(i+1_i4b:n,i) = zero
    end do
!
!$OMP END PARALLEL DO
!
!   GENERATE Q IN mat IF REQUIRED.
!
    if ( do_gen_o ) then
!
        call ortho_gen_qr( mat(:m,:n), e2(:n) )
!
    else
!
        tauo(:n) = e2(:n)
!
    end if
!
!   BIDIAGONALIZE R IN rmat WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
    if (  do_comp_p ) then
!
!       P IS REQUIRED, USE bd_cmp2_a.
!
        call bd_cmp2_a( rmat(:n,:n), d2(:n), e2(:n), p(:n,:n), failure=failure_bd, gen_p=do_gen_p, reortho=do_reortho )
!
    else
!
!       P IS NOT REQUIRED, USE bd_cmp2_b.
!
        call bd_cmp2_b( rmat(:n,:n), d2(:n), e2(:n), failure=failure_bd, reortho=do_reortho )
!
    end if
!
!   SAVE INTERMEDIATE QUANTITIES, IF REQUIRED.
!
    if ( present(d) ) then
        d(:n) = d2(:n)
    end if
!
    if ( present(e) ) then
        e(:n) = e2(:n)
    end if
!
    if ( use_dqds ) then
!
!       COMPUTE SINGULAR VALUES BY THE DQDS ALGORITHM.
!
        call bd_lasq1( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling, info=info )
!
        if ( info==2_i4b ) then
!
            call bd_dqds( d2, e2, failure, maxiter=60_i4b, sort=sort, scaling=scaling )
!
        end if
!
        if ( .not.failure ) then
!
            s(:n) = d2(:n)
!
        else
!
            s(:n) = -one
!
        end if
!
        nsing = n
!
    else
!
!       COMPUTE THE SINGULAR VALUES BY THE BISECTION METHOD APPLIED (IMPLICITLY)
!       TO THE PRODUCT OF THE TRANSPOSE OF THE BIDIAGONAL MATRIX WITH ITSELF.
!
        call bd_singval2( d2, e2, nsing, s, failure, sort=sort,               &
                          vector=vector, abstol=abstol, ls=ls, theta=theta,   &
                          scaling=scaling, init=init                          )
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled .and. nsing>0_i4b ) then
!
        call lascl( s(:nsing), merge(smlnum,bignum,matnorm<smlnum) , matnorm )
!
    end if
!
!
! END OF SUBROUTINE select_singval_cmp4_b
! _______________________________________
!
    end subroutine select_singval_cmp4_b
!
! =========================================================================================
!
    subroutine svd_cmp_a( mat, s, failure, v, sort, mul_size, maxiter,             &
                          max_francis_steps, perfect_shift, bisect, dqds, use_svd2 )
!
! Purpose
! _______
!
!   SVD_CMP computes the Singular Value Decomposition (SVD) of a real
!   m-by-n matrix MAT. The SVD is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix. The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!   The columns of U and V are the left and right
!   singular vectors of MAT.
!
!   SVD_CMP computes only the first min(m,n) columns of U and V (e.g. the left
!   and right singular vectors of MAT in the thin SVD of MAT).
!
!   The routine returns the first min(m,n) singular values and the associated left
!   and right singular vectors.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit, MAT is overwritten with the first min(m,n)
!                       columns of U, the left singular vectors.
!
!   S                   (OUTPUT) real(stnd), dimension(:)
!                       The singular values of MAT.
!
!                       The size of S must verify: size( S ) = min(m,n) .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit
!                       - FAILURE = true  :  indicates that the algorithm did not converge and
!                         that full accuracy was not attained in the bidiagonal
!                         SVD of an intermediate bidiagonal form B of MAT.
!
!   V                   (OUTPUT) real(stnd), dimension(:,:)
!                       On exit, V contains the first min(m,n) columns of V,
!                       the right singular vectors.
!
!                       The shape of V must verify: 
!
!                       - size( V, 1 ) = n,
!                       - size( V, 2 ) = min(m,n).
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the singular values into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd'.
!                       The singular vectors are rearranged accordingly.
!
!   MUL_SIZE            (INPUT, OPTIONAL) integer(i4b)
!                       Internal parameter. MUL_SIZE must verify: 1 <= MUL_SIZE <= max(m,n),
!                       otherwise a default value is used. MUL_SIZE can be increased or decreased
!                       to improve the performance of the algorithm.
!
!                       The default value is 32.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm.
!
!                       The bidiagonal SVD algorithm of an intermediate bidiagonal 
!                       form B of MAT fails to converge if the number of QR sweeps 
!                       exceeds MAXITER * min(m,n). Convergence usually occurs in 
!                       about 2 * min(m,n) QR sweeps.
!
!                       The default is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm.
!
!                       MAX_FRANCIS_STEPS is a strictly positive integer, otherwise 
!                       the default value is used.
!
!                       The default is the minimum of min(m,n) and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   USE_SVD2            (INPUT, OPTIONAL) logical(lgl)
!                       If the optional argument USE_SVD2 is used and is set to true, an alternate
!                       SVD algorithm which used less workspace (but which may be slower) is
!                       automatically used if m is much larger than n or if n is much larger
!                       than m (e.g. if max(m,n)>=1.5 * min(m,n) ).
!                
!
! Further Details
! _______________
!
!   Computing the SVD of a rectangular matrix in subroutine SVD_CMP consists of three steps:
!
!       1) reduction of the rectangular matrix to bidiagonal form via orthogonal
!          transformations (e.g. Householder transformations);
!       2) in place accumulation of the orthogonal transformations used in the reduction
!          to bidiagonal form;
!       3) computation of the SVD of the bidiagonal matrix.
!
!   For further details, on the SVD of a rectangular matrix and the algorithm to
!   compute it, see the references (1) or (2).
!
!   All the three steps of the SVD algorithm (e.g. the reduction to bidiagonal form,
!   accumulation of the Householder transformations used in the reduction to bidiagonal form
!   and computation of the SVD of the bidiagonal matrix) are  parallelized if OPENMP is used.
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!                
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq, lascl
    use QR_Procedures,   only : qr_cmp, apply_q_qr, ortho_gen_qr, lq_cmp, apply_q_lq
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:)   :: s
    real(stnd), intent(out),   dimension(:,:) :: v
!
    integer(i4b), intent(in), optional :: mul_size, maxiter, max_francis_steps
!
    logical(lgl), intent(out)          :: failure
    logical(lgl), intent(in), optional :: perfect_shift, bisect, dqds, use_svd2
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension( size(s) )        :: e, tauq, taup, tau
    real(stnd)                              :: matnorm, tmp
    real(stnd), dimension(:,:), allocatable :: work, qmat
!
    integer(i4b) ::  i, j, k, l, m, n, inc, mnthr, i1, i2
    integer      ::  iok
!
    logical(lgl) :: mat_is_scaled, usesvd2
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='svd_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b), &
                   int(size(v,1),i4b),   &
                   name_proc  )
!
    k = assert_eq( min(m,n),             &
                   int(size(s),i4b),     &
                   int(size(v,2),i4b),   &
                   name_proc  )
!
    if ( present(mul_size) ) then
        i = merge( mul_size, 32_i4b, 1_i4b<=mul_size )
    else
        i = 32_i4b
    end if
!
    inc = min( i, max(m,n) )
!
    usesvd2 = false
    if ( present(use_svd2) ) then
        usesvd2 = use_svd2
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, n )
        tmp = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
        s(:k) = zero
!
        do i = 1_i4b, k
!
            mat(i,i) = one
            v(:n,i) = zero
            v(i,i) = one
!
        end do
!
        return
!
    else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat  TO BIDIAGONAL FORM. IF max(m,n) EXCEEDS mnthr ,
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE
!   m-BY-n MATRIX mat TO A TRIANGULAR FORM.
!    
    mnthr = int( real( k, stnd )*thr_bd, i4b )
!
    if ( m>=mnthr ) then
!
!
!       PATH 1 : m MUCH LARGER THAN n .
!
!
!       FIRST REDUCE mat USING THE QR DECOMPOSITION.
!       SECOND REDUCE R TO UPPER BIDIAGONAL FORM.
!
!       COMPUTE mat = Q * R .
!
        call qr_cmp( mat, s, tau )
!
        if ( usesvd2 ) then
!
!           USE AN ALTERNATE SVD ALGORITHM AUTOMATICALLY (PATH 1b).
!
            iok = -1
!
        else
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM (PATH 1a).
!
            allocate( qmat(m,n), work(n,n), stat = iok )
!
        end if
!
        if ( iok==0 ) then
!
!           PATH 1a : SUFFICIENT WORKSPACE FOR A FAST ALGORITHM.
!
!
!           COPY R TO work , ZEROING OUT BELOW IT.
!           MAKE A COPY OF Q IN qmat.
!
!$OMP PARALLEL DO IF(m*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, n
!
                work(1_i4b:i-1_i4b,i) = mat(1_i4b:i-1_i4b,i)
                work(i,i) = s(i)
                work(i+1_i4b:n,i) = zero
!
                qmat(i:m,i) = mat(i:m,i)
!
            end do
!
!$OMP END PARALLEL DO
!
!           BIDIAGONALIZE R IN work .
!
            call bd_cmp_a( work, s, e, tauq, taup )
!
!           GENERATE RIGHT VECTORS BIDIAGONALIZING R IN v .
!           GENERATE LEFT VECTORS BIDIAGONALIZING R IN work .
!
            call ortho_gen_bd( work, tauq, taup, v )
!
!           PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!           SINGULAR VECTORS OF R IN work AND COMPUTING RIGHT
!           SINGULAR VECTORS OF R IN v .
!
            call bd_svd_a( true, s, e, failure, work, v, sort=sort,              &
                           maxiter=maxiter, max_francis_steps=max_francis_steps, &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!           GENERATE Q IN qmat .
!
            call ortho_gen_qr( qmat, tau )
!
!           MULTIPLY Q IN qmat BY LEFT SINGULAR VECTORS OF R IN
!           work , STORING RESULT IN mat .
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, n, n, one, qmat(1_i4b:m,1_i4b:n), m,        &
                       work(1_i4b:n,1_i4b:n), n, zero, mat(1_i4b:m,1_i4b:n), m  )
#else
            mat(:m,:n) = matmul( qmat(:m,:n), work(:n,:n) )
#endif
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( qmat, work )
!
        else
!
!           PATH 1b : INSUFFICIENT WORKSPACE FOR A FAST ALGORITHM.
!
!
!           COPY R TO v , ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, n
                v(1_i4b:i-1_i4b,i) = mat(1_i4b:i-1_i4b,i)
                v(i,i) = s(i)
                v(i+1_i4b:n,i) = zero
            end do
!
!$OMP END PARALLEL DO
!
!           GENERATE Q IN mat .
!
            call ortho_gen_qr( mat, tau )
!
!           BIDIAGONALIZE R IN v .
!
            call bd_cmp_a( v, s, e, tauq, taup )
!
!           MULTIPLY Q IN mat BY LEFT VECTORS BIDIAGONALIZING R.
!
            call apply_q_qr( v, tauq, mat, left=false, trans=false )
!
!           SHIFT THE VECTORS, IN v, WHICH DEFINE THE ELEMENTARY 
!           REFLECTORS NEEDED TO GENERATE RIGHT VECTORS OF R ONE
!           ROW DOWNWARD AND TRANSPOSE THEM.
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC,1), PRIVATE(i)
!
            do i = 2_i4b, n
                v(i,2_i4b:i) = v(1_i4b:i-1_i4b,i)
            end do
!
!$OMP END PARALLEL DO
!
!           GENERATE RIGHT VECTORS BIDIAGONALIZING R IN v .
!
            v(1_i4b,1_i4b) = one
            v(2_i4b:n,1_i4b) = zero
            v(1_i4b,2_i4b:n) = zero
!
            call ortho_gen_qr( v(2_i4b:n,2_i4b:n), taup(:n-1_i4b) )        
!
!           PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!           SINGULAR VECTORS OF mat IN mat AND COMPUTING RIGHT
!           SINGULAR VECTORS OF mat IN v .
!
            call bd_svd_a( true, s, e, failure, mat, v, sort=sort,               &
                           maxiter=maxiter, max_francis_steps=max_francis_steps, &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
        end if
!
    else if ( n>=mnthr ) then
!
!       PATH 2 : n MUCH LARGER THAN m.
!
!
!       FIRST REDUCE mat USING THE LQ DECOMPOSITION.
!       SECOND REDUCE L TO UPPER BIDIAGONAL FORM.
!
!       COMPUTE mat = L * Q .
!
        call lq_cmp( mat, s, tau )
!
        if ( usesvd2 ) then
!
!           USE AN ALTERNATE SVD ALGORITHM AUTOMATICALLY (PATH 2b).
!
            iok = -1
!
        else
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM (PATH 2a).
!
            allocate( qmat(n,m), work(m,m), stat = iok )
!
        end if
!
        if ( iok==0 ) then
!
!           PATH 2a : SUFFICIENT WORKSPACE FOR A FAST ALGORITHM.
!
!
!           TRANSPOSE THE VECTORS, IN mat, WHICH DEFINE THE ELEMENTARY 
!           REFLECTORS NEEDED TO GENERATE Q AND COPY THEM IN qmat ;
!           RESTORE L IN mat , ZEROING ABOUT ABOVE IT.
!
!$OMP PARALLEL DO IF(n*m>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, n
!
                if ( i<=m ) then
                    qmat(i,1_i4b:i) = mat(1_i4b:i,i)
                    mat(1_i4b:i-1_i4b,i) = zero
                    mat(i,i) = s(i)
                else
                    qmat(i,1_i4b:m) = mat(1_i4b:m,i)
                end if
!
            end do
!
!$OMP END PARALLEL DO
!
!           UPPER BIDIAGONALIZE L IN mat .
!
            call bd_cmp_a( mat(:m,:m), s, e, tauq, taup )
!
!           GENERATE LEFT VECTORS BIDIAGONALIZING L IN mat(:m,:m) .
!           GENERATE RIGHT VECTORS BIDIAGONALIZING L IN work(:m,:m) .
!
            call ortho_gen_bd( mat(:m,:m), tauq, taup, work(:m,:m) )
!
!           PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!           SINGULAR VECTORS OF L IN mat(:m,:m) AND COMPUTING RIGHT
!           SINGULAR VECTORS OF L IN work(:m,:m) .
!
            call bd_svd_a( true, s, e, failure, mat(:m,:m), work(:m,:m),                    &
                           sort=sort, maxiter=maxiter, max_francis_steps=max_francis_steps, &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!           GENERATE Q' IN qmat .
!
            call ortho_gen_qr( qmat, tau )
!
!           MULTIPLY Q' IN qmat by RIGHT SINGULAR VECTORS OF L IN work(:m,:m),
!           STORING RESULT IN v.
!
#ifdef _BLAS
            call gemm( 'N', 'N', n, m, m, one, qmat(1_i4b:n,1_i4b:m), n,        &
                       work(1_i4b:m,1_i4b:m), m, zero, v(1_i4b:n,1_i4b:m), n )
#else
            v(:n,:m) = matmul( qmat(:n,:m), work(:m,:m) )
#endif
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( qmat, work )
!
        else
!
!           PATH 2b : INSUFFICIENT WORKSPACE FOR A FAST ALGORITHM.
!
!
!           TRANSPOSE THE VECTORS, IN mat, WHICH DEFINE THE ELEMENTARY 
!           REFLECTORS NEEDED TO GENERATE Q AND COPY THEM IN v ;
!           RESTORE L IN mat , ZEROING ABOUT ABOVE IT.
!
!$OMP PARALLEL DO IF(n*m>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, n
!
                if ( i<=m ) then
                    v(i,1_i4b:i) = mat(1_i4b:i,i)
                    mat(1_i4b:i-1_i4b,i) = zero
                    mat(i,i) = s(i)
                else
                    v(i,1_i4b:m) = mat(1_i4b:m,i)
                end if
!
            end do
!
!$OMP END PARALLEL DO
!
!           GENERATE Q' IN v .
!
            call ortho_gen_qr( v, tau )
!
!           UPPER BIDIAGONALIZE L IN mat .
!
            call bd_cmp_a( mat(:m,:m), s, e, tauq, taup )
!
!           MULTIPLY RIGHT VECTORS BIDIAGONALIZING L BY Q' IN v.
!
            call apply_q_lq( mat(:m,2_i4b:m), taup(:m-1_i4b), v(:n,2_i4b:m),     &
                             left=false, trans=true )
!
!           GENERATE LEFT VECTORS BIDIAGONALIZING L IN mat(:m,:m) .
!
            call ortho_gen_q_bd( mat(:m,:m), tauq )
!
!           PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!           SINGULAR VECTORS OF mat IN mat(:m,:m) AND COMPUTING RIGHT
!           SINGULAR VECTORS OF mat IN v(:n,:m) .
!
            call bd_svd_a( true, s, e, failure, mat(:m,:m), v(:n,:m), sort=sort,  &
                           maxiter=maxiter, max_francis_steps=max_francis_steps,  &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds  )
!
        end if
!
    else
!
!       PATH 3 : n AT LEAST m, BUT NOT MUCH LARGER OR
!                m AT LEAST n, BUT NOT MUCH LARGER.
!
!       REDUCE mat TO BIDIAGONAL FORM WITHOUT
!       QR OR LQ DECOMPOSITION.
!
!       BIDIAGONALIZE mat .
!
        call bd_cmp_a( mat, s, e, tauq, taup )
!
!       GENERATE RIGHT VECTORS BIDIAGONALIZING mat IN v .
!       GENERATE LEFT VECTORS BIDIAGONALIZING mat IN mat .
!
        call ortho_gen_bd( mat, tauq, taup, v )
!
!       PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!       SINGULAR VECTORS OF mat IN mat AND COMPUTING RIGHT
!       SINGULAR VECTORS OF mat IN v .
!
        call bd_svd_a( merge(true,false,m>=n), s, e, failure, mat(:,:k), v, sort=sort, &
                       maxiter=maxiter, max_francis_steps=max_francis_steps,           &
                       perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
    end if
!
!   SET UNUSED PART OF mat TO ZERO.
!
    mat(:m,k+1_i4b:n) = zero
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled ) then
!
        call lascl( s, merge(smlnum,bignum,matnorm<smlnum), matnorm )
!
    end if
!
!
! END OF SUBROUTINE svd_cmp_a
! ___________________________
!
    end subroutine svd_cmp_a
!
! =========================================================================================
!
    subroutine svd_cmp2_a( mat, s, failure, u_vt, sort, mul_size, maxiter,          &
                           max_francis_steps, perfect_shift, bisect, dqds, use_svd2 )
!
! Purpose
! _______
!
!   SVD_CMP2 computes the Singular Value Decomposition (SVD) of a real
!   m-by-n matrix MAT. The SVD is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix. The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!   The columns of U and V are the left and right
!   singular vectors of MAT.
!
!   SVD_CMP2 computes only the first min(m,n) columns of U and V (e.g. the left
!   and right singular vectors of MAT in the thin SVD of MAT).
!
!   The routine returns the first min(m,n) singular values and the associated left
!   and right singular vectors. The right singular vectors are returned row-wise.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit:
!
!                       - if m>=n, MAT is overwritten with the first min(m,n)
!                         columns of U (the left singular vectors,
!                         stored column-wise);
!
!                       - if m<n, MAT is overwritten with the first min(m,n)
!                         rows of V' (the right singular vectors,
!                         stored row-wise).
!
!   S                   (OUTPUT) real(stnd), dimension(:)
!                       The singular values of MAT.
!
!                       The size of S must verify: size( S ) = min(m,n) .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit
!                       - FAILURE = true  :  indicates that the algorithm did not converge and
!                         that full accuracy was not attained in the bidiagonal
!                         SVD of an intermediate bidiagonal form B of MAT.
!
!   U_VT                (OUTPUT) real(stnd), dimension(:,:)
!                       On exit:
!
!                       - if m>=n,     U_VT contains the n-by-n orthogonal matrix V';
!                       - if m<n,      U_VT contains the m-by-m orthogonal matrix U.
!
!                       The shape of U_VT must verify: 
!                       size( U_VT, 1 ) = size( U_VT, 2 ) = min(m,n).
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the singular values into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd'.
!                       The singular vectors are rearranged accordingly.
!
!   MUL_SIZE            (INPUT, OPTIONAL) integer(i4b)
!                       Internal parameter. MUL_SIZE must verify 1 <= MUL_SIZE <= max(m,n),
!                       otherwise a default value is used. MUL_SIZE can be increased or decreased
!                       to improve the performance of the algorithm.
!
!                       The default value is 32.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm.
!
!                       The bidiagonal SVD algorithm of an intermediate bidiagonal 
!                       form B of MAT fails to converge if the number of QR sweeps 
!                       exceeds MAXITER * min(m,n). Convergence usually occurs in 
!                       about 2 * min(m,n) QR sweeps.
!
!                       The default is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm.
!
!                       MAX_FRANCIS_STEPS is a strictly positive integer, otherwise 
!                       the default value is used.
!
!                       The default is the minimum of min(m,n) and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   USE_SVD2            (INPUT, OPTIONAL) logical(lgl)
!                       If the optional argument USE_SVD2 is used and is set to true, an alternate
!                       SVD algorithm which used less workspace (but which may be slower) is
!                       automatically used if m is much larger than n or if n is much larger
!                       than m (e.g. if max(m,n)>=1.5 * min(m,n) ).
!                
!
! Further Details
! _______________
!
!   Computing the SVD of a rectangular matrix in subroutine SVD_CMP2 consists of three steps:
!
!       1) reduction of the rectangular matrix to bidiagonal form via orthogonal
!          transformations (e.g. Householder transformations);
!       2) in place accumulation of the orthogonal transformations used in the reduction
!          to bidiagonal form;
!       3) computation of the SVD of the bidiagonal matrix.
!
!   For further details, on the SVD of a rectangular matrix and the algorithm to
!   compute it, see the references (1) or (2).
!
!   All the three steps of the SVD algorithm (e.g. the reduction to bidiagonal form,
!   accumulation of the Householder transformations used in the reduction to bidiagonal form
!   and computation of the SVD of the bidiagonal matrix) are  parallelized if OPENMP is used.
!
!   For more informations, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq, lascl
    use QR_Procedures,   only : qr_cmp, lq_cmp, apply_q_lq, apply_q_qr, ortho_gen_lq, ortho_gen_qr
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:)   :: s
    real(stnd), intent(out),   dimension(:,:) :: u_vt
!
    integer(i4b), intent(in), optional :: mul_size, maxiter, max_francis_steps
!
    logical(lgl), intent(out)          :: failure
    logical(lgl), intent(in), optional :: perfect_shift, bisect, dqds, use_svd2
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension( size(s) )        :: e, tauq, taup, tau
    real(stnd)                              :: matnorm, tmp
    real(stnd), dimension(:,:), allocatable :: work, qmat
!
    integer(i4b) ::  i, j, k, l, m, n, inc, mnthr, i1, i2
    integer      ::  iok
!
    logical(lgl) :: mat_is_scaled, usesvd2
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='svd_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
    l = size( u_vt, 1)
    i = size( u_vt, 2)
    j = size( s      )
!
    k = assert_eq( min(m,n),         &
                   j,                &
                   l,                &
                   i,                &
                   name_proc  )
!
    if ( present(mul_size) ) then
        i = merge( mul_size, 32_i4b, 1_i4b<=mul_size )
    else
        i = 32_i4b
    end if
!
    inc = min( i, max(m,n) )
!
    usesvd2 = false
    if ( present(use_svd2) ) then
        usesvd2 = use_svd2
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, n )
        tmp = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
        s(:k) = zero
!
        do i = 1_i4b, k
            mat(i,i)   = one
            u_vt(:k,i) = zero
            u_vt(i,i)  = one
        end do
!
        return
!
    else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat  TO BIDIAGONAL FORM. IF max(m,n) EXCEEDS mnthr ,
!   A QR FACTORIZATION IS USED FIRST TO REDUCE THE MATRIX
!   mat TO A TRIANGULAR FORM.
!    
    mnthr = int( real( k, stnd )*thr_bd, i4b )
!
    if ( m>=n ) then
!
!       mat HAS AT LEAST AS MANY ROWS AS COLUMNS.
!       n LEFT SINGULAR VECTORS TO BE OVERWRITTEN ON mat AND
!       n RIGHT SINGULAR VECTORS TO BE COMPUTED IN u_vt .
!
        if (  m>=mnthr  ) then
!
!
!           PATH 1 : m MUCH LARGER THAN n .
!
!
!           FIRST REDUCE mat USING THE QR DECOMPOSITION.
!           SECOND REDUCE R TO UPPER BIDIAGONAL FORM.
!
!           COMPUTE mat = Q * R .
!
            call qr_cmp( mat, s, tau )
!
            if ( usesvd2 ) then
!
!               USE AN ALTERNATE SVD ALGORITHM AUTOMATICALLY (PATH 1b).
!
                iok = -1
!
            else
!
!               ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM (PATH 1a).
!
                allocate( qmat(m,n), work(n,n), stat = iok )
!
            end if
!
            if ( iok==0 ) then
!
!               PATH 1a : SUFFICIENT WORKSPACE FOR A FAST ALGORITHM.
!
!
!               COPY R TO work , ZEROING OUT BELOW IT.
!               MAKE A COPY OF Q IN qmat.
!
!$OMP PARALLEL DO IF(m*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
                do i = 1_i4b, n
!
                    work(1_i4b:i-1_i4b,i) = mat(1_i4b:i-1_i4b,i)
                    work(i,i) = s(i)
                    work(i+1_i4b:n,i) = zero
!
                    qmat(i:m,i) = mat(i:m,i)
!
                end do
!
!$OMP END PARALLEL DO
!
!               BIDIAGONALIZE R IN work .
!
                call bd_cmp_a( work, s, e, tauq, taup )
!
!               GENERATE RIGHT VECTORS BIDIAGONALIZING R IN u_vt .
!               GENERATE LEFT VECTORS BIDIAGONALIZING R IN work .
!
                call ortho_gen_bd2( work, tauq, taup, u_vt )
!
!               PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!               SINGULAR VECTORS OF R IN work AND COMPUTING RIGHT
!               SINGULAR VECTORS OF R IN u_vt .
!
                call bd_svd2_a( true, s, e, failure, work, u_vt, sort=sort,           &
                                maxiter=maxiter, max_francis_steps=max_francis_steps, &
                                perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!               GENERATE Q IN qmat .
!
                call ortho_gen_qr( qmat, tau )
!
!               MULTIPLY Q IN qmat BY LEFT SINGULAR VECTORS OF R IN
!               work , STORING RESULT IN mat .
!
#ifdef _BLAS
                call gemm( 'N', 'N', m, n, n, one, qmat(1_i4b:m,1_i4b:n), m,        &
                           work(1_i4b:n,1_i4b:n), n, zero, mat(1_i4b:m,1_i4b:n), m  )
#else
                mat(:m,:n) = matmul( qmat(:m,:n), work(:n,:n) )
#endif
!
!               DEALLOCATE WORKSPACE.
!
                deallocate( qmat, work )
!
            else
!
!               PATH 1b : INSUFFICIENT WORKSPACE FOR A FAST ALGORITHM.
!
!
!               COPY R TO u_vt , ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
                do i = 1_i4b, n
                    u_vt(1_i4b:i-1_i4b,i) = mat(1_i4b:i-1_i4b,i)
                    u_vt(i,i) = s(i)
                    u_vt(i+1_i4b:n,i) = zero
                end do
!
!$OMP END PARALLEL DO
!
!               GENERATE Q IN mat .
!
                call ortho_gen_qr( mat, tau )
!
!               BIDIAGONALIZE R IN u_vt .
!
                call bd_cmp_a( u_vt, s, e, tauq, taup )
!
!               MULTIPLY Q IN mat BY LEFT VECTORS BIDIAGONALIZING R.
!
                call apply_q_qr( u_vt, tauq, mat, left=false, trans=false )
!
!               SHIFT THE VECTORS, IN u_vt, WHICH DEFINE THE ELEMENTARY 
!               REFLECTORS NEEDED TO GENERATE RIGHT VECTORS OF R ONE ROW
!               DOWNWARD.
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC,1), PRIVATE(i)
!
                do i = 2_i4b, n
                    u_vt(2_i4b:i,i) = u_vt(1_i4b:i-1_i4b,i)
                end do
!
!$OMP END PARALLEL DO
!
!               GENERATE RIGHT VECTORS BIDIAGONALIZING R IN u_vt .
!
                call ortho_gen_lq( u_vt(2_i4b:,2_i4b:), taup(:n-1_i4b) )        
!
                u_vt(1_i4b,1_i4b) = one
                u_vt(2_i4b:n,1_i4b) = zero
                u_vt(1_i4b,2_i4b:n) = zero
!
!               PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!               SINGULAR VECTORS OF mat IN mat AND COMPUTING RIGHT
!               SINGULAR VECTORS OF mat IN u_vt .
!
                call bd_svd2_a( true, s, e, failure, mat, u_vt, sort=sort,            &
                                maxiter=maxiter, max_francis_steps=max_francis_steps, &
                                perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
            end if
!
        else
!
!           PATH 2 : m AT LEAST n, BUT NOT MUCH LARGER;
!
!           REDUCE mat TO UPPER BIDIAGONAL FORM WITHOUT 
!           QR DECOMPOSITION.
!
!           BIDIAGONALIZE mat .
!
            call bd_cmp_a( mat, s, e, tauq, taup )
!
!           GENERATE RIGHT BIDIAGONALIZING VECTORS IN u_vt .
!           GENERATE LEFT BIDIAGONALIZING VECTORS IN mat .
!
            call ortho_gen_bd2( mat, tauq, taup, u_vt )
!
!           PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!           SINGULAR VECTORS OF mat IN mat AND COMPUTING RIGHT
!           SINGULAR VECTORS OF mat IN u_vt .
!
            call bd_svd2_a( true, s, e, failure, mat, u_vt, sort=sort,            &
                            maxiter=maxiter, max_francis_steps=max_francis_steps, &
                            perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
        end if
!
    else
!
!       mat HAS MORE COLUMNS THAN ROWS.
!       m RIGHT SINGULAR VECTORS TO BE OVERWRITTEN ON mat AND
!       m LEFT SINGULAR VECTORS TO BE COMPUTED IN u_vt .
!
        if (  n>=mnthr ) then
!
!           PATH 3 : n MUCH LARGER THAN m.
!
!           FIRST REDUCE mat USING THE LQ DECOMPOSITION.
!           SECOND REDUCE L TO UPPER BIDIAGONAL FORM.
!
!           COMPUTE mat = L * Q .
!
            call lq_cmp( mat, s, tau )
!
            if ( usesvd2 ) then
!
!               USE AN ALTERNATE SVD ALGORITHM AUTOMATICALLY (PATH 3b).
!
                iok = -1
!
            else
!
!               ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM (PATH 3a).
!
                allocate( qmat(m,n), work(m,m), stat = iok )
!
            end if
!
            if ( iok==0 ) then
!
!               PATH 3a : SUFFICIENT WORKSPACE FOR A FAST ALGORITHM.
!
!               COPY Q TO qmat .
!               COPY L TO u_vt , ZEROING ABOUT ABOVE IT.
!
!$OMP PARALLEL DO IF(n*m>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
                do i = 1_i4b, n
!
                    if ( i<=m ) then
                        qmat(1_i4b:i,i) = mat(1_i4b:i,i)
                        u_vt(1_i4b:i-1_i4b,i) = zero
                        u_vt(i,i) = s(i)
                        u_vt(i+1_i4b:m,i) = mat(i+1_i4b:m,i)
                    else
                        qmat(1_i4b:m,i) = mat(1_i4b:m,i)
                    end if
!
                end do
!
!$OMP END PARALLEL DO
!
!               UPPER BIDIAGONALIZE L IN u_vt .
!
                call bd_cmp_a( u_vt, s, e, tauq, taup )
!
!               GENERATE LEFT VECTORS BIDIAGONALIZING L IN u_vt .
!               GENERATE RIGHT VECTORS BIDIAGONALIZING L IN work .
!
                call ortho_gen_bd2( u_vt, tauq, taup, work )
!
!               PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!               SINGULAR VECTORS OF L IN u_vt AND COMPUTING RIGHT
!               SINGULAR VECTORS OF L IN work .
!
                call bd_svd2_a( true, s, e, failure, u_vt, work, sort=sort,           &
                                maxiter=maxiter, max_francis_steps=max_francis_steps, &
                                perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!               GENERATE Q IN qmat .
!
                call ortho_gen_lq( qmat, tau )
!
!               MULTIPLY RIGHT SINGULAR VECTORS OF L IN work BY Q
!               IN qmat, STORING RESULT IN mat.
!
#ifdef _BLAS
                call gemm( 'N', 'N', m, n, m, one, work(1_i4b:m,1_i4b:m), m,        &
                            qmat(1_i4b:m,1_i4b:n), m, zero, mat(1_i4b:m,1_i4b:n), m )
#else
                mat(:m,:n) = matmul( work(:m,:m), qmat(:m,:n) )
#endif
!
!               DEALLOCATE WORKSPACE.
!
                deallocate( qmat, work )
!
            else
!
!               PATH 3b : INSUFFICIENT WORKSPACE FOR A FAST ALGORITHM.
!
!               COPY L TO u_vt , ZEROING ABOUT ABOVE IT.
!
!$OMP PARALLEL DO IF(m*m>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
                do i = 1_i4b, m
                    u_vt(1_i4b:i-1_i4b,i) = zero
                    u_vt(i,i) = s(i)
                    u_vt(i+1_i4b:m,i) = mat(i+1_i4b:m,i)
                end do
!
!$OMP END PARALLEL DO
!
!               GENERATE Q IN mat .
!
                call ortho_gen_lq( mat, tau )
!
!               UPPER BIDIAGONALIZE L IN u_vt .
!
                call bd_cmp_a( u_vt, s, e, tauq, taup )
!
!               MULTIPLY RIGHT VECTORS BIDIAGONALIZING L BY Q IN mat.
!
                call apply_q_lq( u_vt(:,2_i4b:), taup(:m-1_i4b), mat(2_i4b:,:),     &
                                 left=true, trans=false )
!
!               GENERATE LEFT VECTORS BIDIAGONALIZING L IN u_vt .
!
                call ortho_gen_qr( u_vt, tauq )
!
!               PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!               SINGULAR VECTORS OF mat IN u_vt AND COMPUTING RIGHT
!               SINGULAR VECTORS OF mat IN mat .
!
                call bd_svd2_a( true, s, e, failure, u_vt, mat, sort=sort,            &
                                maxiter=maxiter, max_francis_steps=max_francis_steps, &
                                perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
            end if
!
        else
!
!           PATH 4 : n AT LEAST m, BUT NOT MUCH LARGER.
!
!           REDUCE mat TO LOWER BIDIAGONAL FORM WITHOUT
!           LQ DECOMPOSITION.
!
!           BIDIAGONALIZE mat .
!
            call bd_cmp_a( mat, s, e, tauq, taup )
!
!           GENERATE LEFT BIDIAGONALIZING VECTORS IN u_vt .
!           GENERATE RIGHT BIDIAGONALIZING VECTORS IN mat .
!
            call ortho_gen_bd2( mat, tauq, taup, u_vt )
!
!           PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!           SINGULAR VECTORS OF mat IN u_vt AND COMPUTING RIGHT
!           SINGULAR VECTORS OF mat IN mat .
!
            call bd_svd2_a( false, s, e, failure, u_vt, mat, sort=sort,           &
                            maxiter=maxiter, max_francis_steps=max_francis_steps, &
                            perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
        end if
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled ) then
!
        call lascl( s, merge(smlnum,bignum,matnorm<smlnum), matnorm )
!
    end if
!
!
! END OF SUBROUTINE svd_cmp2_a
! ____________________________
!
    end subroutine svd_cmp2_a
!
! =========================================================================================
!
    subroutine svd_cmp_b( mat, s, failure, sort, mul_size, maxiter, bisect, dqds, &
                          d, e, tauq, taup )
!
! Purpose
! _______
!
!   SVD_CMP computes the singular values of a real m-by-n matrix MAT.
!
!   The Singular Value Decomposition (SVD) is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix. The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!
!   The original matrix MAT is first reduced to upper or lower
!   bidiagonal form BD by an orthogonal transformation:
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal. The singular values SIGMA of the bidiagonal
!   matrix BD are then computed by the bidiagonal implicit QR algorithm (if BISECT=false
!   and DQDS=false), the bisection method (if BISECT=true) or the dqds algorithm (if DQDS=true)
!   applied to the bidiagonal matrix BD.
!
!
!   The routine outputs SIGMA and optionally Q and P (in packed form), and BD for
!   a given matrix MAT. SIGMA, Q, P and BD may then be used to obtain selected
!   singular vectors with subroutines BD_INVITER, BD_INVITER2, BD_DEFLATE or BD_DEFLATE2.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n matrix MAT.
!
!                On exit, MAT is destroyed and if TAUQ or TAUP are present
!                MAT is overwritten as follows:
!
!                - if m >= n, the elements on and below the diagonal,
!                  with the array TAUQ, represent the orthogonal
!                  matrix Q as a product of elementary reflectors, 
!                  and the elements above the diagonal, with the
!                  array TAUP, represent the orthogonal matrix P
!                  as a product of elementary reflectors;
!                - if m < n, the elements below the diagonal, with
!                  the array TAUQ, represent the orthogonal
!                  matrix Q as a product of elementary reflectors,
!                  and the elements on and above the diagonal,
!                  with the array TAUP, represent the orthogonal
!                  matrix P as a product of elementary reflectors.
!
!                See Further Details.
!
!   S            (OUTPUT) real(stnd), dimension(:)
!                On exit, the singular values SIGMA of MAT if FAILURE is equal to false.
!                If FAILURE is equal to true, the sign of the incorrect singular values
!                is set to negative.
!
!                The size of S must be min( size(MAT,1) , size(MAT,2) ) = min(m,n).
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the implicit QR, bisection or dqds
!                  algorithm used to compute the singular values of the bidiagonal
!                  form BD of the input m-by-n matrix MAT did not converge and that
!                  full accuracy was not attained in the bidiagonal SVD of this
!                  intermediate bidiagonal form BD of MAT.
!                  The sign of the incorrect singular values is set to negative.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!                By default, the singular values are not sorted.
!
!   MUL_SIZE     (INPUT, OPTIONAL) integer(i4b)
!                Internal parameter. MUL_SIZE must verify: 1 <= MUL_SIZE <= max(m,n),
!                otherwise a default value is used. For better performance, 
!                at the expense of more workspace, a large value can be used.
!
!                The default is 32.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the
!                bidiagonal implicit QR phase of the SVD algorithm.
!
!                The bidiagonal SVD algorithm of an intermediate bidiagonal 
!                form BD of MAT fails to converge if the number of QR sweeps 
!                exceeds MAXITER * min(m,n). Convergence usually occurs in 
!                about 2 * min(m,n) QR sweeps.
!
!                This argument has no effect if BISECT or DQDS is equal to true.
!
!                The default is 10.
!
!   BISECT       (INPUT, OPTIONAL) logical(lgl)
!                BISECT determines how the singular values of the intermediate min(m,n)-by-min(m,n)
!                bidiagonal matrix BD are computed.
!
!                If BISECT is set to true, singular values are computed with a more accurate
!                bisection algorithm delivering improved accuracy in the final computed SVD
!                decomposition at the expense of a slower execution time.
!
!                If BISECT is set to false, singular values are computed with the bidiagonal implicit
!                QR or dqds algorithm applied to the associated min(m,n)-by-min(m,n) bidiagonal matrix BD.
!
!                The default is false.
!
!   DQDS         (INPUT, OPTIONAL) logical(lgl)
!                DQDS determines how the singular values of the intermediate min(m,n)-by-min(m,n)
!                bidiagonal matrix BD are computed.
!
!                If DQDS is set to true, singular values are computed with a more accurate
!                dqds algorithm delivering improved accuracy in the final computed SVD
!                decomposition. Moreover, the dqds algorithm is usually also faster than the
!                bisection or bidiagonal implicit QR algorithms and is the default method
!                used in SVD_CMP for computing singular values of the bidiagonal form BD of the
!                input matrix MAT.
!
!                If DQDS is set to false, singular values are computed with the bidiagonal implicit
!                QR or bisection algorithm applied to the associated min(m,n)-by-min(m,n) bidiagonal matrix BD.
!
!                If both optional arguments BISECT and DQDS are specified with the value true, the
!                dqds algorithm is used.
!
!                If both optional arguments BISECT and DQDS are specified with the value false, the
!                bidiagonal implicit QR algorithm is used.
!
!                The default is true.
!
!   D            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The diagonal elements of the intermediate bidiagonal matrix BD
!
!                The size of D must be min( size(MAT,1) , size(MAT,2) ) = min(m,n).
!
!   E            (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The off-diagonal elements of the intermediate bidiagonal matrix BD:
!
!                - if m >= n, E(i) = BD(i-1,i) for i = 2,3,...,n;
!                - if m < n,  E(i) = BD(i,i-1) for i = 2,3,...,m.
!
!                The size of E must be min( size(MAT,1) , size(MAT,2) ) = min(m,n).
!
!   TAUQ         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix Q. See Further Details.
!
!                The size of TAUQ must be min( size(MAT,1) , size(MAT,2) ) = min(m,n).
!
!   TAUP         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                The scalar factors of the elementary reflectors which
!                represent the orthogonal matrix P. See Further Details.
!
!                The size of TAUP must be min( size(MAT,1) , size(MAT,2) ) = min(m,n).
!
!
! Further Details
! _______________
!
!   Computing the singular values of a rectangular matrix in subroutine SVD_CMP consists of two steps:
!
!       1) reduction of the rectangular matrix to bidiagonal form BD, see the references (1) and (2);
!       2) computation of the singular values of the  min(m,n)-by-min(m,n) bidiagonal matrix BD by a
!          bidiagonal implicit QR algorithm (if BISECT=false and DQDS=false), a dqds algorithm (if DQDS=true)
!          or a bisection algorithm (if BISECT=true). See the references (1), (2), (3) and (4) for a description
!          of these different algorithms.
!
!   Note that if max(m,n) is much larger than min(m,n) and the optional arguments TAUQ and TAUP are not used,
!   the rectangular matrix is first reduced to upper or lower triangular form by a QR or LQ factorization and
!   the reduction algorithm is applied to the resulting triangular factor. The singular values of the rectangular
!   matrix are then obtained from those of the triangular factor. This usually gives a large speedup in the
!   computations.
!
!   The matrices Q and P in the bidiagonal reduction of the input m-by-n matrix MAT are represented
!   as products of elementary reflectors:
!
!   - If m >= n,
!
!         Q = H(1) * H(2) * . . . * H(n)  and  P = G(1) * G(2) * . . . * G(n-1)
!
!     Each H(i) and G(i) has the form:
!
!         H(i) = I + tauq * u * u'  and G(i) = I + taup * v * v'
!
!     where tauq and taup are real scalars, and u and v are real vectors.
!     Moreover, u(1:i-1) = 0 and v(1:i) = 0.
!
!     If TAUQ or TAUP are present :
!
!     - u(i:m)   is stored on exit in MAT(i:m,i);
!     - v(i+1:n) is stored on exit in MAT(i,i+1:n).
!
!     If TAUQ is present : tauq is stored in TAUQ(i).
!
!     If TAUP is present : taup is stored in TAUP(i).
!
!
!   - If m < n,
!
!         Q = H(1) * H(2) * . . . * H(m-1)  and  P = G(1) * G(2) * . . . * G(m)
!
!     Each H(i) and G(i) has the form:
!
!         H(i) = I + tauq * u * u'  and G(i) = I + taup * v * v'
!
!     where tauq and taup are real scalars, and u and v are real vectors.
!     Moreover, u(1:i) = 0 and v(1:i-1) = 0.
!
!     If TAUQ or TAUP are present :
!
!     - u(i+1:m) is stored on exit in MAT(i+1:m,i);
!     - v(i:n)   is stored on exit in MAT(i,i:n).
!
!     If TAUQ is present : tauq is stored in TAUQ(i).
!
!     If TAUP is present : taup is stored in TAUP(i).
!
!   The contents of MAT on exit, if TAUQ or TAUP are present,
!   are illustrated by the following examples:
!
!   - m = 6 and n = 5 (m >= n):
!
!        (  u1  v1  v1  v1  v1 )
!
!        (  u1  u2  v2  v2  v2 )
!
!        (  u1  u2  u3  v3  v3 )
!
!        (  u1  u2  u3  u4  v4 )
!
!        (  u1  u2  u3  u4  u5 )
!
!        (  u1  u2  u3  u4  u5 )
!
!   - m = 5 and n = 6 (m < n):
!
!        (  v1  v1  v1  v1  v1  v1 )
!
!        (  u1  v2  v2  v2  v2  v2 )
!
!        (  u1  u2  v3  v3  v3  v3 )
!
!        (  u1  u2  u3  v4  v4  v4 )
!
!        (  u1  u2  u3  u4  v5  v5 )
!
!   where ui denotes an element of the vector defining H(i), and vi an element 
!   of the vector defining G(i).
!
!   For further details, on the SVD of a rectangular matrix and the algorithms to
!   compute it, see the references (1) or (2). In SVD_CMP subroutine, the reduction
!   to bidiagonal form by orthogonal transformations is parallelized if OPENMP is used,
!   the computation of the singular values is also parallelized if OPENMP is used and
!   BISECT is used with the value true. Further details on the dqds algorithm
!   are given in the references (3) and (4).
!
!   For more informations, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (4) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert, assert_eq, lascl
    use QR_Procedures,   only : qr_cmp, lq_cmp
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout),  dimension(:,:)         :: mat
    real(stnd),  intent(out),    dimension(:)           :: s
    real(stnd),  intent(out),    dimension(:), optional :: tauq, taup, d, e
!
    integer(i4b), intent(in), optional ::  mul_size, maxiter
!
    logical(lgl), intent(out)          :: failure
    logical(lgl), intent(in), optional :: bisect, dqds
!
    character , intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: matnorm, tmp, abstol
    real(stnd), dimension(size(s)) :: d2, e2, tauq2, taup2
!
    integer(i4b) :: i, j, k, l, m, m2, n, n2, inc, mnthr, i1, i2, info
!
    logical(lgl) :: mat_is_scaled, use_bisect, use_dqds
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='svd_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = assert_eq( min(m,n),         &
                   int(size(s),i4b), &
                   name_proc  )
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat  TO BIDIAGONAL FORM. IF max(m,n) EXCEEDS mnthr ,
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE MATRIX
!   mat TO A TRIANGULAR FORM.
!    
    mnthr = int( real( k, stnd )*thr_bd, i4b )
!
!   TEST OPTIONAL ARGUMENTS.
!
    if ( present(d) ) then
        call assert( logical( k==size(d), lgl ), name_proc )
    end if
!
    if ( present(e) ) then
        call assert( logical( k==size(e), lgl ), name_proc )
    end if
!
    if ( present(tauq) ) then
        call assert( logical( k==size(tauq), lgl ), name_proc )
        mnthr = huge( i )
    end if
!
    if ( present(taup) ) then
        call assert( logical( k==size(taup), lgl ), name_proc )
        mnthr = huge( i )
    end if
!
    if ( present(mul_size) ) then
        inc = merge( mul_size, 32_i4b, 1_i4b<=mul_size .and. mul_size<=max(m,n) )
    else
        inc = 32_i4b
    end if
!
    use_bisect = false
!
    if ( present( bisect ) ) then
!
        use_bisect = bisect
!
    end if
!
    use_dqds = true
!
    if ( present(dqds) ) then
!
        use_dqds = dqds
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, n )
        tmp = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
        s(:k) = zero
!
        if ( present(d) )    d(:k) = zero
        if ( present(e) )    e(:k) = zero
        if ( present(tauq) ) tauq(:k) = zero
        if ( present(taup) ) taup(:k) = zero
!
        return
!
    else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
        call lascl( mat, matnorm, bignum )
!
    end if
!
    m2 = m
    n2 = n
!
    if (  m>mnthr  ) then
!
!       m MUCH LARGER THAN n .
!
!       FIRST REDUCE mat USING THE QR DECOMPOSITION.
!
!       COMPUTE mat = Q * R .
!
        call qr_cmp( mat, tauq2, taup2 )
!
!       RECONSTRUCT R IN mat, ZEROING OUT BELOW IT.
!
!$OMP  PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)         &
!$OMP             ,SCHEDULE(STATIC,1), PRIVATE(i)
!
        do i = 1_i4b, n
!
            mat(i,i) = tauq2(i)
            mat(i+1_i4b:n,i) = zero
!
        end do
!
!$OMP END PARALLEL DO
!
        m2 = n
!
    else if (  n>mnthr ) then
!
!       n MUCH LARGER THAN m.
!
!       FIRST REDUCE mat USING THE LQ DECOMPOSITION.
!
!       COMPUTE mat = L * Q .
!
        call lq_cmp( mat, tauq2, taup2 )
!
!       RECONSTRUCT L IN mat , ZEROING ABOUT ABOVE IT.
!
!$OMP  PARALLEL DO IF((m*(m+1_i4b))/2_i4b>=omp_limit)         &
!$OMP             ,SCHEDULE(STATIC,1), PRIVATE(i)
!
        do i = 1_i4b, m
!
            mat(1_i4b:i-1_i4b,i) = zero
            mat(i,i) = tauq2(i)
!
        end do
!
!$OMP END PARALLEL DO
!
        n2 = m
!
    end if
!
!   REDUCE mat TO LOWER OR UPPER BIDIAGONAL FORM.
!
    call bd_cmp_a( mat(:m2,:n2), d2, e2, tauq2, taup2 )
!
!   SAVE INTERMEDIATE QUANTITIES, IF REQUIRED.
!
    if ( present(d) ) then
        d = d2
    end if
!
    if ( present(e) ) then
        e = e2
    end if
!
    if ( present(tauq) ) then
        tauq = tauq2
    end if
!
    if ( present(taup) ) then
        taup = taup2
    end if
!
    if ( use_dqds ) then
!
!       COMPUTE SINGULAR VALUES BY THE DQDS ALGORITHM.
!
        call bd_lasq1( d2, e2, failure, maxiter=60_i4b, sort=sort, info=info )
!
        if ( info==2_i4b ) then
!
            call bd_dqds( d2, e2, failure, maxiter=60_i4b, sort=sort )
!
        end if
!
        if ( .not.failure ) then
!
            s(:k) = d2(:k)
!
        else
!
            s(:k) = -one
!
        end if
!
    else if ( use_bisect ) then
!
!       COMPUTE SINGULAR VALUES BY A BISECTION ALGORITHM.
!
        abstol = sqrt( safmin )
!
        call bd_singval( d2, e2, i, s, failure, sort=sort, abstol=abstol )
!
    else
!
!       PERFORM BIDIAGONAL QR ITERATIONS FOR COMPUTING THE SINGULAR VALUES.
!
        call bd_svd_t( true, d2, e2, failure, sort=sort, maxiter=maxiter )
!
        s(:k) = d2(:k)
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled ) then
!
        call lascl( s, merge(smlnum,bignum,matnorm<smlnum), matnorm )
!
    end if
!
!
! END OF SUBROUTINE svd_cmp_b
! ___________________________
!
    end subroutine svd_cmp_b
!
! =========================================================================================
!
    subroutine svd_cmp3_a( mat, s, failure, u_v, sort, maxiter, max_francis_steps, &
                           perfect_shift, bisect, dqds, reortho, failure_bd )
!
! Purpose
! _______
!
!   SVD_CMP3 computes the Singular Value Decomposition (SVD) of a real
!   m-by-n matrix MAT. The SVD is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix. The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!   The columns of U and V are, respectively, the left and right
!   singular vectors of MAT.
!
!   The routine returns the first min(m,n) singular values and the associated
!   left and right singular vectors. The right singular vectors are
!   returned row-wise if m<n.
!
!   MAT (or MAT' if m<n) is first reduced to bidiagonal form B with the help of the
!   Ralha-Barlow one-sided bidiagonalization algorithm, see the references (1) and (2).
!
!   The singular values, left and right singular vectors of B are then computed by the
!   bidiagonal implicit QR algorithm applied to B, see the references (3) and (4).
!
!   The singular vectors of MAT are finally computed by a back transformation algorithm.
!
!   In cases of a very large condition number of MAT, SVD_CMP3 may compute left (right if m<n) 
!   singular vectors of MAT, which are not numerically orthogonal (see Further Details).
!   However, the largest left (right if m<n) singular vectors of MAT are always numerically
!   orthogonal even if MAT is singular or nearly singular (see Further Details).
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit:
!
!                       - if m>=n, MAT is overwritten with the first n
!                         columns of U (the left singular vectors,
!                         stored column-wise);
!                       - if m<n, MAT is overwritten with the first m rows
!                         of V' (the first m right singular
!                         vectors, stored row-wise);
!
!   S                   (OUTPUT) real(stnd), dimension(:)
!                       The singular values of MAT.
!
!                       The size of S must verify: size( S ) = min(m,n) .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit.
!                       - FAILURE = true  :  indicates that the bidiagonal SVD algorithm
!                         did not converge and that full accuracy was not attained in the
!                         bidiagonal SVD of an intermediate bidiagonal form B of MAT.
!
!   U_V                 (OUTPUT) real(stnd), dimension(:,:)
!                       On exit:
!
!                       - if m>=n,     U_V contains the n-by-n orthogonal matrix V;
!                       - if m<n,      U_V contains the m-by-m orthogonal matrix U.
!
!                       The shape of U_V must verify: 
!                       size( U_V, 1 ) = size( U_V, 2 ) = min(m,n) .
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the singular values into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd'. If this argument is not
!                       used the singular values are not sorted.
!                       The singular vectors are rearranged accordingly.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm.
!
!                       The bidiagonal SVD algorithm of an intermediate bidiagonal 
!                       form B of MAT fails to converge if the number of QR sweeps 
!                       exceeds MAXITER * min(m,n). Convergence usually occurs in 
!                       about 2 * min(m,n) QR sweeps.
!
!                       The default is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm.
!
!                       MAX_FRANCIS_STEPS is a strictly positive integer, otherwise 
!                       the default value is used.
!
!                       The default is the minimum of min(m,n) and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm applied to the associated symmetric tridiagonal
!                       matrix B * B'  whose eigenvalues are the squares of the singular values of B or MAT.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm applied to the associated symmetric tridiagonal
!                       matrix B * B'  whose eigenvalues are the squares of the singular values of B of MAT.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   REORTHO             (INPUT, OPTIONAL) logical(lgl)
!                       If the optional argument REORTHO is set to true, reorthogonalization is performed
!                       (when needed) in the Ralha-Barlow one-sided bidiagonalization of MAT.
!                       If this optional logical argument is set to false, reorthogonalization is never
!                       performed, which can result in a loss of orthogonality for the left (right)
!                       singular vectors of MAT if m>=n (m<n) and MAT has a large condition number.
!
!                       See description of the BD_CMP2 subroutine for further details.
!
!                       The default is REORTHO = true.
!
!   FAILURE_BD          (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit:
!
!                       - FAILURE_BD = false :  indicates that maximum accuracy was obtained in the
!                         Ralha-Barlow one-sided bidiagonalization of MAT.
!                       - FAILURE_BD = true  :  indicates that MAT is nearly singular and some
!                         loss of orthogonality can be expected in the Ralha-Barlow
!                         bidiagonalization algorithm.
!
!                       See description of the BD_CMP2 subroutine for further details.
!                
!
! Further Details
! _______________
!
!   Computing the SVD of a m-by-n matrix MAT with m>=n in subroutine SVD_CMP3 consists of three steps:
!
!       1) reduction of the rectangular matrix to bidiagonal form B via the Ralha-Barlow
!          one-sided bidiagonalization algorithm, see the references (1) and (2);
!       2) in place accumulation of the orthogonal transformations used in the reduction
!          to bidiagonal form B, see the references (3) and (4);
!       3) computation of the SVD of the bidiagonal matrix B, see the references (3) and (4).
!
!   In cases of a large condition number of MAT, this three-step algorithm may compute
!   left singular vectors of MAT, which are not numerically orthogonal. This is because the
!   left (orthogonal) matrix in the bidiagonal decomposition of MAT (estimated by the Ralha-Barlow
!   one-sided bidiagonalization algorithm) may also not be numerically orthogonal as it is computed
!   by a recurrence relationship, see the references (1) and (2) for details. To correct partly this
!   deficiency, partial reorthogonalization can be performed to ensure orthogonality at the expense
!   of speed of computation. The reorthogonalization uses the Gram-Schmidt method described in the
!   reference (5) and is activated if the optional logical argument REORTHO is set to true.
!
!   However, the largest left singular vectors of MAT are always numerically orthogonal even if MAT
!   is singular or nearly singular, see the reference (1).
!
!   If m<n, this three-step algorithm is applied to MAT', instead of MAT, to get the SVD of MAT and
!   it computes also numerically orthogonal right singular vectors of MAT in that case.
!
!   Note that if max(m,n) is much larger than min(m,n), the rectangular matrix is first reduced to
!   upper or lower triangular form by a QR or LQ factorization and the three-steps reduction algorithm
!   is applied to the resulting triangular factor. The singular vectors of the rectangular matrix
!   are then obtained from those of the triangular factor by a back-transformation algorithm.
!
!   For further details on the SVD of a rectangular matrix and the algorithms to
!   compute it, see the references below.
!
!   The three or four steps of the SVD algorithm used here (e.g., preliminary QR or LQ factorization,
!   reduction to bidiagonal form B, in place accumulation of the orthogonal transformations and 
!   computation of the SVD of the bidiagonal matrix B) are parallelized if OPENMP is used.
!
!   For more details, see:
!
!   (1) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!            A new stable bidiagonal reduction algorithm.
!            Linear Algebra Appl., No 397, pp. 35-84.
!
!   (2) Bosner, N., and Barlow, J.L., 2007:
!            Block and Parallel versions of one-sided bidiagonalization.
!            SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!            Matrix Computations.
!            3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (4) Lawson, C.L., and Hanson, R.J., 1974:
!            Solving least square problems.
!            Prentice-Hall.
!
!   (5) Stewart, G.W., 2007:
!           Block Gram-Schmidt Orthogonalization.
!           Report TR-4823, Department of Computer Science, College Park, University of Maryland.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert_eq, lascl, merror
    use QR_Procedures,   only : qr_cmp, ortho_gen_qr
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:)   :: s
    real(stnd), intent(out),   dimension(:,:) :: u_v
!
    integer(i4b), intent(in), optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: perfect_shift, bisect, dqds, reortho
    logical(lgl), intent(out), optional :: failure_bd
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension( size(s) )        :: e
    real(stnd)                              :: matnorm, tmp
    real(stnd), dimension(:,:), allocatable :: work, mat2
!
    integer(i4b) ::  i, j, k, l, m, n, mnthr
    integer      ::  iok
!
    logical(lgl) :: mat_is_scaled
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='svd_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
    if ( present( failure_bd ) ) then
!
        failure_bd = false
!
    end if
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
    l = size( u_v, 1 )
    i = size( u_v, 2 )
    j = size( s      )
!
    k = assert_eq( min(m,n),  &
                   j,         &
                   l,         &
                   i,         &
                   name_proc  )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
    do i = 1_i4b, n
!
        tmp = maxval( abs( mat(1_i4b:m,i) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       mat IS A ZERO MATRIX.
!
        s(1_i4b:k) = zero
!
        do i = 1_i4b, k
            mat(i,i)       = one
            u_v(1_i4b:k,i) = zero
            u_v(i,i)       = one
        end do
!
        return
!
    else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat TO BIDIAGONAL FORM. IF max( m, n ) EXCEEDS mnthr ,
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE
!   m-BY-n MATRIX mat TO A TRIANGULAR FORM.
!    
    mnthr = int( real( k, stnd )*thr_bd2, i4b )
!
    if ( m>=n ) then
!
!       mat HAS AT LEAST AS MANY ROWS AS COLUMNS.
!       n LEFT SINGULAR VECTORS TO BE OVERWRITTEN ON mat AND
!       n RIGHT SINGULAR VECTORS TO BE COMPUTED IN u_v .
!
        if ( m>=mnthr ) then
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM (PATH 1).
!
            allocate( mat2(m,n), work(n,n), stat = iok )
!
        else
!
!           USE THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!           WITHOUT A PRELIMINARY QR FACTORIZATION (PATH 2).
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!
!           PATH 1 : m MUCH LARGER THAN n AND SUFFICIENT WORKSPACE FOR A FAST ALGORITHM.
!
!           REDUCE mat USING THE QR DECOMPOSITION.
!           SECOND REDUCE R TO UPPER BIDIAGONAL FORM WITH
!           THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!           MAKE FIRST A COPY OF mat .
!
            mat2(:m,:n) = mat(:m,:n)
!
!           COMPUTE mat = Q * R .
!
            call qr_cmp( mat2(:m,:n), s(:n), e(:n) )
!
!           COPY R TO work , ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, n
!
                work(1_i4b:i-1_i4b,i) = mat2(1_i4b:i-1_i4b,i)
                work(i,i) = s(i)
                work(i+1_i4b:n,i) = zero
!
            end do
!
!$OMP END PARALLEL DO
!
!           GENERATE Q IN mat2 .
!
            call ortho_gen_qr( mat2(:m,:n), e(:n) )
!
!           BIDIAGONALIZE R STORED IN work WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
            call bd_cmp2_a( work(:n,:n), s(:n), e(:n), u_v(:n,:n), failure=failure_bd, reortho=reortho )
!
!           PERFORM BIDIAGONAL QR ITERATION AFTER BIDIAGONALIZATION OF R IN work ,
!           COMPUTING n LEFT SINGULAR VECTORS OF R IN work AND COMPUTING n RIGHT
!           SINGULAR VECTORS OF R (AND mat) IN u_v .
!
            call bd_svd_a( true, s(:n), e(:n), failure, work(:n,:n), u_v(:n,:n), sort=sort,   &
                           maxiter=maxiter, max_francis_steps=max_francis_steps,              &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!           MULTIPLY Q IN mat2 BY LEFT SINGULAR VECTORS OF R IN
!           work , TO GET LEFT SINGULAR VECTORS OF mat, STORING
!           RESULT IN mat .
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, n, n, one, mat2, m, work, n, zero, mat, m )
#else
            mat(:m,:n) = matmul( mat2(:m,:n), work(:n,:n) )
#endif
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( mat2, work )
!     
        else
!
!           PATH 2 : mat HAS AT LEAST AS MANY ROWS AS COLUMNS, BUT m IS NOT MUCH LARGER THAN n.
!                    BIDIAGONALIZE mat WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!                    DIRECTLY.
!
            call bd_cmp2_a( mat(:m,:n), s(:k), e(:k), u_v(:k,:k), failure=failure_bd, reortho=reortho )
!
!           PERFORM BIDIAGONAL QR ITERATION AFTER BIDIAGONALIZATION OF mat,
!           COMPUTING n LEFT SINGULAR VECTORS OF mat IN mat AND COMPUTING n RIGHT
!           SINGULAR VECTORS OF mat IN u_v .
!
            call bd_svd_a( true, s(:k), e(:k), failure, mat(:m,:n), u_v(:k,:k), sort=sort,    &
                           maxiter=maxiter, max_francis_steps=max_francis_steps,              &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!     
        end if
!
    else
!
!       mat HAS MORE COLUMNS THAN ROWS.
!       m RIGHT SINGULAR VECTORS TO BE OVERWRITTEN ON mat AND
!       m LEFT SINGULAR VECTORS TO BE COMPUTED IN u_v .
!
!       ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
        allocate( mat2(n,m), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       FIRST TRANSPOSE mat .
!
        mat2(:n,:m) = transpose( mat(:m,:n) )
!
        if ( n>=mnthr ) then
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM (PATH 3).
!
#ifdef _BLAS
            allocate( work(m,m), stat = iok )
#else
            allocate( work(n,m), stat = iok )
#endif
!
        else
!
!           USE THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!           WITHOUT A PRELIMINARY QR FACTORIZATION OF mat**(t) (PATH 4).
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!
!           PATH 3 : n MUCH LARGER THAN m AND SUFFICIENT WORKSPACE FOR A FAST ALGORITHM.
!
!           REDUCE mat**(t) USING THE QR DECOMPOSITION.
!           SECOND REDUCE R TO UPPER BIDIAGONAL FORM WITH
!           THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!           COMPUTE mat**(t) = Q * R .
!
            call qr_cmp( mat2(:n,:m), s(:m), e(:m) )
#ifdef _BLAS
!
!           COPY R TO work , ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(m*m>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, m
!
                work(1_i4b:i-1_i4b,i) = mat2(1_i4b:i-1_i4b,i)
                work(i,i) = s(i)
                work(i+1_i4b:m,i) = zero
!
            end do
!
!$OMP END PARALLEL DO
!
!           GENERATE Q IN mat2 .
!
            call ortho_gen_qr( mat2(:n,:m), e(:m) )
!
!           BIDIAGONALIZE R IN work(:m,:m) WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
            call bd_cmp2_a( work(:m,:m), s(:m), e(:m), u_v(:m,:m), failure=failure_bd, reortho=reortho )
!
!           PERFORM BIDIAGONAL QR ITERATION, COMPUTING m LEFT
!           SINGULAR VECTORS OF R IN work(:m,:m) AND m RIGHT SINGULAR
!           VECTORS OF R IN u_v (THESE SINGULAR VECTORS ARE ALSO
!           THE LEFT SINGULAR VECTORS OF mat).
!
            call bd_svd_a( true, s, e, failure, work(:m,:m), u_v(:m,:m), sort=sort,   &
                           maxiter=maxiter, max_francis_steps=max_francis_steps,      &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!           MULTIPLY THE TRANSPOSE OF THE LEFT SINGULAR VECTORS OF R IN work(:m,:m)
!           BY THE TRANSPOSE OF Q IN mat2 TO GET THE RIGHT SINGULAR VECTORS OF mat
!           STORED ROW-WISE.

            call gemm( 'T', 'T', m, n, m, one, work(1_i4b:m,1_i4b:m), m,        &
                       mat2(1_i4b:n,1_i4b:m), n, zero, mat(1_i4b:m,1_i4b:n), m  )
#else
!
!           COPY R TO mat , ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(m*m>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, m
!
                mat(1_i4b:i-1_i4b,i) = mat2(1_i4b:i-1_i4b,i)
                mat(i,i) = s(i)
                mat(i+1_i4b:m,i) = zero
!
            end do
!
!$OMP END PARALLEL DO
!
!           GENERATE Q IN mat2 .
!
            call ortho_gen_qr( mat2(:n,:m), e(:m) )
!
!           BIDIAGONALIZE R IN mat(:m,:m) WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
            call bd_cmp2_a( mat(:m,:m), s(:m), e(:m), u_v(:m,:m), failure=failure_bd, reortho=reortho )
!
!           PERFORM BIDIAGONAL QR ITERATION, COMPUTING m LEFT
!           SINGULAR VECTORS OF R IN mat(:m,:m) AND m RIGHT SINGULAR
!           VECTORS OF R IN u_v (THESE SINGULAR VECTORS ARE ALSO
!           THE LEFT SINGULAR VECTORS OF mat).
!
            call bd_svd_a( true, s, e, failure, mat(:m,:m), u_v(:m,:m), sort=sort,    &
                           maxiter=maxiter, max_francis_steps=max_francis_steps,      &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!           MULTIPLY Q IN mat2 BY LEFT SINGULAR VECTORS OF R IN mat(:m,:m)
!           TO GET THE RIGHT SINGULAR VECTORS OF mat STORED COLUMN-WISE.
!
            work(:n,:m) = matmul( mat2(:n,:m), mat(:m,:m) )
!
!           TRANSPOSE THE RIGHT SINGULAR VECTORS OF mat .
!            
            mat(:m,:n) = transpose( work(:n,:m) )
#endif
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( work, mat2 )
!     
        else
!
!           PATH 4 : mat HAS AT LEAST AS MANY COLUMNS AS ROWS, BUT n IS NOT MUCH LARGER THAN m.
!                    BIDIAGONALIZE mat**(t) WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
            call bd_cmp2_a( mat2(:n,:m), s(:k), e(:k), u_v(:k,:k), failure=failure_bd, reortho=reortho )
!
            mat(:m,:n) = transpose( mat2(:n,:m) )
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( mat2 )
!
!           PERFORM BIDIAGONAL QR ITERATION, COMPUTING m LEFT
!           SINGULAR VECTORS OF mat IN u_v AND COMPUTING m RIGHT
!           SINGULAR VECTORS OF mat IN mat STORED ROW-WISE.
!
            call bd_svd2_a( false, s, e, failure, u_v(:k,:k), mat(:m,:n), sort=sort,    &
                            maxiter=maxiter, max_francis_steps=max_francis_steps,       &
                            perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
        end if
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled ) then
!
        call lascl( s, merge(smlnum,bignum,matnorm<smlnum) , matnorm )
!
    end if
!
!
! END OF SUBROUTINE svd_cmp3_a
! ____________________________
!
    end subroutine svd_cmp3_a
!
! =========================================================================================
!
    subroutine svd_cmp4_a( mat, s, failure, v, sort, maxiter,               &
                           max_francis_steps, perfect_shift, bisect, dqds,  &
                           sing_vec, gen_p, reortho, failure_bd, d, e )
!
! Purpose
! _______
!
!   SVD_CMP4 computes the Singular Value Decomposition (SVD) of a real
!   m-by-n matrix MAT with m>=n. The SVD is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix. The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!   The columns of U and V are, respectively, the left and right
!   singular vectors of MAT.
!
!   The routine returns the first n singular values and the associated
!   left and right singular vectors.
!
!   MAT is first reduced to bidiagonal form B with the help of the Ralha-Barlow
!   one-sided bidiagonalization algorithm, see the references (1) and (2).
!
!   The singular values and right singular vectors of B are then computed by the
!   bidiagonal implicit QR algorithm applied to B, see the references (3) and (4).
!
!   The singular vectors of MAT are finally computed by a back transformation algorithm
!   and an orthogonalization step for the left singular vectors.
!
!   Optionally, if the logical argument SING_VEC is used with the value false,
!   matrices the routine computes only the singular values by the dqds algorithm and the
!   orthogonal Q and P used to reduce MAT to bidiagonal form B. This is useful for computing
!   a partial SVD of MAT with subroutines BD_INVITER2 or BD_DEFLATE2 for example.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit:
!
!                       - if SING_VEC=true, MAT is overwritten with the first n
!                         columns of U (the left singular vectors, stored column-wise);
!                       - if SING_VEC=false, MAT is overwritten with the first n columns
!                         of Q (stored column-wise), the orthogonal matrix used to reduce
!                         MAT to bidiagonal form as returned by subroutine BD_CMP2 in
!                         its argument MAT.
!
!                       The shape of MAT must verify: 
!                       size( MAT, 1 ) >= size( MAT, 2 ) = n .
!
!   S                   (OUTPUT) real(stnd), dimension(:)
!                       The singular values of MAT.
!
!                       The size of S must verify: size( S ) = size( MAT, 2 ) = n .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit.
!                       - FAILURE = true  :  indicates that the bidiagonal SVD or dqds algorithms
!                         did not converge and that full accuracy was not attained in the
!                         bidiagonal SVD of an intermediate bidiagonal form B of MAT.
!
!   V                   (OUTPUT) real(stnd), dimension(:,:)
!                       On exit:
!
!                       - if SING_VEC=true, V is overwritten with the n-by-n orthogonal matrix V
!                         (the right singular vectors, stored column-wise);
!                       - if SING_VEC=false, V is overwritten with the n-by-n orthogonal matrix P
!                         (stored column-wise or in packed form), the orthogonal
!                         matrix used to reduce MAT to bidiagonal form as returned
!                         by subroutine BD_CMP2 in its argument P.
!
!                       The shape of V must verify: 
!                       size( V, 1 ) = size( V, 2 ) = n .
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the singular values into ascending order if SORT = 'A' or 'a',
!                       or into descending order if SORT = 'D' or 'd'. The singular vectors
!                       are rearranged accordingly.
!
!                       The default is to sort the singular values and vectors into descending order.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm.
!
!                       The bidiagonal SVD algorithm of an intermediate bidiagonal 
!                       form B of MAT fails to converge if the number of QR sweeps 
!                       exceeds MAXITER * min(m,n). Convergence usually occurs in 
!                       about 2 * min(m,n) QR sweeps.
!
!                       The default is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       On entry, this optional argument has an effect only if
!                       the optional argument SING_VEC has the value true.
!
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm.
!
!                       MAX_FRANCIS_STEPS is a strictly positive integer, otherwise 
!                       the default value is used.
!
!                       The default is the minimum of n and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       On entry, this optional argument has an effect only if
!                       the optional argument SING_VEC has the value true.
!
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm applied to the associated symmetric tridiagonal
!                       matrix B * B'  whose eigenvalues are the squares of the singular values of B or MAT.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm applied to the associated symmetric tridiagonal
!                       matrix B * B'  whose eigenvalues are the squares of the singular values of B or MAT.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   SING_VEC            (INPUT, OPTIONAL) logical(lgl)
!                       On entry:
!
!                       - if SING_VEC=true, the routine computes the singular
!                         values and vectors of MAT.
!                       - If SING_VEC=false the routine computes only the singular values of
!                         MAT and the orthogonal matrices Q and P used to reduce MAT to upper
!                         bidiagonal form as returned by subroutine BD_CMP2.
!                         See the description of BD_CMP2 subroutine for more details.
!
!                       The default is true.
!
!   GEN_P               (INPUT, OPTIONAL) logical(lgl)
!                       On entry, this optional argument has an effect only if the optional argument
!                       SING_VEC is also used with the value false.
!
!                       In this case, if the optional argument GEN_P is used and is set to true, the
!                       orthogonal matrix P used to reduce MAT to bidiagonal form is generated on output
!                       of the subroutine in its argument V.
!
!                       If this argument is set to false, the orthogonal matrix P is stored in factored
!                       form as products of elementary reflectors in the lower triangle of the array V.
!
!                       See the description of BD_CMP2 subroutine for more details.
!
!                       The default is true.
!
!   REORTHO             (INPUT, OPTIONAL) logical(lgl)
!                       On entry, this optional argument has an effect only if the optional argument
!                       SING_VEC is also used with the value false.
!
!                       In this case, if the optional argument REORTHO is set to true, reorthogonalization
!                       is performed (when needed) in the Ralha-Barlow one-sided bidiagonalization of MAT.
!                       If this optional logical argument is set to false, reorthogonalization is never
!                       performed, which can result in a loss of orthogonality if MAT has a large condition
!                       number.
!
!                       See description of the BD_CMP2 subroutine for further details.
!
!                       The default is REORTHO = true.
!
!   FAILURE_BD          (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit:
!
!                       - FAILURE_BD = false :  indicates that maximum accuracy was obtained in the
!                         Ralha-Barlow one-sided bidiagonalization of MAT;
!                       - FAILURE_BD = true  :  indicates that MAT is nearly singular and some loss
!                         of orthogonality can be expected in the left orthogonal matrix in the
!                         bidiagonal reduction of MAT obtained by the one-sided Rhala-Barlow algorithm
!                         if SING_VEC=false is used in the call to SVD_CMP4.
!
!                       See description of the BD_CMP2 subroutine for further details.
!
!   D                   (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                       The diagonal elements of the intermediate upper bidiagonal matrix B.
!
!                       The size of D must be size( D ) = size( MAT, 2 ) = n .
!
!   E                   (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                       The off-diagonal elements of the intermediate upper bidiagonal matrix B:
!
!                            E(i) = B(i-1,i) for i = 2,3,...,n;
!
!                       E(1) is arbitrary.
!
!                       The size of E must be size( E ) = size( MAT, 2 ) = n .
!                
!
! Further Details
! _______________
!
!   Computing the SVD of a m-by-n matrix MAT, with m>=n, in subroutine SVD_CMP4 consists of four steps:
!
!       1) reduction of the rectangular matrix to bidiagonal form B via the Ralha-Barlow
!          one-sided bidiagonalization algorithm, see the references (1) and (2);
!       2) in place accumulation of the right orthogonal transformations used in the reduction
!          of MAT to bidiagonal form B, see the references (3) and (4);
!       3) computation of the singular values and right singular vectors of MAT by applying the implicit QR
!          algorithm to the bidiagonal matrix B, see the references (3) and (4);
!       4) computation and orthogonalization of the left singular vectors in the SVD of MAT to avoid
!          the possible loss of orthogonality of the left orthogonal matrix in the bidiagonal factorization
!          of MAT computed by the one-sided bidiagonalization algorithm, see the references (3) and (4).
!
!   This four-step algorithm computes numerically orthogonal left singular vectors of MAT even in
!   cases of large condition number of MAT despite that the left (orthogonal) matrix in the bidiagonal
!   decomposition of MAT computed by the Ralha-Barlow one-sided bidiagonalization algorithm may not
!   be numerically orthogonal if MAT is nearly singular or has a very large condition number (see
!   references (1) and (2) for details).
!
!   If singular vectors are requested (e.g., if the optional logical argument SING_VEC
!   is not used or is used and set to true) and max(m,n) is much larger than min(m,n),
!   the rectangular matrix is first reduced to upper triangular form by a QR factorization
!   and the above four-step algorithm is applied to the resulting triangular factor. The left
!   singular vectors of the rectangular matrix are then obtained from those of the triangular
!   factor by a back-transformation algorithm.
!
!   For further details on the SVD of a rectangular matrix and the different algorithms to
!   compute it, see the references below.

!   The four or five steps of the SVD algorithm used here (e.g., preliminary QR or LQ factorization,
!   the reduction to bidiagonal form B, in place accumulation of the orthogonal transformations, computation
!   of the SVD of the bidiagonal matrix B and computation/orthogonalization of the left singular vectors in
!   the SVD of MAT) are parallelized if OPENMP is used.
!
!   Optionally, the intermediate bidiagonal decomposition of MAT can be output by the subroutine
!   if the optional logical argument SING_VEC is used with the value false and the optional arguments
!   D and E are also specified. Note, however, that in that case the left (orthogonal) matrix in the
!   bidiagonal decomposition of MAT (computed by the Ralha-Barlow one-sided bidiagonalization algorithm)
!   may not be numerically orthogonal if MAT is nearly singular or has a very large condition number (see
!   references (1) and (2) for details). To correct partly this deficiency, partial reorthogonalization
!   can be performed to ensure orthogonality of this left matrix at the expense of speed of computation.
!   The reorthogonalization uses the Gram-Schmidt method described in the reference (5) and
!   is activated if the optional logical argument REORTHO is set to true.
!
!   When the optional logical argument SING_VEC is used with the value false, the singular values of the
!   bidiagonal form B of MAT are computed by the more accurate and faster dqds algorithm. See references
!   (6) and (7) for a description of the dqds algorithm.
!
!   For more details, see:
!
!   (1) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!            A new stable bidiagonal reduction algorithm.
!            Linear Algebra Appl., No 397, pp. 35-84.
!
!   (2) Bosner, N., and Barlow, J.L., 2007:
!            Block and Parallel versions of one-sided bidiagonalization.
!            SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!            Matrix Computations.
!            3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (4) Lawson, C.L., and Hanson, R.J., 1974:
!            Solving least square problems.
!            Prentice-Hall.
!
!   (5) Stewart, G.W., 2007:
!           Block Gram-Schmidt Orthogonalization.
!           Report TR-4823, Department of Computer Science, College Park, University of Maryland.
!
!   (6) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (7) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error, svd_error5
    use String_Procedures, only : to_lower
    use Utilities,         only : assert, assert_eq, lascl, merror
    use QR_Procedures,     only : qr_cmp, ortho_gen_qr
#ifdef _BLAS
    use BLAS_interfaces,   only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:)         :: mat
    real(stnd), intent(out),   dimension(:)           :: s
    real(stnd), intent(out),   dimension(:,:)         :: v
    real(stnd), intent(out),   dimension(:), optional :: d, e
!
    integer(i4b), intent(in), optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: perfect_shift, bisect, dqds, sing_vec, gen_p, reortho
    logical(lgl), intent(out), optional :: failure_bd
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension( size(s) )        :: e2
    real(stnd), dimension(:),   allocatable :: d2
    real(stnd), dimension(:,:), allocatable :: mat2, work
    real(stnd)                              :: matnorm, tmp
!
    integer(i4b) :: i, m, n, mnthr, info
    integer      :: iok
!
    logical(lgl) :: mat_is_scaled, do_sing_vec, do_gen_p
!
    character    :: sort_def
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='svd_cmp4'
!
    character,  parameter :: a_min=achar(97), d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
    if ( present( failure_bd ) ) then
!
        failure_bd = false
!
    end if
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( (/ int(size( mat, 2 ),i4b)    ,  &
                      int(size( v,   1 ),i4b)    ,  &
                      int(size( v,   2 ),i4b)    ,  &
                      int(size( s )     ,i4b) /) ,  &
                      name_proc )
!
!   CHECK DIMENSIONS OF THE INPUT MATRIX.
!
    if ( m<n ) then
!
        call merror( name_proc//svd_error5 )
!
    end if
!
!   TEST OPTIONAL ARGUMENTS.
!
    do_sing_vec = true
!
    if ( present( sing_vec ) ) then
!
        do_sing_vec = sing_vec
!
    end if
!
    do_gen_p = true
!
    if ( .not.do_sing_vec .and. present( gen_p ) ) then
!
        do_gen_p = gen_p
!
    end if
!
    sort_def = d_min
!
    if ( present( sort ) ) then
!
        sort_def = to_lower( sort )
!
    end if
!
    if ( present(d) ) then
!
        call assert( logical( n==size(d), lgl ), name_proc )
!
    end if
!
    if ( present(e) ) then
!
        call assert( logical( n==size(e), lgl ), name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
    do i = 1_i4b, n
!
        tmp = maxval( abs( mat(1_i4b:m,i) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       mat IS A ZERO MATRIX.
!
        s(1_i4b:n) = zero
!
        do i = 1_i4b, n
            mat(i,i)     = one
            v(1_i4b:n,i) = zero
            v(i,i)       = one
        end do
!
        if ( n/=1_i4b .and. .not.do_sing_vec .and. .not.do_gen_p ) then
            v(1_i4b,1_i4b) = -one
        end if
!
        if ( present(d) ) then
            d(:n) = zero
        end if
!
        if ( present(e) ) then
            e(:n) = zero
        end if
!
        return
!
    else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, bignum )
!
    end if
!
    if ( do_sing_vec ) then
!
!       mat HAS AT LEAST AS MANY ROWS AS COLUMNS.
!       n LEFT SINGULAR VECTORS TO BE OVERWRITTEN ON mat
!       AND n RIGHT SINGULAR VECTORS TO BE COMPUTED IN v .
!
!       ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
        allocate( mat2(m,n), d2(n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE FIRST A COPY OF mat .
!
        mat2(:m,:n) = mat(:m,:n)
!
!       DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!       mat TO BIDIAGONAL FORM. IF m EXCEEDS mnthr AND LEFT
!       SINGULAR VECTORS ARE REQUESTED, A QR FACTORIZATION
!       IS USED FIRST TO REDUCE THE m-BY-n MATRIX mat
!       TO A TRIANGULAR FORM.
!    
        mnthr = int( real( n, stnd )*thr_bd2, i4b )
!
        if ( m>=mnthr ) then
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM (PATH 1).
!
            allocate( work(n,n), stat = iok )
!
        else
!
!           USE THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!           WITHOUT A PRELIMINARY QR FACTORIZATION (PATH 2).
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!
!           PATH 1 : m MUCH LARGER THAN n.
!
!           REDUCE mat USING THE QR DECOMPOSITION.
!           SECOND REDUCE R TO UPPER BIDIAGONAL FORM WITH
!           THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!           COMPUTE mat = Q * R .
!
            call qr_cmp( mat2(:m,:n), s(:n), e2(:n) )
!
!           COPY R TO v AND mat, ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, n
 !
                v(1_i4b:i-1_i4b,i) = mat2(1_i4b:i-1_i4b,i)
                v(i,i) = s(i)
                v(i+1_i4b:n,i) = zero
!
                mat(1_i4b:i,i) = v(1_i4b:i,i)
                mat(i+1_i4b:n,i) = zero
!
            end do
!
!$OMP END PARALLEL DO
!
!           GENERATE Q IN mat2 .
!
            call ortho_gen_qr( mat2(:m,:n), e2(:n) )
!
!           BIDIAGONALIZE R STORED IN v WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
            call bd_cmp3( v(:n,:n), s(:n), e2(:n), gen_p=true, failure=failure_bd )
!
!           OUTPUT THE INTERMEDIATE BIDIAGONAL MATRIX IF REQUIRED.
!
            if ( present(d) ) then
!
                d(:n) = s(:n)
!
            end if
!
            if ( present(e) ) then
!
                e(:n) = e2(:n)
!
            end if
!
!           PERFORM BIDIAGONAL QR ITERATION AFTER BIDIAGONALIZATION OF R,
!           COMPUTING n RIGHT SINGULAR VECTORS OF R AND mat IN v .
!
            call bd_svd_b( false, s(:n), e2(:n), failure, v(:n,:n), sort=d_min,  &
                           maxiter=maxiter, max_francis_steps=max_francis_steps, &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!           MULTIPLY R IN mat BY RIGHT SINGULAR VECTORS OF R IN v, TO GET UNNORMALIZED
!           LEFT SINGULAR VECTORS OF R, STORING RESULT IN work .
!
#ifdef _BLAS
            call gemm( 'N', 'N', n, n, n, one, mat, m, v, n, zero, work, n  )
#else
            work(:n,:n) = matmul( mat(:n,:n), v(:n,:n) )
#endif
!
!           ORHOGONALIZE AND NORMALIZE LEFT SINGULAR VECTORS OF R WITH A QR FACTORIZATION.
!
            call qr_cmp( work(:n,:n), d2(:n), e2(:n) )
!
!           GENERATE LEFT SINGULAR VECTORS OF R IN work .
!
            call ortho_gen_qr( work(:n,:n), e2(:n) )
!
!           ADJUST THE SIGN OF THE LEFT SINGULAR VECTORS OF R .
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, n
!
                if ( d2(i)<zero ) then
!
                    work(:n,i) = -work(:n,i)
!        
                end if
!        
            end do
!
!$OMP END PARALLEL DO
!
!           SORT SINGULAR VALUES AND VECTORS OF R IN ASCENDING ORDER IF REQUIRED.
!
            if ( sort_def==a_min ) then
!
                call lr_singvec_sort( a_min, s, work, v )
!
            end if
!
!           MULTIPLY Q IN mat2 BY LEFT SINGULAR VECTORS OF R IN
!           work, TO GET LEFT SINGULAR VECTORS OF mat, STORING
!           RESULT IN mat .
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, n, n, one, mat2, m, work, n, zero, mat, m  )
#else
            mat(:m,:n) = matmul( mat2(:m,:n), work(:n,:n) )
#endif
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( mat2, d2, work )
!     
        else
!
!           PATH 2 : mat HAS AT LEAST AS MANY ROWS AS COLUMNS, BUT m IS NOT MUCH LARGER THAN n.
!                    BIDIAGONALIZE mat WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!                    DIRECTLY.
!
            call bd_cmp3( mat(:m,:n), s(:n), e2(:n), gen_p=true, failure=failure_bd )
!
            v(:n,:n) = mat(:n,:n)
!
!           OUTPUT THE INTERMEDIATE BIDIAGONAL MATRIX IF REQUIRED.
!
            if ( present(d) ) then
!
                d(:n) = s(:n)
!
            end if
!
            if ( present(e) ) then
!
                e(:n) = e2(:n)
!
            end if
!
!           PERFORM BIDIAGONAL QR ITERATION AFTER BIDIAGONALIZATION OF R OR mat,
!           COMPUTING n RIGHT SINGULAR VECTORS OF R OR mat IN v .
!
            call bd_svd_b( false, s(:n), e2(:n), failure, v(:n,:n), sort=d_min,  &
                           maxiter=maxiter, max_francis_steps=max_francis_steps, &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!           MULTIPLY mat IN mat2 BY RIGHT SINGULAR VECTORS OF mat (OR R) IN
!           v, TO GET UNNORMALIZED LEFT SINGULAR VECTORS OF mat,
!           STORING RESULT IN mat .
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, n, n, one, mat2, m, v, n, zero, mat, m )
#else
            mat(:m,:n) = matmul( mat2(:m,:n), v(:n,:n) )
#endif
!
!           ORHOGONALIZE AND NORMALIZE LEFT SINGULAR VECTORS OF mat WITH A QR FACTORIZATION.
!
            call qr_cmp( mat(:m,:n), d2(:n), e2(:n) )
!
!           GENERATE LEFT SINGULAR VECTORS OF mat IN mat .
!
            call ortho_gen_qr( mat(:m,:n), e2(:n) )
!
!           ADJUST THE SIGN OF THE LEFT SINGULAR VECTORS OF mat .
!
!$OMP PARALLEL DO IF(m*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, n
!
                if ( d2(i)<zero ) then
!
                    mat(:m,i) = -mat(:m,i)
!        
                end if
!        
            end do
!
!$OMP END PARALLEL DO
!
!           SORT SINGULAR VALUES AND VECTORS OF mat IN ASCENDING ORDER IF REQUIRED.
!
            if ( sort_def==a_min ) then
!
                call lr_singvec_sort( a_min, s, mat, v )
!
            end if
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( mat2, d2 )
!
        end if
!
    else
!
!       PATH 3 : mat HAS AT LEAST AS MANY ROWS AS COLUMNS.
!       SINGULAR VECTORS ARE NOT REQUESTED, COMPUTE ONLY SINGULAR VALUES.
!
!       BIDIAGONALIZE mat WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM DIRECTLY.
!
        call bd_cmp2_a( mat(:m,:n), s(:n), e2(:n), v(:n,:n), failure=failure_bd,  &
                        gen_p=do_gen_p, reortho=reortho )
!
!       OUTPUT THE INTERMEDIATE BIDIAGONAL MATRIX IF REQUIRED.
!
        if ( present(d) ) then
!
            d(:n) = s(:n)
!
        end if
!
        if ( present(e) ) then
!
            e(:n) = e2(:n)
!
        end if
!
!       COMPUTE SINGULAR VALUES BY THE DQDS ALGORITHM.
!
        call bd_lasq1( s(:n), e2(:n), failure, maxiter=60_i4b, sort=sort_def, info=info )
!
        if ( info==2_i4b ) then
!
            call bd_dqds( s(:n), e2(:n), failure, maxiter=60_i4b, sort=sort_def )
!
        end if
!
!       PERFORM BIDIAGONAL QR ITERATIONS FOR COMPUTING THE SINGULAR VALUES.
!
!        call bd_svd_t( true, s(:n), e2(:n), failure, sort=sort_def, maxiter=maxiter )
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled ) then
!
        call lascl( s, merge(smlnum,bignum,matnorm<smlnum), matnorm )
!
    end if
!
!
! END OF SUBROUTINE svd_cmp4_a
! ____________________________
!
    end subroutine svd_cmp4_a
!
! =========================================================================================
!
    subroutine svd_cmp3_b( mat, s, failure, sort, maxiter, bisect, dqds, save_mat, &
                           reortho, failure_bd )
!
! Purpose
! _______
!
!   SVD_CMP3 computes the singular values of a real m-by-n matrix MAT.
!   The singular value decomposition (SVD) is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix. The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!   The columns of U and V are, respectively, the left and right
!   singular vectors of MAT.
!
!   The original matrix MAT is first reduced to upper or lower
!   bidiagonal form B by an orthogonal transformation:
!
!                       Q' * MAT * P = B
!
!   where Q and P are orthogonal. The singular values SIGMA of the bidiagonal
!   matrix B are then computed by the bidiagonal implicit QR algorithm (if BISECT=false
!   and DQDS=false), the bisection method (if BISECT=true) or the dqds algorithm (if DQDS=true)
!   applied to the bidiagonal matrix B.
!
!   The routine returns only the first min(m,n) singular values of MAT.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit,  the m-by-n matrix MAT is destroyed if m>=n
!                       and the optional argument SAVE_MAT is not used with
!                       the value true.
!
!   S                   (OUTPUT) real(stnd), dimension(:)
!                       The singular values of MAT.
!
!                       The size of S must verify: size( S ) = min(m,n) .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit.
!                       - FAILURE = true  :  indicates that the implicit QR, bisection or dqds
!                         algorithm used to compute the singular values of the bidiagonal
!                         form B of the input m-by-n matrix MAT did not converge and that
!                         full accuracy was not attained in the bidiagonal SVD of this
!                         intermediate bidiagonal form B of MAT.
!                         The sign of the incorrect singular values is set to negative.
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the singular values into ascending order if SORT = 'A' or 'a',
!                       or into descending order if SORT = 'D' or 'd'.
!
!                       The default is to sort the singular values into descending order.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal implicit QR phase of the SVD algorithm.
!
!                       The bidiagonal SVD algorithm of an intermediate bidiagonal 
!                       form B of MAT fails to converge if the number of QR sweeps 
!                       exceeds MAXITER * min(m,n). Convergence usually occurs in 
!                       about 2 * min(m,n) QR sweeps.
!
!                       This argument has no effect if BISECT is equal to true.
!
!                       The default is 10.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values of the intermediate min(m,n)-by-min(m,n)
!                       bidiagonal matrix B are computed.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the bidiagonal implicit
!                       QR or dqds algorithm applied to the associated min(m,n)-by-min(m,n) bidiagonal matrix B.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values of the intermediate min(m,n)-by-min(m,n)
!                       bidiagonal matrix B are computed.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition. Moreover, the dqds algorithm is usually also faster than the
!                       bisection or bidiagonal implicit QR algorithms and is the default method
!                       used in SVD_CMP3 for computing singular values of the bidiagonal form B of the
!                       input matrix MAT.
!
!                       If DQDS is set to false, singular values are computed with the bidiagonal implicit
!                       QR or bisection algorithm applied to the associated min(m,n)-by-min(m,n) bidiagonal matrix B.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       dqds algorithm is used.
!
!                       If both optional arguments BISECT and DQDS are specified with the value false, the
!                       bidiagonal implicit QR algorithm is used.
!
!                       The default is true.
!
!   SAVE_MAT            (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if SAVE_MAT is set to true, the m-by-n matrix MAT is
!                       not modified by the routine.
!                
!                       The default is false.
!
!   REORTHO             (INPUT, OPTIONAL) logical(lgl)
!                       If the optional argument REORTHO is set to true, reorthogonalization is performed
!                       (when needed) in the Ralha-Barlow one-sided bidiagonalization of MAT.
!                       If this optional logical argument is set to false, reorthogonalization is never
!                       performed, which can result in a loss of orthogonality for the left (right)
!                       singular vectors of MAT if m>=n (m<n) if MAT has a large condition number. However,
!                       this loss of orthogonality is not detrimental to the accuracy of the singular values.
!
!                       See description of the BD_CMP2 subroutine for further details.
!
!                       The default is REORTHO = false.
!
!   FAILURE_BD          (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit:
!
!                       - FAILURE_BD = false :  indicates that maximum accuracy was obtained in the
!                         Ralha-Barlow one-sided bidiagonalization of MAT.
!                       - FAILURE_BD = true  :  indicates that MAT is nearly singular.
!                
!
! Further Details
! _______________
!
!   Computing the singular values of a rectangular matrix in subroutine SVD_CMP3 consists of two steps:
!
!       1) reduction of the rectangular matrix to bidiagonal form B via the Ralha-Barlow
!          one-sided bidiagonalization algorithm, see the references (1) and (2);
!       2) computation of the singular values of the  min(m,n)-by-min(m,n) bidiagonal matrix B by a
!          bidiagonal implicit QR algorithm (if BISECT=false and DQDS=false), a dqds algorithm (if DQDS=true)
!          or a bisection algorithm (if BISECT=true). See the references (3), (4), (5) and (6) for a description
!          of these different algorithms.
!
!   Note that if max(m,n) is much larger than min(m,n), the rectangular matrix is first reduced to
!   upper or lower triangular form by a QR or LQ factorization and the reduction algorithm
!   is applied to the resulting triangular factor. The singular values of the rectangular matrix
!   are then obtained from those of the triangular factor.
!
!   For further details on the SVD of a rectangular matrix and the algorithms to
!   compute it, see the references below.
!
!   The different steps of the SVD algorithm used here (e.g., preliminary QR or LQ factorization,
!   reduction to bidiagonal form, computation of the singular values of the bidiagonal form) are
!   parallelized if OPENMP is used.
!
!   For more details, see:
!
!   (1) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!            A new stable bidiagonal reduction algorithm.
!            Linear Algebra Appl., No 397, pp. 35-84.
!
!   (2) Bosner, N., and Barlow, J.L., 2007:
!            Block and Parallel versions of one-sided bidiagonalization.
!            SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!            Matrix Computations.
!            3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (4) Lawson, C.L., and Hanson, R.J., 1974:
!            Solving least square problems.
!            Prentice-Hall.
!
!   (5) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (6) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error
    use String_Procedures, only : to_lower
    use Utilities,         only : assert_eq, lascl, merror
    use QR_Procedures,     only : qr_cmp
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:)   :: s
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: bisect, dqds, save_mat, reortho
    logical(lgl), intent(out), optional :: failure_bd
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension( size(s) )        :: d, e
    real(stnd)                              :: matnorm, tmp, abstol
    real(stnd), dimension(:,:), allocatable :: mat2, work
!
    integer(i4b) ::  i, k, m, n, mnthr, info
    integer      ::  iok
!
    logical(lgl) :: mat_is_scaled, do_save_mat, use_bisect, use_dqds, do_reortho
!
    character    :: sort_def
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='svd_cmp3'
!
    character,  parameter :: d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
    if ( present( failure_bd ) ) then
!
        failure_bd = false
!
    end if
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
    i = size( s      )
!
    k = assert_eq( min(m,n), i, name_proc  )
!
    sort_def = d_min
!
    if ( present( sort ) ) then
!
        sort_def = to_lower( sort )
!
    end if
!
    use_bisect = false
!
    if ( present( bisect ) ) then
!
        use_bisect = bisect
!
    end if
!
    use_dqds = true
!
    if ( present(dqds) ) then
!
        use_dqds = dqds
!
    end if
!
    do_reortho = false
!
    if ( present( reortho ) ) then
!
        do_reortho = reortho
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
    do i = 1_i4b, n
!
        tmp = maxval( abs( mat(1_i4b:m,i) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
    if ( matnorm==zero ) then
!
!       mat IS A ZERO MATRIX.
!
        s(1_i4b:k) = zero
!
        return
!
    end if
!
!   SET OPTIONAL ARGUMENT.
!
    do_save_mat = false
!
    if ( present( save_mat ) ) then
!
        do_save_mat = save_mat
!
    end if
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat TO BIDIAGONAL FORM. IF max( m, n ) EXCEEDS mnthr ,
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE
!   m-BY-n MATRIX mat TO A TRIANGULAR FORM.
!    
    mnthr = int( real( k, stnd )*thr_bd2, i4b )
!
    mat_is_scaled = false
!
    if ( m>=n ) then
!
!       mat HAS AT LEAST AS MANY ROWS AS COLUMNS.
!
        if ( do_save_mat ) then
!
!           ALLOCATE WORKSPACE NEEDED FOR STORING mat.
!
            allocate( mat2(m,n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            mat2(:m,:n) = mat(:m,:n)
!
!           SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
            if ( matnorm<smlnum ) then
!
                mat_is_scaled = true
!
                call lascl( mat2, matnorm, smlnum )
!
            else if ( matnorm>bignum ) then
!
                mat_is_scaled = true
!
                call lascl( mat2, matnorm, bignum )
!
            end if
!
            if ( m>=mnthr ) then
!
!               ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM (PATH 1).
!
                allocate( work(n,n), stat = iok )
!
            else
!
!               USE THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!               WITHOUT A PRELIMINARY QR FACTORIZATION (PATH 2).
!
                iok = -1
!
            end if
!
            if ( iok==0 ) then
!
!               PATH 1 : m MUCH LARGER THAN n AND SUFFICIENT WORKSPACE FOR A FAST ALGORITHM.
!
!               REDUCE mat USING THE QR DECOMPOSITION.
!               SECOND REDUCE R TO UPPER BIDIAGONAL FORM WITH
!               THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!               COMPUTE mat = Q * R .
!
                call qr_cmp( mat2(:m,:n), d(:n), e(:n) )
!
!               COPY R TO work , ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
                do i = 1_i4b, n
!
                    work(1_i4b:i-1_i4b,i) = mat2(1_i4b:i-1_i4b,i)
                    work(i,i) = d(i)
                    work(i+1_i4b:n,i) = zero
!
                end do
!
!$OMP END PARALLEL DO
!
!               BIDIAGONALIZE R IN work WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!                call bd_cmp3( work(:n,:n), d(:n), e(:n), gen_p=false, failure=failure_bd )
                call bd_cmp2_b( work(:n,:n), d(:n), e(:n), failure=failure_bd, reortho=do_reortho )
!
!               DEALLOCATE WORKSPACE.
!
                deallocate( work )
!     
            else
!
!               PATH 2 : mat HAS AT LEAST AS MANY ROWS AS COLUMNS, BUT m IS NOT MUCH LARGER THAN n.
!                        BIDIAGONALIZE mat WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!                call bd_cmp3( mat2(:m,:n), d(:n), e(:n), gen_p=false, failure=failure_bd )
                call bd_cmp2_b( mat2(:m,:n), d(:n), e(:n), failure=failure_bd, reortho=do_reortho )
!
            end if
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( mat2 )
!
        else
!
!           SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
            if ( matnorm<smlnum ) then
!
                mat_is_scaled = true
!
                call lascl( mat, matnorm, smlnum )
!
            else if ( matnorm>bignum ) then
!
                mat_is_scaled = true
!
                call lascl( mat, matnorm, bignum )
!
            end if
!
            if ( m>=mnthr ) then
!
!               ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM (PATH 1).
!
                allocate( work(n,n), stat = iok )
!
            else
!
!               USE THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!               WITHOUT A PRELIMINARY QR FACTORIZATION (PATH 2).
!
                iok = -1
!
            end if
!
            if ( iok==0 ) then
!
!               PATH 1 : m MUCH LARGER THAN n AND SUFFICIENT WORKSPACE FOR A FAST ALGORITHM.
!
!               REDUCE mat USING THE QR DECOMPOSITION.
!               SECOND REDUCE R TO UPPER BIDIAGONAL FORM WITH
!               THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!               COMPUTE mat = Q * R .
!
                call qr_cmp( mat(:m,:n), d(:n), e(:n) )
!
!               COPY R TO work , ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
                do i = 1_i4b, n
!
                    work(1_i4b:i-1_i4b,i) = mat(1_i4b:i-1_i4b,i)
                    work(i,i) = d(i)
                    work(i+1_i4b:n,i) = zero
!
                end do
!
!$OMP END PARALLEL DO
!
!               BIDIAGONALIZE R IN work WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!                call bd_cmp3( work(:n,:n), d(:n), e(:n), gen_p=false, failure=failure_bd )
                call bd_cmp2_b( work(:n,:n), d(:n), e(:n), failure=failure_bd, reortho=do_reortho )
!
!               DEALLOCATE WORKSPACE.
!
                deallocate( work )
!     
            else
!
!               PATH 2 : mat HAS AT LEAST AS MANY ROWS AS COLUMNS, BUT m IS NOT MUCH LARGER THAN n.
!                        BIDIAGONALIZE mat WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!                call bd_cmp3( mat(:m,:n), d(:n), e(:n), gen_p=false, failure=failure_bd )
                call bd_cmp2_b( mat(:m,:n), d(:n), e(:n), failure=failure_bd, reortho=do_reortho )
!
            end if
!
        end if
!
    else
!
!       mat HAS MORE COLUMNS THAN ROWS.
!
!       ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM.
!
        allocate( mat2(n,m), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       FIRST TRANSPOSE mat .
!
        mat2(:n,:m) = transpose( mat(:m,:n) )
!
!       SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
        if ( matnorm<smlnum ) then
!
            mat_is_scaled = true
!
            call lascl( mat2, matnorm, smlnum )
!
        else if ( matnorm>bignum ) then
!
            mat_is_scaled = true
!
            call lascl( mat2, matnorm, bignum )
!
        end if
!
        if ( n>=mnthr ) then
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM (PATH 3).
!
            allocate( work(m,m), stat = iok )
!
        else
!
!           USE THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!           WITHOUT A PRELIMINARY QR FACTORIZATION OF mat**(t) (PATH 4).
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!
!           PATH 3 : n MUCH LARGER THAN m AND SUFFICIENT WORKSPACE FOR A FAST ALGORITHM.
!
!           REDUCE mat**(t) USING THE QR DECOMPOSITION.
!           SECOND REDUCE R TO UPPER BIDIAGONAL FORM WITH
!           THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!           COMPUTE mat**(t) = Q * R .
!
            call qr_cmp( mat2(:n,:m), d(:m), e(:m) )
!
!           COPY R TO work , ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(m*m>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, m
!
                work(1_i4b:i-1_i4b,i) = mat2(1_i4b:i-1_i4b,i)
                work(i,i) = d(i)
                work(i+1_i4b:m,i) = zero
!
            end do
!
!$OMP END PARALLEL DO
!
!           BIDIAGONALIZE R IN work WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!            call bd_cmp3( work(:m,:m), d(:m), e(:m), gen_p=false, failure=failure_bd )
            call bd_cmp2_b( work(:m,:m), d(:m), e(:m), failure=failure_bd, reortho=do_reortho )
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( work )
!     
        else
!
!           PATH 4 : mat HAS AT LEAST AS MANY COLUMNS AS ROWS, BUT n IS NOT MUCH LARGER THAN m.
!                    BIDIAGONALIZE mat**(t) WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!            call bd_cmp3( mat2(:n,:m), d(:m), e(:m), gen_p=false, failure=failure_bd )
            call bd_cmp2_b( mat2(:n,:m), d(:m), e(:m), failure=failure_bd, reortho=do_reortho )
!
        end if
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( mat2 )
!
    end if
!
    if ( use_dqds ) then
!
!       COMPUTE SINGULAR VALUES BY THE DQDS ALGORITHM.
!
        call bd_lasq1( d(:k), e(:k), failure, maxiter=60_i4b, sort=sort_def, info=info )
!
        if ( info==2_i4b ) then
!
            call bd_dqds( d(:k), e(:k), failure, maxiter=60_i4b, sort=sort_def )
!
        end if
!
        if ( .not.failure ) then
!
            s(:k) = d(:k)
!
        else
!
            s(:k) = -one
!
        end if
!
    else if ( use_bisect ) then
!
!       COMPUTE SINGULAR VALUES BY A BISECTION ALGORITHM.
!
        abstol = sqrt( safmin )
!
        call bd_singval( d(:k), e(:k), i, s(:k), failure, sort=sort_def, abstol=abstol )
!
    else
!
!       PERFORM BIDIAGONAL QR ITERATIONS FOR COMPUTING THE SINGULAR VALUES.
!
        call bd_svd_t( true, d(:k), e(:k), failure, sort=sort_def, maxiter=maxiter )
!
        s(:k) = d(:k)
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled ) then
!
        call lascl( s(:k), merge(smlnum,bignum,matnorm<smlnum) , matnorm )
!
    end if
!
!
! END OF SUBROUTINE svd_cmp3_b
! ____________________________
!
    end subroutine svd_cmp3_b
!
! =========================================================================================
!
    subroutine svd_cmp5_a( mat, s, failure, v, sort, maxiter,                        &
                          max_francis_steps, perfect_shift, bisect, dqds, failure_bd )
!
! Purpose
! _______
!
!   SVD_CMP5 computes the Singular Value Decomposition (SVD) of a real
!   m-by-n matrix MAT. The SVD is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix. The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!   The columns of U and V are, respectively, the left and right
!   singular vectors of MAT.
!
!   The routine returns the first min(m,n) singular values and the associated
!   left and right singular vectors.
!
!   MAT (or MAT' if m<n) is first reduced to bidiagonal form B with the help of the
!   Ralha-Barlow one-sided bidiagonalization algorithm, see the references (1) and (2).
!
!   The singular values and right (left if m<n) singular vectors of B are then computed
!   by the bidiagonal implicit QR algorithm applied to B, see the references (3) and (4).
!
!   The singular vectors of MAT are finally computed by a back transformation algorithm and 
!   and an orthogonalization step for the left (right if m<n) singular vectors.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit, MAT is overwritten with the first min(m,n)
!                       columns of U, the left singular vectors.
!
!   S                   (OUTPUT) real(stnd), dimension(:)
!                       The singular values of MAT.
!
!                       The size of S must verify: size( S ) = min(m,n) .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit.
!                       - FAILURE = true  :  indicates that the bidiagonal SVD algorithm
!                         did not converge and that full accuracy was not attained in the
!                         bidiagonal SVD of an intermediate bidiagonal form B of MAT.
!
!   V                   (OUTPUT) real(stnd), dimension(:,:)
!                       On exit, V contains the first min(m,n) columns of V,
!                       the right singular vectors.
!
!                       The shape of V must verify: 
!
!                       - size( V, 1 ) = n,
!                       - size( V, 2 ) = min(m,n).
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the singular values into ascending order if SORT = 'A' or 'a',
!                       or into descending order if SORT = 'D' or 'd'. The singular vectors
!                       are rearranged accordingly.
!
!                       The default is to sort the singular values and vectors into descending order.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm.
!
!                       The bidiagonal SVD algorithm of an intermediate bidiagonal 
!                       form B of MAT fails to converge if the number of QR sweeps 
!                       exceeds MAXITER * min(m,n). Convergence usually occurs in 
!                       about 2 * min(m,n) QR sweeps.
!
!                       The default is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm.
!
!                       MAX_FRANCIS_STEPS is a strictly positive integer, otherwise 
!                       the default value is used.
!
!                       The default is the minimum of min(m,n) and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm applied to the associated symmetric tridiagonal
!                       matrix B * B'  whose eigenvalues are the squares of the singular values of B or MAT.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm applied to the associated symmetric tridiagonal
!                       matrix B * B'  whose eigenvalues are the squares of the singular values of B or MAT.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   FAILURE_BD          (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit:
!
!                       - FAILURE_BD = false :  indicates that maximum accuracy was obtained in the
!                         Ralha-Barlow one-sided bidiagonalization of MAT.
!                       - FAILURE_BD = true  :  indicates that MAT is nearly singular.
!                
!
! Further Details
! _______________
!
!   Computing the SVD of a m-by-n matrix MAT, with m>=n, in subroutine SVD_CMP5 consists of four steps:
!
!       1) reduction of the rectangular matrix to bidiagonal form B via the Ralha-Barlow
!          one-sided bidiagonalization algorithm, see the references (1) and (2);
!       2) in place accumulation of the right orthogonal transformations used in the reduction
!          of MAT to bidiagonal form B, see the references (3) and (4);
!       3) computation of the singular values and right singular vectors of MAT by applying the implicit QR
!          algorithm to the bidiagonal matrix B, see the references (3) and (4);
!       4) computation and orthogonalization of the left singular vectors in the SVD of MAT to avoid
!          the possible loss of orthogonality of the left orthogonal matrix in the bidiagonal factorization
!          of MAT computed by the one-sided bidiagonalization algorithm, see the references (3) and (4).
!
!   This four-step algorithm computes numerically orthogonal left singular vectors of a m-by-n matrix MAT,
!   with m>=n, even in cases of large condition number of MAT despite that the left (orthogonal) matrix in the
!   bidiagonal decomposition of MAT computed by the Ralha-Barlow one-sided bidiagonalization algorithm may not
!   be numerically orthogonal if MAT is nearly singular or has a very large condition number (see references (1)
!   and (2) for details).
!
!   If m<n, this four-step algorithm is applied to MAT', instead of MAT, to get the SVD of MAT and it
!   computes also numerically orthogonal right singular vectors of MAT in that case.
!
!   Note that if max(m,n) is much larger than min(m,n), the rectangular matrix is first reduced to
!   upper or lower triangular form by a QR or LQ factorization and the four-step reduction algorithm
!   is applied to the resulting triangular factor. The singular vectors of the rectangular matrix
!   are then obtained from those of the triangular factor in the QR or LQ factorization by a back-transformation
!   algorithm (see the references (3) and (4) for details).
!
!   For further details on the SVD of a rectangular matrix and the algorithms to compute it, see the references below.
!
!   The four or five steps of the SVD algorithm used here (e.g., preliminary QR or LQ factorization,
!   reduction to bidiagonal form B, in place accumulation of the orthogonal transformations, computation
!   of the SVD of the bidiagonal matrix B and reorthogonalization of the left (or right if m<n) singular
!   vectors in the SVD of MAT) are parallelized if OPENMP is used.
!
!   For more details, see:
!
!   (1) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!            A new stable bidiagonal reduction algorithm.
!            Linear Algebra Appl., No 397, pp. 35-84.
!
!   (2) Bosner, N., and Barlow, J.L., 2007:
!            Block and Parallel versions of one-sided bidiagonalization.
!            SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!            Matrix Computations.
!            3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (4) Lawson, C.L., and Hanson, R.J., 1974:
!            Solving least square problems.
!            Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error
    use String_Procedures, only : to_lower
    use Utilities,         only : assert_eq, lascl, merror
    use QR_Procedures,     only : qr_cmp, ortho_gen_qr
#ifdef _BLAS
    use BLAS_interfaces,   only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:)   :: s
    real(stnd), intent(out),   dimension(:,:) :: v
!
    integer(i4b), intent(in), optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: perfect_shift, bisect, dqds
    logical(lgl), intent(out), optional :: failure_bd
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension( size(s) )        :: d, e
    real(stnd)                              :: matnorm, tmp
    real(stnd), dimension(:,:), allocatable :: mat2, work
!
    integer(i4b) ::  i, k, m, n, mnthr
    integer      ::  iok
!
    logical(lgl) :: mat_is_scaled
!
    character    :: sort_def
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='svd_cmp5'
!
    character,  parameter :: a_min=achar(97), d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
    if ( present( failure_bd ) ) then
!
        failure_bd = false
!
    end if
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b), &
                   int(size(v,1),i4b),   &
                   name_proc  )
!
    k = assert_eq( min(m,n),             &
                   int(size(s),i4b),     &
                   int(size(v,2),i4b),   &
                   name_proc  )
!
    sort_def = d_min
!
    if ( present( sort ) ) then
!
        sort_def = to_lower( sort )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
    do i = 1_i4b, n
!
        tmp = maxval( abs( mat(1_i4b:m,i) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       mat IS A ZERO MATRIX.
!
        s(:k) = zero
!
        do i = 1_i4b, k
!
            mat(i,i) = one
!
            v(:n,i)  = zero
            v(i,i)   = one
!
        end do
!
        return
!
    else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat TO BIDIAGONAL FORM. IF max( m, n ) EXCEEDS mnthr ,
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE
!   m-BY-n MATRIX mat TO A TRIANGULAR FORM.
!    
    mnthr = int( real( k, stnd )*thr_bd2, i4b )
!
    if ( m>=n ) then
!
!       mat HAS AT LEAST AS MANY ROWS AS COLUMNS.
!       n LEFT SINGULAR VECTORS TO BE OVERWRITTEN ON mat AND
!       n RIGHT SINGULAR VECTORS TO BE COMPUTED IN v .
!
!       ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
        allocate( mat2(m,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE FIRST A COPY OF mat .
!
        mat2(:m,:n) = mat(:m,:n)
!
        if ( m>=mnthr ) then
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM (PATH 1).
!
            allocate( work(n,n), stat = iok )
!
        else
!
!           USE THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!           WITHOUT A PRELIMINARY QR FACTORIZATION (PATH 2).
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!
!           PATH 1 : m MUCH LARGER THAN n.
!
!           REDUCE mat USING THE QR DECOMPOSITION.
!           SECOND REDUCE R TO UPPER BIDIAGONAL FORM WITH
!           THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!           COMPUTE mat = Q * R .
!
            call qr_cmp( mat2(:m,:n), s(:n), e(:n) )
!
!           COPY R TO v AND mat, ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, n
!
                v(1_i4b:i-1_i4b,i) = mat2(1_i4b:i-1_i4b,i)
                v(i,i) = s(i)
                v(i+1_i4b:n,i) = zero
!
                mat(1_i4b:i,i) = v(1_i4b:i,i)
                mat(i+1_i4b:n,i) = zero
!
            end do
!
!$OMP END PARALLEL DO
!
!           GENERATE Q IN mat2 .
!
            call ortho_gen_qr( mat2(:m,:n), e(:n) )
!
!           BIDIAGONALIZE R STORED IN v WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
            call bd_cmp3( v(:n,:n), s(:n), e(:n), gen_p=true, failure=failure_bd )
!
!           PERFORM BIDIAGONAL QR ITERATION AFTER BIDIAGONALIZATION OF R,
!           COMPUTING n RIGHT SINGULAR VECTORS OF R AND mat IN v .
!
            call bd_svd_b( false, s(:n), e(:n), failure, v(:n,:n), sort=d_min,   &
                           maxiter=maxiter, max_francis_steps=max_francis_steps, &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!           MULTIPLY R STORED IN mat BY RIGHT SINGULAR VECTORS OF R STORED IN v, TO GET UNNORMALIZED
!           LEFT SINGULAR VECTORS OF R, STORING RESULT IN work .
!
#ifdef _BLAS
            call gemm( 'N', 'N', n, n, n, one, mat, m, v, n, zero, work, n  )
#else
            work(:n,:n) = matmul( mat(:n,:n), v(:n,:n) )
#endif
!
!           ORHOGONALIZE AND NORMALIZE LEFT SINGULAR VECTORS OF R WITH A QR FACTORIZATION.
!
            call qr_cmp( work(:n,:n), d(:n), e(:n) )
!
!           GENERATE LEFT SINGULAR VECTORS OF R IN work .
!
            call ortho_gen_qr( work(:n,:n), e(:n) )
!
!           ADJUST THE SIGN OF THE LEFT SINGULAR VECTORS OF R .
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, n
!
                if ( d(i)<zero ) then
!
                    work(:n,i) = -work(:n,i)
!        
                end if
!        
            end do
!
!$OMP END PARALLEL DO
!
!           SORT SINGULAR VALUES AND VECTORS OF R IN ASCENDING ORDER IF REQUIRED.
!
            if ( sort_def==a_min ) then
!
                call lr_singvec_sort( a_min, s, work, v )
!
            end if
!
!           MULTIPLY Q IN mat2 BY LEFT SINGULAR VECTORS OF R STORED IN
!           work, TO GET LEFT SINGULAR VECTORS OF mat, STORING
!           RESULT IN mat .
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, n, n, one, mat2, m, work, n, zero, mat, m  )
#else
            mat(:m,:n) = matmul( mat2(:m,:n), work(:n,:n) )
#endif
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( mat2, work )
!     
        else
!
!           PATH 2 : mat HAS AT LEAST AS MANY ROWS AS COLUMNS, BUT m IS NOT MUCH LARGER THAN n.
!                    BIDIAGONALIZE mat WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!                    DIRECTLY.
!
            call bd_cmp3( mat(:m,:n), s(:n), e(:n), gen_p=true, failure=failure_bd )
!
            v(:n,:n) = mat(:n,:n)
!
!           PERFORM BIDIAGONAL QR ITERATION AFTER BIDIAGONALIZATION OF mat,
!           COMPUTING n RIGHT SINGULAR VECTORS OF mat IN v .
!
            call bd_svd_b( false, s(:n), e(:n), failure, v(:n,:n), sort=d_min,   &
                           maxiter=maxiter, max_francis_steps=max_francis_steps, &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!           MULTIPLY mat STORED IN mat2 BY RIGHT SINGULAR VECTORS OF mat STORED IN v, TO GET
!           UNNORMALIZED LEFT SINGULAR VECTORS OF mat, STORING RESULT IN mat .
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, n, n, one, mat2, m, v, n, zero, mat, m  )
#else
            mat(:m,:n) = matmul( mat2(:m,:n), v(:n,:n) )
#endif
!
!           ORHOGONALIZE AND NORMALIZE LEFT SINGULAR VECTORS OF mat WITH A QR FACTORIZATION.
!
            call qr_cmp( mat(:m,:n), d(:n), e(:n) )
!
!           GENERATE LEFT SINGULAR VECTORS OF mat IN mat .
!
            call ortho_gen_qr( mat(:m,:n), e(:n) )
!
!           ADJUST THE SIGN OF THE LEFT SINGULAR VECTORS OF mat .
!
!$OMP PARALLEL DO IF(m*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, n
!
                if ( d(i)<zero ) then
!
                    mat(:m,i) = -mat(:m,i)
!        
                end if
!        
            end do
!
!$OMP END PARALLEL DO
!
!           SORT SINGULAR VALUES AND VECTORS OF mat IN ASCENDING ORDER IF REQUIRED.
!
            if ( sort_def==a_min ) then
!
                call lr_singvec_sort( a_min, s, mat, v )
!
            end if
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( mat2 )
!
        end if
!
    else
!
!       mat HAS MORE COLUMNS THAN ROWS.
!       m LEFT SINGULAR VECTORS TO BE OVERWRITTEN ON mat AND
!       m RIGHT SINGULAR VECTORS TO BE COMPUTED IN v .
!
!       ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
        allocate( mat2(n,m), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       FIRST TRANSPOSE mat IN mat2 .
!
        mat2(:n,:m) = transpose( mat(:m,:n) )
!
        if ( n>=mnthr ) then
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM (PATH 3).
!
            allocate( work(m,m), stat = iok )
!
        else
!
!           USE THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!           WITHOUT A PRELIMINARY QR FACTORIZATION OF mat**(t) (PATH 4).
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!
!           PATH 3 : n MUCH LARGER THAN m.
!
!           REDUCE mat**(t) USING THE QR DECOMPOSITION.
!           SECOND REDUCE R TO UPPER BIDIAGONAL FORM WITH
!           THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!           COMPUTE mat**(t) = Q * R .
!
            call qr_cmp( mat2(:n,:m), s(:m), e(:m) )
!
!           COPY R TO v AND mat, ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(m*m>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, m
!
                v(1_i4b:i-1_i4b,i) = mat2(1_i4b:i-1_i4b,i)
                v(i,i) = s(i)
                v(i+1_i4b:m,i) = zero
!
                mat(1_i4b:i,i) = v(1_i4b:i,i)
                mat(i+1_i4b:m,i) = zero
!
            end do
!
!$OMP END PARALLEL DO
!
!           GENERATE Q IN mat2 .
!
            call ortho_gen_qr( mat2(:n,:m), e(:m) )
!
!           BIDIAGONALIZE R STORED IN mat WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
            call bd_cmp3( mat(:m,:m), s(:m), e(:m), gen_p=true, failure=failure_bd )
!
!           PERFORM BIDIAGONAL QR ITERATION AFTER BIDIAGONALIZATION OF R ,
!           COMPUTING m RIGHT SINGULAR VECTORS OF R AND mat**(t) IN mat .
!
            call bd_svd_b( false, s(:m), e(:m), failure, mat(:m,:m), sort=d_min, &
                           maxiter=maxiter, max_francis_steps=max_francis_steps, &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!           MULTIPLY R STORED IN v BY RIGHT SINGULAR VECTORS OF R STORED IN mat, TO GET
!           UNNORMALIZED LEFT SINGULAR VECTORS OF R, STORING RESULT IN work .
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, m, m, one, v, n, mat, m, zero, work, m )
#else
            work(:m,:m) = matmul( v(:m,:m), mat(:m,:m) )
#endif
!
!           ORHOGONALIZE AND NORMALIZE LEFT SINGULAR VECTORS OF R WITH A QR FACTORIZATION.
!
            call qr_cmp( work(:m,:m), d(:m), e(:m) )
!
!           GENERATE LEFT SINGULAR VECTORS OF R IN work .
!
            call ortho_gen_qr( work(:m,:m), e(:m) )
!
!           ADJUST THE SIGN OF THE LEFT SINGULAR VECTORS OF R .
!
!$OMP PARALLEL DO IF(m*m>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, m
!
                if ( d(i)<zero ) then
!
                    work(:m,i) = -work(:m,i)
!        
                end if
!        
            end do
!
!$OMP END PARALLEL DO
!
!           SORT SINGULAR VALUES AND VECTORS OF R IN ASCENDING ORDER IF REQUIRED.
!
            if ( sort_def==a_min ) then
!
                call lr_singvec_sort( a_min, s, work, mat(:m,:m) )
!
            end if
!
!           MULTIPLY Q STORED IN mat2 BY THE LEFT SINGULAR VECTORS OF R STORED IN work
!           TO GET THE RIGHT SINGULAR VECTORS OF mat STORED COLUMN-WISE, STORING THE RESULTS IN v .

#ifdef _BLAS
            call gemm( 'N', 'N', n, m, m, one, mat2, n, work, m, zero, v, n  )
#else
            v(:n,:m) = matmul( mat2(:n,:m), work(:m,:m) )
#endif
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( mat2, work )
!
        else
!
!           PATH 4 : mat HAS AT LEAST AS MANY COLUMNS AS ROWS, BUT n IS NOT MUCH LARGER THAN m.
!                    BIDIAGONALIZE mat**(t) WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!                    DIRECTLY.
!
!           SAVE mat**(t) IN v .
!
            v(:n,:m) = mat2(:n,:m)
!
!           BIDIAGONALIZE mat**(t) WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
            call bd_cmp3( v(:n,:m), s(:m), e(:m), gen_p=true, failure=failure_bd )
!
            mat(:m,:m) = v(:m,:m)
!
!           PERFORM BIDIAGONAL QR ITERATION AFTER BIDIAGONALIZATION mat**(t),
!           COMPUTING m RIGHT SINGULAR VECTORS OF mat**(t) IN mat .
!
            call bd_svd_b( false, s(:m), e(:m), failure, mat(:m,:m), sort=d_min, &
                           maxiter=maxiter, max_francis_steps=max_francis_steps, &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!           MULTIPLY mat**(t) STORED IN mat2 BY RIGHT SINGULAR VECTORS OF mat**(t) STORED IN
!           mat , TO GET UNNORMALIZED LEFT SINGULAR VECTORS OF mat**(t), STORING RESULT IN v .
!
#ifdef _BLAS
            call gemm( 'N', 'N', n, m, m, one, mat2, n, mat, m, zero, v, n  )
#else
            v(:n,:m) = matmul( mat2(:n,:m), mat(:m,:m) )
#endif
!
!           ORHOGONALIZE AND NORMALIZE LEFT SINGULAR VECTORS OF mat**(t) WITH A QR FACTORIZATION.
!
            call qr_cmp( v(:n,:m), d(:m), e(:m) )
!
!           GENERATE LEFT SINGULAR VECTORS OF mat**(t) IN mat2 .
!
            call ortho_gen_qr( v(:n,:m), e(:m) )
!
!           ADJUST THE SIGN OF THE LEFT SINGULAR VECTORS OF mat**(t) .
!
!$OMP PARALLEL DO IF(m*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, m
!
                if ( d(i)<zero ) then
!
                    v(:n,i) = -v(:n,i)
!        
                end if
!        
            end do
!
!$OMP END PARALLEL DO
!
!           SORT SINGULAR VALUES AND VECTORS OF mat IN ASCENDING ORDER IF REQUIRED.
!
            if ( sort_def==a_min ) then
!
                call lr_singvec_sort( a_min, s, mat(:m,:m), v )
!
            end if
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( mat2 )
!
        end if
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled ) then
!
        call lascl( s, merge(smlnum,bignum,matnorm<smlnum), matnorm )
!
    end if
!
!
! END OF SUBROUTINE svd_cmp5_a
! ____________________________
!
    end subroutine svd_cmp5_a
!
! =========================================================================================
!
    subroutine svd_cmp6( mat, s, v, failure, sort, nsvd, maxiter, ortho, backward_sweep, &
                         scaling, initvec, failure_bd, failure_bisect )
!
! Purpose
! _______
!
!   SVD_CMP6 computes a full or partial Singular Value Decomposition (SVD) of a real
!   m-by-n matrix MAT. The full SVD is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix. The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!   The columns of U and V are, respectively, the left and right
!   singular vectors of MAT.
!
!   The routine can return the first min(m,n) singular values and the associated
!   left and right singular vectors or a truncated SVD if the optional integer parameter
!   NSVD is used in the call to SVD_CMP6.
!
!   MAT (or MAT' if m<n) is first reduced to bidiagonal form B with the help of the Ralha-Barlow
!   one-sided bidiagonalization algorithm without reorthogonalisation, see the references (1) and (2).
!
!   The singular values and right (left if m<n) singular vectors of B are then computed
!   by the bisection and inverse iteration methods applied to B and the tridiagonal
!   matrix B' * B, respectively, see the reference (3).
!
!   The singular vectors of MAT are finally computed by a back transformation algorithm
!   and an orthogonalization step for the left (right if m<n) singular vectors.
!
!
! Arguments
! _________
!
!   MAT            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                  On entry, the m-by-n matrix MAT.
!
!                  On exit, MAT is overwritten with the first nsvd columns of U, e.g., the left
!                  singular vectors associated with the first nsvd largest singular values of MAT.
!
!   S              (OUTPUT) real(stnd), dimension(:), pointer
!                  On exit, S(:) contains estimates of the first nsvd largest singular values of MAT.
!                  The singular values are given in decreasing order and are positive or zero.
!
!                  The statut of the pointer S must not be undefined on entry.
!                  If, on entry, the pointer S is already allocated, it will be first deallocated
!                  and then reallocated with the correct size.
!
!                  On exit, the size of the pointer S will verify:
!
!                  - size( S ) = nsvd <= min( size(MAT,1) , size(MAT,2) ) = min(m,n) .
!
!   V              (OUTPUT) real(stnd), dimension(:,:), pointer
!                  On exit, the computed first nsvd columns of V, e.g., the right singular vectors
!                  associated with the first nsvd largest singular values of MAT.
!
!                  The right singular vector associated with the singular value S(j) is stored
!                  in the j-th column of V.
!
!                  The statut of the pointer V must not be undefined on entry.
!                  If, on entry, the pointer V is already allocated, it will be first deallocated
!                  and then reallocated with the correct shape.
!
!                  On exit, the shape of the pointer V will verify:  
!
!                  - size( V, 1 ) = size( MAT, 2 ) = n ,
!                  - size( V, 2 ) = size( S )      = nsvd .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false indicates successful exit.
!                  - FAILURE = true indicates that some singular vectors
!                    failed to converge in MAXITER inverse iterations.
!
!   SORT           (INPUT, OPTIONAL) character
!                  Sort the singular values into ascending order if SORT = 'A' or 'a',
!                  or into descending order if SORT = 'D' or 'd'. The singular vectors
!                  are rearranged accordingly.
!
!                  The default is to sort the singular values and vectors into descending order.
!
!   NSVD           (INPUT/OUTPUT, OPTIONAL) integer(i4b)
!                  On entry, NSVD specifies the number of the top singular triplets which are
!                  requested.
!
!                  On exit, NSVD is the number of singular triplets which have been
!                  computed by the subroutine, which can be greater than the requested
!                  number if multiple singular values at index NSVD make unique selection
!                  impossible.
!
!                  On entry, NSVD must be greater than 0 and less or equal to min(m,n).
!
!                  The default is NSVD = min( size(MAT,1) , size(MAT,2) ) = min(m,n).
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed for computing singular vectors.
!
!                  By default, 2 inverse iterations are performed for all the singular vectors.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, all the singular vectors are orthogonalized by the
!                    Modified Gram-Schmidt or QR algorithm in the inverse iteration algorithm;
!                  - ORTHO=false, the singular vectors are not orthogonalized by 
!                    the Modified Gram-Schmidt or QR algorithm in the inverse iteration algorithm.
!
!                  The default is to orthogonalize the singular vectors only for the singular values,
!                  which are not well-separated.
!
!   BACKWARD_SWEEP (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - BACKWARD_SWEEP=true and the singular vectors are orthogonalized by
!                    the modified Gram-Schmidt algorithm in the inverse iteration algorithm,
!                    a backward sweep of the modified Gram-Schmidt algorithm is also performed;
!                  - BACKWARD_SWEEP=false a backward sweep is not performed.
!
!                  The default is not to perform a backward sweep of the modified Gram-Schmidt
!                  algorithm.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the bidiagonal matrix B and tridiagonal matrix B' * B are scaled
!                    before computing the singular values and vectors, respectively;
!                  - SCALING=false, the bidiagonal matrix B and tridiagonal matrix B' * B are not scaled.
!
!                  The default is to scale the bidiagonal and tridiagonal matrices.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INITVEC=true, Fernando vectors are used to start the inverse
!                    iteration process for computing the singular vectors;
!                  - INITVEC=false, random uniform starting vectors are used.
!
!                  For unreduced tridiagonal matrices B' * B, the default is to use Fernando starting vectors
!                  if the eigenvalues (e.g., the squares of the singular values) are well-separated and random
!                  uniform starting vectors otherwise.
!
!                  For reduced tridiagonal matrices B' * B, the default is to use random uniform starting vectors.
!
!   FAILURE_BD     (OUTPUT, OPTIONAL) logical(lgl)
!                  On exit:
!
!                  - FAILURE_BD = false :  indicates that maximum accuracy was obtained in the
!                    Ralha-Barlow one-sided bidiagonalization of MAT;
!                  - FAILURE_BD = true  :  indicates that MAT is nearly singular.
!
!   FAILURE_BISECT (OUTPUT, OPTIONAL) logical(lgl)
!                  On exit:
!
!                  - FAILURE_BISECT = false :  indicates successful exit and the bisection algorithm converged for all
!                    the computed singular values to the desired accuracy;
!                  - FAILURE_BISECT = true  :  indicates that some or all of the singular values failed to converge
!                    or were not computed. This is generally caused by unexpectedly inaccurate arithmetic.
!                
!
! Further Details
! _______________
!
!   Computing the SVD of a m-by-n matrix MAT, with m>=n, in subroutine SVD_CMP6 consists of four steps:
!
!       1) reduction of the rectangular matrix to bidiagonal form B via the Ralha-Barlow
!          one-sided bidiagonalization algorithm, see the references (1) and (2);
!       2) computation of the singular values and right singular vectors of B by applying the bisection
!          and inverse iteration algorithms to the matrices B and B' * B, respectively, see the reference (3);
!       3) computation of the right singular vectors of MAT from those of B by a back-transformation algorithm,
!          see the references (3) and (4);
!       4) computation and orthogonalization of the left singular vectors in the SVD of MAT to avoid
!          the possible loss of orthogonality of the left orthogonal matrix in the bidiagonal factorization
!          of MAT computed by the one-sided bidiagonalization algorithm, see the references (1) and (2).
!
!   This four-step algorithm computes numerically orthogonal left singular vectors of a m-by-n matrix MAT,
!   with m>=n, even in cases of large condition number of MAT despite that the left (orthogonal) matrix in the
!   bidiagonal decomposition of MAT (computed by the Ralha-Barlow one-sided bidiagonalization algorithm) may not
!   be numerically orthogonal when MAT is nearly singular or has a very large condition number (see references (1)
!   and (2) for details).
!
!   If m<n, this four-step algorithm is applied to MAT', instead of MAT, to get the SVD of MAT and it
!   computes also numerically orthogonal right singular vectors of MAT in that case.
!
!   Note that if max(m,n) is much larger than min(m,n), the rectangular matrix is first reduced to
!   upper or lower triangular form by a preliminary QR or LQ factorization and the four-step reduction algorithm
!   described above is applied to the resulting triangular factor. The singular vectors of the original rectangular
!   matrix are then obtained from those of the triangular factor in the QR or LQ factorization by a back-transformation
!   algorithm (see the references (3) and (4) for details).
!
!   For further details on the SVD of a rectangular matrix and the algorithms to compute it, see the references below.
!
!   The four or five steps of the SVD algorithm used here (e.g., preliminary QR or LQ factorization,
!   reduction to bidiagonal form B, computation of the singular values and right (or left if m<n) singular
!   vectors of B, computation of the right (or left if m<n) singular vectors of MAT and orthogonalization
!   of the left (or right if m<n) singular vectors in the SVD of MAT) are parallelized if OPENMP is used.
!
!   For more details, see:
!
!   (1) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!            A new stable bidiagonal reduction algorithm.
!            Linear Algebra Appl., No 397, pp. 35-84.
!
!   (2) Bosner, N., and Barlow, J.L., 2007:
!            Block and Parallel versions of one-sided bidiagonalization.
!            SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!            Matrix Computations.
!            3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (4) Lawson, C.L., and Hanson, R.J., 1974:
!            Solving least square problems.
!            Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error
    use String_Procedures, only : to_lower
    use Utilities,         only : assert, lascl, merror
    use QR_Procedures,     only : qr_cmp, ortho_gen_qr, apply_q_qr
    use EIG_Procedures,    only : trid_inviter
#ifdef _BLAS
    use BLAS_interfaces,   only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout), dimension(:,:)          :: mat
    real(stnd),  intent(out),   dimension(:),   pointer :: s
    real(stnd),  intent(out),   dimension(:,:), pointer :: v
!
    integer(i4b), intent(in),    optional :: maxiter
    integer(i4b), intent(inout), optional :: nsvd
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: ortho, backward_sweep, scaling, initvec
    logical(lgl), intent(out), optional :: failure_bd, failure_bisect
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension( min(size(mat,1),size(mat,2)) ) :: d, e, d2, e2, eigval
    real(stnd)                                            :: matnorm, tmp, abstol
    real(stnd), dimension(:,:), allocatable               :: mat2, work, work2, w
    real(stnd), dimension(:),   allocatable               :: tau
!
    integer(i4b) ::  i, k, m, n, mnthr, mnthr_nsvd, nsvd2
    integer      ::  iok
!
    logical(lgl) :: mat_is_scaled, gen_p
!
    character    :: sort_def
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='svd_cmp6'
!
    character,  parameter :: a_min=achar(97), d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
    if ( present( failure_bd ) ) then
!
        failure_bd = false
!
    end if
!
    if ( present( failure_bisect ) ) then
!
        failure_bisect = false
!
    end if
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = min( m, n )
!
    sort_def = d_min
!
    if ( present( sort ) ) then
!
        sort_def = to_lower( sort )
!
    end if
!
    nsvd2 = k
!
    if ( present(nsvd) ) then
!
        call assert( logical( nsvd<=k, lgl ),     &
                     logical( 1_i4b<=nsvd, lgl ), &
                     name_proc )
!
        nsvd2 = nsvd
!
    end if
!
!   DEALLOCATE POINTERS ON ENTRY.
!
#ifndef _USE_NAGWARE
    if ( associated( s ) ) then
!
        deallocate( s )
!
    end if
!
    if ( associated( v ) ) then
!
        deallocate( v )
!
    end if
#endif
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
    do i = 1_i4b, n
!
        tmp = maxval( abs( mat(1_i4b:m,i) ) )
        matnorm = max( matnorm, tmp )
!
    end do
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       ALLOCATE INPUT POINTER ARGUMENTS FOR STORING THE (PARTIAL) SVD.
!
        allocate( s(k), v(n,k), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       mat IS A ZERO MATRIX, ALL SINGULAR VALUES ARE ZERO AND
!       U AND V ARE IDENTITY MATRICES.
!
        s(:k) = zero
!
        do i = 1_i4b, k
!
            mat(i,i) = one
!
            v(:n,i)  = zero
            v(i,i)   = one
!
        end do
!
        if ( present(nsvd) ) then
            nsvd = k
        end if
!
        return
!
    else if ( matnorm<smlnum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   DETERMINE SINGULAR VALUES TO MAXIMUM ACCURACY IN SUBROUTINE SYMTRID_BISECT.
!
    abstol = sqrt( safmin )
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat TO BIDIAGONAL FORM. IF max( m, n ) EXCEEDS mnthr ,
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE
!   m-BY-n MATRIX mat TO A TRIANGULAR FORM.
!    
    mnthr = int( real( k, stnd )*thr_bd2, i4b )
!
!   DETERMINE CROSSOVER POINT FOR THE BACK-TRANSFORMATION
!   ALGORITHM WHEN COMPUTING SINGULAR VECTORS. IF nsvd2
!   EXCEEDS mnthr_nsvd, THE RIGHT ORTHOGONAL MATRIX p
!   OF THE ONE-SIDED BIDIAGONAL REDUCTION ALGORITHM IS GENERATED
!   AND BACK-TRANSFORMATION IS DONE BY A MATRIX MULTIPLICATION.
!
    mnthr_nsvd = int( real( k, stnd )*thr_bt, i4b )
!
    gen_p = nsvd2>=mnthr_nsvd
!
    if ( m>=n ) then
!
!       mat HAS AT LEAST AS MANY ROWS AS COLUMNS.
!       n LEFT SINGULAR VECTORS TO BE OVERWRITTEN ON mat AND
!       n RIGHT SINGULAR VECTORS TO BE COMPUTED IN v .
!
!       ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
        allocate( mat2(m,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE FIRST A COPY OF mat .
!
        mat2(:m,:n) = mat(:m,:n)
!
        if ( m>=mnthr ) then
!
!           USE THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!           WITH A PRELIMINARY QR FACTORIZATION (PATH 1).
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM.
!
            allocate( work(n,n), tau(n), stat = iok )
!
        else
!
!           USE THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!           WITHOUT A PRELIMINARY QR FACTORIZATION (PATH 2).
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!
!           PATH 1 : m MUCH LARGER THAN n.
!
!           REDUCE mat USING THE QR DECOMPOSITION.
!           SECOND REDUCE R TO UPPER BIDIAGONAL FORM WITH
!           THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!           COMPUTE mat = Q * R .
!
            call qr_cmp( mat2(:m,:n), d(:n), tau(:n) )
!
!           COPY R TO work AND mat, ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(n*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, n
!
                work(1_i4b:i-1_i4b,i) = mat2(1_i4b:i-1_i4b,i)
                work(i,i) = d(i)
                work(i+1_i4b:n,i) = zero
!
                mat(1_i4b:i,i) = work(1_i4b:i,i)
                mat(i+1_i4b:n,i) = zero
!
            end do
!
!$OMP END PARALLEL DO
!
!           BIDIAGONALIZE R STORED IN work WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
            call bd_cmp3( work(:n,:n), d(:n), e(:n), gen_p=gen_p, failure=failure_bd )
!     
!           COMPUTE SINGULAR VALUES OF THE ASSOCIATED BIDIAGONAL MATRIX B BY BISECTION.
!          
            call bd_singval( d, e, i, eigval, failure_bisect, sort=d_min,  &
                             abstol=abstol, ls=nsvd2, scaling=scaling )
!
!           SAVE THE NUMBER OF SINGULAR TRIPLETS TO COMPUTE.
!
            nsvd2 = i
!
!           ALLOCATE WORK ARRAY AND INPUT POINTER ARGUMENTS FOR STORING THE (PARTIAL) SVD.
!
            if ( gen_p ) then
!
                allocate( s(nsvd2), v(n,nsvd2), w(n,nsvd2), stat=iok )
!
            else
!
                allocate( s(nsvd2), v(n,nsvd2), stat=iok )
!
            end if
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           SAVE SINGULAR VALUES AND COMPUTE EIGENVALUES OF B**(t)*B FROM THE SINGULAR VALUES OF B.
!
            s(:nsvd2) = eigval(:nsvd2)
            eigval(:nsvd2) = s(:nsvd2)*s(:nsvd2)
!     
!           COMPUTE THE ASSOCIATED TRIDIAGONAL MATRIX B**(t)*B.
!     
            d2(1_i4b) = d(1_i4b)*d(1_i4b)
            d2(2_i4b:n) = d(2_i4b:n)*d(2_i4b:n) + e(2_i4b:n)*e(2_i4b:n)
!
            e2(1_i4b:n-1_i4b) = d(1_i4b:n-1_i4b)*e(2_i4b:n)
            e2(n) = zero
!
            if ( gen_p ) then
!     
!               COMPUTE EIGENVECTORS OF THE ASSOCIATED TRIDIAGONAL MATRIX B**(t)*B BY INVERSE ITERATION.
!               THESE EIGENVECTORS ARE ALSO THE RIGHT SINGULAR VECTORS OF B.
!     
                call trid_inviter( d2, e2, eigval(:nsvd2), w(:n,:nsvd2), failure, maxiter=maxiter, ortho=ortho, &
                                   backward_sweep=backward_sweep, scaling=scaling, initvec=initvec )
!
!               COMPUTE RIGHT SINGULAR VECTORS OF R, WHICH ARE ALSO RIGHT SINGULAR VECTORS OF THE FULL m-BY-n MATRIX
!               mat BY MULTIPLICATION.
!
#ifdef _BLAS
                call gemm( 'N', 'N', n, nsvd2, n, one, work, n, w, n, zero, v, n )
#else
                v(:n,:nsvd2) = matmul( work(:n,:n), w(:n,:nsvd2) )
#endif
!
!               DEALLOCATE WORK ARRAY.
!
                deallocate( w )
!
            else
!     
!               COMPUTE EIGENVECTORS OF THE ASSOCIATED TRIDIAGONAL MATRIX B**(t)*B BY INVERSE ITERATION.
!               THESE EIGENVECTORS ARE ALSO THE RIGHT SINGULAR VECTORS OF B.
!     
                call trid_inviter( d2, e2, eigval(:nsvd2), v(:n,:nsvd2), failure, maxiter=maxiter, ortho=ortho, &
                                   backward_sweep=backward_sweep, scaling=scaling, initvec=initvec )
!
!               COMPUTE RIGHT SINGULAR VECTORS OF R, WHICH ARE ALSO RIGHT SINGULAR VECTORS OF THE FULL m-BY-n MATRIX
!               mat BY BACK TRANSFORMATION.
!
                call apply_q_qr( work(2_i4b:n,2_i4b:n), work(2_i4b:n,1_i4b), v(2_i4b:n,:nsvd2), left=true, trans=false )
!
            end if
!
!           MULTIPLY R STORED IN mat BY RIGHT SINGULAR VECTORS OF R STORED IN v, TO GET UNNORMALIZED
!           LEFT SINGULAR VECTORS OF R, STORING RESULT IN work .
!
#ifdef _BLAS
            call gemm( 'N', 'N', n, nsvd2, n, one, mat, m, v, n, zero, work, n  )
#else
            work(:n,:nsvd2) = matmul( mat(:n,:n), v(:n,:nsvd2) )
#endif
!
!           ORHOGONALIZE AND NORMALIZE LEFT SINGULAR VECTORS OF R WITH A QR FACTORIZATION.
!
            call qr_cmp( work(:n,:nsvd2), d(:nsvd2), e(:nsvd2) )
!
!           GENERATE LEFT SINGULAR VECTORS OF R IN work .
!
            call ortho_gen_qr( work(:n,:nsvd2), e(:nsvd2) )
!
!           ADJUST THE SIGN OF THE LEFT SINGULAR VECTORS OF R .
!
!$OMP PARALLEL DO IF(n*nsvd2>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, nsvd2
!
                if ( d(i)<zero ) then
!
                    work(:n,i) = -work(:n,i)
!        
                end if
!        
            end do
!
!$OMP END PARALLEL DO
!
!           SORT SINGULAR VALUES AND VECTORS OF R IN ASCENDING ORDER IF REQUIRED.
!
            if ( sort_def==a_min ) then
!
                call lr_singvec_sort( a_min, s(:nsvd2), work(:n,:nsvd2), v(:n,:nsvd2) )
!
            end if
!
            if ( gen_p ) then
!
!               GENERATE Q IN mat2 .
!
                call ortho_gen_qr( mat2(:m,:n), tau(:n) )
!
!               MULTIPLY Q IN mat2 BY LEFT SINGULAR VECTORS OF R STORED IN work
!               TO GET LEFT SINGULAR VECTORS OF mat, STORING RESULT IN mat .
!
#ifdef _BLAS
                call gemm( 'N', 'N', m, nsvd2, n, one, mat2, m, work, n, zero, mat, m  )
#else
                mat(:m,:nsvd2) = matmul( mat2(:m,:n), work(:n,:nsvd2) )
#endif
!
            else
!
!               MULTIPLY Q IN mat2, STORED IN FACTORED FORM, BY LEFT SINGULAR VECTORS OF R
!               STORED IN work BY A BACK_TRANSFORMATION ALGORITHM TO GET LEFT SINGULAR
!               VECTORS OF mat, STORING RESULT IN mat .
!
                mat(:n,:nsvd2) = work(:n,:nsvd2)
                mat(n+1_i4b:m,:nsvd2) = zero
!
                call apply_q_qr( mat2, tau, mat(:m,:nsvd2), left=true, trans=false )
!
            end if
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( mat2, work, tau )
!     
        else
!
!           PATH 2 : mat HAS AT LEAST AS MANY ROWS AS COLUMNS, BUT m IS NOT MUCH LARGER THAN n.
!                    BIDIAGONALIZE mat WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!                    DIRECTLY.
!
            call bd_cmp3( mat(:m,:n), d(:n), e(:n), gen_p=gen_p, failure=failure_bd )
!     
!           COMPUTE SINGULAR VALUES OF THE ASSOCIATED BIDIAGONAL MATRIX B BY BISECTION.
!          
            call bd_singval( d, e, i, eigval, failure_bisect,  sort=d_min, &
                             abstol=abstol, ls=nsvd2, scaling=scaling )
!
!           SAVE THE NUMBER OF SINGULAR TRIPLETS TO COMPUTE.
!
            nsvd2 = i
!
!           ALLOCATE WORK ARRAY AND INPUT POINTER ARGUMENTS FOR STORING THE (PARTIAL) SVD.
!
            if ( gen_p ) then
!
                allocate( s(nsvd2), v(n,nsvd2), w(n,nsvd2), stat=iok )
!
            else
!
                allocate( s(nsvd2), v(n,nsvd2), stat=iok )
!
            end if
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           SAVE SINGULAR VALUES AND COMPUTE EIGENVALUES OF B**(t)*B FROM THE SINGULAR VALUES OF B.
!
            s(:nsvd2) = eigval(:nsvd2)
            eigval(:nsvd2) = s(:nsvd2)*s(:nsvd2)
!     
!           COMPUTE THE ASSOCIATED TRIDIAGONAL MATRIX B**(t)*B.
!     
            d2(1_i4b) = d(1_i4b)*d(1_i4b)
            d2(2_i4b:n) = d(2_i4b:n)*d(2_i4b:n) + e(2_i4b:n)*e(2_i4b:n)
!
            e2(1_i4b:n-1_i4b) = d(1_i4b:n-1_i4b)*e(2_i4b:n)
            e2(n) = zero
!
            if ( gen_p ) then
!     
!               COMPUTE EIGENVECTORS OF THE ASSOCIATED TRIDIAGONAL MATRIX B**(t)*B BY INVERSE ITERATION.
!               THESE EIGENVECTORS ARE ALSO THE RIGHT SINGULAR VECTORS OF B.
!     
                call trid_inviter( d2, e2, eigval(:nsvd2), w(:n,:nsvd2), failure, maxiter=maxiter, ortho=ortho, &
                                   backward_sweep=backward_sweep, scaling=scaling, initvec=initvec )
!
!               COMPUTE RIGHT SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY MULTIPLICATION.
!
#ifdef _BLAS
                call gemm( 'N', 'N', n, nsvd2, n, one, mat, m, w, n, zero, v, n )
#else
                v(:n,:nsvd2) = matmul( mat(:n,:n), w(:n,:nsvd2) )
#endif
!
!               DEALLOCATE WORK ARRAY.
!
                deallocate( w )
!
            else
!     
!               COMPUTE EIGENVECTORS OF THE ASSOCIATED TRIDIAGONAL MATRIX B**(t)*B BY INVERSE ITERATION.
!               THESE EIGENVECTORS ARE ALSO THE RIGHT SINGULAR VECTORS OF B.
!     
                call trid_inviter( d2, e2, eigval(:nsvd2), v(:n,:nsvd2), failure, maxiter=maxiter, ortho=ortho, &
                                   backward_sweep=backward_sweep, scaling=scaling, initvec=initvec )
!
!               COMPUTE RIGHT SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY BACK TRANSFORMATION.
!
                call apply_q_qr( mat(2_i4b:n,2_i4b:n), mat(2_i4b:n,1_i4b), v(2_i4b:n,:nsvd2), left=true, trans=false )
!
            end if
!
!           MULTIPLY mat STORED IN mat2 BY RIGHT SINGULAR VECTORS OF mat STORED IN v, TO GET
!           UNNORMALIZED LEFT SINGULAR VECTORS OF mat, STORING RESULT IN mat .
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, nsvd2, n, one, mat2, m, v, n, zero, mat, m  )
#else
            mat(:m,:nsvd2) = matmul( mat2(:m,:n), v(:n,:nsvd2) )
#endif
!
!           ORHOGONALIZE AND NORMALIZE LEFT SINGULAR VECTORS OF mat WITH A QR FACTORIZATION.
!
            call qr_cmp( mat(:m,:nsvd2), d(:nsvd2), e(:nsvd2) )
!
!           GENERATE LEFT SINGULAR VECTORS OF mat IN mat .
!
            call ortho_gen_qr( mat(:m,:nsvd2), e(:nsvd2) )
!
!           ADJUST THE SIGN OF THE LEFT SINGULAR VECTORS OF mat .
!
!$OMP PARALLEL DO IF(m*nsvd2>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, nsvd2
!
                if ( d(i)<zero ) then
!
                    mat(:m,i) = -mat(:m,i)
!        
                end if
!        
            end do
!
!$OMP END PARALLEL DO
!
!           SORT SINGULAR VALUES AND VECTORS OF mat IN ASCENDING ORDER IF REQUIRED.
!
            if ( sort_def==a_min ) then
!
                call lr_singvec_sort( a_min, s(:nsvd2), mat(:m,:nsvd2), v(:n,:nsvd2) )
!
            end if
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( mat2 )
!
        end if
!
    else
!
!       mat HAS MORE COLUMNS THAN ROWS.
!       m LEFT SINGULAR VECTORS TO BE OVERWRITTEN ON mat AND
!       m RIGHT SINGULAR VECTORS TO BE COMPUTED IN v .
!
!       ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
        allocate( mat2(n,m), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       FIRST TRANSPOSE mat IN mat2 .
!
        mat2(:n,:m) = transpose( mat(:m,:n) )
!
        if ( n>=mnthr ) then
!
!           USE THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!           WITH A PRELIMINARY QR FACTORIZATION OF mat**(t) (PATH 3).
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM.
!
            allocate( work(m,m), work2(m,m), tau(m), stat = iok )
!
        else
!
!           USE THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!           WITHOUT A PRELIMINARY QR FACTORIZATION OF mat**(t) (PATH 4).
!
!           ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
            allocate( work(n,m), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!
!           PATH 3 : n MUCH LARGER THAN m.
!
!           REDUCE mat**(t) USING THE QR DECOMPOSITION.
!           SECOND REDUCE R TO UPPER BIDIAGONAL FORM WITH
!           THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
!           COMPUTE mat**(t) = Q * R .
!
            call qr_cmp( mat2(:n,:m), d(:m), tau(:m) )
!
!           COPY R TO work AND work2, ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(m*m>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, m
!
                work(1_i4b:i-1_i4b,i) = mat2(1_i4b:i-1_i4b,i)
                work(i,i) = d(i)
                work(i+1_i4b:m,i) = zero
!
                work2(1_i4b:i,i) = work(1_i4b:i,i)
                work2(i+1_i4b:m,i) = zero
!
            end do
!
!$OMP END PARALLEL DO
!
!           BIDIAGONALIZE R STORED IN work WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
            call bd_cmp3( work(:m,:m), d(:m), e(:m), gen_p=gen_p, failure=failure_bd )
!     
!           COMPUTE SINGULAR VALUES OF THE ASSOCIATED BIDIAGONAL MATRIX B BY BISECTION.
!
            call bd_singval( d, e, i, eigval, failure_bisect,  sort=d_min,  &
                             abstol=abstol, ls=nsvd2, scaling=scaling )
!
!           SAVE THE NUMBER OF SINGULAR TRIPLETS TO COMPUTE.
!
            nsvd2 = i
!
!           ALLOCATE WORK ARRAY AND INPUT POINTER ARGUMENTS FOR STORING THE (PARTIAL) SVD.
!
            if ( gen_p ) then
!
                allocate( s(nsvd2), v(n,nsvd2), w(m,nsvd2), stat=iok )
!
            else
!
                allocate( s(nsvd2), v(n,nsvd2), stat=iok )
!
            end if
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           SAVE SINGULAR VALUES AND COMPUTE EIGENVALUES OF B**(t)*B FROM THE SINGULAR VALUES OF B.
!
            s(:nsvd2) = eigval(:nsvd2)
            eigval(:nsvd2) = s(:nsvd2)*s(:nsvd2)
!     
!           COMPUTE THE ASSOCIATED TRIDIAGONAL MATRIX B**(t)*B.
!     
            d2(1_i4b) = d(1_i4b)*d(1_i4b)
            d2(2_i4b:m) = d(2_i4b:m)*d(2_i4b:m) + e(2_i4b:m)*e(2_i4b:m)
!
            e2(1_i4b:m-1_i4b) = d(1_i4b:m-1_i4b)*e(2_i4b:m)
            e2(m) = zero
!
            if ( gen_p ) then
!     
!               COMPUTE EIGENVECTORS OF THE ASSOCIATED TRIDIAGONAL MATRIX B**(t)*B BY INVERSE ITERATION.
!               THESE EIGENVECTORS ARE ALSO THE RIGHT SINGULAR VECTORS OF B.
!     
                call trid_inviter( d2, e2, eigval(:nsvd2), w(:m,:nsvd2), failure, maxiter=maxiter, ortho=ortho, &
                                   backward_sweep=backward_sweep, scaling=scaling, initvec=initvec )
!
!               COMPUTE RIGHT SINGULAR VECTORS OF R, WHICH ARE ALSO RIGHT SINGULAR VECTORS OF THE FULL m-BY-n MATRIX
!               mat BY MULTIPLICATION.
!
#ifdef _BLAS
                call gemm( 'N', 'N', m, nsvd2, m, one, work, m, w, m, zero, mat, m )
#else
                mat(:m,:nsvd2) = matmul( work(:m,:m), w(:m,:nsvd2) )
#endif
!
!               DEALLOCATE WORK ARRAY.
!
                deallocate( w )
!
            else
!     
!               COMPUTE EIGENVECTORS OF THE ASSOCIATED TRIDIAGONAL MATRIX B**(t)*B BY INVERSE ITERATION.
!               THESE EIGENVECTORS ARE ALSO THE RIGHT SINGULAR VECTORS OF B.
!     
                call trid_inviter( d2, e2, eigval(:nsvd2), mat(:m,:nsvd2), failure, maxiter=maxiter, ortho=ortho, &
                                   backward_sweep=backward_sweep, scaling=scaling, initvec=initvec )
!
!               COMPUTE RIGHT SINGULAR VECTORS OF R, WHICH ARE ALSO LEFT SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat
!               BY BACK TRANSFORMATION.
!
                call apply_q_qr( work(2_i4b:m,2_i4b:m), work(2_i4b:m,1_i4b), mat(2_i4b:m,:nsvd2), left=true, trans=false )
!
            end if
!
!           MULTIPLY R STORED IN work2 BY RIGHT SINGULAR VECTORS OF R STORED IN mat, TO GET
!           UNNORMALIZED LEFT SINGULAR VECTORS OF R, STORING RESULT IN work .
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, nsvd2, m, one, work2, m, mat, m, zero, work, m )
#else
            work(:m,:nsvd2) = matmul( work2(:m,:m), mat(:m,:nsvd2) )
#endif
!
!           ORHOGONALIZE AND NORMALIZE LEFT SINGULAR VECTORS OF R WITH A QR FACTORIZATION.
!
            call qr_cmp( work(:m,:nsvd2), d(:nsvd2), e(:nsvd2) )
!
!           GENERATE LEFT SINGULAR VECTORS OF R IN work .
!
            call ortho_gen_qr( work(:m,:nsvd2), e(:nsvd2) )
!
!           ADJUST THE SIGN OF THE LEFT SINGULAR VECTORS OF R .
!
!$OMP PARALLEL DO IF(m*nsvd2>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, nsvd2
!
                if ( d(i)<zero ) then
!
                    work(:m,i) = -work(:m,i)
!        
                end if
!        
            end do
!
!$OMP END PARALLEL DO
!
!           SORT SINGULAR VALUES AND VECTORS OF R IN ASCENDING ORDER IF REQUIRED.
!
            if ( sort_def==a_min ) then
!
                call lr_singvec_sort( a_min, s(:nsvd2), work(:m,:nsvd2), mat(:m,:nsvd2) )
!
            end if
!
            if ( gen_p ) then
!
!               GENERATE Q IN mat2 .
!
                call ortho_gen_qr( mat2(:n,:m), tau(:m) )
!
!               MULTIPLY Q IN mat2 BY THE LEFT SINGULAR VECTORS OF R STORED IN work
!               TO GET THE RIGHT SINGULAR VECTORS OF mat STORED COLUMN-WISE, STORING
!               RESULT IN v .

#ifdef _BLAS
                call gemm( 'N', 'N', n, nsvd2, m, one, mat2, n, work, m, zero, v, n  )
#else
                v(:n,:nsvd2) = matmul( mat2(:n,:m), work(:m,:nsvd2) )
#endif
!
            else
!
!               MULTIPLY Q IN mat2, STORED IN FACTORED FORM, BY LEFT SINGULAR VECTORS OF R
!               STORED IN work BY A BACK_TRANSFORMATION ALGORITHM TO GET THE RIGHT SINGULAR
!               VECTORS OF mat STORED COLUMN-WISE, STORING RESULT IN v .
!
                v(:m,:nsvd2) = work(:m,:nsvd2)
                v(m+1_i4b:n,:nsvd2) = zero
!
                call apply_q_qr( mat2, tau, v(:n,:nsvd2), left=true, trans=false )
!
            end if
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( mat2, work, work2, tau )
!
        else
!
!           PATH 4 : mat HAS AT LEAST AS MANY COLUMNS AS ROWS, BUT n IS NOT MUCH LARGER THAN m.
!                    BIDIAGONALIZE mat**(t) WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM
!                    DIRECTLY.
!
!           SAVE mat**(t) IN work .
!
            work(:n,:m) = mat2(:n,:m)
!
!           BIDIAGONALIZE mat**(t) WITH THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION ALGORITHM.
!
            call bd_cmp3( work(:n,:m), d(:m), e(:m), gen_p=gen_p, failure=failure_bd )
!     
!           COMPUTE SINGULAR VALUES OF THE ASSOCIATED BIDIAGONAL MATRIX B BY BISECTION.
!
            call bd_singval( d, e, i, eigval, failure_bisect,  sort=d_min,  &
                             abstol=abstol, ls=nsvd2, scaling=scaling )
!
!           SAVE THE NUMBER OF SINGULAR TRIPLETS TO COMPUTE.
!
            nsvd2 = i
!
!           ALLOCATE WORK ARRAY AND INPUT POINTER ARGUMENTS FOR STORING THE (PARTIAL) SVD.
!
            if ( gen_p ) then
!
                allocate( s(nsvd2), v(n,nsvd2), w(m,nsvd2), stat=iok )
!
            else
!
                allocate( s(nsvd2), v(n,nsvd2), stat=iok )
!
            end if
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           SAVE SINGULAR VALUES AND COMPUTE EIGENVALUES OF B**(t)*B FROM THE SINGULAR VALUES OF B.
!
            s(:nsvd2) = eigval(:nsvd2)
            eigval(:nsvd2) = s(:nsvd2)*s(:nsvd2)
!     
!           COMPUTE THE ASSOCIATED TRIDIAGONAL MATRIX B**(t)*B.
!     
            d2(1_i4b) = d(1_i4b)*d(1_i4b)
            d2(2_i4b:m) = d(2_i4b:m)*d(2_i4b:m) + e(2_i4b:m)*e(2_i4b:m)
!
            e2(1_i4b:m-1_i4b) = d(1_i4b:m-1_i4b)*e(2_i4b:m)
            e2(m) = zero
!
            if ( gen_p ) then
!     
!               COMPUTE EIGENVECTORS OF THE ASSOCIATED TRIDIAGONAL MATRIX B**(t)*B BY INVERSE ITERATION.
!               THESE EIGENVECTORS ARE ALSO THE RIGHT SINGULAR VECTORS OF B.
!     
                call trid_inviter( d2, e2, eigval(:nsvd2), w(:m,:nsvd2), failure, maxiter=maxiter, ortho=ortho, &
                                   backward_sweep=backward_sweep, scaling=scaling, initvec=initvec )
!
!               COMPUTE RIGHT SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY MULTIPLICATION.
!
#ifdef _BLAS
                call gemm( 'N', 'N', m, nsvd2, m, one, work, n, w, m, zero, mat, m )
#else
                mat(:m,:nsvd2) = matmul( work(:m,:m), w(:m,:nsvd2) )
#endif
!
!               DEALLOCATE WORK ARRAY.
!
                deallocate( w )
!
            else
!     
!               COMPUTE EIGENVECTORS OF THE ASSOCIATED TRIDIAGONAL MATRIX B**(t)*B BY INVERSE ITERATION.
!               THESE EIGENVECTORS ARE ALSO THE RIGHT SINGULAR VECTORS OF B.
!     
                call trid_inviter( d2, e2, eigval(:nsvd2), mat(:m,:nsvd2), failure, maxiter=maxiter, ortho=ortho, &
                                   backward_sweep=backward_sweep, scaling=scaling, initvec=initvec )
!
!               COMPUTE RIGHT SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat**(t) BY BACK TRANSFORMATION.
!
                call apply_q_qr( work(2_i4b:m,2_i4b:m), work(2_i4b:m,1_i4b), mat(2_i4b:m,:nsvd2), left=true, trans=false )
!
            end if
!
!           MULTIPLY mat**(t) STORED IN mat2 BY RIGHT SINGULAR VECTORS OF mat**(t) STORED IN
!           mat , TO GET UNNORMALIZED LEFT SINGULAR VECTORS OF mat**(t), STORING RESULT IN v .
!
#ifdef _BLAS
            call gemm( 'N', 'N', n, nsvd2, m, one, mat2, n, mat, m, zero, v, n  )
#else
            v(:n,:nsvd2) = matmul( mat2(:n,:m), mat(:m,:nsvd2) )
#endif
!
!           ORHOGONALIZE AND NORMALIZE LEFT SINGULAR VECTORS OF mat**(t) WITH A QR FACTORIZATION.
!
            call qr_cmp( v(:n,:nsvd2), d(:nsvd2), e(:nsvd2) )
!
!           GENERATE LEFT SINGULAR VECTORS OF mat**(t) IN v .
!
            call ortho_gen_qr( v(:n,:nsvd2), e(:nsvd2) )
!
!           ADJUST THE SIGN OF THE LEFT SINGULAR VECTORS OF mat**(t) .
!
!$OMP PARALLEL DO IF(n*nsvd2>=omp_limit), SCHEDULE(STATIC), PRIVATE(i)
!
            do i = 1_i4b, nsvd2
!
                if ( d(i)<zero ) then
!
                    v(:n,i) = -v(:n,i)
!        
                end if
!        
            end do
!
!$OMP END PARALLEL DO
!
!           SORT SINGULAR VALUES AND VECTORS OF mat IN ASCENDING ORDER IF REQUIRED.
!
            if ( sort_def==a_min ) then
!
                call lr_singvec_sort( a_min, s(:nsvd2), mat(:m,:nsvd2), v )
!
            end if
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( mat2, work )
!
        end if
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled ) then
!
        call lascl( s, merge(smlnum,bignum,matnorm<smlnum), matnorm )
!
    end if
!
!   OUTPUT THE NUMBER OF COMPUTED SINGULAR TRIPLETS, IF REQUIRED.
!
    if ( present(nsvd) ) then
        nsvd = nsvd2
    end if
!
!
! END OF SUBROUTINE svd_cmp6
! __________________________
!
    end subroutine svd_cmp6
!
! =========================================================================================
!
    subroutine svd_cmp7( mat, s, u, v, failure, sort, maxiter, ortho, backward_sweep, &
                         scaling, initvec, failure_dqds )
!
! Purpose
! _______
!
!   SVD_CMP7 computes a full or partial Singular Value Decomposition (SVD) of a real
!   m-by-n matrix MAT. The full SVD is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix. The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!   The columns of U and V are, respectively, the left and right
!   singular vectors of MAT.
!
!   The routine returns the first min(m,n) singular values and the associated
!   left and right singular vectors (e.g. the left and right singular vectors of MAT
!   in the thin SVD of MAT) or a truncated partial SVD at the user option.
!
!   MAT is first reduced to bidiagonal form B with a one-stage or two-stage algorithm,
!   depending on its shape, see Further Details below and the references (1) and (2).
!
!   The singular values and vectors of B are then computed by the dqds and inverse iteration methods
!   applied to B, see the references (1), (3) and (4).
!
!   The singular vectors of MAT are finally computed by a blocked back-transformation algorithm.
!
!
! Arguments
! _________
!
!   MAT            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                  On entry, the m-by-n matrix MAT.
!
!                  On exit, MAT is overwritten as it is used as workspace inside the subroutine.
!
!   S              (OUTPUT) real(stnd), dimension(:)
!                  The singular values of MAT.
!
!                  The size of S must verify: size( S ) = min(m,n) .
!
!   U              (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed first nsvd columns of U, e.g., the left singular vectors
!                  associated with the first nsvd largest singular values of MAT.
!
!                  The left singular vector associated with the singular value S(j) is stored
!                  in the j-th column of U.
!
!                  The shape of V must verify: 
!
!                  - size( U, 1 ) = size( MAT, 1 )    = m ,
!                  - size( U, 2 ) = nsvd <= size( S ) = min(m,n) .
!
!   V              (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed first nsvd columns of V, e.g., the right singular vectors
!                  associated with the first nsvd largest singular values of MAT.
!
!                  The right singular vector associated with the singular value S(j) is stored
!                  in the j-th column of V.
!
!                  The shape of V must verify: 
!
!                  - size( V, 1 ) = size( MAT, 2 ) = n ,
!                  - size( V, 2 ) = nsvd <= size( S ) = min(m,n) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false indicates successful exit.
!                  - FAILURE = true indicates that the dqds algorithm failed to converge for some
!                    singular values or that some singular vectors failed to converge in MAXITER
!                    inverse iterations. If the dqds algorithm failed to compute some singular values,
!                    the sign of the incorrect singular values in S is set to negative and
!                    arguments U and V are filled with a quiet NAN in output of the subroutine.
!
!   SORT           (INPUT, OPTIONAL) character
!                  Sort the singular values into ascending order if SORT = 'A' or 'a',
!                  or into descending order if SORT = 'D' or 'd'. The singular vectors
!                  are rearranged accordingly.
!
!                  The default is to sort the singular values and vectors into descending order.
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed for computing singular vectors.
!
!                  By default, 2 inverse iterations are performed for all the singular vectors.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, all the singular vectors are orthogonalized by the
!                    Modified Gram-Schmidt or QR algorithm in the inverse iteration algorithm;
!                  - ORTHO=false, the singular vectors are not orthogonalized by 
!                    the Modified Gram-Schmidt or QR algorithm in the inverse iteration algorithm.
!
!                  The default is to orthogonalize the singular vectors only for the singular values,
!                  which are not well-separated.
!
!   BACKWARD_SWEEP (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - BACKWARD_SWEEP=true and the singular vectors are orthogonalized by
!                    the modified Gram-Schmidt algorithm in the inverse iteration algorithm,
!                    a backward sweep of the modified Gram-Schmidt algorithm is also performed;
!                  - BACKWARD_SWEEP=false, a backward sweep is not performed.
!
!                  The default is not to perform a backward sweep of the modified Gram-Schmidt
!                  algorithm.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the bidiagonal matrix B is scaled before computing
!                    the singular values and vectors;
!                  - SCALING=false, the bidiagonal matrix B is not scaled.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INITVEC=true, Fernando vectors are used to start the inverse iteration process
!                    for computing the singular vectors of the bidiagonal matrix BD (e.g. the eigenvectors of the
!                    associated Golub-Kahan tridiagonal matrix);
!                  - INITVEC=false, random uniform starting vectors are used.
!
!                  The default is to use Fernando starting vectors if the singular values are well-separated and
!                  the Golub-Kahan form of the input bidiagonal matrix is unreduced, and random uniform
!                  starting vectors otherwise.
!
!   FAILURE_DQDS   (OUTPUT, OPTIONAL) logical(lgl)
!                  On exit:
!
!                  - FAILURE_DQDS = false :  indicates successful exit and the dqds algorithm converged for all
!                    the computed singular values to the desired accuracy;
!                  - FAILURE_DQDS = true  :  indicates that some or all of the singular values failed to converge
!                    or were not computed. This is generally caused by unexpectedly inaccurate arithmetic.
!                    In that case, U and V are filled with a quiet NAN and the sign of the incorrect
!                    singular values in S is set to negative in output of the subroutine.
!                
!
! Further Details
! _______________
!
!   Computing the SVD of a m-by-n matrix MAT in subroutine SVD_CMP7 consists of three steps:
!
!       1) reduction of the rectangular matrix to bidiagonal form B via orthogonal
!          transformations (e.g. Householder transformations);
!       2) computation of the SVD of the bidiagonal matrix B with the dqds algorithm for the singular values
!          and inverse iteration for all or the leading singular vectors at the user option;
!       3) computation of the singular vectors of MAT from those of its bidiagonal form B by a blocked
!          back-transformation algorithm, see the references (1) and (2).
!
!   Note that if max(m,n) is much larger than min(m,n), the rectangular matrix is first reduced to
!   upper or lower triangular form by a preliminary QR or LQ factorization and the three-step reduction algorithm
!   described above is applied to the resulting triangular factor. The singular vectors of the original rectangular
!   matrix are then obtained from those of the triangular factor in the QR or LQ factorization by another blocked
!   back-transformation algorithm.
!
!   Concerning the inverse iteration algorithm used here, note that a first estimate of the singular vectors
!   is computed by the Fernando method applied to the tridiagonal Golub-Kahan matrix associated
!   with the bidiagonal matrix B (see the references (5) and (6) for details) for the singular values
!   which are well-separated and if the Golub-Kahan form of the input bidiagonal matrix is unreduced.
!   For the other singular values, a random start is used as a first estimate of the singular vectors
!   as in the standard inverse-iteration algorithm.
!
!   The singular vectors of B are then computed or refined using inverse iteration on the
!   tridiagonal Golub-Kahan matrix for all the singular values at one step.
!
!   By default, the singular vectors of B are then orthogonalized by the Modified Gram-Schmidt
!   or QR algorithm only if the singular values are not well-separated.
!
!   Excepted for the computation of the singular values by the dqds algorithm, all the steps of the SVD algorithm used
!   here (e.g., preliminary QR or LQ factorization, reduction to bidiagonal form B, computation of the singular vectors 
!   vectors of B, computation of the singular vectors of MAT by blocked back-transformation algorithms are parallelized
!   if OPENMP is used.
!
!   For more details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!            Matrix Computations.
!            3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!            Solving least square problems.
!            Prentice-Hall.
!
!   (3) Parlett, B.N., and Marques, O.A., 2000:
!           An implementation of the dqds algorithm (positive case).
!           Linear Algebra Appl., Volume 309, pp. 217-259.
!
!   (4) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!   (5) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (6) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error
    use String_Procedures, only : to_lower
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout), dimension(:,:) :: mat
    real(stnd),  intent(out),   dimension(:)   :: s
    real(stnd),  intent(out),   dimension(:,:) :: u, v
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: ortho, backward_sweep, scaling, initvec
    logical(lgl), intent(out), optional :: failure_dqds
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: abstol
    real(stnd), dimension(:,:), allocatable :: rlmat
    real(stnd), dimension(:),   allocatable :: d, e, tauo, tauq, taup
!
    integer(i4b) :: k, m, n, mnthr, mnthr_nsvd, nsvd, nsing
    integer      :: iok
!
    logical(lgl) :: two_stage, gen_q, dqds, failure1, failure2
!
    character    :: sort_def
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='svd_cmp7'
!
    character,  parameter :: a_min=achar(97), d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size(mat,1),i4b), &
                   int(size(u,1),i4b),   &
                   name_proc  )
!
    n = assert_eq( int(size(mat,2),i4b), &
                   int(size(v,1),i4b),   &
                   name_proc  )
!
    k = assert_eq( min(m,n),             &
                   int(size(s),i4b),     &
                   name_proc  )
!
    nsvd = assert_eq( int(size(u,2),i4b), &
                      int(size(v,2),i4b), &
                      name_proc  )
!
    call assert( logical( nsvd<=k, lgl ), &
                 name_proc )
!
!   TEST THE OPTIONAL ARGUMENTS.
!
    sort_def = d_min
!
    if ( present( sort ) ) then
!
        sort_def = to_lower( sort )
!
    end if
!
    if ( present( failure_dqds ) ) then
!
        failure_dqds = false
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   CHOOSE TUNING PARAMETERS FOR THE ALGORITHM.
!
!   DETERMINE HOW SINGULAR VALUES ARE COMPUTED. IF
!   dqds IS SET TO TRUE THE DQDS ALGORITHM IS USED
!   INSTEAD OF BISECTION IN select_singval_cmp
!   SUBROUTINE.
!
    dqds = true
!
!   NEXT CHOOSE TUNING PARAMETERS FOR THE BISECTION ALGORITHM WHICH IS USED
!   AS A BACKUP IF THE DQDS ALGORITHM FAILED OR IF dqds IS SET TO FALSE ABOVE.
!   SINGULAR VALUES WILL BE COMPUTED MOST ACCURATELY WHEN abstol IS SET
!   TO THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD (E.G., sqrt(safmin) )
!   IN THE BISECTION ALGORITHM.
!
    abstol = sqrt( safmin )
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat TO BIDIAGONAL FORM. IF max(n,m) EXCEEDS mnthr ,
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE
!   m-BY-n MATRIX mat TO A TRIANGULAR FORM.
!    
    mnthr = int( real( k, stnd )*thr_bd, i4b )
!
    two_stage = max( m, n )>=mnthr
!    two_stage = true
!    two_stage = false
!
!   DETERMINE CROSSOVER POINT FOR THE BACK-TRANSFORMATION
!   ALGORITHM WHEN COMPUTING SINGULAR VECTORS. IF max(n,m)
!   EXCEEDS mnthr (E.G. A PRELIMINARY QR OR QL IS DONE) AND nsvd
!   EXCEEDS mnthr_nsvd, THE ORTHOGONAL MATRIX q OF THE QR OR
!   QL FACTORIZATION IS GENERATED AND THE SINGULAR VECTORS
!   ARE COMPUTED BY A MATRIX MULTIPLICATION.
!
    mnthr_nsvd = int( real( k, stnd )*thr_bt, i4b )
!
    gen_q = nsvd>=mnthr_nsvd .and. two_stage
!    gen_q = true
!    gen_q = false
!
!   ALLOCATE WORK ARRAYS.
!
    if ( two_stage ) then
!
        if ( gen_q ) then
            allocate( rlmat(k,k), d(k), e(k),    &
                      tauq(k), taup(k), stat=iok )
        else
            allocate( rlmat(k,k), d(k), e(k),             &
                      tauo(k), tauq(k), taup(k), stat=iok )
        end if
!
    else
!
        allocate( d(k), e(k), tauq(k), taup(k), stat=iok )
!
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL m-by-n MATRIX mat
!   IS WRITTEN
!
!                       a = U * S * V**(t)
!
!   WHERE S IS AN m-by-n MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   min(m,n) DIAGONAL ELEMENTS, U IS AN n-BY-n ORTHOGONAL MATRIX, AND
!   V IS AN m-BY-m ORTHOGONAL MATRIX. THE DIAGONAL ELEMENTS OF S
!   ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!   STEP1 : COMPUTE THE min(m,n) LARGEST SINGULAR VALUES OF mat BY THE GOLUB AND REINSCH
!   BIDIAGONAL REDUCTION ALGORITHM APPLIED TO mat, THE DQDS ALGORITHM APPLIED TO THE
!   RESULTING BIDIAGONAL MATRIX AND STORE INTERMEDIATE RESULTS WITH SUBROUTINE select_singval_cmp.
!
    if ( two_stage ) then
!
!       STEP1A : REDUCE THE MATRIX mat TO BIDIAGONAL FORM BY A TWO-STAGE ALGORITHM.
!       mat IS FIRST REDUCED TO TRIANGULAR FORM BY A QR OR LQ FACTORIZATION
!       IN A FIRST STAGE. THE TRIANGULAR MATRIX IS REDUCED TO BIDIAGONAL FORM IN A
!       SECOND STAGE.
!       THE INTERMEDIATE BIDIAGONAL FORM OF mat IS STORED IN d AND e.
!       THE ORTHOGONAL MATRICES USED TO REDUCE mat TO BIDIAGONAL FORM IN THE TWO-STAGE 
!       ALGORITHM ARE STORED IN mat, rlmat, tauo, tauq AND taup.
!       FINALLY, THE SINGULAR VALUES ARE COMPUTED BY THE DQDS ALGORITHM.
!
        if ( gen_q ) then
!
            call select_singval_cmp( mat, rlmat, nsing, s, failure=failure1, sort=d_min, d=d, e=e,   &
                                     abstol=abstol, tauq=tauq, taup=taup, scaling=scaling, dqds=dqds )
!
        else
!
            call select_singval_cmp( mat, rlmat, nsing, s, failure=failure1, sort=d_min, d=d, e=e,    &
                                     abstol=abstol, tauo=tauo, tauq=tauq, taup=taup, scaling=scaling, &
                                     dqds=dqds )
!
        end if
!
    else
!
!       STEP1B : REDUCE THE MATRIX mat TO BIDIAGONAL FORM WITHOUT A QR OR LQ FACTORIZATION.
!       THE INTERMEDIATE BIDIAGONAL FORM OF mat IS STORED IN d AND e.
!       THE ORTHOGONAL MATRICES USED TO REDUCE mat TO BIDIAGONAL FORM
!       ARE STORED IN mat, tauq AND taup.
!       FINALLY, THE SINGULAR VALUES ARE COMPUTED BY THE DQDS ALGORITHM.
!
        call select_singval_cmp( mat, nsing, s, failure=failure1, sort=d_min, d=d, e=e, &
                                 abstol=abstol, tauq=tauq, taup=taup, scaling=scaling,  &
                                 dqds=dqds )
!
    end if
!
!   ON EXIT OF select_singval_cmp :
!
!   failure= false :  INDICATES SUCCESSFUL EXIT IN THE BISECTION OR DQDS ALGORITHM.
!   failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                     THAT FULL ACCURACY WAS NOT ATTAINED IN THE
!                     COMPUTATION OF THE SINGULAR VALUES OF mat.
!                     THE SIGN OF THE INCORRECT SINGULAR VALUES IS SET TO NEGATIVE.
!
!   THE SINGULAR VALUES ARE STORED IN s IN DECREASING ORDER (IF sort='d'). HERE, THIS
!   IS REQUIRED BY SUBROUTINE bd_inviter2 IN THE SECOND STEP.
!
!   THE INTERMEDIATE BIDIAGONAL FORM OF mat IS STORED IN d AND e.
!   THE ORTHOGONAL MATRICES USED TO REDUCE mat TO BIDIAGONAL FORM
!   ARE STORED IN PACKED FORM IN mat, tauq, taup AND ALSO rlmat AND tauo IF A
!   TWO-STAGE ALGORITHM HAS BEEN USED.
!
!   nsing GIVES THE NUMBER OF SINGULAR VALUES REALLY COMPUTED BY select_singval_cmp.
!   HERE, nsing IS ALWAYS EQUAL TO k AS ALL THE SINGULAR VALUES ARE COMPUTED.
!   THE nsing LARGEST SINGULAR VALUES ARE STORED IN THE FIRST nsing POSITIONS OF s
!   IN DECREASING ORDER IF sort='d'.
!
    if ( failure1 ) then
!
!       IF THE DQDS ALGORITHM FAILED, USE THE BISECTION ALGORITHM AS A BACKUP.
!
        call bd_singval( d, e, nsing, s, failure1, sort=d_min, &
                         abstol=abstol, scaling=scaling )
!
    end if
!
!   STEP2 : COMPUTE THE FIRST nsvd LEADING SINGULAR VECTORS OF mat .
!
    if ( .not.failure1 ) then
!
!       COMPUTE THE FIRST nsvd LEFT AND RIGHT SINGULAR VECTORS OF mat BY maxiter
!       INVERSE ITERATIONS ON THE INTERMEDIATE BIDIAGONAL MATRIX
!       d_e AND BACK-TRANSFORMATION WITH SUBROUTINE bd_inviter2 .
!
!       ON ENTRY, PARAMETER s CONTAINS SELECTED SINGULAR VALUES OF THE BIDIAGONAL MATRIX d_e (OR mat).
!       THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
        if ( two_stage ) then
!
            if ( gen_q ) then
!
                call bd_inviter2( mat, tauq, taup, rlmat, d, e, s(:nsvd), u, v,   &
                                  failure=failure2, maxiter=maxiter, ortho=ortho, &
                                  backward_sweep=backward_sweep, scaling=scaling, &
                                  initvec=initvec  )
!
            else
!
                call bd_inviter2( mat, tauq, taup, rlmat, d, e, s(:nsvd), u, v, &
                                  failure=failure2, tauo=tauo, maxiter=maxiter, &
                                  ortho=ortho, backward_sweep=backward_sweep,   &
                                  scaling=scaling, initvec=initvec )
!
            end if
!
        else
!
            call bd_inviter2( mat, tauq, taup, d, e, s(:nsvd), u, v,      &
                              failure=failure2, maxiter=maxiter,          &
                              ortho=ortho, backward_sweep=backward_sweep, &
                              scaling=scaling, initvec=initvec )
!
        end if
!
!       ON EXIT OF bd_inviter2 :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT IN THE INVERSE ITERATION ALGORITHM.
!       failure= true  :  INDICATES THAT THE ALGORITHM FAILS TO CONVERGE FOR SOME SINGULAR VECTORS.
!
!       THE SINGULAR VECTORS OF THE BIDIAGONAL FORM d_e ARE COMPUTED USING maxiter INVERSE ITERATIONS
!       FOR ALL THE SINGULAR VALUES AT ONE STEP. THE SINGULAR VECTORS OF d_e ARE THEN ORTHOGONALIZED BY 
!       THE MODIFIED GRAM-SCHMIDT OR QR ALGORITHM IF THE SINGULAR VALUES ARE NOT WELL SEPARATED.
!       THE SINGULAR VECTORS OF mat ARE FINALLY COMPUTED BY BLOCKED BACK-TRANSFORMATIONS.
!       ON EXIT, u AND v CONTAIN THE FIRST nsvd LEFT AND RIGHT SINGULAR VECTORS OF mat, RESPECTIVELY.
!
    else
!
!       SINGULAR VECTORS CANNOT BE COMPUTED, SET FAILURE INDICATORS.
!
        failure2 = true
!
        u(:m,:nsvd) = nan()
        v(:n,:nsvd) = nan()
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( d, e, tauq, taup )
!
    if ( two_stage ) then
!
        if ( gen_q ) then
            deallocate( rlmat )
        else
            deallocate( rlmat, tauo )
        end if
!
    end if
!
!   SET FAILURE INDICATOR.
!
    failure = failure1 .or. failure2
!
    if ( present( failure_dqds ) ) then
!
        failure_dqds = failure1
!
    end if
!
!   SORT THE SINGULAR VALUES AND VECTORS IF REQUIRED.
!
    if ( sort_def==a_min .and. .not.failure1 ) then
!
        call svd_sort( a_min, s(:nsvd), u(:m,:nsvd), v(:n,:nsvd) )
!
    end if
!
!
! END OF SUBROUTINE svd_cmp7
! __________________________
!
    end subroutine svd_cmp7
!
! =========================================================================================
!
    subroutine svd_cmp8( mat, s, u, v, failure, sort, maxiter, ortho, backward_sweep, &
                         scaling, initvec, reortho, failure_bd, failure_dqds )
!
! Purpose
! _______
!
!   SVD_CMP8 computes a full or partial Singular Value Decomposition (SVD) of a real
!   m-by-n matrix MAT with m>=n. The full SVD is written:
!
!                       MAT = U * SIGMA * V'
!
!   where SIGMA is an m-by-n matrix which is zero except for its
!   n-diagonal elements, U is an m-by-m orthogonal matrix, and
!   V is an n-by-n orthogonal matrix. The diagonal elements of SIGMA
!   are the singular values of MAT; they are real and non-negative.
!   The columns of U and V are, respectively, the left and right
!   singular vectors of MAT.
!
!   The routine returns the first n singular values and the associated
!   left and right singular vectors (e.g. the left and right singular vectors of MAT
!   in the thin SVD of MAT) or a truncated partial SVD at the user option.
!
!   MAT is first reduced to bidiagonal form B with the Ralha-Barlow one-sided
!   bidiagonalization algorithm, see the references (7) and (8).
!   In addition, if m is much larger than n, a preliminary QR factorization is applied
!   to MAT and the one-sided bidiagonalization algorithm is applied to the triangular factor
!   of this QR factorization in order to speedup the computations.
!
!   The singular values and vectors of B are then computed by the dqds and inverse iteration methods
!   applied to B, see the references (1), (3) and (4).
!
!   The singular vectors of MAT are finally computed by a blocked back-transformation algorithm.
!
!
! Arguments
! _________
!
!   MAT            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                  On entry, the m-by-n matrix MAT.
!
!                  On exit, MAT is overwritten as it is used as workspace inside the subroutine.
!
!   S              (OUTPUT) real(stnd), dimension(:)
!                  The singular values of MAT.
!
!                  The size of S must verify: size( S ) = min(m,n) .
!
!   U              (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed first nsvd columns of U, e.g., the left singular vectors
!                  associated with the first nsvd largest singular values of MAT.
!
!                  The left singular vector associated with the singular value S(j) is stored
!                  in the j-th column of U.
!
!                  The shape of V must verify: 
!
!                  - size( U, 1 ) = size( MAT, 1 )    = m ,
!                  - size( U, 2 ) = nsvd <= size( S ) = min(m,n) .
!
!   V              (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed first nsvd columns of V, e.g., the right singular vectors
!                  associated with the first nsvd largest singular values of MAT.
!
!                  The right singular vector associated with the singular value S(j) is stored
!                  in the j-th column of V.
!
!                  The shape of V must verify: 
!
!                  - size( V, 1 ) = size( MAT, 2 ) = n ,
!                  - size( V, 2 ) = nsvd <= size( S ) = min(m,n) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false indicates successful exit.
!                  - FAILURE = true indicates that the dqds algorithm failed to converge for some
!                    singular values or that some singular vectors failed to converge in MAXITER
!                    inverse iterations. If the dqds algorithm failed to compute some singular values,
!                    the sign of the incorrect singular values in S is set to negative and
!                    arguments U and V are filled with a quiet NAN in output of the subroutine.
!
!   SORT           (INPUT, OPTIONAL) character
!                  Sort the singular values into ascending order if SORT = 'A' or 'a',
!                  or into descending order if SORT = 'D' or 'd'. The singular vectors
!                  are rearranged accordingly.
!
!                  The default is to sort the singular values and vectors into descending order.
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed for computing singular vectors.
!
!                  By default, 2 inverse iterations are performed for all the singular vectors.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, all the singular vectors are orthogonalized by the
!                    Modified Gram-Schmidt or QR algorithm in the inverse iteration algorithm;
!                  - ORTHO=false, the singular vectors are not orthogonalized by 
!                    the Modified Gram-Schmidt or QR algorithm in the inverse iteration algorithm.
!
!                  The default is to orthogonalize the singular vectors only for the singular values,
!                  which are not well-separated.
!
!   BACKWARD_SWEEP (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - BACKWARD_SWEEP=true and the singular vectors are orthogonalized by
!                    the modified Gram-Schmidt algorithm in the inverse iteration algorithm,
!                    a backward sweep of the modified Gram-Schmidt algorithm is also performed;
!                  - BACKWARD_SWEEP=false, a backward sweep is not performed.
!
!                  The default is not to perform a backward sweep of the modified Gram-Schmidt
!                  algorithm.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the bidiagonal matrix B is scaled before computing
!                    the singular values and vectors;
!                  - SCALING=false, the bidiagonal matrix B is not scaled.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INITVEC=true, Fernando vectors are used to start the inverse iteration process
!                    for computing the singular vectors of the bidiagonal matrix BD (e.g. the eigenvectors of the
!                    associated Golub-Kahan tridiagonal matrix);
!                  - INITVEC=false, random uniform starting vectors are used.
!
!                  The default is to use Fernando starting vectors if the singular values are well-separated and
!                  the Golub-Kahan form of the input bidiagonal matrix is unreduced, and random uniform
!                  starting vectors otherwise.
!
!   REORTHO        (INPUT, OPTIONAL) logical(lgl)
!                  If the optional argument REORTHO is set to true, reorthogonalization is performed
!                  (when needed) in the Ralha-Barlow one-sided bidiagonalization of MAT.
!                  If this optional logical argument is set to false, reorthogonalization is never
!                  performed, which can result in a loss of orthogonality for the left
!                  singular vectors of MAT if MAT has a large condition number.
!
!                  See description of the BD_CMP2 subroutine for further details.
!
!                  The default is REORTHO = true.
!
!   FAILURE_BD     (OUTPUT, OPTIONAL) logical(lgl)
!                  On exit:
!
!                  - FAILURE_BD = false :  indicates that maximum accuracy was obtained in the
!                    Ralha-Barlow one-sided bidiagonalization of MAT;
!                  - FAILURE_BD = true  :  indicates that MAT is nearly singular and some
!                    loss of orthogonality can be expected in the Ralha-Barlow
!                    bidiagonalization algorithm.
!
!                  See description of the BD_CMP2 subroutine for further details.
!
!   FAILURE_DQDS   (OUTPUT, OPTIONAL) logical(lgl)
!                  On exit:
!
!                  - FAILURE_DQDS = false :  indicates successful exit and the dqds algorithm converged for all
!                    the computed singular values to the desired accuracy;
!                  - FAILURE_DQDS = true  :  indicates that some or all of the singular values failed to converge
!                    or were not computed. This is generally caused by unexpectedly inaccurate arithmetic.
!                    In that case, U and V are filled with a quiet NAN and the sign of the incorrect
!                    singular values in S is set to negative in output of the subroutine.
!                
!
! Further Details
! _______________
!
!   Computing the SVD of a m-by-n matrix MAT, with m>=n, in subroutine SVD_CMP8 consists of three steps:
!
!       1) reduction of the rectangular matrix to bidiagonal form B via the Ralha-Barlow
!          one-sided bidiagonalization algorithm, see the references (7) and (8);
!       2) computation of the SVD of the bidiagonal matrix B with the dqds algorithm for the singular values
!          and inverse iteration for all or the leading singular vectors at the user option;
!       3) computation of the singular vectors of MAT from those of its bidiagonal form B by a blocked
!          back-transformation algorithm, see the references (1) and (2).
!
!   Note that if m is much larger than n, the rectangular matrix is first reduced to
!   upper triangular form by a preliminary QR factorization and the three-step reduction algorithm
!   described above is applied to the resulting triangular factor. The singular vectors of the original rectangular
!   matrix are then obtained from those of the triangular factor in the QR factorization by another blocked
!   back-transformation algorithm.
!
!   Concerning the inverse iteration algorithm used here, note that a first estimate of the singular vectors
!   is computed by the Fernando method applied to the tridiagonal Golub-Kahan matrix associated
!   with the bidiagonal matrix B (see the references (5) and (6) for details) for the singular values
!   which are well-separated and if the Golub-Kahan form of the input bidiagonal matrix is unreduced.
!   For the other singular values, a random start is used as a first estimate of the singular vectors
!   as in the standard inverse-iteration algorithm.
!
!   The singular vectors of B are then computed or refined using inverse iteration on the
!   tridiagonal Golub-Kahan matrix for all the singular values at one step.
!
!   By default, the singular vectors of B are then orthogonalized by the Modified Gram-Schmidt
!   or QR algorithm only if the singular values are not well-separated.
!
!   Excepted for the computation of the singular values by the dqds algorithm, all the steps of the SVD algorithm used
!   here (e.g., preliminary QR factorization, reduction to bidiagonal form B, computation of the singular vectors 
!   vectors of B, computation of the singular vectors of MAT by blocked back-transformation algorithms are parallelized
!   if OPENMP is used.
!
!   For more details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!            Matrix Computations.
!            3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!            Solving least square problems.
!            Prentice-Hall.
!
!   (3) Parlett, B.N., and Marques, O.A., 2000:
!           An implementation of the dqds algorithm (positive case).
!           Linear Algebra Appl., Volume 309, pp. 217-259.
!
!   (4) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!   (5) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (6) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (7) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!            A new stable bidiagonal reduction algorithm.
!            Linear Algebra Appl., No 397, pp. 35-84.
!
!   (8) Bosner, N., and Barlow, J.L., 2007:
!            Block and Parallel versions of one-sided bidiagonalization.
!            SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error
    use String_Procedures, only : to_lower
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout), dimension(:,:) :: mat
    real(stnd),  intent(out),   dimension(:)   :: s
    real(stnd),  intent(out),   dimension(:,:) :: u, v
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: reortho, ortho, backward_sweep, scaling, initvec
    logical(lgl), intent(out), optional :: failure_bd, failure_dqds
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: abstol
    real(stnd), dimension(:,:), allocatable :: rmat, p
    real(stnd), dimension(:),   allocatable :: d, e, tauo
!
    integer(i4b) :: m, n, mnthr, mnthr_nsvd, nsvd, nsing
    integer      :: iok
!
    logical(lgl) :: two_stage, gen_p, gen_q, dqds, failure1, failure2
!
    character    :: sort_def
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='svd_cmp8'
!
    character,  parameter :: a_min=achar(97), d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size(mat,1),i4b), &
                   int(size(u,1),i4b),   &
                   name_proc  )
!
    n = assert_eq( int(size(mat,2),i4b), &
                   int(size(v,1),i4b),   &
                   int(size(s),i4b),     &
                   name_proc  )
!
    nsvd = assert_eq( int(size(u,2),i4b), &
                      int(size(v,2),i4b), &
                      name_proc  )
!
    call assert( logical( nsvd<=n, lgl ), &
                 logical( n<=m, lgl ),    &
                 name_proc )
!
!   TEST THE OPTIONAL ARGUMENTS.
!
    sort_def = d_min
!
    if ( present( sort ) ) then
!
        sort_def = to_lower( sort )
!
    end if
!
    if ( present( failure_bd ) ) then
!
        failure_bd = false
!
    end if
!
    if ( present( failure_dqds ) ) then
!
        failure_dqds = false
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
!   CHOOSE TUNING PARAMETERS FOR THE ALGORITHM.
!
!   DETERMINE HOW SINGULAR VALUES ARE COMPUTED. IF
!   dqds IS SET TO TRUE THE DQDS ALGORITHM IS USED
!   INSTEAD OF BISECTION IN select_singval_cmp3
!   SUBROUTINE.
!
    dqds = true
!
!   NEXT CHOOSE TUNING PARAMETERS FOR THE BISECTION ALGORITHM WHICH IS USED
!   AS A BACKUP IF THE DQDS ALGORITHM FAILED OR IF dqds IS SET TO FALSE ABOVE.
!   SINGULAR VALUES WILL BE COMPUTED MOST ACCURATELY WHEN abstol IS SET
!   TO THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD (E.G., sqrt(safmin) )
!   IN THE BISECTION ALGORITHM.
!
    abstol = sqrt( safmin )
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat TO BIDIAGONAL FORM. IF m EXCEEDS mnthr ,
!   A QR FACTORIZATION IS USED FIRST TO REDUCE THE
!   m-BY-n MATRIX mat TO A TRIANGULAR FORM.
!    
    mnthr = int( real( n, stnd )*thr_bd, i4b )
!
    two_stage = m>=mnthr
!    two_stage = true
!    two_stage = false
!
!   DETERMINE CROSSOVER POINT FOR THE BACK-TRANSFORMATION
!   ALGORITHM WHEN COMPUTING RIGHT SINGULAR VECTORS. IF
!   nsvd EXCEEDS mnthr_nsvd, THE RIGHT ORTHOGONAL MATRIX p
!   OF THE ONE-SIDED BIDIAGONAL REDUCTION ALGORITHM IS GENERATED
!   AND BACK-TRANSFORMATION IS DONE BY A MATRIX MULTIPLICATION.
!
    mnthr_nsvd = int( real( n, stnd )*thr_bt, i4b )
!
    gen_p = nsvd>=mnthr_nsvd
!    gen_p = true
!    gen_p = false
!
!   DETERMINE CROSSOVER POINT FOR THE BACK-TRANSFORMATION
!   ALGORITHM WHEN COMPUTING LEFT SINGULAR VECTORS. IF m
!   EXCEEDS mnthr (E.G. A PRELIMINARY QR IS DONE) AND nsvd
!   EXCEEDS mnthr_nsvd, THE ORTHOGONAL MATRIX q OF THE QR
!   FACTORIZATION IS GENERATED AND THE LEFT SINGULAR VECTORS
!   ARE COMPUTED BY A MATRIX MULTIPLICATION.
!
    gen_q = nsvd>=mnthr_nsvd .and. two_stage
!    gen_q = true
!    gen_q = false
!
!   ALLOCATE WORK ARRAYS.
!
    if ( two_stage ) then
!
        if ( gen_q ) then
            allocate( rmat(n,n), d(n), e(n), p(n,n), stat=iok )
        else
            allocate( rmat(n,n), d(n), e(n), tauo(n), p(n,n), stat=iok )
        end if
!
    else
!
        allocate( d(n), e(n), p(n,n), stat=iok )
!
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL m-BY-n MATRIX mat
!   IS WRITTEN
!
!                       a = U * S * V**(t)
!
!   WHERE S IS AN m-by-n MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   min(m,n) DIAGONAL ELEMENTS, U IS AN n-BY-n ORTHOGONAL MATRIX, AND
!   V IS AN m-BY-m ORTHOGONAL MATRIX. THE DIAGONAL ELEMENTS OF S
!   ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!   STEP1 : COMPUTE THE n LARGEST SINGULAR VALUES OF mat BY THE RHALA-BARLOW ONE-SIDED
!   BIDIAGONAL REDUCTION ALGORITHM APPLIED TO mat, THE DQDS ALGORITHM APPLIED TO THE
!   RESULTING BIDIAGONAL MATRIX AND STORE INTERMEDIATE RESULTS WITH SUBROUTINE select_singval_cmp3.
!
    if ( two_stage ) then
!
!       STEP1A : REDUCE THE MATRIX mat TO BIDIAGONAL FORM BY A TWO-STAGE ALGORITHM.
!       mat IS FIRST REDUCED TO TRIANGULAR FORM BY A QR FACTORIZATION IN A FIRST STAGE.
!       THE TRIANGULAR MATRIX IS THEN REDUCED TO BIDIAGONAL FORM IN A SECOND STAGE BY
!       THE RHALA-BARLOW ONE-SIDED BIDIAGONAL REDUCTION ALGORITHM.
!       THE INTERMEDIATE BIDIAGONAL FORM OF mat IS STORED IN d AND e.
!       THE ORTHOGONAL MATRICES USED TO REDUCE mat TO BIDIAGONAL FORM IN THE TWO-STAGE 
!       ALGORITHM ARE STORED IN mat, rmat, tauo AND p.
!       FINALLY, THE SINGULAR VALUES ARE COMPUTED BY THE DQDS ALGORITHM.
!
        if ( gen_q ) then
!
            call select_singval_cmp3( mat, rmat, nsing, s, failure=failure1, sort=d_min, d=d, &
                                      e=e, abstol=abstol, p=p, gen_p=gen_p, reortho=reortho,  &
                                      scaling=scaling, dqds=dqds, failure_bd=failure_bd )
!
        else
!
            call select_singval_cmp3( mat, rmat, nsing, s, failure=failure1, sort=d_min, d=d, e=e,  &
                                      abstol=abstol, tauo=tauo, p=p, gen_p=gen_p, reortho=reortho,  &
                                      scaling=scaling, dqds=dqds, failure_bd=failure_bd )
!
        end if
!
    else
!
!       STEP1B : REDUCE THE MATRIX mat TO BIDIAGONAL FORM WITHOUT A QR FACTORIZATION
!       BY THE RHALA-BARLOW ONE-SIDED BIDIAGONAL REDUCTION ALGORITHM.
!       THE INTERMEDIATE BIDIAGONAL FORM OF mat IS STORED IN d AND e.
!       THE ORTHOGONAL MATRICES USED TO REDUCE mat TO BIDIAGONAL FORM
!       ARE STORED IN mat AND p .
!       FINALLY, THE SINGULAR VALUES ARE COMPUTED BY THE DQDS ALGORITHM.
!
        call select_singval_cmp3( mat, nsing, s, failure=failure1, sort=d_min, d=d, e=e, &
                                  abstol=abstol, p=p, gen_p=gen_p, reortho=reortho,      &
                                  scaling=scaling, dqds=dqds, failure_bd=failure_bd )
!
    end if
!
!   ON EXIT OF select_singval_cmp3 :
!
!   failure= false :  INDICATES SUCCESSFUL EXIT IN THE BISECTION OR DQDS ALGORITHM.
!   failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                     THAT FULL ACCURACY WAS NOT ATTAINED IN THE
!                     COMPUTATION OF THE SINGULAR VALUES OF mat.
!                     THE SIGN OF THE INCORRECT SINGULAR VALUES IS SET TO NEGATIVE.
!
!   failure_bd= false :  INDICATES SUCCESSFUL EXIT IN THE BIDIAGONAL REDUCTION ALGORITHM.
!   failure_bd= true  :  INDICATES THAT a IS NEARLY SINGULAR AND SOME LOSS OF ORTHOGONALITY
!                        CAN BE EXPECTED IN THE RALHA-BARLOW ONE-SIDED BIDIAGONALIZATION
!                        OF mat FOR THE LAST LEFT SINGULAR VECTORS.
!
!   THE SINGULAR VALUES ARE STORED IN s IN DECREASING ORDER (IF sort='d'). HERE, THIS
!   IS REQUIRED BY SUBROUTINE bd_inviter2 IN THE SECOND STEP.
!
!   THE INTERMEDIATE BIDIAGONAL FORM OF mat IS STORED IN d AND e.
!   THE ORTHOGONAL MATRICES USED TO REDUCE mat TO BIDIAGONAL FORM
!   ARE STORED IN PACKED FORM IN mat, p AND ALSO rmat AND tauo IF A
!   TWO-STAGE ALGORITHM HAS BEEN USED.
!
!   nsing GIVES THE NUMBER OF SINGULAR VALUES REALLY COMPUTED BY select_singval_cmp.
!   HERE, nsing IS ALWAYS EQUAL TO n AS ALL THE SINGULAR VALUES ARE COMPUTED.
!   THE nsing LARGEST SINGULAR VALUES ARE STORED IN THE FIRST nsing POSITIONS OF s
!   IN DECREASING ORDER IF sort='d'.
!
    if ( failure1 ) then
!
!       IF THE DQDS ALGORITHM FAILED, USE THE BISECTION ALGORITHM AS A BACKUP.
!
        call bd_singval( d, e, nsing, s, failure1, sort=d_min, &
                         abstol=abstol, scaling=scaling )
!
    end if
!
!   STEP2 : COMPUTE THE FIRST nsvd LEADING SINGULAR VECTORS OF mat .
!
    if ( .not.failure1 ) then
!
!       COMPUTE THE FIRST nsvd LEFT AND RIGHT SINGULAR VECTORS OF mat BY maxiter
!       INVERSE ITERATIONS ON THE INTERMEDIATE BIDIAGONAL MATRIX
!       d_e AND BACK-TRANSFORMATION WITH SUBROUTINE bd_inviter2 .
!
!       ON ENTRY, PARAMETER s CONTAINS SELECTED SINGULAR VALUES OF THE BIDIAGONAL MATRIX d_e (OR mat).
!       THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
        if ( two_stage ) then
!
            if ( gen_q ) then
!
                call bd_inviter2( mat, rmat, p, d, e, s(:nsvd), u, v,             &
                                  failure=failure2, maxiter=maxiter, ortho=ortho, &
                                  backward_sweep=backward_sweep, scaling=scaling, &
                                  initvec=initvec  )
!
            else
!
                call bd_inviter2( mat, rmat, p, d, e, s(:nsvd), u, v,           &
                                  failure=failure2, tauo=tauo, maxiter=maxiter, &
                                  ortho=ortho, backward_sweep=backward_sweep,   &
                                  scaling=scaling, initvec=initvec )
!
            end if
!
        else
!
            call bd_inviter2( mat, p, d, e, s(:nsvd), u, v,               &
                              failure=failure2, maxiter=maxiter,          &
                              ortho=ortho, backward_sweep=backward_sweep, &
                              scaling=scaling, initvec=initvec )
!
        end if
!
!       ON EXIT OF bd_inviter2 :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT IN THE INVERSE ITERATION ALGORITHM.
!       failure= true  :  INDICATES THAT THE ALGORITHM FAILS TO CONVERGE FOR SOME SINGULAR VECTORS.
!
!       THE SINGULAR VECTORS OF THE BIDIAGONAL FORM d_e ARE COMPUTED USING maxiter INVERSE ITERATIONS
!       FOR ALL THE SINGULAR VALUES AT ONE STEP. THE SINGULAR VECTORS OF d_e ARE THEN ORTHOGONALIZED BY 
!       THE MODIFIED GRAM-SCHMIDT OR QR ALGORITHM IF THE SINGULAR VALUES ARE NOT WELL SEPARATED.
!       THE SINGULAR VECTORS OF mat ARE FINALLY COMPUTED BY BLOCKED BACK-TRANSFORMATIONS.
!       ON EXIT, u AND v CONTAIN THE FIRST nsvd LEFT AND RIGHT SINGULAR VECTORS OF mat, RESPECTIVELY.
!
    else
!
!       SINGULAR VECTORS CANNOT BE COMPUTED, SET FAILURE INDICATORS.
!
        failure2 = true
!
        u(:m,:nsvd) = nan()
        v(:n,:nsvd) = nan()
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( d, e, p )
!
    if ( two_stage ) then
!
        if ( gen_q ) then
            deallocate( rmat )
        else
            deallocate( rmat, tauo )
        end if
!
    end if
!
!   SET FAILURE INDICATORS.
!
    failure = failure1 .or. failure2
!
    if ( present( failure_dqds ) ) then
!
        failure_dqds = failure1
!
    end if
!
!   SORT THE SINGULAR VALUES AND VECTORS IF REQUIRED.
!
    if ( sort_def==a_min .and. .not.failure1 ) then
!
        call svd_sort( a_min, s(:nsvd), u(:m,:nsvd), v(:n,:nsvd) )
!
    end if
!
!
! END OF SUBROUTINE svd_cmp8
! __________________________
!
    end subroutine svd_cmp8
!
!
! =========================================================================================
!      RANDOMIZED SVD OF A RECTANGULAR MATRIX OR EVD OF A POSITIVE SEMI-DEFINITE MATRIX
! =========================================================================================
!
!
    subroutine rsvd_cmp( mat, s, leftvec, rightvec, failure, niter, nover, ortho, &
                         extd_samp, rng_alg, maxiter, max_francis_steps,          &
                         perfect_shift, bisect, dqds )
!
! Purpose
! _______
!
!   RSVD_CMP computes approximations of the nsvd largest singular values and associated
!   left and right singular vectors of a full m-by-n real matrix MAT using randomized power,
!   subspace or block Krylov iterations.
!
!   nsvd is the target rank of the partial Singular Value Decomposition (SVD), which is sought, 
!   and is equal to the size of the output real vector argument S, i.e., nsvd = size( S ).
!
!   
! Arguments
! _________
!
!   MAT                 (INPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!                       MAT is not modified by the routine.
!
!   S                   (OUTPUT) real(stnd), dimension(:)
!                       On exit, S(:) contains the first top nsvd singular values of MAT.
!                       The singular values are given in decreasing order and are
!                       positive or zero.
!
!                       The size of S must verify:
!                       size( S ) = nsvd <= min( size(MAT,1) , size(MAT,2) ) .
!
!   LEFTVEC             (OUTPUT) real(stnd), dimension(:,:)
!                       On exit, the computed nsvd top left singular vectors. The left singular
!                       vector associated with the singular value S(j) is stored in the j-th column
!                       of LEFTVEC.
!
!                       The shape of LEFTVEC must verify:
!
!                       - size( LEFTVEC, 1 ) = size( MAT, 1 ) = m ,
!                       - size( LEFTVEC, 2 ) = size( S ) = nsvd .
!
!   RIGHTVEC            (OUTPUT) real(stnd), dimension(:,:)
!                       On exit, the computed nsvd top right singular vectors. The right singular
!                       vector associated with the singular value S(j) is stored in the j-th column of
!                       RIGHTVEC.
!
!                       The shape of RIGHTVEC must verify:  
!
!                       - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ,
!                       - size( RIGHTVEC, 2 ) = size( S ) = nsvd .
!
!   FAILURE             (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit, if the optional logical argument FAILURE is present, a test of the
!                       accuracy of the computed singular triplets is performed and in that
!                       case:
!
!                       - FAILURE = false : indicates successful exit;
!                       - FAILURE = true  : indicates that some singular values and vectors of MAT
!                         failed to converge in NITER iterations.
!
!                       If FAILURE = true on exit, results are still useful, but some of the
!                       approximated singular triplets have a poor accuracy.
!
!   NITER               (INPUT, OPTIONAL) integer(i4b)
!                       The number of randomized power, subspace or block Krylov iterations performed
!                       in the subroutine for computing the top nsvd singular triplets. NITER must be
!                       positive or null.
!
!                       By default, 5 randomized power, subspace or block Krylov iterations are performed.
!
!   NOVER               (INPUT, OPTIONAL) integer(i4b)
!                       The oversampling size used in the randomized power, subspace or block Krylov
!                       iterations for computing the top nsvd singular triplets.
!
!                       NOVER must be positive or null and verifies the relationship:
!
!                       - NOVER + size( S ) <= min( size(MAT,1) , size(MAT,2) )
!
!                       and is adjusted if necessary to verify this relationship in all cases.
!
!                       See Further Details and the cited references for the meaning and usefulness
!                       of the oversampling size in the randomized power, subspace or block Krylov iterations.
!
!                       By default, the oversampling size is set to 10.
!
!   RNG_ALG             (INPUT, OPTIONAL) integer(i4b)
!                       On entry, a scalar integer to select the random (uniform) number generator used to build
!                       the random gaussian test matrix in the randomized SVD algorithm.
!
!                       The possible values are:
!
!                       - ALG=1  : selects the Marsaglia's KISS random number generator;
!                       - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!                       - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!                       - ALG=4  : selects the Mersenne Twister random number generator;
!                       - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!                       - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!                       - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!                       - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!                       - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!                       - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                         random number generator;
!
!                       For other values, the current random number generator and its current state are not changed.
!                       Note further, that, on exit, the current random number generator is not reset to its previous
!                       value before the call to RSVD_CMP.
!
!                       See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   ORTHO               (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if:
!
!                       - ORTHO=true, orthonormalization is carried out between each step of
!                         the power or block Krylov iterations, to avoid loss of
!                         accuracy due to rounding errors. This means that subspace iterations
!                         are used instead of power iterations;
!                       - ORTHO=false, orthonormalization is not performed. 
!
!                       The default is to use orthonormalization, e.g., ORTHO=true.
!
!   EXTD_SAMP           (INPUT, OPTIONAL) logical(lgl)
!                       The optional argument EXTD_SAMP determines if extended sampling
!                       (e.g., block Krylov iterations) is used or not for computing
!                       the top nsvd singular triplets.
!
!                       On entry, if:
!
!                       - EXTD_SAMP=true, block Krylov iterations are used;
!                       - EXTD_SAMP=false, power or subspace iterations are used. 
!
!                       The default is to use power or subspace iterations, e.g., EXTD_SAMP=false.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm, which is used in the
!                       last phase of the randomized algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       The optional argument MAX_FRANCIS_STEPS controls the maximum number
!                       of Francis sets (e.g., QR sweeps) of Givens rotations which must be
!                       saved before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm, which is used
!                       in the last phase of the randomized algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is the minimum of nsvd and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       The optional argument PERFECT_SHIFT determines if a perfect shift strategy
!                       is used in the implicit QR algorithm in order to minimize the number of QR
!                       sweeps in the bidiagonal SVD algorithm, which is used in the last phase of
!                       the randomized algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   For a good introduction to randomized linear algebra, see the references (1) and (2).
!
!   The randomized power or subspace iteration was proposed in (3; see Algorithm 4.4) to compute
!   an orthonormal matrix whose range approximates the range of MAT. An approximate
!   partial SVD can then be computed using the aforementioned orthonormal matrix, see
!   Algorithm 5.1 in (3).
!
!   The randomized block Krylov iterations for computing an approximate partial SVD
!   was proposed in (5; see Algorithm 2). See also the reference (1).
!
!   For further details, on randomized linear algebra, computing low-rank matrix approximations
!   and partial SVD using randomized power, subspace or block Krylov iterations, see:
!
!   (1) Martinsson, P.G., 2019:
!            Randomized methods for matrix computations.
!            arXiv.1607.01649
!
!   (2) Erichson, N.B., Voronin, S., Brunton, S.L., and Kutz, J.N., 2019:
!            Randomized matrix decompositions using R.
!            arXiv.1608.02148
!
!   (3) Halko, N., Martinsson, P.G., and Tropp, J.A., 2011:
!            Finding structure with randomness:
!            probabilistic algorithms for constructing approximate matrix decompositions.
!            SIAM Rev., 53, 217-288.
!
!   (4) Gu, M., 2015:
!            Subspace iteration randomization and singular value problems.
!            SIAM J. Sci. Comput., 37, A1139-A1173.
!
!   (5) Musco, C., and Musco, C., 2015:
!            Randomized block krylov methods for stronger and faster approximate singular value decomposition.
!            In Proceedings of the 28th International Conference on Neural Information Processing Systems,
!            NIPS 15, pages 1396-1404, Cambridge, MA, USA, 2015. MIT Press.
!
!   (6) Li, H.,Linderman, G.C., Szlam, A., Stanton, K.P., Kluger, Y., and Tygert, M., 2017:
!            Algorithm 971: An implementation of a randomized algorithm for principal component analysis.
!            ACM Trans. Math. Softw. 43, 3, Article 28 (January 2017).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, assert_eq, merror
    use Random,          only : random_seed_, normal_random_number3_
    use QR_Procedures,   only : qr_cmp, ortho_gen_qr
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:,:) :: mat
    real(stnd),  intent(out), dimension(:)   :: s
    real(stnd),  intent(out), dimension(:,:) :: leftvec, rightvec
!
    integer(i4b), intent(in), optional :: niter, nover, rng_alg, maxiter, max_francis_steps
!
    logical(lgl), intent(in),  optional :: ortho, extd_samp, perfect_shift, bisect, dqds
    logical(lgl), intent(out), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: err
!
#ifdef _BLAS
    real(stnd), dimension(:,:), allocatable :: q, b, v, k
#else
    real(stnd), dimension(:,:), allocatable :: q, qt, b, bt, k
#endif
    real(stnd), dimension(:),   allocatable :: diagr, beta
!
    integer      :: iok, alg
    integer(i4b) :: m, n, mn, nsvd, np, npe, i, i1, i2, niter2, nover2
!
    logical(lgl) :: do_test, do_ortho, do_extd_samp, failure_svd
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='rsvd_cmp'
!
!   niter0 IS THE DEFAULT NUMBER OF RANDOMIZED POWER, SUBSPACE OR BLOCK KRYLOV ITERATIONS PERFORMED.
!   nover0 IS THE DEFAULT OVERSAMPLING SIZE.
!
    integer(i4b),  parameter :: niter0=5_i4b, nover0=10_i4b
!   
!   eps0 IS THE THRESHOLD FOR THE ACCURACY TEST OF THE SINGULAR TRIPLETS.
!   
    real(stnd), parameter  :: eps0=0.0001_stnd
!   
    character, parameter   :: sort='d'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
!
    m = assert_eq( int(size( mat,     1 ),i4b)  ,   &
                   int(size( leftvec, 1 ),i4b)  ,   &
                   name_proc )
!
    n = assert_eq( int(size( mat,      2 ),i4b)  ,  &
                   int(size( rightvec, 1 ),i4b)  ,  &
                   name_proc )
!
    nsvd = assert_eq( int(size( s ),i4b)           ,  &
                      int(size( leftvec, 2 ),i4b)  ,  &
                      int(size( rightvec, 2 ),i4b) ,  &
                      name_proc )
!
    mn = min( m, n )
!
!   DETERMINE IF A TEST OF SINGULAR TRIPLETS IS REQUIRED BEFORE EXIT.
!
    do_test = present( failure )
!
    if ( do_test ) then
        failure = false
    end if
!!
!   QUICK RETURN IF POSSIBLE.
!
    if ( mn<=0_i4b  .or. nsvd<=0_i4b ) then
!
        return
!
    end if
!
    call assert( logical( nsvd>0_i4b .and. nsvd<=mn, lgl ), name_proc )
!
!   SET OPTIONAL ARGUMENTS.
!
!   DETERMINE THE NUMBER OF POWER, SUBSPACE OR BLOCK KRYLOV ITERATIONS
!   niter2 TO BE PERFORMED.
!
    if ( present(niter) ) then
        niter2 = max( niter, 0_i4b )
    else
        niter2 = niter0
    end if
!
!   DETERMINE THE OVERSAMPLING SIZE nover2 .
!
    if ( present(nover) ) then
        nover2 = max( nover, 0_i4b )
    else
        nover2 = nover0
    end if
!
!   DETERMINE IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE POWER (SUBSPACE) OR BLOCK KRYLOV ITERATIONS, TO AVOID LOSS
!   OF ACCURACY DUE TO ROUNDING ERRORS.
!
    if ( present(ortho) ) then
        do_ortho = ortho
    else
        do_ortho = true
    end if
!
!   DETERMINE IF POWER, SUBSPACE OR BLOCK KRYLOV ITERATIONS ARE USED.
!
    if ( present(extd_samp) .and. niter2>0_i4b ) then
        do_extd_samp = extd_samp
    else
        do_extd_samp = false
    end if
!
!   CHECK VALIDITY OF THE OVERSAMPLING SIZE PARAMETER
!   AND OF THE NUMBER OF ITERATIONS FOR BLOCK KRYLOV METHOD.
!
    np = min( nover2 + nsvd, mn ) 
!
    if ( do_extd_samp ) then
!
       npe = ( niter2 + 1_i4b )*np
!
       if ( npe>mn ) then
!
           npe = np
!
           do_extd_samp = false
!
       end if
!
!       do i = niter2, 0_i4b, -1_i4b
!
!           if ( npe<=mn ) exit
!
!           npe = npe - np           
!
!       end do
!
!       niter2 = i
!
!       do_extd_samp = niter2>0_i4b
!
    else
!
       npe = np
!
    end if
!
!   SELECT AND SET THE RANDOM NUMBER GENERATOR TO BE USED IF REQUIRED.
!
    if ( present(rng_alg) ) then
!
        if ( rng_alg>=1_i4b .and. rng_alg<=10_i4b ) then
!
            alg = rng_alg
!
!           SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
            call random_seed_( alg=alg )
!
!           RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
            call random_seed_()
!
        end if
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    allocate( q(m,npe), b(n,npe), v(npe,npe),  &
              diagr(npe), beta(npe), stat=iok  )
#else
    allocate( q(m,npe), qt(np,m), b(n,npe), bt(np,n), &
              diagr(npe), beta(npe), stat=iok         )
#endif
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-BY-np IID GAUSSIAN MATRIX.
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
    call normal_random_number3_( b(:n,:np) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
!    do i = 1_i4b, np
!
!        call normal_random_number3_( b(:n,i) )
!
!    end do
!
!   COMPUTE RANDOM SAMPLE MATRIX.
!
#ifdef _BLAS
    call gemm( 'N', 'N', m, np, n, one, mat(1_i4b:m,1_i4b:n), m,     &
               b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:m,1_i4b:np), m  )
#else
    q(:m,:np) = matmul( mat(:m,:n), b(:n,:np) )
#endif
!
    if ( do_extd_samp ) then
!
!       USE BLOCK KRYLOV ITERATIONS.
!
!       ALLOCATE WORK ARRAY.
!
        allocate( k(m,np), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE FIRST KRYLOV BLOCK.
!
        k(:m,:np) = q(:m,:np)
!
        i1 = 1_i4b
        i2 = np
!
    end if
!
!   DO POWER OR SUBSPACE ITERATIONS.
!
    do i = 1_i4b, niter2
!
        if ( do_ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
            call qr_cmp( q(:m,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX
!           TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM COLUMN SUBSPACE.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( q(:m,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'T', 'N', n, np, m, one, mat(1_i4b:m,1_i4b:n), m,     &
                   q(1_i4b:m,1_i4b:np), m, zero, b(1_i4b:n,1_i4b:np), n  )
#else
        qt(:np,:m) = transpose( q(:m,:np) )
!
        bt(:np,:n) = matmul( qt(:np,:m), mat(:m,:n) )
!
        b(:n,:np) = transpose( bt(:np,:n) )
#endif
!
        if ( do_ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
            call qr_cmp( b(:n,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX
!           TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM ROW SUBSPACE.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( b(:n,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM ROW SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, np, n, one, mat(1_i4b:m,1_i4b:n), m,     &
                   b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:m,1_i4b:np), m  )
#else
        q(:m,:np) = matmul( mat(:m,:n), b(:n,:np) )
#endif
!
        if ( do_extd_samp ) then
!
!           SAVE CURRENT BLOCK IF BLOCK KRYLOV ITERATIONS ARE USED.
!
            i1 = i1 + np
            i2 = i2 + np
!
            q(:m,i1:i2) = q(:m,:np)
!
        end if
!
    end do
!
    if ( do_extd_samp ) then
!
!       USE BLOCK KRYLOV ITERATIONS.
!
!       RESTORE FIRST KRYLOV BLOCK.
!
        q(:m,:np) = k(:m,:np)
!
!       DEALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
        deallocate( k )
#else
!
        deallocate( k, qt, bt )
!
!       REALLOCATE WORK ARRAYS.
!
        allocate( qt(npe,m), bt(npe,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
#endif
!
    end if
!
!   COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
    call qr_cmp( q(:m,:npe), diagr(:npe), beta(:npe) )
!
!   GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX
!   TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM COLUMN SUBSPACE.
!   q IS ASSUMED OF FULL RANK.
!
    call ortho_gen_qr( q(:m,:npe), beta(:npe) )
!
!   COMPUTE FINAL RANDOM COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
    call gemm( 'T', 'N', n, npe, m, one, mat(1_i4b:m,1_i4b:n), m,      &
               q(1_i4b:m,1_i4b:npe), m, zero, b(1_i4b:n,1_i4b:npe), n  )
!
!   COMPUTE SVD OF THE FINAL RANDOM MATRIX COLUMN PROJECTION.
!
    call svd_cmp( b(:n,:npe), beta(:npe), failure_svd, v=v(:npe,:npe), sort=sort, &
                  maxiter=maxiter, max_francis_steps=max_francis_steps,           &
                  perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!   COMPUTE THE APPROXIMATE TOP nsvd LEFT SINGULAR VECTORS OF mat .
!
    call gemm( 'N', 'N', m, nsvd, npe, one, q(1_i4b:m,1_i4b:npe), m,              &
               v(1_i4b:npe,1_i4b:npe), npe, zero, leftvec(1_i4b:m,1_i4b:nsvd), m  )
#else
    qt(:npe,:m) = transpose( q(:m,:npe) )
!
    bt(:npe,:n) = matmul( qt(:npe,:m), mat(:m,:n) )
!
!   COMPUTE SVD OF THE FINAL RANDOM MATRIX COLUMN PROJECTION.
!
    call svd_cmp( bt(:npe,:n), beta(:npe), failure_svd, v=b(:n,:npe), sort=sort, &
                  maxiter=maxiter, max_francis_steps=max_francis_steps,          &
                  perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!   COMPUTE THE APPROXIMATE TOP nsvd LEFT SINGULAR VECTORS OF mat .
!
    leftvec(:m,:nsvd) = matmul( q(:m,:npe), bt(:npe,:nsvd) )
#endif
!
!   COMPUTE THE APPROXIMATE TOP nsvd RIGHT SINGULAR VECTORS OF mat .
!
    rightvec(:n,:nsvd) = b(:n,:nsvd)
!
!   EXTRACT THE APPROXIMATE TOP nsvd SINGULAR VALUES OF mat .
!
    s(:nsvd) = beta(:nsvd)
!
!   TEST ACCURACY OF THE APPROXIMATE PARTIAL SVD IF REQUIRED.
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION mat*v(:m,:nsvd) - u(:n,:nsvd)*diag(s(:nsvd)),
!       WHERE u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF mat.
!
        q(:m,:nsvd) = leftvec(:m,:nsvd)*spread( s(:nsvd), dim=1, ncopies=m)
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsvd, n, one, mat(1_i4b:m,1_i4b:n), m,              &
                   rightvec(1_i4b:n,1_i4b:nsvd), n, -one, q(1_i4b:m,1_i4b:nsvd), m  )
#else
        q(:m,:nsvd) = matmul( mat(:m,:n), rightvec(:n,:nsvd) ) - q(:m,:nsvd)
#endif
!
        beta(:nsvd) = norm( q(:m,:nsvd), dim=2_i4b )
!
        err = maxval( beta(:nsvd) )/( norm( mat )*real(mn,stnd) )
!
        failure = err>eps0 .or. failure_svd
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    deallocate( q, b, v, diagr, beta )
#else
    deallocate( q, qt, b, bt, diagr, beta )
#endif
!
!
! END OF SUBROUTINE rsvd_cmp
! __________________________
!
    end subroutine rsvd_cmp
!
! =========================================================================================
!
    subroutine rsvd_cmp_fixed_precision( mat, relerr, s, leftvec, rightvec, failure_relerr,    &
                                         failure, niter, blk_size, maxiter_qb, ortho, reortho, &
                                         niter_qb, rng_alg, maxiter, max_francis_steps,        &
                                         perfect_shift, bisect, dqds )
!
! Purpose
! _______
!
!   RSVD_CMP_FIXED_PRECISION computes approximations of the top nsvd singular values and associated
!   left and right singular vectors of a full m-by-n real matrix MAT using randomized power or
!   subspace iterations.
!
!   nsvd is the target rank of the partial Singular Value Decomposition (SVD), which is sought, 
!   and this partial SVD  must have an approximation error which fulfills:
!
!            ||MAT-rSVD||_F <= ||MAT||_F * relerr
!
!   , where rSVD is the computed partial SVD approximation, || ||_F is the Frobenius norm and
!   relerr is a prescribed accuracy tolerance for the relative error of the computed partial SVD
!   approximation, specified in the input real argument RELERR.
!   
!   In other words, nsvd is not known in advance and is determined in the subroutine. This explains
!   why the output real array arguments S, LEFTVEC and RIGHTVEC, which contain the computed singular
!   triplets of the partial SVD on exit, must be declared in the calling program as pointers.
!   
!   On exit, nsvd is equal to the size of the output real pointer argument S, which contains the
!   computed singular values, i.e., nsvd = size( S ) and the relative error in the Frobenius norm
!   of the computed partial SVD approximation is output in argument RELERR.
!
!   RSVD_CMP_FIXED_PRECISION searches incrementally the best (e.g., smallest) partial SVD approximation,
!   which fulfills the prescribed accuracy tolerance for the relative error. More precisely, the rank of
!   the partial SVD approximation is increased progressively of BLK_SIZE by BLK_SIZE until the prescribed
!   accuracy tolerance is satisfied and then improved and adjusted precisely by additional subspace iterations
!   (as specified by the optional NITER_QB integer argument) to obtain the smallest partial SVD approximation,
!   which satisfies the prescribed tolerance.
!
!   Note that the product of the two integer arguments BLK_SIZE and MAXITER_QB (see below for their meaning),
!   BLK_SIZE*MAXITER_QB, determines the maximum allowable rank of the partial SVD approximation,
!   which is sought. In other words, the subroutine will stop the search for the best (e.g., smallest)
!   partial SVD approximation, which fulfills the requested tolerance, if the rank of this partial
!   SVD approximation exceeds BLK_SIZE*MAXITER_QB. In that case, the subroutine will return the current
!   partial SVD approximation (with a rank less or equal to BLK_SIZE*MAXITER_QB).
!
!   In all cases the relative error of the computed partial SVD approximation is output in argument RELERR.
!
!   If, finally, the optional logical argument FAILURE_RELERR is used, it will be set
!   to true if the computed partial SVD does not fulfill the requested relative error specified
!   on entry in the argument RELERR and to false otherwise.
!   
!   
! Arguments
! _________
!
!   MAT                 (INPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       MAT is not modified by the routine.
!
!   RELERR              (INPUT/OUTPUT) real(stnd)
!                       On entry, the requested accuracy tolerance for the relative error of
!                       the computed restricted SVD approximation.
!
!                       The preset RELERR must be greater than 4*epsilon( RELERR ), less than one
!                       and verifies:
!
!                       - RELERR >= 2 * sqrt( epsilon( RELERR )/RELERR )
!
!                       and is forced to be greater than 2*sqrt( epsilon( RELERR )/RELERR )
!                       if this is not the case to avoid loss of accuracy in the algorithm.
!                       See reference (6) for more details.
!
!                       On exit, RELERR contains the relative error of the computed partial SVD
!                       approximation in the Frobenius norm:
!
!                       - RELERR = ||MAT-rSVD||_F / ||MAT||_F
!
!   S                   (OUTPUT) real(stnd), dimension(:), pointer
!                       On exit, S(:) contains estimates of the first top nsvd singular values of MAT.
!                       The singular values are given in decreasing order and are positive or zero.
!
!                       The statut of the pointer S must not be undefined on entry.
!                       If, on entry, the pointer S is already allocated, it will be first deallocated
!                       and then reallocated with the correct size.
!
!                       On exit, the size of the pointer S will verify:
!
!                       - size( S ) = nsvd <= min( size(MAT,1) , size(MAT,2) ) .
!
!   LEFTVEC             (OUTPUT) real(stnd), dimension(:,:), pointer
!                       On exit, the computed nsvd top left singular vectors. The left singular
!                       vector associated with the singular value S(j) is stored in the j-th column
!                       of LEFTVEC.
!
!                       The statut of the pointer LEFTVEC must not be undefined on entry.
!                       If, on entry, the pointer LEFTVEC is already allocated, it will be first deallocated
!                       and then reallocated with the correct shape.
!
!                       On exit, the shape of the pointer LEFTVEC will verify:
!
!                       - size( LEFTVEC, 1 ) = size( MAT, 1 ) = m ,
!                       - size( LEFTVEC, 2 ) = size( S ) = nsvd .
!
!   RIGHTVEC            (OUTPUT) real(stnd), dimension(:,:), pointer
!                       On exit, the computed nsvd top right singular vectors. The right singular
!                       vector associated with the singular value S(j) is stored in the j-th column of
!                       RIGHTVEC.
!
!                       The statut of the pointer RIGHTVEC must not be undefined on entry.
!                       If, on entry, the pointer RIGHTVEC is already allocated, it will be first deallocated
!                       and then reallocated with the correct shape.
!
!                       On exit, the shape of the pointer RIGHTVEC will verify:  
!
!                       - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ,
!                       - size( RIGHTVEC, 2 ) = size( S )      = nsvd .
!
!   FAILURE_RELERR      (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit, if the optional logical argument FAILURE_RELERR is present, it is
!                       set on exit as follows:
!
!                       - FAILURE_RELERR = false : indicates successful exit and the computed partial
!                         SVD fulfills the requested relative error specified on entry in the argument RELERR,
!                       - FAILURE_RELERR = true  : indicates that the computed partial SVD has a relative error
!                         larger than the requested relative error. This means that the requested accuracy tolerance
!                         for the relative error is too small (i.e., RELERR < 2 * sqrt( epsilon( RELERR )/RELERR )
!                         or that the input parameters BLK_SIZE and/or MAXITER_QB have a too small value, given the
!                         distribution of the singular values of MAT, and must be increased to fullfill the preset
!                         accuracy tolerance for the relative error of the partial SVD approximation.
!
!   FAILURE             (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit, if the optional logical argument FAILURE is present, a test of the
!                       accuracy of the computed singular triplets is performed and in that
!                       case:
!
!                       - FAILURE = false : indicates successful exit;
!                       - FAILURE = true  : indicates that some singular values and vectors of MAT
!                         failed to converge in NITER and NITER_QB power and subspace iterations.
!
!                       If FAILURE = true on exit, results are still useful, but some of the approximated
!                       singular triplets have a poor accuracy.
!
!   NITER               (INPUT, OPTIONAL) integer(i4b)
!                       The number of randomized power or subspace iterations performed
!                       in the first phase of the randomized algorithm for computing the preliminary
!                       QB factorization.
!
!                       NITER must be positive or null.
!
!                       By default, 1 randomized power or subspace iteration is performed.
!
!   BLK_SIZE            (INPUT, OPTIONAL) integer(i4b)
!                       On entry, the block size used in the randomized QB factorization, which is used in
!                       the first phase of the randomized SVD algorithm.
!
!                       BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!                       smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!                       By default, BLK_SIZE is set to min( 10, min(m,n) ).
!
!   MAXITER_QB          (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER_QB controls the maximum number of allowed iterations in the randomized
!                       QB algorithm, which is used in the first phase of the randomized SVD algorithm.
!
!                       MAXITER_QB must be set greater or equal to one and less than int( min(m,n)/BLK_SIZE ).
!
!                       By default, MAXITER_QB is set to max( 1, int( min(m,n)/(4*BLK_SIZE) ) ).
!
!   ORTHO               (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if:
!
!                       - ORTHO=true, orthonormalization is carried out between each step of
!                         the power iterations, to avoid loss of accuracy due to rounding errors.
!                         This means that subspace iterations are used instead of power iterations,
!                       - ORTHO=false, orthonormalization is not performed. 
!
!                       The default is to use orthonormalization, e.g., ORTHO=true.
!
!   REORTHO             (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if:
!
!                       - REORTHO=true, a reorthogonalization step is performed to avoid the loss
!                         of orthogonality in the Gram-Schmidt procedure, which is used in the
!                         randomized QB factorization;
!                       - REORTHO=false, a reorthogonalization step is not performed in
!                         the Gram-Schmidt procedure.
!
!                       The default is to use a reorthogonalization step, e.g., REORTHO=true.
!
!   NITER_QB            (INPUT, OPTIONAL) integer(i4b)
!                       The number of subspace iterations performed in the last phase of the QB algorithm
!                       for improving the QB factorization and computes the top nsvd singular triplets
!                       of MAT.
!
!                       NITER_QB must be greater or equal to 0.
!
!                       By default, 2 final subspace iterations are performed.
!
!   RNG_ALG             (INPUT, OPTIONAL) integer(i4b)
!                       On entry, a scalar integer to select the random (uniform) number generator used to build
!                       the random gaussian test matrix in the randomized SVD algorithm.
!
!                       The possible values are:
!
!                       - ALG=1  : selects the Marsaglia's KISS random number generator;
!                       - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!                       - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!                       - ALG=4  : selects the Mersenne Twister random number generator;
!                       - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!                       - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!                       - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!                       - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!                       - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!                       - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                         random number generator;
!
!                       For other values, the current random number generator and its current state are not changed.
!                       Note further, that, on exit, the current random number generator is not reset to its previous
!                       value before the call to RSVD_CMP_FIXED_PRECISION.
!
!                       See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm, which is used in the
!                       last phase of the randomized algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       The optional argument MAX_FRANCIS_STEPS controls the maximum number
!                       of Francis sets (e.g., QR sweeps) of Givens rotations which must be
!                       saved before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm, which is used
!                       in the last phase of the randomized algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is the minimum of nsvd and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       The optional argument PERFECT_SHIFT determines if a perfect shift strategy
!                       is used in the implicit QR algorithm in order to minimize the number of QR
!                       sweeps in the bidiagonal SVD algorithm, which is used in the last phase of
!                       the randomized algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   For a good introduction to randomized linear algebra , see the references (1) and (2).
!
!   The randomized power or subspace iteration was proposed in (3; see Algorithm 4.4) to compute
!   an orthonormal matrix whose range approximates the range of MAT. An approximate
!   partial SVD can then be computed using the aforementioned orthonormal matrix, see
!   Algorithm 5.1 in (3).
!
!   Usually, the problem of low-rank matrix approximation falls into two categories:
!
!   - the fixed-rank problem, where the rank parameter nsvd is given;
!   - the fixed-precision problem, where we seek a partial SVD factorization, rSVD, as small
!     as possible such that
!     
!     ||MAT-rSVD||_F <= eps
!
!     , where eps is a given accuracy tolerance.
!
!   RSVD_CMP_FIXED_PRECISION is dedicated to solve the fixed-precision problem. The fixed-rank problem
!   can be solved by subroutine RSVD_CMP.
!
!   RSVD_CMP_FIXED_PRECISION uses an improved version of the "randQB_FP" algorithm described in the
!   reference (6) to solve the fixed-precision problem.
!
!   For further details, on randomized linear algebra, computing low-rank matrix approximations, partial
!   SVD using randomized power or subspace iterations or solving the fixed-precision problem, see:
!
!   (1) Martinsson, P.G., 2019:
!            Randomized methods for matrix computations.
!            arXiv.1607.01649
!
!   (2) Erichson, N.B., Voronin, S., Brunton, S.L., and Kutz, J.N., 2019:
!            Randomized matrix decompositions using R.
!            arXiv.1608.02148
!
!   (3) Halko, N., Martinsson, P.G., and Tropp, J.A., 2011:
!            Finding structure with randomness:
!            probabilistic algorithms for constructing approximate matrix decompositions.
!            SIAM Rev., 53, 217-288.
!
!   (4) Gu, M., 2015:
!            Subspace iteration randomization and singular value problems.
!            SIAM J. Sci. Comput., 37, A1139-A1173.
!
!   (5) Martinsson, P.G., and Voronin, S., 2016:
!            A randomized blocked algorithm for efficiently computing
!            rank-revealing factorizations of matrices.
!            SIAM J. Sci. Comput., 38:5, S485-S507.
!
!   (6) Yu, W.,  Gu, Y., and Li, Y., 2018:
!            Efficient randomized algorithms for the fixed-precision low-rank
!            matrix approximation.
!            SIAM J. Mat. Ana. Appl., 39:3, 1339-1359.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error, svd_error9
    use Utilities,       only : merror
    use Random,          only : random_seed_, normal_random_number3_
    use QR_Procedures,   only : qr_cmp, ortho_gen_qr
    use Lin_Procedures,  only : triang_solve
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:,:)          :: mat
    real(stnd),  intent(out), dimension(:),   pointer :: s
    real(stnd),  intent(out), dimension(:,:), pointer :: leftvec, rightvec
    real(stnd),  intent(inout)                        :: relerr
!
    integer(i4b), intent(in), optional :: niter, blk_size, maxiter_qb, niter_qb, &
                                          rng_alg, maxiter, max_francis_steps
!
    logical(lgl), intent(in),  optional :: ortho, reortho, perfect_shift, bisect, dqds
    logical(lgl), intent(out), optional :: failure_relerr, failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: err, threshold, threshold2, norm_mat, tmp, tmp2
!
#ifdef _BLAS
    real(stnd), dimension(:,:), allocatable :: q, b, h, r, r2, ti, ti2, yi, bi
#else
    real(stnd), dimension(:,:), allocatable :: q, qt, b, bt, r, r2, ti, ti2, yi, bi
#endif
    real(stnd), dimension(:),   allocatable :: diagr, beta
!
    integer      :: iok, alg
    integer(i4b) :: m, n, mn, nsvd, np, i, i0, i1, i2, j, niter2, blk_size2, maxiter_qb2, niter_qb2
!
    logical(lgl) :: do_test, do_test_relerr, do_ortho, do_reortho, failure_svd
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='rsvd_cmp_fixed_precision'
!
!   niter0 IS THE DEFAULT NUMBER OF RANDOMIZED POWER OR SUBSPACE ITERATIONS PERFORMED FOR COMPUTING
!   THE PRELIMINARY QB FACTORIZATION IN THE FIRST PHASE OF THE ALGORITHM.
!   blk_size0 IS THE DEFAULT BLOCK SIZE USED IN THE RANDOMIZED QB FACTORIZATION.
!   niter_qb0 IS THE DEFAULT NUMBER OF SUBSPACE ITERATIONS PERFORMED FOR IMPROVING THE INITIAL QB FACTORIZATION
!   BEFORE COMPUTING THE PARTIAL SVD FACTORIZATION IN THE SECOND PHASE OF THE ALGORITHM.
!
    integer(i4b),  parameter :: niter0=1_i4b, blk_size0=10_i4b, niter_qb0=2_i4b
!
!   threshold0 IS THE DEFAULT EXTRA THRESHOLD FOR IMPROVING THE ACCURACY OF THE RESULTS
!   IN THE SVD PHASE OF THE QB-SVD ALGORITHM.
!
    real(stnd),  parameter :: threshold0=0.05_stnd
!   
!   eps0 IS THE THRESHOLD FOR THE ACCURACY TEST OF THE SINGULAR TRIPLETS.
!   
    real(stnd), parameter  :: eps0=0.0001_stnd
!   
    character, parameter   :: sort='d'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
!
    m = size( mat, 1 )
!
    n = size( mat, 2 )
!
    mn = min( m, n )
!
!   DEALLOCATE POINTERS ON ENTRY.
!
#ifndef _USE_NAGWARE
    if ( associated( s ) ) then
!
        deallocate( s )
!
    end if
!
    if ( associated( leftvec ) ) then
!
        deallocate( leftvec )
!
    end if
!
    if ( associated( rightvec ) ) then
!
        deallocate( rightvec )
!
    end if
#endif
!
!   DETERMINE THE TESTS WHICH ARE REQUIRED BEFORE EXIT.
!
    do_test = present( failure )
!
    if ( do_test ) then
        failure = false
    end if
!
    do_test_relerr = present( failure_relerr )
!
    if ( do_test_relerr ) then
        failure_relerr = true
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( mn<=0_i4b  ) then
!
        relerr = one
!
        return
!
    end if
!
!   CHECK VALIDITY OF THE REQUESTED TOLERANCE FOR THE RELATIVE
!   ERROR OF THE COMPUTED SVD APPROXIMATION.
!
    tmp = four*ulp
!    tmp = four*epsilon( relerr )
!
    if ( relerr<=tmp .or. relerr>=one ) then
!
!        call merror( name_proc//' : RELERR must be greater than 4.*epsilon( RELERR ) and less than  one !' )
        call merror( name_proc//svd_error9 )
!
    end if
!
!   CHECK OPTIONAL ARGUMENTS.
!
!   DETERMINE THE NUMBER OF POWER OR SUBSPACE ITERATIONS
!   niter2 TO BE PERFORMED IN THE FIRST PHASE OF THE QB FACTORIZATION.
!
    if ( present(niter) ) then
        niter2 = max( niter, 0_i4b )
    else
        niter2 = niter0
    end if
!
!   DETERMINE THE BLOCK SIZE USED IN THE RANDOMIZED QB FACTORIZATION.
!
    if ( present(blk_size) ) then
        blk_size2 = max( blk_size, 1_i4b )
    else
        blk_size2 = blk_size0
    end if
!
    blk_size2 = min( mn, blk_size2 )
!
!   DETERMINE THE MAXIMUM NUMBER OF ALLOWED ITERATIONS
!   IN THE RANDOMIZED QB FACTORIZATION.
!
    if ( present(maxiter_qb) ) then
!
        maxiter_qb2 = max( maxiter_qb, 1_i4b )
!
        tmp = real( mn, stnd )/real( blk_size2, stnd )
        maxiter_qb2 = min( maxiter_qb2, int(tmp,i4b) )
!
    else
!
        tmp = real( mn, stnd )/(four*real( blk_size2, stnd ))
        maxiter_qb2 = max( int(tmp,i4b), 1_i4b )
!
    end if
!
!   np SETS THE MAXIMUM ALLOWABLE RANK OF THE QB FACTORIZATION AND PARTIAL SVD, WHICH ARE SOUGHT.
!
    np = blk_size2*maxiter_qb2
!
!   DETERMINE IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE POWER ITERATIONS (I.E. SHIFTING TO SUBSPACE ITERATIONS),
!   TO AVOID LOSS OF ACCURACY DUE TO ROUNDING ERRORS.
!
    if ( present(ortho) ) then
        do_ortho = ortho
    else
        do_ortho = true
    end if
!
!   DETERMINE IF REORTHOGONALIZATION IS CARRIED OUT AFTER EACH STEP
!   OF THE GRAM-SCHMIDT PROCEDURE TO AVOID LOSS OF ORTHOGONALITY IN
!   THE QB FACTORIZATION.
!
    if ( present(reortho) ) then
        do_reortho = reortho
    else
        do_reortho = true
    end if
!
!   DETERMINE THE NUMBER OF SUBSPACE ITERATIONS
!   niter_qb2 TO BE PERFORMED IN THE LAST PHASE OF THE
!   ALGORITHM TO IMPROVE ESTIMATION OF THE QB FACTORIZATION
!   AND OF THE SINGULAR TRIPLETS.
!
    if ( present(niter_qb) ) then
        niter_qb2 = max( niter_qb, 1_i4b )
    else
        niter_qb2 = niter_qb0
    end if
!
!   SELECT AND SET THE RANDOM NUMBER GENERATOR TO BE USED IF REQUIRED.
!
    if ( present(rng_alg) ) then
!
        if ( rng_alg>=1_i4b .and. rng_alg<=10_i4b ) then
!
            alg = rng_alg
!
!           SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
            call random_seed_( alg=alg )
!
!           RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
            call random_seed_()
!
        end if
!
    end if
!
!   TEST VALIDITY OF THE PRESET RELATIVE ACCURACY TOLERANCE relerr AND
!   COMPUTE THE REQUIRED THRESHOLD.
!
    tmp = two*sqrt( ulp/relerr )
    threshold = max( tmp, relerr )
!
!   DECREASE SLIGHTLY THE PRESCRIBED TOLERANCE FOR BETTER ACCURACY IN
!   THE SECOND PHASE OF THE QB-SVD ALGORITHM.
!
    if ( threshold>=two*threshold0 ) then
        threshold2 = max( tmp, threshold - threshold0 )
    else
        threshold2 = threshold
    end if
!
!   COMPUTE THE FROBENIUS NORM OF THE INPUT MATRIX.
!
    norm_mat = norm( mat(1_i4b:m,1_i4b:n) )
!
    err = one
!
!   ALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    allocate( q(m,np), b(n,np), diagr(np), beta(np), h(n,np),  &
              r(blk_size2,blk_size2), yi(m,blk_size2),         &
              ti(np,blk_size2), bi(blk_size2,n), stat=iok      )
#else
    allocate( q(m,np), qt(np,m), b(n,np), bt(np,n) ,diagr(np),   &
              beta(np), r(blk_size2,blk_size2), yi(m,blk_size2), &
              ti(np,blk_size2), bi(blk_size2,n), stat=iok        )
#endif
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    if ( do_reortho ) then
!
        allocate( r2(blk_size2,blk_size2), ti2(np,blk_size2), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
    end if
!
!   GENERATE A n-BY-np IID GAUSSIAN MATRIX.
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
    call normal_random_number3_( b(:n,:np) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
!    do i = 1_i4b, np
!
!        call normal_random_number3_( b(:n,i) )
!
!    end do
!
!   COMPUTE RANDOM SAMPLE MATRIX.
!
#ifdef _BLAS
    call gemm( 'N', 'N', m, np, n, one, mat(1_i4b:m,1_i4b:n), m,     &
               b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:m,1_i4b:np), m  )
#else
    q(:m,:np) = matmul( mat(:m,:n), b(:n,:np) )
#endif
!
!   DO POWER OR SUBSPACE ITERATIONS.
!
    do i = 1_i4b, niter2
!
        if ( do_ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
            call qr_cmp( q(:m,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX
!           TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM COLUMN SUBSPACE.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( q(:m,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'T', 'N', n, np, m, one, mat(1_i4b:m,1_i4b:n), m,     &
                   q(1_i4b:m,1_i4b:np), m, zero, b(1_i4b:n,1_i4b:np), n  )
#else
        qt(:np,:m) = transpose( q(:m,:np) )
!
        bt(:np,:n) = matmul( qt(:np,:m), mat(:m,:n) )
!
        b(:n,:np) = transpose( bt(:np,:n) )
#endif
!
        if ( do_ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
            call qr_cmp( b(:n,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX.
!           TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM ROW SUBSPACE.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( b(:n,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM ROW SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, np, n, one, mat(1_i4b:m,1_i4b:n), m,     &
                   b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:m,1_i4b:np), m  )
#else
        q(:m,:np) = matmul( mat(:m,:n), b(:n,:np) )
#endif
!
    end do
!
!   FINAL COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
    call gemm( 'T', 'N', n, np, m, one, mat(1_i4b:m,1_i4b:n), m,     &
               q(1_i4b:m,1_i4b:np), m, zero, h(1_i4b:n,1_i4b:np), n  )
#else
    qt(:np,:m) = transpose( q(:m,:np) )
!
    bt(:np,:n) = matmul( qt(:np,:m), mat(:m,:n) )
#endif
!
    i0 = 0_i4b
!
    r(1_i4b:blk_size2,1_i4b:blk_size2) = zero
    ti(1_i4b:np,1_i4b:blk_size2)       = zero
!
    if ( do_reortho ) then
!
        r2(1_i4b:blk_size2,1_i4b:blk_size2) = zero
        ti2(1_i4b:np,1_i4b:blk_size2)       = zero
!
    end if
!
!   COMPUTE THE QB FACTORIZATION.
!
    do i = 1_i4b, maxiter_qb2
!
        i1 = i0 + 1_i4b
        i2 = i0 + blk_size2
!
        if ( i==1_i4b ) then
!
            yi(:m,:blk_size2) = q(:m,i1:i2)
!
        else
!
!           PERFORM BLOCK GRAM-SCHMIDT ORTHOGONALISATION STEP.
!
#ifdef _BLAS
            call gemm( 'T', 'N', i0, blk_size2, n, one, b(1_i4b:n,1_i4b:i0), n,     &
                       b(1_i4b:n,i1:i2), n, zero, ti(1_i4b:np,1_i4b:blk_size2), np  )
!
            call gemm( 'N', 'N', m, blk_size2, i0, -one, q(1_i4b:m,1_i4b:i0), m,   &
                       ti(1_i4b:np,1_i4b:blk_size2), np, one, q(1_i4b:m,i1:i2), m  )
#else
            ti(:i0,:blk_size2) = matmul( bt(:i0,:n), b(:n,i1:i2) )
!
            q(:m,i1:i2) = q(:m,i1:i2) - matmul( q(:m,:i0), ti(:i0,:blk_size2) )
#endif
!            
            yi(:m,:blk_size2) = q(:m,i1:i2)
!
        end if
!
        call qr_cmp( q(:m,i1:i2), diagr(:blk_size2), beta(:blk_size2) )
!
!       SAVE UPPER TRIANGULAR MATRIX FOR LATER USE.
!
        if ( i==1_i4b .or. .not.do_reortho ) then
!
            do j = 1, blk_size2
!
                r(1_i4b:j-1_i4b,j) = q(1_i4b:j-1_i4b,i0+j)
                r(j,j)             = diagr(j)
!                r(j+1_i4b:blk_size2,j) = zero
!
            end do
!
        else
!
            do j = 1, blk_size2
!
                r2(1_i4b:j-1_i4b,j) = q(1_i4b:j-1_i4b,i0+j)
                r2(j,j)             = diagr(j)
!                r2(j+1_i4b:blk_size2,j) = zero
!
            end do
!
        end if
!
!       GENERATE COLUMNS i1 TO i2 OF THE ORTHOGONAL MATRIX OF THE QB FACTORIZATION.
!
        call ortho_gen_qr( q(:m,i1:i2), beta(:blk_size2) )
!
        if ( i/=1_i4b ) then
!
            if ( do_reortho ) then
!
!               REORTHONORMALIZATION IS CARRIED OUT TO AVOID LOSS OF ORTHOGONALITY
!               IN THE BLOCK GRAM-SCHMIDT ORTHOGONALISATION SCHEME USED TO BUILD
!               THE ORTHOGONAL MATRIX OF THE QB DECOMPOSITION.
!
#ifdef _BLAS
                call gemm( 'T', 'N', i0, blk_size2, m, one, q(1_i4b:m,1_i4b:i0), m,     &
                           q(1_i4b:m,i1:i2), m, zero, ti2(1_i4b:np,1_i4b:blk_size2), np )
!
                call gemm( 'N', 'N', m, blk_size2, i0, -one, q(1_i4b:m,1_i4b:i0), m,    &
                           ti2(1_i4b:np,1_i4b:blk_size2), np, one, q(1_i4b:m,i1:i2), m  )
#else
                ti2(:i0,:blk_size2) = matmul( qt(:i0,:m), q(:m,i1:i2) )
!
                q(:m,i1:i2) = q(:m,i1:i2) - matmul( q(:m,:i0), ti2(:i0,:blk_size2) )
#endif
!
!               IMPROVE ESTIMATES OF COLUMNS i1 TO i2 OF THE ORTHOGONAL MATRIX.
!
                call qr_cmp( q(:m,i1:i2), diagr(:blk_size2), beta(:blk_size2) )
!
!               SAVE UPPER TRIANGULAR MATRIX FOR LATER USE.
!
                do j = 1, blk_size2
!
                    bi(1_i4b:j-1_i4b,j)     = q(1_i4b:j-1_i4b,i0+j)
                    bi(j,j)                 = diagr(j)
                    bi(j+1_i4b:blk_size2,j) = zero
!
                end do
!
!               UPDATE COLUMNS i1 TO i2 OF THE ORTHOGONAL MATRIX.
!
                call ortho_gen_qr( q(:m,i1:i2), beta(:blk_size2) )
!
#ifdef _BLAS
                call gemm( 'N', 'N', blk_size2, blk_size2, blk_size2, one,       &
                           bi(1_i4b:blk_size2,1_i4b:blk_size2), blk_size2,       &
                           r2(1_i4b:blk_size2,1_i4b:blk_size2), blk_size2, zero, &
                           r(1_i4b:blk_size2,1_i4b:blk_size2), blk_size2 )
!
                call gemm( 'T', 'N', i0, blk_size2, m, one, q(1_i4b:m,1_i4b:i0), m,  &
                           yi(1_i4b:m,1_i4b:blk_size2), m, one, ti(1_i4b:np,1_i4b:blk_size2), np )
#else
                r(:blk_size2,:blk_size2) = matmul( bi(:blk_size2,:blk_size2), r2(:blk_size2,:blk_size2) )
!
                ti(:i0,:blk_size2) = ti(:i0,:blk_size2) + matmul( qt(:i0,:m), yi(:m,:blk_size2) )
#endif
!
            end if
!
#ifdef _BLAS
!
            call gemm( 'N', 'N', n, blk_size2, i0, -one, b(1_i4b:n,1_i4b:i0), n,  &
                       ti(1_i4b:np,1_i4b:blk_size2), np, one, h(1_i4b:n,i1:i2), n )
#else
            bi(:blk_size2,:i0) = transpose( ti(:i0,:blk_size2) )
            bt(i1:i2,:n)       = bt(i1:i2,:n) - matmul( bi(:blk_size2,:i0), bt(:i0,:n) )
#endif
!
        end if
!
!       COMPUTE ROWS i1 TO i2 OF THE B FACTOR OF QB DECOMPOSITION.
!
#ifdef _BLAS
        bi(:blk_size2,:n) = transpose( h(:n,i1:i2) )
!
        call triang_solve( r(:blk_size2,:blk_size2), bi(:blk_size2,:n), upper=true, trans=true )
!
        b(:n,i1:i2) = transpose( bi(:blk_size2,:n) )
#else
        bi(:blk_size2,:n) = bt(i1:i2,:n)
!
        call triang_solve( r(:blk_size2,:blk_size2), bi(:blk_size2,:n), upper=true, trans=true )
!
        bt(i1:i2,:n) = bi(:blk_size2,:n)
!
        if ( do_reortho ) then
!
            qt(i1:i2,:m) = transpose( q(:m,i1:i2) )
!
        end if
#endif
!
!       COMPUTE RELATIVE ERROR OF THE CURRENT QB FACTORIZATION.
!
        tmp  = norm( bi(:blk_size2,:n) )/norm_mat
        tmp2 = ( err - tmp )*( err + tmp )
        err  = sqrt( max( zero, tmp2 ) )
!
        i0 = i0 + blk_size2
!
!       CHECK IF THE THRESHOLD IS SATISFIED WITH INCLUSION OF THE CURRENT BLOCK.
!
        if ( err<threshold2 ) exit
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    deallocate( h, r, yi, ti, bi )
#else
    deallocate( r, yi, ti, bi )
#endif
!
    if ( do_reortho ) then
!
        deallocate(  r2, ti2 )
!
    end if
!
!   DO FINAL SUBSPACE ITERATIONS FOR IMPROVING THE QB FACTORIZATION.
!
    do i = 1_i4b, niter_qb2
!
#ifndef _BLAS
        b(:n,:i0) = transpose( bt(:i0,:n) )
#endif
!
!       COMPUTE QR DECOMPOSITION OF COLUMN PROJECTION MATRIX.
!
        call qr_cmp( b(:n,:i0), diagr(:i0), beta(:i0) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF COLUMN PROJECTION
!       MATRIX TO OBTAIN AN ORTHONORMAL BASIS OF ROW SUBSPACE.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( b(:n,:i0), beta(:i0) )
!
!       COMPUTE NEXT ROW SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, i0, n, one, mat(1_i4b:m,1_i4b:n), m,     &
                   b(1_i4b:n,1_i4b:i0), n, zero, q(1_i4b:m,1_i4b:i0), m  )
#else
        q(:m,:i0) = matmul( mat(:m,:n), b(:n,:i0) )
#endif
!
!       COMPUTE QR DECOMPOSITION OF ROW PROJECTION MATRIX.
!
        call qr_cmp( q(:m,:i0), diagr(:i0), beta(:i0) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF ROW PROJECTION
!       MATRIX TO OBTAIN AN ORTHONORMAL BASIS OF COLUMN SUBSPACE.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( q(:m,:i0), beta(:i0) )
!
!       COMPUTE NEXT COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'T', 'N', n, i0, m, one, mat(1_i4b:m,1_i4b:n), m,     &
                   q(1_i4b:m,1_i4b:i0), m, zero, b(1_i4b:n,1_i4b:i0), n  )
#else
        qt(:i0,:m) = transpose( q(:m,:i0) )
!
        bt(:i0,:n) = matmul( qt(:i0,:m), mat(:m,:n) )
!
!        b(:n,:i0) = transpose( bt(:i0,:n) )
#endif
!
    end do
!
#ifdef _BLAS
!
!   ALLOCATE WORK ARRAY.
!
    allocate( h(i0,i0), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
#else
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( qt )
!
#endif
!
!   COMPUTE SVD OF THE FINAL COLUMN PROJECTION MATRIX.
!
#ifdef _BLAS
    call svd_cmp( b(:n,:i0), beta(:i0), failure_svd, v=h(:i0,:i0), sort=sort, &
                  maxiter=maxiter, max_francis_steps=max_francis_steps,       &
                  perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
#else
    call svd_cmp( bt(:i0,:n), beta(:i0), failure_svd, v=b(:n,:i0), sort=sort,  &
                  maxiter=maxiter, max_francis_steps=max_francis_steps,        &
                  perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
#endif
!
!   EXTRACT THE EXACT RANK AND RELATIVE ERROR OF THE PARTIAL SVD, WHICH FULFILLS THE THRESHOLD.
!
    diagr(:i0) = (beta(:i0)/norm_mat)**2
!
    err = one - sum( diagr(:i0) )
!
    tmp2 = threshold*threshold
!
    if ( err<tmp2 ) then
!
        do i = i0, 1_i4b, -1_i4b
!
            err = err + diagr(i)
!
            if ( err>=tmp2 ) exit
!
        end do
!
        nsvd = max( i, 1_i4b )
!
        err = one - sum( diagr(:nsvd) )
!
    else
!
       nsvd = i0
!
    end if
!
!   ALLOCATE INPUT POINTER ARGUMENTS FOR STORING THE PARTIAL SVD.
!
    allocate( s(nsvd), leftvec(m,nsvd), rightvec(n,nsvd), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   COMPUTE THE APPROXIMATE TOP nsvd LEFT SINGULAR VECTORS OF mat .
!
#ifdef _BLAS
    call gemm( 'N', 'N', m, nsvd, i0, one, q(1_i4b:m,1_i4b:i0), m,               &
               h(1_i4b:i0,1_i4b:nsvd), i0, zero, leftvec(1_i4b:m,1_i4b:nsvd), m  )
#else
    leftvec(:m,:nsvd) = matmul( q(:m,:i0), bt(:i0,:nsvd) )
#endif
!
!   EXTRACT THE APPROXIMATE TOP nsvd SINGULAR VALUES OF mat .
!
    s(:nsvd) = beta(:nsvd)
!
!   EXTRACT THE APPROXIMATE TOP nsvd RIGHT SINGULAR VECTORS OF mat .
!
    rightvec(:n,:nsvd) = b(:n,:nsvd)
!
!   CHECK ACCURACY OF THE APPROXIMATE PARTIAL SVD FACTORIZATION IF REQUIRED.
!
    if ( do_test_relerr ) then
!
        failure_relerr = relerr/=threshold .or. err>=tmp2
!
    end if
!
!   OUTPUT RELATIVE ERROR OF THE COMPUTED PARTIAL SVD FACTORIZATION
!   IN FROBENIUS NORM.
!
    relerr = sqrt( max( err, zero ) )
!
!   TEST ACCURACY OF THE APPROXIMATE SVD IF REQUIRED.
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION mat*v(:m,:nsvd) - u(:n,:nsvd)*diag(s(:nsvd)),
!       WHERE u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF mat.
!
        q(:m,:nsvd) = leftvec(:m,:nsvd)*spread( s(:nsvd), dim=1, ncopies=m)
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsvd, n, one, mat(1_i4b:m,1_i4b:n), m,              &
                   rightvec(1_i4b:n,1_i4b:nsvd), n, -one, q(1_i4b:m,1_i4b:nsvd), m  )
#else
        q(:m,:nsvd) = matmul( mat(:m,:n), rightvec(:n,:nsvd) ) - q(:m,:nsvd)
#endif
!
        beta(:nsvd) = norm( q(:m,:nsvd), dim=2_i4b )
!
        err = maxval( beta(:nsvd) )/( norm_mat*real(mn,stnd) )
!
        failure = err>eps0 .or. failure_svd
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    deallocate( q, b, h, beta, diagr )
#else
    deallocate( q, b, bt, beta, diagr )
#endif
!
!
! END OF SUBROUTINE rsvd_cmp_fixed_precision
! __________________________________________
!
    end subroutine rsvd_cmp_fixed_precision
!
! =========================================================================================
!
    subroutine reig_pos_cmp( mat, eigval, eigvec, failure, niter, nover, ortho, extd_samp, &
                             use_nystrom, rng_alg, maxiter, max_francis_steps,             &
                             perfect_shift, bisect, dqds )
!
! Purpose
! _______
!
!   REIG_POS_CMP computes approximations of the neig largest eigenvalues and associated
!   eigenvectors of a full n-by-n real symmetric positive semi-definite matrix MAT using randomized
!   power, subspace or block Krylov iterations and, at the user option, the Nystrom method
!   (see below for details).
!
!   neig is the target rank of the partial EigenValue Decomposition (EVD), which is sought,
!   and is equal to the size of the output real vector argument EIGVAL, i.e., neig = size( EIGVAL ).
!   
!   
! Arguments
! _________
!
!   MAT                 (INPUT) real(stnd), dimension(:,:)
!                       On entry, the n-by-n symmetric positive semi-definite matrix MAT.
!
!                       MAT is not modified by the routine.
!
!   EIGVAL              (OUTPUT) real(stnd), dimension(:)
!                       On exit, EIGVAL(:) contains the first top neig eigenvalues of MAT.
!                       The eigenvalues are given in decreasing order of magnitude.
!
!                       The size of EIGVAL must verify:
!
!                       - size( EIGVAL ) = neig <= size( MAT, 1 ) = size( MAT, 2 ) = n.
!
!   EIGVEC              (OUTPUT) real(stnd), dimension(:,:)
!                       On exit, the computed neig top eigenvectors. The eigenvector
!                       associated with the eigenvalue EIGVAL(j) is stored in the j-th column
!                       of EIGVEC.
!
!                       The shape of EIGVEC must verify:
!
!                       - size( EIGVEC, 1 ) = size( MAT, 1 ) = size( MAT, 2 ) = n,
!                       - size( EIGVEC, 2 ) = size( EIGVEC ) = neig .
!
!                       If FAILURE = true on exit, results are still useful, but some of the
!                       approximated eigen couplets have a poor accuracy.
!
!   FAILURE             (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit, if the optional logical argument FAILURE is present, a test of the
!                       accuracy of the computed partial EVD is performed and in that
!                       case:
!
!                       - FAILURE = false : indicates successful exit;
!                       - FAILURE = true  : indicates that some of the computed eigenvalues
!                         and eigenvectors of MAT failed to converge in NITER iterations.
!
!                       If FAILURE = true on exit, results are still useful, but some of the approximated
!                       eigen couplets have a poor accuracy.
!
!   NITER               (INPUT, OPTIONAL) integer(i4b)
!                       The number of randomized power, subspace or block Krylov iterations performed
!                       in the subroutine for computing the top neig eigen triplets. NITER must be
!                       positive or null.
!
!                       By default, 10 randomized power, subspace or block Krylov iterations are performed.
!
!   NOVER               (INPUT, OPTIONAL) integer(i4b)
!                       The oversampling size used in the randomized power, subspace or block Krylov
!                       iterations for computing the top neig eigen triplets.
!
!                       NOVER must be positive or null and verifies the relationship:
!
!                       -  NOVER + size( EIGVAL ) <= size( MAT, 1 ) = size( MAT, 2 ) = n
!
!                       and is adjusted if necessary to verify this relationship in all cases.
!
!                       See Further Details and the cited references for the meaning and usefulness
!                       of the oversampling size in the randomized power, subspace or block Krylov iterations.
!
!                       By default, the oversampling size is set to 10.
!
!   ORTHO               (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if:
!
!                       - ORTHO=true, orthonormalization is carried out between each step of
!                         the power iterations to avoid loss of accuracy due to rounding errors.
!                         This means that subspace iterations are used instead of power iterations;
!                       - ORTHO=false, orthonormalization is not performed. 
!
!                       The default is to use orthonormalization, e.g., ORTHO=true.
!
!   EXTD_SAMP           (INPUT, OPTIONAL) logical(lgl)
!                       The optional argument EXTD_SAMP determines if extended sampling
!                       (e.g., block Krylov iterations) is used or not for computing
!                       the top neig eigen triplets.
!
!                       On entry, if:
!
!                       - EXTD_SAMP=true, block Krylov iterations are used;
!                       - EXTD_SAMP=false, power or subspace iterations are used. 
!
!                       The default is to use power or subspace iterations, e.g., EXTD_SAMP=false.
!
!   USE_NYSTROM         (INPUT, OPTIONAL) logical(lgl)
!                       If the optional argument USE_NYSTROM is used and is set to:
!
!                       - true, the last step of the randomized algorithm is performed with
!                         the Nystrom method and an SVD decomposition;
!                       - false, an EVD decomposition is used in the final step
!                         of the randomized algorithm.
!
!                       The default is to use the Nystrom method, e.g., USE_NYSTROM=true.
!
!   RNG_ALG             (INPUT, OPTIONAL) integer(i4b)
!                       On entry, a scalar integer to select the random (uniform) number generator used to build
!                       the random gaussian test matrix in the randomized EVD algorithm.
!
!                       The possible values are:
!
!                       - ALG=1  : selects the Marsaglia's KISS random number generator;
!                       - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!                       - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!                       - ALG=4  : selects the Mersenne Twister random number generator;
!                       - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!                       - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!                       - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!                       - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!                       - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!                       - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                         random number generator;
!
!                       For other values, the current random number generator and its current state are not changed.
!                       Note further, that, on exit, the current random number generator is not reset to its previous
!                       value before the call to REIG_POS_CMP.
!
!                       See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the bidiagonal SVD phase
!                       of the SVD algorithm or in the QR phase of the EVD algorithm, which
!                       are used in the last phase of the randomized algorithm.
!
!                       See description of suboutines SVD_CMP and EIG_CMP for further details
!                       about this optional argument.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       The optional argument MAX_FRANCIS_STEPS controls the maximum number
!                       of Francis sets (e.g. QR sweeps) of Givens rotations which must be
!                       saved before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm, which is used
!                       in the last phase of the randomized algorithm if the Nystrom method
!                       is used.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is the minimum of neig and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       The optional argument PERFECT_SHIFT determines if a perfect shift strategy
!                       is used in the implicit QR algorithm in order to minimize the number of QR
!                       sweeps in the bidiagonal SVD algorithm, which is used in the last phase of
!                       the randomized EVD algorithm if the Nystrom method is used.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false
!                       or if the Nystrom method is not used.
!
!                       If PERFECT_SHIFT and BISECT are both set to true, singular values are computed
!                       with a more accurate bisection algorithm delivering improved accuracy in the final
!                       computed EVD decomposition at the expense of a slightly slower execution time if
!                       the Nystrom method is used.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm if the Nystrom method is used and PERFECT_SHIFT
!                       is equal to true.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   For a good introduction to randomized linear algebra, see the references (1) and (2).
!
!   The randomized subspace iteration was proposed in (3; see Algorithm 4.4) to compute
!   an orthonormal matrix whose range approximates the range of MAT. An approximate
!   partial spectral decomposition can then be computed using the aforementioned orthonormal
!   matrix, see Algorithm 5.3 in (3). Moreover, if the input matrix is positive semi-definite, an
!   improved randomized algorithm exists, the Nystrom method, see Algorithm 5.5 in (3) and also
!   references (1) and (5).
!
!   The Nystrom method will be selected in REIG_POS_CMP if the USE_NYSTROM argument is used with
!   the value true (this is the default), otherwise the standard EVD algorithm will be used in the last
!   step of the randomized algorithm. The Nystrom method provides more accurate results for positive
!   (semi-)definite matrices.
!
!   The randomized block Krylov iterations for computing an approximate partial EVD
!   was proposed in (4; see Algorithm 2). See also the reference (1).
!
!   For further details on randomized linear algebra, computing a partial EVD decomposition using randomized
!   power, subspace or block Krylov iterations, or the Nystrom method, see:
!
!   (1) Martinsson, P.G., 2019:
!            Randomized methods for matrix computations.
!            arXiv.1607.01649
!
!   (2) Erichson, N.B., Voronin, S., Brunton, S.L., and Kutz, J.N., 2019:
!            Randomized matrix decompositions using R.
!            arXiv.1608.02148
!
!   (3) Halko, N., Martinsson, P.G., and Tropp, J.A., 2011:
!            Finding structure with randomness:
!            probabilistic algorithms for constructing approximate matrix decompositions.
!            SIAM Rev., 53, 217-288.
!
!   (4) Musco, C., and Musco, C., 2015:
!            Randomized block krylov methods for stronger and faster approximate singular value decomposition.
!            In Proceedings of the 28th International Conference on Neural Information Processing Systems,
!            NIPS 15, pages 1396-1404, Cambridge, MA, USA, 2015. MIT Press.
!
!   (5) Li, H.,Linderman, G.C., Szlam, A., Stanton, K.P., Kluger, Y., and Tygert, M., 2017:
!            Algorithm 971: An implementation of a randomized algorithm for principal component analysis.
!            ACM Trans. Math. Softw. 43, 3, Article 28 (January 2017).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, assert_eq, merror
    use Random,          only : random_seed_, normal_random_number3_
    use QR_Procedures,   only : qr_cmp, ortho_gen_qr
    use Lin_Procedures,  only : chol_cmp
    use Eig_Procedures,  only : eig_cmp, eig_abs_sort
#ifdef _BLAS
    use BLAS_interfaces, only : dot, symm, gemm
!    use BLAS_interfaces, only : dot, gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:,:) :: mat
    real(stnd),  intent(out), dimension(:)   :: eigval
    real(stnd),  intent(out), dimension(:,:) :: eigvec
!
    integer(i4b), intent(in), optional :: niter, nover, rng_alg, maxiter, max_francis_steps
!
    logical(lgl), intent(in),  optional :: ortho, extd_samp, use_nystrom, perfect_shift, bisect, dqds
    logical(lgl), intent(out), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: d1, eps, err
!
#ifdef _BLAS
    real(stnd), dimension(:,:), allocatable :: q, b, v, k
#else
    real(stnd), dimension(:,:), allocatable :: q, qt, b, v, k
#endif
    real(stnd), dimension(:),   allocatable :: diagr, beta
!
    integer      :: iok, alg
    integer(i4b) :: n, neig, np, npe, i, i1, i2, j, niter2, nover2
!
    logical(lgl) :: do_test, do_ortho, do_extd_samp, do_nystrom, failure2
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='reig_pos_cmp'
!
!   niter0 IS THE DEFAULT NUMBER OF RANDOMIZED POWER, SUBSPACE OR BLOCK KRYLOV ITERATIONS PERFORMED.
!   nover0 IS THE DEFAULT OVERSAMPLING SIZE.
!
    integer(i4b),  parameter :: niter0=10_i4b, nover0=10_i4b
!   
!   eps0 IS THE THRESHOLD FOR THE ACCURACY TEST OF THE EIGEN COUPLETS.
!   
    real(stnd), parameter  :: eps0=0.0001_stnd
!   
    character, parameter   :: sort='d'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size( mat,    1 ),i4b)  ,  &
                   int(size( mat,    2 ),i4b)  ,  &
                   int(size( eigvec, 1 ),i4b)  ,  &
                   name_proc )
!
    neig = assert_eq( int(size( eigval ),i4b)     ,  &
                      int(size( eigvec, 2 ),i4b)  ,  &
                      name_proc )
!
!   DETERMINE IF A TEST OF EIGEN COUPLETS IS REQUIRED BEFORE EXIT.
!
    do_test = present( failure )
!
    if ( do_test ) then
        failure = false
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. neig<=0_i4b ) then
!
        return
!
    end if
!
    call assert( logical( neig>0_i4b .and. neig<=n, lgl ), name_proc )
!
!   SET OPTIONAL ARGUMENTS.
!
!   DETERMINE THE NUMBER OF POWER, SUBSPACE OR BLOCK KRYLOV ITERATIONS
!   niter2 TO BE PERFORMED.
!
    if ( present(niter) ) then
        niter2 = max( niter, 0_i4b )
    else
        niter2 = niter0
    end if
!
!   DETERMINE THE OVERSAMPLING SIZE nover2 .
!
    if ( present(nover) ) then
        nover2 = max( nover, 0_i4b )
    else
        nover2 = nover0
    end if
!
!   SPECIFY IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE OR BLOCK KRYLOV ITERATIONS, TO AVOID LOSS OF ACCURACY
!   DUE TO ROUNDING ERRORS. THIS IS NORMALLY NOT NEEDED FOR POSITIVE MATRIX.
!
    if ( present(ortho) ) then
        do_ortho = ortho
    else
        do_ortho = true
    end if
!
!   DETERMINE IF LAST STEP OF THE RANDOMIZED ALGORITHM IS PERFORMED WITH THE
!   NYSTROM METHOD OR THE EIGENVALUE DECOMPOSITION.
!
    if ( present(use_nystrom) ) then
        do_nystrom = use_nystrom
    else
        do_nystrom = true
    end if
!
!   DETERMINE IF POWER, SUBSPACE OR BLOCK KRYLOV ITERATIONS ARE USED.
!
    if ( present(extd_samp) .and. niter2>0_i4b ) then
        do_extd_samp = extd_samp
    else
        do_extd_samp = false
    end if
!
!   CHECK VALIDITY OF THE OVERSAMPLING SIZE PARAMETER
!   AND OF THE NUMBER OF ITERATIONS FOR BLOCK KRYLOV METHOD.
!
    np = min( nover2 + neig, n ) 
!
    if ( do_extd_samp ) then
!
       npe = ( niter2 + 1_i4b )*np
!
       if ( npe>n ) then
!
           npe = np
!
           do_extd_samp = false
!
       end if
!
!       do i = niter2, 0_i4b, -1_i4b
!
!           if ( npe<=n ) exit
!
!           npe = npe - np           
!
!       end do
!
!       niter2 = i
!
!       do_extd_samp = niter2>0_i4b
!
    else
!
       npe = np
!
    end if
!
!   SELECT AND SET THE RANDOM NUMBER GENERATOR TO BE USED IF REQUIRED.
!
    if ( present(rng_alg) ) then
!
        if ( rng_alg>=1_i4b .and. rng_alg<=10_i4b ) then
!
            alg = rng_alg
!
!           SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
            call random_seed_( alg=alg )
!
!           RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
            call random_seed_()
!
        end if
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    allocate( q(n,npe), b(n,npe), v(npe,npe), &
              diagr(npe), beta(npe), stat=iok )
#else
    allocate( q(n,npe), qt(npe,n), b(n,npe), v(npe,npe),  &
              diagr(npe), beta(npe), stat=iok             )
#endif
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-BY-np IID GAUSSIAN MATRIX.
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
    call normal_random_number3_( b(:n,:np) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
!    do i = 1_i4b, np
!
!        call normal_random_number3_( b(:n,i) )
!
!    end do
!
!   COMPUTE RANDOM SAMPLE MATRIX.
!
#ifdef _BLAS
!    call gemm( 'N', 'N', n, np, n, one, mat(1_i4b:n,1_i4b:n), n,     &
!               b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:n,1_i4b:np), n  )
    call symm( 'L', 'U', n, np, one, mat(1_i4b:n,1_i4b:n), n,        &
               b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:n,1_i4b:np), n  )
#else
    q(:n,:np) = matmul( mat(:n,:n), b(:n,:np) )
#endif
!
    if ( do_extd_samp ) then
!
!       USE BLOCK KRYLOV ITERATIONS.
!
!       ALLOCATE WORK ARRAY.
!
        allocate( k(n,np), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE FIRST KRYLOV BLOCK.
!
        k(:n,:np) = q(:n,:np)
!
        i1 = 1_i4b
        i2 = np
!
    end if
!
!   DO POWER OR SUBSPACE ITERATIONS.
!
    do i = 1_i4b, niter2
!
        if ( do_ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
            call qr_cmp( q(:n,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX
!           TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM COLUMN SUBSPACE.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( q(:n,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
!        call gemm( 'N', 'N', n, np, n, one, mat(1_i4b:n,1_i4b:n), n,     &
!                   q(1_i4b:n,1_i4b:np), n, zero, b(1_i4b:n,1_i4b:np), n  )
        call symm( 'L', 'U', n, np, one, mat(1_i4b:n,1_i4b:n), n,        &
                   q(1_i4b:n,1_i4b:np), n, zero, b(1_i4b:n,1_i4b:np), n  )
#else
        b(:n,:np) = matmul( mat(:n,:n), q(:n,:np)  )
#endif
!
        q(:n,:np) = b(:n,:np)
!
        if ( do_extd_samp ) then
!
!           USE BLOCK KRYLOV ITERATIONS.
!
            i1 = i1 + np
            i2 = i2 + np
!
!           SAVE i+1 KRYLOV BLOCK.
!
            q(:n,i1:i2) = b(:n,:np)
!
        end if
!
    end do
!
    if ( do_extd_samp ) then
!
!       USE BLOCK KRYLOV ITERATIONS.
!
!       RESTORE FIRST KRYLOV BLOCK.
!
        q(:n,:np) = k(:n,:np)
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( k )
!
    end if
!
!   COMPUTE QR DECOMPOSITION OF RANDOM SAMPLE MATRIX.
!
    call qr_cmp( q(:n,:npe), diagr(:npe), beta(:npe) )
!
!   GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX
!   TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM COLUMN SUBSPACE.
!   q IS ASSUMED OF FULL RANK.
!
    call ortho_gen_qr( q(:n,:npe), beta(:npe) )
!
    if ( do_nystrom ) then
!
!       USE THE NYSTROM METHOD.
!
!       COMPUTE b = mat*q AND v = q**(t)*( b + ulp*norm(b)*q ).
!
#ifdef _BLAS
!        call gemm( 'N', 'N', n, npe, n, one, mat(1_i4b:n,1_i4b:n), n,      &
!                   q(1_i4b:n,1_i4b:npe), n, zero, b(1_i4b:n,1_i4b:npe), n  )
        call symm( 'L', 'U', n, npe, one, mat(1_i4b:n,1_i4b:n), n,         &
                   q(1_i4b:n,1_i4b:npe), n, zero, b(1_i4b:n,1_i4b:npe), n  )
!
        eps = ulp*norm( b(1_i4b:n,1_i4b:npe) )

        b(1_i4b:n,1_i4b:npe) = b(1_i4b:n,1_i4b:npe) + eps*q(1_i4b:n,1_i4b:npe)
!
        call gemm( 'T', 'N', npe, npe, n, one, q(1_i4b:n,1_i4b:npe), n,        &
                   b(1_i4b:n,1_i4b:npe), n, zero, v(1_i4b:npe,1_i4b:npe), npe  )
#else
        b(:n,:npe) = matmul( mat(:n,:n), q(:n,:npe)  )
!
        eps = ulp*norm( b(1_i4b:n,1_i4b:npe) )

        b(1_i4b:n,1_i4b:npe) = b(1_i4b:n,1_i4b:npe) + eps*q(1_i4b:n,1_i4b:npe)
!
        qt(:npe,:n) = transpose( q(:n,:npe) )
!
        v(:npe,:npe) = matmul( qt(:npe,:n), b(:n,:npe) )
#endif
!
!       COMPUTE CHOLESKY FACTORIZATION OF v = c**(t)*c .
!
        call chol_cmp( v(:npe,:npe), diagr(:npe), d1 )
#ifdef _BLAS
!
!       ALLOCATE WORK ARRAY.
!
        allocate( k(npe,n), stat=iok )
!
        if ( iok/=0 ) then
             d1 = zero
        end if
#endif
!
    else
!
!       USE AN EVD DECOMPOSITION AS A BACKUP.
!
        d1 = zero
!
    end if
!
    if ( d1/=zero ) then
!
!       USE THE NYSTROM METHOD.
!
!       COMPUTE F = b*c**(-1) USING THE CHOLESKY FACTORIZATION, v = c**(t)*c, AND A TRIANGULAR SOLVE.
!
#ifdef _BLAS
        k(:npe,:n) = transpose( b(:n,:npe) )    
!
!$OMP PARALLEL DO IF(npe*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i,j)
!
        do i = 1_i4b, n
!
            do j = 1_i4b, npe
!
                k(j,i) = ( k(j,i) - dot( j-1_i4b, v(1_i4b:j-1_i4b,j), 1_i4b, k(1_i4b:j-1_i4b,i), 1_i4b ) )*diagr(j)
!
            end do
!
        end do
!
!$OMP END PARALLEL DO
!
        b(:n,:npe) = transpose( k(:npe,:n) )    
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( k )
#else
        qt(:npe,:n) = transpose( b(:n,:npe) )    
!
!$OMP PARALLEL DO IF(npe*n>=omp_limit), SCHEDULE(STATIC), PRIVATE(i,j)
!
        do i = 1_i4b, n
!
            do j = 1_i4b, npe
!
                qt(j,i) = ( qt(j,i) - dot_product( v(1_i4b:j-1_i4b,j), qt(1_i4b:j-1_i4b,i) ) )*diagr(j)
!
            end do
!
        end do
!
!$OMP END PARALLEL DO
!
        b(:n,:npe) = transpose( qt(:npe,:n) )    
#endif
!
!       COMPUTE SVD OF THE CHOLESKY FACTOR.
!
        call svd_cmp( b(:n,:npe), beta(:npe), failure2, v=v(:npe,:npe), sort=sort, &
                      maxiter=maxiter, max_francis_steps=max_francis_steps,        &
                      perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!       EXTRACT THE APPROXIMATE TOP neig EIGENVECTORS OF mat .
!
        eigvec(:n,:neig) = b(:n,:neig)
!
!       COMPUTE THE APPROXIMATE TOP neig EIGENVALUES OF mat .
!
        eigval(:neig) = beta(:neig)*beta(:neig) - eps
!
        eigval(:neig) = max( eigval(:neig), zero )
!
    else
!
!       USE A STANDARD SPECTRAL DECOMPOSITION.
!
!       COMPUTE b = mat*q AND v = q**(t)*b .
!
#ifdef _BLAS
!        call gemm( 'N', 'N', n, npe, n, one, mat(1_i4b:n,1_i4b:n), n,      &
!                   q(1_i4b:n,1_i4b:npe), n, zero, b(1_i4b:n,1_i4b:npe), n  )
        call symm( 'L', 'U', n, npe, one, mat(1_i4b:n,1_i4b:n), n,         &
                   q(1_i4b:n,1_i4b:npe), n, zero, b(1_i4b:n,1_i4b:npe), n  )
!
        call gemm( 'T', 'N', npe, npe, n, one, q(1_i4b:n,1_i4b:npe), n,        &
                   b(1_i4b:n,1_i4b:npe), n, zero, v(1_i4b:npe,1_i4b:npe), npe  )
#else
        b(:n,:npe) = matmul( mat(:n,:n), q(:n,:npe)  )
!
        qt(:npe,:n) = transpose( q(:n,:npe) )
!
        v(:npe,:npe) = matmul( qt(:npe,:n), b(:n,:npe) )
#endif
!
!       COMPUTE SPECTRAL DECOMPOSITION.
!
        call eig_cmp( v(:npe,:npe), beta(:npe), failure2, sort=sort, maxiter=maxiter )
!
        call eig_abs_sort( sort, beta(:npe), v(:npe,:npe) )
!
!       COMPUTE THE APPROXIMATE TOP neig EIGENVECTORS OF mat .
!!
#ifdef _BLAS
        call gemm( 'N', 'N', n, neig, npe, one, q(1_i4b:n,1_i4b:npe), n,              &
                   v(1_i4b:npe,1_i4b:neig), npe, zero, eigvec(1_i4b:n,1_i4b:neig), n  )
#else
        eigvec(:n,:neig) = matmul( q(:n,:npe), v(:npe,:neig) )
#endif
!
!       EXTRACT THE APPROXIMATE TOP neig EIGENVALUES OF mat .
!
        eigval(:neig) = max( beta(:neig), zero )
!
    end if
!
!   TEST ACCURACY OF THE APPROXIMATE EVD IF REQUIRED.
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION mat*u(:n,:neig) - u(:n,:neig)*diag(eigval(:neig)),
!       WHERE u ARE THE EIGENVECTORS OF mat.
!
        q(:n,:neig) = eigvec(:n,:neig)*spread(eigval(:neig),dim=1,ncopies=n)
!
#ifdef _BLAS
!        call gemm( 'N', 'N', n, neig, n, one, mat(1_i4b:n,1_i4b:n), n,            &
!                   eigvec(1_i4b:n,1_i4b:neig), n, -one, q(1_i4b:n,1_i4b:neig), n  )
       call symm( 'L', 'U', n, neig, one, mat(1_i4b:n,1_i4b:n), n,        &
                   eigvec(1_i4b:n,1_i4b:neig), n, -one, q(1_i4b:n,1_i4b:neig), n  )
#else
        q(:n,:neig) = matmul( mat(:n,:n), eigvec(:n,:neig)  ) - q(:n,:neig)
#endif
!
        beta(:neig) = norm( q(:n,:neig), dim=2_i4b )
!
        err = maxval( beta(:neig) )/( norm( mat )*real(n,stnd) )
!
        failure = err>eps0 .or. failure2
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    deallocate( q, b, v, diagr, beta )
#else
    deallocate( q, qt, b, v, diagr, beta )
#endif
!
!
! END OF SUBROUTINE reig_pos_cmp
! ______________________________
!
    end subroutine reig_pos_cmp
!
! =========================================================================================
!
    subroutine rqr_svd_cmp( mat, s, failure, v, random_qr, truncated_qr, rng_alg, blk_size,  &
                            nover, nover_svd, maxiter, max_francis_steps, perfect_shift,     &
                            bisect, dqds )
!
! Purpose
! _______
!
!   RQR_SVD_CMP computes approximations of the nsvd largest singular values and associated
!   left and right singular vectors of a full m-by-n real matrix MAT using a three-step procedure,
!   which can be termed a QR-SVD algorithm:
!
!   - first, a partial (or complete) QR factorization with column pivoting of MAT is computed;
!   - in a second step, a Singular Value Decomposition (SVD) of the (permuted) upper triangular
!     or trapezoidal (e.g., if n>m) factor, R, of this QR decomposition is computed. The singular values
!     and right singular vectors of this SVD of R are also estimates of the singular values and right
!     singular vectors of MAT;
!   - Estimates of the  associated left singular vectors of MAT are then obtained by pre-multiplying 
!     the left singular of R by the orthogonal matrix Q in the initial QR decomposition of MAT
!     (or its first k columns if the QR factorization is only partial).
!
!   By default, a standard deterministic BLAS2 QR factorization with column pivoting is used in the first phase
!   of the QR-SVD algorithm. However, if the optional logical argument RANDOM_QR is used with the value true,
!   an alternate fast randomized partial QR factorization is used in the first phase of the QR-SVD algorithm.
!
!   Furthermore if, in addition, the optional logical argument TRUNCATED_QR is used with the value true,
!   an even faster (but less accurate) randomized partial and truncated QR factorization is used in the first
!   phase of the QR-SVD algorithm.
!
!   nsvd is the target rank of the partial SVD, which is sought, and is equal to the size of the output
!   real vector argument S, i.e., nsvd = size( S ). If, nsvd = min( size(MAT,1) , size(MAT,2) ), a full
!   SVD of MAT is obtained with the same or higher accuracy than subroutines SVD_CMP or SVD_CMP2 if the
!   optional logical argument RANDOM_QR is not used (or is set to false).
!
!   
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit, the top nsvd left singular vectors are stored in the first nsvd columns of MAT.
!                       The left singular vector associated with the singular value S(j) is stored in the j-th column
!                       of MAT. The other part of MAT is used as workspace in the algorithm and is destroyed on exit.
!
!   S                   (OUTPUT) real(stnd), dimension(:)
!                       On exit, S(:) contains the first top nsvd singular values of MAT.
!                       The singular values are given in decreasing order and are
!                       positive or zero.
!
!                       The size of S must verify:
!
!                       - size( S ) = nsvd <= min( size(MAT,1) , size(MAT,2) ) .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit, if:
!
!                       - FAILURE = false : indicates successful exit in the SVD of the triangular factor
!                         R of the QR decomposition of MAT.
!                       - FAILURE = true  : indicates that the SVD algorithm did not converge and
!                         that full accuracy was not attained in the bidiagonal SVD of an 
!                         intermediate bidiagonal form B of the triangular factor R in the QR
!                         decomposition of MAT.
!
!                       If on entry, RANDOM_QR=true and TRUNCATED_QR=true, a test of the accuracy of
!                       the randomized partial and truncated QR factorization used in the first phase
!                       is also performed. In that case:
!
!                       - FAILURE = false : indicates also that this randomized partial and truncated QR
!                         factorization seems accurate.
!                       - FAILURE = true  : indicates that this randomized partial and truncated QR
!                         factorization is not accurate.
!
!   V                   (OUTPUT) real(stnd), dimension(:,:)
!                       On exit, the computed top nsvd right singular vectors of MAT. The right singular
!                       vector associated with the singular value S(j) is stored in the j-th column of V.
!
!                       The shape of V must verify:  
!
!                       - size( V, 1 ) = size( MAT, 2 ) = n ,
!                       - size( V, 2 ) = size( S )      = nsvd .
!
!   RANDOM_QR           (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if RANDOM_QR is used with the value true, a fast randomized partial QR
!                       factorization is used in the first phase of the QR-SVD algorithm.
!
!                       By default, RANDOM_QR = false, i.e., a standard deterministic (partial) QR factorization
!                       with column pivoting is used in the first phase of the QR-SVD algorithm.
!
!   TRUNCATED_QR        (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if TRUNCATED_QR is used with the value true in addition to RANDOM_QR also
!                       set to true, a very fast (but less accurate) randomized partial and truncated
!                       QR factorization is used in the first phase of the QR-SVD algorithm.
!
!                       By default, TRUNCATED_QR = false, i.e., a "standard" randomized (partial) QR factorization
!                       with column pivoting is used in the first phase of the QR-SVD algorithm if RANDOM_QR = true.
!
!   RNG_ALG             (INPUT, OPTIONAL) integer(i4b)
!                       On entry, a scalar integer to select the random (uniform) number generator used to build
!                       the random gaussian matrix in the randomized (partial) QR phase of the QR-SVD algorithm
!                       if RANDOM_QR = true.
!
!                       The possible values are:
!
!                       - ALG=1  : selects the Marsaglia's KISS random number generator;
!                       - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!                       - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!                       - ALG=4  : selects the Mersenne Twister random number generator;
!                       - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!                       - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!                       - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!                       - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!                       - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!                       - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                         random number generator;
!
!                       For other values, the current random number generator and its current state are not changed.
!                       Note further, that, on exit, the current random number generator is not reset to its previous
!                       value before the call to RQR_SVD_CMP.
!
!                       See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   BLK_SIZE            (INPUT, OPTIONAL) integer(i4b)
!                       On entry, the block size used in the randomized partial QR phase of the QR-SVD algorithm
!                       if RANDOM_QR = true (and TRUNCATED_QR = false).
!
!                       BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!                       smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!                       See Further Details and the cited references for the meaning of the block size
!                       in the randomized (partial) QR algorithm.
!
!                       By default, BLK_SIZE is set to min( BLKSZ_QR, min(m,n) ), where parameter BLKSZ_QR is the
!                       default block size for QR related algorithms specified in module Select_Parameters.
!
!   NOVER               (INPUT, OPTIONAL) integer(i4b)
!                       The oversampling size used in the randomized partial QR phase of the QR-SVD algorithm
!                       if RANDOM_QR = true.
!
!                       NOVER must be positive or null and verify the relationships:
!
!                       - NOVER + BLK_SIZE              <= size( MAT, 1 ) if TRUNCATED_QR = false;
!                       - NOVER + NOVER_SVD + size( S ) <= size( MAT, 1 ) if TRUNCATED_QR = true.
!
!                       and is adjusted if necessary to verify these relationships in all cases.
!
!                       See Further Details and the cited references for the meaning and usefulness
!                       of the oversampling size in the randomized (partial) QR algorithms.
!
!                       By default, the oversampling size is set to:
!
!                       - 10                                   if TRUNCATED_QR = false;
!                       - max( (NOVER_SVD+size(S))/2_i4b, 10 ) if TRUNCATED_QR = true.
!
!   NOVER_SVD           (INPUT, OPTIONAL) integer(i4b)
!                       The oversampling size used in the SVD phase of the QR-SVD algorithm
!                       for computing the top nsvd singular triplets.
!
!                       NOVER_SVD must be positive or null and verify the relationship:
!
!                       - NOVER_SVD + size( S ) <= min( size(MAT,1) , size(MAT,2) )
!
!                       and is adjusted if necessary to verify this relationship in all cases.
!
!                       See Further Details and the cited references for the meaning and usefulness
!                       of the oversampling size in the SVD phase of the QR-SVD algorithm.
!
!                       By default, the oversampling size in the SVD phase is set to 10.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm, which is used in the
!                       last step of the QR-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       The optional argument MAX_FRANCIS_STEPS controls the maximum number
!                       of Francis sets (e.g., QR sweeps) of Givens rotations which must be
!                       saved before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm, which is used
!                       in the last step of the QR-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is the minimum of nsvd and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       The optional argument PERFECT_SHIFT determines if a perfect shift strategy
!                       is used in the implicit QR algorithm in order to minimize the number of QR
!                       sweeps in the bidiagonal SVD algorithm, which is used in the last step of
!                       the QR-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed QR-SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   The standard deterministic BLAS2 algorithm for computing a QR factorization with column pivoting
!   is described in the reference (1). The randomized partial QR factorization with column pivoting
!   used if the optional logical argument RANDOM_QR is present with the value true is described
!   in the references (3), (4), (5) and (6). Finally, the randomized partial and truncated QR factorization
!   with column pivoting used if both the optional logical arguments RANDOM_QR and TRUNCATED_QR are
!   present with the value true is described in the reference (7). This algorithm is the fastest, but
!   is less accurate than the randomized partial QR factorization with column pivoting described
!   in the references (3), (4), (5) and (6).
!
!   For further details, on computing low-rank matrix approximations from QR factorizations
!   with column pivoting, the QR-SVD or randomized QR algorithms, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Chan, T.F., and Hansen, P.C., 1992:
!           Some applications of the rank revealing QR factorization.
!           SIAM J. Sci. Statist. Comput., Volume 13, 727-741.
!
!   (3) Duersch, J.A., and Gu, M., 2017:
!           Randomized QR with column pivoting.
!           SIAM J. Sci. Comput., Volume 39, C263-C291.
!
!   (4) Martinsson, P.G., Quintana-Orti, G., Heavner, N., and Van de Geijn, R., 2017:
!           Householder QR factorization with randomization for column pivoting (HQRRP).
!           SIAM J. Sci. Comput., Volume 39, C96-C115.
!
!   (5) Xiao, J., Gu, M., and Langou, J., 2017:
!           Fast parallel randomized QR with column pivoting algorithms
!           for reliable low-rank matrix approximations.
!           IEEE 24th International Conference
!           on High Performance Computing (HiPC), IEEE, 2017, 233-242.
!
!   (6) Duersch, J.A., and Gu, M., 2020:
!           Randomized projection for rank-revealing matrix factorizations
!           and low-rank approximations.
!           SIAM Review, Volume 62, Issue 3, 661-682.
!
!   (7) Mary, T., Yamazaki, I., Kurzak, J., Luszczek, P., Tomov, S., and Dongarra, J., 2015:
!           Performance of Random Sampling for Computing Low-rank Approximations of a
!           Dense Matrix on GPUs.
!           International Conference for High Performance Computing, Networking, Storage and Analysis (SC'15).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, assert_eq, merror
    use QR_Procedures,   only : partial_qr_cmp, ortho_gen_qr
    use Random,          only : partial_rqr_cmp, partial_rtqr_cmp
!    use Random,          only : partial_rqr_cmp2, partial_rtqr_cmp
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout), dimension(:,:) :: mat
    real(stnd),  intent(out),   dimension(:)   :: s
    real(stnd),  intent(out),   dimension(:,:) :: v
!
    integer(i4b), intent(in), optional :: rng_alg, blk_size, nover, nover_svd, &
                                          maxiter, max_francis_steps
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: random_qr, truncated_qr, perfect_shift, bisect, dqds
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
!
    real(stnd), dimension(:,:), allocatable :: work, u
    real(stnd), dimension(:),   allocatable :: diagr, beta
    real(stnd)                              :: norm_mat
!
    integer      :: iok
    integer(i4b) :: m, n, mn, nsvd, nover_svd2, np, i, j
    integer(i4b), dimension(:), allocatable :: ip
!
    logical(lgl) :: use_random_qr, use_truncated_qr, failure_rtqr
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='rqr_svd_cmp'
!
!   nover_svd0 IS THE DEFAULT OVERSAMPLING SIZE FOR THE SVD PHASE.
!
    integer(i4b),  parameter :: nover_svd0=10_i4b
!   
    character, parameter   :: sort='d'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
!
    m = size( mat, 1 )
!
    n = assert_eq( int(size(mat,2),i4b) ,  &
                   int(size(v,1),i4b)   ,  &
                   name_proc )
!
    mn = min( m, n )
!
    nsvd = assert_eq( int(size(s),i4b)   ,  &
                      int(size(v,2),i4b) ,  &
                      name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( mn<=0_i4b  .or. nsvd<=0_i4b ) then
!
        failure = false
!
        return
!
    end if
!
    call assert( logical( nsvd>0_i4b .and. nsvd<=mn, lgl ), name_proc )
!
!   SET OPTIONAL ARGUMENTS.
!
!   DETERMINE IF A RANDOMIZED QR ALGORITHM MUST BE USED.
!
    if ( present(random_qr) ) then
        use_random_qr = random_qr
    else
        use_random_qr = false
    end if
!
    if ( present(truncated_qr) .and. use_random_qr ) then
        use_truncated_qr = truncated_qr
    else
        use_truncated_qr = false
    end if
!
!   DETERMINE THE OVERSAMPLING SIZE nover_svd2 FOR THE SVD PHASE.
!
    if ( present(nover_svd) ) then
        nover_svd2 = max( nover_svd, 0_i4b )
    else
        nover_svd2 = nover_svd0
    end if
!
!   CHECK VALIDITY OF THE OVERSAMPLING SIZE PARAMETER FOR THE SVD PHASE.
!
    np = min( nover_svd2 + nsvd, mn ) 
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( ip(n), work(np,n), u(np,np), diagr(np), beta(np), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   FIRST STEP, DO PARTIAL QR WITH COLUMN PIVOTING OF mat AS
!           
!                   mat * P = Q * R
!           
!   WHERE P IS A n-BY-n PERMUTATION MATRIX, Q AN m-BY-m ORTHOGONAL
!   MATRIX AND R AN UPPER TRIANGULAR OR TRAPEZOIDAL MATRIX.
!           
    if ( use_random_qr ) then
!           
        if ( use_truncated_qr ) then
!           
!           COMPUTE FROBENIUS NORM OF THE INPUT MATRIX FOR THE ACCURACY TEST.
!           
            norm_mat = norm( mat(:m,:n) )
!           
!           USE A RANDOMIZED AND TRUNCATED QR ALGORITHM.
!           
            call partial_rtqr_cmp( mat(:m,:n), diagr(:np), beta(:np), ip(:n), i,   &
                                   rng_alg=rng_alg, nover=nover )
!           
        else
!           
!           USE A RANDOMIZED QR ALGORITHM.
!           
            call partial_rqr_cmp( mat(:m,:n), diagr(:np), beta(:np), ip(:n), i,   &
                                  rng_alg=rng_alg, blk_size=blk_size, nover=nover )
!           
!            call partial_rqr_cmp2( mat(:m,:n), diagr(:np), beta(:np), ip(:n), i,   &
!                                   rng_alg=rng_alg, blk_size=blk_size, nover=nover )
!
        end if
!           
    else
!           
!       USE A STANDARD DETERMINISTIC QR ALGORITHM.
!           
        call partial_qr_cmp( mat(:m,:n), diagr(:np), beta(:np), ip(:n), i )
!
    end if
!
!   RESTORE AND APPLY TRANSPOSE PERMUTATION TO THE UPPER TRIANGULAR FACTOR
!   OF THE QR FACTORIZATION OF mat , STORING RESULTS IN work ,
!           
!                   work = R * P'
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  np*n>=omp_limit                     .and.       &
                  n/i>=omp_limit2                     .and.       &
                  i>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC), PRIVATE(j,i)
!
    do j = 1_i4b, n
!
        i = ip(j)
!
        if ( j<=np ) then
!
            work(:j-1_i4b,i)   = mat(:j-1_i4b,j)
            work(j,i)          = diagr(j)
            work(j+1_i4b:np,i) = zero
!
        else
!
            work(:np,i) = mat(:np,j)
!
        end if
!
    end do
!
!$OMP END PARALLEL DO
!
!   GENERATE FIRST np COLUMNS OF ORTHOGONAL MATRIX Q OF QR DECOMPOSITION OF mat .
!
    call ortho_gen_qr( mat(:m,:np), beta(:np) )
!
!   SECOND STEP, COMPUTE SVD OF rmat = R * P'.
!
    call svd_cmp2( work(:np,:n), diagr(:np), failure, u(:np,:np), sort=sort, &
                   maxiter=maxiter, max_francis_steps=max_francis_steps,     &
                   perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!           
    if ( use_truncated_qr ) then
!
!       TEST ACCURACY OF THE RANDOMIZED AND TRUNCATED QR ALGORITHM
!       USED IN THE FIRST STEP IF REQUIRED.
!
        failure_rtqr = sum( (diagr(:np)/norm_mat)**2 )>one
!
        failure = failure .or. failure_rtqr
!
    end if
!
!   SAVE SINGULAR VALUES OF mat .
!
    s(:nsvd) = diagr(:nsvd)
!
    if ( np<n ) then
!
!       SAVE RIGHT SINGULAR VECTORS OF mat .
!
        v(:n,:nsvd) = transpose( work(:nsvd,:n) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( ip, work, diagr, beta )
!
!       REALLOCATE WORKSPACE FOR COMPUTING LEFT SINGULAR VECTORS OF mat .
!
        allocate( work(m,nsvd), stat=iok  )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       NEXT COMPUTE LEFT SINGULAR VECTORS OF mat AS
!
!                Q(:m,:np) * U
!
!       , WHERE Q IS THE ORTHOGONAL MATRIX OF QR DECOMPOSITION OF mat
!       AND U ARE THE LEFT SINGULAR VECTORS OF R * P'.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsvd, np, one, mat(1_i4b:m,1_i4b:np), m, &
                   u(1_i4b:np,1_i4b:nsvd), np, zero, work, m  )
#else
        work(:m,:nsvd) = matmul( mat(:m,:np), u(:np,:nsvd) )
#endif
!
!       SAVE LEFT SINGULAR VECTORS IN mat ON EXIT.
!
        mat(:m,:nsvd) = work(:m,:nsvd)
!
    else
!
!       SAVE RIGHT SINGULAR VECTORS OF mat .
!
        v(:n,:nsvd) = transpose( u(:nsvd,:n) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( ip, u, diagr, beta )
!
!       REALLOCATE WORKSPACE FOR COMPUTING LEFT SINGULAR VECTORS OF mat .
!
        allocate( u(m,nsvd), stat=iok  )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       NEXT COMPUTE LEFT SINGULAR VECTORS OF mat AS
!
!                    Q(:m,:np) * U
!
!       , WHERE Q IS THE ORTHOGONAL MATRIX OF QR DECOMPOSITION OF mat
!       AND U ARE THE LEFT SINGULAR VECTORS OF R * P'.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsvd, np, one, mat(1_i4b:m,1_i4b:np), m, &
                   work(1_i4b:np,1_i4b:nsvd), np, zero, u, m  )
#else
        u(:m,:nsvd) = matmul( mat(:m,:np), work(:np,:nsvd) )
#endif
!
!       SAVE LEFT SINGULAR VECTORS IN mat ON EXIT.
!
        mat(:m,:nsvd) = u(:m,:nsvd)
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( work, u )
!
!
! END OF SUBROUTINE rqr_svd_cmp
! _____________________________
!
    end subroutine rqr_svd_cmp
!
! =========================================================================================
!
    subroutine rqr_svd_cmp_fixed_precision( mat, relerr, s, failure, v, random_qr, rng_alg, &
                                            blk_size, nover, maxiter, max_francis_steps,    &
                                            perfect_shift, bisect, dqds )
!
! Purpose
! _______
!
!   RQR_SVD_CMP_FIXED_PRECISION computes approximations of the nsvd largest singular values and associated
!   left and right singular vectors of a full m-by-n real matrix MAT using a three-step procedure,
!   which can be termed a QR-SVD algorithm:
!
!   - first, a partial QR factorization with column pivoting of MAT is computed;
!   - in a second step, a Singular Value Decomposition (SVD) of the (permuted) upper triangular
!     or trapezoidal (e.g., if n>m) factor, R, of this QR decomposition is computed. The singular values
!     and right singular vectors of this SVD of R are also estimates of the singular values and right
!     singular vectors of MAT.
!   - in a final step, estimates of the associated left singular vectors of MAT are obtained by
!     pre-multiplying the left singular of R by the orthogonal matrix Q of the initial QR decomposition
!     (or its first k columns if the QR factorization is only partial).
!
!   By default, a standard deterministic BLAS2 QR factorization with column pivoting is used in the first phase
!   of the QR-SVD algorithm. However, if the optional logical argument RANDOM_QR is used with the value true,
!   an alternate fast randomized partial QR factorization is used in the first phase of the QR-SVD algorithm.
!
!   nsvd is the target rank of the partial Singular Value Decomposition (SVD), which is sought, 
!   and this partial SVD  must have an approximation error which fulfills:
!
!            ||MAT-rSVD||_F <= ||MAT||_F * relerr
!
!   , where rSVD is the computed partial SVD approximation, || ||_F is the Frobenius norm and
!   relerr is a prescribed accuracy tolerance for the relative error of the computed partial SVD
!   approximation, which is specified in the input real argument RELERR.
!   
!   In other words, nsvd is not known in advance and is determined in the subroutine. This
!   explains why the output real array arguments S and V, which contain the computed singular
!   values and associated right singular vectors of MAT on exit, must be declared in the calling
!   program as pointers.
!   
!   On exit, nsvd is equal to the size of the output real pointer argument S, which contains the
!   computed singular values, i.e., nsvd = size( S ) and the relative error in the Frobenius norm
!   of the computed partial SVD approximation is output in argument RELERR.
!
!   RQR_SVD_CMP_FIXED_PRECISION searches incrementally the best (e.g., smallest) partial SVD approximation,
!   which fulfills the prescribed accuracy tolerance for the relative error. More precisely, the rank of
!   the partial SVD approximation is increased progressively until the prescribed accuracy tolerance is
!   satisfied and then improved and adjusted precisely in a final step to obtain the smallest partial SVD
!   approximation, which satisfies the prescribed tolerance.
!
!   In all cases the relative error of the computed partial SVD approximation is output in argument RELERR.
!
!   
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit, the top nsvd left singular vectors are stored in the first nsvd columns of MAT.
!                       The left singular vector associated with the singular value S(j) is stored in the j-th column
!                       of MAT. The other part of MAT is used as workspace in the algorithm and is destroyed on exit.
!
!   RELERR              (INPUT/OUTPUT) real(stnd)
!                       On entry, the requested accuracy tolerance for the relative error of
!                       the computed partial SVD approximation.
!
!                       The preset value for RELERR must be greater than 4*epsilon( RELERR ) and less than one.
!
!                       On exit, RELERR contains the relative error of the computed partial SVD
!                       approximation in the Frobenius norm:
!
!                       - RELERR = ||MAT-rSVD||_F / ||MAT||_F
!
!   S                   (OUTPUT) real(stnd), dimension(:), pointer
!                       On exit, S(:) contains estimates of the first top nsvd singular values of MAT.
!                       The singular values are given in decreasing order and are positive or zero.
!
!                       The statut of the pointer S must not be undefined on entry.
!                       If, on entry, the pointer S is already allocated, it will be first deallocated
!                       and then reallocated with the correct size.
!
!                       On exit, the size of the pointer S will verify:
!
!                       - size( S ) = nsvd <= min( size(MAT,1) , size(MAT,2) ) .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false : indicates successful exit in the SVD of the triangular factor
!                         R of the QR decomposition of MAT.
!                       - FAILURE = true  : indicates that the SVD algorithm did not converge and
!                         that full accuracy was not attained in the bidiagonal SVD of an 
!                         intermediate bidiagonal form B of the triangular factor R of the QR
!                         decomposition of MAT.
!
!   V                   (OUTPUT) real(stnd), dimension(:,:), pointer
!                       On exit, the computed nsvd top right singular vectors. The right singular
!                       vector associated with the singular value S(j) is stored in the j-th column of V.
!
!                       The statut of the pointer V must not be undefined on entry.
!                       If, on entry, the pointer V is already allocated, it will be first deallocated
!                       and then reallocated with the correct shape.
!
!                       On exit, the shape of the pointer V will verify:  
!
!                       - size( V, 1 ) = size( MAT, 2 ) = n ,
!                       - size( V, 2 ) = size( S )      = nsvd .
!
!   RANDOM_QR           (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if RANDOM_QR is used with the value true, a fast randomized partial QR
!                       factorization is used in the first phase of the QR-SVD algorithm.
!
!                       By default, RANDOM_QR = false, i.e., a standard deterministic (partial) QR factorization
!                       with column pivoting is used in the first phase of the QR-SVD algorithm.
!
!   RNG_ALG             (INPUT, OPTIONAL) integer(i4b)
!                       On entry, a scalar integer to select the random (uniform) number generator used to build
!                       the random gaussian matrix in the randomized (partial) QR phase of the QR-SVD algorithm
!                       if RANDOM_QR = true.
!
!                       The possible values are:
!
!                       - ALG=1  : selects the Marsaglia's KISS random number generator;
!                       - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!                       - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!                       - ALG=4  : selects the Mersenne Twister random number generator;
!                       - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!                       - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!                       - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!                       - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!                       - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!                       - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                         random number generator;
!
!                       For other values, the current random number generator and its current state are not changed.
!                       Note further, that, on exit, the current random number generator is not reset to its previous
!                       value before the call to RQR_SVD_CMP_FIXED_PRECISION.
!
!                       See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   BLK_SIZE            (INPUT, OPTIONAL) integer(i4b)
!                       On entry, the block size used in the randomized partial QR phase of the QR-SVD algorithm
!                       if RANDOM_QR = true.
!
!                       BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!                       smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!                       See Further Details and the cited references for the meaning of the block size
!                       in the randomized (partial) QR algorithm.
!
!                       By default, BLK_SIZE is set to min( BLKSZ_QR, min(m,n) ), where parameter BLKSZ_QR is the default
!                       block size for QR related algorithms specified in module Select_Parameters.
!
!   NOVER               (INPUT, OPTIONAL) integer(i4b)
!                       The oversampling size used in the randomized partial QR phase of the QR-SVD algorithm
!                       if RANDOM_QR = true.
!
!                       NOVER must be positive or null and verify the relationship:
!
!                       - NOVER + BLK_SIZE <= size( MAT, 1 )
!
!                       and is adjusted if necessary to verify this relationship in all cases.
!
!                       See Further Details and the cited references for the meaning and usefulness
!                       of the oversampling size in the randomized (partial) QR algorithm.
!
!                       By default, the oversampling size is set to 10.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm, which is used in the
!                       last step of the QR-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       The optional argument MAX_FRANCIS_STEPS controls the maximum number
!                       of Francis sets (e.g., QR sweeps) of Givens rotations which must be
!                       saved before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm, which is used
!                       in the last step of the QR-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is the minimum of nsvd and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       The optional argument PERFECT_SHIFT determines if a perfect shift strategy
!                       is used in the implicit QR algorithm in order to minimize the number of QR
!                       sweeps in the bidiagonal SVD algorithm, which is used in the last step of
!                       the QR-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed QR-SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   Usually, the problem of low-rank matrix approximation falls into two categories:
!
!   - the fixed-rank problem, where the rank parameter nsvd is given;
!   - the fixed-precision problem, where we seek a partial SVD factorization, rSVD, as small
!     as possible such that
!     
!     ||MAT-rSVD||_F <= eps
!
!     , where eps is a given accuracy tolerance.
!
!   RQR_SVD_CMP_FIXED_PRECISION is dedicated to solve the fixed-precision problem. The fixed-rank problem
!   can be solved by subroutine RQR_SVD_CMP.
!
!   The standard deterministic BLAS2 algorithm for computing a QR factorization with column pivoting
!   is described in the reference (1). The randomized partial QR algorithm with column pivoting
!   used if the optional logical argument RANDOM_QR is present with the value true is described
!   in the references (3), (4), (5) and (6).
!
!   For further details, on computing low-rank matrix approximations from QR factorizations
!   with column pivoting, the QR-SVD or randomized QR algorithms, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Chan, T.F., and Hansen, P.C., 1992:
!           Some applications of the rank revealing QR factorization.
!           SIAM J. Sci. Statist. Comput., Volume 13, 727-741.
!
!   (3) Duersch, J.A., and Gu, M., 2017:
!           Randomized QR with column pivoting.
!           SIAM J. Sci. Comput., Volume 39, C263-C291.
!
!   (4) Martinsson, P.G., Quintana-Orti, G., Heavner, N., and Van de Geijn, R., 2017:
!           Householder QR factorization with randomization for column pivoting (HQRRP).
!           SIAM J. Sci. Comput., Volume 39, C96-C115.
!
!   (5) Xiao, J., Gu, M., and Langou, J., 2017:
!           Fast parallel randomized QR with column pivoting algorithms
!           for reliable low-rank matrix approximations.
!           IEEE 24th International Conference
!           on High Performance Computing (HiPC), IEEE, 2017, 233-242.
!
!   (6) Duersch, J.A., and Gu, M., 2020:
!           Randomized projection for rank-revealing matrix factorizations
!           and low-rank approximations.
!           SIAM Review, Volume 62, Issue 3, 661-682.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error, svd_error9
    use Utilities,       only : merror
    use QR_Procedures,   only : partial_qr_cmp_fixed_precision, ortho_gen_qr
    use Random,          only : partial_rqr_cmp_fixed_precision
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout), dimension(:,:)          :: mat
    real(stnd),  intent(out),   dimension(:),   pointer :: s
    real(stnd),  intent(out),   dimension(:,:), pointer :: v
    real(stnd),  intent(inout)                          :: relerr
!
    integer(i4b), intent(in), optional :: rng_alg, blk_size, nover,   &
                                          maxiter, max_francis_steps
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: random_qr, perfect_shift, bisect, dqds
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
!
    real(stnd)                              :: tmp, err, norm_mat, threshold
    real(stnd), dimension(:,:), allocatable :: work, u
    real(stnd), dimension(:),   allocatable :: diagr, beta
!
    integer      :: iok
    integer(i4b) :: m, n, mn, nsvd, np, i, j
    integer(i4b), dimension(:), allocatable :: ip
!
    logical(lgl) :: use_random_qr
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='rqr_svd_cmp_fixed_precision'
!
!   threshold0 IS THE DEFAULT EXTRA THRESHOLD FOR IMPROVING THE ACCURACY OF THE RESULTS
!   IN THE SVD PHASE OF THE QR-SVD ALGORITHM.
!
    real(stnd),  parameter :: threshold0=0.05_stnd
!   
    character, parameter   :: sort='d'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = size( mat, 2 )
!
    mn = min( m, n )
!
!   DEALLOCATE POINTERS ON ENTRY.
!
#ifndef _USE_NAGWARE
    if ( associated( s ) ) then
!
        deallocate( s )
!
    end if
!
    if ( associated( v ) ) then
!
        deallocate( v )
!
    end if
#endif
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( mn<=0_i4b  ) then
!
        relerr = one
!
        return
!
    end if
!
!   CHECK VALIDITY OF THE REQUESTED TOLERANCE FOR THE RELATIVE
!   ERROR OF THE COMPUTED SVD APPROXIMATION.
!
    tmp = four*ulp
!
    if ( relerr<=tmp .or. relerr>=one ) then
!
        call merror( name_proc//svd_error9 )
!
    end if
!
!   SET OPTIONAL ARGUMENTS.
!
!   DETERMINE IF A RANDOMIZED QR ALGORITHM MUST BE USED.
!
    if ( present(random_qr) ) then
        use_random_qr = random_qr
    else
        use_random_qr = false
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( ip(n), diagr(mn), beta(mn), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   COMPUTE THE FROBENIUS NORM OF THE INPUT MATRIX.
!
    norm_mat = norm( mat(:m,:n) )
!
!   SAVE THE PRESCRIBED TOLERANCE FOR LATER USE.
!
    threshold = relerr
!
!   DECREASE SLIGHTLY THE PRESCRIBED TOLERANCE FOR BETTER ACCURACY IN
!   THE SECOND PHASE OF THE QR-SVD ALGORITHM.
!
    if ( relerr>=two*threshold0 ) then
        relerr = max( relerr - threshold0, tmp )
    end if
!
!   FIRST STEP, DO PARTIAL QR WITH COLUMN PIVOTING OF mat AS
!           
!                   mat * P ≈ Q * R
!           
!   WHERE P IS A n-BY-n PERMUTATION MATRIX, Q AN m-BY-np MATRIX WITH
!   ORTHOGONAL COLUMNS AND R A np-BY-n UPPER TRAPEZOIDAL MATRIX.
!   np IS A FIRST ESTIMATE OF nsvd, THE RANK OF THE PARTIAL SVD APPROXIMATION,
!   WHICH IS SOUGHT.
!           
    if ( use_random_qr ) then
!           
!       USE A RANDOMIZED QR ALGORITHM.
!           
        call partial_rqr_cmp_fixed_precision( mat(:m,:n), relerr, diagr(:mn), beta(:mn), ip(:n),  &
                                              np, rng_alg=rng_alg, blk_size=blk_size, nover=nover )
!           
    else
!           
!       USE A STANDARD DETERMINISTIC QR ALGORITHM.
!           
        call partial_qr_cmp_fixed_precision( mat(:m,:n), relerr, diagr(:mn), beta(:mn), ip(:n), np )
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( work(np,n), u(np,np), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   RESTORE AND APPLY TRANSPOSE PERMUTATION TO THE UPPER TRIANGULAR FACTOR
!   OF THE QR FACTORIZATION OF mat , STORING RESULTS IN work ,
!           
!                   rmat = R * P'
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  np*n>=omp_limit                     .and.       &
                  n/i>=omp_limit2                     .and.       &
                  i>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC), PRIVATE(j,i)
!
    do j = 1_i4b, n
!
        i = ip(j)
!
        if ( j<=np ) then
!
            work(:j-1_i4b,i)   = mat(:j-1_i4b,j)
            work(j,i)          = diagr(j)
            work(j+1_i4b:np,i) = zero
!
        else
!
            work(:np,i) = mat(:np,j)
!
        end if
!
    end do
!
!$OMP END PARALLEL DO
!
!   GENERATE FIRST np COLUMNS OF ORTHOGONAL MATRIX Q OF QR DECOMPOSITION OF mat .
!
    call ortho_gen_qr( mat(:m,:np), beta(:np) )
!
!   SECOND STEP, COMPUTE SVD OF rmat = R * P'.
!
    call svd_cmp2( work(:np,:n), beta(:np), failure, u(:np,:np), sort=sort, &
                   maxiter=maxiter, max_francis_steps=max_francis_steps,    &
                   perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!   EXTRACT THE EXACT RANK AND RELATIVE ERROR OF THE PARTIAL SVD, WHICH FULFILLS THE THRESHOLD.
!
    diagr(:np) = (beta(:np)/norm_mat)**2
!
    tmp = one - sum( diagr(:np) )
    err = max( tmp, zero )
!
    tmp = threshold*threshold
!
    do i = np, 1_i4b, -1_i4b
!
        err = err + diagr(i)
!
        if ( err>=tmp ) exit
!
    end do
!
    nsvd = max( i, 1_i4b )
!
    err = one - sum( diagr(:nsvd) )
!
!   OUTPUT RELATIVE ERROR OF THE COMPUTED PARTIAL SVD FACTORIZATION
!   IN FROBENIUS NORM.
!
    relerr = sqrt( max( err, zero ) )
!
!   ALLOCATE INPUT POINTER ARGUMENTS FOR STORING THE PARTIAL SVD.
!
    allocate( s(nsvd), v(n,nsvd), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SAVE SINGULAR VALUES OF mat .
!
    s(:nsvd) = beta(:nsvd)
!
    if ( np<n ) then
!
!       SAVE RIGHT SINGULAR VECTORS OF mat .
!
        v(:n,:nsvd) = transpose( work(:nsvd,:n) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( ip, work, diagr, beta )
!
!       REALLOCATE WORKSPACE FOR COMPUTING LEFT SINGULAR VECTORS OF mat .
!
        allocate( work(m,nsvd), stat=iok  )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       NEXT COMPUTE LEFT SINGULAR VECTORS OF mat AS
!
!                 Q(:m,:np) * U
!
!       , WHERE Q IS THE ORTHOGONAL MATRIX OF QR DECOMPOSITION OF mat
!       AND U ARE THE LEFT SINGULAR VECTORS OF R * P'.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsvd, np, one, mat(1_i4b:m,1_i4b:np), m, &
                   u(1_i4b:np,1_i4b:nsvd), np, zero, work, m  )
#else
        work(:m,:nsvd) = matmul( mat(:m,:np), u(:np,:nsvd) )
#endif
!
!       SAVE LEFT SINGULAR VECTORS IN mat ON EXIT.
!
        mat(:m,:nsvd) = work(:m,:nsvd)
!
    else
!
!       SAVE RIGHT SINGULAR VECTORS OF mat .
!
        v(:n,:nsvd) = transpose( u(:nsvd,:n) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( ip, u, diagr, beta )
!
!       REALLOCATE WORKSPACE FOR COMPUTING LEFT SINGULAR VECTORS OF mat .
!
        allocate( u(m,nsvd), stat=iok  )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       NEXT COMPUTE LEFT SINGULAR VECTORS OF mat AS
!
!                    Q(:m,:np) * U
!
!       , WHERE Q IS THE ORTHOGONAL MATRIX OF QR DECOMPOSITION OF mat
!       AND U ARE THE LEFT SINGULAR VECTORS OF R * P'.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsvd, np, one, mat(1_i4b:m,1_i4b:np), m, &
                   work(1_i4b:np,1_i4b:nsvd), np, zero, u, m  )
#else
        u(:m,:nsvd) = matmul( mat(:m,:np), work(:np,:nsvd) )
#endif
!
!       SAVE LEFT SINGULAR VECTORS IN mat ON EXIT.
!
        mat(:m,:nsvd) = u(:m,:nsvd)
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( work, u )
!
!
! END OF SUBROUTINE rqr_svd_cmp_fixed_precision
! _____________________________________________
!
    end subroutine rqr_svd_cmp_fixed_precision
!
! =========================================================================================
!
    subroutine rqlp_svd_cmp( mat, s, leftvec, rightvec, failure, niter, random_qr,        &
                             truncated_qr, rng_alg, blk_size, nover, nover_svd, maxiter,  &
                             max_francis_steps, perfect_shift, bisect, dqds )
!
! Purpose
! _______
!
!   RQLP_SVD_CMP computes approximations of the nsvd largest singular values and associated
!   left and right singular vectors of a full m-by-n real matrix MAT using a four-step procedure,
!   which can be termed a QLP-SVD algorithm:
!
!   - First, a partial (or complete) QLP factorization of MAT is computed as
!
!                     MAT = Q * L * P 
!
!     , where Q is a m-by-m (or m-by-k if the factorization is only partial) orthogonal matrix, P is an
!     n-by-n (or k-by-n if the factorization is partial) orthogonal matrix and L is a lower m-by-n
!     (or k-by-k if the factorization is partial) triangular matrix.
!   - In a second step, the matrix product MAT * P' is computed and, at the user option, a number of
!     QR-QL iterations are performed (this is equivalent to subspace iterations) to improve the estimates
!     of the principal row and columns subspaces of MAT.
!   - In a third step, a Singular Value Decomposition (SVD) of the matrix product MAT * P' is
!     computed. The singular values and left singular vectors of this SVD are also estimates of the singular
!     values and left singular vectors of MAT.
!   - In a final step, estimates of the associated right singular vectors of MAT are
!     obtained by pre-multiplying P' by the right singular vectors in the SVD of this matrix product.
!
!   If the optional logical argument RANDOM_QR is used with the value true, a fast randomized (partial) QLP
!   factorization is used in the first phase of the QLP-SVD algorithm.
!   Furthermore if, in addition, the optional logical argument TRUNCATED_QR is used with the value true,
!   an even faster (but slightly less accurate) randomized partial and truncated QLP factorization will be
!   used in the first phase of the QLP-SVD algorithm.
!
!   nsvd is the target rank of the partial SVD, which is sought, and is equal to the size of the output
!   real vector argument S, i.e., nsvd = size( S ). If, nsvd = min( size(MAT,1) , size(MAT,2) ), a full
!   SVD of MAT is obtained with the same or higher accuracy than subroutines SVD_CMP or SVD_CMP2.
!   
!   See Further Details and the cited references for more information.
!
!   
! Arguments
! _________
!
!   MAT                 (INPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit, MAT is not modified by the routine.
!
!   S                   (OUTPUT) real(stnd), dimension(:)
!                       On exit, S(:) contains the first top nsvd singular values of MAT.
!                       The singular values are given in decreasing order and are
!                       positive or zero.
!
!                       The size of S must verify:
!
!                       - size( S ) = nsvd <= min( size(MAT,1) , size(MAT,2) ) .
!
!   LEFTVEC             (OUTPUT) real(stnd), dimension(:,:)
!                       On exit, the computed top nsvd left singular vectors of MAT. The left singular
!                       vector associated with the singular value S(j) is stored in the j-th column
!                       of LEFTVEC.
!
!                       The shape of LEFTVEC must verify:
!
!                       - size( LEFTVEC, 1 ) = size( MAT, 1 ) = m ,
!                       - size( LEFTVEC, 2 ) = size( S )      = nsvd .
!
!   RIGHTVEC            (OUTPUT) real(stnd), dimension(:,:)
!                       On exit, the computed top nsvd right singular vectors of MAT. The right singular
!                       vector associated with the singular value S(j) is stored in the j-th column of
!                       RIGHTVEC.
!
!                       The shape of RIGHTVEC must verify:  
!
!                       - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ,
!                       - size( RIGHTVEC, 2 ) = size( S )      = nsvd .
!
!   FAILURE             (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit, if the optional logical argument FAILURE is present, a test of the
!                       accuracy of the computed singular triplets is performed and in that
!                       case:
!
!                       - FAILURE = false : indicates successful exit;
!                       - FAILURE = true  : indicates that some singular values and vectors of MAT
!                         failed to converge in NITER QR-QL iterations.
!
!                       If FAILURE = true on exit, results are still useful, but some of the
!                       approximated singular triplets have a poor accuracy.
!
!   NITER               (INPUT, OPTIONAL) integer(i4b)
!                       The number of subspace iterations performed in the subroutine after the initial QLP 
!                       factorization and first subspace projection for computing the top nsvd singular triplets.
!
!                       NITER must be positive or null.
!
!                       By default, no subspace iterations are performed after the initial QLP factorization and first
!                       subspace projection.
!
!   RANDOM_QR           (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if RANDOM_QR is used with the value true, a fast randomized partial QLP
!                       factorization is used in the first phase of the QLP-SVD algorithm.
!
!                       By default, RANDOM_QR = false, i.e., a standard (partial) deterministic QLP factorization
!                       is used in the first phase of the QLP-SVD algorithm.
!
!   TRUNCATED_QR        (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if TRUNCATED_QR is used with the value true in addition to RANDOM_QR also
!                       set to true, a very fast (but slightly less accurate) randomized partial QLP factorization
!                       is used in the first phase of the QLP-SVD algorithm.
!
!                       By default, TRUNCATED_QR = false, i.e., a "standard" randomized (partial) QLP factorization
!                       is used in the first phase of the QLP-SVD algorithm if RANDOM_QR = true.
!
!   RNG_ALG             (INPUT, OPTIONAL) integer(i4b)
!                       On entry, a scalar integer to select the random (uniform) number generator used to build
!                       the random gaussian matrix in the randomized (partial) QLP phase of the QLP-SVD algorithm
!                       if RANDOM_QR = true.
!
!                       The possible values are:
!
!                       - ALG=1  : selects the Marsaglia's KISS random number generator;
!                       - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!                       - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!                       - ALG=4  : selects the Mersenne Twister random number generator;
!                       - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!                       - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!                       - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!                       - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!                       - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!                       - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                         random number generator;
!
!                       For other values, the current random number generator and its current state are not changed.
!                       Note further, that, on exit, the current random number generator is not reset to its previous
!                       value before the call to RQLP_SVD_CMP.
!
!                       See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   BLK_SIZE            (INPUT, OPTIONAL) integer(i4b)
!                       On entry, the block size used in the randomized partial QLP phase of the QLP-SVD algorithm
!                       if RANDOM_QR = true (and TRUNCATED_QR = false).
!
!                       BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!                       smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!                       See Further Details and the cited references for the meaning of the block size
!                       in the randomized (partial) QR or QLP algorithm.
!
!                       By default, BLK_SIZE is set to min( BLKSZ_QR, min(m,n) ), where parameter BLKSZ_QR is the default
!                       block size for QR related algorithms specified in module Select_Parameters.
!
!   NOVER               (INPUT, OPTIONAL) integer(i4b)
!                       The oversampling size used in the randomized partial QLP phase of the QLP-SVD algorithm
!                       if RANDOM_QR = true.
!
!                       NOVER must be positive or null and verify the relationships:
!
!                       - NOVER + BLK_SIZE              <= size( MAT, 1 ) if TRUNCATED_QR = false;
!                       - NOVER + NOVER_SVD + size( S ) <= size( MAT, 1 ) if TRUNCATED_QR = true.
!
!                       and is adjusted if necessary to verify these relationships in all cases.
!
!                       See Further Details and the cited references for the meaning and usefulness
!                       of the oversampling size in the partial randomized QR and QLP algorithms.
!
!                       By default, the oversampling size is set to:
!
!                       - 10                                   if TRUNCATED_QR = false;
!                       - max( (NOVER_SVD+size(S))/2_i4b, 10 ) if TRUNCATED_QR = true.
!
!   NOVER_SVD           (INPUT, OPTIONAL) integer(i4b)
!                       The oversampling size used in the SVD phase of the QLP-SVD algorithm
!                       for computing the top nsvd singular triplets.
!
!                       NOVER_SVD must be positive or null and verify the relationship:
!
!                       - NOVER_SVD + size( S ) <= min( size(MAT,1) , size(MAT,2) )
!
!                       and is adjusted if necessary to verify this relationship in all cases.
!
!                       See Further Details and the cited references for the meaning and usefulness
!                       of the oversampling size in the SVD phase of the QLP-SVD algorithm.
!
!                       By default, the oversampling size is set to 10.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm, which is used in the
!                       last step of the QLP-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       The optional argument MAX_FRANCIS_STEPS controls the maximum number
!                       of Francis sets (e.g., QR sweeps) of Givens rotations which must be
!                       saved before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm, which is used
!                       in the last step of the QLP-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is the minimum of nsvd and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       The optional argument PERFECT_SHIFT determines if a perfect shift strategy
!                       is used in the implicit QR algorithm in order to minimize the number of QR
!                       sweeps in the bidiagonal SVD algorithm, which is used in the last step of
!                       the QLP-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed QLP-SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   The QLP-SVD algorithm implemented in RQLP_SVD_CMP subroutine is a variation of the TXUV algorithm
!   described in the references (2), (5) and (6).
!
!   For further details, on computing low-rank matrix approximations with a QLP factorization,
!   the QLP-SVD (e.g., TXUV) algorithm or randomized (partial) QLP and QR factorizations, see:
!
!   (1) Stewart, G.W., 1999:
!           The QLP approximation to the singular value decomposition.
!           SIAM J. Sci. Comput., Volume 20, 1336-1348.
!
!   (2) Duersch, J.A., and Gu, M., 2017:
!           Randomized QR with column pivoting.
!           SIAM J. Sci. Comput., Volume 39, C263-C291.
!
!   (3) Martinsson, P.G., Quintana-Orti, G., Heavner, N., and Van de Geijn, R., 2017:
!           Householder QR factorization with randomization for column pivoting (HQRRP).
!           SIAM J. Sci. Comput., Volume 39, C96-C115.
!
!   (4) Xiao, J., Gu, M., and Langou, J., 2017:
!           Fast parallel randomized QR with column pivoting algorithms
!           for reliable low-rank matrix approximations.
!           IEEE 24th International Conference
!           on High Performance Computing (HiPC), IEEE, 2017, 233-242.
!
!   (5) Feng, Y., Xiao, J., and Gu, M., 2019:
!           Flip-flop spectrum-revealing QR factorizations and its applications
!           to singular value decomposition.
!           Electronic Transactions on Numerical Analysis (ETNA), Volume 51, 469-494.
!
!   (6) Duersch, J.A., and Gu, M., 2020:
!           Randomized projection for rank-revealing matrix factorizations
!           and low-rank approximations.
!           SIAM Review, Volume 62, Issue 3, 661-682.
!
!   (7) Huckaby, D.A., and Chan, T.F., 2003:
!           On the convergence of Stewart's QLP algorithm for approximating the SVD.
!           Numer. Algorithms, Volume 32, 287-316.
!
!   (8) Huckaby, D.A., and Chan, T.F., 2005:
!           Stewart's pivoted QLP decomposition for low-rank matrices
!           Numerical Linear Algebra with Applications, Volume 12, 153-159.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, assert_eq, merror
    use QR_Procedures,   only : qr_cmp, ortho_gen_qr
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout), dimension(:,:) :: mat
    real(stnd),  intent(out),   dimension(:)   :: s
    real(stnd),  intent(out),   dimension(:,:) :: leftvec, rightvec
!
    integer(i4b), intent(in), optional :: niter, rng_alg, blk_size, nover, nover_svd, &
                                          maxiter, max_francis_steps
!
    logical(lgl), intent(out), optional :: failure
    logical(lgl), intent(in),  optional :: random_qr, truncated_qr, perfect_shift, bisect, dqds
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: err
!
    real(stnd), dimension(:,:), allocatable :: work, v, v2
    real(stnd), dimension(:),   allocatable :: beta, tau
!
    integer      :: iok
    integer(i4b) :: m, n, mn, nsvd, niter2, nover_svd2, np, i
!
    logical(lgl) :: do_test, failure_svd
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='rqlp_svd_cmp'
!
!   niter0 IS THE DEFAULT NUMBER OF SUBSPACE ITERATIONS PERFORMED AFTER THE QLP FACTORIZATION.
!   nover_svd0 IS THE DEFAULT OVERSAMPLING SIZE FOR THE SVD PHASE.
!
    integer(i4b),  parameter :: niter0=0_i4b, nover_svd0=10_i4b
!   
!   eps0 IS THE THRESHOLD FOR THE ACCURACY TEST OF THE SINGULAR TRIPLETS.
!   
    real(stnd), parameter  :: eps0=0.0001_stnd
!   
    character, parameter   :: sort='d'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size(mat,1),i4b)       ,  &
                   int(size(leftvec,1),i4b)   ,  &
                   name_proc )
!
    n = assert_eq( int(size(mat,2),i4b)        ,  &
                   int(size(rightvec,1),i4b)   ,  &
                   name_proc )
!
    mn = min( m, n )
!
    nsvd = assert_eq( int(size(s),i4b)   ,         &
                      int(size(leftvec,2),i4b) ,   &
                      int(size(rightvec,2),i4b) ,  &
                      name_proc )
!
!   DETERMINE IF A TEST OF SINGULAR TRIPLETS IS REQUIRED BEFORE EXIT.
!
    do_test = present( failure )
!
    if ( do_test ) then
        failure = false
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( mn<=0_i4b  .or. nsvd<=0_i4b ) then
!
        return
!
    end if
!
    call assert( logical( nsvd>0_i4b .and. nsvd<=mn, lgl ), name_proc )
!
!   SET OPTIONAL ARGUMENTS.
!
!   DETERMINE THE NUMBER OF SUBSPACE ITERATIONS niter2 TO BE PERFORMED.
!
    if ( present(niter) ) then
        niter2 = max( niter, 0_i4b )
    else
        niter2 = niter0
    end if
!
!   DETERMINE THE OVERSAMPLING SIZE nover_svd2 FOR THE SVD PHASE.
!
    if ( present(nover_svd) ) then
        nover_svd2 = max( nover_svd, 0_i4b )
    else
        nover_svd2 = nover_svd0
    end if
!
!   CHECK VALIDITY OF THE OVERSAMPLING SIZE PARAMETER FOR THE SVD PHASE.
!
    np = min( nover_svd2 + nsvd, mn ) 
!
!   ALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    i = np
#else
    if ( niter2>0_i4b ) then
        i = m
    else
        i = np
    end if
#endif
!
    allocate( work(m,n), v(n,np), v2(np,i), tau(np), beta(np), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   MAKE A LOCAL COPY OF mat .
!
    work(:m,:n) = mat(:m,:n)
!
!   FIRST STEP, DO QLP FACTORIZATION OF mat AS
!           
!                   mat = Q * L * P
!           
!   WHERE Q AND P ARE, RESPECTIVELY, m-BY-np AND np-BY-n ORTHOGONAL
!   MATRICES AND L A np-BY-np LOWER TRIANGULAR MATRIX.
!           
    call qlp_cmp( work(:m,:n), beta(:np), tau(:np), random_qr=random_qr,         &
                  truncated_qr=truncated_qr, rng_alg=rng_alg, blk_size=blk_size, &
                  nover=nover         )
!
!   COMPUTE TRANSPOSE OF FIRST np ROWS OF ORTHOGONAL MATRIX P.
!           
    v(:n,:np) = transpose( work(:np,:n) )
!
    call ortho_gen_qr( v(:n,:np), tau(:np) )
!
!   COMPUTE MATRIX PRODUCT mat*P(:np,:n)'.
!
#ifdef _BLAS
    call gemm( 'N', 'N', m, np, n, one, mat(1_i4b:m,1_i4b:n), m,        &
               v(1_i4b:n,1_i4b:np), n, zero, work(1_i4b:m,1_i4b:np), m  )
#else
    work(:m,:np) = matmul( mat(:m,:n), v(:n,:np) )
#endif
!
!   DO niter2 SUBSPACE ITERATIONS.
!
    do i = 1_i4b, niter2
!
!       COMPUTE QR DECOMPOSITION OF MATRIX PRODUCT.
!
        call qr_cmp( work(:m,:np), tau(:np), beta(:np) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF MATRIX PRODUCT
!       TO OBTAIN AN ORTHONORMAL BASIS OF COLUMN SUBSPACE.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( work(:m,:np), beta(:np) )
!
!       COMPUTE COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'T', 'N', n, np, m, one, mat(1_i4b:m,1_i4b:n), m,        &
                   work(1_i4b:m,1_i4b:np), m, zero, v(1_i4b:n,1_i4b:np), n  )
#else
        v2(:np,:m) = transpose( work(:m,:np) )
!
        work(:np,:n) = matmul( v2(:np,:m), mat(:m,:n) )
!
        v(:n,:np) = transpose( work(:np,:n) )
#endif
!
!       COMPUTE QL DECOMPOSITION OF MATRIX PRODUCT.
!
        call qr_cmp( v(:n,:np), tau(:np), beta(:np) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QL DECOMPOSITION OF MATRIX PRODUCT
!       TO OBTAIN AN ORTHONORMAL BASIS OF ROW SUBSPACE.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( v(:n,:np), beta(:np) )
!
!       COMPUTE ROW SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, np, n, one, mat(1_i4b:m,1_i4b:n), m,        &
                   v(1_i4b:n,1_i4b:np), n, zero, work(1_i4b:m,1_i4b:np), m  )
#else
        work(:m,:np) = matmul( mat(:m,:n), v(:n,:np) )
#endif
!
    end do
!
!   COMPUTE SVD OF THE FINAL MATRIX OF ROW SUBSPACE PROJECTION.
!
    call svd_cmp( work(:m,:np), beta(:np), failure_svd, v=v2(:np,:np), sort=sort, &
                  maxiter=maxiter, max_francis_steps=max_francis_steps,           &
                  perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!   EXTRACT THE APPROXIMATE TOP nsvd SINGULAR VALUES OF mat .
!
    s(:nsvd) = beta(:nsvd)
!
!   EXTRACT THE APPROXIMATE TOP nsvd LEFT SINGULAR VECTORS OF mat .
!
    leftvec(:m,:nsvd) = work(:m,:nsvd)
!
!   COMPUTE THE APPROXIMATE TOP nsvd RIGHT SINGULAR VECTORS OF mat .
!
#ifdef _BLAS
    call gemm( 'N', 'N', n, nsvd, np, one, v(1_i4b:n,1_i4b:np), n,                &
               v2(1_i4b:np,1_i4b:nsvd), np, zero, rightvec(1_i4b:n,1_i4b:nsvd), n )
#else
    rightvec(:n,:nsvd) = matmul( v(:n,:np), v2(:np,:nsvd) )
#endif
!
!   TEST ACCURACY OF THE APPROXIMATE PARTIAL SVD IF REQUIRED.
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION mat*v(:m,:nsvd) - u(:n,:nsvd)*diag(s(:nsvd)),
!       WHERE u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF mat.
!
        work(:m,:nsvd) = leftvec(:m,:nsvd)*spread( s(:nsvd), dim=1, ncopies=m)
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsvd, n, one, mat(1_i4b:m,1_i4b:n), m,                 &
                   rightvec(1_i4b:n,1_i4b:nsvd), n, -one, work(1_i4b:m,1_i4b:nsvd), m  )
#else
        work(:m,:nsvd) = matmul( mat(:m,:n), rightvec(:n,:nsvd) ) - work(:m,:nsvd)
#endif
!
        beta(:nsvd) = norm( work(:m,:nsvd), dim=2_i4b )
!
        err = maxval( beta(:nsvd) )/( norm( mat )*real(mn,stnd) )
!
        failure = err>eps0 .or. failure_svd
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( work, v, v2, tau, beta )
!
!
! END OF SUBROUTINE rqlp_svd_cmp
! ______________________________
!
    end subroutine rqlp_svd_cmp
!
! =========================================================================================
!
    subroutine rqlp_svd_cmp2( mat, s, leftvec, rightvec, failure, niter, rng_alg, nover,         &
                              nover_svd, maxiter, max_francis_steps, perfect_shift, bisect, dqds )
!
! Purpose
! _______
!
!   RQLP_SVD_CMP2 computes approximations of the nsvd largest singular values and associated
!   left and right singular vectors of a full m-by-n real matrix MAT using a four-step procedure,
!   which can be termed a randomized QLP-SVD algorithm:
!
!   - First, an approximate and randomized partial QLP factorization of MAT is computed as
!
!                     MAT ≈ Q * L * P 
!
!     , where Q is a m-by-k matrix with orthonormal colmuns, P is an k-by-n matrix with orthonormal rows
!     and L is a lower k-by-k triangular matrix.
!   - In a second step, the matrix product MAT * P' is computed and, at the user option, a number of
!     QR-QL iterations are performed (this is equivalent to subspace iterations) to improve the estimates
!     of the principal row and columns subspaces of MAT.
!   - In a third step, a Singular Value Decomposition (SVD) of the matrix product MAT * P' is
!     computed. The singular values and left singular vectors of this SVD are also estimates of the singular
!     values and left singular vectors of MAT.
!   - In a final step, estimates of the associated right singular vectors of MAT are
!     obtained by pre-multiplying P' by the right singular vectors in the SVD of this matrix product.
!
!   A very fast randomized partial and truncated QLP factorization is used in the first phase of the QLP-SVD
!   algorithm.
!
!   nsvd is the target rank of the partial SVD, which is sought, and is equal to the size of the output
!   real vector argument S, i.e., nsvd = size( S ).
!   
!   See Further Details and the cited references for more information.
!
!   
! Arguments
! _________
!
!   MAT                 (INPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit, MAT is not modified by the routine.
!
!   S                   (OUTPUT) real(stnd), dimension(:)
!                       On exit, S(:) contains the first top nsvd singular values of MAT.
!                       The singular values are given in decreasing order and are
!                       positive or zero.
!
!                       The size of S must verify:
!
!                       - size( S ) = nsvd <= min( size(MAT,1) , size(MAT,2) ) .
!
!   LEFTVEC             (OUTPUT) real(stnd), dimension(:,:)
!                       On exit, the computed top nsvd left singular vectors of MAT. The left singular
!                       vector associated with the singular value S(j) is stored in the j-th column
!                       of LEFTVEC.
!
!                       The shape of LEFTVEC must verify:
!
!                       - size( LEFTVEC, 1 ) = size( MAT, 1 ) = m ,
!                       - size( LEFTVEC, 2 ) = size( S )      = nsvd .
!
!   RIGHTVEC            (OUTPUT) real(stnd), dimension(:,:)
!                       On exit, the computed top nsvd right singular vectors of MAT. The right singular
!                       vector associated with the singular value S(j) is stored in the j-th column of
!                       RIGHTVEC.
!
!                       The shape of RIGHTVEC must verify:  
!
!                       - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ,
!                       - size( RIGHTVEC, 2 ) = size( S )      = nsvd .
!
!   FAILURE             (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit, if the optional logical argument FAILURE is present, a test of the
!                       accuracy of the computed singular triplets is performed and in that
!                       case:
!
!                       - FAILURE = false : indicates successful exit;
!                       - FAILURE = true  : indicates that some singular values and vectors of MAT
!                         failed to converge in NITER QR-QL iterations.
!
!                       If FAILURE = true on exit, results are still useful, but some of the
!                       approximated singular triplets may have a poor accuracy.
!
!   NITER               (INPUT, OPTIONAL) integer(i4b)
!                       The number of subspace iterations performed in the subroutine after the initial QLP 
!                       factorization and first subspace projection for computing the top nsvd singular triplets.
!
!                       NITER must be positive or null.
!
!                       By default, no subspace iterations are performed after the initial QLP factorization and first
!                       subspace projection.
!
!   RNG_ALG             (INPUT, OPTIONAL) integer(i4b)
!                       On entry, a scalar integer to select the random (uniform) number generator used to build
!                       the random gaussian matrix in the randomized (partial) QLP phase of the QLP-SVD algorithm
!                       if RANDOM_QR = true.
!
!                       The possible values are:
!
!                       - ALG=1  : selects the Marsaglia's KISS random number generator;
!                       - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!                       - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!                       - ALG=4  : selects the Mersenne Twister random number generator;
!                       - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!                       - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!                       - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!                       - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!                       - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!                       - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                         random number generator;
!
!                       For other values, the current random number generator and its current state are not changed.
!                       Note further, that, on exit, the current random number generator is not reset to its previous
!                       value before the call to RQLP_SVD_CMP.
!
!                       See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   NOVER               (INPUT, OPTIONAL) integer(i4b)
!                       The oversampling size used in the randomized partial QLP phase of the QLP-SVD algorithm
!                       if RANDOM_QR = true.
!
!                       NOVER must be positive or null and verify the relationship:
!
!                       - NOVER + NOVER_SVD + size( S ) <= size( MAT, 1 ).
!
!                       and is adjusted if necessary to verify this relationship in all cases.
!
!                       See Further Details and the cited references for the meaning and usefulness
!                       of the oversampling size in the partial randomized QR and QLP algorithms.
!
!                       By default, the oversampling size is set to:
!
!                       - max( (NOVER_SVD+size(S))/2_i4b, 10 ).
!
!   NOVER_SVD           (INPUT, OPTIONAL) integer(i4b)
!                       The oversampling size used in the SVD phase of the QLP-SVD algorithm
!                       for computing the top nsvd singular triplets.
!
!                       NOVER_SVD must be positive or null and verify the relationship:
!
!                       - NOVER_SVD + size( S ) <= min( size(MAT,1) , size(MAT,2) )
!
!                       and is adjusted if necessary to verify this relationship in all cases.
!
!                       See Further Details and the cited references for the meaning and usefulness
!                       of the oversampling size in the SVD phase of the QLP-SVD algorithm.
!
!                       By default, the oversampling size is set to 10.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm, which is used in the
!                       last step of the QLP-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       The optional argument MAX_FRANCIS_STEPS controls the maximum number
!                       of Francis sets (e.g., QR sweeps) of Givens rotations which must be
!                       saved before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm, which is used
!                       in the last step of the QLP-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is the minimum of nsvd and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       The optional argument PERFECT_SHIFT determines if a perfect shift strategy
!                       is used in the implicit QR algorithm in order to minimize the number of QR
!                       sweeps in the bidiagonal SVD algorithm, which is used in the last step of
!                       the QLP-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed QLP-SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   The QLP-SVD algorithm implemented in RQLP_SVD_CMP2 subroutine is a variation of the TXUV algorithm
!   described in the references (2), (5) and (6) in which the initial randomized partial QR factorization
!   is replaced by the randomized partial and truncated QR algorithm described in the reference (9).
!
!   With this modification, RQLP_SVD_CMP2 subroutine is less accurate than RQLP_SVD_CMP subroutine,
!   but significantly faster and much less memory demanding.
!
!   For further details, on computing low-rank matrix approximations with a QLP factorization,
!   the QLP-SVD (e.g., TXUV) algorithm or randomized (partial) QLP and QR factorizations, see:
!
!   (1) Stewart, G.W., 1999:
!           The QLP approximation to the singular value decomposition.
!           SIAM J. Sci. Comput., Volume 20, 1336-1348.
!
!   (2) Duersch, J.A., and Gu, M., 2017:
!           Randomized QR with column pivoting.
!           SIAM J. Sci. Comput., Volume 39, C263-C291.
!
!   (3) Martinsson, P.G., Quintana-Orti, G., Heavner, N., and Van de Geijn, R., 2017:
!           Householder QR factorization with randomization for column pivoting (HQRRP).
!           SIAM J. Sci. Comput., Volume 39, C96-C115.
!
!   (4) Xiao, J., Gu, M., and Langou, J., 2017:
!           Fast parallel randomized QR with column pivoting algorithms
!           for reliable low-rank matrix approximations.
!           IEEE 24th International Conference
!           on High Performance Computing (HiPC), IEEE, 2017, 233-242.
!
!   (5) Feng, Y., Xiao, J., and Gu, M., 2019:
!           Flip-flop spectrum-revealing QR factorizations and its applications
!           to singular value decomposition.
!           Electronic Transactions on Numerical Analysis (ETNA), Volume 51, 469-494.
!
!   (6) Duersch, J.A., and Gu, M., 2020:
!           Randomized projection for rank-revealing matrix factorizations
!           and low-rank approximations.
!           SIAM Review, Volume 62, Issue 3, 661-682.
!
!   (7) Huckaby, D.A., and Chan, T.F., 2003:
!           On the convergence of Stewart's QLP algorithm for approximating the SVD.
!           Numer. Algorithms, Volume 32, 287-316.
!
!   (8) Huckaby, D.A., and Chan, T.F., 2005:
!           Stewart's pivoted QLP decomposition for low-rank matrices
!           Numerical Linear Algebra with Applications, Volume 12, 153-159.
!
!   (9) Mary, T., Yamazaki, I., Kurzak, J., Luszczek, P., Tomov, S., and Dongarra, J., 2015:
!           Performance of Random Sampling for Computing Low-rank Approximations of a
!           Dense Matrix on GPUs.
!           International Conference for High Performance Computing, Networking, Storage and Analysis (SC'15).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, assert_eq, merror, arth
    use Hous_Procedures, only : hous1
    use Sort_Procedures, only : reorder
    use QR_Procedures,   only : partial_qr_cmp, qr_cmp, ortho_gen_qr
    use Random,          only : random_seed_, normal_random_number3_
#ifdef _BLAS
    use BLAS_interfaces, only : gemm, axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout), dimension(:,:) :: mat
    real(stnd),  intent(out),   dimension(:)   :: s
    real(stnd),  intent(out),   dimension(:,:) :: leftvec, rightvec
!
    integer(i4b), intent(in), optional :: niter, rng_alg, nover, nover_svd, maxiter, max_francis_steps
!
    logical(lgl), intent(out), optional :: failure
    logical(lgl), intent(in),  optional :: perfect_shift, bisect, dqds
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: beta1, prod, err
!
    real(stnd), dimension(:,:), allocatable :: work, b, omega, rmat, v, v2
    real(stnd), dimension(:),   allocatable :: beta, tau
!
    integer      :: iok, alg
    integer(i4b) :: m, n, mn, nsvd, niter2, nover2, nover_svd2, np, nbr, i, l
    integer(i4b), dimension(:), allocatable :: ip, ipt
!
    logical(lgl) :: do_test, failure_svd
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='rqlp_svd_cmp2'
!
!   niter0 IS THE DEFAULT NUMBER OF SUBSPACE ITERATIONS PERFORMED AFTER THE QLP FACTORIZATION.
!   nover0 IS THE DEFAULT OVERSAMPLING SIZE.
!   nover_svd0 IS THE DEFAULT OVERSAMPLING SIZE FOR THE SVD PHASE.
!
    integer(i4b),  parameter :: niter0=0_i4b, nover0=10_i4b, nover_svd0=10_i4b
!   
!   eps0 IS THE THRESHOLD FOR THE ACCURACY TEST OF THE SINGULAR TRIPLETS.
!   
    real(stnd), parameter  :: eps0=0.0001_stnd
!   
    character, parameter   :: sort='d'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size(mat,1),i4b)       ,  &
                   int(size(leftvec,1),i4b)   ,  &
                   name_proc )
!
    n = assert_eq( int(size(mat,2),i4b)        ,  &
                   int(size(rightvec,1),i4b)   ,  &
                   name_proc )
!
    mn = min( m, n )
!
    nsvd = assert_eq( int(size(s),i4b)   ,         &
                      int(size(leftvec,2),i4b) ,   &
                      int(size(rightvec,2),i4b) ,  &
                      name_proc )
!
!   DETERMINE IF A TEST OF SINGULAR TRIPLETS IS REQUIRED BEFORE EXIT.
!
    do_test = present( failure )
!
    if ( do_test ) then
        failure = false
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( mn<=0_i4b  .or. nsvd<=0_i4b ) then
!
        return
!
    end if
!
    call assert( logical( nsvd>0_i4b .and. nsvd<=mn, lgl ), name_proc )
!
!   SET OPTIONAL ARGUMENTS.
!
!   SELECT AND SET THE RANDOM NUMBER GENERATOR TO BE USED IF REQUIRED.
!
    if ( present(rng_alg) ) then
!
        if ( rng_alg>=1_i4b .and. rng_alg<=10_i4b ) then
!
            alg = rng_alg
!
!           SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
            call random_seed_( alg=alg )
!
!           RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
            call random_seed_()
!
        end if
!
    end if
!
!   DETERMINE THE NUMBER OF SUBSPACE ITERATIONS niter2 TO BE PERFORMED.
!
    if ( present(niter) ) then
        niter2 = max( niter, 0_i4b )
    else
        niter2 = niter0
    end if
!
!   DETERMINE THE OVERSAMPLING SIZE nover_svd2 FOR THE SVD PHASE.
!
    if ( present(nover_svd) ) then
        nover_svd2 = max( nover_svd, 0_i4b )
    else
        nover_svd2 = nover_svd0
    end if
!
!   CHECK VALIDITY OF THE OVERSAMPLING SIZE PARAMETER FOR THE SVD PHASE.
!
    np = min( nover_svd2 + nsvd, mn ) 
!
!   DETERMINE THE OVERSAMPLING SIZE PARAMETER nover2 FOR THE QLP PHASE.
!
    if ( present(nover) ) then
        nover2 = max( nover, 0_i4b )
    else
        nover2 = max( np/2_i4b, nover0 )
    end if
!
!   CHECK VALIDITY OF THE OVERSAMPLING SIZE PARAMETER FOR THE QLP PHASE.
!   nbr IS THE NUMBER OF ROWS IN MATRICES b and omega .
!
    nbr = min( nover2 + np, m )
!
!   START COMPUTATIONS.
!
!   ALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    i = n
#else
    if ( niter2>0_i4b ) then
        i = max( m, n )
    else
        i = n
    end if
#endif
!
    allocate( b(nbr,n), omega(nbr,m), work(m,np), rmat(np,i), v(n,np), v2(np,np), &
              tau(np), beta(np), ip(n), ipt(n), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   FIRST STEP, DO QLP FACTORIZATION OF mat AS
!           
!                   mat = Q * L * P
!           
!   WHERE Q AND P ARE, RESPECTIVELY, m-BY-np AND np-BY-n ORTHOGONAL
!   MATRICES AND L A np-BY-np LOWER TRIANGULAR MATRIX.
!
!   USE A FAST RANDOMIZED PARTIAL QR FACTORIZATION OF mat WITHOUT UPDATES OF THE
!   COMPRESSION AND TRAILING MATRICES IN THE QR PHASE OF THE QLP FACTORIZATION TO
!   OBTAIN THE PARTIAL QR DECOMPOSITION
!
!                     MAT * N ≈ Q * R = Q * [  R11  R12  ]
!                  
!   WHERE N IS A n-BY-n PERMUTATION MATRIX, Q A m-BY-np MATRIX WITH ORTHONORMAL COLUMNS
!   AND R np-BY-n TRIANGULAR MATRIX.
!
!   GENERATE A nbr-BY-m IID GAUSSIAN MATRIX.
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
    call normal_random_number3_( omega(:nbr,:m) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
!    do i = 1_i4b, m
!
!        call normal_random_number3_( omega(:nbr,i) )
!
!    end do
!
!   COMPUTE COMPRESSION MATRIX b(:nbr,:n).
!
#ifdef _BLAS
    call gemm( 'N', 'N', nbr, n, m, one, omega(1_i4b:nbr,1_i4b:m), nbr,  &
               mat(1_i4b:m,1_i4b:n), m, zero, b(1_i4b:nbr,1_i4b:n), nbr  )
#else
    b(:nbr,:n) = matmul( omega(:nbr,:m), mat(:m,:n) )
#endif
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( omega )
!
!   NEXT DO PARTIAL QR WITH COLUMN PIVOTING ON COMPRESSION MATRIX b(:nbr,:n) TO FIND np PIVOTS.
!           
    call partial_qr_cmp( b(:nbr,:n), tau(:np), beta(:np), ip(:n), i )
!
!   SWAP CORRESPONDING COLUMNS OF mat(:m,:n) STORING RESULTS IN work(:m,:np).
!
    do i = 1_i4b, np
!
        work(:m,i) =  mat(:m,ip(i))
!
    end do
!
!   COMPUTE H(1), ... , H(np) AND DO QR DECOMPOSITION 
!   OF THE FIRST BLOCK OF mat .
!
    do i = 1_i4b, np
!    
!       RESTORE UPPER TRIANGULAR FACTOR IN THE QR FACTORIZATION OF COMPRESSION
!       MATRIX IN b(:np,:np) FOR LATER USE.
!
        b(i,i) = tau(i)                
!    
!       GENERATE ELEMENTARY REFLECTOR H(i) TO ANNIHILATE mat(i+1:m,i) .
!    
        call hous1( work(i:m,i), beta1, tau(i) )
!    
!       APPLY H(i) TO mat(i:m,i+1:krank) FROM THE LEFT.
!
        do l = i + 1_i4b, np
!
            prod = beta1*dot_product( work(i:m,i), work(i:m,l) )
#ifdef _BLAS
            call axpy( m-i+1_i4b, prod, work(i:m,i), 1_i4b, work(i:m,l), 1_i4b )
#else
            work(i:m,l) = work(i:m,l) + prod*work(i:m,i)
#endif
!    
        end do
!    
!       START UPDATE OF THE SUBMATRIX R12 BY FORMING
!    
!            R(:np,:np)*inv( b(:np,:np) )
!    
!       WHERE R(:np,:np) AND b(:np,:np) ARE, RESPECTIVELY, THE UPPER TRIANGULAR FACTORS (e.g., R11)
!       IN THE QR FACTORIZATIONS OF FIRST BLOCK OF mat AND COMPRESSION MATRICES, STORING THE RESULT
!       IN b(:np,:np) .
!
        beta(:i-1_i4b) = work(:i-1_i4b,i)
        beta(i)        = tau(i)
!    
        do l = 1_i4b, i-1_i4b
!
!            if ( b(l,i)==zero ) cycle
!
            beta(1_i4b:l) = beta(1_i4b:l) - b(l,i)*b(1_i4b:l,l)
!
        end do
!
        beta1 = sign( max( abs(b(i,i)), safmin ), b(i,i) )
!
        prod = one/beta1
!
        b(1_i4b:i,i) = prod*beta(1_i4b:i)
!
        b(i+1:np,i) = zero
!
    end do
!    
!   FINISH BLOCK UPDATE OF THE SUBMATRIX R12 AND STORED RESULTS IN rmat(:np,np+1:n).
!
    i = np + 1_i4b
    l  = n - np
!
#ifdef _BLAS
    call gemm( 'N', 'N', np, l, np, one, b(1_i4b:nbr,1_i4b:np), nbr,  &
               b(1_i4b:nbr,i:n), nbr, zero, rmat(1_i4b:np,i:n), np  )
#else
    rmat(:np,i:n) = matmul( b(:np,:np), b(:np,i:n) )
#endif
!
!   RESTORE R11 IN rmat(:np,:np).
!
    do i = 1_i4b, np
!
        rmat(:i-1_i4b,i)   = work(:i-1_i4b,i)
        rmat(i,i)          = tau(i)
        rmat(i+1_i4b:np,i) = zero
!
    end do
!
!   COMPUTE TRANSPOSE OF THE PERMUTATION N .     
!                
    ipt(ip) = arth( 1_i4b, 1_i4b, n )
!                
!   PERMUTE BACK IN ORIGINAL ORDER THE COLUMNS OF R, E.G., COMPUTE R * N' .  
!                
    call reorder( ipt(:n), rmat(:np,:n) )
!
!   TRANSPOSE R * N'.
!           
    v(:n,:np) = transpose( rmat(:np,:n) )
!
!   SECOND STEP, COMPUTE A DETERMINISTIC LQ DECOMPOSITION OF R * N' = L * P .
!
    call qr_cmp( v(:n,:np), beta(:np), tau(:np) )
!
!   START SVD PHASE.
!
!   COMPUTE TRANSPOSE OF FIRST np ROWS OF ORTHOGONAL MATRIX P, P(:np,:n)'.
!
    call ortho_gen_qr( v(:n,:np), tau(:np) )
!
!   COMPUTE MATRIX PRODUCT mat*P(:np,:n)'.
!
#ifdef _BLAS
    call gemm( 'N', 'N', m, np, n, one, mat(1_i4b:m,1_i4b:n), m,        &
               v(1_i4b:n,1_i4b:np), n, zero, work(1_i4b:m,1_i4b:np), m  )
#else
    work(:m,:np) = matmul( mat(:m,:n), v(:n,:np) )
#endif
!
!   DO niter2 SUBSPACE ITERATIONS.
!
    do i = 1_i4b, niter2
!
!       COMPUTE QR DECOMPOSITION OF MATRIX PRODUCT.
!
        call qr_cmp( work(:m,:np), tau(:np), beta(:np) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF MATRIX PRODUCT
!       TO OBTAIN AN ORTHONORMAL BASIS OF COLUMN SUBSPACE.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( work(:m,:np), beta(:np) )
!
!       COMPUTE COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'T', 'N', n, np, m, one, mat(1_i4b:m,1_i4b:n), m,        &
                   work(1_i4b:m,1_i4b:np), m, zero, v(1_i4b:n,1_i4b:np), n  )
#else
        rmat(:np,:m) = transpose( work(:m,:np) )
!
        b(:np,:n) = matmul( rmat(:np,:m), mat(:m,:n) )
!
        v(:n,:np) = transpose( b(:np,:n) )
#endif
!
!       COMPUTE QL DECOMPOSITION OF MATRIX PRODUCT.
!
        call qr_cmp( v(:n,:np), tau(:np), beta(:np) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QL DECOMPOSITION OF MATRIX PRODUCT
!       TO OBTAIN AN ORTHONORMAL BASIS OF ROW SUBSPACE.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( v(:n,:np), beta(:np) )
!
!       COMPUTE ROW SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, np, n, one, mat(1_i4b:m,1_i4b:n), m,        &
                   v(1_i4b:n,1_i4b:np), n, zero, work(1_i4b:m,1_i4b:np), m  )
#else
        work(:m,:np) = matmul( mat(:m,:n), v(:n,:np) )
#endif
!
    end do
!
!   COMPUTE SVD OF THE FINAL MATRIX OF ROW SUBSPACE PROJECTION.
!
    call svd_cmp( work(:m,:np), beta(:np), failure_svd, v=v2(:np,:np), sort=sort, &
                  maxiter=maxiter, max_francis_steps=max_francis_steps,           &
                  perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!   EXTRACT THE APPROXIMATE TOP nsvd SINGULAR VALUES OF mat .
!
    s(:nsvd) = beta(:nsvd)
!
!   EXTRACT THE APPROXIMATE TOP nsvd LEFT SINGULAR VECTORS OF mat .
!
    leftvec(:m,:nsvd) = work(:m,:nsvd)
!
!   COMPUTE THE APPROXIMATE TOP nsvd RIGHT SINGULAR VECTORS OF mat .
!
#ifdef _BLAS
    call gemm( 'N', 'N', n, nsvd, np, one, v(1_i4b:n,1_i4b:np), n,                &
               v2(1_i4b:np,1_i4b:nsvd), np, zero, rightvec(1_i4b:n,1_i4b:nsvd), n )
#else
    rightvec(:n,:nsvd) = matmul( v(:n,:np), v2(:np,:nsvd) )
#endif
!
!   TEST ACCURACY OF THE APPROXIMATE PARTIAL SVD IF REQUIRED.
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION mat*v(:m,:nsvd) - u(:n,:nsvd)*diag(s(:nsvd)),
!       WHERE u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF mat.
!
        work(:m,:nsvd) = leftvec(:m,:nsvd)*spread( s(:nsvd), dim=1, ncopies=m)
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsvd, n, one, mat(1_i4b:m,1_i4b:n), m,                 &
                   rightvec(1_i4b:n,1_i4b:nsvd), n, -one, work(1_i4b:m,1_i4b:nsvd), m  )
#else
        work(:m,:nsvd) = matmul( mat(:m,:n), rightvec(:n,:nsvd) ) - work(:m,:nsvd)
#endif
!
        beta(:nsvd) = norm( work(:m,:nsvd), dim=2_i4b )
!
        beta1 = norm( mat )
!
        err = maxval( beta(:nsvd) )/( beta1*real(mn,stnd) )
!
        failure = err>eps0 .or. failure_svd
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( b, work, rmat, v, v2, tau, beta, ip, ipt )
!
!
! END OF SUBROUTINE rqlp_svd_cmp2
! _______________________________
!
    end subroutine rqlp_svd_cmp2
!
! =========================================================================================
!
    subroutine rqlp_svd_cmp_fixed_precision( mat, relerr, s, leftvec, rightvec, failure,  &
                                             niter, random_qr, rng_alg, blk_size, nover,  &
                                             maxiter,  max_francis_steps, perfect_shift,  &
                                             bisect, dqds )
!
! Purpose
! _______
!
!   RQLP_SVD_CMP_FIXED_PRECISION computes approximations of the nsvd largest singular values and
!   associated left and right singular vectors of a full m-by-n real matrix MAT using a four-step
!   procedure, which can be termed a QLP-SVD algorithm:
!
!   - First, a partial QLP factorization of MAT is computed as
!
!                     MAT ≈ Q * L * P 
!
!     , where Q is a m-by-k orthogonal matrix, P' is an n-by-k orthogonal matrix and
!     L is a lower k-by-k triangular matrix.
!   - In a second step, the matrix product MAT * P' is computed and at, the user option, a number of
!     QR-QL iterations are performed (this is equivalent to subspace iterations) to improve the estimates
!     of the principal row and columns subspaces of MAT.
!   - In a third step, a Singular Value Decomposition (SVD) of the matrix product MAT * P' is
!     computed. The singular values and left singular vectors of this SVD are also estimates of the top singular
!     values and left singular vectors of MAT.
!   - In a final step, estimates of the associated right singular vectors of MAT are
!     obtained by pre-multiplying P' by the right singular vectors in the SVD of this matrix product.
!
!   By default, a standard deterministic BLAS2 QR factorization with column pivoting is used in the first phase
!   of the QLP step of the QLP-SVD algorithm. However, if the optional logical argument RANDOM_QR is used with
!   the value true, a fast randomized partial QR factorization is used in the QLP step of the QLP-SVD algorithm.
!
!   nsvd is the target rank of the partial SVD, which is sought, and this partial SVD  must have an
!   approximation error which fulfills:
!
!            ||MAT-rSVD||_F <= ||MAT||_F * relerr
!
!   , where rSVD is the computed partial SVD approximation, || ||_F is the Frobenius norm and
!   relerr is a prescribed accuracy tolerance for the relative error of the computed partial SVD
!   approximation, which is specified in the input real argument RELERR.
!   
!   In other words, nsvd is not known in advance and is determined in the subroutine. This explains
!   why the output real array arguments S, LEFTVEC and RIGHTVEC, which contain the computed singular
!   values and associated singular vectors of MAT on exit, must be declared in the calling program
!   as pointers.
!   
!   On exit, nsvd is equal to the size of the output real pointer argument S, which contains the
!   computed singular values, i.e., nsvd = size( S ) and the relative error in the Frobenius norm
!   of the computed partial SVD approximation is output in argument RELERR.
!
!   RQLP_SVD_CMP_FIXED_PRECISION first searches incrementally the best (e.g., smallest) partial QR approximation,
!   which fulfills the prescribed accuracy tolerance for the relative error. More precisely, the rank of
!   this partial QR approximation is increased progressively until the prescribed accuracy tolerance is
!   satisfied. This partial QR approximation is then transformed in a partial QLP factorization, which is
!   improved and adjusted precisely in a final step to obtain the smallest partial SVD approximation,
!   which satisfies the prescribed tolerance.
!
!   In all cases the relative error of the computed partial SVD approximation is output in argument RELERR.
!   
!   See Further Details and the cited references for more information.
!
!   
! Arguments
! _________
!
!   MAT                 (INPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit, MAT is not modified by the routine.
!
!   RELERR              (INPUT/OUTPUT) real(stnd)
!                       On entry, the requested accuracy tolerance for the relative error of
!                       the computed partial SVD approximation.
!
!                       The preset value for RELERR must be greater than 4*epsilon( RELERR ) and less than one.
!
!                       On exit, RELERR contains the relative error of the computed partial SVD
!                       approximation in the Frobenius norm:
!
!                       - RELERR = ||MAT-rSVD||_F / ||MAT||_F
!
!   S                   (OUTPUT) real(stnd), dimension(:), pointer
!                       On exit, S(:) contains estimates of the first top nsvd singular values of MAT.
!                       The singular values are given in decreasing order and are positive or zero.
!
!                       The statut of the pointer S must not be undefined on entry.
!                       If, on entry, the pointer S is already allocated, it will be first deallocated
!                       and then reallocated with the correct size.
!
!                       On exit, the size of the pointer S will verify:
!
!                       - size( S ) = nsvd <= min( size(MAT,1) , size(MAT,2) ) .
!
!   LEFTVEC             (OUTPUT) real(stnd), dimension(:,:), pointer
!                       On exit, the computed top nsvd left singular vectors of MAT. The left singular
!                       vector associated with the singular value S(j) is stored in the j-th column
!                       of LEFTVEC.
!
!                       The statut of the pointer LEFTVEC must not be undefined on entry.
!                       If, on entry, the pointer LEFTVEC is already allocated, it will be first deallocated
!                       and then reallocated with the correct shape.
!
!                       On exit, the shape of the pointer LEFTVEC will verify:  
!
!                       - size( LEFTVEC, 1 ) = size( MAT, 1 ) = m ,
!                       - size( LEFTVEC, 2 ) = size( S )      = nsvd .
!
!   RIGHTVEC            (OUTPUT) real(stnd), dimension(:,:), pointer
!                       On exit, the computed top nsvd right singular vectors of MAT. The right singular
!                       vector associated with the singular value S(j) is stored in the j-th column of
!                       RIGHTVEC.
!
!                       The statut of the pointer RIGHTVEC must not be undefined on entry.
!                       If, on entry, the pointer RIGHTVEC is already allocated, it will be first deallocated
!                       and then reallocated with the correct shape.
!
!                       On exit, the shape of the pointer RIGHTVEC will verify:  
!
!                       - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ,
!                       - size( RIGHTVEC, 2 ) = size( S )      = nsvd .
!
!   FAILURE             (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit, if the optional logical argument FAILURE is present, a test of the
!                       accuracy of the computed singular triplets is performed and in that
!                       case:
!
!                       - FAILURE = false : indicates successful exit;
!                       - FAILURE = true  : indicates that some singular values and vectors of MAT
!                         failed to converge in NITER QR-QL iterations for the requested accuracy
!                         tolerance for the relative error of the computed partial SVD approximation.
!
!                       If FAILURE = true on exit, results are still useful, but some of the
!                       approximated singular triplets have a poor accuracy.
!
!   NITER               (INPUT, OPTIONAL) integer(i4b)
!                       The number of subspace iterations performed in the subroutine after the initial QLP 
!                       factorization and first subspace projection for computing the top nsvd singular triplets.
!
!                       NITER must be positive or null.
!
!                       By default, no subspace iterations are performed after the initial QLP factorization and first
!                       subspace projection.
!
!   RANDOM_QR           (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if RANDOM_QR is used with the value true, a fast randomized partial QLP
!                       factorization is used in the first phase of the QLP-SVD algorithm.
!
!                       By default, RANDOM_QR = false, i.e., a standard (partial) deterministic QLP factorization
!                       is used in the first phase of the QLP-SVD algorithm.
!
!   RNG_ALG             (INPUT, OPTIONAL) integer(i4b)
!                       On entry, a scalar integer to select the random (uniform) number generator used to build
!                       the random gaussian matrix in the randomized (partial) QLP phase of the QLP-SVD algorithm
!                       if RANDOM_QR = true.
!
!                       The possible values are:
!
!                       - ALG=1  : selects the Marsaglia's KISS random number generator;
!                       - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!                       - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!                       - ALG=4  : selects the Mersenne Twister random number generator;
!                       - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!                       - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!                       - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!                       - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!                       - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!                       - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                         random number generator;
!
!                       For other values, the current random number generator and its current state are not changed.
!                       Note further, that, on exit, the current random number generator is not reset to its previous
!                       value before the call to RQLP_SVD_CMP.
!
!                       See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   BLK_SIZE            (INPUT, OPTIONAL) integer(i4b)
!                       On entry, the block size used in the randomized partial QLP phase of the QLP-SVD algorithm
!                       if RANDOM_QR = true.
!
!                       BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!                       smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!                       See Further Details and the cited references for the meaning of the block size
!                       in the randomized (partial) QR or QLP algorithm.
!
!                       By default, BLK_SIZE is set to min( BLKSZ_QR, min(m,n) ), where parameter BLKSZ_QR is the
!                       default block size for QR related algorithms specified in module Select_Parameters.
!
!   NOVER               (INPUT, OPTIONAL) integer(i4b)
!                       The oversampling size used in the randomized partial QLP phase of the QLP-SVD algorithm
!                       if RANDOM_QR = true.
!
!                       NOVER must be positive or null and verify the relationship:
!
!                       - NOVER + BLK_SIZE <= size( MAT, 1 )
!
!                       and is adjusted if necessary to verify these relationships in all cases.
!
!                       See Further Details and the cited references for the meaning and usefulness
!                       of the oversampling size in the partial randomized QR and QLP algorithms.
!
!                       By default, the oversampling size is set to 10.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm, which is used in the
!                       last step of the QLP-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       The optional argument MAX_FRANCIS_STEPS controls the maximum number
!                       of Francis sets (e.g., QR sweeps) of Givens rotations which must be
!                       saved before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm, which is used
!                       in the last step of the QLP-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is the minimum of nsvd and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       The optional argument PERFECT_SHIFT determines if a perfect shift strategy
!                       is used in the implicit QR algorithm in order to minimize the number of QR
!                       sweeps in the bidiagonal SVD algorithm, which is used in the last step of
!                       the QLP-SVD algorithm.
!
!                       See description of suboutine SVD_CMP for further details about
!                       this optional argument.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed QLP-SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   Usually, the problem of low-rank matrix approximation falls into two categories:
!
!   - the fixed-rank problem, where the rank parameter nsvd is given;
!   - the fixed-precision problem, where we seek a partial SVD factorization, rSVD, as small
!     as possible such that
!     
!     ||MAT-rSVD||_F <= eps
!
!     , where eps is a given accuracy tolerance.
!
!   RQLP_SVD_CMP_FIXED_PRECISION is dedicated to solve the fixed-precision problem. The fixed-rank
!   problem can be solved by subroutines RQLP_SVD_CMP or RQLP_SVD_CMP2.
!
!   The QLP-SVD algorithm implemented in RQLP_SVD_CMP_FIXED_PRECISION subroutine is a variation of
!   the TXUV algorithm described in the references (2), (5) and (6).
!
!   For further details, on computing low-rank matrix approximations with a QLP factorization,
!   the QLP-SVD (e.g., TXUV) algorithm or randomized (partial) QLP and QR factorizations, see:
!
!   (1) Stewart, G.W., 1999:
!           The QLP approximation to the singular value decomposition.
!           SIAM J. Sci. Comput., Volume 20, 1336-1348.
!
!   (2) Duersch, J.A., and Gu, M., 2017:
!           Randomized QR with column pivoting.
!           SIAM J. Sci. Comput., Volume 39, C263-C291.
!
!   (3) Martinsson, P.G., Quintana-Orti, G., Heavner, N., and Van de Geijn, R., 2017:
!           Householder QR factorization with randomization for column pivoting (HQRRP).
!           SIAM J. Sci. Comput., Volume 39, C96-C115.
!
!   (4) Xiao, J., Gu, M., and Langou, J., 2017:
!           Fast parallel randomized QR with column pivoting algorithms
!           for reliable low-rank matrix approximations.
!           IEEE 24th International Conference
!           on High Performance Computing (HiPC), IEEE, 2017, 233-242.
!
!   (5) Feng, Y., Xiao, J., and Gu, M., 2019:
!           Flip-flop spectrum-revealing QR factorizations and its applications
!           to singular value decomposition.
!           Electronic Transactions on Numerical Analysis (ETNA), Volume 51, 469-494.
!
!   (6) Duersch, J.A., and Gu, M., 2020:
!           Randomized projection for rank-revealing matrix factorizations
!           and low-rank approximations.
!           SIAM Review, Volume 62, Issue 3, 661-682.
!
!   (7) Huckaby, D.A., and Chan, T.F., 2003:
!           On the convergence of Stewart's QLP algorithm for approximating the SVD.
!           Numer. Algorithms, Volume 32, 287-316.
!
!   (8) Huckaby, D.A., and Chan, T.F., 2005:
!           Stewart's pivoted QLP decomposition for low-rank matrices
!           Numerical Linear Algebra with Applications, Volume 12, 153-159.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error, svd_error9
    use Utilities,       only : merror
    use QR_Procedures,   only : partial_qr_cmp_fixed_precision, qr_cmp, ortho_gen_qr
    use Random,          only : partial_rqr_cmp_fixed_precision
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout), dimension(:,:)          :: mat
    real(stnd),  intent(out),   dimension(:),   pointer :: s
    real(stnd),  intent(out),   dimension(:,:), pointer :: leftvec, rightvec
    real(stnd),  intent(inout)                          :: relerr
!
    integer(i4b), intent(in), optional :: niter, rng_alg, blk_size, nover,  &
                                          maxiter, max_francis_steps
!
    logical(lgl), intent(out), optional :: failure
    logical(lgl), intent(in),  optional :: random_qr, perfect_shift, bisect, dqds
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp, err, norm_mat, threshold
!
    real(stnd), dimension(:,:), allocatable :: work, rmat, v, v2
    real(stnd), dimension(:),   allocatable :: beta, tau
!
    integer      :: iok
    integer(i4b) :: m, n, mn, nsvd, niter2, np, i, j
    integer(i4b), dimension(:), allocatable :: ip
!
    logical(lgl) :: use_random_qr, do_test, failure_svd
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='rqlp_svd_cmp_fixed_precision'
!
!   niter0 IS THE DEFAULT NUMBER OF SUBSPACE ITERATIONS PERFORMED AFTER THE QLP FACTORIZATION.
!
    integer(i4b),  parameter :: niter0=0_i4b
!
!   threshold0 IS THE DEFAULT EXTRA THRESHOLD FOR IMPROVING THE ACCURACY OF THE RESULTS
!   IN THE SVD PHASE OF THE QLP-SVD ALGORITHM.
!
    real(stnd),  parameter :: threshold0=0.05_stnd
!   
!   eps0 IS THE THRESHOLD FOR THE ACCURACY TEST OF THE SINGULAR TRIPLETS.
!   
    real(stnd), parameter  :: eps0=0.0001_stnd
!   
    character, parameter   :: sort='d'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n = size( mat, 2 )
!
    mn = min( m, n )
!
!   DEALLOCATE POINTERS ON ENTRY.
!
#ifndef _USE_NAGWARE
    if ( associated( s ) ) then
!
        deallocate( s )
!
    end if
!
    if ( associated( leftvec ) ) then
!
        deallocate( leftvec )
!
    end if
!
    if ( associated( rightvec ) ) then
!
        deallocate( rightvec )
!
    end if
#endif
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( mn<=0_i4b  ) then
!
        relerr = one
!
        return
!
    end if
!
!   CHECK VALIDITY OF THE REQUESTED TOLERANCE FOR THE RELATIVE
!   ERROR OF THE COMPUTED SVD APPROXIMATION.
!
    tmp = four*ulp
!
    if ( relerr<=tmp .or. relerr>=one ) then
!
        call merror( name_proc//svd_error9 )
!
    end if
!
!   SET OPTIONAL ARGUMENTS.
!
!   DETERMINE IF A RANDOMIZED QR ALGORITHM MUST BE USED.
!
    if ( present(random_qr) ) then
        use_random_qr = random_qr
    else
        use_random_qr = false
    end if
!
!   DETERMINE IF A TEST OF SINGULAR TRIPLETS IS REQUIRED BEFORE EXIT.
!
    do_test = present( failure )
!
    if ( do_test ) then
        failure = false
    end if
!
!   DETERMINE THE NUMBER OF SUBSPACE ITERATIONS niter2 TO BE PERFORMED.
!
    if ( present(niter) ) then
        niter2 = max( niter, 0_i4b )
    else
        niter2 = niter0
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( work(m,n), ip(n), tau(mn), beta(mn), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   MAKE A LOCAL COPY OF mat .
!
    work(:m,:n) = mat(:m,:n)
!
!   COMPUTE THE FROBENIUS NORM OF THE INPUT MATRIX.
!
    norm_mat = norm( work(:m,:n) )
!
!   SAVE THE PRESCRIBED TOLERANCE FOR LATER USE.
!
    threshold = relerr
!
!   DECREASE SLIGHTLY THE PRESCRIBED TOLERANCE FOR BETTER ACCURACY IN
!   THE SECOND PHASE OF THE QR-SVD ALGORITHM.
!
    if ( relerr>=two*threshold0 ) then
        relerr = max( relerr - threshold0, tmp )
    end if
!
!   FIRST STEP, DO PARTIAL QR WITH COLUMN PIVOTING OF mat AS
!           
!                   mat * N ≈ Q * R
!           
!   WHERE N IS A n-BY-n PERMUTATION MATRIX, Q AN m-BY-np MATRIX WITH
!   ORTHOGONAL COLUMNS AND R A np-BY-n UPPER TRAPEZOIDAL MATRIX.
!   np IS A FIRST ESTIMATE OF nsvd, THE RANK OF THE PARTIAL SVD APPROXIMATION,
!   WHICH IS SOUGHT.
!           
    if ( use_random_qr ) then
!           
!       USE A RANDOMIZED QR ALGORITHM.
!           
        call partial_rqr_cmp_fixed_precision( work(:m,:n), relerr, tau(:mn), beta(:mn), ip(:n),   &
                                              np, rng_alg=rng_alg, blk_size=blk_size, nover=nover )
!           
    else
!           
!       USE A STANDARD DETERMINISTIC QR ALGORITHM.
!           
        call partial_qr_cmp_fixed_precision( work(:m,:n), relerr, tau(:mn), beta(:mn), ip(:n), np )
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( rmat(np,n), v(n,np), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   FINISH THE PARTIAL QLP FACTORIZATION OF mat, STORING THE RESULTS IN ARRAYS rmat, v, beta AND tau.
!   THE RANK OF THE PARTIAL QLP FACTORIZATION IS np, WHICH IS A FIRST ESTIMATE OF nsvd, THE RANK
!   OF THE PARTIAL SVD APPROXIMATION, WHICH IS SOUGHT.
!
!   FIRST STEP, RESTORE AND APPLY TRANSPOSE PERMUTATION TO THE UPPER TRIANGULAR FACTOR
!   OF THE QR FACTORIZATION OF mat , STORING RESULTS IN rmat ,
!           
!                   rmat = R * N'
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  np*n>=omp_limit                     .and.       &
                  n/i>=omp_limit2                     .and.       &
                  i>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC), PRIVATE(j,i)
!
    do j = 1_i4b, n
!
        i = ip(j)
!
        if ( j<=np ) then
!
            rmat(:j-1_i4b,i)   = work(:j-1_i4b,j)
            rmat(j,i)          = tau(j)
            rmat(j+1_i4b:np,i) = zero
!
        else
!
            rmat(:np,i) = work(:np,j)
!
        end if
!
    end do
!
!$OMP END PARALLEL DO
!
!   SECOND STEP, COMPUTE A DETERMINISTIC LQ DECOMPOSITION OF rmat = R * N' = L * P .
!           
    v(:n,:np) = transpose( rmat(:np,:n) )
!
    call qr_cmp( v(:n,:np), beta(:np), tau(:np) )
!
!   COMPUTE TRANSPOSE OF FIRST np ROWS OF ORTHOGONAL MATRIX P.
!
    call ortho_gen_qr( v(:n,:np), tau(:np) )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( rmat, ip )
!
!   ALLOCATE WORK ARRAY.
!
#ifdef _BLAS
    i = np
#else
    if ( niter2>0_i4b ) then
        i = m
    else
        i = np
    end if
#endif
!
    allocate( v2(np,i), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   COMPUTE MATRIX PRODUCT mat*P(:np,:n)'.
!
#ifdef _BLAS
    call gemm( 'N', 'N', m, np, n, one, mat(1_i4b:m,1_i4b:n), m,        &
               v(1_i4b:n,1_i4b:np), n, zero, work(1_i4b:m,1_i4b:np), m  )
#else
    work(:m,:np) = matmul( mat(:m,:n), v(:n,:np) )
#endif
!
!   DO niter2 SUBSPACE ITERATIONS.
!
    do i = 1_i4b, niter2
!
!       COMPUTE QR DECOMPOSITION OF MATRIX PRODUCT.
!
        call qr_cmp( work(:m,:np), tau(:np), beta(:np) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF MATRIX PRODUCT
!       TO OBTAIN AN ORTHONORMAL BASIS OF COLUMN SUBSPACE.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( work(:m,:np), beta(:np) )
!
!       COMPUTE COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'T', 'N', n, np, m, one, mat(1_i4b:m,1_i4b:n), m,        &
                   work(1_i4b:m,1_i4b:np), m, zero, v(1_i4b:n,1_i4b:np), n  )
#else
        v2(:np,:m) = transpose( work(:m,:np) )
!
        work(:np,:n) = matmul( v2(:np,:m), mat(:m,:n) )
!
        v(:n,:np) = transpose( work(:np,:n) )
#endif
!
!       COMPUTE QL DECOMPOSITION OF MATRIX PRODUCT.
!
        call qr_cmp( v(:n,:np), tau(:np), beta(:np) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QL DECOMPOSITION OF MATRIX PRODUCT
!       TO OBTAIN AN ORTHONORMAL BASIS OF ROW SUBSPACE.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( v(:n,:np), beta(:np) )
!
!       COMPUTE ROW SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, np, n, one, mat(1_i4b:m,1_i4b:n), m,        &
                   v(1_i4b:n,1_i4b:np), n, zero, work(1_i4b:m,1_i4b:np), m  )
#else
        work(:m,:np) = matmul( mat(:m,:n), v(:n,:np) )
#endif
!
    end do
!
!   COMPUTE SVD OF THE FINAL MATRIX OF ROW SUBSPACE PROJECTION.
!
    call svd_cmp( work(:m,:np), beta(:np), failure_svd, v=v2(:np,:np), sort=sort, &
                  maxiter=maxiter, max_francis_steps=max_francis_steps,           &
                  perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!   EXTRACT THE EXACT RANK AND RELATIVE ERROR OF THE PARTIAL SVD, WHICH FULFILLS THE THRESHOLD.
!
    tau(:np) = (beta(:np)/norm_mat)**2
!
    tmp = one - sum( tau(:np) )
    err = max( tmp, zero )
!
    tmp = threshold*threshold
!
    do i = np, 1_i4b, -1_i4b
!
        err = err + tau(i)
!
        if ( err>=tmp ) exit
!
    end do
!
    nsvd = max( i, 1_i4b )
!
    err = one - sum( tau(:nsvd) )
!
!   OUTPUT RELATIVE ERROR OF THE COMPUTED PARTIAL SVD FACTORIZATION
!   IN FROBENIUS NORM.
!
    relerr = sqrt( max( err, zero ) )
!
!   ALLOCATE INPUT POINTER ARGUMENTS FOR STORING THE PARTIAL SVD.
!
    allocate( s(nsvd), leftvec(m,nsvd), rightvec(n,nsvd), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   EXTRACT THE APPROXIMATE TOP nsvd SINGULAR VALUES OF mat .
!
    s(:nsvd) = beta(:nsvd)
!
!   EXTRACT THE APPROXIMATE TOP nsvd LEFT SINGULAR VECTORS OF mat .
!
    leftvec(:m,:nsvd) = work(:m,:nsvd)
!
!   COMPUTE THE APPROXIMATE TOP nsvd RIGHT SINGULAR VECTORS OF mat .
!
#ifdef _BLAS
    call gemm( 'N', 'N', n, nsvd, np, one, v(1_i4b:n,1_i4b:np), n,                &
               v2(1_i4b:np,1_i4b:nsvd), np, zero, rightvec(1_i4b:n,1_i4b:nsvd), n )
#else
    rightvec(:n,:nsvd) = matmul( v(:n,:np), v2(:np,:nsvd) )
#endif
!
!   TEST ACCURACY OF THE APPROXIMATE PARTIAL SVD IF REQUIRED.
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION mat*v(:m,:nsvd) - u(:n,:nsvd)*diag(s(:nsvd)),
!       WHERE u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF mat.
!
        work(:m,:nsvd) = leftvec(:m,:nsvd)*spread( s(:nsvd), dim=1, ncopies=m)
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsvd, n, one, mat(1_i4b:m,1_i4b:n), m,                 &
                   rightvec(1_i4b:n,1_i4b:nsvd), n, -one, work(1_i4b:m,1_i4b:nsvd), m  )
#else
        work(:m,:nsvd) = matmul( mat(:m,:n), rightvec(:n,:nsvd) ) - work(:m,:nsvd)
#endif
!
        beta(:nsvd) = norm( work(:m,:nsvd), dim=2_i4b )
!
        err = maxval( beta(:nsvd) )/( norm( mat )*real(mn,stnd) )
!
        failure = err>eps0 .or. failure_svd
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( work, v, v2, tau, beta )
!
!
! END OF SUBROUTINE rqlp_svd_cmp_fixed_precision
! ______________________________________________
!
    end subroutine rqlp_svd_cmp_fixed_precision
!
!
! =========================================================================================
!        RANDOMIZED PARTIAL OR COMPLETE QLP DECOMPOSITION OF A RECTANGULAR MATRIX
! =========================================================================================
!
!
    subroutine qlp_cmp( mat, beta, tau, lmat, qmat, pmat, random_qr, truncated_qr,   &
                        rng_alg, blk_size, nover )
!
! Purpose
! _______
!
!   QLP_CMP computes a partial or complete QLP factorization of a m-by-n matrix MAT:
!
!                     MAT ≈ Q * L * P 
!
!   , where Q is a m-by-krank orthogonal matrix, P is a krank-by-n orthogonal matrix and L is a
!   krank-by-krank lower triangular matrix. If krank = min(m,n), the QLP factorization is complete
!   and  MAT = Q * L * P .
!
!   The QLP factorization is obtained by a two-step algorithm:
!
!   - first, a partial (or complete) QR factorization with column pivoting of MAT is computed;
!   - in a second step, a LQ Decomposition of the (permuted) upper triangular or trapezoidal (e.g., if n>m)
!     factor, R, of this QR decomposition is computed.
!
!   By default, a standard deterministic QR factorization with column pivoting is used in the first phase
!   of the QLP algorithm. However, if the optional logical argument RANDOM_QR is used with the value true,
!   an alternate fast randomized (partial) QR factorization is used in the first phase of the QLP algorithm.
!   Furthermore if, in addition, the optional logical argument TRUNCATED_QR is used with the value true,
!   an even faster (but less accurate) randomized partial and truncated QR factorization will be used in the first
!   phase of the QLP algorithm. In all cases, a deterministic blocked LQ factorization is used in the second step
!   of the QLP factorization.
!
!   At the user option, the QLP factorization can also be only partial, e.g., the subroutine stops the computations
!   when the numbers of columns of Q and of rows of P are equal to a predefined value equals to
!   krank = size( BETA ) = size( TAU ).
!   
!   The QLP decomposition provides a reasonable and cheap estimate of the Singular Value Decomposition (SVD) of a matrix
!   when this matrix has a low rank or a significant gap in its singular values spectrum.
!   
!   See Further Details and the cited references for more information.
!   
!
! Arguments
! _________
!
!   MAT           (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                 On entry, the real m-by-n matrix to be decomposed.
!
!                 On exit, MAT has been overwritten by details of its (partial)
!                 QLP factorization.
!
!                 See Further Details.
!
!   BETA          (OUTPUT) real(stnd), dimension(:)
!                 On exit, the scalars factors of the elementary reflectors defining Q.
!
!                 See Further Details.
!
!                 The size of BETA must verify:
!
!                 - size( BETA ) = krank <= min( m , n ) = min( size(MAT,1) , size(MAT,2) ).
!
!   TAU           (OUTPUT) real(stnd), dimension(:)
!                 On exit, the scalars factors of the elementary reflectors defining P.
!
!                 See Further Details.
!
!                 The size of TAU must verify:
!
!                 - size( TAU ) = krank <= min( m , n ) = min( size(MAT,1) , size(MAT,2) ).
!
!   LMAT          (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!                 On exit, LMAT stores the lower triangular matrix L in the (partial)
!                 QLP factorization of MAT. The diagonal elements of LMAT (called the L-values)
!                 are estimates of the singular values of MAT if there is a significant gap in
!                 the singular values spectrum of MAT.
!
!                 See Further Details.
!
!                 The shape of LMAT must verify:
!
!                 - size( LMAT, 1 ) = size( LMAT, 2 ) = krank.
!
!   QMAT          (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!                 On exit, QMAT stores the first krank columns of the orthogonal matrix Q in the (partial)
!                 QLP factorization of MAT.
!
!                 See Further Details.
!
!                 The shape of QMAT must verify:
!
!                 - size( QMAT, 1 ) = m.
!                 - size( QMAT, 2 ) = krank.
!
!   PMAT          (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!                 On exit, PMAT stores the first krank rows of the orthogonal matrix P in the (partial)
!                 QLP factorization of MAT.
!
!                 See Further Details.
!
!                 The shape of PMAT must verify:
!
!                 - size( QMAT, 1 ) = krank.
!                 - size( QMAT, 2 ) = n.
!
!   RANDOM_QR     (INPUT, OPTIONAL) logical(lgl)
!                 On entry, if RANDOM_QR is used with the value true, a fast randomized (partial) QR
!                 factorization with column pivoting is used in the first phase of the QLP algorithm.
!
!                 By default, RANDOM_QR = false, i.e., A standard deterministic (partial) QR factorization
!                 with column pivoting is used in the first phase of the QLP algorithm.
!
!   TRUNCATED_QR  (INPUT, OPTIONAL) logical(lgl)
!                 On entry, if TRUNCATED_QR is used with the value true in addition to RANDOM_QR also
!                 set to true, a very fast (but less accurate) randomized partial and truncated
!                 QR factorization is used in the first phase of the QLP algorithm.
!
!                 By default, TRUNCATED_QR = false, i.e., a "standard" randomized (partial) QR factorization
!                 with column pivoting is used in the first phase of the QLP algorithm if RANDOM_QR = true.
!
!   RNG_ALG       (INPUT, OPTIONAL) integer(i4b)
!                 On entry, a scalar integer to select the random (uniform) number generator used to build
!                 the random gaussian matrix in the randomized (partial) QR phase of the QLP algorithm,
!                 if RANDOM_QR = true.
!
!                 The possible values are:
!
!                 - ALG=1  : selects the Marsaglia's KISS random number generator;
!                 - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!                 - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!                 - ALG=4  : selects the Mersenne Twister random number generator;
!                 - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!                 - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!                 - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!                 - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!                 - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!                 - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                   random number generator;
!
!                 For other values, the current random number generator and its current state are not changed.
!                 Note further, that, on exit, the current random number generator is not reset to its previous
!                 value before the call to QLP_CMP.
!
!                 See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   BLK_SIZE      (INPUT, OPTIONAL) integer(i4b)
!                 On entry, the block size used in the randomized (partial) QR phase of the QLP algorithm,
!                 if RANDOM_QR = true (and TRUNCATED_QR = false).
!
!                 BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!                 smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!                 See Further Details and the cited references for the meaning of the block size
!                 in the randomized (partial) QR algorithm.
!
!                 By default, BLK_SIZE is set to min( BLKSZ_QR, min(m,n) ), where parameter BLKSZ_QR is the
!                 default block size for QR related algorithms specified in module Select_Parameters.
!
!   NOVER         (INPUT, OPTIONAL) integer(i4b)
!                 The oversampling size used in the randomized (partial) QR phase of the QLP algorithm,
!                 if RANDOM_QR = true.
!
!                 NOVER must be positive or null and verify the relationships:
!
!                 - NOVER + BLK_SIZE     <= size( MAT, 1 ) if TRUNCATED_QR = false;
!                 - NOVER + size( BETA ) <= size( MAT, 1 ) if TRUNCATED_QR = true.
!
!                 and is adjusted if necessary to verify these relationships in all cases.
!
!                 See Further Details and the cited references for the meaning and usefulness
!                 of the oversampling size in the randomized (partial) QR algorithms.
!
!                 By default, the oversampling size is set to:
!
!                 - 10                          if TRUNCATED_QR = false;
!                 - max( size(BETA)/2_i4b, 10 ) if TRUNCATED_QR = true.
!
!
! Further Details
! _______________
!
!   QLP_CMP first computes a (partial or complete) QR factorization with column pivoting
!   of the m-by-n matrix MAT:
!
!                     MAT * N = Q * R
!
!   , where N is a n-by-n permutation matrix, R is a upper triangular or trapezoidal
!   (i.e., if n>m) matrix and Q is a m-by-m orthogonal matrix.
!
!   If the optional logical argument RANDOM_QR is used with the value true, a fast randomized
!   partial (and truncated, if the optional logical argument TRUNCATED_QR is also used with the value
!   true) QR factorization with column pivoting is used in this first phase of the QLP algorithm.
!
!   At the user option, this QR factorization can also be only partial, e.g., the subroutine
!   ends when the numbers of columns of Q is equal to a predefined value equals to
!   krank = size( BETA ) = size( TAU ).
!
!   This leads implicitly to the following partition of Q:
!
!   [  Q1  Q2  ]
!
!   where Q1 is a m-by-krank orthonormal matrix and Q2 is a m-by-(m-krank) orthonormal matrix
!   orthogonal to Q1, and to the following corresponding partition of R:
!
!   [  R11  R12  ]
!
!   [  R21  R22  ]
!
!   where R11 is a krank-by-krank triangular matrix, R21 is zero by construction,
!   R12 is a full krank-by-(n-krank) matrix and R22 is a full (m-krank)-by-(n-krank)
!   matrix.
!
!   In a second step, QLP_CMP computes a deterministic LQ factorization of the matrix
!   product:
!
!   R * N' = L * P
!
!   if the first QR factorization is complete, or of the matrix product:
!
!   [  R11  R12  ] * N' = L * P
!
!   if this first QR factorization is only partial. This leads to the (partial) QLP factorization
!   of MAT:
!
!                     MAT ≈ Q1 * L * P
!
!   where L is a krank-by-krank triangular matrix and P is a krank-by-n matrix with orthonormal rows.
!
!   The properties of the QLP factorization and when it can be used as a good proxy for the (partial
!   or complete) SVD of a matrix are discussed in the references (2), (7), (8) and (9).
!
!   The computations are parallelized if OPENMP is used. However, note that QLP_CMP
!   uses a standard "BLAS2" algorithm without any blocking for performing the first QR factorization
!   with column pivoting if the optional logical argument RANDOM_QR is not used or used with the value
!   false. On the other hand, QLP_CMP uses fast randomized and blocked QR algorithms with column
!   pivoting (see the references (3), (4), (5) and (6)) if RANDOM_QR is used with the value true.
!   These randomized algorithms are thus particularly efficient for large matrices.
!
!   The standard deterministic BLAS2 algorithm for computing a QR factorization with column pivoting
!   is described in the reference (1). The randomized partial QR algorithm with column pivoting
!   used if the optional logical argument RANDOM_QR is present with the value true is described
!   in the references (3), (4) and (5). Finally, the randomized partial and truncated QR algorithm
!   with column pivoting used if both the optional logical arguments RANDOM_QR and TRUNCATED_QR are
!   present with the value true is described in the reference (6). This algorithm is the fastest, but
!   less accurate than the randomized partial QR algorithm with column pivoting described in the
!   references (3), (4) and (5).
!
!   In all cases, QLP_CMP uses an efficient (but deterministic) blocked algorithm for performing the
!   LQ factorization in the second step of the QLP decomposition. The LQ factorization is described
!   in the reference (1).
!
!   On exit, the matrix Q is represented as a product of elementary reflectors
!
!            Q = H(1) * H(2) * ... * H(krank), where krank = size( BETA ) <= min( m , n ).
!
!   Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!   where beta is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in MAT(i:m,i) and beta in BETA(i). Note also that v(i) = 1.
!
!   On exit of QLP_CMP, the orthonormal matrix Q stored in factored form in MAT 
!   can be generated by a call to suboutine ORTHO_GEN_QR with arguments MAT and BETA.
!   Alternatively, QLP_CMP computes the first krank columns of Q explicitly if the optional
!   array argument QMAT is present.
!
!   The matrix P is represented as a product of elementary reflectors
!
!           Q = G(k) * . . . * G(2) * G(1), where  krank = size( TAU ) <= min( m , n ).
!
!   Each G(i) has the form
!
!            G(i) = I + tau * ( u * u' ) ,
!                      
!   where tau is a real scalar and u is a real n-element vector with u(1:i-1) = 0.
!   u(i:n) is stored on exit in MAT(i,i:n) and tau in TAU(i). Note also that u(i) = 1.
!
!   On exit of QLP_CMP, the orthonormal matrix P stored in factored in MAT
!   can be generated by a call to suboutine ORTHO_GEN_LQ with arguments MAT and TAU.
!   Alternatively, QLP_CMP computes the first krank rows of P explicitly if the optional
!   array argument PMAT is present.
!
!   Finally, QLP_CMP outputs the krank-by-krank lower triangular matrix L in the optional
!   array argument LMAT. If LMAT is not specified in the QLP_CMP call, the L factor of the
!   QLP decomposition is not stored on exit.
!
!   For further details on the QLP factorization and its use, or randomized QR and QLP algorithms, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Stewart, G.W., 1999:
!           The QLP approximation to the singular value decomposition.
!           SIAM J. Sci. Comput., Volume 20, 1336-1348.
!
!   (3) Duersch, J.A., and Gu, M., 2017:
!           Randomized QR with column pivoting.
!           SIAM J. Sci. Comput., Volume 39, C263-C291.
!
!   (4) Martinsson, P.G., Quintana-Orti, G., Heavner, N., and Van de Geijn, R., 2017:
!           Householder QR factorization with randomization for column pivoting (HQRRP).
!           SIAM J. Sci. Comput., Volume 39, C96-C115.
!
!   (5) Duersch, J.A., and Gu, M., 2020:
!           Randomized projection for rank-revealing matrix factorizations
!           and low-rank approximations.
!           SIAM Review, Volume 62, Issue 3, 661-682.
!
!   (6) Mary, T., Yamazaki, I., Kurzak, J., Luszczek, P., Tomov, S., and Dongarra, J., 2015:
!           Performance of Random Sampling for Computing Low-rank Approximations of a
!           Dense Matrix on GPUs.
!           International Conference for High Performance Computing, Networking, Storage and Analysis (SC'15).
!
!   (7) Wu, N., and Xiang, H., 2020:
!           Randomized QLP decomposition.
!           Linear algebra and its applications, Volume 599, 18-35
!
!   (8) Huckaby, D.A., and Chan, T.F., 2003:
!           On the convergence of Stewart's QLP algorithm for approximating the SVD.
!           Numer. Algorithms, Volume 32, 287-316.
!
!   (9) Huckaby, D.A., and Chan, T.F., 2005:
!           Stewart's pivoted QLP decomposition for low-rank matrices
!           Numerical Linear Algebra with Applications, Volume 12, 153-159.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : allocate_error
    use Utilities,        only : assert_eq, assert, merror
    use QR_Procedures,    only : lq_cmp, ortho_gen_lq, ortho_gen_qr, partial_qr_cmp
    use Random,           only : partial_rqr_cmp, partial_rtqr_cmp
!    use Random,           only : partial_rqr_cmp2, partial_rtqr_cmp
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:)           :: mat
    real(stnd),   intent(out),   dimension(:)             :: beta, tau
    real(stnd),   intent(out),   dimension(:,:), optional :: lmat, qmat, pmat
!
    integer(i4b), intent(in), optional :: rng_alg, blk_size, nover
!
    logical(lgl), intent(in),  optional :: random_qr, truncated_qr
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(:),   allocatable :: diag
    real(stnd), dimension(:,:), allocatable :: rmat
!
    integer      :: iok
    integer(i4b) :: n, m, ndiag, i, j, krank
    integer(i4b), dimension(:), allocatable :: ip
!
    logical(lgl) :: use_random_qr, use_truncated_qr, comp_lmat, comp_qmat, comp_pmat
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='qlp_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    comp_lmat = present( lmat )
!
    if ( comp_lmat ) then
!
        krank = assert_eq( int(size(beta),i4b)  ,   &
                           int(size(tau),i4b)   ,   &
                           int(size(lmat,1),i4b) ,  &
                           int(size(lmat,2),i4b) ,  &
                           name_proc )
!
    else
!
        krank = assert_eq( int(size(beta),i4b) ,   &
                           int(size(tau),i4b)  ,   &
                           name_proc )
!
    end if
!
    comp_qmat = present( qmat )
!
    if ( comp_qmat ) then
!
        m = assert_eq( int(size(mat,1),i4b) ,   &
                       int(size(qmat,1),i4b) ,  &
                       name_proc )
!
        i = size( qmat, 2 )
!
        call assert( logical( i==krank, lgl ), name_proc )
!
    else
!
        m = size( mat, 1 )
!
    end if
!
    comp_pmat = present( pmat )
!
    if ( comp_pmat ) then
!
        n = assert_eq( int(size(mat,2),i4b) ,   &
                       int(size(pmat,2),i4b) ,  &
                       name_proc )
!
        i = size( pmat, 1 )
!
        call assert( logical( i==krank, lgl ), name_proc )
!
    else
!
        n = size( mat, 2 )
!
    end if
!
    ndiag = min( m, n )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( ndiag<=0_i4b  .or. krank<=0_i4b ) then
!
        return
!
    end if
!
    call assert( logical( krank<=ndiag, lgl ), name_proc )
!
!   DETERMINE IF A RANDOMIZED QR ALGORITHM MUST BE USED.
!
    if ( present(random_qr) ) then
        use_random_qr = random_qr
    else
        use_random_qr = false
    end if
!
    if ( present(truncated_qr) ) then
        use_truncated_qr = truncated_qr
    else
        use_truncated_qr = false
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( ip(n), diag(krank), rmat(krank,n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   FIRST STEP, DO PARTIAL QR WITH COLUMN PIVOTING OF mat AS
!           
!                   mat * N = Q * R
!           
!   WHERE N IS A n-BY-n PERMUTATION MATRIX, Q AN m-BY-m ORTHOGONAL
!   MATRIX AND R AN UPPER TRIANGULAR OR TRAPEZOIDAL MATRIX.
!           
    if ( use_random_qr ) then
!           
        if ( use_truncated_qr ) then
!           
!           USE A RANDOMIZED AND TRUNCATED QR ALGORITHM.
!           
            call partial_rtqr_cmp( mat(:m,:n), diag(:krank), beta(:krank), ip(:n), i,   &
                                   rng_alg=rng_alg, nover=nover )
!           
        else
!           
!           USE A RANDOMIZED QR ALGORITHM.
!           
            call partial_rqr_cmp( mat(:m,:n), diag(:krank), beta(:krank), ip(:n), i, &
                                  rng_alg=rng_alg, blk_size=blk_size, nover=nover    )
!           
!            call partial_rqr_cmp2( mat(:m,:n), diag(:krank), beta(:krank), ip(:n), i, &
!                                   rng_alg=rng_alg, blk_size=blk_size, nover=nover    )
!
        end if
!           
    else
!           
!       USE A STANDARD DETERMINISTIC QR ALGORITHM.
!           
        call partial_qr_cmp( mat(:m,:n), diag(:krank), beta(:krank), ip(:n), i )
!
    end if
!
!   RESTORE AND APPLY TRANSPOSE PERMUTATION TO THE UPPER TRIANGULAR FACTOR
!   OF THE QR FACTORIZATION OF mat , STORING RESULTS IN rmat ,
!           
!                   rmat = R * N'
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  krank*n>=omp_limit                  .and.       &
                  n/i>=omp_limit2                     .and.       &
                  i>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC), PRIVATE(j,i)
!
    do j = 1_i4b, n
!
        i = ip(j)
!
        if ( j<=krank ) then
!
            rmat(:j-1_i4b,i)      = mat(:j-1_i4b,j)
            rmat(j,i)             = diag(j)
            rmat(j+1_i4b:krank,i) = zero
!
        else
!
            rmat(:krank,i) = mat(:krank,j)
!
        end if
!
    end do
!
!$OMP END PARALLEL DO
!
!   SECOND STEP, COMPUTE A DETERMINISTIC LQ DECOMPOSITION OF rmat = R * N' = L * P .
!
    call lq_cmp( rmat, diag, tau )
!
!   SAVE HOUSEHOLDER REFLECTORS DEFINING P IN mat .
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC), PRIVATE(j)
!
    do j = 1_i4b, n
!
        if ( j<=krank ) then
!
            mat(:j-1_i4b,j) = rmat(:j-1_i4b,j)
!
        else
!
            mat(:krank,j) = rmat(:krank,j)
!
        end if
!
    end do
!
!$OMP END PARALLEL DO
!
!   SAVE LOWER TRIANGULAR MATRIX L IF REQUIRED.
!
    if ( comp_lmat ) then
!
        do j = 1_i4b, krank
!
            lmat(:j-1_i4b,j)      = zero
            lmat(j,j)             = diag(j)
            lmat(j+1_i4b:krank,j) = rmat(j+1_i4b:krank,j)
!
        end do
!
    end if 
!
!   COMPUTE FIRST kank COLUMNS OF ORTHOGONAL MATRIX Q IF REQUIRED.
!
    if ( comp_qmat ) then
!
        qmat(:m,:krank) = mat(:m,:krank)
!
        call ortho_gen_qr( qmat, beta )
!
    end if 
!
!   COMPUTE FIRST kank ROWS OF ORTHOGONAL MATRIX P IF REQUIRED.
!
    if ( comp_pmat ) then
!
        pmat(:krank,:n) = mat(:krank,:n)
!
        call ortho_gen_lq( pmat, tau )
!
    end if 
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( ip, diag, rmat )
!
!
! END OF SUBROUTINE qlp_cmp
! _________________________
!
    end subroutine qlp_cmp
!
! =========================================================================================
!
    subroutine qlp_cmp2( mat, lmat, qmat, pmat, niter_qrql, random_qr, truncated_qr,  &
                         rng_alg, blk_size, nover )
!
! Purpose
! _______
!
!   QLP_CMP2 computes a partial or complete QLP factorization of an m-by-n matrix MAT:
!
!                     MAT = Q * L * P 
!
!   , where Q is a m-by-krank orthogonal matrix, P is a krank-by-n orthogonal matrix and L is a
!   krank-by-krank lower triangular matrix. If krank = min(m,n), the QLP factorization is complete
!   and  MAT = Q * L * P .
!
!   The QLP factorization is obtained by a three-step algorithm:
!
!   - first, a partial (or complete) QR factorization with column pivoting of MAT is computed;
!   - in a second step, a LQ Decomposition of the (permuted) upper triangular or trapezoidal (e.g., if n>m)
!     factor, R, in this QR decomposition of MAT is computed.
!   - and, in a final step, NITER_QRQL QR-QL iterations can be performed on the L factor in this LQ decomposition
!     to improve the accuracy of the diagonal elements of L (the so called L-values) as estimates of the singular
!     values of MAT (see references (2), (7), (8) and (9) for details).
!
!   By default, a standard deterministic QR factorization with column pivoting is used in the first phase
!   of the QLP algorithm. However, if the optional logical argument RANDOM_QR is used with the value true,
!   an alternate fast randomized (partial) QR factorization is used in the first phase of the QLP algorithm.
!   Furthermore if, in addition, the optional logical argument TRUNCATED_QR is used with the value true,
!   an even faster (but less accurate) randomized partial and truncated QR factorization will be used in the first
!   phase of the QLP algorithm. In all cases, deterministic blocked LQ and QR factorizations are used in the second
!   and third steps of the QLP factorization.
!
!   At the user option, the QLP factorization can also be only partial, e.g., the subroutine stops the computations
!   when the numbers of columns of Q and of rows of P are equal to a predefined value equals to
!   krank = size( LMAT, 1 ) = size( LMAT, 2 ).
!   
!   The QLP decomposition provides a reasonable and cheap estimate of the Singular Value Decomposition (SVD) of a matrix
!   when this matrix has a low rank or a significant gap in its singular values spectrum.
!   
!   See Further Details and the cited references for more information.
!   
!
! Arguments
! _________
!
!   MAT           (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                 On entry, the real m-by-n matrix to be decomposed.
!
!                 On exit, MAT is destroyed as MAT is used as workspace in the routine.
!
!                 See Further Details.
!
!   LMAT          (OUTPUT) real(stnd), dimension(:,:)
!                 On exit, LMAT stores the lower triangular matrix L in the (partial)
!                 QLP factorization of MAT.
!
!                 See Further Details.
!
!                 The shape of LMAT must verify:
!
!                 - size( LMAT, 1 ) = size( LMAT, 2 ) = krank.
!
!   QMAT          (OUTPUT) real(stnd), dimension(:,:)
!                 On exit, QMAT stores the first krank columns of the orthogonal matrix
!                 Q in the (partial) QLP factorization of MAT.
!
!                 See Further Details.
!
!                 The shape of QMAT must verify:
!
!                 - size( QMAT, 1 ) = m.
!                 - size( QMAT, 2 ) = krank.
!
!   PMAT          (OUTPUT) real(stnd), dimension(:,:)
!                 On exit, PMAT stores the first krank rows of the orthogonal matrix
!                 P in the (partial) QLP factorization of MAT.
!
!                 See Further Details.
!
!                 The shape of PMAT must verify:
!
!                 - size( QMAT, 1 ) = krank.
!                 - size( QMAT, 2 ) = n.
!
!   NITER_QRQL    (INPUT, OPTIONAL) integer(i4b)
!                 The number of QR-QL iterations performed on L after the initial QLP factorization
!                 for improving the accuracy of the L-values. NITER_QRQL must be positive or null.
!
!                 By default, no QR-QL iterations are performed after the initial QLP factorization.
!
!   RANDOM_QR     (INPUT, OPTIONAL) logical(lgl)
!                 On entry, if RANDOM_QR is used with the value true, a fast randomized (partial) QR
!                 factorization with column pivoting is used in the first phase of the QLP algorithm.
!
!                 By default, RANDOM_QR = false, i.e., A standard deterministic (partial) QR factorization
!                 with column pivoting is used in the first phase of the QLP algorithm.
!
!   TRUNCATED_QR  (INPUT, OPTIONAL) logical(lgl)
!                 On entry, if TRUNCATED_QR is used with the value true in addition to RANDOM_QR also
!                 set to true, a very fast (but less accurate) randomized partial and truncated
!                 QR factorization is used in the first phase of the QLP algorithm.
!
!                 By default, TRUNCATED_QR = false, i.e., a "standard" randomized (partial) QR factorization
!                 with column pivoting is used in the first phase of the QLP algorithm if RANDOM_QR = true.
!
!   RNG_ALG       (INPUT, OPTIONAL) integer(i4b)
!                 On entry, a scalar integer to select the random (uniform) number generator used to build
!                 the random gaussian matrix in the randomized (partial) QR phase of the QLP algorithm,
!                 if RANDOM_QR = true.
!
!                 The possible values are:
!
!                 - ALG=1  : selects the Marsaglia's KISS random number generator;
!                 - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!                 - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!                 - ALG=4  : selects the Mersenne Twister random number generator;
!                 - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!                 - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!                 - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!                 - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!                 - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!                 - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                   random number generator;
!
!                 For other values, the current random number generator and its current state are not changed.
!                 Note further, that, on exit, the current random number generator is not reset to its previous
!                 value before the call to QLP_CMP2.
!
!                 See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   BLK_SIZE      (INPUT, OPTIONAL) integer(i4b)
!                 On entry, the block size used in the randomized (partial) QR phase of the QLP algorithm,
!                 if RANDOM_QR = true (and TRUNCATED_QR = false).
!
!                 BLK_SIZE must be greater or equal to one and less than min(m,n) and must be set to a much
!                 smaller value than min(m,n) usually, depending also on the architecture of the computer.
!
!                 See Further Details and the cited references for the meaning of the block size
!                 in the randomized (partial) QR algorithm.
!
!                 By default, BLK_SIZE is set to min( BLKSZ_QR, min(m,n) ), where parameter BLKSZ_QR is the default
!                 block size for QR related algorithms specified in module Select_Parameters.
!
!   NOVER         (INPUT, OPTIONAL) integer(i4b)
!                 The oversampling size used in the randomized (partial) QR phase of the QLP algorithm,
!                 if RANDOM_QR = true.
!
!                 NOVER must be positive or null and verify the relationships:
!
!                 - NOVER + BLK_SIZE        <= size( MAT, 1 ) if TRUNCATED_QR = false;
!                 - NOVER + size( LMAT, 1 ) <= size( MAT, 1 ) if TRUNCATED_QR = true.
!
!                 and is adjusted if necessary to verify these relationships in all cases.
!
!                 See Further Details and the cited references for the meaning and usefulness
!                 of the oversampling size in the randomized partial QR algorithm.
!
!                 By default, the oversampling size is set to:
!
!                 - 10                            if TRUNCATED_QR = false;
!                 - max( size(LMAT,1)/2_i4b, 10 ) if TRUNCATED_QR = true.
!
!
! Further Details
! _______________
!
!   QLP_CMP2 first computes a (partial or complete) QR factorization with column pivoting
!   of the m-by-n matrix MAT:
!
!                     MAT * N = Q * R
!
!   , where N is a n-by-n permutation matrix, R is a upper triangular or trapezoidal
!   (i.e., if n>m) matrix and Q is a m-by-m orthogonal matrix.
!
!   If the optional logical argument RANDOM_QR is used with the value true, a fast randomized
!   partial (and truncated, if the optional logical argument TRUNCATED_QR is also used with the value
!   true) QR factorization with column pivoting is used in this first phase of the QLP algorithm.
!
!   At the user option, this QR factorization can also be only partial, e.g., the subroutine
!   ends when the numbers of columns of Q is equal to a predefined value equals to
!   krank = size( LMAT, 1 ) = size( LMAT, 2 ).
!
!   This leads implicitly to the following partition of Q:
!
!   [  Q1  Q2  ]
!
!   where Q1 is a m-by-krank orthonormal matrix and Q2 is a m-by-(m-krank) orthonormal matrix
!   orthogonal to Q1, and to the following corresponding partition of R:
!
!   [  R11  R12  ]
!
!   [  R21  R22  ]
!
!   where R11 is a krank-by-krank triangular matrix, R21 is zero by construction,
!   R12 is a full krank-by-(n-krank) matrix and R22 is a full (m-krank)-by-(n-krank)
!   matrix.
!
!   In a second step, QLP_CMP2 computes a deterministic LQ factorization of the matrix
!   product:
!
!   R * N' = L * P
!
!   if the first QR factorization is complete, or of the matrix product:
!
!   [  R11  R12  ] * N' = L * P
!
!   if this first QR factorization is only partial. This leads to the (partial) QLP factorization
!   of MAT:
!
!                     MAT ≈ Q1 * L * P
!
!   where L is a krank-by-krank triangular matrix and P is a krank-by-n matrix with orthonormal rows.
!
!   In a final step, NITER_QRQL QR-QL iterations can be performed on L to improve the accuracy of the diagonal
!   elements of L (the so called L-values) as estimates of the singular values of MAT (see references (2), (7),
!   (8) and (9) for details) and the orthogonal matrices Q and P are updated accordingly.
!
!   The properties of the QLP factorization and when it can be used as a good proxy for the (partial
!   or complete) SVD of a matrix are discussed in the references (2), (7), (8) and (9).
!
!   The computations are parallelized if OPENMP is used. However, note that QLP_CMP2
!   uses a standard "BLAS2" algorithm without any blocking for performing the first QR factorization
!   with column pivoting if the optional logical argument RANDOM_QR is not used or used with the value
!   false. On the other hand, QLP_CMP2 uses an efficient randomized and blocked QR algorithm
!   with column pivoting (see the references (2), (3) and (4)) if RANDOM_QR is used with the value true.
!   This randomized algorithm is thus particularly efficient for large matrices.
!
!   The standard deterministic BLAS2 algorithm for computing a QR factorization with column pivoting
!   is described in the reference (1). The randomized partial QR algorithm with column pivoting
!   used if the optional logical argument RANDOM_QR is present with the value true is described
!   in the references (3), (4) and (5). Finally, the randomized partial and truncated QR algorithm
!   with column pivoting used if both the optional logical arguments RANDOM_QR and TRUNCATED_QR are
!   present with the value true is described in the reference (6). This algorithm is the fastest, but
!   less accurate than the randomized partial QR algorithm with column pivoting described in the
!   references (3), (4) and (5).
!
!   In all cases, QLP_CMP2 uses efficient blocked algorithms for performing the LQ factorization
!   in the second step of the QLP algorithm and also in the final QR-QL iterations performed on L.
!   The LQ factorization is described in the reference (1).
!
!   On exit, QLP_CMP2 stores:
!
!   - the krank-by-krank lower triangular matrix L in the array argument LMAT;
!   - the first krank columns of Q in the array argument QMAT;
!   - and the first krank rows of P in the array argument PMAT.
!
!   For further details on the QLP factorization and its use, randomized QR and QLP algorithms or
!   QR-QL iterations, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Stewart, G.W., 1999:
!           The QLP approximation to the singular value decomposition.
!           SIAM J. Sci. Comput., Volume 20, 1336-1348.
!
!   (3) Duersch, J.A., and Gu, M., 2017:
!           Randomized QR with column pivoting.
!           SIAM J. Sci. Comput., Volume 39, C263-C291.
!
!   (4) Martinsson, P.G., Quintana-Orti, G., Heavner, N., and Van de Geijn, R., 2017:
!           Householder QR factorization with randomization for column pivoting (HQRRP).
!           SIAM J. Sci. Comput., Volume 39, C96-C115.
!
!   (5) Duersch, J.A., and Gu, M., 2020:
!           Randomized projection for rank-revealing matrix factorizations
!           and low-rank approximations.
!           SIAM Review, Volume 62, Issue 3, 661-682.
!
!   (6) Mary, T., Yamazaki, I., Kurzak, J., Luszczek, P., Tomov, S., and Dongarra, J., 2015:
!           Performance of Random Sampling for Computing Low-rank Approximations of a
!           Dense Matrix on GPUs.
!           International Conference for High Performance Computing, Networking, Storage and Analysis (SC'15).
!
!   (7) Wu, N., and Xiang, H., 2020:
!           Randomized QLP decomposition.
!           Linear algebra and its applications, Volume 599, 18-35
!
!   (8) Huckaby, D.A., and Chan, T.F., 2003:
!           On the convergence of Stewart's QLP algorithm for approximating the SVD.
!           Numer. Algorithms, Volume 32, 287-316.
!
!   (9) Huckaby, D.A., and Chan, T.F., 2005:
!           Stewart's pivoted QLP decomposition for low-rank matrices
!           Numerical Linear Algebra with Applications, Volume 12, 153-159.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,        only : assert_eq, assert
    use QR_Procedures,    only : lq_cmp, ortho_gen_lq, apply_q_lq, &
                                 qr_cmp, ortho_gen_qr, apply_q_qr, &
                                 partial_qr_cmp
    use Random,           only : partial_rqr_cmp, partial_rtqr_cmp
!    use Random,           only : partial_rqr_cmp2, partial_rtqr_cmp
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:,:) :: lmat, qmat, pmat
!
    integer(i4b), intent(in), optional :: niter_qrql, rng_alg, blk_size, nover
!
    logical(lgl), intent(in),  optional :: random_qr, truncated_qr
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(lmat,1))  :: diag, beta, tau
!
    integer(i4b)                         :: n, m, ndiag, i, j, krank, niter_qrql2
    integer(i4b), dimension(size(mat,2)) :: ip
!
    logical(lgl) :: use_random_qr, use_truncated_qr
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='qlp_cmp2'
!
!   niter_qrql0 IS THE DEFAULT NUMBER OF QR-QL ITERATIONS PERFORMED AFTER THE QLP FACTORIZATION.
!
    integer(i4b),  parameter :: niter_qrql0=0_i4b
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    krank = assert_eq( int(size(lmat,1),i4b) ,  &
                       int(size(lmat,2),i4b) ,  &
                       int(size(qmat,2),i4b) ,  &
                       int(size(pmat,1),i4b) ,  &
                       name_proc )
!
    m = assert_eq( int(size(mat,1),i4b) ,   &
                   int(size(qmat,1),i4b) ,  &
                   name_proc )
!
    n = assert_eq( int(size(mat,2),i4b) ,   &
                   int(size(pmat,2),i4b) ,  &
                   name_proc )
!
    ndiag = min( m, n )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( ndiag<=0_i4b  .or. krank<=0_i4b ) then
!
        return
!
    end if
!
    call assert( logical( krank<=ndiag, lgl ), name_proc )
!
!   SET OPTIONAL ARGUMENTS.
!
!   DETERMINE THE NUMBER OF QR-QL ITERATIONS niter_qrql2 TO BE PERFORMED.
!
    if ( present(niter_qrql) ) then
        niter_qrql2 = max( niter_qrql, 0_i4b )
    else
        niter_qrql2 = niter_qrql0
    end if
!
!   DETERMINE IF A RANDOMIZED QR ALGORITHM MUST BE USED.
!
    if ( present(random_qr) ) then
        use_random_qr = random_qr
    else
        use_random_qr = false
    end if
!
    if ( present(truncated_qr) ) then
        use_truncated_qr = truncated_qr
    else
        use_truncated_qr = false
    end if
!
!   FIRST STEP, DO PARTIAL QR WITH COLUMN PIVOTING OF mat AS
!           
!                   mat * N = Q * R
!           
!   WHERE N IS A n-BY-n PERMUTATION MATRIX, Q AN m-BY-m ORTHOGONAL
!   MATRIX AND R AN UPPER TRIANGULAR OR TRAPEZOIDAL MATRIX.
!           
    if ( use_random_qr ) then
!           
        if ( use_truncated_qr ) then
!           
!           USE A RANDOMIZED AND TRUNCATED QR ALGORITHM.
!           
            call partial_rtqr_cmp( mat(:m,:n), diag(:krank), beta(:krank), ip(:n), i,   &
                                   rng_alg=rng_alg, nover=nover )
!           
        else
!           
!           USE A RANDOMIZED QR ALGORITHM.
!           
            call partial_rqr_cmp( mat(:m,:n), diag(:krank), beta(:krank), ip(:n), i, &
                                  rng_alg=rng_alg, blk_size=blk_size, nover=nover    )
!           
!            call partial_rqr_cmp2( mat(:m,:n), diag(:krank), beta(:krank), ip(:n), i, &
!                                   rng_alg=rng_alg, blk_size=blk_size, nover=nover    )
!
        end if
!           
    else
!           
!       USE A STANDARD DETERMINISTIC QR ALGORITHM.
!           
        call partial_qr_cmp( mat(:m,:n), diag(:krank), beta(:krank), ip(:n), i )
!
    end if
!
!   RESTORE AND APPLY TRANSPOSE PERMUTATION TO THE UPPER TRIANGULAR FACTOR
!   OF THE QR FACTORIZATION OF mat , STORING RESULTS IN pmat ,
!           
!                   pmat = R * N'
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  krank*n>=omp_limit                  .and.       &
                  n/i>=omp_limit2                     .and.       &
                  i>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC), PRIVATE(j,i)
!
    do j = 1_i4b, n
!
        i = ip(j)
!
        if ( j<=krank ) then
!
            pmat(:j-1_i4b,i)      = mat(:j-1_i4b,j)
            pmat(j,i)             = diag(j)
            pmat(j+1_i4b:krank,i) = zero
!
        else
!
            pmat(:krank,i) = mat(:krank,j)
!
        end if
!
    end do
!
!$OMP END PARALLEL DO
!
!   SECOND STEP, COMPUTE A DETERMINISTIC LQ DECOMPOSITION OF pmat = R * N' = L * P .
!
    call lq_cmp( pmat, diag, tau )
!
!   COPY LOWER TRIANGULAR MATRX L IN lmat .
!
    do j = 1_i4b, krank
!
        lmat(:j-1_i4b,j)      = zero
        lmat(j,j)             = diag(j)
        lmat(j+1_i4b:krank,j) = pmat(j+1_i4b:krank,j)
!
    end do
!
!   COMPUTE FIRST kank COLUMNS OF ORTHOGONAL MATRIX Q .
!
    qmat(:m,:krank) = mat(:m,:krank)
!
    call ortho_gen_qr( qmat, beta )
!
!   COMPUTE FIRST kank ROWS OF ORTHOGONAL MATRIX P .
!
    call ortho_gen_lq( pmat, tau )
!
!   DO niter_qrql2 QR-QL ITERATIONS TO IMPROVE THE L-VALUES IF REQUIRED.
!
    do j = 1_i4b, niter_qrql2
!
!       COMPUTE QR DECOMPOSITION OF LOWER TRIANGULAR MATRX L.
!
        call qr_cmp( lmat, diag, beta )
!
!       UPDATE THE ORTHONORMAL MATRIX Q.
!
        call apply_q_qr( lmat, beta, qmat, left=false, trans=false )
!
!       RESTORE THE UPPER TRIANGULAR MATRIX R.
!
        do i = 1_i4b, krank
!
            lmat(i,i) = diag(i)
!
            lmat(i+1_i4b:krank,i) = zero
!
        end do
!
!       COMPUTE LQ DECOMPOSITION OF UPPER TRIANGULAR MATRX R.
!
        call lq_cmp( lmat, diag, beta )
!
!       UPDATE THE ORTHONORMAL MATRIX P.
!
        call apply_q_lq( lmat, beta, pmat, left=true, trans=false )
!
!       RESTORE THE LOWER TRIANGULAR MATRIX L.
!
        do i = 1_i4b, krank
!
            lmat(i,i) = diag(i)
!
            lmat(1_i4b:i-1_i4b,i) = zero
!
        end do
!
    end do
!
!
! END OF SUBROUTINE qlp_cmp2
! __________________________
!
    end subroutine qlp_cmp2
!
! =========================================================================================
!
    subroutine rqlp_cmp( mat, lmat, qmat, pmat, niter, rng_alg, ortho, niter_qrql )
!
! Purpose
! _______
!
!   RQLP_CMP computes a randomized partial QLP factorization of an m-by-n matrix MAT:
!
!                     MAT ≈ Q * L * P 
!
!   , where Q is a m-by-krank matrix with orthonormal columns, P is a krank-by-n matrix with orthonormal
!   rows and L is a krank-by-krank lower triangular matrix.
!
!   The randomized QLP factorization is only partial, e.g., the subroutine stops the computations
!   when the numbers of columns of Q and of rows of P are equal to a predefined value equals to
!   krank = size( LMAT, 1 ) = size( LMAT, 2 ).
!
!   The randomized partial QLP factorization is obtained by a four-step algorithm:
!
!   - first, the routines computes a partial QB factorization of MAT with the help of a randomized algorithm:
!
!                     MAT ≈ Q * B
!
!     , where Q is a m-by-krank orthonormal matrix, B is a krank-by-n matrix and the product Q*B is a good
!     approximation of MAT according to the spectral or Frobenius norm;
!   - second, a QR factorization with column pivoting of B is computed and Q is post-multiplied
!     by the krank-by-krank orthogonal matrix, O, in this QR factorization of B;
!   - in a third step, a LQ Decomposition of the (permuted) upper trapezoidal factor, R, in this
!     QR decomposition of B is computed.
!   - and, in a final step, NITER_QRQL QR-QL iterations are performed on the L matrix in this LQ decomposition to improve
!     the accuracy of the diagonal elements of L (the so called L-values) as estimates of the singular values of MAT
!     (see references (5), (6) and (7) for details).
!   
!   This randomized QLP decomposition provides a reasonable and cheap estimate of the Singular Value Decomposition (SVD) of a matrix
!   when this matrix has a low rank or a significant gap in its singular values spectrum.
!   
!   See Further Details and the cited references for more information.
!   
!
! Arguments
! _________
!
!   MAT        (INPUT) real(stnd), dimension(:,:)
!              On entry, the real m-by-n matrix to be decomposed.
!
!              MAT is not modified by the routine.
!
!   LMAT       (OUTPUT) real(stnd), dimension(:,:)
!              On exit, LMAT stores the lower triangular matrix L in the (partial)
!              QLP factorization of MAT.
!
!              See Further Details.
!
!              The shape of LMAT must verify:
!
!              - size( LMAT, 1 ) = size( LMAT, 2 ) = krank.
!
!   QMAT       (OUTPUT) real(stnd), dimension(:,:)
!              On exit, QMAT stores the first krank columns of the orthogonal matrix
!              Q in the (partial) QLP factorization of MAT.
!
!              See Further Details.
!
!              The shape of QMAT must verify:
!
!              - size( QMAT, 1 ) = m.
!              - size( QMAT, 2 ) = krank.
!
!   PMAT       (OUTPUT) real(stnd), dimension(:,:)
!              On exit, PMAT stores the first krank rows of the orthogonal matrix
!              P in the (partial) QLP factorization of MAT.
!
!              See Further Details.
!
!              The shape of PMAT must verify:
!
!              - size( QMAT, 1 ) = krank.
!              - size( QMAT, 2 ) = n.
!
!   NITER      (INPUT, OPTIONAL) integer(i4b)
!              The number of randomized power or subspace iterations performed
!              in the first phase of the randomized QLP algorithm for computing the preliminary
!              randomized QB factorization.
!
!              NITER must be positive or null.
!
!              By default, 5 randomized power or subspace iterations are performed.
!
!   RNG_ALG    (INPUT, OPTIONAL) integer(i4b)
!              On entry, a scalar integer to select the random (uniform) number generator used to build
!              the random gaussian test matrix in the initial randomized partial QB factorization.
!
!              The possible values are:
!
!              - ALG=1  : selects the Marsaglia's KISS random number generator;
!              - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!              - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!              - ALG=4  : selects the Mersenne Twister random number generator;
!              - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!              - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!              - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!              - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!              - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!              - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                random number generator;
!
!              For other values, the current random number generator and its current state are not changed.
!              Note further, that, on exit, the current random number generator is not reset to its previous
!              value before the call to RQLP_CMP.
!
!              See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   ORTHO      (INPUT, OPTIONAL) logical(lgl)
!              On entry, if:
!
!              - ORTHO=true, orthonormalization is carried out between each step of
!                the power iterations, to avoid loss of accuracy due to rounding errors.
!                This means that subspace iterations are used instead of power iterations
!                in the QB phase of the algorithm,
!              - ORTHO=false, orthonormalization is not performed. 
!
!              The default is to use orthonormalization, e.g., ORTHO=true.
!
!   NITER_QRQL (INPUT, OPTIONAL) integer(i4b)
!              The number of QR-QL iterations performed on L after the initial QLP factorization
!              for improving the accuracy of the L-values. NITER_QRQL must be positive or null.
!
!              By default, no QR-QL iterations are performed after the initial QLP factorization.
!
!
! Further Details
! _______________
!
!   RQLP_CMP first computes a partial QB factorization of MAT with the help of a randomized algorithm:
!
!                     MAT ≈ Q * B
!
!   , where Q is a m-by-krank orthonormal matrix, B is a krank-by-n matrix and the product Q*B is a good
!   approximation of MAT according to the spectral or Frobenius norm. Here, krank = size( LMAT, 1 ) = size( LMAT, 2 ).
!
!   In a second step, RQLP_CMP computes a deterministic QR factorization with column pivoting of the krank-by-n matrix B, 
!   to obtain an approximate QR factorization with column pivoting of MAT :
!
!             MAT * N = Q * ( B * N ) = Q * ( O * R ) = (Q * O ) * R
!
!   , where N is a n-by-n permutation matrix, O is a krank-by-krank orthogonal matrix and R is a
!   krank-by-n upper trapezoidal matrix.
!
!   In a third step, RQLP_CMP computes a deterministic LQ factorization of the matrix
!   product:
!
!   R * N' = L * P
!
!   This leads to the approximate QLP factorization of MAT:
!
!                     MAT ≈ ( Q * O ) * L * P
!
!   where Q is a m-by-krank matrix with orthonormal columns, O is a krank-by-krank orthogonal matrix,
!   L is a krank-by-krank lower triangular matrix and P is a krank-by-n matrix with orthonormal rows.
!
!   In a final step, NITER_QRQL QR-QL iterations can be performed on L to improve the accuracy of the diagonal
!   elements of L (the so called L-values) as estimates of the singular values of MAT (see references (1),
!   (5), (6) and (7) for details) and the orthogonal matrices Q and P are updated accordingly.
!
!   The computations are parallelized if OPENMP is used.
!
!   In all cases, RQLP_CMP uses efficient blocked algorithms for performing the QB, QR and LQ steps
!   in the randomized QLP algorithm and also in the final QR-QL iterations performed on L.
!
!   On exit, RQLP_CMP stores:
!
!   - the krank-by-krank lower triangular matrix L in the array argument LMAT;
!   - the first krank columns of Q in the array argument QMAT;
!   - and the first krank rows of P in the array argument PMAT.
!
!   For further details on the QLP factorization and its use or randomized QB, QR and QLP algorithms or
!   QR-QL iterations, see:
!
!   (1) Stewart, G.W., 1999:
!           The QLP approximation to the singular value decomposition.
!           SIAM J. Sci. Comput., Volume 20, 1336-1348.
!
!   (2) Duersch, J.A., and Gu, M., 2017:
!           Randomized QR with column pivoting.
!           SIAM J. Sci. Comput., Volume 39, C263-C291.
!
!   (3) Martinsson, P.G., Quintana-Orti, G., Heavner, N., and Van de Geijn, R., 2017:
!           Householder QR factorization with randomization for column pivoting (HQRRP).
!           SIAM J. Sci. Comput., Volume 39, C96-C115.
!
!   (4) Martinsson, P.G., and Voronin, S., 2016:
!           A randomized blocked algorithm for efficiently computing
!           rank-revealing factorizations of matrices.
!           SIAM J. Sci. Comput., 38:5, S485-S507.
!
!   (5) Wu, N., and Xiang, H., 2020:
!           Randomized QLP decomposition.
!           Linear algebra and its applications, Volume 599, 18-35
!
!   (6) Huckaby, D.A., and Chan, T.F., 2003::
!           On the convergence of Stewart's QLP algorithm for approximating the SVD.
!           Numer. Algorithms, Volume 32, 287-316.
!
!   (7) Huckaby, D.A., and Chan, T.F., 2005:
!           Stewart's pivoted QLP decomposition for low-rank matrices
!           Numerical Linear Algebra with Applications, Volume 12, 153-159.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,        only : assert_eq, assert, arth
    use QR_Procedures,    only : lq_cmp, ortho_gen_lq, apply_q_lq, &
                                 qr_cmp, apply_q_qr
    use Random,           only : rqb_cmp
#ifdef _BLAS
    use BLAS_interfaces,  only : swap
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),  dimension(:,:) :: mat
    real(stnd),   intent(out), dimension(:,:) :: lmat, qmat, pmat
!
    integer(i4b), intent(in), optional :: niter, niter_qrql, rng_alg
!
    logical(lgl), intent(in),  optional :: ortho
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(lmat,1)) :: diag, beta
#ifndef _BLAS
    real(stnd), dimension(size(pmat,1)) :: tmp
#endif
!
    integer(i4b)                         :: n, m, ndiag, i, j, in, krank, niter_qrql2
    integer(i4b), dimension(size(mat,2)) :: ip, ipt
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='rqlp_cmp'
!
!   niter_qrql0 IS THE DEFAULT NUMBER OF QR-QL ITERATIONS PERFORMED AFTER THE QLP FACTORIZATION.
!
    integer(i4b),  parameter :: niter_qrql0=0_i4b
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    krank = assert_eq( int(size(lmat,1),i4b) ,  &
                       int(size(lmat,2),i4b) ,  &
                       int(size(qmat,2),i4b) ,  &
                       int(size(pmat,1),i4b) ,  &
                       name_proc )
!
    m = assert_eq( int(size(mat,1),i4b) ,   &
                   int(size(qmat,1),i4b) ,  &
                   name_proc )
!
    n = assert_eq( int(size(mat,2),i4b) ,   &
                   int(size(pmat,2),i4b) ,  &
                   name_proc )
!
    ndiag = min( m, n )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( ndiag<=0_i4b  .or. krank<=0_i4b ) then
!
        return
!
    end if
!
    call assert( logical( krank<=ndiag, lgl ), name_proc )
!
!   SET OPTIONAL ARGUMENTS.
!
!   DETERMINE THE NUMBER OF QR-QL ITERATIONS niter_qrql2 TO BE PERFORMED
!   IN THE FINAL STEP FOR IMPROVING THE L-VALUES.
!
    if ( present(niter_qrql) ) then
        niter_qrql2 = max( niter_qrql, 0_i4b )
    else
        niter_qrql2 = niter_qrql0
    end if
!
!   FIRST STEP, DO RANDOMIZED PARTIAL QR WITH COLUMN PIVOTING OF mat AS
!           
!                   mat * N = Q * R
!           
!   WHERE N IS A n-BY-n PERMUTATION MATRIX, Q AN m-BY-krank MATRIX WITH
!   ORTHONORMAL COLUMNS AND R AN krank-BY-n UPPER TRIANGULAR OR TRAPEZOIDAL MATRIX.
!           
    call rqb_cmp( mat(:m,:n), qmat(:m,:krank), pmat(:krank,:n), niter=niter, &
                  rng_alg=rng_alg, ortho=ortho, ip=ip )
!
!   RESTORE AND APPLY TRANSPOSE PERMUTATION TO THE UPPER TRAPEZOIDAL FACTOR
!   OF THE QR FACTORIZATION OF mat , IN PLACE, IN pmat ,
!           
!                   pmat = R * N'
!
!   FIRST COMPUTE N'.
!
    ipt(ip) = arth( 1_i4b, 1_i4b, n )
!
!   NEXT SWAP CORRESPONDING COLUMNS OF pmat(:krank,:n) IN PLACE.
!
    ipt(:n) = -ipt(:n)
!
    do i = 1_i4b, n
!
        if ( ipt(i)>0_i4b ) cycle
!
        j      = i
        ipt(j) = -ipt(j)
        in     = ipt(j)
!
        do
!
            if ( ipt(in)>0_i4b ) exit
!
            if ( j/=in ) then
#ifdef _BLAS
                call swap( krank, pmat(1_i4b:krank,j), 1_i4b, pmat(1_i4b:krank,in), 1_i4b )
#else
                tmp(1_i4b:krank)     = pmat(1_i4b:krank,j)
                pmat(1_i4b:krank,j)  = pmat(1_i4b:krank,in)
                pmat(1_i4b:krank,in) = tmp(1_i4b:krank)
#endif
!
            end if
!
            ipt(in) = -ipt(in)
            j       = in
            in      = ipt(in)
!
        end do
!
    end do
!
!   SECOND STEP, COMPUTE LQ DECOMPOSITION OF pmat = R * N' = L * P .
!
    call lq_cmp( pmat, diag, beta )
!
!   COPY LOWER TRIANGULAR MATRX L IN lmat .
!
    do j = 1_i4b, krank
!
        lmat(:j-1_i4b,j)      = zero
        lmat(j,j)             = diag(j)
        lmat(j+1_i4b:krank,j) = pmat(j+1_i4b:krank,j)
!
    end do
!
!   COMPUTE FIRST kank ROWS OF ORTHOGONAL MATRIX P .
!
    call ortho_gen_lq( pmat, beta )
!
!   DO niter_qrql2 QR-QL ITERATIONS TO IMPROVE THE L-VALUES IF REQUIRED.
!
    do j = 1_i4b, niter_qrql2
!
!       COMPUTE QR DECOMPOSITION OF LOWER TRIANGULAR MATRX L.
!
        call qr_cmp( lmat, diag, beta )
!
!       UPDATE THE ORTHONORMAL MATRIX Q.
!
        call apply_q_qr( lmat, beta, qmat, left=false, trans=false )
!
!       RESTORE THE UPPER TRIANGULAR MATRIX R.
!
        do i = 1_i4b, krank
!
            lmat(i,i) = diag(i)
!
            lmat(i+1_i4b:krank,i) = zero
!
        end do
!
!       COMPUTE LQ DECOMPOSITION OF UPPER TRIANGULAR MATRX R.
!
        call lq_cmp( lmat, diag, beta )
!
!       UPDATE THE ORTHONORMAL MATRIX P.
!
        call apply_q_lq( lmat, beta, pmat, left=true, trans=false )
!
!       RESTORE THE LOWER TRIANGULAR MATRIX L.
!
        do i = 1_i4b, krank
!
            lmat(1_i4b:i-1_i4b,i) = zero
!
            lmat(i,i) = diag(i)
!
        end do
!
    end do
!
!
! END OF SUBROUTINE rqlp_cmp
! __________________________
!
    end subroutine rqlp_cmp
!
!
! =========================================================================================
!                   SINGULAR VECTORS BY INVERSE ITERATION ALGORITHMS
! =========================================================================================
!
!
    function maxdiag_gkinv_qr( e, lambda )
!
! Purpose
! _______
!
!   This function computes the index of the element of maximum absolute value in the
!   diagonal entries of
!
!                     ( GK - LAMBDA * I )**(-1)
!
!   where GK is a n-by-n symmetric tridiagonal matrix with a zero diagonal,
!   I is the identity matrix and LAMBDA is a scalar.
!
!
! Arguments
! _________
!
!   E           (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 off-diagonal elements of the tridiagonal matrix.
!
!   LAMBDA      (INPUT) real(stnd)
!               On entry, the eigenvalue or shift used in the QR factorization.
!
!
! Further Details
! _______________
!
!   The diagonal entries of ( GK - LAMBDA * I )**(-1) are computed by means of the QR
!   factorization of ( GK - LAMBDA * I ). For the latter computation, the semiseparable
!   structure of ( GK - LAMBDA * I )**(-1) is used, see the reference (1). Moreover, it
!   is assumed that GK is unreduced, but no check is done in the subroutine to verify
!   this assumption.
!
!   This subroutine is adapted from the pseudo-code trace_Tinv given in the reference (1).
!
!   For further details, see:
!
!   (1) Bini, D.A., Gemignani, L., and Tisseur, F., 2005:
!           The Ehrlich-Aberth method for the nonsymmetric tridiagonal eigenvalue problem.
!           SIAM J. Matrix Anal. Appl., 27, 153-175.
!
!   (2) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (3) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Giv_Procedures, only : define_rot_givens
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in) :: lambda
    real(stnd), dimension(:),   intent(in) :: e
!
    integer(i4b) :: maxdiag_gkinv_qr
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: dmin, a, b, g, cs, sn, tau
    real(stnd), dimension(size(e)+1) :: r, s, t, v, w
!
    integer(i4b) :: n, i, k
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = int( size(e), i4b) + 1_i4b
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        maxdiag_gkinv_qr = n
!
        return
!
    end if
!
!   REPLACEMENT FOR VERY SMALL PIVOTS IN QR FACTORIZATION.
!
!    g = maxval( abs(e(1_i4b:n-1_i4b)) )
!    dmin = max( safmin, (safmin*g)*g ) 
!
!    dmin = sqrt( safmin )
!
    dmin = safmin
!
!   MAKE A LOCAL COPY OF THE ZERO-DIAGONAL TRIDIAGONAL MATRIX.
!
    s(:n-1_i4b) = e(:n-1_i4b)
    s(n)        = zero
!
!   REPLACE SMALL OFF-DIAGONALS BY safmin.
!
    where ( abs(s(1_i4b:n-1_i4b))<=safmin )                      &
          s(1_i4b:n-1_i4b) = sign( safmin, s(1_i4b:n-1_i4b) )
!
!   COMPUTE THE QR FACTORIZATION AND THE ENTRIES OF THE SCALED FORM OF R
!   OF R IN THE VECTORS r(:), s(:) AND t(:).
!
    a = -lambda
    g = s(1_i4b)
    b = g
!
    do i = 1_i4b, n-1_i4b
!
        call define_rot_givens( a, b, cs, sn )
!
        r(i) = cs*a + sn*b
        s(i) = -sn*( cs*g - sn*lambda )
        a    = -sn*g - cs*lambda
!
!       SAVE COSINUS OF GIVENS ROTATION FOR LATER USE.
!
        v(i) = cs
!
        b    = s(i+1_i4b)
        t(i) = sn*sn*b
        g    = cs*b
!
        if ( i==1_i4b ) cycle
!
        t(i-1_i4b) = sn*t(i-1_i4b)
!
    end do
!
    r(n) = a
    v(n) = one
!
!   REPLACE SMALL DIAGONALS BY dmin.
!
    where ( abs(r(:n))<=dmin ) r(:n) = sign( dmin, r(:n) )
!
!   SOLVE THE TRIANGULAR LINEAR SYSTEM R.w = v .
!
!    w(n)       = v(n)/r(n)
!    w(n-1_i4b) = ( v(n-1_i4b) - w(n)*s(n-1_i4b) )/r(n-1_i4b)
!
!    do i = n-2_i4b, 1_i4b, -1_i4b
!
!       w(i) = ( v(i) - w(i+1_i4b)*s(i) - w(i+2_i4b)*t(i) )/r(i)
!
!    end do
!
!   SOLVE THE TRIANGULAR LINEAR SYSTEM R.w = v WITH SCALING
!   TO AVOID OVERFLOWS.
!
    w(:n) = v(:n)
!
    if ( abs(w(n))>abs(r(n)) ) then
!
        a           = r(n)/w(n)
        w(:n-1_i4b) = a*w(:n-1_i4b)
        w(n)        = one
!
    else
!
        w(n) = w(n)/r(n)
!
    end if
!
    a = w(n-1_i4b) - s(n-1_i4b)*w(n)
!
    if ( abs(a)>abs(r(n-1_i4b)) ) then
!
        a          = r(n-1_i4b)/a
        w(:n)      = a*w(:n)
        w(n-1_i4b) = one
!
    else
!
        w(n-1_i4b) = a/r(n-1_i4b)
!
    end if
!    
    do i = n-2_i4b, 1_i4b, -1_i4b
!
        a = w(i) - s(i)*w(i+1_i4b) - t(i)*w(i+2_i4b)
!
        if ( abs(a)>abs(r(i)) ) then
!
            a     = r(i)/a
            w(:n) = a*w(:n)
            w(i)  = one
!
        else
!
            w(i) = a/r(i)
!
        end if
!
    end do
!
!   FIND OPTIMAL k AND tau .
!
    tau = abs( w(1_i4b) )
    k   = 1_i4b
!
    do i = 2_i4b, n
!
        a = abs( v(i-1_i4b)*w(i) )
!
        if ( a<tau ) cycle
!
        tau = a
        k   = i
!
    end do
!
    maxdiag_gkinv_qr = k
!
!
! END OF function maxdiag_gkinv_qr
! ________________________________
!
    end function maxdiag_gkinv_qr
!
! =========================================================================================
!
    function maxdiag_gkinv_ldu( e, lambda )
!
! Purpose
! _______
!
!   This function computes the index of the element of maximum absolute value in the
!   diagonal entries of
!
!                         ( GK - LAMBDA * I )**(-1)
!
!   where GK is a n-by-n symmetric tridiagonal matrix with a zero diagonal, 
!   I is the identity matrix and LAMBDA is a scalar.
!
!
! Arguments
! _________
!
!   E           (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 off-diagonal elements of the tridiagonal matrix.
!
!   LAMBDA      (INPUT) real(stnd)
!               On entry, the eigenvalue or shift used.
!
!
! Further Details
! _______________
!
!   The diagonal entries of ( GK - LAMBDA * I )**(-1) are computed by means of two
!   triangular factorizations of ( GK - LAMBDA * I ) of the forms L(+) * D(+) * U(+) and
!   U(-) * D(-) * L(-) where L(+) and L(-) are unit lower bidiagonal, U(+) and U(-) are unit
!   upper bidiagonal, and D(+) and D(-) are diagonal.
!
!   It is assumed that GK is unreduced, but no check is done in the subroutine to verify
!   this assumption.
!
!   This subroutine is adapted from the references (1) and (2).
!
!   For further details, on Fernando's method for computing eigenvectors of tridiagonal
!   matrices, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in) :: lambda
    real(stnd), dimension(:),   intent(in) :: e
!
    integer(i4b) :: maxdiag_gkinv_ldu
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: gamma, dmin, tau, x
    real(stnd), dimension(size(e)+1) :: diag, diag2, e2
!
    integer(i4b) :: n, i, k
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = int( size(e), i4b) + 1_i4b
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        maxdiag_gkinv_ldu = n
!
        return
!
    end if
!
!   DETERMINE THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
    gamma  = sqrt( safmin )
!
!   REPLACEMENT FOR VERY SMALL PIVOTS IN LDU AND UDL FACTORIZATIONS.
!
!    x    = maxval( abs( e(1_i4b:n-1_i4b) ) )
!    dmin = max( safmin, (safmin*x)*x ) 
!
    dmin = gamma
!
!   MAKE A LOCAL COPY OF THE OFF DIAGONAL ELEMENTS OF THE TRIDIAGONAL MATRIX T.
!
    e2(:n-1_i4b) = e(:n-1_i4b)
    e2(n)        = zero
!
!   REPLACE SMALL SUBDIAGONALS BY gamma.
!
    where ( abs(e2(1_i4b:n-1_i4b))<=gamma )                      &
          e2(1_i4b:n-1_i4b) = sign( gamma, e2(1_i4b:n-1_i4b) )
!
!   SQUARE THE OFF DIAGONAL ELEMENTS OF THE TRIDIAGONAL MATRIX GK.
!
    e2(:n-1_i4b) = e2(:n-1_i4b)*e2(:n-1_i4b)
!    
!   FERNANDO METHOD.
!
!   LDU FACTORIZATION OF GK - lambda*I .
!
    x = -max( lambda, dmin )
    diag(1_i4b) = x
!
    do i = 2_i4b, n
!
        x = - e2(i-1_i4b)/x - lambda
!
        if ( abs(x)<dmin ) then
            x = -dmin
        end if
!
        diag(i) = x
!
    end do
!
!   UDL FACTORIZATION OF GK - lambda*I .
!
    x = -max( lambda, dmin )
    diag2(n) = x
!
    do i = n-1_i4b, 1_i4b, -1_i4b
!
        x = - e2(i)/x - lambda
!
        if ( abs(x)<dmin ) then
            x = -dmin
        end if
!
        diag2(i) = x
!
    end do
!
!   FIND OPTIMAL k AND tau.
!
    tau = abs( diag(1_i4b) + diag2(1_i4b) + lambda )
    k = 1_i4b
!
    do i = 2_i4b, n
!
        x = abs( diag(i) + diag2(i) + lambda )
!
        if ( x>=tau ) cycle
!
        tau = x
        k   = i
!
    end do
!
    maxdiag_gkinv_ldu = k
!
!
! END OF function maxdiag_gkinv_ldu
! _________________________________
!
    end function maxdiag_gkinv_ldu
!
! =========================================================================================
!
    subroutine gk_qr_cmp_rv( e, lambda, cs, sn, diag, sup1, sup2, maxdiag_gkinv )
!
! Purpose
! _______
!
!  GK_QR_CMP factorizes the symmetric matrix GK - LAMBDA * I, where 
!  GK is an n-by-n symmetric tridiagonal matrix with a zero diagonal,
!  I is the identity matrix and LAMBDA is a scalar, as
!
!                         GK - LAMBDA * I = Q * R
!
!  where Q is an orthogonal matrix represented as the product of n-1 Givens
!  rotations and R is an upper triangular matrix with at most two non-zero
!  super-diagonal elements per column.
!
!  The parameter LAMBDA is included in the routine so that GK_QR_CMP may
!  be used to obtain eigenvectors of GK by inverse iteration.
!
!  The subroutine also computes the index of the entry of maximum absolute value
!  in the diagonal of ( GK - LAMBDA * I )**(-1), which provides a good initial
!  approximation to start the inverse iteration process for computing the eigenvector
!  associated with the eigenvalue LAMBDA, see the references (1), (2) and (3) for
!  further details.
!
!
! Arguments
! _________
!
!   E             (INPUT) real(stnd), dimension(:)
!                 On entry, the n-1 off-diagonal elements of the tridiagonal matrix.
!
!   LAMBDA        (INPUT) real(stnd)
!                 On entry, the eigenvalue or shift used in the QR factorization.
!
!   CS            (OUTPUT) real(stnd), dimension(:)
!                 On exit, the vector of the cosines coefficients of the chain of n-1 Givens
!                 rotations for the QR factorization of GK - LAMBDA * I.
!
!                 The size of CS must be  size( CS ) = size( E ) = n - 1.
!
!   SN            (OUTPUT) real(stnd), dimension(:)
!                 On exit, the vector of the sines coefficients of the chain of n-1 Givens
!                 rotations for the QR factorization of GK - LAMBDA * I.
!
!                 The size of SN must be  size( SN ) = size( E ) = n - 1.
!
!   DIAG          (OUTPUT) real(stnd), dimension(:)
!                 On exit, DIAG(:) contains the n diagonal elements of the
!                 upper triangular matrix R of the QR factorization of GK - LAMBDA * I.
!               
!                 The size of DIAG must verify:  size( DIAG ) = size( E ) + 1 = n .
!
!   SUP1          (OUTPUT) real(stnd), dimension(:)
!                 On exit, SUP1(:n-1) contains the n-1 superdiagonal elements of the
!                 upper triangular matrix R of the QR factorization of GK - LAMBDA * I,
!                 SUP1(n) is arbitrary .
!
!                 The size of SUP1 must verify:  size( SUP1 ) = size( E ) + 1 = n .
!
!   SUP2          (OUTPUT) real(stnd), dimension(:)
!                 On exit, SUP2(:n-2) contains the n-2 second superdiagonal elements of
!                 the upper triangular matrix R of the QR factorization of GK - LAMBDA * I,
!                 SUP2(n-1:n) is arbitrary .
!
!                 The size of SUP2 must verify:  size( SUP2 ) = size( E ) + 1 = n .
!
!   MAXDIAG_GKINV (OUPTPUT) integer(i4b)
!                 On exit, MAXDIAG_GKINV is the index of the entry of maximum modulus
!                 in the main diagonal of ( GK - LAMBDA * I )**(-1).
!
!
! Further Details
! _______________
!
!   The QR factorization of ( GK - LAMBDA * I ) is obtained by means of n-1 unitary
!   Givens rotations.
!
!   The diagonal entries of ( GK - LAMBDA * I )**(-1) are computed by means of this QR
!   factorization of ( GK - LAMBDA * I ). For the latter computation, the semiseparable
!   structure of ( GK - LAMBDA * I )**(-1) is used, see the reference (1). Moreover, it
!   is assumed that GK is unreduced for computing the index of the entry of maximum
!   absolute value in the diagonal of ( GK - LAMBDA * I )**(-1), but no check is done
!   in the subroutine to verify this assumption.
!
!   For further details, see:
!
!   (1) Bini, D.A., Gemignani, L., and Tisseur, F., 2005:
!           The Ehrlich-Aberth method for the nonsymmetric tridiagonal eigenvalue problem.
!           SIAM J. Matrix Anal. Appl., 27, 153-175.
!
!   (2) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (3) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,      only : assert_eq
    use Giv_Procedures, only : define_rot_givens
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                intent(in)  :: lambda
    real(stnd), dimension(:),  intent(in)  :: e
    real(stnd), dimension(:),  intent(out) :: cs, sn, diag, sup1, sup2
!
    integer(i4b), intent(out) :: maxdiag_gkinv
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: dmin, a, b, g, csi, sni, tau
    real(stnd), dimension(size(e)+1) :: s, t, w
!
    integer(i4b) :: n, i, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='gk_qr_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(e),i4b)  + 1_i4b ,   &
                       int(size(cs),i4b) + 1_i4b ,   &
                       int(size(sn),i4b) + 1_i4b ,   &
                       int(size(diag),i4b)       ,   &
                       int(size(sup1),i4b)       ,   &
                       int(size(sup2),i4b) /)    ,   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            diag(1_i4b) = - lambda
            sup1(1_i4b) = zero
            sup2(1_i4b) = zero
!
        end if
!
        maxdiag_gkinv = n
!
        return
!
    end if
!
!   REPLACEMENT FOR VERY SMALL PIVOTS IN QR FACTORIZATION.
!
!    g = maxval( abs( e(1_i4b:n-1_i4b) ) )
!    dmin = max( safmin, (safmin*g)*g ) 
!
!    dmin = sqrt( safmin )
!
    dmin = safmin
!
!   MAKE A LOCAL COPY OF THE SHIFTED TRIDIAGONAL MATRIX.
!
    sup1(:n-1_i4b) = e(:n-1_i4b)
    sup1(n)        = zero
!
!   REPLACE SMALL OFF-DIAGONALS BY safmin.
!
    where ( abs(sup1(1_i4b:n-1_i4b))<=safmin )                      &
          sup1(1_i4b:n-1_i4b) = sign( safmin, sup1(1_i4b:n-1_i4b) )
!
!   COMPUTE THE QR FACTORIZATION SAVING THE ENTRIES OF R IN THE VECTORS
!   diag(:), sup1(:) AND sup2(:).
!
    a = -lambda
    g = sup1(1_i4b)
    b = g
!
    do i = 1_i4b, n-1_i4b
!
        call define_rot_givens( a, b, csi, sni )
!
        diag(i) =  csi*a + sni*b
        sup1(i) =  csi*g - sni*lambda
        a       = -sni*g - csi*lambda
!
        b       = sup1(i+1_i4b)
        sup2(i) = sni*b
        g       = csi*b
!
!       SAVE THE COSINES AND SINES COEFFICIENTS OF THE CHAIN
!       OF n-1 GIVENS ROTATIONS DEFINING Q FOR LATER USE.
!
        cs(i) = csi
        sn(i) = sni
!
    end do
!
    diag(n)         = a
    sup2(n-1_i4b:n) = zero
!
!   REPLACE SMALL DIAGONALS BY dmin.
!
    where ( abs(diag(:n))<=dmin ) diag(:n) = sign( dmin, diag(:n) )
!
!   SCALE THE UPPER TRIANGULAR MATRIX R FOR COMPUTING maxdiag_gkinv.
!
    s(:n-1_i4b) = -sn(:n-1_i4b)*sup1(:n-1_i4b)
    s(n)        = zero
!
    t(:n-2_i4b)  = sn(2_i4b:n-1_i4b)*sn(:n-2_i4b)*sup2(:n-2_i4b)
    t(n-1_i4b:n) = zero
!
    w(:n-1_i4b) = cs(:n-1_i4b)
!
!   SOLVE THE TRIANGULAR LINEAR SYSTEM R.w = v WITH SCALING
!   TO AVOID OVERFLOWS.
!
    if ( one>abs(diag(n)) ) then
!
        a           = diag(n)
        w(:n-1_i4b) = a*w(:n-1_i4b)
        w(n)        = one
!
    else
!
        w(n) = one/diag(n)
!
    end if
!
    a = w(n-1_i4b) - s(n-1_i4b)*w(n)
!
    if ( abs(a)>abs(diag(n-1_i4b)) ) then
!
        a          = diag(n-1_i4b)/a
        w(:n)      = a*w(:n)
        w(n-1_i4b) = one
!
    else
!
        w(n-1_i4b) = a/diag(n-1_i4b)
!
    end if
!    
    do i = n-2_i4b, 1_i4b, -1_i4b
!
        a = w(i) - s(i)*w(i+1_i4b) - t(i)*w(i+2_i4b)
!
        if ( abs(a)>abs(diag(i)) ) then
!
            a     = diag(i)/a
            w(:n) = a*w(:n)
            w(i)  = one
!
        else
!
            w(i) = a/diag(i)
!
        end if
!
    end do
!
!   SOLVE THE TRIANGULAR LINEAR SYSTEM R.w = v .
!
!    w(n)       = one/diag(n)
!    w(n-1_i4b) = ( w(n-1_i4b) - w(n)*s(n-1_i4b) )/diag(n-1_i4b)
!
!    do i = n-2_i4b, 1_i4b, -1_i4b
!
!        w(i) = ( w(i) - w(i+1_i4b)*s(i) - w(i+2_i4b)*t(i) )/diag(i)
!
!    end do
!
!   FIND OPTIMAL k AND tau .
!
    tau = abs( w(1_i4b) )
    k   = 1_i4b
!
    do i = 2_i4b, n
!
        a = abs( cs(i-1_i4b)*w(i) )
!
        if ( a<tau ) cycle
!
        tau = a
        k   = i
!
    end do
!
    maxdiag_gkinv = k
!
!
! END OF subroutine gk_qr_cmp_rv
! ______________________________
!
    end subroutine gk_qr_cmp_rv
!
! =========================================================================================
!
    subroutine bd_inviter_rv( upper, d, e, s, leftvec, rightvec, failure,      &
                              maxiter, scaling, initvec )
!
! Purpose
! _______
!
!   BD_INVITER computes the left and right singular vectors of a real n-by-n bidiagonal
!   matrix BD corresponding to a specified singular value, using Fernando's method and 
!   inverse iteration on the tridiagonal Golub-Kahan (TGK) form of the bidiagonal matrix BD.
!
!
! Arguments
! _________
!
!   UPPER          (INPUT) logical(lgl)
!                  On entry, if:
!
!                  - UPPER = true  : BD is upper bidiagonal ;
!                  - UPPER = false : BD is lower bidiagonal.
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, D contains the diagonal elements of the bidiagonal
!                  matrix BD.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, E contains the off-diagonal elements of the bidiagonal
!                  matrix BD. E(1) is arbitrary.
!
!                  The size of E must verify: size( E ) = size( D ) = n . 
!
!   S              (INPUT) real(stnd)
!                  On entry, the selected singular value of the bidiagonal matrix BD.
!                  The singular value must be positive or zero.
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:)
!                  On exit, the computed left singular vector.
!
!                  The shape of LEFTVEC must verify: size( LEFTVEC ) = size( D ) = n .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:)
!                  On exit, the computed right singular vector.
!
!                  The shape of RIGHTVEC must verify: size( RIGHTVEC ) = size( D ) = n .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = FALSE : indicates successful exit,
!                  - FAILURE = TRUE : indicates that some singular vectors
!                    failed to converge in MAXITER iterations.
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed in the subroutine. By default,
!                  2 inverse iterations are performed.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the bidiagonal matrix BD is scaled before computing
!                    the singular vector;
!                  - SCALING=false, the bidiagonal matrix BD is not scaled.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INITVEC=true, a Fernando vector is used to start the inverse iteration process
!                    for computing the singular vectors of the bidiagonal matrix BD (e.g. the eigenvector
!                    of the associated tridiagonal Golub-Kahan matrix);
!                  - INITVEC=false, a random uniform starting vector is used.
!
!                  The default is to use a Fernando starting vector if the Golub-Kahan form of the input
!                  bidiagonal matrix is unreduced, and a random uniform starting vector otherwise.
!
!
! Further Details
! _______________
!
!   A first estimate of the singular vectors is computed by the Fernando method applied to the
!   tridiagonal Golub-Kahan matrix associated with the bidiagonal matrix BD (see the reference
!   (1) for details) if this Golub-Kahan form of the input bidiagonal matrix is unreduced.
!   Otherwise, a random start is used as a first estimate of the singular vectors as in the
!   standard inverse-iteration algorithm.
!
!   The singular vectors are then computed or refined using inverse iteration on the
!   tridiagonal Golub-Kahan matrix.
!
!   For further details, on Fernando's method for computing eigenvectors of tridiagonal
!   matrices or inverse iteration, see
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (4) Bini, D.A., Gemignani, L., and Tisseur, F., 2005:
!           The Ehrlich-Aberth method for the nonsymmetric tridiagonal eigenvalue problem.
!           SIAM J. Matrix Anal. Appl., 27, 153-175.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : svd_error7
    use Utilities,        only : assert_eq, merror, pythag
    use Eig_Procedures,   only : trid_cmp, trid_solve
    use Random,           only : random_number_
#ifdef _BLAS
    use BLAS_interfaces,  only : dot, axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in)                 :: s
    real(stnd),  intent(in),  dimension(:)  :: d, e
    real(stnd),  intent(out), dimension(:)  :: leftvec, rightvec
!
    integer(i4b), intent(in),  optional :: maxiter
!
    logical(lgl), intent(in)            :: upper
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: scaling, initvec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),  dimension(2*size(d)) :: gk, sub, diag, sup1, sup2, temp
    real(stnd) :: lambda, xj, xjm, tmp, prec, normbd, scalfac, tau,       &
                  gamma, omega, ortol, sqrt2, lim
!
    integer(i4b) :: n, n2, mm, maxit, i, ileft, iright, i1, i2, j, l, nn1, nn2
!
    logical(lgl)                        :: do_scaling, fernando_init, test_initvec
    logical(lgl),  dimension(2*size(d)) :: perm
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_inviter'
!
!   maxiter_init IS THE MAXIMUM NUMBER OF ALLOWED ITERATIONS FOR IMPROVING
!   THE INITIAL GUESS OF THE SINGULAR VECTORS.
!
    integer(i4b),  parameter :: maxiter_init=3_i4b
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)         ,   &
                    int(size(e),i4b)         ,   &
                    int(size(leftvec),i4b)   ,   &
                    int(size(rightvec),i4b)  ,   &
                    name_proc )
!
    n2  = 2_i4b*n
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            leftvec(1_i4b)  = one
            rightvec(1_i4b) = one
!
        end if
!
        return
!
    end if
!
!   TEST IF SINGULAR VALUE IS POSITIVE OR ZERO.
!
    if ( s<zero ) then
!
        call merror( name_proc//svd_error7 )
!
    end if
!
!   SET OPTIONAL ARGUMENTS.
!
    if ( present(maxiter) ) then
        maxit = max( maxiter, 1_i4b )
    else
        maxit = 2_i4b
    end if
!
    test_initvec = present( initvec )
!
    if ( present(scaling) ) then
        do_scaling = scaling
    else
        do_scaling = true
    end if
!    
!   GET MACHINES CONSTANTS.
!
    gamma  = sqrt( safmin )
    ortol  = sqrt( ulp )
    prec   = fudge_prec*ulp
    sqrt2  = sqrt( two )
!
!   CHECK IF THE BIDIAGONAL MATRIX IS THE ZERO MATRIX.
!
    diag(:n)     = d(:n)
    sub(1_i4b)   = zero
    sub(2_i4b:n) = e(2_i4b:n)
!
    do j = n, 1_i4b, -1_i4b
!
        if ( abs(diag(j))>=safmin .or. abs(sub(j))>=safmin ) exit
!
        if ( j/=n ) cycle
!
        leftvec(:n) = zero
        leftvec(j)  = one
!
        rightvec(:n) = zero
        rightvec(j)  = one
!
    end do
!
    if ( j==0_i4b ) then
!
!       THE BIDIAGONAL MATRIX IS THE NULL MATRIX, RETURN.
!
        return
!
    end if
!
!   CHECK THE FIRST PART OF THE BIDIAGONAL MATRIX FOR ZEROS
!   AND REDUCE AGAIN ITS SIZE IF POSSIBLE.
!
    nn2 = j
!
    do j = 1_i4b, n-1_i4b
!
        if ( abs(diag(j))>=safmin .or. abs(sub(j+1_i4b))>=safmin ) exit
!
    end do
!
    nn1 = j
!
    if ( nn1>nn2 ) then
!
!       MISMATCH, WE CANNOT WORK ON THE REDUCED PORTION
!       OF THE TRIDIAGONAL GOLUB-KAHAN MATRIX.
!
        nn1 = 1_i4b
        nn2 = n
!
    end if
!
!   FORM THE OFF DIAGONAL ELEMENTS OF THE SYMMETRIC TRIDIAGONAL GOLUB-KAHAN SYSTEM.
!
    gk(1_i4b:n2:2_i4b)       = diag(1_i4b:n)
    gk(2_i4b:n2-1_i4b:2_i4b) = sub(2_i4b:n)
    gk(n2)                   = zero
!    
!   SCALE THE SYMMETRIC TRIDIAGONAL GOLUB-KAHAN SYSTEM IF REQUIRED.    
!    
    if ( do_scaling ) then
!    
        normbd  = maxval( abs(gk(1_i4b:n2-1_i4b) ) )
!    
#ifdef _SCALE
        omega  = scale(one,maxexp-1_i4b)
#else
        omega  = machbase**(maxexp-1_i4b)
#endif
        tau     = sqrt( gamma*omega )
        j       = exponent( tau ) - exponent( normbd ) - 1_i4b
        scalfac = machbase**j
!
        gk(1_i4b:n2-1_i4b) = scalfac*gk(1_i4b:n2-1_i4b)
!
        where( abs(gk(1_i4b:n2-1_i4b))<gamma )
            gk(1_i4b:n2-1_i4b) = sign( gamma, gk(1_i4b:n2-1_i4b) )
        end where
!    
        lambda = scalfac*max( s, safmin )
!    
    else
!    
        scalfac = one
        lambda  = max( s, safmin )
!    
    end if
!
!   COMPUTE THE UPPER GERSCHGORIN BOUND AND THE NORM OF
!   THE TRIDIAGONAL GOLUB-KAHAN MATRIX.
!
    normbd = zero
    tau    = zero
    xj     = zero
!
    do j=1_i4b, n2
!
        xjm    = abs( gk(j) )
        tmp    = xj + xjm
        normbd = normbd + tmp
        tau    = max( tau, tmp )
        xj     = xjm
!
    end do
!
!   DETERMINE THRESHOLD FOR TINY SINGULAR VALUES AND SMALL DIAGONAL ELEMENTS.
!
!    lim = normbd*real( n, stnd )*prec
    lim = tau*real( n, stnd )*prec
!
    i1 = 2_i4b*nn1 - 1_i4b
    i2 = 2_i4b*nn2
!
    if ( lambda<=lim .and. all( abs( gk(i1:i2:2_i4b) ) > lim ) ) then
!    if ( lambda<=lim .and. all( abs( gk(i1:i2:2_i4b) ) > gamma ) ) then
!
!       IF THE SINGULAR VALUE IS VERY SMALL AND NONE OF THE DIAGONAL ELEMENTS OF THE
!       BIDIAGONAL MATRIX ARE ZERO, WE CANNOT WORK SAFELY ON THE REDUCED PORTION
!       OF THE TRIDIAGONAL GOLUB-KAHAN MATRIX.
!
        i1 = 1_i4b
        i2 = n2
!
    else if ( lambda>lim ) then
!
!       IF THE SINGULAR VALUE IS NOT SMALL, TEST IF THE FIRST OR LAST OFF-DIAGONAL
!       ELEMENTS OF THE TRIDIAGONAL GOLUB-KAHAN MATRIX ARE ZERO AND
!       REDUCE ITS SIZE IF POSSIBLE.
!
        if ( abs( gk(i1) )<=lim ) i1 = i1 + 1_i4b
!        if ( abs( gk(i1) )<=gamma ) i1 = i1 + 1_i4b
!
        if ( abs( gk(i2-1_i4b) )<=lim ) i2 = i2 - 1_i4b
!        if ( abs( gk(i2-1_i4b) )<=gamma ) i2 = i2 - 1_i4b
!
    end if
!
    if ( test_initvec ) then
!
        fernando_init = initvec
!
    else
!
!       TEST IF THE GOLUB-KAHAN FORM OF THE INPUT BIDIAGONAL MATRIX IS UNREDUCED.
!
        fernando_init = all( abs( gk(i1:i2-1_i4b) )>lim )
!        fernando_init = all( abs( gk(i1:i2-1_i4b) )>gamma )
!
    end if
!
    if ( upper ) then
!
        ileft  = 2_i4b
        iright = 1_i4b
!
    else
!
        ileft  = 1_i4b
        iright = 2_i4b
!
    end if
!
!   COMPUTE SINGULAR VECTORS BY INVERSE ITERATION ON THE TRIDIAGONAL
!   GOLUB-KAHAN FORM OF THE BIDIAGONAL MATRIX.
!
    temp(i1:i2) = zero
!
!   COMPUTE LU FACTORS WITH PARTIAL PIVOTING (PT=LU) OF THE TGK MATRIX.
!
    call trid_cmp( temp(i1:i2), gk(i1:i2), lambda, sub(i1:i2), diag(i1:i2),    &
                   sup1(i1:i2), sup2(i1:i2), perm(i1:i2) )
!
!   REPLACE SMALL DIAGONAL ELEMENTS BY safmin.
!
    where( abs(diag(i1:i2))<safmin ) diag(i1:i2) = sign( safmin, diag(i1:i2) )
!    
    do l=1_i4b, maxiter_init
!    
        if ( fernando_init ) then
!    
!           FERNANDO METHOD.  
!
            mm = maxdiag_gkinv_ldu( gk(i1:i2-1_i4b), lambda ) + i1 - 1_i4b
!    
!           BINI METHOD.  
!                
!            mm = maxdiag_gkinv_qr( gk(i1:i2-1_i4b), lambda ) + i1 - 1_i4b
!    
            temp(mm)  = one
!
        else
!    
!           COMPUTE A RANDOM INITIAL VECTOR.  
!
            call random_number_( temp(i1:i2) )
!
            temp(i1:i2) = temp(i1:i2)*two - one
!
        end if
!
!       DO maxit ITERATIONS FOR THE EIGENVECTOR OF THE TGK MATRIX.
!
        do i=1_i4b, maxit
!
!           SOLVE THE SYSTEM LUx = Pb WITH RESCALING TO AVOID OVERFLOW.
!
            call trid_solve( sub(i1:i2), diag(i1:i2), sup1(i1:i2),  &
                             sup2(i1:i2), perm(i1:i2), temp(i1:i2)  )
!
            if ( i/=maxit ) then
!
!               NORMALIZE THE EIGENVECTOR OF THE TGK MATRIX.
!
                tmp = norm( temp(i1:i2) )
!
                if ( tmp/=zero ) then
!
                    temp(i1:i2) = temp(i1:i2)/tmp
!
                else
!
                    call random_number_( temp(i1:i2) )
!
                    temp(i1:i2) = temp(i1:i2)*two - one
!
                end if
!
            end if
!
        end do
!
        temp(1_i4b:i1-1_i4b) = zero
        temp(i2+1_i4b:n2)    = zero
!
!       NORMALIZE THE SINGULAR VECTORS.
!
        xj = norm( temp(ileft:n2:2_i4b) )
!
        if ( xj/=zero ) then
            leftvec(1_i4b:n) = temp(ileft:n2:2_i4b)/xj
        else
            leftvec(1_i4b:n) = zero
        end if
!
        xjm = norm( temp(iright:n2:2_i4b) )
!
        if ( xjm/=zero ) then
            rightvec(1_i4b:n) = temp(iright:n2:2_i4b)/xjm
        else
            rightvec(1_i4b:n) = zero
        end if
!
        tau = pythag( xj, xjm )
        tmp = max( xj, xjm )
!
        if ( abs( tmp - ortol*tau )*sqrt2>tau .or. tmp<gamma ) then
!
!           BAD ESTIMATE OF THE SINGULAR VECTORS, TRY TO IMPROVE THEM.
!
            if ( l/=maxiter_init ) then
!
!               BAD ESTIMATE OF THE SINGULAR VECTORS, TRY AGAIN
!               WITH A RANDOM INITIAL VECTOR.
!
                fernando_init = false        
!
                cycle
!
            else if ( tmp==tau  ) then
!
!               BAD SINGULAR VECTORS, SET FAILURE INDICATOR.
!
                failure = true
!
            end if
!
        end if
!
        exit
!
    end do
!
!   CHECK ACCURACY OF THE SINGULAR VECTORS.
!
    if ( min( xj, xjm )/=zero .and. .not.failure ) then
!
        diag(:n)      = d(:n)
        sub(:n)       = e(:n)
        sub(n+1_i4b)  = zero
        sup1(n+1_i4b) = zero
!
        if ( upper ) then
!
            sup1(1_i4b:n) = rightvec(1_i4b:n)
            sup2(1_i4b:n) = leftvec(1_i4b:n)
!
        else
!
            sup1(1_i4b:n) = leftvec(1_i4b:n)
            sup2(1_i4b:n) = rightvec(1_i4b:n)
!
        end if
!
        temp(1_i4b:n) = diag(1_i4b:n)*sup1(1_i4b:n)             +  &
                        sub(2_i4b:n+1_i4b)*sup1(2_i4b:n+1_i4b)
!
!       COMPUTE RAYLEIGH QUOTIENT.
!
#ifdef _BLAS
        tau = dot( n, sup2(1_i4b:n), 1_i4b, temp(1_i4b:n), 1_i4b )
#else
        tau = dot_product( sup2(1_i4b:n), temp(1_i4b:n) )
#endif
!
        if ( tau<zero ) then
            leftvec(1_i4b:n)  = -leftvec(1_i4b:n)
        end if
!
!       COMPUTE THE RESIDUAL NORM OF THE LEFT AND RIGHT SINGULAR VECTORS
!       WITH RESPECT TO THEIR RAYLEIGH QUOTIENT.
!
#ifdef _BLAS
        call axpy( n, -tau, sup2(1_i4b:n), 1_i4b, temp(1_i4b:n), 1_i4b )
#else
        temp(1_i4b:n) = temp(1_i4b:n) - tau*sup2(1_i4b:n)
#endif
        xj  = sum( abs( temp(:n) ) )
        xjm = abs( abs(tau) - s )
!
!       DETERMINE THRESHOLD FOR THE TEST OF ACCURACY.
!
        lim = ( normbd*real(n,stnd)*prec )/scalfac
!
        failure = max( xj, xjm )>lim
!
    else
!
        failure = true
!
    end if
!
!
! END OF SUBROUTINE bd_inviter_rv
! _______________________________
!
    end subroutine bd_inviter_rv
!
! =========================================================================================
!
    subroutine bd_inviter_rm( upper, d, e, s, leftvec, rightvec, failure,      &
                              maxiter, ortho, backward_sweep, scaling, initvec )
!
! Purpose
! _______
!
!   BD_INVITER computes the left and right singular vectors of a real n-by-n bidiagonal
!   matrix BD corresponding to specified singular values, using Fernando's method and 
!   inverse iteration on the tridiagonal Golub-Kahan (TGK) form of the bidiagonal matrix BD.
!
!
! Arguments
! _________
!
!   UPPER          (INPUT) logical(lgl)
!                  On entry, if:
!
!                  - UPPER = true  : BD is upper bidiagonal ;
!                  - UPPER = false : BD is lower bidiagonal.
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, D contains the diagonal elements of the bidiagonal
!                  matrix BD.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, E contains the off-diagonal elements of the bidiagonal
!                  matrix BD. E(1) is arbitrary.
!
!                  The size of E must verify: size( E ) = size( D ) = n . 
!
!   S              (INPUT) real(stnd), dimension(:)
!                  On entry, selected singular values of the bidiagonal matrix BD.
!                  The singular values must be given in decreasing order and must
!                  be positive or zero.
!
!                  The size of S must verify:  size( S ) <= size( D ) = n .
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed left singular vectors. The left singular vector associated
!                  with the singular value S(j) is stored in the j-th column of LEFTVEC.
!
!                  The shape of LEFTVEC must verify: 
!
!                  - size( LEFTVEC, 1 ) = size( D ) = n ,
!                  - size( LEFTVEC, 2 ) = size( S ) .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed right singular vectors. The right singular vector associated
!                  with the singular value S(j) is stored in the j-th column of RIGHTVEC.
!
!                  The shape of RIGHTVEC must verify:  
!
!                  - size( RIGHTVEC, 1 ) = size( D ) = n ,
!                  - size( RIGHTVEC, 2 ) = size( S ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = FALSE : indicates successful exit,
!                  - FAILURE = TRUE : indicates that some singular vectors
!                    failed to converge in MAXITER iterations.
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed in the subroutine. By default,
!                  2 inverse iterations are performed for all the singular vectors.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, all the singular vectors are orthogonalized by the
!                    Modified Gram-Schmidt or QR algorithm;
!                  - ORTHO=false, the singular vectors are not orthogonalized by the 
!                    Modified Gram-Schmidt or QR algorithm.
!
!                  The default is to orthogonalize the singular vectors only for the singular values,
!                  which are not well-separated.
!
!   BACKWARD_SWEEP (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - BACKWARD_SWEEP=true and the singular vectors are orthogonalized by the
!                    modified Gram-Schmidt algorithm, a backward sweep of the modified Gram-Schmidt algorithm
!                    is also performed;
!                  - BACKWARD_SWEEP=false, a backward sweep is not performed.
!
!                  The default is not to perform a backward sweep of the modified Gram-Schmidt algorithm.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the bidiagonal matrix BD is scaled before computing
!                    the singular vectors;
!                  - SCALING=false, the bidiagonal matrix BD is not scaled.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INITVEC=true, Fernando vectors are used to start the inverse iteration process
!                    for computing the singular vectors of the bidiagonal matrix BD (e.g. the eigenvectors of the
!                    associated Golub-Kahan tridiagonal matrix);
!                  - INITVEC=false, random uniform starting vectors are used.
!
!                  The default is to use Fernando starting vectors if the singular values are well-separated and
!                  the Golub-Kahan form of the input bidiagonal matrix is unreduced, and random uniform
!                  starting vectors otherwise.
!
!
! Further Details
! _______________
!
!   A first estimate of the singular vectors is computed by the Fernando method applied to the
!   tridiagonal Golub-Kahan matrix associated with the bidiagonal matrix BD (see the reference
!   (1) for details) for the singular values which are well-separated and if the Golub-Kahan
!   form of the input bidiagonal matrix is unreduced. For the other singular values,
!   a random start is used as a first estimate of the singular vectors as in the
!   standard inverse-iteration algorithm.
!
!   The singular vectors are then computed or refined using inverse iteration on the
!   tridiagonal Golub-Kahan matrix for all the singular values at one step.
!
!   By default, the singular vectors are then orthogonalized by the Modified Gram-Schmidt
!   or QR algorithm only if the singular values are not well-separated.
!
!   The computation of the singular vectors is parallelized if OPENMP is used.
!
!   BD_INVITER may fail if clusters of tiny singular values are present in parameter S.
!
!   For further details, on Fernando's method for computing eigenvectors of tridiagonal
!   matrices or inverse iteration, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (4) Bini, D.A., Gemignani, L., and Tisseur, F., 2005:
!           The Ehrlich-Aberth method for the nonsymmetric tridiagonal eigenvalue problem.
!           SIAM J. Matrix Anal. Appl., 27, 153-175.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : svd_error3, svd_error7, allocate_error
    use Utilities,        only : assert_eq, assert, merror, pythag
    use Eig_Procedures,   only : trid_cmp, trid_solve
    use QR_Procedures,    only : qr_cmp, ortho_gen_qr
    use Random,           only : random_number_
#ifdef _BLAS
    use BLAS_interfaces,  only : dot, axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)    :: d, e, s
    real(stnd),  intent(out), dimension(:,:)  :: leftvec, rightvec
!
    integer(i4b), intent(in),  optional :: maxiter
!
    logical(lgl), intent(in)            :: upper
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: ortho, backward_sweep, scaling, initvec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),  dimension(size(s))   :: lambda, singval
    real(stnd),  dimension(2*size(d)) :: gk, sub, diag, sup1, sup2, temp, temp2
    real(stnd)                        :: xj, xjm, tmp, prec, pertol, sep,           &
                                         normbd, lim, lim_sing, lim_ortho, scalfac, &
                                         tau, gamma, omega, ortol, sqrt2, eigval
!
    integer(i4b)                             :: n, n2, m, mm, maxit, i, ileft, iright,    &
                                                i1, i2, j, j1, j2, k, l, nn1, nn2, m0, p
    integer(i4b),  dimension(:), allocatable :: jstart, jend
    integer                                  :: iok
!
    logical(lgl)                        :: test_ortho, do_ortho, do_backward_sweep, do_scaling,       &
                                           test_initvec, do_initvec, zero_vec, first_zero, last_zero, &
                                           diag_zero, bd_zero, merge_clusters
    logical(lgl),  dimension(size(s))   :: fernando_init, comp_norm
    logical(lgl),  dimension(2*size(d)) :: perm
!
#ifdef _OPENMP
!
    real(stnd),  dimension(size(s)) :: lambda2
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_inviter'
!
!   maxiter_init IS THE MAXIMUM NUMBER OF ALLOWED ITERATIONS FOR IMPROVING
!   THE INITIAL GUESS OF THE SINGULAR VECTORS BEFORE MODIFIED GRAM-SCHMIDT
!   OR QR ORTHOGONALIZATION.
!
    integer(i4b),  parameter :: maxiter_init=3_i4b
!
!   merge_factor IS A "MERGE FACTOR" FOR CLUSTERS OF SINGULAR VALUES: IF ONE CLUSTER CONTAINS
!   MORE THAN merge_factor*m SINGULAR VALUES, ALL THE CLUSTERS WILL BE MERGE TOGETHER.
!
    real(stnd),  parameter :: merge_factor=c0_95
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure  = false
    zero_vec = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)           ,   &
                    int(size(e),i4b)           ,   &
                    int(size(leftvec,1),i4b)   ,   &
                    int(size(rightvec,1),i4b)  ,   &
                    name_proc )
!
    n2  = 2_i4b*n
!
    m =  assert_eq( int(size(s),i4b)           ,   &
                    int(size(leftvec,2),i4b)   ,   &
                    int(size(rightvec,2),i4b)  ,   &
                    name_proc )
!
    call assert( logical( m<=n, lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b .or. m<=0_i4b) then
!
        if ( n==1_i4b .and. m==1_i4b ) then
!
            leftvec(1_i4b,1_i4b)  = one
            rightvec(1_i4b,1_i4b) = one
!
        end if
!
        return
!
    end if
!
!   TEST IF SINGULAR VALUES ARE POSITIVE OR ZERO.
!
    if ( any( s(1_i4b:m)<zero ) ) then
!
        call merror( name_proc//svd_error7 )
!
    end if
!
!   TEST IF SINGULAR VALUES ARE IN DECREASING ORDER.
!
    if ( m/=1_i4b ) then
!
        temp(1_i4b:m-1_i4b) = s(2_i4b:m) - s(1_i4b:m-1_i4b)
!
        if ( any( temp(1_i4b:m-1_i4b)>zero ) ) then
!
            call merror( name_proc//svd_error3 )
!
        end if
!
    end if
!
!   SET OPTIONAL ARGUMENTS.
!
    if ( present(maxiter) ) then
        maxit = max( maxiter, 1_i4b )
    else
        maxit = 2_i4b
    end if
!
    test_ortho = present( ortho )
    do_ortho = false
!
    test_initvec = present( initvec )
!
    if ( present(backward_sweep) ) then
        do_backward_sweep = backward_sweep
    else
        do_backward_sweep = false
    end if
!
    if ( present(scaling) ) then
        do_scaling = scaling
    else
        do_scaling = true
    end if
!    
!   GET MACHINES CONSTANTS.
!
    gamma  = sqrt( safmin )
    ortol  = sqrt( ulp )
    prec   = fudge_prec*ulp
    sqrt2  = sqrt( two )
!
!   CHECK IF THE BIDIAGONAL MATRIX IS THE ZERO MATRIX.
!
    diag(:n)     = d(:n)
    sub(1_i4b)   = zero
    sub(2_i4b:n) = e(2_i4b:n)
!
    m0 = n
!
    do j = n, 1_i4b, -1_i4b
!
        if ( abs(diag(j))>=safmin .or. abs(sub(j))>=safmin ) exit
!
        if ( m0>0_i4b .and. m0<=m ) then
!
            leftvec(:n,m0) = zero
            leftvec(j,m0)  = one
!
            rightvec(:n,m0) = zero
            rightvec(j,m0)  = one
!
        end if
!
        m0 = m0 - 1_i4b
!
    end do
!
    if ( j==0_i4b ) then
!
!       THE BIDIAGONAL MATRIX IS THE NULL MATRIX, RETURN.
!
        return
!
    end if
!
!   CHECK THE FIRST PART OF THE BIDIAGONAL MATRIX FOR ZEROS
!   AND REDUCE AGAIN ITS SIZE IF POSSIBLE.
!
    nn2 = j
!
    do j = 1_i4b, n-1_i4b
!
        if ( abs(diag(j))>=safmin .or. abs(sub(j+1_i4b))>=safmin ) exit
!
        if ( m0>0_i4b .and. m0<=m ) then
!
            leftvec(:n,m0) = zero
            leftvec(j,m0)  = one
!
            rightvec(:n,m0) = zero
            rightvec(j,m0)  = one
!
        end if
!
        m0 = m0 - 1_i4b
!
    end do
!
    nn1 = j
!
    m0 = min( m0, m )
!
    if (  m0>nn2-nn1+1_i4b .or. m0<=0_i4b ) then
!
!       MISMATCH, WE CANNOT WORK ON THE REDUCED PORTION
!       OF THE TRIDIAGONAL GOLUB-KAHAN MATRIX.
!
        nn1 = 1_i4b
        nn2 = n
!
        m0 = m
!
    end if
!
    lambda(m0+1_i4b:m) = one
!
#ifdef _OPENMP
    lambda2(m0+1_i4b:m) = one
#endif
!
!   FORM THE OFF DIAGONAL ELEMENTS OF THE SYMMETRIC TRIDIAGONAL GOLUB-KAHAN SYSTEM.
!
    gk(1_i4b:n2:2_i4b)       = diag(1_i4b:n)
    gk(2_i4b:n2-1_i4b:2_i4b) = sub(2_i4b:n)
    gk(n2)                   = zero
!    
!   SCALE THE SYMMETRIC TRIDIAGONAL GOLUB-KAHAN SYSTEM IF REQUIRED.    
!    
    if ( do_scaling ) then
!    
        normbd  = maxval( abs(gk(1_i4b:n2-1_i4b) ) )
!    
#ifdef _SCALE
        omega  = scale(one,maxexp-1_i4b)
#else
        omega  = machbase**(maxexp-1_i4b)
#endif
        tau     = sqrt( gamma*omega )
        j       = exponent( tau ) - exponent( normbd ) - 1_i4b
        scalfac = machbase**j
!
        gk(1_i4b:n2-1_i4b) = scalfac*gk(1_i4b:n2-1_i4b)
!
        where( abs(gk(1_i4b:n2-1_i4b))<gamma )
            gk(1_i4b:n2-1_i4b) = sign( gamma, gk(1_i4b:n2-1_i4b) )
        end where
!    
        singval(:m) = scalfac*max( s(:m), safmin )
!    
    else
!    
        scalfac     = one
        singval(:m) = max( s(:m), safmin )
!    
    end if
!
!   COMPUTE THE UPPER GERSCHGORIN BOUND AND THE NORM OF
!   THE TRIDIAGONAL GOLUB-KAHAN MATRIX.
!
    normbd = zero
    tau    = zero
    xj     = zero
!
    do j=1_i4b, n2
!
        xjm    = abs( gk(j) )
        tmp    = xj + xjm
        normbd = normbd + tmp
        tau    = max( tau, tmp )
        xj     = xjm
!
    end do
!
!   DETERMINE THRESHOLD FOR TINY SINGULAR VALUES AND SMALL DIAGONAL ELEMENTS.
!
!    lim_sing = normbd*real( n, stnd )*prec
    lim_sing = tau*real( n, stnd )*prec
!
    i1 = 2_i4b*nn1 - 1_i4b
    i2 = 2_i4b*nn2
!
!   TEST IF THE FIRST OR LAST SUBDIAGONAL ELEMENTS OF THE TRIDIAGONAL GOLUB-KAHAN MATRIX ARE ZERO.
!
    first_zero = abs( gk(i1) )<=lim_sing
!    first_zero = abs( gk(i1) )<=gamma
!
    last_zero = abs( gk(i2-1_i4b) )<=lim_sing
!    last_zero = abs( gk(i2-1_i4b) )<=gamma
!
!   TEST IF THE GOLUB-KAHAN FORM OF THE INPUT BIDIAGONAL MATRIX IS UNREDUCED.
!
    if ( m/=n .and. singval(m)<=lim_sing ) then
!
        m0 = m
!
        do_initvec = all( abs( gk(:n2-1_i4b) )>lim_sing )
!        do_initvec = all( abs( gk(:n2-1_i4b) )>gamma )
!
        diag_zero = false
!
    else
!
        do_initvec = all( abs( gk(i1:i2-1_i4b) )>lim_sing )
!        do_initvec = all( abs( gk(i1:i2-1_i4b) )>gamma )
!
        if ( .not.first_zero .and. .not.last_zero ) then
!
            diag_zero  = any( abs( gk(i1:i2-1_i4b:2_i4b) )<=lim_sing )
!            diag_zero  = any( abs( gk(i1:i2-1_i4b:2_i4b) )<=gamma )
!
        else
!
            diag_zero = true
!
        end if
!
    end if
!
!   DETERMINE THRESHOLDS FOR SEPARATION OF SINGULAR VALUES AND
!   REORTHOGONALIZATION OF SINGULAR VECTORS.
!
    lim_ortho = fudge_ortho*sqrt( macheps )*tau
!    
!    omega = macheps*normbd
    omega = lim_ortho
!    omega = real( n, stnd )*sqrt( macheps )*normbd
!
!   IF SINGULAR VALUES ARE TOO CLOSE, ADD A RELATIVELY SMALL PERTURBATION
!   TO AVOID PARALLEL ITERATES.
!   IF SINGULAR VALUES ARE TOO CLOSE, REORTHOGONALIZE SINGULAR VECTORS BY
!   MODIFIED GRAM-SCHMIDT.
!
    xjm = singval(m0)
!
    do j=m0-1_i4b, 1_i4b, -1_i4b
!
        xj     = singval(j)
        pertol = abs( prec*xj )
        sep    = xj - xjm
!
        comp_norm(j+1_i4b) = sep<lim_ortho
!
        fernando_init(j+1_i4b) = sep>omega
!
!BUG:   THE NAGFOR COMPILER DO NOT IMPLEMENT THE MERGE FUNCTION
!       CORRECTLY IN EXTENDED PRECISION. THUS, IT IS SAFER TO USE
!       A SIMPLE IF THEN ELSE INSTRUCTION.
!
!        xjm = merge( xj, xjm+pertol, sep>=pertol )
!
        if ( sep>=pertol ) then
            xjm = xj
        else
            xjm       = xjm + pertol
            singval(j) = xjm
        end if
!
    end do
!
    if ( diag_zero .and. (m0==nn2-nn1+1_i4b) ) then
!
        singval(m0) = zero
!
    end if
!
!   SET INDICATOR VARIABLES FOR ORTHOGONALIZATION.
!
    comp_norm(1_i4b) = false
!
    if ( test_ortho ) then
!
        if ( m0/=1_i4b ) then
!
            do_ortho = ortho
!
        end if
!
        if ( do_ortho ) then
!
            comp_norm(2_i4b:m0) = true
!
        end if
!
    else
!
        do_ortho = any( comp_norm(2_i4b:m0) )
!
    end if
!
!   SET INDICATOR VARIABLE FOR INITIALIZATION OF THE
!   INVERSE ITERATION PROCESS FOR EACH SINGULAR VECTOR.
!
    if ( test_initvec ) then
        do_initvec = initvec
    end if
!    
    if ( do_initvec ) then
!
        if ( test_initvec ) then
!
!           START THE RIGHT HAND SIDES WITH FERNANDO VECTORS.
!
            fernando_init(1_i4b:m0) = true
!
        else
!
!           START THE RIGHT HAND SIDES WITH FERNANDO VECTORS,
!           ONLY IF THE SINGULAR VALUES ARE WELL SEPARATED.
!
            fernando_init(1_i4b) = true
!
        end if
!
    else
!
!       START THE RIGHT HAND SIDES AT RANDOM VALUES.
!
        fernando_init(1_i4b:m0) = false
!
    end if
!
    if ( upper ) then
!
        ileft  = 2_i4b
        iright = 1_i4b
!
    else
!
        ileft  = 1_i4b
        iright = 2_i4b
!
    end if
!
!   DETERMINE THRESHOLD FOR THE TEST OF ACCURACY
!   OF THE COMPUTED SINGULAR VECTORS.
!
    lim = ( normbd*real(n,stnd)*prec )/scalfac
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m0>1_i4b                            .and.       &
                  i>1_i4b
!
!   COMPUTE SINGULAR VECTORS BY INVERSE ITERATION ON THE TRIDIAGONAL
!   GOLUB-KAHAN FORM OF THE BIDIAGONAL MATRIX.
!
    if ( test_par ) then
!
!       OPENMP PARALLEL PATH.
!
!$OMP PARALLEL PRIVATE(bd_zero,i,j,j1,j2,l,p,mm,xj,xjm,tau,tmp,eigval, &
!$OMP                  sub,diag,sup1,sup2,perm,temp,temp2)
!
!$OMP DO SCHEDULE(STATIC)
!
        do j = 1_i4b, m0
!
!            eigval = max( singval(j), safmin )
            eigval = singval(j)
!
            bd_zero     = false
!
            j1 = i1
            j2 = i2
!
            if ( eigval>lim_sing .or. m==n ) then
!
                if ( m/=n ) then
!
                    if ( first_zero ) j1 = j1 + 1_i4b
                    if ( last_zero )  j2 = j2 - 1_i4b
!
                end if
!
                bd_zero = diag_zero .and. j==nn2-nn1+1_i4b
!
            else if ( m/=n ) then
!
                j1 = 1_i4b
                j2 = n2
!
            end if
!
            temp(1_i4b:n2) = zero
!
!           COMPUTE LU FACTORS WITH PARTIAL PIVOTING (PT=LU) OF THE TGK MATRIX.
!
            call trid_cmp( temp(j1:j2), gk(j1:j2), eigval, sub(j1:j2), diag(j1:j2),    &
                           sup1(j1:j2), sup2(j1:j2), perm(j1:j2) )
!
!           REPLACE SMALL DIAGONAL ELEMENTS BY safmin.
!
            where( abs(diag(j1:j2))<safmin ) diag(j1:j2) = sign( safmin, diag(j1:j2) )
!
            do l = 1_i4b, maxiter_init
!
                if ( fernando_init(j) ) then
!    
!                   FERNANDO METHOD.  
!
                    mm = maxdiag_gkinv_ldu( gk(j1:j2-1_i4b), eigval ) + j1 - 1_i4b
!    
!                   BINI METHOD.  
!                
!                   mm = maxdiag_gkinv_qr( gk(j1:j2-1_i4b), eigval ) + j1 - 1_i4b
!    
                    if ( comp_norm(j) ) then
!
!                       ADD A SMALL PERTURBATION TO ALL COMPONENTS OF THE
!                       INITIAL VECTOR TO PROTECT AGAINST FAILURE OF FERNANDO
!                       METHOD WHICH MAY PRODUCE PARALLEL INITIAL VECTORS IF
!                       SINGULAR VALUES ARE TOO CLOSED.
!
                        call random_number_( temp(j1:j2) )
!
                        temp(j1:j2) = sqrt( macheps )*( temp(j1:j2)*two - one )
!    
                    else
!
                        temp(j1:j2) = zero
!    
                    end if
!    
                    temp(mm) = one
!
                else
!
!                   STANDARD INVERSE ITERATION METHOD.
!    
!                   COMPUTE A RANDOM INITIAL VECTOR.  
!
                    call random_number_( temp(j1:j2) )
!
                    temp(j1:j2) = temp(j1:j2)*two - one
!
                end if
!
                if ( bd_zero .and. .not.fernando_init(j) ) then
!
                    do p=1_i4b, 2_i4b
!    
                        if ( p==1_i4b ) then
!    
                            if ( first_zero ) then
!    
                                temp2(j1)          = one
                                temp2(j1+1_i4b:j2) = zero
!    
                            else
!    
                                temp2(j1:j2:2_i4b)       = temp(j1:j2:2_i4b)
                                temp2(j1+1_i4b:j2:2_i4b) = zero
!    
                            end if
!    
                        else
!    
                            if ( last_zero ) then
!    
                                temp2(j2)          = one
                                temp2(j1:j2-1_i4b) = zero
!    
                            else
!    
                                temp2(j1:j2:2_i4b)       = zero
                                temp2(j1+1_i4b:j2:2_i4b) = temp(j1+1_i4b:j2:2_i4b)
!    
                            end if
!    
                        end if
!
!                       DO maxit ITERATIONS FOR THE EIGENVECTORS OF THE TGK MATRIX.
!
                        do i=1_i4b, maxit
!
!                           SOLVE THE SYSTEM LUx = Pb WITH RESCALING TO AVOID OVERFLOW.
!
                            call trid_solve( sub(j1:j2), diag(j1:j2), sup1(j1:j2),  &
                                             sup2(j1:j2), perm(j1:j2), temp2(j1:j2) )
!
                            if ( i/=maxit ) then
!
!                               NORMALIZE THE EIGENVECTORS OF THE TGK MATRIX.
!
                                xj = norm( temp2(j1:j2) )
!
                                if ( xj/=zero ) then
!
                                    temp2(j1:j2) = temp2(j1:j2)/xj
!
                                else
!
                                    call random_number_( temp2(j1:j2) )
!
                                    if ( p==1_i4b ) then
!    
                                        temp2(j1:j2:2_i4b)       = temp2(j1:j2:2_i4b)*two - one
                                        temp2(j1+1_i4b:j2:2_i4b) = zero
!    
                                    else
!    
                                        temp2(j1:j2:2_i4b) = zero
                                        temp2(j1+1_i4b:j2:2_i4b) = temp2(j1+1_i4b:j2:2_i4b)*two - one
!    
                                    end if
!
                                end if
!
                            end if
!
                        end do
!    
                        if ( p==1_i4b ) then
!    
                            temp(j1:j2:2_i4b) = temp2(j1:j2:2_i4b)
!    
                        else
!    
                            temp(j1+1_i4b:j2:2_i4b) = temp2(j1+1_i4b:j2:2_i4b)
!    
                        end if
!
                    end do
!
                else
!
!                   DO maxit ITERATIONS FOR THE EIGENVECTOR OF THE TGK MATRIX.
!
                    do i=1_i4b, maxit
!
!                       SOLVE THE SYSTEM LUx = Pb WITH RESCALING TO AVOID OVERFLOW.
!
                        call trid_solve( sub(j1:j2), diag(j1:j2), sup1(j1:j2),  &
                                         sup2(j1:j2), perm(j1:j2), temp(j1:j2)  )
!
                        if ( i/=maxit ) then
!
!                           NORMALIZE THE EIGENVECTORS OF THE TGK MATRIX.
!
                            xj = norm( temp(j1:j2) )
!
                            if ( xj/=zero ) then
!
                                temp(j1:j2) = temp(j1:j2)/xj
!
                            else
!
                                call random_number_( temp(j1:j2) )
!
                                temp(j1:j2) = temp(j1:j2)*two - one
!
                            end if
!
                        end if
!
                    end do
!
                end if
!
!               NORMALIZE THE SINGULAR VECTORS.
!
                xj = norm( temp(ileft:n2:2_i4b) )
!
                if ( xj/=zero ) then
                    leftvec(1_i4b:n,j) = temp(ileft:n2:2_i4b)/xj
                else
                    leftvec(1_i4b:n,j) = zero
                end if
!
                xjm = norm( temp(iright:n2:2_i4b) )
!
                if ( xjm/=zero ) then
                    rightvec(1_i4b:n,j) = temp(iright:n2:2_i4b)/xjm
                else
                    rightvec(1_i4b:n,j) = zero
                end if
!
                tau = pythag( xj, xjm )
                tmp = max( xj, xjm )
!
                if ( abs( tmp - ortol*tau )*sqrt2>tau .or. tmp<gamma ) then
!
!                   BAD ESTIMATES OF THE SINGULAR VECTORS, TRY TO IMPROVE THEM.
!
                    if ( l/=maxiter_init ) then
!
!                       BAD ESTIMATES OF THE SINGULAR VECTORS, TRY AGAIN
!                       WITH A NEW RANDOM INITIAL VECTOR.
!
                        fernando_init(j) = false        
!
                        cycle
!
                    else if ( .not.test_ortho .and. tmp/=tau .and. j/=1_i4b ) then
!
!                       BAD ESTIMATES OF THE SINGULAR VECTORS, REORTHOGONALIZE THEM
!                       BY MODIFIED GRAM-SCHMIDT.
!
                        comp_norm(j) = true
!
!$OMP CRITICAL (orth)
                        do_ortho = true
!
!$OMP END CRITICAL (orth)
!
                    end if
!
                end if
!
                exit
!
            end do
!
            lambda(j)  = xj
            lambda2(j) = xjm
!
        end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
        if ( do_ortho ) then
!
!           REORTHOGONALIZE LEFT AND RIGHT SINGULAR VECTORS BY MODIFIED GRAM-SCHMIDT
!           OR A QR FACTORIZATION.
!
!           FIRST DETERMINE THE NUMBER OF CLUSTERS OF SINGULAR VALUES.
!
            i = count( comp_norm(2_i4b:m0) )
!
            if ( i==m0-1_i4b ) then
                k = 1_i4b
            else
                k = i
            end if
!
!           ALLOCATE WORK VARIABLES.
!
            allocate( jstart(k), jend(k), stat = iok  )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           k IS THE NUMBER OF CLUSTERS OF SINGULAR VALUES.
!           jstart(k) IS THE INDICE OF THE FIRST SINGULAR VALUE IN CLUSTER k.
!           jend(k)+1 IS THE INDICE OF THE LAST SINGULAR VALUE IN CLUSTER k.
!
            if ( i==m0-1_i4b ) then
!
!               ALL THE SINGULAR VALUES ARE CLUSTERED TOGETHER.
!
                jstart(1_i4b) = 1_i4b
                jend(1_i4b)   = m0 - 1_i4b
!
            else
!
!               UNKNOWN NUMBER OF CLUSTERS, BUT AT LEAST ONE
!               CLUSTER OF SINGULAR VALUES.
!
                j = 2_i4b
                k = 0_i4b
!
                do
!
                    if ( comp_norm(j) ) then
!
                        k = k + 1_i4b
!
                        jstart(k) = j - 1_i4b
!
                        do i=j+1_i4b, m0
                            if ( .not.comp_norm(i) ) exit
                        end do
!
                        jend(k) = i - 2_i4b
!
                        j = i + 1_i4b
!
                    else
!
                        j = j + 1_i4b
!
                    end if
!
                    if ( j>m0 ) exit
!
                end do
!
!               IF ONE CLUSTER CONTAINS ALMOST ALL THE SINGULAR VALUES, MERGE
!               ALL THE CLUSTERS TOGETHER AS IN THIS CASE USING ONLY ONE QR
!               FACTORIZATION IS EQUALLY FAST.
!
                xj = real( m0, stnd )*merge_factor
                i  = int( xj, i4b )
!
                merge_clusters = false
!
                do l=1_i4b, k
!
                    j = jend(l) - jstart(l) + 2_i4b            
!
                    if ( j>i ) then
!
                        merge_clusters = true
!
                        exit
!
                    end if
!
                end do
!
                if ( merge_clusters ) then
!
                    k = 1_i4b
!
                    jstart(1_i4b) = 1_i4b
                    jend(1_i4b)   = m0 - 1_i4b
!
                    comp_norm(2_i4b:m0) = true
!
                end if
!
            end if
!
!           FIRST, REORTHOGONALIZE LEFT AND RIGHT SINGULAR VECTORS CORRESPONDING TO
!           BIG CLUSTERS OF ALMOST-MULTIPLE SINGULAR VALUES. PARALLELIZATION IS OVER
!           THE NUMBER OF VECTORS IN EACH CLUSTER. REORTHOGONALIZATION IS PERFORMED
!           BY A QR FACTORIZATION.
!
            do l=1_i4b, k
!
                j1 = jstart(l)
                j2 = jend(l)
!
                if ( j2-j1>=nsing_cluster ) then
!
                    j2 = j2 + 1_i4b
!
                    j = j2 - j1 + 1_i4b
!
!                   REORTHOGONALIZE LEFT SINGULAR VECTORS.
!
                    call qr_cmp( leftvec(:n,j1:j2), diag(:j), sub(:j) )
!
                    call ortho_gen_qr( leftvec(:n,j1:j2), sub(:j) )
!
!                   REORTHOGONALIZE RIGHT SINGULAR VECTORS.
!
                    call qr_cmp( rightvec(:n,j1:j2), diag(:j), sub(:j) )
!
                    call ortho_gen_qr( rightvec(:n,j1:j2), sub(:j) )
!
                end if
!
            end do
!
        end if
!
!$OMP PARALLEL PRIVATE(i,j,j1,j2,l,mm,sub,diag,sup1,sup2,temp,xj,tau)
!
        if ( do_ortho ) then
!
!           SECOND, REORTHOGONALIZE LEFT AND RIGHT SINGULAR VECTORS CORRESPONDING TO
!           SMALL CLUSTERS OF ALMOST-MULTIPLE SINGULAR VALUES. EACH THREAD TAKES CARE
!           OF ONE CLUSTER AND PARALLELIZATION IS OVER THE NUMBER OF SMALL CLUSTERS.
!           REORTHOGONALIZATION IS PERFORMED BY MODIFIED GRAM-SCHMIDT.
!
!$OMP DO SCHEDULE(DYNAMIC,1)

            do l=1_i4b, k
!
                j1 = jstart(l)
                j2 = jend(l)
!
                if ( j2-j1<nsing_cluster ) then
!
                    mm = j2 + 1_i4b
!
!                   REORTHOGONALIZE LEFT SINGULAR VECTORS BY MODIFIED GRAM-SCHMIDT.
!
!                   FORWARD SWEEP OF MGS ORTHOGONALIZATION FOR SMALL CLUSTERS.
!
                    do j=j1, j2 
!
!                       MAKE A LOCAL COPY OF THE j LEFT SINGULAR VECTOR.
!
                        sup1(:n) = leftvec(:n,j)
!
                        do i=j+1_i4b, mm
!
#ifdef _BLAS
                            xj = -dot( n, sup1(:n), 1_i4b, leftvec(:n,i), 1_i4b )
                            call axpy( n, xj, sup1(:n), 1_i4b, leftvec(:n,i), 1_i4b )
#else
                            xj = -dot_product( sup1(:n), leftvec(:n,i) )
                            leftvec(:n,i) = leftvec(:n,i) + xj*sup1(:n)
#endif
!
                        end do
!
!                       NORMALIZE THE NEXT j LEFT SINGULAR VECTOR.
!
                        sup1(:n) = leftvec(:n,j+1_i4b)
!
                        xj = norm( sup1(1_i4b:n) )
!
                        if ( xj<=fudge_ortho2 ) then
!
!                           REORTHOGONALIZATION IS REQUIRED.
!
                            call ortho_vec(leftvec(:n,j1:j), sup1(:n), xj )
!
                        else
!
!                           NORMALIZATION IS ENOUGH.
!
                            sup1(:n) = (one/xj)*sup1(:n)
!
                        end if
!
                        leftvec(:n,j+1_i4b) = sup1(:n)
!
                        lambda(j+1_i4b) = xj
!
                    end do
!
                    if ( do_backward_sweep ) then
!
!                       BACKWARD SWEEP OF MGS ORTHOGONALIZATION FOR SMALL CLUSTERS.
!
                        do j=j2, j1, -1_i4b
!
                            sup1(:n) = leftvec(:n,j)
!
                            do i=j+1_i4b, mm
!
#ifdef _BLAS
                                xj = -dot( n, sup1(:n), 1_i4b, leftvec(:n,i), 1_i4b )
                                call axpy( n, xj, sup1(:n), 1_i4b, leftvec(:n,i), 1_i4b )
#else
                                xj = -dot_product( sup1(:n), leftvec(:n,i) )
                                leftvec(:n,i) = leftvec(:n,i) + xj*sup1(:n)
#endif
!
                            end do
!
                        end do
!
                    end if
!
!                   REORTHOGONALIZE RIGHT SINGULAR VECTORS BY MODIFIED GRAM-SCHMIDT.
!
!                   FORWARD SWEEP OF MGS ORTHOGONALIZATION FOR SMALL CLUSTERS.
!
                    do j=j1, j2 
!
!                       MAKE A LOCAL COPY OF THE j RIGHT SINGULAR VECTOR.
!
                        sup1(:n) = rightvec(:n,j)
!
                        do i=j+1_i4b, mm
!
#ifdef _BLAS
                            xj = -dot( n, sup1(:n), 1_i4b, rightvec(:n,i), 1_i4b )
                            call axpy( n, xj, sup1(:n), 1_i4b, rightvec(:n,i), 1_i4b )
#else
                            xj = -dot_product( sup1(:n), rightvec(:n,i) )
                            rightvec(:n,i) = rightvec(:n,i) + xj*sup1(:n)
#endif
!
                        end do
!
!                       NORMALIZE THE NEXT j RIGHT SINGULAR VECTOR.
!
                        sup1(:n) = rightvec(:n,j+1_i4b)
!
                        xj = norm( sup1(1_i4b:n) )
!
                        if ( xj<=fudge_ortho2 ) then
!
!                           REORTHOGONALIZATION IS REQUIRED.
!
                            call ortho_vec(rightvec(:n,j1:j), sup1(:n), xj )
!
                        else
!
!                           NORMALIZATION IS ENOUGH.
!
                            sup1(:n) = (one/xj)*sup1(:n)
!
                        end if
!
                        rightvec(:n,j+1_i4b) = sup1(:n)
!
                        lambda2(j+1_i4b) = xj
!
                    end do
!
                    if ( do_backward_sweep ) then
!
!                       BACKWARD SWEEP OF MGS ORTHOGONALIZATION FOR SMALL CLUSTERS.
!
                        do j=j2, j1, -1_i4b
!
                            sup1(:n) = rightvec(:n,j)
!
                            do i=j+1_i4b, mm
!
#ifdef _BLAS
                                xj = -dot( n, sup1(:n), 1_i4b, rightvec(:n,i), 1_i4b )
                                call axpy( n, xj, sup1(:n), 1_i4b, rightvec(:n,i), 1_i4b )
#else
                                xj = -dot_product( sup1(:n), rightvec(:n,i) )
                                rightvec(:n,i) = rightvec(:n,i) + xj*sup1(:n)
#endif
!
                            end do
!
                        end do
!
                    end if
!
                end if
!
            end do
!
!$OMP END DO
!
        end if
!
!       CHECK ACCURACY OF THE SINGULAR VECTORS.
!
        diag(:n)      = d(:n)
        sub(:n)       = e(:n)
        sub(n+1_i4b)  = zero
        sup1(n+1_i4b) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
        do j = 1_i4b, m
!
            if ( lambda(j)/=zero .and. lambda2(j)/=zero  ) then
!
                if ( upper ) then
                    sup1(1_i4b:n) = rightvec(1_i4b:n,j)
                    sup2(1_i4b:n) = leftvec(1_i4b:n,j)
                else
                    sup1(1_i4b:n) = leftvec(1_i4b:n,j)
                    sup2(1_i4b:n) = rightvec(1_i4b:n,j)
                end if
!
                temp(1_i4b:n) = diag(1_i4b:n)*sup1(1_i4b:n)             +  &
                                sub(2_i4b:n+1_i4b)*sup1(2_i4b:n+1_i4b)
!
!               COMPUTE RAYLEIGH QUOTIENT.
!
#ifdef _BLAS
                xj = dot( n, sup2(1_i4b:n), 1_i4b, temp(1_i4b:n), 1_i4b )
#else
                xj = dot_product( sup2(1_i4b:n), temp(1_i4b:n) )
#endif
!
                if ( xj<zero ) then
                    leftvec(1_i4b:n,j)  = -leftvec(1_i4b:n,j)
                end if
!
                gk(j) = abs( xj )
!
!               COMPUTE THE RESIDUAL NORM OF THE JTH LEFT AND RIGHT SINGULAR VECTORS
!               WITH RESPECT TO THEIR RAYLEIGH QUOTIENT.
!
#ifdef _BLAS
                call axpy( n, -xj, sup2(1_i4b:n), 1_i4b, temp(1_i4b:n), 1_i4b )
#else
                temp(1_i4b:n) = temp(1_i4b:n) - xj*sup2(1_i4b:n)
#endif
                tau = sum( abs( temp(:n) ) )
!
                comp_norm(j) = tau>lim
!
            else
!
                comp_norm(j) = true
!
!$OMP CRITICAL (zerovec)
                zero_vec = true
!$OMP END CRITICAL (zerovec)
!
            end if
!
        end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
!       SET OUTPUT FAILURE PARAMETER.
!
        failure = any( comp_norm(:m) )
!
    else
#endif
!
!       SERIAL PATH.
!
        do j = 1_i4b, m0
!
!            eigval = max( singval(j), safmin )
            eigval = singval(j)
!    
            bd_zero     = false
!
            j1 = i1
            j2 = i2
!
            if ( eigval>lim_sing .or. m==n  ) then
!
                if ( m/=n ) then
!
                    if ( first_zero ) j1 = j1 + 1_i4b
                    if ( last_zero )  j2 = j2 - 1_i4b
!
                end if
!
                bd_zero = diag_zero .and. j==nn2-nn1+1_i4b
!
            else if ( m/=n ) then
!
                j1 = 1_i4b
                j2 = n2
!
            end if
!
            temp(1_i4b:n2) = zero
!
!           COMPUTE LU FACTORS WITH PARTIAL PIVOTING (PT=LU) OF THE TGK MATRIX.
!
            call trid_cmp( temp(j1:j2), gk(j1:j2), eigval, sub(j1:j2), diag(j1:j2),    &
                           sup1(j1:j2), sup2(j1:j2), perm(j1:j2) )
!
!           REPLACE SMALL DIAGONAL ELEMENTS BY safmin.
!
            where( abs(diag(j1:j2))<safmin ) diag(j1:j2) = sign( safmin, diag(j1:j2) )
!
            do l = 1_i4b, maxiter_init
!    
                if ( fernando_init(j) ) then
!    
!                   FERNANDO METHOD.  
!
                    mm = maxdiag_gkinv_ldu( gk(j1:j2-1_i4b), eigval ) + j1 - 1_i4b
!    
!                   BINI METHOD.  
!                
!                   mm = maxdiag_gkinv_qr( gk(j1:j2-1_i4b), eigval ) + j1 - 1_i4b
!    
                    if ( comp_norm(j) ) then
!
!                       ADD A SMALL PERTURBATION TO ALL COMPONENTS OF THE
!                       INITIAL VECTOR TO PROTECT AGAINST FAILURE OF FERNANDO
!                       METHOD WHICH MAY PRODUCE PARALLEL INITIAL VECTORS IF
!                       SINGULAR VALUES ARE TOO CLOSED.
!
                        call random_number_( temp(j1:j2) )
!
                        temp(j1:j2) = sqrt( macheps )*( temp(j1:j2)*two - one )
!    
                    else
!
                        temp(j1:j2) = zero
!    
                    end if
!    
                    temp(mm) = one
!
                else
!
!                   STANDARD INVERSE ITERATION METHOD.
!    
!                   COMPUTE A RANDOM INITIAL VECTOR.  
!
                    call random_number_( temp(j1:j2) )
!
                    temp(j1:j2) = temp(j1:j2)*two - one
!
                end if
!
                if ( bd_zero .and. .not.fernando_init(j) ) then
!
                    do p=1_i4b, 2_i4b
!    
                        if ( p==1_i4b ) then
!    
                            if ( first_zero ) then
!    
                                temp2(j1)          = one
                                temp2(j1+1_i4b:j2) = zero
!    
                            else
!    
                                temp2(j1:j2:2_i4b)       = temp(j1:j2:2_i4b)
                                temp2(j1+1_i4b:j2:2_i4b) = zero
!    
                            end if
!    
                        else
!    
                            if ( last_zero ) then
!    
                                temp2(j2)          = one
                                temp2(j1:j2-1_i4b) = zero
!    
                            else
!    
                                temp2(j1:j2:2_i4b)       = zero
                                temp2(j1+1_i4b:j2:2_i4b) = temp(j1+1_i4b:j2:2_i4b)
!    
                            end if
!    
                        end if
!
!                       DO maxit ITERATIONS FOR THE EIGENVECTORS OF THE TGK MATRIX.
!
                        do i=1_i4b, maxit
!
!                           SOLVE THE SYSTEM LUx = Pb WITH RESCALING TO AVOID OVERFLOW.
!
                            call trid_solve( sub(j1:j2), diag(j1:j2), sup1(j1:j2),  &
                                             sup2(j1:j2), perm(j1:j2), temp2(j1:j2) )
!
                            if ( i/=maxit ) then
!
!                               NORMALIZE THE EIGENVECTORS OF THE TGK MATRIX.
!
                                xj = norm( temp2(j1:j2) )
!
                                if ( xj/=zero ) then
!
                                    temp2(j1:j2) = temp2(j1:j2)/xj
!
                                else
!
                                    call random_number_( temp2(j1:j2) )
!
                                    if ( p==1_i4b ) then
!    
                                        temp2(j1:j2:2_i4b)       = temp2(j1:j2:2_i4b)*two - one
                                        temp2(j1+1_i4b:j2:2_i4b) = zero
!    
                                    else
!    
                                        temp2(j1:j2:2_i4b) = zero
                                        temp2(j1+1_i4b:j2:2_i4b) = temp2(j1+1_i4b:j2:2_i4b)*two - one
!    
                                    end if
!
                                end if
!
                            end if
!
                        end do
!    
                        if ( p==1_i4b ) then
!    
                            temp(j1:j2:2_i4b) = temp2(j1:j2:2_i4b)
!    
                        else
!    
                            temp(j1+1_i4b:j2:2_i4b) = temp2(j1+1_i4b:j2:2_i4b)
!    
                        end if
!
                    end do
!
                else
!
!                   DO maxit ITERATIONS FOR THE EIGENVECTOR OF THE TGK MATRIX.
!
                    do i=1_i4b, maxit
!
!                       SOLVE THE SYSTEM LUx = Pb WITH RESCALING TO AVOID OVERFLOW.
!
                        call trid_solve( sub(j1:j2), diag(j1:j2), sup1(j1:j2),  &
                                         sup2(j1:j2), perm(j1:j2), temp(j1:j2)  )
!
                        if ( i/=maxit ) then
!
!                           NORMALIZE THE EIGENVECTORS OF THE TGK MATRIX.
!
                            xj = norm( temp(j1:j2) )
!
                            if ( xj/=zero ) then
!
                                temp(j1:j2) = temp(j1:j2)/xj
!
                            else
!
                                call random_number_( temp(j1:j2) )
!
                                temp(j1:j2) = temp(j1:j2)*two - one
!
                            end if
!
                        end if
!
                    end do
!
                end if
!
!               NORMALIZE THE SINGULAR VECTORS.
!
                xj = norm( temp(ileft:n2:2_i4b) )
!
                if ( xj/=zero ) then
                    leftvec(1_i4b:n,j) = temp(ileft:n2:2_i4b)/xj
                else
                    leftvec(1_i4b:n,j) = zero
                end if
!
                xjm = norm( temp(iright:n2:2_i4b) )
!
                if ( xjm/=zero ) then
                    rightvec(1_i4b:n,j) = temp(iright:n2:2_i4b)/xjm
                else
                    rightvec(1_i4b:n,j) = zero
                end if
!
                tau = pythag( xj, xjm )
                tmp = max( xj, xjm )
!
                if ( abs( tmp - ortol*tau )*sqrt2>tau .or. tmp<gamma ) then
!
!                   BAD ESTIMATES OF THE SINGULAR VECTORS, TRY TO IMPROVE THEM.
!
                    if ( l/=maxiter_init ) then
!
!                       BAD ESTIMATES OF THE SINGULAR VECTORS, TRY AGAIN
!                       WITH A RANDOM INITIAL VECTOR.
!
                        fernando_init(j) = false        
!
                        cycle
!
                    else if ( .not.test_ortho .and. tmp/=tau .and. j/=1_i4b ) then
!
!                       BAD ESTIMATES OF THE SINGULAR VECTORS, REORTHOGONALIZE THEM
!                       BY MODIFIED GRAM-SCHMIDT.
!
                        comp_norm(j) = true
!
                        do_ortho = true
!
                    end if
!
                end if
!
                exit
!
            end do
!
            lambda(j) = min( xj, xjm )
!
        end do
!
        if ( do_ortho ) then
!
!           REORTHOGONALIZE LEFT AND RIGHT SINGULAR VECTORS BY MODIFIED GRAM-SCHMIDT
!           OR A QR FACTORIZATION.
!
!           FIRST DETERMINE THE NUMBER OF CLUSTERS OF SINGULAR VALUES.
!
            i = count( comp_norm(2_i4b:m0) )
!
            if ( i==m0-1_i4b ) then
                k = 1_i4b
            else
                k = i
            end if
!
!           ALLOCATE WORK VARIABLES.
!
            allocate( jstart(k), jend(k), stat = iok  )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           k IS THE NUMBER OF CLUSTERS OF SINGULAR VALUES.
!           jstart(k) IS THE INDICE OF THE FIRST SINGULAR VALUE IN CLUSTER k.
!           jend(k)+1 IS THE INDICE OF THE LAST SINGULAR VALUE IN CLUSTER k.
!
            if ( i==m0-1_i4b ) then
!
!               ALL THE SINGULAR VALUES ARE CLUSTERED TOGETHER.
!
                jstart(1_i4b) = 1_i4b
                jend(1_i4b)   = m0 - 1_i4b
!
            else
!
!               UNKNOWN NUMBER OF CLUSTERS, BUT AT LEAST ONE
!               CLUSTER OF SINGULAR VALUES.
!
                j = 2_i4b
                k = 0_i4b
!
                do
!
                    if ( comp_norm(j) ) then
!
                        k = k + 1_i4b
!
                        jstart(k) = j - 1_i4b
!
                        do i=j+1_i4b, m0
                            if ( .not.comp_norm(i) ) exit
                        end do
!
                        jend(k) = i - 2_i4b
!
                        j = i + 1_i4b
!
                    else
!
                        j = j + 1_i4b
!
                    end if
!
                    if ( j>m0 ) exit
!
                end do
!
!               IF ONE CLUSTER CONTAINS ALMOST ALL THE SINGULAR VALUES, MERGE
!               ALL THE CLUSTERS TOGETHER AS IN THIS CASE USING ONLY ONE QR
!               FACTORIZATION IS EQUALLY FAST.
!
                xj = real( m0, stnd )*merge_factor
                i  = int( xj, i4b )
!
                merge_clusters = false
!
                do l=1_i4b, k
!
                    j = jend(l) - jstart(l) + 2_i4b            
!
                    if ( j>=i ) then
!
                        merge_clusters = true
!
                        exit
!
                    end if
!
                end do
!
                if ( merge_clusters ) then
!
                    k = 1_i4b
!
                    jstart(1_i4b) = 1_i4b
                    jend(1_i4b)   = m0 - 1_i4b
!
                    comp_norm(2_i4b:m0) = true
!
                end if
!
            end if
!
!           REORTHOGONALIZE LEFT AND RIGHT SINGULAR VECTORS CORRESPONDING TO
!           CLUSTERS OF ALMOST-MULTIPLE SINGULAR VALUES.
!
            do l=1_i4b, k
!
                j1 = jstart(l)
                j2 = jend(l)
!
                mm = j2 + 1_i4b
!
                if ( j2-j1<nsing_cluster ) then
!
!                   REORTHOGONALIZE LEFT SINGULAR VECTORS BY MODIFIED GRAM-SCHMIDT.
!
!                   FORWARD SWEEP OF MGS ORTHOGONALIZATION FOR SMALL CLUSTERS.
!
                    do j=j1, j2
!
                        sup1(:n) = leftvec(:n,j)
!
                        do i=j+1_i4b, mm
!
#ifdef _BLAS
                            xj = -dot( n, sup1(:n), 1_i4b, leftvec(:n,i), 1_i4b )
                            call axpy( n, xj, sup1(:n), 1_i4b, leftvec(:n,i), 1_i4b )
#else
                            xj = -dot_product( sup1(:n), leftvec(:n,i) )
                            leftvec(:n,i) = leftvec(:n,i) + xj*sup1(:n)
#endif
                        end do
!
!                       NORMALIZE THE NEXT j LEFT SINGULAR VECTOR.
!
                        sup1(:n) = leftvec(:n,j+1_i4b)
!
                        xj = norm( sup1(1_i4b:n) )
!
                        if ( xj<=fudge_ortho2 ) then
!
!                           REORTHOGONALIZATION IS REQUIRED.
!
                            call ortho_vec(leftvec(:n,j1:j), sup1(:n), xj )
!
                        else
!
!                           NORMALIZATION IS ENOUGH.
!
                            sup1(:n) = (one/xj)*sup1(:n)
!
                        end if
!
                        leftvec(:n,j+1_i4b) = sup1(:n)
!
                        lambda(j+1_i4b) = xj
!
                    end do
!
                    if ( do_backward_sweep ) then
!
!                       BACKWARD SWEEP OF MGS ORTHOGONALIZATION FOR SMALL CLUSTERS.
!
                        do j=j2, j1, -1_i4b
!
                            sup1(:n) = leftvec(:n,j)
!
                            do i=j+1_i4b, mm
!
#ifdef _BLAS
                                xj = -dot( n, sup1(:n), 1_i4b, leftvec(:n,i), 1_i4b )
                                call axpy( n, xj, sup1(:n), 1_i4b, leftvec(:n,i), 1_i4b )
#else
                                xj = -dot_product( sup1(:n), leftvec(:n,i) )
                                leftvec(:n,i) = leftvec(:n,i) + xj*sup1(:n)
#endif
!
                            end do
!
                        end do
!
                    end if
!
!                   REORTHOGONALIZE RIGHT SINGULAR VECTORS BY MODIFIED GRAM-SCHMIDT.
!
!                   FORWARD SWEEP OF MGS ORTHOGONALIZATION FOR SMALL CLUSTERS.
!
                    do j=j1, j2
!
                        sup1(:n) = rightvec(:n,j)
!
                        do i=j+1_i4b, mm
!
#ifdef _BLAS
                            xj = -dot( n, sup1(:n), 1_i4b, rightvec(:n,i), 1_i4b )
                            call axpy( n, xj, sup1(:n), 1_i4b, rightvec(:n,i), 1_i4b )
#else
                            xj = -dot_product( sup1(:n), rightvec(:n,i) )
                            rightvec(:n,i) = rightvec(:n,i) + xj*sup1(:n)
#endif
                        end do
!
!                       NORMALIZE THE NEXT j RIGHT SINGULAR VECTOR.
!
                        sup1(:n) = rightvec(:n,j+1_i4b)
!
                        xj = norm( sup1(1_i4b:n) )
!
                        if ( xj<=fudge_ortho2 ) then
!
!                           REORTHOGONALIZATION IS REQUIRED.
!
                            call ortho_vec(rightvec(:n,j1:j), sup1(:n), xj )
!
                        else
!
!                           NORMALIZATION IS ENOUGH.
!
                            sup1(:n) = (one/xj)*sup1(:n)
!
                        end if
!
                        rightvec(:n,j+1_i4b) = sup1(:n)
!
                        lambda(j+1_i4b) = min( xj,lambda(j+1_i4b) )
!
                    end do
!
                    if ( do_backward_sweep ) then
!
!                       BACKWARD SWEEP OF MGS ORTHOGONALIZATION FOR SMALL CLUSTERS.
!
                        do j=j2, j1, -1_i4b
!
                            sup1(:n) = rightvec(:n,j)
!
                            do i=j+1_i4b, mm
!
#ifdef _BLAS
                                xj = -dot( n, sup1(:n), 1_i4b, rightvec(:n,i), 1_i4b )
                                call axpy( n, xj, sup1(:n), 1_i4b, rightvec(:n,i), 1_i4b )
#else
                                xj = -dot_product( sup1(:n), rightvec(:n,i) )
                                rightvec(:n,i) = rightvec(:n,i) + xj*sup1(:n)
#endif
                            end do
!
                        end do
!
                    end if
!
                else
!
!                   QR FACTORIZATION FOR BIG CLUSTERS.
!
                    j = mm - j1 + 1_i4b
!
!                   REORTHOGONALIZE LEFT SINGULAR VECTORS.
!
                    call qr_cmp( leftvec(:n,j1:mm), diag(:j), sub(:j) )
!
                    call ortho_gen_qr( leftvec(:n,j1:mm), sub(:j) )
!
!                   REORTHOGONALIZE RIGHT SINGULAR VECTORS.
!
                    call qr_cmp( rightvec(:n,j1:mm), diag(:j), sub(:j) )
!
                    call ortho_gen_qr( rightvec(:n,j1:mm), sub(:j) )
!
                end if
!
            end do
!
        end if
!
!       CHECK ACCURACY OF THE SINGULAR VECTORS.
!
        if ( all( lambda(:m)/=zero ) ) then
!
            diag(:n)      = d(:n)
            sub(:n)       = e(:n)
            sub(n+1_i4b)  = zero
            sup1(n+1_i4b) = zero
!
            do j = 1_i4b, m
!
                if ( upper ) then
                    sup1(1_i4b:n) = rightvec(1_i4b:n,j)
                    sup2(1_i4b:n) = leftvec(1_i4b:n,j)
                else
                    sup1(1_i4b:n) = leftvec(1_i4b:n,j)
                    sup2(1_i4b:n) = rightvec(1_i4b:n,j)
                end if
!
                temp(1_i4b:n) = diag(1_i4b:n)*sup1(1_i4b:n)             +  &
                                sub(2_i4b:n+1_i4b)*sup1(2_i4b:n+1_i4b)
!
!               COMPUTE RAYLEIGH QUOTIENT.
!
#ifdef _BLAS
                xj = dot( n, sup2(1_i4b:n), 1_i4b, temp(1_i4b:n), 1_i4b )
#else
                xj = dot_product( sup2(1_i4b:n), temp(1_i4b:n) )
#endif
!
                if ( xj<zero ) then
                    leftvec(1_i4b:n,j)  = -leftvec(1_i4b:n,j)
                end if
!
                gk(j) = abs( xj )
!
                if ( .not.failure ) then
!
!                   SET OUTPUT FAILURE PARAMETER.
!
!                   COMPUTE THE RESIDUAL NORM OF THE JTH LEFT AND RIGHT SINGULAR VECTORS
!                   WITH RESPECT TO THEIR RAYLEIGH QUOTIENT.
!
#ifdef _BLAS
                    call axpy( n, -xj, sup2(1_i4b:n), 1_i4b, temp(1_i4b:n), 1_i4b )
#else
                    temp(1_i4b:n) = temp(1_i4b:n) - xj*sup2(1_i4b:n)
#endif
                    tau = sum( abs( temp(:n) ) )
!
                    failure = tau>lim
!
                end if
!
            end do
!
        else
!
            failure  = true
            zero_vec = true
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
    if ( .not.zero_vec ) then
!
!       CHECK THE PAIRING OF SINGULAR VALUES AND VECTORS.
!
        if ( do_ortho ) then
!
            do l=1_i4b, k
!
                j1 = jstart(l)
                j2 = jend(l)
!
                do j=j1, j2
!
                    if ( gk(j)<gk(j+1_i4b) ) exit
!
                end do
!
                if ( j<=j2 ) then
!
                    call lr_singvec_sort( 'd', gk(j1:j2+1_i4b), leftvec(:,j1:j2+1_i4b), rightvec(:,j1:j2+1_i4b) )
!
                end if
!
                if ( l==1_i4b ) cycle
!
                if ( .not.failure ) then
!
                    j2 = jend(l-1_i4b) + 1_i4b
!
                    failure = gk(j2)<gk(j1)
!
                end if
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( jstart, jend )
!
        else if ( .not.failure ) then
!
            do i=1_i4b, m0-1_i4b
!
                if ( gk(i)<gk(i+1_i4b) ) exit
!
            end do
!
!            failure = i<m0
!
            if ( i<m0 ) then
!
                call lr_singvec_sort( 'd', gk(:m0), leftvec(:,:m0), rightvec(:,:m0) )
!
            end if
!
        end if
!
!       CHECK FINALLY IF THE RAYLEIGH QUOTIENTS AND THE SINGULAR VALUES AGREE.
!
        if ( .not.failure ) then
!
            xj  = maxval( abs( gk(1_i4b:m) - s(1_i4b:m) ) )
            failure = xj>lim
!
        end if
!
    else if ( do_ortho ) then
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( jstart, jend )
!
    end if
!
!
! END OF SUBROUTINE bd_inviter_rm
! _______________________________
!
    end subroutine bd_inviter_rm
!
! =========================================================================================
!
    subroutine bd_inviter2_a( mat, tauq, taup, d, e, s, leftvec, rightvec, failure,  &
                              maxiter, ortho, backward_sweep, scaling, initvec )
!
! Purpose
! _______
!
!   BD_INVITER2 computes the left and right singular vectors of a full real m-by-n
!   matrix MAT corresponding to specified singular values, using inverse iteration.
!
!   It is required that the original matrix MAT has been reduced to upper or lower
!   bidiagonal form BD by an orthogonal transformation:
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal. This can be done with a call to BD_CMP with parameters
!   TAUQ and TAUP, before calling BD_SVD, BD_SINGVAL or BD_SINGVAL2 subroutines for computing
!   singular values and, finally, BD_INVITER2 for computing selected singular vectors.
!
!   Alternatively and more simply, the user can call SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2
!   subroutines, with parameters TAUQ and TAUP present, to compute the bidiagonal reduction and
!   singular values in one step and, finally, call BD_INVITER2 for computing all or selected
!   singular vectors of MAT.
!
!   If m >= n, BD is upper bidiagonal and if m < n, BD is lower bidiagonal.
!
!   
! Arguments
! _________
!
!   MAT            (INPUT) real(stnd), dimension(:,:)
!                  On entry, the original m-by-n matrix after reduction by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 subroutines.
!                  MAT must contains the vectors which define the elementary reflectors
!                  H(i) and G(i) whose products determine the matrices Q and P,
!                  as returned by BD_CMP, SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2
!                  subroutines when the arguments TAUQ and TAUP are present in the call
!                  to these subroutines. MAT must be specified as returned by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 and is not modified by
!                  the routine.
!                
!   TAUQ           (INPUT) real(stnd), dimension(:)
!                  TAUQ(i) must contain the scalar factor of the elementary
!                  reflector H(i) which determines Q, as returned by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 in their array
!                  argument TAUQ.
!
!                  The size of TAUQ must verify:
!                  size( TAUQ ) = min( size(MAT,1) , size(MAT,2) ) .
!
!   TAUP           (INPUT) real(stnd), dimension(:)
!                  TAUP(i) must contain the scalar factor of the elementary
!                  reflector G(i), which determines P, as returned by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 in their array
!                  argument TAUP.
!
!                  The size of TAUP must verify:
!                  size( TAUP ) = min( size(MAT,1) , size(MAT,2) ) .
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, D contains the diagonal elements of the bidiagonal
!                  matrix BD as returned by BD_CMP, SELECT_SINGVAL_CMP or
!                  SELECT_SINGVAL_CMP2.
!
!                  The size of D must verify: 
!                  size( D ) = min( size(MAT,1) , size(MAT,2) ) . 
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, E contains the off-diagonal elements of the bidiagonal
!                  matrix BD as returned by BD_CMP, SELECT_SINGVAL_CMP or
!                  SELECT_SINGVAL_CMP2:
!
!                  - if m >= n, E(i) = BD(i-1,i) for i = 2,3,...,n;
!                  - if m < n,  E(i) = BD(i,i-1) for i = 2,3,...,m.
!
!                  E(1) is arbitrary.
!
!                  The size of E must verify: 
!                  size( E ) = min( size(MAT,1) , size(MAT,2) ) . 
!
!   S              (INPUT) real(stnd), dimension(:)
!                  On entry, selected singular values of the bidiagonal matrix BD.
!                  The singular values must be given in decreasing order and are
!                  assumed to be positive or zero.
!
!                  The size of S must verify:
!                  size( S ) <= min( size(MAT,1) , size(MAT,2) ) .
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed left singular vectors. The left singular vector
!                  associated with the singular value S(j) is stored in the j-th column
!                  of LEFTVEC.
!
!                  The shape of LEFTVEC must verify: 
!
!                  - size( LEFTVEC, 1 ) = size( MAT, 1 ) = m ,
!                  - size( LEFTVEC, 2 ) = size( S ) .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed right singular vectors. The right singular vector
!                  associated with the singular value S(j) is stored in the j-th column of
!                  RIGHTVEC.
!
!                  The shape of RIGHTVEC must verify: 
!
!                  - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ,
!                  - size( RIGHTVEC, 2 ) = size( S ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = FALSE : indicates successful exit,
!                  - FAILURE = TRUE : indicates that some singular vectors of BD
!                    failed to converge in MAXITER iterations.
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed in the subroutine.
!
!                  By default, 2 inverse iterations are performed for all the singular vectors.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, all the singular vectors of the bidiagonal matrix BD are
!                    orthogonalized by the  Modified Gram-Schmidt or QR algorithm;
!                  - ORTHO=false, the singular vectors of the bidiagonal matrix BD are not
!                    orthogonalized by the Modified Gram-Schmidt or QR algorithm.
!
!                  The default is to orthogonalize the singular vectors only for the singular values,
!                  which are not well-separated.
!
!   BACKWARD_SWEEP (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - BACKWARD_SWEEP=true and the singular vectors of the bidiagonal matrix BD
!                    are orthogonalized by the modified Gram-Schmidt algorithm, a backward sweep of the 
!                    modified Gram-Schmidt algorithm is also performed;
!                  - BACKWARD_SWEEP=false, a backward sweep is not performed.
!
!                  The default is not to perform a backward sweep of the modified Gram-Schmidt algorithm.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the bidiagonal matrix BD is scaled before computing
!                    the singular vectors;
!                  - SCALING=false, the bidiagonal matrix BD is not scaled.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INITVEC=true, Fernando vectors are used to start the inverse iteration process
!                    for computing the singular vectors of the bidiagonal matrix BD (e.g. the eigenvectors of the
!                    associated Golub-Kahan tridiagonal matrix);
!                  - INITVEC=false, random uniform starting vectors are used.
!
!                  The default is to use Fernando starting vectors if the singular values are well-separated and
!                  the Golub-Kahan form of the input bidiagonal matrix is unreduced, and random uniform
!                  starting vectors otherwise.
!
!
! Further Details
! _______________
!
!   A first estimate of the singular vectors is computed by the Fernando method applied to the
!   tridiagonal Golub-Kahan matrix associated with the bidiagonal matrix BD (see the reference
!   (1) for details) for the singular values which are well-separated and if the Golub-Kahan
!   form of the input bidiagonal matrix is unreduced. For the other singular values,
!   a random start is used as a first estimate of the singular vectors as in the
!   standard inverse-iteration algorithm.
!
!   The singular vectors of BD are then computed or refined using inverse iteration on the
!   tridiagonal Golub-Kahan matrix for all the singular values at one step.
!
!   By default, the singular vectors of BD are then orthogonalized by the Modified Gram-Schmidt
!   or QR algorithm only if the singular values are not well-separated.
!
!   The singular vectors of MAT are finally computed by a blocked back-transformation algorithm.
!
!   The computation of the singular vectors of BD and the blocked back-transformation algorithm
!   to find the singular vectors of MAT are parallelized if OPENMP is used.
!
!   BD_INVITER2 may fail if some singular values specified in parameter S are nearly
!   identical for some pathological matrices.
!
!   For further details on Fernando method for computing eigenvectors of tridiagonal
!   matrices, the blocked back-transformation algorithm or inverse iteration, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:,:)   :: mat
    real(stnd),  intent(in),  dimension(:)     :: tauq, taup, d, e, s
    real(stnd),  intent(out), dimension(:,:)   :: leftvec, rightvec
!
    integer(i4b), intent(in),  optional :: maxiter
!
    logical(lgl), intent(in),  optional :: ortho, backward_sweep, scaling, initvec
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(:,:), allocatable :: work, work2
!
    integer(i4b) :: n, m, nm, nsing, n1, n2, mnthr_nsing
    integer      :: iok
!
    logical(lgl) :: upper
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_inviter2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size( mat,     1 ),i4b)  ,   &
                   int(size( leftvec, 1 ),i4b)  ,   &
                   name_proc )
!
    n = assert_eq( int(size( mat,      2 ),i4b)  ,  &
                   int(size( rightvec, 1 ),i4b)  ,  &
                   name_proc )
!
    nm = assert_eq( (/ min(m,n)            ,   &
                       int(size(tauq),i4b) ,   &
                       int(size(taup),i4b) ,   &
                       int(size(d),i4b)    ,   &
                       int(size(e),i4b) /) ,   &
                       name_proc )
!
    nsing = assert_eq( int(size(s),i4b)             ,  &
                       int(size( leftvec, 2 ),i4b)  ,  &
                       int(size( rightvec, 2 ),i4b) ,  &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( nm<=0_i4b ) then
!
        return
!
    end if
!
    upper = m >= n
!
!   COMPUTE SINGULAR VECTORS OF THE BIDIAGONAL FORM BD OF mat.
!
    call bd_inviter( upper, d, e, s, leftvec(:nm,:nsing), rightvec(:nm,:nsing),               &
                     failure, maxiter=maxiter, ortho=ortho, backward_sweep=backward_sweep,    &
                     scaling=scaling, initvec=initvec  )
!
!   COMPUTE SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY BACK-TRANSFORMATION.
!
    mnthr_nsing = int( real( nm, stnd )*thr_bt, i4b )
!
    if ( nsing>=mnthr_nsing  ) then
!
!       ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM IF NEARLY FULL SVD OF mat IS
!       REQUESTED.
!
        if ( upper ) then
            n1 = m
            n2 = nm
        else
            n1 = n
            n2 = nm + 1_i4b
        end if
!
        allocate( work(n1,n2), work2(nm,nsing), stat = iok )
!
    else
!
!       USE DIRECT ALGORITHM OTHERWISE OR IF THE REQUIRED
!       WORKSPACE IS NOT AVAILABLE.
!
        iok = -1
!
    end if
!
    if ( iok==0 ) then
!
        if ( upper ) then
!
!           GENERATE P IN work(:n,:n) .
!
            call ortho_gen_p_bd( mat, taup, work(:n,:n) )
!
!           MULTIPLY P IN work(:n,:n) BY RIGHT SINGULAR VECTORS OF BD IN rightvec
!           FOR COMPUTING RIGHT SINGULAR VECTORS OF mat .
!
            work2(:n,:nsing) = rightvec(:n,:nsing)
!
#ifdef _BLAS
            call gemm( 'N', 'N', n, nsing, n, one, work, m, work2, n, zero, rightvec, n  )
#else
            rightvec(:n,:nsing) = matmul( work(:n,:n), work2(:n,:nsing) )
#endif
!
!           GENERATE Q IN work .
!
            work(:m,:n) = mat(:m,:n)
!
            call ortho_gen_q_bd( work, tauq )
!
!           MULTIPLY Q IN work BY LEFT SINGULAR VECTORS OF BD IN leftvec
!           FOR COMPUTING LEFT SINGULAR VECTORS OF mat .
!
            work2(:n,:nsing) = leftvec(:n,:nsing)
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, nsing, n, one, work, m, work2, n, zero, leftvec, m   )
#else
            leftvec(:m,:nsing) = matmul( work(:m,:n), work2(:n,:nsing) )
#endif
!
        else
!
!           GENERATE Q IN work .
!
            work(:m,:m+1_i4b) = mat(:m,:m+1_i4b)
!
            call ortho_gen_q_bd( work(:m,:m+1_i4b), tauq )
!
!           MULTIPLY Q IN work BY LEFT SINGULAR VECTORS OF BD IN leftvec
!           FOR COMPUTING LEFT SINGULAR VECTORS OF mat.
!
            work2(:m,:nsing) = leftvec(:m,:nsing)
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, nsing, m, one, work, n, work2, m, zero, leftvec, m   )
#else
            leftvec(:m,:nsing) = matmul( work(:m,:m), work2(:m,:nsing) )
#endif
!
!           GENERATE P IN work .
!
            call ortho_gen_p_bd( mat, taup, work(:n,:m) )
!
!           MULTIPLY P IN work BY RIGHT SINGULAR VECTORS OF BD IN rightvec
!           FOR COMPUTING RIGHT SINGULAR VECTORS OF mat.
!
            work2(:m,:nsing) = rightvec(:m,:nsing)
!
#ifdef _BLAS
            call gemm( 'N', 'N', n, nsing, m, one, work, n, work2, m, zero, rightvec, n   )
#else
            rightvec(:n,:nsing) = matmul( work(:n,:m), work2(:m,:nsing) )
#endif
!
        end if
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( work, work2 )
!
    else
!
        if ( upper ) then
            leftvec(nm+1_i4b:m,:nsing) = zero
        else
            rightvec(nm+1_i4b:n,:nsing) = zero
        end if
!
!       GENERATE LEFT SINGULAR VECTORS OF mat IN leftvec .
!
        call apply_q_bd( mat, tauq, leftvec,  left=true, trans=false )
!
!       GENERATE RIGHT SINGULAR VECTORS OF mat IN rightvec .
!
        call apply_p_bd( mat, taup, rightvec, left=true, trans=false )
!
    end if
!
!
! END OF SUBROUTINE bd_inviter2_a
! _______________________________
!
    end subroutine bd_inviter2_a
!
! =========================================================================================
!
    subroutine bd_inviter2_b( mat, p, d, e, s, leftvec, rightvec, failure,         &
                              maxiter, ortho, backward_sweep, scaling, initvec,    &
                              tol_reortho )
!
! Purpose
! _______
!
!   BD_INVITER2 computes the left and right singular vectors of a full real m-by-n
!   matrix MAT with m>=n corresponding to specified singular values, using inverse iteration.
!
!   It is required that the original matrix MAT has been reduced to upper bidiagonal form BD
!   by an orthogonal transformation:
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal. This can be done with a call to BD_CMP2 (or a call to BD_CMP
!   followed by a call to ORTHO_GEN_BD), before calling BD_SVD, BD_SINGVAL or BD_SINGVAL2 subroutines
!   for computing singular values and BD_INVITER2 for computing selected singular vectors.
!
!   Alternatively and more simply, the user can call SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4
!   subroutines, with parameter P present, to compute the bidiagonal reduction and
!   singular values in one step and, finally, call BD_INVITER2 for computing all or selected
!   singular vectors of MAT.
!
!   
! Arguments
! _________
!
!   MAT            (INPUT) real(stnd), dimension(:,:)
!                  On entry, the m-by-n orthogonal matrix Q after reduction by BD_CMP2,
!                  SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4 (or by BD_CMP and ORTHO_GEN_BD).
!                  MAT is not modified by the routine.
!                
!                  The shape of MAT must verify: size( MAT, 1 ) >= size( MAT, 2 ) = n .
!
!   P              (INPUT) real(stnd), dimension(:,:)
!                  On entry, the n-by-n orthogonal matrix P after reduction by BD_CMP2,
!                  SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4 (or by BD_CMP and ORTHO_GEN_BD).        
!                  If P has been computed by BD_CMP2, SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4,
!                  P can be stored in factored form or not. Both cases are handled by the subroutine.
!                  P is not modified by the routine.
!                
!                  The shape of P must verify: size( P, 1 ) = size( P, 2 ) = size( MAT, 2 ) = n .
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, D contains the diagonal elements of the bidiagonal
!                  matrix BD as returned by BD_CMP, BD_CMP2, SELECT_SINGVAL_CMP3
!                  or SELECT_SINGVAL4 subroutines.
!
!                  The size of D must verify: size( D ) = size( MAT, 2 ) = n . 
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, E contains the off-diagonal elements of the bidiagonal
!                  matrix BD as returned by BD_CMP, BD_CMP2, SELECT_SINGVAL_CMP3
!                  or SELECT_SINGVAL4 subroutines:
!
!                          E(i) = BD(i-1,i) for i = 2,3,...,n;
!
!                  E(1) is arbitrary.
!
!                  The size of E must verify: size( E ) = size( MAT, 2 ) = n . 
!
!   S              (INPUT) real(stnd), dimension(:)
!                  On entry, selected singular values of the bidiagonal matrix BD.
!                  The singular values must be given in decreasing order and are
!                  assumed to be positive or zero.
!
!                  The size of S must verify: size( S ) <= size( MAT, 2 ) = n .
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed left singular vectors. The left singular vector
!                  associated with the singular value S(j) is stored in the j-th column
!                  of LEFTVEC.
!
!                  The shape of LEFTVEC must verify:
!
!                  - size( LEFTVEC, 1 ) = size( MAT, 1 ) = m ,
!                  - size( LEFTVEC, 2 ) = size( S ) .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed right singular vectors. The right singular vector
!                  associated with the singular value S(j) is stored in the j-th column of
!                  RIGHTVEC.
!
!                  The shape of RIGHTVEC must verify:  
!
!                  - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ,
!                  - size( RIGHTVEC, 2 ) = size( S ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = FALSE : indicates successful exit,
!                  - FAILURE = TRUE : indicates that some singular vectors of BD
!                    failed to converge in MAXITER iterations.
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed in the subroutine.
!
!                  By default, 2 inverse iterations are performed for all the singular vectors.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, all the singular vectors of the bidiagonal matrix BD are
!                    orthogonalized by the  Modified Gram-Schmidt or QR algorithm;
!                  - ORTHO=false, the singular vectors of the bidiagonal matrix BD are
!                    not orthogonalized by the Modified Gram-Schmidt or QR algorithm.
!
!                  The default is to orthogonalize the singular vectors only for the singular values,
!                  which are not well-separated.
!
!   BACKWARD_SWEEP (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - BACKWARD_SWEEP=true and the singular vectors of the bidiagonal matrix BD
!                    are orthogonalized by the modified Gram-Schmidt algorithm, a backward sweep of the 
!                    modified Gram-Schmidt algorithm is also performed;
!                  - BACKWARD_SWEEP=false, a backward sweep is not performed.
!
!                  The default is not to perform a backward sweep of the modified Gram-Schmidt algorithm.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the bidiagonal matrix BD is scaled before computing
!                    the singular vectors;
!                  - SCALING=false, the bidiagonal matrix BD is not scaled.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INITVEC=true, Fernando vectors are used to start the inverse iteration process
!                    for computing the singular vectors of the bidiagonal matrix BD (e.g. the eigenvectors of the
!                    associated Golub-Kahan tridiagonal matrix);
!                  - INITVEC=false, random uniform starting vectors are used.
!
!                  The default is to use Fernando starting vectors if the singular values are well-separated and
!                  the Golub-Kahan form of the input bidiagonal matrix is unreduced, and random uniform
!                  starting vectors otherwise.
!
!   TOL_REORTHO    (INPUT, OPTIONAL) real(stnd)
!                  On entry, TOL_REORTHO is used to determine if the left singular vectors stored in LEFTVEC
!                  must be reortogonalized on exit in order to correct for the loss of orthogonality in the Ralha-Barlow
!                  one-sided bidiagonal reduction algorithm if MAT is nearly deficient. If one of the singular values,
!                  S(i), verifies the condition
!
!                                              S(i) <= TOL_REORTHO * S(1)
!
!                  all the computed left singular vectors are reorthogonalized with a QR factorization. If S(1) is the
!                  largest singular value of MAT, this condition leads to the assertion that the rank of MAT is less
!                  than size(S) and is thus a nearly singular matrix if TOL_REORTHO is a small positive value of the
!                  order of the machine epsilon.
!
!                  TOL_REORTHO must be greater or equal to zero and less than or equal to one. If TOL_REORTHO = 0. is
!                  used, the left singular vectors are reorthogonalized only if some singular values are almost zero.
!                  On the other hand, If TOL_REORTHO = 1. is used, the left singular vectors are always reorthogonalized.
!                  If TOL_REORTHO is specified as less than zero or greater than one, the default value is used.
!
!                  The default value is the value of the module parameter tol_reortho_def if size( S ) = n and
!                  tol_reortho_partial_def otherwise.
!
! Further Details
! _______________
!
!   A first estimate of the singular vectors is computed by the Fernando method applied to the
!   tridiagonal Golub-Kahan matrix associated with the bidiagonal matrix BD (see the reference
!   (1) for details) for the singular values which are well-separated and if the Golub-Kahan
!   form of the input bidiagonal matrix is unreduced. For the other singular values,
!   a random start is used as a first estimate of the singular vectors as in the
!   standard inverse-iteration algorithm.
!
!   The singular vectors of BD are then computed or refined using inverse iteration on the
!   tridiagonal Golub-Kahan matrix for all the singular values at one step.
!
!   By default, the singular vectors of BD are then orthogonalized by the Modified Gram-Schmidt
!   or QR algorithm only if the singular values are not well-separated.
!
!   The singular vectors of MAT are finally computed by a blocked back-transformation algorithm.
!
!   The computation of the singular vectors of BD and the blocked back-transformation algorithm
!   to find the singular vectors of MAT are parallelized if OPENMP is used.
!
!   BD_INVITER2 may fail if some singular values specified in parameter S are nearly
!   identical for some pathological matrices.
!
!   For further details on Fernando method for computing eigenvectors of tridiagonal
!   matrices, the blocked back-transformation algorithm or inverse iteration, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error5, svd_error8, allocate_error
    use Utilities,       only : assert_eq, merror, ifirstloc
    use QR_Procedures,   only : qr_cmp, ortho_gen_qr, apply_q_qr
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:,:)   :: mat, p
    real(stnd),  intent(in),  dimension(:)     :: d, e, s
    real(stnd),  intent(out), dimension(:,:)   :: leftvec, rightvec
    real(stnd),  intent(in),  optional         :: tol_reortho
!
    integer(i4b), intent(in),  optional :: maxiter
!
    logical(lgl), intent(in),  optional :: ortho, backward_sweep, scaling, initvec
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: tol, tmp
    real(stnd), dimension(:),   allocatable :: diag, beta
    real(stnd), dimension(:,:), allocatable :: leftvec0, rightvec0
!
    integer(i4b) :: i, n, m, nsing, nrank
    integer      :: iok
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_inviter2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size( mat,     1 ),i4b)  ,   &
                   int(size( leftvec, 1 ),i4b)  ,   &
                   name_proc )
!
    n = assert_eq( (/ int(size( mat,      2 ),i4b)    ,  &
                      int(size( rightvec, 1 ),i4b)    ,  &
                      int(size( p,        1 ),i4b)    ,  &
                      int(size( p,        2 ),i4b)    ,  &
                      int(size( d )          ,i4b)    ,  &
                      int(size( e )          ,i4b) /) ,  &
                      name_proc )
!
    if ( m<n ) then
!
        call merror( name_proc//svd_error5 )
!
    end if
!
    nsing = assert_eq( int(size(s),i4b)             ,  &
                       int(size( leftvec, 2 ),i4b)  ,  &
                       int(size( rightvec, 2 ),i4b) ,  &
                       name_proc )
!
    if ( nsing==n ) then
        tol = tol_reortho_def
    else
        tol = tol_reortho_partial_def
    end if
!
    if ( present(tol_reortho) ) then
!
        if ( tol_reortho>=zero .and. tol_reortho<=one ) then
!
            tol = tol_reortho
!
        end if
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
!
        return
!
    end if
!
    if ( p(1_i4b,1_i4b)==one ) then
!
!       THE ORTHOGONAL MATRIX P HAS BEEN GENERATED BY bd_cmp2 OR ortho_gen_bd.
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( leftvec0(n,nsing), rightvec0(n,nsing), stat=iok   )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE SINGULAR VECTORS OF THE BIDIAGONAL FORM BD OF mat.
!
        call bd_inviter( true, d, e, s, leftvec0(:n,:nsing), rightvec0(:n,:nsing),                &
                         failure, maxiter=maxiter, ortho=ortho, backward_sweep=backward_sweep,    &
                         scaling=scaling, initvec=initvec  )
!
!       COMPUTE SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY MULTIPLICATION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsing, n, one, mat, m, leftvec0, n, zero, leftvec, m )
!
        call gemm( 'N', 'N', n, nsing, n, one, p, n, rightvec0, n, zero, rightvec, n )
#else
        leftvec(:m,:nsing) = matmul( mat(:m,:n), leftvec0(:n,:nsing) )
!
        rightvec(:n,:nsing) = matmul( p(:n,:n), rightvec0(:n,:nsing) )
#endif
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( leftvec0, rightvec0 )
!
    else if ( p(1_i4b,1_i4b)==-one ) then
!
!       THE ORTHOGONAL MATRIX P IS STORED IN FACTORED FORM BY bd_cmp2.
!
!       ALLOCATE WORK ARRAY.
!
        allocate( leftvec0(n,nsing), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE SINGULAR VECTORS OF THE BIDIAGONAL FORM BD OF mat.
!
        call bd_inviter( true, d, e, s, leftvec0(:n,:nsing), rightvec(:n,:nsing),                &
                         failure, maxiter=maxiter, ortho=ortho, backward_sweep=backward_sweep,   &
                         scaling=scaling, initvec=initvec  )
!
!       COMPUTE LEFT SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY MULTIPLICATION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsing, n, one, mat, m, leftvec0, n, zero, leftvec, m )
#else
        leftvec(:m,:nsing) = matmul( mat(:m,:n), leftvec0(:n,:nsing) )
#endif
!
!       COMPUTE RIGHT SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY BACK TRANSFORMATION.
!
        call apply_q_qr( p(2_i4b:n,2_i4b:n), p(2_i4b:n,1_i4b), rightvec(2_i4b:n,:nsing),   &
                         left=true, trans=false  )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( leftvec0 )
!
    else
!
        call merror( name_proc//svd_error8 )
!
    end if
!
!   DETERMINE THE EFFECTIVE RANK OF mat USING THE TOLERANCE tol .
!
    tmp = max( tol*s(1_i4b), safmin )
!
    nrank = ifirstloc( logical( s(:nsing)<=tmp, lgl ) ) - 1_i4b
!
    if ( nrank<nsing ) then
!
!       REORTOGONALIZE LEFT SINGULAR VECTORS IF REQUIRED.
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( diag(nsing), beta(nsing), stat=iok   )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        call qr_cmp( leftvec(:m,:nsing), diag(:nsing), beta(:nsing) )
!
        call ortho_gen_qr( leftvec(:m,:nsing), beta(:nsing) )
!
        do i = 1_i4b, nsing
!
            if ( diag(i)>=zero ) cycle
!
            leftvec(:m,i) = -leftvec(:m,i)
!        
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( diag, beta )
!
    end if
!
!
! END OF SUBROUTINE bd_inviter2_b
! _______________________________
!
    end subroutine bd_inviter2_b
!
! =========================================================================================
!
    subroutine bd_inviter2_c( mat, tauq, taup, rlmat, d, e, s, leftvec, rightvec, failure, &
                              tauo, maxiter, ortho, backward_sweep, scaling, initvec )
!
! Purpose
! _______
!
!   BD_INVITER2 computes the left and right singular vectors of a full real m-by-n
!   matrix MAT corresponding to specified singular values, using inverse iteration.
!
!   It is required that the original matrix MAT has been reduced to upper bidiagonal
!   form BD by a two-step algorithm as performed by BD_CMP subroutine with parameters
!   TAUQ, TAUP, RLMAT, and eventually TAUO, or more simply by SELECT_SINGVAL_CMP or
!   SELECT_SINGVAL_CMP2 subroutines with the same arguments:
!
!   - If m >= n, a QR factorization of the real m-by-n matrix MAT is first computed
!
!                        MAT = O * R
!
!   where O is orthogonal and R is upper triangular. In a second step, the n-by-n upper
!   triangular matrix R is reduced to upper bidiagonal form BD
!   by an orthogonal transformation:
!
!                       Q' * R * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix.
!
!   - If m < n, an LQ factorization of the real m-by-n matrix MAT is first computed
!
!                       MAT = L * O
!
!   where O is orthogonal and L is lower triangular. In a second step, the m-by-m lower
!   triangular matrix L is reduced to upper bidiagonal form BD
!   by an orthogonal transformation:
!
!                       Q' * L * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix.
!
!   After this call to BD_CMP with parameters TAUQ, TAUP, RLMAT, and eventually TAUO,
!   the user can call BD_SVD, BD_SINGVAL or BD_SINGVAL2 subroutines for computing singular values
!   of BD and, finally, BD_INVITER2 for computing all or selected singular vectors of MAT.
!
!   Alternatively and more simply, the user can call SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2
!   subroutines, with  parameters TAUQ, TAUP, RLMAT, and eventually TAUO, to perform the two-stage
!   bidiagonal reduction and get singular values in one step and, finally, call BD_INVITER2 for
!   computing all or selected singular vectors of MAT.
!   
!   
! Arguments
! _________
!
!   MAT            (INPUT) real(stnd), dimension(:,:)
!                  On entry, the original m-by-n matrix after a two-stage bidiagonal reduction
!                  by BD_CMP, SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 subroutines (e.g.
!                  when the matrix argument RLMAT is present in the call to these subroutines).
!                  The precise content of MAT is determined by the presence or absence of the
!                  optional argument TAUO:
!
!                  - if the optional argument TAUO is absent, it is assumed that the
!                    first n columns (if m>=n) or the first m rows (if m<n) of the
!                    orthogonal matrix O is stored explicitly in the argument MAT on entry;
!                  - if the optional argument TAUO is present, it is assumed that the
!                    orthogonal matrix O is stored in factored form, as a product of
!                    elementary reflectors, in the argument MAT on entry. In this case,
!                    MAT must contains the vectors which define the elementary reflectors
!                    W(i) whose products determine the matrix O, as returned by BD_CMP,
!                    SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 subroutines when the matrix
!                    argument RLMAT is present in the call to these subroutines.
!
!                  In both cases, MAT must be specified as returned by BD_CMP and is not
!                  modified by the routine.
!                
!   TAUQ           (INPUT) real(stnd), dimension(:)
!                  TAUQ(i) must contain the scalar factor of the elementary
!                  reflector H(i) which determines Q, as returned by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 in their array
!                  argument TAUQ.
!
!                  The size of TAUQ must verify:
!                  size( TAUQ ) = min( size(MAT,1) , size(MAT,2) ) .
!
!   TAUP           (INPUT) real(stnd), dimension(:)
!                  TAUP(i) must contain the scalar factor of the elementary
!                  reflector G(i), which determines P, as returned by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 in their array
!                  argument TAUP.
!
!                  The size of TAUP must verify:
!                  size( TAUP ) = min( size(MAT,1) , size(MAT,2) ) .
!
!   RLMAT          (INPUT) real(stnd), dimension(:,:)
!                  On entry, the elements on and below the diagonal,
!                  with the array TAUQ, represent the orthogonal
!                  matrix Q as a product of elementary reflectors, 
!                  and the elements above the diagonal, with the
!                  array TAUP, represent the orthogonal matrix P
!                  as a product of elementary reflectors.
!                  RLMAT must be specified as returned by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 and
!                  is not modified by the routine.
!
!                  The shape of RLMAT must verify:
!                  size( RLMAT, 1 ) = size( RLMAT, 2 ) = min( size(MAT,1) , size(MAT,2) ) .
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, D contains the diagonal elements of the upper bidiagonal
!                  matrix BD as returned by BD_CMP, SELECT_SINGVAL_CMP or
!                  SELECT_SINGVAL_CMP2.
!
!                  The size of D must verify: size( D ) = min( size(MAT,1) , size(MAT,2) ) . 
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, E contains the off-diagonal elements of the upper bidiagonal
!                  matrix BD as returned by BD_CMP, SELECT_SINGVAL_CMP or
!                  SELECT_SINGVAL_CMP2:
!
!                       E(i) = BD(i-1,i) for i = 2,3,...,min(m,n);
!
!                  E(1) is arbitrary.
!
!                  The size of E must verify: size( E ) = min( size(MAT,1) , size(MAT,2) ) . 
!
!   S              (INPUT) real(stnd), dimension(:)
!                  On entry, selected singular values of the upper bidiagonal matrix BD.
!                  The singular values must be given in decreasing order and are
!                  assumed to be positive or zero.
!
!                  The size of S must verify: size( S ) <= min( size(MAT,1) , size(MAT,2) ) .
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed left singular vectors. The left singular vector
!                  associated with the singular value S(j) is stored in the j-th column
!                  of LEFTVEC.
!
!                  The shape of LEFTVEC must verify: 
!
!                  - size( LEFTVEC, 1 ) = size( MAT, 1 ) = m ,
!                  - size( LEFTVEC, 2 ) = size( S ) .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed right singular vectors. The right singular vector
!                  associated with the singular value S(j) is stored in the j-th column of
!                  RIGHTVEC.
!
!                  The shape of RIGHTVEC must verify: 
!
!                  - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ,
!                  - size( RIGHTVEC, 2 ) = size( S ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = FALSE : indicates successful exit ,
!                  - FAILURE = TRUE : indicates that some singular vectors of BD
!                    failed to converge in MAXITER iterations .
!
!   TAUO           (INPUT, OPTIONAL) real(stnd), dimension(:)
!                  The scalar factors of the elementary reflectors W(i), which
!                  represent the orthogonal matrix O of the QR or LQ decomposition of MAT.
!
!                  If the optional argument TAUO is present, it is assumed that the orthogonal matrix O is stored
!                  in factored form, as a product of elementary reflectors, in the argument MAT on entry.
!
!                  If the optional argument TAUO is absent, it is assumed that the orthogonal matrix O is stored
!                  explicitly in the argument MAT on entry.
!
!                  If the optional argument TAUO has been specified in the initial call to the BD_CMP, SELECT_SINGVAL_CMP
!                  or SELECT_SINGVAL_CMP2 subroutines, this optional argument TAUO must also be specified in the call
!                  to BD_INVITER2 for computing singular vectors, otherwise the results will be incorrect.
!
!                  See description of the argument MAT in the description of the BD_CMP subroutine,
!                  when the argument RLMAT is also present, for further details.
!
!                  The size of TAUO must be min( size(MAT,1) , size(MAT,2) ).
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed in the subroutine.
!
!                  By default, 2 inverse iterations are performed for all the singular vectors.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, all the singular vectors of the bidiagonal matrix BD are
!                    orthogonalized by the  Modified Gram-Schmidt or QR algorithm;
!                  - ORTHO=false, the singular vectors of the bidiagonal matrix BD are not
!                    orthogonalized by the Modified Gram-Schmidt or QR algorithm.
!
!                  The default is to orthogonalize the singular vectors only for the singular values,
!                  which are not well-separated.
!
!   BACKWARD_SWEEP (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - BACKWARD_SWEEP=true and the singular vectors of the bidiagonal matrix BD
!                    are orthogonalized by the modified Gram-Schmidt algorithm, a backward sweep of the 
!                    modified Gram-Schmidt algorithm is also performed;
!                  - BACKWARD_SWEEP=false, a backward sweep is not performed.
!
!                  The default is not to perform a backward sweep of the modified Gram-Schmidt algorithm.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the bidiagonal matrix BD is scaled before computing
!                    the singular vectors;
!                  - SCALING=false, the bidiagonal matrix BD is not scaled.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INITVEC=true, Fernando vectors are used to start the inverse iteration process
!                    for computing the singular vectors of the bidiagonal matrix BD (e.g. the eigenvectors of the
!                    associated Golub-Kahan tridiagonal matrix);
!                  - INITVEC=false, random uniform starting vectors are used.
!
!                  The default is to use Fernando starting vectors if the singular values are well-separated and
!                  the Golub-Kahan form of the input bidiagonal matrix is unreduced, and random uniform
!                  starting vectors otherwise.
!
!
! Further Details
! _______________
!
!   A first estimate of the singular vectors is computed by the Fernando method applied to the
!   tridiagonal Golub-Kahan matrix associated with the bidiagonal matrix BD (see the reference
!   (1) for details) for the singular values which are well-separated and if the Golub-Kahan
!   form of the input bidiagonal matrix is unreduced. For the other singular values,
!   a random start is used as a first estimate of the singular vectors as in the
!   standard inverse-iteration algorithm.
!
!   The singular vectors of BD are then computed or refined using inverse iteration on the
!   tridiagonal Golub-Kahan matrix for all the singular values at one step.
!
!   By default, the singular vectors of BD are then orthogonalized by the Modified Gram-Schmidt
!   or QR algorithm only if the singular values are not well-separated.
!
!   The singular vectors of MAT are finally computed by a blocked back-transformation algorithm.
!
!   The computation of the singular vectors of BD and the blocked back-transformation algorithm
!   to find the singular vectors of MAT are parallelized if OPENMP is used.
!
!   BD_INVITER2 may fail if some singular values specified in parameter S are nearly
!   identical for some pathological matrices.
!
!   For further details on Fernando method for computing eigenvectors of tridiagonal
!   matrices, the blocked back-transformation algorithm or inverse iteration, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, assert_eq, merror
    use QR_Procedures,   only : ortho_gen_qr, apply_q_qr, apply_q_lq
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),           dimension(:,:) :: mat, rlmat
    real(stnd),  intent(in),           dimension(:)   :: tauq, taup, d, e, s
    real(stnd),  intent(in), optional, dimension(:)   :: tauo
    real(stnd),  intent(out),          dimension(:,:) :: leftvec, rightvec
!
    integer(i4b), intent(in),  optional :: maxiter
!
    logical(lgl), intent(in),  optional :: ortho, backward_sweep, scaling, initvec
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(:,:), allocatable :: work, work2
!
    integer(i4b) :: n, m, nm, nsing, n1, mnthr_nsing
    integer      :: iok
!
    logical(lgl) :: upper, use_gen_o
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_inviter2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size( mat,     1 ),i4b)  ,   &
                   int(size( leftvec, 1 ),i4b)  ,   &
                   name_proc )
!
    n = assert_eq( int(size( mat,      2 ),i4b)  ,  &
                   int(size( rightvec, 1 ),i4b)  ,  &
                   name_proc )
!
    nm = assert_eq( (/ min(m,n)                  , &
                       int(size(tauq),i4b)       , &
                       int(size(taup),i4b)       , &
                       int(size(d),i4b)          , &
                       int(size(e),i4b)          , &
                       int(size(rlmat,1),i4b)    , &
                       int(size(rlmat,2),i4b)  /), &
                       name_proc )
!
    nsing = assert_eq( int(size(s),i4b)             ,  &
                       int(size( leftvec, 2 ),i4b)  ,  &
                       int(size( rightvec, 2 ),i4b) ,  &
                       name_proc )
!
    use_gen_o = true
!
    if ( present( tauo ) ) then
!
        call assert( logical( nm==size(tauo), lgl ), name_proc )
!
        use_gen_o = false
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( nm<=0_i4b ) then
!
        return
!
    end if
!
    upper = m >= n
!
!   COMPUTE SINGULAR VECTORS OF THE BIDIAGONAL FORM BD OF mat.
!
    call bd_inviter( true, d, e, s, leftvec(:nm,:nsing), rightvec(:nm,:nsing),                &
                     failure, maxiter=maxiter, ortho=ortho, backward_sweep=backward_sweep,    &
                     scaling=scaling, initvec=initvec  )
!
!   COMPUTE SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY BACK-TRANSFORMATION.
!
!   FIRST GENERATE LEFT AND RIGHT SINGULAR VECTORS OF R OR L IN leftvec AND rightvec, respectively.
!
    mnthr_nsing = int( real( nm, stnd )*thr_bt, i4b )
!
    if ( nsing>=mnthr_nsing  ) then
!
!       ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM IF NEARLY FULL SVD OF mat IS
!       REQUESTED.
!
        allocate( work(nm,nm), work2(nm,nsing), stat = iok )
!
    else
!
!       USE DIRECT ALGORITHM OTHERWISE OR IF THE REQUIRED
!       WORKSPACE IS NOT AVAILABLE.
!
        iok = -1
!
    end if
!
    if ( iok==0 ) then
!
!       GENERATE P IN work(:nm,:nm) .
!
        call ortho_gen_p_bd( rlmat, taup, work(:nm,:nm) )
!
!       MULTIPLY P IN work BY RIGHT SINGULAR VECTORS OF BD IN work2
!       FOR COMPUTING RIGHT SINGULAR VECTORS OF R OR L IN rightvec(:nm,:nsing) .
!
        work2(:nm,:nsing) = rightvec(:nm,:nsing)
!
#ifdef _BLAS
        call gemm( 'N', 'N', nm, nsing, nm, one, work, nm, work2, nm, zero, rightvec, n  )
#else
        rightvec(:nm,:nsing) = matmul( work(:nm,:nm), work2(:nm,:nsing) )
#endif
!
!       GENERATE Q IN work(:nm,:nm) .
!
        work(:nm,:nm) = rlmat(:nm,:nm)
!
        call ortho_gen_q_bd( work, tauq )
!
!       MULTIPLY Q IN work BY LEFT SINGULAR VECTORS OF BD IN work2
!       FOR COMPUTING LEFT SINGULAR VECTORS OF R OR L IN leftvec(:nm,:nsing) .
!
        work2(:nm,:nsing) = leftvec(:nm,:nsing)
!
#ifdef _BLAS
        call gemm( 'N', 'N', nm, nsing, nm, one, work, nm, work2, nm, zero, leftvec, m   )
#else
        leftvec(:nm,:nsing) = matmul( work(:nm,:nm), work2(:nm,:nsing) )
#endif
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( work )
!
    else
!
!       GENERATE LEFT SINGULAR VECTORS OF R OR L IN leftvec .
!
        call apply_q_bd( rlmat, tauq, leftvec(:nm,:nsing),  left=true, trans=false )
!
!       GENERATE RIGHT SINGULAR VECTORS OF R OR L IN rightvec .
!
        call apply_p_bd( rlmat, taup, rightvec(:nm,:nsing), left=true, trans=false )
!
    end if
!
    if ( use_gen_o ) then
!
!       THE ORTHOGONAL MATRIX O IS STORED IN EXPLICIT FORM IN mat .
!
        if ( upper ) then
!
            if ( iok==-1 ) then
!
!               ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
                allocate( work2(n,nsing), stat = iok )
!
                if ( iok/=0 ) then
                    call merror( name_proc//allocate_error )
                end if
!
            end if
!
!           COPY THE LEFT SINGULAR VECTORS OF R IN work .
!
            work2(:n,:nsing) = leftvec(:n,:nsing)
!
!           COMPUTE THE LEFT SINGULAR VECTORS OF mat IN leftvec .
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, nsing, n, one, mat, m, work2, n, zero, leftvec, m )
#else
            leftvec(:m,:nsing) = matmul( mat(:m,:n), work2(:n,:nsing) )
#endif
!
        else
!
#ifdef _BLAS
!
            if ( iok==-1 ) then
!
!               ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
                allocate( work2(m,nsing), stat = iok )
!
                if ( iok/=0 ) then
                    call merror( name_proc//allocate_error )
                end if
!
            end if
!
!           COPY THE RIGHT SINGULAR VECTORS OF R IN work .
!
            work2(:m,:nsing) = rightvec(:m,:nsing)
!
            call gemm( 'T', 'N', n, nsing, m, one, mat, m, work2, m, zero, rightvec, n )
!
#else
!
            if ( iok==0 ) then
!
!               DEALLOCATE WORKSPACE.
!
                deallocate( work2 )
!
            end if
!
!           REALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
            allocate( work(nsing,m), work2(nsing,n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           TRANSPOSE THE RIGHT SINGULAR VECTORS OF R IN work .
!
            work(:nsing,:m) = transpose( rightvec(:m,:nsing) )
!
!           COMPUTE TRANSPOSE OF THE RIGHT SINGULAR VECTORS OF mat IN work2 .
!
            work2(:nsing,:n) = matmul( work(:nsing,:m), mat(:m,:n) )
!
!           SAVE THE RIGHT SINGULAR VECTORS OF mat IN rightvec .
!
            rightvec(:n,:nsing) = transpose( work2(:nsing,:n) )
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( work )
#endif
!
        end if
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( work2 )
!
    else
!
!       THE ORTHOGONAL MATRIX O IS STORED IN FACTORED FORM IN mat AND tauo .
!
        if ( nsing>=mnthr_nsing ) then
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM IF NEARLY FULL SVD OF mat IS
!           REQUESTED.
!
            if ( upper ) then
                n1 = m
            else
                n1 = n
            end if
!
            if ( iok==0 ) then
!
                allocate( work(n1,nm), stat = iok )
!
            else
!
                allocate( work(n1,nm), work2(nm,nsing), stat = iok )
!
            end if
!
        else
!
!           USE DIRECT ALGORITHM OTHERWISE OR IF THE REQUIRED
!           WORKSPACE IS NOT AVAILABLE.
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!
            if ( upper ) then
!
!               GENERATE O IN work .
!
                work(:m,:n) = mat(:m,:n)
!
                call ortho_gen_qr( work, tauo )
!
!               MULTIPLY O IN work BY LEFT SINGULAR VECTORS OF R IN leftvec .
!
                work2(:n,:nsing) = leftvec(:n,:nsing)
!
#ifdef _BLAS
                call gemm( 'N', 'N', m, nsing, n, one, work, m, work2, n, zero, leftvec, m )
#else
                leftvec(:m,:nsing) = matmul( work(:m,:n), work2(:n,:nsing) )
#endif
!
            else
!
!               GENERATE O**(t) IN work .
!
                work(:n,:m) = transpose( mat(:m,:n) )
!
                call ortho_gen_qr( work, tauo )
!
!               MULTIPLY O**(t) IN work BY RIGHT SINGULAR VECTORS OF L IN rightvec .
!
                work2(:m,:nsing) = rightvec(:m,:nsing)
!
#ifdef _BLAS
                call gemm( 'N', 'N', n, nsing, m, one, work, n, work2, m, zero, rightvec, n )
#else
                rightvec(:n,:nsing) = matmul( work(:n,:m), work2(:m,:nsing) )
#endif
!
            end if
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( work, work2 )
!
        else
!
            if ( upper ) then
!
!               MULTIPLY O IN mat BY LEFT SINGULAR VECTORS OF R IN leftvec .
!
                leftvec(nm+1_i4b:m,:nsing) = zero
!
                call apply_q_qr( mat, tauo, leftvec(:m,:nsing), left=true, trans=false )
!
            else
!
!               MULTIPLY O**(t) IN mat BY RIGHT SINGULAR VECTORS OF L IN rightvec .
!
                rightvec(nm+1_i4b:n,:nsing) = zero
!
                call apply_q_lq( mat, tauo, rightvec(:n,:nsing), left=true, trans=true )
!
            end if
!
        end if
!
    end if
!
!
! END OF SUBROUTINE bd_inviter2_c
! _______________________________
!
    end subroutine bd_inviter2_c
!
! =========================================================================================
!
    subroutine bd_inviter2_d( mat, rmat, p, d, e, s, leftvec, rightvec, failure,      &
                              tauo, maxiter, ortho, backward_sweep, scaling, initvec, &
                              tol_reortho )
!
! Purpose
! _______
!
!   BD_INVITER2 computes all or selected left and right singular vectors of a full real m-by-n
!   matrix MAT with m>=n corresponding to specified singular values, using inverse iteration.
!
!   It is required that the original matrix MAT has been reduced to upper bidiagonal
!   form BD by a two-step algorithm as performed by SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4
!   subroutines with parameters P, RMAT, and eventually TAUO:
!
!   A QR factorization of the real m-by-n matrix MAT is first computed
!
!                        MAT = O * R
!
!   where O is orthogonal and R is upper triangular. In a second step, the n-by-n upper
!   triangular matrix R is reduced to upper bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * R * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix. Subroutines
!   SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4 compute O, Q, P, BD and also all or some
!   of the singular values of R, which are also the singular values of MAT. Using this two-step
!   factorization, BD_INVITER2 computes all or selected left and right singular vectors of R and
!   apply to them a back-transformation algorithm to obtain the corresponding left and right singular
!   vectors of MAT.
!
!   
! Arguments
! _________
!
!   MAT            (INPUT) real(stnd), dimension(:,:)
!                  On entry, the original m-by-n matrix after the two-step bidiagonal reduction by
!                  SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4 subroutines with arguments P, RMAT,
!                  and eventually TAUO.
!                  The precise content of MAT is determined by the presence or absence of the
!                  optional argument TAUO:
!
!                  - if the optional argument TAUO is absent, it is assumed that the
!                    first n columns of the orthogonal matrix O is stored explicitly
!                    in the argument MAT on entry,
!                  - if the optional argument TAUO is present, it is assumed that the
!                    orthogonal matrix O is stored in factored form, as a product of
!                    elementary reflectors, in the argument MAT on entry. In this case,
!                    MAT must contains the vectors which define the elementary reflectors
!                    W(i) whose products determine the matrix O, as returned by
!                    SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4 subroutines when the matrix
!                    argument RLMAT is present in the call to these subroutines.
!
!                  In both cases, MAT must be specified as returned by these subroutines and
!                  is not modified by the routine. See the description of the argument TAUO
!                  below for further details.
!                
!                  The shape of MAT must verify: size( MAT, 1 ) >= size( MAT, 2 ) = n .
!
!   RMAT           (INPUT) real(stnd), dimension(:,:)
!                  On entry, the n-by-n orthogonal matrix Q after reduction by SELECT_SINGVAL_CMP3 or
!                  SELECT_SINGVAL_CMP4 subroutines.
!                  RMAT must be specified as returned by these subroutines and is not modified by the routine.
!
!                  The shape of RMAT must verify: size( RMAT, 1 ) = size( RMAT, 2 ) = size( MAT, 2 ) = n.
!
!   P              (INPUT) real(stnd), dimension(:,:)
!                  On entry, the n-by-n orthogonal matrix P after reduction by SELECT_SINGVAL_CMP3 or
!                  SELECT_SINGVAL_CMP4 subroutines. P can be stored in factored form or not.
!                  Both cases are handled by the subroutine and P is not modified by the routine.
!                
!                  The shape of P must verify: size( P, 1 ) = size( P, 2 ) = size( MAT, 2 ) = n .
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, D contains the diagonal elements of the bidiagonal
!                  matrix BD as returned by SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4.
!
!                  The size of D must verify: size( D ) = size( MAT, 2 ) = n . 
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, E contains the off-diagonal elements of the bidiagonal
!                  matrix BD as returned by SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4:
!
!                          E(i) = BD(i-1,i) for i = 2,3,...,n;
!
!                  E(1) is arbitrary.
!
!                  The size of E must verify: size( E ) = size( MAT, 2 ) = n . 
!
!   S              (INPUT) real(stnd), dimension(:)
!                  On entry, selected singular values of the bidiagonal matrix BD.
!                  The singular values must be given in decreasing order and are
!                  assumed to be positive or zero.
!
!                  The size of S must verify: size( S ) <= size( MAT, 2 ) = n .
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed left singular vectors. The left singular vector
!                  associated with the singular value S(j) is stored in the j-th column
!                  of LEFTVEC.
!
!                  The shape of LEFTVEC must verify:
!
!                  - size( LEFTVEC, 1 ) = size( MAT, 1 ) = m ,
!                  - size( LEFTVEC, 2 ) = size( S ) .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed right singular vectors. The right singular vector
!                  associated with the singular value S(j) is stored in the j-th column of
!                  RIGHTVEC.
!
!                  The shape of RIGHTVEC must verify:  
!
!                  - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ,
!                  - size( RIGHTVEC, 2 ) = size( S ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = FALSE : indicates successful exit,
!                  - FAILURE = TRUE : indicates that some singular vectors of BD
!                    failed to converge in MAXITER iterations.
!
!   TAUO           (INPUT, OPTIONAL) real(stnd), dimension(:)
!                  The scalar factors of the elementary reflectors W(i), which
!                  represent the orthogonal matrix O of the QR decomposition of MAT as returned by
!                  SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4.
!
!                  If the optional argument TAUO is present, it is assumed that the orthogonal matrix O is stored
!                  in factored form, as a product of elementary reflectors, in the argument MAT on entry.
!
!                  If the optional argument TAUO is absent, it is assumed that the orthogonal matrix O is stored
!                  explicitly in the argument MAT on entry.
!
!                  If the optional argument TAUO has been specified in the initial call to the SELECT_SINGVAL_CMP3 or
!                  SELECT_SINGVAL_CMP4 subroutines, this optional argument TAUO must also be specified in the call to
!                  BD_INVITER2, otherwise the results will be incorrect.
!
!                  See description of the argument MAT in the description of the SELECT_SINGVAL_CMP3 or
!                  SELECT_SINGVAL_CMP4 subroutines, when the argument RMAT is also present, for further details.
!
!                  The size of TAUO must be size(MAT,2) = n .
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed in the subroutine.
!
!                  By default, 2 inverse iterations are performed for all the singular vectors.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, all the singular vectors of the bidiagonal matrix BD are
!                    orthogonalized by the  Modified Gram-Schmidt or QR algorithm;
!                  - ORTHO=false, the singular vectors of the bidiagonal matrix BD are
!                    not orthogonalized by the Modified Gram-Schmidt or QR algorithm.
!
!                  The default is to orthogonalize the singular vectors only for the singular values,
!                  which are not well-separated.
!
!   BACKWARD_SWEEP (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - BACKWARD_SWEEP=true and the singular vectors of the bidiagonal matrix BD
!                    are orthogonalized by the modified Gram-Schmidt algorithm, a backward sweep of the 
!                    modified Gram-Schmidt algorithm is also performed;
!                  - BACKWARD_SWEEP=false, a backward sweep is not performed.
!
!                  The default is not to perform a backward sweep of the modified Gram-Schmidt algorithm.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the bidiagonal matrix BD is scaled before computing
!                    the singular vectors;
!                  - SCALING=false, the bidiagonal matrix BD is not scaled.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INITVEC=true, Fernando vectors are used to start the inverse iteration process
!                    for computing the singular vectors of the bidiagonal matrix BD (e.g. the eigenvectors of the
!                    associated Golub-Kahan tridiagonal matrix);
!                  - INITVEC=false, random uniform starting vectors are used.
!
!                  The default is to use Fernando starting vectors if the singular values are well-separated and
!                  the Golub-Kahan form of the input bidiagonal matrix is unreduced, and random uniform
!                  starting vectors otherwise.
!
!   TOL_REORTHO    (INPUT, OPTIONAL) real(stnd)
!                  On entry, TOL_REORTHO is used to determine if the left singular vectors stored in LEFTVEC
!                  must be reortogonalized on exit in order to correct for the loss of orthogonality in the Ralha-Barlow
!                  one-sided bidiagonal reduction algorithm if MAT is nearly deficient. If one of the singular values,
!                  S(i), verifies the condition
!
!                                              S(i) <= TOL_REORTHO * S(1)
!
!                  all the computed left singular vectors are reorthogonalized with a QR factorization. If S(1) is the
!                  largest singular value of MAT, this condition leads to the assertion that the rank of MAT is less
!                  than size(S) and is thus a nearly singular matrix if TOL_REORTHO is a small positive value of the
!                  order of the machine epsilon.
!
!                  TOL_REORTHO must be greater or equal to zero and less than or equal to one. If TOL_REORTHO = 0. is
!                  used, the left singular vectors are reorthogonalized only if some singular values are almost zero.
!                  On the other hand, If TOL_REORTHO = 1. is used, the left singular vectors are always reorthogonalized.
!                  If TOL_REORTHO is specified as less than zero or greater than one, the default value is used.
!
!                  The default value is the value of the module parameter tol_reortho_def if size( S ) = n and
!                  tol_reortho_partial_def otherwise.
!
!
! Further Details
! _______________
!
!   A first estimate of the singular vectors is computed by the Fernando method applied to the
!   tridiagonal Golub-Kahan matrix associated with the bidiagonal matrix BD (see the reference
!   (1) for details) for the singular values which are well-separated and if the Golub-Kahan
!   form of the input bidiagonal matrix is unreduced. For the other singular values,
!   a random start is used as a first estimate of the singular vectors as in the
!   standard inverse-iteration algorithm.
!
!   The singular vectors of BD are then computed or refined using inverse iteration on the
!   tridiagonal Golub-Kahan matrix for all the singular values at one step.
!
!   By default, the singular vectors of BD are then orthogonalized by the Modified Gram-Schmidt
!   or QR algorithm only if the singular values are not well-separated.
!
!   The singular vectors of MAT are finally computed by a blocked back-transformation algorithm.
!
!   The computation of the singular vectors of BD and the blocked back-transformation algorithm
!   to find the singular vectors of MAT are parallelized if OPENMP is used.
!
!   BD_INVITER2 may fail if some singular values specified in parameter S are nearly
!   identical for some pathological matrices.
!
!   For further details on Fernando method for computing eigenvectors of tridiagonal
!   matrices, the blocked back-transformation algorithm or inverse iteration, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error5, svd_error8, allocate_error
    use Utilities,       only : assert, assert_eq, merror, ifirstloc
    use QR_Procedures,   only : qr_cmp, ortho_gen_qr, apply_q_qr
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),           dimension(:,:) :: mat, rmat, p
    real(stnd),  intent(in),           dimension(:)   :: d, e, s
    real(stnd),  intent(in), optional, dimension(:)   :: tauo
    real(stnd),  intent(in), optional                 :: tol_reortho
    real(stnd),  intent(out),          dimension(:,:) :: leftvec, rightvec
!
    integer(i4b), intent(in),  optional :: maxiter
!
    logical(lgl), intent(in),  optional :: ortho, backward_sweep, scaling, initvec
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: tol, tmp
    real(stnd), dimension(:),   allocatable :: diag, beta
    real(stnd), dimension(:,:), allocatable :: leftvec0, work
!
    integer(i4b) :: i, n, m, nsing, mnthr_nsing, nrank
    integer      :: iok
!
    logical(lgl) :: use_gen_o
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_inviter2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size( mat,     1 ),i4b)  ,   &
                   int(size( leftvec, 1 ),i4b)  ,   &
                   name_proc )
!
    n = assert_eq( (/ int(size( mat,      2 ),i4b)    ,  &
                      int(size( rightvec, 1 ),i4b)    ,  &
                      int(size( rmat,     1 ),i4b)    ,  &
                      int(size( rmat,     2 ),i4b)    ,  &
                      int(size( p,        1 ),i4b)    ,  &
                      int(size( p,        2 ),i4b)    ,  &
                      int(size( d )          ,i4b)    ,  &
                      int(size( e )          ,i4b) /) ,  &
                      name_proc )
!
    if ( m<n ) then
!
        call merror( name_proc//svd_error5 )
!
    end if
!
    nsing = assert_eq( int(size(s),i4b)             ,  &
                       int(size( leftvec, 2 ),i4b)  ,  &
                       int(size( rightvec, 2 ),i4b) ,  &
                       name_proc )
!
    use_gen_o = true
!
    if ( present( tauo ) ) then
!
        call assert( logical( n==size(tauo), lgl ), name_proc )
!
        use_gen_o = false
!
    end if
!
    if ( nsing==n ) then
        tol = tol_reortho_def
    else
        tol = tol_reortho_partial_def
    end if
!
    if ( present(tol_reortho) ) then
!
        if ( tol_reortho>=zero .and. tol_reortho<=one ) then
!
            tol = tol_reortho
!
        end if
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
!
        return
!
    end if
!
!   FIRST STEP: COMPUTE SINGULAR VECTORS OF THE UPPER n-BY-n TRIANGULAR MATRIX R.
!
    if ( p(1_i4b,1_i4b)==one ) then
!
!       THE ORTHOGONAL MATRIX P HAS BEEN GENERATED BY select_singval_cmp3 OR select_singval_cmp4.
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( leftvec0(n,nsing), work(n,nsing), stat=iok   )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE SINGULAR VECTORS OF THE BIDIAGONAL FORM BD OF mat.
!
        call bd_inviter( true, d, e, s, leftvec(:n,:nsing), work(:n,:nsing),                    &
                         failure, maxiter=maxiter, ortho=ortho, backward_sweep=backward_sweep,  &
                         scaling=scaling, initvec=initvec  )
!
!       COMPUTE RIGHT SINGULAR VECTORS OF THE UPPER n-BY-n TRIANGULAR MATRIX R BY MULTIPLICATION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', n, nsing, n, one, p, n, work, n, zero, rightvec, n )
#else
        rightvec(:n,:nsing) = matmul( p(:n,:n), work(:n,:nsing) )
#endif
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( work )
!
    else if ( p(1_i4b,1_i4b)==-one ) then
!
!       THE ORTHOGONAL MATRIX P IS STORED IN FACTORED FORM BY select_singval_cmp3 OR select_singval_cmp4.
!
!       ALLOCATE WORK ARRAY.
!
        allocate( leftvec0(n,nsing), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE SINGULAR VECTORS OF THE BIDIAGONAL FORM BD OF mat.
!
        call bd_inviter( true, d, e, s, leftvec(:n,:nsing), rightvec(:n,:nsing),                 &
                         failure, maxiter=maxiter, ortho=ortho, backward_sweep=backward_sweep,   &
                         scaling=scaling, initvec=initvec  )
!
!       COMPUTE RIGHT SINGULAR VECTORS OF THE UPPER n-BY-n TRIANGULAR MATRIX R BY BACK TRANSFORMATION.
!
        call apply_q_qr( p(2_i4b:n,2_i4b:n), p(2_i4b:n,1_i4b), rightvec(2_i4b:n,:nsing),   &
                         left=true, trans=false  )
!
    else
!
        call merror( name_proc//svd_error8 )
!
    end if
!
!   COMPUTE LEFT SINGULAR VECTORS OF THE UPPER n-BY-n TRIANGULAR MATRIX R BY MULTIPLICATION.
!
#ifdef _BLAS
    call gemm( 'N', 'N', n, nsing, n, one, rmat, n, leftvec, m, zero, leftvec0, n )
#else
    leftvec0(:n,:nsing) = matmul( rmat(:n,:n), leftvec(:n,:nsing) )
#endif
!
!   SECOND STEP: COMPUTE SINGULAR VECTORS OF THE FULL m-BY-n RECTANGULAR MATRIX MAT
!   BY BACK TRANSFORMATION OF THOSE OF R.
!
    if ( use_gen_o ) then
!
!       THE ORTHOGONAL MATRIX O IS STORED IN EXPLICIT FORM IN mat .
!
!       COMPUTE THE LEFT SINGULAR VECTORS OF mat IN leftvec .
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsing, n, one, mat, m, leftvec0, n, zero, leftvec, m )
#else
        leftvec(:m,:nsing) = matmul( mat(:m,:n), leftvec0(:n,:nsing) )
#endif
!
    else
!
!       THE ORTHOGONAL MATRIX O IS STORED IN FACTORED FORM IN mat AND tauo .
!    
        mnthr_nsing = int( real( n, stnd )*thr_bt, i4b )
!
        if ( nsing>=mnthr_nsing ) then
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM IF NEARLY FULL SVD OF mat IS
!           REQUESTED.
!
            allocate( work(m,n), stat = iok )
!
        else
!
!           USE DIRECT ALGORITHM OTHERWISE OR IF THE REQUIRED
!           WORKSPACE IS NOT AVAILABLE.
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!
!           GENERATE O IN work .
!
            work(:m,:n) = mat(:m,:n)
!
            call ortho_gen_qr( work, tauo )
!
!           MULTIPLY O IN work BY LEFT SINGULAR VECTORS OF R IN leftvec0
!           TO GET LEFT SINGULAR VECTORS OF MAT.
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, nsing, n, one, work, m, leftvec0, n, zero, leftvec, m )
#else
            leftvec(:m,:nsing) = matmul( work(:m,:n), leftvec0(:n,:nsing) )
#endif
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( work )
!
        else
!
!           MULTIPLY O STORED IN FACTORED FORM IN mat BY LEFT SINGULAR VECTORS OF R IN leftvec0
!           TO GET LEFT SINGULAR VECTORS OF MAT.
!
            leftvec(:n,:nsing)        = leftvec0(:n,:nsing)
            leftvec(n+1_i4b:m,:nsing) = zero
!
            call apply_q_qr( mat, tauo, leftvec(:m,:nsing), left=true, trans=false )
!
        end if
!
    end if
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( leftvec0 )
!
!   DETERMINE THE EFFECTIVE RANK OF mat USING THE TOLERANCE tol .
!
    tmp = max( tol*s(1_i4b), safmin )
!
    nrank = ifirstloc( logical( s(:nsing)<=tmp, lgl ) ) - 1_i4b
!
    if ( nrank<nsing ) then
!
!       REORTOGONALIZE LEFT SINGULAR VECTORS IF REQUIRED.
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( diag(nsing), beta(nsing), stat=iok   )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        call qr_cmp( leftvec(:m,:nsing), diag(:nsing), beta(:nsing) )
!
        call ortho_gen_qr( leftvec(:m,:nsing), beta(:nsing) )
!
        do i = 1_i4b, nsing
!
            if ( diag(i)>=zero ) cycle
!
            leftvec(:m,i) = -leftvec(:m,i)
!        
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( diag, beta )
!
    end if
!
!
! END OF SUBROUTINE bd_inviter2_d
! _______________________________
!
    end subroutine bd_inviter2_d
!
!
! =========================================================================================
!                   SINGULAR VECTORS BY DEFLATION ALGORITHMS
! =========================================================================================
!
!
    subroutine upper_bd_dsqd2_a( q2, e2, shift, flip, d )
!
! Purpose
! _______
!
!   UPPER_BD_DSQD2 computes:
!
!   - the L * D * L' factorization of the matrix BD' * BD - shift * I , if FLIP=false;
!
!   - the U * D * U' factorization of the matrix BD * BD' - shift * I , if FLIP=true;
!
!   for a n-by-n (upper) bidiagonal matrix BD and a given shift. L and U are, respectively,
!   unit lower and unit upper bidiagonal matrices and D is a diagonal matrix.
!
!   The differential form of the stationary QD algorithm of Rutishauser is used to compute
!   the factorization from the squared elements of the bidiagonal matrix BD (see the reference
!   (1) below for further details).
!
!   The subroutine outputs the diagonal matrix D of the factorization.
!
!
! Arguments
! _________
!
!   Q2          (INPUT) real(stnd), dimension(:)
!               On entry, Q2 contains the squared diagonal elements of the bidiagonal
!               matrix BD.
!
!   E2          (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 squared off-diagonal elements of the bidiagonal matrix BD.
!
!               The size of E2 must be  size( E2 ) = size( Q2 ) - 1.
!
!   SHIFT       (INPUT) real(stnd)
!               On entry, the shift.
!
!   FLIP        (INPUT) logical(lgl)
!               On entry, if FLIP=false the L * D * L' factorization of
!               the matrix BD' * BD - shift * I is computed. Otherwise,
!               if FLIP=true the U * D * U' factorization of the matrix
!               BD * BD' - shift * I is computed.
!
!   D           (OUTPUT) real(stnd), dimension(:)
!               On exit, the elements of the diagonal matrix D.
!
!               The size of D must be  size( D ) = size( Q2 ).
!
!
! Further Details
! _______________
!
!   The bidiagonal matrix BD must be scaled appropriately before using this subroutine
!   in order to avoid overflows (see the reference (1) below for further details).
!
!   This subroutine is adapted from the algorithms given in reference (1). See:
!
!   (1) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)  :: shift
    real(stnd), dimension(:),   intent(in)  :: q2, e2
    real(stnd), dimension(:),   intent(out) :: d
!
    logical(lgl),               intent(in)  :: flip
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)    :: gamma, di, ti
!
    integer(i4b)  :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='upper_bd_dsqd2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(q2),i4b)         ,   &
                    int(size(e2),i4b) + 1_i4b ,   &
                    int(size(d),i4b)          ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            d(1_i4b) = q2(1_i4b) - shift
!
        end if
!
        return
!
    end if
!
!   GET SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
    gamma  = sqrt( safmin )
!
    if ( flip ) then
!
!       APPLY THE DIFFERENTIAL STATIONARY QD ALGORITHM
!       FROM BOTTOM TO TOP.
!
        ti = shift
        di = q2(n) - shift
!
        do i=n, 2_i4b, -1_i4b
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
!
            ti = shift + e2(i-1_i4b)*(ti/di)
            di = q2(i-1_i4b) - ti
!
        end do
!
        d(1_i4b) = di
!
    else
!
!       APPLY THE DIFFERENTIAL STATIONARY QD ALGORITHM
!       FROM TOP TO BOTTOM.
!
        ti = shift
        di = q2(1_i4b) - shift
!
        do i=1_i4b, n-1_i4b
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
!
            ti = shift + e2(i)*(ti/di)
            di = q2(i+1_i4b) - ti
!
        end do
!
        d(n) = di
!
    end if
!
!
! END OF SUBROUTINE upper_bd_dsqd2_a
! __________________________________
!
    end subroutine upper_bd_dsqd2_a
!
! =========================================================================================
!
    subroutine upper_bd_dpqd2_a( q2, e2, shift, flip, d )
!
! Purpose
! _______
!
!   UPPER_BD_DPQD2 computes:
!
!   - the L * D * L' factorization of the matrix BD * BD' - shift * I  , if FLIP=false;
!
!   - the U * D * U' factorization of the matrix BD' * BD - shift * I  , if FLIP=true;
!
!   for a n-by-n (upper) bidiagonal matrix BD and a given shift. L and U are, respectively,
!   unit lower and unit upper bidiagonal matrices and D is a diagonal matrix.
!
!   The differential form of the progressive QD algorithm of Rutishauser is used to compute
!   the factorization from the squared elements of the bidiagonal matrix BD (see the reference
!   (1) below for further details).
!
!   The subroutine outputs the diagonal matrix D of the factorization.
!
!
! Arguments
! _________
!
!   Q2          (INPUT) real(stnd), dimension(:)
!               On entry, Q2 contains the squared diagonal elements of the bidiagonal
!               matrix BD.
!
!   E2          (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 squared off-diagonal elements of the bidiagonal matrix BD.
!
!               The size of E2 must be  size( E2 ) = size( Q2 ) - 1.
!
!   SHIFT       (INPUT) real(stnd)
!               On entry, the shift.
!
!   FLIP        (INPUT) logical(lgl)
!               On entry, if FLIP=false the L * D * L' factorization of
!               the matrix BD * BD' - shift * I is computed. Otherwise,
!               if FLIP=true the U * D * U' factorization of the matrix
!               BD' * BD - shift * I is computed.
!
!   D           (OUTPUT) real(stnd), dimension(:)
!               On exit, the elements of the diagonal matrix D.
!
!               The size of D must be  size( D ) = size( Q2 ).
!
!
! Further Details
! _______________
!
!   The bidiagonal matrix BD must be scaled appropriately before using this subroutine
!   in order to avoid overflows (see the reference (1) below for further details).
!
!   This subroutine is adapted from the algorithms given in reference (1). See:
!
!   (1) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)  :: shift
    real(stnd), dimension(:),   intent(in)  :: q2, e2
    real(stnd), dimension(:),   intent(out) :: d
!
    logical(lgl),               intent(in)  :: flip
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)    :: gamma, di, si
!
    integer(i4b)  :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='upper_bd_dpqd2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(q2),i4b)         ,   &
                    int(size(e2),i4b) + 1_i4b ,   &
                    int(size(d),i4b)          ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            d(1_i4b) = q2(1_i4b) - shift
!
        end if
!
        return
!
    end if
!
!   GET SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
    gamma  = sqrt( safmin )
!
    if ( flip ) then
!
!       APPLY THE DIFFERENTIAL PROGRESSIVE QD ALGORITHM
!       FROM BOTTOM TO TOP.
!
        si = q2(n) - shift
!
        do i=n, 2_i4b, -1_i4b
!
            di = si + e2(i-1_i4b)
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
!
            si = q2(i-1_i4b)*(si/di) - shift
!
        end do
!
        d(1_i4b) = si
!
    else
!
!       APPLY THE DIFFERENTIAL PROGRESSIVE QD ALGORITHM
!       FROM TOP TO BOTTOM.
!
        si = q2(1_i4b) - shift
!
        do i=1_i4b, n-1_i4b
!
            di = si + e2(i)
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
!
            si = q2(i+1_i4b)*(si/di) - shift
!
        end do
!
        d(n) = si
!
    end if
!
!
! END OF SUBROUTINE upper_bd_dpqd2_a
! __________________________________
!
    end subroutine upper_bd_dpqd2_a
!
! =========================================================================================
!
    subroutine upper_bd_dsqd2_b( q2, e2, shift, flip, d, t )
!
! Purpose
! _______
!
!   UPPER_BD_DSQD2 computes:
!
!   - the L * D * L' factorization of the matrix BD' * BD - shift * I , if FLIP=false;
!
!   - the U * D * U' factorization of the matrix BD * BD' - shift * I , if FLIP=true;
!
!   for a n-by-n (upper) bidiagonal matrix BD and a given shift. L and U are, respectively,
!   unit lower and unit upper bidiagonal matrices and D is a diagonal matrix.
!
!   The differential form of the stationary QD algorithm of Rutishauser is used to compute
!   the factorization from the squared elements of the bidiagonal matrix BD (see the reference
!   (1) below for further details).
!
!   The subroutine outputs the diagonal matrix D of the factorization and the auxiliary
!   variable T in the differential form of the stationary QD algorithm.
!
!
! Arguments
! _________
!
!   Q2          (INPUT) real(stnd), dimension(:)
!               On entry, Q2 contains the squared diagonal elements of the bidiagonal
!               matrix BD.
!
!   E2          (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 squared off-diagonal elements of the bidiagonal matrix BD.
!
!               The size of E2 must be  size( E2 ) = size( Q2 ) - 1.
!
!   SHIFT       (INPUT) real(stnd)
!               On entry, the shift.
!
!   FLIP        (INPUT) logical(lgl)
!               On entry, if FLIP=false the L * D * L' factorization of
!               the matrix BD' * BD - shift * I is computed. Otherwise,
!               if FLIP=true the U * D * U' factorization of the matrix
!               BD * BD' - shift * I is computed.
!
!   D           (OUTPUT) real(stnd), dimension(:)
!               On exit, the elements of the diagonal matrix D.
!
!               The size of D must be  size( D ) = size( Q2 ).
!
!   T           (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the auxiliary values T(i) in the differential form
!               of the stationary QD algorithm.
!
!               The size of T must be  size( T ) = size( D ) = size( Q2 ).
!
!
! Further Details
! _______________
!
!   The bidiagonal matrix BD must be scaled appropriately before using this subroutine
!   in order to avoid overflows (see the reference (1) below for further details).
!
!   This subroutine is adapted from the algorithms given in reference (1). See:
!
!   (1) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)  :: shift
    real(stnd), dimension(:),   intent(in)  :: q2, e2
    real(stnd), dimension(:),   intent(out) :: d, t
!
    logical(lgl),               intent(in)  :: flip
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: gamma, di, ti
!
    integer(i4b) :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='upper_bd_dsqd2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(q2),i4b)         ,   &
                    int(size(e2),i4b) + 1_i4b ,   &
                    int(size(d),i4b)          ,   &
                    int(size(t),i4b)          ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            t(1_i4b) = shift
            d(1_i4b) = q2(1_i4b) - shift
!
        end if
!
        return
!
    end if
!
!   GET SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
    gamma  = sqrt( safmin )
!
    if ( flip ) then
!
!       APPLY THE DIFFERENTIAL STATIONARY QD ALGORITHM
!       FROM BOTTOM TO TOP.
!
        ti = shift
        di = q2(n) - shift
!
        do i=n, 2_i4b, -1_i4b
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
            t(i) = ti
!
            ti = shift + e2(i-1_i4b)*(ti/di)
            di = q2(i-1_i4b) - ti
!
        end do
!
        d(1_i4b) = di
        t(1_i4b) = ti
!
    else
!
!       APPLY THE DIFFERENTIAL STATIONARY QD ALGORITHM
!       FROM TOP TO BOTTOM.
!
        ti = shift
        di = q2(1_i4b) - shift
!
        do i=1_i4b, n-1_i4b
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
            t(i) = ti
!
            ti = shift + e2(i)*(ti/di)
            di = q2(i+1_i4b) - ti
!
        end do
!
        d(n) = di
        t(n) = ti
!
    end if
!
!
! END OF SUBROUTINE upper_bd_dsqd2_b
! __________________________________
!
    end subroutine upper_bd_dsqd2_b
!
! =========================================================================================
!
    subroutine upper_bd_dpqd2_b( q2, e2, shift, flip, d, s )
!
! Purpose
! _______
!
!   UPPER_BD_DPQD2 computes:
!
!   - the L * D * L' factorization of the matrix BD * BD' - shift * I , if FLIP=false;
!
!   - the U * D * U' factorization of the matrix BD' * BD - shift * I , if FLIP=true;
!
!   for a n-by-n (upper) bidiagonal matrix BD and a given shift. L and U are, respectively,
!   unit lower and unit upper bidiagonal matrices and D is a diagonal matrix.
!
!   The differential form of the progressive QD algorithm of Rutishauser is used to compute
!   the factorization from the squared elements of the bidiagonal matrix BD (see the reference
!   (1) below for further details).
!
!   The subroutine outputs the diagonal matrix D of the factorization and the auxiliary
!   variable S in the differential form of the progressive QD algorithm.
!
!
! Arguments
! _________
!
!   Q2          (INPUT) real(stnd), dimension(:)
!               On entry, Q2 contains the squared diagonal elements of the bidiagonal
!               matrix BD.
!
!   E2          (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 squared off-diagonal elements of the bidiagonal matrix BD.
!
!               The size of E2 must be  size( E2 ) = size( Q2 ) - 1.
!
!   SHIFT       (INPUT) real(stnd)
!               On entry, the shift.
!
!   FLIP        (INPUT) logical(lgl)
!               On entry, if FLIP=false the L * D * L' factorization of
!               the matrix BD * BD' - shift * I is computed. Otherwise,
!               if FLIP=true the U * D * U' factorization of the matrix
!               BD' * BD - shift * I is computed.
!
!   D           (OUTPUT) real(stnd), dimension(:)
!               On exit, the elements of the diagonal matrix D.
!
!               The size of D must be  size( D ) = size( Q2 ).
!
!   S           (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the auxiliary values S(i) in the differential form
!               of the progressive QD algorithm.
!
!               The size of S must be  size( S ) = size( D ) = size( Q2 ).
!
!
! Further Details
! _______________
!
!   The bidiagonal matrix BD must be scaled appropriately before using this subroutine
!   in order to avoid overflows (see the reference (1) below for further details).
!
!   This subroutine is adapted from the algorithms given in reference (1). See:
!
!   (1) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)  :: shift
    real(stnd), dimension(:),   intent(in)  :: q2, e2
    real(stnd), dimension(:),   intent(out) :: d, s
!
    logical(lgl),               intent(in)  :: flip
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)     :: gamma, di, si
!
    integer(i4b)   :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='upper_bd_dpqd2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(q2),i4b)         ,   &
                    int(size(e2),i4b) + 1_i4b ,   &
                    int(size(d),i4b)          ,   &
                    int(size(s),i4b)          ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            s(1_i4b) = q2(1_i4b) - shift
            d(1_i4b) = s(1_i4b)
!
        end if
!
        return
!
    end if
!
!   GET SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
    gamma  = sqrt( safmin )
!
    if ( flip ) then
!
!       APPLY THE DIFFERENTIAL PROGRESSIVE QD ALGORITHM
!       FROM BOTTOM TO TOP.
!
        si = q2(n) - shift
!
        do i=n, 2_i4b, -1_i4b
!
            di = si + e2(i-1_i4b)
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
            s(i) = si
!
            si = q2(i-1_i4b)*(si/di) - shift
!
        end do
!
        d(1_i4b) = si
        s(1_i4b) = si
!
    else
!
!       APPLY THE DIFFERENTIAL PROGRESSIVE QD ALGORITHM
!       FROM TOP TO BOTTOM.
!
        si = q2(1_i4b) - shift
!
        do i=1_i4b, n-1_i4b
!
            di = si + e2(i)
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
            s(i) = si
!
            si = q2(i+1_i4b)*(si/di) - shift
!
        end do
!
        d(n) = si
        s(n) = si
!
    end if
!
!
! END OF SUBROUTINE upper_bd_dpqd2_b
! __________________________________
!
    end subroutine upper_bd_dpqd2_b
!
! =========================================================================================
!
    subroutine upper_bd_dsqd_a( a, b, shift, flip, d )
!
! Purpose
! _______
!
!   UPPER_BD_DSQD computes:
!
!   - the L * D * L' factorization of the matrix BD' * BD - shift * I , if FLIP=false;
!
!   - the U * D * U' factorization of the matrix BD * BD' - shift * I , if FLIP=true;
!
!   for a n-by-n (upper) bidiagonal matrix BD and a given shift. L and U are, respectively,
!   unit lower and unit upper bidiagonal matrices and D is a diagonal matrix.
!
!   The differential form of the stationary QD algorithm of Rutishauser is used to compute
!   the factorization (see the reference (1) below for further details).
!
!   The subroutine outputs the diagonal matrix D of the factorization.
!
!
! Arguments
! _________
!
!   A           (INPUT) real(stnd), dimension(:)
!               On entry, A contains the diagonal elements of the bidiagonal
!               matrix BD.
!
!   B           (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 off-diagonal elements of the bidiagonal matrix BD.
!
!               The size of B must be  size( B ) = size( A ) - 1.
!
!   SHIFT       (INPUT) real(stnd)
!               On entry, the shift.
!
!   FLIP        (INPUT) logical(lgl)
!               On entry, if FLIP=false the L * D * L' factorization of
!               the matrix BD' * BD - shift * I is computed. Otherwise,
!               if FLIP=true the U * D * U' factorization of the matrix
!               BD * BD' - shift * I is computed.
!
!   D           (OUTPUT) real(stnd), dimension(:)
!               On exit, the elements of the diagonal matrix D.
!
!               The size of D must be  size( D ) = size( A ).
!
!
! Further Details
! _______________
!
!   The bidiagonal matrix BD must be scaled appropriately before using this subroutine
!   in order to avoid overflows (see the reference (1) below for further details).
!
!   This subroutine is adapted from the algorithms given in reference (1). See:
!
!   (1) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)  :: shift
    real(stnd), dimension(:),   intent(in)  :: a, b
    real(stnd), dimension(:),   intent(out) :: d
!
    logical(lgl),               intent(in)  :: flip
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                      :: gamma, di, ti
    real(stnd), dimension(size(a))  :: q2
    real(stnd), dimension(size(b))  :: e2
!
    integer(i4b)  :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='upper_bd_dsqd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(a),i4b)         ,   &
                    int(size(b),i4b) + 1_i4b ,   &
                    int(size(d),i4b)         ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            d(1_i4b) = a(1_i4b)*a(1_i4b) - shift
!
        end if
!
        return
!
    end if
!
!   GET SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
    gamma  = sqrt( safmin )
!
!   COMPUTE THE QD VARIABLES.
!
    q2(1_i4b:n)       = a(1_i4b:n)*a(1_i4b:n)
    e2(1_i4b:n-1_i4b) = b(1_i4b:n-1_i4b)*b(1_i4b:n-1_i4b)
!
    if ( flip ) then
!
!       APPLY THE DIFFERENTIAL STATIONARY QD ALGORITHM
!       FROM BOTTOM TO TOP.
!
        ti = shift
        di = q2(n) - shift
!
        do i=n, 2_i4b, -1_i4b
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
!
            ti = shift + e2(i-1_i4b)*(ti/di)
            di = q2(i-1_i4b) - ti
!
        end do
!
        d(1_i4b) = di
!
    else
!
!       APPLY THE DIFFERENTIAL STATIONARY QD ALGORITHM
!       FROM TOP TO BOTTOM.
!
        ti = shift
        di = q2(1_i4b) - shift
!
        do i=1_i4b, n-1_i4b
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
!
            ti = shift + e2(i)*(ti/di)
            di = q2(i+1_i4b) - ti
!
        end do
!
        d(n) = di
!
    end if
!
!
! END OF SUBROUTINE upper_bd_dsqd_a
! _________________________________
!
    end subroutine upper_bd_dsqd_a
!
! =========================================================================================
!
    subroutine upper_bd_dpqd_a( a, b, shift, flip, d )
!
! Purpose
! _______
!
!   UPPER_BD_DPQD computes:
!
!   - the L * D * L' factorization of the matrix BD * BD' - shift * I , if FLIP=false;
!
!   - the U * D * U' factorization of the matrix BD' * BD - shift * I , if FLIP=true;
!
!   for a n-by-n (upper) bidiagonal matrix BD and a given shift. L and U are, respectively,
!   unit lower and unit upper bidiagonal matrices and D is a diagonal matrix.
!
!   The differential form of the progressive QD algorithm of Rutishauser is used to compute
!   the factorization (see the reference (1) below for further details).
!
!   The subroutine outputs the diagonal matrix D of the factorization.
!
!
! Arguments
! _________
!
!   A           (INPUT) real(stnd), dimension(:)
!               On entry, A contains the diagonal elements of the bidiagonal
!               matrix BD.
!
!   B           (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 off-diagonal elements of the bidiagonal matrix BD.
!
!               The size of B must be  size( B ) = size( A ) - 1.
!
!   SHIFT       (INPUT) real(stnd)
!               On entry, the shift.
!
!   FLIP        (INPUT) logical(lgl)
!               On entry, if FLIP=false the L * D * L' factorization of
!               the matrix BD * BD' - shift * I is computed. Otherwise,
!               if FLIP=true the U * D * U' factorization of the matrix
!               BD' * BD - shift * I is computed.
!
!   D           (OUTPUT) real(stnd), dimension(:)
!               On exit, the elements of the diagonal matrix D.
!
!               The size of D must be  size( D ) = size( A ).
!
!
! Further Details
! _______________
!
!   The bidiagonal matrix BD must be scaled appropriately before using this subroutine
!   in order to avoid overflows (see the reference (1) below for further details).
!
!   This subroutine is adapted from the algorithms given in reference (1). See:
!
!   (1) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)  :: shift
    real(stnd), dimension(:),   intent(in)  :: a, b
    real(stnd), dimension(:),   intent(out) :: d
!
    logical(lgl),               intent(in)  :: flip
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                      :: gamma, di, si
    real(stnd), dimension(size(a))  :: q2
    real(stnd), dimension(size(b))  :: e2
!
    integer(i4b)  :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='upper_bd_dpqd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(a),i4b)         ,   &
                    int(size(b),i4b) + 1_i4b ,   &
                    int(size(d),i4b)         ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            d(1_i4b) = a(1_i4b)*a(1_i4b) - shift
!
        end if
!
        return
!
    end if
!
!   GET SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
    gamma  = sqrt( safmin )
!
!   COMPUTE THE QD VARIABLES.
!
    q2(1_i4b:n)       = a(1_i4b:n)*a(1_i4b:n)
    e2(1_i4b:n-1_i4b) = b(1_i4b:n-1_i4b)*b(1_i4b:n-1_i4b)
!
    if ( flip ) then
!
!       APPLY THE DIFFERENTIAL PROGRESSIVE QD ALGORITHM
!       FROM BOTTOM TO TOP.
!
        si = q2(n) - shift
!
        do i=n, 2_i4b, -1_i4b
!
            di = si + e2(i-1_i4b)
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
!
            si = q2(i-1_i4b)*(si/di) - shift
!
        end do
!
        d(1_i4b) = si
!
    else
!
!       APPLY THE DIFFERENTIAL PROGRESSIVE QD ALGORITHM
!       FROM TOP TO BOTTOM.
!
        si = q2(1_i4b) - shift
!
        do i=1_i4b, n-1_i4b
!
            di = si + e2(i)
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
!
            si = q2(i+1_i4b)*(si/di) - shift
!
        end do
!
        d(n) = si
!
    end if
!
!
! END OF SUBROUTINE upper_bd_dpqd_a
! _________________________________
!
    end subroutine upper_bd_dpqd_a
!
! =========================================================================================
!
    subroutine upper_bd_dsqd_b( a, b, shift, flip, d, t )
!
! Purpose
! _______
!
!   UPPER_BD_DSQD computes:
!
!   - the L * D * L' factorization of the matrix BD' * BD - shift * I , if FLIP=false;
!
!   - the U * D * U' factorization of the matrix BD * BD' - shift * I , if FLIP=true;
!
!   for a n-by-n (upper) bidiagonal matrix BD and a given shift. L and U are, respectively,
!   unit lower and unit upper bidiagonal matrices and D is a diagonal matrix.
!
!   The differential form of the stationary QD algorithm of Rutishauser is used to compute
!   the factorization (see the reference (1) below for further details).
!
!   The subroutine outputs the diagonal matrix D of the factorization and the auxiliary
!   variable T in the differential form of the stationary QD algorithm.
!
!
! Arguments
! _________
!
!   A           (INPUT) real(stnd), dimension(:)
!               On entry, A contains the diagonal elements of the bidiagonal
!               matrix BD.
!
!   B           (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 off-diagonal elements of the bidiagonal matrix BD.
!
!               The size of B must be  size( B ) = size( A ) - 1.
!
!   SHIFT       (INPUT) real(stnd)
!               On entry, the shift.
!
!   FLIP        (INPUT) logical(lgl)
!               On entry, if FLIP=false the L * D * L' factorization of
!               the matrix BD' * BD - shift * I is computed. Otherwise,
!               if FLIP=true the U * D * U' factorization of the matrix
!               BD * BD' - shift * I is computed.
!
!   D           (OUTPUT) real(stnd), dimension(:)
!               On exit, the elements of the diagonal matrix D.
!
!               The size of D must be  size( D ) = size( A ).
!
!   T           (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the auxiliary values T(i) in the differential form
!               of the stationary QD algorithm.
!
!               The size of T must be  size( T ) = size( D ) = size( A ).
!
!
! Further Details
! _______________
!
!   The bidiagonal matrix BD must be scaled appropriately before using this subroutine
!   in order to avoid overflows (see the reference (1) below for further details).
!
!   This subroutine is adapted from the algorithms given in reference (1). See:
!
!   (1) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)  :: shift
    real(stnd), dimension(:),   intent(in)  :: a, b
    real(stnd), dimension(:),   intent(out) :: d, t
!
    logical(lgl),               intent(in)  :: flip
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                      :: gamma, di, ti
    real(stnd), dimension(size(a))  :: q2
    real(stnd), dimension(size(b))  :: e2
!
    integer(i4b)                    :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='upper_bd_dsqd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(a),i4b)         ,   &
                    int(size(b),i4b) + 1_i4b ,   &
                    int(size(d),i4b)         ,   &
                    int(size(t),i4b)         ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            t(1_i4b) = shift
            d(1_i4b) = a(1_i4b)*a(1_i4b) - shift
!
        end if
!
        return
!
    end if
!
!   GET SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
    gamma  = sqrt( safmin )
!
!   COMPUTE THE QD VARIABLES.
!
    q2(1_i4b:n)       = a(1_i4b:n)*a(1_i4b:n)
    e2(1_i4b:n-1_i4b) = b(1_i4b:n-1_i4b)*b(1_i4b:n-1_i4b)
!
    if ( flip ) then
!
!       APPLY THE DIFFERENTIAL STATIONARY QD ALGORITHM
!       FROM BOTTOM TO TOP.
!
        ti = shift
        di = q2(n) - shift
!
        do i=n, 2_i4b, -1_i4b
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
            t(i) = ti
!
            ti = shift + e2(i-1_i4b)*(ti/di)
            di = q2(i-1_i4b) - ti
!
        end do
!
        d(1_i4b) = di
        t(1_i4b) = ti
!
    else
!
!       APPLY THE DIFFERENTIAL STATIONARY QD ALGORITHM
!       FROM TOP TO BOTTOM.
!
        ti = shift
        di = q2(1_i4b) - shift
!
        do i=1_i4b, n-1_i4b
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
            t(i) = ti
!
            ti = shift + e2(i)*(ti/di)
            di = q2(i+1_i4b) - ti
!
        end do
!
        d(n) = di
        t(n) = ti
!
    end if
!
!
! END OF SUBROUTINE upper_bd_dsqd_b
! _________________________________
!
    end subroutine upper_bd_dsqd_b
!
! =========================================================================================
!
    subroutine upper_bd_dpqd_b( a, b, shift, flip, d, s )
!
! Purpose
! _______
!
!   UPPER_BD_DPQD computes:
!
!   - the L * D * L' factorization of the matrix BD * BD' - shift * I  , if FLIP=false;
!
!   - the U * D * U' factorization of the matrix BD' * BD - shift * I  , if FLIP=true;
!
!   for a n-by-n (upper) bidiagonal matrix BD and a given shift. L and U are, respectively,
!   unit lower and unit upper bidiagonal matrices and D is a diagonal matrix.
!
!   The differential form of the progressive QD algorithm of Rutishauser is used to compute
!   the factorization (see the reference (1) below for further details).
!
!   The subroutine outputs the diagonal matrix D of the factorization and the auxiliary
!   variable S in the differential form of the progressive QD algorithm.
!
!
! Arguments
! _________
!
!   A           (INPUT) real(stnd), dimension(:)
!               On entry, A contains the diagonal elements of the bidiagonal
!               matrix BD.
!
!   B           (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 off-diagonal elements of the bidiagonal matrix BD.
!
!               The size of B must be  size( B ) = size( A ) - 1.
!
!   SHIFT       (INPUT) real(stnd)
!               On entry, the shift.
!
!   FLIP        (INPUT) logical(lgl)
!               On entry, if FLIP=false the L * D * L' factorization of
!               the matrix BD * BD' - shift * I is computed. Otherwise,
!               if FLIP=true the U * D * U' factorization of the matrix
!               BD' * BD - shift * I is computed.
!
!   D           (OUTPUT) real(stnd), dimension(:)
!               On exit, the elements of the diagonal matrix D.
!
!               The size of D must be  size( D ) = size( A ).
!
!   S           (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the auxiliary values S(i) in the differential form
!               of the progressive QD algorithm.
!
!               The size of S must be  size( S ) = size( D ) = size( A ).
!
!
! Further Details
! _______________
!
!   The bidiagonal matrix BD must be scaled appropriately before using this subroutine
!   in order to avoid overflows (see the reference (1) below for further details).
!
!   This subroutine is adapted from the algorithms given in reference (1). See:
!
!   (1) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)  :: shift
    real(stnd), dimension(:),   intent(in)  :: a, b
    real(stnd), dimension(:),   intent(out) :: d, s
!
    logical(lgl),               intent(in)  :: flip
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: gamma, di, si
    real(stnd), dimension(size(a)) :: q2
    real(stnd), dimension(size(b)) :: e2
!
    integer(i4b)                   :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='upper_bd_dpqd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(a),i4b)         ,   &
                    int(size(b),i4b) + 1_i4b ,   &
                    int(size(d),i4b)         ,   &
                    int(size(s),i4b)         ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            s(1_i4b) = a(1_i4b)*a(1_i4b) - shift
            d(1_i4b) = s(1_i4b)
!
        end if
!
        return
!
    end if
!
!   GET SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
    gamma  = sqrt( safmin )
!
!   COMPUTE THE QD VARIABLES.
!
    q2(1_i4b:n)       = a(1_i4b:n)*a(1_i4b:n)
    e2(1_i4b:n-1_i4b) = b(1_i4b:n-1_i4b)*b(1_i4b:n-1_i4b)
!
    if ( flip ) then
!
!       APPLY THE DIFFERENTIAL PROGRESSIVE QD ALGORITHM
!       FROM BOTTOM TO TOP.
!
        si = q2(n) - shift
!
        do i=n, 2_i4b, -1_i4b
!
            di = si + e2(i-1_i4b)
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
            s(i) = si
!
            si = q2(i-1_i4b)*(si/di) - shift
!
        end do
!
        d(1_i4b) = si
        s(1_i4b) = si
!
    else
!
!       APPLY THE DIFFERENTIAL PROGRESSIVE QD ALGORITHM
!       FROM TOP TO BOTTOM.
!
        si = q2(1_i4b) - shift
!
        do i=1_i4b, n-1_i4b
!
            di = si + e2(i)
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            d(i) = di
            s(i) = si
!
            si = q2(i+1_i4b)*(si/di) - shift
!
        end do
!
        d(n) = si
        s(n) = si
!
    end if
!
!
! END OF SUBROUTINE upper_bd_dpqd_b
! _________________________________
!
    end subroutine upper_bd_dpqd_b
!
! =========================================================================================
!
    subroutine upper_bd_dsqd_c( a, b, shift, flip, d, t, l )
!
! Purpose
! _______
!
!   UPPER_BD_DSQD computes:
!
!   - the L * D * L' factorization of the matrix BD' * BD - shift * I  , if FLIP=false;
!
!   - the U * D * U' factorization of the matrix BD * BD' - shift * I  , if FLIP=true;
!
!   for a n-by-n (upper) bidiagonal matrix BD and a given shift. L and U are, respectively,
!   unit lower and unit upper bidiagonal matrices and D is a diagonal matrix.
!
!   The differential form of the stationary QD algorithm of Rutishauser is used to compute
!   the factorization (see the reference (1) below for further details).
!
!   The subroutine outputs the diagonal matrix D of the factorization, the off-diagonal
!   entries of L (or of U if FLIP=true) and the auxiliary variable T in the differential 
!   form of the stationary QD algorithm.
!
!
! Arguments
! _________
!
!   A           (INPUT) real(stnd), dimension(:)
!               On entry, A contains the diagonal elements of the bidiagonal
!               matrix BD.
!
!   B           (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 off-diagonal elements of the bidiagonal matrix BD.
!
!               The size of B must be  size( B ) = size( A ) - 1.
!
!   SHIFT       (INPUT) real(stnd)
!               On entry, the shift.
!
!   FLIP        (INPUT) logical(lgl)
!               On entry, if FLIP=false the L * D * L' factorization of
!               the matrix BD' * BD - shift * I is computed. Otherwise,
!               if FLIP=true the U * D * U' factorization of the matrix
!               BD * BD' - shift * I is computed.
!
!   D           (OUTPUT) real(stnd), dimension(:)
!               On exit, the elements of the diagonal matrix D.
!
!               The size of D must be  size( D ) = size( A ).
!
!   T           (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the auxiliary values T(i) in the differential form
!               of the stationary QD algorithm.
!
!               The size of T must be  size( T ) = size( D ) = size( A ).
!
!   L           (OUTPUT) real(stnd), dimension(:)
!               On exit, the off-diagonal entries of L if FLIP=false or the off-diagonal
!               entries of U if FLIP=true.
!
!               The size of L must be  size( L ) = size( B ) = size( A ) - 1.
!
!
! Further Details
! _______________
!
!   The bidiagonal matrix BD must be scaled appropriately before using this subroutine
!   in order to avoid overflows (see the reference (1) below for further details).
!
!   This subroutine is adapted from the algorithms given in reference (1). See:
!
!   (1) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)  :: shift
    real(stnd), dimension(:),   intent(in)  :: a, b
    real(stnd), dimension(:),   intent(out) :: d, t, l
!
    logical(lgl),               intent(in)  :: flip
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                      :: gamma, di, ti
    real(stnd), dimension(size(a))  :: q2
    real(stnd), dimension(size(b))  :: e2
!
    integer(i4b)                    :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='upper_bd_dsqd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(a),i4b)           ,   &
                       int(size(b),i4b) + 1_i4b   ,   &
                       int(size(d),i4b)           ,   &
                       int(size(t),i4b)           ,   &
                       int(size(l),i4b) + 1_i4b /),   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            t(1_i4b) = shift
            d(1_i4b) = a(1_i4b)*a(1_i4b) - shift
!
        end if
!
        return
!
    end if
!
!   GET SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
    gamma  = sqrt( safmin )
!
!   COMPUTE THE QD VARIABLES.
!
    q2(1_i4b:n)       = a(1_i4b:n)*a(1_i4b:n)
    e2(1_i4b:n-1_i4b) = b(1_i4b:n-1_i4b)*b(1_i4b:n-1_i4b)
!
    if ( flip ) then
!
!       APPLY THE DIFFERENTIAL STATIONARY QD ALGORITHM
!       FROM BOTTOM TO TOP.
!
        ti = shift
        di = q2(n) - shift
!
        do i=n, 2_i4b, -1_i4b
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            l(i-1_i4b) = (b(i-1_i4b)*a(i))/di
!            l(i-1_i4b) = sign(sign(sqrt(e(i-1_i4b)*q(i))/di,b(i-1_i4b)),a(i))
!
            d(i) = di
            t(i) = ti
!
            ti = shift + e2(i-1_i4b)*(ti/di)
            di = q2(i-1_i4b) - ti
!
        end do
!
        d(1_i4b) = di
        t(1_i4b) = ti
!
    else
!
!       APPLY THE DIFFERENTIAL STATIONARY QD ALGORITHM
!       FROM TOP TO BOTTOM.
!
        ti = shift
        di = q2(1_i4b) - shift
!
        do i=1_i4b, n-1_i4b
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            l(i) = (b(i)*a(i))/di
!            l(i) = sign(sign(sqrt(e(i)*q(i))/di,b(i)),a(i))
!
            d(i) = di
            t(i) = ti
!
            ti = shift + e2(i)*(ti/di)
            di = q2(i+1_i4b) - ti
!
        end do
!
        d(n) = di
        t(n) = ti
!
    end if
!
!
! END OF SUBROUTINE upper_bd_dsqd_c
! _________________________________
!
    end subroutine upper_bd_dsqd_c
!
! =========================================================================================
!
    subroutine upper_bd_dpqd_c( a, b, shift, flip, d, s, l )
!
! Purpose
! _______
!
!   UPPER_BD_DPQD computes:
!
!   - the L * D * L' factorization of the matrix BD * BD' - shift * I  , if FLIP=false;
!
!   - the U * D * U' factorization of the matrix BD' * BD - shift * I  , if FLIP=true;
!
!   for a n-by-n (upper) bidiagonal matrix BD and a given shift. L and U are, respectively,
!   unit lower and unit upper bidiagonal matrices and D is a diagonal matrix.
!
!   The differential form of the progressive QD algorithm of Rutishauser is used to compute
!   the factorization (see the reference (1) below for further details).
!
!   The subroutine outputs the diagonal matrix D of the factorization, the off-diagonal
!   entries of L (or of U if FLIP=true) and the auxiliary variable S in the differential 
!   form of the progressive QD algorithm.
!
!
! Arguments
! _________
!
!   A           (INPUT) real(stnd), dimension(:)
!               On entry, A contains the diagonal elements of the bidiagonal
!               matrix BD.
!
!   B           (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 off-diagonal elements of the bidiagonal matrix BD.
!
!               The size of B must be  size( B ) = size( A ) - 1.
!
!   SHIFT       (INPUT) real(stnd)
!               On entry, the shift.
!
!   FLIP        (INPUT) logical(lgl)
!               On entry, if FLIP=false the L * D * L' factorization of
!               the matrix BD * BD' - shift * I is computed. Otherwise,
!               if FLIP=true the U * D * U' factorization of the matrix
!               BD' * BD - shift * I is computed.
!
!   D           (OUTPUT) real(stnd), dimension(:)
!               On exit, the elements of the diagonal matrix D.
!
!               The size of D must be  size( D ) = size( A ).
!
!   S           (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the auxiliary values S(i) in the differential form
!               of the progressive QD algorithm.
!
!               The size of S must be  size( S ) = size( D ) = size( A ).
!
!   L           (OUTPUT) real(stnd), dimension(:)
!               On exit, the off-diagonal entries of L if FLIP=false or the off-diagonal
!               entries of U if FLIP=true.
!
!               The size of L must be  size( L ) = size( B ) = size( A ) - 1.
!
!
! Further Details
! _______________
!
!   The bidiagonal matrix BD must be scaled appropriately before using this subroutine
!   in order to avoid overflows (see the reference (1) below for further details).
!
!   This subroutine is adapted from the algorithms given in reference (1). See:
!
!   (1) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)  :: shift
    real(stnd), dimension(:),   intent(in)  :: a, b
    real(stnd), dimension(:),   intent(out) :: d, s, l
!
    logical(lgl),               intent(in)  :: flip
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: gamma, di, si
    real(stnd), dimension(size(a)) :: q2
    real(stnd), dimension(size(b)) :: e2
!
    integer(i4b)                   :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='upper_bd_dpqd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(a),i4b)           ,   &
                       int(size(b),i4b) + 1_i4b   ,   &
                       int(size(d),i4b)           ,   &
                       int(size(l),i4b)           ,   &
                       int(size(s),i4b) + 1_i4b /),   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            s(1_i4b) = a(1_i4b)*a(1_i4b) - shift
            d(1_i4b) = s(1_i4b)
!
        end if
!
        return
!
    end if
!
!   GET SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
    gamma  = sqrt( safmin )
!
!   COMPUTE THE QD VARIABLES.
!
    q2(1_i4b:n)       = a(1_i4b:n)*a(1_i4b:n)
    e2(1_i4b:n-1_i4b) = b(1_i4b:n-1_i4b)*b(1_i4b:n-1_i4b)
!
    if ( flip ) then
!
!       APPLY THE DIFFERENTIAL PROGRESSIVE QD ALGORITHM
!       FROM BOTTOM TO TOP.
!
        si = q2(n) - shift
!
        do i=n, 2_i4b, -1_i4b
!
            di = si + e2(i-1_i4b)
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            l(i-1_i4b) = (b(i-1_i4b)*a(i-1_i4b))/di
!            l(i-1_i4b) = sign(sign(sqrt(e(i-1_i4b)*q(i-1_i4b))/di,b(i-1_i4b)),a(i-1_i4b))
!
            d(i) = di
            s(i) = si
!
            si = q2(i-1_i4b)*(si/di) - shift
!
        end do
!
        d(1_i4b) = si
        s(1_i4b) = si
!
    else
!
!       APPLY THE DIFFERENTIAL PROGRESSIVE QD ALGORITHM
!       FROM TOP TO BOTTOM.
!
        si = q2(1_i4b) - shift
!
        do i=1_i4b, n-1_i4b
!
            di = si + e2(i)
!
            if ( abs(di)<gamma ) then
                di = -gamma
            end if
!
            l(i) = (b(i)*a(i+1_i4b))/di
!            l(i) = sign(sign(sqrt(e(i)*q(i+1_i4b))/di,b(i)),a(i+1_i4b))
!
            d(i) = di
            s(i) = si
!
            si = q2(i+1_i4b)*(si/di) - shift
!
        end do
!
        d(n) = si
        s(n) = si
!
    end if
!
!
! END OF SUBROUTINE upper_bd_dpqd_c
! _________________________________
!
    end subroutine upper_bd_dpqd_c
!
! =========================================================================================
!
    subroutine dflgen_bd( d, e, lambda, cs_left, sn_left, cs_right, sn_right, scaling )
!
! Purpose
! _______
!
!   DFLGEN_BD computes deflation parameters (e.g. two chains of Givens rotations) for
!   a n-by-n (upper) bidiagonal matrix BD and a given singular value of BD.
!
!   On output, the arguments CS_LEFT, SN_LEFT, CS_RIGHT and SN_RIGHT contain, respectively,
!   the vectors of the cosines and sines coefficients of the chain of n-1 planar rotations that
!   deflates the real n-by-n bidiagonal matrix BD corresponding to a singular value LAMBDA.
!
!
!
! Arguments
! _________
!
!   D           (INPUT) real(stnd), dimension(:)
!               On entry, D contains the diagonal elements of the bidiagonal
!               matrix BD.
!
!   E           (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 off-diagonal elements of the bidiagonal matrix BD.
!
!               The size of E must be  size( E ) = size( D ) - 1.
!
!   LAMBDA      (INPUT) real(stnd)
!               On entry, a singular value of the bidiagonal matrix BD.
!
!   CS_LEFT     (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations that deflates the bidiagonal matrix BD on the left.
!
!               The size of CS_LEFT must be  size( CS_LEFT ) = size( E ) = size( D ) - 1.
!
!   SN_LEFT     (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations that deflates the bidiagonal matrix BD on the left.
!
!               The size of SN_LEFT must be  size( SN_LEFT ) = size( E ) = size( D ) - 1.
!
!   CS_RIGHT    (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations that deflates the bidiagonal matrix BD on the right.
!
!               The size of CS_RIGHT must be  size( CS_RIGHT ) = size( E ) = size( D ) - 1.
!
!   SN_RIGHT    (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations that deflates the bidiagonal matrix BD on the right.
!
!               The size of SN_RIGHT must be  size( SN_RIGHT ) = size( E ) = size( D ) - 1.
!
!   SCALING     (INPUT, OPTIONAL) logical(lgl)
!               On entry, if SCALING=true the bidiagonal matrix BD is scaled before computing
!               the deflation parameters in order to avoid overflows.
!
!               The default is to scale the bidiagonal matrix.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the matlab routine DFLGEN in the reference (1) and
!   algorithms given in reference (2).
!
!   For further details, see:
!
!   (1) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (2) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : machtiny
    use Utilities,        only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)  :: lambda
    real(stnd), dimension(:),   intent(in)  :: d, e
    real(stnd), dimension(:),   intent(out) :: cs_left, sn_left, cs_right, sn_right
!
    logical(lgl), intent(in), optional :: scaling
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: lambda2, maxbd, maxd, maxe, maxde, maxde2, tau, &
                                      gamma, omega, scalfac, cf, sfi, f, xim, u1, l2
    real(stnd), dimension(size(d)) :: diag, q2, d1, d2, s1, s2
    real(stnd), dimension(size(e)) :: sub, e2, dl, sf
!
    integer(i4b)                     :: n, i, k, j, j1, j2, j3, ei, ce, sei
    integer(i4b), dimension(size(e)) :: se
!
    logical(lgl) :: do_scaling
!
#ifdef _OPENMP
!
    logical :: test_par, ompparallel
!
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='dflgen_bd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(d),i4b)                  ,   &
                       int(size(e),i4b)  + 1_i4b         ,   &
                       int(size(cs_left),i4b) + 1_i4b    ,   &
                       int(size(sn_left),i4b) + 1_i4b    ,   &
                       int(size(cs_right),i4b) + 1_i4b   ,   &
                       int(size(sn_right),i4b) + 1_i4b /),   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   MAKE A LOCAL COPY OF THE BIDIAGONAL MATRIX.
!
    diag(1_i4b:n)      = d(1_i4b:n)
    sub(1_i4b:n-1_i4b) = e(1_i4b:n-1_i4b)
!
    maxd  = maxval( abs( diag(1_i4b:n) ) )
    maxe  = maxval( abs( sub(1_i4b:n-1_i4b) ) )
    maxbd = max( maxd, maxe )
!
    if ( maxbd<=zero ) then
!
        cs_left(1_i4b:n-1_i4b)  = one
        sn_left(1_i4b:n-1_i4b)  = zero
!
        cs_right(1_i4b:n-1_i4b) = one
        sn_right(1_i4b:n-1_i4b) = zero
!
        return
!
    end if
!
    do_scaling = true
    if ( present(scaling) ) then
        do_scaling = scaling
    end if
!
!   GET MACHINE PARAMETERS.
!
    gamma  = sqrt( safmin )
!
    if ( do_scaling ) then
!
        maxde  = maxval( abs( diag(1_i4b:n-1_i4b)*sub(1_i4b:n-1_i4b) ) )
        maxde2 = maxval( abs( diag(2_i4b:n)*sub(1_i4b:n-1_i4b) ) )
!
!       COMPUTE SCALING FACTOR TO AVOID OVERFLOW.
!
#ifdef _SCALE
        omega  = scale(one,maxexp-1_i4b)
#else
        omega  = machbase**(maxexp-1_i4b)
#endif
        tau    = sqrt( omega*gamma )
!
        j  = exponent( tau )
        j1 = j - exponent( maxbd ) - 1_i4b
        j2 = ( j - exponent( maxde ) - 1_i4b )/2_i4b
        j3 = ( j - exponent( maxde2 ) - 1_i4b )/2_i4b
        j  = min( j1, j2, j3 )
!
        scalfac = machbase**j
!
        lambda2 = (scalfac*lambda)**2
!
    else
!
        scalfac = one
!
        lambda2 = lambda*lambda
!
    end if
!
!   COMPUTE AND SCALE THE QD VARIABLES ASSOCIATED WITH THE BIDIAGONAL MATRIX.
!   IF ANY ELEMENT OF abs(d(:)) OR abs(e(:)) IS LESS THAN gamma/scalfac THAN THE CORRESPONDING
!   SCALED AND SQUARED ELEMENTS (EG THE QD VARIABLES) WILL UNDERFLOW TO zero AND
!   HIGH RELATIVE ACCURACY WILL BE LOST.
!
    tau = gamma/scalfac
!
    where( abs( diag(1_i4b:n) )>=tau )
        diag(1_i4b:n) = scalfac*diag(1_i4b:n)
        q2(1_i4b:n)   = diag(1_i4b:n)*diag(1_i4b:n)
    elsewhere
        diag(1_i4b:n) = sign( gamma, diag(1_i4b:n) )
        q2(1_i4b:n)   = safmin
    end where
!
    where( abs( sub(1_i4b:n-1_i4b) )>=tau )
        sub(1_i4b:n-1_i4b) = scalfac*sub(1_i4b:n-1_i4b)
        e2(1_i4b:n-1_i4b)  = sub(1_i4b:n-1_i4b)*sub(1_i4b:n-1_i4b)
    elsewhere
        sub(1_i4b:n-1_i4b) = sign( gamma, sub(1_i4b:n-1_i4b) )
        e2(1_i4b:n-1_i4b)  = safmin
    end where
!    
!   FERNANDO METHOD.  
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    test_par    = .not.ompparallel .and. i>1_i4b
#endif
!
!$OMP PARALLEL SECTIONS IF(test_par), NUM_THREADS( 2 ),                                  &
!$OMP                   PRIVATE(d1,d2,s1,s2,dl,sf,u1,l2,se,tau,xim,f,cf,ce,sfi,sei,ei,k,i)
!
!$OMP SECTION
!
!   COMPUTE RIGHT DEFLATION PARAMETERS.
!
!   L1*D1*U1 FACTORIZATION OF BD'*BD - lambda2*I .
!
    call upper_bd_dsqd2_b( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, false, d1(1_i4b:n), s1(1_i4b:n) )
!
!   U2*D2*L2 FACTORIZATION OF BD'*BD - lambda2*I .
!
    call upper_bd_dpqd2_b( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, true, d2(1_i4b:n), s2(1_i4b:n) )
!
!   FIND OPTIMAL k AND tau .
!
    tau = abs( s2(1_i4b) - s1(1_i4b) + lambda2 )
    k = 1_i4b
!
    do i=2_i4b, n
!
        cf = abs( s2(i) - s1(i) + lambda2 )
        if ( cf>=tau ) cycle
        tau = cf
        k   = i
!
    end do
!
!   LDU FACTORIZATION OF BD'*BD - lambda2*I .
!
!    call upper_bd_dsqd2_a( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, false, d1(1_i4b:n) )
!
!   UDL FACTORIZATION OF BD'*BD - lambda2*I .
!
!    call upper_bd_dpqd2_a( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, true, d2(1_i4b:n) )
!
!   FIND OPTIMAL k AND tau .
!
!    tau = abs( (d1(1_i4b) - q2(1_i4b)) + d2(1_i4b) + lambda2 )
!    k = 1_i4b
!
!    do i=2_i4b, n
!
!        cf = abs( (d1(i) - q2(i)) + (d2(i) - e2(i-1_i4b)) + lambda2 )
!        if ( cf>=tau ) cycle
!        tau = cf
!        k   = i
!
!    end do
!
!   COMPUTE OFF-DIAGONAL ELEMENTS OF BD'*BD.
!
    dl(1_i4b:n-1_i4b) = diag(1_i4b:n-1_i4b)*sub(1_i4b:n-1_i4b)
!
!    dl(1_i4b:n-1_i4b) = sqrt( q2(1_i4b:n-1_i4b)*e2(1_i4b:n-1_i4b) )
!    dl(1_i4b:n-1_i4b) = sign( dl(1_i4b:n-1_i4b), diag(1_i4b:n-1_i4b) )
!    dl(1_i4b:n-1_i4b) = sign( dl(1_i4b:n-1_i4b), sub(1_i4b:n-1_i4b) )
!
!   FROM HERE USE FLOATING POINT NUMBERS WITH EXTENDED
!   EXPONENTS FOR THE COMPUTATIONS.
!
    do i=1_i4b, k-1_i4b
!
        u1 = dl(i)/d1(i)
!
        if ( abs(u1)>machtiny ) then
!
!           REPRESENT u1 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = fraction( u1 )
            se(i) = exponent( u1 )
!
        else
!
!           RECOMPUTE u1 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = fraction( dl(i) )/fraction( d1(i) )
            se(i) = exponent( dl(i) ) - exponent( d1(i) )
!
        end if
!
    end do
!
    do i=k, n-1_i4b
!
        l2 = dl(i)/d2(i+1_i4b)
!
        if ( abs(l2)>machtiny ) then
!
!           REPRESENT 1/l2 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = one/fraction( l2 )
            se(i) = -exponent( l2 )                
!
        else
!
!           RECOMPUTE 1/l2 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = fraction( d2(i+1_i4b) )/fraction( dl(i) )
            se(i) = exponent( d2(i+1_i4b) ) - exponent( dl(i) )
!
        end if
!
    end do
!
!    u1(1_i4b:k-1_i4b) = dl(1_i4b:k-1_i4b)/d1(1_i4b:k-1_i4b)
!    l2(k:n-1_i4b)     = dl(k:n-1_i4b)/d2(k+1_i4b:n)
!
!   FROM HERE USE FLOATING POINT NUMBERS WITH EXTENDED
!   EXPONENTS FOR THE COMPUTATIONS.
!
!    where ( abs(u1(1_i4b:k-1_i4b))>machtiny )
!
!       REPRESENT u1(:k-1) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(1_i4b:k-1_i4b) = fraction( u1(1_i4b:k-1_i4b) )
!        se(1_i4b:k-1_i4b) = exponent( u1(1_i4b:k-1_i4b) )
!
!    elsewhere
!
!       RECOMPUTE u1(:k-1) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(1_i4b:k-1_i4b) = fraction( dl(1_i4b:k-1_i4b) )/fraction( d1(1_i4b:k-1_i4b) )
!        se(1_i4b:k-1_i4b) = exponent( dl(1_i4b:k-1_i4b) ) - exponent( d1(1_i4b:k-1_i4b) )
!
!    end where    
!
!    where ( abs(l2(k:n-1_i4b))>machtiny )
!
!       REPRESENT 1/l2(k:) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(k:n-1_i4b) = one/fraction( l2(k:n-1_i4b) )
!        se(k:n-1_i4b) = -exponent( l2(k:n-1_i4b) )                
!
!    elsewhere
!
!       RECOMPUTE 1/l2(k:) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(k:n-1_i4b) = fraction( d2(k+1_i4b:n) )/fraction( dl(k:n-1_i4b) )
!        se(k:n-1_i4b) = exponent( d2(k+1_i4b:n) ) - exponent( dl(k:n-1_i4b) )
!
!    end where
!
    cf = fraction( one )
    ce = exponent( one )
!
    do i=1_i4b, n-1_i4b
!
!       CONVERT cf*(base**ce) TO cs_right(i) AND sf(i)*(base**se(i)) TO sn_right(i).
!
        sfi = sf(i)
        sei = se(i)
        ei  = max( ce, sei )
        ce  = ce - ei
        sei = sei - ei
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
#ifdef _SCALE
        f   = sqrt( (scale(cf,ce))**2 + (scale(sfi,sei))**2 )
#else
        f   = sqrt( (cf*(machbase**ce))**2 + (sfi*(machbase**sei))**2 )
#endif
!
!       SAVE cs_right(i) IN THE FORM WITH EXTENDED EXPONENTS.
!
        xim = cf/f
        cf  = fraction( xim )
        ce  = ce + exponent( xim )
!
!       SAVE COSINES AND SINES OF GIVENS ROTATIONS.
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
#ifdef _SCALE
        cs_right(i) = scale( cf, ce )
        sn_right(i) = scale( sfi/f, sei )
#else
        cs_right(i) = cf*(machbase**ce)
        sn_right(i) = (sfi/f)*(machbase**sei)
#endif
!
    end do
!
!$OMP SECTION
!
!   COMPUTE LEFT DEFLATION PARAMETERS.
!
!   L1*D1*U1 FACTORIZATION OF BD*BD' - lambda2*I .
!
    call upper_bd_dpqd2_b( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, false, d1(1_i4b:n), s1(1_i4b:n)  )
!
!   U2*D2*L2 FACTORIZATION OF BD*BD' - lambda2*I .
!
    call upper_bd_dsqd2_b( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, true, d2(1_i4b:n), s2(1_i4b:n) )
!
!   FIND OPTIMAL k AND tau .
!
    tau = abs( s1(n) - s2(n) + lambda2 )
    k = n
!
    do i=n-1_i4b, 1_i4b, -1_i4b
!
        cf = abs( s1(i) - s2(i) + lambda2 )
        if ( cf>=tau ) cycle
        tau = cf
        k   = i
!
    end do
!
!   LDU FACTORIZATION OF BD*BD' - lambda2*I .
!
!    call upper_bd_dpqd2_a( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, false, d1(1_i4b:n) )
!
!   UDL FACTORIZATION OF BD*BD' - lambda2*I .
!
!    call upper_bd_dsqd2_a( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, true, d2(1_i4b:n) )
!
!   FIND OPTIMAL k AND tau .
!
!    tau = abs( (d1(n) - q2(n)) + d2(n) + lambda2 )
!    k = n
!
!    do i=n-1_i4b, 1_i4b, -1_i4b
!
!        cf = abs( (d1(i) - q2(i)) + (d2(i) - e2(i)) + lambda2 )
!        if ( cf>=tau ) cycle
!        tau = cf
!        k   = i
!
!    end do
!
!   COMPUTE OFF-DIAGONAL ELEMENTS OF BD*BD'.
!
    dl(1_i4b:n-1_i4b) = diag(2_i4b:n)*sub(1_i4b:n-1_i4b)
!
!    dl(1_i4b:n-1_i4b) = sqrt( q2(2_i4b:n)*e2(1_i4b:n-1_i4b) )
!    dl(1_i4b:n-1_i4b) = sign( dl(1_i4b:n-1_i4b), diag(2_i4b:n) )
!    dl(1_i4b:n-1_i4b) = sign( dl(1_i4b:n-1_i4b), sub(1_i4b:n-1_i4b) )
!
!   FROM HERE USE FLOATING POINT NUMBERS WITH EXTENDED
!   EXPONENTS FOR THE COMPUTATIONS IN ORDER TO AVOID UNDERFLOWS.
!
    do i=1_i4b, k-1_i4b
!
        u1 = dl(i)/d1(i)
!
        if ( abs(u1)>machtiny ) then
!
!           REPRESENT u1 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = fraction( u1 )
            se(i) = exponent( u1 )
!
        else
!
!           RECOMPUTE u1 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = fraction( dl(i) )/fraction( d1(i) )
            se(i) = exponent( dl(i) ) - exponent( d1(i) )
!
        end if
!
    end do
!
    do i=k, n-1_i4b
!
        l2 = dl(i)/d2(i+1_i4b)
!
        if ( abs(l2)>machtiny ) then
!
!           REPRESENT 1/l2 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = one/fraction( l2 )
            se(i) = -exponent( l2 )                
!
        else
!
!           RECOMPUTE 1/l2 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = fraction( d2(i+1_i4b) )/fraction( dl(i) )
            se(i) = exponent( d2(i+1_i4b) ) - exponent( dl(i) )
!
        end if
!
    end do
!
!    u1(1_i4b:k-1_i4b) = dl(1_i4b:k-1_i4b)/d1(1_i4b:k-1_i4b)
!    l2(k:n-1_i4b)     = dl(k:n-1_i4b)/d2(k+1_i4b:n)
!
!   FROM HERE USE FLOATING POINT NUMBERS WITH EXTENDED
!   EXPONENTS FOR THE COMPUTATIONS IN ORDER TO AVOID UNDERFLOWS.
!
!    where ( abs(u1(1_i4b:k-1_i4b))>machtiny )
!
!       REPRESENT u1(:k-1) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(1_i4b:k-1_i4b) = fraction( u1(1_i4b:k-1_i4b) )
!        se(1_i4b:k-1_i4b) = exponent( u1(1_i4b:k-1_i4b) )
!
!    elsewhere
!
!       RECOMPUTE u1(:k-1) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(1_i4b:k-1_i4b) = fraction( dl(1_i4b:k-1_i4b) )/fraction( d1(1_i4b:k-1_i4b) )
!        se(1_i4b:k-1_i4b) = exponent( dl(1_i4b:k-1_i4b) ) - exponent( d1(1_i4b:k-1_i4b) )
!
!    end where    
!
!    where ( abs(l2(k:n-1_i4b))>machtiny )
!
!       REPRESENT 1/l2(k:) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(k:n-1_i4b) = one/fraction( l2(k:n-1_i4b) )
!        se(k:n-1_i4b) = -exponent( l2(k:n-1_i4b) )                
!
!    elsewhere
!
!       RECOMPUTE 1/l2(k:) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(k:n-1_i4b) = fraction( d2(k+1_i4b:n) )/fraction( dl(k:n-1_i4b) )
!        se(k:n-1_i4b) = exponent( d2(k+1_i4b:n) ) - exponent( dl(k:n-1_i4b) )
!
!    end where
!
    cf = fraction( one )
    ce = exponent( one )
!
    do i=1_i4b, n-1_i4b
!
!       CONVERT cf*(base**ce) TO cs_right(i) AND sf(i)*(base**se(i)) TO sn_left(i).
!
        sfi = sf(i)
        sei = se(i)
        ei  = max( ce, sei )
        ce  = ce - ei
        sei = sei - ei
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
#ifdef _SCALE
        f = sqrt( (scale(cf,ce))**2 + (scale(sfi,sei))**2 )
#else
        f = sqrt( (cf*(machbase**ce))**2 + (sfi*(machbase**sei))**2 )
#endif
!
!       SAVE cs_left(i) IN THE FORM WITH EXTENDED EXPONENTS.
!
        xim = cf/f
        cf  = fraction( xim )
        ce  = ce + exponent( xim )
!
!       SAVE COSINES AND SINES OF GIVENS ROTATIONS.
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
#ifdef _SCALE
        cs_left(i) = scale( cf, ce )
        sn_left(i) = scale( sfi/f, sei )
#else
        cs_left(i) = cf*(machbase**ce)
        sn_left(i) = (sfi/f)*(machbase**sei)
#endif
!
    end do
!
!$OMP END PARALLEL SECTIONS
!
!
! END OF SUBROUTINE dflgen_bd
! ___________________________
!
    end subroutine dflgen_bd
!
! =========================================================================================
!
    subroutine dflgen2_bd( d, e, lambda, cs_left, sn_left, cs_right, sn_right, deflate, scaling )
!
! Purpose
! _______
!
!   DFLGEN2_BD computes and applies deflation parameters (e.g. two chains of Givens rotations) for
!   a n-by-n (upper) bidiagonal matrix BD and a given singular value of BD.
!
!   On input:
!
!       The arguments D and E contain, respectively, the main diagonal
!       and off-diagonal of the bidiagonal matrix, and the argument LAMBDA
!       contains an estimate of the singular value.
!
!   On output:
!
!       The arguments D and E contain, respectively, the new main diagonal
!       and off-diagonal of the deflated bidiagonal matrix if DEFLATE is set
!       to true, otherwise D and E are not changed.
!
!       The arguments CS_LEFT, SN_LEFT, CS_RIGHT and SN_RIGHT contain, respectively,
!       the vectors of the cosines and sines coefficients of the chain of n-1 planar
!       rotations that deflates the real n-by-n bidiagonal matrix BD corresponding to
!       the singular value LAMBDA. One chain is applied to the left of BD (CS_LEFT, SN_LEFT)
!       and the other is applied to the right of BD (CS_RIGHT, SN_RIGHT).
!
!
! Arguments
! _________
!
!   D           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, D contains the diagonal elements of the bidiagonal
!               matrix BD.
!
!               On exit, the new main diagonal of the bidiagonal matrix
!               if DEFLATE=true. Otherwise, D is not changed.
!
!   E           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the n-1 off-diagonal elements of the bidiagonal matrix BD.
!
!               On exit, the new off-diagonal of the bidiagonal matrix if DEFLATE=true.
!               Otherwise, E is not changed.
!
!               The size of E must be  size( E ) = size( D ) - 1.
!
!   LAMBDA      (INPUT) real(stnd)
!               On entry, a singular value of the bidiagonal matrix BD.
!
!   CS_LEFT     (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations that deflates the bidiagonal matrix BD on the left.
!
!               The size of CS_LEFT must be  size( CS_LEFT ) = size( E ) = size( D ) - 1.
!
!   SN_LEFT     (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations that deflates the bidiagonal matrix BD on the left.
!
!               The size of SN_LEFT must be  size( SN_LEFT ) = size( E ) = size( D ) - 1.
!
!   CS_RIGHT    (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations that deflates the bidiagonal matrix BD on the right.
!
!               The size of CS_RIGHT must be  size( CS_RIGHT ) = size( E ) = size( D ) - 1.
!
!   SN_RIGHT    (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations that deflates the bidiagonal matrix BD on the right.
!
!               The size of SN_RIGHT must be  size( SN_RIGHT ) = size( E ) = size( D ) - 1.
!
!   DEFLATE     (OUTPUT) logical(lgl)
!               On exit:
!
!               - DEFLATE = true :  indicates successful exit.
!               - DEFLATE = false:  indicates that full accuracy was not attained
!                 in the deflation of the bidiagonal matrix.
!
!   SCALING     (INPUT, OPTIONAL) logical(lgl)
!               On entry, if SCALING=true the bidiagonal matrix BD is scaled before computing
!               the deflation parameters in order to avoid overflows.
!
!               The default is to scale the bidiagonal matrix.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the matlab routine DFLGEN in the reference (1) and
!   algorithms given in reference (2).
!
!   For further details, see:
!
!   (1) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (2) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : machtiny
    use Utilities,        only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)    :: lambda
    real(stnd), dimension(:),   intent(inout) :: d, e
    real(stnd), dimension(:),   intent(out)   :: cs_left, sn_left, cs_right, sn_right
!
    logical(lgl), intent(in), optional :: scaling
    logical(lgl), intent(out)          :: deflate
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: eps, lambda2, maxbd, maxd, maxe, maxde, maxde2, tau, gamma,   &
                                      omega, scalfac, cf, sfi, f, x, cs, sn, g, h, h_max, y, u1, l2
    real(stnd), dimension(size(d)) :: diag, q2, d1, d2, s1, s2
    real(stnd), dimension(size(e)) :: sub, e2, dl, sf
!
    integer(i4b)                     :: n, i, im1, k, j, j1, j2, j3, ei, ce, sei
    integer(i4b), dimension(size(e)) :: se
!
    logical(lgl) :: do_scaling
!
#ifdef _OPENMP
!
    logical :: test_par, ompparallel
!
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='dflgen2_bd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(d),i4b)                  ,   &
                       int(size(e),i4b)  + 1_i4b         ,   &
                       int(size(cs_left),i4b) + 1_i4b    ,   &
                       int(size(sn_left),i4b) + 1_i4b    ,   &
                       int(size(cs_right),i4b) + 1_i4b   ,   &
                       int(size(sn_right),i4b) + 1_i4b /),   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   COMPUTE THE NORM OF THE BIDIAGONAL MATRIX.
!
    d1(1_i4b:n) = abs( d(1_i4b:n) )
    s1(1_i4b)   = zero
    s1(2_i4b:n) = abs( e(1_i4b:n-1_i4b) )
!
    maxbd = maxval( d1(1_i4b:n) + s1(1_i4b:n) )
!
!   TEST IF THE BIDIAGONAL MATRIX IS THE ZERO MATRIX.
!
    if ( maxbd<=zero ) then
!
        cs_left(1_i4b:n-1_i4b)  = one
        sn_left(1_i4b:n-1_i4b)  = zero
!
        cs_right(1_i4b:n-1_i4b) = one
        sn_right(1_i4b:n-1_i4b) = zero
!
        return
!
    end if
!
    do_scaling = true
!
    if ( present(scaling) ) then
!
        do_scaling = scaling
!
    end if
!
!   DETERMINE THE UNIT ROUNDOFF AND UNDERFLOW THRESHOLD.
!
    eps = macheps*maxbd
!
    gamma  = sqrt( safmin )
!
    if ( do_scaling ) then
!
        maxd  = maxval( d1(1_i4b:n) )
        maxe  = maxval( s1(2_i4b:n) )
        maxbd = max( maxd, maxe )
!
        maxde  = maxval( d1(1_i4b:n-1_i4b)*s1(2_i4b:n) )
        maxde2 = maxval( d1(2_i4b:n)*s1(2_i4b:n) )
!
!       COMPUTE SCALING FACTOR TO AVOID OVERFLOW.
!
#ifdef _SCALE
        omega  = scale(one,maxexp-1_i4b)
#else
        omega  = machbase**(maxexp-1_i4b)
#endif
        tau    = sqrt( omega*gamma )
!
        j  = exponent( tau )
        j1 = j - exponent( maxbd ) - 1_i4b
        j2 = ( j - exponent( maxde ) - 1_i4b )/2_i4b
        j3 = ( j - exponent( maxde2 ) - 1_i4b )/2_i4b
        j  = min( j1, j2, j3 )
!
        scalfac = machbase**j
!
        lambda2 = (scalfac*lambda)**2
!
    else
!
        scalfac = one
!
        lambda2 = lambda*lambda
!
    end if
!
!   COMPUTE AND SCALE THE QD VARIABLES ASSOCIATED WITH THE BIDIAGONAL MATRIX.
!   IF ANY ELEMENT OF abs(d(:)) OR abs(e(:)) IS LESS THAN gamma/scalfac THAN THE CORRESPONDING
!   SCALED AND SQUARED ELEMENTS (EG THE QD VARIABLES) WILL UNDERFLOW TO zero AND
!   HIGH RELATIVE ACCURACY WILL BE LOST.
!
    tau = gamma/scalfac
!
    where( d1(1_i4b:n)>=tau )
        diag(1_i4b:n) = scalfac*d(1_i4b:n)
        q2(1_i4b:n)   = diag(1_i4b:n)*diag(1_i4b:n)
    elsewhere
        diag(1_i4b:n) = sign( gamma, d(1_i4b:n) )
        q2(1_i4b:n)   = safmin
    end where
!
    where( s1(2_i4b:n)>=tau )
        sub(1_i4b:n-1_i4b) = scalfac*e(1_i4b:n-1_i4b)
        e2(1_i4b:n-1_i4b)  = sub(1_i4b:n-1_i4b)*sub(1_i4b:n-1_i4b)
    elsewhere
        sub(1_i4b:n-1_i4b) = sign( gamma, e(1_i4b:n-1_i4b) )
        e2(1_i4b:n-1_i4b)  = safmin
    end where
!    
!   FERNANDO METHOD.
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    test_par    = .not.ompparallel .and. i>1_i4b
#endif
!
!$OMP PARALLEL SECTIONS IF(test_par), NUM_THREADS( 2 ),                                &
!$OMP                   PRIVATE(d1,d2,s1,s2,dl,sf,u1,l2,se,tau,x,f,cf,ce,sfi,sei,ei,k,i)
!
!$OMP SECTION
!
!   COMPUTE RIGHT DEFLATION PARAMETERS.
!
!   L1*D1*U1 FACTORIZATION OF BD'*BD - lambda2*I .
!
    call upper_bd_dsqd2_b( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, false, d1(1_i4b:n), s1(1_i4b:n) )
!
!   U2*D2*L2 FACTORIZATION OF BD'*BD - lambda2*I .
!
    call upper_bd_dpqd2_b( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, true, d2(1_i4b:n), s2(1_i4b:n) )
!
!   FIND OPTIMAL k AND tau .
!
    tau = abs( s2(1_i4b) - s1(1_i4b) + lambda2 )
    k = 1_i4b
!
    do i=2_i4b, n
!
        cf = abs( s2(i) - s1(i) + lambda2 )
!
        if ( cf>=tau ) cycle
!
        tau = cf
        k   = i
!
    end do
!
!   LDU FACTORIZATION OF BD'*BD - lambda2*I .
!
!    call upper_bd_dsqd2_a( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, false, d1(1_i4b:n) )
!
!   UDL FACTORIZATION OF BD'*BD - lambda2*I .
!
!    call upper_bd_dpqd2_a( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, true, d2(1_i4b:n) )
!
!   FIND OPTIMAL k AND tau .
!
!    tau = abs( (d1(1_i4b) - q2(1_i4b)) + d2(1_i4b) + lambda2 )
!    k = 1_i4b
!
!    do i=2_i4b, n
!
!        cf = abs( (d1(i) - q2(i)) + (d2(i) - e2(i-1_i4b)) + lambda2 )
!        if ( cf>=tau ) cycle
!        tau = cf
!        k   = i
!
!    end do
!
!   COMPUTE OFF-DIAGONAL ELEMENTS OF BD'*BD.
!
    dl(1_i4b:n-1_i4b) = diag(1_i4b:n-1_i4b)*sub(1_i4b:n-1_i4b)
!
!    dl(1_i4b:n-1_i4b) = sqrt( q2(1_i4b:n-1_i4b)*e2(1_i4b:n-1_i4b) )
!    dl(1_i4b:n-1_i4b) = sign( dl(1_i4b:n-1_i4b), diag(1_i4b:n-1_i4b) )
!    dl(1_i4b:n-1_i4b) = sign( dl(1_i4b:n-1_i4b), sub(1_i4b:n-1_i4b) )
!
!   FROM HERE USE FLOATING POINT NUMBERS WITH EXTENDED
!   EXPONENTS FOR THE COMPUTATIONS.
!
    do i=1_i4b, k-1_i4b
!
        u1 = dl(i)/d1(i)
!
        if ( abs(u1)>machtiny ) then
!
!           REPRESENT u1 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = fraction( u1 )
            se(i) = exponent( u1 )
!
        else
!
!           RECOMPUTE u1 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = fraction( dl(i) )/fraction( d1(i) )
            se(i) = exponent( dl(i) ) - exponent( d1(i) )
!
        end if
!
    end do
!
    do i=k, n-1_i4b
!
        l2 = dl(i)/d2(i+1_i4b)
!
        if ( abs(l2)>machtiny ) then
!
!           REPRESENT 1/l2 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = one/fraction( l2 )
            se(i) = -exponent( l2 )                
!
        else
!
!           RECOMPUTE 1/l2 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = fraction( d2(i+1_i4b) )/fraction( dl(i) )
            se(i) = exponent( d2(i+1_i4b) ) - exponent( dl(i) )
!
        end if
!
    end do
!
!    u1(1_i4b:k-1_i4b) = dl(1_i4b:k-1_i4b)/d1(1_i4b:k-1_i4b)
!    l2(k:n-1_i4b)     = dl(k:n-1_i4b)/d2(k+1_i4b:n)
!
!   FROM HERE USE FLOATING POINT NUMBERS WITH EXTENDED
!   EXPONENTS FOR THE COMPUTATIONS.
!
!    where ( abs(u1(1_i4b:k-1_i4b))>machtiny )
!
!       REPRESENT u1(:k-1) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(1_i4b:k-1_i4b) = fraction( u1(1_i4b:k-1_i4b) )
!        se(1_i4b:k-1_i4b) = exponent( u1(1_i4b:k-1_i4b) )
!
!    elsewhere
!
!       RECOMPUTE u1(:k-1) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(1_i4b:k-1_i4b) = fraction( dl(1_i4b:k-1_i4b) )/fraction( d1(1_i4b:k-1_i4b) )
!        se(1_i4b:k-1_i4b) = exponent( dl(1_i4b:k-1_i4b) ) - exponent( d1(1_i4b:k-1_i4b) )
!
!    end where    
!
!    where ( abs(l2(k:n-1_i4b))>machtiny )
!
!       REPRESENT 1/l2(k:) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(k:n-1_i4b) = one/fraction( l2(k:n-1_i4b) )
!        se(k:n-1_i4b) = -exponent( l2(k:n-1_i4b) )                
!
!    elsewhere
!
!       RECOMPUTE 1/l2(k:) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(k:n-1_i4b) = fraction( d2(k+1_i4b:n) )/fraction( dl(k:n-1_i4b) )
!        se(k:n-1_i4b) = exponent( d2(k+1_i4b:n) ) - exponent( dl(k:n-1_i4b) )
!
!    end where
!
    cf = fraction( one )
    ce = exponent( one )
!
    do i=1_i4b, n-1_i4b
!
!       CONVERT cf*(base**ce) TO cs_right(i) AND sf(i)*(base**se(i)) TO sn_right(i).
!
        sfi = sf(i)
        sei = se(i)
        ei  = max( ce, sei )
        ce  = ce - ei
        sei = sei - ei
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
#ifdef _SCALE
        f = sqrt( (scale(cf,ce))**2 + (scale(sfi,sei))**2 )
#else
        f = sqrt( (cf*(machbase**ce))**2 + (sfi*(machbase**sei))**2 )
#endif
!
!       SAVE cs_right(i) IN THE FORM WITH EXTENDED EXPONENTS.
!
        x  = cf/f
        cf = fraction( x )
        ce = ce + exponent( x )
!
!       SAVE COSINES AND SINES OF GIVENS ROTATIONS.
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
#ifdef _SCALE
        cs_right(i) = scale( cf, ce )
        sn_right(i) = scale( sfi/f, sei )
#else
        cs_right(i) = cf*(machbase**ce)
        sn_right(i) = (sfi/f)*(machbase**sei)
#endif
!
    end do
!
!$OMP SECTION
!
!   COMPUTE LEFT DEFLATION PARAMETERS.
!
!   L1*D1*U1 FACTORIZATION OF BD*BD' - lambda2*I .
!
    call upper_bd_dpqd2_b( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, false, d1(1_i4b:n), s1(1_i4b:n)  )
!
!   U2*D2*L2 FACTORIZATION OF BD*BD' - lambda2*I .
!
    call upper_bd_dsqd2_b( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, true, d2(1_i4b:n), s2(1_i4b:n) )
!
!   FIND OPTIMAL k AND tau .
!
    tau = abs( s1(n) - s2(n) + lambda2 )
    k = n
!
    do i=n-1_i4b, 1_i4b, -1_i4b
!
        cf = abs( s1(i) - s2(i) + lambda2 )
!
        if ( cf>=tau ) cycle
!
        tau = cf
        k   = i
!
    end do
!
!   LDU FACTORIZATION OF BD*BD' - lambda2*I .
!
!    call upper_bd_dpqd2_a( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, false, d1(1_i4b:n) )
!
!   UDL FACTORIZATION OF BD*BD' - lambda2*I .
!
!    call upper_bd_dsqd2_a( q2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda2, true, d2(1_i4b:n) )
!
!   FIND OPTIMAL k AND tau .
!
!    tau = abs( (d1(n) - q2(n)) + d2(n) + lambda2 )
!    k = n
!
!    do i=n-1_i4b, 1_i4b, -1_i4b
!
!        cf = abs( (d1(i) - q2(i)) + (d2(i) - e2(i)) + lambda2 )
!        if ( cf>=tau ) cycle
!        tau = cf
!        k   = i
!
!    end do
!
!   COMPUTE OFF-DIAGONAL ELEMENTS OF BD*BD'.
!
    dl(1_i4b:n-1_i4b) = diag(2_i4b:n)*sub(1_i4b:n-1_i4b)
!
!    dl(1_i4b:n-1_i4b) = sqrt( q2(2_i4b:n)*e2(1_i4b:n-1_i4b) )
!    dl(1_i4b:n-1_i4b) = sign( dl(1_i4b:n-1_i4b), diag(2_i4b:n) )
!    dl(1_i4b:n-1_i4b) = sign( dl(1_i4b:n-1_i4b), sub(1_i4b:n-1_i4b) )
!
!   FROM HERE USE FLOATING POINT NUMBERS WITH EXTENDED
!   EXPONENTS FOR THE COMPUTATIONS.
!
    do i=1_i4b, k-1_i4b
!
        u1 = dl(i)/d1(i)
!
        if ( abs(u1)>machtiny ) then
!
!           REPRESENT u1 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = fraction( u1 )
            se(i) = exponent( u1 )
!
        else
!
!           RECOMPUTE u1 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = fraction( dl(i) )/fraction( d1(i) )
            se(i) = exponent( dl(i) ) - exponent( d1(i) )
!
        end if
!
    end do
!
    do i=k, n-1_i4b
!
        l2 = dl(i)/d2(i+1_i4b)
!
        if ( abs(l2)>machtiny ) then
!
!           REPRESENT 1/l2 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = one/fraction( l2 )
            se(i) = -exponent( l2 )                
!
        else
!
!           RECOMPUTE 1/l2 IN THE FORM WITH EXTENDED EXPONENTS.
!
            sf(i) = fraction( d2(i+1_i4b) )/fraction( dl(i) )
            se(i) = exponent( d2(i+1_i4b) ) - exponent( dl(i) )
!
        end if
!
    end do
!
!    u1(1_i4b:k-1_i4b) = dl(1_i4b:k-1_i4b)/d1(1_i4b:k-1_i4b)
!    l2(k:n-1_i4b)     = dl(k:n-1_i4b)/d2(k+1_i4b:n)
!
!   FROM HERE USE FLOATING POINT NUMBERS WITH EXTENDED
!   EXPONENTS FOR THE COMPUTATIONS IN ORDER TO AVOID UNDERFLOWS.
!
!    where ( abs(u1(1_i4b:k-1_i4b))>machtiny )
!
!       REPRESENT u1(:k-1) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(1_i4b:k-1_i4b) = fraction( u1(1_i4b:k-1_i4b) )
!        se(1_i4b:k-1_i4b) = exponent( u1(1_i4b:k-1_i4b) )
!
!    elsewhere
!
!       RECOMPUTE u1(:k-1) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(1_i4b:k-1_i4b) = fraction( dl(1_i4b:k-1_i4b) )/fraction( d1(1_i4b:k-1_i4b) )
!        se(1_i4b:k-1_i4b) = exponent( dl(1_i4b:k-1_i4b) ) - exponent( d1(1_i4b:k-1_i4b) )
!
!    end where    
!
!    where ( abs(l2(k:n-1_i4b))>machtiny )
!
!       REPRESENT 1/l2(k:) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(k:n-1_i4b) = one/fraction( l2(k:n-1_i4b) )
!        se(k:n-1_i4b) = -exponent( l2(k:n-1_i4b) )                
!
!    elsewhere
!
!       RECOMPUTE 1/l2(k:) IN THE FORM WITH EXTENDED EXPONENTS.
!
!        sf(k:n-1_i4b) = fraction( d2(k+1_i4b:n) )/fraction( dl(k:n-1_i4b) )
!        se(k:n-1_i4b) = exponent( d2(k+1_i4b:n) ) - exponent( dl(k:n-1_i4b) )
!
!    end where
!
    cf = fraction( one )
    ce = exponent( one )
!
    do i=1_i4b, n-1_i4b
!
!       CONVERT cf*(base**ce) TO cs_right(i) AND sf(i)*(base**se(i)) TO sn_left(i).
!
        sfi = sf(i)
        sei = se(i)
        ei  = max( ce, sei )
        ce  = ce - ei
        sei = sei - ei
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
#ifdef _SCALE
        f = sqrt( (scale(cf,ce))**2 + (scale(sfi,sei))**2 )
#else
        f = sqrt( (cf*(machbase**ce))**2 + (sfi*(machbase**sei))**2 )
#endif
!
!       SAVE cs_left(i) IN THE FORM WITH EXTENDED EXPONENTS.
!
        x  = cf/f
        cf = fraction( x )
        ce = ce + exponent( x )
!
!       SAVE COSINES AND SINES OF GIVENS ROTATIONS.
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
#ifdef _SCALE
        cs_left(i) = scale( cf, ce )
        sn_left(i) = scale( sfi/f, sei )
#else
        cs_left(i) = cf*(machbase**ce)
        sn_left(i) = (sfi/f)*(machbase**sei)
#endif
!
    end do
!
!$OMP END PARALLEL SECTIONS
!
!   APPLY DEFLATION PARAMETERS.
!
    d1(1_i4b:n) = d(1_i4b:n)
    s1(1_i4b)   = zero
    s1(2_i4b:n) = e(1_i4b:n-1_i4b)
!
    cs = one
    sn = zero
!
    f = zero
    x = d1(1_i4b)
!
    h_max = zero
!
    do i=2_i4b, n
!
        im1 = i - 1_i4b
!
        g = s1(i)
        y = d1(i)
        h = sn*g
        g = cs*g
!
!       APPLY THE RIGHT DEFLATION PARAMETERS TO THE BIDIAGONAL MATRIX.
!
        cs = cs_right(im1)
        sn = sn_right(im1)
!
        s1(im1) =  cs*f + sn*h
        h       = -sn*f + cs*h
!
        h_max = max( h_max, abs(h) )
!
        f =  cs*x + sn*g
        g = -sn*x + cs*g
!
        h = sn*y
        y = cs*y
!
!       APPLY THE LEFT DEFLATION PARAMETERS TO THE BIDIAGONAL MATRIX.
!
        cs = cs_left(im1)
        sn = sn_left(im1)
!
        d1(im1) =  cs*f + sn*h
        h       = -sn*f + cs*h
!
        h_max = max( h_max, abs(h) )
!
        f =  cs*g + sn*y
        x = -sn*g + cs*y
!
    end do
!
    s1(n) = f
    d1(n) = x
!
!   CHECK ACCURACY OF THE DEFLATION.
!
    f = max( abs(f), h_max )
!
    deflate = f<=eps*fudge_deflate
!
!   OUTPUT THE NEW BIDIAGONAL MATRIX IF DEFLATION IS SUCCESSFUL.
!
    if ( deflate ) then
!
        d(1_i4b:n)       = d1(1_i4b:n)
        e(1_i4b:n-1_i4b) = s1(2_i4b:n)
!
    end if
!
!
! END OF SUBROUTINE dflgen2_bd
! ____________________________
!
    end subroutine dflgen2_bd
!
! =========================================================================================
!
    subroutine dflapp_bd( d, e, cs_left, sn_left, cs_right, sn_right, deflate )
!
! Purpose
! _______
!
!   DFLAPP_BD deflates a real n-by-n (upper) bidiagonal matrix BD by two chains
!   of planar rotations produced by DFLGEN_BD or DFLGEN2_BD.
!
!   On entry, the arguments D and E contain, respectively, the main diagonal
!   and off-diagonal of the bidiagonal matrix.
!
!   On output, the arguments D and E contain, respectively, the new main diagonal
!   and off-diagonal of the deflated bidiagonal matrix if DEFLATE is set
!   to true.
!
!
! Arguments
! _________
!
!   D           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, D contains the diagonal elements of the bidiagonal
!               matrix BD.
!
!               On exit, the new main diagonal of the bidiagonal matrix
!               if DEFLATE=true. Otherwise, D is not changed.
!
!   E           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the n-1 off-diagonal elements of the bidiagonal matrix BD.
!
!               On exit, the new off-diagonal of the bidiagonal matrix if DEFLATE=true.
!               Otherwise, E is not changed.
!
!               The size of E must be  size( E ) = size( D ) - 1.
!
!   CS_LEFT     (INPUT) real(stnd), dimension(:)
!               On entry, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations that deflates the bidiagonal matrix BD on the left.
!
!               The size of CS_LEFT must be  size( CS_LEFT ) = size( E ) = size( D ) - 1.
!
!   SN_LEFT     (INPUT) real(stnd), dimension(:)
!               On entry, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations that deflates the bidiagonal matrix BD on the left.
!
!               The size of SN_LEFT must be  size( SN_LEFT ) = size( E ) = size( D ) - 1.
!
!   CS_RIGHT    (INPUT) real(stnd), dimension(:)
!               On entry, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations that deflates the bidiagonal matrix BD on the right.
!
!               The size of CS_RIGHT must be  size( CS_RIGHT ) = size( E ) = size( D ) - 1.
!
!   SN_RIGHT    (INPUT) real(stnd), dimension(:)
!               On entry, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations that deflates the bidiagonal matrix BD on the right.
!
!               The size of SN_RIGHT must be  size( SN_RIGHT ) = size( E ) = size( D ) - 1.
!
!   DEFLATE     (OUTPUT) logical(lgl)
!               On exit:
!
!               - DEFLATE = true :  indicates successful exit.
!               - DEFLATE = false:  indicates that full accuracy was not attained
!                 in the deflation of the bidiagonal matrix.
!
!
! Further Details
! _______________
!
!   For further details, see:
!
!   (1) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (2) Dhillon, I.S., 1998:
!           Reliable computation of the condition number of a tridiagonal matrix in O(n) time.
!           SIAM J. MATRIX ANAL. APPL, Vol. 19, 776-796.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),   intent(inout) :: d, e
    real(stnd), dimension(:),   intent(in)    :: cs_left, sn_left, cs_right, sn_right
!
    logical(lgl),               intent(out) :: deflate
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: eps, normbd, cs, sn, f, x, h, h_max, g, y
    real(stnd), dimension(size(d)) :: diag, sub
!
    integer(i4b)                   :: n, i, im1
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='dflapp_bd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(d),i4b)                  ,   &
                       int(size(e),i4b)  + 1_i4b         ,   &
                       int(size(cs_left),i4b) + 1_i4b    ,   &
                       int(size(sn_left),i4b) + 1_i4b    ,   &
                       int(size(cs_right),i4b) + 1_i4b   ,   &
                       int(size(sn_right),i4b) + 1_i4b /),   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   COMPUTE THE NORM OF THE BIDIAGONAL MATRIX.
!
    diag(1_i4b:n) = d(1_i4b:n)
    sub(1_i4b)    = zero
    sub(2_i4b:n)  = e(1_i4b:n-1_i4b)
!
    normbd = maxval( abs(diag(1_i4b:n)) + abs(sub(1_i4b:n)) )
!
!   TEST IF THE BIDIAGONAL MATRIX IS THE ZERO MATRIX.
!
    if ( normbd==zero ) then
!
        return
!
    end if
!
!   DETERMINE THE UNIT ROUNDOFF AND UNDERFLOW THRESHOLD.
!
    eps = macheps*normbd
!
!   APPLY DEFLATION PARAMETERS.
!
    cs = one
    sn = zero
!
    f = zero
    x = diag(1_i4b)
!
    h_max = zero
!
    do i=2_i4b, n
!
        im1 = i - 1_i4b
!
        g = sub(i)
        y = diag(i)
        h = sn*g
        g = cs*g
!
!       APPLY THE RIGHT DEFLATION PARAMETERS TO THE BIDIAGONAL MATRIX.
!
        cs = cs_right(im1)
        sn = sn_right(im1)
!
        sub(im1) =  cs*f + sn*h
        h        = -sn*f + cs*h
!
        h_max = max( h_max, abs(h) )
!
        f =  cs*x + sn*g
        g = -sn*x + cs*g
!
        h = sn*y
        y = cs*y
!
!       APPLY THE LEFT DEFLATION PARAMETERS TO THE BIDIAGONAL MATRIX.
!
        cs = cs_left(im1)
        sn = sn_left(im1)
!
        diag(im1) =  cs*f + sn*h
        h         = -sn*f + cs*h
!
        h_max = max( h_max, abs(h) )
!
        f =  cs*g + sn*y
        x = -sn*g + cs*y
!
    end do
!
    sub(n)  = f
    diag(n) = x
!
!   CHECK ACCURACY OF THE DEFLATION.
!
    f = max( abs(f), h_max )
!
    deflate = f<=eps*fudge_deflate
!
!   OUTPUT THE NEW BIDIAGONAL MATRIX IF DEFLATION IS SUCCESSFUL.
!
    if ( deflate ) then
!
        d(1_i4b:n)       = diag(1_i4b:n)
        e(1_i4b:n-1_i4b) = sub(2_i4b:n)
!
    end if
!
!
! END OF SUBROUTINE dflapp_bd
! ___________________________
!
    end subroutine dflapp_bd
!
! =========================================================================================
!
    subroutine qrstep_bd( d, e, lambda, cs_left, sn_left, cs_right, sn_right, deflate, update_bd )
!
! Purpose
! _______
!
!   QRSTEP_BD performs one QR step with a given shift LAMBDA on a n-by-n real (upper)
!   bidiagonal matrix BD.
!
!   On entry, the arguments D and E contain, respectively, the main diagonal
!   and superdiagonal of the bidiagonal matrix.
!
!   On output, the arguments D and E contain, respectively, the new main diagonal
!   and superdiagonal of the updated (e.g. deflated) bidiagonal matrix, if DEFLATE is set
!   to true or if the optional logical argument UPDATE_BD is used with the value true,
!   otherwise they are not changed.
!
!   The two chains of n-1 planar rotations produced during the QR step with shift LAMBDA
!   are saved in the arguments CS_LEFT, SN_LEFT, CS_RIGHT, SN_RIGHT.
!
!
! Arguments
! _________
!
!   D           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, D contains the diagonal elements of the bidiagonal
!               matrix BD.
!
!               On exit, the new main diagonal of the bidiagonal matrix
!               if DEFLATE=true or if UPDATE_BD=true. Otherwise, D is not changed.
!
!   E           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the n-1 superdiagonal elements of the bidiagonal matrix BD.
!
!               On exit, the new superdiagonal of the bidiagonal matrix if DEFLATE=true
!               or if UPDATE_BD=true. Otherwise, E is not changed.
!
!               The size of E must be  size( E ) = size( D ) - 1.
!
!   LAMBDA      (INPUT) real(stnd)
!               On entry, the shift used in the current QR step.
!
!   CS_LEFT     (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations applied to the bidiagonal matrix BD on the left in the current
!               QR step.
!
!               The size of CS_LEFT must be  size( CS_LEFT ) = size( E ) = size( D ) - 1.
!
!   SN_LEFT     (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations applied to the bidiagonal matrix BD on the left in the current
!               QR step.
!
!               The size of SN_LEFT must be  size( SN_LEFT ) = size( E ) = size( D ) - 1.
!
!   CS_RIGHT    (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations applied to the bidiagonal matrix BD on the right in the current
!               QR step.
!
!               The size of CS_RIGHT must be  size( CS_RIGHT ) = size( E ) = size( D ) - 1.
!
!   SN_RIGHT    (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations applied to the bidiagonal matrix BD on the right in the current
!               QR step.
!
!               The size of SN_RIGHT must be  size( SN_RIGHT ) = size( E ) = size( D ) - 1.
!
!   DEFLATE     (OUTPUT) logical(lgl)
!               On exit:
!
!               - DEFLATE = true :  indicates that deflation occured at the end of the step.
!               - DEFLATE = false:  indicates that the last superdiagonal element of the
!                 bidiagonal matrix is not small.
!
!   UPDATE_BD   (INPUT, OPTIONAL) logical(lgl)
!               On entry:
!
!               - UPDATE_BD = true :  indicates that the bidiagonal matrix will be updated on exit.
!               - UPDATE_BD = false:  indicates that the bidiagonal matrix will be updated on exit
!                 only if DEFLATE = true.
!
!               The default value for UPDATE_BD is false.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the matlab routine QRSTEP given in the reference (1).
!   The bidiagonal matrix BD is assumed to be unreduced, but no checks are done in the
!   subroutine to verify this hypothesis.
!
!   For further details, see:
!
!   (1) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!   (2) Demmel, J.W., and Kahan, W., 1990:
!           Accurate singular values of bidiagonal matrices.
!           SIAM J. Sci. Statist. Comput., 11:5, 873-912.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),               intent(in)    :: lambda
    real(stnd), dimension(:), intent(inout) :: d, e
    real(stnd), dimension(:), intent(out)   :: cs_left, sn_left, cs_right, sn_right
!
    logical(lgl), intent(in), optional :: update_bd
    logical(lgl), intent(out)          :: deflate
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: eps, f, x, cs, sn, g, y, h, t
    real(stnd), dimension(size(d)) :: d2, e2
!
    integer(i4b) :: n, i, j
!
    logical(lgl) :: update_bd2
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='qrstep_bd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    deflate    = true
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(d),i4b),                     &
                       int(size(e),i4b)  + 1_i4b,            &
                       int(size(cs_left),i4b) + 1_i4b,       &
                       int(size(sn_left),i4b) + 1_i4b,       &
                       int(size(cs_right),i4b) + 1_i4b,      &
                       int(size(sn_right),i4b) + 1_i4b /),   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
    update_bd2 = false
!
    if ( present(update_bd) ) then
!
        update_bd2 = update_bd
!
    end if
!
!   MAKE A SHIFTED COPY OF THE BIDIAGONAL MATRIX.
!
    d2(1_i4b:n) = d(1_i4b:n)
!
    e2(1_i4b)   = zero
    e2(2_i4b:n) = e(1_i4b:n-1_i4b)
!
!   DETERMINE THE DEFLATION THRESHOLD.
!
    t = max( ten, min( c100, macheps**(-eighth) ) )
!
    eps = macheps*t
!
!    eps = macheps*maxval( abs(d(:n)) + abs(e2(:n)) )
!
!   QR SWEEP.
!
    x = d2(1_i4b)
    f = (x-lambda)*(x+lambda)
!
!   INNER LOOP.
!
    cs = one
    sn = x
!
    do i = 2_i4b, n
!
        j = i - 1_i4b
!
        g = e2(i)
        y = d2(i)
        h = sn*g
        g = cs*g
!
!       DEFINE THE GIVENS ROTATION.
!
        if ( abs(f)>abs(h) ) then
!
            t     = h/f
            sn    = sign( sqrt(one+t*t), f )
            e2(j) = f*sn
            cs    = one/sn
            sn    = t*cs
!
        else if ( h/=zero ) then
!
            t     = f/h
            cs    = sign( sqrt(one+t*t), h )
            e2(j) = h*cs
            sn    = one/cs
            cs    = t*sn
!
        else
!
            cs    = zero
            sn    = one
            e2(j) = zero
!
        end if
!
!       APPLY THE GIVENS ROTATION FROM THE RIGHT.
!
        f = x*cs + g*sn
        g = -x*sn + g*cs
        h = y*sn
        y = y*cs
!
!       STORE THE GIVENS ROTATION.
!
        cs_right(j) = cs
        sn_right(j) = sn
!
!       DEFINE THE GIVENS ROTATION.
!
        if ( abs(f)>abs(h) ) then
!
            t     = h/f
            sn    = sign( sqrt(one+t*t), f )
            d2(j) = f*sn
            cs    = one/sn
            sn    = t*cs
!
        else if ( h/=zero ) then
!
            t     = f/h
            cs    = sign( sqrt(one+t*t), h )
            d2(j) = h*cs
            sn    = one/cs
            cs    = t*sn
!
        else
!
            cs    = zero
            sn    = one
            d2(j) = zero
!
        end if
!
!       APPLY THE GIVENS ROTATION FROM THE LEFT.
!
        f =  cs*g + sn*y
        x = -sn*g + cs*y
!
!       STORE THE GIVENS ROTATION.
!
        cs_left(j) = cs
        sn_left(j) = sn
!
    end do
!
    e2(n) = f
    d2(n) = x
!
!   CHECK IF DEFLATION OCCURED.
!
    deflate = abs( f )<=( eps*abs( x ) + safmin )
!
!    deflate = abs( f )<=eps*fudge_deflate
!
!   OUTPUT THE NEW BIDIAGONAL MATRIX IF DEFLATION IS SUCCESSFUL OR
!   IF UPDATING IS REQUESTED.
!
    if ( deflate .or. update_bd2 ) then
!
        d(1_i4b:n)       = d2(1_i4b:n)
        e(1_i4b:n-1_i4b) = e2(2_i4b:n)
!
    end if
!
!
! END OF SUBROUTINE qrstep_bd
! ___________________________
!
    end subroutine qrstep_bd
!
! =========================================================================================
!
    subroutine qrstep_zero_bd( d, e, cs_left, sn_left, cs_right, sn_right, deflate, update_bd )
!
! Purpose
! _______
!
!   QRSTEP_ZERO_BD performs one implicit QR step with a zero shift on a n-by-n real
!   (upper) bidiagonal matrix BD.
!
!   On entry, the arguments D and E contain, respectively, the main diagonal
!   and superdiagonal of the bidiagonal matrix.
!
!   On output, the arguments D and E contain, respectively, the new main diagonal
!   and superdiagonal of the updated (e.g. deflated) bidiagonal matrix, if DEFLATE is set
!   to true or if the optional logical argument UPDATE_BD is used with the value true,
!   otherwise they are not changed.
!
!   The two chains of n-1 planar rotations produced during the QR step with zero shift 
!   are saved in the arguments CS_LEFT, SN_LEFT, CS_RIGHT, SN_RIGHT.
!
!
! Arguments
! _________
!
!   D           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, D contains the diagonal elements of the bidiagonal
!               matrix BD.
!
!               On exit, the new main diagonal of the bidiagonal matrix
!               if DEFLATE=true or if UPDATE_BD=true. Otherwise, D is not changed.
!
!   E           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the n-1 superdiagonal elements of the bidiagonal matrix BD.
!
!               On exit, the new superdiagonal of the bidiagonal matrix if DEFLATE=true
!               or if UPDATE_BD=true. Otherwise, E is not changed.
!
!               The size of E must be  size( E ) = size( D ) - 1.
!
!   CS_LEFT     (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations applied to the bidiagonal matrix BD on the left in the current
!               QR step.
!
!               The size of CS_LEFT must be  size( CS_LEFT ) = size( E ) = size( D ) - 1.
!
!   SN_LEFT     (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations applied to the bidiagonal matrix BD on the left in the current
!               QR step.
!
!               The size of SN_LEFT must be  size( SN_LEFT ) = size( E ) = size( D ) - 1.
!
!   CS_RIGHT    (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations applied to the bidiagonal matrix BD on the right in the current
!               QR step.
!
!               The size of CS_RIGHT must be  size( CS_RIGHT ) = size( E ) = size( D ) - 1.
!
!   SN_RIGHT    (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations applied to the bidiagonal matrix BD on the right in the current
!               QR step.
!
!               The size of SN_RIGHT must be  size( SN_RIGHT ) = size( E ) = size( D ) - 1.
!
!   DEFLATE     (OUTPUT) logical(lgl)
!               On exit:
!
!               - DEFLATE = true :  indicates that deflation occured at the end of the step.
!               - DEFLATE = false:  indicates that the last superdiagonal element of the
!                 bidiagonal matrix is not small.
!
!   UPDATE_BD   (INPUT, OPTIONAL) logical(lgl)
!               On entry:
!
!               - UPDATE_BD = true :  indicates that the bidiagonal matrix will be updated on exit.
!               - UPDATE_BD = false:  indicates that the bidiagonal matrix will be updated on exit
!                 only if DEFLATE = true.
!
!               The default value for UPDATE_BD is false.
!
! Further Details
! _______________
!
!   This subroutine is adapted from the implicit zero-shift QR algorithm given in
!   the reference (1).
!
!   For further details, see:
!
!   (1) Demmel, J.W., and Kahan, W., 1990:
!           Accurate singular values of bidiagonal matrices.
!           SIAM J. Sci. Statist. Comput., 11:5, 873-912.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:), intent(inout) :: d, e
    real(stnd), dimension(:), intent(out)   :: cs_left, sn_left, cs_right, sn_right
!
    logical(lgl), intent(in), optional :: update_bd
    logical(lgl), intent(out)          :: deflate
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: maxbd, eps, cs, sn, cs2, sn2, f, g, h, r, t, tt
    real(stnd), dimension(size(d)) :: d2, e2
!
    integer(i4b) :: n, i, j
!
    logical(lgl) :: update_bd2
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='qrstep_zero_bd'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    deflate    = true
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(d),i4b),                     &
                       int(size(e),i4b)  + 1_i4b,            &
                       int(size(cs_left),i4b) + 1_i4b,       &
                       int(size(sn_left),i4b) + 1_i4b,       &
                       int(size(cs_right),i4b) + 1_i4b,      &
                       int(size(sn_right),i4b) + 1_i4b /),   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
    update_bd2 = false
!
    if ( present(update_bd) ) then
!
        update_bd2 = update_bd
!
    end if
!
!   MAKE A SHIFTED COPY OF THE BIDIAGONAL MATRIX.
!
    d2(1_i4b:n) = d(1_i4b:n)
!
    e2(1_i4b)   = zero
    e2(2_i4b:n) = e(1_i4b:n-1_i4b)
!
!   COMPUTE THE NORM OF THE BIDIAGONAL MATRIX.
!
    maxbd = maxval( abs( d2(1_i4b:n) ) + abs( e2(1_i4b:n) ) )
!
!   TEST IF THE BIDIAGONAL MATRIX IS THE ZERO MATRIX.
!
    if ( maxbd<=zero ) then
!
        cs_left(1_i4b:n-1_i4b)  = one
        sn_left(1_i4b:n-1_i4b)  = zero
!
        cs_right(1_i4b:n-1_i4b) = one
        sn_right(1_i4b:n-1_i4b) = zero
!
        return
!
    end if
!
!   DETERMINE THE DEFLATION THRESHOLD.
!
!    t = max( ten, min( c100, macheps**(-eighth) ) )
!
!    eps = macheps*t
!
    eps = macheps*maxbd
!
!   QR SWEEP.
!
!   INNER LOOP.
!
    cs2 = one
    sn2 = zero
!
    h = d2(1_i4b)
!
    do i = 2_i4b, n
!
        j = i - 1_i4b
!
        g = e2(i)
!
!       DEFINE THE GIVENS ROTATION.
!
        if ( h==zero ) then
!
            cs = zero
            sn = one
            r  = g
!
        else if ( abs(h)>abs(g) ) then
!
            t  = g/h
            tt = sqrt( one + t*t )
            cs = one/tt
            sn = t*cs
            r  = h*tt
!
        else
!
            t  = h/g
            tt = sqrt( one + t*t )
            sn = one/tt
            cs = t*sn
            r  = g*tt
!
        end if
!
!       APPLY THE GIVENS ROTATION FROM THE RIGHT.
!
        e2(j) = sn2*r
        f     = cs2*r
        g     = d2(i)*sn
        h     = d2(i)*cs
!
!       STORE THE GIVENS ROTATION.
!
        cs_right(j) = cs
        sn_right(j) = sn
!
!       DEFINE THE GIVENS ROTATION.
!
        if ( f==zero ) then
!
            cs2 = zero
            sn2 = one
            r   = g
!
        else if ( abs(f)>abs(g) ) then
!
            t   = g/f
            tt  = sqrt( one + t*t )
            cs2 = one/tt
            sn2 = t*cs2
            r   = f*tt
!
        else
!
            t   = f/g
            tt  = sqrt( one + t*t )
            sn2 = one/tt
            cs2 = t*sn2
            r   = g*tt
!
        end if
!
!       APPLY THE GIVENS ROTATION FROM THE LEFT.
!
        d2(j) = r
!
!       STORE THE GIVENS ROTATION.
!
        cs_left(j) = cs2
        sn_left(j) = sn2
!
    end do
!
    e2(n) = h*sn2
    d2(n) = h*cs2
!
!   CHECK IF DEFLATION OCCURED.
!
!    deflate = abs( e2(n) )<=( eps*abs( d2(n) ) + safmin )
!
    deflate = abs( e2(n) )<=eps*fudge_deflate
!
!   OUTPUT THE NEW BIDIAGONAL MATRIX IF DEFLATION IS SUCCESSFUL OR
!   IF UPDATING IS REQUESTED.
!
    if ( deflate .or. update_bd2 ) then
!
        d(1_i4b:n)       = d2(1_i4b:n)
        e(1_i4b:n-1_i4b) = e2(2_i4b:n)
!
    end if
!
!
! END OF SUBROUTINE qrstep_zero_bd
! ________________________________
!
    end subroutine qrstep_zero_bd
!
! =========================================================================================
!
    subroutine upper_bd_deflate_rv( d, e, singval, leftvec, rightvec, failure, &
                                    max_qr_steps, scaling )
!
! Purpose
! _______
!
!   UPPER_BD_DEFLATE computes the left and right singular vectors of a real (upper) bidiagonal
!   matrix BD corresponding to a specified singular value, using a deflation technique.
!
!
! Arguments
! _________
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, the diagonal elements of the bidiagonal matrix BD.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, the n-1 superdiagonal elements of the bidiagonal
!                  matrix BD.
!
!                  The size of E must be  size( E ) = size( D ) - 1 = n - 1.
!
!   SINGVAL        (INPUT) real(stnd)
!                  On entry, a singular value of the bidiagonal matrix. SINGVAL is
!                  assumed to be positive or zero.
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:)
!                  On exit, the computed left singular vector associated with the
!                  singular value SINGVAL.
!
!                  The shape of LEFTVEC must verify: size( LEFTVEC ) = size( D ) = n .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:)
!                  On exit, the computed right singular vector associated with the
!                  singular value SINGVAL.
!
!                  The shape of RIGHTVEC must verify: size( RIGHTVEC ) = size( D ) = n .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false :  indicates successful exit.
!                  - FAILURE = true  :  indicates that the algorithm did not converge and
!                    that full accuracy was not attained in the deflation procedure
!                    of the bidiagonal matrix.
!
!   MAX_QR_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                  MAX_QR_STEPS controls the maximum number of QR sweeps for deflating
!                  the bidiagonal matrix for a given singular value.
!
!                  The algorithm fails to converge if the total number of QR sweeps
!                  exceeds MAX_QR_STEPS.
!
!                  The default is 4.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if SCALING=true the bidiagonal matrix BD is scaled before computing
!                  the deflation parameters in order to avoid overflows.
!
!                  The default is to scale the bidiagonal matrix.
!
!
! Further Details
! _______________
!
!   UPPER_BD_DEFLATE is a low-level subroutine used by BD_DEFLATE subroutines. Its use
!   as a stand-alone method for computing singular vectors of a bidiagonal matrix is not
!   recommended.
!
!   Note also that the sign of the singular vectors computed by this subroutine is arbitrary
!   and not necessarily consistent between the left and right singular vectors. In order
!   to compute consistent singular triplets, subroutine BD_DEFLATE must be used instead.
!
!   For further details, see:
!
!   (1) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (2) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!   (3) Demmel, J.W., and Kahan, W., 1990:
!           Accurate singular values of bidiagonal matrices.
!           SIAM J. Sci. Statist. Comput., 11:5, 873-912.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert_eq, merror
    use Eig_Procedures,  only : prodgiv_eigvec, prodgiv
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in)                :: singval
    real(stnd),  intent(in),  dimension(:) :: d, e
    real(stnd),  intent(out), dimension(:) :: leftvec, rightvec
!
    integer(i4b), intent(in), optional :: max_qr_steps
!
    logical(lgl), intent(out)          :: failure
    logical(lgl), intent(in), optional :: scaling
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                  :: gamma, eps2, prec, g, singval2, lim
    real(stnd),  dimension(size(d))             :: d2, e2
    real(stnd),  dimension(:,:),    allocatable :: csmat_left, snmat_left, csmat_right, snmat_right
!
    integer(i4b) :: n, i, i1, i2, i3, k, maxqrs
    integer      :: iok
!
    logical(lgl) :: deflate
!
#ifdef _OPENMP
!
    logical :: test_par, ompparallel
!
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='upper_bd_deflate'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)         ,   &
                    int(size(e),i4b) + 1_i4b ,   &
                    int(size(leftvec),i4b)   ,   &
                    int(size(rightvec),i4b)  ,   &
                    name_proc )
!
    failure = false
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            leftvec(1_i4b)  = one
            rightvec(1_i4b) = one
!
        end if
!
        return
!
    end if
!
!   INITIALIZE COUNTERS FOR QR ITERATIONS.
!
    if ( present(max_qr_steps) ) then
        maxqrs = merge( max_qr_steps, 4_i4b , max_qr_steps>=1_i4b )
    else
        maxqrs = 4_i4b
    end if
!
!   DETERMINE THE UNIT ROUNDOFF AND UNDERFLOW THRESHOLD.
!
    prec   = fudge_prec*ulp
    gamma  = sqrt( safmin )
!
!   DETERMINE THE DEFLATION THRESHOLD.
!
    g = max( ten, min( c100, macheps**(-eighth) ) )
!
    eps2 = macheps*g
!
!   MAKE A LOCAL COPY OF THE BIDIAGONAL MATRIX AND SINGULAR VALUE.
!
    d2(1_i4b:n) = d(1_i4b:n)
    e2(1_i4b)   = zero
    e2(2_i4b:n) = e(1_i4b:n-1_i4b)
!
    singval2 = abs( singval )
!
    lim = real( n, stnd )*prec*sum( abs(d2(1_i4b:n)) + abs(e2(1_i4b:n)) )
!    lim = real( n, stnd )*prec*maxval( abs(d2(1_i4b:n)) + abs(e2(1_i4b:n)) )
!
!   TEST IF THE UPPER PART OF THE BIDIAGONAL MATRIX IS NOT ZERO OR DIAGONAL.
!
    do i1=1_i4b, n-1_i4b
!
!        if ( abs( e2(i1+1_i4b) )>gamma .or. abs( d2(i1) )>gamma ) exit
!
        if ( abs( e2(i1+1_i4b) )>gamma ) exit
        if ( abs( d2(i1) )>gamma .and. abs( d2(i1) - singval2 )<=lim ) exit
!
    end do
!
!   TEST IF THE LOWER PART OF THE BIDIAGONAL MATRIX IS NOT ZERO OR DIAGONAL.
!
    do i2=n, 2_i4b, -1_i4b
!
!        if ( abs( e2(i2) )>gamma .or. abs( d2(i2) )>gamma ) exit
!
        if ( abs( e2(i2) )>gamma ) exit
        if ( abs( d2(i2) )>gamma .and. abs( d2(i2) - singval2 )<=lim ) exit
!
    end do
!
!    if ( ( singval2<=gamma .and. ( i1/=1_i4b .or. i2/=n ) ) &
    if ( ( singval2<=lim .and. ( i1/=1_i4b .or. i2/=n ) ) &
         .or. ( i2-i1<0_i4b ) ) then
!
!       IF singval IS VERY SMALL, WE CANNOT WORK ON THE REDUCED PORTION
!       OF THE BIDIAGONAL MATRIX AND THE RESULTS ARE PROBABLY NOT ACCURATE.
!
        i1 = 1_i4b
        i2 = n
!
    end if
!
    i3 = i2 - i1
!
!   ALLOCATE WORK VARIABLES NEEDED TO STORE THE GIVENS ROTATIONS.
!
    allocate( csmat_left(i3,maxqrs),     &
              snmat_left(i3,maxqrs),     &
              csmat_right(i3,maxqrs),    &
              snmat_right(i3,maxqrs),    &
              stat = iok      )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    csmat_left(1_i4b:i3,1_i4b:maxqrs)  = one
    snmat_left(1_i4b:i3,1_i4b:maxqrs)  = zero
    csmat_right(1_i4b:i3,1_i4b:maxqrs) = one
    snmat_right(1_i4b:i3,1_i4b:maxqrs) = zero
!
!   COMPUTE DEFLATION PARAMETERS.
!
    deflate = false
    k = 1_i4b
!
    if ( singval2<=lim ) then
!
!       IF SINGULAR VALUE IS SMALL, TRY FIRST IMPLICIT ZERO-SHIFT QR ALGORITHM.
!
        call qrstep_zero_bd( d2(i1:i2), e2(i1+1_i4b:i2),                               &
                             csmat_left(1_i4b:i3,1_i4b),  snmat_left(1_i4b:i3,1_i4b),  &
                             csmat_right(1_i4b:i3,1_i4b), snmat_right(1_i4b:i3,1_i4b), &
                             deflate )
!
    end if
!
    if ( .not.deflate ) then
!
!       OTHERWISE, TRY FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION.
!
        call dflgen2_bd( d2(i1:i2), e2(i1+1_i4b:i2), singval2,                      &
                         csmat_left(1_i4b:i3,1_i4b),  snmat_left(1_i4b:i3,1_i4b),   &
                         csmat_right(1_i4b:i3,1_i4b), snmat_right(1_i4b:i3,1_i4b),  &
                         deflate, scaling=scaling )
!
        if ( .not.deflate ) then
!
!           USE QR STEPS IF DEFLATION FAILED WITH GODUNOV DEFLATION.
!
            do k=1_i4b, maxqrs
!
                if ( singval2>lim ) then
!
!                   IF SINGULAR VALUE IS NOT SMALL, USE STANDARD IMPLICIT SHIFT QR ALGORITHM.
!
                    g = eps2*abs( d2(i2) ) + safmin
!
!                   REPLACE SMALL SUPERDIAGONALS.
!
                    where ( abs(e2(i1+1_i4b:i2-1_i4b))<g )                         &
                            e2(i1+1_i4b:i2-1_i4b) = sign( g, e2(i1+1_i4b:i2-1_i4b) )
!
!                   REPLACE SMALL DIAGONALS.
!
                    where ( abs(d2(i1:i2-1_i4b))<g )                   &
                            d2(i1:i2-1_i4b) = sign( g, d2(i1:i2-1_i4b) )
!
                    call qrstep_bd( d2(i1:i2), e2(i1+1_i4b:i2), singval2,             &
                                    csmat_left(1_i4b:i3,k),  snmat_left(1_i4b:i3,k),  &
                                    csmat_right(1_i4b:i3,k), snmat_right(1_i4b:i3,k), &
                                    deflate, update_bd=true )
!
                else
!
!                   IF SINGULAR VALUE IS VERY SMALL, USE IMPLICIT ZERO-SHIFT QR ALGORITHM.
!
                    call qrstep_zero_bd( d2(i1:i2), e2(i1+1_i4b:i2),                       &
                                         csmat_left(1_i4b:i3,k),  snmat_left(1_i4b:i3,k),  &
                                         csmat_right(1_i4b:i3,k), snmat_right(1_i4b:i3,k), &
                                         deflate, update_bd=true )
!
                end if
!
                if ( deflate ) exit
!
            end do
!
            k = min( k, maxqrs )
!
        end if
!
    end if
!
    failure = .not.deflate
!
!   NOW COMPUTE THE SINGULAR VECTORS.
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    test_par    = .not.ompparallel .and. i>1_i4b
#endif
!
!$OMP PARALLEL SECTIONS IF(test_par), PRIVATE(i), NUM_THREADS( 2 )
!
!$OMP SECTION
!
    leftvec(1_i4b:i1-1_i4b) = zero
    leftvec(i1:i2)          = prodgiv_eigvec( csmat_left(1_i4b:i3,k), snmat_left(1_i4b:i3,k) )
    leftvec(i2+1_i4b:n)     = zero
!
    do i=k-1_i4b, 1_i4b, -1_i4b
        call prodgiv( csmat_left(1_i4b:i3,i), snmat_left(1_i4b:i3,i), leftvec(i1:i2) )
    end do
!
!$OMP SECTION
!
    rightvec(1_i4b:i1-1_i4b) = zero
    rightvec(i1:i2)          = prodgiv_eigvec( csmat_right(1_i4b:i3,k), snmat_right(1_i4b:i3,k) )
    rightvec(i2+1_i4b:n)     = zero
!
    do i=k-1_i4b, 1_i4b, -1_i4b
        call prodgiv( csmat_right(1_i4b:i3,i), snmat_right(1_i4b:i3,i), rightvec(i1:i2) )
    end do
!
!$OMP END PARALLEL SECTIONS
!    
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( csmat_left, snmat_left, csmat_right, snmat_right )
!
!
! END OF SUBROUTINE upper_bd_deflate_rv
! _____________________________________
!
    end subroutine upper_bd_deflate_rv
!
! =========================================================================================
!
    subroutine upper_bd_deflate_rm( d, e, singval, leftvec, rightvec, failure,  &
                                    max_qr_steps, scaling )
!
! Purpose
! _______
!
!   UPPER_BD_DEFLATE computes the left and right singular vectors of a real (upper) bidiagonal
!   matrix BD corresponding to specified singular values, using a deflation technique.
!
!
! Arguments
! _________
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, the diagonal elements of the bidiagonal matrix BD.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, the n-1 superdiagonal elements of the bidiagonal
!                  matrix BD.
!
!                  The size of E must be  size( E ) = size( D ) - 1 = n - 1.
!
!   SINGVAL        (INPUT) real(stnd), dimension(:)
!                  On entry, selected singular values of the bidiagonal matrix.
!                  The singular values can be given in any order, but are assumed to
!                  be positive or zero.
!
!                  The size of SINGVAL must verify:  size( SINGVAL ) <= size( D ) = n .
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed left singular vectors. The left singular vector associated
!                  with the singular value SINGVAL(j) is stored in the j-th column of LEFTVEC.
!
!                  The shape of LEFTVEC must verify:
!
!                  - size( LEFTVEC, 1 ) = size( D ) = n ,
!                  - size( LEFTVEC, 2 ) = size( SINGVAL ) .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed right singular vectors. The right singular vector associated
!                  with the singular value SINGVAL(j) is stored in the j-th column of RIGHTVEC.
!
!                  The shape of RIGHTVEC must verify:
!
!                  - size( RIGHTVEC, 1 ) = size( D ) = n ,
!                  - size( RIGHTVEC, 2 ) = size( SINGVAL ) .
!
!   FAILURE        (OUTPUT) logical(lgl), dimension(:)
!                  On exit:
!
!                  - FAILURE(j) = FALSE : indicates successful exit for the jth singular triplet.
!                  - FAILURE(j) = TRUE : indicates  that the algorithm did not converge and
!                    full accuracy was not attained in the deflation procedure
!                    of the bidiagonal matrix for the jth singular triplet.
!
!                  The size of FAILURE must verify:  size( FAILURE ) = size( SINGVAL ) .
!
!   MAX_QR_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                  MAX_QR_STEPS controls the maximum number of QR sweeps for deflating
!                  the bidiagonal matrix for a given singular value.
!                  The algorithm fails to converge if the total number of QR sweeps for
!                  all eigenvalues exceeds MAX_QR_STEPS * size(EIGVAL).
!
!                  The default is 4.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if SCALING=true the bidiagonal matrix BD is scaled before computing
!                  the deflation parameters in order to avoid overflows.
!
!                  The default is to scale the bidiagonal matrix.
!
!
! Further Details
! _______________
!
!   UPPER_BD_DEFLATE is a low-level subroutine used by BD_DEFLATE subroutines. Its use
!   as a stand-alone method for computing singular vectors of a bidiagonal matrix is not
!   recommended.
!
!   Note also that the sign of the singular vectors computed by this subroutine is arbitrary
!   and not necessarily consistent between the left and right singular vectors. In order
!   to compute consistent singular triplets, subroutine BD_DEFLATE must be used instead.
!
!   For further details, see:
!
!   (1) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (2) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!   (3) Demmel, J.W., and Kahan, W., 1990:
!           Accurate singular values of bidiagonal matrices.
!           SIAM J. Sci. Statist. Comput., 11:5, 873-912.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert_eq, assert, merror
!    use Eig_Procedures,  only : prodgiv_eigvec, prodgiv
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)   :: d, e, singval
    real(stnd),  intent(out), dimension(:,:) :: leftvec, rightvec
!
    integer(i4b), intent(in), optional :: max_qr_steps
!
    logical(lgl), intent(out),  dimension(:)          :: failure
    logical(lgl), intent(in),                optional :: scaling
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                  :: eps2, prec, gamma, g, tmp, c, s, singval2, lim
    real(stnd),  dimension(size(d))             :: d2, e2, vec
    real(stnd),  dimension(:,:),    allocatable :: csmat_left, snmat_left, csmat_right, snmat_right
!
    integer(i4b)                           :: n, m, i, i1, i2, j, l, k, p, q, maxqrs, maxqrs2, ndeflate
    integer(i4b), dimension(size(singval)) :: jstart, jend, ind
    integer                                :: iok
!
    logical(lgl)                           :: done
    logical(lgl), dimension(size(singval)) :: deflate
!
#ifdef _OPENMP
!
    logical :: test_par, ompparallel
!
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='upper_bd_deflate'
!
! NUMBER OF ITERATIONS FOR DEFLATING THE TRIDIAGONAL MATRIX WITH FERNANDO-GODUNOV METHOD.
!
    integer(i4b), parameter :: maxiter=2_i4b
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)           ,   &
                    int(size(e),i4b) + 1_i4b   ,   &
                    int(size(leftvec,1),i4b)   ,   &
                    int(size(rightvec,1),i4b)  ,   &
                    name_proc )
!
    m =  assert_eq( int(size(singval),i4b)     ,   &
                    int(size(failure),i4b)     ,   &
                    int(size(leftvec,2),i4b)   ,   &
                    int(size(rightvec,2),i4b)  ,   &
                    name_proc )
!
    call assert( logical( m<=n, lgl ), name_proc )
!
    failure(1_i4b:m) = false
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b .or. m<=0_i4b) then
!
        if ( n==1_i4b .and. m==1_i4b ) then
!
            leftvec(1_i4b,1_i4b)  = one
            rightvec(1_i4b,1_i4b) = one
!
        end if
!
        return
!
    end if
!
!   INITIALIZE COUNTERS FOR QR ITERATIONS.
!
    if ( present(max_qr_steps) ) then
        maxqrs = merge( max_qr_steps, 4_i4b , max_qr_steps>=1_i4b )
    else
        maxqrs = 4_i4b
    end if
!
    jstart(1_i4b:m) = 0_i4b
    jend(1_i4b:m)   = -1_i4b
!
!   DETERMINE THE UNIT ROUNDOFF AND UNDERFLOW THRESHOLD.
!
    prec   = fudge_prec*ulp
    gamma  = sqrt( safmin )
!
!   DETERMINE THE DEFLATION THRESHOLD.
!
    g = max( ten, min( c100, macheps**(-eighth) ) )
!
    eps2 = macheps*g
!
!   MAKE A LOCAL COPY OF THE BIDIAGONAL MATRIX.
!
    d2(1_i4b:n) = d(1_i4b:n)
    e2(1_i4b)   = zero
    e2(2_i4b:n) = e(1_i4b:n-1_i4b)
!
    lim = real( n, stnd )*prec*sum( abs(d2(1_i4b:n)) + abs(e2(1_i4b:n)) )
!    lim = real( n, stnd )*prec*maxval( abs(d2(1_i4b:n)) + abs(e2(1_i4b:n)) )
!
!   TEST IF THE UPPER PART OF THE BIDIAGONAL MATRIX IS NOT ZERO OR DIAGONAL.
!
    do i1=1_i4b, n-1_i4b
!
!        if ( abs( e2(i1+1_i4b) )>gamma .or. abs( d2(i1) )>gamma ) exit
!
        if ( abs( e2(i1+1_i4b) )>gamma ) exit
        if ( abs( d2(i1) )>gamma .and. any( abs( d2(i1) - singval(1_i4b:m) )<=lim )  ) exit
!
    end do
!
!   TEST IF THE LOWER PART OF THE BIDIAGONAL MATRIX IS NOT ZERO OR DIAGONAL.
!
    do i2=n, 2_i4b, -1_i4b
!
!        if ( abs( e2(i2) )>gamma .or. abs( d2(i2) )>gamma ) exit
!
        if ( abs( e2(i2) )>gamma ) exit
        if ( abs( d2(i2) )>gamma .and. any( abs( d2(i2) - singval(1_i4b:m) )<=lim ) ) exit
!
    end do
!
!    if ( ( any( abs( singval(:m) )<=gamma ) .and. ( i1/=1_i4b .or. i2/=n ) ) &
    if ( ( any( abs( singval(:m) )<=lim ) .and. ( i1/=1_i4b .or. i2/=n ) ) &
          .or. ( m>i2-i1+1_i4b ) ) then
!
!       IF SOME SINGULAR VALUES ARE VERY SMALL OR IF WE MISS SOME SINGULAR VALUES
!       WHEN COMPUTING i1 and i2 ABOVE, WE CANNOT WORK ON THE REDUCED PORTION OF 
!       THE BIDIAGONAL MATRIX AND THE RESULTS ARE PROBABLY NOT ACCURATE.
!
        i1 = 1_i4b
        i2 = n
!
    end if
!
!   ALLOCATE WORK VARIABLES NEEDED TO STORE THE GIVENS ROTATIONS.
!
    k = maxqrs*m
!
    allocate( csmat_left(n-1_i4b,k),     &
              snmat_left(n-1_i4b,k),     &
              csmat_right(n-1_i4b,k),    &
              snmat_right(n-1_i4b,k),    &
              stat = iok      )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    csmat_left(1_i4b:n-1_i4b,1_i4b:k)  = one
    snmat_left(1_i4b:n-1_i4b,1_i4b:k)  = zero
    csmat_right(1_i4b:n-1_i4b,1_i4b:k) = one
    snmat_right(1_i4b:n-1_i4b,1_i4b:k) = zero
!
    l        = i2
    k        = 1_i4b
    ndeflate = 0_i4b
!
    done        = false
    deflate(:m) = false
!
!   COMPUTE DEFLATION PARAMETERS.
!
    do i=1_i4b, maxiter
!
        done = i==1_i4b
!
        do j=1_i4b, m
!
            if ( deflate(j) ) cycle
!
            singval2 = abs( singval(j) )
!
            if ( singval2<=lim .and. done ) then
!
!               IF SINGULAR VALUE IS VERY SMALL, TRY FIRST IMPLICIT ZERO-SHIFT QR ALGORITHM.
!
                call qrstep_zero_bd( d2(i1:l), e2(i1+1_i4b:l),                             &
                                     csmat_left(i1:l-1_i4b,k),  snmat_left(i1:l-1_i4b,k),  &
                                     csmat_right(i1:l-1_i4b,k), snmat_right(i1:l-1_i4b,k), &
                                     deflate(j) )
!
            else
!
!               OTHERWISE, TRY FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION.
!
                call dflgen2_bd( d2(i1:l), e2(i1+1_i4b:l),  singval2,                   &
                                 csmat_left(i1:l-1_i4b,k),  snmat_left(i1:l-1_i4b,k),   &
                                 csmat_right(i1:l-1_i4b,k), snmat_right(i1:l-1_i4b,k),  &
                                 deflate(j), scaling=scaling )
!
            end if
!
            if ( .not.deflate(j) ) cycle
!
            ndeflate = ndeflate + 1_i4b
            ind(j)   = ndeflate
!
            jstart(ndeflate) = k
            jend(ndeflate)   = k
!
            k = k + 1_i4b
            l = l - 1_i4b
!
        end do
!
        done = all( deflate(:m) )
!
        if ( done ) exit
!
    end do
!
    if ( .not.done ) then
!
!       RECOMPUTE DEFLATION PARAMETERS.
!
        do j=1_i4b, m
!
            if ( deflate(j) ) cycle
!
            ndeflate = ndeflate + 1_i4b
            ind(j)   = ndeflate
            jstart(ndeflate) = k
!
            singval2 = abs( singval(j) )
!
!           TRY AGAIN FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION.
!
            call dflgen2_bd( d2(i1:l), e2(i1+1_i4b:l),  singval2,                   &
                             csmat_left(i1:l-1_i4b,k),  snmat_left(i1:l-1_i4b,k),   &
                             csmat_right(i1:l-1_i4b,k), snmat_right(i1:l-1_i4b,k),  &
                             deflate(j), scaling=scaling )
!
            if ( .not.deflate(j) ) then
!
!               USE QR STEPS IF DEFLATION FAILED WITH FERNANDO-GODUNOV DEFLATION.
!
                maxqrs2 = maxqrs*(ndeflate - 1_i4b ) - k + 1_i4b + maxqrs
!
                do i=1_i4b, maxqrs2
!
                    if ( singval2>lim ) then
!
!                       IF SINGULAR VALUE IS NOT SMALL, USE STANDARD IMPLICIT SHIFT QR ALGORITHM.
!
                        g = eps2*abs( d2(l) ) + safmin
!
!                       REPLACE SMALL SUPERDIAGONALS.
!
                        where ( abs(e2(i1+1_i4b:l-1_i4b))<g )                      &
                              e2(i1+1_i4b:l-1_i4b) = sign( g, e2(i1+1_i4b:l-1_i4b) )
!
!                       REPLACE SMALL DIAGONALS.
!
                        where ( abs(d2(i1:l-1_i4b))<g )                &
                              d2(i1:l-1_i4b) = sign( g, d2(i1:l-1_i4b) )
!
                        call qrstep_bd( d2(i1:l), e2(i1+1_i4b:l),  singval2,                  &
                                        csmat_left(i1:l-1_i4b,k),  snmat_left(i1:l-1_i4b,k),  &
                                        csmat_right(i1:l-1_i4b,k), snmat_right(i1:l-1_i4b,k), &
                                        deflate(j), update_bd=true )
!
                    else
!
!                       IF SINGULAR VALUE IS VERY SMALL, USE IMPLICIT ZERO-SHIFT QR ALGORITHM.
!
                        call qrstep_zero_bd( d2(i1:l), e2(i1+1_i4b:l),                             &
                                             csmat_left(i1:l-1_i4b,k),  snmat_left(i1:l-1_i4b,k),  &
                                             csmat_right(i1:l-1_i4b,k), snmat_right(i1:l-1_i4b,k), &
                                             deflate(j), update_bd=true )
!
                    end if
!
                    if ( deflate(j) ) exit
!
                    k = k + 1_i4b
!
                end do
!
                if ( .not.deflate(j) ) then
!
                    k = k - 1_i4b
!
                end if
!
            end if
!
            jend(ndeflate) = k
!
            k = k + 1_i4b
            l = l - 1_i4b
!
        end do
!
    end if
!
    failure(1_i4b:m) = .not.deflate(1_i4b:m)
!
!   NOW COMPUTE THE SINGULAR VECTORS.
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    test_par    = .not.ompparallel .and.      &
                  m>1_i4b          .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                            &
!$OMP         ,PRIVATE(j,l,i,k,p,q,ndeflate,tmp,c,s,vec)
!
!   COMPUTE LEFT SINGULAR VECTORS.
!
!$OMP DO SCHEDULE(STATIC,1)
!
    do j=1_i4b, m
!
        ndeflate = ind(j)
!
        l = i2 - ndeflate + 1_i4b
!
        p = jstart(ndeflate)
        k = jend(ndeflate)
!
!BUG:   SOME VERSIONS OF PGF95 WITH OPENMP HAVE TROUBLES WITH THE FOLLOWING CALL.
!       THUS, IT IS SAFER TO INLINE THE FUNCTION.
!
!        leftvec(i1:l,j) = prodgiv_eigvec( csmat_left(i1:l-1_i4b,k), snmat_left(i1:l-1_i4b,k) )
!
!       COMPUTE THE PRODUCT OF THE GIVENS ROTATIONS TIMES THE VECTOR El.
!
        tmp = one
!
        do i =l-1_i4b, i1, -1_i4b
!
!           APPLY THE GIVENS ROTATION TO THE VECTOR El.
!
            vec(i+1_i4b) = csmat_left(i,k)*tmp
            tmp          = -snmat_left(i,k)*tmp
!
        end do
!
        vec(i1) = tmp
!
        do i=k-1_i4b, p, -1_i4b
!
!BUG:       SOME VERSIONS OF PGF95 WITH OPENMP HAVE TROUBLES WITH THE FOLLOWING CALL.
!           THUS, IT IS SAFER TO INLINE THE SUBROUTINE.
!
!            call prodgiv( csmat_left(i1:l-1_i4b,i), snmat_left(i1:l-1_i4b,i), leftvec(i1:l,j) )
!
!           COMPUTE THE PRODUCT OF THE GIVENS ROTATIONS TIMES THE VECTOR leftvec(i1:l,j).
!
            tmp = vec(l)
!
            do q =l-1_i4b, i1, -1_i4b
!
                c = csmat_left(q,i)
                s = snmat_left(q,i)
!
!               APPLY THE GIVENS ROTATION TO THE VECTOR leftvec(i1:l,j).
!
                vec(q+1_i4b) = s*vec(q) + c*tmp
                tmp          = c*vec(q) - s*tmp
!
            end do
!
            vec(i1) = tmp
!
        end do
!
        vec(l+1_i4b:n) = zero
!
        do i=ndeflate-1_i4b, 1_i4b, -1_i4b
!
            do k=jend(i), jstart(i), -1_i4b
!
!BUG:           SOME VERSIONS OF PGF95 WITH OPENMP HAVE TROUBLES WITH THE FOLLOWING CALL.
!               THUS, IT IS SAFER TO INLINE THE SUBROUTINE.
!
!                call prodgiv( csmat_left(i1:l,k), snmat_left(i1:l,k), leftvec(i1:l+1_i4b,j) )
!
!               COMPUTE THE PRODUCT OF THE GIVENS ROTATIONS TIMES THE VECTOR leftvec(i1:l+1,j).
!
                tmp = vec(l+1_i4b)
!
                do q =l, i1, -1_i4b
!
                    c = csmat_left(q,k)
                    s = snmat_left(q,k)
!
!                   APPLY THE GIVENS ROTATION TO THE VECTOR leftvec(i1:l+1,j).
!
                    vec(q+1_i4b) = s*vec(q) + c*tmp
                    tmp          = c*vec(q) - s*tmp
!
                end do
!
                vec(i1) = tmp
!
            end do
!
            l = l + 1_i4b
!
        end do
!
        vec(1_i4b:i1-1_i4b) = zero
!
        leftvec(1_i4b:n,j) = vec(1_i4b:n)
!
    end do
!
!$OMP END DO NOWAIT
!
!   COMPUTE RIGHT SINGULAR VECTORS.
!
!$OMP DO SCHEDULE(STATIC,1)
!
    do j=1_i4b, m
!
        ndeflate = ind(j)
!
        l = i2 - ndeflate + 1_i4b
!
        p = jstart(ndeflate)
        k = jend(ndeflate)
!
!BUG:   SOME VERSIONS OF PGF95 WITH OPENMP HAVE TROUBLES WITH THE FOLLOWING CALL.
!       THUS, IT IS SAFER TO INLINE THE FUNCTION.
!
!        rightvec(i1:l,j) = prodgiv_eigvec( csmat_right(i1:l-1_i4b,k), snmat_right(i1:l-1_i4b,k) )
!
!       COMPUTE THE PRODUCT OF THE GIVENS ROTATIONS TIMES THE VECTOR El.
!
        tmp = one
!
        do i =l-1_i4b, i1, -1_i4b
!
!           APPLY THE GIVENS ROTATION TO THE VECTOR El.
!
            vec(i+1_i4b) = csmat_right(i,k)*tmp
            tmp          = -snmat_right(i,k)*tmp
!
        end do
!
        vec(i1) = tmp
!
        do i=k-1_i4b, p, -1_i4b
!
!BUG:       SOME VERSIONS OF PGF95 WITH OPENMP HAVE TROUBLES WITH THE FOLLOWING CALL.
!           THUS, IT IS SAFER TO INLINE THE SUBROUTINE.
!
!            call prodgiv( csmat_right(i1:l-1_i4b,i), snmat_right(i1:l-1_i4b,i), rightvec(i1:l,j) )
!
!           COMPUTE THE PRODUCT OF THE GIVENS ROTATIONS TIMES THE VECTOR rightvec(i1:l,j).
!
            tmp = vec(l)
!
            do q =l-1_i4b, i1, -1_i4b
!
                c = csmat_right(q,i)
                s = snmat_right(q,i)
!
!               APPLY THE GIVENS ROTATION TO THE VECTOR rightvec(i1:l,j).
!
                vec(q+1_i4b) = s*vec(q) + c*tmp
                tmp          = c*vec(q) - s*tmp
!
            end do
!
            vec(i1) = tmp
!
        end do
!
        vec(l+1_i4b:n) = zero
!
        do i=ndeflate-1_i4b, 1_i4b, -1_i4b
!
            do k=jend(i), jstart(i), -1_i4b
!
!BUG:           SOME VERSIONS OF PGF95 WITH OPENMP HAVE TROUBLES WITH THE FOLLOWING CALL.
!               THUS, IT IS SAFER TO INLINE THE SUBROUTINE.
!
!                call prodgiv( csmat_right(i1:l,k), snmat_right(i1:l,k), rightvec(i1:l+1_i4b,j) )
!
!               COMPUTE THE PRODUCT OF THE GIVENS ROTATIONS TIMES THE VECTOR rightvec(i1:l+1,j).
!
                tmp = vec(l+1_i4b)
!
                do q =l, i1, -1_i4b
!
                    c = csmat_right(q,k)
                    s = snmat_right(q,k)
!
!                   APPLY THE GIVENS ROTATION TO THE VECTOR rightvec(i1:l+1,j).
!
                    vec(q+1_i4b) = s*vec(q) + c*tmp
                    tmp          = c*vec(q) - s*tmp
!
                end do
!
                vec(i1) = tmp
!
            end do
!
            l = l + 1_i4b
!
        end do
!
        vec(1_i4b:i1-1_i4b) = zero
!
        rightvec(1_i4b:n,j) = vec(1_i4b:n)
!
    end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!    
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( csmat_left, snmat_left, csmat_right, snmat_right )
!
!
! END OF SUBROUTINE upper_bd_deflate_rm
! _____________________________________
!
    end subroutine upper_bd_deflate_rm
!
! =========================================================================================
!
    subroutine bd_deflate( upper, d, e, s, leftvec, rightvec, failure, max_qr_steps,      &
                           ortho, scaling, inviter )
!
! Purpose
! _______
!
!   BD_DEFLATE computes the left and right singular vectors of a real n-by-n bidiagonal
!   matrix BD corresponding to specified singular values, using deflation techniques on
!   the bidiagonal matrix BD.
!
!
! Arguments
! _________
!
!   UPPER          (INPUT) logical(lgl)
!                  On entry, if:
!
!                  - UPPER = true  : BD is upper bidiagonal ;
!                  - UPPER = false : BD is lower bidiagonal.
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, D contains the diagonal elements of the bidiagonal
!                  matrix BD.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, E contains the off-diagonal elements of the bidiagonal
!                  matrix BD. E(1) is arbitrary.
!
!                  The size of E must verify: size( E ) = size( D ) = n . 
!
!   S              (INPUT) real(stnd), dimension(:)
!                  On entry, selected singular values of the bidiagonal matrix BD.
!                  The singular values must be given in decreasing order and are assumed
!                  to be positive or zero.
!
!                  The size of S must verify:  size( S ) <= size( D ) = n .
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed left singular vectors. The left singular vector associated
!                  with the singular value S(j) is stored in the j-th column of LEFTVEC.
!
!                  The shape of LEFTVEC must verify:
!
!                  - size( LEFTVEC, 1 ) = size( D ) = n ,
!                  - size( LEFTVEC, 2 ) = size( S ) .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed right singular vectors. The right singular vector associated
!                  with the singular value S(j) is stored in the j-th column of RIGHTVEC.
!
!                  The shape of RIGHTVEC must verify:
!
!                  - size( RIGHTVEC, 1 ) = size( D ) = n ,
!                  - size( RIGHTVEC, 2 ) = size( S ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false :  indicates successful exit.
!                  - FAILURE = true  :  indicates that the algorithm did not converge and
!                    that full accuracy was not attained in the deflation procedure
!                    of the bidiagonal matrix.
!
!   MAX_QR_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                  MAX_QR_STEPS controls the maximum number of QR sweeps for deflating
!                  the bidiagonal matrix for a given singular value.
!                  The algorithm fails to converge if the total number of QR sweeps for
!                  all singular values exceeds MAX_QR_STEPS * size(S).
!
!                  The default is 4.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, the bidiagonal matrix BD is deflated sequentially
!                    for all the specified singular values; this implies that the singular
!                    vectors of the bidiagonal matrix BD will be automatically orthogonal
!                    on exit.
!                  - ORTHO=false, the bidiagonal matrix BD is deflated in parallel for the
!                    different clusters of singular values or isolated singular values; this
!                    implies that orthogonality of the singular vectors of bidiagonal
!                    matrix BD is preserved inside each cluster, but not
!                    automatically between clusters.
!
!                  The default is ORTHO=false.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the bidiagonal matrix BD is scaled before computing
!                    the deflation parameters in order to avoid overflows;
!                  - SCALING=false, the bidiagonal matrix BD is not scaled.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INVITER        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INVITER=true, singular vectors corresponding to isolated
!                    singular values or singular vectors of bidiagonal matrices with
!                    zeros are computed by inverse iteration instead of deflation.
!                  - INVITER=false, singular vectors corresponding to isolated
!                    singular values or singular vectors of bidiagonal matrices
!                    with zeros are computed by deflation.
!
!                  The default is INVITER=true.
!
!
! Further Details
! _______________
!
!   The singular vectors are computed using deflation techniques applied to the
!   bidiagonal matrix BD.
!   The first deflation technique used in BD_DEFLATE combines an extension to
!   bidiagonal matrices of Fernando's approach for computing eigenvectors
!   of tridiagonal matrices with a deflation procedure by Givens rotations originally
!   developed by Godunov and his collaborators (see references (1) and (2) for more
!   details). If this deflation technique failed, QR iterations are used instead as
!   described in (3) and (4).
!
!   Optionally, singular vectors corresponding to isolated singular values or singular
!   vectors of bidiagonal matrices with zeros may be also computed by inverse iteration
!   on the Golub-Kahan tridiagonal form of the bidiagonal matrix BD.
!   This is the default since in these cases inverse iteration is safer and faster
!   than the deflation algorithms.
!
!   The computation of the singular vectors is parallelized if OPENMP is used.
!
!   It is essential that singular values given on entry of BD_DEFLATE are computed to
!   high relative accuracy. Subroutines BD_SINGVAL or BD_SINGVAL2 may be used for this
!   purpose.
!
!   BD_DEFLATE may fail if some the singular values specified in parameter S are nearly
!   identical or for clusters of small singular values.
!
!   For further details, on the deflation techniques used in BD_DEFLATE, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (3) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!   (4) Demmel, J.W., and Kahan, W., 1990:
!           Accurate singular values of bidiagonal matrices.
!           SIAM J. Sci. Statist. Comput., 11:5, 873-912.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error3, svd_error7, allocate_error
    use Utilities,       only : assert_eq, assert, merror
    use Eig_Procedures,  only : prodgiv_eigvec, prodgiv
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)    :: d, e, s
    real(stnd),  intent(out), dimension(:,:)  :: leftvec, rightvec
!
    integer(i4b), intent(in), optional :: max_qr_steps
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(out)          :: failure
    logical(lgl), intent(in), optional :: ortho, scaling, inviter
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),  dimension(:), allocatable :: csvec_left, snvec_left, csvec_right, snvec_right
    real(stnd),  dimension(size(s))        :: lambda, rayleigh
    real(stnd),  dimension(size(d))        :: diag, sup2, temp
    real(stnd),  dimension(size(d)+1)      :: sub, sup1
    real(stnd)                             :: xj, xjm, prec, bdnorm, bdnorm2, lim, lim2, g
!
    integer(i4b)                            :: n, n1, n2, n3, m, m0, j, j1, j2, k, l
    integer(i4b), dimension(:), allocatable :: jstart, jend
    integer                                 :: iok
!
    logical(lgl)                     :: do_ortho, do_inviter, diag_zero, super_zero, deflate, last_vec
    logical(lgl), dimension(size(s)) :: failure2
!
#ifdef _OPENMP
!
    logical :: test_par, ompparallel, ompnested
!
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_deflate'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)           ,   &
                    int(size(e),i4b)           ,   &
                    int(size(leftvec,1),i4b)   ,   &
                    int(size(rightvec,1),i4b)  ,   &
                    name_proc )
!
    m =  assert_eq( int(size(s),i4b)           ,   &
                    int(size(leftvec,2),i4b)   ,   &
                    int(size(rightvec,2),i4b)  ,   &
                    name_proc )
!
    call assert( logical( m<=n, lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b .or. m<=0_i4b) then
!
        if ( n==1_i4b .and. m==1_i4b ) then
!
            leftvec(1_i4b,1_i4b)  = one
            rightvec(1_i4b,1_i4b) = one
!
        end if
!
        return
!
    end if
!
!   CHECK OPTIONAL PARAMETERS.
!
    do_ortho = false
!
    if ( present(ortho) ) then
!
        if ( m>1_i4b ) then
!
            do_ortho = ortho
!
        end if
!
    end if
!
    do_inviter = true
!
    if ( present(inviter) ) then
!
        do_inviter = inviter
!
    end if
!
!   TEST IF SINGULAR VALUES ARE POSITIVE OR ZERO.
!
    if ( any( s(1_i4b:m)<zero ) ) then
!
        call merror( name_proc//svd_error7 )
!
    end if
!
!   TEST IF SINGULAR VALUES ARE IN DECREASING ORDER.
!
    if ( m/=1_i4b ) then
!
        temp(1_i4b:m-1_i4b) = s(2_i4b:m) - s(1_i4b:m-1_i4b)
!
        if ( any( temp(1_i4b:m-1_i4b)>zero ) ) then
!
            call merror( name_proc//svd_error3 )
!
        end if
!
    end if
!
!   DETERMINE MACHINE PARAMETERS.
!
    prec = fudge_prec*ulp
!
!   MAKE LOCAL COPIES OF THE BIDIAGONAL MATRIX AND SINGULAR VALUES.
!
    diag(1_i4b:n) = d(1_i4b:n)
    sub(1_i4b)    = zero
    sub(2_i4b:n)  = e(2_i4b:n)
!
    lambda(1_i4b:m) = s(1_i4b:m)
!
!   CHECK IF THE BIDIAGONAL MATRIX IS THE ZERO MATRIX.
!
    m0 = n
!
    do n2 = n, 1_i4b, -1_i4b
!
        if ( abs(diag(n2))>=safmin .or. abs(sub(n2))>=safmin ) exit
!
        if ( m0>0_i4b .and. m0<=m ) then
!
            leftvec(:n,m0) = zero
            leftvec(n2,m0)  = one
!
            rightvec(:n,m0) = zero
            rightvec(n2,m0) = one
!
        end if
!
        m0 = m0 - 1_i4b
!
    end do
!
    if ( n2==0_i4b ) then
!
!       THE BIDIAGONAL MATRIX IS THE NULL MATRIX, RETURN.
!
        return
!
    end if
!
!   COMPUTE THE NORMS OF THE BIDIAGONAL MATRIX.
!
    temp(1_i4b:n) = abs( diag(1_i4b:n) ) + abs( sub(1_i4b:n) )
    bdnorm        = sum( temp(1_i4b:n)  )
    bdnorm2       = maxval( temp(1_i4b:n) )
!
!   COMPUTE THRESHOLD FOR THE ACCURACY TESTS.
!
!    lim = bdnorm*real( n, stnd )*prec
    lim = bdnorm2*real( n, stnd )*prec
!
!   COMPUTE THRESHOLD FOR SEPARATION OF SINGULAR VALUES.
!
    lim2 = fudge_ortho*sqrt( macheps )*bdnorm2
!
!   TEST IF THE BIDIAGONAL MATRIX HAS SOME ZERO ELEMENTS.
!
    diag_zero  = any( abs( diag(1_i4b:n2) )<=lim )
    super_zero = any( abs( sub(2_i4b:n2) )<=lim )
!
    if ( diag_zero .or. super_zero ) then
!
!       TEST IF THE BIDIAGONAL MATRIX HAS ALMOST A ZERO DIAGONAL.
!
        deflate = all( abs(diag(1_i4b:n))<=lim )
!
        if ( deflate .and. m==n ) then
!
!           THE BIDIAGONAL MATRIX HAS ALMOST A ZERO DIAGONAL, USE
!           A FAST DIRECT METHOD FOR COMPUTING ALL SINGULAR VECTORS.
!
            leftvec(:n,:n)  = zero
            rightvec(:n,:n) = zero
!
            if ( upper ) then
!
                do j = 2_i4b, n
!
                    if ( sub(j) >= zero ) then
                        rightvec(j,j) = one
                    else
                        rightvec(j,j) = -one
                    end if
!
                    leftvec(j-1_i4b,j) = one
!
                end do
!
                rightvec(1_i4b,1_i4b) = one
                leftvec(n,1_i4b)      = one
!
            else
!
                do j = 2_i4b, n
!
                    if ( sub(j) >= zero ) then
                        leftvec(j,j) = one
                    else
                        leftvec(j,j) = -one
                    end if
!
                    rightvec(j-1_i4b,j) = one
!
                end do
!
                leftvec(1_i4b,1_i4b) = one
                rightvec(n,1_i4b)    = one
!
            end if
!
            diag(:n) = abs( sub(:n) )
!
            call lr_singvec_sort( 'd', diag(:n), leftvec(:n,:n), rightvec(:n,:n) )
!
            failure = any( abs( diag(:n) - lambda(:n) )>lim )
!
            return
!
        else if ( deflate .or. do_inviter ) then
!
!           USE INVERSE ITERATION FOR COMPUTING SINGULAR VECTORS TO AVOID PRECISION PROBLEMS
!           WITH THE DEFLATION ALGORITHM WHEN ZEROS ARE PRESENT IN THE BIDIAGONAL MATRIX.
!
            if ( do_ortho ) then
!
                call bd_inviter_rm( upper, diag(:n), sub(:n), lambda(:m), leftvec(:n,:m),    &
                                    rightvec(:n,:m), failure, ortho=true, scaling=scaling,   &
                                    initvec=false )
!
            else
!
                call bd_inviter_rm( upper, diag(:n), sub(:n), lambda(:m), leftvec(:n,:m),    &
                                    rightvec(:n,:m), failure, scaling=scaling, initvec=false )
!
            end if
!
            return
!
        end if
!
    end if
!
!   CHECK THE FIRST PART OF THE BIDIAGONAL MATRIX FOR ZEROS
!   AND REDUCE AGAIN ITS SIZE IF POSSIBLE.
!
    do n1 = 1_i4b, n-1_i4b
!
        if ( abs(diag(n1))>=safmin .or. abs(sub(n1+1_i4b))>=safmin ) exit
!
        if ( m0>0_i4b .and. m0<=m ) then
!
            leftvec(:n,m0) = zero
            leftvec(n1,m0)  = one
!
            rightvec(:n,m0) = zero
            rightvec(n1,m0)  = one
!
        end if
!
        m0 = m0 - 1_i4b
!
    end do
!
    m0 = min( m0, m )
!
    if (  ( m/=n .and. lambda(m)<=lim ) .or. m0>n2-n1+1_i4b .or. m0<=0_i4b ) then
!
!       IF SOME SELECTED SINGULAR VALUES ARE TINY AND WE DONT COMPUTE THE FULL SVD,
!       WE CANNOT WORK ON THE REDUCED BIDIAGONAL MATRIX.
!
        n1 = 1_i4b
        n2 = n
!
        m0 = m
!
    end if
!
    n3 = -1_i4b        
!
    if ( n2>n1 ) then
!
        n3 = n2 - n1           
!
        last_vec = m0==n3+1_i4b
!
!       CHECK FOR ZEROS IN THE DIAGONAL OF THE BIDIAGONAL MATRIX AND IF
!       A FULL SVD IS REQUESTED.
!
        diag_zero = diag_zero .and. last_vec
!
!       SINCE NO SPLITTING STRATEGY IS CURRENTLY USED IN THE SUBROUTINE,
!       PREPROCESSING THE BIDIAGONAL MATRIX WITH THE IMPLICIT ZERO-SHIFT
!       QR ALGORITHM IS USED HERE ONLY IF WE COMPUTE THE FULL SVD SINCE
!       IN THAT CASE WE DON'T NEED TO COMPUTE THE SINGULAR VECTORS COUPLET
!       ASSOCIATED WITH THE ZERO SINGULAR VALUE BY GODUNOV-FERNANDO DEFLATION.
!
        if ( diag_zero ) then
!
!           IF ZEROS ARE PRESENT IN THE DIAGONAL, PROCESS THE BIDIAGONAL MATRIX
!           WITH THE IMPLICIT ZERO-SHIFT QR ALGORITHM TO AVOID PROBLEMS IN THE
!           DEFLATION ALGORITHM.
!
!           ALLOCATE WORK VARIABLES NEEDED TO STORE THE GIVENS ROTATIONS.
!
            allocate( csvec_left(n3),  snvec_left(n3),     &
                      csvec_right(n3), snvec_right(n3),    &
                      stat = iok      )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           COMPUTE DEFLATION PARAMETERS WITH IMPLICIT ZERO-SHIFT QR ALGORITHM.
!
            call qrstep_zero_bd( diag(n1:n2), sub(n1+1_i4b:n2),                &
                                 csvec_left(1_i4b:n3),  snvec_left(1_i4b:n3),  &
                                 csvec_right(1_i4b:n3), snvec_right(1_i4b:n3), &
                                 deflate )
!
            if ( deflate ) then
!
!               UPDATE THE NUMBER OF SINGULAR COUPLETS TO COMPUTE.
!
                if ( last_vec ) m0 = m0 - 1_i4b
!
                n2 = n2 - 1_i4b
!
            else
!
                diag_zero = false
                last_vec  = false
!
!               DEALLOCATE WORK ARRAYS IF DEFLATION WAS USED.
!
                deallocate( csvec_left, snvec_left, csvec_right, snvec_right )
!
            end if
!
        end if
!
    else
!
        diag_zero = false
        last_vec  = false
!
    end if
!
    leftvec(1_i4b:n1-1_i4b,1_i4b:m0) = zero
    leftvec(n2+1_i4b:n,1_i4b:m0)     = zero
!
    rightvec(1_i4b:n1-1_i4b,1_i4b:m0) = zero
    rightvec(n2+1_i4b:n,1_i4b:m0)     = zero
!
    failure2(m0+1_i4b:m) = false
!    
    lim = bdnorm*real( n, stnd )*prec
!    
#ifdef _OPENMP
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m>1_i4b                             .and.       &
                  j>1_i4b
#endif
!
!   COMPUTE THE SINGULAR VECTORS.
!
    if ( m==1_i4b ) then
!
!       ONLY ONE SINGULAR VECTORS COUPLET IS REQUIRED.
!
        if ( do_inviter ) then
!
!           USE INVERSE ITERATION.
!
            call upper_bd_inviter( diag(n1:n2), sub(n1:n2), lambda(1_i4b),         &
                                   leftvec(n1:n2,1_i4b), rightvec(n1:n2,1_i4b),    &
                                   failure2(1_i4b), scaling=scaling, initvec=false )
!
        else
!
!           USE DEFLATION.
!
            call upper_bd_deflate_rv( diag(n1:n2), sub(n1+1_i4b:n2), lambda(1_i4b),  &
                                      leftvec(n1:n2,1_i4b), rightvec(n1:n2,1_i4b),   &
                                      failure2(1_i4b), max_qr_steps=max_qr_steps,    &
                                      scaling=scaling   )
!
        end if
!
    else
!
        if ( do_ortho ) then
!
!           SEVERAL SINGULAR VECTORS WITH GUARANTEED ORTHOGONALITY ARE REQUIRED.
!           USE SEQUENTIAL DEFLATION ON ALL SINGULAR VALUES.
!
            call upper_bd_deflate_rm( diag(n1:n2), sub(n1+1_i4b:n2), lambda(1_i4b:m0),   &
                                      leftvec(n1:n2,1_i4b:m0), rightvec(n1:n2,1_i4b:m0), &
                                      failure2(1_i4b:m0), max_qr_steps=max_qr_steps,     &
                                      scaling=scaling    )
!
        else
!
!           SEVERAL SINGULAR VECTORS WITH GUARANTEED ORTHOGONALITY
!           INSIDE EACH CLUSTER OF SINGULAR VALUES ARE REQUIRED.
!
!           FIRST DETERMINE CLUSTERS OF SINGULAR VALUES.
!
!           ALLOCATE WORK VARIABLES.
!
            allocate( jstart(m0), jend(m0), stat = iok  )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           DETERMINE THE NUMBER OF CLUSTERS OF SINGULAR VALUES.
!           k IS THE NUMBER OF CLUSTERS OF SINGULAR VALUES
!           jstart(k) IS THE INDICE OF THE FIRST SINGULAR VALUE IN CLUSTER k.
!           jend(k)   IS THE INDICE OF THE LAST SINGULAR VALUE IN CLUSTER k.
!
            k = 1_i4b
            jstart(1_i4b) = 1_i4b
!
            xj = lambda(1_i4b)
!
            do j=2_i4b, m0
!
                xjm = lambda(j)
                g   = xj - xjm
                xj  = xjm
!
                if ( g<lim2 ) cycle
!
                jend(k)   = j - 1_i4b
                k         = k + 1_i4b
                jstart(k) = j
!
            end do
!
            jend(k) = m0
!
!           MERGE SOME CLUSTERS FOR SPECIAL CASES.
!
            if ( k==2_i4b) then
!
!               IF THERE IS ONLY TWO CLUSTERS AND ONE OF THEM IS AN ISOLATED SINGULAR VALUE
!               MERGE THE TWO CLUSTERS.
!
                if ( ( jstart(1_i4b)==jend(1_i4b) ) .or. ( jstart(2_i4b)==jend(2_i4b) ) ) then
!
                    k = 1_i4b
!
                    jstart(1_i4b) = 1_i4b
                    jend(1_i4b)   = m0
!
                end if
!
            else if (  k>2_i4b ) then
!
!               IF THE LAST CLUSTER IS AN ISOLATED AND SMALL SINGULAR
!               VALUE MERGE IT WITH THE PREVIOUS CLUSTER.
!
                j1 = jstart(k)
                j2 = jend(k)
!
                if (  lambda(j1)<=lim2 .and. j1==j2 ) then
!
                    k = k - 1_i4b
!
                    jend(k) = m0
!
                end if
!
            end if
!
#ifdef _OPENMP
!
            if ( test_par .and. k>1_i4b ) then
!
!               OPENMP PARALLEL PATH.
!
!               FIRST, COMPUTE LEFT AND RIGHT SINGULAR VECTORS CORRESPONDING TO ISOLATED
!               SINGULAR VALUES AND SMALL CLUSTERS OF ALMOST-MULTIPLE SINGULAR VALUES.
!               EACH THREAD TAKES CARE OF ONE ISOLATED SINGULAR VALUE OR SMALL CLUSTER
!               AND PARALLELIZATION IS OVER THE NUMBER OF SMALL CLUSTERS AND ISOLATED
!               SINGULAR VALUES.
!
!$OMP PARALLEL DO SCHEDULE(DYNAMIC,1)     &
!$OMP            ,PRIVATE(j,j1,j2)
!
                do j=1_i4b, k
!
                    j1 = jstart(j)
                    j2 = jend(j)
!
                    if ( j2-j1<nsing_cluster ) then
!
                        if ( j1==j2 ) then
!
                            if ( do_inviter ) then
!
!                               USE INVERSE ITERATION.
!
                                call upper_bd_inviter( diag(n1:n2), sub(n1:n2), lambda(j1),         &
                                                       leftvec(n1:n2,j1), rightvec(n1:n2,j1),       &
                                                       failure2(j1), scaling=scaling, initvec=false )
!
                            else
!
!                               USE DEFLATION.
!
                                call upper_bd_deflate_rv( diag(n1:n2), sub(n1+1_i4b:n2), lambda(j1),   &
                                                          leftvec(n1:n2,j1), rightvec(n1:n2,j1),       &
                                                          failure2(j1), max_qr_steps=max_qr_steps,     &
                                                          scaling=scaling    )
!
                            end if
!
                        else
!
!                           USE SEQUENTIAL DEFLATION ON SMALL CLUSTERS.
!
                            call upper_bd_deflate_rm( diag(n1:n2), sub(n1+1_i4b:n2), lambda(j1:j2), &
                                                      leftvec(n1:n2,j1:j2), rightvec(n1:n2,j1:j2),  &
                                                      failure2(j1:j2), max_qr_steps=max_qr_steps,   &
                                                      scaling=scaling    )
!
                        end if
!
                    end if
!
                end do
!
!$OMP END PARALLEL DO
!
!               SECOND, COMPUTE LEFT AND RIGHT SINGULAR VECTORS CORRESPONDING TO
!               BIG CLUSTERS OF ALMOST-MULTIPLE SINGULAR VALUES. PARALLELIZATION IS OVER
!               THE NUMBER OF VECTORS IN EACH CLUSTER.
!    
                do j=1_i4b, k
!
                    j1 = jstart(j)
                    j2 = jend(j)
!
                    if ( j2-j1<nsing_cluster ) cycle
!
!                   USE SEQUENTIAL DEFLATION ON ALL SINGULAR VALUES IN THE CLUSTER.
!
                    call upper_bd_deflate_rm( diag(n1:n2), sub(n1+1_i4b:n2), lambda(j1:j2), &
                                              leftvec(n1:n2,j1:j2), rightvec(n1:n2,j1:j2),  &
                                              failure2(j1:j2), max_qr_steps=max_qr_steps,   &
                                              scaling=scaling    )
!
                end do
!
            else
#endif
!
!               SERIAL PATH.
!
                do j=1_i4b, k
!
                    j1 = jstart(j)
                    j2 = jend(j)
!
                    if ( j1==j2 ) then
!
!                       COMPUTE SINGULAR VECTORS ASSOCIATED WITH AN ISOLATED SINGULAR VALUE.
!
                        if ( do_inviter ) then
!
!                           USE INVERSE ITERATION.
!
                            call upper_bd_inviter( diag(n1:n2), sub(n1:n2), lambda(j1),         &
                                                   leftvec(n1:n2,j1), rightvec(n1:n2,j1),       &
                                                   failure2(j1), scaling=scaling, initvec=false )
!
                        else
!
!                           USE DEFLATION.
!
                            call upper_bd_deflate_rv( diag(n1:n2), sub(n1+1_i4b:n2), lambda(j1),   &
                                                      leftvec(n1:n2,j1), rightvec(n1:n2,j1),       &
                                                      failure2(j1), max_qr_steps=max_qr_steps,     &
                                                      scaling=scaling    )
!
                        end if
!
                    else
!
!                       COMPUTE SINGULAR VECTORS ASSOCIATED WITH SINGULAR VALUES BELONGING
!                       TO A CLUSTER OF SINGULAR VALUES.
!                       USE SEQUENTIAL DEFLATION ON ALL SINGULAR VALUES IN THE CLUSTER.
!
                        call upper_bd_deflate_rm( diag(n1:n2), sub(n1+1_i4b:n2), lambda(j1:j2), &
                                                  leftvec(n1:n2,j1:j2), rightvec(n1:n2,j1:j2),  &
                                                  failure2(j1:j2), max_qr_steps=max_qr_steps,   &
                                                  scaling=scaling    )
!
                    end if
!
                end do
!
#ifdef _OPENMP
            end if
#endif
!
        end if
!
    end if
!
    if ( diag_zero ) then
!
        n2 = n2 + 1_i4b
!
        if ( last_vec ) m0 = m0 + 1_i4b
!
    end if
!
!   CHECK ACCURACY OF THE COMPUTED SINGULAR VECTORS.
!
!$OMP PARALLEL  IF(test_par)                        &
!$OMP          ,PRIVATE(j,xj,temp,diag,sub,sup1,sup2)
!
    diag(:n)      = d(:n)
    sub(:n)       = e(:n)
    sub(n+1_i4b)  = zero
    sup1(n+1_i4b) = zero
!
!$OMP DO SCHEDULE(STATIC,1)
!
    do j=1_i4b, m
!
        if ( diag_zero .and. j<=m0 ) then
!
            if ( last_vec .and. j==m0 ) then
!
!               COMPUTE THE SINGULAR VECTORS ASSOCIATED WITH THE DEFLATED ZERO SINGULAR VALUE.
!
                leftvec(1_i4b:n1-1_i4b,j) = zero
                leftvec(n1:n2,j)          = prodgiv_eigvec( csvec_left(1_i4b:n3), snvec_left(1_i4b:n3) )
                leftvec(n2+1_i4b:n,j)     = zero
!
                rightvec(1_i4b:n1-1_i4b,j) = zero
                rightvec(n1:n2,j)          = prodgiv_eigvec( csvec_right(1_i4b:n3), snvec_right(1_i4b:n3) )
                rightvec(n2+1_i4b:n,j)     = zero
!
            else
!
!               APPLY GIVENS ROTATIONS TO OTHERS SINGULAR VECTORS COUPLETS.
!
                call prodgiv( csvec_left(1_i4b:n3),  snvec_left(1_i4b:n3),  leftvec(n1:n2,j) )
                call prodgiv( csvec_right(1_i4b:n3), snvec_right(1_i4b:n3), rightvec(n1:n2,j) )
!
            end if
!
        end if
!
        sup1(1_i4b:n) = rightvec(1_i4b:n,j)
        sup2(1_i4b:n) = leftvec(1_i4b:n,j)
!
!       SWAP THE SINGULAR VECTORS IF THE BIDIAGONAL MATRIX IS LOWER BIDIAGONAL.
!
        if ( .not.upper ) then
            leftvec(1_i4b:n,j)  = sup1(1_i4b:n)
            rightvec(1_i4b:n,j) = sup2(1_i4b:n)
        end if
!
        temp(1_i4b:n) = diag(1_i4b:n)*sup1(1_i4b:n)             +  &
                        sub(2_i4b:n+1_i4b)*sup1(2_i4b:n+1_i4b)
!
!       COMPUTE RAYLEIGH QUOTIENT.
!
        xj = dot_product( sup2(1_i4b:n), temp(1_i4b:n) )
!
        rayleigh(j) = abs( xj )
!
        if ( xj<zero ) then
            leftvec(1_i4b:n,j)  = -leftvec(1_i4b:n,j)
        end if
!
!       COMPUTE THE RESIDUAL NORM OF THE JTH LEFT AND RIGHT SINGULAR VECTORS
!       WITH RESPECT TO THEIR RAYLEIGH QUOTIENT.
!
#ifdef _BLAS
        call axpy( n, -xj, sup2(1_i4b:n), 1_i4b, temp(1_i4b:n), 1_i4b )
#else
        temp(1_i4b:n) = temp(1_i4b:n) - xj*sup2(1_i4b:n)
#endif
        lambda(j) = sum( abs( temp(:n) ) )
!
    end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
    if ( diag_zero ) then
!
!       DEALLOCATE WORK ARRAYS IF THE IMPLICIT ZERO-SHIFT QR ALGORITHM WAS USED.
!
        deallocate( csvec_left, snvec_left, csvec_right, snvec_right )
!
    end if
!
!   CHECK NORMS OF RESIDUAL VECTORS.
!
!    failure = any( lambda(1_i4b:m)>lim ) .or. any( failure2(1_i4b:m) )
    failure = any( lambda(1_i4b:m)>lim )
!
!   CHECK THE PAIRING OF SINGULAR VALUES AND SINGULAR VECTORS.
!
    if ( m/=1_i4b .and. .not.do_ortho ) then
!
        do l=1_i4b, k
!
            j1 = jstart(l)
            j2 = jend(l)
!
            do j=j1, j2-1_i4b
!
                if ( rayleigh(j)<rayleigh(j+1_i4b) ) exit
!
            end do
!
            if ( j<j2 ) then
!
                call lr_singvec_sort( 'd', rayleigh(j1:j2), leftvec(:,j1:j2), rightvec(:,j1:j2) )
!
            end if
!
            if ( l==1_i4b ) cycle
!
            if ( .not.failure ) then
!
                j2 = jend(l-1_i4b)
!
                failure = rayleigh(j2)<rayleigh(j1)
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS
!
        deallocate( jstart, jend )
!
    else
!
        do j=1_i4b, m-1_i4b
!
            if ( rayleigh(j)<rayleigh(j+1_i4b) ) exit
!
        end do
!
!        failure = failure .or. j<m
!
        if ( j<m ) then
!
            call lr_singvec_sort( 'd', rayleigh(1_i4b:m), leftvec(:,:m), rightvec(:,:m) )
!
        end if
!
    end if
!
!   CHECK FINALLY IF THE RAYLEIGH QUOTIENTS AND THE SINGULAR VALUES AGREE.
!
    if ( .not.failure ) then
!
        xj  = maxval( abs( rayleigh(1_i4b:m) - s(1_i4b:m) ) )
        failure = xj>lim
!
    end if
!
!
! END OF SUBROUTINE bd_deflate
! ____________________________
!
    end subroutine bd_deflate
!
! =========================================================================================
!
    subroutine bd_deflate2_a( mat, tauq, taup, d, e, s, leftvec, rightvec, failure,  &
                              max_qr_steps, ortho, scaling, inviter )
!
! Purpose
! _______
!
!   BD_DEFLATE2 computes the left and right singular vectors of a full real m-by-n
!   matrix MAT corresponding to specified singular values, using deflation techniques.
!
!   It is required that the original matrix MAT has been reduced to upper or lower
!   bidiagonal form BD by an orthogonal transformation:
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal. This can be done with a call to BD_CMP with parameters
!   TAUQ and TAUP, before calling BD_SINGVAL (or BD_SINGVAL2) for computing singular values
!   and a call to BD_DEFLATE2 for computing selected singular vectors.
!
!   Alternatively and more simply, the user can call SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2
!   subroutines, with parameters TAUQ and TAUP present, to compute the bidiagonal reduction and
!   singular values in one step and, finally, call DEFLATE2 for computing all or selected
!   singular vectors of MAT.
!
!   If m >= n, BD is upper bidiagonal and if m < n, BD is lower bidiagonal.
!
!   
! Arguments
! _________
!
!   MAT            (INPUT) real(stnd), dimension(:,:)
!                  On entry, the original m-by-n matrix after reduction by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 subroutines.
!                  MAT must contains the vectors which define the elementary reflectors
!                  H(i) and G(i) whose products determine the matrices Q and P,
!                  as returned by BD_CMP, SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2
!                  subroutines when the arguments TAUQ and TAUP are present in the call
!                  to these subroutines. MAT must be specified as returned by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 and is not modified by
!                  the routine.
!                
!   TAUQ           (INPUT) real(stnd), dimension(:)
!                  TAUQ(i) must contain the scalar factor of the elementary
!                  reflector H(i) which determines Q, as returned by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 in their array
!                  argument TAUQ.
!
!                  The size of TAUQ must verify:
!                        size( TAUQ ) = min( size(MAT,1) , size(MAT,2) ) .
!
!   TAUP           (INPUT) real(stnd), dimension(:)
!                  TAUP(i) must contain the scalar factor of the elementary
!                  reflector G(i), which determines P, as returned by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 in their array
!                  argument TAUP.
!
!                  The size of TAUP must verify:
!                        size( TAUP ) = min( size(MAT,1) , size(MAT,2) ) .
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, D contains the diagonal elements of the bidiagonal
!                  matrix BD as returned by BD_CMP, SELECT_SINGVAL_CMP or
!                  SELECT_SINGVAL_CMP2.
!
!                  The size of D must verify: 
!                          size( D ) = min( size(MAT,1) , size(MAT,2) ) . 
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, E contains the off-diagonal elements of the bidiagonal
!                  matrix BD as returned by BD_CMP, SELECT_SINGVAL_CMP or
!                  SELECT_SINGVAL_CMP2:
!
!                  - if m >= n, E(i) = BD(i-1,i) for i = 2,3,...,n;
!                  - if m < n,  E(i) = BD(i,i-1) for i = 2,3,...,m.
!
!                  E(1) is arbitrary.
!
!                  The size of E must verify: 
!                          size( E ) = min( size(MAT,1) , size(MAT,2) ) . 
!
!   S              (INPUT) real(stnd), dimension(:)
!                  On entry, selected singular values of the bidiagonal matrix BD.
!                  The singular values must be given in decreasing order and are
!                  assumed to be positive or zero.
!
!                  The size of S must verify:
!                  size( S ) <= min( size(MAT,1) , size(MAT,2) ) .
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed left singular vectors. The left singular vector associated
!                  with the singular value S(j) is stored in the j-th column of LEFTVEC.
!
!                  The shape of LEFTVEC must verify: 
!
!                  - size( LEFTVEC, 1 ) = size( MAT, 1 ) = m ,
!                  - size( LEFTVEC, 2 ) = size( S ) .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed right singular vectors. The right singular vector associated
!                  with the singular value S(j) is stored in the j-th column of RIGHTVEC.
!
!                  The shape of RIGHTVEC must verify:
!
!                  - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ,
!                  - size( RIGHTVEC, 2 ) = size( S ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false :  indicates successful exit.
!                  - FAILURE = true  :  indicates that the algorithm did not converge and
!                    that full accuracy was not attained in the deflation procedure
!                    of the bidiagonal matrix BD.
!
!   MAX_QR_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                  MAX_QR_STEPS controls the maximum number of QR sweeps for deflating
!                  the bidiagonal matrix BD for a given singular value.
!                  The algorithm fails to converge if the total number of QR sweeps for
!                  all singular values exceeds MAX_QR_STEPS * size(S).
!
!                  The default is 4.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry:
!
!                  - ORTHO=true, the bidiagonal matrix BD is deflated sequentially
!                    for all the specified singular values; this implies that the singular
!                    vectors of the bidiagonal matrix BD will be automatically orthogonal
!                    on exit.
!                  - ORTHO=false, the bidiagonal matrix BD is deflated in parallel for the
!                    different clusters of singular values or isolated singular values; this
!                    implies that orthogonality of the singular vectors of bidiagonal
!                    matrix BD is preserved inside each cluster, but not
!                    automatically between clusters.
!
!                  The default is ORTHO=false.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the intermediate bidiagonal matrix BD is scaled before computing
!                    the deflation parameters in order to avoid overflows;
!                  - SCALING=false, the intermediate bidiagonal matrix BD is not scaled.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INVITER        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INVITER=true, singular vectors corresponding to isolated
!                    singular values or singular vectors of bidiagonal matrices with
!                    zeros are computed by inverse iteration instead of deflation.
!                  - INVITER=false, singular vectors corresponding to isolated
!                    singular values or singular vectors of bidiagonal matrices
!                    with zeros are computed by deflation.
!
!                  The default is INVITER=true.
!
!
! Further Details
! _______________
!
!   The singular vectors are computed using deflation techniques applied implicitly to
!   the associated tridiagonal forms BD' * BD and BD * BD' of the bidiagonal matrix BD. See
!   description of the BD_DEFLATE subroutine for more details.
!
!   The computation of the singular vectors is parallelized if OPENMP is used.
!
!   It is essential that singular values given on entry of BD_DEFLATE2 are computed to
!   high (relative) accuracy. Subroutines BD_SINGVAL or BD_SINGVAL2 may be used for this
!   purpose.
!
!   BD_DEFLATE2 may fail if some the singular values specified in parameter S are nearly
!   identical or for clusters of small singular values for some pathological matrices.
!
!   The deflation algorithms used in BD_DEFLATE2 are competitive with the inverse iteration
!   procedure implemented in BD_INVITER2.
!
!   For further details on the deflation techniques or the blocked back-transformation
!   algorithm used in BD_DEFLATE2, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (3) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!   (4) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:,:)   :: mat
    real(stnd),  intent(in),  dimension(:)     :: tauq, taup, d, e, s
    real(stnd),  intent(out), dimension(:,:)   :: leftvec, rightvec
!
    integer(i4b), intent(in), optional :: max_qr_steps
!
    logical(lgl), intent(in),  optional :: ortho, scaling, inviter
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(:,:), allocatable :: work, work2
!
    integer(i4b) :: n, m, nm, nsing, n1, n2, mnthr_nsing
    integer      :: iok
!
    logical(lgl) :: upper
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_deflate2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size( mat,     1 ),i4b)  ,   &
                   int(size( leftvec, 1 ),i4b)  ,   &
                   name_proc )
!
    n = assert_eq( int(size( mat,      2 ),i4b)  ,   &
                   int(size( rightvec, 1 ),i4b)  ,   &
                   name_proc )
!
    nm = assert_eq( (/ min(m,n)            ,   &
                       int(size(tauq),i4b) ,   &
                       int(size(taup),i4b) ,   &
                       int(size(d),i4b)    ,   &
                       int(size(e),i4b) /) ,   &
                       name_proc )
!
    nsing = assert_eq( int(size(s),i4b)             ,  &
                       int(size( leftvec, 2 ),i4b)  ,  &
                       int(size( rightvec, 2 ),i4b) ,  &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( nm<=0_i4b ) then
!
        return
!
    end if
!
    upper = m >= n
!
!   COMPUTE SINGULAR VECTORS OF THE BIDIAGONAL FORM BD OF mat.
!
    call bd_deflate( upper, d, e, s, leftvec(:nm,:nsing), rightvec(:nm,:nsing),        &
                     failure, max_qr_steps=max_qr_steps, ortho=ortho, scaling=scaling, &
                     inviter=inviter  )
!
!   COMPUTE SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY BACK-TRANSFORMATION.
!
    mnthr_nsing = int( real( nm, stnd )*thr_bt, i4b )
!
    if ( nsing>=mnthr_nsing ) then
!
!       ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM IF NEARLY FULL SVD OF mat IS
!       REQUESTED.
!
        if ( upper ) then
            n1 = m
            n2 = nm
        else
            n1 = n
            n2 = nm + 1_i4b
        end if
!
        allocate( work(n1,n2), work2(nm,nsing), stat = iok )
!
    else
!
!       USE DIRECT ALGORITHM OTHERWISE OR IF THE REQUIRED
!       WORKSPACE IS NOT AVAILABLE.
!
        iok = -1
!
    end if
!
    if ( iok==0 ) then
!
        if ( upper ) then
!
!           GENERATE P IN work(:n,:n) .
!
            call ortho_gen_p_bd( mat, taup, work(:n,:n) )
!
!           MULTIPLY P IN work(:n,:n) BY RIGHT SINGULAR VECTORS OF BD IN rightvec
!           FOR COMPUTING RIGHT SINGULAR VECTORS OF mat .
!
            work2(:n,:nsing) = rightvec(:n,:nsing)
!
#ifdef _BLAS
            call gemm( 'N', 'N', n, nsing, n, one, work, m, work2, n, zero, rightvec, n  )
#else
            rightvec(:n,:nsing) = matmul( work(:n,:n), work2(:n,:nsing) )
#endif
!
!           GENERATE Q IN work .
!
            work(:m,:n) = mat(:m,:n)
!
            call ortho_gen_q_bd( work, tauq )
!
!           MULTIPLY Q IN work BY LEFT SINGULAR VECTORS OF BD IN leftvec
!           FOR COMPUTING LEFT SINGULAR VECTORS OF mat .
!
            work2(:n,:nsing) = leftvec(:n,:nsing)
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, nsing, n, one, work, m, work2, n, zero, leftvec, m   )
#else
            leftvec(:m,:nsing) = matmul( work(:m,:n), work2(:n,:nsing) )
#endif
!
        else
!
!           GENERATE Q IN work .
!
            work(:m,:m+1_i4b) = mat(:m,:m+1_i4b)
!
            call ortho_gen_q_bd( work(:m,:m+1_i4b), tauq )
!
!           MULTIPLY Q IN work BY LEFT SINGULAR VECTORS OF BD IN leftvec
!           FOR COMPUTING LEFT SINGULAR VECTORS OF mat.
!
            work2(:m,:nsing) = leftvec(:m,:nsing)
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, nsing, m, one, work, n, work2, m, zero, leftvec, m   )
#else
            leftvec(:m,:nsing) = matmul( work(:m,:m), work2(:m,:nsing) )
#endif
!
!           GENERATE P IN work .
!
            call ortho_gen_p_bd( mat, taup, work(:n,:m) )
!
!           MULTIPLY P IN work BY RIGHT SINGULAR VECTORS OF BD IN rightvec
!           FOR COMPUTING RIGHT SINGULAR VECTORS OF mat.
!
            work2(:m,:nsing) = rightvec(:m,:nsing)
!
#ifdef _BLAS
            call gemm( 'N', 'N', n, nsing, m, one, work, n, work2, m, zero, rightvec, n   )
#else
            rightvec(:n,:nsing) = matmul( work(:n,:m), work2(:m,:nsing) )
#endif
!
        end if
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( work, work2 )
!
    else
!
        if ( upper ) then
            leftvec(nm+1_i4b:m,:nsing) = zero
        else
            rightvec(nm+1_i4b:n,:nsing) = zero
        end if
!
!       GENERATE LEFT SINGULAR VECTORS OF mat IN leftvec .
!
        call apply_q_bd( mat, tauq, leftvec,  left=true, trans=false )
!
!       GENERATE RIGHT SINGULAR VECTORS OF mat IN rightvec .
!
        call apply_p_bd( mat, taup, rightvec, left=true, trans=false )
!
    end if
!
!
! END OF SUBROUTINE bd_deflate2_a
! _______________________________
!
    end subroutine bd_deflate2_a
!
! =========================================================================================
!
    subroutine bd_deflate2_b( mat, p, d, e, s, leftvec, rightvec, failure,       &
                              max_qr_steps, ortho, scaling, inviter, tol_reortho )
!
! Purpose
! _______
!
!   BD_DEFLATE2 computes the left and right singular vectors of a full real m-by-n
!   matrix MAT with m>=n corresponding to specified singular values, using deflation techniques.
!
!   It is required that the original matrix MAT has been reduced to upper bidiagonal form BD
!   by an orthogonal transformation:
!
!                       Q' * MAT * P = BD
!
!   where Q and P are orthogonal. This can be done with a call to BD_CMP2 (or a call to BD_CMP
!   followed by a call to ORTHO_GEN_BD), before calling BD_SVD, BD_SINGVAL (or BD_SINGVAL2) subroutines
!   for computing singular values and BD_DEFLATE2 for computing selected singular vectors.
!
!   Alternatively and more simply, the user can call SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4
!   subroutines, with parameter P present, to compute the bidiagonal reduction and
!   singular values in one step and, finally, call DEFLATE2 for computing all or selected
!   singular vectors of MAT.
!
!   
! Arguments
! _________
!
!   MAT            (INPUT) real(stnd), dimension(:,:)
!                  On entry, the m-by-n orthogonal matrix Q after reduction by BD_CMP2,
!                  SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4 (or by BD_CMP and ORTHO_GEN_BD).
!                  MAT is not modified by the routine.
!                
!                  The shape of MAT must verify: size( MAT, 1 ) >= size( MAT, 2 ) = n .
!
!   P              (INPUT) real(stnd), dimension(:,:)
!                  On entry, the n-by-n orthogonal matrix P after reduction by BD_CMP2,
!                  SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4 (or by BD_CMP and ORTHO_GEN_BD).        
!                  If P has been computed by BD_CMP2, SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4,
!                  P can be stored in factored form or not. Both cases are handled by the subroutine.
!                  P is not modified by the routine.
!                
!                  The shape of P must verify: size( P, 1 ) = size( P, 2 ) = size( MAT, 2 ) = n .
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, D contains the diagonal elements of the bidiagonal
!                  matrix BD as returned by BD_CMP, BD_CMP2, SELECT_SINGVAL_CMP3
!                  or SELECT_SINGVAL4 subroutines.
!
!                  The size of D must verify: size( D ) = size( MAT, 2 ) = n . 
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, E contains the off-diagonal elements of the bidiagonal
!                  matrix BD as returned by BD_CMP, BD_CMP2, SELECT_SINGVAL_CMP3
!                  or SELECT_SINGVAL4 subroutines:
!
!                        E(i) = BD(i-1,i) for i = 2,3,...,n;
!
!                  E(1) is arbitrary.
!
!                  The size of E must verify: size( E ) = size( MAT, 2 ) = n . 
!
!   S              (INPUT) real(stnd), dimension(:)
!                  On entry, selected singular values of the bidiagonal matrix BD.
!                  The singular values must be given in decreasing order and are
!                  assumed to be positive or zero.
!
!                  The size of S must verify: size( S ) <= size( MAT, 2 ) = n .
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed left singular vectors. The left singular vector
!                  associated with the singular value S(j) is stored in the j-th column
!                  of LEFTVEC.
!
!                  The shape of LEFTVEC must verify:  
!
!                  - size( LEFTVEC, 1 ) = size( MAT, 1 ) = m ,
!                  - size( LEFTVEC, 2 ) = size( S ) .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed right singular vectors. The right singular vector
!                  associated with the singular value S(j) is stored in the j-th column of
!                  RIGHTVEC.
!
!                  The shape of RIGHTVEC must verify:  
!
!                  - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ,
!                  - size( RIGHTVEC, 2 ) = size( S ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false :  indicates successful exit.
!                  - FAILURE = true  :  indicates that the algorithm did not converge and
!                    that full accuracy was not attained in the deflation procedure
!                    of the bidiagonal matrix BD.
!
!   MAX_QR_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                  MAX_QR_STEPS controls the maximum number of QR sweeps for deflating
!                  the bidiagonal matrix BD for a given singular value.
!                  The algorithm fails to converge if the total number of QR sweeps for
!                  all singular values exceeds MAX_QR_STEPS * size(S).
!
!                  The default is 4.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry:
!
!                  - ORTHO=true, the bidiagonal matrix BD is deflated sequentially
!                    for all the specified singular values; this implies that the singular
!                    vectors of the bidiagonal matrix BD will be automatically orthogonal
!                    on exit.
!                  - ORTHO=false, the bidiagonal matrix BD is deflated in parallel for the
!                    different clusters of singular values or isolated singular values; this
!                    implies that orthogonality of the singular vectors of bidiagonal
!                    matrix BD is preserved inside each cluster, but not
!                    automatically between clusters.
!
!                  The default is ORTHO=false.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the intermediate bidiagonal matrix BD is scaled before computing
!                    the deflation parameters in order to avoid overflows;
!                  - SCALING=false, the intermediate bidiagonal matrix BD is not scaled.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INVITER        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INVITER=true, singular vectors corresponding to isolated
!                    singular values or singular vectors of bidiagonal matrices with
!                    zeros are computed by inverse iteration instead of deflation.
!                  - INVITER=false, singular vectors corresponding to isolated
!                    singular values or singular vectors of bidiagonal matrices
!                    with zeros are computed by deflation.
!
!                  The default is INVITER=true.
!
!   TOL_REORTHO    (INPUT, OPTIONAL) real(stnd)
!                  On entry, TOL_REORTHO is used to determine if the left singular vectors stored in LEFTVEC
!                  must be reortogonalized on exit in order to correct for the loss of orthogonality in the Ralha-Barlow
!                  one-sided bidiagonal reduction algorithm if MAT is nearly deficient. If one of the singular values,
!                  S(i), verifies the condition
!
!                                              S(i) <= TOL_REORTHO * S(1)
!
!                  all the computed left singular vectors are reorthogonalized with a QR factorization. If S(1) is the
!                  largest singular value of MAT, this condition leads to the assertion that the rank of MAT is less
!                  than size(S) and is thus a nearly singular matrix if TOL_REORTHO is a small positive value of the
!                  order of the machine epsilon.
!
!                  TOL_REORTHO must be greater or equal to zero and less than or equal to one. If TOL_REORTHO = 0. is
!                  used, the left singular vectors are reorthogonalized only if some singular values are almost zero.
!                  On the other hand, If TOL_REORTHO = 1. is used, the left singular vectors are always reorthogonalized.
!                  If TOL_REORTHO is specified as less than zero or greater than one, the default value is used.
!
!                  The default value is the value of the module parameter tol_reortho_def if size( S ) = n and
!                  tol_reortho_partial_def otherwise.
!
!
! Further Details
! _______________
!
!   The singular vectors are computed using deflation techniques applied implicitly to
!   the associated tridiagonal forms BD' * BD and BD * BD' of the bidiagonal matrix BD. See
!   description of the BD_DEFLATE subroutine for more details.
!
!   The computation of the singular vectors is parallelized if OPENMP is used.
!
!   It is essential that singular values given on entry of BD_DEFLATE2 are computed to
!   high (relative) accuracy. Subroutines BD_SINGVAL or BD_SINGVAL2 may be used for this
!   purpose.
!
!   BD_DEFLATE2 may fail if some the singular values specified in parameter S are nearly
!   identical or for clusters of small singular values for some pathological matrices.
!
!   The deflation algorithms used in BD_DEFLATE2 are competitive with the inverse iteration
!   procedure implemented in BD_INVITER2.
!
!   For further details on the deflation techniques or the blocked back-transformation
!   algorithm used in BD_DEFLATE2, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (3) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!   (4) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error5, svd_error8, allocate_error
    use Utilities,       only : assert_eq, merror, ifirstloc
    use QR_Procedures,   only : qr_cmp, ortho_gen_qr, apply_q_qr
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:,:)   :: mat, p
    real(stnd),  intent(in),  dimension(:)     :: d, e, s
    real(stnd),  intent(in),  optional         :: tol_reortho
    real(stnd),  intent(out), dimension(:,:)   :: leftvec, rightvec
!
    integer(i4b), intent(in), optional :: max_qr_steps
!
    logical(lgl), intent(in),  optional :: ortho, scaling, inviter
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: tol, tmp
    real(stnd), dimension(:),   allocatable :: diag, beta
    real(stnd), dimension(:,:), allocatable :: leftvec0, rightvec0
!
    integer(i4b) :: i, n, m, nsing, nrank
    integer      :: iok
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_deflate2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size( mat,     1 ),i4b)  ,   &
                   int(size( leftvec, 1 ),i4b)  ,   &
                   name_proc )
!
    n = assert_eq( (/ int(size( mat,      2 ),i4b)    ,  &
                      int(size( rightvec, 1 ),i4b)    ,  &
                      int(size( p,        1 ),i4b)    ,  &
                      int(size( p,        2 ),i4b)    ,  &
                      int(size( d )          ,i4b)    ,  &
                      int(size( e )          ,i4b) /) ,  &
                      name_proc )
!
    if ( m<n ) then
!
        call merror( name_proc//svd_error5 )
!
    end if
!
    nsing = assert_eq( int(size(s),i4b)             ,  &
                       int(size( leftvec, 2 ),i4b)  ,  &
                       int(size( rightvec, 2 ),i4b) ,  &
                       name_proc )
!
    if ( nsing==n ) then
        tol = tol_reortho_def
    else
        tol = tol_reortho_partial_def
    end if
!
    if ( present(tol_reortho) ) then
!
        if ( tol_reortho>=zero .and. tol_reortho<=one ) then
!
            tol = tol_reortho
!
        end if
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
!
        return
!
    end if
!
    if ( p(1_i4b,1_i4b)==one ) then
!
!       THE ORTHOGONAL MATRIX P HAS BEEN GENERATED BY bd_cmp2 OR ortho_gen_bd.
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( leftvec0(n,nsing), rightvec0(n,nsing), stat=iok   )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE SINGULAR VECTORS OF THE BIDIAGONAL FORM BD OF mat.
!
        call bd_deflate( true, d, e, s, leftvec0(:n,:nsing), rightvec0(:n,:nsing),         &
                         failure, max_qr_steps=max_qr_steps, ortho=ortho, scaling=scaling, &
                         inviter=inviter  )
!
!       COMPUTE SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY MULTIPLICATION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsing, n, one, mat, m, leftvec0, n, zero, leftvec, m )
!
        call gemm( 'N', 'N', n, nsing, n, one, p, n, rightvec0, n, zero, rightvec, n )
#else
        leftvec(:m,:nsing) = matmul( mat(:m,:n), leftvec0(:n,:nsing) )
!
        rightvec(:n,:nsing) = matmul( p(:n,:n), rightvec0(:n,:nsing) )
#endif
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( leftvec0, rightvec0 )
!
    else if ( p(1_i4b,1_i4b)==-one ) then
!
!       THE ORTHOGONAL MATRIX P IS STORED IN FACTORED FORM BY bd_cmp2.
!
!       ALLOCATE WORK ARRAY.
!
        allocate( leftvec0(n,nsing), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE SINGULAR VECTORS OF THE BIDIAGONAL FORM BD OF mat.
!
        call bd_deflate( true, d, e, s, leftvec0(:n,:nsing), rightvec(:n,:nsing),          &
                         failure, max_qr_steps=max_qr_steps, ortho=ortho, scaling=scaling, &
                         inviter=inviter  )
!
!       COMPUTE LEFT SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY MULTIPLICATION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsing, n, one, mat, m, leftvec0, n, zero, leftvec, m )
#else
        leftvec(:m,:nsing) = matmul( mat(:m,:n), leftvec0(:n,:nsing) )
#endif
!
!       COMPUTE RIGHT SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY BACK TRANSFORMATION.
!
        call apply_q_qr( p(2_i4b:n,2_i4b:n), p(2_i4b:n,1_i4b), rightvec(2_i4b:n,:nsing),   &
                         left=true, trans=false  )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( leftvec0 )
!
    else
!
        call merror( name_proc//svd_error8 )
!
    end if
!
!   DETERMINE THE EFFECTIVE RANK OF mat USING THE TOLERANCE tol .
!
    tmp = max( tol*s(1_i4b), safmin )
!
    nrank = ifirstloc( logical( s(:nsing)<=tmp, lgl ) ) - 1_i4b
!
    if ( nrank<nsing ) then
!
!       REORTOGONALIZE LEFT SINGULAR VECTORS IF REQUIRED.
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( diag(nsing), beta(nsing), stat=iok   )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        call qr_cmp( leftvec(:m,:nsing), diag(:nsing), beta(:nsing) )
!
        call ortho_gen_qr( leftvec(:m,:nsing), beta(:nsing) )
!
        do i = 1_i4b, nsing
!
            if ( diag(i)>=zero ) cycle
!
            leftvec(:m,i) = -leftvec(:m,i)
!        
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( diag, beta )
!
    end if
!
!
! END OF SUBROUTINE bd_deflate2_b
! _______________________________
!
    end subroutine bd_deflate2_b
!
! =========================================================================================
!
    subroutine bd_deflate2_c( mat, tauq, taup, rlmat, d, e, s, leftvec, rightvec, failure,  &
                              tauo, max_qr_steps, ortho, scaling, inviter )
!
! Purpose
! _______
!
!   BD_DEFLATE2 computes the left and right singular vectors of a full real m-by-n
!   matrix MAT corresponding to specified singular values, using deflation techniques.
!
!   It is required that the original matrix MAT has been reduced to upper bidiagonal
!   form BD by a two-step algorithm as performed by BD_CMP subroutine with parameters
!   TAUQ, TAUP, RLMAT, and eventually TAUO, or more simply by SELECT_SINGVAL_CMP or
!   SELECT_SINGVAL_CMP2 subroutines with the same arguments:
!
!   - If m >= n, a QR factorization of the real m-by-n matrix MAT is first computed
!
!                        MAT = O * R
!
!   where O is orthogonal and R is upper triangular. In a second step, the n-by-n upper
!   triangular matrix R is reduced to upper bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * R * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix.
!
!   - If m < n, an LQ factorization of the real m-by-n matrix MAT is first computed
!
!                       MAT = L * O
!
!   where O is orthogonal and L is lower triangular. In a second step, the m-by-m lower
!   triangular matrix L is reduced to upper bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * L * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix.
!
!   After this call to BD_CMP with parameters TAUQ, TAUP, RLMAT, and eventually TAUO,
!   the user can call BD_SINGVAL or BD_SINGVAL2 subroutines for computing singular values
!   of BD and, finally, BD_DEFLATE2 for computing all or selected singular vectors of MAT.
!
!   Alternatively and more simply, the user can call SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2
!   subroutines, with  parameters TAUQ, TAUP, RLMAT, and eventually TAUO, to perform the two-stage
!   bidiagonal reduction and get singular values in one step and, finally, call BD_DEFLATE2 for
!   computing all or selected singular vectors of MAT.
!
!   
! Arguments
! _________
!                
!   MAT            (INPUT) real(stnd), dimension(:,:)
!                  On entry, the original m-by-n matrix after a two-stage bidiagonal reduction
!                  by BD_CMP, SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 subroutines (e.g.
!                  when the matrix argument RLMAT is present in the call to these subroutines).
!                  The precise content of MAT is determined by the presence or absence of the
!                  optional argument TAUO:
!
!                  - if the optional argument TAUO is absent, it is assumed that the
!                    first n columns (if m>=n) or the first m rows (if m<n) of the
!                    orthogonal matrix O is stored explicitly in the argument MAT on entry;
!                  - if the optional argument TAUO is present, it is assumed that the
!                    orthogonal matrix O is stored in factored form, as a product of
!                    elementary reflectors, in the argument MAT on entry. In this case,
!                    MAT must contains the vectors which define the elementary reflectors
!                    W(i) whose products determine the matrix O, as returned by BD_CMP,
!                    SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 subroutines when the matrix
!                    argument RLMAT is present in the call to these subroutines.
!
!                  In both cases, MAT must be specified as returned by BD_CMP and is not
!                  modified by the routine.
!                
!   TAUQ           (INPUT) real(stnd), dimension(:)
!                  TAUQ(i) must contain the scalar factor of the elementary
!                  reflector H(i) which determines Q, as returned by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 in their array
!                  argument TAUQ.
!
!                  The size of TAUQ must verify:
!                  size( TAUQ ) = min( size(MAT,1) , size(MAT,2) ) .
!
!   TAUP           (INPUT) real(stnd), dimension(:)
!                  TAUP(i) must contain the scalar factor of the elementary
!                  reflector G(i), which determines P, as returned by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 in their array
!                  argument TAUP.
!
!                  The size of TAUP must verify:
!                  size( TAUP ) = min( size(MAT,1) , size(MAT,2) ) .
!
!   RLMAT          (INPUT) real(stnd), dimension(:,:)
!                  On entry, the elements on and below the diagonal,
!                  with the array TAUQ, represent the orthogonal
!                  matrix Q as a product of elementary reflectors, 
!                  and the elements above the diagonal, with the
!                  array TAUP, represent the orthogonal matrix P
!                  as a product of elementary reflectors.
!                  RLMAT must be specified as returned by BD_CMP,
!                  SELECT_SINGVAL_CMP or SELECT_SINGVAL_CMP2 and
!                  is not modified by the routine.
!
!                  The shape of RLMAT must verify:
!                  size( RLMAT, 1 ) = size( RLMAT, 2 ) = min( size(MAT,1) , size(MAT,2) ).
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, D contains the diagonal elements of the upper bidiagonal
!                  matrix BD as returned by BD_CMP, SELECT_SINGVAL_CMP or
!                  SELECT_SINGVAL_CMP2.
!
!                  The size of D must verify: size( D ) = min( size(MAT,1) , size(MAT,2) ) . 
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, E contains the off-diagonal elements of the upper bidiagonal
!                  matrix BD as returned by BD_CMP, SELECT_SINGVAL_CMP or
!                  SELECT_SINGVAL_CMP2:
!
!                       E(i) = BD(i-1,i) for i = 2,3,...,min(m,n);
!
!                  E(1) is arbitrary.
!
!                  The size of E must verify: size( E ) = min( size(MAT,1) , size(MAT,2) ) . 
!
!   S              (INPUT) real(stnd), dimension(:)
!                  On entry, selected singular values of the upper bidiagonal matrix BD.
!                  The singular values must be given in decreasing order and are
!                  assumed to be positive or zero.
!
!                  The size of S must verify: size( S ) <= min( size(MAT,1) , size(MAT,2) ) .
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed left singular vectors. The left singular vector associated
!                  with the singular value S(j) is stored in the j-th column of LEFTVEC.
!
!                  The shape of LEFTVEC must verify: 
!
!                  - size( LEFTVEC, 1 ) = size( MAT, 1 ) = m ,
!                  - size( LEFTVEC, 2 ) = size( S ) .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed right singular vectors. The right singular vector associated
!                  with the singular value S(j) is stored in the j-th column of RIGHTVEC.
!
!                  The shape of RIGHTVEC must verify:
!
!                  - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ,
!                  - size( RIGHTVEC, 2 ) = size( S ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false :  indicates successful exit ;
!                  - FAILURE = true  :  indicates that the algorithm did not converge and
!                    that full accuracy was not attained in the deflation procedure
!                    of the bidiagonal matrix BD .
!
!   TAUO           (INPUT, OPTIONAL) real(stnd), dimension(:)
!                  The scalar factors of the elementary reflectors W(i), which
!                  represent the orthogonal matrix O of the QR or LQ decomposition of MAT.
!
!                  If the optional argument TAUO is present, it is assumed that the orthogonal matrix O is stored
!                  in factored form, as a product of elementary reflectors, in the argument MAT on entry.
!
!                  If the optional argument TAUO is absent, it is assumed that the orthogonal matrix O is stored
!                  explicitly in the argument MAT on entry.
!
!                  If the optional argument TAUO has been specified in the initial call to the BD_CMP subroutine,
!                  this optional argument TAUO must also be specified in the call to BD_DEFLATE2, otherwise the results
!                  will be incorrect.
!
!                  See description of the argument MAT in the description of the BD_CMP subroutine,
!                  when the argument RLMAT is also present, for further details.
!
!                  The size of TAUO must be min( size(MAT,1) , size(MAT,2) ).
!
!   MAX_QR_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                  MAX_QR_STEPS controls the maximum number of QR sweeps for deflating
!                  the bidiagonal matrix BD for a given singular value.
!                  The algorithm fails to converge if the total number of QR sweeps for
!                  all singular values exceeds MAX_QR_STEPS * size(S).
!
!                  The default is 4.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry:
!
!                  - ORTHO=true, the bidiagonal matrix BD is deflated sequentially
!                    for all the specified singular values; this implies that the singular
!                    vectors of the bidiagonal matrix BD will be automatically orthogonal
!                    on exit.
!                  - ORTHO=false, the bidiagonal matrix BD is deflated in parallel for the
!                    different clusters of singular values or isolated singular values; this
!                    implies that orthogonality of the singular vectors of bidiagonal
!                    matrix BD is preserved inside each cluster, but not
!                    automatically between clusters.
!
!                  The default is ORTHO=false.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the intermediate bidiagonal matrix BD is scaled before computing
!                    the deflation parameters in order to avoid overflows;
!                  - SCALING=false, the intermediate bidiagonal matrix BD is not scaled.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INVITER        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INVITER=true, singular vectors corresponding to isolated
!                    singular values or singular vectors of bidiagonal matrices with
!                    zeros are computed by inverse iteration instead of deflation.
!                  - INVITER=false, singular vectors corresponding to isolated
!                    singular values or singular vectors of bidiagonal matrices
!                    with zeros are computed by deflation.
!
!                  The default is INVITER=true.
!
!
! Further Details
! _______________
!
!   The singular vectors of BD are computed using deflation techniques applied implicitly to
!   the associated tridiagonal forms BD' * BD and BD * BD' of the bidiagonal matrix BD. See
!   description of the BD_DEFLATE subroutine for more details.
!
!   The singular vectors of MAT are finally computed by a blocked back-transformation algorithm.
!
!   The computation of the singular vectors of BD and the blocked back-transformation algorithm
!   to find the singular vectors of MAT are parallelized if OPENMP is used.
!
!   It is essential that singular values given on entry of BD_DEFLATE2 are computed to
!   high (relative) accuracy. Subroutines BD_SINGVAL or BD_SINGVAL2 may be used for this
!   purpose.
!
!   BD_DEFLATE2 may fail if some the singular values specified in parameter S are nearly
!   identical or for clusters of small singular values for some pathological matrices.
!
!   The deflation algorithms used in BD_DEFLATE2 are competitive with the inverse iteration
!   procedure implemented in BD_INVITER2.
!
!   For further details on the deflation techniques or the blocked back-transformation
!   algorithm used in BD_DEFLATE2, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (3) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!   (4) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, assert_eq, merror
    use QR_Procedures,   only : ortho_gen_qr, apply_q_qr, apply_q_lq
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),           dimension(:,:) :: mat, rlmat
    real(stnd),  intent(in),           dimension(:)   :: tauq, taup, d, e, s
    real(stnd),  intent(in), optional, dimension(:)   :: tauo
    real(stnd),  intent(out),          dimension(:,:) :: leftvec, rightvec
!
    integer(i4b), intent(in), optional :: max_qr_steps
!
    logical(lgl), intent(in),  optional :: ortho, scaling, inviter
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(:,:), allocatable :: work, work2
!
    integer(i4b) :: n, m, nm, nsing, n1, mnthr_nsing
    integer      :: iok
!
    logical(lgl) :: upper, use_gen_o
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_deflate2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size( mat,     1 ),i4b)  ,   &
                   int(size( leftvec, 1 ),i4b)  ,   &
                   name_proc )
!
    n = assert_eq( int(size( mat,      2 ),i4b)  ,   &
                   int(size( rightvec, 1 ),i4b)  ,   &
                   name_proc )
!
    nm = assert_eq( (/ min(m,n)                  , &
                       int(size(tauq),i4b)       , &
                       int(size(taup),i4b)       , &
                       int(size(d),i4b)          , &
                       int(size(e),i4b)          , &
                       int(size(rlmat,1),i4b)    , &
                       int(size(rlmat,2),i4b)  /), &
                       name_proc )
!
    nsing = assert_eq( int(size(s),i4b)             ,  &
                       int(size( leftvec, 2 ),i4b)  ,  &
                       int(size( rightvec, 2 ),i4b) ,  &
                       name_proc )
!
    use_gen_o = true
!
    if ( present( tauo ) ) then
!
        call assert( logical( nm==size(tauo), lgl ), name_proc )
!
        use_gen_o = false
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( nm<=0_i4b ) then
!
        return
!
    end if
!
    upper = m >= n
!
!   COMPUTE SINGULAR VECTORS OF THE BIDIAGONAL FORM BD OF mat.
!
    call bd_deflate( true, d, e, s, leftvec(:nm,:nsing), rightvec(:nm,:nsing),         &
                     failure, max_qr_steps=max_qr_steps, ortho=ortho, scaling=scaling, &
                     inviter=inviter  )
!
!   COMPUTE SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY BACK-TRANSFORMATION.
!
!   FIRST GENERATE LEFT AND RIGHT SINGULAR VECTORS OF R OR L IN leftvec AND rightvec, respectively.
!
    mnthr_nsing = int( real( nm, stnd )*thr_bt, i4b )
!
    if ( nsing>=mnthr_nsing  ) then
!
!       ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM IF NEARLY FULL SVD OF mat IS
!       REQUESTED.
!
        allocate( work(nm,nm), work2(nm,nsing), stat = iok )
!
    else
!
!       USE DIRECT ALGORITHM OTHERWISE OR IF THE REQUIRED
!       WORKSPACE IS NOT AVAILABLE.
!
        iok = -1
!
    end if
!
    if ( iok==0 ) then
!
!       GENERATE P IN work(:nm,:nm) .
!
        call ortho_gen_p_bd( rlmat, taup, work(:nm,:nm) )
!
!       MULTIPLY P IN work BY RIGHT SINGULAR VECTORS OF BD IN work2
!       FOR COMPUTING RIGHT SINGULAR VECTORS OF R OR L IN rightvec(:nm,:nsing) .
!
        work2(:nm,:nsing) = rightvec(:nm,:nsing)
!
#ifdef _BLAS
        call gemm( 'N', 'N', nm, nsing, nm, one, work, nm, work2, nm, zero, rightvec, n  )
#else
        rightvec(:nm,:nsing) = matmul( work(:nm,:nm), work2(:nm,:nsing) )
#endif
!
!       GENERATE Q IN work(:nm,:nm) .
!
        work(:nm,:nm) = rlmat(:nm,:nm)
!
        call ortho_gen_q_bd( work, tauq )
!
!       MULTIPLY Q IN work BY LEFT SINGULAR VECTORS OF BD IN work2
!       FOR COMPUTING LEFT SINGULAR VECTORS OF R OR L IN leftvec(:nm,:nsing) .
!
        work2(:nm,:nsing) = leftvec(:nm,:nsing)
!
#ifdef _BLAS
        call gemm( 'N', 'N', nm, nsing, nm, one, work, nm, work2, nm, zero, leftvec, m   )
#else
        leftvec(:nm,:nsing) = matmul( work(:nm,:nm), work2(:nm,:nsing) )
#endif
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( work )
!
    else
!
!       GENERATE LEFT SINGULAR VECTORS OF R OR L IN leftvec .
!
        call apply_q_bd( rlmat, tauq, leftvec(:nm,:nsing),  left=true, trans=false )
!
!       GENERATE RIGHT SINGULAR VECTORS OF R OR L IN rightvec .
!
        call apply_p_bd( rlmat, taup, rightvec(:nm,:nsing), left=true, trans=false )
!
    end if
!
    if ( use_gen_o ) then
!
!       THE ORTHOGONAL MATRIX O IS STORED IN EXPLICIT FORM IN mat .
!
        if ( upper ) then
!
            if ( iok==-1 ) then
!
!               ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
                allocate( work2(n,nsing), stat = iok )
!
                if ( iok/=0 ) then
                    call merror( name_proc//allocate_error )
                end if
!
            end if
!
!           COPY THE LEFT SINGULAR VECTORS OF R IN work .
!
            work2(:n,:nsing) = leftvec(:n,:nsing)
!
!           COMPUTE THE LEFT SINGULAR VECTORS OF mat IN leftvec .
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, nsing, n, one, mat, m, work2, n, zero, leftvec, m )
#else
            leftvec(:m,:nsing) = matmul( mat(:m,:n), work2(:n,:nsing) )
#endif
!
        else
!
#ifdef _BLAS
!
            if ( iok==-1 ) then
!
!               ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
                allocate( work2(m,nsing), stat = iok )
!
                if ( iok/=0 ) then
                    call merror( name_proc//allocate_error )
                end if
!
            end if
!
!           COPY THE RIGHT SINGULAR VECTORS OF R IN work .
!
            work2(:m,:nsing) = rightvec(:m,:nsing)
!
            call gemm( 'T', 'N', n, nsing, m, one, mat, m, work2, m, zero, rightvec, n )
!
#else
!
            if ( iok==0 ) then
!
!               DEALLOCATE WORKSPACE.
!
                deallocate( work2 )
!
            end if
!
!           REALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
            allocate( work(nsing,m), work2(nsing,n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           TRANSPOSE THE RIGHT SINGULAR VECTORS OF R IN work .
!
            work(:nsing,:m) = transpose( rightvec(:m,:nsing) )
!
!           COMPUTE TRANSPOSE OF THE RIGHT SINGULAR VECTORS OF mat IN work2 .
!
            work2(:nsing,:n) = matmul( work(:nsing,:m), mat(:m,:n) )
!
!           SAVE THE RIGHT SINGULAR VECTORS OF mat IN rightvec .
!
            rightvec(:n,:nsing) = transpose( work2(:nsing,:n) )
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( work )
#endif
!
        end if
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( work2 )
!
    else
!
!       THE ORTHOGONAL MATRIX O IS STORED IN FACTORED FORM IN mat AND tauo .
!
        if ( nsing>=mnthr_nsing ) then
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM IF NEARLY FULL SVD OF mat IS
!           REQUESTED.
!
            if ( upper ) then
                n1 = m
            else
                n1 = n
            end if
!
            if ( iok==0 ) then
!
                allocate( work(n1,nm), stat = iok )
!
            else
!
                allocate( work(n1,nm), work2(nm,nsing), stat = iok )
!
            end if
!
        else
!
!           USE DIRECT ALGORITHM OTHERWISE OR IF THE REQUIRED
!           WORKSPACE IS NOT AVAILABLE.
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!
            if ( upper ) then
!
!               GENERATE O IN work .
!
                work(:m,:n) = mat(:m,:n)
!
                call ortho_gen_qr( work, tauo )
!
!               MULTIPLY O IN work BY LEFT SINGULAR VECTORS OF R IN leftvec .
!
                work2(:n,:nsing) = leftvec(:n,:nsing)
!
#ifdef _BLAS
                call gemm( 'N', 'N', m, nsing, n, one, work, m, work2, n, zero, leftvec, m )
#else
                leftvec(:m,:nsing) = matmul( work(:m,:n), work2(:n,:nsing) )
#endif
!
            else
!
!               GENERATE O**(t) IN work .
!
                work(:n,:m) = transpose( mat(:m,:n) )
!
                call ortho_gen_qr( work, tauo )
!
!               MULTIPLY O**(t) IN work BY RIGHT SINGULAR VECTORS OF L IN rightvec .
!
                work2(:m,:nsing) = rightvec(:m,:nsing)
!
#ifdef _BLAS
                call gemm( 'N', 'N', n, nsing, m, one, work, n, work2, m, zero, rightvec, n )
#else
                rightvec(:n,:nsing) = matmul( work(:n,:m), work2(:m,:nsing) )
#endif
!
            end if
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( work, work2 )
!
        else
!
            if ( upper ) then
!
!               MULTIPLY O IN mat BY LEFT SINGULAR VECTORS OF R IN leftvec .
!
                leftvec(nm+1_i4b:m,:nsing) = zero
!
                call apply_q_qr( mat, tauo, leftvec(:m,:nsing), left=true, trans=false )
!
            else
!
!               MULTIPLY O**(t) IN mat BY RIGHT SINGULAR VECTORS OF L IN rightvec .
!
                rightvec(nm+1_i4b:n,:nsing) = zero
!
                call apply_q_lq( mat, tauo, rightvec(:n,:nsing), left=true, trans=true )
!
            end if
!
        end if
!
    end if
!
!
! END OF SUBROUTINE bd_deflate2_c
! _______________________________
!
    end subroutine bd_deflate2_c
!
! =========================================================================================
!
    subroutine bd_deflate2_d( mat, rmat, p, d, e, s, leftvec, rightvec, failure,       &
                             tauo,  max_qr_steps, ortho, scaling, inviter, tol_reortho )
!
! Purpose
! _______
!
!   BD_DEFLATE2 computes all or selected left and right singular vectors of a full real m-by-n
!   matrix MAT with m>=n corresponding to specified singular values, using deflation techniques.
!
!   It is required that the original matrix MAT has been reduced to upper bidiagonal
!   form BD by a two-step algorithm as performed by SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4
!   subroutines with parameters P, RMAT, and eventually TAUO:
!
!   A QR factorization of the real m-by-n matrix MAT is first computed
!
!                        MAT = O * R
!
!   where O is orthogonal and R is upper triangular. In a second step, the n-by-n upper
!   triangular matrix R is reduced to upper bidiagonal form BD
!   by an orthogonal transformation :
!
!                       Q' * R * P = BD
!
!   where Q and P are orthogonal and BD is an upper bidiagonal matrix. Subroutines
!   SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4 computes O, Q, P, BD and also all or some
!   of the singular values of R, which are also the singular values of MAT. Using this two-step
!   factorization, BD_DEFLATE2 computes all or selected left and right singular vectors of R and
!   apply to them a back-transformation algorithm to obtain the corresponding left and right singular
!   vectors of MAT.
!
!   
! Arguments
! _________
!
!   MAT            (INPUT) real(stnd), dimension(:,:)
!                  On entry, the original m-by-n matrix after the two-step bidiagonal reduction by
!                  SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4 subroutines with arguments P, RMAT,
!                  and eventually TAUO.
!                  The precise content of MAT is determined by the presence or absence of the
!                  optional argument TAUO:
!
!                  - if the optional argument TAUO is absent, it is assumed that the
!                    first n columns of the orthogonal matrix O is stored explicitly
!                    in the argument MAT on entry,
!                  - if the optional argument TAUO is present, it is assumed that the
!                    orthogonal matrix O is stored in factored form, as a product of
!                    elementary reflectors, in the argument MAT on entry. In this case,
!                    MAT must contains the vectors which define the elementary reflectors
!                    W(i) whose products determine the matrix O, as returned by
!                    SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4 subroutines when the matrix
!                    argument RLMAT is present in the call to these subroutines.
!
!                  In both cases, MAT must be specified as returned by these subroutines and
!                  is not modified by the routine. See the description of the argument TAUO
!                  below for further details.
!                
!                  The shape of MAT must verify: size( MAT, 1 ) >= size( MAT, 2 ) = n .
!
!   RMAT           (INPUT) real(stnd), dimension(:,:)
!                  On entry, the n-by-n orthogonal matrix Q after reduction by SELECT_SINGVAL_CMP3 or
!                  SELECT_SINGVAL_CMP4 subroutines.
!                  RMAT must be specified as returned by these subroutines and is not modified by the routine.
!
!                  The shape of RMAT must verify: size( RMAT, 1 ) = size( RMAT, 2 ) = size( MAT, 2 ) = n.
!
!   P              (INPUT) real(stnd), dimension(:,:)
!                  On entry, the n-by-n orthogonal matrix P after reduction by SELECT_SINGVAL_CMP3 or
!                  SELECT_SINGVAL_CMP4 subroutines. P can be stored in factored form or not.
!                  Both cases are handled by the subroutine and P is not modified by the routine.
!                
!                  The shape of P must verify: size( P, 1 ) = size( P, 2 ) = size( MAT, 2 ) = n .
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, D contains the diagonal elements of the bidiagonal
!                  matrix BD as returned by SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4.
!
!                  The size of D must verify: size( D ) = size( MAT, 2 ) = n . 
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, E contains the off-diagonal elements of the bidiagonal
!                  matrix BD as returned by SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4:
!
!                          E(i) = BD(i-1,i) for i = 2,3,...,n;
!
!                  E(1) is arbitrary.
!
!                  The size of E must verify: size( E ) = size( MAT, 2 ) = n . 
!
!   S              (INPUT) real(stnd), dimension(:)
!                  On entry, selected singular values of the bidiagonal matrix BD.
!                  The singular values must be given in decreasing order and are
!                  assumed to be positive or zero.
!
!                  The size of S must verify: size( S ) <= size( MAT, 2 ) .
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed left singular vectors. The left singular vector
!                  associated with the singular value S(j) is stored in the j-th column
!                  of LEFTVEC.
!
!                  The shape of LEFTVEC must verify:  
!
!                  - size( LEFTVEC, 1 ) = size( MAT, 1 ) = m ,
!                  - size( LEFTVEC, 2 ) = size( S ) .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed right singular vectors. The right singular vector
!                  associated with the singular value S(j) is stored in the j-th column of
!                  RIGHTVEC.
!
!                  The shape of RIGHTVEC must verify:  
!
!                  - size( RIGHTVEC, 1 ) = size( MAT, 2 ) = n ,
!                  - size( RIGHTVEC, 2 ) = size( S ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false :  indicates successful exit,
!                  - FAILURE = true  :  indicates that the algorithm did not converge and
!                    that full accuracy was not attained in the deflation procedure
!                    of the bidiagonal matrix BD.
!
!   TAUO           (INPUT, OPTIONAL) real(stnd), dimension(:)
!                  The scalar factors of the elementary reflectors W(i), which
!                  represent the orthogonal matrix O of the QR decomposition of MAT as returned by
!                  SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4.
!
!                  If the optional argument TAUO is present, it is assumed that the orthogonal matrix O is stored
!                  in factored form, as a product of elementary reflectors, in the argument MAT on entry.
!
!                  If the optional argument TAUO is absent, it is assumed that the orthogonal matrix O is stored
!                  explicitly in the argument MAT on entry.
!
!                  If the optional argument TAUO has been specified in the initial call to the SELECT_SINGVAL_CMP3 or
!                  SELECT_SINGVAL_CMP4 subroutines, this optional argument TAUO must also be specified in the call to
!                  BD_DEFLATE2, otherwise the results will be incorrect.
!
!                  See description of the argument MAT in the description of the SELECT_SINGVAL_CMP3 or
!                  SELECT_SINGVAL_CMP4 subroutines, when the argument RMAT is also present, for further details.
!
!                  The size of TAUO must be size(MAT,2) = n .
!
!   MAX_QR_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                  MAX_QR_STEPS controls the maximum number of QR sweeps for deflating
!                  the bidiagonal matrix BD for a given singular value.
!                  The algorithm fails to converge if the total number of QR sweeps for
!                  all singular values exceeds MAX_QR_STEPS * size(S).
!
!                  The default is 4.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry:
!
!                  - ORTHO=true, the bidiagonal matrix BD is deflated sequentially
!                    for all the specified singular values; this implies that the singular
!                    vectors of the bidiagonal matrix BD will be automatically orthogonal
!                    on exit;
!                  - ORTHO=false, the bidiagonal matrix BD is deflated in parallel for the
!                    different clusters of singular values or isolated singular values; this
!                    implies that orthogonality of the singular vectors of bidiagonal
!                    matrix BD is preserved inside each cluster, but not
!                    automatically between clusters.
!
!                  The default is ORTHO=false.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the intermediate bidiagonal matrix BD is scaled before computing
!                    the deflation parameters in order to avoid overflows;
!                  - SCALING=false, the intermediate bidiagonal matrix BD is not scaled.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INVITER        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INVITER=true, singular vectors corresponding to isolated
!                    singular values or singular vectors of bidiagonal matrices with
!                    zeros are computed by inverse iteration instead of deflation.
!                  - INVITER=false, singular vectors corresponding to isolated
!                    singular values or singular vectors of bidiagonal matrices
!                    with zeros are computed by deflation.
!
!                  The default is INVITER=true.
!
!   TOL_REORTHO    (INPUT, OPTIONAL) real(stnd)
!                  On entry, TOL_REORTHO is used to determine if the left singular vectors stored in LEFTVEC
!                  must be reortogonalized on exit in order to correct for the loss of orthogonality in the Ralha-Barlow
!                  one-sided bidiagonal reduction algorithm if MAT is nearly deficient. If one of the singular values,
!                  S(i), verifies the condition
!
!                                              S(i) <= TOL_REORTHO * S(1)
!
!                  all the computed left singular vectors are reorthogonalized with a QR factorization. If S(1) is the
!                  largest singular value of MAT, this condition leads to the assertion that the rank of MAT is less
!                  than size(S) and is thus a nearly singular matrix if TOL_REORTHO is a small positive value of the
!                  order of the machine epsilon.
!
!                  TOL_REORTHO must be greater or equal to zero and less than or equal to one. If TOL_REORTHO = 0. is
!                  used, the left singular vectors are reorthogonalized only if some singular values are almost zero.
!                  On the other hand, If TOL_REORTHO = 1. is used, the left singular vectors are always reorthogonalized.
!                  If TOL_REORTHO is specified as less than zero or greater than one, the default value is used.
!
!                  The default value is the value of the module parameter tol_reortho_def if size( S ) = n and
!                  tol_reortho_partial_def otherwise.
!
!
! Further Details
! _______________
!
!   The singular vectors are computed using deflation techniques applied implicitly to
!   the associated tridiagonal forms BD' * BD and BD * BD' of the bidiagonal matrix BD. See
!   description of the BD_DEFLATE subroutine for more details.
!
!   The computation of the singular vectors is parallelized if OPENMP is used.
!
!   It is essential that singular values given on entry of BD_DEFLATE2 are computed to
!   high (relative) accuracy. Subroutines SELECT_SINGVAL_CMP3 or SELECT_SINGVAL_CMP4 may
!   be used for this purpose.
!
!   BD_DEFLATE2 may fail if some the singular values specified in parameter S are nearly
!   identical or for clusters of small singular values for some pathological matrices.
!
!   The deflation algorithms used in BD_DEFLATE2 are competitive with the inverse iteration
!   procedure implemented in BD_INVITER2.
!
!   For further details on the deflation techniques or the blocked back-transformation
!   algorithm used in BD_DEFLATE2, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (3) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!   (4) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : svd_error5, svd_error8, allocate_error
    use Utilities,       only : assert, assert_eq, merror, ifirstloc
    use QR_Procedures,   only : qr_cmp, ortho_gen_qr, apply_q_qr
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),           dimension(:,:) :: mat, rmat, p
    real(stnd),  intent(in),           dimension(:)   :: d, e, s
    real(stnd),  intent(in), optional, dimension(:)   :: tauo
    real(stnd),  intent(in), optional                 :: tol_reortho
    real(stnd),  intent(out),          dimension(:,:) :: leftvec, rightvec
!
    integer(i4b), intent(in), optional :: max_qr_steps
!
    logical(lgl), intent(in),  optional :: ortho, scaling, inviter
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: tol, tmp
    real(stnd), dimension(:),   allocatable :: diag, beta
    real(stnd), dimension(:,:), allocatable :: leftvec0, work
!
    integer(i4b) :: i, n, m, nsing, mnthr_nsing, nrank
    integer      :: iok
!
    logical(lgl) :: use_gen_o
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_deflate2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size( mat,     1 ),i4b)  ,   &
                   int(size( leftvec, 1 ),i4b)  ,   &
                   name_proc )
!
    n = assert_eq( (/ int(size( mat,      2 ),i4b)    ,  &
                      int(size( rightvec, 1 ),i4b)    ,  &
                      int(size( rmat,     1 ),i4b)    ,  &
                      int(size( rmat,     2 ),i4b)    ,  &
                      int(size( p,        1 ),i4b)    ,  &
                      int(size( p,        2 ),i4b)    ,  &
                      int(size( d )          ,i4b)    ,  &
                      int(size( e )          ,i4b) /) ,  &
                      name_proc )
!
    if ( m<n ) then
!
        call merror( name_proc//svd_error5 )
!
    end if
!
    nsing = assert_eq( int(size(s),i4b)             ,  &
                       int(size( leftvec, 2 ),i4b)  ,  &
                       int(size( rightvec, 2 ),i4b) ,  &
                       name_proc )
!
    use_gen_o = true
!
    if ( present( tauo ) ) then
!
        call assert( logical( n==size(tauo), lgl ), name_proc )
!
        use_gen_o = false
!
    end if
!
    if ( nsing==n ) then
        tol = tol_reortho_def
    else
        tol = tol_reortho_partial_def
    end if
!
    if ( present(tol_reortho) ) then
!
        if ( tol_reortho>=zero .and. tol_reortho<=one ) then
!
            tol = tol_reortho
!
        end if
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
!
        return
!
    end if
!
!   FIRST STEP: COMPUTE SINGULAR VECTORS OF THE UPPER n-BY-n TRIANGULAR MATRIX R.
!
    if ( p(1_i4b,1_i4b)==one ) then
!
!       THE ORTHOGONAL MATRIX P HAS BEEN GENERATED BY select_singval_cmp3 OR select_singval_cmp4.
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( leftvec0(n,nsing), work(n,nsing), stat=iok   )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE SINGULAR VECTORS OF THE BIDIAGONAL FORM BD OF mat.
!
        call bd_deflate( true, d, e, s, leftvec(:n,:nsing), work(:n,:nsing),               &
                         failure, max_qr_steps=max_qr_steps, ortho=ortho, scaling=scaling, &
                         inviter=inviter  )
!
!       COMPUTE RIGHT SINGULAR VECTORS OF THE UPPER n-BY-n TRIANGULAR MATRIX R BY MULTIPLICATION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', n, nsing, n, one, p, n, work, n, zero, rightvec, n )
#else
        rightvec(:n,:nsing) = matmul( p(:n,:n), work(:n,:nsing) )
#endif
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( work )
!
    else if ( p(1_i4b,1_i4b)==-one ) then
!
!       THE ORTHOGONAL MATRIX P IS STORED IN FACTORED FORM BY select_singval_cmp3 OR select_singval_cmp4.
!
!       ALLOCATE WORK ARRAY.
!
        allocate( leftvec0(n,nsing), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE SINGULAR VECTORS OF THE BIDIAGONAL FORM BD OF mat.
!
        call bd_deflate( true, d, e, s, leftvec(:n,:nsing), rightvec(:n,:nsing),           &
                         failure, max_qr_steps=max_qr_steps, ortho=ortho, scaling=scaling, &
                         inviter=inviter  )
!
!       COMPUTE RIGHT SINGULAR VECTORS OF THE FULL m-BY-n MATRIX mat BY BACK TRANSFORMATION.
!
        call apply_q_qr( p(2_i4b:n,2_i4b:n), p(2_i4b:n,1_i4b), rightvec(2_i4b:n,:nsing),   &
                         left=true, trans=false  )
!
    else
!
        call merror( name_proc//svd_error8 )
!
    end if
!
!   COMPUTE LEFT SINGULAR VECTORS OF THE UPPER n-BY-n TRIANGULAR MATRIX R BY MULTIPLICATION.
!
#ifdef _BLAS
    call gemm( 'N', 'N', n, nsing, n, one, rmat, n, leftvec, m, zero, leftvec0, n )
#else
    leftvec0(:n,:nsing) = matmul( rmat(:n,:n), leftvec(:n,:nsing) )
#endif
!
!   SECOND STEP: COMPUTE SINGULAR VECTORS OF THE FULL m-BY-n RECTANGULAR MATRIX MAT
!   BY BACK TRANSFORMATION OF THOSE OF R.
!
    if ( use_gen_o ) then
!
!       THE ORTHOGONAL MATRIX O IS STORED IN EXPLICIT FORM IN mat .
!
!       COMPUTE THE LEFT SINGULAR VECTORS OF mat IN leftvec .
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsing, n, one, mat, m, leftvec0, n, zero, leftvec, m )
#else
        leftvec(:m,:nsing) = matmul( mat(:m,:n), leftvec0(:n,:nsing) )
#endif
!
    else
!
!       THE ORTHOGONAL MATRIX O IS STORED IN FACTORED FORM IN mat AND tauo .
!    
        mnthr_nsing = int( real( n, stnd )*thr_bt, i4b )
!
        if ( nsing>=mnthr_nsing ) then
!
!           ALLOCATE WORKSPACE NEEDED FOR A FAST ALGORITHM IF NEARLY FULL SVD OF mat IS
!           REQUESTED.
!
            allocate( work(m,n), stat = iok )
!
        else
!
!           USE DIRECT ALGORITHM OTHERWISE OR IF THE REQUIRED
!           WORKSPACE IS NOT AVAILABLE.
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!
!           GENERATE O IN work .
!
            work(:m,:n) = mat(:m,:n)
!
            call ortho_gen_qr( work, tauo )
!
!           MULTIPLY O IN work BY LEFT SINGULAR VECTORS OF R IN leftvec0
!           TO GET LEFT SINGULAR VECTORS OF MAT.
!
#ifdef _BLAS
            call gemm( 'N', 'N', m, nsing, n, one, work, m, leftvec0, n, zero, leftvec, m )
#else
            leftvec(:m,:nsing) = matmul( work(:m,:n), leftvec0(:n,:nsing) )
#endif
!
!           DEALLOCATE WORKSPACE.
!
            deallocate( work )
!
        else
!
!           MULTIPLY O STORED IN FACTORED FORM IN mat BY LEFT SINGULAR VECTORS OF R IN leftvec0
!           TO GET LEFT SINGULAR VECTORS OF MAT.
!
            leftvec(:n,:nsing)        = leftvec0(:n,:nsing)
            leftvec(n+1_i4b:m,:nsing) = zero
!
            call apply_q_qr( mat, tauo, leftvec(:m,:nsing), left=true, trans=false )
!
        end if
!
    end if
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( leftvec0 )
!
!   DETERMINE THE EFFECTIVE RANK OF mat USING THE TOLERANCE tol .
!
    tmp = max( tol*s(1_i4b), safmin )
!
    nrank = ifirstloc( logical( s(:nsing)<=tmp, lgl ) ) - 1_i4b
!
    if ( nrank<nsing ) then
!
!       REORTOGONALIZE LEFT SINGULAR VECTORS IF REQUIRED.
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( diag(nsing), beta(nsing), stat=iok   )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        call qr_cmp( leftvec(:m,:nsing), diag(:nsing), beta(:nsing) )
!
        call ortho_gen_qr( leftvec(:m,:nsing), beta(:nsing) )
!
        do i = 1_i4b, nsing
!
            if ( diag(i)>=zero ) cycle
!
            leftvec(:m,i) = -leftvec(:m,i)
!        
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( diag, beta )
!
    end if
!
!
! END OF SUBROUTINE bd_deflate2_d
! _______________________________
!
    end subroutine bd_deflate2_d
!
!
! =========================================================================================
!          SORTING PROCEDURES FOR SINGULAR VALUES AND ASSOCIATED SINGULAR VECTORS
! =========================================================================================
!
!
    subroutine lr_singvec_sort( sort, d, u, v )
!
! Purpose
! _______
!
!   Given the singular values D and singular vectors U and V as output from
!   BD_SVD, SVD_CMP or SVD_CMP3, this subroutine sorts the singular values into ascending
!   or descending order, and, rearranges the columns of U and V correspondingly.
!
!
! Arguments
! _________
!
!   SORT         (INPUT) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!                The singular vectors are rearranged accordingly.
!
!   D            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the singular values.
!
!                On exit, the singular values in ascending or decreasing order.
!
!   U            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the columns of U are the (left) singular vectors. 
!
!                On exit, U contains the rearranged (left) singular vectors.
!
!                The shape of U must verify:  size(U,2) = size( D ) .
!
!   V            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the columns of V are the (right) singular vectors. 
!
!                On exit, V contains the rearranged (right) singular vectors.
!
!                The shape of V must verify:  size(V,2) = size( D ) .
!
!
! Further Details
! _______________
!
!   The method is straight insertion.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq, iminloc
#ifdef _BLAS
    use BLAS_interfaces, only : swap
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)  :: d
    real(stnd), intent(inout), dimension(:,:) :: u, v
!
    character , intent(in)  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _BLAS
    real(stnd) :: tmp
#else
    real(stnd) :: tmp_vec(max(size(u,1),size(v,1))), tmp
#endif
!
    integer(i4b) :: i, l, n, nru, nrv
!   
!   
! PARAMETERS 
! __________
!
    character , parameter ::  a_maj=achar(65),  a_min=achar(97),       &
                              d_maj=achar(68),  d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   CHECK THE ARGUMENTS.
!
    n = assert_eq(    int( size(d), i4b ),     &
                      int( size(u,2), i4b ),   &
                      int( size(v,2), i4b ),   &
                      'svd_sort' )
    nru = size( u, 1 )
    nrv = size( v, 1 )
!
!   SORT THE SINGULAR VALUES.
!   INSERTION SORT ON SINGULAR VALUES, BUT ONLY ONE
!   TRANSPOSITION PER SINGULAR VECTOR.
!
    select case( sort )
!
        case( a_maj, a_min ) 
!
!           SORT IN ASCENDING ORDER.
!
            do l = 1, n - 1_i4b
!
!               SCAN FOR SMALLEST d(l) .
!
                i = iminloc( d(l:n) ) + l - 1_i4b
                if ( i==l ) cycle
!
!               SWAP SINGULAR VALUES AND VECTORS.
!
                tmp  = d(i)
                d(i) = d(l)
                d(l) = tmp
!
#ifdef _BLAS
                call swap( nrv, v(:nrv,l), 1_i4b, v(:nrv,i), 1_i4b )
!
                call swap( nru, u(:nru,l), 1_i4b, u(:nru,i), 1_i4b )
#else
                tmp_vec(:nrv) = v(:nrv,l)
                v(:nrv,l)     = v(:nrv,i)
                v(:nrv,i)     = tmp_vec(:nrv)
!
                tmp_vec(:nru) = u(:nru,l)
                u(:nru,l)     = u(:nru,i)
                u(:nru,i)     = tmp_vec(:nru)
#endif
!
            end do
!
        case( d_maj, d_min ) 
!
!           SORT IN DESCENDING ORDER.
!
            do l = n, 2_i4b, -1_i4b
!
!               SCAN FOR SMALLEST d(l) .
!
                i = iminloc( d(:l) )
                if ( i==l ) cycle
!
!               SWAP SINGULAR VALUES AND VECTORS.
!
                tmp  = d(i)
                d(i) = d(l)
                d(l) = tmp
!
#ifdef _BLAS
                call swap( nrv, v(:nrv,l), 1_i4b, v(:nrv,i), 1_i4b )
!
                call swap( nru, u(:nru,l), 1_i4b, u(:nru,i), 1_i4b )
#else
                tmp_vec(:nrv) = v(:nrv,l)
                v(:nrv,l)     = v(:nrv,i)
                v(:nrv,i)     = tmp_vec(:nrv)
!
                tmp_vec(:nru) = u(:nru,l)
                u(:nru,l)     = u(:nru,i)
                u(:nru,i)     = tmp_vec(:nru)
#endif
!
            end do
!
    end select
!
!
! END OF SUBROUTINE lr_singvec_sort
! _________________________________
!
    end subroutine lr_singvec_sort
!
! =========================================================================================
!
    subroutine lr_singvec_sort2( sort, d, u, vt )
!
! Purpose
! _______
!
!   Given the singular values D and singular vectors U and VT as output from
!   BD_SVD2 or SVD_CMP2, this subroutine sorts the singular values into ascending
!   or descending order, and, rearranges the columns of U and the rows of VT
!   correspondingly.
!
!
! Arguments
! _________
!
!   SORT         (INPUT) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!                The singular vectors are rearranged accordingly.
!
!   D            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the singular values.
!
!                On exit, the singular values in ascending or decreasing order.
!
!   U            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the columns of U are the (left) singular vectors. 
!
!                On exit, U contains the rearranged (left) singular vectors.
!
!                The shape of U must verify:  size(U,2) = size( D ) .
!
!   VT           (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the rows of VT are the (right) singular vectors. 
!
!                On exit, VT contains the rearranged (right) singular vectors.
!
!                The shape of VT must verify:  size(VT,1) = size( D ) .
!
!
! Further Details
! _______________
!
!   The method is straight insertion.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq, iminloc
#ifdef _BLAS
    use BLAS_interfaces, only : swap
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)   :: d
    real(stnd), intent(inout), dimension(:,:) :: u, vt
!
    character , intent(in)  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _BLAS
    real(stnd) :: tmp_vec(size(vt,2)), tmp
#else
    real(stnd) :: tmp_vec(max(size(u,1),size(vt,2))), tmp
#endif
!
    integer(i4b) :: i, l, n, nru, ncvt
!   
!   
! PARAMETERS 
! __________
!
    character , parameter ::  a_maj=achar(65),  a_min=achar(97),       &
                              d_maj=achar(68),  d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   CHECK THE ARGUMENTS.
!
    n = assert_eq(    int( size(d), i4b ),     &
                      int( size(u,2), i4b ),   &
                      int( size(vt,1), i4b ),  &
                      'svd_sort2' )
    nru = size(  u, 1 )
    ncvt = size( vt, 2 )
!
!   SORT THE SINGULAR VALUES.
!   INSERTION SORT ON SINGULAR VALUES, BUT ONLY ONE
!   TRANSPOSITION PER SINGULAR VECTOR.
!
    select case( sort )
!
        case( a_maj, a_min ) 
!
!           SORT IN ASCENDING ORDER.
!
            do l = 1, n - 1_i4b
!
!               SCAN FOR SMALLEST d(l) .
!
                i = iminloc( d(l:n) ) + l - 1_i4b
                if ( i==l ) cycle
!
!               SWAP SINGULAR VALUES AND VECTORS.
!
                tmp  = d(i)
                d(i) = d(l)
                d(l) = tmp
!
                tmp_vec(:ncvt) = vt(l,:ncvt)
                vt(l,:ncvt)    = vt(i,:ncvt)
                vt(i,:ncvt)    = tmp_vec(:ncvt)
!
#ifdef _BLAS
                call swap( nru, u(:nru,l), 1_i4b, u(:nru,i), 1_i4b )
#else
                tmp_vec(:nru) = u(:nru,l)
                u(:nru,l)     = u(:nru,i)
                u(:nru,i)     = tmp_vec(:nru)
#endif
!
            end do
!
        case( d_maj, d_min ) 
!
!           SORT IN DESCENDING ORDER.
!
            do l = n, 2_i4b, -1_i4b
!
!               SCAN FOR SMALLEST d(l) .
!
                i = iminloc( d(:l) )
                if ( i==l ) cycle
!
!               SWAP SINGULAR VALUES AND VECTORS.
!
                tmp  = d(i)
                d(i) = d(l)
                d(l) = tmp
!
                tmp_vec(:ncvt) = vt(l,:ncvt)
                vt(l,:ncvt)    = vt(i,:ncvt)
                vt(i,:ncvt)    = tmp_vec(:ncvt)
!
#ifdef _BLAS
                call swap( nru, u(:nru,l), 1_i4b, u(:nru,i), 1_i4b )
#else
                tmp_vec(:nru) = u(:nru,l)
                u(:nru,l)     = u(:nru,i)
                u(:nru,i)     = tmp_vec(:nru)
#endif
!
            end do
!
    end select
!
!
! END OF SUBROUTINE lr_singvec_sort2
! __________________________________
!
    end subroutine lr_singvec_sort2
!
! =========================================================================================
!
    subroutine singvec_sort( sort, d, u )
!
! Purpose
! _______
!
!   Given the singular values D and singular vectors U, stored columwise,
!   as output from BD_SVD, SVD_CMP, BD_SVD2, SVD_CMP2 or SVD_CMP3,
!   this subroutine sorts the singular values into ascending or descending
!   order, and, rearranges the columns of U correspondingly.
!
!
! Arguments
! _________
!
!   SORT         (INPUT) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!                The singular vectors are reordered accordingly.
!
!   D            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the singular values.
!
!                On exit, the singular values in ascending or decreasing order.
!
!   U            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the columns of U are the singular vectors. 
!
!                On exit, U contains the reordered singular vectors.
!
!                The shape of U must verify:  size(U,2) = size( D ) = n .
!
!
! Further Details
! _______________
!
!   The method is straight insertion.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq, iminloc
#ifdef _BLAS
    use BLAS_interfaces, only : swap
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)   :: d
    real(stnd), intent(inout), dimension(:,:) :: u
!
    character , intent(in)  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _BLAS
    real(stnd) :: tmp
#else
    real(stnd) :: tmp_vec(size(u,1)), tmp
#endif
!
    integer(i4b) :: i, l, n, nru
!   
!   
! PARAMETERS 
! __________
!
    character , parameter ::  a_maj=achar(65),  a_min=achar(97),       &
                              d_maj=achar(68),  d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   CHECK THE ARGUMENTS.
!
    n = assert_eq(    int( size(d), i4b ),    &
                      int( size(u,2), i4b ),  &
                      'singvec_sort' )
    nru = size( u, 1 ) 
!
!   SORT THE SINGULAR VALUES.
!   INSERTION SORT ON SINGULAR VALUE, BUT ONLY ONE
!   TRANSPOSITION PER SINGULAR VECTOR.
!
    select case( sort )
!
        case( a_maj, a_min ) 
!
!           SORT IN ASCENDING ORDER.
!
            do l = 1, n - 1_i4b
!
!               SCAN FOR SMALLEST d(l) .
!
                i = iminloc( d(l:) ) + l - 1_i4b
                if ( i==l ) cycle
!
!               SWAP SINGULAR VALUES AND VECTORS.
!
                tmp  = d(i)
                d(i) = d(l)
                d(l) = tmp
!
#ifdef _BLAS
                call swap( nru, u(:nru,l), 1_i4b, u(:nru,i), 1_i4b )
#else
                tmp_vec(:nru) = u(:nru,i)
                u(:nru,i)     = u(:nru,l)
                u(:nru,l)     = tmp_vec(:nru)
#endif
!
            end do
!
        case( d_maj, d_min ) 
!
!           SORT IN DESCENDING ORDER.
!
            do l = n, 2_i4b, -1_i4b
!
!               SCAN FOR SMALLEST d(l) .
!
                i = iminloc( d(:l) )
                if ( i==l ) cycle
!
!               SWAP SINGULAR VALUES AND VECTORS.
!
                tmp  = d(i)
                d(i) = d(l)
                d(l) = tmp
!
#ifdef _BLAS
                call swap( nru, u(:nru,l), 1_i4b, u(:nru,i), 1_i4b )
#else
                tmp_vec(:nru) = u(:nru,i)
                u(:nru,i)     = u(:nru,l)
                u(:nru,l)     = tmp_vec(:nru)
#endif
!
            end do
!
    end select
!
!
! END OF SUBROUTINE singvec_sort
! ______________________________
!
    end subroutine singvec_sort
!
! =========================================================================================
!
    subroutine singval_sort( sort, d )
!
! Purpose
! _______
!
!   Given the singular values D as output from BD_SVD, BD_SVD2, SVD_CMP,
!   SVD_CMP2 or SVD_CMP3, this routine sorts the singular values into ascending or
!   descending order.
!
!
! Arguments
! _________
!
!   SORT         (INPUT) character
!                Sort the singular values into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   D            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the singular values.
!
!                On exit, the singular values in ascending or decreasing order.
!
!
! Further Details
! _______________
!
!   The method is quick sort.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Sort_Procedures, only : quick_sort
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: d(:)
!
    character , intent(in)  :: sort
!   
!   
! PARAMETERS 
! __________
!
    character , parameter ::  a_maj=achar(65),  a_min=achar(97),       &
                              d_maj=achar(68),  d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   SORT SINGULAR VALUES .
!
    select case( sort )
!
        case( a_maj, a_min ) 
!
!           SORT SINGULAR VALUES IN ASCENDING ORDER.
!
            call quick_sort( d, ascending=true )
!
        case( d_maj, d_min ) 
!
!           SORT SINGULAR VALUES IN DESCENDING ORDER.
!
            call quick_sort( d, ascending=false )
!
    end select
!
!
! END OF SUBROUTINE singval_sort
! ______________________________
!
    end subroutine singval_sort
!
!
! =========================================================================================
!                           SVD OF A PRODUCT OF TWO MATRICES
! =========================================================================================
!
!
    subroutine product_svd_cmp( a, b, s, failure, sort, maxiter,               &
                                max_francis_steps, perfect_shift, bisect, dqds )
!
! Purpose
! _______
!
!   This subroutine computes the singular value decomposition of the product of a
!   m-by-n matrix A by the transpose of a p-by-n matrix B:
!
!                           A * B' = U * SIGMA * V'
!
!   where A and B have more rows than columns ( n<=min(m,p) ), SIGMA is an n-by-n matrix
!   which is zero except for its diagonal elements, U is an m-by-n orthogonal
!   matrix, and V is an p-by-n orthogonal matrix. The diagonal elements of SIGMA
!   are the singular values of A * B'; they are real and non-negative.
!   The columns of U and V are the left and right singular vectors
!   of A * B', respectively.
!
!
! Arguments
! _________
!
!   A                   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix A.
!
!                       On exit,  the m-by-n left-singular matrix U.
!
!   B                   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the p-by-n matrix B.
!
!                       On exit,  the p-by-n right-singular matrix V.
!
!   S                   (OUTPUT) real(stnd), dimension(:)
!                       The singular values of A * B'.
!
!                       The size of S must verify: size( S ) = n .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit
!                       - FAILURE = true  :  indicates that the algorithm did not converge and
!                         that full accuracy was not attained in the SVD.
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the singular values into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd'.
!                       The singular vectors are rearranged accordingly.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm.
!
!                       The bidiagonal SVD algorithm of an intermediate bidiagonal 
!                       form of A * B' fails to converge if the number of QR sweeps 
!                       exceeds MAXITER * n. Convergence usually occurs in 
!                       about 2 * n QR sweeps.
!
!                       The default is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm.
!
!                       MAX_FRANCIS_STEPS is a strictly positive integer, otherwise 
!                       the default value is used.
!
!                       The default is the minimum of n and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   The size of S must match:  size( S ) = size( A, 2 ) = size( B, 2 ) .
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq, merror
    use Char_Constants,  only : svd_error1
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:), intent(inout) :: a, b
    real(stnd), dimension(:),   intent(out)   :: s
!
    integer(i4b), intent(in), optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(in), optional :: perfect_shift, bisect, dqds
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(s),   size(s)) :: tmp
    real(stnd), dimension(size(a,1), size(s)) :: a2
    real(stnd), dimension(size(b,1), size(s)) :: b2
!
    integer(i4b) :: n, m, p, i
!
    logical(lgl) :: fail1, fail2
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='product_svd_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   CHECK THE ARGUMENTS.
!
    n = assert_eq( int( size(s), i4b ),        &
                   int( size(b,2), i4b ),      &
                   int( size(a,2), i4b ),      &
                   name_proc)
!
    m = size( a, 1 )
    p = size( b, 1 )
!
    if ( m<n .or. p<n ) then
        call merror( name_proc//svd_error1 )
    end if
!
!   MAKE LOCAL COPIES OF THE INPUT MATRICES.
!
    a2(:m,:n) = a(:m,:n)
    b2(:p,:n) = b(:p,:n)
!
!   FIRST COMPUTE SVD OF a.
!
    call svd_cmp_a( a2(:m,:n), s(:n), fail1, tmp(:n,:n),                  &
                    maxiter=maxiter, max_francis_steps=max_francis_steps, &
                    perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!   COMPUTE g = b * ( tmp*diag(s) ) .
!
    do i = 1_i4b, n
!
        tmp(:n,i) = tmp(:n,i)*s(i)
!
    end do
!
#ifdef _BLAS
    call gemm( 'N', 'N', p, n, n, one, b2, p, tmp, n, zero, b, p )
#else
    b = matmul( b2, tmp )
#endif
!
!   COMPUTE SVD OF g, THE SINGULAR VALUES AND THE LEFT-SINGULAR MATRIX OF g
!   ARE THE SINGULAR VALUES AND THE RIGHT-SINGULAR MATRIX OF THE PRODUCT.
!
    call svd_cmp_a( b(:p,:n), s(:n), fail2, tmp(:n,:n), sort=sort,        &
                    maxiter=maxiter, max_francis_steps=max_francis_steps, &
                    perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!   LAST COMPUTE THE LEFT-SINGULAR MATRIX OF THE PRODUCT.
!
#ifdef _BLAS
    call gemm( 'N', 'N', m, n, n, one, a2, m, tmp, n, zero, a, m )
#else
    a = matmul( a2, tmp )
#endif
!
    failure = fail1 .or. fail2
!
!
! END OF SUBROUTINE product_svd_cmp
! _________________________________
!
    end subroutine product_svd_cmp
!
!
! =========================================================================================
!                           GENERALIZED INVERSE OF A MATRIX
! =========================================================================================
!
!
    function ginv( mat, tol, maxiter, max_francis_steps, perfect_shift, bisect, dqds ) result( matginv )
!
! Purpose
! _______
!
!   GINV returns the generalized inverse of a m-by-n real matrix, MAT.
!   The generalized inverse of MAT is a n-by-m matrix.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!   TOL                 (INPUT, OPTIONAL) real(stnd)
!                       On entry:
!
!                       - If TOL is less than or equal to zero or is absent,
!                         the function computes the generalized inverse of MAT.
!                       - If TOL is greater than zero, the subroutine computes the
!                         generalized inverse of a matrix close to MAT, but having 
!                         condition number in the 2-norm less than 1/TOL.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm.
!
!                       The bidiagonal SVD phase of an intermediate bidiagonal 
!                       form B of MAT fails to converge if the number of QR sweeps 
!                       exceeds MAXITER * min(m,n). Convergence usually occurs in 
!                       about 2 * min(m,n) QR sweeps.
!
!                       The default is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm.
!
!                       MAX_FRANCIS_STEPS is a strictly positive integer, otherwise 
!                       the default value is used.
!
!                       The default is the minimum of min(m,n) and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed
!                       generalized inverse at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   If MAT is the null matrix or the SVD algorithm used to compute the generalized
!   inverse of MAT did not converge and full accuracy was not attained in the bidiagonal 
!   SVD of an intermediate bidiagonal form of MAT, function GINV
!   returns a n-by-m matrix filled with NAN() function.
!
!   The computation of the generalized inverse is parallelized 
!   if OPENMP is used.
!
!   For further details, on the generalized inverse of a rectangular matrix
!   and the algorithm to compute it, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore, Maryland.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:,:)  :: mat
    real(stnd), intent(in),        optional :: tol
!
    integer(i4b), intent(in),  optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(in), optional :: perfect_shift, bisect, dqds
!
    real(stnd), dimension(size(mat,2),size(mat,1)) :: matginv
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension( size(mat,1), size(mat,2) )  :: mat2
    real(stnd), dimension(size(mat,2),size(mat,1))     :: matginv2
!
    integer(i4b) :: m, n, k

    logical(lgl) ::  failure
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = min( m, n )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b )  return
!
!   MAKE A COPY OF mat .
!
    mat2(1_i4b:m,1_i4b:n) = mat(1_i4b:m,1_i4b:n)
!
!   CALL comp_ginv SUBROUTINE WITH APPROPRIATE ARGUMENTS.
!
    call comp_ginv_a( mat2, failure, matginv2, tol=tol, maxiter=maxiter,    &
                      max_francis_steps=max_francis_steps,                  &
                      perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!   OUTPUT THE RESULT.
!
    if ( failure ) then
        matginv(:n,:m) = nan()
    else 
        matginv(:n,:m) = matginv2(:n,:m)
    end if
!
!
! END OF FUNCTION ginv
! ____________________
!
    end function ginv
!
! =========================================================================================
!
    subroutine comp_ginv_a( mat, failure, matginv, tol, singvalues, krank, mul_size, &
                            maxiter, max_francis_steps, perfect_shift, bisect, dqds  )
!
! Purpose
! _______
!
!   COMP_GINV computes the generalized inverse of a m-by-n real matrix, MAT.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit, MAT is destroyed.
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit.
!                       - FAILURE = true  :  indicates that MAT is the null matrix or that the SVD
!                         algorithm which is used to compute the generalized inverse of MAT
!                         did not converge and that full accuracy was not attained in the
!                         bidiagonal SVD of an intermediate bidiagonal form B of MAT.
!
!   MATGINV             (OUTPUT) real(stnd), dimension(:,:)
!                       On exit, MATGINV contains the generalized inverse of MAT
!                       or the generalized inverse of a matrix close to MAT.
!
!                       The shape of MATGINV must verify:
!
!                       - size(MATGINV,1) = size(MAT,2) = n ,
!                       - size(MATGINV,2) = size(MAT,1) = m .
!
!   TOL                 (INPUT, OPTIONAL) real(stnd)
!                       On entry, if:
!
!                       - TOL is less than or equal to zero or is absent,
!                         the subroutine computes the generalized inverse of MAT.
!                       - TOL is greater than zero, the subroutine computes the
!                         generalized inverse of a matrix close to MAT, but having 
!                         condition number in the 2-norm less than 1/TOL.
!
!   SINGVALUES          (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                       The singular values of MAT in decreasing order.
!                       The condition number of MAT in the 2-norm is
!
!                             SINGVALUES(1)/SINGVALUES(min(m,n)).
!
!                       The size of SINGVALUES must verify : size( SINGVALUES ) = min(m,n) .
!
!   KRANK               (OUTPUT, OPTIONAL) integer(i4b)
!                       On exit, the effective rank of MAT, i.e., the number
!                       of singular values which are greater than TOL * SINGVALUES(1).
!
!   MUL_SIZE            (INPUT, OPTIONAL) integer(i4b)
!                       Internal parameter. MUL_SIZE must verify: 1 <= MUL_SIZE <= max(m,n),
!                       otherwise a default value is used. MUL_SIZE can be increased or decreased
!                       to improve the performance of the algorithm.
!
!                       The default value is 32.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm.
!
!                       The bidiagonal SVD algorithm of an intermediate bidiagonal 
!                       form B of MAT fails to converge if the number of QR sweeps 
!                       exceeds MAXITER * min(m,n). Convergence usually occurs in 
!                       about 2 * min(m,n) QR sweeps.
!
!                       The default is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm.
!
!                       MAX_FRANCIS_STEPS is a strictly positive integer, otherwise 
!                       the default value is used.
!
!                       The default is the minimum of min(m,n) and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed
!                       generalized inverse at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   If all the elements of MAT are equal to zero, subroutine COMP_GINV
!   returns a n-by-m matrix filled with NAN() function in argument MATGINV
!   and the logical argument FAILURE is set to .true. .
!
!   The computation of the generalized inverse is parallelized 
!   if OPENMP is used.
!
!   For further details, on the generalized inverse of a rectangular matrix
!   and the algorithm to compute it, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore, Maryland.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert_eq, assert, lascl, ifirstloc, merror
    use QR_Procedures,   only : qr_cmp, lq_cmp, apply_q_lq, apply_q_qr
#ifdef _BLAS
    use BLAS_interfaces, only : scal, gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:,:) :: matginv
    real(stnd), intent(in),                 optional  :: tol
    real(stnd), intent(out), dimension(:) , optional  :: singvalues
!
    integer(i4b), intent(in),  optional :: mul_size, maxiter, max_francis_steps
    integer(i4b), intent(out), optional :: krank
!
    logical(lgl), intent(in), optional  :: perfect_shift, bisect, dqds
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: matnorm, thr
    real(stnd), dimension( min(size(mat,1),size(mat,2)) ) :: s, e, tau, tauq, taup
    real(stnd), dimension(:,:), allocatable :: work, work2
!
    integer(i4b) :: m, n, mn, i, j, k, l, mnthr, inc, i1, i2
    integer      ::  iok
!
    logical(lgl) ::  mat_is_scaled
!
#ifdef _OPENMP
!
    logical :: test_par, test_par2, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_ginv'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size(mat,1)    ,i4b),   &
                   int(size(matginv,2),i4b),   &
                   name_proc  )
!
    n = assert_eq( int(size(mat,2)    ,i4b),   &
                   int(size(matginv,1),i4b),   &
                   name_proc  )
!
    mn = min( m, n )
!
    if ( present(singvalues) ) then
!
        call assert( logical(size(singvalues)==mn,lgl), name_proc )
!
    end if
!
    if ( present(mul_size) ) then
!
        inc = merge( mul_size, 32_i4b, 1_i4b<=mul_size .and. mul_size<=max(m,n) )
!
    else
!
        inc = 32_i4b
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( mn<=0_i4b ) then
!
        if ( present(krank) ) krank = 0_i4b
!
        return
!
    end if
!
#ifdef _OPENMP
!
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested  ) .and. i>1_i4b
    test_par2   = test_par .and. n*m>=omp_limit .and. inc+inc<l
!
#endif
!
!   COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
!$OMP PARALLEL DO IF(test_par2), SCHEDULE(STATIC)              &
!$OMP            ,PRIVATE(i,i1,i2,thr), REDUCTION(max:matnorm)
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, n )
        thr = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, thr )
!
    end do
!
!$OMP END PARALLEL DO
!
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       MATRIX mat ALL ZERO .
!
        failure = true
        matginv = nan()
!
        if ( present(krank) )      krank      = 0_i4b
        if ( present(singvalues) ) singvalues = zero
!
        return
!
    else if ( matnorm<smlnum ) then
!
!       SCALE MATRIX NORM UP TO smlnum.
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
!       SCALE MATRIX NORM DOWN TO bignum.
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat  TO BIDIAGONAL FORM. IF max(m,n) EXCEEDS mnthr ,
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE
!   MATRIX mat TO A TRIANGULAR FORM.
!    
    mnthr = mn*2_i4b
!
    if ( m>=n ) then
!
!       mat HAS AT LEAST AS MANY ROWS AS COLUMNS.
!
        if (  m>=mnthr  ) then
!
!           PATH 1 : m MUCH LARGER THAN n .
!
!           FIRST REDUCE mat USING THE QR DECOMPOSITION.
!           SECOND REDUCE R TO UPPER BIDIAGONAL FORM.
!
!           COMPUTE mat = Q * R .
!
            call qr_cmp( mat, s, tau )
!
!           COPY R' TO matginv(:n,:n) , ZEROING OUT ABOVE IT.
!
#ifdef _OPENMP
            test_par2 = test_par .and. n*n>=omp_limit
#endif
!
!$OMP PARALLEL DO IF(test_par2)      &
!$OMP            ,SCHEDULE(STATIC,1) &
!$OMP            ,PRIVATE(i)
!
            do i = 1_i4b, n
!
                matginv(i,1_i4b:i-1_i4b) = mat(1_i4b:i-1_i4b,i)
                matginv(i,i) = s(i)
                matginv(1_i4b:i-1_i4b,i) = zero
!
            end do
!
!$OMP END PARALLEL DO
!
!           BIDIAGONALIZE R' IN matginv(:n,:n) .
!
            call bd_cmp_a( matginv(:n,:n), s, e, tauq, taup )
!
!           GENERATE LEFT VECTORS BIDIAGONALIZING R IN matginv(:n,n+1_i4b:mnthr) .
!           GENERATE RIGHT VECTORS BIDIAGONALIZING R IN matginv(:n,:n) .
!
            call ortho_gen_bd( matginv(:n,:n), tauq, taup, matginv(:n,n+1_i4b:mnthr) )
!
!           PERFORM BIDIAGONAL QR ITERATION, COMPUTING RIGHT SINGULAR 
!           VECTORS OF R IN matginv(:n,:n) AND LEFT SINGULAR VECTORS OF
!           R IN matginv(:n,n+1_i4b:mnthr) .
!
            call bd_svd_a( true, s, e, failure, matginv(:n,:n), matginv(:n,n+1_i4b:mnthr), &
                           sort='d', maxiter=maxiter, max_francis_steps=max_francis_steps, &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds           )
!
!           DETERMINE THE EFFECTIVE RANK OF R .
!
            if ( present(tol) ) then
                thr = max( tol*s(1_i4b), safmin )
            else
                thr = safmin
            end if
!
            k = ifirstloc( logical( s<=thr, lgl ) ) - 1_i4b
!
!           COMPUTE RECIPROCALS OF SINGULAR VALUES.
!
            e(:k) = one/s(:k)
!
!           MULTIPLY LEFT SINGULAR VECTORS OF R STORED IN matginv(:n,n+1_i4b:mnthr) 
!           BY RECIPROCALS OF SINGULAR VALUES .
!
#ifdef _OPENMP
            test_par2 = test_par .and. n*k>=omp_limit
#endif
!
!$OMP PARALLEL DO IF(test_par2), SCHEDULE(STATIC,1), PRIVATE(i)
!
            do i = 1_i4b, k
!
#ifdef _BLAS
                call scal( n, e(i), matginv(:n,n+i), 1_i4b )
#else
                matginv(:n,n+i) = e(i)*matginv(:n,n+i)
#endif
!
            end do
!
!$OMP END PARALLEL DO
!
#ifdef _BLAS
!           ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
            allocate( work(n,k), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           MULTIPLY THE TRANSPOSE OF SCALED LEFT SINGULAR OF R BY RIGHT SINGULAR
!           VECTORS OF R STORED COLUMN-WISE IN matginv(:n,:n), STORING THE FINAL
!           RESULT IN matginv(:n,:n) .
!
            work(:n,:k) = matginv(:n,:k)
!
            call gemm( 'N', 'T', n, n, k, one, work, n, matginv(:n,n+1_i4b:n+k), n, zero, matginv, n )
#else
!           ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
            allocate( work(k,n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           MULTIPLY THE TRANSPOSE OF SCALED LEFT SINGULAR OF R BY RIGHT SINGULAR
!           VECTORS OF R STORED COLUMN-WISE IN matginv(:n,:n), STORING THE FINAL
!           RESULT IN matginv(:n,:n) .
!
            work(:k,:n) = transpose( matginv(:n,n+1_i4b:n+k) )
!
            matginv(:n,n+1_i4b:mnthr) = matmul( matginv(:n,:k), work(:k,:n) )
            matginv(:n,:n) = matginv(:n,n+1_i4b:mnthr)
#endif
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( work )
!
!           MULTIPLY Q' BY THE RESULTING MATRIX, STORING THE RESULT IN matginv.
!
            matginv(:n,n+1_i4b:m) = zero
!
            call apply_q_qr( mat, tau, matginv, left=false, trans=true )
!
        else
!
!           PATH 2 : m AT LEAST n, BUT NOT MUCH LARGER;
!
!           REDUCE mat TO UPPER BIDIAGONAL FORM WITHOUT 
!           QR DECOMPOSITION.
!
!           BIDIAGONALIZE mat .
!
            call bd_cmp_a( mat, s, e, tauq, taup )
!
!           GENERATE RIGHT VECTORS BIDIAGONALIZING mat IN matginv(:n,:n) .
!           GENERATE LEFT VECTORS BIDIAGONALIZING mat IN mat .
!
            call ortho_gen_bd( mat, tauq, taup, matginv(:n,:n) )
!
!           PERFORM BIDIAGONAL QR ITERATION, COMPUTE LEFT SINGULAR
!           VECTORS OF mat IN mat AND RIGHT SINGULAR VECTORS OF mat
!           IN matginv(:n,:n) .
!
            call bd_svd_a( true, s, e, failure, mat, matginv(:n,:n), sort='d',   &
                           maxiter=maxiter, max_francis_steps=max_francis_steps, &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!           DETERMINE THE EFFECTIVE RANK OF mat .
!
            if ( present(tol) ) then
                thr = max( tol*s(1_i4b), safmin )
            else
                thr = safmin
            end if
!
            k = ifirstloc( logical( s<=thr, lgl ) ) - 1_i4b
!
!           COMPUTE RECIPROCALS OF SINGULAR VALUES.
!
            e(:k) = one/s(:k)
!
!           MULTIPLY LEFT SINGULAR VECTORS OF mat STORED COLUMN-WISE IN mat BY 
!           RECIPROCALS OF SINGULAR VALUES .
!
#ifdef _OPENMP
            test_par2 = test_par .and. m*k>=omp_limit
#endif
!
!$OMP PARALLEL DO IF(test_par2), SCHEDULE(STATIC,1), PRIVATE(i)
!
            do i = 1_i4b, k
!
#ifdef _BLAS
                call scal( m, e(i), mat(:m,i), 1_i4b )
#else
                mat(:m,i) = e(i)*mat(:m,i)
#endif
!
            end do
!
!$OMP END PARALLEL DO
!
#ifdef _BLAS
!           ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
            allocate( work(n,k),  stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           MULTIPLY TRANSPOSE OF SCALED LEFT SINGULAR VECTORS OF mat STORED COLUMN-WISE
!           IN mat BY RIGHT SINGULAR VECTORS OF mat STORED IN matginv(:n,:n), STORING THE
!           FINAL RESULT IN matginv .
!
            work(:n,:k) = matginv(:n,:k)
!
            call gemm( 'N', 'T', n, m, k, one, work, n, mat, m, zero, matginv, n )
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( work )
#else
!           ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
            allocate( work(k,m), work2(n,k), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           MULTIPLY TRANSPOSE OF SCALED LEFT SINGULAR VECTORS OF mat STORED COLUMN-WISE
!           IN mat BY RIGHT SINGULAR VECTORS OF mat STORED IN matginv(:n,:n), STORING THE
!           FINAL RESULT IN matginv .
!
            work(:k,:m)  = transpose( mat(:m,:k) )
            work2(:n,:k) = matginv(:n,:k)
!
            matginv(:n,:m) = matmul( work2(:n,:k), work(:k,:m) )
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( work, work2 )
#endif
!
        end if
!
    else
!
!       mat HAS MORE COLUMNS THAN ROWS.
!
        if (  n>=mnthr ) then
!
!           PATH 3 : n MUCH LARGER THAN m.
!
!           FIRST REDUCE mat USING THE LQ DECOMPOSITION.
!           SECOND REDUCE L TO UPPER BIDIAGONAL FORM.
!
!           COMPUTE mat = L * Q .
!
            call lq_cmp( mat, s, tau )
!
!           COPY L' TO matginv(:m,:m) , ZEROING OUT BELOW IT.
!
#ifdef _OPENMP
            test_par2 = test_par .and. m*m>=omp_limit
#endif
!
!$OMP PARALLEL DO IF(test_par2)       &
!$OMP            ,SCHEDULE(STATIC,1)  &
!$OMP            ,PRIVATE(i)
!
            do i = 1_i4b, m
!
                matginv(i,i+1_i4b:m) = mat(i+1_i4b:m,i)
                matginv(i,i) = s(i)
                matginv(i+1_i4b:m,i) = zero
!
            end do
!
!$OMP END PARALLEL DO
!
!           BIDIAGONALIZE L' IN matginv(:m,:m) .
!
            call bd_cmp_a( matginv(:m,:m), s, e, tauq, taup )
!
!           GENERATE LEFT VECTORS BIDIAGONALIZING L IN matginv(m+1_i4b:mnthr,:m) .
!           GENERATE RIGHT VECTORS BIDIAGONALIZING L IN matginv(:m,:m) .
!
            call ortho_gen_bd( matginv(:m,:m), tauq, taup, matginv(m+1_i4b:mnthr,:m) )
!
!           PERFORM BIDIAGONAL QR ITERATION, COMPUTING RIGHT SINGULAR 
!           VECTORS OF L IN matginv(:m,:m) AND LEFT SINGULAR VECTORS OF
!           L IN matginv(m+1_i4b:mnthr,:m) .
!
            call bd_svd_a( true, s, e, failure, matginv(:m,:m), matginv(m+1_i4b:mnthr,:m), &
                           sort='d', maxiter=maxiter, max_francis_steps=max_francis_steps, &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds           )
!
!           DETERMINE THE EFFECTIVE RANK OF mat .
!
            if ( present(tol) ) then
                thr = max( tol*s(1_i4b), safmin )
            else
                thr = safmin
            end if
!
            k = ifirstloc( logical( s<=thr, lgl ) ) - 1_i4b
!
!           COMPUTE RECIPROCALS OF SINGULAR VALUES.
!
            e(:k) = one/s(:k)
!
!           MULTIPLY LEFT SINGULAR VECTORS OF L STORED IN COLUMN-WISE IN
!           matginv(m+1_i4b:mnthr,:m) BY RECIPROCALS OF SINGULAR VALUES .
!
#ifdef _OPENMP
            test_par2 = test_par .and. m*k>=omp_limit
#endif
!
!$OMP PARALLEL DO IF(test_par2), SCHEDULE(STATIC,1), PRIVATE(i)
!
            do i = 1_i4b, k
!
#ifdef _BLAS
                call scal( mnthr-m, e(i), matginv(m+1_i4b:mnthr,i), 1_i4b )
#else
                matginv(m+1_i4b:mnthr,i) = e(i)*matginv(m+1_i4b:mnthr,i)
#endif
!
            end do
!
!$OMP END PARALLEL DO
!
#ifdef _BLAS
!           ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
            allocate( work(m,k), work2(m,k), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           MULTIPLY THE TRANSPOSE OF SCALED LEFT SINGULAR VECTORS OF L BY
!           RIGHT SINGULAR VECTORS OF L STORED COLUMN-WISE IN  matginv(:m,:m) .
!
            work(:m,:k)  = matginv(:m,:k)
            work2(:m,:k) = matginv(m+1_i4b:mnthr,:k)
!
            call gemm( 'N', 'T', m, m, k, one, work, m, work2, m, zero, matginv, n )
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( work, work2 )
#else
!           ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
            allocate( work(k,m), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           MULTIPLY THE TRANSPOSE OF SCALED LEFT SINGULAR VECTORS OF L BY
!           RIGHT SINGULAR VECTORS OF L STORED COLUMN-WISE IN  matginv(:m,:m) .
!
            work(:k,:m) = transpose( matginv(m+1_i4b:mnthr,:k) )
!
            matginv(m+1_i4b:mnthr,:m) = matmul( matginv(:m,:k), work(:k,:m) )
!
            matginv(:m,:m) = matginv(m+1_i4b:mnthr,:m)
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( work )
#endif
!
!           MULTIPLY Q' BY THE RESULTING MATRIX, STORING THE RESULT IN matginv.
!
            matginv(m+1_i4b:n,:m) = zero
!
            call apply_q_lq( mat, tau, matginv, left=true, trans=true )
!
        else
!
!           PATH 4 :  n NOT MUCH LARGER THAN m.
!
!           REDUCE mat TO LOWER BIDIAGONAL FORM WITHOUT 
!           LQ DECOMPOSITION.
!
!           BIDIAGONALIZE mat .
!
            call bd_cmp_a( mat, s, e, tauq, taup )
!
!           GENERATE RIGHT VECTORS BIDIAGONALIZING mat IN matginv .
!           GENERATE LEFT VECTORS BIDIAGONALIZING mat IN mat(:m,:m) .
!
            call ortho_gen_bd( mat, tauq, taup, matginv )
!
!           PERFORM BIDIAGONAL QR ITERATION, COMPUTE LEFT SINGULAR
!           VECTORS OF mat IN mat(:m,:m) AND RIGHT SINGULAR VECTORS OF mat
!           IN matginv .
!
            call bd_svd_a( false, s, e, failure, mat(:m,:m), matginv, sort='d',  &
                           maxiter=maxiter, max_francis_steps=max_francis_steps, &
                           perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!           DETERMINE THE EFFECTIVE RANK OF mat .
!
            if ( present(tol) ) then
                thr = max( tol*s(1_i4b), safmin )
            else
                thr = safmin
            end if
!
            k = ifirstloc( logical( s<=thr, lgl ) ) - 1_i4b
!
!           COMPUTE RECIPROCALS OF SINGULAR VALUES.
!
            e(:k) = one/s(:k)
!
!           MULTIPLY LEFT SINGULAR VECTORS OF mat BY RECIPROCALS OF  
!           SINGULAR VALUES.
!
#ifdef _OPENMP
            test_par2 = test_par .and. m*k>=omp_limit
#endif
!
!$OMP PARALLEL DO IF(test_par2), SCHEDULE(STATIC,1), PRIVATE(i)
!
            do i = 1_i4b, k
!
#ifdef _BLAS
                call scal( m, e(i), mat(:m,i), 1_i4b )
#else
                mat(:m,i) = e(i)*mat(:m,i)
#endif
!
            end do
!
!$OMP END PARALLEL DO
!
#ifdef _BLAS
!           ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
            allocate( work(n,k), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           MULTIPLY TRANSPOSE OF SCALED LEFT SINGULAR VECTORS OF mat STORED
!           COLUMN-WISE IN mat(:m,:m) BY RIGHT SINGULAR VECTORS OF mat STORED IN matginv,
!           STORING FINAL RESULT IN matginv .
!
            work(:n,:k) = matginv(:n,:k)
!
            call gemm( 'N', 'T', n, m, k, one, work, n, mat, m, zero, matginv, n )
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( work )
#else
!           ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
            allocate( work(k,m), work2(n,k), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           MULTIPLY TRANSPOSE OF SCALED LEFT SINGULAR VECTORS OF mat STORED
!           COLUMN-WISE IN mat(:m,:m) BY RIGHT SINGULAR VECTORS OF mat STORED IN matginv,
!           STORING FINAL RESULT IN matginv .
!
            work(:k,:m) = transpose( mat(:m,:k) )
            work2(:n,:k) = matginv(:n,:k)
!
            matginv(:n,:m) = matmul( work2(:n,:k), work(:k,:m) )
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( work, work2 )
#endif
!
        end if
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled ) then
!
        call lascl( matginv, matnorm, merge( smlnum, bignum, matnorm<smlnum ) )
!
    end if
!    
!   OUTPUT RANK OF mat, IF REQUIRED.
!    
    if ( present(krank) ) then
!    
        krank = k
!    
    end if
!    
!   OUTPUT SINGULAR VALUES OF mat, IF REQUIRED.
!    
    if ( present(singvalues) )  then
!
!       UNDO SCALING IF NECESSARY.
!
        if ( mat_is_scaled ) then
!
            call lascl( s, merge( smlnum, bignum, matnorm<smlnum ), matnorm )
!
        end if

        singvalues = s
!
    end if
!
!
! END OF SUBROUTINE comp_ginv_a
! _____________________________
!
    end subroutine comp_ginv_a
!
! =========================================================================================
!
    subroutine comp_ginv_b( mat, failure, tol, singvalues, krank, mul_size,         &
                            maxiter, max_francis_steps, perfect_shift, bisect, dqds )
!
! Purpose
! _______
!
!   COMP_GINV computes the generalized inverse of a m-by-n real matrix, MAT.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit, MAT contains the transpose of the generalized
!                       inverse of MAT or of the generalized inverse of a matrix
!                       close to MAT.
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit.
!                       - FAILURE = true  :  indicates that MAT is the null matrix or that the SVD
!                         algorithm which is used to compute the generalized inverse of MAT
!                         did not converge and that full accuracy was not attained in the
!                         bidiagonal SVD of an intermediate bidiagonal form B of MAT.
!
!   TOL                 (INPUT, OPTIONAL) real(stnd)
!                       On entry, if:
!
!                       - TOL is less than or equal to zero or is absent,
!                         the subroutine computes the generalized inverse of MAT.
!                       - TOL is greater than zero, the subroutine computes the
!                         generalized inverse of a matrix close to MAT, but having 
!                         condition number in the 2-norm less than 1/TOL.
!
!   SINGVALUES          (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                       The singular values of MAT in decreasing order.
!                       The condition number of MAT in the 2-norm is
!
!                             SINGVALUES(1)/SINGVALUES(min(m,n)).
!
!                       The size of SINGVALUES must verify: size( SINGVALUES ) = min(m,n) .
!
!   KRANK               (OUTPUT, OPTIONAL) integer(i4b)
!                       On exit, the effective rank of MAT, i.e., the number
!                       of singular values which are greater than TOL * SINGVALUES(1).
!
!   MUL_SIZE            (INPUT, OPTIONAL) integer(i4b)
!                       Internal parameter. MUL_SIZE must verify: 1 <= MUL_SIZE <= max(m,n),
!                       otherwise a default value is used. MUL_SIZE can be increased or decreased
!                       to improve the performance of the algorithm.
!
!                       The default value is 32.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm.
!
!                       The bidiagonal SVD algorithm of an intermediate bidiagonal 
!                       form B of MAT fails to converge if the number of QR sweeps 
!                       exceeds MAXITER * min(m,n). Convergence usually occurs in 
!                       about 2 * min(m,n) QR sweeps.
!
!                       The default is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm.
!
!                       MAX_FRANCIS_STEPS is a strictly positive integer, otherwise 
!                       the default value is used.
!
!                       The default is the minimum of n and the integer parameter
!                       MAX_FRANCIS_STEPS_SVD specified in the module Select_Parameters.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!   BISECT              (INPUT, OPTIONAL) logical(lgl)
!                       BISECT determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If BISECT is set to true, singular values are computed with a more accurate
!                       bisection algorithm delivering improved accuracy in the final computed
!                       generalized inverse at the expense of a slightly slower execution time.
!
!                       If BISECT is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!   DQDS                (INPUT, OPTIONAL) logical(lgl)
!                       DQDS determines how the singular values are computed if a perfect shift
!                       strategy is used in the bidiagonal SVD algorithm (e.g., if PERFECT_SHIFT is
!                       equal to TRUE). This argument has no effect if PERFECT_SHIFT is equal to false.
!
!                       If DQDS is set to true, singular values are computed with a more accurate
!                       dqds algorithm delivering improved accuracy in the final computed SVD
!                       decomposition at the expense of a slightly slower execution time.
!
!                       If DQDS is set to false, singular values are computed with the fast
!                       Pal-Walker-Kahan algorithm.
!
!                       If both optional arguments BISECT and DQDS are specified with the value true, the
!                       bisection algorithm is used.
!
!                       The default is false.
!
!
! Further Details
! _______________
!
!   If all the elements of MAT are equal to zero, subroutine COMP_GINV
!   returns a m-by-n matrix filled with NAN() function in argument MAT
!   and the logical argument FAILURE is set to .true. .
!
!   The computation of the generalized inverse is parallelized 
!   if OPENMP is used.
!
!   For further details, on the generalized inverse of a rectangular matrix
!   and the algorithm to compute it, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore, Maryland.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, merror, lascl, ifirstloc
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:)           :: mat
    real(stnd), intent(in),                   optional  :: tol
    real(stnd), intent(out),  dimension(:) ,  optional  :: singvalues
!
    integer(i4b), intent(in),  optional :: mul_size, maxiter, max_francis_steps
    integer(i4b), intent(out), optional :: krank
!
    logical(lgl), intent(in), optional  :: perfect_shift, bisect, dqds
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: matnorm, thr
    real(stnd), dimension( min(size(mat,1),size(mat,2)) ) :: s, e, tauq, taup
    real(stnd), dimension(:,:), allocatable  :: work, work2
!
    integer(i4b) :: m, n, mn, i, j, k, l, inc, i1, i2
    integer      :: iok
!
    logical(lgl) ::  mat_is_scaled
!
#ifdef _OPENMP
!
    logical :: test_par, test_par2, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_ginv'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    mn = min( m, n )
!
    if ( present(singvalues) ) then
!
        call assert( logical(size(singvalues)==mn,lgl), name_proc )
!
    end if
!
    if ( present(mul_size) ) then
!
        inc = merge( mul_size, 32_i4b, 1_i4b<=mul_size .and. mul_size<=max(m,n) )
!
    else
!
        inc = 32_i4b
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( mn<=0_i4b ) then
!
        if ( present(krank) ) krank = 0_i4b
!
        return
!
    end if
!
#ifdef _OPENMP
!
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_max_active_levels()>1 .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.       &
                  m*n>=omp_limit                      .and.       &
                  i>1_i4b
    test_par2   = test_par .and.  inc+inc<l
!
#endif
!
!   COMPUTE MAX ELEMENT OF mat .
!
    j = 1_i4b - inc
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
!
    matnorm = zero
!
!$OMP PARALLEL DO IF(test_par2), SCHEDULE(STATIC)              &
!$OMP            ,PRIVATE(i,i1,i2,thr), REDUCTION(max:matnorm)
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, n )
        thr = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, thr )
!
    end do
!
!$OMP END PARALLEL DO
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    mat_is_scaled = false
!
    if ( matnorm==zero ) then
!
!       MATRIX mat ALL ZERO .
!
        failure = true
        mat     = nan()
!
        if ( present(krank) )      krank      = 0_i4b
        if ( present(singvalues) ) singvalues = zero
!
        return
!
    else if ( matnorm<smlnum ) then
!
!       SCALE MATRIX NORM UP TO smlnum.
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
!       SCALE MATRIX NORM DOWN TO bignum.
!
        mat_is_scaled = true
!
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   REDUCE mat TO UPPER BIDIAGONAL FORM IF mat
!   HAS AT LEAST AS MANY ROWS AS COLUMNS, E.G. IF m>=n .
!
!   REDUCE mat TO LOWER BIDIAGONAL FORM IF mat
!   HAS MORE COLUMNS THAN ROWS, E.G. IF m<n .
!
!   BIDIAGONALIZE mat .
!
    call bd_cmp_a( mat, s, e, tauq, taup )
!
    if ( m>=n ) then
!
!       PATH 1 : mat HAS AT LEAST AS MANY ROWS AS COLUMNS, E.G. m AT LEAST n.
!
!       ALLOCATE WORK ARRAY.
!
        allocate( work(n,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       GENERATE LEFT VECTORS BIDIAGONALIZING mat IN mat .
!       GENERATE TRANSPOSE OF RIGHT VECTORS BIDIAGONALIZING mat IN work(:n,:n) .
!
        call ortho_gen_bd2( mat, tauq, taup, work )
!
!       PERFORM BIDIAGONAL QR ITERATION, COMPUTE LEFT SINGULAR
!       VECTORS OF mat IN mat AND TRANSPOSE OF RIGHT SINGULAR
!       VECTORS OF mat IN work(:n,:n) .
!
        call bd_svd2_a( true, s, e, failure, mat, work(:n,:n), sort='d',      &
                        maxiter=maxiter, max_francis_steps=max_francis_steps, &
                        perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!       DETERMINE THE EFFECTIVE RANK OF mat .
!
        if ( present(tol) ) then
            thr = max( tol*s(1_i4b), safmin )
        else
            thr = safmin
        end if
!
        k = ifirstloc( logical( s<=thr, lgl ) ) - 1_i4b
!
!       COMPUTE RECIPROCALS OF SINGULAR VALUES.
!
        e(:k) = one/s(:k)
!
!       ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
        allocate( work2(m,k), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MULTIPLY LEFT SINGULAR VECTORS OF mat STORED COLUMN-WISE IN mat BY 
!       RECIPROCALS OF SINGULAR VALUES, STORING RESULT IN work2 .
!
#ifdef _OPENMP
        test_par2 = test_par .and. m*k>=omp_limit
#endif
!
!$OMP PARALLEL DO IF(test_par2), SCHEDULE(STATIC,1), PRIVATE(i)
!
        do i = 1_i4b, k
!
            work2(:m,i) = e(i)*mat(:m,i)
!
        end do
!
!$OMP END PARALLEL DO
!
!       MULTIPLY SCALED LEFT SINGULAR VECTORS OF mat STORED COLUMN-WISE
!       IN work2 BY TRANSPOSE OF RIGHT SINGULAR VECTORS OF mat STORED IN work(:n,:n).
!
#ifdef _BLAS
!
        call gemm( 'N', 'N', m, n, k, one, work2, m, work, n, zero, mat, m )
#else
!
        mat(:m,:n) = matmul( work2(:m,:k), work(:k,:n) )
#endif
!
    else
!
!       PATH 2 : mat HAS MORE COLUMNS THAN ROWS, E.G. n GREATER THAN m.
!
!       ALLOCATE WORK ARRAY.
!
        allocate( work(m,m), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       GENERATE TRANSPOSE OF RIGHT VECTORS BIDIAGONALIZING mat IN mat(:m,:n) .
!       GENERATE LEFT VECTORS BIDIAGONALIZING mat IN work(:m,:m) .
!
        call ortho_gen_bd2( mat, tauq, taup, work )
!
!       PERFORM BIDIAGONAL QR ITERATION, COMPUTE LEFT SINGULAR
!       VECTORS OF mat IN work(:m,:m) AND TRANSPOSE OF RIGHT SINGULAR
!       VECTORS OF mat IN mat(:m,:n) .
!
        call bd_svd2_a( false, s, e, failure, work, mat, sort='d',            &
                        maxiter=maxiter, max_francis_steps=max_francis_steps, &
                        perfect_shift=perfect_shift, bisect=bisect, dqds=dqds )
!
!       DETERMINE THE EFFECTIVE RANK OF mat .
!
        if ( present(tol) ) then
            thr = max( tol*s(1_i4b), safmin )
        else
            thr = safmin
        end if
!
        k = ifirstloc( logical( s<=thr, lgl ) ) - 1_i4b
!
!       COMPUTE RECIPROCALS OF SINGULAR VALUES.
!
        e(:k) = one/s(:k)
!
!       ALLOCATE WORKSPACE NEEDED FOR THE ALGORITHM.
!
        allocate( work2(k,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MULTIPLY TRANSPOSE OF RIGHT SINGULAR VECTORS OF mat BY RECIPROCALS OF  
!       SINGULAR VALUES, STORING RESULT IN work2 .
!
#ifdef _OPENMP
        test_par2 = test_par .and. n*k>=omp_limit
#endif
!
!$OMP PARALLEL DO IF(test_par2), SCHEDULE(STATIC,1), PRIVATE(i)
!
        do i = 1_i4b, n
!
            work2(:k,i) = e(:k)*mat(:k,i)
!
        end do
!
!$OMP END PARALLEL DO
!
!       MULTIPLY LEFT SINGULAR VECTORS OF mat STORED COLUMN-WISE
!       IN work BY TRANSPOSE OF SCALED RIGHT SINGULAR VECTORS OF mat STORED IN work2(:m,:n).
!
#ifdef _BLAS
!
        call gemm( 'N', 'N', m, n, k, one, work, m, work2, k, zero, mat, m )
#else
!
        mat(:m,:n) = matmul( work(:m,:k), work2(:k,:n) )
#endif
!
    end if
!
!   DEALLOCATE WORKSPACE.
!
    deallocate( work, work2 )
!
!   UNDO SCALING IF NECESSARY.
!
    if ( mat_is_scaled ) then
!
        call lascl( mat, matnorm, merge( smlnum, bignum, matnorm<smlnum ) )
!
    end if
!    
!   OUTPUT RANK OF mat, IF REQUIRED.
!    
    if ( present(krank) )  then
!
        krank = k
!
    end if
!    
!   OUTPUT SINGULAR VALUES OF mat, IF REQUIRED.
!    
    if ( present(singvalues) )  then
!
!       UNDO SCALING IF NECESSARY.
!
        if ( mat_is_scaled ) then
!
            call lascl( s, merge( smlnum, bignum, matnorm<smlnum ), matnorm )
!
        end if

        singvalues = s
!
    end if
!
!
! END OF SUBROUTINE comp_ginv_b
! _____________________________
!
    end subroutine comp_ginv_b
!
!
! =========================================================================================
!                   PROCEDURES FOR GENERATION OF BIDIAGONAL MATRICES
! =========================================================================================
!
!
    subroutine gen_bd_mat( type, d, e, failure, known_singval, from_tridiag,  &
                           singval, sort, val1, val2, l0, glu0 )
!
! Purpose
! _______
!
!   GEN_BD_MAT generates different types of bidiagonal matrices with known singular values
!   or specific numerical properties such as clustered singular values for testing
!   purposes of singular value decomposition bidiagonal solvers.
!
!   Optionally, the singular values of the selected bidiagonal matrix can be computed
!   analytically, if possible, or by a bisection algorithm with high absolute and relative
!   accuracies.
!
!
! Arguments
! _________
!
!   TYPE                (INPUT) integer(i4b)
!                       Select the type of bidiagonal matrix BD to be generated by the subroutine.
!
!                       If TYPE is between 1 and 56, the subroutine generates a specific bidiagonal
!                       matrix as described in the comments inside the code of the subroutine.
!                       For other values of TYPE, all diagonal and off-diagonal elements of the bidiagonal
!                       matrix are generated from an uniform random numbers distribution between
!                       0 and 1.
!
!                       For TYPE between 1 and 17, the singular values of the bidiagonal matrix are
!                       known analytically. For other values of TYPE, the singular values may be estimated by a
!                       bisection algorithm with high accuracy.
!                       In all cases, the singular values may be output in the optional parameter SINGVAL.
!
!                       For TYPE between 1 and 11 or 52 and 56, the bidiagonal matrix BD is computed as the
!                       Cholesky factor of symmetric positive-definite tridiagonal matrices.
!
!   D                   (OUTPUT) real(stnd), dimension(:)
!                       On exit, D contains the diagonal elements of the bidiagonal matrix BD.
!
!                       The size of D must verify: size( D )>=2 . 
!
!   E                   (OUTPUT) real(stnd), dimension(:)
!                       On exit, E(2:) contains the off-diagonal elements of the bidiagonal matrix BD.
!                       E(1) is arbitrary, but is set to zero.
!
!                       The size of E must verify: size( E ) = size( D ) . 
!
!   FAILURE             (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates that the singular values of BD are known analytically
!                         or have been computed with high accuracy;
!                       - FAILURE = true  :  indicates that the singular values of BD are not known analytically
!                         and have not been computed with maximum accuracy with the bisection
!                         algorithm.
!
!   KNOWN_SINGVAL       (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit:
!
!                       - KNOWN_SINGVAL = true  :  indicates that the singular values of BD are known analytically for
!                         the selected TYPE.
!                       - KNOWN_SINGVAL = false :  indicates that the eigenvalues of BD are not known analytically for
!                         the selected TYPE.
!
!   FROM_TRIDIAG        (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit:
!
!                       - FROM_TRIDIAG = true  :  indicates that the bidiagonal matrix BD has been computed as the
!                         Cholesky factor of a positive-definite tridiagonal matrix for
!                         the selected TYPE.
!                       - FROM_TRIDIAG = false  : indicates that the bidiagonal matrix BD has not been computed as the
!                         Cholesky factor of a positive-definite tridiagonal matrix for
!                         the selected TYPE.
!
!   SINGVAL             (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                       On exit, the singular values of BD computed analytically or estimated to high
!                       accuracy with a bisection algorithm.
!
!                       The size of SINGVAL must verify: size( SINGVAL ) = size( D ) . 
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the singular values into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd', if the optional
!                       argument SINGVAL is present. For other values of SORT
!                       nothing is done and SINGVAL(:) may not be sorted.
!
!   VAL1                (INPUT, OPTIONAL) real(stnd)
!                       On entry, specifies the parameter d0 for parametrized
!                       bidiagonal matrices (e.g. TYPE= 2-8, 10, 15, 32-35).
!
!                       If this parameter is changed for TYPE between 2 and 8,
!                       care must be taken to insure that the initial symmetric
!                       tridiagonal matrix, which is used to derive the bidiagonal
!                       matrix BD, is positive-definite. If this is not the case,
!                       the subroutine will issue an error message and stop the
!                       program.
!
!                       Also, if this parameter is changed for TYPE between 32 and 35,
!                       which correspond to graded (or reversely graded) matrices
!                       with an arithmetic or geometric progression, care must be taken
!                       to insure that some elements of the arithmetic or geometric progression
!                       will not underflow or overflow as no checks are done in the subroutine
!                       for such errors.
!
!                       The default for VAL1 is:
!
!                       - 2. for TYPE between 2 and 7;
!                       - 3. for TYPE equal to 8;
!                       - 1. for TYPE equal to 10;
!                       - 1. for TYPE equal to 15;
!                       - 1. for TYPE between 32 and 35.
!
!   VAL2                (INPUT, OPTIONAL) real(stnd)
!                       On entry, specifies the parameter e0 for parametrized
!                       bidiagonal matrices (e.g. TYPE= 2-8, 10, 32-35).
!
!                       If this parameter is changed for TYPE between 2 and 8,
!                       care must be taken to insure that the initial symmetric
!                       tridiagonal matrix, which is used to derive the bidiagonal
!                       matrix BD, is positive-definite. If this is not the case,
!                       the subroutine will issue an error message and stop the
!                       program.
!
!                       Also, if this parameter is changed for TYPE between 32 and 35,
!                       which correspond to graded (or reversely graded) matrices
!                       with an arithmetic or geometric progression, care must be taken
!                       to insure that some elements of the arithmetic or geometric progression
!                       will not underflow or overflow as no checks are done in the subroutine
!                       for such errors.
!
!                       The default for VAL2 is:
!
!                       - 1. for TYPE between 2 and 7;
!                       - 2. for TYPE equal to 8;
!                       - 2. for TYPE equal to 10;
!                       - 2. for TYPE between 32 and 35.
!
!   L0                  (INPUT, OPTIONAL) integer(i4b)
!                       On entry, specify the radius of the initial matrix for parametrized
!                       form of glued bidiagonal matrices (e.g. for TYPE equal to 44, 46, 48,
!                       53, 55).
!
!                       L0 must be greater than 0 and preferably less or equal to size( D )/2 .
!                       The default is 5.
!
!   GLU0                (INPUT, OPTIONAL) real(stnd)
!                       On entry, specify the glue parameter for parametrized form of glued 
!                       bidiagonal matrices (e.g. for TYPE equal to 44, 46, 48, 53, 55).
!
!                       The default is sqrt( epsilon(GLU0) ).
!
!
! Further Details
! _______________
!
!   This subroutine tries to take care of imprecisions in intrinsic subroutines (e.g. like the cos
!   function in the gfortran compiler) when computing singular values by analytic formulae.
!
!   For further details on the bidiagonal matrices used for testing in GEN_BD_MAT subroutine, see:
!
!   (1) Gladwell, G.M.L., Jones, T.H., Willms N.B., 2014:
!           A test matrix for an inverse eigenvalue problem.
!           Journal of Applied Mathematics, 14, 6 pages, Article ID 515082, DOI 10.1155/2014/515082.
!
!   (2) Clement, P.A., 1959:
!           A class of triple-diagonal matrices for test purposes.
!           SIAM Review, 1(1):50-52, DOI 10.1137/1001006.
!
!   (3) Gregory, R.T., Karney, D.L., 1969:
!           A collection of matrices for testing computational algorithms.
!           New York: Wiley. Reprinted with corrections by Robert E. Krieger,
!           Huntington, New York, 1978.
!
!   (4) Higham, N.J., 1991: Algorithm 694:
!           A collection of test matrices in MATLAB.
!           ACM Transactions on Mathematical Software 17(3):289-305 DOI 10.1145/114697.116805.
!
!   (5) Godunov, S.K., Antonov, A.G., Kirillyuk, O.P., and Kostin, V.I., 1993:
!           Guaranteed Accuracy in numerical linear algebra.
!           Kluwer Academic Publishers.
!
!   (6) Parlett, B.N., and Vomel, C., 2005:
!           How the MRRR algorithm can fail on tight eigenvalue clusters.
!           Lapack Working Note 163.
!
!   (7) Nakatsukasa, Y., Aishima, K., and Yamazaki, I., 2012:
!           dqds with agressive early deflation.
!           SIAM J. Matrix Anal. Appl., 33(1): 22-51.
!
!   (8) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differenial qd algorithms.
!           Numer. Math., 67: 191-229.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error, svd_error6
    use Reals_Constants,   only : c16, c200, c900, c1_m2, f4_3, pi
    use Utilities,         only : assert_eq, assert, merror, arth, geop
    use Random,            only : random_number_
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out),           dimension(:) :: d, e
    real(stnd), intent(out), optional, dimension(:) :: singval
    real(stnd), intent(in),  optional               :: val1, val2, glu0
!
    integer(i4b), intent(in)           :: type
    integer(i4b), intent(in), optional :: l0
!
    logical(lgl), intent(out), optional :: failure, known_singval, from_tridiag
!
    character,   intent(in), optional :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: abstol, a, b, c, d0, e0, lambda, glu, tmp
    real(stnd), allocatable, dimension(:) :: temp, temp2, x, y, xy
!
    integer      :: iok
    integer(i4b) :: n, i, i1, i2, j, l, p
!
    logical(lgl)  :: known_singval0, from_tridiag0, test_singval, fail
!
!
! PARAMETERS
! __________
!
!
    character(len=*),  parameter :: name_proc='gen_bd_mat'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b),       &
                   int(size(e),i4b),       & 
                   name_proc )
!
    if ( present(singval) ) then
!
        call assert( logical( int(size(singval),i4b)==n, lgl ), &
                     logical( 2_i4b<=n, lgl ),                  &
                     name_proc )
!
        test_singval = true
!
        singval(:n) = zero
!
    else
!
        call assert( logical( 2_i4b<=n, lgl ),  &
                     name_proc )
!
        test_singval = false
!
    end if
!
    if ( present(failure) ) then
!
        failure = false
!
    end if
!
    known_singval0 = 1_i4b<=type .and. type<=17_i4b
!
    if ( present(known_singval) ) then
!
        known_singval = known_singval0
!
    end if
!
    from_tridiag0 =  ( 1_i4b<=type  .and. type<=11_i4b ) .or.  &
                     ( 52_i4b<=type .and. type<=56_i4b )
!
    if ( present(from_tridiag) ) then
!
        from_tridiag =  from_tridiag0
!
    end if
!
!   SET DEFAULT VALUES FOR PARAMETRIZED FORMS OF BIDIAGONAL MATRICES.
!
    select case( type )
!
        case( 2_i4b:7_i4b )
!
            d0 = two
            e0 = one
!
        case( 8_i4b )
!
            d0 = three
            e0 = two
!
        case( 10_i4b )
!
            d0 = one
            e0 = two
!
        case( 15_i4b )
!
            d0 = one
!
        case( 32_i4b:35_i4b )
!
            d0 = one
            e0 = two
!
    end select
!
!   CHANGE THE DEFAULT VALUES OF d0 AND e0 IF REQUIRED.
!
    if ( present(val1) ) then
        d0 = val1
    end if
!
    if ( present(val2) ) then
        e0 = val2
    end if
!
!   GET MACHINE PARAMETERS.
!
    abstol = sqrt( safmin )
!
!   SPECIFY THE RADIUS l FOR PARAMETRIZED FORM OF GLUED
!   BIDIAGONAL MATRICES.
!
    l = 5_i4b
!
    if ( present(l0) ) then
!
        if ( l0>0_i4b ) then
            l = l0
        end if
!
    end if
!
!   SPECIFY THE GLUE PARAMETER FOR PARAMETRIZED FORM OF GLUED
!   BIDIAGONAL MATRICES.
!
    glu = sqrt( ulp )
!
    if ( present(glu0) ) then
!
        glu = glu0
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    i = max( n, 21_i4b, (2_i4b*l)+1_i4b )
!
    if ( from_tridiag0 .and. known_singval0 ) then
!
        allocate( temp(i), temp2(i), x(n), stat=iok )
!
    else
!
        allocate( temp(i), temp2(i), stat=iok )
!
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    select case( type )
!
        case( 1_i4b )
!
!           CASE 1: GENERATE A BIDIAGONAL MATRIX FROM FROM THE CHOLESKY FACTOR OF A PSEUDO
!                   JACOBIAN TRIDIAGONAL MATRIX WITH THE FIRST DIAGONAL ELEMENT d(1) EQUAL TO n
!                   AND ALL OTHER DIAGONAL ELEMENTS OF THE TRIDIAGONAL MATRIX ARE EQUAL TO n-1 .
!                   FIRST OFF-DIAGONAL ELEMENT e(1) IS EQUAL TO 0 AND OFF-DIAGONAL ELEMENT e(j)
!                   IS EQUAL 0.5*sqrt( (j-1)*(2*n-j) ) FOR j=2 to n-1 .
!                   SEE GLADWELL ET AL.
!                   EIGENVALUES OF THE PSEUDO JACOBIAN TRIDIAGONAL ARE EQUAL TO 1, 3, ..., 2*n-3
!                   AND n. THE SINGULAR VALUES ARE THE SQUARE ROOTS OF THESE EIGENVALUES.
!
            d(1_i4b)   = real( n, stnd )
            d(2_i4b:n) = real( n-1_i4b, stnd )
!
            e(1_i4b) = zero
!
            do j = 2_i4b, n-1_i4b
                e(j) = half*sqrt( real( (j-1_i4b)*(2_i4b*n-j), stnd ) )
            end do
!
            x(:n-1_i4b) = arth( one, two, n-1_i4b )
            x(n)        = d(1_i4b)
!
        case( 2_i4b )
!
!           CASE 2: GENERATE A BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR OF A
!                   JACOBIAN TRIDIAGONAL MATRIX WITH EIGENVALUES d0, d0+2*e0,
!                   d0+4*e0, d0+6*e0, ..., d0+(2*n-2)*e0.
!                   SEE GLADWELL ET AL.
!                   THE SINGULAR VALUES ARE THE SQUARE ROOTS OF THESE EIGENVALUES.
!
            a  = d0 + e0*real( n-1_i4b, stnd ) 
!
            d(:n) = a
!
            do j = 1_i4b, n-2_i4b
                e(j) = -e0*half*sqrt( real( j*(2_i4b*n-j-1_i4b), stnd ) )
            end do
!
            e(n-1_i4b) = -e0*sqrt( real( n*(n-1_i4b), stnd )*half )
!
            x(:n) = arth( d0, two*e0, n )
!
        case( 3_i4b )
!
!           CASE 3: GENERATE A BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR OF A PSEUDO
!                   JACOBIAN TRIDIAGONAL MATRIX WITH EIGENVALUES d0+e0, d0+e0+2*e0,
!                   d0+e0+4*e0, d0+e0+6*e0, ..., d0+e0+(2*n-4)*e0 AND lambda=d0+e0+2*e0 .
!                   SEE GLADWELL ET AL.
!                   THE SINGULAR VALUES ARE THE SQUARE ROOTS OF THESE EIGENVALUES.
!
            lambda = d0 + e0 + two*e0
!
            a  = d0 + e0*real( n-1_i4b, stnd ) 
!
            d(1_i4b)   = lambda
            d(2_i4b:n) = a
!
            e(1_i4b) = zero
!
            do j = 2_i4b, n-1_i4b
                e(j) = -e0*half*sqrt( real( (j-1_i4b)*(2_i4b*n-j), stnd ) )
            end do
!
            x(:n-1_i4b) = arth( d0+e0, two*e0, n-1_i4b )
            x(n)        = lambda
!
        case( 4_i4b )
!
!           CASE 4: GENERATE A BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR OF EXAMPLE 7.4
!                   OF GREGORY AND KARNEY. (P. 137) OR EXAMPLE 2 OF GODUNOV ET AL. (P. 398),
!                   WHICH IS A TOEPLITZ TRIDIAGONAL MATRIX WITH EIGENVALUES
!                   d0 + 2*e0*cos(j*pi/(n+1)), j=1:n, WHERE d0 AND e0
!                   ARE SCALARS, ALL THE DIAGONAL ELEMENTS ARE EQUAL TO d0
!                   AND ALL SUBDIAGONAL AND SUPERDIAGONAL ELEMENTS ARE EQUAL TO e0.
!                   THE SINGULAR VALUES ARE THE SQUARE ROOTS OF THESE EIGENVALUES.
!
            d(1_i4b:n)       = d0     
            e(1_i4b:n-1_i4b) = e0
!
            c = real( n, stnd) + one
!
            do j = 1_i4b, n
!
                tmp = real( j, stnd)/c
!
                if ( tmp==half ) then
                    x(j) = d0
                else if ( tmp==(four/six) ) then
                    x(j) = d0 - e0
                else if ( tmp==one ) then
                    x(j) = d0 - two*e0
                else if ( tmp==zero ) then
                    x(j) = d0 + two*e0
                else
                    x(j) = d0 + two*e0*cos( tmp*pi )
                end if
!
            end do
!
        case( 5_i4b )
!
!           CASE 5: EXAMPLE 7.5 OF GREGORY AND KARNEY. P. 137.
!                   GENERATE A BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR
!                   OF A TRIDIAGONAL MATRIX WITH EIGENVALUES
!                   d0 + 2*e0*cos(2*j*pi/(2*n+1)), j=1:n, WHERE d0 AND e0
!                   ARE SCALARS. ALL THE DIAGONAL ELEMENTS OF THE TRIDIAGONAL
!                   MATRIX ARE EQUAL TO d0, EXCEPTED THE FIRST, WHICH IS EQUAL
!                   TO d0-e0, AND ALL SUBDIAGONAL AND SUPERDIAGONAL ELEMENTS
!                   ARE EQUAL TO e0.
!                   THE SINGULAR VALUES ARE THE SQUARE ROOTS OF THESE EIGENVALUES.
!
            d(1_i4b)         = d0 - e0   
            d(2_i4b:n)       = d0     
            e(1_i4b:n-1_i4b) = e0
!
            c = two*real( n, stnd) + one
!
            do j = 1_i4b, n
!
                tmp = (two*real( j, stnd))/c
!
                if ( tmp==half ) then
                    x(j) = d0
                else if ( tmp==(four/six) ) then
                    x(j) = d0 - e0
                else if ( tmp==one ) then
                    x(j) = d0 - two*e0
                else if ( tmp==zero ) then
                    x(j) = d0 + two*e0
                else
                    x(j) = d0 + two*e0*cos( tmp*pi )
                end if
!
            end do
!
        case( 6_i4b )
!
!           CASE 6: EXAMPLE 7.6 OF GREGORY AND KARNEY. P. 138.
!                   GENERATE A BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR OF
!                   A TRIDIAGONAL MATRIX WITH EIGENVALUES
!                   d0 + 2*e0*cos((2*j-1)*pi/(2*n)), j=1:n, WHERE d0 AND e0
!                   ARE SCALARS, ALL THE DIAGONAL ELEMENTS OF THE TRIDIAGONAL
!                   MATRIX ARE EQUAL TO d0, EXCEPTED THE FIRST, WHICH IS EQUAL
!                   TO d0-e0, AND THE LAST, WHICH IS EQUAL TO d0+e0,
!                   AND ALL SUBDIAGONAL AND SUPERDIAGONAL ELEMENTS ARE EQUAL TO e0.
!
            d(1_i4b)         = d0 - e0
            d(n)             = d0 + e0
            d(2_i4b:n-1_i4b) = d0
            e(1_i4b:n-1_i4b) = e0
!
            c = two*real( n, stnd)
!
            do j = 1_i4b, n
!
                tmp = (two*real( j, stnd) - one)/c
!
                if ( tmp==half ) then
                    x(j) = d0
                else if ( tmp==(four/six) ) then
                    x(j) = d0 - e0
                else if ( tmp==one ) then
                    x(j) = d0 - two*e0
                else if ( tmp==zero ) then
                    x(j) = d0 + two*e0
                else
                    x(j) = d0 + two*e0*cos( tmp*pi )
                end if
!
            end do
!        
        case( 7_i4b )
!
!           CASE 7: EXAMPLE 7.7 OF GREGORY AND KARNEY. P. 138.
!                   GENERATE A BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR
!                   OF A TRIDIAGONAL MATRIX WITH EIGENVALUES
!                   d0 + 2*e0*cos((pi/n)*(j-1)), j=1:n, WHERE d0 AND e0
!                   ARE SCALARS, ALL THE DIAGONAL ELEMENTS OF THE TRIDIAGONAL
!                   MATRIX ARE EQUAL TO d0, EXCEPTED THE FIRST AND THE LAST,
!                   WHICH ARE EQUAL TO d0+e0, AND ALL SUBDIAGONAL AND
!                   SUPERDIAGONAL ELEMENTS ARE EQUAL TO e0.
!                   THE SINGULAR VALUES ARE THE SQUARE ROOTS OF THESE EIGENVALUES.
!
            d(1_i4b)         = d0 + e0   
            d(n)             = d0 + e0  
            d(2_i4b:n-1_i4b) = d0     
            e(1_i4b:n-1_i4b) = e0
!
            c = real( n, stnd)
!
            do j = 1_i4b, n
!
                tmp = (real( j, stnd) - one)/c
!
                if ( tmp==half ) then
                    x(j) = d0
                else if ( tmp==(four/six) ) then
                    x(j) = d0 - e0
                else if ( tmp==one ) then
                    x(j) = d0 - two*e0
                else if ( tmp==zero ) then
                    x(j) = d0 + two*e0
                else
                    x(j) = d0 + two*e0*cos( tmp*pi )
                end if
!
            end do            
!        
        case( 8_i4b )
!
!           CASE 8: EXAMPLE 7.8 AND 7.9 OF GREGORY AND KARNEY. P. 139.
!                   IF N IS EVEN GENERATE A BIDIAGONAL MATRIX AS THE CHOLESKY FACTOR OF
!                   A TRIDIAGONAL MATRIX WITH EIGENVALUES:
!                   (d0 + e0 + sqrt( (d0-e0)**2 + 16*( cos((j*pi)/(n+1)) )**2 ) )/2  FOR j=1:n/2
!                   (d0 + e0 - sqrt( (d0-e0)**2 + 16*( cos((j*pi)/(n+1)) )**2 ) )/2  FOR j=1:n/2
!
!                   AND IF N IS ODD GENERATE A BIDIAGONAL MATRIX AS THE CHOLESKY FACTOR OF
!                   A TRIDIAGONAL MATRIX WITH EIGENVALUES:
!                   (d0 + e0 + sqrt( (d0-e0)**2 + 16*( cos((j*pi)/(n+2)) )**2 ) )/2  F0R j=1:n/2
!                   (d0 + e0 - sqrt( (d0-e0)**2 + 16*( cos((j*pi)/(n+2)) )**2 ) )/2  FOR j=1:n/2 AND d0
!                   , WHERE d0 AND e0 ARE SCALARS.
!                   ALL SUBDIAGONAL AND SUPERDIAGONAL ELEMENTS OF THE TRIDIAGONAL MATRIX
!                   ARE EQUAL TO 1, AND DIAGONAL ELEMENTS ARE ALTERNATIVELY EQUAL TO d0 AND e0,
!                   STARTING BY d0.
!                   THE SINGULAR VALUES ARE THE SQUARE ROOTS OF THESE EIGENVALUES.
!
            d(1_i4b:n:2_i4b) = d0     
            d(2_i4b:n:2_i4b) = e0     
            e(1_i4b:n-1_i4b) = one
!
            a = ( d0 - e0 )**2_i4b
            b = d0 + e0
            c = real( n, stnd) + one
!
            do j = 1_i4b, n/2_i4b
!
                tmp = real(j,stnd)/c
!
                if  ( tmp==zero  ) then
                    lambda = sqrt( a + c16 )
                else if  ( tmp==quarter  ) then
                    lambda = sqrt( a + eight )
                else if ( tmp==half  ) then
                    lambda = abs( d0 - e0 )
                else
                    lambda = sqrt( a + c16*( cos( tmp*pi ) )**2 )
                end if
!
                x(j)         = ( b + lambda )/two
                x(n/2_i4b+j) = ( b - lambda )/two
!
            end do
!
            if ( 2_i4b*(n/2_i4b)/=n ) x(n) = d0
!
        case( 9_i4b )
!
!           CASE 9: GENERATE A BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR OF
!                   TRIDIAGONAL MATRIX EXAMPLE 1 OF GODUNOV ET AL. P. 391.
!                   ALL DIAGONAL ELEMENTS OF THE TRIDIAGONAL MATRIX ARE EQUAL TO 1 AND
!                   OFF-DIAGONAL ELEMENTS e(j) ARE EQUAL TO 1/(e0**((j+1)/2_i4b)) FOR j=1, n, 2
!                   AND 0 FOR FOR j=2, n, 2 .
!
!                   IF n IS EVEN THE EIGENVALUES ARE :
!                   1 - 1/(e0**j) FOR j=1, ..., n/2 AND 1 + 1/(e0**j) FOR j=1, ..., n/2
!                   IF N IS ODD THE EIGENVALUES ARE :
!                   1 - 1/(e0**j) FOR j=1, ..., (n-1)/2 , 1 + 1/(e0**j) FOR j=1, ..., (n-1)/2 AND 1
!                   THE SINGULAR VALUES ARE THE SQUARE ROOTS OF THESE EIGENVALUES.
!
            e0 = quarter
!
            d(1_i4b:n) = one                     
!
            e(2_i4b:n:2_i4b) = zero
!
            do j = 2_i4b, n, 2_i4b
                e(j-1_i4b) = e0**(j/2_i4b)
            end do
!
            do j = 1_i4b, n/2_i4b
                temp(j) = e0**j
            end do
!
            x(:n/2_i4b) = one - temp(:n/2_i4b)
!
            if ( n == 2_i4b*(n/2_i4b) ) then
                x((n/2_i4b)+1_i4b:n) = one + temp(:n/2_i4b)
            else
                x((n/2_i4b)+1_i4b)   = one
                x((n/2_i4b)+2_i4b:n) = one + temp(:n/2_i4b)
            end if
!
        case( 10_i4b )
!
!           CASE 10: PARAMETRIZED FORM OF A BIDIAGONAL MATRIX GENERATED FROM
!                    THE CHOLESKY FACTOR OF TRIDIAGONAL MATRIX EXAMPLE 1 OF GODUNOV ET AL. P. 391.
!                    ALL DIAGONAL ELEMENTS OF THE TRIDIAGONAL MATRIX ARE EQUAL TO d0 AND 
!                    OFF-DIAGONAL ELEMENTS e(j)  ARE EQUAL TO d0/(e0**((j+1)/2_i4b)) FOR j=1, n, 2
!                    AND 0 FOR FOR j=2, n, 2 .
!
!                    IF n IS EVEN THE EIGENVALUES ARE :
!                    d0 - d0/(e0**j) FOR j=1, ..., n/2 AND d0 + d0/(e0**j) FOR j=1, ..., n/2
!                    IF N IS ODD THE EIGENVALUES ARE :
!                    d0 - d0/(e0**j) FOR j=1, ..., (n-1)/2 , d0 + d0/(e0**j) FOR j=1, ..., (n-1)/2 AND d0 .
!                   THE SINGULAR VALUES ARE THE SQUARE ROOTS OF THESE EIGENVALUES.
!
            d(1_i4b:n) = d0                     
!
            e(2_i4b:n:2_i4b) = zero
!
            tmp = one/e0
!
            do j = 2_i4b, n, 2_i4b
                e(j-1_i4b) = d0*(tmp**(j/2_i4b))
            end do
!
            do j = 1_i4b, n/2_i4b
                temp(j) = d0*(tmp**j)
            end do
!
            x(:n/2_i4b) = d0 - temp(:n/2_i4b)
!
            if ( n == 2_i4b*(n/2_i4b) ) then
                x((n/2_i4b)+1_i4b:n) = d0 + temp(:n/2_i4b)
            else
                x((n/2_i4b)+1_i4b)   = d0
                x((n/2_i4b)+2_i4b:n) = d0 + temp(:n/2_i4b)
            end if
!
        case( 11_i4b )
!
!           CASE 11: GENERATE A BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR OF
!                    A (1-2-1) TRIDIAGONAL MATRIX.
!                    THE EIGENVALUES ARE EQUAL TO  2 + 2*cos(j*pi/(n+1)), j=1:n
!                    THE SINGULAR VALUES ARE THE SQUARE ROOTS OF THESE EIGENVALUES.
!
            d(:n) = two
            e(:n) = one
!
            c = real( n, stnd) + one
!
            do j = 1_i4b, n
!
                tmp = real( j, stnd)/c
!
                if ( tmp==half ) then
                    x(j) = two
                else if ( tmp==(four/six) ) then
                    x(j) = one
                else if ( tmp==one ) then
                    x(j) = zero
                else
                    x(j) = two + two*cos( tmp*pi )
                end if
!
            end do
!
        case( 12_i4b )
!
!           CASE 12: GENERATE A BIDIAGONAL MATRIX FROM A CLEMENT BIDIAGONAL
!                    MATRIX WITH SINGULAR VALUES
!                    EQUAL TO 2*n-1, 2*n-3, 2*n-5, ..., 1.
!
            p = 2_i4b*n - 1_i4b
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( x(p), y(p), xy(p), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            x(:p)  = arth( one, one, p )
            y(:p)  = x(p:1_i4b:-1_i4b)
            xy(:p) = sqrt( x(:p)*y(:p) )
!
            d(:n)      = xy(1_i4b:p:2_i4b)
            e(2_i4b:n) = xy(2_i4b:p:2_i4b)
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( x, y, xy )
!
            if ( test_singval ) then
!
                a = real( p, stnd )
!
                singval(:n) = arth( a, -two, n )
!
            end if
!
        case( 13_i4b )
!
!           CASE 13: GENERATE A BIDIAGONAL MATRIX FROM TRIDIAGONAL EXAMPLE 2 OF  
!                    OF GODUNOV ET AL. P. 400.
!                    ALL DIAGONAL AND OFF-DIAGONAL ELEMENTS OF THE BIDIAGONAL
!                    MATRIX ARE EQUAL TO 0.5.
!                    SINGULAR VALUES ARE THE POSITIVE ROOTS OF THE CHEBYSHEV
!                    POLYNOMIALS OF SECOND KIND AND ORDER 4*n, WITH
!                    THE LEADING COEFFICIENT EQUAL TO ONE AND ARE EQUAL TO
!                    cos(j*pi/(2*n+1)), j=1:n.
!
            d(1_i4b:n) = half   
            e(2_i4b:n) = half
!
            if ( test_singval ) then
!
                c = two*real( n, stnd) + one
!
                do j = 1_i4b, n
!
                    tmp = real( j, stnd)
!
                    if ( (tmp/c)==half ) then
                        singval(j) = zero
                    else
                        singval(j) = cos( (tmp*pi)/c )
                    end if
!
                end do
!
            end if
!
        case( 14_i4b )
!
!           CASE 14: GENERATE A BIDIAGONAL MATRIX FROM TRIDIAGONAL EXAMPLE 7.4 OF GREGORY
!                    AND KARNEY. (P. 137) OR TRIDIAGONAL EXAMPLE 2 OF GODUNOV ET AL. (P. 398).
!                    GENERATE A TOEPLITZ BIDIAGONAL MATRIX WITH SINGULAR VALUES
!                    2*d0*cos(j*pi/(2*n+1)), j=1:n, WHERE d0 IS EQUAL TO 2.
!                    ALL THE DIAGONAL ELEMENTS AND ALL OFF-DIAGONAL
!                    ELEMENTS ARE EQUAL TO d0.
!
            d0 = two
!
            d(1_i4b:n) = d0     
            e(2_i4b:n) = d0
!
            if ( test_singval ) then
!
                c = two*real( n, stnd) + one
!
                do j = 1_i4b, n
!
                    tmp = real( j, stnd)
!
                    if ( (tmp/c)==half ) then
                        singval(j) = zero
                    else
                        singval(j) = two*d0*cos( (tmp*pi)/c )
                    end if
!
                end do
!
            end if
!
        case( 15_i4b )
!
!           CASE 15: PARAMETRIZED FORM OF A BIDIAGONAL MATRIX FROM TRIDIAGONAL EXAMPLE 7.4
!                    OF GREGORY AND KARNEY. (P. 137) OR TRIDIAGONAL EXAMPLE 2 OF GODUNOV ET
!                    AL. (P. 398). GENERATE A TOEPLITZ BIDIAGONAL MATRIX WITH SINGULAR VALUES
!                    2*d0*cos(j*pi/(2*n+1)), j=1:n, WHERE d0 IS SCALAR.
!                    ALL THE DIAGONAL ELEMENTS AND ALL OFF-DIAGONAL ELEMENTS OF THE 
!                    BIDIAGONAL MATRIX ARE EQUAL TO d0.
!
            d(1_i4b:n) = d0     
            e(2_i4b:n) = d0
!
            if ( test_singval ) then
!
                c = two*real( n, stnd) + one
!
                do j = 1_i4b, n
!
                    tmp = real( j, stnd)
!
                    if ( (tmp/c)==half ) then
                        singval(j) = zero
                    else
                        singval(j) = two*d0*cos( (tmp*pi)/c )
                    end if
!
                end do
!
            end if
!
        case( 16_i4b )
!
!           CASE 16: GENERATE A ZERO BIDIAGONAL MATRIX.
!                    THE SINGULAR VALUES ARE ALL EQUAL TO ZERO.
!
            d(:n) = zero
            e(:n) = zero
!
            if ( test_singval ) then
!
                singval(:n) = zero
!
            end if
!
        case( 17_i4b )
!
!           CASE 17: GENERATE A UNIT BIDIAGONAL MATRIX.
!                    THE SINGULAR VALUES ARE ALL EQUAL TO ONE.
!
            d(:n) = one
            e(:n) = zero
!
            if ( test_singval ) then
!
                singval(:n) = one
!
            end if
!
        case( 18_i4b )
!
!           CASE 18: GENERATE A BIDIAGONAL MATRIX FROM TRIDIAGONAL EXAMPLE 3
!                    OF GODUNOV ET AL. P. 401.
!                    SINGULAR VALUES ARE THE POSITIVE ROOTS OF THE ORTHOGONAL
!                    LEGENDRE POLYNOMIALS OF ORDER 4*n.
!
            do l = 2_i4b, 2_i4b*n, 2_i4b
                j          = ( (2_i4b*l) - 1_i4b )*( (2_i4b*l) + 1_i4b )
                tmp        = real( j, stnd) 
                d(l/2_i4b) = real( l, stnd)/sqrt( tmp )
            end do
!
            do l = 3_i4b, (2_i4b*n)-1_i4b, 2_i4b
                j                  = ( (2_i4b*l) -1_i4b )*( (2_i4b*l) + 1_i4b )
                tmp                = real( j, stnd) 
                e((l+1_i4b)/2_i4b) = real( l, stnd)/sqrt( tmp )
            end do
!
        case( 19_i4b )
!
!           CASE 19: BIDIAGONAL EXAMPLE 1 OF GODUNOV ET AL. P. 417.
!
            d(:n) = one
            e(:n) = ten
!
        case( 20_i4b )
!
!           CASE 20: BIDIAGONAL EXAMPLE 2 OF GODUNOV ET AL. P. 417.
!        
            d(:n) = c1_m2
            e(:n) = c900
!
        case( 21_i4b )
!
!           CASE 21: BIDIAGONAL EXAMPLE 3 OF GODUNOV ET AL. P. 419.
!        
            j = n/3_i4b
            l = n - j
!
            d(:j) = c1_m2
            e(:j) = c900
!
            d(j+1_i4b:l) = c900
            e(j+1_i4b:l) = c1_m2
!
            d(l+1_i4b:n) = c1_m2
            e(l+1_i4b:n) = c900
!
        case( 22_i4b:26_i4b )
!
!           CASES 22 to 26: GENERATE SELECTED TOEPLITZ-LIKE BIDIAGONAL MATRICES.
!
            select case( type )
!
                  case( 22_i4b )
!
                      d0 = quarter
!
                  case( 23_i4b )
!
                      d0 = half
!
                  case( 24_i4b )
!
                      d0 = f4_3
!
                  case( 25_i4b )
!
                      d0 = 0.875_stnd
!
                  case( 26_i4b )
!
                      d0 = two
!
              end select
!
              e0 = one
!
              d(:n)      = d0
              e(2_i4b:n) = e0
!
        case( 27_i4b:31_i4b )
!
!           CASES 27 to 31: GENERATE SELECTED GRADED BIDIAGONAL MATRICES WITH
!                           A GEOMETRIC PROGRESSION.
!
            d0 = one
!
            select case( type )
!
                case( 27_i4b )
!
                    e0 = quarter
!
                case( 28_i4b )
!
                    e0 = half
!
                case( 29_i4b )
!
                    e0 = f4_3
!
                case( 30_i4b )
!
                    e0 = 0.875_stnd
!
                case( 31_i4b )
!
                    e0 = 1.5_stnd
!
            end select
!
            d(:n)      = geop( d0, e0, n )
            e(2_i4b:n) = d(:n-1_i4b)
!
        case( 32_i4b )
!
!           CASE 32: GENERATE A GRADED BIDIAGONAL MATRIX WITH A GEOMETRIC PROGRESSION
!                    (PARAMETRIZED FORM).
!
            d(:n)      = geop( d0, e0, n )
            e(2_i4b:n) = d(:n-1_i4b)
!
        case( 33_i4b )
!
!           CASE 33: GENERATE A REVERSELY GRADED BIDIAGONAL MATRIX WITH A GEOMETRIC PROGRESSION.
!                    (PARAMETRIZED FORM).
!
            d(n:1_i4b:-1_i4b) = geop( d0, e0, n )
            e(2_i4b:n)        = d(:n-1_i4b)
!
        case( 34_i4b )
!
!           CASE 34: GENERATE A GRADED BIDIAGONAL MATRIX WITH AN ARITHMETIC PROGRESSION.
!                    (PARAMETRIZED FORM).
!
            d(:n)      = arth( d0, e0, n )
            e(2_i4b:n) = d(:n-1_i4b)
!
        case( 35_i4b )
!
!           CASE 35: GENERATE A REVERSELY GRADED BIDIAGONAL MATRIX WITH AN ARITHMETIC PROGRESSION.
!                    (PARAMETRIZED FORM).
!
            d(n:1_i4b:-1_i4b) = arth( d0, e0, n )
            e(2_i4b:n)        = d(:n-1_i4b)
!
        case( 36_i4b )
!
!           CASE 36: EXAMPLE 1 OF NAKATSUKASA ET AL. P. 47 TABLE 1.
!
            d0    = real( n, stnd )
            d(:n) = arth( d0, -one, n )
!       
            e(2_i4b:n) = one
!
        case( 37_i4b )
!
!           CASE 37: EXAMPLE 4 OF NAKATSUKASA ET AL. P. 47 TABLE 1.
!
            d0 = real( n, stnd) + one
!
            do j = 2_i4b, n, 2_i4b
                d(j)       = real( j/2_i4b, stnd )
                d(j-1_i4b) = d0 - d(j)
            end do
!
            if ( n/=2_i4b*(n/2_i4b) ) then
                d(n) = d0*half
            end if
!
            do j = 1_i4b, n-1_i4b
                e(j+1_i4b) = real( n-j, stnd )/five
            end do
!
        case( 38_i4b )
!
!           CASE 38: EXAMPLE 5 OF NAKATSUKASA ET AL. P. 47 TABLE 1.
!
            i = n/2_i4b
            if ( n/=2_i4b*i ) then
                i = i + 1_i4b
            end if
!
            d0 = real( i, stnd )
            e0 = 1.01_stnd
!
            d(i)                    = one
            d(i+1_i4b:n)            = geop( e0, e0, n-i)
            d(i-1_i4b:1_i4b:-1_i4b) = geop( e0, e0, i-1_i4b)
!       
            e(2_i4b:n) = one
!
        case( 39_i4b )
!
!           CASE 39: GENERATE A BIDIAGONAL MATRIX WITH A SINGULAR VALUE
!                    NEAR ONE AND A CLUSTER OF SMALL SINGULAR VALUES.
!
            d0 = c200*ulp
!
            d(1_i4b)   = one
            d(2_i4b:n) = d0
!
            e(2_i4b:n) = d0
!
        case( 40_i4b )
!
!           CASE 40: GENERATE A PSEUDO HILBERT BIDIAGONAL MATRIX.
!                    THE HILBERT MATRIX HAS (i,j) ELEMENT EQUAL TO
!                    1/(i+j-1). 
!
            do j = 1_i4b, n-1_i4b
                d(j) = one/( real( 2_i4b*j - 1_i4b, stnd ) )
                e(j+1_i4b) = one/( real( 2_i4b*j, stnd ) )
            end do
!       
            d(n) = one/( real( 2_i4b*n - 1_i4b, stnd ) )
!
        case( 41_i4b )
!
!           CASE 41: FROM EXAMPLE 6 OF FERNANDO AND PARLETT P. 218,
!                    GENERATE A WILKINSON-TYPE BIDIAGONAL MATRIX .
!
            i = n/2_i4b
!
            do j = 1_i4b, n
!
                d(j) = real( abs( j-i+1_i4b ), stnd )
!
            end do
!
            e(2_i4b:n) = one
!
        case( 42_i4b )
!
!           CASE 42: FROM EXAMPLE 7 OF FERNANDO AND PARLETT P. 218,
!                    GENERATE A WILKINSON-TYPE BIDIAGONAL MATRIX .
!
            i = n/2_i4b
            l = n/4_i4b
!
            do j = 1_i4b, i
!
                d(j) = real( abs( j-l+1_i4b ), stnd )
!
            end do
!
            do j = i+1_i4b, n
!
                d(j) = d(j-i)
!
            end do
!
            e(2_i4b:n) = one
!
        case( 43_i4b )
!
!           CASE 43: FROM EXAMPLE 7.2 OF GREGORY AND KARNEY. P. 135,
!                    GENERATE A WILKINSON-LIKE BIDIAGONAL MATRIX .
!
            i = n/2_i4b
            d0 = real( i, stnd )
!
            d(1_i4b:i)  = arth( d0, -one, i )
            if ( n/=2_i4b*i ) then
                d(i+1_i4b)          = zero
                d(n:i+2_i4b:-1_i4b) = d(1_i4b:i)
            else
                d(n:i+1_i4b:-1_i4b) = d(1_i4b:i)
            end if          
!
            e(2_i4b:n) = one
!        
        case( 44_i4b )
!
!           CASE 44: GENERATE A GLUED WILKINSON-LIKE BIDIAGONAL MATRIX FROM
!                    EXAMPLE 7.2 OF GREGORY AND KARNEY. P. 135.
!         
            d0  = real( l, stnd )
!
            temp(1_i4b:l)                      = arth( d0, -one, l )
            temp(l+1_i4b)                      = zero
            temp(2_i4b*l+1_i4b:l+2_i4b:-1_i4b) = temp(1_i4b:l)
!
            temp2(1_i4b:2_i4b*l) = one
            temp2(2_i4b*l+1_i4b) = glu     
!
            do i=0_i4b, n-1_i4b, 2_i4b*l+1_i4b
!
                i1 = i + 1_i4b
                i2 = min( i + 2_i4b*l+1_i4b, n)
                j  = i2 - i1 + 1_i4b
!
                d(i1:i2) = temp(1_i4b:j)
                e(i1:i2) = temp2(1_i4b:j)
!
                if ( i/=0_i4b ) then
                    d(i)  = d(i)  + glu
                    d(i1) = d(i1) + glu
                end if
!
            end do
!
            e(2_i4b:n) = e(1_i4b:n-1_i4b)
!        
        case( 45_i4b )
!
!           CASE 45: FROM EXAMPLE 7.3 OF GREGORY AND KARNEY. P. 136,
!                    GENERATE A WILKINSON-LIKE BIDIAGONAL MATRIX .
!
            i = n/2_i4b
            d0 = real( i, stnd )
!
            d(1_i4b:i)  = arth( d0, -one, i )
            if ( n/=2_i4b*i ) then
                d(i+1_i4b)          = zero
                d(n:i+2_i4b:-1_i4b) = -d(1_i4b:i)
            else
                d(n:i+1_i4b:-1_i4b) = -d(1_i4b:i)
            end if          
!
            e(2_i4b:n) = one
!        
        case( 46_i4b )
!
!           CASE 46: GENERATE A GLUED WILKINSON-LIKE BIDIAGONAL MATRIX FROM
!                    EXAMPLE 7.3 OF GREGORY AND KARNEY. P. 136.
!         
            d0  = real( l, stnd )
!
            temp(1_i4b:l)                      = arth( d0, -one, l )
            temp(l+1_i4b)                      = zero
            temp(2_i4b*l+1_i4b:l+2_i4b:-1_i4b) = -temp(1_i4b:l)
!
            temp2(1_i4b:2_i4b*l) = one
            temp2(2_i4b*l+1_i4b) = glu        
!
            do i=0_i4b, n-1_i4b, 2_i4b*l+1_i4b
!
                i1 = i + 1_i4b
                i2 = min( i + 2_i4b*l+1_i4b, n)
                j  = i2 - i1 + 1_i4b
!
                d(i1:i2) = temp(1_i4b:j)
                e(i1:i2) = temp2(1_i4b:j)
!
                if ( i/=0_i4b ) then
                    d(i)  = d(i)  + glu
                    d(i1) = d(i1) + glu
                end if
!
            end do
!
            e(2_i4b:n) = e(1_i4b:n-1_i4b)
!
        case( 47_i4b )
!
!           CASE 47: GENERATE A GLUED KIMURA BIDIAGONAL MATRIX WITH RADIUS EQUAL TO 11.
!
            d0  = real( 11_i4b, stnd )
            glu = 10.**(-4_i4b)
!
            temp(1_i4b:11_i4b)         = arth( d0, -one, 11_i4b )
            temp(21_i4b:12_i4b:-1_i4b) = temp(1_i4b:10_i4b)
!
            temp2(1_i4b:20_i4b) = one
            temp2(21_i4b)       = glu
!
            do i=0_i4b, n-1_i4b, 21_i4b
!
                i1 = i + 1_i4b
                i2 = min( i + 21_i4b, n)
                j  = i2 - i1 + 1_i4b
!
                d(i1:i2) = temp(1_i4b:j)
                e(i1:i2) = temp2(1_i4b:j)
!
            end do
!
            e(2_i4b:n) = e(1_i4b:n-1_i4b)
!
        case( 48_i4b )
!
!           CASE 48: GENERATE A GLUED KIMURA BIDIAGONAL MATRIX (PARAMETRIZED FORM).
!
            d0  = real( l, stnd )
!
            temp(1_i4b:l)                      = arth( d0, -one, l )
            temp(2_i4b*l-1_i4b:l+1_i4b:-1_i4b) = temp(1_i4b:l-1_i4b)
!
            temp2(1_i4b:2_i4b*l-2_i4b) = one
            temp2(2_i4b*l-1_i4b)       = glu
!
            do i=0_i4b, n-1_i4b, 2_i4b*l-1_i4b
!
                i1 = i + 1_i4b
                i2 = min( i +  2_i4b*l-1_i4b, n)
                j  = i2 - i1 + 1_i4b
!
                d(i1:i2) = temp(1_i4b:j)
                e(i1:i2) = temp2(1_i4b:j)
!
            end do
!
            e(2_i4b:n) = e(1_i4b:n-1_i4b)
!
        case( 49_i4b )
!
!           CASE 49: GENERATE A BIDIAGONAL MATRIX FROM LAGUERRE RECURRENCE.
!
            do j = 1_i4b, n
                lambda  =  real( j, stnd)
                e(j) = lambda + one
                lambda  = two*lambda + one
                d(j) = lambda
            end do
!
            e(2_i4b:n) = e(1_i4b:n-1_i4b)
!
        case( 50_i4b )
!
!           TEST 50: GENERATE A BIDIAGONAL MATRIX FROM HERMITE RECURRENCE.
!
            d(:n) = zero
!
            do j = 1_i4b, n-1_i4b
                lambda =  sqrt( real( j, stnd) )
                e(j+1_i4b) = lambda
            end do
!
        case( 51_i4b )
!
!           TEST 51: GENERATE A BIDIAGONAL MATRIX FROM HERMITE RECURRENCE.
!
            p = 2_i4b*n - 1_i4b
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( x(p), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            do j = 1_i4b, p
                x(j) = sqrt( real( j, stnd) )
            end do
!
            d(1_i4b:n) = x(1_i4b:p:2_i4b)
            e(2_i4b:n) = x(2_i4b:p:2_i4b)
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( x )
!        
        case( 52_i4b )
!
!           CASE 52: GENERATE A BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR
!           OF A WILKINSON DEFINITE-POSITIVE TRIDIAGONAL MATRIX.
!
            i = n/2_i4b
            d0 = real( i, stnd )
!
            d(1_i4b:i)  = arth( d0, -one, i )
            if ( n/=2_i4b*i ) then
                d(i+1_i4b)          = zero
                d(n:i+2_i4b:-1_i4b) = d(1_i4b:i)
            else
                d(n:i+1_i4b:-1_i4b) = d(1_i4b:i)
            end if
!       
            d(:n) = d(:n) + two
!
            e(1_i4b:n-1_i4b) = one
!        
        case( 53_i4b )
!
!           CASE 53: GENERATE A BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR OF
!                    A GLUED WILKINSON DEFINITE-POSITIVE TRIDIAGONAL MATRIX FROM
!                    EXAMPLE 7.2 OF GREGORY AND KARNEY. P. 135.
!         
            d0  = real( l, stnd )
!
            temp(1_i4b:l)                      = arth( d0, -one, l )
            temp(l+1_i4b)                      = zero
            temp(2_i4b*l+1_i4b:l+2_i4b:-1_i4b) = temp(1_i4b:l)
!       
            temp(1_i4b:2_i4b*l+1_i4b) = temp(1_i4b:2_i4b*l+1_i4b) + two
!
            temp2(1_i4b:2_i4b*l) = one
            temp2(2_i4b*l+1_i4b) = glu     
!
            do i=0_i4b, n-1_i4b, 2_i4b*l+1_i4b
!
                i1 = i + 1_i4b
                i2 = min( i + 2_i4b*l+1_i4b, n)
                j  = i2 - i1 + 1_i4b
!
                d(i1:i2) = temp(1_i4b:j)
                e(i1:i2) = temp2(1_i4b:j)
!
                if ( i/=0_i4b ) then
                    d(i)  = d(i)  + glu
                    d(i1) = d(i1) + glu
                end if
!
            end do
!        
        case( 54_i4b )
!
!           CASE 54: GENERATE A BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR OF
!                    A WILKINSON DEFINITE-POSITIVE TRIDIAGONAL MATRIX.
!
            i = n/2_i4b
            d0 = real( i, stnd )
!
            d(1_i4b:i)  = arth( d0, -one, i )
            if ( n/=2_i4b*i ) then
                d(i+1_i4b)          = zero
                d(n:i+2_i4b:-1_i4b) = -d(1_i4b:i)
            else
                d(n:i+1_i4b:-1_i4b) = -d(1_i4b:i)
            end if
!        
            d(:n) = d(:n) + real( i+1_i4b, stnd )
!
            e(1_i4b:n-1_i4b) = one
!        
        case( 55_i4b )
!
!           CASE 55: GENERATE A BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR OF
!                    A GLUED DEFINITE-POSITIVE WILKINSON TRIDIAGONAL MATRIX
!                    FROM EXAMPLE 7.3 OF GREGORY AND KARNEY. P. 136.
!         
            d0  = real( l, stnd )
!
            temp(1_i4b:l)                      = arth( d0, -one, l )
            temp(l+1_i4b)                      = zero
            temp(2_i4b*l+1_i4b:l+2_i4b:-1_i4b) = -temp(1_i4b:l)
!       
            temp(1_i4b:2_i4b*l+1_i4b) = temp(1_i4b:2_i4b*l+1_i4b) + real( l+1_i4b, stnd )
!
            temp2(1_i4b:2_i4b*l) = one
            temp2(2_i4b*l+1_i4b) = glu        
!
            do i=0_i4b, n-1_i4b, 2_i4b*l+1_i4b
!
                i1 = i + 1_i4b
                i2 = min( i + 2_i4b*l+1_i4b, n)
                j  = i2 - i1 + 1_i4b
!
                d(i1:i2) = temp(1_i4b:j)
                e(i1:i2) = temp2(1_i4b:j)
!
                if ( i/=0_i4b ) then
                    d(i)  = d(i)  + glu
                    d(i1) = d(i1) + glu
                end if
!
            end do
!
        case( 56_i4b )
!
!           CASE 56: GENERATE A BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR OF
!           A POSITIVE-DEFINITE TRIDIAGONAL MATRIX COMPUTED FROM LAGUERRE RECURRENCE.
!
            do j = 1_i4b, n
                lambda = real( j, stnd)
                e(j)   = lambda + one
                lambda = two*lambda + one
                d(j)   = lambda
            end do
!
        case default
!
!           DEFAULT CASE: GENERATE A RANDOM BIDIAGONAL MATRIX.
!
            call random_number_( d(:n) )
!
            call random_number_( e(2_i4b:n) )
!
    end select
!
    if ( from_tridiag0 ) then
!
        if ( known_singval0 ) then
!
!           IF EIGENVALUES OF THE TRIDIAGONAL MATRIX ARE KNOWN, CHECK
!           IF THEY ARE STRICTLY POSITIVE.
!
            if ( any( x(:n)<=zero ) ) then
!
                call merror( name_proc//svd_error6 )
!
            end if
!
            if ( test_singval ) then
!
!               COMPUTE SINGULAR VALUES FROM EIGENVALUES IF REQUIRED.
!
                singval(:n) = sqrt( x(:n) )
!
            end if
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( x )
!
        end if
!
!       FORM THE UPPER BIDIAGONAL MATRIX FROM THE CHOLESKY FACTOR OF THE
!       TRIDIAGONAL POSITIVE-DEFINITE MATRIX IF NECESSARY.
!
        temp(1_i4b) = d(1_i4b)
!
        do l = 2_i4b, n
            temp(l) = d(l) - e(l-1_i4b)**2/temp(l-1_i4b)
            temp(l) = max( temp(l), safmin )
        end do
!
        d(:n) = sqrt( temp(:n) )
!
        e(2_i4b:n) = e(1_i4b:n-1_i4b)/d(:n-1_i4b)
!
    end if
!
    if ( test_singval ) then
!
!       SINGULAR VALUES ARE REQUESTED.
!
        if ( known_singval0 ) then
!
!           SORT THE SINGULAR VALUES IF REQUIRED.
!
            if ( present(sort) ) then
!
                call singval_sort( sort, singval(:n) )
!
            end if
!
        else
!
!           COMPUTE THE SINGULAR VALUES OF THE BIDIAGONAL MATRIX BY A BISECTION METHOD WITH HIGH ACCURACY.
!           THE SINGULAR VALUES ARE COMPUTED WITH MAXIMUM ACCURACY WHEN OPTIONAL PARAMETER abstol IS SET TO
!           sqrt( lamch('s') ), WHICH IS THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
            call bd_singval( d(:n), e(:n), i, temp(:n), failure=fail, sort=sort, abstol=abstol )
!
            singval(:n) = temp(:n)
!
            if ( present(failure) ) then
!
                failure = fail
!
            end if
!
        end if
!
    end if
!
    e(1_i4b) = zero
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( temp, temp2 )
!
!
! END OF SUBROUTINE gen_bd_mat
! ____________________________
!
    end subroutine gen_bd_mat
!
!
! =========================================================================================
!                                 PRIVATE SUBROUTINES
! =========================================================================================
!
!
    subroutine bd_lasq2( n, maxiter, ieee, z, info )
!
! Purpose
! _______
!
!   BD_LASQ2 computes all the eigenvalues of the N-by-N symmetric positive
!   definite tridiagonal matrix associated with the qd array Z to high
!   relative accuracy, in the absence of denormalization, underflow and overflow.
!
!   To see the relation of Z to the tridiagonal matrix, let L be a
!   unit lower bidiagonal matrix with subdiagonals Z(2,4,6,,..) and
!   let U be an upper bidiagonal matrix with 1's above and diagonal
!   Z(1,3,5,,..). The tridiagonal is L*U or, if you prefer, the
!   symmetric tridiagonal to which it is similar.
!
!
! Arguments
! _________
!
!   N        (INPUT) integer(i4b)
!            The number of rows and columns in the tridiagonal matrix.
!            It is assumed that N>2, but this is not checked in the subroutine.
!            This assumption is automatically verified when BD_LASQ2 is called from
!            BD_LASQ1 subroutine.
!
!   MAXITER  (INPUT) integer(i4b)
!            MAXITER controls the maximum number of iterations in the inner loop
!            of the dqds algorithm. MAXITER must be greater than 10,
!            otherwise the default value is used.
!
!            The default is 30.
!
!   IEEE     (INPUT) logical(lgl)
!            On entry, logical flag for IEEE or non IEEE arithmetic.
!            When IEEE is set to true, the division in the dqds loop is not protected from
!            incurring an exception (division by zero or overflow) since the powerful feature of
!            arithmetic units conforming to IEEE floating point standard 754 is that computation
!            is not held up by an exception. At the end of the loop, the code tests whether an
!            infinite or an NaN (not a number) occurred and acts appropriately.
!
!   Z        (INPUT/OUTPUT) real(stnd), dimension(4*n)
!            On entry Z(:2*N-1) holds the qd array. The other part of Z can be arbitrary.
!            On normal exit, entries 1 to N hold the eigenvalues (e.g. if INFO is equal to 0).
!            The eigenvalues are not sorted.      
!            On abnormal exit, if INFO is equal to 2, Z(:2*N-1) holds a qd array with the same
!            eigenvalues as the given Z(:2*N-1).
!
!            It is assumed that all the elements of Z(:2*N-1) are greater or equal to zero on
!            input, but this is not checked in the subroutine.
!            This assumption is automatically verified when BD_LASQ2 is called from BD_LASQ1 subroutine.
!
!   INFO     (OUTPUT) integer(i4b)
!            On exit:
!
!            - INFO = 0 :  indicates successful exit;
!            - INFO = 1 :  indicates a failure because a split was marked by a positive
!              value in the subdiagonal of the given tridiagonal matrix;
!            - INFO = 2 :  indicates that the algorithm did not converge
!              and that the maximum number of iterations in the inner loop
!              of the dqds algorithm is exceeded.
!              On exit, Z(:2*N-1) holds a qd array with the same eigenvalues as the
!              given Z(:2*N-1);
!            - INFO = 3 :  indicates that the termination criterion of the outer while
!              loop of the algorithm was not met (e.g. the program created more than
!              N unreduced blocks).
!            - INFO = 4 :  indicates that the maximum number of iterations in the inner
!              loop of the dqds algorithm is exceeded and that a qd array with
!              the same eigenvalues as the original one cannot be recovered.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the DLASQ2 subroutine in LAPACK version 3.12.0.
!
!   Local Variables: I0:N0 defines a current unreduced segment of Z.
!   The shifts are accumulated in SIGMA.
!   Ping-pong is controlled by PP (alternates between 0 and 1).
!
!   This subroutine is part of the dqds algorithm described
!   in the references (2) and (3).
!
!   For further details on the dqds algorithm, see:
!
!   (1) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (2) Parlett, B.N., and Marques, O.A., 2000:
!           An implementation of the dqds algorithm (positive case).
!           Linear Algebra Appl., Volume 309, pp. 217-259.
!
!   (3) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: n, maxiter
    integer(i4b), intent(out) :: info
!
    real(stnd),   intent(inout), dimension(4*n) :: z
!
    logical(lgl), intent(in) :: ieee
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: d, dee, deemin, desig, dmn, dmn1, dmn2, dn,  &
                  dn1, dn2, emax, emin, g, oldemn, qmax, qmin, &
                  sigma, temp, tol, tol2, tempe, tempq
!
    integer(i4b) :: maxiter2, i0, i1, i4, ipn4, iwhila, iwhilb,   &
                    k, kmin, n0, n1, nbig, pp, splt, ttype
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   INPUT ARGUMENTS ARE NOT TESTED AS IT IS ASSUMED THAT bd_lasq2 IS CALLED FROM bd_lasq1,
!   IN WHICH CASE n>2 .
!
    info = 0_i4b
!
!   COMPUTE TOLERANCES.
!
    tol  = ulp*fudge_lasq
    tol2 = tol**2
!
    maxiter2 = merge( maxiter, 30_i4b, maxiter>=10_i4b )
!
    z(2_i4b*n) = zero
!
!   REARRANGE DATA FOR LOCALITY: Z=(Q1,QQ1,E1,EE1,Q2,QQ2,E2,EE2,...).
!
    do k=2_i4b*n, 2_i4b, -2_i4b
!
        z(2_i4b*k)       = zero
        z(2_i4b*k-1_i4b) = z(k)
        z(2_i4b*k-2_i4b) = zero
        z(2_i4b*k-3_i4b) = z(k-1_i4b)
!
    end do
!
    i0 = 1_i4b
    n0 = n
!
!   REVERSE THE QD-ARRAY, IF WARRANTED.
!
    if ( c1_5*z( 4_i4b*i0-3_i4b )<z( 4*n0-3_i4b ) ) then
!
        ipn4 = 4_i4b*( i0+n0 )
!
        do i4=4_i4b*i0, 2_i4b*( i0+n0-1_i4b ), 4_i4b
!
            temp = z(i4-3_i4b)
            z(i4-3_i4b) = z(ipn4-i4-3_i4b)
            z(ipn4-i4-3_i4b) = temp
!
            temp = z(i4-1_i4b)
            z(i4-1_i4b) = z(ipn4-i4-5_i4b)
            z(ipn4-i4-5_i4b) = temp
!
        end do
!
    end if
!
!   INITIAL SPLIT CHECKING VIA DQD AND LI'S TEST.
!
    pp = 0_i4b
!
    do k = 1_i4b, 2_i4b
!
        d = z(4_i4b*n0+pp-3_i4b)
!
        do i4 = 4_i4b*( n0-1_i4b ) + pp, 4_i4b*i0 + pp, -4_i4b
!
            if ( z(i4-1_i4b)<=tol2*d ) then
                z(i4-1_i4b) = -zero
                d = z(i4-3_i4b)
            else
                d = z(i4-3_i4b)*( d/( d + z(i4-1_i4b) ) )
            end if
!
        end do
!
!       DQD MAPS Z TO ZZ PLUS LI'S TEST.
!
        d = z(4_i4b*i0+pp-3_i4b)
!
        do i4 = 4_i4b*i0 + pp, 4_i4b*( n0-1_i4b ) + pp, 4_i4b
!
            z(i4-2_i4b*pp-2_i4b) = d + z(i4-1_i4b)
!
            if ( z(i4-1_i4b)<=tol2*d ) then
               z(i4-1_i4b) = -zero
               z(i4-2_i4b*pp-2_i4b) = d
               z(i4-2_i4b*pp) = zero
               d = z(i4+1_i4b)
            else if ( safmin*z(i4+1_i4b)<z(i4-2_i4b*pp-2_i4b) .and. safmin*z(i4-2_i4b*pp-2_i4b)<z(i4+1_i4b) ) then
               temp = z(i4+1_i4b)/z(i4-2_i4b*pp-2_i4b)
               z(i4-2_i4b*pp) = z(i4-1_i4b)*temp
               d = d*temp
            else
               z(i4-2_i4b*pp) = z(i4+1_i4b)*(z( i4-1_i4b)/z(i4-2_i4b*pp-2_i4b) )
               d = z(i4+1_i4b)*( d/z(i4-2_i4b*pp-2_i4b) )
            end if
!
        end do
!
        z(4_i4b*n0-pp-2_i4b) = d
!
!       PREPARE FOR THE NEXT ITERATION ON k.
!
        pp = 1_i4b - pp
!
   end do
!
!  INITIALISE VARIABLES TO PASS TO bd_lasq3.
!
    ttype = 0_i4b
    dmn1  = zero
    dmn2  = zero
    dn    = zero
    dn1   = zero
    dn2   = zero
    g     = zero
!
    do iwhila=1_i4b, n + 1_i4b
!
!       CHECK IF ALL EIGENVALUES HAVE BEEN FOUND.
!
        if ( n0<1_i4b )  exit
!
!       WHILE ARRAY UNFINISHED DO.
!
!       e(n0) HOLDS THE VALUE OF sigma WHEN SUBMATRIX IN i0:n0
!       SPLITS FROM THE REST OF THE ARRAY, BUT IS NEGATED.
!
        desig = zero
!
        if ( n0==n ) then
            sigma = zero
        else
            sigma = -z(4_i4b*n0-1_i4b)
        end if
!
        if ( sigma<zero ) then
!
            info = 1_i4b
!
            exit
!
        end if
!
!       FIND LAST UNREDUCED SUBMATRIX'S TOP INDEX i0, FIND qmax AND
!       emin. FIND GERSHGORIN-TYPE BOUND IF Q'S MUCH GREATER THAN E'S.
!
        emax = zero
        if ( n0>i0 ) then
            emin = abs( z(4_i4b*n0-5_i4b) )
        else
            emin = zero
        end if
!
        qmin = z(4_i4b*n0-3_i4b)
        qmax = qmin
!
        do i4=4_i4b*n0, 8_i4b, -4_i4b
!
            if ( z(i4-5_i4b)<=zero ) exit
!
            if( qmin>=four*emax ) then
               qmin = min( qmin, z(i4-3_i4b) )
               emax = max( emax, z(i4-5_i4b) )
            end if
!
            qmax = max( qmax, z(i4-7_i4b)+z(i4-5_i4b) )
            emin = min( emin, z(i4-5_i4b) )
!
        end do
!
        if ( i4<8_i4b ) then
            i4 = 4_i4b
        end if
!
        i0 = i4/4_i4b
        pp = 0_i4b
!
        if ( n0-i0>1_i4b ) then
!
            dee = z(4_i4b*i0-3_i4b)
            deemin = dee
            kmin = i0
!
            do i4=4_i4b*i0+1_i4b, 4_i4b*n0-3_i4b, 4_i4b
!
                dee = z(i4)*( dee/( dee + z(i4-2_i4b) ) )
!
                if ( dee<=deemin ) then
                    deemin = dee
                    kmin = ( i4 + 3_i4b )/4_i4b
                end if
!
            end do 
!
            if ( (kmin-i0)*2_i4b<n0-kmin .and. deemin<=half*z(4_i4b*n0-3_i4b) ) then
!
                ipn4 = 4_i4b*( i0 + n0 )
                pp = 2_i4b
!
                do i4=4_i4b*i0, 2_i4b*( i0+n0-1_i4b ), 4_i4b
!
                    temp = z(i4-3_i4b)
                    z(i4-3_i4b) = z(ipn4-i4-3_i4b)
                    z(ipn4-i4-3_i4b) = temp
!
                    temp = z(i4-2_i4b)
                    z(i4-2_i4b) = z(ipn4-i4-2_i4b)
                    z(ipn4-i4-2_i4b) = temp
!
                    temp = z(i4-1_i4b)
                    z(i4-1_i4b) = z(ipn4-i4-5_i4b)
                    z(ipn4-i4-5_i4b) = temp
!
                    temp = z(i4)
                    z(i4) = z(ipn4-i4-4_i4b)
                    z(ipn4-i4-4_i4b) = temp
!
                end do
!
            end if
!
        end if
!
!       PUT -(INITIAL SHIFT) INTO dmn.
!
        dmn = -max( zero, qmin-two*sqrt( qmin )*sqrt( emax ) )
!
!       NOW i0:n0 IS UNREDUCED.
!       pp = 0 FOR PING, pp = 1 FOR PONG.
!       pp = 2 INDICATES THAT FLIPPING WAS APPLIED TO THE Z ARRAY AND
!              AND THAT THE TESTS FOR DEFLATION UPON ENTRY IN bd_lasq3
!              SHOULD NOT BE PERFORMED.
!
        nbig = maxiter2*( n0 - i0 + 1_i4b )
!
        do iwhilb = 1_i4b, nbig
!
            if ( i0>n0 ) exit
!
!           WHILE SUBMATRIX UNFINISHED TAKE A GOOD DQDS STEP.
!
            call bd_lasq3( i0, n0, z, pp, sigma, desig, qmax, ieee, &
                           ttype, dmn, dmn1, dmn2, dn, dn1, dn2, g  )
!
            pp = 1_i4b - pp
!
!           WHEN emin IS VERY SMALL CHECK FOR SPLITS.
!
            if ( pp==0_i4b .and. n0-i0>=3_i4b ) then
!
                if ( z(4_i4b*n0)<=tol2*qmax .or. z(4_i4b*n0-1_i4b)<=tol2*sigma ) then
!
                    splt = i0 - 1_i4b
                    qmax = z(4*i0-3_i4b)
                    emin = z(4*i0-1_i4b)
                    oldemn = z(4*i0)
!
                    do i4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
                        if ( z(i4)<=tol2*z(i4-3_i4b) .or. z(i4-1_i4b)<=tol2*sigma ) then
                            z(i4-1_i4b) = -sigma
                            splt = i4/4_i4b
                            qmax = zero
                            emin = z(i4+3_i4b)
                            oldemn = z(i4+4_i4b)
                        else
                            qmax = max( qmax, z(i4+1_i4b) )
                            emin = min( emin, z(i4-1_i4b) )
                            oldemn = min( oldemn, z(i4) )
                        end if
!
                    end do
!
                    z(4_i4b*n0-1_i4b) = emin
                    z(4_i4b*n0) = oldemn
!
                    i0 = splt + 1_i4b
!
                end if
!
            end if
!
        end do
!
        if ( iwhilb>nbig ) then
!
!           MAXIMUM NUMBER OF INNER ITERATIONS EXCEEDED.
!
            info = 2_i4b
!
            exit
!
        end if
!
    end do
!
!   POST-PROCESSING.
!
    if ( iwhila>n+1_i4b ) then
!
!       MAXIMUM NUMBER OF OUTER ITERATIONS EXCEEDED.
!
        info = 3_i4b
!
    else if ( info==0_i4b ) then
!
!       ALL EIGENVALUES FOUND, MOVE Q'S TO THE FRONT.
!
        do k=2_i4b, n
            z(k) = z(4_i4b*k-3_i4b)
        end do
!
    else if ( info==2_i4b ) then
!
!       MAXIMUM NUMBER OF INNER ITERATIONS EXCEEDED, RESTORE THE SHIFT
!       sigma AND PLACE THE NEW D'S AND E'S IN A QD ARRAY.
!       THIS MIGHT NEED TO BE DONE FOR SEVERAL BLOCKS.
!
        i1 = i0
        n1 = n0
!
       do
!
            tempq = z(4_i4b*i0-3_i4b)
            z(4_i4b*i0-3_i4b) = z(4_i4b*i0-3_i4b) + sigma
!
            do k=i0+1_i4b, n0
!
                tempe = z(4_i4b*k-5_i4b)
                z(4_i4b*k-5_i4b) = z(4_i4b*k-5_i4b)*( tempq/z(4_i4b*k-7_i4b) )
!
                tempq = z(4_i4b*k-3_i4b)
                z(4_i4b*k-3_i4b) = z(4_i4b*k-3_i4b) + sigma + tempe - z(4_i4b*k-5_i4b)
!
            end do
!
!           PREPARE TO DO THIS ON THE PREVIOUS BLOCK IF THERE IS ONE.
!
            if ( i1>1_i4b ) then
!
                n1 = i1 - 1_i4b
!
                i4 = i1
!
                do while( i1>=2_i4b .and. z(4_i4b*i1-5_i4b)>=zero )
                   i1 = i1 - 1_i4b
                end do
!
                if ( i1/=i4 ) then
!
                   sigma = -z(4_i4b*n1-1_i4b)
!
                   cycle
!
                end if
!
            end if
!
            exit
!
        end do
!
        do k=1_i4b, n
!
            z(2_i4b*k-1_i4b) = z(4_i4b*k-3_i4b)
!
!           ONLY THE BLOCK 1..n0 IS UNFINISHED.  THE REST OF THE E'S
!           MUST BE ESSENTIALLY ZERO, ALTHOUGH SOMETIMES OTHER DATA
!           HAS BEEN STORED IN THEM.
!
            if( k<n0 ) then
               z(2_i4b*k) = z(4_i4b*k-1_i4b)
            else
               z(2_i4b*k) = 0_i4b
            end if
!
        end do
!
!       CHECK IF ANY ELEMENT IN THE NEW QD ARRAY IS NEGATIVE. 
!      
        if ( any( z(1_i4b:2_i4b*n)<zero ) ) then
!      
            info = 4_i4b
!                    
        end if
!
    end if
!
!
! END OF SUBROUTINE bd_lasq2
! __________________________
!
    end subroutine bd_lasq2
!
! ===========================================================================================
!
    subroutine bd_lasq3( i0, n0, z, pp, sigma, desig, qmax, ieee,   &
                         ttype, dmin, dmin1, dmin2, dn, dn1, dn2, g )
!
! Purpose
! _______
!
!   BD_LASQ3 checks for deflation, computes a shift (tau) and calls dqds.
!   In case of failure it changes shifts, and tries again until output
!   is positive.
!
!
! Arguments
! _________
!
!
!   I0      (INPUT) integer(i4b)
!           On entry, I0 is the first index.
!
!   N0      (INPUT/OUTPUT) integer(i4b)
!           On entry, N0 is the last index.
!           On exit, N0 is upddated with the new last index of the segment.
!
!   Z       (INPUT/OUTPUT) real(stnd), dimension(4*n)
!           Z holds the qd array of size 4*n where n is the dimension of the
!           associated symmetric positive definite tridiagonal matrix.
!
!   PP      (INPUT/OUTPUT) integer(i4b)
!           PP=0 for ping, PP=1 for pong.
!           PP=2 indicates that flipping was applied to the Z array
!           and that the initial tests for deflation should not be
!           performed.
!
!   SIGMA   (INPUT/OUTPUT) real(stnd)
!           SIGMA is the sum of shifts used in current segment.
!
!   DESIG   (INPUT/OUTPUT) real(stnd)
!           DESIG is the lower order part of SIGMA.
!
!   QMAX    (INPUT/OUTPUT) real(stnd)
!           QMAX is the maximum value of q.
!
!   IEEE    (INPUT) logical(lgl)
!           On entry, logical flag for IEEE or non IEEE arithmetic (IEEE is passed to BD_LASQ5).
!
!   TTYPE   (INPUT/OUTPUT) integer(i4b)
!           On entry, the preceding shift type.
!           On exit, the new shift type.
!
!   DMIN    (INPUT/OUTPUT) real(stnd)
!           Minimum value of D.
!
!   DMIN1   (INPUT/OUTPUT) real(stnd)
!           Minimum value of D, excluding D(N0).
!
!   DMIN2   (INPUT/OUTPUT) real(stnd)
!           Minimum value of D, excluding D(N0) and D(N0-1).
!
!   DN      (INPUT/OUTPUT) real(stnd)
!           On entry, D(N0), the last value of D. 
!
!   DN1     (INPUT/OUTPUT) real(stnd)
!           On entry, D(N0-1).
!
!   DN2     (INPUT/OUTPUT) real(stnd)
!           On entry, D(N0-2).
!
!   G       (INPUT/OUTPUT) real(stnd)
!           G is passed as an argument in order to save its value between
!           calls to BD_LASQ3.
!
!   SIGMA, DESIG, QMAX, DMIN, DMIN1, DMIN2, DN, DN1, DN2, G and TTYPE are passed as arguments in order
!   to save their values between calls to BD_LASQ3 (and also BD_LASQ4 and BD_LASQ5 inside of BD_LASQ3).
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the DLASQ3 subroutine from LAPACK version 3.12.0.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
    use Num_Constants, only : is_nan
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: qmax, sigma, desig, dmin, dmin1, dmin2, dn, dn1, dn2, g
    real(stnd), intent(inout), dimension(*) :: z
!
    integer(i4b), intent(in)    :: i0
    integer(i4b), intent(inout) :: n0, pp, ttype
!
    logical(lgl), intent(in) :: ieee
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tau, s, t, temp, tol, tol2
!
    integer(i4b) :: ipn4, j4, n0in, nn
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   COMPUTE TOLERANCES AND SAVE INITIAL VALUE OF n0 .
!
    n0in = n0
!
    tol = ulp*fudge_lasq
    tol2 = tol**2
!
!   CHECK FOR DEFLATION.
!
    do
!
        if ( n0<=i0 ) then
!
            if ( n0==i0 ) then
!
!               1 EIGENVALUE FOUND.
!
                z( 4_i4b*n0-3_i4b ) = z( 4_i4b*n0+pp-3_i4b ) + sigma
!
                n0 = n0 - 1_i4b
!
            end if
!
            return
!
        end if
!
        nn = 4_i4b*n0 + pp
!
        if ( n0/=i0+1_i4b ) then
!
!           CHECK WHETHER e(n0-1) IS NEGLIGIBLE, 1 EIGENVALUE FOUND.
!
            if ( z( nn-5_i4b )<=tol2*( sigma+z( nn-3_i4b ) ) .or. z( nn-2_i4b*pp-4_i4b )<=tol2*z( nn-7_i4b ) ) then
!
                z( 4_i4b*n0-3_i4b ) = z( 4_i4b*n0+pp-3_i4b ) + sigma
!
                n0 = n0 - 1_i4b
!
                cycle
!
            end if
!
!           CHECK  WHETHER e(n0-2) IS NEGLIGIBLE.
!
            if ( z( nn-9_i4b )>tol2*sigma .and. z( nn-2_i4b*pp-8_i4b )>tol2*z( nn-11_i4b ) ) exit
!
        end if
!
!       2 EIGENVALUES FOUND.
!
        if ( z( nn-3_i4b )>z( nn-7_i4b ) ) then
!
            s = z( nn-3_i4b )
            z( nn-3_i4b ) = z( nn-7_i4b )
            z( nn-7_i4b ) = s
!
        end if
!
        t = half*( ( z( nn-7_i4b ) - z( nn-3_i4b ) ) + z( nn-5_i4b ) )
!
        if ( z( nn-5_i4b )>z( nn-3_i4b )*tol2 .and. t/=zero ) then
!
            s = z( nn-3_i4b )*( z( nn-5_i4b )/t )
!
            if ( s<=t ) then
                s = z( nn-3_i4b )*( z( nn-5_i4b )/( t*( one+sqrt( one+s/t ) ) ) )
            else
                s = z( nn-3_i4b )*( z( nn-5_i4b )/( t+sqrt( t )*sqrt( t+s ) ) )
            end if
!
            t = z( nn-7_i4b ) + ( s+z( nn-5_i4b ) )
            z( nn-3_i4b ) = z( nn-3_i4b )*( z( nn-7_i4b )/t )
            z( nn-7_i4b ) = t
!
        end if
!
        z( 4_i4b*n0-7_i4b ) = z( nn-7_i4b ) + sigma
        z( 4_i4b*n0-3_i4b ) = z( nn-3_i4b ) + sigma
!
        n0 = n0 - 2_i4b
!
    end do
!
    if ( pp==2_i4b ) pp = 0_i4b
!
!   REVERSE THE QD-ARRAY, IF WARRANTED.
!
    if ( dmin<=zero .or. n0<n0in ) then
!
        if ( c1_5*z( 4_i4b*i0+pp-3_i4b )<z( 4_i4b*n0+pp-3_i4b ) ) then
!
            ipn4 = 4_i4b*( i0+n0 )
!
            do j4=4_i4b*i0, 2_i4b*( i0+n0-1_i4b ), 4_i4b
!
               temp = z( j4-3_i4b )
               z( j4-3_i4b ) = z( ipn4-j4-3_i4b )
               z( ipn4-j4-3_i4b ) = temp
!
               temp = z( j4-2_i4b )
               z( j4-2_i4b ) = z( ipn4-j4-2_i4b )
               z( ipn4-j4-2_i4b ) = temp
!
               temp = z( j4-1_i4b )
               z( j4-1_i4b ) = z( ipn4-j4-5_i4b )
               z( ipn4-j4-5_i4b ) = temp
!
               temp = z( j4 )
               z( j4 ) = z( ipn4-j4-4_i4b )
               z( ipn4-j4-4_i4b ) = temp
!
            end do
!
            if ( n0-i0<=4_i4b ) then
                z( 4_i4b*n0+pp-1_i4b ) = z( 4_i4b*i0+pp-1_i4b )
                z( 4_i4b*n0-pp ) = z( 4_i4b*i0-pp )
            end if
!
            dmin2 = min( dmin2, z( 4*n0+pp-1 ) )
!
            z( 4_i4b*n0+pp-1_i4b ) = min( z( 4_i4b*n0+pp-1_i4b ), z( 4_i4b*i0+pp-1_i4b ), z( 4_i4b*i0+pp+3_i4b ) )
            z( 4_i4b*n0-pp )       = min( z( 4_i4b*n0-pp ), z( 4_i4b*i0-pp ), z( 4_i4b*i0-pp+4_i4b ) )
!
            qmax = max( qmax, z( 4*i0+pp-3_i4b ), z( 4*i0+pp+1 ) )
            dmin = -zero
!
        end if
!
    end if
!
!   CHOOSE A SHIFT.
!
    call bd_lasq4( i0, n0, z, pp, n0in, dmin, dmin1, dmin2, dn, dn1, dn2, tau, ttype, g )
!
!   CALL DQDS UNTIL dmin > 0.
!
    do
!
        call bd_lasq5( i0, n0, z, pp, tau, sigma, dmin, dmin1, dmin2, dn, dn1, dn2, ieee )
!
!       CHECK STATUS.
!
        if ( dmin>=zero .and. dmin1>=zero ) then
!
!           SUCCESS.
!
            exit
!
        else if ( dmin<zero .and. dmin1>zero                   .and. &
                  z( 4_i4b*( n0-1_i4b )-pp )<tol*( sigma+dn1 ) .and. &
                  abs( dn )<tol*sigma ) then
!
!           CONVERGENCE HIDDEN BY NEGATIVE dn.
!
            z( 4_i4b*( n0-1_i4b )-pp+2_i4b ) = zero
            dmin = zero
!
            exit
!
        else if ( dmin<zero ) then
!
!           tau TOO BIG. SELECT NEW tau AND TRY AGAIN.
!
            if ( ttype<-22_i4b ) then
!
!               FAILED TWICE. PLAY IT SAFE.
!
                tau = zero
!
            else if ( dmin1>zero ) then
!
!               LATE FAILURE. GIVES EXCELLENT SHIFT.
!
                tau = ( tau + dmin )*( one - two*ulp )
                ttype = ttype - 11_i4b
!
            else
!
!               EARLY FAILURE. DIVIDE BY 4.
!
                tau = quarter*tau
                ttype = ttype - 12_i4b
!
            end if
!
            cycle
!
        else if ( is_nan( dmin ) .and. tau/=zero ) then
!
!           NaN.
!
            tau = zero
!
            cycle
!
        end if
!
!       POSSIBLE UNDERFLOW. PLAY IT SAFE.
!
        call bd_lasq6( i0, n0, z, pp, dmin, dmin1, dmin2, dn, dn1, dn2 )
!
        tau = zero
!
    end do
!
    if ( tau<sigma ) then
        desig = desig + tau
        t = sigma + desig
        desig = desig - ( t - sigma )
    else
        t = sigma + tau
        desig = sigma - ( t - tau ) + desig
    end if
!
    sigma = t
!
!
! END OF SUBROUTINE bd_lasq3
! __________________________
!
    end subroutine bd_lasq3
!
! ===========================================================================================
!
    subroutine bd_lasq4( i0, n0, z, pp, n0in, dmin, dmin1, dmin2, dn, dn1, dn2, tau, ttype, g )
!
! Purpose
! _______
!
!   BD_LASQ4 computes an approximation TAU to the smallest eigenvalue
!   using values of D from the previous dqds transform.
!
!
! Arguments
! _________
!
!
!   I0      (INPUT) integer(i4b)
!           On entry, I0 is the first index.
!
!   N0      (INPUT) integer(i4b)
!           On entry, N0 is the last index.
!
!   Z       (INPUT) real(stnd), dimension(4*n)
!           Z holds the qd array of size 4*n where n is the dimension of the
!           associated symmetric positive definite tridiagonal matrix.
!
!   PP      (INPUT) integer(i4b)
!           On entry, PP=0 for ping, PP=1 for pong.
!
!   N0IN    (INPUT) integer(i4b)
!           On entry, the value of N0 at start of BD_LASQ3.
!
!   DMIN    (INPUT) real(stnd)
!           On entry, minimum value of D.
!
!   DMIN1   (INPUT) real(stnd)
!           On entry, minimum value of D, excluding D(N0).
!
!   DMIN2   (INPUT) real(stnd)
!           On entry, minimum value of D, excluding D(N0) and D(N0-1).
!
!   DN      (INPUT) real(stnd)
!           On entry, D(N0), the last value of D. 
!
!   DN1     (INPUT) real(stnd)
!           On entry, D(N0-1).
!
!   DN2     (INPUT) real(stnd)
!           On entry, D(N0-2).
!
!   TAU     (OUTPUT) real(stnd)
!           On exit, TAU is the shift to be used.
!
!   TTYPE   (INPUT/OUTPUT) integer(i4b)
!           On entry, the preceding shift type.
!           On exit, the new shift type.
!
!   G       (INPUT/OUTPUT) real(stnd)
!           G is passed as an argument in order to save its value between
!           calls to BD_LASQ4.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the DLASQ4 subroutine from LAPACK version 3.12.0.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
    use Select_Parameters, only : i4b, stnd
    use Reals_Constants,   only : zero, quarter, third, half, one, two, c100
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: dmin, dmin1, dmin2, dn, dn1, dn2
    real(stnd), intent(in), dimension(*) :: z
    real(stnd), intent(inout) :: g
    real(stnd), intent(out)   :: tau
!
    integer(i4b), intent(in)    :: i0, n0, pp, n0in
    integer(i4b), intent(inout) :: ttype
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: a2, b1, b2, gam, gap1, gap2, s
!
    integer(i4b) :: i4, nn, np
!
!
! PARAMETERS
! __________
!
!   cnst1 = 9/16 .
!
    real(stnd),  parameter :: cnst1=0.563_stnd, cnst2=1.01_stnd, cnst3=1.05_stnd
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( dmin<=zero ) then
!
!       A NEGATIVE dmin FORCES THE SHIFT TO TAKE THAT ABSOLUTE VALUE.
!
        tau = -dmin
        ttype = -1_i4b
!
        return
!
    end if
!
    nn = 4_i4b*n0 + pp
!
    if ( n0in==n0 ) then
!
!       NO EIGENVALUES DEFLATED.
!
        if ( dmin==dn .or. dmin==dn1 ) then
!
            b1 = sqrt( z( nn-3_i4b ) )*sqrt( z( nn-5_i4b ) )
            b2 = sqrt( z( nn-7_i4b ) )*sqrt( z( nn-9_i4b ) )
            a2 = z( nn-7_i4b ) + z( nn-5_i4b )
!
            if ( dmin==dn .and. dmin1==dn1 ) then
!
!               CASES 2 AND 3.
!
                gap2 = dmin2 - a2 - dmin2*quarter
!
                if ( gap2>zero .and. gap2>b2 ) then
                    gap1 = a2 - dn - ( b2/gap2 )*b2
                else
                    gap1 = a2 - dn - ( b1+b2 )
                end if
!
                if ( gap1>zero .and. gap1>b1 ) then
                    s = max( dn-( b1/gap1 )*b1, half*dmin )
                    ttype = -2_i4b
                else
                    s = zero
                    if ( dn>b1 ) s = dn - b1
                    if ( a2>( b1+b2 ) ) s = min( s, a2-( b1+b2 ) )
                    s = max( s, third*dmin )
                    ttype = -3_i4b
                end if
!
            else
!
!               CASE 4.
!
                ttype = -4_i4b
                s = quarter*dmin
!
                if ( dmin==dn ) then
                    gam = dn
                    a2 = zero
!
                    if ( z( nn-5_i4b ) > z( nn-7_i4b ) ) return
!
                    b2 = z( nn-5_i4b )/z( nn-7_i4b )
                    np = nn - 9_i4b
               else
                    np = nn - 2_i4b*pp
                    gam = dn1
!
                    if ( z( np-4_i4b ) > z( np-2_i4b ) ) return
!
                    a2 = z( np-4_i4b )/z( np-2_i4b )
!
                    if ( z( nn-9_i4b ) > z( nn-11_i4b ) ) return
!
                    b2 = z( nn-9_i4b )/z( nn-11_i4b )
                    np = nn - 13_i4b
               end if
!
!              APPROXIMATE CONTRIBUTION TO NORM SQUARED FROM i < nn-1.
!
               a2 = a2 + b2
!
               do i4=np, 4_i4b*i0 - 1_i4b + pp, -4_i4b
!
                   if ( b2==zero ) exit
!
                   b1 = b2
!
                   if ( z( i4 ) > z( i4-2_i4b ) ) return
!
                   b2 = b2*( z( i4 )/z( i4-2_i4b ) )
                   a2 = a2 + b2
!
                   if ( c100*max( b2, b1 )<a2 .or. cnst1<a2 ) exit
!
               end do
!
               a2 = cnst3*a2
!
!              RAYLEIGH QUOTIENT RESIDUAL BOUND.
!
               if ( a2<cnst1 ) s = gam*( one-sqrt( a2 ) )/( one+a2 )
!
            end if
!
        else if ( dmin==dn2 ) then
!
!           CASE 5.
!
            ttype = -5_i4b
            s = quarter*dmin
!
!           COMPUTE CONTRIBUTION TO NORM SQUARED FROM i > nn-2.
!
            np = nn - 2_i4b*pp
            b1 = z( np-2_i4b )
            b2 = z( np-6_i4b )
            gam = dn2
!
            if ( z( np-8_i4b )>b2 .or. z( np-4_i4b )>b1 ) return
!
            a2 = ( z( np-8_i4b )/b2 )*( one + z( np-4_i4b )/b1 )
!
!           APPROXIMATE CONTRIBUTION TO NORM SQUARED FROM i < nn-2. 
!
            if ( n0-i0>2_i4b ) then
!
                b2 = z( nn-13_i4b )/z( nn-15_i4b )
                a2 = a2 + b2
!
                do i4=nn - 17_i4b, 4_i4b*i0 - 1_i4b + pp, -4_i4b
!
                    if ( b2==zero ) exit
!
                    b1 = b2
!
                    if ( z( i4 )>z( i4-2_i4b ) ) return
!
                    b2 = b2*( z( i4 )/z( i4-2_i4b ) )
                    a2 = a2 + b2
!
                    if ( c100*max( b2, b1 )<a2 .or. cnst1<a2 ) exit
!
                end do
!
                a2 = cnst3*a2
!
            end if
!
            if ( a2<cnst1 ) s = gam*( one - sqrt( a2 ) )/( one + a2 )
!
        else
!
!           CASE 6, NO INFORMATION TO GUIDE US.
!
            if ( ttype==-6_i4b ) then
                g = g + third*( one - g )
            else if ( ttype==-18_i4b ) then
                g = quarter*third
            else
                g = quarter
            end if
!
            s = g*dmin
            ttype = -6_i4b
!
        end if
!
    else if ( n0in==n0+1_i4b ) then
!
!       ONE EIGENVALUE JUST DEFLATED. USE dmin1, dn1 FOR dmin AND dn.
!
        if ( dmin1==dn1 .and. dmin2==dn2 ) then
!
!           CASES 7 AND 8.
!
            ttype = -7_i4b
            s = third*dmin1
!
            if ( z( nn-5_i4b )>z( nn-7_i4b ) ) return
!
            b1 = z( nn-5_i4b )/z( nn-7_i4b )
            b2 = b1
!
            if ( b2/=zero ) then
!
                do i4=4_i4b*n0 - 9_i4b + pp, 4_i4b*i0 - 1_i4b + pp, -4_i4b
!
                    a2 = b1
!
                    if ( z( i4 )>z( i4-2_i4b ) ) return
!
                    b1 = b1*( z( i4 )/z( i4-2_i4b ) )
                    b2 = b2 + b1
!
                    if ( c100*max( b1, a2 )<b2 ) exit
!
                end do
!
            end if
!
            b2 = sqrt( cnst3*b2 )
            a2 = dmin1/( one + b2**2 )
            gap2 = half*dmin2 - a2
!
            if ( gap2>zero .and. gap2>b2*a2 ) then
                s = max( s, a2*( one-cnst2*a2*( b2/gap2 )*b2 ) )
            else
                s = max( s, a2*( one - cnst2*b2 ) )
                ttype = -8_i4b
            end if
!
        else
!
!           CASE 9.
!
            s = quarter*dmin1
            if ( dmin1==dn1 ) s = half*dmin1
            ttype = -9_i4b
!
        end if
!
    else if ( n0in==n0+2_i4b ) then
!
!       TWO EIGENVALUES DEFLATED. USE dmin2, dn2 FOR dmin AND dn.
!
!       CASES 10 AND 11.
!
        if ( dmin2==dn2 .and. two*z( nn-5_i4b )<z( nn-7_i4b ) ) then
!
            ttype = -10_i4b
            s = third*dmin2
!
            if ( z( nn-5_i4b )>z( nn-7_i4b ) )  return
!
            b1 = z( nn-5_i4b )/z( nn-7_i4b )
            b2 = b1
!
            if ( b2/=zero ) then
!
                do i4=4_i4b*n0 - 9_i4b + pp, 4_i4b*i0 - 1_i4b + pp, -4_i4b
!
                    if ( z( i4 )>z( i4-2_i4b ) ) return
!
                    b1 = b1*( z( i4 )/z( i4-2_i4b ) )
                    b2 = b2 + b1
!
                    if ( c100*b1<b2 ) exit
!
                end do
!
            end if
!
            b2 = sqrt( cnst3*b2 )
            a2 = dmin2/( one + b2**2 )
            gap2 = z( nn-7_i4b ) + z( nn-9_i4b ) - sqrt( z( nn-11_i4b ) )*sqrt( z( nn-9_i4b ) ) - a2
!
            if ( gap2>zero .and. gap2>b2*a2 ) then
                s = max( s, a2*( one-cnst2*a2*( b2/gap2 )*b2 ) )
            else
                s = max( s, a2*( one-cnst2*b2 ) )
            end if
!
        else
!
            s = quarter*dmin2
            ttype = -11_i4b
!
        end if
!
    else if ( n0in>( n0+2_i4b ) ) then
!
!       CASE 12, MORE THAN TWO EIGENVALUES DEFLATED. NO INFORMATION.
!
        s = zero
        ttype = -12_i4b
!
    end if
!
    tau = s
!
!
! END OF SUBROUTINE bd_lasq4
! __________________________
!
    end subroutine bd_lasq4
!
! ===========================================================================================
!
    subroutine bd_lasq5( i0, n0, z, pp, tau, sigma, dmin, dmin1, dmin2, dn, dn1, dn2, ieee )
!
! Purpose
! _______
!
!   BD_LASQ5 computes one dqds transform in ping-pong form, one
!   version for IEEE machines another for non IEEE machines.
!
!
! Arguments
! _________
!
!
!   I0      (INPUT) integer(i4b)
!           On entry, I0 is the first index.
!
!   N0      (INPUT) integer(i4b)
!           On entry, N0 is the last index.
!
!   Z       (INPUT/OUTPUT) real(stnd), dimension(4*n)
!           Z holds the qd array of size 4*n where n is the dimension of the
!           associated symmetric positive definite tridiagonal matrix.
!
!           EMIN is stored in Z(4*N0) to avoid an extra argument.
!
!   PP      (INPUT) integer(i4b)
!           On entry, PP=0 for ping, PP=1 for pong.
!
!   TAU     (INPUT/OUTPUT) real(stnd)
!           On entry, TAU is the shift to be used.
!           On exit, TAU is set to zero if it is small compared to
!           the accumulated shift used in this step.
!
!   SIGMA   (INPUT) real(stnd)
!           On entry, SIGMA is the accumulated shift up to this step.
!
!   DMIN    (OUTPUT) real(stnd)
!           On exit, minimum value of D.
!
!   DMIN1   (OUTPUT) real(stnd)
!           On exit, minimum value of D, excluding D(N0).
!
!   DMIN2   (OUTPUT) real(stnd)
!           On exit, minimum value of D, excluding D(N0) and D(N0-1).
!
!   DN      (OUTPUT) real(stnd)
!           On exit, D(N0), the last value of D. 
!
!   DN1     (OUTPUT) real(stnd)
!           On exit, D(N0-1).
!
!   DN2     (OUTPUT) real(stnd)
!           On exit, D(N0-2).
!
!   IEEE    (INPUT) logical(lgl)
!           On entry, logical flag for IEEE or non IEEE arithmetic.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the DLASQ5 subroutine from LAPACK version 3.12.0.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)    :: sigma
    real(stnd), intent(inout) :: tau
    real(stnd), intent(inout), dimension(*) :: z
!
    real(stnd), intent(out) :: dmin, dmin1, dmin2, dn, dn1, dn2
!
    integer(i4b), intent(in) :: i0, n0, pp
!
    logical(lgl), intent(in) :: ieee
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: d, emin, temp, dthresh
!
    integer(i4b) :: j4, j4p2
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n0-i0-1_i4b<=0_i4b ) then
!
        return
!
    end if
!
    dthresh = ulp*( sigma + tau )
!
    if ( tau<dthresh*half ) tau = zero
!
    if ( tau/=zero ) then
!
        j4 = 4_i4b*i0 + pp - 3_i4b
        emin = z( j4+4_i4b )
        d = z( j4 ) - tau
        dmin = d
        dmin1 = -z( j4 )
!
        if ( ieee ) then
!
!           CODE FOR IEEE ARITHMETIC.
!
            if ( pp==0_i4b ) then
!
                do j4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
                    z( j4-2_i4b ) = d + z( j4-1_i4b )
                    temp = z( j4+1_i4b )/z( j4-2_i4b )
                    d = d*temp - tau
                    dmin = min( dmin, d )
                    z( j4 ) = z( j4-1_i4b )*temp
                    emin = min( z( j4 ), emin )
!
                end do
!
            else
!
                do j4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
                    z( j4-3_i4b ) = d + z( j4 )
                    temp = z( j4+2_i4b )/z( j4-3_i4b )
                    d = d*temp - tau
                    dmin = min( dmin, d )
                    z( j4-1_i4b ) = z( j4 )*temp
                    emin = min( z( j4-1_i4b ), emin )
!
                end do
!
            end if
!
!           UNROLL LAST TWO STEPS.
!
            dn2 = d
            dmin2 = dmin
            j4 = 4_i4b*( n0-2_i4b ) - pp
            j4p2 = j4 + 2_i4b*pp - 1_i4b
            z( j4-2_i4b ) = dn2 + z( j4p2 )
            z( j4 ) = z( j4p2+2_i4b )*( z( j4p2 )/z( j4-2_i4b ) )
            dn1 = z( j4p2+2_i4b )*( dn2/z( j4-2_i4b ) ) - tau
            dmin = min( dmin, dn1 )
!
            dmin1 = dmin
            j4 = j4 + 4_i4b
            j4p2 = j4 + 2_i4b*pp - 1_i4b
            z( j4-2_i4b ) = dn1 + z( j4p2 )
            z( j4 ) = z( j4p2+2_i4b )*( z( j4p2 )/z( j4-2_i4b ) )
            dn = z( j4p2+2_i4b )*( dn1/z( j4-2_i4b ) ) - tau
            dmin = min( dmin, dn )
!
        else
!
!           CODE FOR NON IEEE ARITHMETIC.
!
            if ( pp==0_i4b ) then
!
                do j4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
                    z( j4-2_i4b ) = d + z( j4-1_i4b )
!
                    if ( d<zero ) then
                        return
                    else
                        z( j4 ) = z( j4+1_i4b )*( z( j4-1_i4b )/z( j4-2_i4b ) )
                        d = z( j4+1_i4b )*( d/z( j4-2_i4b ) ) - tau
                    end if
!
                    dmin = min( dmin, d )
                    emin = min( emin, z( j4 ) )
!
                end do
!
            else
!
                do j4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
                    z( j4-3_i4b ) = d + z( j4 )
!
                    if ( d<zero ) then
                        return
                    else
!
                        z( j4-1_i4b ) = z( j4+2_i4b )*( z( j4 )/z( j4-3_i4b ) )
                        d = z( j4+2_i4b )*( d/z( j4-3_i4b ) ) - tau
                    end if
!
                    dmin = min( dmin, d )
                    emin = min( emin, z( j4-1_i4b ) )
!
                end do
!
            end if
!
!           UNROLL LAST TWO STEPS.
!
            dn2 = d
            dmin2 = dmin
            j4 = 4_i4b*( n0 - 2_i4b ) - pp
            j4p2 = j4 + 2_i4b*pp - 1_i4b
            z( j4-2_i4b ) = dn2 + z( j4p2 )
!
            if ( dn2<zero ) then
                return
            else
                z( j4 ) = z( j4p2+2_i4b )*( z( j4p2 )/z( j4-2_i4b ) )
                dn1 = z( j4p2+2_i4b )*( dn2/z( j4-2_i4b ) ) - tau
            end if
!
            dmin = min( dmin, dn1 )
!
            dmin1 = dmin
            j4 = j4 + 4_i4b
            j4p2 = j4 + 2*pp - 1_i4b
            z( j4-2_i4b ) = dn1 + z( j4p2 )
!
            if ( dn1<zero ) then
                return
            else
                z( j4 ) = z( j4p2+2_i4b )*( z( j4p2 )/z( j4-2_i4b ) )
                dn = z( j4p2+2_i4b )*( dn1/z( j4-2_i4b ) ) - tau
            end if
!
            dmin = min( dmin, dn )
!
        end if
!
    else
!
!       THIS IS THE VERSION THAT SETS D'S TO ZERO IF THEY ARE SMALL ENOUGH.
!
        j4 = 4_i4b*i0 + pp - 3_i4b
        emin = z( j4+4_i4b )
        d = z( j4 ) - tau
        dmin = d
        dmin1 = -z( j4 )
!
        if ( ieee ) then
!
!           CODE FOR IEEE ARITHMETIC.
!
            if ( pp==0_i4b ) then
!
                do j4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
                    z( j4-2_i4b ) = d + z( j4-1_i4b )
                    temp = z( j4+1_i4b )/z( j4-2_i4b )
                    d = d*temp
!                    d = d*temp - tau
                    if ( d<dthresh ) d = zero
                    dmin = min( dmin, d )
                    z( j4 ) = z( j4-1_i4b )*temp
                    emin = min( z( j4 ), emin )
!
                end do
!
            else
!
                do j4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
                    z( j4-3_i4b ) = d + z( j4 )
                    temp = z( j4+2_i4b )/z( j4-3_i4b )
                    d = d*temp
!                    d = d*temp - tau
                    if ( d<dthresh ) d = zero
                    dmin = min( dmin, d )
                    z( j4-1_i4b ) = z( j4 )*temp
                    emin = min( z( j4-1_i4b ), emin )
!
                end do
!
            end if
!
!           UNROLL LAST TWO STEPS.
!
            dn2 = d
            dmin2 = dmin
            j4 = 4_i4b*( n0-2_i4b ) - pp
            j4p2 = j4 + 2_i4b*pp - 1_i4b
            z( j4-2_i4b ) = dn2 + z( j4p2 )
            z( j4 ) = z( j4p2+2_i4b )*( z( j4p2 )/z( j4-2_i4b ) )
            dn1 = z( j4p2+2_i4b )*( dn2/z( j4-2_i4b ) )
!            dn1 = z( j4p2+2_i4b )*( dn2/z( j4-2_i4b ) ) - tau
            dmin = min( dmin, dn1 )
!
            dmin1 = dmin
            j4 = j4 + 4_i4b
            j4p2 = j4 + 2_i4b*pp - 1_i4b
            z( j4-2_i4b ) = dn1 + z( j4p2 )
            z( j4 ) = z( j4p2+2_i4b )*( z( j4p2 )/z( j4-2_i4b ) )
            dn = z( j4p2+2_i4b )*( dn1/z( j4-2_i4b ) )
!            dn = z( j4p2+2_i4b )*( dn1/z( j4-2_i4b ) ) - tau
            dmin = min( dmin, dn )
!
        else
!
!           CODE FOR NON IEEE ARITHMETIC.
!
            if ( pp==0_i4b ) then
!
                do j4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
                    z( j4-2_i4b ) = d + z( j4-1_i4b )
!
                    if ( d<zero ) then
                        return
                    else
                        z( j4 ) = z( j4+1_i4b )*( z( j4-1_i4b )/z( j4-2_i4b ) )
                        d = z( j4+1_i4b )*( d/z( j4-2_i4b ) )
!                        d = z( j4+1_i4b )*( d/z( j4-2_i4b ) ) - tau
                    end if
!
                    if ( d<dthresh) d = zero
                    dmin = min( dmin, d )
                    emin = min( emin, z( j4 ) )
!
                end do
!
            else
!
                do j4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
                    z( j4-3_i4b ) = d + z( j4 )
!
                    if ( d<zero ) then
                        return
                    else
                        z( j4-1_i4b ) = z( j4+2_i4b )*( z( j4 )/z( j4-3_i4b ) )
                        d = z( j4+2_i4b )*( d/z( j4-3_i4b ) )
!                        d = z( j4+2_i4b )*( d/z( j4-3_i4b ) ) - tau
                    end if
!
                    if ( d<dthresh) d = zero
                    dmin = min( dmin, d )
                    emin = min( emin, z( j4-1_i4b ) )
!
                end do
!
            end if
!
!           UNROLL LAST TWO STEPS.
!
            dn2 = d
            dmin2 = dmin
            j4 = 4_i4b*( n0 - 2_i4b ) - pp
            j4p2 = j4 + 2_i4b*pp - 1_i4b
            z( j4-2_i4b ) = dn2 + z( j4p2 )
!
            if ( dn2<zero ) then
                return
            else
                z( j4 ) = z( j4p2+2_i4b )*( z( j4p2 )/z( j4-2_i4b ) )
                dn1 = z( j4p2+2_i4b )*( dn2/z( j4-2_i4b ) )
!                dn1 = z( j4p2+2_i4b )*( dn2/z( j4-2_i4b ) ) - tau
            end if
!
            dmin = min( dmin, dn1 )
!
            dmin1 = dmin
            j4 = j4 + 4_i4b
            j4p2 = j4 + 2_i4b*pp - 1_i4b
            z( j4-2_i4b ) = dn1 + z( j4p2 )
!
            if ( dn1<zero ) then
               return
            else
                z( j4 ) = z( j4p2+2_i4b )*( z( j4p2 )/z( j4-2_i4b ) )
                dn = z( j4p2+2_i4b )*( dn1/z( j4-2_i4b ) )
!                dn = z( j4p2+2_i4b )*( dn1/z( j4-2_i4b ) ) - tau
            end if
!
            dmin = min( dmin, dn )
!
        end if
!
    end if
!
    z( j4+2_i4b ) = dn
    z( 4*n0-pp ) = emin
!
!
! END OF SUBROUTINE bd_lasq5
! __________________________
!
    end subroutine bd_lasq5
!
! ===========================================================================================
!
      subroutine bd_lasq6( i0, n0, z, pp, dmin, dmin1, dmin2, dn, dn1, dn2 )
!
! Purpose
! _______
!
!   BD_LASQ6 computes one dqd (e.g., shift is equal to zero) transform in
!   ping-pong form, with protection against underflow and overflow.
!
!
! Arguments
! _________
!
!
!   I0      (INPUT) integer(i4b)
!           On entry, I0 is the first index.
!
!   N0      (INPUT) integer(i4b)
!           On entry, N0 is the last index.
!
!   Z       (INPUT/OUTPUT) real(stnd), dimension(4*n)
!           Z holds the qd array of size 4*n where n is the dimension of the
!           associated symmetric positive definite tridiagonal matrix.
!
!           EMIN is stored in Z(4*N0) to avoid an extra argument.
!
!   PP      (INPUT) integer(i4b)
!           On entry, PP=0 for ping, PP=1 for pong.
!
!   DMIN    (OUTPUT) real(stnd)
!           On exit, minimum value of D.
!
!   DMIN1   (OUTPUT) real(stnd)
!           On exit, minimum value of D, excluding D( N0 ).
!
!   DMIN2   (OUTPUT) real(stnd)
!           On exit, minimum value of D, excluding D( N0 ) and D( N0-1 ).
!
!   DN      (OUTPUT) real(stnd)
!           On exit, D(N0), the last value of D. 
!
!   DN1     (OUTPUT) real(stnd)
!           On exit, D(N0-1).
!
!   DN2     (OUTPUT) real(stnd)
!           On exit, D(N0-2).
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the DLASQ6 subroutine from LAPACK version 3.12.0.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(*) :: z
!
    real(stnd), intent(out) :: dmin, dmin1, dmin2, dn, dn1, dn2
!
    integer(i4b), intent(in) :: i0, n0, pp
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: d, emin, temp
!
    integer(i4b) :: j4, j4p2
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n0-i0-1_i4b<=0_i4b ) then
!
        return
!
    end if
!
    j4 = 4_i4b*i0 + pp - 3_i4b
    emin = z( j4+4_i4b )
    d = z( j4 )
    dmin = d
!
    if ( pp==0_i4b ) then
!
        do j4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
            z( j4-2_i4b ) = d + z( j4-1_i4b )
!
            if ( z( j4-2_i4b )==zero ) then
                z( j4 ) = zero
                d = z( j4+1_i4b )
                dmin = d
                emin = zero
            else if ( safmin*z( j4+1_i4b )<z( j4-2_i4b ) .and. safmin*z( j4-2_i4b )<z( j4+1_i4b ) ) then
                temp = z( j4+1_i4b )/z( j4-2_i4b )
                z( j4 ) = z( j4-1_i4b )*temp
                d = d*temp
            else
                z( j4 ) = z( j4+1_i4b )*( z( j4-1_i4b )/z( j4-2_i4b ) )
                d = z( j4+1_i4b )*( d/z( j4-2_i4b ) )
            end if
!
            dmin = min( dmin, d )
            emin = min( emin, z( j4 ) )
!
        end do
!
    else
!
        do j4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
            z( j4-3_i4b ) = d + z( j4 )
!
            if ( z( j4-3_i4b )==zero ) then
                z( j4-1_i4b ) = zero
                d = z( j4+2_i4b )
                dmin = d
                emin = zero
            else if ( safmin*z( j4+2_i4b )<z( j4-3_i4b ) .and. safmin*z( j4-3_i4b )<z( j4+2_i4b ) ) then
                temp = z( j4+2_i4b )/z( j4-3_i4b )
                z( j4-1_i4b ) = z( j4 )*temp
                d = d*temp
            else
                z( j4-1_i4b ) = z( j4+2_i4b )*( z( j4 )/z( j4-3_i4b ) )
                d = z( j4+2_i4b )*( d/z( j4-3_i4b ) )
            end if
!
            dmin = min( dmin, d )
            emin = min( emin, z( j4-1_i4b ) )
!
        end do
!
    end if
!
!   UNROLL LAST TWO STEPS.
!
    dn2 = d
    dmin2 = dmin
    j4 = 4_i4b*( n0 - 2_i4b ) - pp
    j4p2 = j4 + 2_i4b*pp - 1_i4b
    z( j4-2_i4b ) = dn2 + z( j4p2 )
!
    if ( z( j4-2_i4b )==zero ) then
       z( j4 ) = zero
       dn1 = z( j4p2+2_i4b )
       dmin = dn1
       emin = zero
    else if ( safmin*z( j4p2+2_i4b )<z( j4-2_i4b ) .and. safmin*z( j4-2_i4b )<z( j4p2+2_i4b ) ) then
       temp = z( j4p2+2_i4b )/z( j4-2_i4b )
       z( j4 ) = z( j4p2 )*temp
       dn1 = dn2*temp
    else
       z( j4 ) = z( j4p2+2_i4b )*( z( j4p2 )/z( j4-2_i4b ) )
       dn1 = z( j4p2+2_i4b )*( dn2/z( j4-2_i4b ) )
    end if
!
    dmin = min( dmin, dn1 )
!
    dmin1 = dmin
    j4 = j4 + 4_i4b
    j4p2 = j4 + 2_i4b*pp - 1_i4b
    z( j4-2_i4b ) = dn1 + z( j4p2 )
!
    if ( z( j4-2_i4b )==zero ) then
       z( j4 ) = zero
       dn = z( j4p2+2_i4b )
       dmin = dn
       emin = zero
    else if ( safmin*z( j4p2+2_i4b )<z( j4-2_i4b ) .and. safmin*z( j4-2_i4b )<z( j4p2+2_i4b ) ) then
       temp = z( j4p2+2_i4b )/z( j4-2_i4b )
       z( j4 ) = z( j4p2 )*temp
       dn = dn1*temp
    else
       z( j4 ) = z( j4p2+2_i4b )*( z( j4p2 )/z( j4-2_i4b ) )
       dn = z( j4p2+2_i4b )*( dn1/z( j4-2_i4b ) )
    end if
!
    dmin = min( dmin, dn )
!
    z( j4+2_i4b ) = dn
    z( 4*n0-pp ) = emin
!
!
! END OF SUBROUTINE bd_lasq6
! __________________________
!
    end subroutine bd_lasq6
!
! ===========================================================================================
!
    subroutine bd_lasq2_aggdef2( n, rk, rp, maxiter, ieee, z, info )
!
! Purpose
! _______
!
!   BD_LASQ2_AGGDEF2 computes all the eigenvalues of the N-by-N symmetric positive
!   definite tridiagonal matrix associated with the qd array Z to high
!   relative accuracy, in the absence of denormalization, underflow and overflow.,
!   with the dqds algorithm and an early aggressive deflation strategy.
!
!   To see the relation of Z to the tridiagonal matrix, let L be a
!   unit lower bidiagonal matrix with subdiagonals Z(2,4,6,,..) and
!   let U be an upper bidiagonal matrix with 1's above and diagonal
!   Z(1,3,5,,..). The tridiagonal is L*U or, if you prefer, the
!   symmetric tridiagonal to which it is similar.
!
!
! Arguments
! _________
!
!   N        (INPUT) integer(i4b)
!            The number of rows and columns in the tridiagonal matrix.
!            It is assumed that N>2, but this is not checked in the subroutine.
!            This assumption is automatically verified when BD_LASQ2_AGGDEF2 is
!            called from BD_LASQ1_AGGDEF2 subroutine.
!
!   RK       (INPUT) integer(i4b)
!            On entry, the maximum window size to be used in early aggressive deflation.
!
!   RP       (INPUT) integer(i4b)
!            On entry, the frequency with which early aggressive deflation will be invoked.
!
!   MAXITER  (INPUT) integer(i4b)
!            MAXITER controls the maximum number of iterations in the inner loop
!            of the dqds algorithm. MAXITER must be greater than 10,
!            otherwise the default value is used.
!
!            The default is 30.
!
!   IEEE     (INPUT) logical(lgl)
!            On entry, logical flag for IEEE or non IEEE arithmetic.
!            When IEEE is set to true, the division in the dqds loop is not protected from
!            incurring an exception (division by zero or overflow) since the powerful feature of
!            arithmetic units conforming to IEEE floating point standard 754 is that computation
!            is not held up by an exception. At the end of the loop, the code tests whether an
!            infinite or an NaN (not a number) occurred and acts appropriately.
!
!   Z        (INPUT/OUTPUT) real(stnd), dimension(4*n)
!            On entry Z(:2*N-1) holds the qd array. The other part of Z can be arbitrary and
!            is used as workspace.
!            On normal exit, entries 1 to N hold the eigenvalues (e.g. if INFO is equal to 0).
!            The eigenvalues are not sorted.      
!            On abnormal exit, if INFO is equal to 2, Z(:2*N-1) holds a qd array with the same
!            eigenvalues as the given qd array Z(:2*N-1).
!
!            It is assumed that all the elements of Z(:2*N-1) are greater or equal to zero on
!            input, but this is not checked in the subroutine.
!            This assumption is automatically verified when BD_LASQ2_AGGDEF2 is called
!            from BD_LASQ1_AGGDEF2 subroutine.
!
!   INFO     (OUTPUT) integer(i4b)
!            On exit:
!
!            - INFO = 0 :  indicates successful exit;
!            - INFO = 1 :  indicates a failure because a split was marked by a positive
!              value in the subdiagonal of the given tridiagonal matrix;
!            - INFO = 2 :  indicates that the algorithm did not converge
!              and that the maximum number of iterations in the inner loop
!              of the dqds algorithm is exceeded.
!              On exit, Z(:2*N-1) holds a qd array with the same eigenvalues as the
!              given Z(:2*N-1);
!            - INFO = 3 :  indicates that the termination criterion of the outer while
!              loop of the algorithm was not met (e.g. the program created more than
!              N unreduced blocks).
!            - INFO = 4 :  indicates that the maximum number of iterations in the inner
!              loop of the dqds algorithm is exceeded and that a qd array with
!              the same eigenvalues as the original one cannot be recovered.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the DLASQ2 subroutine in LAPACK version 3.12.0
!   and the test codes (e.g., DLASQDD2 and AGGDEF3I subroutines) associated with the reference (4)
!   for performing early aggressive deflation in conjunction with the dqds algorithm.
!
!   Local Variables: I0:N0 defines a current unreduced segment of Z.
!   The shifts are accumulated in SIGMA.
!   Ping-pong is controlled by PP (alternates between 0 and 1).
!
!   This subroutine is part of the dqds algorithm with early aggressive deflation described
!   in the references (2), (3) and (4).
!
!   For further details on the dqds algorithm, see:
!
!   (1) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (2) Parlett, B.N., and Marques, O.A., 2000:
!           An implementation of the dqds algorithm (positive case).
!           Linear Algebra Appl., Volume 309, pp. 217-259.
!
!   (3) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!   (4) Nakatsukasa, Y., Aishima, K., and Yamazaki, I., 2012:
!           dqds with aggressive early deflation.
!           SIAM Journal on Matrix Analysis Applications, Volume 33, No 1, pp. 22-51.
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: n, rk, rp, maxiter
    integer(i4b), intent(out) :: info
!
    real(stnd),   intent(inout), dimension(4*n) :: z
!
    logical(lgl), intent(in) :: ieee
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: d, dee, deemin, desig, dmn, dmn1, dmn2, dn,    &
                  dn1, dn2, emax, emin, g, oldemn, qmax, qmin,   &
                  sigma, temp, tol, tol2, tempe, tempq, rr
    real(stnd), dimension(4*n) :: z2
!
    integer(i4b) :: maxiter2, i, i0, i1, i4, ipn4, iwhila, iwhilb,   &
                    k, kmin, n0, n1, nbig, pp, splt, ttype
    integer(i4b) :: rkhere, ri, n0_old, i00, info2, n_aggdef
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   INPUT ARGUMENTS ARE NOT TESTED AS IT IS ASSUMED THAT bd_lasq2_aggdef2 IS CALLED FROM bd_lasq1_aggdef2.
!   IN WHICH CASE n>2 .
!
    info = 0_i4b
!
!   COMPUTE TOLERANCES.
!
    tol  = ulp*fudge_lasq
    tol2 = tol**2
!
    maxiter2 = merge( maxiter, 30_i4b, maxiter>=10_i4b )
!
    z(2_i4b*n) = zero
!
!   REARRANGE DATA FOR LOCALITY: Z=(Q1,QQ1,E1,EE1,Q2,QQ2,E2,EE2,...).
!
    do k=2_i4b*n, 2_i4b, -2_i4b
!
        z(2_i4b*k)       = zero
        z(2_i4b*k-1_i4b) = z(k)
        z(2_i4b*k-2_i4b) = zero
        z(2_i4b*k-3_i4b) = z(k-1_i4b)
!
    end do
!
    i0 = 1_i4b
    n0 = n
!
!   REVERSE THE QD-ARRAY, IF WARRANTED.
!
    if ( c1_5*z( 4_i4b*i0-3_i4b )<z( 4*n0-3_i4b ) ) then
!
        ipn4 = 4_i4b*( i0+n0 )
!
        do i4=4_i4b*i0, 2_i4b*( i0+n0-1_i4b ), 4_i4b
!
            temp = z(i4-3_i4b)
            z(i4-3_i4b) = z(ipn4-i4-3_i4b)
            z(ipn4-i4-3_i4b) = temp
!
            temp = z(i4-1_i4b)
            z(i4-1_i4b) = z(ipn4-i4-5_i4b)
            z(ipn4-i4-5_i4b) = temp
!
        end do
!
    end if
!
!   INITIAL SPLIT CHECKING VIA DQD AND LI'S TEST.
!
    pp = 0_i4b
!
    do k = 1_i4b, 2_i4b
!
        d = z(4_i4b*n0+pp-3_i4b)
!
        do i4 = 4_i4b*( n0-1_i4b ) + pp, 4_i4b*i0 + pp, -4_i4b
!
            if ( z(i4-1_i4b)<=tol2*d ) then
                z(i4-1_i4b) = -zero
                d = z(i4-3_i4b)
            else
                d = z(i4-3_i4b)*( d/( d + z(i4-1_i4b) ) )
            end if
!
        end do
!
!       DQD MAPS Z TO ZZ PLUS LI'S TEST.
!
        d = z(4_i4b*i0+pp-3_i4b)
!
        do i4 = 4_i4b*i0 + pp, 4_i4b*( n0-1_i4b ) + pp, 4_i4b
!
            z(i4-2_i4b*pp-2_i4b) = d + z(i4-1_i4b)
!
            if ( z(i4-1_i4b)<=tol2*d ) then
               z(i4-1_i4b) = -zero
               z(i4-2_i4b*pp-2_i4b) = d
               z(i4-2_i4b*pp) = zero
               d = z(i4+1_i4b)
            else if ( safmin*z(i4+1_i4b)<z(i4-2_i4b*pp-2_i4b) .and. safmin*z(i4-2_i4b*pp-2_i4b)<z(i4+1_i4b) ) then
               temp = z(i4+1_i4b)/z(i4-2_i4b*pp-2_i4b)
               z(i4-2_i4b*pp) = z(i4-1_i4b)*temp
               d = d*temp
            else
               z(i4-2_i4b*pp) = z(i4+1_i4b)*(z( i4-1_i4b)/z(i4-2_i4b*pp-2_i4b) )
               d = z(i4+1_i4b)*( d/z(i4-2_i4b*pp-2_i4b) )
            end if
!
        end do
!
        z(4_i4b*n0-pp-2_i4b) = d
!
!       PREPARE FOR THE NEXT ITERATION ON k.
!
        pp = 1_i4b - pp
!
   end do
!
!  INITIALISE VARIABLES TO PASS TO bd_lasq3.
!
    ttype = 0_i4b
    dmn1  = zero
    dmn2  = zero
    dn    = zero
    dn1   = zero
    dn2   = zero
    g     = zero
!
    do iwhila=1_i4b, n + 1_i4b
!
!       CHECK IF ALL EIGENVALUES HAVE BEEN FOUND.
!
        if ( n0<1_i4b )  exit
!
!       WHILE ARRAY UNFINISHED DO.
!
!       e(n0) HOLDS THE VALUE OF sigma WHEN SUBMATRIX IN i0:n0
!       SPLITS FROM THE REST OF THE ARRAY, BUT IS NEGATED.
!
        desig = zero
!
        if ( n0==n ) then
            sigma = zero
        else
            sigma = -z(4_i4b*n0-1_i4b)
        end if
!
        if ( sigma<zero ) then
!
            info = 1_i4b
!
            exit
!
        end if
!
!       FIND LAST UNREDUCED SUBMATRIX'S TOP INDEX i0, FIND qmax AND
!       emin. FIND GERSHGORIN-TYPE BOUND IF Q'S MUCH GREATER THAN E'S.
!
        emax = zero
        if ( n0>i0 ) then
            emin = abs( z(4_i4b*n0-5_i4b) )
        else
            emin = zero
        end if
!
        qmin = z(4_i4b*n0-3_i4b)
        qmax = qmin
!
        do i4=4_i4b*n0, 8_i4b, -4_i4b
!
            if ( z(i4-5_i4b)<=zero ) exit
!
            if( qmin>=four*emax ) then
               qmin = min( qmin, z(i4-3_i4b) )
               emax = max( emax, z(i4-5_i4b) )
            end if
!
            qmax = max( qmax, z(i4-7_i4b)+z(i4-5_i4b) )
            emin = min( emin, z(i4-5_i4b) )
!
        end do
!
        if ( i4<8_i4b ) then
            i4 = 4_i4b
        end if
!
        i0 = i4/4_i4b
        pp = 0_i4b
!
        if ( n0-i0>1_i4b ) then
!
            dee = z(4_i4b*i0-3_i4b)
            deemin = dee
            kmin = i0
!
            do i4=4_i4b*i0+1_i4b, 4_i4b*n0-3_i4b, 4_i4b
!
                dee = z(i4)*( dee/( dee + z(i4-2_i4b) ) )
!
                if ( dee<=deemin ) then
                    deemin = dee
                    kmin = ( i4 + 3_i4b )/4_i4b
                end if
!
            end do 
!
            if ( (kmin-i0)*2_i4b<n0-kmin .and. deemin<=half*z(4_i4b*n0-3_i4b) ) then
!
                ipn4 = 4_i4b*( i0 + n0 )
                pp = 2_i4b
!
                do i4=4_i4b*i0, 2_i4b*( i0+n0-1_i4b ), 4_i4b
!
                    temp = z(i4-3_i4b)
                    z(i4-3_i4b) = z(ipn4-i4-3_i4b)
                    z(ipn4-i4-3_i4b) = temp
!
                    temp = z(i4-2_i4b)
                    z(i4-2_i4b) = z(ipn4-i4-2_i4b)
                    z(ipn4-i4-2_i4b) = temp
!
                    temp = z(i4-1_i4b)
                    z(i4-1_i4b) = z(ipn4-i4-5_i4b)
                    z(ipn4-i4-5_i4b) = temp
!
                    temp = z(i4)
                    z(i4) = z(ipn4-i4-4_i4b)
                    z(ipn4-i4-4_i4b) = temp
!
                end do
!
            end if
!
        end if
!
!       PUT -(INITIAL SHIFT) INTO dmn.
!
        dmn = -max( zero, qmin-two*sqrt( qmin )*sqrt( emax ) )
!
!       NOW i0:n0 IS UNREDUCED.
!       pp = 0 FOR PING, pp = 1 FOR PONG.
!       pp = 2 INDICATES THAT FLIPPING WAS APPLIED TO THE Z ARRAY AND
!              AND THAT THE TESTS FOR DEFLATION UPON ENTRY IN bd_lasq3
!              SHOULD NOT BE PERFORMED.
!
        ri   = 0_i4b
        nbig = maxiter2*( n0 - i0 + 1_i4b )
!
        do iwhilb = 1_i4b, nbig
!
            if ( i0>n0 ) exit
!
!           WHILE SUBMATRIX UNFINISHED TAKE A GOOD DQDS STEP.
!
            call bd_lasq3( i0, n0, z, pp, sigma, desig, qmax, ieee, &
                           ttype, dmn, dmn1, dmn2, dn, dn1, dn2, g  )
!
            ri = 1_i4b + ri
            pp = 1_i4b - pp
!
!           WHEN emin IS VERY SMALL CHECK FOR SPLITS.
!
            i00 = i0
!
            if ( pp==0_i4b ) then
!
                if ( n0-i0>=3_i4b ) then
!
                  if ( z(4_i4b*n0)<=tol2*qmax .or. z(4_i4b*n0-1_i4b)<=tol2*sigma ) then
!
                    splt   = i0 - 1_i4b
                    qmax   = z(4*i0-3_i4b)
                    emin   = z(4*i0-1_i4b)
                    oldemn = z(4*i0)
!
                    do i4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
                        if ( z(i4)<=tol2*z(i4-3_i4b) .or. z(i4-1_i4b)<=tol2*sigma ) then
                            z(i4-1_i4b) = -sigma
                            splt = i4/4_i4b
                            qmax = zero
                            emin = z(i4+3_i4b)
                            oldemn = z(i4+4_i4b)
                        else
                            qmax = max( qmax, z(i4+1_i4b) )
                            emin = min( emin, z(i4-1_i4b) )
                            oldemn = min( oldemn, z(i4) )
                        end if
!
                    end do
!
                    z(4_i4b*n0-1_i4b) = emin
                    z(4_i4b*n0) = oldemn
!
                    i0 = splt + 1_i4b
!
                  end if
!
                end if
!
                if ( i00==i0 ) then 
!
                    if ( ri>=rp .and. n0-i0+1_i4b>2_i4b*rk ) then
!
!                       PERFORM AGGRESSIVE DEFLATION AT A FREQUENCY OF rp .
!                 
!                       FIRST ESTIMATE THE CURRENT WINDOW SIZE (rk IS THE MAXIMUM POSSIBLE WINDOW)
!                       USING THE INFORMATION FROM THE BOTTOM RIGHT OF THE BIDIAGONAL MATRIX.
!
!                       FIRST COMPARE E(n0-i+1) AND D(n0-i+2) FOR i=1,2,... AND FIND THE LARGEST i
!                       SUCH THAT abs(E(n0-i+1))<abs(D(n0-i+2)). WHEN THIS RESULTS IN A WINDOW SIZE
!                       LESS THAN OR EQUAL TO 10, WE SKIP AGGRESSIVE DEFLATION TO AVOID WASTING EFFORT.
!
	                do i=2_i4b, 10_i4b
!
	                    if ( z(4_i4b*(n0-i)-3_i4b)<z(4_i4b*(n0-i)-1_i4b) ) then 
!
!	                        DONT DO AGGRESSIVE DEFLATION IN THIS CASE TO AVOID WASTING EFFORT.
!       
!                               RESET THE ITERATION COUNT AND EXIT.
!       
                                ri = 0_i4b
!
	                        exit	
!
	                     end if
!
	                end do
!                 
                        if ( i>10_i4b ) cycle
!                 
!                       NOW DO AGGRESSIVE DEFLATION WITH A WINDOW SIZE OF AT LEAST 10.
!                 
                        n0_old = n0
!                 
                        do
!                 
!                           ESTIMATE THE CURRENT WINDOW SIZE rkhere .
!                 
	                    rr = one
!                 
                 	    do i=11_i4b, min(rk,n0-10_i4b)
!                 
                	        rr = rr*z(4_i4b*(n0-i)-1_i4b)/z(4_i4b*(n0-i)-3_i4b)
!                 
                                if ( z(4_i4b*(n0-i)-3_i4b)<z(4_i4b*(n0-i)-1_i4b) .or. rr<ulp*ulp ) then 
	                            exit
	                        end if
!                 
	                    end do
!                 
                            rkhere = i
!                 
!                           NOW PERFORM AGGRESSIVE DEFLATION WITH A WINDOW SIZE OF rkhere .
!                            
!                           FIRST MAKE A LOCAL COPY OF THE BIDIAGONAL MATRIX IN THE WINDOW OF SIZE rkhere
!                           IN ORDER TO PERFORM AGGRESSIVE DEFLATION.
!                            
                            i = 0_i4b
!                            
                            do i4=n0-rkhere+1_i4b, n0
!                            
                                z2(i+1_i4b) = z(4_i4b*i4-3_i4b)
                                z2(i+2_i4b) = z(4_i4b*i4-1_i4b)
!                            
                                i = i + 2_i4b
!                            
                            end do
!
!                           DO AGGRESSIVE DEFLATION.
!                            
                            call bd_aggdef2( rkhere, maxiter2, sigma, ieee, z2, z(4_i4b*(n0-rkhere)+1_i4b), &
                                             n_aggdef, info2 )
!                 
!                           REPEAT IMMEDIATELY AGGRESSIVE DEFLATION IF MANY EIGENVALUES ARE DEFLATED.
!                 
                            if ( n_aggdef>2_i4b .and. n0>3_i4b*rk ) then
!                            
                                n0 = n0 - n_aggdef
!                            
                                cycle
!                            
                            end if
!                 
                            exit
!                 
                        end do
!       
!                       CORRECT THE CURRENT SIZE n0 OF THE BIDIAGONAL MATRIX.
!       
                        n0 = n0_old
!
!                       END OF AGGRESSIVE DEFLATION.
!       
!                       RESET THE ITERATION COUNT.
!       
                        ri = 0_i4b
!       
                    end if
!       
                else
!       
!                   RESET THE ITERATION COUNT.
!       
                    ri = 0_i4b
!       
                end if
!       
            end if
!
        end do
!
        if ( iwhilb>nbig ) then
!
            info = 2_i4b
!
            exit
!
        end if
!
    end do
!
!   POST-PROCESSING.
!
    if ( iwhila>n+1_i4b ) then
!
!       MAXIMUM NUMBER OF OUTER ITERATIONS EXCEEDED.
!
        info = 3_i4b
!
    else if ( info==0_i4b ) then
!
!       ALL EIGENVALUES FOUND, MOVE Q'S TO THE FRONT.
!
        do k=2_i4b, n
            z(k) = z(4_i4b*k-3_i4b)
        end do
!
    else if ( info==2_i4b ) then
!
!       MAXIMUM NUMBER OF INNER ITERATIONS EXCEEDED, RESTORE THE SHIFT
!       SIGMA AND PLACE THE NEW D'S AND E'S IN A QD ARRAY.
!       THIS MIGHT NEED TO BE DONE FOR SEVERAL BLOCKS.
!
        i1 = i0
        n1 = n0
!
        do
!
            tempq = z(4_i4b*i0-3_i4b)
            z(4_i4b*i0-3_i4b) = z(4_i4b*i0-3_i4b) + sigma
!
            do k=i0+1_i4b, n0
!
                tempe = z(4_i4b*k-5_i4b)
                z(4_i4b*k-5_i4b) = z(4_i4b*k-5_i4b)*( tempq/z(4_i4b*k-7_i4b) )
!
                tempq = z(4_i4b*k-3_i4b)
                z(4_i4b*k-3_i4b) = z(4_i4b*k-3_i4b) + sigma + tempe - z(4_i4b*k-5_i4b)
!
            end do
!
!           PREPARE TO DO THIS ON THE PREVIOUS BLOCK IF THERE IS ONE.
!
            if ( i1>1_i4b ) then
!
                n1 = i1 - 1_i4b
!
                i4 = i1
!
                do while( i1>=2_i4b .and. z(4_i4b*i1-5_i4b)>=zero )
                   i1 = i1 - 1_i4b
                end do
!
                if ( i1/=i4 ) then
!
                   sigma = -z(4_i4b*n1-1_i4b)
!
                   cycle
!
                end if
!
            end if
!
            exit
!
        end do
!
        do k=1_i4b, n
!
            z(2_i4b*k-1_i4b) = z(4_i4b*k-3_i4b)
!
!           ONLY THE BLOCK 1..n0 IS UNFINISHED.  THE REST OF THE E'S
!           MUST BE ESSENTIALLY ZERO, ALTHOUGH SOMETIMES OTHER DATA
!           HAS BEEN STORED IN THEM.
!
            if( k<n0 ) then
               z(2_i4b*k) = z(4_i4b*k-1_i4b)
            else
               z(2_i4b*k) = 0_i4b
            end if
!
        end do
!
!       CHECK IF ANY ELEMENT IN THE NEW QD ARRAY IS NEGATIVE. 
!      
        if ( any( z(1_i4b:2_i4b*n)<zero ) ) then
!      
            info = 4_i4b
!                    
        end if
!
    end if
!
!
! END OF SUBROUTINE bd_lasq2_aggdef2
! __________________________________
!      
    end subroutine bd_lasq2_aggdef2
!
! ===========================================================================================
!
    subroutine bd_aggdef2( n, maxiter, sigmag, ieee, z, w, stat, info ) 
!
! Purpose
! _______
!
!   BD_AGGDEF2 computes implicitly all the eigenvalues of the N-by-N symmetric positive
!   definite tridiagonal matrix associated with the qd array Z of the current window of size N
!   to high relative accuracy, in the absence of denormalization, underflow and overflow. This
!   is done to perform early aggressive deflation for the dqds algorithm as described in the
!   reference (4).
!
!   To see the relation of Z to the tridiagonal matrix, let L be a
!   unit lower bidiagonal matrix with subdiagonals Z(2,4,6,,..) and
!   let U be an upper bidiagonal matrix with 1's above and diagonal
!   Z(1,3,5,,..). The tridiagonal is L*U or, if you prefer, the
!   symmetric tridiagonal to which it is similar.
!
!
! Arguments
! _________
!
!   N        (INPUT) integer(i4b)
!            The number of rows and columns in the tridiagonal matrix associated with the current
!            window or, equivalently, the current window size.
!
!            It is assumed that N>2, but this is not checked in the subroutine.
!            This assumption is automatically verified when BD_AGGDEF2 is called from
!            BD_LASQ2_AGGDEF2 subroutine.
!
!   MAXITER  (INPUT) integer(i4b)
!            MAXITER controls the maximum number of iterations in the inner loop
!            of the dqds algorithm. MAXITER must be greater than 10,
!            otherwise the default value is used.
!
!   SIGMAG   (INPUT) real(stnd)
!            On entry, the accumulated shifts by the previous dqds steps.
!
!   IEEE     (INPUT) logical(lgl)
!            On entry, logical flag for IEEE or non IEEE arithmetic (IEEE is passed to BD_LASQ3).
!
!            The default is 30.
!
!   Z        (INPUT/OUTPUT) real(stnd), dimension(4*n)
!            On entry Z(:2*N-1) holds a copy of the qd array associated with the current window. The other
!            part of Z can be arbitrary and is used as workspace.
!
!            It is assumed that all the elements of Z(:2*N-1) are greater or equal to zero on
!            input, but this is not checked in the subroutine. This assumption is automatically
!            verified when BD_AGGDEF2 is called from BD_LASQ2_AGGDEF2 subroutine.
!
!   W        (INPUT/OUTPUT) real(stnd), dimension(*)
!            On entry W holds the qd array associated with the current window.
!
!   STAT     (OUTPUT) integer(i4b)
!            On exit, the number of deflated elements in the current early aggressive deflation step.
!
!   INFO     (OUTPUT) integer(i4b)
!            On exit:
!
!            - INFO = 0 :  indicates successful exit;
!            - INFO = 1 :  indicates a failure because a split was marked by a positive
!              value in the subdiagonal of the given tridiagonal matrix;
!            - INFO = 2 :  indicates that the algorithm did not converge
!              and that the maximum number of iterations in the inner loop
!              of the dqds algorithm is exceeded.
!            - INFO = 3 :  indicates that the termination criterion of the outer while
!              loop of the algorithm was not met (e.g. the program created more than
!              N unreduced blocks).
!            - INFO = 4 :  indicates that the maximum number of iterations in the inner
!              loop of the dqds algorithm is exceeded and that a qd array with
!              the same eigenvalues as the original one cannot be recovered.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the DLASQ2 subroutine in LAPACK version 3.12.0
!   and the codes (e.g., DLASQ2ADI subroutine) associated with the reference (4).
!
!   Local Variables: I0:N0 defines a current unreduced segment of Z.
!   The accumulated shifts are in SIGMAG.
!   Ping-pong is controlled by PP (e.g., alternates between 0 and 1).
!
!   This subroutine is part of the dqds algorithm with early aggressive deflation
!   described in the reference (4) and is adapted from the test codes provided by
!   the authors of reference (4).
!
!   For further details on the dqds algorithm and early aggressive deflation, see:
!
!   (1) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (2) Parlett, B.N., and Marques, O.A., 2000:
!           An implementation of the dqds algorithm (positive case).
!           Linear Algebra Appl., Volume 309, pp. 217-259.
!
!   (3) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
!   (4) Nakatsukasa, Y., Aishima, K., and Yamazaki, I., 2012:
!           dqds with aggressive early deflation.
!           SIAM Journal on Matrix Analysis Applications, Volume 33, No 1, pp. 22-51.
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: n, maxiter
    integer(i4b), intent(out) :: stat, info
!
    real(stnd),   intent(in) :: sigmag
    real(stnd),   intent(inout), dimension(4*n) :: z
    real(stnd),   intent(inout), dimension(*)   :: w
!
    logical(lgl), intent(in) :: ieee
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
!
    real(stnd) :: d, dee, deemin, desig, dmn, dmn1, dmn2, dn,    &
                  dn1, dn2, emax, emin, g, oldemn, qmax, qmin,   &
                  s, sigma, temp, tol, tol2, s1, s2, s2pre,      &
                  nrm, offv, offv2, offv3
    real(stnd), dimension(2*n+1) :: zpre
!
    integer(i4b) :: maxiter2, i, i0, i0in, i3, i4, ipn4, iwhila, iwhilb,  &
                    j, k, kmin, n0, n0in, nbig, nn, pp, splt, ttype, offi
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   INPUT ARGUMENTS ARE NOT TESTED AS IT IS ASSUMED THAT bd_aggdef2 IS CALLED FROM bd_lasq2_aggdef2,
!   IN WHICH CASE n>2 .
!
    info = 0_i4b
!
!   COMPUTE TOLERANCES.
!
    tol  = ulp*fudge_lasq
    tol2 = tol**2
!
    maxiter2 = merge( maxiter, 30_i4b, maxiter>=10_i4b )
!
    z(2_i4b*n) = zero
!
!   REARRANGE DATA FOR LOCALITY: Z=(Q1,QQ1,E1,EE1,Q2,QQ2,E2,EE2,...).
!
    do k=2_i4b*n, 2_i4b, -2_i4b
!
        z(2_i4b*k)       = zero
        z(2_i4b*k-1_i4b) = z(k)
        z(2_i4b*k-2_i4b) = zero
        z(2_i4b*k-3_i4b) = z(k-1_i4b)
!
    end do
!
    i0 = 1_i4b
    n0 = n
!
!   REVERSE THE QD-ARRAY, IF WARRANTED.
!
    if ( c1_5*z( 4_i4b*i0-3_i4b )<z( 4*n0-3_i4b ) ) then
!
        ipn4 = 4_i4b*( i0+n0 )
!
        do i4=4_i4b*i0, 2_i4b*( i0+n0-1_i4b ), 4_i4b
!
            temp = z(i4-3_i4b)
            z(i4-3_i4b) = z(ipn4-i4-3_i4b)
            z(ipn4-i4-3_i4b) = temp
!
            temp = z(i4-1_i4b)
            z(i4-1_i4b) = z(ipn4-i4-5_i4b)
            z(ipn4-i4-5_i4b) = temp
!
        end do
!
    end if
!
!   INITIAL SPLIT CHECKING VIA DQD AND LI'S TEST.
!
    pp = 0_i4b
!
    do k = 1_i4b, 2_i4b
!
        d = z(4_i4b*n0+pp-3_i4b)
!
        do i4 = 4_i4b*( n0-1_i4b ) + pp, 4_i4b*i0 + pp, -4_i4b
!
            if ( z(i4-1_i4b)<=tol2*d ) then
                z(i4-1_i4b) = -zero
                d = z(i4-3_i4b)
            else
                d = z(i4-3_i4b)*( d/( d + z(i4-1_i4b) ) )
            end if
!
        end do
!
!       DQD MAPS Z TO ZZ PLUS LI'S TEST.
!
        d = z(4_i4b*i0+pp-3_i4b)
!
        do i4 = 4_i4b*i0 + pp, 4_i4b*( n0-1_i4b ) + pp, 4_i4b
!
            z(i4-2_i4b*pp-2_i4b) = d + z(i4-1_i4b)
!
            if ( z(i4-1_i4b)<=tol2*d ) then
               z(i4-1_i4b) = -zero
               z(i4-2_i4b*pp-2_i4b) = d
               z(i4-2_i4b*pp) = zero
               d = z(i4+1_i4b)
            else if ( safmin*z(i4+1_i4b)<z(i4-2_i4b*pp-2_i4b) .and. safmin*z(i4-2_i4b*pp-2_i4b)<z(i4+1_i4b) ) then
               temp = z(i4+1_i4b)/z(i4-2_i4b*pp-2_i4b)
               z(i4-2_i4b*pp) = z(i4-1_i4b)*temp
               d = d*temp
            else
               z(i4-2_i4b*pp) = z(i4+1_i4b)*(z( i4-1_i4b)/z(i4-2_i4b*pp-2_i4b) )
               d = z(i4+1_i4b)*( d/z(i4-2_i4b*pp-2_i4b) )
            end if
!
        end do
!
        z(4_i4b*n0-pp-2_i4b) = d
!
!       PREPARE FOR THE NEXT ITERATION ON k.
!
        pp = 1_i4b - pp
!
   end do
!
!  INITIALISE VARIABLES TO PASS TO bd_lasq3.
!
    ttype = 0_i4b
    dmn1  = zero
    dmn2  = zero
    dn    = zero
    dn1   = zero
    dn2   = zero
    g     = zero
!
!   INITIALIZE AGGRESSIVE DEFLATION COUNTERS.
!
    stat = 0_i4b
    j    = n
    s2   = zero
!
    whila_iter : do iwhila=1_i4b, n + 1_i4b
!
!       CHECK IF ALL EIGENVALUES HAVE BEEN FOUND.
!
        if ( n0<1_i4b )  exit
!
!       WHILE ARRAY UNFINISHED DO.
!
!       e(n0) HOLDS THE VALUE OF sigma WHEN SUBMATRIX IN i0:n0
!       SPLITS FROM THE REST OF THE ARRAY, BUT IS NEGATED.
!
        desig = zero
!
        if ( n0==n ) then
            sigma = zero
        else
            sigma = -z(4_i4b*n0-1_i4b)
        end if
!
        if ( sigma<zero ) then
!
            info = 1_i4b
!
            exit
!
        end if
!
!       FIND LAST UNREDUCED SUBMATRIX'S TOP INDEX i0, FIND qmax AND
!       emin. FIND GERSHGORIN-TYPE BOUND IF Q'S MUCH GREATER THAN E'S.
!
        emax = zero
        if ( n0>i0 ) then
            emin = abs( z(4_i4b*n0-5_i4b) )
        else
            emin = zero
        end if
!
        qmin = z(4_i4b*n0-3_i4b)
        qmax = qmin
!
        do i4=4_i4b*n0, 8_i4b, -4_i4b
!
            if ( z(i4-5_i4b)<=zero ) exit
!
            if( qmin>=four*emax ) then
               qmin = min( qmin, z(i4-3_i4b) )
               emax = max( emax, z(i4-5_i4b) )
            end if
!
            qmax = max( qmax, z(i4-7_i4b)+z(i4-5_i4b) )
            emin = min( emin, z(i4-5_i4b) )
!
        end do
!
        if ( i4<8_i4b ) then
            i4 = 4_i4b
        end if
!
        i0 = i4/4_i4b
        pp = 0_i4b
!
        if ( n0-i0>1_i4b ) then
!
            dee = z(4_i4b*i0-3_i4b)
            deemin = dee
            kmin = i0
!
            do i4=4_i4b*i0+1_i4b, 4_i4b*n0-3_i4b, 4_i4b
!
                dee = z(i4)*( dee/( dee + z(i4-2_i4b) ) )
!
                if ( dee<=deemin ) then
                    deemin = dee
                    kmin = ( i4 + 3_i4b )/4_i4b
                end if
!
            end do 
!
            if ( (kmin-i0)*2_i4b<n0-kmin .and. deemin<=half*z(4_i4b*n0-3_i4b) ) then
!
                ipn4 = 4_i4b*( i0 + n0 )
                pp = 2_i4b
!
                do i4=4_i4b*i0, 2_i4b*( i0+n0-1_i4b ), 4_i4b
!
                    temp = z(i4-3_i4b)
                    z(i4-3_i4b) = z(ipn4-i4-3_i4b)
                    z(ipn4-i4-3_i4b) = temp
!
                    temp = z(i4-2_i4b)
                    z(i4-2_i4b) = z(ipn4-i4-2_i4b)
                    z(ipn4-i4-2_i4b) = temp
!
                    temp = z(i4-1_i4b)
                    z(i4-1_i4b) = z(ipn4-i4-5_i4b)
                    z(ipn4-i4-5_i4b) = temp
!
                    temp = z(i4)
                    z(i4) = z(ipn4-i4-4_i4b)
                    z(ipn4-i4-4_i4b) = temp
!
                end do
!
            end if
!
        end if
!
!       PUT -(INITIAL SHIFT) INTO dmn.
!
        dmn = -max( zero, qmin-two*sqrt( qmin )*sqrt( emax ) )
!
!       NOW i0:n0 IS UNREDUCED.
!       pp = 0 FOR PING, pp = 1 FOR PONG.
!       pp = 2 INDICATES THAT FLIPPING WAS APPLIED TO THE Z ARRAY AND
!              AND THAT THE TESTS FOR DEFLATION UPON ENTRY IN bd_lasq3
!              SHOULD NOT BE PERFORMED.
!
        nbig = maxiter2*( n0 - i0 + 1_i4b )
!
        do iwhilb = 1_i4b, nbig
!
            if ( i0>n0 ) exit
!
!           WHILE SUBMATRIX UNFINISHED TAKE A GOOD DQDS STEP.
!
            i0in = i0
            n0in = n0
!
            call bd_lasq3( i0, n0, z, pp, sigma, desig, qmax, ieee, &
                           ttype, dmn, dmn1, dmn2, dn, dn1, dn2, g  )
!
            if ( i0in==i0 .and. n0in==n0 ) then
!
!               PERFORM AGGRESSIVE DEFLATION.
!
                nn = 4_i4b*n0in
                s1 = z(nn-3_i4b)
!
!               SCALING.
!
                s     = s1 - s2
	        s2pre = s2
                s2    = s1
!
!               SAVE INITIAL VALUES OF THE QD ARRAY.
!
                do i4=1_i4b, j-1_i4b
                    zpre(2_i4b*i4-1_i4b) = w(4_i4b*i4-3_i4b)
                    zpre(2_i4b*i4)       = w(4_i4b*i4-1_i4b)
                end do
!
                zpre(2_i4b*j-1_i4b) = w(4_i4b*j-3_i4b)
!
!               APPLY A DIFFERENTIAL STATIONARY QDS TRANSFORM TO DEFLATE s .
!
                call stqdsi( s, j, w(1_i4b), w(3_i4b), 4_i4b )
!
!               ESTIMATE TOLERANCE.
!
                tol = ulp*sqrt( real(n, stnd) )*( s2 + sigmag )
!
!               IF NEGATIVE ELEMENTS EXIST OTHER THAN AT THE BOTTOM DIAGONAL,
!               THIS INDICATES A BREAKDOWN OF THE CHOLESKY DECOMPOSITION. IN THIS CASE,
!               WE EXIT AND WE RETURN TO DQDS ITERATIONS.
!
                do i4=1_i4b, j-1_i4b
!
                    if ( w(4_i4b*i4-3_i4b)<=zero .or. w(4_i4b*i4-1_i4b)<=zero ) then
!
                        do i3=1_i4b, j-1_i4b
                            w(4_i4b*i3-3_i4b) = zpre(2_i4b*i3-1_i4b)
                            w(4_i4b*i3-1_i4b) = zpre(2_i4b*i3)
                        end do
!
                        w(4_i4b*j-3_i4b) = zpre(2_i4b*j-1_i4b)
	                s2               = s2pre
!
                        exit whila_iter
!
                    end if
!
                end do
!
!               IF BOTTOM ELEMENT NOT SMALL, EXIT AS s IS NOT DEFLATED
!               BY THE PREVIOUS DIFFERENTIAL STATIONARY QDS TRANSFORM.
!
	        if ( abs(w(4*j-3))>=tol ) then
!
                    do i3=1_i4b, j-1_i4b
                        w(4_i4b*i3-3_i4b) = zpre(2_i4b*i3-1_i4b)
                        w(4_i4b*i3-1_i4b) = zpre(2_i4b*i3)
                    end do
!
                    w(4_i4b*j-3_i4b) = zpre(2_i4b*j-1_i4b)
		    s2               = s2pre
!
                    exit whila_iter
!
	        end if
!
!               NOW CHECK IF sqrt( s + sigma ) CAN BE DEFLATED AS A CONVERGED SINGULAR VALUE
!               BY APPLYING A SEQUENCE OF GIVENS TRANSFORMATIONS.
!
                if ( abs(w(4_i4b*j))<tol ) then
!
                    j = j - 1_i4b
!
                    offi = j - 1_i4b
                    offv = w(4_i4b*offi+3_i4b)
!
                    do
!
                        if ( offv/tol*( w(4_i4b*offi-3_i4b) + w(4_i4b*offi-1_i4b) )<(real(n, stnd)**2)*tol ) then
!
                            offv = zero
!
                        else if ( offi>=1_i4b ) then
!
                            offv2 = w(4_i4b*offi+1_i4b)
                            nrm = offv + offv2

                            offi = offi - 1_i4b
                            offv = w(4_i4b*offi+3_i4b)*( offv/nrm )
!
                            cycle
!
                        end if
!
                        exit
!
                    end do
!
                    if ( offv==zero ) then
!
                        stat = stat + 1_i4b
                        offv = w(4_i4b*(j-1_i4b)+3_i4b)
!
                        do i=j-1_i4b, offi+1_i4b, -1_i4b
!
                            offv2 = w(4_i4b*i+1_i4b)
                            offv3 = w(4_i4b*(i-1_i4b)+3_i4b)
                            nrm   = offv + offv2
!
                            w(4_i4b*i+1_i4b)        = nrm
                            w(4_i4b*(i-1_i4b)+3_i4b)= offv3*( offv2/nrm )
!
                            offv = offv3*( offv/nrm )
!
                        end do
!
                        w(4_i4b*j+1_i4b)         = s2
                        w(4_i4b*(j-1_i4b)+3_i4b) = zero
!
                        if ( j<=2_i4b ) then
                            exit whila_iter
                        endif
!
                    else
!
!                       NO DEFLATION.
!
                        j = j + 1_i4b
!
                        exit whila_iter
!
                    end if
                else
!
!                   NO DEFLATION.
!
                    exit whila_iter
!
                end if
!
            end if
!
            pp = 1_i4b - pp
!
!           WHEN emin IS VERY SMALL CHECK FOR SPLITS.
!
            if ( pp==0_i4b .and. n0-i0>=3_i4b ) then
!
                if ( z(4_i4b*n0)<=tol2*qmax .or. z(4_i4b*n0-1_i4b)<=tol2*sigma ) then
!
                    splt   = i0 - 1_i4b
                    qmax   = z(4*i0-3_i4b)
                    emin   = z(4*i0-1_i4b)
                    oldemn = z(4*i0)
!
                    do i4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
                        if ( z(i4)<=tol2*z(i4-3_i4b) .or. z(i4-1_i4b)<=tol2*sigma ) then
                            z(i4-1_i4b) = -sigma
                            splt = i4/4_i4b
                            qmax = zero
                            emin = z(i4+3_i4b)
                            oldemn = z(i4+4_i4b)
                        else
                            qmax = max( qmax, z(i4+1_i4b) )
                            emin = min( emin, z(i4-1_i4b) )
                            oldemn = min( oldemn, z(i4) )
                        end if
!
                    end do
!
                    z(4_i4b*n0-1_i4b) = emin
                    z(4_i4b*n0)       = oldemn
!
                    i0 = splt + 1_i4b
!
                end if
!
            end if
!
        end do
!
        if ( iwhilb>nbig ) then
!
!           MAXIMUM NUMBER OF INNER ITERATIONS EXCEEDED.
!
            info = 2_i4b
!
            exit
!
        end if
!
    end do whila_iter
!
!   POST-PROCESSING.
!
    if ( iwhila>n+1_i4b ) then
!
!       MAXIMUM NUMBER OF OUTER ITERATIONS EXCEEDED.
!
        info = 3_i4b
!
    else if ( info==0_i4b .and. s2/=zero ) then
!
        call stqdsi( -s2, j, w(1_i4b), w(3_i4b), 4_i4b )
!
    end if
!
!
! END OF SUBROUTINE bd_aggdef2
! ____________________________
!      
    end subroutine bd_aggdef2
!
! ===========================================================================================
!
    subroutine stqdsi( sigma, n, q, e, inc )
!
! Purpose
! _______
!
!   STQDSI implements the differential form of the stationary
!   QD algorithm of Rutishauser with shift (dstqds) on the qd variables
!   associated with a bidiagonal matrix of size N for dqds algorithms
!   used to compute singular values of a bidiagonal matrix to high
!   relative accuracy.
!
!
! Arguments
! _________
!
!   SIGMA  (INPUT) real(stnd)
!          On entry, the shift.
!
!   N      (INPUT) integer(i4b)
!          On entry, the dimension of the bidiagonal matrix.
!
!   Q      (INPUT/OUTPUT) real(stnd), dimension(*)
!          On entry, Q contains the qd variable associated with
!          the diagonal of the bidiagonal matrix.
!          On exit, Q contains the qd variable after the dstqds
!          transform.
!
!   E      (INPUT/OUTPUT) real(stnd), dimension(*)
!          On entry, E contains the qd variable associated with
!          the subdiagonal of the bidiagonal matrix.
!          On exit, E contains the qd variable after the dstqds
!          transform.
!
!   INC    (INPUT) integer(i4b)
!          On entry, the increment between the elements of Q and E.
!          INC must be greater or equal to 1, but this is not check
!          in the subroutine.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the codes (e.g., STQDSI subroutine) given in reference (1).
!   For more details, see:
!
!   (1) Nakatsukasa, Y., Aishima, K., and Yamazaki, I., 2012:
!           dqds with aggressive early deflation.
!           SIAM Journal on Matrix Analysis Applications, Volume 33, No 1, pp. 22-51.
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
!   
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),               intent(in)    :: sigma
    real(stnd), dimension(*), intent(inout) :: q, e
!
    integer(i4b), intent(in)  :: n, inc
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: d, qd, qe
!
    integer(i4b) :: i, j
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
      i = 1_i4b
      d = -sigma
!
      do j=1_i4b, n-1_i4b
!
        qd = q(i)
        qe = e(i)
!
        q(i) = qd + d
        e(i) = ( qd/q(i) )*qe
!
        d = ( qe/q(i) )*d - sigma
!
        i = i + inc
!
      end do
!
      q(i) = q(i) + d
!
!
! END OF SUBROUTINE stqdsi
! ________________________
!
    end subroutine stqdsi
!
! ===========================================================================================
!
    subroutine bd_one_singval_bisect( d, e, singval, failure, scaling )
!
! Purpose
! _______
!
!   BD_ONE_SINGVAL_BISECT computes one singular value of a real n-by-n
!   (upper or lower) bidiagonal matrix B by a bisection algorithm.
!
!   The Singular Value Decomposition of a bidiagonal matrix B is:
!
!                           B = Q * S * P'
!
!   where S is a diagonal matrix with non-negative diagonal elements (the singular
!   values of B), and, Q and P are orthogonal matrices (P' denotes the transpose of P).
!
!
! Arguments
! _________
!
!   D            (INPUT) real(stnd), dimension(:)
!                On entry, D contains the diagonal elements of the bidiagonal
!                matrix B.
!
!   E            (INPUT) real(stnd), dimension(:)
!                On entry, E contains the off-diagonal elements of the bidiagonal
!                matrix whose singular values are desired. E(1) is arbitrary.
!
!                The size of E must verify: size( E ) = size( D ) . 
!
!   SINGVAL      (OUTPUT) real(stnd)
!                On exit, SINGVAL contains one singular value of B if FAILURE is equal to false.
!                On the other hand, if FAILURE is true, SINGVAL is set to zero.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection algorithm converged for
!                  the computed singular value to the desired accuracy;
!                - FAILURE = true  :  indicates that the singular value failed to converge
!                  or is not computed at all. This is generally caused by unexpectedly inaccurate
!                  arithmetic. In that case, SINGVAL is set to zero.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix B is scaled before computing
!                the singular value.
!
!                The default is to scale the bidiagonal matrix.
!
!
! Further Details
! _______________
!
!   BD_ONE_SINGVAL_BISECT computes the singular value of B by a bisection method (see the reference (1)
!   below, Sec.8.5 ).
!
!   A singular value is considered to be located if its square has been determined to lie in
!   an interval whose width is ULP * | B' * B | or less, where | B' * B | means the 1-norm of
!   the tridiagonal matrix B' * B ( B' means the transpose of B) and ULP is the machine precision
!   (distance from 1 to the next larger floating point number).
!
!   The bisection method is applied (implicitly) to the associated n-by-n symmetric tridiagonal matrix
!   B' * B  whose eigenvalues are the squares of the singular values of B by using the differential stationary
!   form of the qd algorithm of Rutishauser (see the reference (2) below, Sec.3.1 ).
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Fernando, K.V., 1998:
!          Accurately counting singular values of bidiagonal matrices and
!          eigenvalues of skew-symmetric tridiagonal matrices.
!          SIAM J. Matrix Anal. Appl., Vol. 20, no 2, pp.373-399.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
    use Num_Constants,   only : is_nan
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),  dimension(:) :: d, e
    real(stnd), intent(out)               :: singval
!
    logical(lgl), intent(in), optional :: scaling
    logical(lgl), intent(out)          :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: gamma, omega, tau, scalfac, maxd, maxe, maxde, maxbd, &
                  tnorm, atol1, rtol1, tmp1, tmp2, y, z, th, gu, gl
    real(stnd),  dimension(size(d)) :: d2, e2
!
    integer(i4b) :: n, num, nth, nz, ny, ngu, ngl, i, j, j1, j2
!
    logical(lgl) :: do_scaling
!
!
! PARAMETERS
! __________
!
!   relfac IS THE RELATIVE TOLERANCE. AN INTERVAL (a,b] LIES WITHIN "RELATIVE TOLERANCE" IF
!   b-a < relfac*ulp*max(|a|,|b|), WHERE "ulp" IS THE MACHINE PRECISION.
!
    real(stnd),  parameter :: relfac=1.95_stnd
!
!   A "FUDGE FACTOR" TO WIDEN THE UPPER BOUND OF THE GERSHGORIN INTERVAL. IDEALLY, A VALUE OF 1 SHOULD WORK,
!   BUT ON MACHINES WITH SLOPPY ARITHMETIC, THIS NEEDS TO BE LARGER. 
!   THE DEFAULT FOR PUBLICLY RELEASED VERSIONS SHOULD BE LARGE ENOUGH TO HANDLE THE WORST MACHINE AROUND.
!   NOTE THAT THIS PARAMETER HAS NO EFFECT ON ACCURACY OF THE SOLUTION.
!
    real(stnd),  parameter :: fudge=two
!
    character(len=*),  parameter :: name_proc='bd_one_singval_bisect'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure    = false
    do_scaling = true
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b),    &
                   int(size(e),i4b),    & 
                   name_proc )
!
    if ( present(scaling) ) then
        do_scaling = scaling
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=2_i4b ) then
!
        if ( n==1_i4b ) then
!
            singval = abs( d(1_i4b) )
!
        else if ( n==2_i4b ) then
!
            call las2( d(1_i4b), e(2_i4b), d(2_i4b), y, singval )
!
        end if
!
        return
!
    end if
!
    d2(1_i4b:n) =  abs( d(1_i4b:n) )
    maxd = maxval( d2(1_i4b:n) )
!
    e2(1_i4b)   = zero
    e2(2_i4b:n) =  abs( e(2_i4b:n) )
    maxe = maxval( e2(2_i4b:n) )
!
    maxde = maxval( d2(1_i4b:n-1_i4b)*e2(2_i4b:n) )
!
    maxbd = max( maxd, maxe )
!
    if ( maxbd<=zero ) then
!
        singval = zero
!
       return
!
    end if
!
!   GET MACHINE PARAMETERS AND SCALING FACTOR.
!
    rtol1 = ulp*relfac
!
    gamma  = sqrt( safmin )
!
    if ( do_scaling ) then
!
#ifdef _SCALE
        omega  = scale(one,maxexp-1_i4b)
#else
        omega  = machbase**(maxexp-1_i4b)
#endif
        tau    = sqrt( omega*gamma )
!
!       COMPUTE SCALING FACTOR TO ESTIMATE THE (TINY) SINGULAR VALUES
!       TO RELATIVELY HIGH ACCURACY AND TO AVOID OVERFLOW.
!
        j  = exponent( tau )
        j1 = j - exponent( maxbd ) - 1_i4b
        j2 = ( j - exponent( maxde ) - 1_i4b )/2_i4b
        j  = min( j1, j2 )
        scalfac = machbase**j
!
    else
!
        scalfac = one
!
    end if
!
!   SCALE AND SQUARE THE OFF DIAGONAL ELEMENTS OF THE GOLUB-KAHAN TRIDIAGONAL MATRIX.
!   IF ANY ELEMENT OF abs(d2(:)) OR abs(e2(:)) IS LESS THAN gamma/scalfac THAN THE CORRESPONDING
!   SCALED AND SQUARED ELEMENTS WILL UNDERFLOW TO zero AND HIGH RELATIVE ACCURACY WILL BE LOST.
!
    tmp1 = gamma/scalfac
!
    where( d2(1_i4b:n)>=tmp1 )
        d2(1_i4b:n) = (scalfac*d2(1_i4b:n))**2
    elsewhere
        d2(1_i4b:n) = safmin
    end where
!
    where( e2(2_i4b:n)>=tmp1 )
        e2(2_i4b:n) = (scalfac*e2(2_i4b:n))**2
    elsewhere
        e2(2_i4b:n) = safmin
    end where
!
!   COMPUTE THE UPPER GERSCHGORIN BOUND FOR EIGENVALUES OF B' * B.
!
    gu   = zero
    tmp1 = zero
!
    do i=1_i4b, n-1_i4b
!
        y    = d2(i) + e2(i)
        tmp2 = sqrt( d2(i) )*sqrt( e2(i+1_i4b) )
        gu   = max( gu, y+tmp1+tmp2 )
        tmp1 = tmp2
!
    end do
!
    y    = d2(n) + e2(n)
    gu   = max( gu, y+tmp1 )
!
    tnorm = gu
    gu    = gu + (real( n, stnd)*fudge + eight)*gu*ulp
    ngu   = 0_i4b
!
!   DEFINE ABSOLUTE TOLERANCE.
!
    tmp1 = ulp*tnorm
    atol1 = max( gamma, tmp1 )
!
!   FIRST GUESS FOR THE LOWER BOUND gl IF SINGULAR VALUES ARE WELL-SEPARATED.
!
    if ( n==2_i4b ) then
        gl = minval( abs(d(1_i4b:2_i4b)) )
    else
        gl = abs( d(2_i4b) )
    end if
!
    gl = (gl*scalfac)**2
!
!   CHECK QUALITY OF INITIAL ESTIMATE OF gl.
!
    if ( gl>zero ) then
!
        ngl  = 0_i4b
        tmp1 = one
        tmp2 = one
!
        do j=1_i4b, n
!
            tmp2 = gl + e2(j)*(tmp2/tmp1)
            tmp1 = d2(j) - tmp2
            if ( tmp1>=gamma ) then
                ngl = ngl + 1_i4b
            else
                tmp1 = min( tmp1, -gamma )
            end if
!
        end do
!
        if ( ngl==0_i4b ) then
!
            gl  = zero
            ngl = n
!
        end if
!
    else
!
        ngl = n
!
    end if
!
!   INITIALIZE THE LOWER AND UPPER BOUNDS FOR THE SQUARE OF THE SINGULAR VALUE.
!
    z  = gu
    nz = ngu
!
    y  = gl
    ny = ngl
!
    do
!
        tmp1 = abs( z - y )
        tmp2 = max( abs(z), abs(y) )
!
!       CHECK FOR CONVERGENCE.
!
        if ( tmp1<max(rtol1*tmp2,atol1) .or. nz>=ny ) exit
!
!       NEXT BISECTION STEP.
!
        th = ( y + z )*half
!
!       COMPUTE nth=COUNT(th)  THE NUMBER OF SINGULAR VALUES GREATER OR EQUAL
!       TO sqrt(th) BY STURM SEQUENCE.
!
        if ( th>zero ) then
!
            num  = 0_i4b
            tmp1 = one
            tmp2 = one
!
            do j=1_i4b, n
!
                tmp2 = th + e2(j)*(tmp2/tmp1)
                tmp1 = d2(j) - tmp2
!
                if ( tmp1>=gamma ) then
                    num = num + 1_i4b
                else
                    tmp1 = min( tmp1, -gamma )
                end if
!
            end do
!
        else
!
            num = n
!
        end if
!
!       INSURE THAT COUNT(th) IS MONOTONE.
!
        nth = min( max(num,nz),ny ) 
!
!       KEEP ONLY THE INTERVAL CONTAINING w S.T. COUNT(w)=1
!
        if ( nth<1_i4b ) then
!
            z  = th
            nz = nth
!
        else
!
            y  = th
            ny = nth
!
        end if
!
    end do        
!
!   SINGULAR VALUE FOUND.
!
    singval = min( max(( y + z )*half,y), z )
!
!   CHECK FOR CONVERGENCE.
!
    failure =  y>z .or. nz>=ny .or. singval>=gu .or. singval<gl
!
    tmp1 = sqrt( abs( singval ) )
!
!   UNSCALED THE SINGULAR VALUE.
!
    singval = tmp1/scalfac
!
#ifdef _F2003
    if ( ieee_support_datatype( singval ) ) then
!
        failure = .not.ieee_is_normal( singval )
!
    else
#endif
!
        failure = is_nan( singval )
!
#ifdef _F2003
    end if
#endif
!
    if ( failure ) then
!
        singval = zero
!
    end if
!
!
! END OF SUBROUTINE bd_one_singval_bisect
! _______________________________________
!
    end subroutine bd_one_singval_bisect
!
! =========================================================================================
!
    subroutine bd_one_singval_pwk( d, e, singval, failure, scaling )
!
! Purpose
! _______
!
!   BD_ONE_SINGVAL_PWK computes one singular value of a real n-by-n
!   (upper or lower) bidiagonal matrix B by the Pal-Walker-Kahan algorithm.
!
!   The Pal-Walker-Kahan method is applied to the associated n-by-n symmetric tridiagonal
!   matrix B * B'  whose eigenvalues are the squares of the singular values of B.
!
!   The Singular Value Decomposition of a bidiagonal matrix B is:
!
!                           B = Q * S * P'
!
!   where S is a diagonal matrix with non-negative diagonal elements (the singular
!   values of B), and, Q and P are orthogonal matrices (P' denotes the transpose of P).
!
!
! Arguments
! _________
!
!   D            (INPUT) real(stnd), dimension(:)
!                On entry, D contains the diagonal elements of the bidiagonal
!                matrix B.
!
!   E            (INPUT) real(stnd), dimension(:)
!                On entry, E contains the off-diagonal elements of the bidiagonal
!                matrix whose singular values are desired. E(1) is arbitrary.
!
!                The size of E must verify: size( E ) = size( D ) . 
!
!   SINGVAL      (OUTPUT) real(stnd)
!                On exit, SINGVAL contains one singular value of B if FAILURE is equal to false.
!                On the other hand, if FAILURE is true, SINGVAL is set to zero.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the Pal-Walker-Kahan algorithm
!                  converged for the computed singular value to the desired accuracy;
!                - FAILURE = true  :  indicates that the singular value failed to converge to
!                  the desired accuracy or is not computed at all. In that case, SINGVAL is
!                  set to zero.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix B is scaled before computing
!                the singular value.
!
!                The default is to scale the bidiagonal matrix.
!
!
! Further Details
! _______________
!
!   For further details on the Pal-Walker-Kahan algorithm for computing eigenvalues
!   of a symmetric tridiagonal matrix, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Parlett, B.N., 1998:
!           The Symmetric Eigenvalue Problem.
!           Revised edition, SIAM, Philadelphia.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
    use Num_Constants,   only : is_nan
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),  dimension(:) :: d, e
    real(stnd), intent(out)               :: singval
!
    logical(lgl), intent(in), optional :: scaling
    logical(lgl), intent(out)          :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: eps, eps2, x, y, f, g, beta, p, r, sn, cs, b, oldcs, oldg, alpha
    real(stnd),  dimension(size(d)) :: d2, e2, diag, sub
!
    integer(i4b) :: n, i, l, niter, maxqrs
!
    logical(lgl) :: do_scaling
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_one_singval_pwk'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure    = false
    do_scaling = true
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b),    &
                   int(size(e),i4b),    & 
                   name_proc )
!
    if ( present(scaling) ) then
        do_scaling = scaling
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=2_i4b ) then
!
        if ( n==1_i4b ) then
!
            singval = abs( d(1_i4b) )
!
        else if ( n==2_i4b ) then
!
            call las2( d(1_i4b), e(2_i4b), d(2_i4b), x, singval )
!
        end if
!
        return
!
    end if
!
!   DETERMINE THE TOLERANCE FOR CONVERGENCE OF THE QR ITERATIONS.
!
    eps  = fudge_eps*macheps
    eps2 = eps*eps
!
!   SET COUNTER AND THRESHOLD FOR THE NUMBER OF QR ITERATIONS.
!
    niter  = 0_i4b
    maxqrs = n*10_i4b
!
!   MAKE A LOCAL COPY OF THE BIDIAGONAL MATRIX.
!
    d2(1_i4b:n) = d(1_i4b:n)
!
    e2(1_i4b)   = zero
    e2(2_i4b:n) = e(2_i4b:n)
!
    x = maxval( abs( d2(1_i4b:n) ) )
    y = maxval( abs( e2(2_i4b:n) ) )
    f = max( x, y )
!
!   COMPUTE ONE SINGULAR VALUE BY THE PAL-WALKER-KAHAN ALGORITHM.
!
    if ( f/=zero ) then
!
!       SCALE THE BIDIAGONAL MATRIX IF REQUIRED.
!
        if ( do_scaling ) then
!
            d2(1_i4b:n) = d2(1_i4b:n)/f
            e2(2_i4b:n) = e2(2_i4b:n)/f
!
        end if 
!
!       COMPUTE THE ASSOCIATED TRIDIAGONAL MATRIX B*B'.
!
        diag(1_i4b:n-1_i4b) = d2(1_i4b:n-1_i4b)*d2(1_i4b:n-1_i4b) + e2(2_i4b:n)*e2(2_i4b:n)
        diag(n)             = d2(n)*d2(n)
!
        sub(1_i4b)   = zero
        sub(2_i4b:n) = d2(2_i4b:n)*e2(2_i4b:n)
!
!       STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!       THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL IN ORDER
!       TO MAINTAIN ACCURACY OF THE QR ITERATIONS.
!
        if ( reverse( diag(1_i4b:n) ) ) then
!
            diag(1_i4b:n) = diag(n:1_i4b:-1_i4b)
            sub(2_i4b:n)  = sub(n:2_i4b:-1_i4b)
!
        end if
!
!       SQUARE THE OFF-DIAGONAL ELEMENTS.
!
        sub(2_i4b:n) = sub(2_i4b:n)*sub(2_i4b:n)
!
        l = 1_i4b
!
        do
!
!           CUTOFF FOR CONVERGENCE FAILURE.
!
            if ( niter>=maxqrs .and. l/=n ) then
!
                failure = true
                l       = n
!
            end if
!
!           EIGENVALUE FOUND.
!
            if ( l==n ) exit
!
!           COMPUTE EIGENVALUE WITH PAL-WALKER-KAHAN ALGORITHM.
!
            niter = niter + 1_i4b
!
!           FORM WILKINSON SHIFT.
!
            g = sqrt( abs( sub(n) ) )
            beta = ( diag(n-1_i4b) - diag(n) )/( two*g )
!
            if ( abs(beta)>one ) then
!
                p = one/beta
                r = beta*sqrt( one + p*p )
!
            else
!
                r = sign( sqrt( one + beta*beta ), beta )
!
            end if
!
            beta = diag(n) - ( g/( beta + r ) )
!
            sn = zero
            cs = one
            g  = diag(l) - beta
            p  = g*g
!
!           INNER LOOP.
!
            do i = l, n-1_i4b
!
                b       = sub(i+1_i4b)
                r       = p + b
                sub(i)  = sn*r
                oldcs   = cs
                cs      = p/r
                sn      = b/r
                oldg    = g
                alpha   = diag(i+1_i4b)
                g       = cs*( alpha - beta ) - sn*oldg
                diag(i) = oldg + ( alpha - g )
!
                if ( cs/=zero ) then
!
                    p = ( g*g )/cs
!
                else
!
                    p = oldcs*b
!
                end if
!
            end do
!
            sub(n)  = sn*p
            diag(n) = beta + g
            sub(l)  = zero
!
!           TEST FOR SPLITTING : LOOK FOR SMALL OFF-DIAGONAL ELEMENT.
!
            do l = n, 2_i4b, -1_i4b
!
                g = abs( sub(l) )
!
                if ( g <= eps2*abs( diag(l)*diag(l-1_i4b) ) + safmin ) exit
!
            end do
!
        end do
!
        if ( .not.failure .and. diag(n)>=zero ) then
!
            singval = sqrt( abs( diag(n) ) )
!
            if ( do_scaling ) then
!
                singval = singval*f
!
            end if
!
!
#ifdef _F2003
            if ( ieee_support_datatype( singval ) ) then
!
                failure = .not.ieee_is_normal( singval )
!
            else
#endif
!
                failure = is_nan( singval )
!
#ifdef _F2003
            end if
#endif
!
            if ( failure ) then
!
                singval = zero
!
            end if
!
        else
!
            singval = zero
!
            failure = true
!
        end if
!
    else
!
        singval = zero
!
    end if
!
!
! END OF SUBROUTINE bd_one_singval_pwk
! ____________________________________
!
    end subroutine bd_one_singval_pwk
!
! =========================================================================================
!
    subroutine bd_one_singval_lasq1( d, e, singval, failure, scaling )
!
! Purpose
! _______
!
!   BD_ONE_SINGVAL_LASQ1 computes one singular value of a real n-by-n bidiagonal
!   matrix BD (with diagonal D and off-diagonal E) by the differential quotient
!   difference with shifts (dqds) algorithm.
!
!   The singular value is computed to high relative accuracy, in the absence of
!   denormalization, underflow and overflow. The algorithm was first presented
!   in the reference (1) and the present implementation is described in the
!   references (2) and (3).
!
!
! Arguments
! _________
!
!   D            (INPUT) real(stnd), dimension(:)
!                On entry, D contain the diagonal elements of the bidiagonal
!                matrix BD whose SVD is desired.
!
!   E            (INPUT) real(stnd), dimension(:)
!                On entry, E(2:) contain the off-diagonal elements of the bidiagonal
!                matrix BD whose singular values are desired. E(1) is arbitrary.
!
!                The size of E must verify: size( E ) = size( D ) = n . 
!
!   SINGVAL      (OUTPUT) real(stnd)
!                On exit, SINGVAL contains one singular value of BD if FAILURE is false.
!                On the other hand, if FAILURE is true, SINGVAL is set to zero.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit;
!                - FAILURE = true  :  indicates that the dqds algorithm did not converge and that
!                  the maximum number of iterations in the inner loop of the dqds algorithm
!                  is exceeded or that the singular value cannot be estimated due to denormalization,
!                  underflow or overflow.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix BD is scaled before computing
!                the singular values.
!
!                The default is to scale the bidiagonal matrix.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the DLASQ1 subroutine in LAPACK version 3.12.0.
!
!   This subroutine is a variant of the dqds algorithm described
!   in the references (2) and (3).
!
!   For further details on the dqds algorithm, see:
!
!   (1) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (2) Parlett, B.N., and Marques, O.A., 2000:
!           An implementation of the dqds algorithm (positive case).
!           Linear Algebra Appl., Volume 309, pp. 217-259.
!
!   (3) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
    use Utilities, only : assert_eq, lascl
!   
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:) :: d, e
    real(stnd), intent(out)              :: singval
!
    logical(lgl), intent(out) :: failure
    logical(lgl), intent(in), optional :: scaling
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: scale, sigmn, sigmx, maxbd
    real(stnd),  dimension(4*size(d)) :: work
!
    integer(i4b) :: n, info
!
    logical(lgl) :: do_scaling
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_one_singval_lasq1'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure    = false
    do_scaling = true
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b)  ,    &
                   int(size(e),i4b),      &
                   name_proc )
!
    if ( present(scaling) ) then
        do_scaling = scaling
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=2_i4b ) then
!
        if ( n==1_i4b ) then
!
            singval = abs( d(1_i4b) )
!
        else if ( n==2_i4b ) then
!
            call las2( d(1_i4b), e(2_i4b), d(2_i4b), sigmn, singval )
!
        end if
!
        return
!
    end if
!
!   MAKE A COPY OF THE BIDIAGONAL MATRIX.
!
    work(2_i4b*n+1_i4b:3_i4b*n) = abs( d(1_i4b:n) )
    work(3_i4b*n+1_i4b:4_i4b*n-1_i4b) = abs( e(2_i4b:n) )
!
    sigmx = maxval( work(3_i4b*n+1_i4b:4_i4b*n-1_i4b) )
!
    if ( sigmx==zero ) then
!
!       EARLY RETURN IF sigmx IS ZERO AS MATRIX IS ALREADY DIAGONAL.
!
        singval = maxval( work(2_i4b*n+1_i4b:3_i4b*n) )
!
        return
!
    end if
!
!   SCALE THE BIDIAGONAL MATRIX IF REQUIRED.
!   NOTE THAT SQUARING THE INPUT DATA MAKES
!   SCALING BY A POWER OF THE RADIX POINTLESS.
!  
    if ( do_scaling ) then
!
!       ESTIMATE THE LARGEST SINGULAR VALUE.
!
        sigmn = maxval( work(2_i4b*n+1_i4b:3_i4b*n) )
        maxbd = max( sigmn, sigmx )
!
        scale = sqrt( ulp/safmin )
!
        call lascl(  work(2_i4b*n+1_i4b:4_i4b*n-1_i4b), maxbd, scale )
!  
    end if
!  
!   COPY d AND e INTO work(1:2*n-1) IN THE Z FORMAT.
!  
    work(1_i4b:2_i4b*n-1_i4b:2_i4b) = work(2_i4b*n+1_i4b:3_i4b*n)
    work(2_i4b:2_i4b*n-1_i4b:2_i4b) = work(3_i4b*n+1_i4b:4_i4b*n-1_i4b) 
!  
!   SQUARE THE ELEMENTS OF THE BIDIAGONAL MATRIX (E.G., COMPUTE THE Q'S AND E'S).
!  
    work(1_i4b:2_i4b*n-1_i4b) = work(1_i4b:2_i4b*n-1_i4b)**2
    work(2_i4b*n:4_i4b*n)     = zero
!  
!   PERFORM DQDS ITERATIONS TO COMPUTE ONE EIGENVALUE OF THE ASSOCIATED
!   SYMMETRIC POSITIVE DEFINITE TRIDIAGONAL MATRIX.
!  
    call bd_one_singval_lasq2( n, work, sigmx, info )
!  
    if ( info==0_i4b ) then
!  
!       COMPUTE THE SINGULAR VALUE FROM THE EIGENVALUE.
!  
        singval = sqrt( sigmx )
!      
!       UNSCALED IT IF REQUIRED.
!      
        if ( do_scaling ) then
            call lascl(  singval, scale, maxbd )
        end if
!      
    else
!      
        failure = true
!      
        singval = zero
!      
    end if
!
!
! END OF SUBROUTINE bd_one_singval_lasq1
! ______________________________________
!      
    end subroutine bd_one_singval_lasq1
!
! ===========================================================================================
!
    subroutine bd_one_singval_lasq2( n, z, eigval, info )
!
! Purpose
! _______
!
!   BD_ONE_SINGVAL_LASQ2 computes one eigenvalue of the N-by-N symmetric positive
!   definite tridiagonal matrix associated with the qd array Z to high
!   relative accuracy, in the absence of denormalization, underflow and overflow.
!
!   To see the relation of Z to the tridiagonal matrix, let L be a
!   unit lower bidiagonal matrix with subdiagonals Z(2,4,6,,..) and
!   let U be an upper bidiagonal matrix with 1's above and diagonal
!   Z(1,3,5,,..). The tridiagonal is L*U or, if you prefer, the
!   symmetric tridiagonal to which it is similar.
!
!
! Arguments
! _________
!
!   N        (INPUT) integer(i4b)
!            The number of rows and columns in the tridiagonal matrix.
!            It is assumed that N>2, but this is not checked in the subroutine.
!            This assumption is automatically verified when BD_ONE_SINGVAL_LASQ2
!            is called from BD_ONE_SINGVAL_LASQ1 subroutine.
!
!   Z        (INPUT/OUTPUT) real(stnd), dimension(4*n)
!            On entry Z(:2*N-1) holds the qd array. The other part of Z can be arbitrary.
!            Z is used as workspace and is overwritten.
!
!            It is assumed that all the elements of Z(:2*N-1) are greater or equal to zero on
!            input, but this is not checked in the subroutine.
!            This assumption is automatically verified when BD_ONE_SINGVAL_LASQ2
!            is called from BD_ONE_SINGVAL_LASQ1 subroutine.
!
!   EIGVAL   (OUTPUT) real(stnd)
!            On exit, EIGVAL contains one eigenvalue of the N-by-N symmetric positive
!            definite tridiagonal matrix associated with the qd array Z if INFO
!            is equal to 0. Otherwise, EIGVAL is set to EIGVAL_FAILURE parameter
!            defined below.
!
!   INFO     (OUTPUT) integer(i4b)
!            On exit:
!
!            - INFO = 0 :  indicates successful exit;
!            - INFO = 1 :  indicates a failure because
!              a split was marked by a positive value in E;
!            - INFO = 2 :  indicates that the algorithm did not converge
!              and that the maximum number of iterations in the inner loop
!              of the dqds algorithm is exceeded;
!            - INFO = 3 :  indicates that the termination criterion of the outer while
!              loop of the algorithm was not met (e.g. the program created more than
!              N unreduced blocks).
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the DLASQ2 subroutine in LAPACK version 3.12.0.
!
!   Local Variables: I0:N0 defines a current unreduced segment of Z.
!   The shifts are accumulated in SIGMA. Iteration count is in ITER.
!   Ping-pong is controlled by PP (alternates between 0 and 1).
!
!   This subroutine is part of the dqds algorithm described
!   in the references (2) and (3).
!
!   For further details on the dqds algorithm, see:
!
!   (1) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (2) Parlett, B.N., and Marques, O.A., 2000:
!           An implementation of the dqds algorithm (positive case).
!           Linear Algebra Appl., Volume 309, pp. 217-259.
!
!   (3) Li, S., Gu, M., and  Parlett, B. N., 2014:
!           An Improved DQDS Algorithm.
!           SIAM Journal on Scientific Computing, Volume 36, No 3, C290-C308.
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
    use Num_Constants, only : test_ieee
!   
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: n
    integer(i4b), intent(out) :: info
!
    real(stnd), intent(inout), dimension(4*n) :: z
    real(stnd), intent(out)                   :: eigval
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: d, dee, deemin, desig, dmn, dmn1, dmn2, dn,  &
                  dn1, dn2, emax, emin, g, oldemn, qmax, qmin, &
                  sigma, temp, tol, tol2
!
    integer(i4b) :: i0, i4, ipn4, iwhila, iwhilb,   &
                    k, kmin, n0, nbig, pp, splt, ttype
!
    logical(lgl) :: ieee
!
!
! PARAMETERS
! __________
!
!   FAILURE INDICATOR FOR THE EIGENVALUE.
!
    real(stnd),  parameter :: eigval_failure = -c100
!
!   MAXIMUM NUMBER OF ITERATIONS IN THE INNER LOOP OF THE DQDS ALGORITHM.
!
    integer(i4b) :: maxiter=10_i4b    
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   INPUT ARGUMENTS ARE NOT TESTED AS IT IS ASSUMED THAT bd_one_singval_lasq2
!   IS CALLED FROM bd_one_singval_lasq1 IN WHICH CASE n>2 .
!
    info = 0_i4b
!
!   SET FAILURE INDICATOR FOR THE EIGENVALUE.
!
    eigval = eigval_failure
!
!   COMPUTE TOLERANCES.
!
    tol  = ulp*fudge_lasq
    tol2 = tol**2
!
!   TEST IF IEEE ARITHMETIC IS SUPPORTED.
!
    ieee = test_ieee()
!    ieee = false
!
    z(2_i4b*n) = zero
!
!   REARRANGE DATA FOR LOCALITY: Z=(Q1,QQ1,E1,EE1,Q2,QQ2,E2,EE2,...).
!
    do k=2_i4b*n, 2_i4b, -2_i4b
!
        z(2_i4b*k)       = zero
        z(2_i4b*k-1_i4b) = z(k)
        z(2_i4b*k-2_i4b) = zero
        z(2_i4b*k-3_i4b) = z(k-1_i4b)
!
    end do
!
    i0 = 1_i4b
    n0 = n
!
!   REVERSE THE QD-ARRAY, IF WARRANTED.
!
    if ( c1_5*z( 4_i4b*i0-3_i4b )<z( 4*n0-3_i4b ) ) then
!
        ipn4 = 4_i4b*( i0 + n0 )
!
        do i4=4_i4b*i0, 2_i4b*( i0+n0-1_i4b ), 4_i4b
!
            temp = z(i4-3_i4b)
            z(i4-3_i4b) = z(ipn4-i4-3_i4b)
            z(ipn4-i4-3_i4b) = temp
!
            temp = z(i4-1_i4b)
            z(i4-1_i4b) = z(ipn4-i4-5_i4b)
            z(ipn4-i4-5_i4b) = temp
!
        end do
!
    end if
!
!   INITIAL SPLIT CHECKING VIA DQD AND LI'S TEST.
!
    pp = 0_i4b
!
    do k = 1_i4b, 2_i4b
!
        d = z(4_i4b*n0+pp-3_i4b)
!
        do i4=4_i4b*( n0-1_i4b ) + pp, 4_i4b*i0 + pp, -4_i4b
!
            if ( z(i4-1_i4b)<=tol2*d ) then
                z(i4-1_i4b) = -zero
                d = z(i4-3_i4b)
            else
                d = z(i4-3_i4b)*( d/( d + z(i4-1_i4b) ) )
            end if
!
        end do
!
!       DQD MAPS Z TO ZZ PLUS LI'S TEST.
!
        d = z(4_i4b*i0+pp-3_i4b)
!
        do i4=4_i4b*i0 + pp, 4_i4b*( n0-1_i4b ) + pp, 4_i4b
!
            z(i4-2_i4b*pp-2_i4b) = d + z(i4-1_i4b)
!
            if ( z(i4-1_i4b)<=tol2*d ) then
               z(i4-1_i4b) = -zero
               z(i4-2_i4b*pp-2_i4b) = d
               z(i4-2_i4b*pp) = zero
               d = z(i4+1_i4b)
            else if ( safmin*z(i4+1_i4b)<z(i4-2_i4b*pp-2_i4b) .and. safmin*z(i4-2_i4b*pp-2_i4b)<z(i4+1_i4b) ) then
               temp = z(i4+1_i4b)/z(i4-2_i4b*pp-2_i4b)
               z(i4-2_i4b*pp) = z(i4-1_i4b)*temp
               d = d*temp
            else
               z(i4-2_i4b*pp) = z(i4+1_i4b)*(z( i4-1_i4b)/z(i4-2_i4b*pp-2_i4b) )
               d = z(i4+1_i4b)*( d/z(i4-2_i4b*pp-2_i4b) )
            end if
!
        end do
!
        z(4_i4b*n0-pp-2_i4b) = d
!
!       PREPARE FOR THE NEXT ITERATION ON k.
!
        pp = 1_i4b - pp
!
   end do
!
!  INITIALISE VARIABLES TO PASS TO bd_one_singval_lasq3.
!
    ttype = 0_i4b
    dmn1 = zero
    dmn2 = zero
    dn    = zero
    dn1   = zero
    dn2   = zero
    g     = zero
!
    do iwhila=1_i4b, n + 1_i4b
!
!       CHECK IF ALL EIGENVALUES HAVE BEEN FOUND.
!
        if ( n0<1_i4b )  exit
!
!       WHILE ARRAY UNFINISHED DO.
!
!       e(n0) HOLDS THE VALUE OF sigma WHEN SUBMATRIX IN i0:n0
!       SPLITS FROM THE REST OF THE ARRAY, BUT IS NEGATED.
!
        desig = zero
!
        if ( n0==n ) then
            sigma = zero
        else
            sigma = -z(4_i4b*n0-1_i4b)
        end if
!
        if ( sigma<zero ) then
!
            info = 1_i4b
!
            exit
!
        end if
!
!       FIND LAST UNREDUCED SUBMATRIX'S TOP INDEX i0, FIND qmax AND
!       emin. FIND GERSHGORIN-TYPE BOUND IF Q'S MUCH GREATER THAN E'S.
!
        emax = zero
        if ( n0>i0 ) then
            emin = abs( z(4_i4b*n0-5_i4b) )
        else
            emin = zero
        end if
!
        qmin = z(4_i4b*n0-3_i4b)
        qmax = qmin
!
        do i4=4_i4b*n0, 8_i4b, -4_i4b
!
            if ( z(i4-5_i4b)<=zero ) exit
!
            if( qmin>=four*emax ) then
               qmin = min( qmin, z(i4-3_i4b) )
               emax = max( emax, z(i4-5_i4b) )
            end if
!
            qmax = max( qmax, z(i4-7_i4b)+z(i4-5_i4b) )
            emin = min( emin, z(i4-5_i4b) )
!
        end do
!
        if ( i4<8_i4b ) then
            i4 = 4_i4b
        end if
!
        i0 = i4/4_i4b
        pp = 0_i4b
!
        if ( n0-i0>1_i4b ) then
!
            dee = z(4_i4b*i0-3_i4b)
            deemin = dee
            kmin = i0
!
            do i4=4_i4b*i0+1_i4b, 4_i4b*n0-3_i4b, 4_i4b
!
                dee = z(i4)*( dee/( dee + z(i4-2_i4b) ) )
!
                if ( dee<=deemin ) then
                    deemin = dee
                    kmin = ( i4 + 3_i4b )/4_i4b
                end if
!
            end do 
!
            if ( (kmin-i0)*2_i4b<n0-kmin .and. deemin<=half*z(4_i4b*n0-3_i4b) ) then
!
                ipn4 = 4_i4b*( i0 + n0 )
                pp = 2_i4b
!
                do i4=4_i4b*i0, 2_i4b*( i0+n0-1_i4b ), 4_i4b
!
                    temp = z(i4-3_i4b)
                    z(i4-3_i4b) = z(ipn4-i4-3_i4b)
                    z(ipn4-i4-3_i4b) = temp
!
                    temp = z(i4-2_i4b)
                    z(i4-2_i4b) = z(ipn4-i4-2_i4b)
                    z(ipn4-i4-2_i4b) = temp
!
                    temp = z(i4-1_i4b)
                    z(i4-1_i4b) = z(ipn4-i4-5_i4b)
                    z(ipn4-i4-5_i4b) = temp
!
                    temp = z(i4)
                    z(i4) = z(ipn4-i4-4_i4b)
                    z(ipn4-i4-4_i4b) = temp
!
                end do
!
            end if
!
        end if
!
!       PUT -(INITIAL SHIFT) INTO dmn.
!
        dmn = -max( zero, qmin-two*sqrt( qmin )*sqrt( emax ) )
!
!       NOW i0:n0 IS UNREDUCED.
!       pp = 0 FOR PING, pp = 1 FOR PONG.
!       pp = 2 INDICATES THAT FLIPPING WAS APPLIED TO THE Z ARRAY AND
!              AND THAT THE TESTS FOR DEFLATION UPON ENTRY IN bd_one_singval_lasq3
!              SHOULD NOT BE PERFORMED.
!
        nbig = maxiter*( n0 - i0 + 1_i4b )
!
        do iwhilb = 1_i4b, nbig
!
            if ( i0>n0 ) exit
!
!           WHILE SUBMATRIX UNFINISHED TAKE A GOOD DQDS STEP.
!
            call bd_one_singval_lasq3( i0, n0, z, eigval, pp, sigma, desig, qmax, ieee, &
                                       ttype, dmn, dmn1, dmn2, dn, dn1, dn2, g )
!
            if ( eigval/=eigval_failure ) then
                return
            end if
!
            pp = 1_i4b - pp
!
!           WHEN emin IS VERY SMALL CHECK FOR SPLITS.
!
            if ( pp==0_i4b .and. n0-i0>=3_i4b ) then
!
                if ( z(4_i4b*n0)<=tol2*qmax .or. z(4_i4b*n0-1_i4b)<=tol2*sigma ) then
!
                    splt = i0 - 1_i4b
                    qmax = z( 4*i0-3_i4b )
                    emin = z( 4*i0-1_i4b )
                    oldemn = z( 4*i0 )
!
                    do i4=4_i4b*i0, 4_i4b*( n0-3_i4b ), 4_i4b
!
                        if ( z(i4)<=tol2*z(i4-3_i4b) .or. z(i4-1_i4b)<=tol2*sigma ) then
                            z(i4-1_i4b) = -sigma
                            splt = i4/4_i4b
                            qmax = zero
                            emin = z(i4+3_i4b)
                            oldemn = z(i4+4_i4b)
                        else
                            qmax = max( qmax, z(i4+1_i4b) )
                            emin = min( emin, z(i4-1_i4b) )
                            oldemn = min( oldemn, z(i4) )
                        end if
!
                    end do
!
                    z(4_i4b*n0-1_i4b) = emin
                    z(4_i4b*n0) = oldemn
!
                    i0 = splt + 1_i4b
!
                end if
!
            end if
!
        end do
!
        if ( iwhilb>nbig ) then
!
            info = 2_i4b
!
            exit 
!
        end if
!
    end do
!
!   EIGENVALUE FOUND IF info=0 AND eigval/=eigval_failure.
!
!   POST-PROCESSING.
!
    if ( iwhila>n+1_i4b ) then
!
!       MAXIMUM NUMBER OF OUTER ITERATIONS EXCEEDED.
!
        info = 3_i4b
!
     end if
!
!
! END OF SUBROUTINE bd_one_singval_lasq2
! ______________________________________
!
    end subroutine bd_one_singval_lasq2
!
! ===========================================================================================
!
    subroutine bd_one_singval_lasq3( i0, n0, z, eigval, pp, sigma, desig, qmax, ieee, &
                                     ttype, dmin, dmin1, dmin2, dn, dn1, dn2, g )
!
! Purpose
! _______
!
!   BD_ONE_SINGVAL_LASQ3 checks for deflation, computes a shift (tau) and calls dqds.
!   In case of failure it changes shifts, and tries again until output is positive.
!
!
! Arguments
! _________
!
!
!   I0      (INPUT) integer(i4b)
!           On entry, I0 is the first index.
!
!   N0      (INPUT/OUTPUT) integer(i4b)
!           On entry, N0 is the last index.
!           On exit, N0 is upddated with the new last index of the segment.
!
!   Z       (INPUT/OUTPUT) real(stnd), dimension(4*n)
!           Z holds the qd array of size 4*n where n is the dimension of the
!           associated symmetric positive definite tridiagonal matrix.
!
!   EIGVAL  (INPUT/OUTPUT) real(stnd)
!           On exit, EIGVAL contains one eigenvalue of the n-by-n symmetric positive
!           definite tridiagonal matrix associated with the qd array Z if INFO
!           is equal to 0. Otherwise, EIGVAL is not overwritten.
!
!   PP      (INPUT/OUTPUT) integer(i4b)
!           PP=0 for ping, PP=1 for pong.
!           PP=2 indicates that flipping was applied to the Z array
!           and that the initial tests for deflation should not be
!           performed.
!
!   SIGMA   (INPUT/OUTPUT) real(stnd)
!           SIGMA is the sum of shifts used in current segment.
!
!   DESIG   (INPUT/OUTPUT) real(stnd)
!           DESIG is the lower order part of SIGMA.
!
!   QMAX    (INPUT/OUTPUT) real(stnd)
!           QMAX is the maximum value of q.
!
!   IEEE    (INPUT) logical(lgl)
!           On entry, logical flag for IEEE or non IEEE arithmetic (IEEE is passed to BD_ONE_SINGVAL_LASQ5).
!
!   TTYPE   (INPUT/OUTPUT) integer(i4b)
!           On entry, the preceding shift type.
!           On exit, the new shift type.
!
!   DMIN    (INPUT/OUTPUT) real(stnd)
!           Minimum value of D.
!
!   DMIN1   (INPUT/OUTPUT) real(stnd)
!           Minimum value of D, excluding D(N0).
!
!   DMIN2   (INPUT/OUTPUT) real(stnd)
!           Minimum value of D, excluding D(N0) and D(N0-1).
!
!   DN      (INPUT/OUTPUT) real(stnd)
!           On entry, D(N0), the last value of D. 
!
!   DN1     (INPUT/OUTPUT) real(stnd)
!           On entry, D(N0-1).
!
!   DN2     (INPUT/OUTPUT) real(stnd)
!           On entry, D(N0-2).
!
!   G       (INPUT/OUTPUT) real(stnd)
!           G is passed as an argument in order to save its value between
!           calls to BD_ONE_SINGVAL_LASQ3.
!
!   SIGMA, DESIG, QMAX, DMIN, DMIN1, DMIN2, DN, DN1, DN2, G and TTYPE are passed as arguments in order
!   to save their values between calls to BD_ONE_SINGVAL_LASQ3.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the DLASQ3 subroutine from LAPACK version 3.12.0.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
    use Num_Constants, only : is_nan
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: qmax, sigma, desig, dmin, dmin1, dmin2, dn, dn1, dn2, g, eigval
    real(stnd), intent(inout), dimension(*) :: z
!
    integer(i4b), intent(in)    :: i0
    integer(i4b), intent(inout) :: n0, pp, ttype
!
    logical(lgl), intent(in) :: ieee
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tau, s, t, temp, tol, tol2
!
    integer(i4b) :: ipn4, j4, n0in, nn
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   COMPUTE TOLERANCES AND SAVE INITIAL VALUE OF n0 .
!
    n0in = n0
!
    tol = ulp*fudge_lasq
    tol2 = tol**2
!
!   CHECK FOR DEFLATION.
!
    do
!
        if ( n0<=i0 ) then
!
            if ( n0==i0 ) then
!
!               1 EIGENVALUE FOUND.
!
                eigval = z( 4_i4b*n0+pp-3_i4b ) + sigma
!
            end if
!
            return
!
        end if
!
        nn = 4_i4b*n0 + pp
!
        if ( n0/=i0+1_i4b ) then
!
!           CHECK WHETHER e(n0-1) IS NEGLIGIBLE, 1 EIGENVALUE FOUND.
!
            if ( z( nn-5_i4b )<=tol2*( sigma+z( nn-3_i4b ) ) .or. z( nn-2_i4b*pp-4_i4b )<=tol2*z( nn-7_i4b ) ) then
!
                eigval = z( 4_i4b*n0+pp-3_i4b ) + sigma
!
                return
!
            end if
!
!           CHECK  WHETHER e(n0-2) IS NEGLIGIBLE, 2 EIGENVALUES FOUND.
!
            if ( z( nn-9_i4b )>tol2*sigma .and. z( nn-2_i4b*pp-8_i4b )>tol2*z( nn-11_i4b ) ) exit
!
        end if
!
!       2 EIGENVALUES FOUND.
!
        if ( z( nn-3_i4b )>z( nn-7_i4b ) ) then
!
            s = z( nn-3_i4b )
            z( nn-3_i4b ) = z( nn-7_i4b )
            z( nn-7_i4b ) = s
!
        end if
!
        t = half*( ( z( nn-7_i4b )-z( nn-3_i4b ) )+z( nn-5_i4b ) )
!
        if ( z( nn-5_i4b )>z( nn-3_i4b )*tol2 .and. t/=zero ) then
!
            s = z( nn-3_i4b )*( z( nn-5_i4b )/t )
!
            if ( s<=t ) then
                s = z( nn-3_i4b )*( z( nn-5_i4b )/( t*( one+sqrt( one+s/t ) ) ) )
            else
                s = z( nn-3_i4b )*( z( nn-5_i4b )/( t+sqrt( t )*sqrt( t+s ) ) )
            end if
!
            t = z( nn-7_i4b ) + ( s+z( nn-5_i4b ) )
!
            z( nn-3_i4b ) = z( nn-3_i4b )*( z( nn-7_i4b )/t )
            z( nn-7_i4b ) = t
!
        end if
!
        t = z( nn-7_i4b ) + sigma
        s = z( nn-3_i4b ) + sigma
!
        eigval = max( t, s )
!
        return
!
    end do
!
    if ( pp==2_i4b ) pp = 0_i4b
!
!   REVERSE THE QD-ARRAY, IF WARRANTED.
!
    if ( dmin<=zero .or. n0<n0in ) then
!
        if ( c1_5*z( 4_i4b*i0+pp-3_i4b )<z( 4_i4b*n0+pp-3_i4b ) ) then
!
            ipn4 = 4_i4b*( i0+n0 )
!
            do j4=4_i4b*i0, 2_i4b*( i0+n0-1_i4b ), 4_i4b
!
               temp = z( j4-3_i4b )
               z( j4-3_i4b ) = z( ipn4-j4-3_i4b )
               z( ipn4-j4-3_i4b ) = temp
!
               temp = z( j4-2_i4b )
               z( j4-2_i4b ) = z( ipn4-j4-2_i4b )
               z( ipn4-j4-2_i4b ) = temp
!
               temp = z( j4-1_i4b )
               z( j4-1_i4b ) = z( ipn4-j4-5_i4b )
               z( ipn4-j4-5_i4b ) = temp
!
               temp = z( j4 )
               z( j4 ) = z( ipn4-j4-4_i4b )
               z( ipn4-j4-4_i4b ) = temp
!
            end do
!
            if ( n0-i0<=4_i4b ) then
                z( 4_i4b*n0+pp-1_i4b ) = z( 4_i4b*i0+pp-1_i4b )
                z( 4_i4b*n0-pp ) = z( 4_i4b*i0-pp )
            end if
!
            dmin2 = min( dmin2, z( 4*n0+pp-1 ) )
!
            z( 4_i4b*n0+pp-1_i4b ) = min( z( 4_i4b*n0+pp-1_i4b ), z( 4_i4b*i0+pp-1_i4b ), z( 4_i4b*i0+pp+3_i4b ) )
            z( 4_i4b*n0-pp )       = min( z( 4_i4b*n0-pp ), z( 4_i4b*i0-pp ), z( 4_i4b*i0-pp+4_i4b ) )
!
            qmax = max( qmax, z( 4*i0+pp-3_i4b ), z( 4*i0+pp+1 ) )
            dmin = -zero
!
        end if
!
    end if
!
!   CHOOSE A SHIFT.
!
    call bd_lasq4( i0, n0, z, pp, n0in, dmin, dmin1, dmin2, dn, dn1, dn2, tau, ttype, g )
!
!   CALL DQDS UNTIL dmin > 0.
!
    do
!
        call bd_lasq5( i0, n0, z, pp, tau, sigma, dmin, dmin1, dmin2, dn, dn1, dn2, ieee )
!
!       CHECK STATUS.
!
        if ( dmin>=zero .and. dmin1>=zero ) then
!
!           SUCCESS.
!
            exit
!
        else if ( dmin<zero .and. dmin1>zero                   .and. &
                  z( 4_i4b*( n0-1_i4b )-pp )<tol*( sigma+dn1 ) .and. &
                  abs( dn )<tol*sigma ) then
!
!           CONVERGENCE HIDDEN BY NEGATIVE dn.
!
            z( 4_i4b*( n0-1_i4b )-pp+2_i4b ) = zero
            dmin = zero
!
            exit
!
        else if ( dmin<zero ) then
!
!           tau TOO BIG. SELECT NEW tau AND TRY AGAIN.
!
            if ( ttype<-22_i4b ) then
!
!               FAILED TWICE. PLAY IT SAFE.
!
                tau = zero
!
            else if ( dmin1>zero ) then
!
!               LATE FAILURE. GIVES EXCELLENT SHIFT.
!
                tau = ( tau + dmin )*( one - two*ulp )
                ttype = ttype - 11_i4b
!
            else
!
!               EARLY FAILURE. DIVIDE BY 4.
!
                tau = quarter*tau
                ttype = ttype - 12_i4b
!
            end if
!
            cycle
!
        else if ( is_nan( dmin ) .and. tau/=zero ) then
!
!           NaN.
!
            tau = zero
!
            cycle
!
        end if
!
!       POSSIBLE UNDERFLOW. PLAY IT SAFE.
!
        call bd_lasq6( i0, n0, z, pp, dmin, dmin1, dmin2, dn, dn1, dn2 )
!
        tau = zero
!
    end do
!
    if ( tau<sigma ) then
        desig = desig + tau
        t = sigma + desig
        desig = desig - ( t - sigma )
    else
        t = sigma + tau
        desig = sigma - ( t - tau ) + desig
    end if
!
    sigma = t
!
!
! END OF SUBROUTINE bd_one_singval_lasq3
! ______________________________________
!
    end subroutine bd_one_singval_lasq3
!
! =========================================================================================
!
    subroutine bd_one_singval_dqds( d, e, singval, failure, scaling )
!
! Purpose
! _______
!
!   BD_ONE_SINGVAL_DQDS computes one singular value of a real n-by-n 
!   (upper or lower) bidiagonal matrix BD by the differential quotient
!   difference with shifts (dqds) algorithm.
!
!   The singular value is computed to high relative precision in the absence
!   of denormalization, underflow and overflow with a variant of the
!   differential quotient difference with shifts (dqds) algorithm.
!
!
! Arguments
! _________
!
!   D            (INPUT) real(stnd), dimension(:)
!                On entry, D contains the diagonal elements of the bidiagonal
!                matrix BD.
!
!   E            (INPUT) real(stnd), dimension(:)
!                On entry, E contains the off-diagonal elements of the bidiagonal
!                matrix BD whose singular values are desired. E(1) is arbitrary.
!
!                The size of E must verify: size( E ) = size( D ) = n . 
!
!   SINGVAL      (OUTPUT) real(stnd)
!                On exit, SINGVAL contains one singular value of BD if FAILURE is false.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit ;
!                - FAILURE = true  :  indicates that the algorithm did not converge and that
!                  the maximum number of iterations in the inner loop of the dqds algorithm
!                  is exceeded or that the singular value cannot be estimated due to denormalization,
!                  underflow or overflow.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the bidiagonal matrix BD is scaled before computing
!                the singular value.
!
!                The default is to scale the bidiagonal matrix.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the variant of the dqds algorithm described
!   in the reference (3), which used a new shift strategy for the dqds algorithm.
!
!   It is a Fortran95/2003 translation of a Fortran77 subroutine available at:
!
!   http://syskiso.fuee.u-fukui.ac.jp/~kkimur/LAPROGNC/LAPROGNC.html
!
!   For further details on the dqds algorithm, see:
!
!   (1) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differential qd algorithms.
!           Numer. Math., Volume 67, No 2, pp. 191-229.
!
!   (2) Parlett, B.N., and Marques, O.A., 2000:
!           An implementation of the dqds algorithm (positive case).
!           Linear Algebra Appl., Volume 309, pp. 217-259.
!
!   (3) Yamashita, T., Kimura, K., Takata, M., and Nakamura, Y., 2013:
!           An application of the Kato-Temple inequality on matrix
!           eigenvalues to the dqds algorithm for singular values.
!           JSIAM  Letters, Volume 5, pp. 21-24.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
#ifdef _F2003
    use ieee_arithmetic,   only : ieee_support_datatype, ieee_is_normal
#endif
    use Num_Constants,     only : is_nan
    use Utilities,         only : assert_eq, lascl
!   
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:) :: d, e
    real(stnd), intent(out)              :: singval
!
    logical(lgl), intent(out) :: failure
    logical(lgl), intent(in), optional :: scaling
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7, tmp8, tmp9,      &
                  tau, tau1, tau2, tau3, s, t, scale, maxbd, tol, tol2,      &
                  sigma, sigma1, desig, desig0, scale1, dmin2, dn0, dn1, dn2
    real(stnd),  dimension(4*size(d)) :: work
    real(stnd),  dimension(size(d))   :: work5, work6, work7, work8
!
    integer(i4b) :: n, n2, n3, n4, m, i, j, k, l, p, maxit
!
    logical(lgl) :: do_scaling, test
#ifdef _F2003
    logical(lgl) :: use_ieee
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bd_one_singval_dqds'
!
    real(stnd),  parameter :: const=0.75_stnd
!
!   MAXIMUM NUMBER OF ITERATIONS IN THE INNER LOOP OF THE DQDS ALGORITHM.
!
    integer(i4b) :: maxiter=10_i4b    
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure    = false
    do_scaling = true
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b)  ,    &
                   int(size(e),i4b),      &
                   name_proc )
!
    if ( present(scaling) ) then
        do_scaling = scaling
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=2_i4b ) then
!
        if ( n==1_i4b ) then
!
            singval = abs( d(1_i4b) )
!
        else if ( n==2_i4b ) then
!
            call las2( d(1_i4b), e(2_i4b), d(2_i4b), tmp2, singval )
!
        end if
!
        return
!
    end if
!
!   SET INDICES.
!
    n2 = 2_i4b*n
    n3 = 3_i4b*n
    n4 = 4_i4b*n
!
    m = 1_i4b
!
!   MAKE A COPY OF THE BIDIAGONAL MATRIX.
!
    work(1_i4b:n)          = abs( d(1_i4b:n) )
    work(n+1_i4b:n2-1_i4b) = abs( e(2_i4b:n) )
!
    tmp2 = maxval( work(n+1_i4b:n2-1_i4b) )
!
    if ( tmp2==zero ) then
!
!       EARLY RETURN IF tmp2 IS ZERO AS MATRIX IS ALREADY DIAGONAL.
!
        singval = maxval( work(1_i4b:n) )
!
        return
!
    end if
!
!   ESTIMATE THE LARGEST SINGULAR VALUE.
!
    tmp1  = maxval( work(1_i4b:n) )
    maxbd = max( tmp1, tmp2 )
!
!   DETERMINE TOLERANCES AND SCALING FACTORS.
!
    tol    = fudge_dqds*ulp
    tol2   = tol**2
!
    scale  = sqrt( half/safmin )
    scale1 = one/scale
!
!   TEST IF IEEE ARITHMETIC IS SUPPORTED.
!
#ifdef _F2003
    use_ieee = ieee_support_datatype( dn0 )
#endif
!
!   SCALE THE BIDIAGONAL MATRIX IF REQUIRED.
!  
    if ( do_scaling ) then
!
        call lascl(  work(1_i4b:n2-1_i4b), maxbd, scale )
!  
    end if
!  
!   SQUARE THE ELEMENTS OF THE BIDIAGONAL MATRIX (E.G., COMPUTE THE Q'S AND E'S).
!  
    work(1_i4b:n2-1_i4b) = work(1_i4b:n2-1_i4b)**2
!
!   REVERSE THE QD-ARRAY, IF WARRANTED.
!
    if ( c1_5*work(1_i4b)<work(n) ) then
!    if ( reverse( work(1_i4b:n) ) ) then
!
!        work8(1_i4b:n) = work(n:1_i4b:-1_i4b)
!        work(1_i4b:n)  = work8(1_i4b:n)
!
!        work8(1_i4b:n-1_i4b)   = work(n2-1_i4b:n+1_i4b:-1_i4b)
!        work(n+1_i4b:n2-1_i4b) = work8(1_i4b:n-1_i4b)
!  
        k=n/2_i4b
!
        do j=1_i4b, k
!
            tmp1            = work(j)
            work(j)         = work(n+1_i4b-j)
            work(n+1_i4b-j) = tmp1
!
            tmp1       = work(n+j)
            work(n+j)  = work(n2-j)
            work(n2-j) = tmp1
!
        end do
!
    end if
!
    tmp1 = work(1_i4b)
!
    do j=1_i4b, n-3_i4b
!
        if ( work(n+j)<=tol2*tmp1 ) then
!
            work(n+j)  = -zero
            work(n3+j) = -zero
!
            m  = j + 1_i4b
!
            tmp1 = work(j+1_i4b)
!
        else
!
            tmp1 = work(j+1_i4b)*( tmp1/( tmp1 + work(n+j) ) )
!
        end if
!
    end do
!
    if ( work(n2-2_i4b)<=tol2*tmp1 ) then
         work(n2-2_i4b) = zero
         tmp1 = work(n-1_i4b)
    else
         tmp1 = work(n-1_i4b)*( tmp1/( tmp1 + work(n2-2_i4b) ) )
    end if
!
    if ( work(n2-1_i4b)<=tol2*tmp1 ) then
         work(n2-1_i4b) = zero
    end if
!
    work(n2) = zero
    work(n4) = zero
!
!   DQDS ITERATIONS.
!
    sigma  = -work(n2)
    sigma1 = tol2*sigma
    desig  = -work(n4)
!
    maxit = maxiter*( n - m + 1_i4b )
!
    do i=1_i4b, maxit
!  
        if ( n==m+1_i4b .or. ( work(n2-2_i4b)<=sigma1 .and. n/=m) ) then
!
            call sub_2( 0_i4b, n )
!
            exit
!
        else if ( n==m .or. work(n2-1_i4b)<=sigma1 ) then
!            
            call dqds2( sigma, desig, work(n), t, desig0 )
!
            exit
!
        end if
!
!       REVERSE THE QD-ARRAY, IF WARRANTED.
!         
!        if ( reverse( work(m:n) ) ) then
        if ( c1_5*work(m)<work(n) ) then
!
!            work8(m:n) = work(n:m:-1_i4b)
!            work(m:n)  = work8(m:n)
!
!            work8(m:n-1_i4b)   = work(n2-1_i4b:n+m:-1_i4b)
!            work(n+m:n2-1_i4b) = work8(m:n-1_i4b)
!
            k = ( n - m + 1_i4b )/2_i4b + m - 1_i4b
!         
            do j=m, k
!         
                tmp1        = work(j)
                work(j)     = work(n+m-j)
                work(n+m-j) = tmp1
!         
                tmp1               = work(n+j)
                work(n+j)          = work(n2-1_i4b+m-j)
                work(n2-1_i4b+m-j) = tmp1
!         
            end do
!         
        end if
!
        tau  = work(n)
        tmp3 = work(n-1_i4b)
!         
        if ( tau>tmp3 ) then
!         
            s    = tau
            tau  = tmp3
            tmp3 = s
!         
        end if
!         
        t = half*( ( tmp3 - tau ) + work(n2-1_i4b) )
!         
        if ( work(n2-1_i4b)>tau*tol2 .and. t/=zero ) then
!         
            s = tau*( work(n2-1_i4b)/t )
!         
            if( s<=t ) then
               s = tau*( work(n2-1_i4b)/( t*( one + sqrt( one + s/t ) ) ) )
            else
               s = tau*( work(n2-1_i4b)/( t + sqrt( t )*sqrt( t + s ) ) )
            end if
!         
            t    = tmp3 + ( s + work(n2-1_i4b) )
            tau  = tau*( tmp3/t )
            tmp3 = t
!         
        end if
!         
        do
!
            tau  = min( tau, work(n) )
            tau2 = minval( work(m:n-1_i4b) )
!         
            if ( min(tau,tau2)==zero ) then
!         
                call sub_1()
!         
                exit
!         
            end if
!
            test = true
!         
            do         
!         
                if ( tau2<=tau ) then
!         
                    tau1 = tau2
!         
                    exit
!         
                else
                    tau1 = tau
                end if
!
                dn2 = work(m) - tau
!         
                if ( dn2<=zero ) exit
!         
                do j=m, n-3_i4b
!         
                    dn2 = dn2*( work(j+1_i4b)/( dn2 + work(n+j) ) ) - tau
!         
                    if ( dn2<=zero ) exit
!         
                end do
!         
                if ( j>n-3_i4b ) exit
!         
                tmp1 = dn2 + work(n2-2_i4b)
!         
                if ( safmin*tmp1<work(n-1_i4b) .and. safmin*work(n-1)<tmp1 ) then
                    dn1 = dn2*( work(n-1_i4b)/tmp1 ) - tau
                else
                    dn1 = work(n-1_i4b)*( dn2/tmp1 ) - tau
                end if
!         
                if ( dn1<=zero ) exit
!
                tmp1 = dn1 + work(n2-1_i4b)
!         
                if ( safmin*tmp1<work(n) .and. safmin*work(n)<tmp1 ) then
                    dn0 = dn1*( work(n)/tmp1 ) - tau
                else
                    dn0 = work(n)*( dn1/tmp1 ) - tau
                end if
!
                if( dn0<zero .and. tau+dn0==tau ) dn0 = zero
!
                if ( dn0<zero ) then
!
                    tau2 = max( dn0+tau, zero )
                    if ( tau2==tau ) tau2 = zero
!
                    tau3 = const*tau
                    if ( tau3==tau ) tau3 = half*tau
!
                    tau = max( tau2, tau3 )
!
                end if
!
                test = tau<=zero
!         
                exit
!         
            end do
!
            if ( test ) then
!           
                tmp7  = scale/work(m)
                tmp1  = tmp7
                tmp5  = tmp7**2
                tmp3  = tmp5
                tmp2  = tmp3
                dmin2 = tmp3
!           
                do j=m+1_i4b, n
!           
                    tmp4  = scale/work(j)
                    tmp8  = ( work(n+j-1_i4b)*scale1 )*tmp4
                    tmp7  = tmp8*tmp7 + tmp4
                    tmp1  = tmp1 + tmp7
                    tmp6  = tmp7**2
                    tmp3  = tmp8*( tmp3 + tmp5) + tmp6
                    tmp2  = tmp2 + tmp3
                    dmin2 = max( dmin2, tmp3 )
                    tmp5  = tmp6
!           
                end do
!
                tau  = max( zero, scale/tmp1, scale/sqrt(tmp2) )
                tmp8 = real( n-m+1_i4b, stnd )
                tmp9 = scale*( tmp8/( tmp1 + sqrt( tmp8-one )*sqrt( tmp8*tmp2 - tmp1**2 ) ) )
!
                if ( tmp9==tmp9 ) then
                    tau = max( tau, tmp9 )
                end if
!
                tau2 = tau1
                tmp8 = ceiling( (tmp1/tmp2)*tmp1 )
!
                if ( tmp8>one ) then
                    tmp9 = scale*( tmp8/( tmp1 + sqrt( tmp8*tmp2 - tmp1**2 )/sqrt( tmp8 - one ) ) )
                    if ( tmp9==tmp9 ) then
                        tau2 = min( tmp9, scale/sqrt( dmin2 ), tau2 )
                    else
                        tau2 = min( scale/sqrt( dmin2 ), tau2 )
                    end if
                else
                    tau2 = min( scale/sqrt( dmin2 ), tau2 )
                end if
!
                tau = min( tau, tau2 )
!
                if ( tau2>=two*tau ) then
!
                    work7(m:n)       = one/sqrt( work(m:n) )
                    work8(m:n-1_i4b) = sqrt( work(n+m:n2-1_i4b) )
!         
                    work5(n) = work7(n)
                    tmp3 = work5(n)
!           
                    do j=n-1_i4b, m, -1_i4b
!           
                        work(n2+j) = work8(j)*work7(j)
                        work5(j)   = work7(j) + work(n2+j)*work5(j+1_i4b)
!           
                        tmp3 = max( tmp3, work5(j) )
!           
                    end do
!           
                    tmp3     = one/tmp3
                    work5(m) = (work5(m)*tmp3)*work7(m)
                    tmp1     = work5(m)
!
                    do j=m+1_i4b, n
!
                        work(n3+j) = work8(j-1_i4b)*work7(j)
                        work5(j)   = ( work5(j)*tmp3 )*work7(j) + work(n3+j)*work5(j-1_i4b)
!
                        tmp1 = max( tmp1, work5(j) )
!
                    end do
!
                    tau = zero
!
                    if ( tmp1>zero ) then
                        tau = max( tau, tmp3/tmp1 )
                    end if
!
                    tmp1     = one/tmp1
                    work5(n) = work5(n)*tmp1
                    work6(n) = work5(n)*work7(n)
                    tmp3     = work6(n)
!           
                    do j=n-1_i4b, m, -1_i4b
!
                        work5(j) = work5(j)*tmp1
                        work6(j) = work5(j)*work7(j) + work(n2+j)*work6(j+1_i4b)
!
                        tmp3 = max( tmp3, work6(j) )
!
                    end do
!           
                    tmp3 = one/tmp3
                    tmp2 = ( work6(m)*tmp3 )*work7(m)
                    tmp1 = tmp2/work5(m)
!           
                    do j=m+1_i4b, n
!
                        tmp2 = ( work6(j)*tmp3 )*work7(j) + work(n3+j)*tmp2
                        tmp1 = max( tmp1, tmp2/work5(j) )
!
                    end do
!           
                    if ( tmp1>zero ) then
                        tau = max( tau, tmp3/tmp1 )
                    end if
!
                    tau = min( tau, tau1 )
!
                    if ( tau<=zero ) then
!
                        tau = sqrt( work(n) ) - half*work8(n-1_i4b)
!
                        if ( tau<=zero ) then
                            call sub_1()
                            exit
                        end if
!           
                        do j=n-1_i4b, m+1_i4b, -1_i4b
!
                            tmp2 = sqrt( work(j) ) - half*( work8(j) + work8(j-1_i4b) )
!
                            if ( tmp2<=zero ) then
                                call sub_1()
                                exit
                            end if
!
                            tau = min( tau, tmp2 )
!
                        end do
!
                        if ( j>m ) exit
!                    
                        tmp2 = sqrt( work(m) ) - half*work8(m)
!           
                        if ( tmp2<=zero )  then
                            call sub_1()
                            exit
                        end if
!         
                        tau = min( tau, tmp2 )**2
!
                    end if
!
                end if
!
            end if
!           
            do
!           
                if ( tau==zero ) then
                    call sub_1()
                    exit
                end if
!
                dn2   = work(m)-tau
                dmin2 = dn2
!
                do j=m, n-3_i4b
!
                    work(n2+j) = dn2 + work(n+j)
                    tmp3       = work(j+1_i4b)/work(n2+j)
                    work(n3+j) = work(n+j)*tmp3
                    dn2        = dn2*tmp3 - tau
                    dmin2      = min( dmin2, dn2 )
!
                end do
!
                if ( dmin2<=zero ) then 
!
                    tau3 = const*tau
                    if ( tau3==tau ) tau3 = half*tau
                    tau = tau3
!
                    cycle
!
                end if
!
                work(n3-2_i4b) = dn2+work(n2-2_i4b)
!
                if ( safmin*work(n3-2_i4b)<work(n-1_i4b) .and. safmin*work(n-1_i4b)<work(n3-2_i4b) ) then
                    tmp3           = work(n-1_i4b)/work(n3-2_i4b)
                    work(n4-2_i4b) = work(n2-2_i4b)*tmp3
                    dn1            = dn2*tmp3 - tau
                else
                    work(n4-2_i4b) = work(n-1_i4b)*( work(n2-2_i4b)/work(n3-2_i4b) )
                    dn1            = work(n-1_i4b)*( dn2/work(n3-2_i4b) ) - tau
                end if
!
                if ( dn1<=zero ) then
!           
                    tau2 = max( dn1+tau, zero )
                    if ( tau2==tau ) tau2 = zero
!
                    tau3 = const*tau
                    if ( tau3==tau ) tau3 = half*tau
!
                    tau = max( tau2, tau3 )
!
                    cycle
!
                end if
!
                work(n3-1_i4b) = dn1+work(n2-1_i4b)
!
                if ( safmin*work(n3-1_i4b)<work(n) .and.  safmin*work(n)<work(n3-1_i4b) ) then
                    tmp3           = work(n)/work(n3-1_i4b)
                    work(n4-1_i4b) = work(n2-1_i4b)*tmp3
                    dn0            = dn1*tmp3 - tau
                else
                    work(n4-1_i4b) = work(n)*( work(n2-1_i4b)/work(n3-1_i4b) )
                    dn0            = work(n)*( dn1/work(n3-1_i4b) ) - tau
                end if
!
                if ( dn0<zero .and. tau+dn0==tau ) dn0 = zero
!
                if ( dn0<zero ) then
!
                    tau2 = max( dn0+tau, zero )
                    if ( tau2==tau ) tau2 = zero
!
                    tau3 = const*tau
                    if ( tau3==tau ) tau3 = half*tau
!
                    tau = max( tau2, tau3 )
!
                    cycle
!
                end if
!
                if ( dn0>=work(n) ) then
!
                    tmp3 = dn1/work(n3-1_i4b) - tau/work(n)
!
                    if ( tmp3>=zero ) then
                        dn0 = work(n)*tmp3
                    end if
!
                end if
!
                work(n3) = dn0
!
#ifdef _F2003
                if ( use_ieee ) then
                    test = .not.ieee_is_normal( dn0 )
                else
#endif
                    test = is_nan( dn0 )
#ifdef _F2003
                end if
#endif
!
                if ( test ) then
!
                    tau3 = const*tau
                    if ( tau3==tau ) tau3 = half*tau
                    tau = tau3
!
                    cycle
!
                end if
!
                call dqds2( sigma, desig, tau, t, desig0 )
!
                sigma  = t
                sigma1 = tol2*sigma
                desig  = desig0
!
                exit
!
            end do
!
            exit
!
        end do
!
        if ( work(n4-2_i4b)<=sigma1 .or. work(n2-2_i4b)<=tol2*work(n3-2_i4b) ) then
!
            call sub_2( n2, n3 )
!
            exit
!
        end if
!
        if ( work(n4-1_i4b)<=sigma1 .or. work(n2-1_i4b)<=tol2*work(n3-1_i4b) ) then
!
            call dqds2( sigma, desig, work(n3), t, desig0 )
!
            exit
!
        end if
!
        dn2   = work(n2+m)
        dmin2 = dn2
!
        do j=m, n-3_i4b
!
            work(j)   = dn2 + work(n3+j)
            tmp2      = work(n2+j+1_i4b)/work(j)
            work(n+j) = work(n3+j)*tmp2
            dn2       = dn2*tmp2
            dmin2     = min( dmin2, dn2 )
!
        end do
!
        call sub_3( 0_i4b, n, n2, n3 )
!
        do p=1_i4b, 2_i4b
!
#ifdef _F2003
            if ( use_ieee ) then
                test = .not.ieee_is_normal( dn2 )
            else
#endif
                test = is_nan( dn2 )
#ifdef _F2003
            end if
#endif
            if ( test ) then
!
                dn2 = work(n2+m)
!
                if ( sigma==dn2+sigma ) dn2 = zero
!
                do j=m, n-3_i4b
!
                    work(j) = dn2+work(n3+j)
!
                    if ( work(j)==zero ) then
                        work(n+j) = zero
                        dn2       = work(n2+j+1_i4b)
                    else if ( safmin*work(j)<work(n2+j+1_i4b) .and. safmin*work(n2+j+1_i4b)<work(j) ) then
                        tmp2      = work(n2+j+1_i4b)/work(j)
                        work(n+j) = work(n3+j)*tmp2
                        dn2       = dn2*tmp2
                    else
                        work(n+j) = work(n2+j+1_i4b)*( work(n3+j)/work(j) )
                        dn2       = work(n2+j+1_i4b)*( dn2/work(j) )
                    end if
!
                    if ( sigma==dn2+sigma ) dn2 = zero
!
                end do
!
                call sub_3( 0_i4b, n, n2, n3 )
!
                exit       
!
            end if
!
            if ( dmin2+sigma/=sigma .or. p==2_i4b ) exit
!
            dn2 = work(n2+m)
!
            if ( sigma==dn2+sigma ) dn2 = zero
!
            do j=m, n-3_i4b
!
                work(j)   = dn2+work(n3+j)
                tmp2      = work(n2+j+1_i4b)/work(j)
                work(n+j) = work(n3+j)*tmp2
                dn2       = dn2*tmp2
!
                if ( sigma==dn2+sigma ) dn2 = zero
!
            end do
!
            call sub_3( 0_i4b, n, n2, n3 )
!
        end do
!
        do j=m, n-3_i4b
!
            if ( work(n+j)<=sigma1 .or. work(n3+j)<=tol2*work(j) ) then
!
                work(n+j)  = -sigma
                work(n3+j) = -desig
!
                m = j + 1_i4b
!
            end if
!
        end do
!
        if ( work(n4-2_i4b)<=tol2*work(n-2_i4b) ) then
            work(n2-2_i4b) = zero
        end if
!
        if ( work(n4-1_i4b)<=tol2*work(n-1_i4b) ) then
            work(n2-1_i4b) = zero
        end if
!
    end do
!
!   MAXIMUM NUMBER OF ITERATIONS EXCEEDED, SET FAILURE INDICATOR AND EXIT.
!
    failure = i>maxit
!
    if ( .not.failure ) then
!
        singval = sqrt( t )
!
        if ( do_scaling ) then
            call lascl( singval, scale, maxbd )
        end if
!
#ifdef _F2003
        if ( use_ieee ) then
            failure = .not.ieee_is_normal( singval )
        else
#endif
            failure = is_nan( singval )
#ifdef _F2003
        end if
#endif
!
    end if
!
    if ( failure ) then
!
        singval = zero
!
    end if
!
! ___________________________________________________________________________________________
!
                             contains
!                            ________
!
! ___________________________________________________________________________________________
!
        subroutine sub_1(  )
!
!
!   EXECUTABLE STATEMENTS
!   ---------------------
!
!
        dn2 = work(m)
!
        if ( sigma==dn2+sigma ) dn2 = zero
!
        do l=m, n-3_i4b
!
            work(n2+l) = dn2 + work(n+l)
            tmp3       = work(l+1_i4b)/work(n2+l)
            work(n3+l) = work(n+l)*tmp3
            dn2        = dn2*tmp3
!
            if ( sigma==dn2+sigma ) dn2 = zero
!
        end do
!
!        call sub_3( n2, n3, 0_i4b, n )
!
        do l=2_i4b, 1_i4b, -1_i4b
!
            work(n3-l) = dn2+work(n2-l)
!
            if ( safmin*work(n3-l)<work(n+1_i4b-l) .and. safmin*work(n+1_i4b-l)<work(n3-l) ) then
                tmp2       = work(n+1-l)/work(n3-l)
                work(n4-l) = work(n2-l)*tmp2
                dn2        = dn2*tmp2
            else
                work(n4-l) = work(n+1_i4b-l)*( work(n2-l)/work(n3-l) )
                dn2        = work(n+1_i4b-l)*( dn2/work(n3-l) )
            end if
!         
            if ( sigma==dn2+sigma ) dn2 = zero
! 
        end do
!
        work(n3) = dn2
!
#ifdef _F2003
        if ( use_ieee ) then
            test = .not.ieee_is_normal( dn2 )
        else
#endif
            test = is_nan( dn2 )
#ifdef _F2003
        end if
#endif
        if ( test ) then
!
            dn2 = work(m)
!
            if ( sigma==dn2+sigma ) dn2 = zero
!
            do l=m, n-3_i4b
!
                work(n2+l) = dn2 + work(n+l)
!
                if ( safmin*work(n2+l)<work(l+1_i4b) .and. safmin*work(l+1_i4b)<work(n2+l) ) then
                    tmp2       = work(l+1_i4b)/work(n2+l)
                    work(n3+l) = work(n+l)*tmp2
                    dn2        = dn2*tmp2
                else
                    work(n3+l) = work(l+1_i4b)*( work(n+l)/work(n2+l) )
                    dn2        = work(l+1_i4b)*( dn2/work(n2+l) )
                end if
!
                if ( sigma==dn2+sigma ) dn2 = zero
!
            end do
!
!            call sub_3( n2, n3, 0_i4b, n )
!
            do l=2_i4b, 1_i4b, -1_i4b
!
                work(n3-l) = dn2+work(n2-l)
!
                if ( safmin*work(n3-l)<work(n+1_i4b-l) .and. safmin*work(n+1_i4b-l)<work(n3-l) ) then
                    tmp2       = work(n+1-l)/work(n3-l)
                    work(n4-l) = work(n2-l)*tmp2
                    dn2        = dn2*tmp2
                else
                    work(n4-l) = work(n+1_i4b-l)*( work(n2-l)/work(n3-l) )
                    dn2        = work(n+1_i4b-l)*( dn2/work(n3-l) )
                end if
!         
                if ( sigma==dn2+sigma ) dn2 = zero
! 
            end do
!
            work(n3) = dn2
!
        end if
!
!
!   END OF SUBROUTINE sub_1
!   -----------------------
!
        end subroutine sub_1
!
! ___________________________________________________________________________________________
!
        subroutine sub_2( i1, i2 )
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ----------------------------
!
        integer(i4b), intent(in) :: i1, i2
!
!
!   SPECIFICATIONS FOR LOCAL VARIABLES
!   __________________________________
!
!
        real(stnd) :: t1, t2
!
!
!   EXECUTABLE STATEMENTS
!   ---------------------
!
!
        if ( work(i1+n)>work(i1+n-1_i4b) ) then
            s                = work(i1+n)
            work(i1+n)       = work(i1+n-1_i4b)
            work(i1+n-1_i4b) = s
        end if
!
        t = half*( ( work(i1+n-1_i4b) - work(i1+n) ) + work(i2+n-1_i4b) )
!
        if ( work(i2+n-1_i4b)>work(i1+n)*tol2 .and. t/=zero ) then
!
            s = work(i1+n)*( work(i2+n-1_i4b)/t )
!
            if ( s<=t ) then
                s = work(i1+n)*( work(i2+n-1_i4b)/( t*( one + sqrt( one + s/t ) ) ) )
            else
                s = work(i1+n)*( work(i2+n-1_i4b)/( t + sqrt( t )*sqrt( t + s ) ) )
            end if
!
            t = work(i1+n-1_i4b) + ( s + work(i2+n-1_i4b) )
!
            work(i1+n)       = work(i1+n)*( work(i1+n-1_i4b)/t )
            work(i1+n-1_i4b) = t
!
        end if
!
        call dqds2( sigma, desig, work(i1+n-1_i4b), t1, desig0 )
!
        call dqds2( sigma, desig, work(i1+n), t2, desig0 )
!
        t = max( t1, t2 )
!
!
!   END OF SUBROUTINE sub_2
!   -----------------------
!
        end subroutine sub_2
!
! ___________________________________________________________________________________________
!
        subroutine sub_3( j1, j2, j3, j4 )
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ----------------------------
!
        integer(i4b), intent(in) :: j1, j2, j3, j4
!
!
!   EXECUTABLE STATEMENTS
!   ---------------------
!
!
        do l=2_i4b, 1_i4b, -1_i4b
!
            work(j1+n-l) = dn2+work(j4+n-l)
!
            if ( safmin*work(j1+n-l)<work(j3+n+1_i4b-l) .and. safmin*work(j3+n+1_i4b-l)<work(j1+n-l) ) then
                tmp2         = work(j3+n+1-l)/work(j1+n-l)
                work(j2+n-l) = work(j4+n-l)*tmp2
                dn2          = dn2*tmp2
            else
                work(j2+n-l) = work(j3+n+1_i4b-l)*( work(j4+n-l)/work(j1+n-l) )
                dn2          = work(j3+n+1_i4b-l)*( dn2/work(j1+n-l) )
            end if
!         
            if ( sigma==dn2+sigma ) dn2 = zero
! 
        end do
!
        work(j1+n) = dn2
!
!
!   END OF SUBROUTINE sub_3
!   -----------------------
!
        end subroutine sub_3
!
! ___________________________________________________________________________________________
!
!
! END OF SUBROUTINE bd_one_singval_dqds
! _____________________________________
!      
    end subroutine bd_one_singval_dqds
!
! ===========================================================================================
!
    subroutine upper_bd_inviter( d, e, s, leftvec, rightvec, failure,      &
                                 maxiter, scaling, initvec )
!
! Purpose
! _______
!
!   UPPER_BD_INVITER is a private subroutine for computing the left and right singular vectors of
!   a real n-by-n upper bidiagonal matrix BD corresponding to a specified singular value, using 
!   Fernando's method and inverse iteration on the tridiagonal Golub-Kahan (TGK) form of the
!   bidiagonal matrix BD.
!
!
! Arguments
! _________
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, D contains the diagonal elements of the bidiagonal
!                  matrix BD.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, E contains the off-diagonal elements of the bidiagonal
!                  matrix BD. E(1) is arbitrary.
!
!                  The size of E must verify: size( E ) = size( D ) = n . 
!
!   S              (INPUT) real(stnd)
!                  On entry, the selected singular value of the bidiagonal matrix BD.
!                  The singular value must be positive or zero.
!
!   LEFTVEC        (OUTPUT) real(stnd), dimension(:)
!                  On exit, the computed left singular vector.
!
!                  The shape of LEFTVEC must verify: size(LEFTVEC) = size( D ) = n .
!
!   RIGHTVEC       (OUTPUT) real(stnd), dimension(:)
!                  On exit, the computed right singular vector.
!
!                  The shape of RIGHTVEC must verify: size(RIGHTVEC) = size( D ) = n .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit, FAILURE=FALSE indicates successful exit. If FAILURE is set to TRUE
!                  then the singular vectors failed to converge in MAXITER iterations.
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed in the subroutine.
!
!                  By default, 2 inverse iterations are performed.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if SCALING=true the bidiagonal matrix BD is scaled before computing
!                  the singular vectors.
!
!                  The default is to scale the bidiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if INITVEC=true a Fernando vector is used to start the inverse iteration process
!                  for computing the singular vectors of the bidiagonal matrix BD (e.g. the eigenvector of the
!                  associated tridiagonal Golub-Kahan matrix); if INITVEC=false a random uniform starting vector
!                  is used.
!
!                  The default is to use a Fernando starting vector if the Golub-Kahan form of the input
!                  bidiagonal matrix is unreduced, and a random uniform starting vector otherwise.
!
!
! Further Details
! _______________
!
!   A first estimate of the singular vectors is computed by the Fernando method applied to the
!   tridiagonal Golub-Kahan matrix associated with the bidiagonal matrix BD (see the reference
!   (1) for details) if this Golub-Kahan form of the input bidiagonal matrix is unreduced.
!   Otherwise, a random start is used as a first estimate of the singular vectors as in the
!   standard inverse-iteration algorithm.
!
!   The singular vectors are then computed or refined using inverse iteration on the
!   tridiagonal Golub-Kahan matrix.
!
!   For further details, on Fernando's method for computing eigenvectors of tridiagonal
!   matrices or inverse iteration, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (4) Bini, D.A., Gemignani, L., and Tisseur, F., 2005:
!           The Ehrlich-Aberth method for the nonsymmetric tridiagonal eigenvalue problem.
!           SIAM J. Matrix Anal. Appl., 27, 153-175.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : svd_error7
    use Utilities,        only : assert_eq, merror, pythag
    use Eig_Procedures,   only : trid_cmp, trid_solve
    use Random,           only : random_number_
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in)                 :: s
    real(stnd),  intent(in),  dimension(:)  :: d, e
    real(stnd),  intent(out), dimension(:)  :: leftvec, rightvec
!
    integer(i4b), intent(in),  optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: scaling, initvec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),  dimension(2*size(d)) :: gk, sub, diag, sup1, sup2, temp
    real(stnd) :: lambda, xj, xjm, tmp, prec, normbd, scalfac, tau, gamma,  &
                  omega, ortol, sqrt2, lim
!
    integer(i4b) :: n, n2, mm, maxit, i, ileft, iright, i1, i2, j, l, nn1, nn2
!
    logical(lgl)                        :: do_scaling, fernando_init, test_initvec
    logical(lgl),  dimension(2*size(d)) :: perm
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='upper_bd_inviter'
!
!   maxiter_init IS THE MAXIMUM NUMBER OF ALLOWED ITERATIONS FOR IMPROVING
!   THE INITIAL GUESS OF THE SINGULAR VECTORS.
!
    integer(i4b),  parameter :: maxiter_init=3_i4b
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)         ,   &
                    int(size(e),i4b)         ,   &
                    int(size(leftvec),i4b)   ,   &
                    int(size(rightvec),i4b)  ,   &
                    name_proc )
!
    n2  = 2_i4b*n
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            leftvec(1_i4b)  = one
            rightvec(1_i4b) = one
!
        end if
!
        return
!
    end if
!
!   TEST IF SINGULAR VALUE IS POSITIVE OR ZERO.
!
    if ( s<zero ) then
!
        call merror( name_proc//svd_error7 )
!
    end if
!
!   SET OPTIONAL ARGUMENTS.
!
    if ( present(maxiter) ) then
        maxit = max( maxiter, 1_i4b )
    else
        maxit = 2_i4b
    end if
!
    test_initvec = present( initvec )
!
    if ( present(scaling) ) then
        do_scaling = scaling
    else
        do_scaling = true
    end if
!    
!   GET MACHINE CONSTANTS.
!
    gamma  = sqrt( safmin )
    ortol  = sqrt( ulp )
    prec   = fudge_prec*ulp
    sqrt2  = sqrt( two )
!
!   CHECK IF THE BIDIAGONAL MATRIX IS THE ZERO MATRIX.
!
    diag(:n)     = d(:n)
    sub(1_i4b)   = zero
    sub(2_i4b:n) = e(2_i4b:n)
!
    do j = n, 1_i4b, -1_i4b
!
        if ( abs(diag(j))>=safmin .or. abs(sub(j))>=safmin ) exit
!
        if ( j/=n ) cycle
!
        leftvec(:n) = zero
        leftvec(j)  = one
!
        rightvec(:n) = zero
        rightvec(j)  = one
!
    end do
!
    if ( j==0_i4b ) then
!
!       THE BIDIAGONAL MATRIX IS THE NULL MATRIX, RETURN.
!
        return
!
    end if
!
!   CHECK THE FIRST PART OF THE BIDIAGONAL MATRIX FOR ZEROS
!   AND REDUCE AGAIN ITS SIZE IF POSSIBLE.
!
    nn2 = j
!
    do j = 1_i4b, n-1_i4b
!
        if ( abs(diag(j))>=safmin .or. abs(sub(j+1_i4b))>=safmin ) exit
!
    end do
!
    nn1 = j
!
    if ( nn1>nn2 ) then
!
!       MISMATCH, WE CANNOT WORK ON THE REDUCED PORTION
!       OF THE TRIDIAGONAL GOLUB-KAHAN MATRIX.
!
        nn1 = 1_i4b
        nn2 = n
!
    end if
!
!   FORM THE OFF DIAGONAL ELEMENTS OF THE SYMMETRIC TRIDIAGONAL GOLUB-KAHAN SYSTEM.
!
    gk(1_i4b:n2:2_i4b)       = diag(1_i4b:n)
    gk(2_i4b:n2-1_i4b:2_i4b) = sub(2_i4b:n)
    gk(n2)                   = zero
!    
!   SCALE THE SYMMETRIC TRIDIAGONAL GOLUB-KAHAN SYSTEM IF REQUIRED.    
!    
    if ( do_scaling ) then
!    
        normbd  = maxval( abs(gk(1_i4b:n2-1_i4b) ) )
!    
#ifdef _SCALE
        omega   = scale(one,maxexp-1_i4b)
#else
        omega   = machbase**(maxexp-1_i4b)
#endif
        tau     = sqrt( gamma*omega )
        j       = exponent( tau ) - exponent( normbd ) - 1_i4b
        scalfac = machbase**j
!
        gk(1_i4b:n2-1_i4b) = scalfac*gk(1_i4b:n2-1_i4b)
!
        where( abs(gk(1_i4b:n2-1_i4b))<gamma )
            gk(1_i4b:n2-1_i4b) = sign( gamma, gk(1_i4b:n2-1_i4b) )
        end where
!    
        lambda = scalfac*max( s, safmin )
!    
    else
!    
        scalfac = one
        lambda  = max( s, safmin )
!    
    end if
!
!   COMPUTE THE UPPER GERSCHGORIN BOUND AND THE NORM OF
!   THE TRIDIAGONAL GOLUB-KAHAN MATRIX.
!
    normbd = zero
    tau    = zero
    xj     = zero
!
    do j=1_i4b, n2
!
        xjm    = abs( gk(j) )
        tmp    = xj + xjm
        normbd = normbd + tmp
        tau    = max( tau, tmp )
        xj     = xjm
!
    end do
!
!   DETERMINE THRESHOLD FOR TINY SINGULAR VALUES AND SMALL DIAGONAL ELEMENTS.
!
!    lim = normbd*real( n, stnd )*prec
    lim = tau*real( n, stnd )*prec
!
    i1 = 2_i4b*nn1 - 1_i4b
    i2 = 2_i4b*nn2
!
    if ( lambda<=lim .and. all( abs( gk(i1:i2:2_i4b) ) > lim ) ) then
!    if ( lambda<=lim .and. all( abs( gk(i1:i2:2_i4b) ) > gamma ) ) then
!
!       IF THE SINGULAR VALUE IS VERY SMALL AND NONE OF THE DIAGONAL ELEMENTS OF THE
!       BIDIAGONAL MATRIX ARE ZERO, WE CANNOT WORK SAFELY ON THE REDUCED PORTION
!       OF THE TRIDIAGONAL GOLUB-KAHAN MATRIX.
!
        i1 = 1_i4b
        i2 = n2
!
    else if ( lambda>lim ) then
!
!       IF THE SINGULAR VALUE IS NOT SMALL, TEST IF THE FIRST OR LAST OFF-DIAGONAL
!       ELEMENTS OF THE TRIDIAGONAL GOLUB-KAHAN MATRIX ARE ZERO AND
!       REDUCE ITS SIZE IF POSSIBLE.
!
        if ( abs( gk(i1) )<=lim ) i1 = i1 + 1_i4b
!        if ( abs( gk(i1) )<=gamma ) i1 = i1 + 1_i4b
!
        if ( abs( gk(i2-1_i4b) )<=lim ) i2 = i2 - 1_i4b
!        if ( abs( gk(i2-1_i4b) )<=gamma ) i2 = i2 - 1_i4b
!
    end if
!
    if ( test_initvec ) then
!
        fernando_init = initvec
!
    else
!
!       TEST IF THE GOLUB-KAHAN FORM OF THE INPUT BIDIAGONAL MATRIX IS UNREDUCED.
!
        fernando_init = all( abs( gk(i1:i2-1_i4b) )>lim )
!        fernando_init = all( abs( gk(i1:i2-1_i4b) )>gamma )
!
    end if
!
    ileft  = 2_i4b
    iright = 1_i4b
!
!   COMPUTE SINGULAR VECTORS BY INVERSE ITERATION ON THE TRIDIAGONAL
!   GOLUB-KAHAN FORM OF THE BIDIAGONAL MATRIX.
!    
    temp(i1:i2) = zero
!
!   COMPUTE LU FACTORS WITH PARTIAL PIVOTING (PT=LU) OF THE TGK MATRIX.
!
    call trid_cmp( temp(i1:i2), gk(i1:i2), lambda, sub(i1:i2), diag(i1:i2),    &
                   sup1(i1:i2), sup2(i1:i2), perm(i1:i2) )
!
!   REPLACE SMALL DIAGONAL ELEMENTS BY safmin.
!
    where( abs(diag(i1:i2))<safmin ) diag(i1:i2) = sign( safmin, diag(i1:i2) )
!    
    do l=1_i4b, maxiter_init
!    
        if ( fernando_init ) then
!    
!           FERNANDO METHOD.  
!
            mm = maxdiag_gkinv_ldu( gk(i1:i2-1_i4b), lambda ) + i1 - 1_i4b
!    
!           BINI METHOD.  
!                
!            mm = maxdiag_gkinv_qr( gk(i1:i2-1_i4b), lambda ) + i1 - 1_i4b
!    
            temp(mm) = one
!
        else
!    
!           COMPUTE A RANDOM INITIAL VECTOR.  
!
            call random_number_( temp(i1:i2) )
!
            temp(i1:i2) = temp(i1:i2)*two - one
!
        end if
!
!       DO maxit ITERATIONS FOR THE EIGENVECTOR OF THE TGK MATRIX.
!
        do i=1_i4b, maxit
!
!           SOLVE THE SYSTEM LUx = Pb WITH RESCALING TO AVOID OVERFLOW.
!
            call trid_solve( sub(i1:i2), diag(i1:i2), sup1(i1:i2),  &
                             sup2(i1:i2), perm(i1:i2), temp(i1:i2)  )
!
            if ( i/=maxit ) then
!
!               NORMALIZE THE EIGENVECTOR OF THE TGK MATRIX.
!
                tmp = norm( temp(i1:i2) )
!
                if ( tmp/=zero ) then
!
                    temp(i1:i2) = temp(i1:i2)/tmp
!
                else
!
                    call random_number_( temp(i1:i2) )
!
                    temp(i1:i2) = temp(i1:i2)*two - one
!
                end if
!
            end if
!
        end do
!
        temp(1_i4b:i1-1_i4b) = zero
        temp(i2+1_i4b:n2)    = zero
!
!       NORMALIZE THE SINGULAR VECTORS.
!
        xj = norm( temp(ileft:n2:2_i4b) )
!
        if ( xj/=zero ) then
            leftvec(1_i4b:n) = temp(ileft:n2:2_i4b)/xj
        else
            leftvec(1_i4b:n) = zero
        end if
!
        xjm = norm( temp(iright:n2:2_i4b) )
!
        if ( xjm/=zero ) then
            rightvec(1_i4b:n) = temp(iright:n2:2_i4b)/xjm
        else
            rightvec(1_i4b:n) = zero
        end if
!
        tau = pythag( xj, xjm )
        tmp = max( xj, xjm )
!
        if ( abs( tmp - ortol*tau )*sqrt2>tau .or. tmp<gamma ) then
!
!           BAD ESTIMATE OF THE SINGULAR VECTORS, TRY TO IMPROVE THEM.
!
            if ( l/=maxiter_init ) then
!
!               BAD ESTIMATE OF THE SINGULAR VECTORS, TRY AGAIN
!               WITH A RANDOM INITIAL VECTOR.
!
                fernando_init = false        
!
                cycle
!
            else if ( tmp==tau  ) then
!
!               BAD SINGULAR VECTORS, SET FAILURE INDICATOR.
!
                failure = true
!
            end if
!
        end if
!
        exit
!
    end do
!
    failure = min( xj, xjm )==zero .or. failure
!
!
! END OF SUBROUTINE upper_bd_inviter
! __________________________________
!
    end subroutine upper_bd_inviter
!
! =========================================================================================
!
    subroutine ortho_vec( q, vec, normvec, normvec_orig )
!
! Purpose
! _______
!
!   ORTHO_VEC is a private subroutine used by Gram-Schmidt algorithms for updating
!   orthogonal vectors when reorthogonalization is required. ORTHO_VEC implements 
!   a reorthogonalization procedure for the modified Gram-Schmidt algorithm used in
!   some inverse iteration or bidiagonal reduction algorithms in this module.
!
!
! Arguments
! _________
!
!   Q            (INPUT) real(stnd), dimension(:,:)
!                On entry, a n-by-nq matrix Q of previous singular or normalized
!                orthogonal vectors.
!
!   VEC          (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the current estimate of the nq+1 unnormalized singular
!                or orthogonal vector.
!
!                On exit, the reorthogonalized and normalized nq+1 singular or
!                orthogonal vector.
!
!                The size of VEC must be equal to size(Q,1).
!
!   NORMVEC      (INPUT/OUTPUT) real(stnd)
!                On entry, the current 2-norm of the vector VEC.
!
!                On exit, the 2-norm of the nq+1 orthogonal vector, before normalization.
!
!   NORMVEC_ORIG (INPUT, OPTIONAL) real(stnd)
!                On entry, the original 2-norm of the vector VEC.
!
!                IF NORMVEC_ORIG is not specified, the default value is
!                assumed to be 1.
!
!
! Further Details
! _______________
!
!   ORTHO_VEC is an implementation of an orthogonalization procedure using
!   the Gram-Schmidt method described in the reference (1).
!
!   If size( q, 2) <= 0, ORTHO_VEC returns the normalized vector VEC if
!   NORMVEC is not zero and a nonzero normalized random vector VEC otherwise,
!   and NORMVEC is not modified.
!
!   For more details, see
!
!   (1) Stewart, G.W., 2007:
!       Block Gram-Schmidt Orthogonalization.
!       Report TR-4823, Department of Computer Science, College Park, University of Maryland.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
    use Random,          only : random_number_
#ifdef _BLAS
    use BLAS_interfaces, only : dot, axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:) :: q
    real(stnd), intent(in),    optional       :: normvec_orig
    real(stnd), intent(inout), dimension(:)   :: vec
    real(stnd), intent(inout)                 :: normvec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(vec)) :: vec1
    real(stnd)                       :: tmp, newnorm, oldnorm, normvec0
!
    integer(i4b) :: i, j, n, nq
!   
!   
! PARAMETERS 
! __________
!
    integer(i4b), parameter :: maxi=10
!
    real(stnd), parameter :: fudge=one
!
    character(len=*),  parameter :: name_proc='ortho_vec'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   CHECK THE ARGUMENTS.
!
    n = assert_eq(    int( size(vec), i4b ),   &
                      int( size(q,1), i4b ),   &
                      name_proc )
!
    nq = size( q, 2 ) 
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
    if ( nq<=0_i4b ) then
!
!       SPECIAL CASE WHEN size( q, 2) = 0.
!
        if ( normvec==zero ) then
!
!           IF THE NORM OF THE CURRENT ESTIMATE OF THE
!           VECTOR vec IS ZERO, SET IT TO A NONZERO
!           NORMALIZED VECTOR.
!
            call random_number_( vec1(:n) ) 
!
            vec1(:n) = vec1(:n) - half
            tmp      = norm( vec1(:n) )
            vec(:n)  = vec1(:n)/tmp
!
        else
!
!           NORMALIZED THE CURRENT ESTIMATE OF THE
!           ORTHOGONAL VECTOR OTHERWISE.
!
            vec(:n) = vec(:n)/normvec
!
        end if
!
        return
!
    end if
!
    if ( present(normvec_orig) ) then
!
!       GET TNE NORM OF THE ORIGINAL VECTOR.
!
        normvec0 = normvec_orig
!
    else
!
!       BY DEFAULT, WE ASSUME THAT THE ORIGINAL
!       VECTOR IS NORMALIZED.
!
        normvec0 = one
!
    end if
!
!    if ( normvec<=fudge*macheps*normvec0 ) then
    if ( normvec==zero ) then
!
!       IF THE NORM OF THE CURRENT ESTIMATE OF THE
!       ORTHOGONAL VECTOR IS ZERO, SET IT TO A NONZERO
!       NORMALIZED RANDOM VECTOR AND PROCEED.
!
        call random_number_( vec1(:n) ) 
!
        vec1(:n) = vec1(:n) - half
        tmp      = norm( vec1(:n) )
        vec1(:n) = vec1(:n)/tmp
!
        normvec0 = one
!
    else
!
!       MAKE A LOCAL COPY OF THE CURRENT ESTIMATE OF THE
!       ORTHOGONAL VECTOR AND NORMALIZE IT OTHERWISE.
!
        vec1(:n) = vec(:n)/normvec
!
        normvec0 = normvec0/normvec
!
    end if
!
!   ORTHOGONALIZATION LOOP.
!
!    oldnorm = one
    oldnorm = normvec0
!
    do i=1_i4b, maxi
!
        do j=1_i4b, nq
!
#ifdef _BLAS
            tmp = -dot( n, vec1(:n), 1_i4b, q(:n,j), 1_i4b )
            call axpy( n, tmp, q(:n,j), 1_i4b, vec1(:n), 1_i4b )
#else
            tmp = -dot_product( vec1(:n), q(:n,j) )
            vec1(:n) = vec1(:n) + tmp*q(:n,j)
#endif
!
        end do
!
        newnorm = norm( vec1(:n) )
!
!       RETURN IF vec1 IS ORTHOGONAL TO q.
!
        if ( newnorm>fudge_ortho2*oldnorm ) exit
!
        if ( newnorm>fudge*macheps*normvec0 ) then
!
!           CONTINUE ORTHOGONALIZING IF newnorm IS NOT TOO SMALL.
!
            oldnorm = newnorm
!
        else
!
!           REPLACE vec1 BY A SCALED RANDOM VECTOR OTHERWISE
!           AND PROCEED.
!
            normvec0 = macheps*normvec0
            oldnorm  = normvec0
!
            call random_number_( vec1(:n) ) 
!
            vec1(:n) = vec1(:n) - half
            tmp      = norm( vec1(:n) )
            vec1(:n) = normvec0*( vec1(:n)/tmp )
!
        end if
!
    end do
!
!   RETURN THE CURRENT ESTIMATE OF THE SINGULAR OR ORTHOGONAL VECTOR
!   AND ITS NORM BEFORE NORMALIZATION.
!
    if ( i<=maxi ) then
!
        tmp = newnorm
!
    else
!
        tmp = oldnorm
!
    end if
!
    if ( tmp/=zero ) then
!
        vec(:n) = vec1(:n)/tmp
!
    else
!
        vec(:n) = zero
!
    end if
!
    normvec = normvec*tmp
!
!
! END OF SUBROUTINE ortho_vec
! ___________________________
!
    end subroutine ortho_vec
!
! =========================================================================================
!
    function reverse( diag )
!
! Purpose
! _______
!
!   Private function REVERSE determines if the elements of the vector DIAG are increasing
!   in magnitude by computing the covariance coefficient of abs( DIAG(:) ) with the
!   vector (1:n) where n is the size of DIAG.
!
!   The function returns TRUE if the covariance coefficient is positive
!   and FALSE otherwise.
!
!
! Arguments
! _________
!
!   DIAG    (INPUT) real(stnd), dimension(:)
!           On entry, the vector DIAG . 
!
!
! Further Details
! _______________
!
!   This function is used to determine if the elements of a symmetric tridiagonal matrix
!   must be stored in reversed order before computing eigenvalues of this tridiagonal matrix.
!
!   This is important for maintaining the accuracy of the QR steps for graded matrices.
!
!   For further details, see
!
!   (1) Greenbaum, A., and J. Dongarra, J., 1989:
!           Experiments with QR/QL Methods for the Symmetric Tridiagonal Eigenproblem.
!           LAPACK Working Note No 17, November 1989.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : arth
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: diag(:)
!
    logical(lgl)  :: reverse
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(diag))  :: xt, yt
    real(stnd)                         :: sxy
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = size( diag )
!
    reverse = false
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   COMPUTE COVARIANCE COEFFICIENT.
!
    xt(:n) = arth( one, one, n ) - real( n+1_i4b, stnd)*half
    yt(:n) = abs( diag(:n) )
    yt(:n) = yt(:n) - sum( yt(:n) )/real( n, stnd)
    sxy    = dot_product( xt, yt )
!
    reverse = sxy>zero
!
!
! END OF FUNCTION reverse
! _______________________
!
    end function reverse
!
! =========================================================================================
!
    subroutine dqds2( sigma, desig, tau, sigma2, desig2 )
!
! Purpose
! _______
!
!   DQDS2 is a private subroutine used by dqds algorithms for computing singular values by
!   shift accumulation.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!   
!
! STRONG TYPING IMPOSED 
! _____________________
!    
    implicit none
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)  :: sigma, desig, tau
    real(stnd), intent(out) :: sigma2, desig2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
!
    real(stnd) :: t
!
! EXECUTABLE STATEMENTS
! _____________________
!
!      
    if ( tau<sigma ) then
        desig2 = desig + tau
        t = sigma + desig2
        desig2 = desig2 - ( t - sigma )
    else
        t = sigma + tau
        desig2 = sigma - ( t - tau ) + desig
    end if
!      
    sigma2 = t
!
!
! END OF SUBROUTINE dqds2
! _______________________
!
    end subroutine dqds2
!
!
! =========================================================================================
!
!
! ******************************
! END OF MODULE SVD_Procedures *
! ******************************
!    
end module SVD_Procedures
